return {
 ["hashes"]={
  { "mkstatus", "38dd7ee3f5b372550e9d5e47466ac6c4" },
  { "mkglyphlist", "62bc2c915a0fc936f9249deb0eddcaf3" },
  { "mkcharacters", "ec694a87f77067912eb73402ec99040d" },
  { "luaotfload-tool.lua", "c48f1995e9e74155aad6a745db16bcab" },
  { "luaotfload-parsers.lua", "cc86ac664ce5aedd81218e59b7180e83" },
  { "luaotfload-override.lua", "e6f67a268fa9362d5ebc5cbecad1a900" },
  { "luaotfload-fontloader.lua", "5bce3c40d6841694d6be5ad3680d5b10" },
  { "luaotfload-main.lua", "4a3ac1af4af76b6a7ba60246d90a08c0" },
  { "luaotfload-log.lua", "1f8aee13a8b8a7a20cd8b03285ab425a" },
  { "luaotfload-loaders.lua", "c364151f603514433fa3128939825d16" },
  { "luaotfload-letterspace.lua", "06cfb1e742d496b99c8980fdfd3f29f4" },
  { "luaotfload-glyphlist.lua", "ff440162d1b8a78a586375ee65630c21" },
  { "luaotfload-fonts-tfm.lua", "8fd3865240e4e87e99e0739abeda2322" },
  { "luaotfload-fonts-lua.lua", "50b8edb1db7009b6c661ab71ff24a466" },
  { "luaotfload-fonts-ext.lua", "0eee87fb26b7d135da88ac0a43a8037a" },
  { "luaotfload-fonts-enc.lua", "32775266bfc9eb6bbd9a16e3a7de398d" },
  { "luaotfload-fonts-def.lua", "8dc05f2aa96d2fd0245242c8fd3320eb" },
  { "luaotfload-fonts-cbk.lua", "c1b7a4e5508d68f45a99cdc38c2d0885" },
  { "luaotfload-features.lua", "9a0414d3bcd26497f8082c7803b38676" },
  { "luaotfload-diagnostics.lua", "273afde55dff486fb222a50981b3b1d7" },
  { "luaotfload-database.lua", "3392acc5a193a4988c156d794ed6933f" },
  { "luaotfload-colors.lua", "99878378ef0954d0b67f13db8456dea7" },
  { "luaotfload-characters.lua", "f89547bf3a33bea16a94c04ca8ef7f35" },
  { "luaotfload-basics-nod.lua", "7fdaff58d8c32e3115286504cbd7bcac" },
  { "luaotfload-basics-gen.lua", "9d5c3fc031af110ba67f3faa1a569779" },
  { "luaotfload-auxiliary.lua", "b3c893f68217b4147e5d3ba694648583" },
 },
 ["notes"]={
  ["committer"]="Philipp Gesang <phg42.2a@gmail.com>",
  ["revision"]="fa525b6b60f23a95f550aa737d07b9919720e7fc",
  ["timestamp"]="2014-08-10 23:15:22 +0200",
 },
}