return {
 [0x0]={
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="s",
 },
 {
  category="cc",
  direction="b",
 },
 {
  category="cc",
  direction="s",
 },
 {
  category="cc",
  direction="ws",
 },
 {
  category="cc",
  direction="b",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="b",
 },
 {
  category="cc",
  direction="b",
 },
 {
  category="cc",
  direction="b",
 },
 {
  category="cc",
  direction="s",
 },
 {
  category="zs",
  direction="ws",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="po",
  direction="et",
 },
 {
  category="sc",
  direction="et",
 },
 {
  category="po",
  direction="et",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="ps",
  direction="on",
  mirror=0x29,
  textclass="open",
 },
 {
  category="pe",
  direction="on",
  mirror=0x28,
  textclass="close",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="sm",
  direction="es",
 },
 {
  category="po",
  direction="cs",
 },
 {
  category="pd",
  direction="es",
 },
 {
  category="po",
  direction="cs",
 },
 {
  category="po",
  direction="cs",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="nd",
  direction="en",
 },
 {
  category="po",
  direction="cs",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="sm",
  direction="on",
  mirror=0x3E,
  textclass="open",
 },
 {
  category="sm",
  direction="on",
 },
 {
  category="sm",
  direction="on",
  mirror=0x3C,
  textclass="close",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ps",
  direction="on",
  mirror=0x5D,
  textclass="open",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="pe",
  direction="on",
  mirror=0x5B,
  textclass="close",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="pc",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ps",
  direction="on",
  mirror=0x7D,
  textclass="open",
 },
 {
  category="sm",
  direction="on",
 },
 {
  category="pe",
  direction="on",
  mirror=0x7B,
  textclass="close",
 },
 {
  category="sm",
  direction="on",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="b",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="cc",
  direction="bn",
 },
 {
  category="zs",
  direction="cs",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="sc",
  direction="et",
 },
 {
  category="sc",
  direction="et",
 },
 {
  category="sc",
  direction="et",
 },
 {
  category="sc",
  direction="et",
 },
 {
  category="so",
  direction="on",
 },
 {
  category="so",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="so",
  direction="on",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="pi",
  direction="on",
  mirror=0xBB,
  textclass="open",
 },
 {
  category="sm",
  direction="on",
 },
 {
  category="cf",
  direction="bn",
 },
 {
  category="so",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="so",
  direction="et",
 },
 {
  category="sm",
  direction="et",
 },
 {
  category="no",
  direction="en",
 },
 {
  category="no",
  direction="en",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="so",
  direction="on",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="no",
  direction="en",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="pf",
  direction="on",
  mirror=0xAB,
  textclass="close",
 },
 {
  category="no",
  direction="on",
 },
 {
  category="no",
  direction="on",
 },
 {
  category="no",
  direction="on",
 },
 {
  category="po",
  direction="on",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="sm",
  direction="on",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="sm",
  direction="on",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lo",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lo",
  direction="l",
 },
 {
  category="lo",
  direction="l",
 },
 {
  category="lo",
  direction="l",
 },
 {
  category="lo",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lt",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lt",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lt",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lt",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lo",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="on",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="lm",
  direction="l",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="mn",
  direction="nsm",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="sk",
  direction="on",
 },
 {
  category="lu",
  direction="l",
 },
 {
  category="ll",
  direction="l",
 },
 [0x37A]={
  category="lm",
  direction="l",
 },
 [0x37B]={
  category="ll",
  direction="l",
 },
 [0x37C]={
  category="ll",
  direction="l",
 },
 [0x37D]={
  category="ll",
  direction="l",
 },
 [0x37E]={
  category="po",
  direction="on",
 },
 [0x37F]={
  category="lu",
  direction="l",
 },
 [0x384]={
  category="sk",
  direction="on",
 },
 [0x385]={
  category="sk",
  direction="on",
 },
 [0x386]={
  category="lu",
  direction="l",
 },
 [0x387]={
  category="po",
  direction="on",
 },
 [0x388]={
  category="lu",
  direction="l",
 },
 [0x389]={
  category="lu",
  direction="l",
 },
 [0x38A]={
  category="lu",
  direction="l",
 },
 [0x38C]={
  category="lu",
  direction="l",
 },
 [0x38E]={
  category="lu",
  direction="l",
 },
 [0x38F]={
  category="lu",
  direction="l",
 },
 [0x390]={
  category="ll",
  direction="l",
 },
 [0x391]={
  category="lu",
  direction="l",
 },
 [0x392]={
  category="lu",
  direction="l",
 },
 [0x393]={
  category="lu",
  direction="l",
 },
 [0x394]={
  category="lu",
  direction="l",
 },
 [0x395]={
  category="lu",
  direction="l",
 },
 [0x396]={
  category="lu",
  direction="l",
 },
 [0x397]={
  category="lu",
  direction="l",
 },
 [0x398]={
  category="lu",
  direction="l",
 },
 [0x399]={
  category="lu",
  direction="l",
 },
 [0x39A]={
  category="lu",
  direction="l",
 },
 [0x39B]={
  category="lu",
  direction="l",
 },
 [0x39C]={
  category="lu",
  direction="l",
 },
 [0x39D]={
  category="lu",
  direction="l",
 },
 [0x39E]={
  category="lu",
  direction="l",
 },
 [0x39F]={
  category="lu",
  direction="l",
 },
 [0x3A0]={
  category="lu",
  direction="l",
 },
 [0x3A1]={
  category="lu",
  direction="l",
 },
 [0x3A3]={
  category="lu",
  direction="l",
 },
 [0x3A4]={
  category="lu",
  direction="l",
 },
 [0x3A5]={
  category="lu",
  direction="l",
 },
 [0x3A6]={
  category="lu",
  direction="l",
 },
 [0x3A7]={
  category="lu",
  direction="l",
 },
 [0x3A8]={
  category="lu",
  direction="l",
 },
 [0x3A9]={
  category="lu",
  direction="l",
 },
 [0x3AA]={
  category="lu",
  direction="l",
 },
 [0x3AB]={
  category="lu",
  direction="l",
 },
 [0x3AC]={
  category="ll",
  direction="l",
 },
 [0x3AD]={
  category="ll",
  direction="l",
 },
 [0x3AE]={
  category="ll",
  direction="l",
 },
 [0x3AF]={
  category="ll",
  direction="l",
 },
 [0x3B0]={
  category="ll",
  direction="l",
 },
 [0x3B1]={
  category="ll",
  direction="l",
 },
 [0x3B2]={
  category="ll",
  direction="l",
 },
 [0x3B3]={
  category="ll",
  direction="l",
 },
 [0x3B4]={
  category="ll",
  direction="l",
 },
 [0x3B5]={
  category="ll",
  direction="l",
 },
 [0x3B6]={
  category="ll",
  direction="l",
 },
 [0x3B7]={
  category="ll",
  direction="l",
 },
 [0x3B8]={
  category="ll",
  direction="l",
 },
 [0x3B9]={
  category="ll",
  direction="l",
 },
 [0x3BA]={
  category="ll",
  direction="l",
 },
 [0x3BB]={
  category="ll",
  direction="l",
 },
 [0x3BC]={
  category="ll",
  direction="l",
 },
 [0x3BD]={
  category="ll",
  direction="l",
 },
 [0x3BE]={
  category="ll",
  direction="l",
 },
 [0x3BF]={
  category="ll",
  direction="l",
 },
 [0x3C0]={
  category="ll",
  direction="l",
 },
 [0x3C1]={
  category="ll",
  direction="l",
 },
 [0x3C2]={
  category="ll",
  direction="l",
 },
 [0x3C3]={
  category="ll",
  direction="l",
 },
 [0x3C4]={
  category="ll",
  direction="l",
 },
 [0x3C5]={
  category="ll",
  direction="l",
 },
 [0x3C6]={
  category="ll",
  direction="l",
 },
 [0x3C7]={
  category="ll",
  direction="l",
 },
 [0x3C8]={
  category="ll",
  direction="l",
 },
 [0x3C9]={
  category="ll",
  direction="l",
 },
 [0x3CA]={
  category="ll",
  direction="l",
 },
 [0x3CB]={
  category="ll",
  direction="l",
 },
 [0x3CC]={
  category="ll",
  direction="l",
 },
 [0x3CD]={
  category="ll",
  direction="l",
 },
 [0x3CE]={
  category="ll",
  direction="l",
 },
 [0x3CF]={
  category="lu",
  direction="l",
 },
 [0x3D0]={
  category="ll",
  direction="l",
 },
 [0x3D1]={
  category="ll",
  direction="l",
 },
 [0x3D2]={
  category="lu",
  direction="l",
 },
 [0x3D3]={
  category="lu",
  direction="l",
 },
 [0x3D4]={
  category="lu",
  direction="l",
 },
 [0x3D5]={
  category="ll",
  direction="l",
 },
 [0x3D6]={
  category="ll",
  direction="l",
 },
 [0x3D7]={
  category="ll",
  direction="l",
 },
 [0x3D8]={
  category="lu",
  direction="l",
 },
 [0x3D9]={
  category="ll",
  direction="l",
 },
 [0x3DA]={
  category="lu",
  direction="l",
 },
 [0x3DB]={
  category="ll",
  direction="l",
 },
 [0x3DC]={
  category="lu",
  direction="l",
 },
 [0x3DD]={
  category="ll",
  direction="l",
 },
 [0x3DE]={
  category="lu",
  direction="l",
 },
 [0x3DF]={
  category="ll",
  direction="l",
 },
 [0x3E0]={
  category="lu",
  direction="l",
 },
 [0x3E1]={
  category="ll",
  direction="l",
 },
 [0x3E2]={
  category="lu",
  direction="l",
 },
 [0x3E3]={
  category="ll",
  direction="l",
 },
 [0x3E4]={
  category="lu",
  direction="l",
 },
 [0x3E5]={
  category="ll",
  direction="l",
 },
 [0x3E6]={
  category="lu",
  direction="l",
 },
 [0x3E7]={
  category="ll",
  direction="l",
 },
 [0x3E8]={
  category="lu",
  direction="l",
 },
 [0x3E9]={
  category="ll",
  direction="l",
 },
 [0x3EA]={
  category="lu",
  direction="l",
 },
 [0x3EB]={
  category="ll",
  direction="l",
 },
 [0x3EC]={
  category="lu",
  direction="l",
 },
 [0x3ED]={
  category="ll",
  direction="l",
 },
 [0x3EE]={
  category="lu",
  direction="l",
 },
 [0x3EF]={
  category="ll",
  direction="l",
 },
 [0x3F0]={
  category="ll",
  direction="l",
 },
 [0x3F1]={
  category="ll",
  direction="l",
 },
 [0x3F2]={
  category="ll",
  direction="l",
 },
 [0x3F3]={
  category="ll",
  direction="l",
 },
 [0x3F4]={
  category="lu",
  direction="l",
 },
 [0x3F5]={
  category="ll",
  direction="l",
 },
 [0x3F6]={
  category="sm",
  direction="on",
 },
 [0x3F7]={
  category="lu",
  direction="l",
 },
 [0x3F8]={
  category="ll",
  direction="l",
 },
 [0x3F9]={
  category="lu",
  direction="l",
 },
 [0x3FA]={
  category="lu",
  direction="l",
 },
 [0x3FB]={
  category="ll",
  direction="l",
 },
 [0x3FC]={
  category="ll",
  direction="l",
 },
 [0x3FD]={
  category="lu",
  direction="l",
 },
 [0x3FE]={
  category="lu",
  direction="l",
 },
 [0x3FF]={
  category="lu",
  direction="l",
 },
 [0x400]={
  category="lu",
  direction="l",
 },
 [0x401]={
  category="lu",
  direction="l",
 },
 [0x402]={
  category="lu",
  direction="l",
 },
 [0x403]={
  category="lu",
  direction="l",
 },
 [0x404]={
  category="lu",
  direction="l",
 },
 [0x405]={
  category="lu",
  direction="l",
 },
 [0x406]={
  category="lu",
  direction="l",
 },
 [0x407]={
  category="lu",
  direction="l",
 },
 [0x408]={
  category="lu",
  direction="l",
 },
 [0x409]={
  category="lu",
  direction="l",
 },
 [0x40A]={
  category="lu",
  direction="l",
 },
 [0x40B]={
  category="lu",
  direction="l",
 },
 [0x40C]={
  category="lu",
  direction="l",
 },
 [0x40D]={
  category="lu",
  direction="l",
 },
 [0x40E]={
  category="lu",
  direction="l",
 },
 [0x40F]={
  category="lu",
  direction="l",
 },
 [0x410]={
  category="lu",
  direction="l",
 },
 [0x411]={
  category="lu",
  direction="l",
 },
 [0x412]={
  category="lu",
  direction="l",
 },
 [0x413]={
  category="lu",
  direction="l",
 },
 [0x414]={
  category="lu",
  direction="l",
 },
 [0x415]={
  category="lu",
  direction="l",
 },
 [0x416]={
  category="lu",
  direction="l",
 },
 [0x417]={
  category="lu",
  direction="l",
 },
 [0x418]={
  category="lu",
  direction="l",
 },
 [0x419]={
  category="lu",
  direction="l",
 },
 [0x41A]={
  category="lu",
  direction="l",
 },
 [0x41B]={
  category="lu",
  direction="l",
 },
 [0x41C]={
  category="lu",
  direction="l",
 },
 [0x41D]={
  category="lu",
  direction="l",
 },
 [0x41E]={
  category="lu",
  direction="l",
 },
 [0x41F]={
  category="lu",
  direction="l",
 },
 [0x420]={
  category="lu",
  direction="l",
 },
 [0x421]={
  category="lu",
  direction="l",
 },
 [0x422]={
  category="lu",
  direction="l",
 },
 [0x423]={
  category="lu",
  direction="l",
 },
 [0x424]={
  category="lu",
  direction="l",
 },
 [0x425]={
  category="lu",
  direction="l",
 },
 [0x426]={
  category="lu",
  direction="l",
 },
 [0x427]={
  category="lu",
  direction="l",
 },
 [0x428]={
  category="lu",
  direction="l",
 },
 [0x429]={
  category="lu",
  direction="l",
 },
 [0x42A]={
  category="lu",
  direction="l",
 },
 [0x42B]={
  category="lu",
  direction="l",
 },
 [0x42C]={
  category="lu",
  direction="l",
 },
 [0x42D]={
  category="lu",
  direction="l",
 },
 [0x42E]={
  category="lu",
  direction="l",
 },
 [0x42F]={
  category="lu",
  direction="l",
 },
 [0x430]={
  category="ll",
  direction="l",
 },
 [0x431]={
  category="ll",
  direction="l",
 },
 [0x432]={
  category="ll",
  direction="l",
 },
 [0x433]={
  category="ll",
  direction="l",
 },
 [0x434]={
  category="ll",
  direction="l",
 },
 [0x435]={
  category="ll",
  direction="l",
 },
 [0x436]={
  category="ll",
  direction="l",
 },
 [0x437]={
  category="ll",
  direction="l",
 },
 [0x438]={
  category="ll",
  direction="l",
 },
 [0x439]={
  category="ll",
  direction="l",
 },
 [0x43A]={
  category="ll",
  direction="l",
 },
 [0x43B]={
  category="ll",
  direction="l",
 },
 [0x43C]={
  category="ll",
  direction="l",
 },
 [0x43D]={
  category="ll",
  direction="l",
 },
 [0x43E]={
  category="ll",
  direction="l",
 },
 [0x43F]={
  category="ll",
  direction="l",
 },
 [0x440]={
  category="ll",
  direction="l",
 },
 [0x441]={
  category="ll",
  direction="l",
 },
 [0x442]={
  category="ll",
  direction="l",
 },
 [0x443]={
  category="ll",
  direction="l",
 },
 [0x444]={
  category="ll",
  direction="l",
 },
 [0x445]={
  category="ll",
  direction="l",
 },
 [0x446]={
  category="ll",
  direction="l",
 },
 [0x447]={
  category="ll",
  direction="l",
 },
 [0x448]={
  category="ll",
  direction="l",
 },
 [0x449]={
  category="ll",
  direction="l",
 },
 [0x44A]={
  category="ll",
  direction="l",
 },
 [0x44B]={
  category="ll",
  direction="l",
 },
 [0x44C]={
  category="ll",
  direction="l",
 },
 [0x44D]={
  category="ll",
  direction="l",
 },
 [0x44E]={
  category="ll",
  direction="l",
 },
 [0x44F]={
  category="ll",
  direction="l",
 },
 [0x450]={
  category="ll",
  direction="l",
 },
 [0x451]={
  category="ll",
  direction="l",
 },
 [0x452]={
  category="ll",
  direction="l",
 },
 [0x453]={
  category="ll",
  direction="l",
 },
 [0x454]={
  category="ll",
  direction="l",
 },
 [0x455]={
  category="ll",
  direction="l",
 },
 [0x456]={
  category="ll",
  direction="l",
 },
 [0x457]={
  category="ll",
  direction="l",
 },
 [0x458]={
  category="ll",
  direction="l",
 },
 [0x459]={
  category="ll",
  direction="l",
 },
 [0x45A]={
  category="ll",
  direction="l",
 },
 [0x45B]={
  category="ll",
  direction="l",
 },
 [0x45C]={
  category="ll",
  direction="l",
 },
 [0x45D]={
  category="ll",
  direction="l",
 },
 [0x45E]={
  category="ll",
  direction="l",
 },
 [0x45F]={
  category="ll",
  direction="l",
 },
 [0x460]={
  category="lu",
  direction="l",
 },
 [0x461]={
  category="ll",
  direction="l",
 },
 [0x462]={
  category="lu",
  direction="l",
 },
 [0x463]={
  category="ll",
  direction="l",
 },
 [0x464]={
  category="lu",
  direction="l",
 },
 [0x465]={
  category="ll",
  direction="l",
 },
 [0x466]={
  category="lu",
  direction="l",
 },
 [0x467]={
  category="ll",
  direction="l",
 },
 [0x468]={
  category="lu",
  direction="l",
 },
 [0x469]={
  category="ll",
  direction="l",
 },
 [0x46A]={
  category="lu",
  direction="l",
 },
 [0x46B]={
  category="ll",
  direction="l",
 },
 [0x46C]={
  category="lu",
  direction="l",
 },
 [0x46D]={
  category="ll",
  direction="l",
 },
 [0x46E]={
  category="lu",
  direction="l",
 },
 [0x46F]={
  category="ll",
  direction="l",
 },
 [0x470]={
  category="lu",
  direction="l",
 },
 [0x471]={
  category="ll",
  direction="l",
 },
 [0x472]={
  category="lu",
  direction="l",
 },
 [0x473]={
  category="ll",
  direction="l",
 },
 [0x474]={
  category="lu",
  direction="l",
 },
 [0x475]={
  category="ll",
  direction="l",
 },
 [0x476]={
  category="lu",
  direction="l",
 },
 [0x477]={
  category="ll",
  direction="l",
 },
 [0x478]={
  category="lu",
  direction="l",
 },
 [0x479]={
  category="ll",
  direction="l",
 },
 [0x47A]={
  category="lu",
  direction="l",
 },
 [0x47B]={
  category="ll",
  direction="l",
 },
 [0x47C]={
  category="lu",
  direction="l",
 },
 [0x47D]={
  category="ll",
  direction="l",
 },
 [0x47E]={
  category="lu",
  direction="l",
 },
 [0x47F]={
  category="ll",
  direction="l",
 },
 [0x480]={
  category="lu",
  direction="l",
 },
 [0x481]={
  category="ll",
  direction="l",
 },
 [0x482]={
  category="so",
  direction="l",
 },
 [0x483]={
  category="mn",
  direction="nsm",
 },
 [0x484]={
  category="mn",
  direction="nsm",
 },
 [0x485]={
  category="mn",
  direction="nsm",
 },
 [0x486]={
  category="mn",
  direction="nsm",
 },
 [0x487]={
  category="mn",
  direction="nsm",
 },
 [0x488]={
  category="me",
  direction="nsm",
 },
 [0x489]={
  category="me",
  direction="nsm",
 },
 [0x48A]={
  category="lu",
  direction="l",
 },
 [0x48B]={
  category="ll",
  direction="l",
 },
 [0x48C]={
  category="lu",
  direction="l",
 },
 [0x48D]={
  category="ll",
  direction="l",
 },
 [0x48E]={
  category="lu",
  direction="l",
 },
 [0x48F]={
  category="ll",
  direction="l",
 },
 [0x490]={
  category="lu",
  direction="l",
 },
 [0x491]={
  category="ll",
  direction="l",
 },
 [0x492]={
  category="lu",
  direction="l",
 },
 [0x493]={
  category="ll",
  direction="l",
 },
 [0x494]={
  category="lu",
  direction="l",
 },
 [0x495]={
  category="ll",
  direction="l",
 },
 [0x496]={
  category="lu",
  direction="l",
 },
 [0x497]={
  category="ll",
  direction="l",
 },
 [0x498]={
  category="lu",
  direction="l",
 },
 [0x499]={
  category="ll",
  direction="l",
 },
 [0x49A]={
  category="lu",
  direction="l",
 },
 [0x49B]={
  category="ll",
  direction="l",
 },
 [0x49C]={
  category="lu",
  direction="l",
 },
 [0x49D]={
  category="ll",
  direction="l",
 },
 [0x49E]={
  category="lu",
  direction="l",
 },
 [0x49F]={
  category="ll",
  direction="l",
 },
 [0x4A0]={
  category="lu",
  direction="l",
 },
 [0x4A1]={
  category="ll",
  direction="l",
 },
 [0x4A2]={
  category="lu",
  direction="l",
 },
 [0x4A3]={
  category="ll",
  direction="l",
 },
 [0x4A4]={
  category="lu",
  direction="l",
 },
 [0x4A5]={
  category="ll",
  direction="l",
 },
 [0x4A6]={
  category="lu",
  direction="l",
 },
 [0x4A7]={
  category="ll",
  direction="l",
 },
 [0x4A8]={
  category="lu",
  direction="l",
 },
 [0x4A9]={
  category="ll",
  direction="l",
 },
 [0x4AA]={
  category="lu",
  direction="l",
 },
 [0x4AB]={
  category="ll",
  direction="l",
 },
 [0x4AC]={
  category="lu",
  direction="l",
 },
 [0x4AD]={
  category="ll",
  direction="l",
 },
 [0x4AE]={
  category="lu",
  direction="l",
 },
 [0x4AF]={
  category="ll",
  direction="l",
 },
 [0x4B0]={
  category="lu",
  direction="l",
 },
 [0x4B1]={
  category="ll",
  direction="l",
 },
 [0x4B2]={
  category="lu",
  direction="l",
 },
 [0x4B3]={
  category="ll",
  direction="l",
 },
 [0x4B4]={
  category="lu",
  direction="l",
 },
 [0x4B5]={
  category="ll",
  direction="l",
 },
 [0x4B6]={
  category="lu",
  direction="l",
 },
 [0x4B7]={
  category="ll",
  direction="l",
 },
 [0x4B8]={
  category="lu",
  direction="l",
 },
 [0x4B9]={
  category="ll",
  direction="l",
 },
 [0x4BA]={
  category="lu",
  direction="l",
 },
 [0x4BB]={
  category="ll",
  direction="l",
 },
 [0x4BC]={
  category="lu",
  direction="l",
 },
 [0x4BD]={
  category="ll",
  direction="l",
 },
 [0x4BE]={
  category="lu",
  direction="l",
 },
 [0x4BF]={
  category="ll",
  direction="l",
 },
 [0x4C0]={
  category="lu",
  direction="l",
 },
 [0x4C1]={
  category="lu",
  direction="l",
 },
 [0x4C2]={
  category="ll",
  direction="l",
 },
 [0x4C3]={
  category="lu",
  direction="l",
 },
 [0x4C4]={
  category="ll",
  direction="l",
 },
 [0x4C5]={
  category="lu",
  direction="l",
 },
 [0x4C6]={
  category="ll",
  direction="l",
 },
 [0x4C7]={
  category="lu",
  direction="l",
 },
 [0x4C8]={
  category="ll",
  direction="l",
 },
 [0x4C9]={
  category="lu",
  direction="l",
 },
 [0x4CA]={
  category="ll",
  direction="l",
 },
 [0x4CB]={
  category="lu",
  direction="l",
 },
 [0x4CC]={
  category="ll",
  direction="l",
 },
 [0x4CD]={
  category="lu",
  direction="l",
 },
 [0x4CE]={
  category="ll",
  direction="l",
 },
 [0x4CF]={
  category="ll",
  direction="l",
 },
 [0x4D0]={
  category="lu",
  direction="l",
 },
 [0x4D1]={
  category="ll",
  direction="l",
 },
 [0x4D2]={
  category="lu",
  direction="l",
 },
 [0x4D3]={
  category="ll",
  direction="l",
 },
 [0x4D4]={
  category="lu",
  direction="l",
 },
 [0x4D5]={
  category="ll",
  direction="l",
 },
 [0x4D6]={
  category="lu",
  direction="l",
 },
 [0x4D7]={
  category="ll",
  direction="l",
 },
 [0x4D8]={
  category="lu",
  direction="l",
 },
 [0x4D9]={
  category="ll",
  direction="l",
 },
 [0x4DA]={
  category="lu",
  direction="l",
 },
 [0x4DB]={
  category="ll",
  direction="l",
 },
 [0x4DC]={
  category="lu",
  direction="l",
 },
 [0x4DD]={
  category="ll",
  direction="l",
 },
 [0x4DE]={
  category="lu",
  direction="l",
 },
 [0x4DF]={
  category="ll",
  direction="l",
 },
 [0x4E0]={
  category="lu",
  direction="l",
 },
 [0x4E1]={
  category="ll",
  direction="l",
 },
 [0x4E2]={
  category="lu",
  direction="l",
 },
 [0x4E3]={
  category="ll",
  direction="l",
 },
 [0x4E4]={
  category="lu",
  direction="l",
 },
 [0x4E5]={
  category="ll",
  direction="l",
 },
 [0x4E6]={
  category="lu",
  direction="l",
 },
 [0x4E7]={
  category="ll",
  direction="l",
 },
 [0x4E8]={
  category="lu",
  direction="l",
 },
 [0x4E9]={
  category="ll",
  direction="l",
 },
 [0x4EA]={
  category="lu",
  direction="l",
 },
 [0x4EB]={
  category="ll",
  direction="l",
 },
 [0x4EC]={
  category="lu",
  direction="l",
 },
 [0x4ED]={
  category="ll",
  direction="l",
 },
 [0x4EE]={
  category="lu",
  direction="l",
 },
 [0x4EF]={
  category="ll",
  direction="l",
 },
 [0x4F0]={
  category="lu",
  direction="l",
 },
 [0x4F1]={
  category="ll",
  direction="l",
 },
 [0x4F2]={
  category="lu",
  direction="l",
 },
 [0x4F3]={
  category="ll",
  direction="l",
 },
 [0x4F4]={
  category="lu",
  direction="l",
 },
 [0x4F5]={
  category="ll",
  direction="l",
 },
 [0x4F6]={
  category="lu",
  direction="l",
 },
 [0x4F7]={
  category="ll",
  direction="l",
 },
 [0x4F8]={
  category="lu",
  direction="l",
 },
 [0x4F9]={
  category="ll",
  direction="l",
 },
 [0x4FA]={
  category="lu",
  direction="l",
 },
 [0x4FB]={
  category="ll",
  direction="l",
 },
 [0x4FC]={
  category="lu",
  direction="l",
 },
 [0x4FD]={
  category="ll",
  direction="l",
 },
 [0x4FE]={
  category="lu",
  direction="l",
 },
 [0x4FF]={
  category="ll",
  direction="l",
 },
 [0x500]={
  category="lu",
  direction="l",
 },
 [0x501]={
  category="ll",
  direction="l",
 },
 [0x502]={
  category="lu",
  direction="l",
 },
 [0x503]={
  category="ll",
  direction="l",
 },
 [0x504]={
  category="lu",
  direction="l",
 },
 [0x505]={
  category="ll",
  direction="l",
 },
 [0x506]={
  category="lu",
  direction="l",
 },
 [0x507]={
  category="ll",
  direction="l",
 },
 [0x508]={
  category="lu",
  direction="l",
 },
 [0x509]={
  category="ll",
  direction="l",
 },
 [0x50A]={
  category="lu",
  direction="l",
 },
 [0x50B]={
  category="ll",
  direction="l",
 },
 [0x50C]={
  category="lu",
  direction="l",
 },
 [0x50D]={
  category="ll",
  direction="l",
 },
 [0x50E]={
  category="lu",
  direction="l",
 },
 [0x50F]={
  category="ll",
  direction="l",
 },
 [0x510]={
  category="lu",
  direction="l",
 },
 [0x511]={
  category="ll",
  direction="l",
 },
 [0x512]={
  category="lu",
  direction="l",
 },
 [0x513]={
  category="ll",
  direction="l",
 },
 [0x514]={
  category="lu",
  direction="l",
 },
 [0x515]={
  category="ll",
  direction="l",
 },
 [0x516]={
  category="lu",
  direction="l",
 },
 [0x517]={
  category="ll",
  direction="l",
 },
 [0x518]={
  category="lu",
  direction="l",
 },
 [0x519]={
  category="ll",
  direction="l",
 },
 [0x51A]={
  category="lu",
  direction="l",
 },
 [0x51B]={
  category="ll",
  direction="l",
 },
 [0x51C]={
  category="lu",
  direction="l",
 },
 [0x51D]={
  category="ll",
  direction="l",
 },
 [0x51E]={
  category="lu",
  direction="l",
 },
 [0x51F]={
  category="ll",
  direction="l",
 },
 [0x520]={
  category="lu",
  direction="l",
 },
 [0x521]={
  category="ll",
  direction="l",
 },
 [0x522]={
  category="lu",
  direction="l",
 },
 [0x523]={
  category="ll",
  direction="l",
 },
 [0x524]={
  category="lu",
  direction="l",
 },
 [0x525]={
  category="ll",
  direction="l",
 },
 [0x526]={
  category="lu",
  direction="l",
 },
 [0x527]={
  category="ll",
  direction="l",
 },
 [0x528]={
  category="lu",
  direction="l",
 },
 [0x529]={
  category="ll",
  direction="l",
 },
 [0x52A]={
  category="lu",
  direction="l",
 },
 [0x52B]={
  category="ll",
  direction="l",
 },
 [0x52C]={
  category="lu",
  direction="l",
 },
 [0x52D]={
  category="ll",
  direction="l",
 },
 [0x52E]={
  category="lu",
  direction="l",
 },
 [0x52F]={
  category="ll",
  direction="l",
 },
 [0x531]={
  category="lu",
  direction="l",
 },
 [0x532]={
  category="lu",
  direction="l",
 },
 [0x533]={
  category="lu",
  direction="l",
 },
 [0x534]={
  category="lu",
  direction="l",
 },
 [0x535]={
  category="lu",
  direction="l",
 },
 [0x536]={
  category="lu",
  direction="l",
 },
 [0x537]={
  category="lu",
  direction="l",
 },
 [0x538]={
  category="lu",
  direction="l",
 },
 [0x539]={
  category="lu",
  direction="l",
 },
 [0x53A]={
  category="lu",
  direction="l",
 },
 [0x53B]={
  category="lu",
  direction="l",
 },
 [0x53C]={
  category="lu",
  direction="l",
 },
 [0x53D]={
  category="lu",
  direction="l",
 },
 [0x53E]={
  category="lu",
  direction="l",
 },
 [0x53F]={
  category="lu",
  direction="l",
 },
 [0x540]={
  category="lu",
  direction="l",
 },
 [0x541]={
  category="lu",
  direction="l",
 },
 [0x542]={
  category="lu",
  direction="l",
 },
 [0x543]={
  category="lu",
  direction="l",
 },
 [0x544]={
  category="lu",
  direction="l",
 },
 [0x545]={
  category="lu",
  direction="l",
 },
 [0x546]={
  category="lu",
  direction="l",
 },
 [0x547]={
  category="lu",
  direction="l",
 },
 [0x548]={
  category="lu",
  direction="l",
 },
 [0x549]={
  category="lu",
  direction="l",
 },
 [0x54A]={
  category="lu",
  direction="l",
 },
 [0x54B]={
  category="lu",
  direction="l",
 },
 [0x54C]={
  category="lu",
  direction="l",
 },
 [0x54D]={
  category="lu",
  direction="l",
 },
 [0x54E]={
  category="lu",
  direction="l",
 },
 [0x54F]={
  category="lu",
  direction="l",
 },
 [0x550]={
  category="lu",
  direction="l",
 },
 [0x551]={
  category="lu",
  direction="l",
 },
 [0x552]={
  category="lu",
  direction="l",
 },
 [0x553]={
  category="lu",
  direction="l",
 },
 [0x554]={
  category="lu",
  direction="l",
 },
 [0x555]={
  category="lu",
  direction="l",
 },
 [0x556]={
  category="lu",
  direction="l",
 },
 [0x559]={
  category="lm",
  direction="l",
 },
 [0x55A]={
  category="po",
  direction="l",
 },
 [0x55B]={
  category="po",
  direction="l",
 },
 [0x55C]={
  category="po",
  direction="l",
 },
 [0x55D]={
  category="po",
  direction="l",
 },
 [0x55E]={
  category="po",
  direction="l",
 },
 [0x55F]={
  category="po",
  direction="l",
 },
 [0x561]={
  category="ll",
  direction="l",
 },
 [0x562]={
  category="ll",
  direction="l",
 },
 [0x563]={
  category="ll",
  direction="l",
 },
 [0x564]={
  category="ll",
  direction="l",
 },
 [0x565]={
  category="ll",
  direction="l",
 },
 [0x566]={
  category="ll",
  direction="l",
 },
 [0x567]={
  category="ll",
  direction="l",
 },
 [0x568]={
  category="ll",
  direction="l",
 },
 [0x569]={
  category="ll",
  direction="l",
 },
 [0x56A]={
  category="ll",
  direction="l",
 },
 [0x56B]={
  category="ll",
  direction="l",
 },
 [0x56C]={
  category="ll",
  direction="l",
 },
 [0x56D]={
  category="ll",
  direction="l",
 },
 [0x56E]={
  category="ll",
  direction="l",
 },
 [0x56F]={
  category="ll",
  direction="l",
 },
 [0x570]={
  category="ll",
  direction="l",
 },
 [0x571]={
  category="ll",
  direction="l",
 },
 [0x572]={
  category="ll",
  direction="l",
 },
 [0x573]={
  category="ll",
  direction="l",
 },
 [0x574]={
  category="ll",
  direction="l",
 },
 [0x575]={
  category="ll",
  direction="l",
 },
 [0x576]={
  category="ll",
  direction="l",
 },
 [0x577]={
  category="ll",
  direction="l",
 },
 [0x578]={
  category="ll",
  direction="l",
 },
 [0x579]={
  category="ll",
  direction="l",
 },
 [0x57A]={
  category="ll",
  direction="l",
 },
 [0x57B]={
  category="ll",
  direction="l",
 },
 [0x57C]={
  category="ll",
  direction="l",
 },
 [0x57D]={
  category="ll",
  direction="l",
 },
 [0x57E]={
  category="ll",
  direction="l",
 },
 [0x57F]={
  category="ll",
  direction="l",
 },
 [0x580]={
  category="ll",
  direction="l",
 },
 [0x581]={
  category="ll",
  direction="l",
 },
 [0x582]={
  category="ll",
  direction="l",
 },
 [0x583]={
  category="ll",
  direction="l",
 },
 [0x584]={
  category="ll",
  direction="l",
 },
 [0x585]={
  category="ll",
  direction="l",
 },
 [0x586]={
  category="ll",
  direction="l",
 },
 [0x587]={
  category="ll",
  direction="l",
 },
 [0x589]={
  category="po",
  direction="l",
 },
 [0x58A]={
  category="pd",
  direction="on",
 },
 [0x58D]={
  category="so",
  direction="on",
 },
 [0x58E]={
  category="so",
  direction="on",
 },
 [0x58F]={
  category="sc",
  direction="et",
 },
 [0x591]={
  category="mn",
  direction="nsm",
 },
 [0x592]={
  category="mn",
  direction="nsm",
 },
 [0x593]={
  category="mn",
  direction="nsm",
 },
 [0x594]={
  category="mn",
  direction="nsm",
 },
 [0x595]={
  category="mn",
  direction="nsm",
 },
 [0x596]={
  category="mn",
  direction="nsm",
 },
 [0x597]={
  category="mn",
  direction="nsm",
 },
 [0x598]={
  category="mn",
  direction="nsm",
 },
 [0x599]={
  category="mn",
  direction="nsm",
 },
 [0x59A]={
  category="mn",
  direction="nsm",
 },
 [0x59B]={
  category="mn",
  direction="nsm",
 },
 [0x59C]={
  category="mn",
  direction="nsm",
 },
 [0x59D]={
  category="mn",
  direction="nsm",
 },
 [0x59E]={
  category="mn",
  direction="nsm",
 },
 [0x59F]={
  category="mn",
  direction="nsm",
 },
 [0x5A0]={
  category="mn",
  direction="nsm",
 },
 [0x5A1]={
  category="mn",
  direction="nsm",
 },
 [0x5A2]={
  category="mn",
  direction="nsm",
 },
 [0x5A3]={
  category="mn",
  direction="nsm",
 },
 [0x5A4]={
  category="mn",
  direction="nsm",
 },
 [0x5A5]={
  category="mn",
  direction="nsm",
 },
 [0x5A6]={
  category="mn",
  direction="nsm",
 },
 [0x5A7]={
  category="mn",
  direction="nsm",
 },
 [0x5A8]={
  category="mn",
  direction="nsm",
 },
 [0x5A9]={
  category="mn",
  direction="nsm",
 },
 [0x5AA]={
  category="mn",
  direction="nsm",
 },
 [0x5AB]={
  category="mn",
  direction="nsm",
 },
 [0x5AC]={
  category="mn",
  direction="nsm",
 },
 [0x5AD]={
  category="mn",
  direction="nsm",
 },
 [0x5AE]={
  category="mn",
  direction="nsm",
 },
 [0x5AF]={
  category="mn",
  direction="nsm",
 },
 [0x5B0]={
  category="mn",
  direction="nsm",
 },
 [0x5B1]={
  category="mn",
  direction="nsm",
 },
 [0x5B2]={
  category="mn",
  direction="nsm",
 },
 [0x5B3]={
  category="mn",
  direction="nsm",
 },
 [0x5B4]={
  category="mn",
  direction="nsm",
 },
 [0x5B5]={
  category="mn",
  direction="nsm",
 },
 [0x5B6]={
  category="mn",
  direction="nsm",
 },
 [0x5B7]={
  category="mn",
  direction="nsm",
 },
 [0x5B8]={
  category="mn",
  direction="nsm",
 },
 [0x5B9]={
  category="mn",
  direction="nsm",
 },
 [0x5BA]={
  category="mn",
  direction="nsm",
 },
 [0x5BB]={
  category="mn",
  direction="nsm",
 },
 [0x5BC]={
  category="mn",
  direction="nsm",
 },
 [0x5BD]={
  category="mn",
  direction="nsm",
 },
 [0x5BE]={
  category="po",
  direction="r",
 },
 [0x5BF]={
  category="mn",
  direction="nsm",
 },
 [0x5C0]={
  category="po",
  direction="r",
 },
 [0x5C1]={
  category="mn",
  direction="nsm",
 },
 [0x5C2]={
  category="mn",
  direction="nsm",
 },
 [0x5C3]={
  category="po",
  direction="r",
 },
 [0x5C4]={
  category="mn",
  direction="nsm",
 },
 [0x5C5]={
  category="mn",
  direction="nsm",
 },
 [0x5C6]={
  category="po",
  direction="r",
 },
 [0x5C7]={
  category="mn",
  direction="nsm",
 },
 [0x5D0]={
  category="lo",
  direction="r",
 },
 [0x5D1]={
  category="lo",
  direction="r",
 },
 [0x5D2]={
  category="lo",
  direction="r",
 },
 [0x5D3]={
  category="lo",
  direction="r",
 },
 [0x5D4]={
  category="lo",
  direction="r",
 },
 [0x5D5]={
  category="lo",
  direction="r",
 },
 [0x5D6]={
  category="lo",
  direction="r",
 },
 [0x5D7]={
  category="lo",
  direction="r",
 },
 [0x5D8]={
  category="lo",
  direction="r",
 },
 [0x5D9]={
  category="lo",
  direction="r",
 },
 [0x5DA]={
  category="lo",
  direction="r",
 },
 [0x5DB]={
  category="lo",
  direction="r",
 },
 [0x5DC]={
  category="lo",
  direction="r",
 },
 [0x5DD]={
  category="lo",
  direction="r",
 },
 [0x5DE]={
  category="lo",
  direction="r",
 },
 [0x5DF]={
  category="lo",
  direction="r",
 },
 [0x5E0]={
  category="lo",
  direction="r",
 },
 [0x5E1]={
  category="lo",
  direction="r",
 },
 [0x5E2]={
  category="lo",
  direction="r",
 },
 [0x5E3]={
  category="lo",
  direction="r",
 },
 [0x5E4]={
  category="lo",
  direction="r",
 },
 [0x5E5]={
  category="lo",
  direction="r",
 },
 [0x5E6]={
  category="lo",
  direction="r",
 },
 [0x5E7]={
  category="lo",
  direction="r",
 },
 [0x5E8]={
  category="lo",
  direction="r",
 },
 [0x5E9]={
  category="lo",
  direction="r",
 },
 [0x5EA]={
  category="lo",
  direction="r",
 },
 [0x5F0]={
  category="lo",
  direction="r",
 },
 [0x5F1]={
  category="lo",
  direction="r",
 },
 [0x5F2]={
  category="lo",
  direction="r",
 },
 [0x5F3]={
  category="po",
  direction="r",
 },
 [0x5F4]={
  category="po",
  direction="r",
 },
 [0x600]={
  category="cf",
  direction="an",
 },
 [0x601]={
  category="cf",
  direction="an",
 },
 [0x602]={
  category="cf",
  direction="an",
 },
 [0x603]={
  category="cf",
  direction="an",
 },
 [0x604]={
  category="cf",
  direction="an",
 },
 [0x605]={
  category="cf",
  direction="an",
 },
 [0x606]={
  category="sm",
  direction="on",
 },
 [0x607]={
  category="sm",
  direction="on",
 },
 [0x608]={
  category="sm",
  direction="al",
 },
 [0x609]={
  category="po",
  direction="et",
 },
 [0x60A]={
  category="po",
  direction="et",
 },
 [0x60B]={
  category="sc",
  direction="al",
 },
 [0x60C]={
  category="po",
  direction="cs",
 },
 [0x60D]={
  category="po",
  direction="al",
 },
 [0x60E]={
  category="so",
  direction="on",
 },
 [0x60F]={
  category="so",
  direction="on",
 },
 [0x610]={
  category="mn",
  direction="nsm",
 },
 [0x611]={
  category="mn",
  direction="nsm",
 },
 [0x612]={
  category="mn",
  direction="nsm",
 },
 [0x613]={
  category="mn",
  direction="nsm",
 },
 [0x614]={
  category="mn",
  direction="nsm",
 },
 [0x615]={
  category="mn",
  direction="nsm",
 },
 [0x616]={
  category="mn",
  direction="nsm",
 },
 [0x617]={
  category="mn",
  direction="nsm",
 },
 [0x618]={
  category="mn",
  direction="nsm",
 },
 [0x619]={
  category="mn",
  direction="nsm",
 },
 [0x61A]={
  category="mn",
  direction="nsm",
 },
 [0x61B]={
  category="po",
  direction="al",
 },
 [0x61C]={
  category="cf",
  direction="al",
 },
 [0x61E]={
  category="po",
  direction="al",
 },
 [0x61F]={
  category="po",
  direction="al",
 },
 [0x620]={
  category="lo",
  direction="al",
 },
 [0x621]={
  category="lo",
  direction="al",
 },
 [0x622]={
  category="lo",
  direction="al",
 },
 [0x623]={
  category="lo",
  direction="al",
 },
 [0x624]={
  category="lo",
  direction="al",
 },
 [0x625]={
  category="lo",
  direction="al",
 },
 [0x626]={
  category="lo",
  direction="al",
 },
 [0x627]={
  category="lo",
  direction="al",
 },
 [0x628]={
  category="lo",
  direction="al",
 },
 [0x629]={
  category="lo",
  direction="al",
 },
 [0x62A]={
  category="lo",
  direction="al",
 },
 [0x62B]={
  category="lo",
  direction="al",
 },
 [0x62C]={
  category="lo",
  direction="al",
 },
 [0x62D]={
  category="lo",
  direction="al",
 },
 [0x62E]={
  category="lo",
  direction="al",
 },
 [0x62F]={
  category="lo",
  direction="al",
 },
 [0x630]={
  category="lo",
  direction="al",
 },
 [0x631]={
  category="lo",
  direction="al",
 },
 [0x632]={
  category="lo",
  direction="al",
 },
 [0x633]={
  category="lo",
  direction="al",
 },
 [0x634]={
  category="lo",
  direction="al",
 },
 [0x635]={
  category="lo",
  direction="al",
 },
 [0x636]={
  category="lo",
  direction="al",
 },
 [0x637]={
  category="lo",
  direction="al",
 },
 [0x638]={
  category="lo",
  direction="al",
 },
 [0x639]={
  category="lo",
  direction="al",
 },
 [0x63A]={
  category="lo",
  direction="al",
 },
 [0x63B]={
  category="lo",
  direction="al",
 },
 [0x63C]={
  category="lo",
  direction="al",
 },
 [0x63D]={
  category="lo",
  direction="al",
 },
 [0x63E]={
  category="lo",
  direction="al",
 },
 [0x63F]={
  category="lo",
  direction="al",
 },
 [0x640]={
  category="lm",
  direction="al",
 },
 [0x641]={
  category="lo",
  direction="al",
 },
 [0x642]={
  category="lo",
  direction="al",
 },
 [0x643]={
  category="lo",
  direction="al",
 },
 [0x644]={
  category="lo",
  direction="al",
 },
 [0x645]={
  category="lo",
  direction="al",
 },
 [0x646]={
  category="lo",
  direction="al",
 },
 [0x647]={
  category="lo",
  direction="al",
 },
 [0x648]={
  category="lo",
  direction="al",
 },
 [0x649]={
  category="lo",
  direction="al",
 },
 [0x64A]={
  category="lo",
  direction="al",
 },
 [0x64B]={
  category="mn",
  direction="nsm",
 },
 [0x64C]={
  category="mn",
  direction="nsm",
 },
 [0x64D]={
  category="mn",
  direction="nsm",
 },
 [0x64E]={
  category="mn",
  direction="nsm",
 },
 [0x64F]={
  category="mn",
  direction="nsm",
 },
 [0x650]={
  category="mn",
  direction="nsm",
 },
 [0x651]={
  category="mn",
  direction="nsm",
 },
 [0x652]={
  category="mn",
  direction="nsm",
 },
 [0x653]={
  category="mn",
  direction="nsm",
 },
 [0x654]={
  category="mn",
  direction="nsm",
 },
 [0x655]={
  category="mn",
  direction="nsm",
 },
 [0x656]={
  category="mn",
  direction="nsm",
 },
 [0x657]={
  category="mn",
  direction="nsm",
 },
 [0x658]={
  category="mn",
  direction="nsm",
 },
 [0x659]={
  category="mn",
  direction="nsm",
 },
 [0x65A]={
  category="mn",
  direction="nsm",
 },
 [0x65B]={
  category="mn",
  direction="nsm",
 },
 [0x65C]={
  category="mn",
  direction="nsm",
 },
 [0x65D]={
  category="mn",
  direction="nsm",
 },
 [0x65E]={
  category="mn",
  direction="nsm",
 },
 [0x65F]={
  category="mn",
  direction="nsm",
 },
 [0x660]={
  category="nd",
  direction="an",
 },
 [0x661]={
  category="nd",
  direction="an",
 },
 [0x662]={
  category="nd",
  direction="an",
 },
 [0x663]={
  category="nd",
  direction="an",
 },
 [0x664]={
  category="nd",
  direction="an",
 },
 [0x665]={
  category="nd",
  direction="an",
 },
 [0x666]={
  category="nd",
  direction="an",
 },
 [0x667]={
  category="nd",
  direction="an",
 },
 [0x668]={
  category="nd",
  direction="an",
 },
 [0x669]={
  category="nd",
  direction="an",
 },
 [0x66A]={
  category="po",
  direction="et",
 },
 [0x66B]={
  category="po",
  direction="an",
 },
 [0x66C]={
  category="po",
  direction="an",
 },
 [0x66D]={
  category="po",
  direction="al",
 },
 [0x66E]={
  category="lo",
  direction="al",
 },
 [0x66F]={
  category="lo",
  direction="al",
 },
 [0x670]={
  category="mn",
  direction="nsm",
 },
 [0x671]={
  category="lo",
  direction="al",
 },
 [0x672]={
  category="lo",
  direction="al",
 },
 [0x673]={
  category="lo",
  direction="al",
 },
 [0x674]={
  category="lo",
  direction="al",
 },
 [0x675]={
  category="lo",
  direction="al",
 },
 [0x676]={
  category="lo",
  direction="al",
 },
 [0x677]={
  category="lo",
  direction="al",
 },
 [0x678]={
  category="lo",
  direction="al",
 },
 [0x679]={
  category="lo",
  direction="al",
 },
 [0x67A]={
  category="lo",
  direction="al",
 },
 [0x67B]={
  category="lo",
  direction="al",
 },
 [0x67C]={
  category="lo",
  direction="al",
 },
 [0x67D]={
  category="lo",
  direction="al",
 },
 [0x67E]={
  category="lo",
  direction="al",
 },
 [0x67F]={
  category="lo",
  direction="al",
 },
 [0x680]={
  category="lo",
  direction="al",
 },
 [0x681]={
  category="lo",
  direction="al",
 },
 [0x682]={
  category="lo",
  direction="al",
 },
 [0x683]={
  category="lo",
  direction="al",
 },
 [0x684]={
  category="lo",
  direction="al",
 },
 [0x685]={
  category="lo",
  direction="al",
 },
 [0x686]={
  category="lo",
  direction="al",
 },
 [0x687]={
  category="lo",
  direction="al",
 },
 [0x688]={
  category="lo",
  direction="al",
 },
 [0x689]={
  category="lo",
  direction="al",
 },
 [0x68A]={
  category="lo",
  direction="al",
 },
 [0x68B]={
  category="lo",
  direction="al",
 },
 [0x68C]={
  category="lo",
  direction="al",
 },
 [0x68D]={
  category="lo",
  direction="al",
 },
 [0x68E]={
  category="lo",
  direction="al",
 },
 [0x68F]={
  category="lo",
  direction="al",
 },
 [0x690]={
  category="lo",
  direction="al",
 },
 [0x691]={
  category="lo",
  direction="al",
 },
 [0x692]={
  category="lo",
  direction="al",
 },
 [0x693]={
  category="lo",
  direction="al",
 },
 [0x694]={
  category="lo",
  direction="al",
 },
 [0x695]={
  category="lo",
  direction="al",
 },
 [0x696]={
  category="lo",
  direction="al",
 },
 [0x697]={
  category="lo",
  direction="al",
 },
 [0x698]={
  category="lo",
  direction="al",
 },
 [0x699]={
  category="lo",
  direction="al",
 },
 [0x69A]={
  category="lo",
  direction="al",
 },
 [0x69B]={
  category="lo",
  direction="al",
 },
 [0x69C]={
  category="lo",
  direction="al",
 },
 [0x69D]={
  category="lo",
  direction="al",
 },
 [0x69E]={
  category="lo",
  direction="al",
 },
 [0x69F]={
  category="lo",
  direction="al",
 },
 [0x6A0]={
  category="lo",
  direction="al",
 },
 [0x6A1]={
  category="lo",
  direction="al",
 },
 [0x6A2]={
  category="lo",
  direction="al",
 },
 [0x6A3]={
  category="lo",
  direction="al",
 },
 [0x6A4]={
  category="lo",
  direction="al",
 },
 [0x6A5]={
  category="lo",
  direction="al",
 },
 [0x6A6]={
  category="lo",
  direction="al",
 },
 [0x6A7]={
  category="lo",
  direction="al",
 },
 [0x6A8]={
  category="lo",
  direction="al",
 },
 [0x6A9]={
  category="lo",
  direction="al",
 },
 [0x6AA]={
  category="lo",
  direction="al",
 },
 [0x6AB]={
  category="lo",
  direction="al",
 },
 [0x6AC]={
  category="lo",
  direction="al",
 },
 [0x6AD]={
  category="lo",
  direction="al",
 },
 [0x6AE]={
  category="lo",
  direction="al",
 },
 [0x6AF]={
  category="lo",
  direction="al",
 },
 [0x6B0]={
  category="lo",
  direction="al",
 },
 [0x6B1]={
  category="lo",
  direction="al",
 },
 [0x6B2]={
  category="lo",
  direction="al",
 },
 [0x6B3]={
  category="lo",
  direction="al",
 },
 [0x6B4]={
  category="lo",
  direction="al",
 },
 [0x6B5]={
  category="lo",
  direction="al",
 },
 [0x6B6]={
  category="lo",
  direction="al",
 },
 [0x6B7]={
  category="lo",
  direction="al",
 },
 [0x6B8]={
  category="lo",
  direction="al",
 },
 [0x6B9]={
  category="lo",
  direction="al",
 },
 [0x6BA]={
  category="lo",
  direction="al",
 },
 [0x6BB]={
  category="lo",
  direction="al",
 },
 [0x6BC]={
  category="lo",
  direction="al",
 },
 [0x6BD]={
  category="lo",
  direction="al",
 },
 [0x6BE]={
  category="lo",
  direction="al",
 },
 [0x6BF]={
  category="lo",
  direction="al",
 },
 [0x6C0]={
  category="lo",
  direction="al",
 },
 [0x6C1]={
  category="lo",
  direction="al",
 },
 [0x6C2]={
  category="lo",
  direction="al",
 },
 [0x6C3]={
  category="lo",
  direction="al",
 },
 [0x6C4]={
  category="lo",
  direction="al",
 },
 [0x6C5]={
  category="lo",
  direction="al",
 },
 [0x6C6]={
  category="lo",
  direction="al",
 },
 [0x6C7]={
  category="lo",
  direction="al",
 },
 [0x6C8]={
  category="lo",
  direction="al",
 },
 [0x6C9]={
  category="lo",
  direction="al",
 },
 [0x6CA]={
  category="lo",
  direction="al",
 },
 [0x6CB]={
  category="lo",
  direction="al",
 },
 [0x6CC]={
  category="lo",
  direction="al",
 },
 [0x6CD]={
  category="lo",
  direction="al",
 },
 [0x6CE]={
  category="lo",
  direction="al",
 },
 [0x6CF]={
  category="lo",
  direction="al",
 },
 [0x6D0]={
  category="lo",
  direction="al",
 },
 [0x6D1]={
  category="lo",
  direction="al",
 },
 [0x6D2]={
  category="lo",
  direction="al",
 },
 [0x6D3]={
  category="lo",
  direction="al",
 },
 [0x6D4]={
  category="po",
  direction="al",
 },
 [0x6D5]={
  category="lo",
  direction="al",
 },
 [0x6D6]={
  category="mn",
  direction="nsm",
 },
 [0x6D7]={
  category="mn",
  direction="nsm",
 },
 [0x6D8]={
  category="mn",
  direction="nsm",
 },
 [0x6D9]={
  category="mn",
  direction="nsm",
 },
 [0x6DA]={
  category="mn",
  direction="nsm",
 },
 [0x6DB]={
  category="mn",
  direction="nsm",
 },
 [0x6DC]={
  category="mn",
  direction="nsm",
 },
 [0x6DD]={
  category="cf",
  direction="an",
 },
 [0x6DE]={
  category="me",
  direction="on",
 },
 [0x6DF]={
  category="mn",
  direction="nsm",
 },
 [0x6E0]={
  category="mn",
  direction="nsm",
 },
 [0x6E1]={
  category="mn",
  direction="nsm",
 },
 [0x6E2]={
  category="mn",
  direction="nsm",
 },
 [0x6E3]={
  category="mn",
  direction="nsm",
 },
 [0x6E4]={
  category="mn",
  direction="nsm",
 },
 [0x6E5]={
  category="lm",
  direction="al",
 },
 [0x6E6]={
  category="lm",
  direction="al",
 },
 [0x6E7]={
  category="mn",
  direction="nsm",
 },
 [0x6E8]={
  category="mn",
  direction="nsm",
 },
 [0x6E9]={
  category="so",
  direction="on",
 },
 [0x6EA]={
  category="mn",
  direction="nsm",
 },
 [0x6EB]={
  category="mn",
  direction="nsm",
 },
 [0x6EC]={
  category="mn",
  direction="nsm",
 },
 [0x6ED]={
  category="mn",
  direction="nsm",
 },
 [0x6EE]={
  category="lo",
  direction="al",
 },
 [0x6EF]={
  category="lo",
  direction="al",
 },
 [0x6F0]={
  category="nd",
  direction="en",
 },
 [0x6F1]={
  category="nd",
  direction="en",
 },
 [0x6F2]={
  category="nd",
  direction="en",
 },
 [0x6F3]={
  category="nd",
  direction="en",
 },
 [0x6F4]={
  category="nd",
  direction="en",
 },
 [0x6F5]={
  category="nd",
  direction="en",
 },
 [0x6F6]={
  category="nd",
  direction="en",
 },
 [0x6F7]={
  category="nd",
  direction="en",
 },
 [0x6F8]={
  category="nd",
  direction="en",
 },
 [0x6F9]={
  category="nd",
  direction="en",
 },
 [0x6FA]={
  category="lo",
  direction="al",
 },
 [0x6FB]={
  category="lo",
  direction="al",
 },
 [0x6FC]={
  category="lo",
  direction="al",
 },
 [0x6FD]={
  category="so",
  direction="al",
 },
 [0x6FE]={
  category="so",
  direction="al",
 },
 [0x6FF]={
  category="lo",
  direction="al",
 },
 [0x700]={
  category="po",
  direction="al",
 },
 [0x701]={
  category="po",
  direction="al",
 },
 [0x702]={
  category="po",
  direction="al",
 },
 [0x703]={
  category="po",
  direction="al",
 },
 [0x704]={
  category="po",
  direction="al",
 },
 [0x705]={
  category="po",
  direction="al",
 },
 [0x706]={
  category="po",
  direction="al",
 },
 [0x707]={
  category="po",
  direction="al",
 },
 [0x708]={
  category="po",
  direction="al",
 },
 [0x709]={
  category="po",
  direction="al",
 },
 [0x70A]={
  category="po",
  direction="al",
 },
 [0x70B]={
  category="po",
  direction="al",
 },
 [0x70C]={
  category="po",
  direction="al",
 },
 [0x70D]={
  category="po",
  direction="al",
 },
 [0x70F]={
  category="cf",
  direction="al",
 },
 [0x710]={
  category="lo",
  direction="al",
 },
 [0x711]={
  category="mn",
  direction="nsm",
 },
 [0x712]={
  category="lo",
  direction="al",
 },
 [0x713]={
  category="lo",
  direction="al",
 },
 [0x714]={
  category="lo",
  direction="al",
 },
 [0x715]={
  category="lo",
  direction="al",
 },
 [0x716]={
  category="lo",
  direction="al",
 },
 [0x717]={
  category="lo",
  direction="al",
 },
 [0x718]={
  category="lo",
  direction="al",
 },
 [0x719]={
  category="lo",
  direction="al",
 },
 [0x71A]={
  category="lo",
  direction="al",
 },
 [0x71B]={
  category="lo",
  direction="al",
 },
 [0x71C]={
  category="lo",
  direction="al",
 },
 [0x71D]={
  category="lo",
  direction="al",
 },
 [0x71E]={
  category="lo",
  direction="al",
 },
 [0x71F]={
  category="lo",
  direction="al",
 },
 [0x720]={
  category="lo",
  direction="al",
 },
 [0x721]={
  category="lo",
  direction="al",
 },
 [0x722]={
  category="lo",
  direction="al",
 },
 [0x723]={
  category="lo",
  direction="al",
 },
 [0x724]={
  category="lo",
  direction="al",
 },
 [0x725]={
  category="lo",
  direction="al",
 },
 [0x726]={
  category="lo",
  direction="al",
 },
 [0x727]={
  category="lo",
  direction="al",
 },
 [0x728]={
  category="lo",
  direction="al",
 },
 [0x729]={
  category="lo",
  direction="al",
 },
 [0x72A]={
  category="lo",
  direction="al",
 },
 [0x72B]={
  category="lo",
  direction="al",
 },
 [0x72C]={
  category="lo",
  direction="al",
 },
 [0x72D]={
  category="lo",
  direction="al",
 },
 [0x72E]={
  category="lo",
  direction="al",
 },
 [0x72F]={
  category="lo",
  direction="al",
 },
 [0x730]={
  category="mn",
  direction="nsm",
 },
 [0x731]={
  category="mn",
  direction="nsm",
 },
 [0x732]={
  category="mn",
  direction="nsm",
 },
 [0x733]={
  category="mn",
  direction="nsm",
 },
 [0x734]={
  category="mn",
  direction="nsm",
 },
 [0x735]={
  category="mn",
  direction="nsm",
 },
 [0x736]={
  category="mn",
  direction="nsm",
 },
 [0x737]={
  category="mn",
  direction="nsm",
 },
 [0x738]={
  category="mn",
  direction="nsm",
 },
 [0x739]={
  category="mn",
  direction="nsm",
 },
 [0x73A]={
  category="mn",
  direction="nsm",
 },
 [0x73B]={
  category="mn",
  direction="nsm",
 },
 [0x73C]={
  category="mn",
  direction="nsm",
 },
 [0x73D]={
  category="mn",
  direction="nsm",
 },
 [0x73E]={
  category="mn",
  direction="nsm",
 },
 [0x73F]={
  category="mn",
  direction="nsm",
 },
 [0x740]={
  category="mn",
  direction="nsm",
 },
 [0x741]={
  category="mn",
  direction="nsm",
 },
 [0x742]={
  category="mn",
  direction="nsm",
 },
 [0x743]={
  category="mn",
  direction="nsm",
 },
 [0x744]={
  category="mn",
  direction="nsm",
 },
 [0x745]={
  category="mn",
  direction="nsm",
 },
 [0x746]={
  category="mn",
  direction="nsm",
 },
 [0x747]={
  category="mn",
  direction="nsm",
 },
 [0x748]={
  category="mn",
  direction="nsm",
 },
 [0x749]={
  category="mn",
  direction="nsm",
 },
 [0x74A]={
  category="mn",
  direction="nsm",
 },
 [0x74D]={
  category="lo",
  direction="al",
 },
 [0x74E]={
  category="lo",
  direction="al",
 },
 [0x74F]={
  category="lo",
  direction="al",
 },
 [0x750]={
  category="lo",
  direction="al",
 },
 [0x751]={
  category="lo",
  direction="al",
 },
 [0x752]={
  category="lo",
  direction="al",
 },
 [0x753]={
  category="lo",
  direction="al",
 },
 [0x754]={
  category="lo",
  direction="al",
 },
 [0x755]={
  category="lo",
  direction="al",
 },
 [0x756]={
  category="lo",
  direction="al",
 },
 [0x757]={
  category="lo",
  direction="al",
 },
 [0x758]={
  category="lo",
  direction="al",
 },
 [0x759]={
  category="lo",
  direction="al",
 },
 [0x75A]={
  category="lo",
  direction="al",
 },
 [0x75B]={
  category="lo",
  direction="al",
 },
 [0x75C]={
  category="lo",
  direction="al",
 },
 [0x75D]={
  category="lo",
  direction="al",
 },
 [0x75E]={
  category="lo",
  direction="al",
 },
 [0x75F]={
  category="lo",
  direction="al",
 },
 [0x760]={
  category="lo",
  direction="al",
 },
 [0x761]={
  category="lo",
  direction="al",
 },
 [0x762]={
  category="lo",
  direction="al",
 },
 [0x763]={
  category="lo",
  direction="al",
 },
 [0x764]={
  category="lo",
  direction="al",
 },
 [0x765]={
  category="lo",
  direction="al",
 },
 [0x766]={
  category="lo",
  direction="al",
 },
 [0x767]={
  category="lo",
  direction="al",
 },
 [0x768]={
  category="lo",
  direction="al",
 },
 [0x769]={
  category="lo",
  direction="al",
 },
 [0x76A]={
  category="lo",
  direction="al",
 },
 [0x76B]={
  category="lo",
  direction="al",
 },
 [0x76C]={
  category="lo",
  direction="al",
 },
 [0x76D]={
  category="lo",
  direction="al",
 },
 [0x76E]={
  category="lo",
  direction="al",
 },
 [0x76F]={
  category="lo",
  direction="al",
 },
 [0x770]={
  category="lo",
  direction="al",
 },
 [0x771]={
  category="lo",
  direction="al",
 },
 [0x772]={
  category="lo",
  direction="al",
 },
 [0x773]={
  category="lo",
  direction="al",
 },
 [0x774]={
  category="lo",
  direction="al",
 },
 [0x775]={
  category="lo",
  direction="al",
 },
 [0x776]={
  category="lo",
  direction="al",
 },
 [0x777]={
  category="lo",
  direction="al",
 },
 [0x778]={
  category="lo",
  direction="al",
 },
 [0x779]={
  category="lo",
  direction="al",
 },
 [0x77A]={
  category="lo",
  direction="al",
 },
 [0x77B]={
  category="lo",
  direction="al",
 },
 [0x77C]={
  category="lo",
  direction="al",
 },
 [0x77D]={
  category="lo",
  direction="al",
 },
 [0x77E]={
  category="lo",
  direction="al",
 },
 [0x77F]={
  category="lo",
  direction="al",
 },
 [0x780]={
  category="lo",
  direction="al",
 },
 [0x781]={
  category="lo",
  direction="al",
 },
 [0x782]={
  category="lo",
  direction="al",
 },
 [0x783]={
  category="lo",
  direction="al",
 },
 [0x784]={
  category="lo",
  direction="al",
 },
 [0x785]={
  category="lo",
  direction="al",
 },
 [0x786]={
  category="lo",
  direction="al",
 },
 [0x787]={
  category="lo",
  direction="al",
 },
 [0x788]={
  category="lo",
  direction="al",
 },
 [0x789]={
  category="lo",
  direction="al",
 },
 [0x78A]={
  category="lo",
  direction="al",
 },
 [0x78B]={
  category="lo",
  direction="al",
 },
 [0x78C]={
  category="lo",
  direction="al",
 },
 [0x78D]={
  category="lo",
  direction="al",
 },
 [0x78E]={
  category="lo",
  direction="al",
 },
 [0x78F]={
  category="lo",
  direction="al",
 },
 [0x790]={
  category="lo",
  direction="al",
 },
 [0x791]={
  category="lo",
  direction="al",
 },
 [0x792]={
  category="lo",
  direction="al",
 },
 [0x793]={
  category="lo",
  direction="al",
 },
 [0x794]={
  category="lo",
  direction="al",
 },
 [0x795]={
  category="lo",
  direction="al",
 },
 [0x796]={
  category="lo",
  direction="al",
 },
 [0x797]={
  category="lo",
  direction="al",
 },
 [0x798]={
  category="lo",
  direction="al",
 },
 [0x799]={
  category="lo",
  direction="al",
 },
 [0x79A]={
  category="lo",
  direction="al",
 },
 [0x79B]={
  category="lo",
  direction="al",
 },
 [0x79C]={
  category="lo",
  direction="al",
 },
 [0x79D]={
  category="lo",
  direction="al",
 },
 [0x79E]={
  category="lo",
  direction="al",
 },
 [0x79F]={
  category="lo",
  direction="al",
 },
 [0x7A0]={
  category="lo",
  direction="al",
 },
 [0x7A1]={
  category="lo",
  direction="al",
 },
 [0x7A2]={
  category="lo",
  direction="al",
 },
 [0x7A3]={
  category="lo",
  direction="al",
 },
 [0x7A4]={
  category="lo",
  direction="al",
 },
 [0x7A5]={
  category="lo",
  direction="al",
 },
 [0x7A6]={
  category="mn",
  direction="nsm",
 },
 [0x7A7]={
  category="mn",
  direction="nsm",
 },
 [0x7A8]={
  category="mn",
  direction="nsm",
 },
 [0x7A9]={
  category="mn",
  direction="nsm",
 },
 [0x7AA]={
  category="mn",
  direction="nsm",
 },
 [0x7AB]={
  category="mn",
  direction="nsm",
 },
 [0x7AC]={
  category="mn",
  direction="nsm",
 },
 [0x7AD]={
  category="mn",
  direction="nsm",
 },
 [0x7AE]={
  category="mn",
  direction="nsm",
 },
 [0x7AF]={
  category="mn",
  direction="nsm",
 },
 [0x7B0]={
  category="mn",
  direction="nsm",
 },
 [0x7B1]={
  category="lo",
  direction="al",
 },
 [0x7C0]={
  category="nd",
  direction="r",
 },
 [0x7C1]={
  category="nd",
  direction="r",
 },
 [0x7C2]={
  category="nd",
  direction="r",
 },
 [0x7C3]={
  category="nd",
  direction="r",
 },
 [0x7C4]={
  category="nd",
  direction="r",
 },
 [0x7C5]={
  category="nd",
  direction="r",
 },
 [0x7C6]={
  category="nd",
  direction="r",
 },
 [0x7C7]={
  category="nd",
  direction="r",
 },
 [0x7C8]={
  category="nd",
  direction="r",
 },
 [0x7C9]={
  category="nd",
  direction="r",
 },
 [0x7CA]={
  category="lo",
  direction="r",
 },
 [0x7CB]={
  category="lo",
  direction="r",
 },
 [0x7CC]={
  category="lo",
  direction="r",
 },
 [0x7CD]={
  category="lo",
  direction="r",
 },
 [0x7CE]={
  category="lo",
  direction="r",
 },
 [0x7CF]={
  category="lo",
  direction="r",
 },
 [0x7D0]={
  category="lo",
  direction="r",
 },
 [0x7D1]={
  category="lo",
  direction="r",
 },
 [0x7D2]={
  category="lo",
  direction="r",
 },
 [0x7D3]={
  category="lo",
  direction="r",
 },
 [0x7D4]={
  category="lo",
  direction="r",
 },
 [0x7D5]={
  category="lo",
  direction="r",
 },
 [0x7D6]={
  category="lo",
  direction="r",
 },
 [0x7D7]={
  category="lo",
  direction="r",
 },
 [0x7D8]={
  category="lo",
  direction="r",
 },
 [0x7D9]={
  category="lo",
  direction="r",
 },
 [0x7DA]={
  category="lo",
  direction="r",
 },
 [0x7DB]={
  category="lo",
  direction="r",
 },
 [0x7DC]={
  category="lo",
  direction="r",
 },
 [0x7DD]={
  category="lo",
  direction="r",
 },
 [0x7DE]={
  category="lo",
  direction="r",
 },
 [0x7DF]={
  category="lo",
  direction="r",
 },
 [0x7E0]={
  category="lo",
  direction="r",
 },
 [0x7E1]={
  category="lo",
  direction="r",
 },
 [0x7E2]={
  category="lo",
  direction="r",
 },
 [0x7E3]={
  category="lo",
  direction="r",
 },
 [0x7E4]={
  category="lo",
  direction="r",
 },
 [0x7E5]={
  category="lo",
  direction="r",
 },
 [0x7E6]={
  category="lo",
  direction="r",
 },
 [0x7E7]={
  category="lo",
  direction="r",
 },
 [0x7E8]={
  category="lo",
  direction="r",
 },
 [0x7E9]={
  category="lo",
  direction="r",
 },
 [0x7EA]={
  category="lo",
  direction="r",
 },
 [0x7EB]={
  category="mn",
  direction="nsm",
 },
 [0x7EC]={
  category="mn",
  direction="nsm",
 },
 [0x7ED]={
  category="mn",
  direction="nsm",
 },
 [0x7EE]={
  category="mn",
  direction="nsm",
 },
 [0x7EF]={
  category="mn",
  direction="nsm",
 },
 [0x7F0]={
  category="mn",
  direction="nsm",
 },
 [0x7F1]={
  category="mn",
  direction="nsm",
 },
 [0x7F2]={
  category="mn",
  direction="nsm",
 },
 [0x7F3]={
  category="mn",
  direction="nsm",
 },
 [0x7F4]={
  category="lm",
  direction="r",
 },
 [0x7F5]={
  category="lm",
  direction="r",
 },
 [0x7F6]={
  category="so",
  direction="on",
 },
 [0x7F7]={
  category="po",
  direction="on",
 },
 [0x7F8]={
  category="po",
  direction="on",
 },
 [0x7F9]={
  category="po",
  direction="on",
 },
 [0x7FA]={
  category="lm",
  direction="r",
 },
 [0x800]={
  category="lo",
  direction="r",
 },
 [0x801]={
  category="lo",
  direction="r",
 },
 [0x802]={
  category="lo",
  direction="r",
 },
 [0x803]={
  category="lo",
  direction="r",
 },
 [0x804]={
  category="lo",
  direction="r",
 },
 [0x805]={
  category="lo",
  direction="r",
 },
 [0x806]={
  category="lo",
  direction="r",
 },
 [0x807]={
  category="lo",
  direction="r",
 },
 [0x808]={
  category="lo",
  direction="r",
 },
 [0x809]={
  category="lo",
  direction="r",
 },
 [0x80A]={
  category="lo",
  direction="r",
 },
 [0x80B]={
  category="lo",
  direction="r",
 },
 [0x80C]={
  category="lo",
  direction="r",
 },
 [0x80D]={
  category="lo",
  direction="r",
 },
 [0x80E]={
  category="lo",
  direction="r",
 },
 [0x80F]={
  category="lo",
  direction="r",
 },
 [0x810]={
  category="lo",
  direction="r",
 },
 [0x811]={
  category="lo",
  direction="r",
 },
 [0x812]={
  category="lo",
  direction="r",
 },
 [0x813]={
  category="lo",
  direction="r",
 },
 [0x814]={
  category="lo",
  direction="r",
 },
 [0x815]={
  category="lo",
  direction="r",
 },
 [0x816]={
  category="mn",
  direction="nsm",
 },
 [0x817]={
  category="mn",
  direction="nsm",
 },
 [0x818]={
  category="mn",
  direction="nsm",
 },
 [0x819]={
  category="mn",
  direction="nsm",
 },
 [0x81A]={
  category="lm",
  direction="r",
 },
 [0x81B]={
  category="mn",
  direction="nsm",
 },
 [0x81C]={
  category="mn",
  direction="nsm",
 },
 [0x81D]={
  category="mn",
  direction="nsm",
 },
 [0x81E]={
  category="mn",
  direction="nsm",
 },
 [0x81F]={
  category="mn",
  direction="nsm",
 },
 [0x820]={
  category="mn",
  direction="nsm",
 },
 [0x821]={
  category="mn",
  direction="nsm",
 },
 [0x822]={
  category="mn",
  direction="nsm",
 },
 [0x823]={
  category="mn",
  direction="nsm",
 },
 [0x824]={
  category="lm",
  direction="r",
 },
 [0x825]={
  category="mn",
  direction="nsm",
 },
 [0x826]={
  category="mn",
  direction="nsm",
 },
 [0x827]={
  category="mn",
  direction="nsm",
 },
 [0x828]={
  category="lm",
  direction="r",
 },
 [0x829]={
  category="mn",
  direction="nsm",
 },
 [0x82A]={
  category="mn",
  direction="nsm",
 },
 [0x82B]={
  category="mn",
  direction="nsm",
 },
 [0x82C]={
  category="mn",
  direction="nsm",
 },
 [0x82D]={
  category="mn",
  direction="nsm",
 },
 [0x830]={
  category="po",
  direction="r",
 },
 [0x831]={
  category="po",
  direction="r",
 },
 [0x832]={
  category="po",
  direction="r",
 },
 [0x833]={
  category="po",
  direction="r",
 },
 [0x834]={
  category="po",
  direction="r",
 },
 [0x835]={
  category="po",
  direction="r",
 },
 [0x836]={
  category="po",
  direction="r",
 },
 [0x837]={
  category="po",
  direction="r",
 },
 [0x838]={
  category="po",
  direction="r",
 },
 [0x839]={
  category="po",
  direction="r",
 },
 [0x83A]={
  category="po",
  direction="r",
 },
 [0x83B]={
  category="po",
  direction="r",
 },
 [0x83C]={
  category="po",
  direction="r",
 },
 [0x83D]={
  category="po",
  direction="r",
 },
 [0x83E]={
  category="po",
  direction="r",
 },
 [0x840]={
  category="lo",
  direction="r",
 },
 [0x841]={
  category="lo",
  direction="r",
 },
 [0x842]={
  category="lo",
  direction="r",
 },
 [0x843]={
  category="lo",
  direction="r",
 },
 [0x844]={
  category="lo",
  direction="r",
 },
 [0x845]={
  category="lo",
  direction="r",
 },
 [0x846]={
  category="lo",
  direction="r",
 },
 [0x847]={
  category="lo",
  direction="r",
 },
 [0x848]={
  category="lo",
  direction="r",
 },
 [0x849]={
  category="lo",
  direction="r",
 },
 [0x84A]={
  category="lo",
  direction="r",
 },
 [0x84B]={
  category="lo",
  direction="r",
 },
 [0x84C]={
  category="lo",
  direction="r",
 },
 [0x84D]={
  category="lo",
  direction="r",
 },
 [0x84E]={
  category="lo",
  direction="r",
 },
 [0x84F]={
  category="lo",
  direction="r",
 },
 [0x850]={
  category="lo",
  direction="r",
 },
 [0x851]={
  category="lo",
  direction="r",
 },
 [0x852]={
  category="lo",
  direction="r",
 },
 [0x853]={
  category="lo",
  direction="r",
 },
 [0x854]={
  category="lo",
  direction="r",
 },
 [0x855]={
  category="lo",
  direction="r",
 },
 [0x856]={
  category="lo",
  direction="r",
 },
 [0x857]={
  category="lo",
  direction="r",
 },
 [0x858]={
  category="lo",
  direction="r",
 },
 [0x859]={
  category="mn",
  direction="nsm",
 },
 [0x85A]={
  category="mn",
  direction="nsm",
 },
 [0x85B]={
  category="mn",
  direction="nsm",
 },
 [0x85E]={
  category="po",
  direction="r",
 },
 [0x8A0]={
  category="lo",
  direction="al",
 },
 [0x8A1]={
  category="lo",
  direction="al",
 },
 [0x8A2]={
  category="lo",
  direction="al",
 },
 [0x8A3]={
  category="lo",
  direction="al",
 },
 [0x8A4]={
  category="lo",
  direction="al",
 },
 [0x8A5]={
  category="lo",
  direction="al",
 },
 [0x8A6]={
  category="lo",
  direction="al",
 },
 [0x8A7]={
  category="lo",
  direction="al",
 },
 [0x8A8]={
  category="lo",
  direction="al",
 },
 [0x8A9]={
  category="lo",
  direction="al",
 },
 [0x8AA]={
  category="lo",
  direction="al",
 },
 [0x8AB]={
  category="lo",
  direction="al",
 },
 [0x8AC]={
  category="lo",
  direction="al",
 },
 [0x8AD]={
  category="lo",
  direction="al",
 },
 [0x8AE]={
  category="lo",
  direction="al",
 },
 [0x8AF]={
  category="lo",
  direction="al",
 },
 [0x8B0]={
  category="lo",
  direction="al",
 },
 [0x8B1]={
  category="lo",
  direction="al",
 },
 [0x8B2]={
  category="lo",
  direction="al",
 },
 [0x8E4]={
  category="mn",
  direction="nsm",
 },
 [0x8E5]={
  category="mn",
  direction="nsm",
 },
 [0x8E6]={
  category="mn",
  direction="nsm",
 },
 [0x8E7]={
  category="mn",
  direction="nsm",
 },
 [0x8E8]={
  category="mn",
  direction="nsm",
 },
 [0x8E9]={
  category="mn",
  direction="nsm",
 },
 [0x8EA]={
  category="mn",
  direction="nsm",
 },
 [0x8EB]={
  category="mn",
  direction="nsm",
 },
 [0x8EC]={
  category="mn",
  direction="nsm",
 },
 [0x8ED]={
  category="mn",
  direction="nsm",
 },
 [0x8EE]={
  category="mn",
  direction="nsm",
 },
 [0x8EF]={
  category="mn",
  direction="nsm",
 },
 [0x8F0]={
  category="mn",
  direction="nsm",
 },
 [0x8F1]={
  category="mn",
  direction="nsm",
 },
 [0x8F2]={
  category="mn",
  direction="nsm",
 },
 [0x8F3]={
  category="mn",
  direction="nsm",
 },
 [0x8F4]={
  category="mn",
  direction="nsm",
 },
 [0x8F5]={
  category="mn",
  direction="nsm",
 },
 [0x8F6]={
  category="mn",
  direction="nsm",
 },
 [0x8F7]={
  category="mn",
  direction="nsm",
 },
 [0x8F8]={
  category="mn",
  direction="nsm",
 },
 [0x8F9]={
  category="mn",
  direction="nsm",
 },
 [0x8FA]={
  category="mn",
  direction="nsm",
 },
 [0x8FB]={
  category="mn",
  direction="nsm",
 },
 [0x8FC]={
  category="mn",
  direction="nsm",
 },
 [0x8FD]={
  category="mn",
  direction="nsm",
 },
 [0x8FE]={
  category="mn",
  direction="nsm",
 },
 [0x8FF]={
  category="mn",
  direction="nsm",
 },
 [0x900]={
  category="mn",
  direction="nsm",
 },
 [0x901]={
  category="mn",
  direction="nsm",
 },
 [0x902]={
  category="mn",
  direction="nsm",
 },
 [0x903]={
  category="mc",
  direction="l",
 },
 [0x904]={
  category="lo",
  direction="l",
 },
 [0x905]={
  category="lo",
  direction="l",
 },
 [0x906]={
  category="lo",
  direction="l",
 },
 [0x907]={
  category="lo",
  direction="l",
 },
 [0x908]={
  category="lo",
  direction="l",
 },
 [0x909]={
  category="lo",
  direction="l",
 },
 [0x90A]={
  category="lo",
  direction="l",
 },
 [0x90B]={
  category="lo",
  direction="l",
 },
 [0x90C]={
  category="lo",
  direction="l",
 },
 [0x90D]={
  category="lo",
  direction="l",
 },
 [0x90E]={
  category="lo",
  direction="l",
 },
 [0x90F]={
  category="lo",
  direction="l",
 },
 [0x910]={
  category="lo",
  direction="l",
 },
 [0x911]={
  category="lo",
  direction="l",
 },
 [0x912]={
  category="lo",
  direction="l",
 },
 [0x913]={
  category="lo",
  direction="l",
 },
 [0x914]={
  category="lo",
  direction="l",
 },
 [0x915]={
  category="lo",
  direction="l",
 },
 [0x916]={
  category="lo",
  direction="l",
 },
 [0x917]={
  category="lo",
  direction="l",
 },
 [0x918]={
  category="lo",
  direction="l",
 },
 [0x919]={
  category="lo",
  direction="l",
 },
 [0x91A]={
  category="lo",
  direction="l",
 },
 [0x91B]={
  category="lo",
  direction="l",
 },
 [0x91C]={
  category="lo",
  direction="l",
 },
 [0x91D]={
  category="lo",
  direction="l",
 },
 [0x91E]={
  category="lo",
  direction="l",
 },
 [0x91F]={
  category="lo",
  direction="l",
 },
 [0x920]={
  category="lo",
  direction="l",
 },
 [0x921]={
  category="lo",
  direction="l",
 },
 [0x922]={
  category="lo",
  direction="l",
 },
 [0x923]={
  category="lo",
  direction="l",
 },
 [0x924]={
  category="lo",
  direction="l",
 },
 [0x925]={
  category="lo",
  direction="l",
 },
 [0x926]={
  category="lo",
  direction="l",
 },
 [0x927]={
  category="lo",
  direction="l",
 },
 [0x928]={
  category="lo",
  direction="l",
 },
 [0x929]={
  category="lo",
  direction="l",
 },
 [0x92A]={
  category="lo",
  direction="l",
 },
 [0x92B]={
  category="lo",
  direction="l",
 },
 [0x92C]={
  category="lo",
  direction="l",
 },
 [0x92D]={
  category="lo",
  direction="l",
 },
 [0x92E]={
  category="lo",
  direction="l",
 },
 [0x92F]={
  category="lo",
  direction="l",
 },
 [0x930]={
  category="lo",
  direction="l",
 },
 [0x931]={
  category="lo",
  direction="l",
 },
 [0x932]={
  category="lo",
  direction="l",
 },
 [0x933]={
  category="lo",
  direction="l",
 },
 [0x934]={
  category="lo",
  direction="l",
 },
 [0x935]={
  category="lo",
  direction="l",
 },
 [0x936]={
  category="lo",
  direction="l",
 },
 [0x937]={
  category="lo",
  direction="l",
 },
 [0x938]={
  category="lo",
  direction="l",
 },
 [0x939]={
  category="lo",
  direction="l",
 },
 [0x93A]={
  category="mn",
  direction="nsm",
 },
 [0x93B]={
  category="mc",
  direction="l",
 },
 [0x93C]={
  category="mn",
  direction="nsm",
 },
 [0x93D]={
  category="lo",
  direction="l",
 },
 [0x93E]={
  category="mc",
  direction="l",
 },
 [0x93F]={
  category="mc",
  direction="l",
 },
 [0x940]={
  category="mc",
  direction="l",
 },
 [0x941]={
  category="mn",
  direction="nsm",
 },
 [0x942]={
  category="mn",
  direction="nsm",
 },
 [0x943]={
  category="mn",
  direction="nsm",
 },
 [0x944]={
  category="mn",
  direction="nsm",
 },
 [0x945]={
  category="mn",
  direction="nsm",
 },
 [0x946]={
  category="mn",
  direction="nsm",
 },
 [0x947]={
  category="mn",
  direction="nsm",
 },
 [0x948]={
  category="mn",
  direction="nsm",
 },
 [0x949]={
  category="mc",
  direction="l",
 },
 [0x94A]={
  category="mc",
  direction="l",
 },
 [0x94B]={
  category="mc",
  direction="l",
 },
 [0x94C]={
  category="mc",
  direction="l",
 },
 [0x94D]={
  category="mn",
  direction="nsm",
 },
 [0x94E]={
  category="mc",
  direction="l",
 },
 [0x94F]={
  category="mc",
  direction="l",
 },
 [0x950]={
  category="lo",
  direction="l",
 },
 [0x951]={
  category="mn",
  direction="nsm",
 },
 [0x952]={
  category="mn",
  direction="nsm",
 },
 [0x953]={
  category="mn",
  direction="nsm",
 },
 [0x954]={
  category="mn",
  direction="nsm",
 },
 [0x955]={
  category="mn",
  direction="nsm",
 },
 [0x956]={
  category="mn",
  direction="nsm",
 },
 [0x957]={
  category="mn",
  direction="nsm",
 },
 [0x958]={
  category="lo",
  direction="l",
 },
 [0x959]={
  category="lo",
  direction="l",
 },
 [0x95A]={
  category="lo",
  direction="l",
 },
 [0x95B]={
  category="lo",
  direction="l",
 },
 [0x95C]={
  category="lo",
  direction="l",
 },
 [0x95D]={
  category="lo",
  direction="l",
 },
 [0x95E]={
  category="lo",
  direction="l",
 },
 [0x95F]={
  category="lo",
  direction="l",
 },
 [0x960]={
  category="lo",
  direction="l",
 },
 [0x961]={
  category="lo",
  direction="l",
 },
 [0x962]={
  category="mn",
  direction="nsm",
 },
 [0x963]={
  category="mn",
  direction="nsm",
 },
 [0x964]={
  category="po",
  direction="l",
 },
 [0x965]={
  category="po",
  direction="l",
 },
 [0x966]={
  category="nd",
  direction="l",
 },
 [0x967]={
  category="nd",
  direction="l",
 },
 [0x968]={
  category="nd",
  direction="l",
 },
 [0x969]={
  category="nd",
  direction="l",
 },
 [0x96A]={
  category="nd",
  direction="l",
 },
 [0x96B]={
  category="nd",
  direction="l",
 },
 [0x96C]={
  category="nd",
  direction="l",
 },
 [0x96D]={
  category="nd",
  direction="l",
 },
 [0x96E]={
  category="nd",
  direction="l",
 },
 [0x96F]={
  category="nd",
  direction="l",
 },
 [0x970]={
  category="po",
  direction="l",
 },
 [0x971]={
  category="lm",
  direction="l",
 },
 [0x972]={
  category="lo",
  direction="l",
 },
 [0x973]={
  category="lo",
  direction="l",
 },
 [0x974]={
  category="lo",
  direction="l",
 },
 [0x975]={
  category="lo",
  direction="l",
 },
 [0x976]={
  category="lo",
  direction="l",
 },
 [0x977]={
  category="lo",
  direction="l",
 },
 [0x978]={
  category="lo",
  direction="l",
 },
 [0x979]={
  category="lo",
  direction="l",
 },
 [0x97A]={
  category="lo",
  direction="l",
 },
 [0x97B]={
  category="lo",
  direction="l",
 },
 [0x97C]={
  category="lo",
  direction="l",
 },
 [0x97D]={
  category="lo",
  direction="l",
 },
 [0x97E]={
  category="lo",
  direction="l",
 },
 [0x97F]={
  category="lo",
  direction="l",
 },
 [0x980]={
  category="lo",
  direction="l",
 },
 [0x981]={
  category="mn",
  direction="nsm",
 },
 [0x982]={
  category="mc",
  direction="l",
 },
 [0x983]={
  category="mc",
  direction="l",
 },
 [0x985]={
  category="lo",
  direction="l",
 },
 [0x986]={
  category="lo",
  direction="l",
 },
 [0x987]={
  category="lo",
  direction="l",
 },
 [0x988]={
  category="lo",
  direction="l",
 },
 [0x989]={
  category="lo",
  direction="l",
 },
 [0x98A]={
  category="lo",
  direction="l",
 },
 [0x98B]={
  category="lo",
  direction="l",
 },
 [0x98C]={
  category="lo",
  direction="l",
 },
 [0x98F]={
  category="lo",
  direction="l",
 },
 [0x990]={
  category="lo",
  direction="l",
 },
 [0x993]={
  category="lo",
  direction="l",
 },
 [0x994]={
  category="lo",
  direction="l",
 },
 [0x995]={
  category="lo",
  direction="l",
 },
 [0x996]={
  category="lo",
  direction="l",
 },
 [0x997]={
  category="lo",
  direction="l",
 },
 [0x998]={
  category="lo",
  direction="l",
 },
 [0x999]={
  category="lo",
  direction="l",
 },
 [0x99A]={
  category="lo",
  direction="l",
 },
 [0x99B]={
  category="lo",
  direction="l",
 },
 [0x99C]={
  category="lo",
  direction="l",
 },
 [0x99D]={
  category="lo",
  direction="l",
 },
 [0x99E]={
  category="lo",
  direction="l",
 },
 [0x99F]={
  category="lo",
  direction="l",
 },
 [0x9A0]={
  category="lo",
  direction="l",
 },
 [0x9A1]={
  category="lo",
  direction="l",
 },
 [0x9A2]={
  category="lo",
  direction="l",
 },
 [0x9A3]={
  category="lo",
  direction="l",
 },
 [0x9A4]={
  category="lo",
  direction="l",
 },
 [0x9A5]={
  category="lo",
  direction="l",
 },
 [0x9A6]={
  category="lo",
  direction="l",
 },
 [0x9A7]={
  category="lo",
  direction="l",
 },
 [0x9A8]={
  category="lo",
  direction="l",
 },
 [0x9AA]={
  category="lo",
  direction="l",
 },
 [0x9AB]={
  category="lo",
  direction="l",
 },
 [0x9AC]={
  category="lo",
  direction="l",
 },
 [0x9AD]={
  category="lo",
  direction="l",
 },
 [0x9AE]={
  category="lo",
  direction="l",
 },
 [0x9AF]={
  category="lo",
  direction="l",
 },
 [0x9B0]={
  category="lo",
  direction="l",
 },
 [0x9B2]={
  category="lo",
  direction="l",
 },
 [0x9B6]={
  category="lo",
  direction="l",
 },
 [0x9B7]={
  category="lo",
  direction="l",
 },
 [0x9B8]={
  category="lo",
  direction="l",
 },
 [0x9B9]={
  category="lo",
  direction="l",
 },
 [0x9BC]={
  category="mn",
  direction="nsm",
 },
 [0x9BD]={
  category="lo",
  direction="l",
 },
 [0x9BE]={
  category="mc",
  direction="l",
 },
 [0x9BF]={
  category="mc",
  direction="l",
 },
 [0x9C0]={
  category="mc",
  direction="l",
 },
 [0x9C1]={
  category="mn",
  direction="nsm",
 },
 [0x9C2]={
  category="mn",
  direction="nsm",
 },
 [0x9C3]={
  category="mn",
  direction="nsm",
 },
 [0x9C4]={
  category="mn",
  direction="nsm",
 },
 [0x9C7]={
  category="mc",
  direction="l",
 },
 [0x9C8]={
  category="mc",
  direction="l",
 },
 [0x9CB]={
  category="mc",
  direction="l",
 },
 [0x9CC]={
  category="mc",
  direction="l",
 },
 [0x9CD]={
  category="mn",
  direction="nsm",
 },
 [0x9CE]={
  category="lo",
  direction="l",
 },
 [0x9D7]={
  category="mc",
  direction="l",
 },
 [0x9DC]={
  category="lo",
  direction="l",
 },
 [0x9DD]={
  category="lo",
  direction="l",
 },
 [0x9DF]={
  category="lo",
  direction="l",
 },
 [0x9E0]={
  category="lo",
  direction="l",
 },
 [0x9E1]={
  category="lo",
  direction="l",
 },
 [0x9E2]={
  category="mn",
  direction="nsm",
 },
 [0x9E3]={
  category="mn",
  direction="nsm",
 },
 [0x9E6]={
  category="nd",
  direction="l",
 },
 [0x9E7]={
  category="nd",
  direction="l",
 },
 [0x9E8]={
  category="nd",
  direction="l",
 },
 [0x9E9]={
  category="nd",
  direction="l",
 },
 [0x9EA]={
  category="nd",
  direction="l",
 },
 [0x9EB]={
  category="nd",
  direction="l",
 },
 [0x9EC]={
  category="nd",
  direction="l",
 },
 [0x9ED]={
  category="nd",
  direction="l",
 },
 [0x9EE]={
  category="nd",
  direction="l",
 },
 [0x9EF]={
  category="nd",
  direction="l",
 },
 [0x9F0]={
  category="lo",
  direction="l",
 },
 [0x9F1]={
  category="lo",
  direction="l",
 },
 [0x9F2]={
  category="sc",
  direction="et",
 },
 [0x9F3]={
  category="sc",
  direction="et",
 },
 [0x9F4]={
  category="no",
  direction="l",
 },
 [0x9F5]={
  category="no",
  direction="l",
 },
 [0x9F6]={
  category="no",
  direction="l",
 },
 [0x9F7]={
  category="no",
  direction="l",
 },
 [0x9F8]={
  category="no",
  direction="l",
 },
 [0x9F9]={
  category="no",
  direction="l",
 },
 [0x9FA]={
  category="so",
  direction="l",
 },
 [0x9FB]={
  category="sc",
  direction="et",
 },
 [0xA01]={
  category="mn",
  direction="nsm",
 },
 [0xA02]={
  category="mn",
  direction="nsm",
 },
 [0xA03]={
  category="mc",
  direction="l",
 },
 [0xA05]={
  category="lo",
  direction="l",
 },
 [0xA06]={
  category="lo",
  direction="l",
 },
 [0xA07]={
  category="lo",
  direction="l",
 },
 [0xA08]={
  category="lo",
  direction="l",
 },
 [0xA09]={
  category="lo",
  direction="l",
 },
 [0xA0A]={
  category="lo",
  direction="l",
 },
 [0xA0F]={
  category="lo",
  direction="l",
 },
 [0xA10]={
  category="lo",
  direction="l",
 },
 [0xA13]={
  category="lo",
  direction="l",
 },
 [0xA14]={
  category="lo",
  direction="l",
 },
 [0xA15]={
  category="lo",
  direction="l",
 },
 [0xA16]={
  category="lo",
  direction="l",
 },
 [0xA17]={
  category="lo",
  direction="l",
 },
 [0xA18]={
  category="lo",
  direction="l",
 },
 [0xA19]={
  category="lo",
  direction="l",
 },
 [0xA1A]={
  category="lo",
  direction="l",
 },
 [0xA1B]={
  category="lo",
  direction="l",
 },
 [0xA1C]={
  category="lo",
  direction="l",
 },
 [0xA1D]={
  category="lo",
  direction="l",
 },
 [0xA1E]={
  category="lo",
  direction="l",
 },
 [0xA1F]={
  category="lo",
  direction="l",
 },
 [0xA20]={
  category="lo",
  direction="l",
 },
 [0xA21]={
  category="lo",
  direction="l",
 },
 [0xA22]={
  category="lo",
  direction="l",
 },
 [0xA23]={
  category="lo",
  direction="l",
 },
 [0xA24]={
  category="lo",
  direction="l",
 },
 [0xA25]={
  category="lo",
  direction="l",
 },
 [0xA26]={
  category="lo",
  direction="l",
 },
 [0xA27]={
  category="lo",
  direction="l",
 },
 [0xA28]={
  category="lo",
  direction="l",
 },
 [0xA2A]={
  category="lo",
  direction="l",
 },
 [0xA2B]={
  category="lo",
  direction="l",
 },
 [0xA2C]={
  category="lo",
  direction="l",
 },
 [0xA2D]={
  category="lo",
  direction="l",
 },
 [0xA2E]={
  category="lo",
  direction="l",
 },
 [0xA2F]={
  category="lo",
  direction="l",
 },
 [0xA30]={
  category="lo",
  direction="l",
 },
 [0xA32]={
  category="lo",
  direction="l",
 },
 [0xA33]={
  category="lo",
  direction="l",
 },
 [0xA35]={
  category="lo",
  direction="l",
 },
 [0xA36]={
  category="lo",
  direction="l",
 },
 [0xA38]={
  category="lo",
  direction="l",
 },
 [0xA39]={
  category="lo",
  direction="l",
 },
 [0xA3C]={
  category="mn",
  direction="nsm",
 },
 [0xA3E]={
  category="mc",
  direction="l",
 },
 [0xA3F]={
  category="mc",
  direction="l",
 },
 [0xA40]={
  category="mc",
  direction="l",
 },
 [0xA41]={
  category="mn",
  direction="nsm",
 },
 [0xA42]={
  category="mn",
  direction="nsm",
 },
 [0xA47]={
  category="mn",
  direction="nsm",
 },
 [0xA48]={
  category="mn",
  direction="nsm",
 },
 [0xA4B]={
  category="mn",
  direction="nsm",
 },
 [0xA4C]={
  category="mn",
  direction="nsm",
 },
 [0xA4D]={
  category="mn",
  direction="nsm",
 },
 [0xA51]={
  category="mn",
  direction="nsm",
 },
 [0xA59]={
  category="lo",
  direction="l",
 },
 [0xA5A]={
  category="lo",
  direction="l",
 },
 [0xA5B]={
  category="lo",
  direction="l",
 },
 [0xA5C]={
  category="lo",
  direction="l",
 },
 [0xA5E]={
  category="lo",
  direction="l",
 },
 [0xA66]={
  category="nd",
  direction="l",
 },
 [0xA67]={
  category="nd",
  direction="l",
 },
 [0xA68]={
  category="nd",
  direction="l",
 },
 [0xA69]={
  category="nd",
  direction="l",
 },
 [0xA6A]={
  category="nd",
  direction="l",
 },
 [0xA6B]={
  category="nd",
  direction="l",
 },
 [0xA6C]={
  category="nd",
  direction="l",
 },
 [0xA6D]={
  category="nd",
  direction="l",
 },
 [0xA6E]={
  category="nd",
  direction="l",
 },
 [0xA6F]={
  category="nd",
  direction="l",
 },
 [0xA70]={
  category="mn",
  direction="nsm",
 },
 [0xA71]={
  category="mn",
  direction="nsm",
 },
 [0xA72]={
  category="lo",
  direction="l",
 },
 [0xA73]={
  category="lo",
  direction="l",
 },
 [0xA74]={
  category="lo",
  direction="l",
 },
 [0xA75]={
  category="mn",
  direction="nsm",
 },
 [0xA81]={
  category="mn",
  direction="nsm",
 },
 [0xA82]={
  category="mn",
  direction="nsm",
 },
 [0xA83]={
  category="mc",
  direction="l",
 },
 [0xA85]={
  category="lo",
  direction="l",
 },
 [0xA86]={
  category="lo",
  direction="l",
 },
 [0xA87]={
  category="lo",
  direction="l",
 },
 [0xA88]={
  category="lo",
  direction="l",
 },
 [0xA89]={
  category="lo",
  direction="l",
 },
 [0xA8A]={
  category="lo",
  direction="l",
 },
 [0xA8B]={
  category="lo",
  direction="l",
 },
 [0xA8C]={
  category="lo",
  direction="l",
 },
 [0xA8D]={
  category="lo",
  direction="l",
 },
 [0xA8F]={
  category="lo",
  direction="l",
 },
 [0xA90]={
  category="lo",
  direction="l",
 },
 [0xA91]={
  category="lo",
  direction="l",
 },
 [0xA93]={
  category="lo",
  direction="l",
 },
 [0xA94]={
  category="lo",
  direction="l",
 },
 [0xA95]={
  category="lo",
  direction="l",
 },
 [0xA96]={
  category="lo",
  direction="l",
 },
 [0xA97]={
  category="lo",
  direction="l",
 },
 [0xA98]={
  category="lo",
  direction="l",
 },
 [0xA99]={
  category="lo",
  direction="l",
 },
 [0xA9A]={
  category="lo",
  direction="l",
 },
 [0xA9B]={
  category="lo",
  direction="l",
 },
 [0xA9C]={
  category="lo",
  direction="l",
 },
 [0xA9D]={
  category="lo",
  direction="l",
 },
 [0xA9E]={
  category="lo",
  direction="l",
 },
 [0xA9F]={
  category="lo",
  direction="l",
 },
 [0xAA0]={
  category="lo",
  direction="l",
 },
 [0xAA1]={
  category="lo",
  direction="l",
 },
 [0xAA2]={
  category="lo",
  direction="l",
 },
 [0xAA3]={
  category="lo",
  direction="l",
 },
 [0xAA4]={
  category="lo",
  direction="l",
 },
 [0xAA5]={
  category="lo",
  direction="l",
 },
 [0xAA6]={
  category="lo",
  direction="l",
 },
 [0xAA7]={
  category="lo",
  direction="l",
 },
 [0xAA8]={
  category="lo",
  direction="l",
 },
 [0xAAA]={
  category="lo",
  direction="l",
 },
 [0xAAB]={
  category="lo",
  direction="l",
 },
 [0xAAC]={
  category="lo",
  direction="l",
 },
 [0xAAD]={
  category="lo",
  direction="l",
 },
 [0xAAE]={
  category="lo",
  direction="l",
 },
 [0xAAF]={
  category="lo",
  direction="l",
 },
 [0xAB0]={
  category="lo",
  direction="l",
 },
 [0xAB2]={
  category="lo",
  direction="l",
 },
 [0xAB3]={
  category="lo",
  direction="l",
 },
 [0xAB5]={
  category="lo",
  direction="l",
 },
 [0xAB6]={
  category="lo",
  direction="l",
 },
 [0xAB7]={
  category="lo",
  direction="l",
 },
 [0xAB8]={
  category="lo",
  direction="l",
 },
 [0xAB9]={
  category="lo",
  direction="l",
 },
 [0xABC]={
  category="mn",
  direction="nsm",
 },
 [0xABD]={
  category="lo",
  direction="l",
 },
 [0xABE]={
  category="mc",
  direction="l",
 },
 [0xABF]={
  category="mc",
  direction="l",
 },
 [0xAC0]={
  category="mc",
  direction="l",
 },
 [0xAC1]={
  category="mn",
  direction="nsm",
 },
 [0xAC2]={
  category="mn",
  direction="nsm",
 },
 [0xAC3]={
  category="mn",
  direction="nsm",
 },
 [0xAC4]={
  category="mn",
  direction="nsm",
 },
 [0xAC5]={
  category="mn",
  direction="nsm",
 },
 [0xAC7]={
  category="mn",
  direction="nsm",
 },
 [0xAC8]={
  category="mn",
  direction="nsm",
 },
 [0xAC9]={
  category="mc",
  direction="l",
 },
 [0xACB]={
  category="mc",
  direction="l",
 },
 [0xACC]={
  category="mc",
  direction="l",
 },
 [0xACD]={
  category="mn",
  direction="nsm",
 },
 [0xAD0]={
  category="lo",
  direction="l",
 },
 [0xAE0]={
  category="lo",
  direction="l",
 },
 [0xAE1]={
  category="lo",
  direction="l",
 },
 [0xAE2]={
  category="mn",
  direction="nsm",
 },
 [0xAE3]={
  category="mn",
  direction="nsm",
 },
 [0xAE6]={
  category="nd",
  direction="l",
 },
 [0xAE7]={
  category="nd",
  direction="l",
 },
 [0xAE8]={
  category="nd",
  direction="l",
 },
 [0xAE9]={
  category="nd",
  direction="l",
 },
 [0xAEA]={
  category="nd",
  direction="l",
 },
 [0xAEB]={
  category="nd",
  direction="l",
 },
 [0xAEC]={
  category="nd",
  direction="l",
 },
 [0xAED]={
  category="nd",
  direction="l",
 },
 [0xAEE]={
  category="nd",
  direction="l",
 },
 [0xAEF]={
  category="nd",
  direction="l",
 },
 [0xAF0]={
  category="po",
  direction="l",
 },
 [0xAF1]={
  category="sc",
  direction="et",
 },
 [0xB01]={
  category="mn",
  direction="nsm",
 },
 [0xB02]={
  category="mc",
  direction="l",
 },
 [0xB03]={
  category="mc",
  direction="l",
 },
 [0xB05]={
  category="lo",
  direction="l",
 },
 [0xB06]={
  category="lo",
  direction="l",
 },
 [0xB07]={
  category="lo",
  direction="l",
 },
 [0xB08]={
  category="lo",
  direction="l",
 },
 [0xB09]={
  category="lo",
  direction="l",
 },
 [0xB0A]={
  category="lo",
  direction="l",
 },
 [0xB0B]={
  category="lo",
  direction="l",
 },
 [0xB0C]={
  category="lo",
  direction="l",
 },
 [0xB0F]={
  category="lo",
  direction="l",
 },
 [0xB10]={
  category="lo",
  direction="l",
 },
 [0xB13]={
  category="lo",
  direction="l",
 },
 [0xB14]={
  category="lo",
  direction="l",
 },
 [0xB15]={
  category="lo",
  direction="l",
 },
 [0xB16]={
  category="lo",
  direction="l",
 },
 [0xB17]={
  category="lo",
  direction="l",
 },
 [0xB18]={
  category="lo",
  direction="l",
 },
 [0xB19]={
  category="lo",
  direction="l",
 },
 [0xB1A]={
  category="lo",
  direction="l",
 },
 [0xB1B]={
  category="lo",
  direction="l",
 },
 [0xB1C]={
  category="lo",
  direction="l",
 },
 [0xB1D]={
  category="lo",
  direction="l",
 },
 [0xB1E]={
  category="lo",
  direction="l",
 },
 [0xB1F]={
  category="lo",
  direction="l",
 },
 [0xB20]={
  category="lo",
  direction="l",
 },
 [0xB21]={
  category="lo",
  direction="l",
 },
 [0xB22]={
  category="lo",
  direction="l",
 },
 [0xB23]={
  category="lo",
  direction="l",
 },
 [0xB24]={
  category="lo",
  direction="l",
 },
 [0xB25]={
  category="lo",
  direction="l",
 },
 [0xB26]={
  category="lo",
  direction="l",
 },
 [0xB27]={
  category="lo",
  direction="l",
 },
 [0xB28]={
  category="lo",
  direction="l",
 },
 [0xB2A]={
  category="lo",
  direction="l",
 },
 [0xB2B]={
  category="lo",
  direction="l",
 },
 [0xB2C]={
  category="lo",
  direction="l",
 },
 [0xB2D]={
  category="lo",
  direction="l",
 },
 [0xB2E]={
  category="lo",
  direction="l",
 },
 [0xB2F]={
  category="lo",
  direction="l",
 },
 [0xB30]={
  category="lo",
  direction="l",
 },
 [0xB32]={
  category="lo",
  direction="l",
 },
 [0xB33]={
  category="lo",
  direction="l",
 },
 [0xB35]={
  category="lo",
  direction="l",
 },
 [0xB36]={
  category="lo",
  direction="l",
 },
 [0xB37]={
  category="lo",
  direction="l",
 },
 [0xB38]={
  category="lo",
  direction="l",
 },
 [0xB39]={
  category="lo",
  direction="l",
 },
 [0xB3C]={
  category="mn",
  direction="nsm",
 },
 [0xB3D]={
  category="lo",
  direction="l",
 },
 [0xB3E]={
  category="mc",
  direction="l",
 },
 [0xB3F]={
  category="mn",
  direction="nsm",
 },
 [0xB40]={
  category="mc",
  direction="l",
 },
 [0xB41]={
  category="mn",
  direction="nsm",
 },
 [0xB42]={
  category="mn",
  direction="nsm",
 },
 [0xB43]={
  category="mn",
  direction="nsm",
 },
 [0xB44]={
  category="mn",
  direction="nsm",
 },
 [0xB47]={
  category="mc",
  direction="l",
 },
 [0xB48]={
  category="mc",
  direction="l",
 },
 [0xB4B]={
  category="mc",
  direction="l",
 },
 [0xB4C]={
  category="mc",
  direction="l",
 },
 [0xB4D]={
  category="mn",
  direction="nsm",
 },
 [0xB56]={
  category="mn",
  direction="nsm",
 },
 [0xB57]={
  category="mc",
  direction="l",
 },
 [0xB5C]={
  category="lo",
  direction="l",
 },
 [0xB5D]={
  category="lo",
  direction="l",
 },
 [0xB5F]={
  category="lo",
  direction="l",
 },
 [0xB60]={
  category="lo",
  direction="l",
 },
 [0xB61]={
  category="lo",
  direction="l",
 },
 [0xB62]={
  category="mn",
  direction="nsm",
 },
 [0xB63]={
  category="mn",
  direction="nsm",
 },
 [0xB66]={
  category="nd",
  direction="l",
 },
 [0xB67]={
  category="nd",
  direction="l",
 },
 [0xB68]={
  category="nd",
  direction="l",
 },
 [0xB69]={
  category="nd",
  direction="l",
 },
 [0xB6A]={
  category="nd",
  direction="l",
 },
 [0xB6B]={
  category="nd",
  direction="l",
 },
 [0xB6C]={
  category="nd",
  direction="l",
 },
 [0xB6D]={
  category="nd",
  direction="l",
 },
 [0xB6E]={
  category="nd",
  direction="l",
 },
 [0xB6F]={
  category="nd",
  direction="l",
 },
 [0xB70]={
  category="so",
  direction="l",
 },
 [0xB71]={
  category="lo",
  direction="l",
 },
 [0xB72]={
  category="no",
  direction="l",
 },
 [0xB73]={
  category="no",
  direction="l",
 },
 [0xB74]={
  category="no",
  direction="l",
 },
 [0xB75]={
  category="no",
  direction="l",
 },
 [0xB76]={
  category="no",
  direction="l",
 },
 [0xB77]={
  category="no",
  direction="l",
 },
 [0xB82]={
  category="mn",
  direction="nsm",
 },
 [0xB83]={
  category="lo",
  direction="l",
 },
 [0xB85]={
  category="lo",
  direction="l",
 },
 [0xB86]={
  category="lo",
  direction="l",
 },
 [0xB87]={
  category="lo",
  direction="l",
 },
 [0xB88]={
  category="lo",
  direction="l",
 },
 [0xB89]={
  category="lo",
  direction="l",
 },
 [0xB8A]={
  category="lo",
  direction="l",
 },
 [0xB8E]={
  category="lo",
  direction="l",
 },
 [0xB8F]={
  category="lo",
  direction="l",
 },
 [0xB90]={
  category="lo",
  direction="l",
 },
 [0xB92]={
  category="lo",
  direction="l",
 },
 [0xB93]={
  category="lo",
  direction="l",
 },
 [0xB94]={
  category="lo",
  direction="l",
 },
 [0xB95]={
  category="lo",
  direction="l",
 },
 [0xB99]={
  category="lo",
  direction="l",
 },
 [0xB9A]={
  category="lo",
  direction="l",
 },
 [0xB9C]={
  category="lo",
  direction="l",
 },
 [0xB9E]={
  category="lo",
  direction="l",
 },
 [0xB9F]={
  category="lo",
  direction="l",
 },
 [0xBA3]={
  category="lo",
  direction="l",
 },
 [0xBA4]={
  category="lo",
  direction="l",
 },
 [0xBA8]={
  category="lo",
  direction="l",
 },
 [0xBA9]={
  category="lo",
  direction="l",
 },
 [0xBAA]={
  category="lo",
  direction="l",
 },
 [0xBAE]={
  category="lo",
  direction="l",
 },
 [0xBAF]={
  category="lo",
  direction="l",
 },
 [0xBB0]={
  category="lo",
  direction="l",
 },
 [0xBB1]={
  category="lo",
  direction="l",
 },
 [0xBB2]={
  category="lo",
  direction="l",
 },
 [0xBB3]={
  category="lo",
  direction="l",
 },
 [0xBB4]={
  category="lo",
  direction="l",
 },
 [0xBB5]={
  category="lo",
  direction="l",
 },
 [0xBB6]={
  category="lo",
  direction="l",
 },
 [0xBB7]={
  category="lo",
  direction="l",
 },
 [0xBB8]={
  category="lo",
  direction="l",
 },
 [0xBB9]={
  category="lo",
  direction="l",
 },
 [0xBBE]={
  category="mc",
  direction="l",
 },
 [0xBBF]={
  category="mc",
  direction="l",
 },
 [0xBC0]={
  category="mn",
  direction="nsm",
 },
 [0xBC1]={
  category="mc",
  direction="l",
 },
 [0xBC2]={
  category="mc",
  direction="l",
 },
 [0xBC6]={
  category="mc",
  direction="l",
 },
 [0xBC7]={
  category="mc",
  direction="l",
 },
 [0xBC8]={
  category="mc",
  direction="l",
 },
 [0xBCA]={
  category="mc",
  direction="l",
 },
 [0xBCB]={
  category="mc",
  direction="l",
 },
 [0xBCC]={
  category="mc",
  direction="l",
 },
 [0xBCD]={
  category="mn",
  direction="nsm",
 },
 [0xBD0]={
  category="lo",
  direction="l",
 },
 [0xBD7]={
  category="mc",
  direction="l",
 },
 [0xBE6]={
  category="nd",
  direction="l",
 },
 [0xBE7]={
  category="nd",
  direction="l",
 },
 [0xBE8]={
  category="nd",
  direction="l",
 },
 [0xBE9]={
  category="nd",
  direction="l",
 },
 [0xBEA]={
  category="nd",
  direction="l",
 },
 [0xBEB]={
  category="nd",
  direction="l",
 },
 [0xBEC]={
  category="nd",
  direction="l",
 },
 [0xBED]={
  category="nd",
  direction="l",
 },
 [0xBEE]={
  category="nd",
  direction="l",
 },
 [0xBEF]={
  category="nd",
  direction="l",
 },
 [0xBF0]={
  category="no",
  direction="l",
 },
 [0xBF1]={
  category="no",
  direction="l",
 },
 [0xBF2]={
  category="no",
  direction="l",
 },
 [0xBF3]={
  category="so",
  direction="on",
 },
 [0xBF4]={
  category="so",
  direction="on",
 },
 [0xBF5]={
  category="so",
  direction="on",
 },
 [0xBF6]={
  category="so",
  direction="on",
 },
 [0xBF7]={
  category="so",
  direction="on",
 },
 [0xBF8]={
  category="so",
  direction="on",
 },
 [0xBF9]={
  category="sc",
  direction="et",
 },
 [0xBFA]={
  category="so",
  direction="on",
 },
 [0xC00]={
  category="mn",
  direction="nsm",
 },
 [0xC01]={
  category="mc",
  direction="l",
 },
 [0xC02]={
  category="mc",
  direction="l",
 },
 [0xC03]={
  category="mc",
  direction="l",
 },
 [0xC05]={
  category="lo",
  direction="l",
 },
 [0xC06]={
  category="lo",
  direction="l",
 },
 [0xC07]={
  category="lo",
  direction="l",
 },
 [0xC08]={
  category="lo",
  direction="l",
 },
 [0xC09]={
  category="lo",
  direction="l",
 },
 [0xC0A]={
  category="lo",
  direction="l",
 },
 [0xC0B]={
  category="lo",
  direction="l",
 },
 [0xC0C]={
  category="lo",
  direction="l",
 },
 [0xC0E]={
  category="lo",
  direction="l",
 },
 [0xC0F]={
  category="lo",
  direction="l",
 },
 [0xC10]={
  category="lo",
  direction="l",
 },
 [0xC12]={
  category="lo",
  direction="l",
 },
 [0xC13]={
  category="lo",
  direction="l",
 },
 [0xC14]={
  category="lo",
  direction="l",
 },
 [0xC15]={
  category="lo",
  direction="l",
 },
 [0xC16]={
  category="lo",
  direction="l",
 },
 [0xC17]={
  category="lo",
  direction="l",
 },
 [0xC18]={
  category="lo",
  direction="l",
 },
 [0xC19]={
  category="lo",
  direction="l",
 },
 [0xC1A]={
  category="lo",
  direction="l",
 },
 [0xC1B]={
  category="lo",
  direction="l",
 },
 [0xC1C]={
  category="lo",
  direction="l",
 },
 [0xC1D]={
  category="lo",
  direction="l",
 },
 [0xC1E]={
  category="lo",
  direction="l",
 },
 [0xC1F]={
  category="lo",
  direction="l",
 },
 [0xC20]={
  category="lo",
  direction="l",
 },
 [0xC21]={
  category="lo",
  direction="l",
 },
 [0xC22]={
  category="lo",
  direction="l",
 },
 [0xC23]={
  category="lo",
  direction="l",
 },
 [0xC24]={
  category="lo",
  direction="l",
 },
 [0xC25]={
  category="lo",
  direction="l",
 },
 [0xC26]={
  category="lo",
  direction="l",
 },
 [0xC27]={
  category="lo",
  direction="l",
 },
 [0xC28]={
  category="lo",
  direction="l",
 },
 [0xC2A]={
  category="lo",
  direction="l",
 },
 [0xC2B]={
  category="lo",
  direction="l",
 },
 [0xC2C]={
  category="lo",
  direction="l",
 },
 [0xC2D]={
  category="lo",
  direction="l",
 },
 [0xC2E]={
  category="lo",
  direction="l",
 },
 [0xC2F]={
  category="lo",
  direction="l",
 },
 [0xC30]={
  category="lo",
  direction="l",
 },
 [0xC31]={
  category="lo",
  direction="l",
 },
 [0xC32]={
  category="lo",
  direction="l",
 },
 [0xC33]={
  category="lo",
  direction="l",
 },
 [0xC34]={
  category="lo",
  direction="l",
 },
 [0xC35]={
  category="lo",
  direction="l",
 },
 [0xC36]={
  category="lo",
  direction="l",
 },
 [0xC37]={
  category="lo",
  direction="l",
 },
 [0xC38]={
  category="lo",
  direction="l",
 },
 [0xC39]={
  category="lo",
  direction="l",
 },
 [0xC3D]={
  category="lo",
  direction="l",
 },
 [0xC3E]={
  category="mn",
  direction="nsm",
 },
 [0xC3F]={
  category="mn",
  direction="nsm",
 },
 [0xC40]={
  category="mn",
  direction="nsm",
 },
 [0xC41]={
  category="mc",
  direction="l",
 },
 [0xC42]={
  category="mc",
  direction="l",
 },
 [0xC43]={
  category="mc",
  direction="l",
 },
 [0xC44]={
  category="mc",
  direction="l",
 },
 [0xC46]={
  category="mn",
  direction="nsm",
 },
 [0xC47]={
  category="mn",
  direction="nsm",
 },
 [0xC48]={
  category="mn",
  direction="nsm",
 },
 [0xC4A]={
  category="mn",
  direction="nsm",
 },
 [0xC4B]={
  category="mn",
  direction="nsm",
 },
 [0xC4C]={
  category="mn",
  direction="nsm",
 },
 [0xC4D]={
  category="mn",
  direction="nsm",
 },
 [0xC55]={
  category="mn",
  direction="nsm",
 },
 [0xC56]={
  category="mn",
  direction="nsm",
 },
 [0xC58]={
  category="lo",
  direction="l",
 },
 [0xC59]={
  category="lo",
  direction="l",
 },
 [0xC60]={
  category="lo",
  direction="l",
 },
 [0xC61]={
  category="lo",
  direction="l",
 },
 [0xC62]={
  category="mn",
  direction="nsm",
 },
 [0xC63]={
  category="mn",
  direction="nsm",
 },
 [0xC66]={
  category="nd",
  direction="l",
 },
 [0xC67]={
  category="nd",
  direction="l",
 },
 [0xC68]={
  category="nd",
  direction="l",
 },
 [0xC69]={
  category="nd",
  direction="l",
 },
 [0xC6A]={
  category="nd",
  direction="l",
 },
 [0xC6B]={
  category="nd",
  direction="l",
 },
 [0xC6C]={
  category="nd",
  direction="l",
 },
 [0xC6D]={
  category="nd",
  direction="l",
 },
 [0xC6E]={
  category="nd",
  direction="l",
 },
 [0xC6F]={
  category="nd",
  direction="l",
 },
 [0xC78]={
  category="no",
  direction="on",
 },
 [0xC79]={
  category="no",
  direction="on",
 },
 [0xC7A]={
  category="no",
  direction="on",
 },
 [0xC7B]={
  category="no",
  direction="on",
 },
 [0xC7C]={
  category="no",
  direction="on",
 },
 [0xC7D]={
  category="no",
  direction="on",
 },
 [0xC7E]={
  category="no",
  direction="on",
 },
 [0xC7F]={
  category="so",
  direction="l",
 },
 [0xC81]={
  category="mn",
  direction="nsm",
 },
 [0xC82]={
  category="mc",
  direction="l",
 },
 [0xC83]={
  category="mc",
  direction="l",
 },
 [0xC85]={
  category="lo",
  direction="l",
 },
 [0xC86]={
  category="lo",
  direction="l",
 },
 [0xC87]={
  category="lo",
  direction="l",
 },
 [0xC88]={
  category="lo",
  direction="l",
 },
 [0xC89]={
  category="lo",
  direction="l",
 },
 [0xC8A]={
  category="lo",
  direction="l",
 },
 [0xC8B]={
  category="lo",
  direction="l",
 },
 [0xC8C]={
  category="lo",
  direction="l",
 },
 [0xC8E]={
  category="lo",
  direction="l",
 },
 [0xC8F]={
  category="lo",
  direction="l",
 },
 [0xC90]={
  category="lo",
  direction="l",
 },
 [0xC92]={
  category="lo",
  direction="l",
 },
 [0xC93]={
  category="lo",
  direction="l",
 },
 [0xC94]={
  category="lo",
  direction="l",
 },
 [0xC95]={
  category="lo",
  direction="l",
 },
 [0xC96]={
  category="lo",
  direction="l",
 },
 [0xC97]={
  category="lo",
  direction="l",
 },
 [0xC98]={
  category="lo",
  direction="l",
 },
 [0xC99]={
  category="lo",
  direction="l",
 },
 [0xC9A]={
  category="lo",
  direction="l",
 },
 [0xC9B]={
  category="lo",
  direction="l",
 },
 [0xC9C]={
  category="lo",
  direction="l",
 },
 [0xC9D]={
  category="lo",
  direction="l",
 },
 [0xC9E]={
  category="lo",
  direction="l",
 },
 [0xC9F]={
  category="lo",
  direction="l",
 },
 [0xCA0]={
  category="lo",
  direction="l",
 },
 [0xCA1]={
  category="lo",
  direction="l",
 },
 [0xCA2]={
  category="lo",
  direction="l",
 },
 [0xCA3]={
  category="lo",
  direction="l",
 },
 [0xCA4]={
  category="lo",
  direction="l",
 },
 [0xCA5]={
  category="lo",
  direction="l",
 },
 [0xCA6]={
  category="lo",
  direction="l",
 },
 [0xCA7]={
  category="lo",
  direction="l",
 },
 [0xCA8]={
  category="lo",
  direction="l",
 },
 [0xCAA]={
  category="lo",
  direction="l",
 },
 [0xCAB]={
  category="lo",
  direction="l",
 },
 [0xCAC]={
  category="lo",
  direction="l",
 },
 [0xCAD]={
  category="lo",
  direction="l",
 },
 [0xCAE]={
  category="lo",
  direction="l",
 },
 [0xCAF]={
  category="lo",
  direction="l",
 },
 [0xCB0]={
  category="lo",
  direction="l",
 },
 [0xCB1]={
  category="lo",
  direction="l",
 },
 [0xCB2]={
  category="lo",
  direction="l",
 },
 [0xCB3]={
  category="lo",
  direction="l",
 },
 [0xCB5]={
  category="lo",
  direction="l",
 },
 [0xCB6]={
  category="lo",
  direction="l",
 },
 [0xCB7]={
  category="lo",
  direction="l",
 },
 [0xCB8]={
  category="lo",
  direction="l",
 },
 [0xCB9]={
  category="lo",
  direction="l",
 },
 [0xCBC]={
  category="mn",
  direction="nsm",
 },
 [0xCBD]={
  category="lo",
  direction="l",
 },
 [0xCBE]={
  category="mc",
  direction="l",
 },
 [0xCBF]={
  category="mn",
  direction="l",
 },
 [0xCC0]={
  category="mc",
  direction="l",
 },
 [0xCC1]={
  category="mc",
  direction="l",
 },
 [0xCC2]={
  category="mc",
  direction="l",
 },
 [0xCC3]={
  category="mc",
  direction="l",
 },
 [0xCC4]={
  category="mc",
  direction="l",
 },
 [0xCC6]={
  category="mn",
  direction="l",
 },
 [0xCC7]={
  category="mc",
  direction="l",
 },
 [0xCC8]={
  category="mc",
  direction="l",
 },
 [0xCCA]={
  category="mc",
  direction="l",
 },
 [0xCCB]={
  category="mc",
  direction="l",
 },
 [0xCCC]={
  category="mn",
  direction="nsm",
 },
 [0xCCD]={
  category="mn",
  direction="nsm",
 },
 [0xCD5]={
  category="mc",
  direction="l",
 },
 [0xCD6]={
  category="mc",
  direction="l",
 },
 [0xCDE]={
  category="lo",
  direction="l",
 },
 [0xCE0]={
  category="lo",
  direction="l",
 },
 [0xCE1]={
  category="lo",
  direction="l",
 },
 [0xCE2]={
  category="mn",
  direction="nsm",
 },
 [0xCE3]={
  category="mn",
  direction="nsm",
 },
 [0xCE6]={
  category="nd",
  direction="l",
 },
 [0xCE7]={
  category="nd",
  direction="l",
 },
 [0xCE8]={
  category="nd",
  direction="l",
 },
 [0xCE9]={
  category="nd",
  direction="l",
 },
 [0xCEA]={
  category="nd",
  direction="l",
 },
 [0xCEB]={
  category="nd",
  direction="l",
 },
 [0xCEC]={
  category="nd",
  direction="l",
 },
 [0xCED]={
  category="nd",
  direction="l",
 },
 [0xCEE]={
  category="nd",
  direction="l",
 },
 [0xCEF]={
  category="nd",
  direction="l",
 },
 [0xCF1]={
  category="so",
  direction="l",
 },
 [0xCF2]={
  category="so",
  direction="l",
 },
 [0xD01]={
  category="mn",
  direction="nsm",
 },
 [0xD02]={
  category="mc",
  direction="l",
 },
 [0xD03]={
  category="mc",
  direction="l",
 },
 [0xD05]={
  category="lo",
  direction="l",
 },
 [0xD06]={
  category="lo",
  direction="l",
 },
 [0xD07]={
  category="lo",
  direction="l",
 },
 [0xD08]={
  category="lo",
  direction="l",
 },
 [0xD09]={
  category="lo",
  direction="l",
 },
 [0xD0A]={
  category="lo",
  direction="l",
 },
 [0xD0B]={
  category="lo",
  direction="l",
 },
 [0xD0C]={
  category="lo",
  direction="l",
 },
 [0xD0E]={
  category="lo",
  direction="l",
 },
 [0xD0F]={
  category="lo",
  direction="l",
 },
 [0xD10]={
  category="lo",
  direction="l",
 },
 [0xD12]={
  category="lo",
  direction="l",
 },
 [0xD13]={
  category="lo",
  direction="l",
 },
 [0xD14]={
  category="lo",
  direction="l",
 },
 [0xD15]={
  category="lo",
  direction="l",
 },
 [0xD16]={
  category="lo",
  direction="l",
 },
 [0xD17]={
  category="lo",
  direction="l",
 },
 [0xD18]={
  category="lo",
  direction="l",
 },
 [0xD19]={
  category="lo",
  direction="l",
 },
 [0xD1A]={
  category="lo",
  direction="l",
 },
 [0xD1B]={
  category="lo",
  direction="l",
 },
 [0xD1C]={
  category="lo",
  direction="l",
 },
 [0xD1D]={
  category="lo",
  direction="l",
 },
 [0xD1E]={
  category="lo",
  direction="l",
 },
 [0xD1F]={
  category="lo",
  direction="l",
 },
 [0xD20]={
  category="lo",
  direction="l",
 },
 [0xD21]={
  category="lo",
  direction="l",
 },
 [0xD22]={
  category="lo",
  direction="l",
 },
 [0xD23]={
  category="lo",
  direction="l",
 },
 [0xD24]={
  category="lo",
  direction="l",
 },
 [0xD25]={
  category="lo",
  direction="l",
 },
 [0xD26]={
  category="lo",
  direction="l",
 },
 [0xD27]={
  category="lo",
  direction="l",
 },
 [0xD28]={
  category="lo",
  direction="l",
 },
 [0xD29]={
  category="lo",
  direction="l",
 },
 [0xD2A]={
  category="lo",
  direction="l",
 },
 [0xD2B]={
  category="lo",
  direction="l",
 },
 [0xD2C]={
  category="lo",
  direction="l",
 },
 [0xD2D]={
  category="lo",
  direction="l",
 },
 [0xD2E]={
  category="lo",
  direction="l",
 },
 [0xD2F]={
  category="lo",
  direction="l",
 },
 [0xD30]={
  category="lo",
  direction="l",
 },
 [0xD31]={
  category="lo",
  direction="l",
 },
 [0xD32]={
  category="lo",
  direction="l",
 },
 [0xD33]={
  category="lo",
  direction="l",
 },
 [0xD34]={
  category="lo",
  direction="l",
 },
 [0xD35]={
  category="lo",
  direction="l",
 },
 [0xD36]={
  category="lo",
  direction="l",
 },
 [0xD37]={
  category="lo",
  direction="l",
 },
 [0xD38]={
  category="lo",
  direction="l",
 },
 [0xD39]={
  category="lo",
  direction="l",
 },
 [0xD3A]={
  category="lo",
  direction="l",
 },
 [0xD3D]={
  category="lo",
  direction="l",
 },
 [0xD3E]={
  category="mc",
  direction="l",
 },
 [0xD3F]={
  category="mc",
  direction="l",
 },
 [0xD40]={
  category="mc",
  direction="l",
 },
 [0xD41]={
  category="mn",
  direction="nsm",
 },
 [0xD42]={
  category="mn",
  direction="nsm",
 },
 [0xD43]={
  category="mn",
  direction="nsm",
 },
 [0xD44]={
  category="mn",
  direction="nsm",
 },
 [0xD46]={
  category="mc",
  direction="l",
 },
 [0xD47]={
  category="mc",
  direction="l",
 },
 [0xD48]={
  category="mc",
  direction="l",
 },
 [0xD4A]={
  category="mc",
  direction="l",
 },
 [0xD4B]={
  category="mc",
  direction="l",
 },
 [0xD4C]={
  category="mc",
  direction="l",
 },
 [0xD4D]={
  category="mn",
  direction="nsm",
 },
 [0xD4E]={
  category="lo",
  direction="l",
 },
 [0xD57]={
  category="mc",
  direction="l",
 },
 [0xD60]={
  category="lo",
  direction="l",
 },
 [0xD61]={
  category="lo",
  direction="l",
 },
 [0xD62]={
  category="mn",
  direction="nsm",
 },
 [0xD63]={
  category="mn",
  direction="nsm",
 },
 [0xD66]={
  category="nd",
  direction="l",
 },
 [0xD67]={
  category="nd",
  direction="l",
 },
 [0xD68]={
  category="nd",
  direction="l",
 },
 [0xD69]={
  category="nd",
  direction="l",
 },
 [0xD6A]={
  category="nd",
  direction="l",
 },
 [0xD6B]={
  category="nd",
  direction="l",
 },
 [0xD6C]={
  category="nd",
  direction="l",
 },
 [0xD6D]={
  category="nd",
  direction="l",
 },
 [0xD6E]={
  category="nd",
  direction="l",
 },
 [0xD6F]={
  category="nd",
  direction="l",
 },
 [0xD70]={
  category="no",
  direction="l",
 },
 [0xD71]={
  category="no",
  direction="l",
 },
 [0xD72]={
  category="no",
  direction="l",
 },
 [0xD73]={
  category="no",
  direction="l",
 },
 [0xD74]={
  category="no",
  direction="l",
 },
 [0xD75]={
  category="no",
  direction="l",
 },
 [0xD79]={
  category="so",
  direction="l",
 },
 [0xD7A]={
  category="lo",
  direction="l",
 },
 [0xD7B]={
  category="lo",
  direction="l",
 },
 [0xD7C]={
  category="lo",
  direction="l",
 },
 [0xD7D]={
  category="lo",
  direction="l",
 },
 [0xD7E]={
  category="lo",
  direction="l",
 },
 [0xD7F]={
  category="lo",
  direction="l",
 },
 [0xD82]={
  category="mc",
  direction="l",
 },
 [0xD83]={
  category="mc",
  direction="l",
 },
 [0xD85]={
  category="lo",
  direction="l",
 },
 [0xD86]={
  category="lo",
  direction="l",
 },
 [0xD87]={
  category="lo",
  direction="l",
 },
 [0xD88]={
  category="lo",
  direction="l",
 },
 [0xD89]={
  category="lo",
  direction="l",
 },
 [0xD8A]={
  category="lo",
  direction="l",
 },
 [0xD8B]={
  category="lo",
  direction="l",
 },
 [0xD8C]={
  category="lo",
  direction="l",
 },
 [0xD8D]={
  category="lo",
  direction="l",
 },
 [0xD8E]={
  category="lo",
  direction="l",
 },
 [0xD8F]={
  category="lo",
  direction="l",
 },
 [0xD90]={
  category="lo",
  direction="l",
 },
 [0xD91]={
  category="lo",
  direction="l",
 },
 [0xD92]={
  category="lo",
  direction="l",
 },
 [0xD93]={
  category="lo",
  direction="l",
 },
 [0xD94]={
  category="lo",
  direction="l",
 },
 [0xD95]={
  category="lo",
  direction="l",
 },
 [0xD96]={
  category="lo",
  direction="l",
 },
 [0xD9A]={
  category="lo",
  direction="l",
 },
 [0xD9B]={
  category="lo",
  direction="l",
 },
 [0xD9C]={
  category="lo",
  direction="l",
 },
 [0xD9D]={
  category="lo",
  direction="l",
 },
 [0xD9E]={
  category="lo",
  direction="l",
 },
 [0xD9F]={
  category="lo",
  direction="l",
 },
 [0xDA0]={
  category="lo",
  direction="l",
 },
 [0xDA1]={
  category="lo",
  direction="l",
 },
 [0xDA2]={
  category="lo",
  direction="l",
 },
 [0xDA3]={
  category="lo",
  direction="l",
 },
 [0xDA4]={
  category="lo",
  direction="l",
 },
 [0xDA5]={
  category="lo",
  direction="l",
 },
 [0xDA6]={
  category="lo",
  direction="l",
 },
 [0xDA7]={
  category="lo",
  direction="l",
 },
 [0xDA8]={
  category="lo",
  direction="l",
 },
 [0xDA9]={
  category="lo",
  direction="l",
 },
 [0xDAA]={
  category="lo",
  direction="l",
 },
 [0xDAB]={
  category="lo",
  direction="l",
 },
 [0xDAC]={
  category="lo",
  direction="l",
 },
 [0xDAD]={
  category="lo",
  direction="l",
 },
 [0xDAE]={
  category="lo",
  direction="l",
 },
 [0xDAF]={
  category="lo",
  direction="l",
 },
 [0xDB0]={
  category="lo",
  direction="l",
 },
 [0xDB1]={
  category="lo",
  direction="l",
 },
 [0xDB3]={
  category="lo",
  direction="l",
 },
 [0xDB4]={
  category="lo",
  direction="l",
 },
 [0xDB5]={
  category="lo",
  direction="l",
 },
 [0xDB6]={
  category="lo",
  direction="l",
 },
 [0xDB7]={
  category="lo",
  direction="l",
 },
 [0xDB8]={
  category="lo",
  direction="l",
 },
 [0xDB9]={
  category="lo",
  direction="l",
 },
 [0xDBA]={
  category="lo",
  direction="l",
 },
 [0xDBB]={
  category="lo",
  direction="l",
 },
 [0xDBD]={
  category="lo",
  direction="l",
 },
 [0xDC0]={
  category="lo",
  direction="l",
 },
 [0xDC1]={
  category="lo",
  direction="l",
 },
 [0xDC2]={
  category="lo",
  direction="l",
 },
 [0xDC3]={
  category="lo",
  direction="l",
 },
 [0xDC4]={
  category="lo",
  direction="l",
 },
 [0xDC5]={
  category="lo",
  direction="l",
 },
 [0xDC6]={
  category="lo",
  direction="l",
 },
 [0xDCA]={
  category="mn",
  direction="nsm",
 },
 [0xDCF]={
  category="mc",
  direction="l",
 },
 [0xDD0]={
  category="mc",
  direction="l",
 },
 [0xDD1]={
  category="mc",
  direction="l",
 },
 [0xDD2]={
  category="mn",
  direction="nsm",
 },
 [0xDD3]={
  category="mn",
  direction="nsm",
 },
 [0xDD4]={
  category="mn",
  direction="nsm",
 },
 [0xDD6]={
  category="mn",
  direction="nsm",
 },
 [0xDD8]={
  category="mc",
  direction="l",
 },
 [0xDD9]={
  category="mc",
  direction="l",
 },
 [0xDDA]={
  category="mc",
  direction="l",
 },
 [0xDDB]={
  category="mc",
  direction="l",
 },
 [0xDDC]={
  category="mc",
  direction="l",
 },
 [0xDDD]={
  category="mc",
  direction="l",
 },
 [0xDDE]={
  category="mc",
  direction="l",
 },
 [0xDDF]={
  category="mc",
  direction="l",
 },
 [0xDE6]={
  category="nd",
  direction="l",
 },
 [0xDE7]={
  category="nd",
  direction="l",
 },
 [0xDE8]={
  category="nd",
  direction="l",
 },
 [0xDE9]={
  category="nd",
  direction="l",
 },
 [0xDEA]={
  category="nd",
  direction="l",
 },
 [0xDEB]={
  category="nd",
  direction="l",
 },
 [0xDEC]={
  category="nd",
  direction="l",
 },
 [0xDED]={
  category="nd",
  direction="l",
 },
 [0xDEE]={
  category="nd",
  direction="l",
 },
 [0xDEF]={
  category="nd",
  direction="l",
 },
 [0xDF2]={
  category="mc",
  direction="l",
 },
 [0xDF3]={
  category="mc",
  direction="l",
 },
 [0xDF4]={
  category="po",
  direction="l",
 },
 [0xE01]={
  category="lo",
  direction="l",
 },
 [0xE02]={
  category="lo",
  direction="l",
 },
 [0xE03]={
  category="lo",
  direction="l",
 },
 [0xE04]={
  category="lo",
  direction="l",
 },
 [0xE05]={
  category="lo",
  direction="l",
 },
 [0xE06]={
  category="lo",
  direction="l",
 },
 [0xE07]={
  category="lo",
  direction="l",
 },
 [0xE08]={
  category="lo",
  direction="l",
 },
 [0xE09]={
  category="lo",
  direction="l",
 },
 [0xE0A]={
  category="lo",
  direction="l",
 },
 [0xE0B]={
  category="lo",
  direction="l",
 },
 [0xE0C]={
  category="lo",
  direction="l",
 },
 [0xE0D]={
  category="lo",
  direction="l",
 },
 [0xE0E]={
  category="lo",
  direction="l",
 },
 [0xE0F]={
  category="lo",
  direction="l",
 },
 [0xE10]={
  category="lo",
  direction="l",
 },
 [0xE11]={
  category="lo",
  direction="l",
 },
 [0xE12]={
  category="lo",
  direction="l",
 },
 [0xE13]={
  category="lo",
  direction="l",
 },
 [0xE14]={
  category="lo",
  direction="l",
 },
 [0xE15]={
  category="lo",
  direction="l",
 },
 [0xE16]={
  category="lo",
  direction="l",
 },
 [0xE17]={
  category="lo",
  direction="l",
 },
 [0xE18]={
  category="lo",
  direction="l",
 },
 [0xE19]={
  category="lo",
  direction="l",
 },
 [0xE1A]={
  category="lo",
  direction="l",
 },
 [0xE1B]={
  category="lo",
  direction="l",
 },
 [0xE1C]={
  category="lo",
  direction="l",
 },
 [0xE1D]={
  category="lo",
  direction="l",
 },
 [0xE1E]={
  category="lo",
  direction="l",
 },
 [0xE1F]={
  category="lo",
  direction="l",
 },
 [0xE20]={
  category="lo",
  direction="l",
 },
 [0xE21]={
  category="lo",
  direction="l",
 },
 [0xE22]={
  category="lo",
  direction="l",
 },
 [0xE23]={
  category="lo",
  direction="l",
 },
 [0xE24]={
  category="lo",
  direction="l",
 },
 [0xE25]={
  category="lo",
  direction="l",
 },
 [0xE26]={
  category="lo",
  direction="l",
 },
 [0xE27]={
  category="lo",
  direction="l",
 },
 [0xE28]={
  category="lo",
  direction="l",
 },
 [0xE29]={
  category="lo",
  direction="l",
 },
 [0xE2A]={
  category="lo",
  direction="l",
 },
 [0xE2B]={
  category="lo",
  direction="l",
 },
 [0xE2C]={
  category="lo",
  direction="l",
 },
 [0xE2D]={
  category="lo",
  direction="l",
 },
 [0xE2E]={
  category="lo",
  direction="l",
 },
 [0xE2F]={
  category="lo",
  direction="l",
 },
 [0xE30]={
  category="lo",
  direction="l",
 },
 [0xE31]={
  category="mn",
  direction="nsm",
 },
 [0xE32]={
  category="lo",
  direction="l",
 },
 [0xE33]={
  category="lo",
  direction="l",
 },
 [0xE34]={
  category="mn",
  direction="nsm",
 },
 [0xE35]={
  category="mn",
  direction="nsm",
 },
 [0xE36]={
  category="mn",
  direction="nsm",
 },
 [0xE37]={
  category="mn",
  direction="nsm",
 },
 [0xE38]={
  category="mn",
  direction="nsm",
 },
 [0xE39]={
  category="mn",
  direction="nsm",
 },
 [0xE3A]={
  category="mn",
  direction="nsm",
 },
 [0xE3F]={
  category="sc",
  direction="et",
 },
 [0xE40]={
  category="lo",
  direction="l",
 },
 [0xE41]={
  category="lo",
  direction="l",
 },
 [0xE42]={
  category="lo",
  direction="l",
 },
 [0xE43]={
  category="lo",
  direction="l",
 },
 [0xE44]={
  category="lo",
  direction="l",
 },
 [0xE45]={
  category="lo",
  direction="l",
 },
 [0xE46]={
  category="lm",
  direction="l",
 },
 [0xE47]={
  category="mn",
  direction="nsm",
 },
 [0xE48]={
  category="mn",
  direction="nsm",
 },
 [0xE49]={
  category="mn",
  direction="nsm",
 },
 [0xE4A]={
  category="mn",
  direction="nsm",
 },
 [0xE4B]={
  category="mn",
  direction="nsm",
 },
 [0xE4C]={
  category="mn",
  direction="nsm",
 },
 [0xE4D]={
  category="mn",
  direction="nsm",
 },
 [0xE4E]={
  category="mn",
  direction="nsm",
 },
 [0xE4F]={
  category="po",
  direction="l",
 },
 [0xE50]={
  category="nd",
  direction="l",
 },
 [0xE51]={
  category="nd",
  direction="l",
 },
 [0xE52]={
  category="nd",
  direction="l",
 },
 [0xE53]={
  category="nd",
  direction="l",
 },
 [0xE54]={
  category="nd",
  direction="l",
 },
 [0xE55]={
  category="nd",
  direction="l",
 },
 [0xE56]={
  category="nd",
  direction="l",
 },
 [0xE57]={
  category="nd",
  direction="l",
 },
 [0xE58]={
  category="nd",
  direction="l",
 },
 [0xE59]={
  category="nd",
  direction="l",
 },
 [0xE5A]={
  category="po",
  direction="l",
 },
 [0xE5B]={
  category="po",
  direction="l",
 },
 [0xE81]={
  category="lo",
  direction="l",
 },
 [0xE82]={
  category="lo",
  direction="l",
 },
 [0xE84]={
  category="lo",
  direction="l",
 },
 [0xE87]={
  category="lo",
  direction="l",
 },
 [0xE88]={
  category="lo",
  direction="l",
 },
 [0xE8A]={
  category="lo",
  direction="l",
 },
 [0xE8D]={
  category="lo",
  direction="l",
 },
 [0xE94]={
  category="lo",
  direction="l",
 },
 [0xE95]={
  category="lo",
  direction="l",
 },
 [0xE96]={
  category="lo",
  direction="l",
 },
 [0xE97]={
  category="lo",
  direction="l",
 },
 [0xE99]={
  category="lo",
  direction="l",
 },
 [0xE9A]={
  category="lo",
  direction="l",
 },
 [0xE9B]={
  category="lo",
  direction="l",
 },
 [0xE9C]={
  category="lo",
  direction="l",
 },
 [0xE9D]={
  category="lo",
  direction="l",
 },
 [0xE9E]={
  category="lo",
  direction="l",
 },
 [0xE9F]={
  category="lo",
  direction="l",
 },
 [0xEA1]={
  category="lo",
  direction="l",
 },
 [0xEA2]={
  category="lo",
  direction="l",
 },
 [0xEA3]={
  category="lo",
  direction="l",
 },
 [0xEA5]={
  category="lo",
  direction="l",
 },
 [0xEA7]={
  category="lo",
  direction="l",
 },
 [0xEAA]={
  category="lo",
  direction="l",
 },
 [0xEAB]={
  category="lo",
  direction="l",
 },
 [0xEAD]={
  category="lo",
  direction="l",
 },
 [0xEAE]={
  category="lo",
  direction="l",
 },
 [0xEAF]={
  category="lo",
  direction="l",
 },
 [0xEB0]={
  category="lo",
  direction="l",
 },
 [0xEB1]={
  category="mn",
  direction="nsm",
 },
 [0xEB2]={
  category="lo",
  direction="l",
 },
 [0xEB3]={
  category="lo",
  direction="l",
 },
 [0xEB4]={
  category="mn",
  direction="nsm",
 },
 [0xEB5]={
  category="mn",
  direction="nsm",
 },
 [0xEB6]={
  category="mn",
  direction="nsm",
 },
 [0xEB7]={
  category="mn",
  direction="nsm",
 },
 [0xEB8]={
  category="mn",
  direction="nsm",
 },
 [0xEB9]={
  category="mn",
  direction="nsm",
 },
 [0xEBB]={
  category="mn",
  direction="nsm",
 },
 [0xEBC]={
  category="mn",
  direction="nsm",
 },
 [0xEBD]={
  category="lo",
  direction="l",
 },
 [0xEC0]={
  category="lo",
  direction="l",
 },
 [0xEC1]={
  category="lo",
  direction="l",
 },
 [0xEC2]={
  category="lo",
  direction="l",
 },
 [0xEC3]={
  category="lo",
  direction="l",
 },
 [0xEC4]={
  category="lo",
  direction="l",
 },
 [0xEC6]={
  category="lm",
  direction="l",
 },
 [0xEC8]={
  category="mn",
  direction="nsm",
 },
 [0xEC9]={
  category="mn",
  direction="nsm",
 },
 [0xECA]={
  category="mn",
  direction="nsm",
 },
 [0xECB]={
  category="mn",
  direction="nsm",
 },
 [0xECC]={
  category="mn",
  direction="nsm",
 },
 [0xECD]={
  category="mn",
  direction="nsm",
 },
 [0xED0]={
  category="nd",
  direction="l",
 },
 [0xED1]={
  category="nd",
  direction="l",
 },
 [0xED2]={
  category="nd",
  direction="l",
 },
 [0xED3]={
  category="nd",
  direction="l",
 },
 [0xED4]={
  category="nd",
  direction="l",
 },
 [0xED5]={
  category="nd",
  direction="l",
 },
 [0xED6]={
  category="nd",
  direction="l",
 },
 [0xED7]={
  category="nd",
  direction="l",
 },
 [0xED8]={
  category="nd",
  direction="l",
 },
 [0xED9]={
  category="nd",
  direction="l",
 },
 [0xEDC]={
  category="lo",
  direction="l",
 },
 [0xEDD]={
  category="lo",
  direction="l",
 },
 [0xEDE]={
  category="lo",
  direction="l",
 },
 [0xEDF]={
  category="lo",
  direction="l",
 },
 [0xF00]={
  category="lo",
  direction="l",
 },
 [0xF01]={
  category="so",
  direction="l",
 },
 [0xF02]={
  category="so",
  direction="l",
 },
 [0xF03]={
  category="so",
  direction="l",
 },
 [0xF04]={
  category="po",
  direction="l",
 },
 [0xF05]={
  category="po",
  direction="l",
 },
 [0xF06]={
  category="po",
  direction="l",
 },
 [0xF07]={
  category="po",
  direction="l",
 },
 [0xF08]={
  category="po",
  direction="l",
 },
 [0xF09]={
  category="po",
  direction="l",
 },
 [0xF0A]={
  category="po",
  direction="l",
 },
 [0xF0B]={
  category="po",
  direction="l",
 },
 [0xF0C]={
  category="po",
  direction="l",
 },
 [0xF0D]={
  category="po",
  direction="l",
 },
 [0xF0E]={
  category="po",
  direction="l",
 },
 [0xF0F]={
  category="po",
  direction="l",
 },
 [0xF10]={
  category="po",
  direction="l",
 },
 [0xF11]={
  category="po",
  direction="l",
 },
 [0xF12]={
  category="po",
  direction="l",
 },
 [0xF13]={
  category="so",
  direction="l",
 },
 [0xF14]={
  category="so",
  direction="l",
 },
 [0xF15]={
  category="so",
  direction="l",
 },
 [0xF16]={
  category="so",
  direction="l",
 },
 [0xF17]={
  category="so",
  direction="l",
 },
 [0xF18]={
  category="mn",
  direction="nsm",
 },
 [0xF19]={
  category="mn",
  direction="nsm",
 },
 [0xF1A]={
  category="so",
  direction="l",
 },
 [0xF1B]={
  category="so",
  direction="l",
 },
 [0xF1C]={
  category="so",
  direction="l",
 },
 [0xF1D]={
  category="so",
  direction="l",
 },
 [0xF1E]={
  category="so",
  direction="l",
 },
 [0xF1F]={
  category="so",
  direction="l",
 },
 [0xF20]={
  category="nd",
  direction="l",
 },
 [0xF21]={
  category="nd",
  direction="l",
 },
 [0xF22]={
  category="nd",
  direction="l",
 },
 [0xF23]={
  category="nd",
  direction="l",
 },
 [0xF24]={
  category="nd",
  direction="l",
 },
 [0xF25]={
  category="nd",
  direction="l",
 },
 [0xF26]={
  category="nd",
  direction="l",
 },
 [0xF27]={
  category="nd",
  direction="l",
 },
 [0xF28]={
  category="nd",
  direction="l",
 },
 [0xF29]={
  category="nd",
  direction="l",
 },
 [0xF2A]={
  category="no",
  direction="l",
 },
 [0xF2B]={
  category="no",
  direction="l",
 },
 [0xF2C]={
  category="no",
  direction="l",
 },
 [0xF2D]={
  category="no",
  direction="l",
 },
 [0xF2E]={
  category="no",
  direction="l",
 },
 [0xF2F]={
  category="no",
  direction="l",
 },
 [0xF30]={
  category="no",
  direction="l",
 },
 [0xF31]={
  category="no",
  direction="l",
 },
 [0xF32]={
  category="no",
  direction="l",
 },
 [0xF33]={
  category="no",
  direction="l",
 },
 [0xF34]={
  category="so",
  direction="l",
 },
 [0xF35]={
  category="mn",
  direction="nsm",
 },
 [0xF36]={
  category="so",
  direction="l",
 },
 [0xF37]={
  category="mn",
  direction="nsm",
 },
 [0xF38]={
  category="so",
  direction="l",
 },
 [0xF39]={
  category="mn",
  direction="nsm",
 },
 [0xF3A]={
  category="ps",
  direction="on",
  mirror=0xF3B,
 },
 [0xF3B]={
  category="pe",
  direction="on",
  mirror=0xF3A,
 },
 [0xF3C]={
  category="ps",
  direction="on",
  mirror=0xF3D,
 },
 [0xF3D]={
  category="pe",
  direction="on",
  mirror=0xF3C,
 },
 [0xF3E]={
  category="mc",
  direction="l",
 },
 [0xF3F]={
  category="mc",
  direction="l",
 },
 [0xF40]={
  category="lo",
  direction="l",
 },
 [0xF41]={
  category="lo",
  direction="l",
 },
 [0xF42]={
  category="lo",
  direction="l",
 },
 [0xF43]={
  category="lo",
  direction="l",
 },
 [0xF44]={
  category="lo",
  direction="l",
 },
 [0xF45]={
  category="lo",
  direction="l",
 },
 [0xF46]={
  category="lo",
  direction="l",
 },
 [0xF47]={
  category="lo",
  direction="l",
 },
 [0xF49]={
  category="lo",
  direction="l",
 },
 [0xF4A]={
  category="lo",
  direction="l",
 },
 [0xF4B]={
  category="lo",
  direction="l",
 },
 [0xF4C]={
  category="lo",
  direction="l",
 },
 [0xF4D]={
  category="lo",
  direction="l",
 },
 [0xF4E]={
  category="lo",
  direction="l",
 },
 [0xF4F]={
  category="lo",
  direction="l",
 },
 [0xF50]={
  category="lo",
  direction="l",
 },
 [0xF51]={
  category="lo",
  direction="l",
 },
 [0xF52]={
  category="lo",
  direction="l",
 },
 [0xF53]={
  category="lo",
  direction="l",
 },
 [0xF54]={
  category="lo",
  direction="l",
 },
 [0xF55]={
  category="lo",
  direction="l",
 },
 [0xF56]={
  category="lo",
  direction="l",
 },
 [0xF57]={
  category="lo",
  direction="l",
 },
 [0xF58]={
  category="lo",
  direction="l",
 },
 [0xF59]={
  category="lo",
  direction="l",
 },
 [0xF5A]={
  category="lo",
  direction="l",
 },
 [0xF5B]={
  category="lo",
  direction="l",
 },
 [0xF5C]={
  category="lo",
  direction="l",
 },
 [0xF5D]={
  category="lo",
  direction="l",
 },
 [0xF5E]={
  category="lo",
  direction="l",
 },
 [0xF5F]={
  category="lo",
  direction="l",
 },
 [0xF60]={
  category="lo",
  direction="l",
 },
 [0xF61]={
  category="lo",
  direction="l",
 },
 [0xF62]={
  category="lo",
  direction="l",
 },
 [0xF63]={
  category="lo",
  direction="l",
 },
 [0xF64]={
  category="lo",
  direction="l",
 },
 [0xF65]={
  category="lo",
  direction="l",
 },
 [0xF66]={
  category="lo",
  direction="l",
 },
 [0xF67]={
  category="lo",
  direction="l",
 },
 [0xF68]={
  category="lo",
  direction="l",
 },
 [0xF69]={
  category="lo",
  direction="l",
 },
 [0xF6A]={
  category="lo",
  direction="l",
 },
 [0xF6B]={
  category="lo",
  direction="l",
 },
 [0xF6C]={
  category="lo",
  direction="l",
 },
 [0xF71]={
  category="mn",
  direction="nsm",
 },
 [0xF72]={
  category="mn",
  direction="nsm",
 },
 [0xF73]={
  category="mn",
  direction="nsm",
 },
 [0xF74]={
  category="mn",
  direction="nsm",
 },
 [0xF75]={
  category="mn",
  direction="nsm",
 },
 [0xF76]={
  category="mn",
  direction="nsm",
 },
 [0xF77]={
  category="mn",
  direction="nsm",
 },
 [0xF78]={
  category="mn",
  direction="nsm",
 },
 [0xF79]={
  category="mn",
  direction="nsm",
 },
 [0xF7A]={
  category="mn",
  direction="nsm",
 },
 [0xF7B]={
  category="mn",
  direction="nsm",
 },
 [0xF7C]={
  category="mn",
  direction="nsm",
 },
 [0xF7D]={
  category="mn",
  direction="nsm",
 },
 [0xF7E]={
  category="mn",
  direction="nsm",
 },
 [0xF7F]={
  category="mc",
  direction="l",
 },
 [0xF80]={
  category="mn",
  direction="nsm",
 },
 [0xF81]={
  category="mn",
  direction="nsm",
 },
 [0xF82]={
  category="mn",
  direction="nsm",
 },
 [0xF83]={
  category="mn",
  direction="nsm",
 },
 [0xF84]={
  category="mn",
  direction="nsm",
 },
 [0xF85]={
  category="po",
  direction="l",
 },
 [0xF86]={
  category="mn",
  direction="nsm",
 },
 [0xF87]={
  category="mn",
  direction="nsm",
 },
 [0xF88]={
  category="lo",
  direction="l",
 },
 [0xF89]={
  category="lo",
  direction="l",
 },
 [0xF8A]={
  category="lo",
  direction="l",
 },
 [0xF8B]={
  category="lo",
  direction="l",
 },
 [0xF8C]={
  category="lo",
  direction="l",
 },
 [0xF8D]={
  category="mn",
  direction="nsm",
 },
 [0xF8E]={
  category="mn",
  direction="nsm",
 },
 [0xF8F]={
  category="mn",
  direction="nsm",
 },
 [0xF90]={
  category="mn",
  direction="nsm",
 },
 [0xF91]={
  category="mn",
  direction="nsm",
 },
 [0xF92]={
  category="mn",
  direction="nsm",
 },
 [0xF93]={
  category="mn",
  direction="nsm",
 },
 [0xF94]={
  category="mn",
  direction="nsm",
 },
 [0xF95]={
  category="mn",
  direction="nsm",
 },
 [0xF96]={
  category="mn",
  direction="nsm",
 },
 [0xF97]={
  category="mn",
  direction="nsm",
 },
 [0xF99]={
  category="mn",
  direction="nsm",
 },
 [0xF9A]={
  category="mn",
  direction="nsm",
 },
 [0xF9B]={
  category="mn",
  direction="nsm",
 },
 [0xF9C]={
  category="mn",
  direction="nsm",
 },
 [0xF9D]={
  category="mn",
  direction="nsm",
 },
 [0xF9E]={
  category="mn",
  direction="nsm",
 },
 [0xF9F]={
  category="mn",
  direction="nsm",
 },
 [0xFA0]={
  category="mn",
  direction="nsm",
 },
 [0xFA1]={
  category="mn",
  direction="nsm",
 },
 [0xFA2]={
  category="mn",
  direction="nsm",
 },
 [0xFA3]={
  category="mn",
  direction="nsm",
 },
 [0xFA4]={
  category="mn",
  direction="nsm",
 },
 [0xFA5]={
  category="mn",
  direction="nsm",
 },
 [0xFA6]={
  category="mn",
  direction="nsm",
 },
 [0xFA7]={
  category="mn",
  direction="nsm",
 },
 [0xFA8]={
  category="mn",
  direction="nsm",
 },
 [0xFA9]={
  category="mn",
  direction="nsm",
 },
 [0xFAA]={
  category="mn",
  direction="nsm",
 },
 [0xFAB]={
  category="mn",
  direction="nsm",
 },
 [0xFAC]={
  category="mn",
  direction="nsm",
 },
 [0xFAD]={
  category="mn",
  direction="nsm",
 },
 [0xFAE]={
  category="mn",
  direction="nsm",
 },
 [0xFAF]={
  category="mn",
  direction="nsm",
 },
 [0xFB0]={
  category="mn",
  direction="nsm",
 },
 [0xFB1]={
  category="mn",
  direction="nsm",
 },
 [0xFB2]={
  category="mn",
  direction="nsm",
 },
 [0xFB3]={
  category="mn",
  direction="nsm",
 },
 [0xFB4]={
  category="mn",
  direction="nsm",
 },
 [0xFB5]={
  category="mn",
  direction="nsm",
 },
 [0xFB6]={
  category="mn",
  direction="nsm",
 },
 [0xFB7]={
  category="mn",
  direction="nsm",
 },
 [0xFB8]={
  category="mn",
  direction="nsm",
 },
 [0xFB9]={
  category="mn",
  direction="nsm",
 },
 [0xFBA]={
  category="mn",
  direction="nsm",
 },
 [0xFBB]={
  category="mn",
  direction="nsm",
 },
 [0xFBC]={
  category="mn",
  direction="nsm",
 },
 [0xFBE]={
  category="so",
  direction="l",
 },
 [0xFBF]={
  category="so",
  direction="l",
 },
 [0xFC0]={
  category="so",
  direction="l",
 },
 [0xFC1]={
  category="so",
  direction="l",
 },
 [0xFC2]={
  category="so",
  direction="l",
 },
 [0xFC3]={
  category="so",
  direction="l",
 },
 [0xFC4]={
  category="so",
  direction="l",
 },
 [0xFC5]={
  category="so",
  direction="l",
 },
 [0xFC6]={
  category="mn",
  direction="nsm",
 },
 [0xFC7]={
  category="so",
  direction="l",
 },
 [0xFC8]={
  category="so",
  direction="l",
 },
 [0xFC9]={
  category="so",
  direction="l",
 },
 [0xFCA]={
  category="so",
  direction="l",
 },
 [0xFCB]={
  category="so",
  direction="l",
 },
 [0xFCC]={
  category="so",
  direction="l",
 },
 [0xFCE]={
  category="so",
  direction="l",
 },
 [0xFCF]={
  category="so",
  direction="l",
 },
 [0xFD0]={
  category="po",
  direction="l",
 },
 [0xFD1]={
  category="po",
  direction="l",
 },
 [0xFD2]={
  category="po",
  direction="l",
 },
 [0xFD3]={
  category="po",
  direction="l",
 },
 [0xFD4]={
  category="po",
  direction="l",
 },
 [0xFD5]={
  category="so",
  direction="l",
 },
 [0xFD6]={
  category="so",
  direction="l",
 },
 [0xFD7]={
  category="so",
  direction="l",
 },
 [0xFD8]={
  category="so",
  direction="l",
 },
 [0xFD9]={
  category="po",
  direction="l",
 },
 [0xFDA]={
  category="po",
  direction="l",
 },
 [0x1000]={
  category="lo",
  direction="l",
 },
 [0x1001]={
  category="lo",
  direction="l",
 },
 [0x1002]={
  category="lo",
  direction="l",
 },
 [0x1003]={
  category="lo",
  direction="l",
 },
 [0x1004]={
  category="lo",
  direction="l",
 },
 [0x1005]={
  category="lo",
  direction="l",
 },
 [0x1006]={
  category="lo",
  direction="l",
 },
 [0x1007]={
  category="lo",
  direction="l",
 },
 [0x1008]={
  category="lo",
  direction="l",
 },
 [0x1009]={
  category="lo",
  direction="l",
 },
 [0x100A]={
  category="lo",
  direction="l",
 },
 [0x100B]={
  category="lo",
  direction="l",
 },
 [0x100C]={
  category="lo",
  direction="l",
 },
 [0x100D]={
  category="lo",
  direction="l",
 },
 [0x100E]={
  category="lo",
  direction="l",
 },
 [0x100F]={
  category="lo",
  direction="l",
 },
 [0x1010]={
  category="lo",
  direction="l",
 },
 [0x1011]={
  category="lo",
  direction="l",
 },
 [0x1012]={
  category="lo",
  direction="l",
 },
 [0x1013]={
  category="lo",
  direction="l",
 },
 [0x1014]={
  category="lo",
  direction="l",
 },
 [0x1015]={
  category="lo",
  direction="l",
 },
 [0x1016]={
  category="lo",
  direction="l",
 },
 [0x1017]={
  category="lo",
  direction="l",
 },
 [0x1018]={
  category="lo",
  direction="l",
 },
 [0x1019]={
  category="lo",
  direction="l",
 },
 [0x101A]={
  category="lo",
  direction="l",
 },
 [0x101B]={
  category="lo",
  direction="l",
 },
 [0x101C]={
  category="lo",
  direction="l",
 },
 [0x101D]={
  category="lo",
  direction="l",
 },
 [0x101E]={
  category="lo",
  direction="l",
 },
 [0x101F]={
  category="lo",
  direction="l",
 },
 [0x1020]={
  category="lo",
  direction="l",
 },
 [0x1021]={
  category="lo",
  direction="l",
 },
 [0x1022]={
  category="lo",
  direction="l",
 },
 [0x1023]={
  category="lo",
  direction="l",
 },
 [0x1024]={
  category="lo",
  direction="l",
 },
 [0x1025]={
  category="lo",
  direction="l",
 },
 [0x1026]={
  category="lo",
  direction="l",
 },
 [0x1027]={
  category="lo",
  direction="l",
 },
 [0x1028]={
  category="lo",
  direction="l",
 },
 [0x1029]={
  category="lo",
  direction="l",
 },
 [0x102A]={
  category="lo",
  direction="l",
 },
 [0x102B]={
  category="mc",
  direction="l",
 },
 [0x102C]={
  category="mc",
  direction="l",
 },
 [0x102D]={
  category="mn",
  direction="nsm",
 },
 [0x102E]={
  category="mn",
  direction="nsm",
 },
 [0x102F]={
  category="mn",
  direction="nsm",
 },
 [0x1030]={
  category="mn",
  direction="nsm",
 },
 [0x1031]={
  category="mc",
  direction="l",
 },
 [0x1032]={
  category="mn",
  direction="nsm",
 },
 [0x1033]={
  category="mn",
  direction="nsm",
 },
 [0x1034]={
  category="mn",
  direction="nsm",
 },
 [0x1035]={
  category="mn",
  direction="nsm",
 },
 [0x1036]={
  category="mn",
  direction="nsm",
 },
 [0x1037]={
  category="mn",
  direction="nsm",
 },
 [0x1038]={
  category="mc",
  direction="l",
 },
 [0x1039]={
  category="mn",
  direction="nsm",
 },
 [0x103A]={
  category="mn",
  direction="nsm",
 },
 [0x103B]={
  category="mc",
  direction="l",
 },
 [0x103C]={
  category="mc",
  direction="l",
 },
 [0x103D]={
  category="mn",
  direction="nsm",
 },
 [0x103E]={
  category="mn",
  direction="nsm",
 },
 [0x103F]={
  category="lo",
  direction="l",
 },
 [0x1040]={
  category="nd",
  direction="l",
 },
 [0x1041]={
  category="nd",
  direction="l",
 },
 [0x1042]={
  category="nd",
  direction="l",
 },
 [0x1043]={
  category="nd",
  direction="l",
 },
 [0x1044]={
  category="nd",
  direction="l",
 },
 [0x1045]={
  category="nd",
  direction="l",
 },
 [0x1046]={
  category="nd",
  direction="l",
 },
 [0x1047]={
  category="nd",
  direction="l",
 },
 [0x1048]={
  category="nd",
  direction="l",
 },
 [0x1049]={
  category="nd",
  direction="l",
 },
 [0x104A]={
  category="po",
  direction="l",
 },
 [0x104B]={
  category="po",
  direction="l",
 },
 [0x104C]={
  category="po",
  direction="l",
 },
 [0x104D]={
  category="po",
  direction="l",
 },
 [0x104E]={
  category="po",
  direction="l",
 },
 [0x104F]={
  category="po",
  direction="l",
 },
 [0x1050]={
  category="lo",
  direction="l",
 },
 [0x1051]={
  category="lo",
  direction="l",
 },
 [0x1052]={
  category="lo",
  direction="l",
 },
 [0x1053]={
  category="lo",
  direction="l",
 },
 [0x1054]={
  category="lo",
  direction="l",
 },
 [0x1055]={
  category="lo",
  direction="l",
 },
 [0x1056]={
  category="mc",
  direction="l",
 },
 [0x1057]={
  category="mc",
  direction="l",
 },
 [0x1058]={
  category="mn",
  direction="nsm",
 },
 [0x1059]={
  category="mn",
  direction="nsm",
 },
 [0x105A]={
  category="lo",
  direction="l",
 },
 [0x105B]={
  category="lo",
  direction="l",
 },
 [0x105C]={
  category="lo",
  direction="l",
 },
 [0x105D]={
  category="lo",
  direction="l",
 },
 [0x105E]={
  category="mn",
  direction="nsm",
 },
 [0x105F]={
  category="mn",
  direction="nsm",
 },
 [0x1060]={
  category="mn",
  direction="nsm",
 },
 [0x1061]={
  category="lo",
  direction="l",
 },
 [0x1062]={
  category="mc",
  direction="l",
 },
 [0x1063]={
  category="mc",
  direction="l",
 },
 [0x1064]={
  category="mc",
  direction="l",
 },
 [0x1065]={
  category="lo",
  direction="l",
 },
 [0x1066]={
  category="lo",
  direction="l",
 },
 [0x1067]={
  category="mc",
  direction="l",
 },
 [0x1068]={
  category="mc",
  direction="l",
 },
 [0x1069]={
  category="mc",
  direction="l",
 },
 [0x106A]={
  category="mc",
  direction="l",
 },
 [0x106B]={
  category="mc",
  direction="l",
 },
 [0x106C]={
  category="mc",
  direction="l",
 },
 [0x106D]={
  category="mc",
  direction="l",
 },
 [0x106E]={
  category="lo",
  direction="l",
 },
 [0x106F]={
  category="lo",
  direction="l",
 },
 [0x1070]={
  category="lo",
  direction="l",
 },
 [0x1071]={
  category="mn",
  direction="nsm",
 },
 [0x1072]={
  category="mn",
  direction="nsm",
 },
 [0x1073]={
  category="mn",
  direction="nsm",
 },
 [0x1074]={
  category="mn",
  direction="nsm",
 },
 [0x1075]={
  category="lo",
  direction="l",
 },
 [0x1076]={
  category="lo",
  direction="l",
 },
 [0x1077]={
  category="lo",
  direction="l",
 },
 [0x1078]={
  category="lo",
  direction="l",
 },
 [0x1079]={
  category="lo",
  direction="l",
 },
 [0x107A]={
  category="lo",
  direction="l",
 },
 [0x107B]={
  category="lo",
  direction="l",
 },
 [0x107C]={
  category="lo",
  direction="l",
 },
 [0x107D]={
  category="lo",
  direction="l",
 },
 [0x107E]={
  category="lo",
  direction="l",
 },
 [0x107F]={
  category="lo",
  direction="l",
 },
 [0x1080]={
  category="lo",
  direction="l",
 },
 [0x1081]={
  category="lo",
  direction="l",
 },
 [0x1082]={
  category="mn",
  direction="nsm",
 },
 [0x1083]={
  category="mc",
  direction="l",
 },
 [0x1084]={
  category="mc",
  direction="l",
 },
 [0x1085]={
  category="mn",
  direction="nsm",
 },
 [0x1086]={
  category="mn",
  direction="nsm",
 },
 [0x1087]={
  category="mc",
  direction="l",
 },
 [0x1088]={
  category="mc",
  direction="l",
 },
 [0x1089]={
  category="mc",
  direction="l",
 },
 [0x108A]={
  category="mc",
  direction="l",
 },
 [0x108B]={
  category="mc",
  direction="l",
 },
 [0x108C]={
  category="mc",
  direction="l",
 },
 [0x108D]={
  category="mn",
  direction="nsm",
 },
 [0x108E]={
  category="lo",
  direction="l",
 },
 [0x108F]={
  category="mc",
  direction="l",
 },
 [0x1090]={
  category="nd",
  direction="l",
 },
 [0x1091]={
  category="nd",
  direction="l",
 },
 [0x1092]={
  category="nd",
  direction="l",
 },
 [0x1093]={
  category="nd",
  direction="l",
 },
 [0x1094]={
  category="nd",
  direction="l",
 },
 [0x1095]={
  category="nd",
  direction="l",
 },
 [0x1096]={
  category="nd",
  direction="l",
 },
 [0x1097]={
  category="nd",
  direction="l",
 },
 [0x1098]={
  category="nd",
  direction="l",
 },
 [0x1099]={
  category="nd",
  direction="l",
 },
 [0x109A]={
  category="mc",
  direction="l",
 },
 [0x109B]={
  category="mc",
  direction="l",
 },
 [0x109C]={
  category="mc",
  direction="l",
 },
 [0x109D]={
  category="mn",
  direction="nsm",
 },
 [0x109E]={
  category="so",
  direction="l",
 },
 [0x109F]={
  category="so",
  direction="l",
 },
 [0x10A0]={
  category="lu",
  direction="l",
 },
 [0x10A1]={
  category="lu",
  direction="l",
 },
 [0x10A2]={
  category="lu",
  direction="l",
 },
 [0x10A3]={
  category="lu",
  direction="l",
 },
 [0x10A4]={
  category="lu",
  direction="l",
 },
 [0x10A5]={
  category="lu",
  direction="l",
 },
 [0x10A6]={
  category="lu",
  direction="l",
 },
 [0x10A7]={
  category="lu",
  direction="l",
 },
 [0x10A8]={
  category="lu",
  direction="l",
 },
 [0x10A9]={
  category="lu",
  direction="l",
 },
 [0x10AA]={
  category="lu",
  direction="l",
 },
 [0x10AB]={
  category="lu",
  direction="l",
 },
 [0x10AC]={
  category="lu",
  direction="l",
 },
 [0x10AD]={
  category="lu",
  direction="l",
 },
 [0x10AE]={
  category="lu",
  direction="l",
 },
 [0x10AF]={
  category="lu",
  direction="l",
 },
 [0x10B0]={
  category="lu",
  direction="l",
 },
 [0x10B1]={
  category="lu",
  direction="l",
 },
 [0x10B2]={
  category="lu",
  direction="l",
 },
 [0x10B3]={
  category="lu",
  direction="l",
 },
 [0x10B4]={
  category="lu",
  direction="l",
 },
 [0x10B5]={
  category="lu",
  direction="l",
 },
 [0x10B6]={
  category="lu",
  direction="l",
 },
 [0x10B7]={
  category="lu",
  direction="l",
 },
 [0x10B8]={
  category="lu",
  direction="l",
 },
 [0x10B9]={
  category="lu",
  direction="l",
 },
 [0x10BA]={
  category="lu",
  direction="l",
 },
 [0x10BB]={
  category="lu",
  direction="l",
 },
 [0x10BC]={
  category="lu",
  direction="l",
 },
 [0x10BD]={
  category="lu",
  direction="l",
 },
 [0x10BE]={
  category="lu",
  direction="l",
 },
 [0x10BF]={
  category="lu",
  direction="l",
 },
 [0x10C0]={
  category="lu",
  direction="l",
 },
 [0x10C1]={
  category="lu",
  direction="l",
 },
 [0x10C2]={
  category="lu",
  direction="l",
 },
 [0x10C3]={
  category="lu",
  direction="l",
 },
 [0x10C4]={
  category="lu",
  direction="l",
 },
 [0x10C5]={
  category="lu",
  direction="l",
 },
 [0x10C7]={
  category="lu",
  direction="l",
 },
 [0x10CD]={
  category="lu",
  direction="l",
 },
 [0x10D0]={
  category="lo",
  direction="l",
 },
 [0x10D1]={
  category="lo",
  direction="l",
 },
 [0x10D2]={
  category="lo",
  direction="l",
 },
 [0x10D3]={
  category="lo",
  direction="l",
 },
 [0x10D4]={
  category="lo",
  direction="l",
 },
 [0x10D5]={
  category="lo",
  direction="l",
 },
 [0x10D6]={
  category="lo",
  direction="l",
 },
 [0x10D7]={
  category="lo",
  direction="l",
 },
 [0x10D8]={
  category="lo",
  direction="l",
 },
 [0x10D9]={
  category="lo",
  direction="l",
 },
 [0x10DA]={
  category="lo",
  direction="l",
 },
 [0x10DB]={
  category="lo",
  direction="l",
 },
 [0x10DC]={
  category="lo",
  direction="l",
 },
 [0x10DD]={
  category="lo",
  direction="l",
 },
 [0x10DE]={
  category="lo",
  direction="l",
 },
 [0x10DF]={
  category="lo",
  direction="l",
 },
 [0x10E0]={
  category="lo",
  direction="l",
 },
 [0x10E1]={
  category="lo",
  direction="l",
 },
 [0x10E2]={
  category="lo",
  direction="l",
 },
 [0x10E3]={
  category="lo",
  direction="l",
 },
 [0x10E4]={
  category="lo",
  direction="l",
 },
 [0x10E5]={
  category="lo",
  direction="l",
 },
 [0x10E6]={
  category="lo",
  direction="l",
 },
 [0x10E7]={
  category="lo",
  direction="l",
 },
 [0x10E8]={
  category="lo",
  direction="l",
 },
 [0x10E9]={
  category="lo",
  direction="l",
 },
 [0x10EA]={
  category="lo",
  direction="l",
 },
 [0x10EB]={
  category="lo",
  direction="l",
 },
 [0x10EC]={
  category="lo",
  direction="l",
 },
 [0x10ED]={
  category="lo",
  direction="l",
 },
 [0x10EE]={
  category="lo",
  direction="l",
 },
 [0x10EF]={
  category="lo",
  direction="l",
 },
 [0x10F0]={
  category="lo",
  direction="l",
 },
 [0x10F1]={
  category="lo",
  direction="l",
 },
 [0x10F2]={
  category="lo",
  direction="l",
 },
 [0x10F3]={
  category="lo",
  direction="l",
 },
 [0x10F4]={
  category="lo",
  direction="l",
 },
 [0x10F5]={
  category="lo",
  direction="l",
 },
 [0x10F6]={
  category="lo",
  direction="l",
 },
 [0x10F7]={
  category="lo",
  direction="l",
 },
 [0x10F8]={
  category="lo",
  direction="l",
 },
 [0x10F9]={
  category="lo",
  direction="l",
 },
 [0x10FA]={
  category="lo",
  direction="l",
 },
 [0x10FB]={
  category="po",
  direction="l",
 },
 [0x10FC]={
  category="lm",
  direction="l",
 },
 [0x10FD]={
  category="lo",
  direction="l",
 },
 [0x10FE]={
  category="lo",
  direction="l",
 },
 [0x10FF]={
  category="lo",
  direction="l",
 },
 [0x1100]={
  category="lo",
  direction="l",
 },
 [0x1101]={
  category="lo",
  direction="l",
 },
 [0x1102]={
  category="lo",
  direction="l",
 },
 [0x1103]={
  category="lo",
  direction="l",
 },
 [0x1104]={
  category="lo",
  direction="l",
 },
 [0x1105]={
  category="lo",
  direction="l",
 },
 [0x1106]={
  category="lo",
  direction="l",
 },
 [0x1107]={
  category="lo",
  direction="l",
 },
 [0x1108]={
  category="lo",
  direction="l",
 },
 [0x1109]={
  category="lo",
  direction="l",
 },
 [0x110A]={
  category="lo",
  direction="l",
 },
 [0x110B]={
  category="lo",
  direction="l",
 },
 [0x110C]={
  category="lo",
  direction="l",
 },
 [0x110D]={
  category="lo",
  direction="l",
 },
 [0x110E]={
  category="lo",
  direction="l",
 },
 [0x110F]={
  category="lo",
  direction="l",
 },
 [0x1110]={
  category="lo",
  direction="l",
 },
 [0x1111]={
  category="lo",
  direction="l",
 },
 [0x1112]={
  category="lo",
  direction="l",
 },
 [0x1113]={
  category="lo",
  direction="l",
 },
 [0x1114]={
  category="lo",
  direction="l",
 },
 [0x1115]={
  category="lo",
  direction="l",
 },
 [0x1116]={
  category="lo",
  direction="l",
 },
 [0x1117]={
  category="lo",
  direction="l",
 },
 [0x1118]={
  category="lo",
  direction="l",
 },
 [0x1119]={
  category="lo",
  direction="l",
 },
 [0x111A]={
  category="lo",
  direction="l",
 },
 [0x111B]={
  category="lo",
  direction="l",
 },
 [0x111C]={
  category="lo",
  direction="l",
 },
 [0x111D]={
  category="lo",
  direction="l",
 },
 [0x111E]={
  category="lo",
  direction="l",
 },
 [0x111F]={
  category="lo",
  direction="l",
 },
 [0x1120]={
  category="lo",
  direction="l",
 },
 [0x1121]={
  category="lo",
  direction="l",
 },
 [0x1122]={
  category="lo",
  direction="l",
 },
 [0x1123]={
  category="lo",
  direction="l",
 },
 [0x1124]={
  category="lo",
  direction="l",
 },
 [0x1125]={
  category="lo",
  direction="l",
 },
 [0x1126]={
  category="lo",
  direction="l",
 },
 [0x1127]={
  category="lo",
  direction="l",
 },
 [0x1128]={
  category="lo",
  direction="l",
 },
 [0x1129]={
  category="lo",
  direction="l",
 },
 [0x112A]={
  category="lo",
  direction="l",
 },
 [0x112B]={
  category="lo",
  direction="l",
 },
 [0x112C]={
  category="lo",
  direction="l",
 },
 [0x112D]={
  category="lo",
  direction="l",
 },
 [0x112E]={
  category="lo",
  direction="l",
 },
 [0x112F]={
  category="lo",
  direction="l",
 },
 [0x1130]={
  category="lo",
  direction="l",
 },
 [0x1131]={
  category="lo",
  direction="l",
 },
 [0x1132]={
  category="lo",
  direction="l",
 },
 [0x1133]={
  category="lo",
  direction="l",
 },
 [0x1134]={
  category="lo",
  direction="l",
 },
 [0x1135]={
  category="lo",
  direction="l",
 },
 [0x1136]={
  category="lo",
  direction="l",
 },
 [0x1137]={
  category="lo",
  direction="l",
 },
 [0x1138]={
  category="lo",
  direction="l",
 },
 [0x1139]={
  category="lo",
  direction="l",
 },
 [0x113A]={
  category="lo",
  direction="l",
 },
 [0x113B]={
  category="lo",
  direction="l",
 },
 [0x113C]={
  category="lo",
  direction="l",
 },
 [0x113D]={
  category="lo",
  direction="l",
 },
 [0x113E]={
  category="lo",
  direction="l",
 },
 [0x113F]={
  category="lo",
  direction="l",
 },
 [0x1140]={
  category="lo",
  direction="l",
 },
 [0x1141]={
  category="lo",
  direction="l",
 },
 [0x1142]={
  category="lo",
  direction="l",
 },
 [0x1143]={
  category="lo",
  direction="l",
 },
 [0x1144]={
  category="lo",
  direction="l",
 },
 [0x1145]={
  category="lo",
  direction="l",
 },
 [0x1146]={
  category="lo",
  direction="l",
 },
 [0x1147]={
  category="lo",
  direction="l",
 },
 [0x1148]={
  category="lo",
  direction="l",
 },
 [0x1149]={
  category="lo",
  direction="l",
 },
 [0x114A]={
  category="lo",
  direction="l",
 },
 [0x114B]={
  category="lo",
  direction="l",
 },
 [0x114C]={
  category="lo",
  direction="l",
 },
 [0x114D]={
  category="lo",
  direction="l",
 },
 [0x114E]={
  category="lo",
  direction="l",
 },
 [0x114F]={
  category="lo",
  direction="l",
 },
 [0x1150]={
  category="lo",
  direction="l",
 },
 [0x1151]={
  category="lo",
  direction="l",
 },
 [0x1152]={
  category="lo",
  direction="l",
 },
 [0x1153]={
  category="lo",
  direction="l",
 },
 [0x1154]={
  category="lo",
  direction="l",
 },
 [0x1155]={
  category="lo",
  direction="l",
 },
 [0x1156]={
  category="lo",
  direction="l",
 },
 [0x1157]={
  category="lo",
  direction="l",
 },
 [0x1158]={
  category="lo",
  direction="l",
 },
 [0x1159]={
  category="lo",
  direction="l",
 },
 [0x115A]={
  category="lo",
  direction="l",
 },
 [0x115B]={
  category="lo",
  direction="l",
 },
 [0x115C]={
  category="lo",
  direction="l",
 },
 [0x115D]={
  category="lo",
  direction="l",
 },
 [0x115E]={
  category="lo",
  direction="l",
 },
 [0x115F]={
  category="lo",
  direction="l",
 },
 [0x1160]={
  category="lo",
  direction="l",
 },
 [0x1161]={
  category="lo",
  direction="l",
 },
 [0x1162]={
  category="lo",
  direction="l",
 },
 [0x1163]={
  category="lo",
  direction="l",
 },
 [0x1164]={
  category="lo",
  direction="l",
 },
 [0x1165]={
  category="lo",
  direction="l",
 },
 [0x1166]={
  category="lo",
  direction="l",
 },
 [0x1167]={
  category="lo",
  direction="l",
 },
 [0x1168]={
  category="lo",
  direction="l",
 },
 [0x1169]={
  category="lo",
  direction="l",
 },
 [0x116A]={
  category="lo",
  direction="l",
 },
 [0x116B]={
  category="lo",
  direction="l",
 },
 [0x116C]={
  category="lo",
  direction="l",
 },
 [0x116D]={
  category="lo",
  direction="l",
 },
 [0x116E]={
  category="lo",
  direction="l",
 },
 [0x116F]={
  category="lo",
  direction="l",
 },
 [0x1170]={
  category="lo",
  direction="l",
 },
 [0x1171]={
  category="lo",
  direction="l",
 },
 [0x1172]={
  category="lo",
  direction="l",
 },
 [0x1173]={
  category="lo",
  direction="l",
 },
 [0x1174]={
  category="lo",
  direction="l",
 },
 [0x1175]={
  category="lo",
  direction="l",
 },
 [0x1176]={
  category="lo",
  direction="l",
 },
 [0x1177]={
  category="lo",
  direction="l",
 },
 [0x1178]={
  category="lo",
  direction="l",
 },
 [0x1179]={
  category="lo",
  direction="l",
 },
 [0x117A]={
  category="lo",
  direction="l",
 },
 [0x117B]={
  category="lo",
  direction="l",
 },
 [0x117C]={
  category="lo",
  direction="l",
 },
 [0x117D]={
  category="lo",
  direction="l",
 },
 [0x117E]={
  category="lo",
  direction="l",
 },
 [0x117F]={
  category="lo",
  direction="l",
 },
 [0x1180]={
  category="lo",
  direction="l",
 },
 [0x1181]={
  category="lo",
  direction="l",
 },
 [0x1182]={
  category="lo",
  direction="l",
 },
 [0x1183]={
  category="lo",
  direction="l",
 },
 [0x1184]={
  category="lo",
  direction="l",
 },
 [0x1185]={
  category="lo",
  direction="l",
 },
 [0x1186]={
  category="lo",
  direction="l",
 },
 [0x1187]={
  category="lo",
  direction="l",
 },
 [0x1188]={
  category="lo",
  direction="l",
 },
 [0x1189]={
  category="lo",
  direction="l",
 },
 [0x118A]={
  category="lo",
  direction="l",
 },
 [0x118B]={
  category="lo",
  direction="l",
 },
 [0x118C]={
  category="lo",
  direction="l",
 },
 [0x118D]={
  category="lo",
  direction="l",
 },
 [0x118E]={
  category="lo",
  direction="l",
 },
 [0x118F]={
  category="lo",
  direction="l",
 },
 [0x1190]={
  category="lo",
  direction="l",
 },
 [0x1191]={
  category="lo",
  direction="l",
 },
 [0x1192]={
  category="lo",
  direction="l",
 },
 [0x1193]={
  category="lo",
  direction="l",
 },
 [0x1194]={
  category="lo",
  direction="l",
 },
 [0x1195]={
  category="lo",
  direction="l",
 },
 [0x1196]={
  category="lo",
  direction="l",
 },
 [0x1197]={
  category="lo",
  direction="l",
 },
 [0x1198]={
  category="lo",
  direction="l",
 },
 [0x1199]={
  category="lo",
  direction="l",
 },
 [0x119A]={
  category="lo",
  direction="l",
 },
 [0x119B]={
  category="lo",
  direction="l",
 },
 [0x119C]={
  category="lo",
  direction="l",
 },
 [0x119D]={
  category="lo",
  direction="l",
 },
 [0x119E]={
  category="lo",
  direction="l",
 },
 [0x119F]={
  category="lo",
  direction="l",
 },
 [0x11A0]={
  category="lo",
  direction="l",
 },
 [0x11A1]={
  category="lo",
  direction="l",
 },
 [0x11A2]={
  category="lo",
  direction="l",
 },
 [0x11A3]={
  category="lo",
  direction="l",
 },
 [0x11A4]={
  category="lo",
  direction="l",
 },
 [0x11A5]={
  category="lo",
  direction="l",
 },
 [0x11A6]={
  category="lo",
  direction="l",
 },
 [0x11A7]={
  category="lo",
  direction="l",
 },
 [0x11A8]={
  category="lo",
  direction="l",
 },
 [0x11A9]={
  category="lo",
  direction="l",
 },
 [0x11AA]={
  category="lo",
  direction="l",
 },
 [0x11AB]={
  category="lo",
  direction="l",
 },
 [0x11AC]={
  category="lo",
  direction="l",
 },
 [0x11AD]={
  category="lo",
  direction="l",
 },
 [0x11AE]={
  category="lo",
  direction="l",
 },
 [0x11AF]={
  category="lo",
  direction="l",
 },
 [0x11B0]={
  category="lo",
  direction="l",
 },
 [0x11B1]={
  category="lo",
  direction="l",
 },
 [0x11B2]={
  category="lo",
  direction="l",
 },
 [0x11B3]={
  category="lo",
  direction="l",
 },
 [0x11B4]={
  category="lo",
  direction="l",
 },
 [0x11B5]={
  category="lo",
  direction="l",
 },
 [0x11B6]={
  category="lo",
  direction="l",
 },
 [0x11B7]={
  category="lo",
  direction="l",
 },
 [0x11B8]={
  category="lo",
  direction="l",
 },
 [0x11B9]={
  category="lo",
  direction="l",
 },
 [0x11BA]={
  category="lo",
  direction="l",
 },
 [0x11BB]={
  category="lo",
  direction="l",
 },
 [0x11BC]={
  category="lo",
  direction="l",
 },
 [0x11BD]={
  category="lo",
  direction="l",
 },
 [0x11BE]={
  category="lo",
  direction="l",
 },
 [0x11BF]={
  category="lo",
  direction="l",
 },
 [0x11C0]={
  category="lo",
  direction="l",
 },
 [0x11C1]={
  category="lo",
  direction="l",
 },
 [0x11C2]={
  category="lo",
  direction="l",
 },
 [0x11C3]={
  category="lo",
  direction="l",
 },
 [0x11C4]={
  category="lo",
  direction="l",
 },
 [0x11C5]={
  category="lo",
  direction="l",
 },
 [0x11C6]={
  category="lo",
  direction="l",
 },
 [0x11C7]={
  category="lo",
  direction="l",
 },
 [0x11C8]={
  category="lo",
  direction="l",
 },
 [0x11C9]={
  category="lo",
  direction="l",
 },
 [0x11CA]={
  category="lo",
  direction="l",
 },
 [0x11CB]={
  category="lo",
  direction="l",
 },
 [0x11CC]={
  category="lo",
  direction="l",
 },
 [0x11CD]={
  category="lo",
  direction="l",
 },
 [0x11CE]={
  category="lo",
  direction="l",
 },
 [0x11CF]={
  category="lo",
  direction="l",
 },
 [0x11D0]={
  category="lo",
  direction="l",
 },
 [0x11D1]={
  category="lo",
  direction="l",
 },
 [0x11D2]={
  category="lo",
  direction="l",
 },
 [0x11D3]={
  category="lo",
  direction="l",
 },
 [0x11D4]={
  category="lo",
  direction="l",
 },
 [0x11D5]={
  category="lo",
  direction="l",
 },
 [0x11D6]={
  category="lo",
  direction="l",
 },
 [0x11D7]={
  category="lo",
  direction="l",
 },
 [0x11D8]={
  category="lo",
  direction="l",
 },
 [0x11D9]={
  category="lo",
  direction="l",
 },
 [0x11DA]={
  category="lo",
  direction="l",
 },
 [0x11DB]={
  category="lo",
  direction="l",
 },
 [0x11DC]={
  category="lo",
  direction="l",
 },
 [0x11DD]={
  category="lo",
  direction="l",
 },
 [0x11DE]={
  category="lo",
  direction="l",
 },
 [0x11DF]={
  category="lo",
  direction="l",
 },
 [0x11E0]={
  category="lo",
  direction="l",
 },
 [0x11E1]={
  category="lo",
  direction="l",
 },
 [0x11E2]={
  category="lo",
  direction="l",
 },
 [0x11E3]={
  category="lo",
  direction="l",
 },
 [0x11E4]={
  category="lo",
  direction="l",
 },
 [0x11E5]={
  category="lo",
  direction="l",
 },
 [0x11E6]={
  category="lo",
  direction="l",
 },
 [0x11E7]={
  category="lo",
  direction="l",
 },
 [0x11E8]={
  category="lo",
  direction="l",
 },
 [0x11E9]={
  category="lo",
  direction="l",
 },
 [0x11EA]={
  category="lo",
  direction="l",
 },
 [0x11EB]={
  category="lo",
  direction="l",
 },
 [0x11EC]={
  category="lo",
  direction="l",
 },
 [0x11ED]={
  category="lo",
  direction="l",
 },
 [0x11EE]={
  category="lo",
  direction="l",
 },
 [0x11EF]={
  category="lo",
  direction="l",
 },
 [0x11F0]={
  category="lo",
  direction="l",
 },
 [0x11F1]={
  category="lo",
  direction="l",
 },
 [0x11F2]={
  category="lo",
  direction="l",
 },
 [0x11F3]={
  category="lo",
  direction="l",
 },
 [0x11F4]={
  category="lo",
  direction="l",
 },
 [0x11F5]={
  category="lo",
  direction="l",
 },
 [0x11F6]={
  category="lo",
  direction="l",
 },
 [0x11F7]={
  category="lo",
  direction="l",
 },
 [0x11F8]={
  category="lo",
  direction="l",
 },
 [0x11F9]={
  category="lo",
  direction="l",
 },
 [0x11FA]={
  category="lo",
  direction="l",
 },
 [0x11FB]={
  category="lo",
  direction="l",
 },
 [0x11FC]={
  category="lo",
  direction="l",
 },
 [0x11FD]={
  category="lo",
  direction="l",
 },
 [0x11FE]={
  category="lo",
  direction="l",
 },
 [0x11FF]={
  category="lo",
  direction="l",
 },
 [0x1200]={
  category="lo",
  direction="l",
 },
 [0x1201]={
  category="lo",
  direction="l",
 },
 [0x1202]={
  category="lo",
  direction="l",
 },
 [0x1203]={
  category="lo",
  direction="l",
 },
 [0x1204]={
  category="lo",
  direction="l",
 },
 [0x1205]={
  category="lo",
  direction="l",
 },
 [0x1206]={
  category="lo",
  direction="l",
 },
 [0x1207]={
  category="lo",
  direction="l",
 },
 [0x1208]={
  category="lo",
  direction="l",
 },
 [0x1209]={
  category="lo",
  direction="l",
 },
 [0x120A]={
  category="lo",
  direction="l",
 },
 [0x120B]={
  category="lo",
  direction="l",
 },
 [0x120C]={
  category="lo",
  direction="l",
 },
 [0x120D]={
  category="lo",
  direction="l",
 },
 [0x120E]={
  category="lo",
  direction="l",
 },
 [0x120F]={
  category="lo",
  direction="l",
 },
 [0x1210]={
  category="lo",
  direction="l",
 },
 [0x1211]={
  category="lo",
  direction="l",
 },
 [0x1212]={
  category="lo",
  direction="l",
 },
 [0x1213]={
  category="lo",
  direction="l",
 },
 [0x1214]={
  category="lo",
  direction="l",
 },
 [0x1215]={
  category="lo",
  direction="l",
 },
 [0x1216]={
  category="lo",
  direction="l",
 },
 [0x1217]={
  category="lo",
  direction="l",
 },
 [0x1218]={
  category="lo",
  direction="l",
 },
 [0x1219]={
  category="lo",
  direction="l",
 },
 [0x121A]={
  category="lo",
  direction="l",
 },
 [0x121B]={
  category="lo",
  direction="l",
 },
 [0x121C]={
  category="lo",
  direction="l",
 },
 [0x121D]={
  category="lo",
  direction="l",
 },
 [0x121E]={
  category="lo",
  direction="l",
 },
 [0x121F]={
  category="lo",
  direction="l",
 },
 [0x1220]={
  category="lo",
  direction="l",
 },
 [0x1221]={
  category="lo",
  direction="l",
 },
 [0x1222]={
  category="lo",
  direction="l",
 },
 [0x1223]={
  category="lo",
  direction="l",
 },
 [0x1224]={
  category="lo",
  direction="l",
 },
 [0x1225]={
  category="lo",
  direction="l",
 },
 [0x1226]={
  category="lo",
  direction="l",
 },
 [0x1227]={
  category="lo",
  direction="l",
 },
 [0x1228]={
  category="lo",
  direction="l",
 },
 [0x1229]={
  category="lo",
  direction="l",
 },
 [0x122A]={
  category="lo",
  direction="l",
 },
 [0x122B]={
  category="lo",
  direction="l",
 },
 [0x122C]={
  category="lo",
  direction="l",
 },
 [0x122D]={
  category="lo",
  direction="l",
 },
 [0x122E]={
  category="lo",
  direction="l",
 },
 [0x122F]={
  category="lo",
  direction="l",
 },
 [0x1230]={
  category="lo",
  direction="l",
 },
 [0x1231]={
  category="lo",
  direction="l",
 },
 [0x1232]={
  category="lo",
  direction="l",
 },
 [0x1233]={
  category="lo",
  direction="l",
 },
 [0x1234]={
  category="lo",
  direction="l",
 },
 [0x1235]={
  category="lo",
  direction="l",
 },
 [0x1236]={
  category="lo",
  direction="l",
 },
 [0x1237]={
  category="lo",
  direction="l",
 },
 [0x1238]={
  category="lo",
  direction="l",
 },
 [0x1239]={
  category="lo",
  direction="l",
 },
 [0x123A]={
  category="lo",
  direction="l",
 },
 [0x123B]={
  category="lo",
  direction="l",
 },
 [0x123C]={
  category="lo",
  direction="l",
 },
 [0x123D]={
  category="lo",
  direction="l",
 },
 [0x123E]={
  category="lo",
  direction="l",
 },
 [0x123F]={
  category="lo",
  direction="l",
 },
 [0x1240]={
  category="lo",
  direction="l",
 },
 [0x1241]={
  category="lo",
  direction="l",
 },
 [0x1242]={
  category="lo",
  direction="l",
 },
 [0x1243]={
  category="lo",
  direction="l",
 },
 [0x1244]={
  category="lo",
  direction="l",
 },
 [0x1245]={
  category="lo",
  direction="l",
 },
 [0x1246]={
  category="lo",
  direction="l",
 },
 [0x1247]={
  category="lo",
  direction="l",
 },
 [0x1248]={
  category="lo",
  direction="l",
 },
 [0x124A]={
  category="lo",
  direction="l",
 },
 [0x124B]={
  category="lo",
  direction="l",
 },
 [0x124C]={
  category="lo",
  direction="l",
 },
 [0x124D]={
  category="lo",
  direction="l",
 },
 [0x1250]={
  category="lo",
  direction="l",
 },
 [0x1251]={
  category="lo",
  direction="l",
 },
 [0x1252]={
  category="lo",
  direction="l",
 },
 [0x1253]={
  category="lo",
  direction="l",
 },
 [0x1254]={
  category="lo",
  direction="l",
 },
 [0x1255]={
  category="lo",
  direction="l",
 },
 [0x1256]={
  category="lo",
  direction="l",
 },
 [0x1258]={
  category="lo",
  direction="l",
 },
 [0x125A]={
  category="lo",
  direction="l",
 },
 [0x125B]={
  category="lo",
  direction="l",
 },
 [0x125C]={
  category="lo",
  direction="l",
 },
 [0x125D]={
  category="lo",
  direction="l",
 },
 [0x1260]={
  category="lo",
  direction="l",
 },
 [0x1261]={
  category="lo",
  direction="l",
 },
 [0x1262]={
  category="lo",
  direction="l",
 },
 [0x1263]={
  category="lo",
  direction="l",
 },
 [0x1264]={
  category="lo",
  direction="l",
 },
 [0x1265]={
  category="lo",
  direction="l",
 },
 [0x1266]={
  category="lo",
  direction="l",
 },
 [0x1267]={
  category="lo",
  direction="l",
 },
 [0x1268]={
  category="lo",
  direction="l",
 },
 [0x1269]={
  category="lo",
  direction="l",
 },
 [0x126A]={
  category="lo",
  direction="l",
 },
 [0x126B]={
  category="lo",
  direction="l",
 },
 [0x126C]={
  category="lo",
  direction="l",
 },
 [0x126D]={
  category="lo",
  direction="l",
 },
 [0x126E]={
  category="lo",
  direction="l",
 },
 [0x126F]={
  category="lo",
  direction="l",
 },
 [0x1270]={
  category="lo",
  direction="l",
 },
 [0x1271]={
  category="lo",
  direction="l",
 },
 [0x1272]={
  category="lo",
  direction="l",
 },
 [0x1273]={
  category="lo",
  direction="l",
 },
 [0x1274]={
  category="lo",
  direction="l",
 },
 [0x1275]={
  category="lo",
  direction="l",
 },
 [0x1276]={
  category="lo",
  direction="l",
 },
 [0x1277]={
  category="lo",
  direction="l",
 },
 [0x1278]={
  category="lo",
  direction="l",
 },
 [0x1279]={
  category="lo",
  direction="l",
 },
 [0x127A]={
  category="lo",
  direction="l",
 },
 [0x127B]={
  category="lo",
  direction="l",
 },
 [0x127C]={
  category="lo",
  direction="l",
 },
 [0x127D]={
  category="lo",
  direction="l",
 },
 [0x127E]={
  category="lo",
  direction="l",
 },
 [0x127F]={
  category="lo",
  direction="l",
 },
 [0x1280]={
  category="lo",
  direction="l",
 },
 [0x1281]={
  category="lo",
  direction="l",
 },
 [0x1282]={
  category="lo",
  direction="l",
 },
 [0x1283]={
  category="lo",
  direction="l",
 },
 [0x1284]={
  category="lo",
  direction="l",
 },
 [0x1285]={
  category="lo",
  direction="l",
 },
 [0x1286]={
  category="lo",
  direction="l",
 },
 [0x1287]={
  category="lo",
  direction="l",
 },
 [0x1288]={
  category="lo",
  direction="l",
 },
 [0x128A]={
  category="lo",
  direction="l",
 },
 [0x128B]={
  category="lo",
  direction="l",
 },
 [0x128C]={
  category="lo",
  direction="l",
 },
 [0x128D]={
  category="lo",
  direction="l",
 },
 [0x1290]={
  category="lo",
  direction="l",
 },
 [0x1291]={
  category="lo",
  direction="l",
 },
 [0x1292]={
  category="lo",
  direction="l",
 },
 [0x1293]={
  category="lo",
  direction="l",
 },
 [0x1294]={
  category="lo",
  direction="l",
 },
 [0x1295]={
  category="lo",
  direction="l",
 },
 [0x1296]={
  category="lo",
  direction="l",
 },
 [0x1297]={
  category="lo",
  direction="l",
 },
 [0x1298]={
  category="lo",
  direction="l",
 },
 [0x1299]={
  category="lo",
  direction="l",
 },
 [0x129A]={
  category="lo",
  direction="l",
 },
 [0x129B]={
  category="lo",
  direction="l",
 },
 [0x129C]={
  category="lo",
  direction="l",
 },
 [0x129D]={
  category="lo",
  direction="l",
 },
 [0x129E]={
  category="lo",
  direction="l",
 },
 [0x129F]={
  category="lo",
  direction="l",
 },
 [0x12A0]={
  category="lo",
  direction="l",
 },
 [0x12A1]={
  category="lo",
  direction="l",
 },
 [0x12A2]={
  category="lo",
  direction="l",
 },
 [0x12A3]={
  category="lo",
  direction="l",
 },
 [0x12A4]={
  category="lo",
  direction="l",
 },
 [0x12A5]={
  category="lo",
  direction="l",
 },
 [0x12A6]={
  category="lo",
  direction="l",
 },
 [0x12A7]={
  category="lo",
  direction="l",
 },
 [0x12A8]={
  category="lo",
  direction="l",
 },
 [0x12A9]={
  category="lo",
  direction="l",
 },
 [0x12AA]={
  category="lo",
  direction="l",
 },
 [0x12AB]={
  category="lo",
  direction="l",
 },
 [0x12AC]={
  category="lo",
  direction="l",
 },
 [0x12AD]={
  category="lo",
  direction="l",
 },
 [0x12AE]={
  category="lo",
  direction="l",
 },
 [0x12AF]={
  category="lo",
  direction="l",
 },
 [0x12B0]={
  category="lo",
  direction="l",
 },
 [0x12B2]={
  category="lo",
  direction="l",
 },
 [0x12B3]={
  category="lo",
  direction="l",
 },
 [0x12B4]={
  category="lo",
  direction="l",
 },
 [0x12B5]={
  category="lo",
  direction="l",
 },
 [0x12B8]={
  category="lo",
  direction="l",
 },
 [0x12B9]={
  category="lo",
  direction="l",
 },
 [0x12BA]={
  category="lo",
  direction="l",
 },
 [0x12BB]={
  category="lo",
  direction="l",
 },
 [0x12BC]={
  category="lo",
  direction="l",
 },
 [0x12BD]={
  category="lo",
  direction="l",
 },
 [0x12BE]={
  category="lo",
  direction="l",
 },
 [0x12C0]={
  category="lo",
  direction="l",
 },
 [0x12C2]={
  category="lo",
  direction="l",
 },
 [0x12C3]={
  category="lo",
  direction="l",
 },
 [0x12C4]={
  category="lo",
  direction="l",
 },
 [0x12C5]={
  category="lo",
  direction="l",
 },
 [0x12C8]={
  category="lo",
  direction="l",
 },
 [0x12C9]={
  category="lo",
  direction="l",
 },
 [0x12CA]={
  category="lo",
  direction="l",
 },
 [0x12CB]={
  category="lo",
  direction="l",
 },
 [0x12CC]={
  category="lo",
  direction="l",
 },
 [0x12CD]={
  category="lo",
  direction="l",
 },
 [0x12CE]={
  category="lo",
  direction="l",
 },
 [0x12CF]={
  category="lo",
  direction="l",
 },
 [0x12D0]={
  category="lo",
  direction="l",
 },
 [0x12D1]={
  category="lo",
  direction="l",
 },
 [0x12D2]={
  category="lo",
  direction="l",
 },
 [0x12D3]={
  category="lo",
  direction="l",
 },
 [0x12D4]={
  category="lo",
  direction="l",
 },
 [0x12D5]={
  category="lo",
  direction="l",
 },
 [0x12D6]={
  category="lo",
  direction="l",
 },
 [0x12D8]={
  category="lo",
  direction="l",
 },
 [0x12D9]={
  category="lo",
  direction="l",
 },
 [0x12DA]={
  category="lo",
  direction="l",
 },
 [0x12DB]={
  category="lo",
  direction="l",
 },
 [0x12DC]={
  category="lo",
  direction="l",
 },
 [0x12DD]={
  category="lo",
  direction="l",
 },
 [0x12DE]={
  category="lo",
  direction="l",
 },
 [0x12DF]={
  category="lo",
  direction="l",
 },
 [0x12E0]={
  category="lo",
  direction="l",
 },
 [0x12E1]={
  category="lo",
  direction="l",
 },
 [0x12E2]={
  category="lo",
  direction="l",
 },
 [0x12E3]={
  category="lo",
  direction="l",
 },
 [0x12E4]={
  category="lo",
  direction="l",
 },
 [0x12E5]={
  category="lo",
  direction="l",
 },
 [0x12E6]={
  category="lo",
  direction="l",
 },
 [0x12E7]={
  category="lo",
  direction="l",
 },
 [0x12E8]={
  category="lo",
  direction="l",
 },
 [0x12E9]={
  category="lo",
  direction="l",
 },
 [0x12EA]={
  category="lo",
  direction="l",
 },
 [0x12EB]={
  category="lo",
  direction="l",
 },
 [0x12EC]={
  category="lo",
  direction="l",
 },
 [0x12ED]={
  category="lo",
  direction="l",
 },
 [0x12EE]={
  category="lo",
  direction="l",
 },
 [0x12EF]={
  category="lo",
  direction="l",
 },
 [0x12F0]={
  category="lo",
  direction="l",
 },
 [0x12F1]={
  category="lo",
  direction="l",
 },
 [0x12F2]={
  category="lo",
  direction="l",
 },
 [0x12F3]={
  category="lo",
  direction="l",
 },
 [0x12F4]={
  category="lo",
  direction="l",
 },
 [0x12F5]={
  category="lo",
  direction="l",
 },
 [0x12F6]={
  category="lo",
  direction="l",
 },
 [0x12F7]={
  category="lo",
  direction="l",
 },
 [0x12F8]={
  category="lo",
  direction="l",
 },
 [0x12F9]={
  category="lo",
  direction="l",
 },
 [0x12FA]={
  category="lo",
  direction="l",
 },
 [0x12FB]={
  category="lo",
  direction="l",
 },
 [0x12FC]={
  category="lo",
  direction="l",
 },
 [0x12FD]={
  category="lo",
  direction="l",
 },
 [0x12FE]={
  category="lo",
  direction="l",
 },
 [0x12FF]={
  category="lo",
  direction="l",
 },
 [0x1300]={
  category="lo",
  direction="l",
 },
 [0x1301]={
  category="lo",
  direction="l",
 },
 [0x1302]={
  category="lo",
  direction="l",
 },
 [0x1303]={
  category="lo",
  direction="l",
 },
 [0x1304]={
  category="lo",
  direction="l",
 },
 [0x1305]={
  category="lo",
  direction="l",
 },
 [0x1306]={
  category="lo",
  direction="l",
 },
 [0x1307]={
  category="lo",
  direction="l",
 },
 [0x1308]={
  category="lo",
  direction="l",
 },
 [0x1309]={
  category="lo",
  direction="l",
 },
 [0x130A]={
  category="lo",
  direction="l",
 },
 [0x130B]={
  category="lo",
  direction="l",
 },
 [0x130C]={
  category="lo",
  direction="l",
 },
 [0x130D]={
  category="lo",
  direction="l",
 },
 [0x130E]={
  category="lo",
  direction="l",
 },
 [0x130F]={
  category="lo",
  direction="l",
 },
 [0x1310]={
  category="lo",
  direction="l",
 },
 [0x1312]={
  category="lo",
  direction="l",
 },
 [0x1313]={
  category="lo",
  direction="l",
 },
 [0x1314]={
  category="lo",
  direction="l",
 },
 [0x1315]={
  category="lo",
  direction="l",
 },
 [0x1318]={
  category="lo",
  direction="l",
 },
 [0x1319]={
  category="lo",
  direction="l",
 },
 [0x131A]={
  category="lo",
  direction="l",
 },
 [0x131B]={
  category="lo",
  direction="l",
 },
 [0x131C]={
  category="lo",
  direction="l",
 },
 [0x131D]={
  category="lo",
  direction="l",
 },
 [0x131E]={
  category="lo",
  direction="l",
 },
 [0x131F]={
  category="lo",
  direction="l",
 },
 [0x1320]={
  category="lo",
  direction="l",
 },
 [0x1321]={
  category="lo",
  direction="l",
 },
 [0x1322]={
  category="lo",
  direction="l",
 },
 [0x1323]={
  category="lo",
  direction="l",
 },
 [0x1324]={
  category="lo",
  direction="l",
 },
 [0x1325]={
  category="lo",
  direction="l",
 },
 [0x1326]={
  category="lo",
  direction="l",
 },
 [0x1327]={
  category="lo",
  direction="l",
 },
 [0x1328]={
  category="lo",
  direction="l",
 },
 [0x1329]={
  category="lo",
  direction="l",
 },
 [0x132A]={
  category="lo",
  direction="l",
 },
 [0x132B]={
  category="lo",
  direction="l",
 },
 [0x132C]={
  category="lo",
  direction="l",
 },
 [0x132D]={
  category="lo",
  direction="l",
 },
 [0x132E]={
  category="lo",
  direction="l",
 },
 [0x132F]={
  category="lo",
  direction="l",
 },
 [0x1330]={
  category="lo",
  direction="l",
 },
 [0x1331]={
  category="lo",
  direction="l",
 },
 [0x1332]={
  category="lo",
  direction="l",
 },
 [0x1333]={
  category="lo",
  direction="l",
 },
 [0x1334]={
  category="lo",
  direction="l",
 },
 [0x1335]={
  category="lo",
  direction="l",
 },
 [0x1336]={
  category="lo",
  direction="l",
 },
 [0x1337]={
  category="lo",
  direction="l",
 },
 [0x1338]={
  category="lo",
  direction="l",
 },
 [0x1339]={
  category="lo",
  direction="l",
 },
 [0x133A]={
  category="lo",
  direction="l",
 },
 [0x133B]={
  category="lo",
  direction="l",
 },
 [0x133C]={
  category="lo",
  direction="l",
 },
 [0x133D]={
  category="lo",
  direction="l",
 },
 [0x133E]={
  category="lo",
  direction="l",
 },
 [0x133F]={
  category="lo",
  direction="l",
 },
 [0x1340]={
  category="lo",
  direction="l",
 },
 [0x1341]={
  category="lo",
  direction="l",
 },
 [0x1342]={
  category="lo",
  direction="l",
 },
 [0x1343]={
  category="lo",
  direction="l",
 },
 [0x1344]={
  category="lo",
  direction="l",
 },
 [0x1345]={
  category="lo",
  direction="l",
 },
 [0x1346]={
  category="lo",
  direction="l",
 },
 [0x1347]={
  category="lo",
  direction="l",
 },
 [0x1348]={
  category="lo",
  direction="l",
 },
 [0x1349]={
  category="lo",
  direction="l",
 },
 [0x134A]={
  category="lo",
  direction="l",
 },
 [0x134B]={
  category="lo",
  direction="l",
 },
 [0x134C]={
  category="lo",
  direction="l",
 },
 [0x134D]={
  category="lo",
  direction="l",
 },
 [0x134E]={
  category="lo",
  direction="l",
 },
 [0x134F]={
  category="lo",
  direction="l",
 },
 [0x1350]={
  category="lo",
  direction="l",
 },
 [0x1351]={
  category="lo",
  direction="l",
 },
 [0x1352]={
  category="lo",
  direction="l",
 },
 [0x1353]={
  category="lo",
  direction="l",
 },
 [0x1354]={
  category="lo",
  direction="l",
 },
 [0x1355]={
  category="lo",
  direction="l",
 },
 [0x1356]={
  category="lo",
  direction="l",
 },
 [0x1357]={
  category="lo",
  direction="l",
 },
 [0x1358]={
  category="lo",
  direction="l",
 },
 [0x1359]={
  category="lo",
  direction="l",
 },
 [0x135A]={
  category="lo",
  direction="l",
 },
 [0x135D]={
  category="mn",
  direction="nsm",
 },
 [0x135E]={
  category="mn",
  direction="nsm",
 },
 [0x135F]={
  category="mn",
  direction="nsm",
 },
 [0x1360]={
  category="so",
  direction="l",
 },
 [0x1361]={
  category="po",
  direction="l",
 },
 [0x1362]={
  category="po",
  direction="l",
 },
 [0x1363]={
  category="po",
  direction="l",
 },
 [0x1364]={
  category="po",
  direction="l",
 },
 [0x1365]={
  category="po",
  direction="l",
 },
 [0x1366]={
  category="po",
  direction="l",
 },
 [0x1367]={
  category="po",
  direction="l",
 },
 [0x1368]={
  category="po",
  direction="l",
 },
 [0x1369]={
  category="no",
  direction="l",
 },
 [0x136A]={
  category="no",
  direction="l",
 },
 [0x136B]={
  category="no",
  direction="l",
 },
 [0x136C]={
  category="no",
  direction="l",
 },
 [0x136D]={
  category="no",
  direction="l",
 },
 [0x136E]={
  category="no",
  direction="l",
 },
 [0x136F]={
  category="no",
  direction="l",
 },
 [0x1370]={
  category="no",
  direction="l",
 },
 [0x1371]={
  category="no",
  direction="l",
 },
 [0x1372]={
  category="no",
  direction="l",
 },
 [0x1373]={
  category="no",
  direction="l",
 },
 [0x1374]={
  category="no",
  direction="l",
 },
 [0x1375]={
  category="no",
  direction="l",
 },
 [0x1376]={
  category="no",
  direction="l",
 },
 [0x1377]={
  category="no",
  direction="l",
 },
 [0x1378]={
  category="no",
  direction="l",
 },
 [0x1379]={
  category="no",
  direction="l",
 },
 [0x137A]={
  category="no",
  direction="l",
 },
 [0x137B]={
  category="no",
  direction="l",
 },
 [0x137C]={
  category="no",
  direction="l",
 },
 [0x1380]={
  category="lo",
  direction="l",
 },
 [0x1381]={
  category="lo",
  direction="l",
 },
 [0x1382]={
  category="lo",
  direction="l",
 },
 [0x1383]={
  category="lo",
  direction="l",
 },
 [0x1384]={
  category="lo",
  direction="l",
 },
 [0x1385]={
  category="lo",
  direction="l",
 },
 [0x1386]={
  category="lo",
  direction="l",
 },
 [0x1387]={
  category="lo",
  direction="l",
 },
 [0x1388]={
  category="lo",
  direction="l",
 },
 [0x1389]={
  category="lo",
  direction="l",
 },
 [0x138A]={
  category="lo",
  direction="l",
 },
 [0x138B]={
  category="lo",
  direction="l",
 },
 [0x138C]={
  category="lo",
  direction="l",
 },
 [0x138D]={
  category="lo",
  direction="l",
 },
 [0x138E]={
  category="lo",
  direction="l",
 },
 [0x138F]={
  category="lo",
  direction="l",
 },
 [0x1390]={
  category="so",
  direction="on",
 },
 [0x1391]={
  category="so",
  direction="on",
 },
 [0x1392]={
  category="so",
  direction="on",
 },
 [0x1393]={
  category="so",
  direction="on",
 },
 [0x1394]={
  category="so",
  direction="on",
 },
 [0x1395]={
  category="so",
  direction="on",
 },
 [0x1396]={
  category="so",
  direction="on",
 },
 [0x1397]={
  category="so",
  direction="on",
 },
 [0x1398]={
  category="so",
  direction="on",
 },
 [0x1399]={
  category="so",
  direction="on",
 },
 [0x13A0]={
  category="lo",
  direction="l",
 },
 [0x13A1]={
  category="lo",
  direction="l",
 },
 [0x13A2]={
  category="lo",
  direction="l",
 },
 [0x13A3]={
  category="lo",
  direction="l",
 },
 [0x13A4]={
  category="lo",
  direction="l",
 },
 [0x13A5]={
  category="lo",
  direction="l",
 },
 [0x13A6]={
  category="lo",
  direction="l",
 },
 [0x13A7]={
  category="lo",
  direction="l",
 },
 [0x13A8]={
  category="lo",
  direction="l",
 },
 [0x13A9]={
  category="lo",
  direction="l",
 },
 [0x13AA]={
  category="lo",
  direction="l",
 },
 [0x13AB]={
  category="lo",
  direction="l",
 },
 [0x13AC]={
  category="lo",
  direction="l",
 },
 [0x13AD]={
  category="lo",
  direction="l",
 },
 [0x13AE]={
  category="lo",
  direction="l",
 },
 [0x13AF]={
  category="lo",
  direction="l",
 },
 [0x13B0]={
  category="lo",
  direction="l",
 },
 [0x13B1]={
  category="lo",
  direction="l",
 },
 [0x13B2]={
  category="lo",
  direction="l",
 },
 [0x13B3]={
  category="lo",
  direction="l",
 },
 [0x13B4]={
  category="lo",
  direction="l",
 },
 [0x13B5]={
  category="lo",
  direction="l",
 },
 [0x13B6]={
  category="lo",
  direction="l",
 },
 [0x13B7]={
  category="lo",
  direction="l",
 },
 [0x13B8]={
  category="lo",
  direction="l",
 },
 [0x13B9]={
  category="lo",
  direction="l",
 },
 [0x13BA]={
  category="lo",
  direction="l",
 },
 [0x13BB]={
  category="lo",
  direction="l",
 },
 [0x13BC]={
  category="lo",
  direction="l",
 },
 [0x13BD]={
  category="lo",
  direction="l",
 },
 [0x13BE]={
  category="lo",
  direction="l",
 },
 [0x13BF]={
  category="lo",
  direction="l",
 },
 [0x13C0]={
  category="lo",
  direction="l",
 },
 [0x13C1]={
  category="lo",
  direction="l",
 },
 [0x13C2]={
  category="lo",
  direction="l",
 },
 [0x13C3]={
  category="lo",
  direction="l",
 },
 [0x13C4]={
  category="lo",
  direction="l",
 },
 [0x13C5]={
  category="lo",
  direction="l",
 },
 [0x13C6]={
  category="lo",
  direction="l",
 },
 [0x13C7]={
  category="lo",
  direction="l",
 },
 [0x13C8]={
  category="lo",
  direction="l",
 },
 [0x13C9]={
  category="lo",
  direction="l",
 },
 [0x13CA]={
  category="lo",
  direction="l",
 },
 [0x13CB]={
  category="lo",
  direction="l",
 },
 [0x13CC]={
  category="lo",
  direction="l",
 },
 [0x13CD]={
  category="lo",
  direction="l",
 },
 [0x13CE]={
  category="lo",
  direction="l",
 },
 [0x13CF]={
  category="lo",
  direction="l",
 },
 [0x13D0]={
  category="lo",
  direction="l",
 },
 [0x13D1]={
  category="lo",
  direction="l",
 },
 [0x13D2]={
  category="lo",
  direction="l",
 },
 [0x13D3]={
  category="lo",
  direction="l",
 },
 [0x13D4]={
  category="lo",
  direction="l",
 },
 [0x13D5]={
  category="lo",
  direction="l",
 },
 [0x13D6]={
  category="lo",
  direction="l",
 },
 [0x13D7]={
  category="lo",
  direction="l",
 },
 [0x13D8]={
  category="lo",
  direction="l",
 },
 [0x13D9]={
  category="lo",
  direction="l",
 },
 [0x13DA]={
  category="lo",
  direction="l",
 },
 [0x13DB]={
  category="lo",
  direction="l",
 },
 [0x13DC]={
  category="lo",
  direction="l",
 },
 [0x13DD]={
  category="lo",
  direction="l",
 },
 [0x13DE]={
  category="lo",
  direction="l",
 },
 [0x13DF]={
  category="lo",
  direction="l",
 },
 [0x13E0]={
  category="lo",
  direction="l",
 },
 [0x13E1]={
  category="lo",
  direction="l",
 },
 [0x13E2]={
  category="lo",
  direction="l",
 },
 [0x13E3]={
  category="lo",
  direction="l",
 },
 [0x13E4]={
  category="lo",
  direction="l",
 },
 [0x13E5]={
  category="lo",
  direction="l",
 },
 [0x13E6]={
  category="lo",
  direction="l",
 },
 [0x13E7]={
  category="lo",
  direction="l",
 },
 [0x13E8]={
  category="lo",
  direction="l",
 },
 [0x13E9]={
  category="lo",
  direction="l",
 },
 [0x13EA]={
  category="lo",
  direction="l",
 },
 [0x13EB]={
  category="lo",
  direction="l",
 },
 [0x13EC]={
  category="lo",
  direction="l",
 },
 [0x13ED]={
  category="lo",
  direction="l",
 },
 [0x13EE]={
  category="lo",
  direction="l",
 },
 [0x13EF]={
  category="lo",
  direction="l",
 },
 [0x13F0]={
  category="lo",
  direction="l",
 },
 [0x13F1]={
  category="lo",
  direction="l",
 },
 [0x13F2]={
  category="lo",
  direction="l",
 },
 [0x13F3]={
  category="lo",
  direction="l",
 },
 [0x13F4]={
  category="lo",
  direction="l",
 },
 [0x1400]={
  category="pd",
  direction="on",
 },
 [0x1401]={
  category="lo",
  direction="l",
 },
 [0x1402]={
  category="lo",
  direction="l",
 },
 [0x1403]={
  category="lo",
  direction="l",
 },
 [0x1404]={
  category="lo",
  direction="l",
 },
 [0x1405]={
  category="lo",
  direction="l",
 },
 [0x1406]={
  category="lo",
  direction="l",
 },
 [0x1407]={
  category="lo",
  direction="l",
 },
 [0x1408]={
  category="lo",
  direction="l",
 },
 [0x1409]={
  category="lo",
  direction="l",
 },
 [0x140A]={
  category="lo",
  direction="l",
 },
 [0x140B]={
  category="lo",
  direction="l",
 },
 [0x140C]={
  category="lo",
  direction="l",
 },
 [0x140D]={
  category="lo",
  direction="l",
 },
 [0x140E]={
  category="lo",
  direction="l",
 },
 [0x140F]={
  category="lo",
  direction="l",
 },
 [0x1410]={
  category="lo",
  direction="l",
 },
 [0x1411]={
  category="lo",
  direction="l",
 },
 [0x1412]={
  category="lo",
  direction="l",
 },
 [0x1413]={
  category="lo",
  direction="l",
 },
 [0x1414]={
  category="lo",
  direction="l",
 },
 [0x1415]={
  category="lo",
  direction="l",
 },
 [0x1416]={
  category="lo",
  direction="l",
 },
 [0x1417]={
  category="lo",
  direction="l",
 },
 [0x1418]={
  category="lo",
  direction="l",
 },
 [0x1419]={
  category="lo",
  direction="l",
 },
 [0x141A]={
  category="lo",
  direction="l",
 },
 [0x141B]={
  category="lo",
  direction="l",
 },
 [0x141C]={
  category="lo",
  direction="l",
 },
 [0x141D]={
  category="lo",
  direction="l",
 },
 [0x141E]={
  category="lo",
  direction="l",
 },
 [0x141F]={
  category="lo",
  direction="l",
 },
 [0x1420]={
  category="lo",
  direction="l",
 },
 [0x1421]={
  category="lo",
  direction="l",
 },
 [0x1422]={
  category="lo",
  direction="l",
 },
 [0x1423]={
  category="lo",
  direction="l",
 },
 [0x1424]={
  category="lo",
  direction="l",
 },
 [0x1425]={
  category="lo",
  direction="l",
 },
 [0x1426]={
  category="lo",
  direction="l",
 },
 [0x1427]={
  category="lo",
  direction="l",
 },
 [0x1428]={
  category="lo",
  direction="l",
 },
 [0x1429]={
  category="lo",
  direction="l",
 },
 [0x142A]={
  category="lo",
  direction="l",
 },
 [0x142B]={
  category="lo",
  direction="l",
 },
 [0x142C]={
  category="lo",
  direction="l",
 },
 [0x142D]={
  category="lo",
  direction="l",
 },
 [0x142E]={
  category="lo",
  direction="l",
 },
 [0x142F]={
  category="lo",
  direction="l",
 },
 [0x1430]={
  category="lo",
  direction="l",
 },
 [0x1431]={
  category="lo",
  direction="l",
 },
 [0x1432]={
  category="lo",
  direction="l",
 },
 [0x1433]={
  category="lo",
  direction="l",
 },
 [0x1434]={
  category="lo",
  direction="l",
 },
 [0x1435]={
  category="lo",
  direction="l",
 },
 [0x1436]={
  category="lo",
  direction="l",
 },
 [0x1437]={
  category="lo",
  direction="l",
 },
 [0x1438]={
  category="lo",
  direction="l",
 },
 [0x1439]={
  category="lo",
  direction="l",
 },
 [0x143A]={
  category="lo",
  direction="l",
 },
 [0x143B]={
  category="lo",
  direction="l",
 },
 [0x143C]={
  category="lo",
  direction="l",
 },
 [0x143D]={
  category="lo",
  direction="l",
 },
 [0x143E]={
  category="lo",
  direction="l",
 },
 [0x143F]={
  category="lo",
  direction="l",
 },
 [0x1440]={
  category="lo",
  direction="l",
 },
 [0x1441]={
  category="lo",
  direction="l",
 },
 [0x1442]={
  category="lo",
  direction="l",
 },
 [0x1443]={
  category="lo",
  direction="l",
 },
 [0x1444]={
  category="lo",
  direction="l",
 },
 [0x1445]={
  category="lo",
  direction="l",
 },
 [0x1446]={
  category="lo",
  direction="l",
 },
 [0x1447]={
  category="lo",
  direction="l",
 },
 [0x1448]={
  category="lo",
  direction="l",
 },
 [0x1449]={
  category="lo",
  direction="l",
 },
 [0x144A]={
  category="lo",
  direction="l",
 },
 [0x144B]={
  category="lo",
  direction="l",
 },
 [0x144C]={
  category="lo",
  direction="l",
 },
 [0x144D]={
  category="lo",
  direction="l",
 },
 [0x144E]={
  category="lo",
  direction="l",
 },
 [0x144F]={
  category="lo",
  direction="l",
 },
 [0x1450]={
  category="lo",
  direction="l",
 },
 [0x1451]={
  category="lo",
  direction="l",
 },
 [0x1452]={
  category="lo",
  direction="l",
 },
 [0x1453]={
  category="lo",
  direction="l",
 },
 [0x1454]={
  category="lo",
  direction="l",
 },
 [0x1455]={
  category="lo",
  direction="l",
 },
 [0x1456]={
  category="lo",
  direction="l",
 },
 [0x1457]={
  category="lo",
  direction="l",
 },
 [0x1458]={
  category="lo",
  direction="l",
 },
 [0x1459]={
  category="lo",
  direction="l",
 },
 [0x145A]={
  category="lo",
  direction="l",
 },
 [0x145B]={
  category="lo",
  direction="l",
 },
 [0x145C]={
  category="lo",
  direction="l",
 },
 [0x145D]={
  category="lo",
  direction="l",
 },
 [0x145E]={
  category="lo",
  direction="l",
 },
 [0x145F]={
  category="lo",
  direction="l",
 },
 [0x1460]={
  category="lo",
  direction="l",
 },
 [0x1461]={
  category="lo",
  direction="l",
 },
 [0x1462]={
  category="lo",
  direction="l",
 },
 [0x1463]={
  category="lo",
  direction="l",
 },
 [0x1464]={
  category="lo",
  direction="l",
 },
 [0x1465]={
  category="lo",
  direction="l",
 },
 [0x1466]={
  category="lo",
  direction="l",
 },
 [0x1467]={
  category="lo",
  direction="l",
 },
 [0x1468]={
  category="lo",
  direction="l",
 },
 [0x1469]={
  category="lo",
  direction="l",
 },
 [0x146A]={
  category="lo",
  direction="l",
 },
 [0x146B]={
  category="lo",
  direction="l",
 },
 [0x146C]={
  category="lo",
  direction="l",
 },
 [0x146D]={
  category="lo",
  direction="l",
 },
 [0x146E]={
  category="lo",
  direction="l",
 },
 [0x146F]={
  category="lo",
  direction="l",
 },
 [0x1470]={
  category="lo",
  direction="l",
 },
 [0x1471]={
  category="lo",
  direction="l",
 },
 [0x1472]={
  category="lo",
  direction="l",
 },
 [0x1473]={
  category="lo",
  direction="l",
 },
 [0x1474]={
  category="lo",
  direction="l",
 },
 [0x1475]={
  category="lo",
  direction="l",
 },
 [0x1476]={
  category="lo",
  direction="l",
 },
 [0x1477]={
  category="lo",
  direction="l",
 },
 [0x1478]={
  category="lo",
  direction="l",
 },
 [0x1479]={
  category="lo",
  direction="l",
 },
 [0x147A]={
  category="lo",
  direction="l",
 },
 [0x147B]={
  category="lo",
  direction="l",
 },
 [0x147C]={
  category="lo",
  direction="l",
 },
 [0x147D]={
  category="lo",
  direction="l",
 },
 [0x147E]={
  category="lo",
  direction="l",
 },
 [0x147F]={
  category="lo",
  direction="l",
 },
 [0x1480]={
  category="lo",
  direction="l",
 },
 [0x1481]={
  category="lo",
  direction="l",
 },
 [0x1482]={
  category="lo",
  direction="l",
 },
 [0x1483]={
  category="lo",
  direction="l",
 },
 [0x1484]={
  category="lo",
  direction="l",
 },
 [0x1485]={
  category="lo",
  direction="l",
 },
 [0x1486]={
  category="lo",
  direction="l",
 },
 [0x1487]={
  category="lo",
  direction="l",
 },
 [0x1488]={
  category="lo",
  direction="l",
 },
 [0x1489]={
  category="lo",
  direction="l",
 },
 [0x148A]={
  category="lo",
  direction="l",
 },
 [0x148B]={
  category="lo",
  direction="l",
 },
 [0x148C]={
  category="lo",
  direction="l",
 },
 [0x148D]={
  category="lo",
  direction="l",
 },
 [0x148E]={
  category="lo",
  direction="l",
 },
 [0x148F]={
  category="lo",
  direction="l",
 },
 [0x1490]={
  category="lo",
  direction="l",
 },
 [0x1491]={
  category="lo",
  direction="l",
 },
 [0x1492]={
  category="lo",
  direction="l",
 },
 [0x1493]={
  category="lo",
  direction="l",
 },
 [0x1494]={
  category="lo",
  direction="l",
 },
 [0x1495]={
  category="lo",
  direction="l",
 },
 [0x1496]={
  category="lo",
  direction="l",
 },
 [0x1497]={
  category="lo",
  direction="l",
 },
 [0x1498]={
  category="lo",
  direction="l",
 },
 [0x1499]={
  category="lo",
  direction="l",
 },
 [0x149A]={
  category="lo",
  direction="l",
 },
 [0x149B]={
  category="lo",
  direction="l",
 },
 [0x149C]={
  category="lo",
  direction="l",
 },
 [0x149D]={
  category="lo",
  direction="l",
 },
 [0x149E]={
  category="lo",
  direction="l",
 },
 [0x149F]={
  category="lo",
  direction="l",
 },
 [0x14A0]={
  category="lo",
  direction="l",
 },
 [0x14A1]={
  category="lo",
  direction="l",
 },
 [0x14A2]={
  category="lo",
  direction="l",
 },
 [0x14A3]={
  category="lo",
  direction="l",
 },
 [0x14A4]={
  category="lo",
  direction="l",
 },
 [0x14A5]={
  category="lo",
  direction="l",
 },
 [0x14A6]={
  category="lo",
  direction="l",
 },
 [0x14A7]={
  category="lo",
  direction="l",
 },
 [0x14A8]={
  category="lo",
  direction="l",
 },
 [0x14A9]={
  category="lo",
  direction="l",
 },
 [0x14AA]={
  category="lo",
  direction="l",
 },
 [0x14AB]={
  category="lo",
  direction="l",
 },
 [0x14AC]={
  category="lo",
  direction="l",
 },
 [0x14AD]={
  category="lo",
  direction="l",
 },
 [0x14AE]={
  category="lo",
  direction="l",
 },
 [0x14AF]={
  category="lo",
  direction="l",
 },
 [0x14B0]={
  category="lo",
  direction="l",
 },
 [0x14B1]={
  category="lo",
  direction="l",
 },
 [0x14B2]={
  category="lo",
  direction="l",
 },
 [0x14B3]={
  category="lo",
  direction="l",
 },
 [0x14B4]={
  category="lo",
  direction="l",
 },
 [0x14B5]={
  category="lo",
  direction="l",
 },
 [0x14B6]={
  category="lo",
  direction="l",
 },
 [0x14B7]={
  category="lo",
  direction="l",
 },
 [0x14B8]={
  category="lo",
  direction="l",
 },
 [0x14B9]={
  category="lo",
  direction="l",
 },
 [0x14BA]={
  category="lo",
  direction="l",
 },
 [0x14BB]={
  category="lo",
  direction="l",
 },
 [0x14BC]={
  category="lo",
  direction="l",
 },
 [0x14BD]={
  category="lo",
  direction="l",
 },
 [0x14BE]={
  category="lo",
  direction="l",
 },
 [0x14BF]={
  category="lo",
  direction="l",
 },
 [0x14C0]={
  category="lo",
  direction="l",
 },
 [0x14C1]={
  category="lo",
  direction="l",
 },
 [0x14C2]={
  category="lo",
  direction="l",
 },
 [0x14C3]={
  category="lo",
  direction="l",
 },
 [0x14C4]={
  category="lo",
  direction="l",
 },
 [0x14C5]={
  category="lo",
  direction="l",
 },
 [0x14C6]={
  category="lo",
  direction="l",
 },
 [0x14C7]={
  category="lo",
  direction="l",
 },
 [0x14C8]={
  category="lo",
  direction="l",
 },
 [0x14C9]={
  category="lo",
  direction="l",
 },
 [0x14CA]={
  category="lo",
  direction="l",
 },
 [0x14CB]={
  category="lo",
  direction="l",
 },
 [0x14CC]={
  category="lo",
  direction="l",
 },
 [0x14CD]={
  category="lo",
  direction="l",
 },
 [0x14CE]={
  category="lo",
  direction="l",
 },
 [0x14CF]={
  category="lo",
  direction="l",
 },
 [0x14D0]={
  category="lo",
  direction="l",
 },
 [0x14D1]={
  category="lo",
  direction="l",
 },
 [0x14D2]={
  category="lo",
  direction="l",
 },
 [0x14D3]={
  category="lo",
  direction="l",
 },
 [0x14D4]={
  category="lo",
  direction="l",
 },
 [0x14D5]={
  category="lo",
  direction="l",
 },
 [0x14D6]={
  category="lo",
  direction="l",
 },
 [0x14D7]={
  category="lo",
  direction="l",
 },
 [0x14D8]={
  category="lo",
  direction="l",
 },
 [0x14D9]={
  category="lo",
  direction="l",
 },
 [0x14DA]={
  category="lo",
  direction="l",
 },
 [0x14DB]={
  category="lo",
  direction="l",
 },
 [0x14DC]={
  category="lo",
  direction="l",
 },
 [0x14DD]={
  category="lo",
  direction="l",
 },
 [0x14DE]={
  category="lo",
  direction="l",
 },
 [0x14DF]={
  category="lo",
  direction="l",
 },
 [0x14E0]={
  category="lo",
  direction="l",
 },
 [0x14E1]={
  category="lo",
  direction="l",
 },
 [0x14E2]={
  category="lo",
  direction="l",
 },
 [0x14E3]={
  category="lo",
  direction="l",
 },
 [0x14E4]={
  category="lo",
  direction="l",
 },
 [0x14E5]={
  category="lo",
  direction="l",
 },
 [0x14E6]={
  category="lo",
  direction="l",
 },
 [0x14E7]={
  category="lo",
  direction="l",
 },
 [0x14E8]={
  category="lo",
  direction="l",
 },
 [0x14E9]={
  category="lo",
  direction="l",
 },
 [0x14EA]={
  category="lo",
  direction="l",
 },
 [0x14EB]={
  category="lo",
  direction="l",
 },
 [0x14EC]={
  category="lo",
  direction="l",
 },
 [0x14ED]={
  category="lo",
  direction="l",
 },
 [0x14EE]={
  category="lo",
  direction="l",
 },
 [0x14EF]={
  category="lo",
  direction="l",
 },
 [0x14F0]={
  category="lo",
  direction="l",
 },
 [0x14F1]={
  category="lo",
  direction="l",
 },
 [0x14F2]={
  category="lo",
  direction="l",
 },
 [0x14F3]={
  category="lo",
  direction="l",
 },
 [0x14F4]={
  category="lo",
  direction="l",
 },
 [0x14F5]={
  category="lo",
  direction="l",
 },
 [0x14F6]={
  category="lo",
  direction="l",
 },
 [0x14F7]={
  category="lo",
  direction="l",
 },
 [0x14F8]={
  category="lo",
  direction="l",
 },
 [0x14F9]={
  category="lo",
  direction="l",
 },
 [0x14FA]={
  category="lo",
  direction="l",
 },
 [0x14FB]={
  category="lo",
  direction="l",
 },
 [0x14FC]={
  category="lo",
  direction="l",
 },
 [0x14FD]={
  category="lo",
  direction="l",
 },
 [0x14FE]={
  category="lo",
  direction="l",
 },
 [0x14FF]={
  category="lo",
  direction="l",
 },
 [0x1500]={
  category="lo",
  direction="l",
 },
 [0x1501]={
  category="lo",
  direction="l",
 },
 [0x1502]={
  category="lo",
  direction="l",
 },
 [0x1503]={
  category="lo",
  direction="l",
 },
 [0x1504]={
  category="lo",
  direction="l",
 },
 [0x1505]={
  category="lo",
  direction="l",
 },
 [0x1506]={
  category="lo",
  direction="l",
 },
 [0x1507]={
  category="lo",
  direction="l",
 },
 [0x1508]={
  category="lo",
  direction="l",
 },
 [0x1509]={
  category="lo",
  direction="l",
 },
 [0x150A]={
  category="lo",
  direction="l",
 },
 [0x150B]={
  category="lo",
  direction="l",
 },
 [0x150C]={
  category="lo",
  direction="l",
 },
 [0x150D]={
  category="lo",
  direction="l",
 },
 [0x150E]={
  category="lo",
  direction="l",
 },
 [0x150F]={
  category="lo",
  direction="l",
 },
 [0x1510]={
  category="lo",
  direction="l",
 },
 [0x1511]={
  category="lo",
  direction="l",
 },
 [0x1512]={
  category="lo",
  direction="l",
 },
 [0x1513]={
  category="lo",
  direction="l",
 },
 [0x1514]={
  category="lo",
  direction="l",
 },
 [0x1515]={
  category="lo",
  direction="l",
 },
 [0x1516]={
  category="lo",
  direction="l",
 },
 [0x1517]={
  category="lo",
  direction="l",
 },
 [0x1518]={
  category="lo",
  direction="l",
 },
 [0x1519]={
  category="lo",
  direction="l",
 },
 [0x151A]={
  category="lo",
  direction="l",
 },
 [0x151B]={
  category="lo",
  direction="l",
 },
 [0x151C]={
  category="lo",
  direction="l",
 },
 [0x151D]={
  category="lo",
  direction="l",
 },
 [0x151E]={
  category="lo",
  direction="l",
 },
 [0x151F]={
  category="lo",
  direction="l",
 },
 [0x1520]={
  category="lo",
  direction="l",
 },
 [0x1521]={
  category="lo",
  direction="l",
 },
 [0x1522]={
  category="lo",
  direction="l",
 },
 [0x1523]={
  category="lo",
  direction="l",
 },
 [0x1524]={
  category="lo",
  direction="l",
 },
 [0x1525]={
  category="lo",
  direction="l",
 },
 [0x1526]={
  category="lo",
  direction="l",
 },
 [0x1527]={
  category="lo",
  direction="l",
 },
 [0x1528]={
  category="lo",
  direction="l",
 },
 [0x1529]={
  category="lo",
  direction="l",
 },
 [0x152A]={
  category="lo",
  direction="l",
 },
 [0x152B]={
  category="lo",
  direction="l",
 },
 [0x152C]={
  category="lo",
  direction="l",
 },
 [0x152D]={
  category="lo",
  direction="l",
 },
 [0x152E]={
  category="lo",
  direction="l",
 },
 [0x152F]={
  category="lo",
  direction="l",
 },
 [0x1530]={
  category="lo",
  direction="l",
 },
 [0x1531]={
  category="lo",
  direction="l",
 },
 [0x1532]={
  category="lo",
  direction="l",
 },
 [0x1533]={
  category="lo",
  direction="l",
 },
 [0x1534]={
  category="lo",
  direction="l",
 },
 [0x1535]={
  category="lo",
  direction="l",
 },
 [0x1536]={
  category="lo",
  direction="l",
 },
 [0x1537]={
  category="lo",
  direction="l",
 },
 [0x1538]={
  category="lo",
  direction="l",
 },
 [0x1539]={
  category="lo",
  direction="l",
 },
 [0x153A]={
  category="lo",
  direction="l",
 },
 [0x153B]={
  category="lo",
  direction="l",
 },
 [0x153C]={
  category="lo",
  direction="l",
 },
 [0x153D]={
  category="lo",
  direction="l",
 },
 [0x153E]={
  category="lo",
  direction="l",
 },
 [0x153F]={
  category="lo",
  direction="l",
 },
 [0x1540]={
  category="lo",
  direction="l",
 },
 [0x1541]={
  category="lo",
  direction="l",
 },
 [0x1542]={
  category="lo",
  direction="l",
 },
 [0x1543]={
  category="lo",
  direction="l",
 },
 [0x1544]={
  category="lo",
  direction="l",
 },
 [0x1545]={
  category="lo",
  direction="l",
 },
 [0x1546]={
  category="lo",
  direction="l",
 },
 [0x1547]={
  category="lo",
  direction="l",
 },
 [0x1548]={
  category="lo",
  direction="l",
 },
 [0x1549]={
  category="lo",
  direction="l",
 },
 [0x154A]={
  category="lo",
  direction="l",
 },
 [0x154B]={
  category="lo",
  direction="l",
 },
 [0x154C]={
  category="lo",
  direction="l",
 },
 [0x154D]={
  category="lo",
  direction="l",
 },
 [0x154E]={
  category="lo",
  direction="l",
 },
 [0x154F]={
  category="lo",
  direction="l",
 },
 [0x1550]={
  category="lo",
  direction="l",
 },
 [0x1551]={
  category="lo",
  direction="l",
 },
 [0x1552]={
  category="lo",
  direction="l",
 },
 [0x1553]={
  category="lo",
  direction="l",
 },
 [0x1554]={
  category="lo",
  direction="l",
 },
 [0x1555]={
  category="lo",
  direction="l",
 },
 [0x1556]={
  category="lo",
  direction="l",
 },
 [0x1557]={
  category="lo",
  direction="l",
 },
 [0x1558]={
  category="lo",
  direction="l",
 },
 [0x1559]={
  category="lo",
  direction="l",
 },
 [0x155A]={
  category="lo",
  direction="l",
 },
 [0x155B]={
  category="lo",
  direction="l",
 },
 [0x155C]={
  category="lo",
  direction="l",
 },
 [0x155D]={
  category="lo",
  direction="l",
 },
 [0x155E]={
  category="lo",
  direction="l",
 },
 [0x155F]={
  category="lo",
  direction="l",
 },
 [0x1560]={
  category="lo",
  direction="l",
 },
 [0x1561]={
  category="lo",
  direction="l",
 },
 [0x1562]={
  category="lo",
  direction="l",
 },
 [0x1563]={
  category="lo",
  direction="l",
 },
 [0x1564]={
  category="lo",
  direction="l",
 },
 [0x1565]={
  category="lo",
  direction="l",
 },
 [0x1566]={
  category="lo",
  direction="l",
 },
 [0x1567]={
  category="lo",
  direction="l",
 },
 [0x1568]={
  category="lo",
  direction="l",
 },
 [0x1569]={
  category="lo",
  direction="l",
 },
 [0x156A]={
  category="lo",
  direction="l",
 },
 [0x156B]={
  category="lo",
  direction="l",
 },
 [0x156C]={
  category="lo",
  direction="l",
 },
 [0x156D]={
  category="lo",
  direction="l",
 },
 [0x156E]={
  category="lo",
  direction="l",
 },
 [0x156F]={
  category="lo",
  direction="l",
 },
 [0x1570]={
  category="lo",
  direction="l",
 },
 [0x1571]={
  category="lo",
  direction="l",
 },
 [0x1572]={
  category="lo",
  direction="l",
 },
 [0x1573]={
  category="lo",
  direction="l",
 },
 [0x1574]={
  category="lo",
  direction="l",
 },
 [0x1575]={
  category="lo",
  direction="l",
 },
 [0x1576]={
  category="lo",
  direction="l",
 },
 [0x1577]={
  category="lo",
  direction="l",
 },
 [0x1578]={
  category="lo",
  direction="l",
 },
 [0x1579]={
  category="lo",
  direction="l",
 },
 [0x157A]={
  category="lo",
  direction="l",
 },
 [0x157B]={
  category="lo",
  direction="l",
 },
 [0x157C]={
  category="lo",
  direction="l",
 },
 [0x157D]={
  category="lo",
  direction="l",
 },
 [0x157E]={
  category="lo",
  direction="l",
 },
 [0x157F]={
  category="lo",
  direction="l",
 },
 [0x1580]={
  category="lo",
  direction="l",
 },
 [0x1581]={
  category="lo",
  direction="l",
 },
 [0x1582]={
  category="lo",
  direction="l",
 },
 [0x1583]={
  category="lo",
  direction="l",
 },
 [0x1584]={
  category="lo",
  direction="l",
 },
 [0x1585]={
  category="lo",
  direction="l",
 },
 [0x1586]={
  category="lo",
  direction="l",
 },
 [0x1587]={
  category="lo",
  direction="l",
 },
 [0x1588]={
  category="lo",
  direction="l",
 },
 [0x1589]={
  category="lo",
  direction="l",
 },
 [0x158A]={
  category="lo",
  direction="l",
 },
 [0x158B]={
  category="lo",
  direction="l",
 },
 [0x158C]={
  category="lo",
  direction="l",
 },
 [0x158D]={
  category="lo",
  direction="l",
 },
 [0x158E]={
  category="lo",
  direction="l",
 },
 [0x158F]={
  category="lo",
  direction="l",
 },
 [0x1590]={
  category="lo",
  direction="l",
 },
 [0x1591]={
  category="lo",
  direction="l",
 },
 [0x1592]={
  category="lo",
  direction="l",
 },
 [0x1593]={
  category="lo",
  direction="l",
 },
 [0x1594]={
  category="lo",
  direction="l",
 },
 [0x1595]={
  category="lo",
  direction="l",
 },
 [0x1596]={
  category="lo",
  direction="l",
 },
 [0x1597]={
  category="lo",
  direction="l",
 },
 [0x1598]={
  category="lo",
  direction="l",
 },
 [0x1599]={
  category="lo",
  direction="l",
 },
 [0x159A]={
  category="lo",
  direction="l",
 },
 [0x159B]={
  category="lo",
  direction="l",
 },
 [0x159C]={
  category="lo",
  direction="l",
 },
 [0x159D]={
  category="lo",
  direction="l",
 },
 [0x159E]={
  category="lo",
  direction="l",
 },
 [0x159F]={
  category="lo",
  direction="l",
 },
 [0x15A0]={
  category="lo",
  direction="l",
 },
 [0x15A1]={
  category="lo",
  direction="l",
 },
 [0x15A2]={
  category="lo",
  direction="l",
 },
 [0x15A3]={
  category="lo",
  direction="l",
 },
 [0x15A4]={
  category="lo",
  direction="l",
 },
 [0x15A5]={
  category="lo",
  direction="l",
 },
 [0x15A6]={
  category="lo",
  direction="l",
 },
 [0x15A7]={
  category="lo",
  direction="l",
 },
 [0x15A8]={
  category="lo",
  direction="l",
 },
 [0x15A9]={
  category="lo",
  direction="l",
 },
 [0x15AA]={
  category="lo",
  direction="l",
 },
 [0x15AB]={
  category="lo",
  direction="l",
 },
 [0x15AC]={
  category="lo",
  direction="l",
 },
 [0x15AD]={
  category="lo",
  direction="l",
 },
 [0x15AE]={
  category="lo",
  direction="l",
 },
 [0x15AF]={
  category="lo",
  direction="l",
 },
 [0x15B0]={
  category="lo",
  direction="l",
 },
 [0x15B1]={
  category="lo",
  direction="l",
 },
 [0x15B2]={
  category="lo",
  direction="l",
 },
 [0x15B3]={
  category="lo",
  direction="l",
 },
 [0x15B4]={
  category="lo",
  direction="l",
 },
 [0x15B5]={
  category="lo",
  direction="l",
 },
 [0x15B6]={
  category="lo",
  direction="l",
 },
 [0x15B7]={
  category="lo",
  direction="l",
 },
 [0x15B8]={
  category="lo",
  direction="l",
 },
 [0x15B9]={
  category="lo",
  direction="l",
 },
 [0x15BA]={
  category="lo",
  direction="l",
 },
 [0x15BB]={
  category="lo",
  direction="l",
 },
 [0x15BC]={
  category="lo",
  direction="l",
 },
 [0x15BD]={
  category="lo",
  direction="l",
 },
 [0x15BE]={
  category="lo",
  direction="l",
 },
 [0x15BF]={
  category="lo",
  direction="l",
 },
 [0x15C0]={
  category="lo",
  direction="l",
 },
 [0x15C1]={
  category="lo",
  direction="l",
 },
 [0x15C2]={
  category="lo",
  direction="l",
 },
 [0x15C3]={
  category="lo",
  direction="l",
 },
 [0x15C4]={
  category="lo",
  direction="l",
 },
 [0x15C5]={
  category="lo",
  direction="l",
 },
 [0x15C6]={
  category="lo",
  direction="l",
 },
 [0x15C7]={
  category="lo",
  direction="l",
 },
 [0x15C8]={
  category="lo",
  direction="l",
 },
 [0x15C9]={
  category="lo",
  direction="l",
 },
 [0x15CA]={
  category="lo",
  direction="l",
 },
 [0x15CB]={
  category="lo",
  direction="l",
 },
 [0x15CC]={
  category="lo",
  direction="l",
 },
 [0x15CD]={
  category="lo",
  direction="l",
 },
 [0x15CE]={
  category="lo",
  direction="l",
 },
 [0x15CF]={
  category="lo",
  direction="l",
 },
 [0x15D0]={
  category="lo",
  direction="l",
 },
 [0x15D1]={
  category="lo",
  direction="l",
 },
 [0x15D2]={
  category="lo",
  direction="l",
 },
 [0x15D3]={
  category="lo",
  direction="l",
 },
 [0x15D4]={
  category="lo",
  direction="l",
 },
 [0x15D5]={
  category="lo",
  direction="l",
 },
 [0x15D6]={
  category="lo",
  direction="l",
 },
 [0x15D7]={
  category="lo",
  direction="l",
 },
 [0x15D8]={
  category="lo",
  direction="l",
 },
 [0x15D9]={
  category="lo",
  direction="l",
 },
 [0x15DA]={
  category="lo",
  direction="l",
 },
 [0x15DB]={
  category="lo",
  direction="l",
 },
 [0x15DC]={
  category="lo",
  direction="l",
 },
 [0x15DD]={
  category="lo",
  direction="l",
 },
 [0x15DE]={
  category="lo",
  direction="l",
 },
 [0x15DF]={
  category="lo",
  direction="l",
 },
 [0x15E0]={
  category="lo",
  direction="l",
 },
 [0x15E1]={
  category="lo",
  direction="l",
 },
 [0x15E2]={
  category="lo",
  direction="l",
 },
 [0x15E3]={
  category="lo",
  direction="l",
 },
 [0x15E4]={
  category="lo",
  direction="l",
 },
 [0x15E5]={
  category="lo",
  direction="l",
 },
 [0x15E6]={
  category="lo",
  direction="l",
 },
 [0x15E7]={
  category="lo",
  direction="l",
 },
 [0x15E8]={
  category="lo",
  direction="l",
 },
 [0x15E9]={
  category="lo",
  direction="l",
 },
 [0x15EA]={
  category="lo",
  direction="l",
 },
 [0x15EB]={
  category="lo",
  direction="l",
 },
 [0x15EC]={
  category="lo",
  direction="l",
 },
 [0x15ED]={
  category="lo",
  direction="l",
 },
 [0x15EE]={
  category="lo",
  direction="l",
 },
 [0x15EF]={
  category="lo",
  direction="l",
 },
 [0x15F0]={
  category="lo",
  direction="l",
 },
 [0x15F1]={
  category="lo",
  direction="l",
 },
 [0x15F2]={
  category="lo",
  direction="l",
 },
 [0x15F3]={
  category="lo",
  direction="l",
 },
 [0x15F4]={
  category="lo",
  direction="l",
 },
 [0x15F5]={
  category="lo",
  direction="l",
 },
 [0x15F6]={
  category="lo",
  direction="l",
 },
 [0x15F7]={
  category="lo",
  direction="l",
 },
 [0x15F8]={
  category="lo",
  direction="l",
 },
 [0x15F9]={
  category="lo",
  direction="l",
 },
 [0x15FA]={
  category="lo",
  direction="l",
 },
 [0x15FB]={
  category="lo",
  direction="l",
 },
 [0x15FC]={
  category="lo",
  direction="l",
 },
 [0x15FD]={
  category="lo",
  direction="l",
 },
 [0x15FE]={
  category="lo",
  direction="l",
 },
 [0x15FF]={
  category="lo",
  direction="l",
 },
 [0x1600]={
  category="lo",
  direction="l",
 },
 [0x1601]={
  category="lo",
  direction="l",
 },
 [0x1602]={
  category="lo",
  direction="l",
 },
 [0x1603]={
  category="lo",
  direction="l",
 },
 [0x1604]={
  category="lo",
  direction="l",
 },
 [0x1605]={
  category="lo",
  direction="l",
 },
 [0x1606]={
  category="lo",
  direction="l",
 },
 [0x1607]={
  category="lo",
  direction="l",
 },
 [0x1608]={
  category="lo",
  direction="l",
 },
 [0x1609]={
  category="lo",
  direction="l",
 },
 [0x160A]={
  category="lo",
  direction="l",
 },
 [0x160B]={
  category="lo",
  direction="l",
 },
 [0x160C]={
  category="lo",
  direction="l",
 },
 [0x160D]={
  category="lo",
  direction="l",
 },
 [0x160E]={
  category="lo",
  direction="l",
 },
 [0x160F]={
  category="lo",
  direction="l",
 },
 [0x1610]={
  category="lo",
  direction="l",
 },
 [0x1611]={
  category="lo",
  direction="l",
 },
 [0x1612]={
  category="lo",
  direction="l",
 },
 [0x1613]={
  category="lo",
  direction="l",
 },
 [0x1614]={
  category="lo",
  direction="l",
 },
 [0x1615]={
  category="lo",
  direction="l",
 },
 [0x1616]={
  category="lo",
  direction="l",
 },
 [0x1617]={
  category="lo",
  direction="l",
 },
 [0x1618]={
  category="lo",
  direction="l",
 },
 [0x1619]={
  category="lo",
  direction="l",
 },
 [0x161A]={
  category="lo",
  direction="l",
 },
 [0x161B]={
  category="lo",
  direction="l",
 },
 [0x161C]={
  category="lo",
  direction="l",
 },
 [0x161D]={
  category="lo",
  direction="l",
 },
 [0x161E]={
  category="lo",
  direction="l",
 },
 [0x161F]={
  category="lo",
  direction="l",
 },
 [0x1620]={
  category="lo",
  direction="l",
 },
 [0x1621]={
  category="lo",
  direction="l",
 },
 [0x1622]={
  category="lo",
  direction="l",
 },
 [0x1623]={
  category="lo",
  direction="l",
 },
 [0x1624]={
  category="lo",
  direction="l",
 },
 [0x1625]={
  category="lo",
  direction="l",
 },
 [0x1626]={
  category="lo",
  direction="l",
 },
 [0x1627]={
  category="lo",
  direction="l",
 },
 [0x1628]={
  category="lo",
  direction="l",
 },
 [0x1629]={
  category="lo",
  direction="l",
 },
 [0x162A]={
  category="lo",
  direction="l",
 },
 [0x162B]={
  category="lo",
  direction="l",
 },
 [0x162C]={
  category="lo",
  direction="l",
 },
 [0x162D]={
  category="lo",
  direction="l",
 },
 [0x162E]={
  category="lo",
  direction="l",
 },
 [0x162F]={
  category="lo",
  direction="l",
 },
 [0x1630]={
  category="lo",
  direction="l",
 },
 [0x1631]={
  category="lo",
  direction="l",
 },
 [0x1632]={
  category="lo",
  direction="l",
 },
 [0x1633]={
  category="lo",
  direction="l",
 },
 [0x1634]={
  category="lo",
  direction="l",
 },
 [0x1635]={
  category="lo",
  direction="l",
 },
 [0x1636]={
  category="lo",
  direction="l",
 },
 [0x1637]={
  category="lo",
  direction="l",
 },
 [0x1638]={
  category="lo",
  direction="l",
 },
 [0x1639]={
  category="lo",
  direction="l",
 },
 [0x163A]={
  category="lo",
  direction="l",
 },
 [0x163B]={
  category="lo",
  direction="l",
 },
 [0x163C]={
  category="lo",
  direction="l",
 },
 [0x163D]={
  category="lo",
  direction="l",
 },
 [0x163E]={
  category="lo",
  direction="l",
 },
 [0x163F]={
  category="lo",
  direction="l",
 },
 [0x1640]={
  category="lo",
  direction="l",
 },
 [0x1641]={
  category="lo",
  direction="l",
 },
 [0x1642]={
  category="lo",
  direction="l",
 },
 [0x1643]={
  category="lo",
  direction="l",
 },
 [0x1644]={
  category="lo",
  direction="l",
 },
 [0x1645]={
  category="lo",
  direction="l",
 },
 [0x1646]={
  category="lo",
  direction="l",
 },
 [0x1647]={
  category="lo",
  direction="l",
 },
 [0x1648]={
  category="lo",
  direction="l",
 },
 [0x1649]={
  category="lo",
  direction="l",
 },
 [0x164A]={
  category="lo",
  direction="l",
 },
 [0x164B]={
  category="lo",
  direction="l",
 },
 [0x164C]={
  category="lo",
  direction="l",
 },
 [0x164D]={
  category="lo",
  direction="l",
 },
 [0x164E]={
  category="lo",
  direction="l",
 },
 [0x164F]={
  category="lo",
  direction="l",
 },
 [0x1650]={
  category="lo",
  direction="l",
 },
 [0x1651]={
  category="lo",
  direction="l",
 },
 [0x1652]={
  category="lo",
  direction="l",
 },
 [0x1653]={
  category="lo",
  direction="l",
 },
 [0x1654]={
  category="lo",
  direction="l",
 },
 [0x1655]={
  category="lo",
  direction="l",
 },
 [0x1656]={
  category="lo",
  direction="l",
 },
 [0x1657]={
  category="lo",
  direction="l",
 },
 [0x1658]={
  category="lo",
  direction="l",
 },
 [0x1659]={
  category="lo",
  direction="l",
 },
 [0x165A]={
  category="lo",
  direction="l",
 },
 [0x165B]={
  category="lo",
  direction="l",
 },
 [0x165C]={
  category="lo",
  direction="l",
 },
 [0x165D]={
  category="lo",
  direction="l",
 },
 [0x165E]={
  category="lo",
  direction="l",
 },
 [0x165F]={
  category="lo",
  direction="l",
 },
 [0x1660]={
  category="lo",
  direction="l",
 },
 [0x1661]={
  category="lo",
  direction="l",
 },
 [0x1662]={
  category="lo",
  direction="l",
 },
 [0x1663]={
  category="lo",
  direction="l",
 },
 [0x1664]={
  category="lo",
  direction="l",
 },
 [0x1665]={
  category="lo",
  direction="l",
 },
 [0x1666]={
  category="lo",
  direction="l",
 },
 [0x1667]={
  category="lo",
  direction="l",
 },
 [0x1668]={
  category="lo",
  direction="l",
 },
 [0x1669]={
  category="lo",
  direction="l",
 },
 [0x166A]={
  category="lo",
  direction="l",
 },
 [0x166B]={
  category="lo",
  direction="l",
 },
 [0x166C]={
  category="lo",
  direction="l",
 },
 [0x166D]={
  category="po",
  direction="l",
 },
 [0x166E]={
  category="po",
  direction="l",
 },
 [0x166F]={
  category="lo",
  direction="l",
 },
 [0x1670]={
  category="lo",
  direction="l",
 },
 [0x1671]={
  category="lo",
  direction="l",
 },
 [0x1672]={
  category="lo",
  direction="l",
 },
 [0x1673]={
  category="lo",
  direction="l",
 },
 [0x1674]={
  category="lo",
  direction="l",
 },
 [0x1675]={
  category="lo",
  direction="l",
 },
 [0x1676]={
  category="lo",
  direction="l",
 },
 [0x1677]={
  category="lo",
  direction="l",
 },
 [0x1678]={
  category="lo",
  direction="l",
 },
 [0x1679]={
  category="lo",
  direction="l",
 },
 [0x167A]={
  category="lo",
  direction="l",
 },
 [0x167B]={
  category="lo",
  direction="l",
 },
 [0x167C]={
  category="lo",
  direction="l",
 },
 [0x167D]={
  category="lo",
  direction="l",
 },
 [0x167E]={
  category="lo",
  direction="l",
 },
 [0x167F]={
  category="lo",
  direction="l",
 },
 [0x1680]={
  category="zs",
  direction="ws",
 },
 [0x1681]={
  category="lo",
  direction="l",
 },
 [0x1682]={
  category="lo",
  direction="l",
 },
 [0x1683]={
  category="lo",
  direction="l",
 },
 [0x1684]={
  category="lo",
  direction="l",
 },
 [0x1685]={
  category="lo",
  direction="l",
 },
 [0x1686]={
  category="lo",
  direction="l",
 },
 [0x1687]={
  category="lo",
  direction="l",
 },
 [0x1688]={
  category="lo",
  direction="l",
 },
 [0x1689]={
  category="lo",
  direction="l",
 },
 [0x168A]={
  category="lo",
  direction="l",
 },
 [0x168B]={
  category="lo",
  direction="l",
 },
 [0x168C]={
  category="lo",
  direction="l",
 },
 [0x168D]={
  category="lo",
  direction="l",
 },
 [0x168E]={
  category="lo",
  direction="l",
 },
 [0x168F]={
  category="lo",
  direction="l",
 },
 [0x1690]={
  category="lo",
  direction="l",
 },
 [0x1691]={
  category="lo",
  direction="l",
 },
 [0x1692]={
  category="lo",
  direction="l",
 },
 [0x1693]={
  category="lo",
  direction="l",
 },
 [0x1694]={
  category="lo",
  direction="l",
 },
 [0x1695]={
  category="lo",
  direction="l",
 },
 [0x1696]={
  category="lo",
  direction="l",
 },
 [0x1697]={
  category="lo",
  direction="l",
 },
 [0x1698]={
  category="lo",
  direction="l",
 },
 [0x1699]={
  category="lo",
  direction="l",
 },
 [0x169A]={
  category="lo",
  direction="l",
 },
 [0x169B]={
  category="ps",
  direction="on",
  mirror=0x169C,
 },
 [0x169C]={
  category="pe",
  direction="on",
  mirror=0x169B,
 },
 [0x16A0]={
  category="lo",
  direction="l",
 },
 [0x16A1]={
  category="lo",
  direction="l",
 },
 [0x16A2]={
  category="lo",
  direction="l",
 },
 [0x16A3]={
  category="lo",
  direction="l",
 },
 [0x16A4]={
  category="lo",
  direction="l",
 },
 [0x16A5]={
  category="lo",
  direction="l",
 },
 [0x16A6]={
  category="lo",
  direction="l",
 },
 [0x16A7]={
  category="lo",
  direction="l",
 },
 [0x16A8]={
  category="lo",
  direction="l",
 },
 [0x16A9]={
  category="lo",
  direction="l",
 },
 [0x16AA]={
  category="lo",
  direction="l",
 },
 [0x16AB]={
  category="lo",
  direction="l",
 },
 [0x16AC]={
  category="lo",
  direction="l",
 },
 [0x16AD]={
  category="lo",
  direction="l",
 },
 [0x16AE]={
  category="lo",
  direction="l",
 },
 [0x16AF]={
  category="lo",
  direction="l",
 },
 [0x16B0]={
  category="lo",
  direction="l",
 },
 [0x16B1]={
  category="lo",
  direction="l",
 },
 [0x16B2]={
  category="lo",
  direction="l",
 },
 [0x16B3]={
  category="lo",
  direction="l",
 },
 [0x16B4]={
  category="lo",
  direction="l",
 },
 [0x16B5]={
  category="lo",
  direction="l",
 },
 [0x16B6]={
  category="lo",
  direction="l",
 },
 [0x16B7]={
  category="lo",
  direction="l",
 },
 [0x16B8]={
  category="lo",
  direction="l",
 },
 [0x16B9]={
  category="lo",
  direction="l",
 },
 [0x16BA]={
  category="lo",
  direction="l",
 },
 [0x16BB]={
  category="lo",
  direction="l",
 },
 [0x16BC]={
  category="lo",
  direction="l",
 },
 [0x16BD]={
  category="lo",
  direction="l",
 },
 [0x16BE]={
  category="lo",
  direction="l",
 },
 [0x16BF]={
  category="lo",
  direction="l",
 },
 [0x16C0]={
  category="lo",
  direction="l",
 },
 [0x16C1]={
  category="lo",
  direction="l",
 },
 [0x16C2]={
  category="lo",
  direction="l",
 },
 [0x16C3]={
  category="lo",
  direction="l",
 },
 [0x16C4]={
  category="lo",
  direction="l",
 },
 [0x16C5]={
  category="lo",
  direction="l",
 },
 [0x16C6]={
  category="lo",
  direction="l",
 },
 [0x16C7]={
  category="lo",
  direction="l",
 },
 [0x16C8]={
  category="lo",
  direction="l",
 },
 [0x16C9]={
  category="lo",
  direction="l",
 },
 [0x16CA]={
  category="lo",
  direction="l",
 },
 [0x16CB]={
  category="lo",
  direction="l",
 },
 [0x16CC]={
  category="lo",
  direction="l",
 },
 [0x16CD]={
  category="lo",
  direction="l",
 },
 [0x16CE]={
  category="lo",
  direction="l",
 },
 [0x16CF]={
  category="lo",
  direction="l",
 },
 [0x16D0]={
  category="lo",
  direction="l",
 },
 [0x16D1]={
  category="lo",
  direction="l",
 },
 [0x16D2]={
  category="lo",
  direction="l",
 },
 [0x16D3]={
  category="lo",
  direction="l",
 },
 [0x16D4]={
  category="lo",
  direction="l",
 },
 [0x16D5]={
  category="lo",
  direction="l",
 },
 [0x16D6]={
  category="lo",
  direction="l",
 },
 [0x16D7]={
  category="lo",
  direction="l",
 },
 [0x16D8]={
  category="lo",
  direction="l",
 },
 [0x16D9]={
  category="lo",
  direction="l",
 },
 [0x16DA]={
  category="lo",
  direction="l",
 },
 [0x16DB]={
  category="lo",
  direction="l",
 },
 [0x16DC]={
  category="lo",
  direction="l",
 },
 [0x16DD]={
  category="lo",
  direction="l",
 },
 [0x16DE]={
  category="lo",
  direction="l",
 },
 [0x16DF]={
  category="lo",
  direction="l",
 },
 [0x16E0]={
  category="lo",
  direction="l",
 },
 [0x16E1]={
  category="lo",
  direction="l",
 },
 [0x16E2]={
  category="lo",
  direction="l",
 },
 [0x16E3]={
  category="lo",
  direction="l",
 },
 [0x16E4]={
  category="lo",
  direction="l",
 },
 [0x16E5]={
  category="lo",
  direction="l",
 },
 [0x16E6]={
  category="lo",
  direction="l",
 },
 [0x16E7]={
  category="lo",
  direction="l",
 },
 [0x16E8]={
  category="lo",
  direction="l",
 },
 [0x16E9]={
  category="lo",
  direction="l",
 },
 [0x16EA]={
  category="lo",
  direction="l",
 },
 [0x16EB]={
  category="po",
  direction="l",
 },
 [0x16EC]={
  category="po",
  direction="l",
 },
 [0x16ED]={
  category="po",
  direction="l",
 },
 [0x16EE]={
  category="nl",
  direction="l",
 },
 [0x16EF]={
  category="nl",
  direction="l",
 },
 [0x16F0]={
  category="nl",
  direction="l",
 },
 [0x16F1]={
  category="lo",
  direction="l",
 },
 [0x16F2]={
  category="lo",
  direction="l",
 },
 [0x16F3]={
  category="lo",
  direction="l",
 },
 [0x16F4]={
  category="lo",
  direction="l",
 },
 [0x16F5]={
  category="lo",
  direction="l",
 },
 [0x16F6]={
  category="lo",
  direction="l",
 },
 [0x16F7]={
  category="lo",
  direction="l",
 },
 [0x16F8]={
  category="lo",
  direction="l",
 },
 [0x1700]={
  category="lo",
  direction="l",
 },
 [0x1701]={
  category="lo",
  direction="l",
 },
 [0x1702]={
  category="lo",
  direction="l",
 },
 [0x1703]={
  category="lo",
  direction="l",
 },
 [0x1704]={
  category="lo",
  direction="l",
 },
 [0x1705]={
  category="lo",
  direction="l",
 },
 [0x1706]={
  category="lo",
  direction="l",
 },
 [0x1707]={
  category="lo",
  direction="l",
 },
 [0x1708]={
  category="lo",
  direction="l",
 },
 [0x1709]={
  category="lo",
  direction="l",
 },
 [0x170A]={
  category="lo",
  direction="l",
 },
 [0x170B]={
  category="lo",
  direction="l",
 },
 [0x170C]={
  category="lo",
  direction="l",
 },
 [0x170E]={
  category="lo",
  direction="l",
 },
 [0x170F]={
  category="lo",
  direction="l",
 },
 [0x1710]={
  category="lo",
  direction="l",
 },
 [0x1711]={
  category="lo",
  direction="l",
 },
 [0x1712]={
  category="mn",
  direction="nsm",
 },
 [0x1713]={
  category="mn",
  direction="nsm",
 },
 [0x1714]={
  category="mn",
  direction="nsm",
 },
 [0x1720]={
  category="lo",
  direction="l",
 },
 [0x1721]={
  category="lo",
  direction="l",
 },
 [0x1722]={
  category="lo",
  direction="l",
 },
 [0x1723]={
  category="lo",
  direction="l",
 },
 [0x1724]={
  category="lo",
  direction="l",
 },
 [0x1725]={
  category="lo",
  direction="l",
 },
 [0x1726]={
  category="lo",
  direction="l",
 },
 [0x1727]={
  category="lo",
  direction="l",
 },
 [0x1728]={
  category="lo",
  direction="l",
 },
 [0x1729]={
  category="lo",
  direction="l",
 },
 [0x172A]={
  category="lo",
  direction="l",
 },
 [0x172B]={
  category="lo",
  direction="l",
 },
 [0x172C]={
  category="lo",
  direction="l",
 },
 [0x172D]={
  category="lo",
  direction="l",
 },
 [0x172E]={
  category="lo",
  direction="l",
 },
 [0x172F]={
  category="lo",
  direction="l",
 },
 [0x1730]={
  category="lo",
  direction="l",
 },
 [0x1731]={
  category="lo",
  direction="l",
 },
 [0x1732]={
  category="mn",
  direction="nsm",
 },
 [0x1733]={
  category="mn",
  direction="nsm",
 },
 [0x1734]={
  category="mn",
  direction="nsm",
 },
 [0x1735]={
  category="po",
  direction="l",
 },
 [0x1736]={
  category="po",
  direction="l",
 },
 [0x1740]={
  category="lo",
  direction="l",
 },
 [0x1741]={
  category="lo",
  direction="l",
 },
 [0x1742]={
  category="lo",
  direction="l",
 },
 [0x1743]={
  category="lo",
  direction="l",
 },
 [0x1744]={
  category="lo",
  direction="l",
 },
 [0x1745]={
  category="lo",
  direction="l",
 },
 [0x1746]={
  category="lo",
  direction="l",
 },
 [0x1747]={
  category="lo",
  direction="l",
 },
 [0x1748]={
  category="lo",
  direction="l",
 },
 [0x1749]={
  category="lo",
  direction="l",
 },
 [0x174A]={
  category="lo",
  direction="l",
 },
 [0x174B]={
  category="lo",
  direction="l",
 },
 [0x174C]={
  category="lo",
  direction="l",
 },
 [0x174D]={
  category="lo",
  direction="l",
 },
 [0x174E]={
  category="lo",
  direction="l",
 },
 [0x174F]={
  category="lo",
  direction="l",
 },
 [0x1750]={
  category="lo",
  direction="l",
 },
 [0x1751]={
  category="lo",
  direction="l",
 },
 [0x1752]={
  category="mn",
  direction="nsm",
 },
 [0x1753]={
  category="mn",
  direction="nsm",
 },
 [0x1760]={
  category="lo",
  direction="l",
 },
 [0x1761]={
  category="lo",
  direction="l",
 },
 [0x1762]={
  category="lo",
  direction="l",
 },
 [0x1763]={
  category="lo",
  direction="l",
 },
 [0x1764]={
  category="lo",
  direction="l",
 },
 [0x1765]={
  category="lo",
  direction="l",
 },
 [0x1766]={
  category="lo",
  direction="l",
 },
 [0x1767]={
  category="lo",
  direction="l",
 },
 [0x1768]={
  category="lo",
  direction="l",
 },
 [0x1769]={
  category="lo",
  direction="l",
 },
 [0x176A]={
  category="lo",
  direction="l",
 },
 [0x176B]={
  category="lo",
  direction="l",
 },
 [0x176C]={
  category="lo",
  direction="l",
 },
 [0x176E]={
  category="lo",
  direction="l",
 },
 [0x176F]={
  category="lo",
  direction="l",
 },
 [0x1770]={
  category="lo",
  direction="l",
 },
 [0x1772]={
  category="mn",
  direction="nsm",
 },
 [0x1773]={
  category="mn",
  direction="nsm",
 },
 [0x1780]={
  category="lo",
  direction="l",
 },
 [0x1781]={
  category="lo",
  direction="l",
 },
 [0x1782]={
  category="lo",
  direction="l",
 },
 [0x1783]={
  category="lo",
  direction="l",
 },
 [0x1784]={
  category="lo",
  direction="l",
 },
 [0x1785]={
  category="lo",
  direction="l",
 },
 [0x1786]={
  category="lo",
  direction="l",
 },
 [0x1787]={
  category="lo",
  direction="l",
 },
 [0x1788]={
  category="lo",
  direction="l",
 },
 [0x1789]={
  category="lo",
  direction="l",
 },
 [0x178A]={
  category="lo",
  direction="l",
 },
 [0x178B]={
  category="lo",
  direction="l",
 },
 [0x178C]={
  category="lo",
  direction="l",
 },
 [0x178D]={
  category="lo",
  direction="l",
 },
 [0x178E]={
  category="lo",
  direction="l",
 },
 [0x178F]={
  category="lo",
  direction="l",
 },
 [0x1790]={
  category="lo",
  direction="l",
 },
 [0x1791]={
  category="lo",
  direction="l",
 },
 [0x1792]={
  category="lo",
  direction="l",
 },
 [0x1793]={
  category="lo",
  direction="l",
 },
 [0x1794]={
  category="lo",
  direction="l",
 },
 [0x1795]={
  category="lo",
  direction="l",
 },
 [0x1796]={
  category="lo",
  direction="l",
 },
 [0x1797]={
  category="lo",
  direction="l",
 },
 [0x1798]={
  category="lo",
  direction="l",
 },
 [0x1799]={
  category="lo",
  direction="l",
 },
 [0x179A]={
  category="lo",
  direction="l",
 },
 [0x179B]={
  category="lo",
  direction="l",
 },
 [0x179C]={
  category="lo",
  direction="l",
 },
 [0x179D]={
  category="lo",
  direction="l",
 },
 [0x179E]={
  category="lo",
  direction="l",
 },
 [0x179F]={
  category="lo",
  direction="l",
 },
 [0x17A0]={
  category="lo",
  direction="l",
 },
 [0x17A1]={
  category="lo",
  direction="l",
 },
 [0x17A2]={
  category="lo",
  direction="l",
 },
 [0x17A3]={
  category="lo",
  direction="l",
 },
 [0x17A4]={
  category="lo",
  direction="l",
 },
 [0x17A5]={
  category="lo",
  direction="l",
 },
 [0x17A6]={
  category="lo",
  direction="l",
 },
 [0x17A7]={
  category="lo",
  direction="l",
 },
 [0x17A8]={
  category="lo",
  direction="l",
 },
 [0x17A9]={
  category="lo",
  direction="l",
 },
 [0x17AA]={
  category="lo",
  direction="l",
 },
 [0x17AB]={
  category="lo",
  direction="l",
 },
 [0x17AC]={
  category="lo",
  direction="l",
 },
 [0x17AD]={
  category="lo",
  direction="l",
 },
 [0x17AE]={
  category="lo",
  direction="l",
 },
 [0x17AF]={
  category="lo",
  direction="l",
 },
 [0x17B0]={
  category="lo",
  direction="l",
 },
 [0x17B1]={
  category="lo",
  direction="l",
 },
 [0x17B2]={
  category="lo",
  direction="l",
 },
 [0x17B3]={
  category="lo",
  direction="l",
 },
 [0x17B4]={
  category="cf",
  direction="nsm",
 },
 [0x17B5]={
  category="cf",
  direction="nsm",
 },
 [0x17B6]={
  category="mc",
  direction="l",
 },
 [0x17B7]={
  category="mn",
  direction="nsm",
 },
 [0x17B8]={
  category="mn",
  direction="nsm",
 },
 [0x17B9]={
  category="mn",
  direction="nsm",
 },
 [0x17BA]={
  category="mn",
  direction="nsm",
 },
 [0x17BB]={
  category="mn",
  direction="nsm",
 },
 [0x17BC]={
  category="mn",
  direction="nsm",
 },
 [0x17BD]={
  category="mn",
  direction="nsm",
 },
 [0x17BE]={
  category="mc",
  direction="l",
 },
 [0x17BF]={
  category="mc",
  direction="l",
 },
 [0x17C0]={
  category="mc",
  direction="l",
 },
 [0x17C1]={
  category="mc",
  direction="l",
 },
 [0x17C2]={
  category="mc",
  direction="l",
 },
 [0x17C3]={
  category="mc",
  direction="l",
 },
 [0x17C4]={
  category="mc",
  direction="l",
 },
 [0x17C5]={
  category="mc",
  direction="l",
 },
 [0x17C6]={
  category="mn",
  direction="nsm",
 },
 [0x17C7]={
  category="mc",
  direction="l",
 },
 [0x17C8]={
  category="mc",
  direction="l",
 },
 [0x17C9]={
  category="mn",
  direction="nsm",
 },
 [0x17CA]={
  category="mn",
  direction="nsm",
 },
 [0x17CB]={
  category="mn",
  direction="nsm",
 },
 [0x17CC]={
  category="mn",
  direction="nsm",
 },
 [0x17CD]={
  category="mn",
  direction="nsm",
 },
 [0x17CE]={
  category="mn",
  direction="nsm",
 },
 [0x17CF]={
  category="mn",
  direction="nsm",
 },
 [0x17D0]={
  category="mn",
  direction="nsm",
 },
 [0x17D1]={
  category="mn",
  direction="nsm",
 },
 [0x17D2]={
  category="mn",
  direction="nsm",
 },
 [0x17D3]={
  category="mn",
  direction="nsm",
 },
 [0x17D4]={
  category="po",
  direction="l",
 },
 [0x17D5]={
  category="po",
  direction="l",
 },
 [0x17D6]={
  category="po",
  direction="l",
 },
 [0x17D7]={
  category="lm",
  direction="l",
 },
 [0x17D8]={
  category="po",
  direction="l",
 },
 [0x17D9]={
  category="po",
  direction="l",
 },
 [0x17DA]={
  category="po",
  direction="l",
 },
 [0x17DB]={
  category="sc",
  direction="et",
 },
 [0x17DC]={
  category="lo",
  direction="l",
 },
 [0x17DD]={
  category="mn",
  direction="nsm",
 },
 [0x17E0]={
  category="nd",
  direction="l",
 },
 [0x17E1]={
  category="nd",
  direction="l",
 },
 [0x17E2]={
  category="nd",
  direction="l",
 },
 [0x17E3]={
  category="nd",
  direction="l",
 },
 [0x17E4]={
  category="nd",
  direction="l",
 },
 [0x17E5]={
  category="nd",
  direction="l",
 },
 [0x17E6]={
  category="nd",
  direction="l",
 },
 [0x17E7]={
  category="nd",
  direction="l",
 },
 [0x17E8]={
  category="nd",
  direction="l",
 },
 [0x17E9]={
  category="nd",
  direction="l",
 },
 [0x17F0]={
  category="no",
  direction="on",
 },
 [0x17F1]={
  category="no",
  direction="on",
 },
 [0x17F2]={
  category="no",
  direction="on",
 },
 [0x17F3]={
  category="no",
  direction="on",
 },
 [0x17F4]={
  category="no",
  direction="on",
 },
 [0x17F5]={
  category="no",
  direction="on",
 },
 [0x17F6]={
  category="no",
  direction="on",
 },
 [0x17F7]={
  category="no",
  direction="on",
 },
 [0x17F8]={
  category="no",
  direction="on",
 },
 [0x17F9]={
  category="no",
  direction="on",
 },
 [0x1800]={
  category="po",
  direction="on",
 },
 [0x1801]={
  category="po",
  direction="on",
 },
 [0x1802]={
  category="po",
  direction="on",
 },
 [0x1803]={
  category="po",
  direction="on",
 },
 [0x1804]={
  category="po",
  direction="on",
 },
 [0x1805]={
  category="po",
  direction="on",
 },
 [0x1806]={
  category="pd",
  direction="on",
 },
 [0x1807]={
  category="po",
  direction="on",
 },
 [0x1808]={
  category="po",
  direction="on",
 },
 [0x1809]={
  category="po",
  direction="on",
 },
 [0x180A]={
  category="po",
  direction="on",
 },
 [0x180B]={
  category="mn",
  direction="nsm",
 },
 [0x180C]={
  category="mn",
  direction="nsm",
 },
 [0x180D]={
  category="mn",
  direction="nsm",
 },
 [0x180E]={
  category="zs",
  direction="bn",
 },
 [0x1810]={
  category="nd",
  direction="l",
 },
 [0x1811]={
  category="nd",
  direction="l",
 },
 [0x1812]={
  category="nd",
  direction="l",
 },
 [0x1813]={
  category="nd",
  direction="l",
 },
 [0x1814]={
  category="nd",
  direction="l",
 },
 [0x1815]={
  category="nd",
  direction="l",
 },
 [0x1816]={
  category="nd",
  direction="l",
 },
 [0x1817]={
  category="nd",
  direction="l",
 },
 [0x1818]={
  category="nd",
  direction="l",
 },
 [0x1819]={
  category="nd",
  direction="l",
 },
 [0x1820]={
  category="lo",
  direction="l",
 },
 [0x1821]={
  category="lo",
  direction="l",
 },
 [0x1822]={
  category="lo",
  direction="l",
 },
 [0x1823]={
  category="lo",
  direction="l",
 },
 [0x1824]={
  category="lo",
  direction="l",
 },
 [0x1825]={
  category="lo",
  direction="l",
 },
 [0x1826]={
  category="lo",
  direction="l",
 },
 [0x1827]={
  category="lo",
  direction="l",
 },
 [0x1828]={
  category="lo",
  direction="l",
 },
 [0x1829]={
  category="lo",
  direction="l",
 },
 [0x182A]={
  category="lo",
  direction="l",
 },
 [0x182B]={
  category="lo",
  direction="l",
 },
 [0x182C]={
  category="lo",
  direction="l",
 },
 [0x182D]={
  category="lo",
  direction="l",
 },
 [0x182E]={
  category="lo",
  direction="l",
 },
 [0x182F]={
  category="lo",
  direction="l",
 },
 [0x1830]={
  category="lo",
  direction="l",
 },
 [0x1831]={
  category="lo",
  direction="l",
 },
 [0x1832]={
  category="lo",
  direction="l",
 },
 [0x1833]={
  category="lo",
  direction="l",
 },
 [0x1834]={
  category="lo",
  direction="l",
 },
 [0x1835]={
  category="lo",
  direction="l",
 },
 [0x1836]={
  category="lo",
  direction="l",
 },
 [0x1837]={
  category="lo",
  direction="l",
 },
 [0x1838]={
  category="lo",
  direction="l",
 },
 [0x1839]={
  category="lo",
  direction="l",
 },
 [0x183A]={
  category="lo",
  direction="l",
 },
 [0x183B]={
  category="lo",
  direction="l",
 },
 [0x183C]={
  category="lo",
  direction="l",
 },
 [0x183D]={
  category="lo",
  direction="l",
 },
 [0x183E]={
  category="lo",
  direction="l",
 },
 [0x183F]={
  category="lo",
  direction="l",
 },
 [0x1840]={
  category="lo",
  direction="l",
 },
 [0x1841]={
  category="lo",
  direction="l",
 },
 [0x1842]={
  category="lo",
  direction="l",
 },
 [0x1843]={
  category="lm",
  direction="l",
 },
 [0x1844]={
  category="lo",
  direction="l",
 },
 [0x1845]={
  category="lo",
  direction="l",
 },
 [0x1846]={
  category="lo",
  direction="l",
 },
 [0x1847]={
  category="lo",
  direction="l",
 },
 [0x1848]={
  category="lo",
  direction="l",
 },
 [0x1849]={
  category="lo",
  direction="l",
 },
 [0x184A]={
  category="lo",
  direction="l",
 },
 [0x184B]={
  category="lo",
  direction="l",
 },
 [0x184C]={
  category="lo",
  direction="l",
 },
 [0x184D]={
  category="lo",
  direction="l",
 },
 [0x184E]={
  category="lo",
  direction="l",
 },
 [0x184F]={
  category="lo",
  direction="l",
 },
 [0x1850]={
  category="lo",
  direction="l",
 },
 [0x1851]={
  category="lo",
  direction="l",
 },
 [0x1852]={
  category="lo",
  direction="l",
 },
 [0x1853]={
  category="lo",
  direction="l",
 },
 [0x1854]={
  category="lo",
  direction="l",
 },
 [0x1855]={
  category="lo",
  direction="l",
 },
 [0x1856]={
  category="lo",
  direction="l",
 },
 [0x1857]={
  category="lo",
  direction="l",
 },
 [0x1858]={
  category="lo",
  direction="l",
 },
 [0x1859]={
  category="lo",
  direction="l",
 },
 [0x185A]={
  category="lo",
  direction="l",
 },
 [0x185B]={
  category="lo",
  direction="l",
 },
 [0x185C]={
  category="lo",
  direction="l",
 },
 [0x185D]={
  category="lo",
  direction="l",
 },
 [0x185E]={
  category="lo",
  direction="l",
 },
 [0x185F]={
  category="lo",
  direction="l",
 },
 [0x1860]={
  category="lo",
  direction="l",
 },
 [0x1861]={
  category="lo",
  direction="l",
 },
 [0x1862]={
  category="lo",
  direction="l",
 },
 [0x1863]={
  category="lo",
  direction="l",
 },
 [0x1864]={
  category="lo",
  direction="l",
 },
 [0x1865]={
  category="lo",
  direction="l",
 },
 [0x1866]={
  category="lo",
  direction="l",
 },
 [0x1867]={
  category="lo",
  direction="l",
 },
 [0x1868]={
  category="lo",
  direction="l",
 },
 [0x1869]={
  category="lo",
  direction="l",
 },
 [0x186A]={
  category="lo",
  direction="l",
 },
 [0x186B]={
  category="lo",
  direction="l",
 },
 [0x186C]={
  category="lo",
  direction="l",
 },
 [0x186D]={
  category="lo",
  direction="l",
 },
 [0x186E]={
  category="lo",
  direction="l",
 },
 [0x186F]={
  category="lo",
  direction="l",
 },
 [0x1870]={
  category="lo",
  direction="l",
 },
 [0x1871]={
  category="lo",
  direction="l",
 },
 [0x1872]={
  category="lo",
  direction="l",
 },
 [0x1873]={
  category="lo",
  direction="l",
 },
 [0x1874]={
  category="lo",
  direction="l",
 },
 [0x1875]={
  category="lo",
  direction="l",
 },
 [0x1876]={
  category="lo",
  direction="l",
 },
 [0x1877]={
  category="lo",
  direction="l",
 },
 [0x1880]={
  category="lo",
  direction="l",
 },
 [0x1881]={
  category="lo",
  direction="l",
 },
 [0x1882]={
  category="lo",
  direction="l",
 },
 [0x1883]={
  category="lo",
  direction="l",
 },
 [0x1884]={
  category="lo",
  direction="l",
 },
 [0x1885]={
  category="lo",
  direction="l",
 },
 [0x1886]={
  category="lo",
  direction="l",
 },
 [0x1887]={
  category="lo",
  direction="l",
 },
 [0x1888]={
  category="lo",
  direction="l",
 },
 [0x1889]={
  category="lo",
  direction="l",
 },
 [0x188A]={
  category="lo",
  direction="l",
 },
 [0x188B]={
  category="lo",
  direction="l",
 },
 [0x188C]={
  category="lo",
  direction="l",
 },
 [0x188D]={
  category="lo",
  direction="l",
 },
 [0x188E]={
  category="lo",
  direction="l",
 },
 [0x188F]={
  category="lo",
  direction="l",
 },
 [0x1890]={
  category="lo",
  direction="l",
 },
 [0x1891]={
  category="lo",
  direction="l",
 },
 [0x1892]={
  category="lo",
  direction="l",
 },
 [0x1893]={
  category="lo",
  direction="l",
 },
 [0x1894]={
  category="lo",
  direction="l",
 },
 [0x1895]={
  category="lo",
  direction="l",
 },
 [0x1896]={
  category="lo",
  direction="l",
 },
 [0x1897]={
  category="lo",
  direction="l",
 },
 [0x1898]={
  category="lo",
  direction="l",
 },
 [0x1899]={
  category="lo",
  direction="l",
 },
 [0x189A]={
  category="lo",
  direction="l",
 },
 [0x189B]={
  category="lo",
  direction="l",
 },
 [0x189C]={
  category="lo",
  direction="l",
 },
 [0x189D]={
  category="lo",
  direction="l",
 },
 [0x189E]={
  category="lo",
  direction="l",
 },
 [0x189F]={
  category="lo",
  direction="l",
 },
 [0x18A0]={
  category="lo",
  direction="l",
 },
 [0x18A1]={
  category="lo",
  direction="l",
 },
 [0x18A2]={
  category="lo",
  direction="l",
 },
 [0x18A3]={
  category="lo",
  direction="l",
 },
 [0x18A4]={
  category="lo",
  direction="l",
 },
 [0x18A5]={
  category="lo",
  direction="l",
 },
 [0x18A6]={
  category="lo",
  direction="l",
 },
 [0x18A7]={
  category="lo",
  direction="l",
 },
 [0x18A8]={
  category="lo",
  direction="l",
 },
 [0x18A9]={
  category="mn",
  direction="nsm",
 },
 [0x18AA]={
  category="lo",
  direction="l",
 },
 [0x18B0]={
  category="lo",
  direction="l",
 },
 [0x18B1]={
  category="lo",
  direction="l",
 },
 [0x18B2]={
  category="lo",
  direction="l",
 },
 [0x18B3]={
  category="lo",
  direction="l",
 },
 [0x18B4]={
  category="lo",
  direction="l",
 },
 [0x18B5]={
  category="lo",
  direction="l",
 },
 [0x18B6]={
  category="lo",
  direction="l",
 },
 [0x18B7]={
  category="lo",
  direction="l",
 },
 [0x18B8]={
  category="lo",
  direction="l",
 },
 [0x18B9]={
  category="lo",
  direction="l",
 },
 [0x18BA]={
  category="lo",
  direction="l",
 },
 [0x18BB]={
  category="lo",
  direction="l",
 },
 [0x18BC]={
  category="lo",
  direction="l",
 },
 [0x18BD]={
  category="lo",
  direction="l",
 },
 [0x18BE]={
  category="lo",
  direction="l",
 },
 [0x18BF]={
  category="lo",
  direction="l",
 },
 [0x18C0]={
  category="lo",
  direction="l",
 },
 [0x18C1]={
  category="lo",
  direction="l",
 },
 [0x18C2]={
  category="lo",
  direction="l",
 },
 [0x18C3]={
  category="lo",
  direction="l",
 },
 [0x18C4]={
  category="lo",
  direction="l",
 },
 [0x18C5]={
  category="lo",
  direction="l",
 },
 [0x18C6]={
  category="lo",
  direction="l",
 },
 [0x18C7]={
  category="lo",
  direction="l",
 },
 [0x18C8]={
  category="lo",
  direction="l",
 },
 [0x18C9]={
  category="lo",
  direction="l",
 },
 [0x18CA]={
  category="lo",
  direction="l",
 },
 [0x18CB]={
  category="lo",
  direction="l",
 },
 [0x18CC]={
  category="lo",
  direction="l",
 },
 [0x18CD]={
  category="lo",
  direction="l",
 },
 [0x18CE]={
  category="lo",
  direction="l",
 },
 [0x18CF]={
  category="lo",
  direction="l",
 },
 [0x18D0]={
  category="lo",
  direction="l",
 },
 [0x18D1]={
  category="lo",
  direction="l",
 },
 [0x18D2]={
  category="lo",
  direction="l",
 },
 [0x18D3]={
  category="lo",
  direction="l",
 },
 [0x18D4]={
  category="lo",
  direction="l",
 },
 [0x18D5]={
  category="lo",
  direction="l",
 },
 [0x18D6]={
  category="lo",
  direction="l",
 },
 [0x18D7]={
  category="lo",
  direction="l",
 },
 [0x18D8]={
  category="lo",
  direction="l",
 },
 [0x18D9]={
  category="lo",
  direction="l",
 },
 [0x18DA]={
  category="lo",
  direction="l",
 },
 [0x18DB]={
  category="lo",
  direction="l",
 },
 [0x18DC]={
  category="lo",
  direction="l",
 },
 [0x18DD]={
  category="lo",
  direction="l",
 },
 [0x18DE]={
  category="lo",
  direction="l",
 },
 [0x18DF]={
  category="lo",
  direction="l",
 },
 [0x18E0]={
  category="lo",
  direction="l",
 },
 [0x18E1]={
  category="lo",
  direction="l",
 },
 [0x18E2]={
  category="lo",
  direction="l",
 },
 [0x18E3]={
  category="lo",
  direction="l",
 },
 [0x18E4]={
  category="lo",
  direction="l",
 },
 [0x18E5]={
  category="lo",
  direction="l",
 },
 [0x18E6]={
  category="lo",
  direction="l",
 },
 [0x18E7]={
  category="lo",
  direction="l",
 },
 [0x18E8]={
  category="lo",
  direction="l",
 },
 [0x18E9]={
  category="lo",
  direction="l",
 },
 [0x18EA]={
  category="lo",
  direction="l",
 },
 [0x18EB]={
  category="lo",
  direction="l",
 },
 [0x18EC]={
  category="lo",
  direction="l",
 },
 [0x18ED]={
  category="lo",
  direction="l",
 },
 [0x18EE]={
  category="lo",
  direction="l",
 },
 [0x18EF]={
  category="lo",
  direction="l",
 },
 [0x18F0]={
  category="lo",
  direction="l",
 },
 [0x18F1]={
  category="lo",
  direction="l",
 },
 [0x18F2]={
  category="lo",
  direction="l",
 },
 [0x18F3]={
  category="lo",
  direction="l",
 },
 [0x18F4]={
  category="lo",
  direction="l",
 },
 [0x18F5]={
  category="lo",
  direction="l",
 },
 [0x1900]={
  category="lo",
  direction="l",
 },
 [0x1901]={
  category="lo",
  direction="l",
 },
 [0x1902]={
  category="lo",
  direction="l",
 },
 [0x1903]={
  category="lo",
  direction="l",
 },
 [0x1904]={
  category="lo",
  direction="l",
 },
 [0x1905]={
  category="lo",
  direction="l",
 },
 [0x1906]={
  category="lo",
  direction="l",
 },
 [0x1907]={
  category="lo",
  direction="l",
 },
 [0x1908]={
  category="lo",
  direction="l",
 },
 [0x1909]={
  category="lo",
  direction="l",
 },
 [0x190A]={
  category="lo",
  direction="l",
 },
 [0x190B]={
  category="lo",
  direction="l",
 },
 [0x190C]={
  category="lo",
  direction="l",
 },
 [0x190D]={
  category="lo",
  direction="l",
 },
 [0x190E]={
  category="lo",
  direction="l",
 },
 [0x190F]={
  category="lo",
  direction="l",
 },
 [0x1910]={
  category="lo",
  direction="l",
 },
 [0x1911]={
  category="lo",
  direction="l",
 },
 [0x1912]={
  category="lo",
  direction="l",
 },
 [0x1913]={
  category="lo",
  direction="l",
 },
 [0x1914]={
  category="lo",
  direction="l",
 },
 [0x1915]={
  category="lo",
  direction="l",
 },
 [0x1916]={
  category="lo",
  direction="l",
 },
 [0x1917]={
  category="lo",
  direction="l",
 },
 [0x1918]={
  category="lo",
  direction="l",
 },
 [0x1919]={
  category="lo",
  direction="l",
 },
 [0x191A]={
  category="lo",
  direction="l",
 },
 [0x191B]={
  category="lo",
  direction="l",
 },
 [0x191C]={
  category="lo",
  direction="l",
 },
 [0x191D]={
  category="lo",
  direction="l",
 },
 [0x191E]={
  category="lo",
  direction="l",
 },
 [0x1920]={
  category="mn",
  direction="nsm",
 },
 [0x1921]={
  category="mn",
  direction="nsm",
 },
 [0x1922]={
  category="mn",
  direction="nsm",
 },
 [0x1923]={
  category="mc",
  direction="l",
 },
 [0x1924]={
  category="mc",
  direction="l",
 },
 [0x1925]={
  category="mc",
  direction="l",
 },
 [0x1926]={
  category="mc",
  direction="l",
 },
 [0x1927]={
  category="mn",
  direction="nsm",
 },
 [0x1928]={
  category="mn",
  direction="nsm",
 },
 [0x1929]={
  category="mc",
  direction="l",
 },
 [0x192A]={
  category="mc",
  direction="l",
 },
 [0x192B]={
  category="mc",
  direction="l",
 },
 [0x1930]={
  category="mc",
  direction="l",
 },
 [0x1931]={
  category="mc",
  direction="l",
 },
 [0x1932]={
  category="mn",
  direction="nsm",
 },
 [0x1933]={
  category="mc",
  direction="l",
 },
 [0x1934]={
  category="mc",
  direction="l",
 },
 [0x1935]={
  category="mc",
  direction="l",
 },
 [0x1936]={
  category="mc",
  direction="l",
 },
 [0x1937]={
  category="mc",
  direction="l",
 },
 [0x1938]={
  category="mc",
  direction="l",
 },
 [0x1939]={
  category="mn",
  direction="nsm",
 },
 [0x193A]={
  category="mn",
  direction="nsm",
 },
 [0x193B]={
  category="mn",
  direction="nsm",
 },
 [0x1940]={
  category="so",
  direction="on",
 },
 [0x1944]={
  category="po",
  direction="on",
 },
 [0x1945]={
  category="po",
  direction="on",
 },
 [0x1946]={
  category="nd",
  direction="l",
 },
 [0x1947]={
  category="nd",
  direction="l",
 },
 [0x1948]={
  category="nd",
  direction="l",
 },
 [0x1949]={
  category="nd",
  direction="l",
 },
 [0x194A]={
  category="nd",
  direction="l",
 },
 [0x194B]={
  category="nd",
  direction="l",
 },
 [0x194C]={
  category="nd",
  direction="l",
 },
 [0x194D]={
  category="nd",
  direction="l",
 },
 [0x194E]={
  category="nd",
  direction="l",
 },
 [0x194F]={
  category="nd",
  direction="l",
 },
 [0x1950]={
  category="lo",
  direction="l",
 },
 [0x1951]={
  category="lo",
  direction="l",
 },
 [0x1952]={
  category="lo",
  direction="l",
 },
 [0x1953]={
  category="lo",
  direction="l",
 },
 [0x1954]={
  category="lo",
  direction="l",
 },
 [0x1955]={
  category="lo",
  direction="l",
 },
 [0x1956]={
  category="lo",
  direction="l",
 },
 [0x1957]={
  category="lo",
  direction="l",
 },
 [0x1958]={
  category="lo",
  direction="l",
 },
 [0x1959]={
  category="lo",
  direction="l",
 },
 [0x195A]={
  category="lo",
  direction="l",
 },
 [0x195B]={
  category="lo",
  direction="l",
 },
 [0x195C]={
  category="lo",
  direction="l",
 },
 [0x195D]={
  category="lo",
  direction="l",
 },
 [0x195E]={
  category="lo",
  direction="l",
 },
 [0x195F]={
  category="lo",
  direction="l",
 },
 [0x1960]={
  category="lo",
  direction="l",
 },
 [0x1961]={
  category="lo",
  direction="l",
 },
 [0x1962]={
  category="lo",
  direction="l",
 },
 [0x1963]={
  category="lo",
  direction="l",
 },
 [0x1964]={
  category="lo",
  direction="l",
 },
 [0x1965]={
  category="lo",
  direction="l",
 },
 [0x1966]={
  category="lo",
  direction="l",
 },
 [0x1967]={
  category="lo",
  direction="l",
 },
 [0x1968]={
  category="lo",
  direction="l",
 },
 [0x1969]={
  category="lo",
  direction="l",
 },
 [0x196A]={
  category="lo",
  direction="l",
 },
 [0x196B]={
  category="lo",
  direction="l",
 },
 [0x196C]={
  category="lo",
  direction="l",
 },
 [0x196D]={
  category="lo",
  direction="l",
 },
 [0x1970]={
  category="lo",
  direction="l",
 },
 [0x1971]={
  category="lo",
  direction="l",
 },
 [0x1972]={
  category="lo",
  direction="l",
 },
 [0x1973]={
  category="lo",
  direction="l",
 },
 [0x1974]={
  category="lo",
  direction="l",
 },
 [0x1980]={
  category="lo",
  direction="l",
 },
 [0x1981]={
  category="lo",
  direction="l",
 },
 [0x1982]={
  category="lo",
  direction="l",
 },
 [0x1983]={
  category="lo",
  direction="l",
 },
 [0x1984]={
  category="lo",
  direction="l",
 },
 [0x1985]={
  category="lo",
  direction="l",
 },
 [0x1986]={
  category="lo",
  direction="l",
 },
 [0x1987]={
  category="lo",
  direction="l",
 },
 [0x1988]={
  category="lo",
  direction="l",
 },
 [0x1989]={
  category="lo",
  direction="l",
 },
 [0x198A]={
  category="lo",
  direction="l",
 },
 [0x198B]={
  category="lo",
  direction="l",
 },
 [0x198C]={
  category="lo",
  direction="l",
 },
 [0x198D]={
  category="lo",
  direction="l",
 },
 [0x198E]={
  category="lo",
  direction="l",
 },
 [0x198F]={
  category="lo",
  direction="l",
 },
 [0x1990]={
  category="lo",
  direction="l",
 },
 [0x1991]={
  category="lo",
  direction="l",
 },
 [0x1992]={
  category="lo",
  direction="l",
 },
 [0x1993]={
  category="lo",
  direction="l",
 },
 [0x1994]={
  category="lo",
  direction="l",
 },
 [0x1995]={
  category="lo",
  direction="l",
 },
 [0x1996]={
  category="lo",
  direction="l",
 },
 [0x1997]={
  category="lo",
  direction="l",
 },
 [0x1998]={
  category="lo",
  direction="l",
 },
 [0x1999]={
  category="lo",
  direction="l",
 },
 [0x199A]={
  category="lo",
  direction="l",
 },
 [0x199B]={
  category="lo",
  direction="l",
 },
 [0x199C]={
  category="lo",
  direction="l",
 },
 [0x199D]={
  category="lo",
  direction="l",
 },
 [0x199E]={
  category="lo",
  direction="l",
 },
 [0x199F]={
  category="lo",
  direction="l",
 },
 [0x19A0]={
  category="lo",
  direction="l",
 },
 [0x19A1]={
  category="lo",
  direction="l",
 },
 [0x19A2]={
  category="lo",
  direction="l",
 },
 [0x19A3]={
  category="lo",
  direction="l",
 },
 [0x19A4]={
  category="lo",
  direction="l",
 },
 [0x19A5]={
  category="lo",
  direction="l",
 },
 [0x19A6]={
  category="lo",
  direction="l",
 },
 [0x19A7]={
  category="lo",
  direction="l",
 },
 [0x19A8]={
  category="lo",
  direction="l",
 },
 [0x19A9]={
  category="lo",
  direction="l",
 },
 [0x19AA]={
  category="lo",
  direction="l",
 },
 [0x19AB]={
  category="lo",
  direction="l",
 },
 [0x19B0]={
  category="mc",
  direction="l",
 },
 [0x19B1]={
  category="mc",
  direction="l",
 },
 [0x19B2]={
  category="mc",
  direction="l",
 },
 [0x19B3]={
  category="mc",
  direction="l",
 },
 [0x19B4]={
  category="mc",
  direction="l",
 },
 [0x19B5]={
  category="mc",
  direction="l",
 },
 [0x19B6]={
  category="mc",
  direction="l",
 },
 [0x19B7]={
  category="mc",
  direction="l",
 },
 [0x19B8]={
  category="mc",
  direction="l",
 },
 [0x19B9]={
  category="mc",
  direction="l",
 },
 [0x19BA]={
  category="mc",
  direction="l",
 },
 [0x19BB]={
  category="mc",
  direction="l",
 },
 [0x19BC]={
  category="mc",
  direction="l",
 },
 [0x19BD]={
  category="mc",
  direction="l",
 },
 [0x19BE]={
  category="mc",
  direction="l",
 },
 [0x19BF]={
  category="mc",
  direction="l",
 },
 [0x19C0]={
  category="mc",
  direction="l",
 },
 [0x19C1]={
  category="lo",
  direction="l",
 },
 [0x19C2]={
  category="lo",
  direction="l",
 },
 [0x19C3]={
  category="lo",
  direction="l",
 },
 [0x19C4]={
  category="lo",
  direction="l",
 },
 [0x19C5]={
  category="lo",
  direction="l",
 },
 [0x19C6]={
  category="lo",
  direction="l",
 },
 [0x19C7]={
  category="lo",
  direction="l",
 },
 [0x19C8]={
  category="mc",
  direction="l",
 },
 [0x19C9]={
  category="mc",
  direction="l",
 },
 [0x19D0]={
  category="nd",
  direction="l",
 },
 [0x19D1]={
  category="nd",
  direction="l",
 },
 [0x19D2]={
  category="nd",
  direction="l",
 },
 [0x19D3]={
  category="nd",
  direction="l",
 },
 [0x19D4]={
  category="nd",
  direction="l",
 },
 [0x19D5]={
  category="nd",
  direction="l",
 },
 [0x19D6]={
  category="nd",
  direction="l",
 },
 [0x19D7]={
  category="nd",
  direction="l",
 },
 [0x19D8]={
  category="nd",
  direction="l",
 },
 [0x19D9]={
  category="nd",
  direction="l",
 },
 [0x19DA]={
  category="no",
  direction="l",
 },
 [0x19DE]={
  category="po",
  direction="on",
 },
 [0x19DF]={
  category="po",
  direction="on",
 },
 [0x19E0]={
  category="so",
  direction="on",
 },
 [0x19E1]={
  category="so",
  direction="on",
 },
 [0x19E2]={
  category="so",
  direction="on",
 },
 [0x19E3]={
  category="so",
  direction="on",
 },
 [0x19E4]={
  category="so",
  direction="on",
 },
 [0x19E5]={
  category="so",
  direction="on",
 },
 [0x19E6]={
  category="so",
  direction="on",
 },
 [0x19E7]={
  category="so",
  direction="on",
 },
 [0x19E8]={
  category="so",
  direction="on",
 },
 [0x19E9]={
  category="so",
  direction="on",
 },
 [0x19EA]={
  category="so",
  direction="on",
 },
 [0x19EB]={
  category="so",
  direction="on",
 },
 [0x19EC]={
  category="so",
  direction="on",
 },
 [0x19ED]={
  category="so",
  direction="on",
 },
 [0x19EE]={
  category="so",
  direction="on",
 },
 [0x19EF]={
  category="so",
  direction="on",
 },
 [0x19F0]={
  category="so",
  direction="on",
 },
 [0x19F1]={
  category="so",
  direction="on",
 },
 [0x19F2]={
  category="so",
  direction="on",
 },
 [0x19F3]={
  category="so",
  direction="on",
 },
 [0x19F4]={
  category="so",
  direction="on",
 },
 [0x19F5]={
  category="so",
  direction="on",
 },
 [0x19F6]={
  category="so",
  direction="on",
 },
 [0x19F7]={
  category="so",
  direction="on",
 },
 [0x19F8]={
  category="so",
  direction="on",
 },
 [0x19F9]={
  category="so",
  direction="on",
 },
 [0x19FA]={
  category="so",
  direction="on",
 },
 [0x19FB]={
  category="so",
  direction="on",
 },
 [0x19FC]={
  category="so",
  direction="on",
 },
 [0x19FD]={
  category="so",
  direction="on",
 },
 [0x19FE]={
  category="so",
  direction="on",
 },
 [0x19FF]={
  category="so",
  direction="on",
 },
 [0x1A00]={
  category="lo",
  direction="l",
 },
 [0x1A01]={
  category="lo",
  direction="l",
 },
 [0x1A02]={
  category="lo",
  direction="l",
 },
 [0x1A03]={
  category="lo",
  direction="l",
 },
 [0x1A04]={
  category="lo",
  direction="l",
 },
 [0x1A05]={
  category="lo",
  direction="l",
 },
 [0x1A06]={
  category="lo",
  direction="l",
 },
 [0x1A07]={
  category="lo",
  direction="l",
 },
 [0x1A08]={
  category="lo",
  direction="l",
 },
 [0x1A09]={
  category="lo",
  direction="l",
 },
 [0x1A0A]={
  category="lo",
  direction="l",
 },
 [0x1A0B]={
  category="lo",
  direction="l",
 },
 [0x1A0C]={
  category="lo",
  direction="l",
 },
 [0x1A0D]={
  category="lo",
  direction="l",
 },
 [0x1A0E]={
  category="lo",
  direction="l",
 },
 [0x1A0F]={
  category="lo",
  direction="l",
 },
 [0x1A10]={
  category="lo",
  direction="l",
 },
 [0x1A11]={
  category="lo",
  direction="l",
 },
 [0x1A12]={
  category="lo",
  direction="l",
 },
 [0x1A13]={
  category="lo",
  direction="l",
 },
 [0x1A14]={
  category="lo",
  direction="l",
 },
 [0x1A15]={
  category="lo",
  direction="l",
 },
 [0x1A16]={
  category="lo",
  direction="l",
 },
 [0x1A17]={
  category="mn",
  direction="nsm",
 },
 [0x1A18]={
  category="mn",
  direction="nsm",
 },
 [0x1A19]={
  category="mc",
  direction="l",
 },
 [0x1A1A]={
  category="mc",
  direction="l",
 },
 [0x1A1B]={
  category="mc",
  direction="nsm",
 },
 [0x1A1E]={
  category="po",
  direction="l",
 },
 [0x1A1F]={
  category="po",
  direction="l",
 },
 [0x1A20]={
  category="lo",
  direction="l",
 },
 [0x1A21]={
  category="lo",
  direction="l",
 },
 [0x1A22]={
  category="lo",
  direction="l",
 },
 [0x1A23]={
  category="lo",
  direction="l",
 },
 [0x1A24]={
  category="lo",
  direction="l",
 },
 [0x1A25]={
  category="lo",
  direction="l",
 },
 [0x1A26]={
  category="lo",
  direction="l",
 },
 [0x1A27]={
  category="lo",
  direction="l",
 },
 [0x1A28]={
  category="lo",
  direction="l",
 },
 [0x1A29]={
  category="lo",
  direction="l",
 },
 [0x1A2A]={
  category="lo",
  direction="l",
 },
 [0x1A2B]={
  category="lo",
  direction="l",
 },
 [0x1A2C]={
  category="lo",
  direction="l",
 },
 [0x1A2D]={
  category="lo",
  direction="l",
 },
 [0x1A2E]={
  category="lo",
  direction="l",
 },
 [0x1A2F]={
  category="lo",
  direction="l",
 },
 [0x1A30]={
  category="lo",
  direction="l",
 },
 [0x1A31]={
  category="lo",
  direction="l",
 },
 [0x1A32]={
  category="lo",
  direction="l",
 },
 [0x1A33]={
  category="lo",
  direction="l",
 },
 [0x1A34]={
  category="lo",
  direction="l",
 },
 [0x1A35]={
  category="lo",
  direction="l",
 },
 [0x1A36]={
  category="lo",
  direction="l",
 },
 [0x1A37]={
  category="lo",
  direction="l",
 },
 [0x1A38]={
  category="lo",
  direction="l",
 },
 [0x1A39]={
  category="lo",
  direction="l",
 },
 [0x1A3A]={
  category="lo",
  direction="l",
 },
 [0x1A3B]={
  category="lo",
  direction="l",
 },
 [0x1A3C]={
  category="lo",
  direction="l",
 },
 [0x1A3D]={
  category="lo",
  direction="l",
 },
 [0x1A3E]={
  category="lo",
  direction="l",
 },
 [0x1A3F]={
  category="lo",
  direction="l",
 },
 [0x1A40]={
  category="lo",
  direction="l",
 },
 [0x1A41]={
  category="lo",
  direction="l",
 },
 [0x1A42]={
  category="lo",
  direction="l",
 },
 [0x1A43]={
  category="lo",
  direction="l",
 },
 [0x1A44]={
  category="lo",
  direction="l",
 },
 [0x1A45]={
  category="lo",
  direction="l",
 },
 [0x1A46]={
  category="lo",
  direction="l",
 },
 [0x1A47]={
  category="lo",
  direction="l",
 },
 [0x1A48]={
  category="lo",
  direction="l",
 },
 [0x1A49]={
  category="lo",
  direction="l",
 },
 [0x1A4A]={
  category="lo",
  direction="l",
 },
 [0x1A4B]={
  category="lo",
  direction="l",
 },
 [0x1A4C]={
  category="lo",
  direction="l",
 },
 [0x1A4D]={
  category="lo",
  direction="l",
 },
 [0x1A4E]={
  category="lo",
  direction="l",
 },
 [0x1A4F]={
  category="lo",
  direction="l",
 },
 [0x1A50]={
  category="lo",
  direction="l",
 },
 [0x1A51]={
  category="lo",
  direction="l",
 },
 [0x1A52]={
  category="lo",
  direction="l",
 },
 [0x1A53]={
  category="lo",
  direction="l",
 },
 [0x1A54]={
  category="lo",
  direction="l",
 },
 [0x1A55]={
  category="mc",
  direction="l",
 },
 [0x1A56]={
  category="mn",
  direction="nsm",
 },
 [0x1A57]={
  category="mc",
  direction="l",
 },
 [0x1A58]={
  category="mn",
  direction="nsm",
 },
 [0x1A59]={
  category="mn",
  direction="nsm",
 },
 [0x1A5A]={
  category="mn",
  direction="nsm",
 },
 [0x1A5B]={
  category="mn",
  direction="nsm",
 },
 [0x1A5C]={
  category="mn",
  direction="nsm",
 },
 [0x1A5D]={
  category="mn",
  direction="nsm",
 },
 [0x1A5E]={
  category="mn",
  direction="nsm",
 },
 [0x1A60]={
  category="mn",
  direction="nsm",
 },
 [0x1A61]={
  category="mc",
  direction="l",
 },
 [0x1A62]={
  category="mn",
  direction="nsm",
 },
 [0x1A63]={
  category="mc",
  direction="l",
 },
 [0x1A64]={
  category="mc",
  direction="l",
 },
 [0x1A65]={
  category="mn",
  direction="nsm",
 },
 [0x1A66]={
  category="mn",
  direction="nsm",
 },
 [0x1A67]={
  category="mn",
  direction="nsm",
 },
 [0x1A68]={
  category="mn",
  direction="nsm",
 },
 [0x1A69]={
  category="mn",
  direction="nsm",
 },
 [0x1A6A]={
  category="mn",
  direction="nsm",
 },
 [0x1A6B]={
  category="mn",
  direction="nsm",
 },
 [0x1A6C]={
  category="mn",
  direction="nsm",
 },
 [0x1A6D]={
  category="mc",
  direction="l",
 },
 [0x1A6E]={
  category="mc",
  direction="l",
 },
 [0x1A6F]={
  category="mc",
  direction="l",
 },
 [0x1A70]={
  category="mc",
  direction="l",
 },
 [0x1A71]={
  category="mc",
  direction="l",
 },
 [0x1A72]={
  category="mc",
  direction="l",
 },
 [0x1A73]={
  category="mn",
  direction="nsm",
 },
 [0x1A74]={
  category="mn",
  direction="nsm",
 },
 [0x1A75]={
  category="mn",
  direction="nsm",
 },
 [0x1A76]={
  category="mn",
  direction="nsm",
 },
 [0x1A77]={
  category="mn",
  direction="nsm",
 },
 [0x1A78]={
  category="mn",
  direction="nsm",
 },
 [0x1A79]={
  category="mn",
  direction="nsm",
 },
 [0x1A7A]={
  category="mn",
  direction="nsm",
 },
 [0x1A7B]={
  category="mn",
  direction="nsm",
 },
 [0x1A7C]={
  category="mn",
  direction="nsm",
 },
 [0x1A7F]={
  category="mn",
  direction="nsm",
 },
 [0x1A80]={
  category="nd",
  direction="l",
 },
 [0x1A81]={
  category="nd",
  direction="l",
 },
 [0x1A82]={
  category="nd",
  direction="l",
 },
 [0x1A83]={
  category="nd",
  direction="l",
 },
 [0x1A84]={
  category="nd",
  direction="l",
 },
 [0x1A85]={
  category="nd",
  direction="l",
 },
 [0x1A86]={
  category="nd",
  direction="l",
 },
 [0x1A87]={
  category="nd",
  direction="l",
 },
 [0x1A88]={
  category="nd",
  direction="l",
 },
 [0x1A89]={
  category="nd",
  direction="l",
 },
 [0x1A90]={
  category="nd",
  direction="l",
 },
 [0x1A91]={
  category="nd",
  direction="l",
 },
 [0x1A92]={
  category="nd",
  direction="l",
 },
 [0x1A93]={
  category="nd",
  direction="l",
 },
 [0x1A94]={
  category="nd",
  direction="l",
 },
 [0x1A95]={
  category="nd",
  direction="l",
 },
 [0x1A96]={
  category="nd",
  direction="l",
 },
 [0x1A97]={
  category="nd",
  direction="l",
 },
 [0x1A98]={
  category="nd",
  direction="l",
 },
 [0x1A99]={
  category="nd",
  direction="l",
 },
 [0x1AA0]={
  category="po",
  direction="l",
 },
 [0x1AA1]={
  category="po",
  direction="l",
 },
 [0x1AA2]={
  category="po",
  direction="l",
 },
 [0x1AA3]={
  category="po",
  direction="l",
 },
 [0x1AA4]={
  category="po",
  direction="l",
 },
 [0x1AA5]={
  category="po",
  direction="l",
 },
 [0x1AA6]={
  category="po",
  direction="l",
 },
 [0x1AA7]={
  category="lm",
  direction="l",
 },
 [0x1AA8]={
  category="po",
  direction="l",
 },
 [0x1AA9]={
  category="po",
  direction="l",
 },
 [0x1AAA]={
  category="po",
  direction="l",
 },
 [0x1AAB]={
  category="po",
  direction="l",
 },
 [0x1AAC]={
  category="po",
  direction="l",
 },
 [0x1AAD]={
  category="po",
  direction="l",
 },
 [0x1AB0]={
  category="mn",
  direction="nsm",
 },
 [0x1AB1]={
  category="mn",
  direction="nsm",
 },
 [0x1AB2]={
  category="mn",
  direction="nsm",
 },
 [0x1AB3]={
  category="mn",
  direction="nsm",
 },
 [0x1AB4]={
  category="mn",
  direction="nsm",
 },
 [0x1AB5]={
  category="mn",
  direction="nsm",
 },
 [0x1AB6]={
  category="mn",
  direction="nsm",
 },
 [0x1AB7]={
  category="mn",
  direction="nsm",
 },
 [0x1AB8]={
  category="mn",
  direction="nsm",
 },
 [0x1AB9]={
  category="mn",
  direction="nsm",
 },
 [0x1ABA]={
  category="mn",
  direction="nsm",
 },
 [0x1ABB]={
  category="mn",
  direction="nsm",
 },
 [0x1ABC]={
  category="mn",
  direction="nsm",
 },
 [0x1ABD]={
  category="mn",
  direction="nsm",
 },
 [0x1ABE]={
  category="me",
  direction="nsm",
 },
 [0x1B00]={
  category="mn",
  direction="nsm",
 },
 [0x1B01]={
  category="mn",
  direction="nsm",
 },
 [0x1B02]={
  category="mn",
  direction="nsm",
 },
 [0x1B03]={
  category="mn",
  direction="nsm",
 },
 [0x1B04]={
  category="mc",
  direction="l",
 },
 [0x1B05]={
  category="lo",
  direction="l",
 },
 [0x1B06]={
  category="lo",
  direction="l",
 },
 [0x1B07]={
  category="lo",
  direction="l",
 },
 [0x1B08]={
  category="lo",
  direction="l",
 },
 [0x1B09]={
  category="lo",
  direction="l",
 },
 [0x1B0A]={
  category="lo",
  direction="l",
 },
 [0x1B0B]={
  category="lo",
  direction="l",
 },
 [0x1B0C]={
  category="lo",
  direction="l",
 },
 [0x1B0D]={
  category="lo",
  direction="l",
 },
 [0x1B0E]={
  category="lo",
  direction="l",
 },
 [0x1B0F]={
  category="lo",
  direction="l",
 },
 [0x1B10]={
  category="lo",
  direction="l",
 },
 [0x1B11]={
  category="lo",
  direction="l",
 },
 [0x1B12]={
  category="lo",
  direction="l",
 },
 [0x1B13]={
  category="lo",
  direction="l",
 },
 [0x1B14]={
  category="lo",
  direction="l",
 },
 [0x1B15]={
  category="lo",
  direction="l",
 },
 [0x1B16]={
  category="lo",
  direction="l",
 },
 [0x1B17]={
  category="lo",
  direction="l",
 },
 [0x1B18]={
  category="lo",
  direction="l",
 },
 [0x1B19]={
  category="lo",
  direction="l",
 },
 [0x1B1A]={
  category="lo",
  direction="l",
 },
 [0x1B1B]={
  category="lo",
  direction="l",
 },
 [0x1B1C]={
  category="lo",
  direction="l",
 },
 [0x1B1D]={
  category="lo",
  direction="l",
 },
 [0x1B1E]={
  category="lo",
  direction="l",
 },
 [0x1B1F]={
  category="lo",
  direction="l",
 },
 [0x1B20]={
  category="lo",
  direction="l",
 },
 [0x1B21]={
  category="lo",
  direction="l",
 },
 [0x1B22]={
  category="lo",
  direction="l",
 },
 [0x1B23]={
  category="lo",
  direction="l",
 },
 [0x1B24]={
  category="lo",
  direction="l",
 },
 [0x1B25]={
  category="lo",
  direction="l",
 },
 [0x1B26]={
  category="lo",
  direction="l",
 },
 [0x1B27]={
  category="lo",
  direction="l",
 },
 [0x1B28]={
  category="lo",
  direction="l",
 },
 [0x1B29]={
  category="lo",
  direction="l",
 },
 [0x1B2A]={
  category="lo",
  direction="l",
 },
 [0x1B2B]={
  category="lo",
  direction="l",
 },
 [0x1B2C]={
  category="lo",
  direction="l",
 },
 [0x1B2D]={
  category="lo",
  direction="l",
 },
 [0x1B2E]={
  category="lo",
  direction="l",
 },
 [0x1B2F]={
  category="lo",
  direction="l",
 },
 [0x1B30]={
  category="lo",
  direction="l",
 },
 [0x1B31]={
  category="lo",
  direction="l",
 },
 [0x1B32]={
  category="lo",
  direction="l",
 },
 [0x1B33]={
  category="lo",
  direction="l",
 },
 [0x1B34]={
  category="mn",
  direction="nsm",
 },
 [0x1B35]={
  category="mc",
  direction="l",
 },
 [0x1B36]={
  category="mn",
  direction="nsm",
 },
 [0x1B37]={
  category="mn",
  direction="nsm",
 },
 [0x1B38]={
  category="mn",
  direction="nsm",
 },
 [0x1B39]={
  category="mn",
  direction="nsm",
 },
 [0x1B3A]={
  category="mn",
  direction="nsm",
 },
 [0x1B3B]={
  category="mc",
  direction="l",
 },
 [0x1B3C]={
  category="mn",
  direction="nsm",
 },
 [0x1B3D]={
  category="mc",
  direction="l",
 },
 [0x1B3E]={
  category="mc",
  direction="l",
 },
 [0x1B3F]={
  category="mc",
  direction="l",
 },
 [0x1B40]={
  category="mc",
  direction="l",
 },
 [0x1B41]={
  category="mc",
  direction="l",
 },
 [0x1B42]={
  category="mn",
  direction="nsm",
 },
 [0x1B43]={
  category="mc",
  direction="l",
 },
 [0x1B44]={
  category="mc",
  direction="l",
 },
 [0x1B45]={
  category="lo",
  direction="l",
 },
 [0x1B46]={
  category="lo",
  direction="l",
 },
 [0x1B47]={
  category="lo",
  direction="l",
 },
 [0x1B48]={
  category="lo",
  direction="l",
 },
 [0x1B49]={
  category="lo",
  direction="l",
 },
 [0x1B4A]={
  category="lo",
  direction="l",
 },
 [0x1B4B]={
  category="lo",
  direction="l",
 },
 [0x1B50]={
  category="nd",
  direction="l",
 },
 [0x1B51]={
  category="nd",
  direction="l",
 },
 [0x1B52]={
  category="nd",
  direction="l",
 },
 [0x1B53]={
  category="nd",
  direction="l",
 },
 [0x1B54]={
  category="nd",
  direction="l",
 },
 [0x1B55]={
  category="nd",
  direction="l",
 },
 [0x1B56]={
  category="nd",
  direction="l",
 },
 [0x1B57]={
  category="nd",
  direction="l",
 },
 [0x1B58]={
  category="nd",
  direction="l",
 },
 [0x1B59]={
  category="nd",
  direction="l",
 },
 [0x1B5A]={
  category="po",
  direction="l",
 },
 [0x1B5B]={
  category="po",
  direction="l",
 },
 [0x1B5C]={
  category="po",
  direction="l",
 },
 [0x1B5D]={
  category="po",
  direction="l",
 },
 [0x1B5E]={
  category="po",
  direction="l",
 },
 [0x1B5F]={
  category="po",
  direction="l",
 },
 [0x1B60]={
  category="po",
  direction="l",
 },
 [0x1B61]={
  category="so",
  direction="l",
 },
 [0x1B62]={
  category="so",
  direction="l",
 },
 [0x1B63]={
  category="so",
  direction="l",
 },
 [0x1B64]={
  category="so",
  direction="l",
 },
 [0x1B65]={
  category="so",
  direction="l",
 },
 [0x1B66]={
  category="so",
  direction="l",
 },
 [0x1B67]={
  category="so",
  direction="l",
 },
 [0x1B68]={
  category="so",
  direction="l",
 },
 [0x1B69]={
  category="so",
  direction="l",
 },
 [0x1B6A]={
  category="so",
  direction="l",
 },
 [0x1B6B]={
  category="mn",
  direction="nsm",
 },
 [0x1B6C]={
  category="mn",
  direction="nsm",
 },
 [0x1B6D]={
  category="mn",
  direction="nsm",
 },
 [0x1B6E]={
  category="mn",
  direction="nsm",
 },
 [0x1B6F]={
  category="mn",
  direction="nsm",
 },
 [0x1B70]={
  category="mn",
  direction="nsm",
 },
 [0x1B71]={
  category="mn",
  direction="nsm",
 },
 [0x1B72]={
  category="mn",
  direction="nsm",
 },
 [0x1B73]={
  category="mn",
  direction="nsm",
 },
 [0x1B74]={
  category="so",
  direction="l",
 },
 [0x1B75]={
  category="so",
  direction="l",
 },
 [0x1B76]={
  category="so",
  direction="l",
 },
 [0x1B77]={
  category="so",
  direction="l",
 },
 [0x1B78]={
  category="so",
  direction="l",
 },
 [0x1B79]={
  category="so",
  direction="l",
 },
 [0x1B7A]={
  category="so",
  direction="l",
 },
 [0x1B7B]={
  category="so",
  direction="l",
 },
 [0x1B7C]={
  category="so",
  direction="l",
 },
 [0x1B80]={
  category="mn",
  direction="nsm",
 },
 [0x1B81]={
  category="mn",
  direction="nsm",
 },
 [0x1B82]={
  category="mc",
  direction="l",
 },
 [0x1B83]={
  category="lo",
  direction="l",
 },
 [0x1B84]={
  category="lo",
  direction="l",
 },
 [0x1B85]={
  category="lo",
  direction="l",
 },
 [0x1B86]={
  category="lo",
  direction="l",
 },
 [0x1B87]={
  category="lo",
  direction="l",
 },
 [0x1B88]={
  category="lo",
  direction="l",
 },
 [0x1B89]={
  category="lo",
  direction="l",
 },
 [0x1B8A]={
  category="lo",
  direction="l",
 },
 [0x1B8B]={
  category="lo",
  direction="l",
 },
 [0x1B8C]={
  category="lo",
  direction="l",
 },
 [0x1B8D]={
  category="lo",
  direction="l",
 },
 [0x1B8E]={
  category="lo",
  direction="l",
 },
 [0x1B8F]={
  category="lo",
  direction="l",
 },
 [0x1B90]={
  category="lo",
  direction="l",
 },
 [0x1B91]={
  category="lo",
  direction="l",
 },
 [0x1B92]={
  category="lo",
  direction="l",
 },
 [0x1B93]={
  category="lo",
  direction="l",
 },
 [0x1B94]={
  category="lo",
  direction="l",
 },
 [0x1B95]={
  category="lo",
  direction="l",
 },
 [0x1B96]={
  category="lo",
  direction="l",
 },
 [0x1B97]={
  category="lo",
  direction="l",
 },
 [0x1B98]={
  category="lo",
  direction="l",
 },
 [0x1B99]={
  category="lo",
  direction="l",
 },
 [0x1B9A]={
  category="lo",
  direction="l",
 },
 [0x1B9B]={
  category="lo",
  direction="l",
 },
 [0x1B9C]={
  category="lo",
  direction="l",
 },
 [0x1B9D]={
  category="lo",
  direction="l",
 },
 [0x1B9E]={
  category="lo",
  direction="l",
 },
 [0x1B9F]={
  category="lo",
  direction="l",
 },
 [0x1BA0]={
  category="lo",
  direction="l",
 },
 [0x1BA1]={
  category="mc",
  direction="l",
 },
 [0x1BA2]={
  category="mn",
  direction="nsm",
 },
 [0x1BA3]={
  category="mn",
  direction="nsm",
 },
 [0x1BA4]={
  category="mn",
  direction="nsm",
 },
 [0x1BA5]={
  category="mn",
  direction="nsm",
 },
 [0x1BA6]={
  category="mc",
  direction="l",
 },
 [0x1BA7]={
  category="mc",
  direction="l",
 },
 [0x1BA8]={
  category="mn",
  direction="nsm",
 },
 [0x1BA9]={
  category="mn",
  direction="nsm",
 },
 [0x1BAA]={
  category="mc",
  direction="l",
 },
 [0x1BAB]={
  category="mn",
  direction="nsm",
 },
 [0x1BAC]={
  category="mc",
  direction="nsm",
 },
 [0x1BAD]={
  category="mc",
  direction="nsm",
 },
 [0x1BAE]={
  category="lo",
  direction="l",
 },
 [0x1BAF]={
  category="lo",
  direction="l",
 },
 [0x1BB0]={
  category="nd",
  direction="l",
 },
 [0x1BB1]={
  category="nd",
  direction="l",
 },
 [0x1BB2]={
  category="nd",
  direction="l",
 },
 [0x1BB3]={
  category="nd",
  direction="l",
 },
 [0x1BB4]={
  category="nd",
  direction="l",
 },
 [0x1BB5]={
  category="nd",
  direction="l",
 },
 [0x1BB6]={
  category="nd",
  direction="l",
 },
 [0x1BB7]={
  category="nd",
  direction="l",
 },
 [0x1BB8]={
  category="nd",
  direction="l",
 },
 [0x1BB9]={
  category="nd",
  direction="l",
 },
 [0x1BBA]={
  category="lo",
  direction="l",
 },
 [0x1BBB]={
  category="lo",
  direction="l",
 },
 [0x1BBC]={
  category="lo",
  direction="l",
 },
 [0x1BBD]={
  category="lo",
  direction="l",
 },
 [0x1BBE]={
  category="lo",
  direction="l",
 },
 [0x1BBF]={
  category="lo",
  direction="l",
 },
 [0x1BC0]={
  category="lo",
  direction="l",
 },
 [0x1BC1]={
  category="lo",
  direction="l",
 },
 [0x1BC2]={
  category="lo",
  direction="l",
 },
 [0x1BC3]={
  category="lo",
  direction="l",
 },
 [0x1BC4]={
  category="lo",
  direction="l",
 },
 [0x1BC5]={
  category="lo",
  direction="l",
 },
 [0x1BC6]={
  category="lo",
  direction="l",
 },
 [0x1BC7]={
  category="lo",
  direction="l",
 },
 [0x1BC8]={
  category="lo",
  direction="l",
 },
 [0x1BC9]={
  category="lo",
  direction="l",
 },
 [0x1BCA]={
  category="lo",
  direction="l",
 },
 [0x1BCB]={
  category="lo",
  direction="l",
 },
 [0x1BCC]={
  category="lo",
  direction="l",
 },
 [0x1BCD]={
  category="lo",
  direction="l",
 },
 [0x1BCE]={
  category="lo",
  direction="l",
 },
 [0x1BCF]={
  category="lo",
  direction="l",
 },
 [0x1BD0]={
  category="lo",
  direction="l",
 },
 [0x1BD1]={
  category="lo",
  direction="l",
 },
 [0x1BD2]={
  category="lo",
  direction="l",
 },
 [0x1BD3]={
  category="lo",
  direction="l",
 },
 [0x1BD4]={
  category="lo",
  direction="l",
 },
 [0x1BD5]={
  category="lo",
  direction="l",
 },
 [0x1BD6]={
  category="lo",
  direction="l",
 },
 [0x1BD7]={
  category="lo",
  direction="l",
 },
 [0x1BD8]={
  category="lo",
  direction="l",
 },
 [0x1BD9]={
  category="lo",
  direction="l",
 },
 [0x1BDA]={
  category="lo",
  direction="l",
 },
 [0x1BDB]={
  category="lo",
  direction="l",
 },
 [0x1BDC]={
  category="lo",
  direction="l",
 },
 [0x1BDD]={
  category="lo",
  direction="l",
 },
 [0x1BDE]={
  category="lo",
  direction="l",
 },
 [0x1BDF]={
  category="lo",
  direction="l",
 },
 [0x1BE0]={
  category="lo",
  direction="l",
 },
 [0x1BE1]={
  category="lo",
  direction="l",
 },
 [0x1BE2]={
  category="lo",
  direction="l",
 },
 [0x1BE3]={
  category="lo",
  direction="l",
 },
 [0x1BE4]={
  category="lo",
  direction="l",
 },
 [0x1BE5]={
  category="lo",
  direction="l",
 },
 [0x1BE6]={
  category="mn",
  direction="nsm",
 },
 [0x1BE7]={
  category="mc",
  direction="l",
 },
 [0x1BE8]={
  category="mn",
  direction="nsm",
 },
 [0x1BE9]={
  category="mn",
  direction="nsm",
 },
 [0x1BEA]={
  category="mc",
  direction="l",
 },
 [0x1BEB]={
  category="mc",
  direction="l",
 },
 [0x1BEC]={
  category="mc",
  direction="l",
 },
 [0x1BED]={
  category="mn",
  direction="nsm",
 },
 [0x1BEE]={
  category="mc",
  direction="l",
 },
 [0x1BEF]={
  category="mn",
  direction="nsm",
 },
 [0x1BF0]={
  category="mn",
  direction="nsm",
 },
 [0x1BF1]={
  category="mn",
  direction="nsm",
 },
 [0x1BF2]={
  category="mc",
  direction="l",
 },
 [0x1BF3]={
  category="mc",
  direction="l",
 },
 [0x1BFC]={
  category="po",
  direction="l",
 },
 [0x1BFD]={
  category="po",
  direction="l",
 },
 [0x1BFE]={
  category="po",
  direction="l",
 },
 [0x1BFF]={
  category="po",
  direction="l",
 },
 [0x1C00]={
  category="lo",
  direction="l",
 },
 [0x1C01]={
  category="lo",
  direction="l",
 },
 [0x1C02]={
  category="lo",
  direction="l",
 },
 [0x1C03]={
  category="lo",
  direction="l",
 },
 [0x1C04]={
  category="lo",
  direction="l",
 },
 [0x1C05]={
  category="lo",
  direction="l",
 },
 [0x1C06]={
  category="lo",
  direction="l",
 },
 [0x1C07]={
  category="lo",
  direction="l",
 },
 [0x1C08]={
  category="lo",
  direction="l",
 },
 [0x1C09]={
  category="lo",
  direction="l",
 },
 [0x1C0A]={
  category="lo",
  direction="l",
 },
 [0x1C0B]={
  category="lo",
  direction="l",
 },
 [0x1C0C]={
  category="lo",
  direction="l",
 },
 [0x1C0D]={
  category="lo",
  direction="l",
 },
 [0x1C0E]={
  category="lo",
  direction="l",
 },
 [0x1C0F]={
  category="lo",
  direction="l",
 },
 [0x1C10]={
  category="lo",
  direction="l",
 },
 [0x1C11]={
  category="lo",
  direction="l",
 },
 [0x1C12]={
  category="lo",
  direction="l",
 },
 [0x1C13]={
  category="lo",
  direction="l",
 },
 [0x1C14]={
  category="lo",
  direction="l",
 },
 [0x1C15]={
  category="lo",
  direction="l",
 },
 [0x1C16]={
  category="lo",
  direction="l",
 },
 [0x1C17]={
  category="lo",
  direction="l",
 },
 [0x1C18]={
  category="lo",
  direction="l",
 },
 [0x1C19]={
  category="lo",
  direction="l",
 },
 [0x1C1A]={
  category="lo",
  direction="l",
 },
 [0x1C1B]={
  category="lo",
  direction="l",
 },
 [0x1C1C]={
  category="lo",
  direction="l",
 },
 [0x1C1D]={
  category="lo",
  direction="l",
 },
 [0x1C1E]={
  category="lo",
  direction="l",
 },
 [0x1C1F]={
  category="lo",
  direction="l",
 },
 [0x1C20]={
  category="lo",
  direction="l",
 },
 [0x1C21]={
  category="lo",
  direction="l",
 },
 [0x1C22]={
  category="lo",
  direction="l",
 },
 [0x1C23]={
  category="lo",
  direction="l",
 },
 [0x1C24]={
  category="mc",
  direction="l",
 },
 [0x1C25]={
  category="mc",
  direction="l",
 },
 [0x1C26]={
  category="mc",
  direction="l",
 },
 [0x1C27]={
  category="mc",
  direction="l",
 },
 [0x1C28]={
  category="mc",
  direction="l",
 },
 [0x1C29]={
  category="mc",
  direction="l",
 },
 [0x1C2A]={
  category="mc",
  direction="l",
 },
 [0x1C2B]={
  category="mc",
  direction="l",
 },
 [0x1C2C]={
  category="mn",
  direction="nsm",
 },
 [0x1C2D]={
  category="mn",
  direction="nsm",
 },
 [0x1C2E]={
  category="mn",
  direction="nsm",
 },
 [0x1C2F]={
  category="mn",
  direction="nsm",
 },
 [0x1C30]={
  category="mn",
  direction="nsm",
 },
 [0x1C31]={
  category="mn",
  direction="nsm",
 },
 [0x1C32]={
  category="mn",
  direction="nsm",
 },
 [0x1C33]={
  category="mn",
  direction="nsm",
 },
 [0x1C34]={
  category="mc",
  direction="l",
 },
 [0x1C35]={
  category="mc",
  direction="l",
 },
 [0x1C36]={
  category="mn",
  direction="nsm",
 },
 [0x1C37]={
  category="mn",
  direction="nsm",
 },
 [0x1C3B]={
  category="po",
  direction="l",
 },
 [0x1C3C]={
  category="po",
  direction="l",
 },
 [0x1C3D]={
  category="po",
  direction="l",
 },
 [0x1C3E]={
  category="po",
  direction="l",
 },
 [0x1C3F]={
  category="po",
  direction="l",
 },
 [0x1C40]={
  category="nd",
  direction="l",
 },
 [0x1C41]={
  category="nd",
  direction="l",
 },
 [0x1C42]={
  category="nd",
  direction="l",
 },
 [0x1C43]={
  category="nd",
  direction="l",
 },
 [0x1C44]={
  category="nd",
  direction="l",
 },
 [0x1C45]={
  category="nd",
  direction="l",
 },
 [0x1C46]={
  category="nd",
  direction="l",
 },
 [0x1C47]={
  category="nd",
  direction="l",
 },
 [0x1C48]={
  category="nd",
  direction="l",
 },
 [0x1C49]={
  category="nd",
  direction="l",
 },
 [0x1C4D]={
  category="lo",
  direction="l",
 },
 [0x1C4E]={
  category="lo",
  direction="l",
 },
 [0x1C4F]={
  category="lo",
  direction="l",
 },
 [0x1C50]={
  category="nd",
  direction="l",
 },
 [0x1C51]={
  category="nd",
  direction="l",
 },
 [0x1C52]={
  category="nd",
  direction="l",
 },
 [0x1C53]={
  category="nd",
  direction="l",
 },
 [0x1C54]={
  category="nd",
  direction="l",
 },
 [0x1C55]={
  category="nd",
  direction="l",
 },
 [0x1C56]={
  category="nd",
  direction="l",
 },
 [0x1C57]={
  category="nd",
  direction="l",
 },
 [0x1C58]={
  category="nd",
  direction="l",
 },
 [0x1C59]={
  category="nd",
  direction="l",
 },
 [0x1C5A]={
  category="lo",
  direction="l",
 },
 [0x1C5B]={
  category="lo",
  direction="l",
 },
 [0x1C5C]={
  category="lo",
  direction="l",
 },
 [0x1C5D]={
  category="lo",
  direction="l",
 },
 [0x1C5E]={
  category="lo",
  direction="l",
 },
 [0x1C5F]={
  category="lo",
  direction="l",
 },
 [0x1C60]={
  category="lo",
  direction="l",
 },
 [0x1C61]={
  category="lo",
  direction="l",
 },
 [0x1C62]={
  category="lo",
  direction="l",
 },
 [0x1C63]={
  category="lo",
  direction="l",
 },
 [0x1C64]={
  category="lo",
  direction="l",
 },
 [0x1C65]={
  category="lo",
  direction="l",
 },
 [0x1C66]={
  category="lo",
  direction="l",
 },
 [0x1C67]={
  category="lo",
  direction="l",
 },
 [0x1C68]={
  category="lo",
  direction="l",
 },
 [0x1C69]={
  category="lo",
  direction="l",
 },
 [0x1C6A]={
  category="lo",
  direction="l",
 },
 [0x1C6B]={
  category="lo",
  direction="l",
 },
 [0x1C6C]={
  category="lo",
  direction="l",
 },
 [0x1C6D]={
  category="lo",
  direction="l",
 },
 [0x1C6E]={
  category="lo",
  direction="l",
 },
 [0x1C6F]={
  category="lo",
  direction="l",
 },
 [0x1C70]={
  category="lo",
  direction="l",
 },
 [0x1C71]={
  category="lo",
  direction="l",
 },
 [0x1C72]={
  category="lo",
  direction="l",
 },
 [0x1C73]={
  category="lo",
  direction="l",
 },
 [0x1C74]={
  category="lo",
  direction="l",
 },
 [0x1C75]={
  category="lo",
  direction="l",
 },
 [0x1C76]={
  category="lo",
  direction="l",
 },
 [0x1C77]={
  category="lo",
  direction="l",
 },
 [0x1C78]={
  category="lm",
  direction="l",
 },
 [0x1C79]={
  category="lm",
  direction="l",
 },
 [0x1C7A]={
  category="lm",
  direction="l",
 },
 [0x1C7B]={
  category="lm",
  direction="l",
 },
 [0x1C7C]={
  category="lm",
  direction="l",
 },
 [0x1C7D]={
  category="lm",
  direction="l",
 },
 [0x1C7E]={
  category="po",
  direction="l",
 },
 [0x1C7F]={
  category="po",
  direction="l",
 },
 [0x1CC0]={
  category="po",
  direction="l",
 },
 [0x1CC1]={
  category="po",
  direction="l",
 },
 [0x1CC2]={
  category="po",
  direction="l",
 },
 [0x1CC3]={
  category="po",
  direction="l",
 },
 [0x1CC4]={
  category="po",
  direction="l",
 },
 [0x1CC5]={
  category="po",
  direction="l",
 },
 [0x1CC6]={
  category="po",
  direction="l",
 },
 [0x1CC7]={
  category="po",
  direction="l",
 },
 [0x1CD0]={
  category="mn",
  direction="nsm",
 },
 [0x1CD1]={
  category="mn",
  direction="nsm",
 },
 [0x1CD2]={
  category="mn",
  direction="nsm",
 },
 [0x1CD3]={
  category="po",
  direction="l",
 },
 [0x1CD4]={
  category="mn",
  direction="nsm",
 },
 [0x1CD5]={
  category="mn",
  direction="nsm",
 },
 [0x1CD6]={
  category="mn",
  direction="nsm",
 },
 [0x1CD7]={
  category="mn",
  direction="nsm",
 },
 [0x1CD8]={
  category="mn",
  direction="nsm",
 },
 [0x1CD9]={
  category="mn",
  direction="nsm",
 },
 [0x1CDA]={
  category="mn",
  direction="nsm",
 },
 [0x1CDB]={
  category="mn",
  direction="nsm",
 },
 [0x1CDC]={
  category="mn",
  direction="nsm",
 },
 [0x1CDD]={
  category="mn",
  direction="nsm",
 },
 [0x1CDE]={
  category="mn",
  direction="nsm",
 },
 [0x1CDF]={
  category="mn",
  direction="nsm",
 },
 [0x1CE0]={
  category="mn",
  direction="nsm",
 },
 [0x1CE1]={
  category="mc",
  direction="l",
 },
 [0x1CE2]={
  category="mn",
  direction="nsm",
 },
 [0x1CE3]={
  category="mn",
  direction="nsm",
 },
 [0x1CE4]={
  category="mn",
  direction="nsm",
 },
 [0x1CE5]={
  category="mn",
  direction="nsm",
 },
 [0x1CE6]={
  category="mn",
  direction="nsm",
 },
 [0x1CE7]={
  category="mn",
  direction="nsm",
 },
 [0x1CE8]={
  category="mn",
  direction="nsm",
 },
 [0x1CE9]={
  category="lo",
  direction="l",
 },
 [0x1CEA]={
  category="lo",
  direction="l",
 },
 [0x1CEB]={
  category="lo",
  direction="l",
 },
 [0x1CEC]={
  category="lo",
  direction="l",
 },
 [0x1CED]={
  category="mn",
  direction="nsm",
 },
 [0x1CEE]={
  category="lo",
  direction="l",
 },
 [0x1CEF]={
  category="lo",
  direction="l",
 },
 [0x1CF0]={
  category="lo",
  direction="l",
 },
 [0x1CF1]={
  category="lo",
  direction="l",
 },
 [0x1CF2]={
  category="mc",
  direction="l",
 },
 [0x1CF3]={
  category="mc",
  direction="l",
 },
 [0x1CF4]={
  category="mn",
  direction="nsm",
 },
 [0x1CF5]={
  category="lo",
  direction="l",
 },
 [0x1CF6]={
  category="lo",
  direction="l",
 },
 [0x1CF8]={
  category="mn",
  direction="nsm",
 },
 [0x1CF9]={
  category="mn",
  direction="nsm",
 },
 [0x1D00]={
  category="ll",
  direction="l",
 },
 [0x1D01]={
  category="ll",
  direction="l",
 },
 [0x1D02]={
  category="ll",
  direction="l",
 },
 [0x1D03]={
  category="ll",
  direction="l",
 },
 [0x1D04]={
  category="ll",
  direction="l",
 },
 [0x1D05]={
  category="ll",
  direction="l",
 },
 [0x1D06]={
  category="ll",
  direction="l",
 },
 [0x1D07]={
  category="ll",
  direction="l",
 },
 [0x1D08]={
  category="ll",
  direction="l",
 },
 [0x1D09]={
  category="ll",
  direction="l",
 },
 [0x1D0A]={
  category="ll",
  direction="l",
 },
 [0x1D0B]={
  category="ll",
  direction="l",
 },
 [0x1D0C]={
  category="ll",
  direction="l",
 },
 [0x1D0D]={
  category="ll",
  direction="l",
 },
 [0x1D0E]={
  category="ll",
  direction="l",
 },
 [0x1D0F]={
  category="ll",
  direction="l",
 },
 [0x1D10]={
  category="ll",
  direction="l",
 },
 [0x1D11]={
  category="ll",
  direction="l",
 },
 [0x1D12]={
  category="ll",
  direction="l",
 },
 [0x1D13]={
  category="ll",
  direction="l",
 },
 [0x1D14]={
  category="ll",
  direction="l",
 },
 [0x1D15]={
  category="ll",
  direction="l",
 },
 [0x1D16]={
  category="ll",
  direction="l",
 },
 [0x1D17]={
  category="ll",
  direction="l",
 },
 [0x1D18]={
  category="ll",
  direction="l",
 },
 [0x1D19]={
  category="ll",
  direction="l",
 },
 [0x1D1A]={
  category="ll",
  direction="l",
 },
 [0x1D1B]={
  category="ll",
  direction="l",
 },
 [0x1D1C]={
  category="ll",
  direction="l",
 },
 [0x1D1D]={
  category="ll",
  direction="l",
 },
 [0x1D1E]={
  category="ll",
  direction="l",
 },
 [0x1D1F]={
  category="ll",
  direction="l",
 },
 [0x1D20]={
  category="ll",
  direction="l",
 },
 [0x1D21]={
  category="ll",
  direction="l",
 },
 [0x1D22]={
  category="ll",
  direction="l",
 },
 [0x1D23]={
  category="ll",
  direction="l",
 },
 [0x1D24]={
  category="ll",
  direction="l",
 },
 [0x1D25]={
  category="ll",
  direction="l",
 },
 [0x1D26]={
  category="ll",
  direction="l",
 },
 [0x1D27]={
  category="ll",
  direction="l",
 },
 [0x1D28]={
  category="ll",
  direction="l",
 },
 [0x1D29]={
  category="ll",
  direction="l",
 },
 [0x1D2A]={
  category="ll",
  direction="l",
 },
 [0x1D2B]={
  category="ll",
  direction="l",
 },
 [0x1D2C]={
  category="lm",
  direction="l",
 },
 [0x1D2D]={
  category="lm",
  direction="l",
 },
 [0x1D2E]={
  category="lm",
  direction="l",
 },
 [0x1D2F]={
  category="lm",
  direction="l",
 },
 [0x1D30]={
  category="lm",
  direction="l",
 },
 [0x1D31]={
  category="lm",
  direction="l",
 },
 [0x1D32]={
  category="lm",
  direction="l",
 },
 [0x1D33]={
  category="lm",
  direction="l",
 },
 [0x1D34]={
  category="lm",
  direction="l",
 },
 [0x1D35]={
  category="lm",
  direction="l",
 },
 [0x1D36]={
  category="lm",
  direction="l",
 },
 [0x1D37]={
  category="lm",
  direction="l",
 },
 [0x1D38]={
  category="lm",
  direction="l",
 },
 [0x1D39]={
  category="lm",
  direction="l",
 },
 [0x1D3A]={
  category="lm",
  direction="l",
 },
 [0x1D3B]={
  category="lm",
  direction="l",
 },
 [0x1D3C]={
  category="lm",
  direction="l",
 },
 [0x1D3D]={
  category="lm",
  direction="l",
 },
 [0x1D3E]={
  category="lm",
  direction="l",
 },
 [0x1D3F]={
  category="lm",
  direction="l",
 },
 [0x1D40]={
  category="lm",
  direction="l",
 },
 [0x1D41]={
  category="lm",
  direction="l",
 },
 [0x1D42]={
  category="lm",
  direction="l",
 },
 [0x1D43]={
  category="lm",
  direction="l",
 },
 [0x1D44]={
  category="lm",
  direction="l",
 },
 [0x1D45]={
  category="lm",
  direction="l",
 },
 [0x1D46]={
  category="lm",
  direction="l",
 },
 [0x1D47]={
  category="lm",
  direction="l",
 },
 [0x1D48]={
  category="lm",
  direction="l",
 },
 [0x1D49]={
  category="lm",
  direction="l",
 },
 [0x1D4A]={
  category="lm",
  direction="l",
 },
 [0x1D4B]={
  category="lm",
  direction="l",
 },
 [0x1D4C]={
  category="lm",
  direction="l",
 },
 [0x1D4D]={
  category="lm",
  direction="l",
 },
 [0x1D4E]={
  category="lm",
  direction="l",
 },
 [0x1D4F]={
  category="lm",
  direction="l",
 },
 [0x1D50]={
  category="lm",
  direction="l",
 },
 [0x1D51]={
  category="lm",
  direction="l",
 },
 [0x1D52]={
  category="lm",
  direction="l",
 },
 [0x1D53]={
  category="lm",
  direction="l",
 },
 [0x1D54]={
  category="lm",
  direction="l",
 },
 [0x1D55]={
  category="lm",
  direction="l",
 },
 [0x1D56]={
  category="lm",
  direction="l",
 },
 [0x1D57]={
  category="lm",
  direction="l",
 },
 [0x1D58]={
  category="lm",
  direction="l",
 },
 [0x1D59]={
  category="lm",
  direction="l",
 },
 [0x1D5A]={
  category="lm",
  direction="l",
 },
 [0x1D5B]={
  category="lm",
  direction="l",
 },
 [0x1D5C]={
  category="lm",
  direction="l",
 },
 [0x1D5D]={
  category="lm",
  direction="l",
 },
 [0x1D5E]={
  category="lm",
  direction="l",
 },
 [0x1D5F]={
  category="lm",
  direction="l",
 },
 [0x1D60]={
  category="lm",
  direction="l",
 },
 [0x1D61]={
  category="lm",
  direction="l",
 },
 [0x1D62]={
  category="ll",
  direction="l",
 },
 [0x1D63]={
  category="ll",
  direction="l",
 },
 [0x1D64]={
  category="ll",
  direction="l",
 },
 [0x1D65]={
  category="ll",
  direction="l",
 },
 [0x1D66]={
  category="ll",
  direction="l",
 },
 [0x1D67]={
  category="ll",
  direction="l",
 },
 [0x1D68]={
  category="ll",
  direction="l",
 },
 [0x1D69]={
  category="ll",
  direction="l",
 },
 [0x1D6A]={
  category="ll",
  direction="l",
 },
 [0x1D6B]={
  category="ll",
  direction="l",
 },
 [0x1D6C]={
  category="ll",
  direction="l",
 },
 [0x1D6D]={
  category="ll",
  direction="l",
 },
 [0x1D6E]={
  category="ll",
  direction="l",
 },
 [0x1D6F]={
  category="ll",
  direction="l",
 },
 [0x1D70]={
  category="ll",
  direction="l",
 },
 [0x1D71]={
  category="ll",
  direction="l",
 },
 [0x1D72]={
  category="ll",
  direction="l",
 },
 [0x1D73]={
  category="ll",
  direction="l",
 },
 [0x1D74]={
  category="ll",
  direction="l",
 },
 [0x1D75]={
  category="ll",
  direction="l",
 },
 [0x1D76]={
  category="ll",
  direction="l",
 },
 [0x1D77]={
  category="ll",
  direction="l",
 },
 [0x1D78]={
  category="lm",
  direction="l",
 },
 [0x1D79]={
  category="ll",
  direction="l",
 },
 [0x1D7A]={
  category="ll",
  direction="l",
 },
 [0x1D7B]={
  category="ll",
  direction="l",
 },
 [0x1D7C]={
  category="ll",
  direction="l",
 },
 [0x1D7D]={
  category="ll",
  direction="l",
 },
 [0x1D7E]={
  category="ll",
  direction="l",
 },
 [0x1D7F]={
  category="ll",
  direction="l",
 },
 [0x1D80]={
  category="ll",
  direction="l",
 },
 [0x1D81]={
  category="ll",
  direction="l",
 },
 [0x1D82]={
  category="ll",
  direction="l",
 },
 [0x1D83]={
  category="ll",
  direction="l",
 },
 [0x1D84]={
  category="ll",
  direction="l",
 },
 [0x1D85]={
  category="ll",
  direction="l",
 },
 [0x1D86]={
  category="ll",
  direction="l",
 },
 [0x1D87]={
  category="ll",
  direction="l",
 },
 [0x1D88]={
  category="ll",
  direction="l",
 },
 [0x1D89]={
  category="ll",
  direction="l",
 },
 [0x1D8A]={
  category="ll",
  direction="l",
 },
 [0x1D8B]={
  category="ll",
  direction="l",
 },
 [0x1D8C]={
  category="ll",
  direction="l",
 },
 [0x1D8D]={
  category="ll",
  direction="l",
 },
 [0x1D8E]={
  category="ll",
  direction="l",
 },
 [0x1D8F]={
  category="ll",
  direction="l",
 },
 [0x1D90]={
  category="ll",
  direction="l",
 },
 [0x1D91]={
  category="ll",
  direction="l",
 },
 [0x1D92]={
  category="ll",
  direction="l",
 },
 [0x1D93]={
  category="ll",
  direction="l",
 },
 [0x1D94]={
  category="ll",
  direction="l",
 },
 [0x1D95]={
  category="ll",
  direction="l",
 },
 [0x1D96]={
  category="ll",
  direction="l",
 },
 [0x1D97]={
  category="ll",
  direction="l",
 },
 [0x1D98]={
  category="ll",
  direction="l",
 },
 [0x1D99]={
  category="ll",
  direction="l",
 },
 [0x1D9A]={
  category="ll",
  direction="l",
 },
 [0x1D9B]={
  category="lm",
  direction="l",
 },
 [0x1D9C]={
  category="lm",
  direction="l",
 },
 [0x1D9D]={
  category="lm",
  direction="l",
 },
 [0x1D9E]={
  category="lm",
  direction="l",
 },
 [0x1D9F]={
  category="lm",
  direction="l",
 },
 [0x1DA0]={
  category="lm",
  direction="l",
 },
 [0x1DA1]={
  category="lm",
  direction="l",
 },
 [0x1DA2]={
  category="lm",
  direction="l",
 },
 [0x1DA3]={
  category="lm",
  direction="l",
 },
 [0x1DA4]={
  category="lm",
  direction="l",
 },
 [0x1DA5]={
  category="lm",
  direction="l",
 },
 [0x1DA6]={
  category="lm",
  direction="l",
 },
 [0x1DA7]={
  category="lm",
  direction="l",
 },
 [0x1DA8]={
  category="lm",
  direction="l",
 },
 [0x1DA9]={
  category="lm",
  direction="l",
 },
 [0x1DAA]={
  category="lm",
  direction="l",
 },
 [0x1DAB]={
  category="lm",
  direction="l",
 },
 [0x1DAC]={
  category="lm",
  direction="l",
 },
 [0x1DAD]={
  category="lm",
  direction="l",
 },
 [0x1DAE]={
  category="lm",
  direction="l",
 },
 [0x1DAF]={
  category="lm",
  direction="l",
 },
 [0x1DB0]={
  category="lm",
  direction="l",
 },
 [0x1DB1]={
  category="lm",
  direction="l",
 },
 [0x1DB2]={
  category="lm",
  direction="l",
 },
 [0x1DB3]={
  category="lm",
  direction="l",
 },
 [0x1DB4]={
  category="lm",
  direction="l",
 },
 [0x1DB5]={
  category="lm",
  direction="l",
 },
 [0x1DB6]={
  category="lm",
  direction="l",
 },
 [0x1DB7]={
  category="lm",
  direction="l",
 },
 [0x1DB8]={
  category="lm",
  direction="l",
 },
 [0x1DB9]={
  category="lm",
  direction="l",
 },
 [0x1DBA]={
  category="lm",
  direction="l",
 },
 [0x1DBB]={
  category="lm",
  direction="l",
 },
 [0x1DBC]={
  category="lm",
  direction="l",
 },
 [0x1DBD]={
  category="lm",
  direction="l",
 },
 [0x1DBE]={
  category="lm",
  direction="l",
 },
 [0x1DBF]={
  category="lm",
  direction="l",
 },
 [0x1DC0]={
  category="mn",
  direction="nsm",
 },
 [0x1DC1]={
  category="mn",
  direction="nsm",
 },
 [0x1DC2]={
  category="mn",
  direction="nsm",
 },
 [0x1DC3]={
  category="mn",
  direction="nsm",
 },
 [0x1DC4]={
  category="mn",
  direction="nsm",
 },
 [0x1DC5]={
  category="mn",
  direction="nsm",
 },
 [0x1DC6]={
  category="mn",
  direction="nsm",
 },
 [0x1DC7]={
  category="mn",
  direction="nsm",
 },
 [0x1DC8]={
  category="mn",
  direction="nsm",
 },
 [0x1DC9]={
  category="mn",
  direction="nsm",
 },
 [0x1DCA]={
  category="mn",
  direction="nsm",
 },
 [0x1DCB]={
  category="mn",
  direction="nsm",
 },
 [0x1DCC]={
  category="mn",
  direction="nsm",
 },
 [0x1DCD]={
  category="mn",
  direction="nsm",
 },
 [0x1DCE]={
  category="mn",
  direction="nsm",
 },
 [0x1DCF]={
  category="mn",
  direction="nsm",
 },
 [0x1DD0]={
  category="mn",
  direction="nsm",
 },
 [0x1DD1]={
  category="mn",
  direction="nsm",
 },
 [0x1DD2]={
  category="mn",
  direction="nsm",
 },
 [0x1DD3]={
  category="mn",
  direction="nsm",
 },
 [0x1DD4]={
  category="mn",
  direction="nsm",
 },
 [0x1DD5]={
  category="mn",
  direction="nsm",
 },
 [0x1DD6]={
  category="mn",
  direction="nsm",
 },
 [0x1DD7]={
  category="mn",
  direction="nsm",
 },
 [0x1DD8]={
  category="mn",
  direction="nsm",
 },
 [0x1DD9]={
  category="mn",
  direction="nsm",
 },
 [0x1DDA]={
  category="mn",
  direction="nsm",
 },
 [0x1DDB]={
  category="mn",
  direction="nsm",
 },
 [0x1DDC]={
  category="mn",
  direction="nsm",
 },
 [0x1DDD]={
  category="mn",
  direction="nsm",
 },
 [0x1DDE]={
  category="mn",
  direction="nsm",
 },
 [0x1DDF]={
  category="mn",
  direction="nsm",
 },
 [0x1DE0]={
  category="mn",
  direction="nsm",
 },
 [0x1DE1]={
  category="mn",
  direction="nsm",
 },
 [0x1DE2]={
  category="mn",
  direction="nsm",
 },
 [0x1DE3]={
  category="mn",
  direction="nsm",
 },
 [0x1DE4]={
  category="mn",
  direction="nsm",
 },
 [0x1DE5]={
  category="mn",
  direction="nsm",
 },
 [0x1DE6]={
  category="mn",
  direction="nsm",
 },
 [0x1DE7]={
  category="mn",
  direction="nsm",
 },
 [0x1DE8]={
  category="mn",
  direction="nsm",
 },
 [0x1DE9]={
  category="mn",
  direction="nsm",
 },
 [0x1DEA]={
  category="mn",
  direction="nsm",
 },
 [0x1DEB]={
  category="mn",
  direction="nsm",
 },
 [0x1DEC]={
  category="mn",
  direction="nsm",
 },
 [0x1DED]={
  category="mn",
  direction="nsm",
 },
 [0x1DEE]={
  category="mn",
  direction="nsm",
 },
 [0x1DEF]={
  category="mn",
  direction="nsm",
 },
 [0x1DF0]={
  category="mn",
  direction="nsm",
 },
 [0x1DF1]={
  category="mn",
  direction="nsm",
 },
 [0x1DF2]={
  category="mn",
  direction="nsm",
 },
 [0x1DF3]={
  category="mn",
  direction="nsm",
 },
 [0x1DF4]={
  category="mn",
  direction="nsm",
 },
 [0x1DF5]={
  category="mn",
  direction="nsm",
 },
 [0x1DFC]={
  category="mn",
  direction="nsm",
 },
 [0x1DFD]={
  category="mn",
  direction="nsm",
 },
 [0x1DFE]={
  category="mn",
  direction="nsm",
 },
 [0x1DFF]={
  category="mn",
  direction="nsm",
 },
 [0x1E00]={
  category="lu",
  direction="l",
 },
 [0x1E01]={
  category="ll",
  direction="l",
 },
 [0x1E02]={
  category="lu",
  direction="l",
 },
 [0x1E03]={
  category="ll",
  direction="l",
 },
 [0x1E04]={
  category="lu",
  direction="l",
 },
 [0x1E05]={
  category="ll",
  direction="l",
 },
 [0x1E06]={
  category="lu",
  direction="l",
 },
 [0x1E07]={
  category="ll",
  direction="l",
 },
 [0x1E08]={
  category="lu",
  direction="l",
 },
 [0x1E09]={
  category="ll",
  direction="l",
 },
 [0x1E0A]={
  category="lu",
  direction="l",
 },
 [0x1E0B]={
  category="ll",
  direction="l",
 },
 [0x1E0C]={
  category="lu",
  direction="l",
 },
 [0x1E0D]={
  category="ll",
  direction="l",
 },
 [0x1E0E]={
  category="lu",
  direction="l",
 },
 [0x1E0F]={
  category="ll",
  direction="l",
 },
 [0x1E10]={
  category="lu",
  direction="l",
 },
 [0x1E11]={
  category="ll",
  direction="l",
 },
 [0x1E12]={
  category="lu",
  direction="l",
 },
 [0x1E13]={
  category="ll",
  direction="l",
 },
 [0x1E14]={
  category="lu",
  direction="l",
 },
 [0x1E15]={
  category="ll",
  direction="l",
 },
 [0x1E16]={
  category="lu",
  direction="l",
 },
 [0x1E17]={
  category="ll",
  direction="l",
 },
 [0x1E18]={
  category="lu",
  direction="l",
 },
 [0x1E19]={
  category="ll",
  direction="l",
 },
 [0x1E1A]={
  category="lu",
  direction="l",
 },
 [0x1E1B]={
  category="ll",
  direction="l",
 },
 [0x1E1C]={
  category="lu",
  direction="l",
 },
 [0x1E1D]={
  category="ll",
  direction="l",
 },
 [0x1E1E]={
  category="lu",
  direction="l",
 },
 [0x1E1F]={
  category="ll",
  direction="l",
 },
 [0x1E20]={
  category="lu",
  direction="l",
 },
 [0x1E21]={
  category="ll",
  direction="l",
 },
 [0x1E22]={
  category="lu",
  direction="l",
 },
 [0x1E23]={
  category="ll",
  direction="l",
 },
 [0x1E24]={
  category="lu",
  direction="l",
 },
 [0x1E25]={
  category="ll",
  direction="l",
 },
 [0x1E26]={
  category="lu",
  direction="l",
 },
 [0x1E27]={
  category="ll",
  direction="l",
 },
 [0x1E28]={
  category="lu",
  direction="l",
 },
 [0x1E29]={
  category="ll",
  direction="l",
 },
 [0x1E2A]={
  category="lu",
  direction="l",
 },
 [0x1E2B]={
  category="ll",
  direction="l",
 },
 [0x1E2C]={
  category="lu",
  direction="l",
 },
 [0x1E2D]={
  category="ll",
  direction="l",
 },
 [0x1E2E]={
  category="lu",
  direction="l",
 },
 [0x1E2F]={
  category="ll",
  direction="l",
 },
 [0x1E30]={
  category="lu",
  direction="l",
 },
 [0x1E31]={
  category="ll",
  direction="l",
 },
 [0x1E32]={
  category="lu",
  direction="l",
 },
 [0x1E33]={
  category="ll",
  direction="l",
 },
 [0x1E34]={
  category="lu",
  direction="l",
 },
 [0x1E35]={
  category="ll",
  direction="l",
 },
 [0x1E36]={
  category="lu",
  direction="l",
 },
 [0x1E37]={
  category="ll",
  direction="l",
 },
 [0x1E38]={
  category="lu",
  direction="l",
 },
 [0x1E39]={
  category="ll",
  direction="l",
 },
 [0x1E3A]={
  category="lu",
  direction="l",
 },
 [0x1E3B]={
  category="ll",
  direction="l",
 },
 [0x1E3C]={
  category="lu",
  direction="l",
 },
 [0x1E3D]={
  category="ll",
  direction="l",
 },
 [0x1E3E]={
  category="lu",
  direction="l",
 },
 [0x1E3F]={
  category="ll",
  direction="l",
 },
 [0x1E40]={
  category="lu",
  direction="l",
 },
 [0x1E41]={
  category="ll",
  direction="l",
 },
 [0x1E42]={
  category="lu",
  direction="l",
 },
 [0x1E43]={
  category="ll",
  direction="l",
 },
 [0x1E44]={
  category="lu",
  direction="l",
 },
 [0x1E45]={
  category="ll",
  direction="l",
 },
 [0x1E46]={
  category="lu",
  direction="l",
 },
 [0x1E47]={
  category="ll",
  direction="l",
 },
 [0x1E48]={
  category="lu",
  direction="l",
 },
 [0x1E49]={
  category="ll",
  direction="l",
 },
 [0x1E4A]={
  category="lu",
  direction="l",
 },
 [0x1E4B]={
  category="ll",
  direction="l",
 },
 [0x1E4C]={
  category="lu",
  direction="l",
 },
 [0x1E4D]={
  category="ll",
  direction="l",
 },
 [0x1E4E]={
  category="lu",
  direction="l",
 },
 [0x1E4F]={
  category="ll",
  direction="l",
 },
 [0x1E50]={
  category="lu",
  direction="l",
 },
 [0x1E51]={
  category="ll",
  direction="l",
 },
 [0x1E52]={
  category="lu",
  direction="l",
 },
 [0x1E53]={
  category="ll",
  direction="l",
 },
 [0x1E54]={
  category="lu",
  direction="l",
 },
 [0x1E55]={
  category="ll",
  direction="l",
 },
 [0x1E56]={
  category="lu",
  direction="l",
 },
 [0x1E57]={
  category="ll",
  direction="l",
 },
 [0x1E58]={
  category="lu",
  direction="l",
 },
 [0x1E59]={
  category="ll",
  direction="l",
 },
 [0x1E5A]={
  category="lu",
  direction="l",
 },
 [0x1E5B]={
  category="ll",
  direction="l",
 },
 [0x1E5C]={
  category="lu",
  direction="l",
 },
 [0x1E5D]={
  category="ll",
  direction="l",
 },
 [0x1E5E]={
  category="lu",
  direction="l",
 },
 [0x1E5F]={
  category="ll",
  direction="l",
 },
 [0x1E60]={
  category="lu",
  direction="l",
 },
 [0x1E61]={
  category="ll",
  direction="l",
 },
 [0x1E62]={
  category="lu",
  direction="l",
 },
 [0x1E63]={
  category="ll",
  direction="l",
 },
 [0x1E64]={
  category="lu",
  direction="l",
 },
 [0x1E65]={
  category="ll",
  direction="l",
 },
 [0x1E66]={
  category="lu",
  direction="l",
 },
 [0x1E67]={
  category="ll",
  direction="l",
 },
 [0x1E68]={
  category="lu",
  direction="l",
 },
 [0x1E69]={
  category="ll",
  direction="l",
 },
 [0x1E6A]={
  category="lu",
  direction="l",
 },
 [0x1E6B]={
  category="ll",
  direction="l",
 },
 [0x1E6C]={
  category="lu",
  direction="l",
 },
 [0x1E6D]={
  category="ll",
  direction="l",
 },
 [0x1E6E]={
  category="lu",
  direction="l",
 },
 [0x1E6F]={
  category="ll",
  direction="l",
 },
 [0x1E70]={
  category="lu",
  direction="l",
 },
 [0x1E71]={
  category="ll",
  direction="l",
 },
 [0x1E72]={
  category="lu",
  direction="l",
 },
 [0x1E73]={
  category="ll",
  direction="l",
 },
 [0x1E74]={
  category="lu",
  direction="l",
 },
 [0x1E75]={
  category="ll",
  direction="l",
 },
 [0x1E76]={
  category="lu",
  direction="l",
 },
 [0x1E77]={
  category="ll",
  direction="l",
 },
 [0x1E78]={
  category="lu",
  direction="l",
 },
 [0x1E79]={
  category="ll",
  direction="l",
 },
 [0x1E7A]={
  category="lu",
  direction="l",
 },
 [0x1E7B]={
  category="ll",
  direction="l",
 },
 [0x1E7C]={
  category="lu",
  direction="l",
 },
 [0x1E7D]={
  category="ll",
  direction="l",
 },
 [0x1E7E]={
  category="lu",
  direction="l",
 },
 [0x1E7F]={
  category="ll",
  direction="l",
 },
 [0x1E80]={
  category="lu",
  direction="l",
 },
 [0x1E81]={
  category="ll",
  direction="l",
 },
 [0x1E82]={
  category="lu",
  direction="l",
 },
 [0x1E83]={
  category="ll",
  direction="l",
 },
 [0x1E84]={
  category="lu",
  direction="l",
 },
 [0x1E85]={
  category="ll",
  direction="l",
 },
 [0x1E86]={
  category="lu",
  direction="l",
 },
 [0x1E87]={
  category="ll",
  direction="l",
 },
 [0x1E88]={
  category="lu",
  direction="l",
 },
 [0x1E89]={
  category="ll",
  direction="l",
 },
 [0x1E8A]={
  category="lu",
  direction="l",
 },
 [0x1E8B]={
  category="ll",
  direction="l",
 },
 [0x1E8C]={
  category="lu",
  direction="l",
 },
 [0x1E8D]={
  category="ll",
  direction="l",
 },
 [0x1E8E]={
  category="lu",
  direction="l",
 },
 [0x1E8F]={
  category="ll",
  direction="l",
 },
 [0x1E90]={
  category="lu",
  direction="l",
 },
 [0x1E91]={
  category="ll",
  direction="l",
 },
 [0x1E92]={
  category="lu",
  direction="l",
 },
 [0x1E93]={
  category="ll",
  direction="l",
 },
 [0x1E94]={
  category="lu",
  direction="l",
 },
 [0x1E95]={
  category="ll",
  direction="l",
 },
 [0x1E96]={
  category="ll",
  direction="l",
 },
 [0x1E97]={
  category="ll",
  direction="l",
 },
 [0x1E98]={
  category="ll",
  direction="l",
 },
 [0x1E99]={
  category="ll",
  direction="l",
 },
 [0x1E9A]={
  category="ll",
  direction="l",
 },
 [0x1E9B]={
  category="ll",
  direction="l",
 },
 [0x1E9C]={
  category="ll",
  direction="l",
 },
 [0x1E9D]={
  category="ll",
  direction="l",
 },
 [0x1E9E]={
  category="lu",
  direction="l",
 },
 [0x1E9F]={
  category="ll",
  direction="l",
 },
 [0x1EA0]={
  category="lu",
  direction="l",
 },
 [0x1EA1]={
  category="ll",
  direction="l",
 },
 [0x1EA2]={
  category="lu",
  direction="l",
 },
 [0x1EA3]={
  category="ll",
  direction="l",
 },
 [0x1EA4]={
  category="lu",
  direction="l",
 },
 [0x1EA5]={
  category="ll",
  direction="l",
 },
 [0x1EA6]={
  category="lu",
  direction="l",
 },
 [0x1EA7]={
  category="ll",
  direction="l",
 },
 [0x1EA8]={
  category="lu",
  direction="l",
 },
 [0x1EA9]={
  category="ll",
  direction="l",
 },
 [0x1EAA]={
  category="lu",
  direction="l",
 },
 [0x1EAB]={
  category="ll",
  direction="l",
 },
 [0x1EAC]={
  category="lu",
  direction="l",
 },
 [0x1EAD]={
  category="ll",
  direction="l",
 },
 [0x1EAE]={
  category="lu",
  direction="l",
 },
 [0x1EAF]={
  category="ll",
  direction="l",
 },
 [0x1EB0]={
  category="lu",
  direction="l",
 },
 [0x1EB1]={
  category="ll",
  direction="l",
 },
 [0x1EB2]={
  category="lu",
  direction="l",
 },
 [0x1EB3]={
  category="ll",
  direction="l",
 },
 [0x1EB4]={
  category="lu",
  direction="l",
 },
 [0x1EB5]={
  category="ll",
  direction="l",
 },
 [0x1EB6]={
  category="lu",
  direction="l",
 },
 [0x1EB7]={
  category="ll",
  direction="l",
 },
 [0x1EB8]={
  category="lu",
  direction="l",
 },
 [0x1EB9]={
  category="ll",
  direction="l",
 },
 [0x1EBA]={
  category="lu",
  direction="l",
 },
 [0x1EBB]={
  category="ll",
  direction="l",
 },
 [0x1EBC]={
  category="lu",
  direction="l",
 },
 [0x1EBD]={
  category="ll",
  direction="l",
 },
 [0x1EBE]={
  category="lu",
  direction="l",
 },
 [0x1EBF]={
  category="ll",
  direction="l",
 },
 [0x1EC0]={
  category="lu",
  direction="l",
 },
 [0x1EC1]={
  category="ll",
  direction="l",
 },
 [0x1EC2]={
  category="lu",
  direction="l",
 },
 [0x1EC3]={
  category="ll",
  direction="l",
 },
 [0x1EC4]={
  category="lu",
  direction="l",
 },
 [0x1EC5]={
  category="ll",
  direction="l",
 },
 [0x1EC6]={
  category="lu",
  direction="l",
 },
 [0x1EC7]={
  category="ll",
  direction="l",
 },
 [0x1EC8]={
  category="lu",
  direction="l",
 },
 [0x1EC9]={
  category="ll",
  direction="l",
 },
 [0x1ECA]={
  category="lu",
  direction="l",
 },
 [0x1ECB]={
  category="ll",
  direction="l",
 },
 [0x1ECC]={
  category="lu",
  direction="l",
 },
 [0x1ECD]={
  category="ll",
  direction="l",
 },
 [0x1ECE]={
  category="lu",
  direction="l",
 },
 [0x1ECF]={
  category="ll",
  direction="l",
 },
 [0x1ED0]={
  category="lu",
  direction="l",
 },
 [0x1ED1]={
  category="ll",
  direction="l",
 },
 [0x1ED2]={
  category="lu",
  direction="l",
 },
 [0x1ED3]={
  category="ll",
  direction="l",
 },
 [0x1ED4]={
  category="lu",
  direction="l",
 },
 [0x1ED5]={
  category="ll",
  direction="l",
 },
 [0x1ED6]={
  category="lu",
  direction="l",
 },
 [0x1ED7]={
  category="ll",
  direction="l",
 },
 [0x1ED8]={
  category="lu",
  direction="l",
 },
 [0x1ED9]={
  category="ll",
  direction="l",
 },
 [0x1EDA]={
  category="lu",
  direction="l",
 },
 [0x1EDB]={
  category="ll",
  direction="l",
 },
 [0x1EDC]={
  category="lu",
  direction="l",
 },
 [0x1EDD]={
  category="ll",
  direction="l",
 },
 [0x1EDE]={
  category="lu",
  direction="l",
 },
 [0x1EDF]={
  category="ll",
  direction="l",
 },
 [0x1EE0]={
  category="lu",
  direction="l",
 },
 [0x1EE1]={
  category="ll",
  direction="l",
 },
 [0x1EE2]={
  category="lu",
  direction="l",
 },
 [0x1EE3]={
  category="ll",
  direction="l",
 },
 [0x1EE4]={
  category="lu",
  direction="l",
 },
 [0x1EE5]={
  category="ll",
  direction="l",
 },
 [0x1EE6]={
  category="lu",
  direction="l",
 },
 [0x1EE7]={
  category="ll",
  direction="l",
 },
 [0x1EE8]={
  category="lu",
  direction="l",
 },
 [0x1EE9]={
  category="ll",
  direction="l",
 },
 [0x1EEA]={
  category="lu",
  direction="l",
 },
 [0x1EEB]={
  category="ll",
  direction="l",
 },
 [0x1EEC]={
  category="lu",
  direction="l",
 },
 [0x1EED]={
  category="ll",
  direction="l",
 },
 [0x1EEE]={
  category="lu",
  direction="l",
 },
 [0x1EEF]={
  category="ll",
  direction="l",
 },
 [0x1EF0]={
  category="lu",
  direction="l",
 },
 [0x1EF1]={
  category="ll",
  direction="l",
 },
 [0x1EF2]={
  category="lu",
  direction="l",
 },
 [0x1EF3]={
  category="ll",
  direction="l",
 },
 [0x1EF4]={
  category="lu",
  direction="l",
 },
 [0x1EF5]={
  category="ll",
  direction="l",
 },
 [0x1EF6]={
  category="lu",
  direction="l",
 },
 [0x1EF7]={
  category="ll",
  direction="l",
 },
 [0x1EF8]={
  category="lu",
  direction="l",
 },
 [0x1EF9]={
  category="ll",
  direction="l",
 },
 [0x1EFA]={
  category="lu",
  direction="l",
 },
 [0x1EFB]={
  category="ll",
  direction="l",
 },
 [0x1EFC]={
  category="lu",
  direction="l",
 },
 [0x1EFD]={
  category="ll",
  direction="l",
 },
 [0x1EFE]={
  category="lu",
  direction="l",
 },
 [0x1EFF]={
  category="ll",
  direction="l",
 },
 [0x1F00]={
  category="ll",
  direction="l",
 },
 [0x1F01]={
  category="ll",
  direction="l",
 },
 [0x1F02]={
  category="ll",
  direction="l",
 },
 [0x1F03]={
  category="ll",
  direction="l",
 },
 [0x1F04]={
  category="ll",
  direction="l",
 },
 [0x1F05]={
  category="ll",
  direction="l",
 },
 [0x1F06]={
  category="ll",
  direction="l",
 },
 [0x1F07]={
  category="ll",
  direction="l",
 },
 [0x1F08]={
  category="lu",
  direction="l",
 },
 [0x1F09]={
  category="lu",
  direction="l",
 },
 [0x1F0A]={
  category="lu",
  direction="l",
 },
 [0x1F0B]={
  category="lu",
  direction="l",
 },
 [0x1F0C]={
  category="lu",
  direction="l",
 },
 [0x1F0D]={
  category="lu",
  direction="l",
 },
 [0x1F0E]={
  category="lu",
  direction="l",
 },
 [0x1F0F]={
  category="lu",
  direction="l",
 },
 [0x1F10]={
  category="ll",
  direction="l",
 },
 [0x1F11]={
  category="ll",
  direction="l",
 },
 [0x1F12]={
  category="ll",
  direction="l",
 },
 [0x1F13]={
  category="ll",
  direction="l",
 },
 [0x1F14]={
  category="ll",
  direction="l",
 },
 [0x1F15]={
  category="ll",
  direction="l",
 },
 [0x1F18]={
  category="lu",
  direction="l",
 },
 [0x1F19]={
  category="lu",
  direction="l",
 },
 [0x1F1A]={
  category="lu",
  direction="l",
 },
 [0x1F1B]={
  category="lu",
  direction="l",
 },
 [0x1F1C]={
  category="lu",
  direction="l",
 },
 [0x1F1D]={
  category="lu",
  direction="l",
 },
 [0x1F20]={
  category="ll",
  direction="l",
 },
 [0x1F21]={
  category="ll",
  direction="l",
 },
 [0x1F22]={
  category="ll",
  direction="l",
 },
 [0x1F23]={
  category="ll",
  direction="l",
 },
 [0x1F24]={
  category="ll",
  direction="l",
 },
 [0x1F25]={
  category="ll",
  direction="l",
 },
 [0x1F26]={
  category="ll",
  direction="l",
 },
 [0x1F27]={
  category="ll",
  direction="l",
 },
 [0x1F28]={
  category="lu",
  direction="l",
 },
 [0x1F29]={
  category="lu",
  direction="l",
 },
 [0x1F2A]={
  category="lu",
  direction="l",
 },
 [0x1F2B]={
  category="lu",
  direction="l",
 },
 [0x1F2C]={
  category="lu",
  direction="l",
 },
 [0x1F2D]={
  category="lu",
  direction="l",
 },
 [0x1F2E]={
  category="lu",
  direction="l",
 },
 [0x1F2F]={
  category="lu",
  direction="l",
 },
 [0x1F30]={
  category="ll",
  direction="l",
 },
 [0x1F31]={
  category="ll",
  direction="l",
 },
 [0x1F32]={
  category="ll",
  direction="l",
 },
 [0x1F33]={
  category="ll",
  direction="l",
 },
 [0x1F34]={
  category="ll",
  direction="l",
 },
 [0x1F35]={
  category="ll",
  direction="l",
 },
 [0x1F36]={
  category="ll",
  direction="l",
 },
 [0x1F37]={
  category="ll",
  direction="l",
 },
 [0x1F38]={
  category="lu",
  direction="l",
 },
 [0x1F39]={
  category="lu",
  direction="l",
 },
 [0x1F3A]={
  category="lu",
  direction="l",
 },
 [0x1F3B]={
  category="lu",
  direction="l",
 },
 [0x1F3C]={
  category="lu",
  direction="l",
 },
 [0x1F3D]={
  category="lu",
  direction="l",
 },
 [0x1F3E]={
  category="lu",
  direction="l",
 },
 [0x1F3F]={
  category="lu",
  direction="l",
 },
 [0x1F40]={
  category="ll",
  direction="l",
 },
 [0x1F41]={
  category="ll",
  direction="l",
 },
 [0x1F42]={
  category="ll",
  direction="l",
 },
 [0x1F43]={
  category="ll",
  direction="l",
 },
 [0x1F44]={
  category="ll",
  direction="l",
 },
 [0x1F45]={
  category="ll",
  direction="l",
 },
 [0x1F48]={
  category="lu",
  direction="l",
 },
 [0x1F49]={
  category="lu",
  direction="l",
 },
 [0x1F4A]={
  category="lu",
  direction="l",
 },
 [0x1F4B]={
  category="lu",
  direction="l",
 },
 [0x1F4C]={
  category="lu",
  direction="l",
 },
 [0x1F4D]={
  category="lu",
  direction="l",
 },
 [0x1F50]={
  category="ll",
  direction="l",
 },
 [0x1F51]={
  category="ll",
  direction="l",
 },
 [0x1F52]={
  category="ll",
  direction="l",
 },
 [0x1F53]={
  category="ll",
  direction="l",
 },
 [0x1F54]={
  category="ll",
  direction="l",
 },
 [0x1F55]={
  category="ll",
  direction="l",
 },
 [0x1F56]={
  category="ll",
  direction="l",
 },
 [0x1F57]={
  category="ll",
  direction="l",
 },
 [0x1F59]={
  category="lu",
  direction="l",
 },
 [0x1F5B]={
  category="lu",
  direction="l",
 },
 [0x1F5D]={
  category="lu",
  direction="l",
 },
 [0x1F5F]={
  category="lu",
  direction="l",
 },
 [0x1F60]={
  category="ll",
  direction="l",
 },
 [0x1F61]={
  category="ll",
  direction="l",
 },
 [0x1F62]={
  category="ll",
  direction="l",
 },
 [0x1F63]={
  category="ll",
  direction="l",
 },
 [0x1F64]={
  category="ll",
  direction="l",
 },
 [0x1F65]={
  category="ll",
  direction="l",
 },
 [0x1F66]={
  category="ll",
  direction="l",
 },
 [0x1F67]={
  category="ll",
  direction="l",
 },
 [0x1F68]={
  category="lu",
  direction="l",
 },
 [0x1F69]={
  category="lu",
  direction="l",
 },
 [0x1F6A]={
  category="lu",
  direction="l",
 },
 [0x1F6B]={
  category="lu",
  direction="l",
 },
 [0x1F6C]={
  category="lu",
  direction="l",
 },
 [0x1F6D]={
  category="lu",
  direction="l",
 },
 [0x1F6E]={
  category="lu",
  direction="l",
 },
 [0x1F6F]={
  category="lu",
  direction="l",
 },
 [0x1F70]={
  category="ll",
  direction="l",
 },
 [0x1F71]={
  category="ll",
  direction="l",
 },
 [0x1F72]={
  category="ll",
  direction="l",
 },
 [0x1F73]={
  category="ll",
  direction="l",
 },
 [0x1F74]={
  category="ll",
  direction="l",
 },
 [0x1F75]={
  category="ll",
  direction="l",
 },
 [0x1F76]={
  category="ll",
  direction="l",
 },
 [0x1F77]={
  category="ll",
  direction="l",
 },
 [0x1F78]={
  category="ll",
  direction="l",
 },
 [0x1F79]={
  category="ll",
  direction="l",
 },
 [0x1F7A]={
  category="ll",
  direction="l",
 },
 [0x1F7B]={
  category="ll",
  direction="l",
 },
 [0x1F7C]={
  category="ll",
  direction="l",
 },
 [0x1F7D]={
  category="ll",
  direction="l",
 },
 [0x1F80]={
  category="ll",
  direction="l",
 },
 [0x1F81]={
  category="ll",
  direction="l",
 },
 [0x1F82]={
  category="ll",
  direction="l",
 },
 [0x1F83]={
  category="ll",
  direction="l",
 },
 [0x1F84]={
  category="ll",
  direction="l",
 },
 [0x1F85]={
  category="ll",
  direction="l",
 },
 [0x1F86]={
  category="ll",
  direction="l",
 },
 [0x1F87]={
  category="ll",
  direction="l",
 },
 [0x1F88]={
  category="lt",
  direction="l",
 },
 [0x1F89]={
  category="lt",
  direction="l",
 },
 [0x1F8A]={
  category="lt",
  direction="l",
 },
 [0x1F8B]={
  category="lt",
  direction="l",
 },
 [0x1F8C]={
  category="lt",
  direction="l",
 },
 [0x1F8D]={
  category="lt",
  direction="l",
 },
 [0x1F8E]={
  category="lt",
  direction="l",
 },
 [0x1F8F]={
  category="lt",
  direction="l",
 },
 [0x1F90]={
  category="ll",
  direction="l",
 },
 [0x1F91]={
  category="ll",
  direction="l",
 },
 [0x1F92]={
  category="ll",
  direction="l",
 },
 [0x1F93]={
  category="ll",
  direction="l",
 },
 [0x1F94]={
  category="ll",
  direction="l",
 },
 [0x1F95]={
  category="ll",
  direction="l",
 },
 [0x1F96]={
  category="ll",
  direction="l",
 },
 [0x1F97]={
  category="ll",
  direction="l",
 },
 [0x1F98]={
  category="lt",
  direction="l",
 },
 [0x1F99]={
  category="lt",
  direction="l",
 },
 [0x1F9A]={
  category="lt",
  direction="l",
 },
 [0x1F9B]={
  category="lt",
  direction="l",
 },
 [0x1F9C]={
  category="lt",
  direction="l",
 },
 [0x1F9D]={
  category="lt",
  direction="l",
 },
 [0x1F9E]={
  category="lt",
  direction="l",
 },
 [0x1F9F]={
  category="lt",
  direction="l",
 },
 [0x1FA0]={
  category="ll",
  direction="l",
 },
 [0x1FA1]={
  category="ll",
  direction="l",
 },
 [0x1FA2]={
  category="ll",
  direction="l",
 },
 [0x1FA3]={
  category="ll",
  direction="l",
 },
 [0x1FA4]={
  category="ll",
  direction="l",
 },
 [0x1FA5]={
  category="ll",
  direction="l",
 },
 [0x1FA6]={
  category="ll",
  direction="l",
 },
 [0x1FA7]={
  category="ll",
  direction="l",
 },
 [0x1FA8]={
  category="lt",
  direction="l",
 },
 [0x1FA9]={
  category="lt",
  direction="l",
 },
 [0x1FAA]={
  category="lt",
  direction="l",
 },
 [0x1FAB]={
  category="lt",
  direction="l",
 },
 [0x1FAC]={
  category="lt",
  direction="l",
 },
 [0x1FAD]={
  category="lt",
  direction="l",
 },
 [0x1FAE]={
  category="lt",
  direction="l",
 },
 [0x1FAF]={
  category="lt",
  direction="l",
 },
 [0x1FB0]={
  category="ll",
  direction="l",
 },
 [0x1FB1]={
  category="ll",
  direction="l",
 },
 [0x1FB2]={
  category="ll",
  direction="l",
 },
 [0x1FB3]={
  category="ll",
  direction="l",
 },
 [0x1FB4]={
  category="ll",
  direction="l",
 },
 [0x1FB6]={
  category="ll",
  direction="l",
 },
 [0x1FB7]={
  category="ll",
  direction="l",
 },
 [0x1FB8]={
  category="lu",
  direction="l",
 },
 [0x1FB9]={
  category="lu",
  direction="l",
 },
 [0x1FBA]={
  category="lu",
  direction="l",
 },
 [0x1FBB]={
  category="lu",
  direction="l",
 },
 [0x1FBC]={
  category="lt",
  direction="l",
 },
 [0x1FBD]={
  category="sk",
  direction="on",
 },
 [0x1FBE]={
  category="ll",
  direction="l",
 },
 [0x1FBF]={
  category="sk",
  direction="on",
 },
 [0x1FC0]={
  category="sk",
  direction="on",
 },
 [0x1FC1]={
  category="sk",
  direction="on",
 },
 [0x1FC2]={
  category="ll",
  direction="l",
 },
 [0x1FC3]={
  category="ll",
  direction="l",
 },
 [0x1FC4]={
  category="ll",
  direction="l",
 },
 [0x1FC6]={
  category="ll",
  direction="l",
 },
 [0x1FC7]={
  category="ll",
  direction="l",
 },
 [0x1FC8]={
  category="lu",
  direction="l",
 },
 [0x1FC9]={
  category="lu",
  direction="l",
 },
 [0x1FCA]={
  category="lu",
  direction="l",
 },
 [0x1FCB]={
  category="lu",
  direction="l",
 },
 [0x1FCC]={
  category="lt",
  direction="l",
 },
 [0x1FCD]={
  category="sk",
  direction="on",
 },
 [0x1FCE]={
  category="sk",
  direction="on",
 },
 [0x1FCF]={
  category="sk",
  direction="on",
 },
 [0x1FD0]={
  category="ll",
  direction="l",
 },
 [0x1FD1]={
  category="ll",
  direction="l",
 },
 [0x1FD2]={
  category="ll",
  direction="l",
 },
 [0x1FD3]={
  category="ll",
  direction="l",
 },
 [0x1FD6]={
  category="ll",
  direction="l",
 },
 [0x1FD7]={
  category="ll",
  direction="l",
 },
 [0x1FD8]={
  category="lu",
  direction="l",
 },
 [0x1FD9]={
  category="lu",
  direction="l",
 },
 [0x1FDA]={
  category="lu",
  direction="l",
 },
 [0x1FDB]={
  category="lu",
  direction="l",
 },
 [0x1FDD]={
  category="sk",
  direction="on",
 },
 [0x1FDE]={
  category="sk",
  direction="on",
 },
 [0x1FDF]={
  category="sk",
  direction="on",
 },
 [0x1FE0]={
  category="ll",
  direction="l",
 },
 [0x1FE1]={
  category="ll",
  direction="l",
 },
 [0x1FE2]={
  category="ll",
  direction="l",
 },
 [0x1FE3]={
  category="ll",
  direction="l",
 },
 [0x1FE4]={
  category="ll",
  direction="l",
 },
 [0x1FE5]={
  category="ll",
  direction="l",
 },
 [0x1FE6]={
  category="ll",
  direction="l",
 },
 [0x1FE7]={
  category="ll",
  direction="l",
 },
 [0x1FE8]={
  category="lu",
  direction="l",
 },
 [0x1FE9]={
  category="lu",
  direction="l",
 },
 [0x1FEA]={
  category="lu",
  direction="l",
 },
 [0x1FEB]={
  category="lu",
  direction="l",
 },
 [0x1FEC]={
  category="lu",
  direction="l",
 },
 [0x1FED]={
  category="sk",
  direction="on",
 },
 [0x1FEE]={
  category="sk",
  direction="on",
 },
 [0x1FEF]={
  category="sk",
  direction="on",
 },
 [0x1FF2]={
  category="ll",
  direction="l",
 },
 [0x1FF3]={
  category="ll",
  direction="l",
 },
 [0x1FF4]={
  category="ll",
  direction="l",
 },
 [0x1FF6]={
  category="ll",
  direction="l",
 },
 [0x1FF7]={
  category="ll",
  direction="l",
 },
 [0x1FF8]={
  category="lu",
  direction="l",
 },
 [0x1FF9]={
  category="lu",
  direction="l",
 },
 [0x1FFA]={
  category="lu",
  direction="l",
 },
 [0x1FFB]={
  category="lu",
  direction="l",
 },
 [0x1FFC]={
  category="lt",
  direction="l",
 },
 [0x1FFD]={
  category="sk",
  direction="on",
 },
 [0x1FFE]={
  category="sk",
  direction="on",
 },
 [0x2000]={
  category="zs",
  direction="ws",
 },
 [0x2001]={
  category="zs",
  direction="ws",
 },
 [0x2002]={
  category="zs",
  direction="ws",
 },
 [0x2003]={
  category="zs",
  direction="ws",
 },
 [0x2004]={
  category="zs",
  direction="ws",
 },
 [0x2005]={
  category="zs",
  direction="ws",
 },
 [0x2006]={
  category="zs",
  direction="ws",
 },
 [0x2007]={
  category="zs",
  direction="ws",
 },
 [0x2008]={
  category="zs",
  direction="ws",
 },
 [0x2009]={
  category="zs",
  direction="ws",
 },
 [0x200A]={
  category="zs",
  direction="ws",
 },
 [0x200B]={
  category="cf",
  direction="bn",
 },
 [0x200C]={
  category="cf",
  direction="bn",
 },
 [0x200D]={
  category="cf",
  direction="bn",
 },
 [0x200E]={
  category="cf",
  direction="l",
 },
 [0x200F]={
  category="cf",
  direction="r",
 },
 [0x2010]={
  category="pd",
  direction="on",
 },
 [0x2011]={
  category="pd",
  direction="on",
 },
 [0x2012]={
  category="pd",
  direction="on",
 },
 [0x2013]={
  category="pd",
  direction="on",
 },
 [0x2014]={
  category="pd",
  direction="on",
 },
 [0x2015]={
  category="pd",
  direction="on",
 },
 [0x2016]={
  category="po",
  direction="on",
 },
 [0x2017]={
  category="po",
  direction="on",
 },
 [0x2018]={
  category="pi",
  direction="on",
 },
 [0x2019]={
  category="pf",
  direction="on",
 },
 [0x201A]={
  category="ps",
  direction="on",
 },
 [0x201B]={
  category="pi",
  direction="on",
 },
 [0x201C]={
  category="pi",
  direction="on",
 },
 [0x201D]={
  category="pf",
  direction="on",
 },
 [0x201E]={
  category="ps",
  direction="on",
 },
 [0x201F]={
  category="pi",
  direction="on",
 },
 [0x2020]={
  category="po",
  direction="on",
 },
 [0x2021]={
  category="po",
  direction="on",
 },
 [0x2022]={
  category="po",
  direction="on",
 },
 [0x2023]={
  category="po",
  direction="on",
 },
 [0x2024]={
  category="po",
  direction="on",
 },
 [0x2025]={
  category="po",
  direction="on",
 },
 [0x2026]={
  category="po",
  direction="on",
 },
 [0x2027]={
  category="po",
  direction="on",
 },
 [0x2028]={
  category="zl",
  direction="ws",
 },
 [0x2029]={
  category="zp",
  direction="b",
 },
 [0x202A]={
  category="cf",
  direction="lre",
 },
 [0x202B]={
  category="cf",
  direction="rle",
 },
 [0x202C]={
  category="cf",
  direction="pdf",
 },
 [0x202D]={
  category="cf",
  direction="lro",
 },
 [0x202E]={
  category="cf",
  direction="rlo",
 },
 [0x202F]={
  category="zs",
  direction="cs",
 },
 [0x2030]={
  category="po",
  direction="et",
 },
 [0x2031]={
  category="po",
  direction="et",
 },
 [0x2032]={
  category="po",
  direction="et",
 },
 [0x2033]={
  category="po",
  direction="et",
 },
 [0x2034]={
  category="po",
  direction="et",
 },
 [0x2035]={
  category="po",
  direction="on",
 },
 [0x2036]={
  category="po",
  direction="on",
 },
 [0x2037]={
  category="po",
  direction="on",
 },
 [0x2038]={
  category="po",
  direction="on",
 },
 [0x2039]={
  category="pi",
  direction="on",
  mirror=0x203A,
  textclass="open",
 },
 [0x203A]={
  category="pf",
  direction="on",
  mirror=0x2039,
  textclass="close",
 },
 [0x203B]={
  category="po",
  direction="on",
 },
 [0x203C]={
  category="po",
  direction="on",
 },
 [0x203D]={
  category="po",
  direction="on",
 },
 [0x203E]={
  category="po",
  direction="on",
 },
 [0x203F]={
  category="pc",
  direction="on",
 },
 [0x2040]={
  category="pc",
  direction="on",
 },
 [0x2041]={
  category="po",
  direction="on",
 },
 [0x2042]={
  category="po",
  direction="on",
 },
 [0x2043]={
  category="po",
  direction="on",
 },
 [0x2044]={
  category="sm",
  direction="cs",
 },
 [0x2045]={
  category="ps",
  direction="on",
  mirror=0x2046,
 },
 [0x2046]={
  category="pe",
  direction="on",
  mirror=0x2045,
 },
 [0x2047]={
  category="po",
  direction="on",
 },
 [0x2048]={
  category="po",
  direction="on",
 },
 [0x2049]={
  category="po",
  direction="on",
 },
 [0x204A]={
  category="po",
  direction="on",
 },
 [0x204B]={
  category="po",
  direction="on",
 },
 [0x204C]={
  category="po",
  direction="on",
 },
 [0x204D]={
  category="po",
  direction="on",
 },
 [0x204E]={
  category="po",
  direction="on",
 },
 [0x204F]={
  category="po",
  direction="on",
 },
 [0x2050]={
  category="po",
  direction="on",
 },
 [0x2051]={
  category="po",
  direction="on",
 },
 [0x2052]={
  category="sm",
  direction="on",
 },
 [0x2053]={
  category="po",
  direction="on",
 },
 [0x2054]={
  category="pc",
  direction="on",
 },
 [0x2055]={
  category="po",
  direction="on",
 },
 [0x2056]={
  category="po",
  direction="on",
 },
 [0x2057]={
  category="po",
  direction="on",
 },
 [0x2058]={
  category="po",
  direction="on",
 },
 [0x2059]={
  category="po",
  direction="on",
 },
 [0x205A]={
  category="po",
  direction="on",
 },
 [0x205B]={
  category="po",
  direction="on",
 },
 [0x205C]={
  category="po",
  direction="on",
 },
 [0x205D]={
  category="po",
  direction="on",
 },
 [0x205E]={
  category="po",
  direction="on",
 },
 [0x205F]={
  category="zs",
  direction="ws",
 },
 [0x2060]={
  category="cf",
  direction="bn",
 },
 [0x2061]={
  category="cf",
  direction="bn",
 },
 [0x2062]={
  category="cf",
  direction="bn",
 },
 [0x2063]={
  category="cf",
  direction="bn",
 },
 [0x2064]={
  category="cf",
  direction="bn",
 },
 [0x2066]={
  category="cf",
  direction="lri",
 },
 [0x2067]={
  category="cf",
  direction="rli",
 },
 [0x2068]={
  category="cf",
  direction="fsi",
 },
 [0x2069]={
  category="cf",
  direction="pdi",
 },
 [0x206A]={
  category="cf",
  direction="bn",
 },
 [0x206B]={
  category="cf",
  direction="bn",
 },
 [0x206C]={
  category="cf",
  direction="bn",
 },
 [0x206D]={
  category="cf",
  direction="bn",
 },
 [0x206E]={
  category="cf",
  direction="bn",
 },
 [0x206F]={
  category="cf",
  direction="bn",
 },
 [0x2070]={
  category="no",
  direction="en",
 },
 [0x2071]={
  category="ll",
  direction="l",
 },
 [0x2074]={
  category="no",
  direction="en",
 },
 [0x2075]={
  category="no",
  direction="en",
 },
 [0x2076]={
  category="no",
  direction="en",
 },
 [0x2077]={
  category="no",
  direction="en",
 },
 [0x2078]={
  category="no",
  direction="en",
 },
 [0x2079]={
  category="no",
  direction="en",
 },
 [0x207A]={
  category="sm",
  direction="es",
 },
 [0x207B]={
  category="sm",
  direction="es",
 },
 [0x207C]={
  category="sm",
  direction="on",
 },
 [0x207D]={
  category="ps",
  direction="on",
  mirror=0x207E,
  textclass="open",
 },
 [0x207E]={
  category="pe",
  direction="on",
  mirror=0x207D,
  textclass="close",
 },
 [0x207F]={
  category="ll",
  direction="l",
 },
 [0x2080]={
  category="no",
  direction="en",
 },
 [0x2081]={
  category="no",
  direction="en",
 },
 [0x2082]={
  category="no",
  direction="en",
 },
 [0x2083]={
  category="no",
  direction="en",
 },
 [0x2084]={
  category="no",
  direction="en",
 },
 [0x2085]={
  category="no",
  direction="en",
 },
 [0x2086]={
  category="no",
  direction="en",
 },
 [0x2087]={
  category="no",
  direction="en",
 },
 [0x2088]={
  category="no",
  direction="en",
 },
 [0x2089]={
  category="no",
  direction="en",
 },
 [0x208A]={
  category="sm",
  direction="es",
 },
 [0x208B]={
  category="sm",
  direction="es",
 },
 [0x208C]={
  category="sm",
  direction="on",
 },
 [0x208D]={
  category="ps",
  direction="on",
  mirror=0x208E,
  textclass="open",
 },
 [0x208E]={
  category="pe",
  direction="on",
  mirror=0x208D,
  textclass="close",
 },
 [0x2090]={
  category="lm",
  direction="l",
 },
 [0x2091]={
  category="lm",
  direction="l",
 },
 [0x2092]={
  category="lm",
  direction="l",
 },
 [0x2093]={
  category="lm",
  direction="l",
 },
 [0x2094]={
  category="lm",
  direction="l",
 },
 [0x2095]={
  category="lm",
  direction="l",
 },
 [0x2096]={
  category="lm",
  direction="l",
 },
 [0x2097]={
  category="lm",
  direction="l",
 },
 [0x2098]={
  category="lm",
  direction="l",
 },
 [0x2099]={
  category="lm",
  direction="l",
 },
 [0x209A]={
  category="lm",
  direction="l",
 },
 [0x209B]={
  category="lm",
  direction="l",
 },
 [0x209C]={
  category="lm",
  direction="l",
 },
 [0x20A0]={
  category="sc",
  direction="et",
 },
 [0x20A1]={
  category="sc",
  direction="et",
 },
 [0x20A2]={
  category="sc",
  direction="et",
 },
 [0x20A3]={
  category="sc",
  direction="et",
 },
 [0x20A4]={
  category="sc",
  direction="et",
 },
 [0x20A5]={
  category="sc",
  direction="et",
 },
 [0x20A6]={
  category="sc",
  direction="et",
 },
 [0x20A7]={
  category="sc",
  direction="et",
 },
 [0x20A8]={
  category="sc",
  direction="et",
 },
 [0x20A9]={
  category="sc",
  direction="et",
 },
 [0x20AA]={
  category="sc",
  direction="et",
 },
 [0x20AB]={
  category="sc",
  direction="et",
 },
 [0x20AC]={
  category="sc",
  direction="et",
 },
 [0x20AD]={
  category="sc",
  direction="et",
 },
 [0x20AE]={
  category="sc",
  direction="et",
 },
 [0x20AF]={
  category="sc",
  direction="et",
 },
 [0x20B0]={
  category="sc",
  direction="et",
 },
 [0x20B1]={
  category="sc",
  direction="et",
 },
 [0x20B2]={
  category="sc",
  direction="et",
 },
 [0x20B3]={
  category="sc",
  direction="et",
 },
 [0x20B4]={
  category="sc",
  direction="et",
 },
 [0x20B5]={
  category="sc",
  direction="et",
 },
 [0x20B6]={
  category="sc",
  direction="et",
 },
 [0x20B7]={
  category="sc",
  direction="et",
 },
 [0x20B8]={
  category="sc",
  direction="et",
 },
 [0x20B9]={
  category="sc",
  direction="et",
 },
 [0x20BA]={
  category="sc",
  direction="et",
 },
 [0x20BB]={
  category="sc",
  direction="et",
 },
 [0x20BC]={
  category="sc",
  direction="et",
 },
 [0x20BD]={
  category="sc",
  direction="et",
 },
 [0x20D0]={
  category="mn",
  direction="nsm",
 },
 [0x20D1]={
  category="mn",
  direction="nsm",
 },
 [0x20D2]={
  category="mn",
  direction="nsm",
 },
 [0x20D3]={
  category="mn",
  direction="nsm",
 },
 [0x20D4]={
  category="mn",
  direction="nsm",
 },
 [0x20D5]={
  category="mn",
  direction="nsm",
 },
 [0x20D6]={
  category="mn",
  direction="nsm",
 },
 [0x20D7]={
  category="mn",
  direction="nsm",
 },
 [0x20D8]={
  category="mn",
  direction="nsm",
 },
 [0x20D9]={
  category="mn",
  direction="nsm",
 },
 [0x20DA]={
  category="mn",
  direction="nsm",
 },
 [0x20DB]={
  category="mn",
  direction="nsm",
 },
 [0x20DC]={
  category="mn",
  direction="nsm",
 },
 [0x20DD]={
  category="me",
  direction="nsm",
 },
 [0x20DE]={
  category="me",
  direction="nsm",
 },
 [0x20DF]={
  category="me",
  direction="nsm",
 },
 [0x20E0]={
  category="me",
  direction="nsm",
 },
 [0x20E1]={
  category="mn",
  direction="nsm",
 },
 [0x20E2]={
  category="me",
  direction="nsm",
 },
 [0x20E3]={
  category="me",
  direction="nsm",
 },
 [0x20E4]={
  category="me",
  direction="nsm",
 },
 [0x20E5]={
  category="mn",
  direction="nsm",
 },
 [0x20E6]={
  category="mn",
  direction="nsm",
 },
 [0x20E7]={
  category="mn",
  direction="nsm",
 },
 [0x20E8]={
  category="mn",
  direction="nsm",
 },
 [0x20E9]={
  category="mn",
  direction="nsm",
 },
 [0x20EA]={
  category="mn",
  direction="nsm",
 },
 [0x20EB]={
  category="mn",
  direction="nsm",
 },
 [0x20EC]={
  category="mn",
  direction="nsm",
 },
 [0x20ED]={
  category="mn",
  direction="nsm",
 },
 [0x20EE]={
  category="mn",
  direction="nsm",
 },
 [0x20EF]={
  category="mn",
  direction="nsm",
 },
 [0x20F0]={
  category="mn",
  direction="nsm",
 },
 [0x2100]={
  category="so",
  direction="on",
 },
 [0x2101]={
  category="so",
  direction="on",
 },
 [0x2102]={
  category="lu",
  direction="l",
 },
 [0x2103]={
  category="so",
  direction="on",
 },
 [0x2104]={
  category="so",
  direction="on",
 },
 [0x2105]={
  category="so",
  direction="on",
 },
 [0x2106]={
  category="so",
  direction="on",
 },
 [0x2107]={
  category="lu",
  direction="l",
 },
 [0x2108]={
  category="so",
  direction="on",
 },
 [0x2109]={
  category="so",
  direction="on",
 },
 [0x210A]={
  category="ll",
  direction="l",
 },
 [0x210B]={
  category="lu",
  direction="l",
 },
 [0x210C]={
  category="lu",
  direction="l",
 },
 [0x210D]={
  category="lu",
  direction="l",
 },
 [0x210E]={
  category="ll",
  direction="l",
 },
 [0x210F]={
  category="ll",
  direction="l",
 },
 [0x2110]={
  category="lu",
  direction="l",
 },
 [0x2111]={
  category="lu",
  direction="l",
 },
 [0x2112]={
  category="lu",
  direction="l",
 },
 [0x2113]={
  category="ll",
  direction="l",
 },
 [0x2114]={
  category="so",
  direction="on",
 },
 [0x2115]={
  category="lu",
  direction="l",
 },
 [0x2116]={
  category="so",
  direction="on",
 },
 [0x2117]={
  category="so",
  direction="on",
 },
 [0x2118]={
  category="so",
  direction="on",
 },
 [0x2119]={
  category="lu",
  direction="l",
 },
 [0x211A]={
  category="lu",
  direction="l",
 },
 [0x211B]={
  category="lu",
  direction="l",
 },
 [0x211C]={
  category="lu",
  direction="l",
 },
 [0x211D]={
  category="lu",
  direction="l",
 },
 [0x211E]={
  category="so",
  direction="on",
 },
 [0x211F]={
  category="so",
  direction="on",
 },
 [0x2120]={
  category="so",
  direction="on",
 },
 [0x2121]={
  category="so",
  direction="on",
 },
 [0x2122]={
  category="so",
  direction="on",
 },
 [0x2123]={
  category="so",
  direction="on",
 },
 [0x2124]={
  category="lu",
  direction="l",
 },
 [0x2125]={
  category="so",
  direction="on",
 },
 [0x2126]={
  category="lu",
  direction="l",
 },
 [0x2127]={
  category="so",
  direction="on",
 },
 [0x2128]={
  category="lu",
  direction="l",
 },
 [0x2129]={
  category="so",
  direction="on",
 },
 [0x212A]={
  category="lu",
  direction="l",
 },
 [0x212B]={
  category="lu",
  direction="l",
 },
 [0x212C]={
  category="lu",
  direction="l",
 },
 [0x212D]={
  category="lu",
  direction="l",
 },
 [0x212E]={
  category="so",
  direction="et",
 },
 [0x212F]={
  category="ll",
  direction="l",
 },
 [0x2130]={
  category="lu",
  direction="l",
 },
 [0x2131]={
  category="lu",
  direction="l",
 },
 [0x2132]={
  category="lu",
  direction="l",
 },
 [0x2133]={
  category="lu",
  direction="l",
 },
 [0x2134]={
  category="ll",
  direction="l",
 },
 [0x2135]={
  category="lo",
  direction="l",
 },
 [0x2136]={
  category="lo",
  direction="l",
 },
 [0x2137]={
  category="lo",
  direction="l",
 },
 [0x2138]={
  category="lo",
  direction="l",
 },
 [0x2139]={
  category="ll",
  direction="l",
 },
 [0x213A]={
  category="so",
  direction="on",
 },
 [0x213B]={
  category="so",
  direction="on",
 },
 [0x213C]={
  category="ll",
  direction="l",
 },
 [0x213D]={
  category="ll",
  direction="l",
 },
 [0x213E]={
  category="lu",
  direction="l",
 },
 [0x213F]={
  category="lu",
  direction="l",
 },
 [0x2140]={
  category="sm",
  direction="on",
 },
 [0x2141]={
  category="sm",
  direction="on",
 },
 [0x2142]={
  category="sm",
  direction="on",
 },
 [0x2143]={
  category="sm",
  direction="on",
 },
 [0x2144]={
  category="sm",
  direction="on",
 },
 [0x2145]={
  category="lu",
  direction="l",
 },
 [0x2146]={
  category="ll",
  direction="l",
 },
 [0x2147]={
  category="ll",
  direction="l",
 },
 [0x2148]={
  category="ll",
  direction="l",
 },
 [0x2149]={
  category="ll",
  direction="l",
 },
 [0x214A]={
  category="so",
  direction="on",
 },
 [0x214B]={
  category="sm",
  direction="on",
 },
 [0x214C]={
  category="so",
  direction="on",
 },
 [0x214D]={
  category="so",
  direction="on",
 },
 [0x214E]={
  category="ll",
  direction="l",
 },
 [0x214F]={
  category="so",
  direction="l",
 },
 [0x2150]={
  category="no",
  direction="on",
 },
 [0x2151]={
  category="no",
  direction="on",
 },
 [0x2152]={
  category="no",
  direction="on",
 },
 [0x2153]={
  category="no",
  direction="on",
 },
 [0x2154]={
  category="no",
  direction="on",
 },
 [0x2155]={
  category="no",
  direction="on",
 },
 [0x2156]={
  category="no",
  direction="on",
 },
 [0x2157]={
  category="no",
  direction="on",
 },
 [0x2158]={
  category="no",
  direction="on",
 },
 [0x2159]={
  category="no",
  direction="on",
 },
 [0x215A]={
  category="no",
  direction="on",
 },
 [0x215B]={
  category="no",
  direction="on",
 },
 [0x215C]={
  category="no",
  direction="on",
 },
 [0x215D]={
  category="no",
  direction="on",
 },
 [0x215E]={
  category="no",
  direction="on",
 },
 [0x215F]={
  category="no",
  direction="on",
 },
 [0x2160]={
  category="nl",
  direction="l",
 },
 [0x2161]={
  category="nl",
  direction="l",
 },
 [0x2162]={
  category="nl",
  direction="l",
 },
 [0x2163]={
  category="nl",
  direction="l",
 },
 [0x2164]={
  category="nl",
  direction="l",
 },
 [0x2165]={
  category="nl",
  direction="l",
 },
 [0x2166]={
  category="nl",
  direction="l",
 },
 [0x2167]={
  category="nl",
  direction="l",
 },
 [0x2168]={
  category="nl",
  direction="l",
 },
 [0x2169]={
  category="nl",
  direction="l",
 },
 [0x216A]={
  category="nl",
  direction="l",
 },
 [0x216B]={
  category="nl",
  direction="l",
 },
 [0x216C]={
  category="nl",
  direction="l",
 },
 [0x216D]={
  category="nl",
  direction="l",
 },
 [0x216E]={
  category="nl",
  direction="l",
 },
 [0x216F]={
  category="nl",
  direction="l",
 },
 [0x2170]={
  category="nl",
  direction="l",
 },
 [0x2171]={
  category="nl",
  direction="l",
 },
 [0x2172]={
  category="nl",
  direction="l",
 },
 [0x2173]={
  category="nl",
  direction="l",
 },
 [0x2174]={
  category="nl",
  direction="l",
 },
 [0x2175]={
  category="nl",
  direction="l",
 },
 [0x2176]={
  category="nl",
  direction="l",
 },
 [0x2177]={
  category="nl",
  direction="l",
 },
 [0x2178]={
  category="nl",
  direction="l",
 },
 [0x2179]={
  category="nl",
  direction="l",
 },
 [0x217A]={
  category="nl",
  direction="l",
 },
 [0x217B]={
  category="nl",
  direction="l",
 },
 [0x217C]={
  category="nl",
  direction="l",
 },
 [0x217D]={
  category="nl",
  direction="l",
 },
 [0x217E]={
  category="nl",
  direction="l",
 },
 [0x217F]={
  category="nl",
  direction="l",
 },
 [0x2180]={
  category="nl",
  direction="l",
 },
 [0x2181]={
  category="nl",
  direction="l",
 },
 [0x2182]={
  category="nl",
  direction="l",
 },
 [0x2183]={
  category="lu",
  direction="l",
 },
 [0x2184]={
  category="ll",
  direction="l",
 },
 [0x2185]={
  category="nl",
  direction="l",
 },
 [0x2186]={
  category="nl",
  direction="l",
 },
 [0x2187]={
  category="nl",
  direction="l",
 },
 [0x2188]={
  category="nl",
  direction="l",
 },
 [0x2189]={
  category="no",
  direction="on",
 },
 [0x2190]={
  category="sm",
  direction="on",
 },
 [0x2191]={
  category="sm",
  direction="on",
 },
 [0x2192]={
  category="sm",
  direction="on",
 },
 [0x2193]={
  category="sm",
  direction="on",
 },
 [0x2194]={
  category="sm",
  direction="on",
 },
 [0x2195]={
  category="so",
  direction="on",
 },
 [0x2196]={
  category="so",
  direction="on",
 },
 [0x2197]={
  category="so",
  direction="on",
 },
 [0x2198]={
  category="so",
  direction="on",
 },
 [0x2199]={
  category="so",
  direction="on",
 },
 [0x219A]={
  category="sm",
  direction="on",
 },
 [0x219B]={
  category="sm",
  direction="on",
 },
 [0x219C]={
  category="so",
  direction="on",
 },
 [0x219D]={
  category="so",
  direction="on",
 },
 [0x219E]={
  category="so",
  direction="on",
 },
 [0x219F]={
  category="so",
  direction="on",
 },
 [0x21A0]={
  category="sm",
  direction="on",
 },
 [0x21A1]={
  category="so",
  direction="on",
 },
 [0x21A2]={
  category="so",
  direction="on",
 },
 [0x21A3]={
  category="sm",
  direction="on",
 },
 [0x21A4]={
  category="so",
  direction="on",
 },
 [0x21A5]={
  category="so",
  direction="on",
 },
 [0x21A6]={
  category="sm",
  direction="on",
 },
 [0x21A7]={
  category="so",
  direction="on",
 },
 [0x21A8]={
  category="so",
  direction="on",
 },
 [0x21A9]={
  category="so",
  direction="on",
 },
 [0x21AA]={
  category="so",
  direction="on",
 },
 [0x21AB]={
  category="so",
  direction="on",
 },
 [0x21AC]={
  category="so",
  direction="on",
 },
 [0x21AD]={
  category="so",
  direction="on",
 },
 [0x21AE]={
  category="sm",
  direction="on",
 },
 [0x21AF]={
  category="so",
  direction="on",
 },
 [0x21B0]={
  category="so",
  direction="on",
 },
 [0x21B1]={
  category="so",
  direction="on",
 },
 [0x21B2]={
  category="so",
  direction="on",
 },
 [0x21B3]={
  category="so",
  direction="on",
 },
 [0x21B4]={
  category="so",
  direction="on",
 },
 [0x21B5]={
  category="so",
  direction="on",
 },
 [0x21B6]={
  category="so",
  direction="on",
 },
 [0x21B7]={
  category="so",
  direction="on",
 },
 [0x21B8]={
  category="so",
  direction="on",
 },
 [0x21B9]={
  category="so",
  direction="on",
 },
 [0x21BA]={
  category="so",
  direction="on",
 },
 [0x21BB]={
  category="so",
  direction="on",
 },
 [0x21BC]={
  category="so",
  direction="on",
 },
 [0x21BD]={
  category="so",
  direction="on",
 },
 [0x21BE]={
  category="so",
  direction="on",
 },
 [0x21BF]={
  category="so",
  direction="on",
 },
 [0x21C0]={
  category="so",
  direction="on",
 },
 [0x21C1]={
  category="so",
  direction="on",
 },
 [0x21C2]={
  category="so",
  direction="on",
 },
 [0x21C3]={
  category="so",
  direction="on",
 },
 [0x21C4]={
  category="so",
  direction="on",
 },
 [0x21C5]={
  category="so",
  direction="on",
 },
 [0x21C6]={
  category="so",
  direction="on",
 },
 [0x21C7]={
  category="so",
  direction="on",
 },
 [0x21C8]={
  category="so",
  direction="on",
 },
 [0x21C9]={
  category="so",
  direction="on",
 },
 [0x21CA]={
  category="so",
  direction="on",
 },
 [0x21CB]={
  category="so",
  direction="on",
 },
 [0x21CC]={
  category="so",
  direction="on",
 },
 [0x21CD]={
  category="so",
  direction="on",
 },
 [0x21CE]={
  category="sm",
  direction="on",
 },
 [0x21CF]={
  category="sm",
  direction="on",
 },
 [0x21D0]={
  category="so",
  direction="on",
 },
 [0x21D1]={
  category="so",
  direction="on",
 },
 [0x21D2]={
  category="sm",
  direction="on",
 },
 [0x21D3]={
  category="so",
  direction="on",
 },
 [0x21D4]={
  category="sm",
  direction="on",
 },
 [0x21D5]={
  category="so",
  direction="on",
 },
 [0x21D6]={
  category="so",
  direction="on",
 },
 [0x21D7]={
  category="so",
  direction="on",
 },
 [0x21D8]={
  category="so",
  direction="on",
 },
 [0x21D9]={
  category="so",
  direction="on",
 },
 [0x21DA]={
  category="so",
  direction="on",
 },
 [0x21DB]={
  category="so",
  direction="on",
 },
 [0x21DC]={
  category="so",
  direction="on",
 },
 [0x21DD]={
  category="so",
  direction="on",
 },
 [0x21DE]={
  category="so",
  direction="on",
 },
 [0x21DF]={
  category="so",
  direction="on",
 },
 [0x21E0]={
  category="so",
  direction="on",
 },
 [0x21E1]={
  category="so",
  direction="on",
 },
 [0x21E2]={
  category="so",
  direction="on",
 },
 [0x21E3]={
  category="so",
  direction="on",
 },
 [0x21E4]={
  category="so",
  direction="on",
 },
 [0x21E5]={
  category="so",
  direction="on",
 },
 [0x21E6]={
  category="so",
  direction="on",
 },
 [0x21E7]={
  category="so",
  direction="on",
 },
 [0x21E8]={
  category="so",
  direction="on",
 },
 [0x21E9]={
  category="so",
  direction="on",
 },
 [0x21EA]={
  category="so",
  direction="on",
 },
 [0x21EB]={
  category="so",
  direction="on",
 },
 [0x21EC]={
  category="so",
  direction="on",
 },
 [0x21ED]={
  category="so",
  direction="on",
 },
 [0x21EE]={
  category="so",
  direction="on",
 },
 [0x21EF]={
  category="so",
  direction="on",
 },
 [0x21F0]={
  category="so",
  direction="on",
 },
 [0x21F1]={
  category="so",
  direction="on",
 },
 [0x21F2]={
  category="so",
  direction="on",
 },
 [0x21F3]={
  category="so",
  direction="on",
 },
 [0x21F4]={
  category="sm",
  direction="on",
 },
 [0x21F5]={
  category="sm",
  direction="on",
 },
 [0x21F6]={
  category="sm",
  direction="on",
 },
 [0x21F7]={
  category="sm",
  direction="on",
 },
 [0x21F8]={
  category="sm",
  direction="on",
 },
 [0x21F9]={
  category="sm",
  direction="on",
 },
 [0x21FA]={
  category="sm",
  direction="on",
 },
 [0x21FB]={
  category="sm",
  direction="on",
 },
 [0x21FC]={
  category="sm",
  direction="on",
 },
 [0x21FD]={
  category="sm",
  direction="on",
 },
 [0x21FE]={
  category="sm",
  direction="on",
 },
 [0x21FF]={
  category="sm",
  direction="on",
 },
 [0x2200]={
  category="sm",
  direction="on",
 },
 [0x2201]={
  category="sm",
  direction="on",
 },
 [0x2202]={
  category="sm",
  direction="on",
 },
 [0x2203]={
  category="sm",
  direction="on",
 },
 [0x2204]={
  category="sm",
  direction="on",
 },
 [0x2205]={
  category="sm",
  direction="on",
 },
 [0x2206]={
  category="sm",
  direction="on",
 },
 [0x2207]={
  category="sm",
  direction="on",
 },
 [0x2208]={
  category="sm",
  direction="on",
  mirror=0x220B,
 },
 [0x2209]={
  category="sm",
  direction="on",
  mirror=0x220C,
 },
 [0x220A]={
  category="sm",
  direction="on",
  mirror=0x220D,
 },
 [0x220B]={
  category="sm",
  direction="on",
  mirror=0x2208,
 },
 [0x220C]={
  category="sm",
  direction="on",
  mirror=0x2209,
 },
 [0x220D]={
  category="sm",
  direction="on",
  mirror=0x220A,
 },
 [0x220E]={
  category="sm",
  direction="on",
 },
 [0x220F]={
  category="sm",
  direction="on",
 },
 [0x2210]={
  category="sm",
  direction="on",
 },
 [0x2211]={
  category="sm",
  direction="on",
 },
 [0x2212]={
  category="sm",
  direction="es",
 },
 [0x2213]={
  category="sm",
  direction="et",
 },
 [0x2214]={
  category="sm",
  direction="on",
 },
 [0x2215]={
  category="sm",
  direction="on",
  mirror=0x29F5,
 },
 [0x2216]={
  category="sm",
  direction="on",
 },
 [0x2217]={
  category="sm",
  direction="on",
 },
 [0x2218]={
  category="sm",
  direction="on",
 },
 [0x2219]={
  category="sm",
  direction="on",
 },
 [0x221A]={
  category="sm",
  direction="on",
 },
 [0x221B]={
  category="sm",
  direction="on",
 },
 [0x221C]={
  category="sm",
  direction="on",
 },
 [0x221D]={
  category="sm",
  direction="on",
 },
 [0x221E]={
  category="sm",
  direction="on",
 },
 [0x221F]={
  category="sm",
  direction="on",
 },
 [0x2220]={
  category="sm",
  direction="on",
 },
 [0x2221]={
  category="sm",
  direction="on",
 },
 [0x2222]={
  category="sm",
  direction="on",
 },
 [0x2223]={
  category="sm",
  direction="on",
 },
 [0x2224]={
  category="sm",
  direction="on",
 },
 [0x2225]={
  category="sm",
  direction="on",
 },
 [0x2226]={
  category="sm",
  direction="on",
 },
 [0x2227]={
  category="sm",
  direction="on",
 },
 [0x2228]={
  category="sm",
  direction="on",
 },
 [0x2229]={
  category="sm",
  direction="on",
 },
 [0x222A]={
  category="sm",
  direction="on",
 },
 [0x222B]={
  category="sm",
  direction="on",
 },
 [0x222C]={
  category="sm",
  direction="on",
 },
 [0x222D]={
  category="sm",
  direction="on",
 },
 [0x222E]={
  category="sm",
  direction="on",
 },
 [0x222F]={
  category="sm",
  direction="on",
 },
 [0x2230]={
  category="sm",
  direction="on",
 },
 [0x2231]={
  category="sm",
  direction="on",
 },
 [0x2232]={
  category="sm",
  direction="on",
 },
 [0x2233]={
  category="sm",
  direction="on",
 },
 [0x2234]={
  category="sm",
  direction="on",
 },
 [0x2235]={
  category="sm",
  direction="on",
 },
 [0x2236]={
  category="sm",
  direction="on",
 },
 [0x2237]={
  category="sm",
  direction="on",
 },
 [0x2238]={
  category="sm",
  direction="on",
 },
 [0x2239]={
  category="sm",
  direction="on",
 },
 [0x223A]={
  category="sm",
  direction="on",
 },
 [0x223B]={
  category="sm",
  direction="on",
 },
 [0x223C]={
  category="sm",
  direction="on",
  mirror=0x223D,
 },
 [0x223D]={
  category="sm",
  direction="on",
  mirror=0x223C,
 },
 [0x223E]={
  category="sm",
  direction="on",
 },
 [0x223F]={
  category="sm",
  direction="on",
 },
 [0x2240]={
  category="sm",
  direction="on",
 },
 [0x2241]={
  category="sm",
  direction="on",
 },
 [0x2242]={
  category="sm",
  direction="on",
 },
 [0x2243]={
  category="sm",
  direction="on",
  mirror=0x22CD,
 },
 [0x2244]={
  category="sm",
  direction="on",
 },
 [0x2245]={
  category="sm",
  direction="on",
 },
 [0x2246]={
  category="sm",
  direction="on",
 },
 [0x2247]={
  category="sm",
  direction="on",
 },
 [0x2248]={
  category="sm",
  direction="on",
 },
 [0x2249]={
  category="sm",
  direction="on",
 },
 [0x224A]={
  category="sm",
  direction="on",
 },
 [0x224B]={
  category="sm",
  direction="on",
 },
 [0x224C]={
  category="sm",
  direction="on",
 },
 [0x224D]={
  category="sm",
  direction="on",
 },
 [0x224E]={
  category="sm",
  direction="on",
 },
 [0x224F]={
  category="sm",
  direction="on",
 },
 [0x2250]={
  category="sm",
  direction="on",
 },
 [0x2251]={
  category="sm",
  direction="on",
 },
 [0x2252]={
  category="sm",
  direction="on",
  mirror=0x2253,
 },
 [0x2253]={
  category="sm",
  direction="on",
  mirror=0x2252,
 },
 [0x2254]={
  category="sm",
  direction="on",
  mirror=0x2255,
 },
 [0x2255]={
  category="sm",
  direction="on",
  mirror=0x2254,
 },
 [0x2256]={
  category="sm",
  direction="on",
 },
 [0x2257]={
  category="sm",
  direction="on",
 },
 [0x2258]={
  category="sm",
  direction="on",
 },
 [0x2259]={
  category="sm",
  direction="on",
 },
 [0x225A]={
  category="sm",
  direction="on",
 },
 [0x225B]={
  category="sm",
  direction="on",
 },
 [0x225C]={
  category="sm",
  direction="on",
 },
 [0x225D]={
  category="sm",
  direction="on",
 },
 [0x225E]={
  category="sm",
  direction="on",
 },
 [0x225F]={
  category="sm",
  direction="on",
 },
 [0x2260]={
  category="sm",
  direction="on",
 },
 [0x2261]={
  category="sm",
  direction="on",
 },
 [0x2262]={
  category="sm",
  direction="on",
 },
 [0x2263]={
  category="sm",
  direction="on",
 },
 [0x2264]={
  category="sm",
  direction="on",
  mirror=0x2265,
 },
 [0x2265]={
  category="sm",
  direction="on",
  mirror=0x2264,
 },
 [0x2266]={
  category="sm",
  direction="on",
  mirror=0x2267,
 },
 [0x2267]={
  category="sm",
  direction="on",
  mirror=0x2266,
 },
 [0x2268]={
  category="sm",
  direction="on",
  mirror=0x2269,
 },
 [0x2269]={
  category="sm",
  direction="on",
  mirror=0x2268,
 },
 [0x226A]={
  category="sm",
  direction="on",
  mirror=0x226B,
 },
 [0x226B]={
  category="sm",
  direction="on",
  mirror=0x226A,
 },
 [0x226C]={
  category="sm",
  direction="on",
 },
 [0x226D]={
  category="sm",
  direction="on",
 },
 [0x226E]={
  category="sm",
  direction="on",
  mirror=0x226F,
 },
 [0x226F]={
  category="sm",
  direction="on",
  mirror=0x226E,
 },
 [0x2270]={
  category="sm",
  direction="on",
  mirror=0x2271,
 },
 [0x2271]={
  category="sm",
  direction="on",
  mirror=0x2270,
 },
 [0x2272]={
  category="sm",
  direction="on",
  mirror=0x2273,
 },
 [0x2273]={
  category="sm",
  direction="on",
  mirror=0x2272,
 },
 [0x2274]={
  category="sm",
  direction="on",
  mirror=0x2275,
 },
 [0x2275]={
  category="sm",
  direction="on",
  mirror=0x2274,
 },
 [0x2276]={
  category="sm",
  direction="on",
  mirror=0x2277,
 },
 [0x2277]={
  category="sm",
  direction="on",
  mirror=0x2276,
 },
 [0x2278]={
  category="sm",
  direction="on",
  mirror=0x2279,
 },
 [0x2279]={
  category="sm",
  direction="on",
  mirror=0x2278,
 },
 [0x227A]={
  category="sm",
  direction="on",
  mirror=0x227B,
 },
 [0x227B]={
  category="sm",
  direction="on",
  mirror=0x227A,
 },
 [0x227C]={
  category="sm",
  direction="on",
  mirror=0x227D,
 },
 [0x227D]={
  category="sm",
  direction="on",
  mirror=0x227C,
 },
 [0x227E]={
  category="sm",
  direction="on",
  mirror=0x227F,
 },
 [0x227F]={
  category="sm",
  direction="on",
  mirror=0x227E,
 },
 [0x2280]={
  category="sm",
  direction="on",
  mirror=0x2281,
 },
 [0x2281]={
  category="sm",
  direction="on",
  mirror=0x2280,
 },
 [0x2282]={
  category="sm",
  direction="on",
  mirror=0x2283,
 },
 [0x2283]={
  category="sm",
  direction="on",
  mirror=0x2282,
 },
 [0x2284]={
  category="sm",
  direction="on",
  mirror=0x2285,
 },
 [0x2285]={
  category="sm",
  direction="on",
  mirror=0x2284,
 },
 [0x2286]={
  category="sm",
  direction="on",
  mirror=0x2287,
 },
 [0x2287]={
  category="sm",
  direction="on",
  mirror=0x2286,
 },
 [0x2288]={
  category="sm",
  direction="on",
  mirror=0x2289,
 },
 [0x2289]={
  category="sm",
  direction="on",
  mirror=0x2288,
 },
 [0x228A]={
  category="sm",
  direction="on",
  mirror=0x228B,
 },
 [0x228B]={
  category="sm",
  direction="on",
  mirror=0x228A,
 },
 [0x228C]={
  category="sm",
  direction="on",
 },
 [0x228D]={
  category="sm",
  direction="on",
 },
 [0x228E]={
  category="sm",
  direction="on",
 },
 [0x228F]={
  category="sm",
  direction="on",
  mirror=0x2290,
 },
 [0x2290]={
  category="sm",
  direction="on",
  mirror=0x228F,
 },
 [0x2291]={
  category="sm",
  direction="on",
  mirror=0x2292,
 },
 [0x2292]={
  category="sm",
  direction="on",
  mirror=0x2291,
 },
 [0x2293]={
  category="sm",
  direction="on",
 },
 [0x2294]={
  category="sm",
  direction="on",
 },
 [0x2295]={
  category="sm",
  direction="on",
 },
 [0x2296]={
  category="sm",
  direction="on",
 },
 [0x2297]={
  category="sm",
  direction="on",
 },
 [0x2298]={
  category="sm",
  direction="on",
  mirror=0x29B8,
 },
 [0x2299]={
  category="sm",
  direction="on",
 },
 [0x229A]={
  category="sm",
  direction="on",
 },
 [0x229B]={
  category="sm",
  direction="on",
 },
 [0x229C]={
  category="sm",
  direction="on",
 },
 [0x229D]={
  category="sm",
  direction="on",
 },
 [0x229E]={
  category="sm",
  direction="on",
 },
 [0x229F]={
  category="sm",
  direction="on",
 },
 [0x22A0]={
  category="sm",
  direction="on",
 },
 [0x22A1]={
  category="sm",
  direction="on",
 },
 [0x22A2]={
  category="sm",
  direction="on",
  mirror=0x22A3,
 },
 [0x22A3]={
  category="sm",
  direction="on",
  mirror=0x22A2,
 },
 [0x22A4]={
  category="sm",
  direction="on",
 },
 [0x22A5]={
  category="sm",
  direction="on",
 },
 [0x22A6]={
  category="sm",
  direction="on",
  mirror=0x2ADE,
 },
 [0x22A7]={
  category="sm",
  direction="on",
 },
 [0x22A8]={
  category="sm",
  direction="on",
  mirror=0x2AE4,
 },
 [0x22A9]={
  category="sm",
  direction="on",
  mirror=0x2AE3,
 },
 [0x22AA]={
  category="sm",
  direction="on",
 },
 [0x22AB]={
  category="sm",
  direction="on",
  mirror=0x2AE5,
 },
 [0x22AC]={
  category="sm",
  direction="on",
 },
 [0x22AD]={
  category="sm",
  direction="on",
 },
 [0x22AE]={
  category="sm",
  direction="on",
 },
 [0x22AF]={
  category="sm",
  direction="on",
 },
 [0x22B0]={
  category="sm",
  direction="on",
  mirror=0x22B1,
 },
 [0x22B1]={
  category="sm",
  direction="on",
  mirror=0x22B0,
 },
 [0x22B2]={
  category="sm",
  direction="on",
  mirror=0x22B3,
 },
 [0x22B3]={
  category="sm",
  direction="on",
  mirror=0x22B2,
 },
 [0x22B4]={
  category="sm",
  direction="on",
  mirror=0x22B5,
 },
 [0x22B5]={
  category="sm",
  direction="on",
  mirror=0x22B4,
 },
 [0x22B6]={
  category="sm",
  direction="on",
  mirror=0x22B7,
 },
 [0x22B7]={
  category="sm",
  direction="on",
  mirror=0x22B6,
 },
 [0x22B8]={
  category="sm",
  direction="on",
 },
 [0x22B9]={
  category="sm",
  direction="on",
 },
 [0x22BA]={
  category="sm",
  direction="on",
 },
 [0x22BB]={
  category="sm",
  direction="on",
 },
 [0x22BC]={
  category="sm",
  direction="on",
 },
 [0x22BD]={
  category="sm",
  direction="on",
 },
 [0x22BE]={
  category="sm",
  direction="on",
 },
 [0x22BF]={
  category="sm",
  direction="on",
 },
 [0x22C0]={
  category="sm",
  direction="on",
 },
 [0x22C1]={
  category="sm",
  direction="on",
 },
 [0x22C2]={
  category="sm",
  direction="on",
 },
 [0x22C3]={
  category="sm",
  direction="on",
 },
 [0x22C4]={
  category="sm",
  direction="on",
 },
 [0x22C5]={
  category="sm",
  direction="on",
 },
 [0x22C6]={
  category="sm",
  direction="on",
 },
 [0x22C7]={
  category="sm",
  direction="on",
 },
 [0x22C8]={
  category="sm",
  direction="on",
 },
 [0x22C9]={
  category="sm",
  direction="on",
  mirror=0x22CA,
 },
 [0x22CA]={
  category="sm",
  direction="on",
  mirror=0x22C9,
 },
 [0x22CB]={
  category="sm",
  direction="on",
  mirror=0x22CC,
 },
 [0x22CC]={
  category="sm",
  direction="on",
  mirror=0x22CB,
 },
 [0x22CD]={
  category="sm",
  direction="on",
  mirror=0x2243,
 },
 [0x22CE]={
  category="sm",
  direction="on",
 },
 [0x22CF]={
  category="sm",
  direction="on",
 },
 [0x22D0]={
  category="sm",
  direction="on",
  mirror=0x22D1,
 },
 [0x22D1]={
  category="sm",
  direction="on",
  mirror=0x22D0,
 },
 [0x22D2]={
  category="sm",
  direction="on",
 },
 [0x22D3]={
  category="sm",
  direction="on",
 },
 [0x22D4]={
  category="sm",
  direction="on",
 },
 [0x22D5]={
  category="sm",
  direction="on",
 },
 [0x22D6]={
  category="sm",
  direction="on",
  mirror=0x22D7,
 },
 [0x22D7]={
  category="sm",
  direction="on",
  mirror=0x22D6,
 },
 [0x22D8]={
  category="sm",
  direction="on",
  mirror=0x22D9,
 },
 [0x22D9]={
  category="sm",
  direction="on",
  mirror=0x22D8,
 },
 [0x22DA]={
  category="sm",
  direction="on",
  mirror=0x22DB,
 },
 [0x22DB]={
  category="sm",
  direction="on",
  mirror=0x22DA,
 },
 [0x22DC]={
  category="sm",
  direction="on",
  mirror=0x22DD,
 },
 [0x22DD]={
  category="sm",
  direction="on",
  mirror=0x22DC,
 },
 [0x22DE]={
  category="sm",
  direction="on",
  mirror=0x22DF,
 },
 [0x22DF]={
  category="sm",
  direction="on",
  mirror=0x22DE,
 },
 [0x22E0]={
  category="sm",
  direction="on",
  mirror=0x22E1,
 },
 [0x22E1]={
  category="sm",
  direction="on",
  mirror=0x22E0,
 },
 [0x22E2]={
  category="sm",
  direction="on",
  mirror=0x22E3,
 },
 [0x22E3]={
  category="sm",
  direction="on",
  mirror=0x22E2,
 },
 [0x22E4]={
  category="sm",
  direction="on",
  mirror=0x22E5,
 },
 [0x22E5]={
  category="sm",
  direction="on",
  mirror=0x22E4,
 },
 [0x22E6]={
  category="sm",
  direction="on",
  mirror=0x22E7,
 },
 [0x22E7]={
  category="sm",
  direction="on",
  mirror=0x22E6,
 },
 [0x22E8]={
  category="sm",
  direction="on",
  mirror=0x22E9,
 },
 [0x22E9]={
  category="sm",
  direction="on",
  mirror=0x22E8,
 },
 [0x22EA]={
  category="sm",
  direction="on",
  mirror=0x22EB,
 },
 [0x22EB]={
  category="sm",
  direction="on",
  mirror=0x22EA,
 },
 [0x22EC]={
  category="sm",
  direction="on",
  mirror=0x22ED,
 },
 [0x22ED]={
  category="sm",
  direction="on",
  mirror=0x22EC,
 },
 [0x22EE]={
  category="sm",
  direction="on",
 },
 [0x22EF]={
  category="sm",
  direction="on",
 },
 [0x22F0]={
  category="sm",
  direction="on",
  mirror=0x22F1,
 },
 [0x22F1]={
  category="sm",
  direction="on",
  mirror=0x22F0,
 },
 [0x22F2]={
  category="sm",
  direction="on",
  mirror=0x22FA,
 },
 [0x22F3]={
  category="sm",
  direction="on",
  mirror=0x22FB,
 },
 [0x22F4]={
  category="sm",
  direction="on",
  mirror=0x22FC,
 },
 [0x22F5]={
  category="sm",
  direction="on",
 },
 [0x22F6]={
  category="sm",
  direction="on",
  mirror=0x22FD,
 },
 [0x22F7]={
  category="sm",
  direction="on",
  mirror=0x22FE,
 },
 [0x22F8]={
  category="sm",
  direction="on",
 },
 [0x22F9]={
  category="sm",
  direction="on",
 },
 [0x22FA]={
  category="sm",
  direction="on",
  mirror=0x22F2,
 },
 [0x22FB]={
  category="sm",
  direction="on",
  mirror=0x22F3,
 },
 [0x22FC]={
  category="sm",
  direction="on",
  mirror=0x22F4,
 },
 [0x22FD]={
  category="sm",
  direction="on",
  mirror=0x22F6,
 },
 [0x22FE]={
  category="sm",
  direction="on",
  mirror=0x22F7,
 },
 [0x22FF]={
  category="sm",
  direction="on",
 },
 [0x2300]={
  category="so",
  direction="on",
 },
 [0x2301]={
  category="so",
  direction="on",
 },
 [0x2302]={
  category="so",
  direction="on",
 },
 [0x2303]={
  category="so",
  direction="on",
 },
 [0x2304]={
  category="so",
  direction="on",
 },
 [0x2305]={
  category="so",
  direction="on",
 },
 [0x2306]={
  category="so",
  direction="on",
 },
 [0x2307]={
  category="so",
  direction="on",
 },
 [0x2308]={
  category="sm",
  direction="on",
  mirror=0x2309,
 },
 [0x2309]={
  category="sm",
  direction="on",
  mirror=0x2308,
 },
 [0x230A]={
  category="sm",
  direction="on",
  mirror=0x230B,
 },
 [0x230B]={
  category="sm",
  direction="on",
  mirror=0x230A,
 },
 [0x230C]={
  category="so",
  direction="on",
 },
 [0x230D]={
  category="so",
  direction="on",
 },
 [0x230E]={
  category="so",
  direction="on",
 },
 [0x230F]={
  category="so",
  direction="on",
 },
 [0x2310]={
  category="so",
  direction="on",
 },
 [0x2311]={
  category="so",
  direction="on",
 },
 [0x2312]={
  category="so",
  direction="on",
 },
 [0x2313]={
  category="so",
  direction="on",
 },
 [0x2314]={
  category="so",
  direction="on",
 },
 [0x2315]={
  category="so",
  direction="on",
 },
 [0x2316]={
  category="so",
  direction="on",
 },
 [0x2317]={
  category="so",
  direction="on",
 },
 [0x2318]={
  category="so",
  direction="on",
 },
 [0x2319]={
  category="so",
  direction="on",
 },
 [0x231A]={
  category="so",
  direction="on",
 },
 [0x231B]={
  category="so",
  direction="on",
 },
 [0x231C]={
  category="so",
  direction="on",
 },
 [0x231D]={
  category="so",
  direction="on",
 },
 [0x231E]={
  category="so",
  direction="on",
 },
 [0x231F]={
  category="so",
  direction="on",
 },
 [0x2320]={
  category="sm",
  direction="on",
 },
 [0x2321]={
  category="sm",
  direction="on",
 },
 [0x2322]={
  category="so",
  direction="on",
 },
 [0x2323]={
  category="so",
  direction="on",
 },
 [0x2324]={
  category="so",
  direction="on",
 },
 [0x2325]={
  category="so",
  direction="on",
 },
 [0x2326]={
  category="so",
  direction="on",
 },
 [0x2327]={
  category="so",
  direction="on",
 },
 [0x2328]={
  category="so",
  direction="on",
 },
 [0x2329]={
  category="ps",
  direction="on",
  mirror=0x232A,
 },
 [0x232A]={
  category="pe",
  direction="on",
  mirror=0x2329,
 },
 [0x232B]={
  category="so",
  direction="on",
 },
 [0x232C]={
  category="so",
  direction="on",
 },
 [0x232D]={
  category="so",
  direction="on",
 },
 [0x232E]={
  category="so",
  direction="on",
 },
 [0x232F]={
  category="so",
  direction="on",
 },
 [0x2330]={
  category="so",
  direction="on",
 },
 [0x2331]={
  category="so",
  direction="on",
 },
 [0x2332]={
  category="so",
  direction="on",
 },
 [0x2333]={
  category="so",
  direction="on",
 },
 [0x2334]={
  category="so",
  direction="on",
 },
 [0x2335]={
  category="so",
  direction="on",
 },
 [0x2336]={
  category="so",
  direction="l",
 },
 [0x2337]={
  category="so",
  direction="l",
 },
 [0x2338]={
  category="so",
  direction="l",
 },
 [0x2339]={
  category="so",
  direction="l",
 },
 [0x233A]={
  category="so",
  direction="l",
 },
 [0x233B]={
  category="so",
  direction="l",
 },
 [0x233C]={
  category="so",
  direction="l",
 },
 [0x233D]={
  category="so",
  direction="l",
 },
 [0x233E]={
  category="so",
  direction="l",
 },
 [0x233F]={
  category="so",
  direction="l",
 },
 [0x2340]={
  category="so",
  direction="l",
 },
 [0x2341]={
  category="so",
  direction="l",
 },
 [0x2342]={
  category="so",
  direction="l",
 },
 [0x2343]={
  category="so",
  direction="l",
 },
 [0x2344]={
  category="so",
  direction="l",
 },
 [0x2345]={
  category="so",
  direction="l",
 },
 [0x2346]={
  category="so",
  direction="l",
 },
 [0x2347]={
  category="so",
  direction="l",
 },
 [0x2348]={
  category="so",
  direction="l",
 },
 [0x2349]={
  category="so",
  direction="l",
 },
 [0x234A]={
  category="so",
  direction="l",
 },
 [0x234B]={
  category="so",
  direction="l",
 },
 [0x234C]={
  category="so",
  direction="l",
 },
 [0x234D]={
  category="so",
  direction="l",
 },
 [0x234E]={
  category="so",
  direction="l",
 },
 [0x234F]={
  category="so",
  direction="l",
 },
 [0x2350]={
  category="so",
  direction="l",
 },
 [0x2351]={
  category="so",
  direction="l",
 },
 [0x2352]={
  category="so",
  direction="l",
 },
 [0x2353]={
  category="so",
  direction="l",
 },
 [0x2354]={
  category="so",
  direction="l",
 },
 [0x2355]={
  category="so",
  direction="l",
 },
 [0x2356]={
  category="so",
  direction="l",
 },
 [0x2357]={
  category="so",
  direction="l",
 },
 [0x2358]={
  category="so",
  direction="l",
 },
 [0x2359]={
  category="so",
  direction="l",
 },
 [0x235A]={
  category="so",
  direction="l",
 },
 [0x235B]={
  category="so",
  direction="l",
 },
 [0x235C]={
  category="so",
  direction="l",
 },
 [0x235D]={
  category="so",
  direction="l",
 },
 [0x235E]={
  category="so",
  direction="l",
 },
 [0x235F]={
  category="so",
  direction="l",
 },
 [0x2360]={
  category="so",
  direction="l",
 },
 [0x2361]={
  category="so",
  direction="l",
 },
 [0x2362]={
  category="so",
  direction="l",
 },
 [0x2363]={
  category="so",
  direction="l",
 },
 [0x2364]={
  category="so",
  direction="l",
 },
 [0x2365]={
  category="so",
  direction="l",
 },
 [0x2366]={
  category="so",
  direction="l",
 },
 [0x2367]={
  category="so",
  direction="l",
 },
 [0x2368]={
  category="so",
  direction="l",
 },
 [0x2369]={
  category="so",
  direction="l",
 },
 [0x236A]={
  category="so",
  direction="l",
 },
 [0x236B]={
  category="so",
  direction="l",
 },
 [0x236C]={
  category="so",
  direction="l",
 },
 [0x236D]={
  category="so",
  direction="l",
 },
 [0x236E]={
  category="so",
  direction="l",
 },
 [0x236F]={
  category="so",
  direction="l",
 },
 [0x2370]={
  category="so",
  direction="l",
 },
 [0x2371]={
  category="so",
  direction="l",
 },
 [0x2372]={
  category="so",
  direction="l",
 },
 [0x2373]={
  category="so",
  direction="l",
 },
 [0x2374]={
  category="so",
  direction="l",
 },
 [0x2375]={
  category="so",
  direction="l",
 },
 [0x2376]={
  category="so",
  direction="l",
 },
 [0x2377]={
  category="so",
  direction="l",
 },
 [0x2378]={
  category="so",
  direction="l",
 },
 [0x2379]={
  category="so",
  direction="l",
 },
 [0x237A]={
  category="so",
  direction="l",
 },
 [0x237B]={
  category="so",
  direction="on",
 },
 [0x237C]={
  category="sm",
  direction="on",
 },
 [0x237D]={
  category="so",
  direction="on",
 },
 [0x237E]={
  category="so",
  direction="on",
 },
 [0x237F]={
  category="so",
  direction="on",
 },
 [0x2380]={
  category="so",
  direction="on",
 },
 [0x2381]={
  category="so",
  direction="on",
 },
 [0x2382]={
  category="so",
  direction="on",
 },
 [0x2383]={
  category="so",
  direction="on",
 },
 [0x2384]={
  category="so",
  direction="on",
 },
 [0x2385]={
  category="so",
  direction="on",
 },
 [0x2386]={
  category="so",
  direction="on",
 },
 [0x2387]={
  category="so",
  direction="on",
 },
 [0x2388]={
  category="so",
  direction="on",
 },
 [0x2389]={
  category="so",
  direction="on",
 },
 [0x238A]={
  category="so",
  direction="on",
 },
 [0x238B]={
  category="so",
  direction="on",
 },
 [0x238C]={
  category="so",
  direction="on",
 },
 [0x238D]={
  category="so",
  direction="on",
 },
 [0x238E]={
  category="so",
  direction="on",
 },
 [0x238F]={
  category="so",
  direction="on",
 },
 [0x2390]={
  category="so",
  direction="on",
 },
 [0x2391]={
  category="so",
  direction="on",
 },
 [0x2392]={
  category="so",
  direction="on",
 },
 [0x2393]={
  category="so",
  direction="on",
 },
 [0x2394]={
  category="so",
  direction="on",
 },
 [0x2395]={
  category="so",
  direction="l",
 },
 [0x2396]={
  category="so",
  direction="on",
 },
 [0x2397]={
  category="so",
  direction="on",
 },
 [0x2398]={
  category="so",
  direction="on",
 },
 [0x2399]={
  category="so",
  direction="on",
 },
 [0x239A]={
  category="so",
  direction="on",
 },
 [0x239B]={
  category="sm",
  direction="on",
 },
 [0x239C]={
  category="sm",
  direction="on",
 },
 [0x239D]={
  category="sm",
  direction="on",
 },
 [0x239E]={
  category="sm",
  direction="on",
 },
 [0x239F]={
  category="sm",
  direction="on",
 },
 [0x23A0]={
  category="sm",
  direction="on",
 },
 [0x23A1]={
  category="sm",
  direction="on",
 },
 [0x23A2]={
  category="sm",
  direction="on",
 },
 [0x23A3]={
  category="sm",
  direction="on",
 },
 [0x23A4]={
  category="sm",
  direction="on",
 },
 [0x23A5]={
  category="sm",
  direction="on",
 },
 [0x23A6]={
  category="sm",
  direction="on",
 },
 [0x23A7]={
  category="sm",
  direction="on",
 },
 [0x23A8]={
  category="sm",
  direction="on",
 },
 [0x23A9]={
  category="sm",
  direction="on",
 },
 [0x23AA]={
  category="sm",
  direction="on",
 },
 [0x23AB]={
  category="sm",
  direction="on",
 },
 [0x23AC]={
  category="sm",
  direction="on",
 },
 [0x23AD]={
  category="sm",
  direction="on",
 },
 [0x23AE]={
  category="sm",
  direction="on",
 },
 [0x23AF]={
  category="sm",
  direction="on",
 },
 [0x23B0]={
  category="sm",
  direction="on",
 },
 [0x23B1]={
  category="sm",
  direction="on",
 },
 [0x23B2]={
  category="sm",
  direction="on",
 },
 [0x23B3]={
  category="sm",
  direction="on",
 },
 [0x23B4]={
  category="so",
  direction="on",
 },
 [0x23B5]={
  category="so",
  direction="on",
 },
 [0x23B6]={
  category="so",
  direction="on",
 },
 [0x23B7]={
  category="so",
  direction="on",
 },
 [0x23B8]={
  category="so",
  direction="on",
 },
 [0x23B9]={
  category="so",
  direction="on",
 },
 [0x23BA]={
  category="so",
  direction="on",
 },
 [0x23BB]={
  category="so",
  direction="on",
 },
 [0x23BC]={
  category="so",
  direction="on",
 },
 [0x23BD]={
  category="so",
  direction="on",
 },
 [0x23BE]={
  category="so",
  direction="on",
 },
 [0x23BF]={
  category="so",
  direction="on",
 },
 [0x23C0]={
  category="so",
  direction="on",
 },
 [0x23C1]={
  category="so",
  direction="on",
 },
 [0x23C2]={
  category="so",
  direction="on",
 },
 [0x23C3]={
  category="so",
  direction="on",
 },
 [0x23C4]={
  category="so",
  direction="on",
 },
 [0x23C5]={
  category="so",
  direction="on",
 },
 [0x23C6]={
  category="so",
  direction="on",
 },
 [0x23C7]={
  category="so",
  direction="on",
 },
 [0x23C8]={
  category="so",
  direction="on",
 },
 [0x23C9]={
  category="so",
  direction="on",
 },
 [0x23CA]={
  category="so",
  direction="on",
 },
 [0x23CB]={
  category="so",
  direction="on",
 },
 [0x23CC]={
  category="so",
  direction="on",
 },
 [0x23CD]={
  category="so",
  direction="on",
 },
 [0x23CE]={
  category="so",
  direction="on",
 },
 [0x23CF]={
  category="so",
  direction="on",
 },
 [0x23D0]={
  category="so",
  direction="on",
 },
 [0x23D1]={
  category="so",
  direction="on",
 },
 [0x23D2]={
  category="so",
  direction="on",
 },
 [0x23D3]={
  category="so",
  direction="on",
 },
 [0x23D4]={
  category="so",
  direction="on",
 },
 [0x23D5]={
  category="so",
  direction="on",
 },
 [0x23D6]={
  category="so",
  direction="on",
 },
 [0x23D7]={
  category="so",
  direction="on",
 },
 [0x23D8]={
  category="so",
  direction="on",
 },
 [0x23D9]={
  category="so",
  direction="on",
 },
 [0x23DA]={
  category="so",
  direction="on",
 },
 [0x23DB]={
  category="so",
  direction="on",
 },
 [0x23DC]={
  category="sm",
  direction="on",
 },
 [0x23DD]={
  category="sm",
  direction="on",
 },
 [0x23DE]={
  category="sm",
  direction="on",
 },
 [0x23DF]={
  category="sm",
  direction="on",
 },
 [0x23E0]={
  category="sm",
  direction="on",
 },
 [0x23E1]={
  category="sm",
  direction="on",
 },
 [0x23E2]={
  category="so",
  direction="on",
 },
 [0x23E3]={
  category="so",
  direction="on",
 },
 [0x23E4]={
  category="so",
  direction="on",
 },
 [0x23E5]={
  category="so",
  direction="on",
 },
 [0x23E6]={
  category="so",
  direction="on",
 },
 [0x23E7]={
  category="so",
  direction="on",
 },
 [0x23E8]={
  category="so",
  direction="on",
 },
 [0x23E9]={
  category="so",
  direction="on",
 },
 [0x23EA]={
  category="so",
  direction="on",
 },
 [0x23EB]={
  category="so",
  direction="on",
 },
 [0x23EC]={
  category="so",
  direction="on",
 },
 [0x23ED]={
  category="so",
  direction="on",
 },
 [0x23EE]={
  category="so",
  direction="on",
 },
 [0x23EF]={
  category="so",
  direction="on",
 },
 [0x23F0]={
  category="so",
  direction="on",
 },
 [0x23F1]={
  category="so",
  direction="on",
 },
 [0x23F2]={
  category="so",
  direction="on",
 },
 [0x23F3]={
  category="so",
  direction="on",
 },
 [0x23F4]={
  category="so",
  direction="on",
 },
 [0x23F5]={
  category="so",
  direction="on",
 },
 [0x23F6]={
  category="so",
  direction="on",
 },
 [0x23F7]={
  category="so",
  direction="on",
 },
 [0x23F8]={
  category="so",
  direction="on",
 },
 [0x23F9]={
  category="so",
  direction="on",
 },
 [0x23FA]={
  category="so",
  direction="on",
 },
 [0x2400]={
  category="so",
  direction="on",
 },
 [0x2401]={
  category="so",
  direction="on",
 },
 [0x2402]={
  category="so",
  direction="on",
 },
 [0x2403]={
  category="so",
  direction="on",
 },
 [0x2404]={
  category="so",
  direction="on",
 },
 [0x2405]={
  category="so",
  direction="on",
 },
 [0x2406]={
  category="so",
  direction="on",
 },
 [0x2407]={
  category="so",
  direction="on",
 },
 [0x2408]={
  category="so",
  direction="on",
 },
 [0x2409]={
  category="so",
  direction="on",
 },
 [0x240A]={
  category="so",
  direction="on",
 },
 [0x240B]={
  category="so",
  direction="on",
 },
 [0x240C]={
  category="so",
  direction="on",
 },
 [0x240D]={
  category="so",
  direction="on",
 },
 [0x240E]={
  category="so",
  direction="on",
 },
 [0x240F]={
  category="so",
  direction="on",
 },
 [0x2410]={
  category="so",
  direction="on",
 },
 [0x2411]={
  category="so",
  direction="on",
 },
 [0x2412]={
  category="so",
  direction="on",
 },
 [0x2413]={
  category="so",
  direction="on",
 },
 [0x2414]={
  category="so",
  direction="on",
 },
 [0x2415]={
  category="so",
  direction="on",
 },
 [0x2416]={
  category="so",
  direction="on",
 },
 [0x2417]={
  category="so",
  direction="on",
 },
 [0x2418]={
  category="so",
  direction="on",
 },
 [0x2419]={
  category="so",
  direction="on",
 },
 [0x241A]={
  category="so",
  direction="on",
 },
 [0x241B]={
  category="so",
  direction="on",
 },
 [0x241C]={
  category="so",
  direction="on",
 },
 [0x241D]={
  category="so",
  direction="on",
 },
 [0x241E]={
  category="so",
  direction="on",
 },
 [0x241F]={
  category="so",
  direction="on",
 },
 [0x2420]={
  category="so",
  direction="on",
 },
 [0x2421]={
  category="so",
  direction="on",
 },
 [0x2422]={
  category="so",
  direction="on",
 },
 [0x2423]={
  category="so",
  direction="on",
 },
 [0x2424]={
  category="so",
  direction="on",
 },
 [0x2425]={
  category="so",
  direction="on",
 },
 [0x2426]={
  category="so",
  direction="on",
 },
 [0x2440]={
  category="so",
  direction="on",
 },
 [0x2441]={
  category="so",
  direction="on",
 },
 [0x2442]={
  category="so",
  direction="on",
 },
 [0x2443]={
  category="so",
  direction="on",
 },
 [0x2444]={
  category="so",
  direction="on",
 },
 [0x2445]={
  category="so",
  direction="on",
 },
 [0x2446]={
  category="so",
  direction="on",
 },
 [0x2447]={
  category="so",
  direction="on",
 },
 [0x2448]={
  category="so",
  direction="on",
 },
 [0x2449]={
  category="so",
  direction="on",
 },
 [0x244A]={
  category="so",
  direction="on",
 },
 [0x2460]={
  category="no",
  direction="on",
 },
 [0x2461]={
  category="no",
  direction="on",
 },
 [0x2462]={
  category="no",
  direction="on",
 },
 [0x2463]={
  category="no",
  direction="on",
 },
 [0x2464]={
  category="no",
  direction="on",
 },
 [0x2465]={
  category="no",
  direction="on",
 },
 [0x2466]={
  category="no",
  direction="on",
 },
 [0x2467]={
  category="no",
  direction="on",
 },
 [0x2468]={
  category="no",
  direction="on",
 },
 [0x2469]={
  category="no",
  direction="on",
 },
 [0x246A]={
  category="no",
  direction="on",
 },
 [0x246B]={
  category="no",
  direction="on",
 },
 [0x246C]={
  category="no",
  direction="on",
 },
 [0x246D]={
  category="no",
  direction="on",
 },
 [0x246E]={
  category="no",
  direction="on",
 },
 [0x246F]={
  category="no",
  direction="on",
 },
 [0x2470]={
  category="no",
  direction="on",
 },
 [0x2471]={
  category="no",
  direction="on",
 },
 [0x2472]={
  category="no",
  direction="on",
 },
 [0x2473]={
  category="no",
  direction="on",
 },
 [0x2474]={
  category="no",
  direction="on",
 },
 [0x2475]={
  category="no",
  direction="on",
 },
 [0x2476]={
  category="no",
  direction="on",
 },
 [0x2477]={
  category="no",
  direction="on",
 },
 [0x2478]={
  category="no",
  direction="on",
 },
 [0x2479]={
  category="no",
  direction="on",
 },
 [0x247A]={
  category="no",
  direction="on",
 },
 [0x247B]={
  category="no",
  direction="on",
 },
 [0x247C]={
  category="no",
  direction="on",
 },
 [0x247D]={
  category="no",
  direction="on",
 },
 [0x247E]={
  category="no",
  direction="on",
 },
 [0x247F]={
  category="no",
  direction="on",
 },
 [0x2480]={
  category="no",
  direction="on",
 },
 [0x2481]={
  category="no",
  direction="on",
 },
 [0x2482]={
  category="no",
  direction="on",
 },
 [0x2483]={
  category="no",
  direction="on",
 },
 [0x2484]={
  category="no",
  direction="on",
 },
 [0x2485]={
  category="no",
  direction="on",
 },
 [0x2486]={
  category="no",
  direction="on",
 },
 [0x2487]={
  category="no",
  direction="on",
 },
 [0x2488]={
  category="no",
  direction="en",
 },
 [0x2489]={
  category="no",
  direction="en",
 },
 [0x248A]={
  category="no",
  direction="en",
 },
 [0x248B]={
  category="no",
  direction="en",
 },
 [0x248C]={
  category="no",
  direction="en",
 },
 [0x248D]={
  category="no",
  direction="en",
 },
 [0x248E]={
  category="no",
  direction="en",
 },
 [0x248F]={
  category="no",
  direction="en",
 },
 [0x2490]={
  category="no",
  direction="en",
 },
 [0x2491]={
  category="no",
  direction="en",
 },
 [0x2492]={
  category="no",
  direction="en",
 },
 [0x2493]={
  category="no",
  direction="en",
 },
 [0x2494]={
  category="no",
  direction="en",
 },
 [0x2495]={
  category="no",
  direction="en",
 },
 [0x2496]={
  category="no",
  direction="en",
 },
 [0x2497]={
  category="no",
  direction="en",
 },
 [0x2498]={
  category="no",
  direction="en",
 },
 [0x2499]={
  category="no",
  direction="en",
 },
 [0x249A]={
  category="no",
  direction="en",
 },
 [0x249B]={
  category="no",
  direction="en",
 },
 [0x249C]={
  category="so",
  direction="l",
 },
 [0x249D]={
  category="so",
  direction="l",
 },
 [0x249E]={
  category="so",
  direction="l",
 },
 [0x249F]={
  category="so",
  direction="l",
 },
 [0x24A0]={
  category="so",
  direction="l",
 },
 [0x24A1]={
  category="so",
  direction="l",
 },
 [0x24A2]={
  category="so",
  direction="l",
 },
 [0x24A3]={
  category="so",
  direction="l",
 },
 [0x24A4]={
  category="so",
  direction="l",
 },
 [0x24A5]={
  category="so",
  direction="l",
 },
 [0x24A6]={
  category="so",
  direction="l",
 },
 [0x24A7]={
  category="so",
  direction="l",
 },
 [0x24A8]={
  category="so",
  direction="l",
 },
 [0x24A9]={
  category="so",
  direction="l",
 },
 [0x24AA]={
  category="so",
  direction="l",
 },
 [0x24AB]={
  category="so",
  direction="l",
 },
 [0x24AC]={
  category="so",
  direction="l",
 },
 [0x24AD]={
  category="so",
  direction="l",
 },
 [0x24AE]={
  category="so",
  direction="l",
 },
 [0x24AF]={
  category="so",
  direction="l",
 },
 [0x24B0]={
  category="so",
  direction="l",
 },
 [0x24B1]={
  category="so",
  direction="l",
 },
 [0x24B2]={
  category="so",
  direction="l",
 },
 [0x24B3]={
  category="so",
  direction="l",
 },
 [0x24B4]={
  category="so",
  direction="l",
 },
 [0x24B5]={
  category="so",
  direction="l",
 },
 [0x24B6]={
  category="so",
  direction="l",
 },
 [0x24B7]={
  category="so",
  direction="l",
 },
 [0x24B8]={
  category="so",
  direction="l",
 },
 [0x24B9]={
  category="so",
  direction="l",
 },
 [0x24BA]={
  category="so",
  direction="l",
 },
 [0x24BB]={
  category="so",
  direction="l",
 },
 [0x24BC]={
  category="so",
  direction="l",
 },
 [0x24BD]={
  category="so",
  direction="l",
 },
 [0x24BE]={
  category="so",
  direction="l",
 },
 [0x24BF]={
  category="so",
  direction="l",
 },
 [0x24C0]={
  category="so",
  direction="l",
 },
 [0x24C1]={
  category="so",
  direction="l",
 },
 [0x24C2]={
  category="so",
  direction="l",
 },
 [0x24C3]={
  category="so",
  direction="l",
 },
 [0x24C4]={
  category="so",
  direction="l",
 },
 [0x24C5]={
  category="so",
  direction="l",
 },
 [0x24C6]={
  category="so",
  direction="l",
 },
 [0x24C7]={
  category="so",
  direction="l",
 },
 [0x24C8]={
  category="so",
  direction="l",
 },
 [0x24C9]={
  category="so",
  direction="l",
 },
 [0x24CA]={
  category="so",
  direction="l",
 },
 [0x24CB]={
  category="so",
  direction="l",
 },
 [0x24CC]={
  category="so",
  direction="l",
 },
 [0x24CD]={
  category="so",
  direction="l",
 },
 [0x24CE]={
  category="so",
  direction="l",
 },
 [0x24CF]={
  category="so",
  direction="l",
 },
 [0x24D0]={
  category="so",
  direction="l",
 },
 [0x24D1]={
  category="so",
  direction="l",
 },
 [0x24D2]={
  category="so",
  direction="l",
 },
 [0x24D3]={
  category="so",
  direction="l",
 },
 [0x24D4]={
  category="so",
  direction="l",
 },
 [0x24D5]={
  category="so",
  direction="l",
 },
 [0x24D6]={
  category="so",
  direction="l",
 },
 [0x24D7]={
  category="so",
  direction="l",
 },
 [0x24D8]={
  category="so",
  direction="l",
 },
 [0x24D9]={
  category="so",
  direction="l",
 },
 [0x24DA]={
  category="so",
  direction="l",
 },
 [0x24DB]={
  category="so",
  direction="l",
 },
 [0x24DC]={
  category="so",
  direction="l",
 },
 [0x24DD]={
  category="so",
  direction="l",
 },
 [0x24DE]={
  category="so",
  direction="l",
 },
 [0x24DF]={
  category="so",
  direction="l",
 },
 [0x24E0]={
  category="so",
  direction="l",
 },
 [0x24E1]={
  category="so",
  direction="l",
 },
 [0x24E2]={
  category="so",
  direction="l",
 },
 [0x24E3]={
  category="so",
  direction="l",
 },
 [0x24E4]={
  category="so",
  direction="l",
 },
 [0x24E5]={
  category="so",
  direction="l",
 },
 [0x24E6]={
  category="so",
  direction="l",
 },
 [0x24E7]={
  category="so",
  direction="l",
 },
 [0x24E8]={
  category="so",
  direction="l",
 },
 [0x24E9]={
  category="so",
  direction="l",
 },
 [0x24EA]={
  category="no",
  direction="on",
 },
 [0x24EB]={
  category="no",
  direction="on",
 },
 [0x24EC]={
  category="no",
  direction="on",
 },
 [0x24ED]={
  category="no",
  direction="on",
 },
 [0x24EE]={
  category="no",
  direction="on",
 },
 [0x24EF]={
  category="no",
  direction="on",
 },
 [0x24F0]={
  category="no",
  direction="on",
 },
 [0x24F1]={
  category="no",
  direction="on",
 },
 [0x24F2]={
  category="no",
  direction="on",
 },
 [0x24F3]={
  category="no",
  direction="on",
 },
 [0x24F4]={
  category="no",
  direction="on",
 },
 [0x24F5]={
  category="no",
  direction="on",
 },
 [0x24F6]={
  category="no",
  direction="on",
 },
 [0x24F7]={
  category="no",
  direction="on",
 },
 [0x24F8]={
  category="no",
  direction="on",
 },
 [0x24F9]={
  category="no",
  direction="on",
 },
 [0x24FA]={
  category="no",
  direction="on",
 },
 [0x24FB]={
  category="no",
  direction="on",
 },
 [0x24FC]={
  category="no",
  direction="on",
 },
 [0x24FD]={
  category="no",
  direction="on",
 },
 [0x24FE]={
  category="no",
  direction="on",
 },
 [0x24FF]={
  category="no",
  direction="on",
 },
 [0x2500]={
  category="so",
  direction="on",
 },
 [0x2501]={
  category="so",
  direction="on",
 },
 [0x2502]={
  category="so",
  direction="on",
 },
 [0x2503]={
  category="so",
  direction="on",
 },
 [0x2504]={
  category="so",
  direction="on",
 },
 [0x2505]={
  category="so",
  direction="on",
 },
 [0x2506]={
  category="so",
  direction="on",
 },
 [0x2507]={
  category="so",
  direction="on",
 },
 [0x2508]={
  category="so",
  direction="on",
 },
 [0x2509]={
  category="so",
  direction="on",
 },
 [0x250A]={
  category="so",
  direction="on",
 },
 [0x250B]={
  category="so",
  direction="on",
 },
 [0x250C]={
  category="so",
  direction="on",
 },
 [0x250D]={
  category="so",
  direction="on",
 },
 [0x250E]={
  category="so",
  direction="on",
 },
 [0x250F]={
  category="so",
  direction="on",
 },
 [0x2510]={
  category="so",
  direction="on",
 },
 [0x2511]={
  category="so",
  direction="on",
 },
 [0x2512]={
  category="so",
  direction="on",
 },
 [0x2513]={
  category="so",
  direction="on",
 },
 [0x2514]={
  category="so",
  direction="on",
 },
 [0x2515]={
  category="so",
  direction="on",
 },
 [0x2516]={
  category="so",
  direction="on",
 },
 [0x2517]={
  category="so",
  direction="on",
 },
 [0x2518]={
  category="so",
  direction="on",
 },
 [0x2519]={
  category="so",
  direction="on",
 },
 [0x251A]={
  category="so",
  direction="on",
 },
 [0x251B]={
  category="so",
  direction="on",
 },
 [0x251C]={
  category="so",
  direction="on",
 },
 [0x251D]={
  category="so",
  direction="on",
 },
 [0x251E]={
  category="so",
  direction="on",
 },
 [0x251F]={
  category="so",
  direction="on",
 },
 [0x2520]={
  category="so",
  direction="on",
 },
 [0x2521]={
  category="so",
  direction="on",
 },
 [0x2522]={
  category="so",
  direction="on",
 },
 [0x2523]={
  category="so",
  direction="on",
 },
 [0x2524]={
  category="so",
  direction="on",
 },
 [0x2525]={
  category="so",
  direction="on",
 },
 [0x2526]={
  category="so",
  direction="on",
 },
 [0x2527]={
  category="so",
  direction="on",
 },
 [0x2528]={
  category="so",
  direction="on",
 },
 [0x2529]={
  category="so",
  direction="on",
 },
 [0x252A]={
  category="so",
  direction="on",
 },
 [0x252B]={
  category="so",
  direction="on",
 },
 [0x252C]={
  category="so",
  direction="on",
 },
 [0x252D]={
  category="so",
  direction="on",
 },
 [0x252E]={
  category="so",
  direction="on",
 },
 [0x252F]={
  category="so",
  direction="on",
 },
 [0x2530]={
  category="so",
  direction="on",
 },
 [0x2531]={
  category="so",
  direction="on",
 },
 [0x2532]={
  category="so",
  direction="on",
 },
 [0x2533]={
  category="so",
  direction="on",
 },
 [0x2534]={
  category="so",
  direction="on",
 },
 [0x2535]={
  category="so",
  direction="on",
 },
 [0x2536]={
  category="so",
  direction="on",
 },
 [0x2537]={
  category="so",
  direction="on",
 },
 [0x2538]={
  category="so",
  direction="on",
 },
 [0x2539]={
  category="so",
  direction="on",
 },
 [0x253A]={
  category="so",
  direction="on",
 },
 [0x253B]={
  category="so",
  direction="on",
 },
 [0x253C]={
  category="so",
  direction="on",
 },
 [0x253D]={
  category="so",
  direction="on",
 },
 [0x253E]={
  category="so",
  direction="on",
 },
 [0x253F]={
  category="so",
  direction="on",
 },
 [0x2540]={
  category="so",
  direction="on",
 },
 [0x2541]={
  category="so",
  direction="on",
 },
 [0x2542]={
  category="so",
  direction="on",
 },
 [0x2543]={
  category="so",
  direction="on",
 },
 [0x2544]={
  category="so",
  direction="on",
 },
 [0x2545]={
  category="so",
  direction="on",
 },
 [0x2546]={
  category="so",
  direction="on",
 },
 [0x2547]={
  category="so",
  direction="on",
 },
 [0x2548]={
  category="so",
  direction="on",
 },
 [0x2549]={
  category="so",
  direction="on",
 },
 [0x254A]={
  category="so",
  direction="on",
 },
 [0x254B]={
  category="so",
  direction="on",
 },
 [0x254C]={
  category="so",
  direction="on",
 },
 [0x254D]={
  category="so",
  direction="on",
 },
 [0x254E]={
  category="so",
  direction="on",
 },
 [0x254F]={
  category="so",
  direction="on",
 },
 [0x2550]={
  category="so",
  direction="on",
 },
 [0x2551]={
  category="so",
  direction="on",
 },
 [0x2552]={
  category="so",
  direction="on",
 },
 [0x2553]={
  category="so",
  direction="on",
 },
 [0x2554]={
  category="so",
  direction="on",
 },
 [0x2555]={
  category="so",
  direction="on",
 },
 [0x2556]={
  category="so",
  direction="on",
 },
 [0x2557]={
  category="so",
  direction="on",
 },
 [0x2558]={
  category="so",
  direction="on",
 },
 [0x2559]={
  category="so",
  direction="on",
 },
 [0x255A]={
  category="so",
  direction="on",
 },
 [0x255B]={
  category="so",
  direction="on",
 },
 [0x255C]={
  category="so",
  direction="on",
 },
 [0x255D]={
  category="so",
  direction="on",
 },
 [0x255E]={
  category="so",
  direction="on",
 },
 [0x255F]={
  category="so",
  direction="on",
 },
 [0x2560]={
  category="so",
  direction="on",
 },
 [0x2561]={
  category="so",
  direction="on",
 },
 [0x2562]={
  category="so",
  direction="on",
 },
 [0x2563]={
  category="so",
  direction="on",
 },
 [0x2564]={
  category="so",
  direction="on",
 },
 [0x2565]={
  category="so",
  direction="on",
 },
 [0x2566]={
  category="so",
  direction="on",
 },
 [0x2567]={
  category="so",
  direction="on",
 },
 [0x2568]={
  category="so",
  direction="on",
 },
 [0x2569]={
  category="so",
  direction="on",
 },
 [0x256A]={
  category="so",
  direction="on",
 },
 [0x256B]={
  category="so",
  direction="on",
 },
 [0x256C]={
  category="so",
  direction="on",
 },
 [0x256D]={
  category="so",
  direction="on",
 },
 [0x256E]={
  category="so",
  direction="on",
 },
 [0x256F]={
  category="so",
  direction="on",
 },
 [0x2570]={
  category="so",
  direction="on",
 },
 [0x2571]={
  category="so",
  direction="on",
 },
 [0x2572]={
  category="so",
  direction="on",
 },
 [0x2573]={
  category="so",
  direction="on",
 },
 [0x2574]={
  category="so",
  direction="on",
 },
 [0x2575]={
  category="so",
  direction="on",
 },
 [0x2576]={
  category="so",
  direction="on",
 },
 [0x2577]={
  category="so",
  direction="on",
 },
 [0x2578]={
  category="so",
  direction="on",
 },
 [0x2579]={
  category="so",
  direction="on",
 },
 [0x257A]={
  category="so",
  direction="on",
 },
 [0x257B]={
  category="so",
  direction="on",
 },
 [0x257C]={
  category="so",
  direction="on",
 },
 [0x257D]={
  category="so",
  direction="on",
 },
 [0x257E]={
  category="so",
  direction="on",
 },
 [0x257F]={
  category="so",
  direction="on",
 },
 [0x2580]={
  category="so",
  direction="on",
 },
 [0x2581]={
  category="so",
  direction="on",
 },
 [0x2582]={
  category="so",
  direction="on",
 },
 [0x2583]={
  category="so",
  direction="on",
 },
 [0x2584]={
  category="so",
  direction="on",
 },
 [0x2585]={
  category="so",
  direction="on",
 },
 [0x2586]={
  category="so",
  direction="on",
 },
 [0x2587]={
  category="so",
  direction="on",
 },
 [0x2588]={
  category="so",
  direction="on",
 },
 [0x2589]={
  category="so",
  direction="on",
 },
 [0x258A]={
  category="so",
  direction="on",
 },
 [0x258B]={
  category="so",
  direction="on",
 },
 [0x258C]={
  category="so",
  direction="on",
 },
 [0x258D]={
  category="so",
  direction="on",
 },
 [0x258E]={
  category="so",
  direction="on",
 },
 [0x258F]={
  category="so",
  direction="on",
 },
 [0x2590]={
  category="so",
  direction="on",
 },
 [0x2591]={
  category="so",
  direction="on",
 },
 [0x2592]={
  category="so",
  direction="on",
 },
 [0x2593]={
  category="so",
  direction="on",
 },
 [0x2594]={
  category="so",
  direction="on",
 },
 [0x2595]={
  category="so",
  direction="on",
 },
 [0x2596]={
  category="so",
  direction="on",
 },
 [0x2597]={
  category="so",
  direction="on",
 },
 [0x2598]={
  category="so",
  direction="on",
 },
 [0x2599]={
  category="so",
  direction="on",
 },
 [0x259A]={
  category="so",
  direction="on",
 },
 [0x259B]={
  category="so",
  direction="on",
 },
 [0x259C]={
  category="so",
  direction="on",
 },
 [0x259D]={
  category="so",
  direction="on",
 },
 [0x259E]={
  category="so",
  direction="on",
 },
 [0x259F]={
  category="so",
  direction="on",
 },
 [0x25A0]={
  category="so",
  direction="on",
 },
 [0x25A1]={
  category="so",
  direction="on",
 },
 [0x25A2]={
  category="so",
  direction="on",
 },
 [0x25A3]={
  category="so",
  direction="on",
 },
 [0x25A4]={
  category="so",
  direction="on",
 },
 [0x25A5]={
  category="so",
  direction="on",
 },
 [0x25A6]={
  category="so",
  direction="on",
 },
 [0x25A7]={
  category="so",
  direction="on",
 },
 [0x25A8]={
  category="so",
  direction="on",
 },
 [0x25A9]={
  category="so",
  direction="on",
 },
 [0x25AA]={
  category="so",
  direction="on",
 },
 [0x25AB]={
  category="so",
  direction="on",
 },
 [0x25AC]={
  category="so",
  direction="on",
 },
 [0x25AD]={
  category="so",
  direction="on",
 },
 [0x25AE]={
  category="so",
  direction="on",
 },
 [0x25AF]={
  category="so",
  direction="on",
 },
 [0x25B0]={
  category="so",
  direction="on",
 },
 [0x25B1]={
  category="so",
  direction="on",
 },
 [0x25B2]={
  category="so",
  direction="on",
 },
 [0x25B3]={
  category="so",
  direction="on",
 },
 [0x25B4]={
  category="so",
  direction="on",
 },
 [0x25B5]={
  category="so",
  direction="on",
 },
 [0x25B6]={
  category="so",
  direction="on",
 },
 [0x25B7]={
  category="sm",
  direction="on",
 },
 [0x25B8]={
  category="so",
  direction="on",
 },
 [0x25B9]={
  category="so",
  direction="on",
 },
 [0x25BA]={
  category="so",
  direction="on",
 },
 [0x25BB]={
  category="so",
  direction="on",
 },
 [0x25BC]={
  category="so",
  direction="on",
 },
 [0x25BD]={
  category="so",
  direction="on",
 },
 [0x25BE]={
  category="so",
  direction="on",
 },
 [0x25BF]={
  category="so",
  direction="on",
 },
 [0x25C0]={
  category="so",
  direction="on",
 },
 [0x25C1]={
  category="sm",
  direction="on",
 },
 [0x25C2]={
  category="so",
  direction="on",
 },
 [0x25C3]={
  category="so",
  direction="on",
 },
 [0x25C4]={
  category="so",
  direction="on",
 },
 [0x25C5]={
  category="so",
  direction="on",
 },
 [0x25C6]={
  category="so",
  direction="on",
 },
 [0x25C7]={
  category="so",
  direction="on",
 },
 [0x25C8]={
  category="so",
  direction="on",
 },
 [0x25C9]={
  category="so",
  direction="on",
 },
 [0x25CA]={
  category="so",
  direction="on",
 },
 [0x25CB]={
  category="so",
  direction="on",
 },
 [0x25CC]={
  category="so",
  direction="on",
 },
 [0x25CD]={
  category="so",
  direction="on",
 },
 [0x25CE]={
  category="so",
  direction="on",
 },
 [0x25CF]={
  category="so",
  direction="on",
 },
 [0x25D0]={
  category="so",
  direction="on",
 },
 [0x25D1]={
  category="so",
  direction="on",
 },
 [0x25D2]={
  category="so",
  direction="on",
 },
 [0x25D3]={
  category="so",
  direction="on",
 },
 [0x25D4]={
  category="so",
  direction="on",
 },
 [0x25D5]={
  category="so",
  direction="on",
 },
 [0x25D6]={
  category="so",
  direction="on",
 },
 [0x25D7]={
  category="so",
  direction="on",
 },
 [0x25D8]={
  category="so",
  direction="on",
 },
 [0x25D9]={
  category="so",
  direction="on",
 },
 [0x25DA]={
  category="so",
  direction="on",
 },
 [0x25DB]={
  category="so",
  direction="on",
 },
 [0x25DC]={
  category="so",
  direction="on",
 },
 [0x25DD]={
  category="so",
  direction="on",
 },
 [0x25DE]={
  category="so",
  direction="on",
 },
 [0x25DF]={
  category="so",
  direction="on",
 },
 [0x25E0]={
  category="so",
  direction="on",
 },
 [0x25E1]={
  category="so",
  direction="on",
 },
 [0x25E2]={
  category="so",
  direction="on",
 },
 [0x25E3]={
  category="so",
  direction="on",
 },
 [0x25E4]={
  category="so",
  direction="on",
 },
 [0x25E5]={
  category="so",
  direction="on",
 },
 [0x25E6]={
  category="so",
  direction="on",
 },
 [0x25E7]={
  category="so",
  direction="on",
 },
 [0x25E8]={
  category="so",
  direction="on",
 },
 [0x25E9]={
  category="so",
  direction="on",
 },
 [0x25EA]={
  category="so",
  direction="on",
 },
 [0x25EB]={
  category="so",
  direction="on",
 },
 [0x25EC]={
  category="so",
  direction="on",
 },
 [0x25ED]={
  category="so",
  direction="on",
 },
 [0x25EE]={
  category="so",
  direction="on",
 },
 [0x25EF]={
  category="so",
  direction="on",
 },
 [0x25F0]={
  category="so",
  direction="on",
 },
 [0x25F1]={
  category="so",
  direction="on",
 },
 [0x25F2]={
  category="so",
  direction="on",
 },
 [0x25F3]={
  category="so",
  direction="on",
 },
 [0x25F4]={
  category="so",
  direction="on",
 },
 [0x25F5]={
  category="so",
  direction="on",
 },
 [0x25F6]={
  category="so",
  direction="on",
 },
 [0x25F7]={
  category="so",
  direction="on",
 },
 [0x25F8]={
  category="sm",
  direction="on",
 },
 [0x25F9]={
  category="sm",
  direction="on",
 },
 [0x25FA]={
  category="sm",
  direction="on",
 },
 [0x25FB]={
  category="sm",
  direction="on",
 },
 [0x25FC]={
  category="sm",
  direction="on",
 },
 [0x25FD]={
  category="sm",
  direction="on",
 },
 [0x25FE]={
  category="sm",
  direction="on",
 },
 [0x25FF]={
  category="sm",
  direction="on",
 },
 [0x2600]={
  category="so",
  direction="on",
 },
 [0x2601]={
  category="so",
  direction="on",
 },
 [0x2602]={
  category="so",
  direction="on",
 },
 [0x2603]={
  category="so",
  direction="on",
 },
 [0x2604]={
  category="so",
  direction="on",
 },
 [0x2605]={
  category="so",
  direction="on",
 },
 [0x2606]={
  category="so",
  direction="on",
 },
 [0x2607]={
  category="so",
  direction="on",
 },
 [0x2608]={
  category="so",
  direction="on",
 },
 [0x2609]={
  category="so",
  direction="on",
 },
 [0x260A]={
  category="so",
  direction="on",
 },
 [0x260B]={
  category="so",
  direction="on",
 },
 [0x260C]={
  category="so",
  direction="on",
 },
 [0x260D]={
  category="so",
  direction="on",
 },
 [0x260E]={
  category="so",
  direction="on",
 },
 [0x260F]={
  category="so",
  direction="on",
 },
 [0x2610]={
  category="so",
  direction="on",
 },
 [0x2611]={
  category="so",
  direction="on",
 },
 [0x2612]={
  category="so",
  direction="on",
 },
 [0x2613]={
  category="so",
  direction="on",
 },
 [0x2614]={
  category="so",
  direction="on",
 },
 [0x2615]={
  category="so",
  direction="on",
 },
 [0x2616]={
  category="so",
  direction="on",
 },
 [0x2617]={
  category="so",
  direction="on",
 },
 [0x2618]={
  category="so",
  direction="on",
 },
 [0x2619]={
  category="so",
  direction="on",
 },
 [0x261A]={
  category="so",
  direction="on",
 },
 [0x261B]={
  category="so",
  direction="on",
 },
 [0x261C]={
  category="so",
  direction="on",
 },
 [0x261D]={
  category="so",
  direction="on",
 },
 [0x261E]={
  category="so",
  direction="on",
 },
 [0x261F]={
  category="so",
  direction="on",
 },
 [0x2620]={
  category="so",
  direction="on",
 },
 [0x2621]={
  category="so",
  direction="on",
 },
 [0x2622]={
  category="so",
  direction="on",
 },
 [0x2623]={
  category="so",
  direction="on",
 },
 [0x2624]={
  category="so",
  direction="on",
 },
 [0x2625]={
  category="so",
  direction="on",
 },
 [0x2626]={
  category="so",
  direction="on",
 },
 [0x2627]={
  category="so",
  direction="on",
 },
 [0x2628]={
  category="so",
  direction="on",
 },
 [0x2629]={
  category="so",
  direction="on",
 },
 [0x262A]={
  category="so",
  direction="on",
 },
 [0x262B]={
  category="so",
  direction="on",
 },
 [0x262C]={
  category="so",
  direction="on",
 },
 [0x262D]={
  category="so",
  direction="on",
 },
 [0x262E]={
  category="so",
  direction="on",
 },
 [0x262F]={
  category="so",
  direction="on",
 },
 [0x2630]={
  category="so",
  direction="on",
 },
 [0x2631]={
  category="so",
  direction="on",
 },
 [0x2632]={
  category="so",
  direction="on",
 },
 [0x2633]={
  category="so",
  direction="on",
 },
 [0x2634]={
  category="so",
  direction="on",
 },
 [0x2635]={
  category="so",
  direction="on",
 },
 [0x2636]={
  category="so",
  direction="on",
 },
 [0x2637]={
  category="so",
  direction="on",
 },
 [0x2638]={
  category="so",
  direction="on",
 },
 [0x2639]={
  category="so",
  direction="on",
 },
 [0x263A]={
  category="so",
  direction="on",
 },
 [0x263B]={
  category="so",
  direction="on",
 },
 [0x263C]={
  category="so",
  direction="on",
 },
 [0x263D]={
  category="so",
  direction="on",
 },
 [0x263E]={
  category="so",
  direction="on",
 },
 [0x263F]={
  category="so",
  direction="on",
 },
 [0x2640]={
  category="so",
  direction="on",
 },
 [0x2641]={
  category="so",
  direction="on",
 },
 [0x2642]={
  category="so",
  direction="on",
 },
 [0x2643]={
  category="so",
  direction="on",
 },
 [0x2644]={
  category="so",
  direction="on",
 },
 [0x2645]={
  category="so",
  direction="on",
 },
 [0x2646]={
  category="so",
  direction="on",
 },
 [0x2647]={
  category="so",
  direction="on",
 },
 [0x2648]={
  category="so",
  direction="on",
 },
 [0x2649]={
  category="so",
  direction="on",
 },
 [0x264A]={
  category="so",
  direction="on",
 },
 [0x264B]={
  category="so",
  direction="on",
 },
 [0x264C]={
  category="so",
  direction="on",
 },
 [0x264D]={
  category="so",
  direction="on",
 },
 [0x264E]={
  category="so",
  direction="on",
 },
 [0x264F]={
  category="so",
  direction="on",
 },
 [0x2650]={
  category="so",
  direction="on",
 },
 [0x2651]={
  category="so",
  direction="on",
 },
 [0x2652]={
  category="so",
  direction="on",
 },
 [0x2653]={
  category="so",
  direction="on",
 },
 [0x2654]={
  category="so",
  direction="on",
 },
 [0x2655]={
  category="so",
  direction="on",
 },
 [0x2656]={
  category="so",
  direction="on",
 },
 [0x2657]={
  category="so",
  direction="on",
 },
 [0x2658]={
  category="so",
  direction="on",
 },
 [0x2659]={
  category="so",
  direction="on",
 },
 [0x265A]={
  category="so",
  direction="on",
 },
 [0x265B]={
  category="so",
  direction="on",
 },
 [0x265C]={
  category="so",
  direction="on",
 },
 [0x265D]={
  category="so",
  direction="on",
 },
 [0x265E]={
  category="so",
  direction="on",
 },
 [0x265F]={
  category="so",
  direction="on",
 },
 [0x2660]={
  category="so",
  direction="on",
 },
 [0x2661]={
  category="so",
  direction="on",
 },
 [0x2662]={
  category="so",
  direction="on",
 },
 [0x2663]={
  category="so",
  direction="on",
 },
 [0x2664]={
  category="so",
  direction="on",
 },
 [0x2665]={
  category="so",
  direction="on",
 },
 [0x2666]={
  category="so",
  direction="on",
 },
 [0x2667]={
  category="so",
  direction="on",
 },
 [0x2668]={
  category="so",
  direction="on",
 },
 [0x2669]={
  category="so",
  direction="on",
 },
 [0x266A]={
  category="so",
  direction="on",
 },
 [0x266B]={
  category="so",
  direction="on",
 },
 [0x266C]={
  category="so",
  direction="on",
 },
 [0x266D]={
  category="so",
  direction="on",
 },
 [0x266E]={
  category="so",
  direction="on",
 },
 [0x266F]={
  category="sm",
  direction="on",
 },
 [0x2670]={
  category="so",
  direction="on",
 },
 [0x2671]={
  category="so",
  direction="on",
 },
 [0x2672]={
  category="so",
  direction="on",
 },
 [0x2673]={
  category="so",
  direction="on",
 },
 [0x2674]={
  category="so",
  direction="on",
 },
 [0x2675]={
  category="so",
  direction="on",
 },
 [0x2676]={
  category="so",
  direction="on",
 },
 [0x2677]={
  category="so",
  direction="on",
 },
 [0x2678]={
  category="so",
  direction="on",
 },
 [0x2679]={
  category="so",
  direction="on",
 },
 [0x267A]={
  category="so",
  direction="on",
 },
 [0x267B]={
  category="so",
  direction="on",
 },
 [0x267C]={
  category="so",
  direction="on",
 },
 [0x267D]={
  category="so",
  direction="on",
 },
 [0x267E]={
  category="so",
  direction="on",
 },
 [0x267F]={
  category="so",
  direction="on",
 },
 [0x2680]={
  category="so",
  direction="on",
 },
 [0x2681]={
  category="so",
  direction="on",
 },
 [0x2682]={
  category="so",
  direction="on",
 },
 [0x2683]={
  category="so",
  direction="on",
 },
 [0x2684]={
  category="so",
  direction="on",
 },
 [0x2685]={
  category="so",
  direction="on",
 },
 [0x2686]={
  category="so",
  direction="on",
 },
 [0x2687]={
  category="so",
  direction="on",
 },
 [0x2688]={
  category="so",
  direction="on",
 },
 [0x2689]={
  category="so",
  direction="on",
 },
 [0x268A]={
  category="so",
  direction="on",
 },
 [0x268B]={
  category="so",
  direction="on",
 },
 [0x268C]={
  category="so",
  direction="on",
 },
 [0x268D]={
  category="so",
  direction="on",
 },
 [0x268E]={
  category="so",
  direction="on",
 },
 [0x268F]={
  category="so",
  direction="on",
 },
 [0x2690]={
  category="so",
  direction="on",
 },
 [0x2691]={
  category="so",
  direction="on",
 },
 [0x2692]={
  category="so",
  direction="on",
 },
 [0x2693]={
  category="so",
  direction="on",
 },
 [0x2694]={
  category="so",
  direction="on",
 },
 [0x2695]={
  category="so",
  direction="on",
 },
 [0x2696]={
  category="so",
  direction="on",
 },
 [0x2697]={
  category="so",
  direction="on",
 },
 [0x2698]={
  category="so",
  direction="on",
 },
 [0x2699]={
  category="so",
  direction="on",
 },
 [0x269A]={
  category="so",
  direction="on",
 },
 [0x269B]={
  category="so",
  direction="on",
 },
 [0x269C]={
  category="so",
  direction="on",
 },
 [0x269D]={
  category="so",
  direction="on",
 },
 [0x269E]={
  category="so",
  direction="on",
 },
 [0x269F]={
  category="so",
  direction="on",
 },
 [0x26A0]={
  category="so",
  direction="on",
 },
 [0x26A1]={
  category="so",
  direction="on",
 },
 [0x26A2]={
  category="so",
  direction="on",
 },
 [0x26A3]={
  category="so",
  direction="on",
 },
 [0x26A4]={
  category="so",
  direction="on",
 },
 [0x26A5]={
  category="so",
  direction="on",
 },
 [0x26A6]={
  category="so",
  direction="on",
 },
 [0x26A7]={
  category="so",
  direction="on",
 },
 [0x26A8]={
  category="so",
  direction="on",
 },
 [0x26A9]={
  category="so",
  direction="on",
 },
 [0x26AA]={
  category="so",
  direction="on",
 },
 [0x26AB]={
  category="so",
  direction="on",
 },
 [0x26AC]={
  category="so",
  direction="l",
 },
 [0x26AD]={
  category="so",
  direction="on",
 },
 [0x26AE]={
  category="so",
  direction="on",
 },
 [0x26AF]={
  category="so",
  direction="on",
 },
 [0x26B0]={
  category="so",
  direction="on",
 },
 [0x26B1]={
  category="so",
  direction="on",
 },
 [0x26B2]={
  category="so",
  direction="on",
 },
 [0x26B3]={
  category="so",
  direction="on",
 },
 [0x26B4]={
  category="so",
  direction="on",
 },
 [0x26B5]={
  category="so",
  direction="on",
 },
 [0x26B6]={
  category="so",
  direction="on",
 },
 [0x26B7]={
  category="so",
  direction="on",
 },
 [0x26B8]={
  category="so",
  direction="on",
 },
 [0x26B9]={
  category="so",
  direction="on",
 },
 [0x26BA]={
  category="so",
  direction="on",
 },
 [0x26BB]={
  category="so",
  direction="on",
 },
 [0x26BC]={
  category="so",
  direction="on",
 },
 [0x26BD]={
  category="so",
  direction="on",
 },
 [0x26BE]={
  category="so",
  direction="on",
 },
 [0x26BF]={
  category="so",
  direction="on",
 },
 [0x26C0]={
  category="so",
  direction="on",
 },
 [0x26C1]={
  category="so",
  direction="on",
 },
 [0x26C2]={
  category="so",
  direction="on",
 },
 [0x26C3]={
  category="so",
  direction="on",
 },
 [0x26C4]={
  category="so",
  direction="on",
 },
 [0x26C5]={
  category="so",
  direction="on",
 },
 [0x26C6]={
  category="so",
  direction="on",
 },
 [0x26C7]={
  category="so",
  direction="on",
 },
 [0x26C8]={
  category="so",
  direction="on",
 },
 [0x26C9]={
  category="so",
  direction="on",
 },
 [0x26CA]={
  category="so",
  direction="on",
 },
 [0x26CB]={
  category="so",
  direction="on",
 },
 [0x26CC]={
  category="so",
  direction="on",
 },
 [0x26CD]={
  category="so",
  direction="on",
 },
 [0x26CE]={
  category="so",
  direction="on",
 },
 [0x26CF]={
  category="so",
  direction="on",
 },
 [0x26D0]={
  category="so",
  direction="on",
 },
 [0x26D1]={
  category="so",
  direction="on",
 },
 [0x26D2]={
  category="so",
  direction="on",
 },
 [0x26D3]={
  category="so",
  direction="on",
 },
 [0x26D4]={
  category="so",
  direction="on",
 },
 [0x26D5]={
  category="so",
  direction="on",
 },
 [0x26D6]={
  category="so",
  direction="on",
 },
 [0x26D7]={
  category="so",
  direction="on",
 },
 [0x26D8]={
  category="so",
  direction="on",
 },
 [0x26D9]={
  category="so",
  direction="on",
 },
 [0x26DA]={
  category="so",
  direction="on",
 },
 [0x26DB]={
  category="so",
  direction="on",
 },
 [0x26DC]={
  category="so",
  direction="on",
 },
 [0x26DD]={
  category="so",
  direction="on",
 },
 [0x26DE]={
  category="so",
  direction="on",
 },
 [0x26DF]={
  category="so",
  direction="on",
 },
 [0x26E0]={
  category="so",
  direction="on",
 },
 [0x26E1]={
  category="so",
  direction="on",
 },
 [0x26E2]={
  category="so",
  direction="on",
 },
 [0x26E3]={
  category="so",
  direction="on",
 },
 [0x26E4]={
  category="so",
  direction="on",
 },
 [0x26E5]={
  category="so",
  direction="on",
 },
 [0x26E6]={
  category="so",
  direction="on",
 },
 [0x26E7]={
  category="so",
  direction="on",
 },
 [0x26E8]={
  category="so",
  direction="on",
 },
 [0x26E9]={
  category="so",
  direction="on",
 },
 [0x26EA]={
  category="so",
  direction="on",
 },
 [0x26EB]={
  category="so",
  direction="on",
 },
 [0x26EC]={
  category="so",
  direction="on",
 },
 [0x26ED]={
  category="so",
  direction="on",
 },
 [0x26EE]={
  category="so",
  direction="on",
 },
 [0x26EF]={
  category="so",
  direction="on",
 },
 [0x26F0]={
  category="so",
  direction="on",
 },
 [0x26F1]={
  category="so",
  direction="on",
 },
 [0x26F2]={
  category="so",
  direction="on",
 },
 [0x26F3]={
  category="so",
  direction="on",
 },
 [0x26F4]={
  category="so",
  direction="on",
 },
 [0x26F5]={
  category="so",
  direction="on",
 },
 [0x26F6]={
  category="so",
  direction="on",
 },
 [0x26F7]={
  category="so",
  direction="on",
 },
 [0x26F8]={
  category="so",
  direction="on",
 },
 [0x26F9]={
  category="so",
  direction="on",
 },
 [0x26FA]={
  category="so",
  direction="on",
 },
 [0x26FB]={
  category="so",
  direction="on",
 },
 [0x26FC]={
  category="so",
  direction="on",
 },
 [0x26FD]={
  category="so",
  direction="on",
 },
 [0x26FE]={
  category="so",
  direction="on",
 },
 [0x26FF]={
  category="so",
  direction="on",
 },
 [0x2700]={
  category="so",
  direction="on",
 },
 [0x2701]={
  category="so",
  direction="on",
 },
 [0x2702]={
  category="so",
  direction="on",
 },
 [0x2703]={
  category="so",
  direction="on",
 },
 [0x2704]={
  category="so",
  direction="on",
 },
 [0x2705]={
  category="so",
  direction="on",
 },
 [0x2706]={
  category="so",
  direction="on",
 },
 [0x2707]={
  category="so",
  direction="on",
 },
 [0x2708]={
  category="so",
  direction="on",
 },
 [0x2709]={
  category="so",
  direction="on",
 },
 [0x270A]={
  category="so",
  direction="on",
 },
 [0x270B]={
  category="so",
  direction="on",
 },
 [0x270C]={
  category="so",
  direction="on",
 },
 [0x270D]={
  category="so",
  direction="on",
 },
 [0x270E]={
  category="so",
  direction="on",
 },
 [0x270F]={
  category="so",
  direction="on",
 },
 [0x2710]={
  category="so",
  direction="on",
 },
 [0x2711]={
  category="so",
  direction="on",
 },
 [0x2712]={
  category="so",
  direction="on",
 },
 [0x2713]={
  category="so",
  direction="on",
 },
 [0x2714]={
  category="so",
  direction="on",
 },
 [0x2715]={
  category="so",
  direction="on",
 },
 [0x2716]={
  category="so",
  direction="on",
 },
 [0x2717]={
  category="so",
  direction="on",
 },
 [0x2718]={
  category="so",
  direction="on",
 },
 [0x2719]={
  category="so",
  direction="on",
 },
 [0x271A]={
  category="so",
  direction="on",
 },
 [0x271B]={
  category="so",
  direction="on",
 },
 [0x271C]={
  category="so",
  direction="on",
 },
 [0x271D]={
  category="so",
  direction="on",
 },
 [0x271E]={
  category="so",
  direction="on",
 },
 [0x271F]={
  category="so",
  direction="on",
 },
 [0x2720]={
  category="so",
  direction="on",
 },
 [0x2721]={
  category="so",
  direction="on",
 },
 [0x2722]={
  category="so",
  direction="on",
 },
 [0x2723]={
  category="so",
  direction="on",
 },
 [0x2724]={
  category="so",
  direction="on",
 },
 [0x2725]={
  category="so",
  direction="on",
 },
 [0x2726]={
  category="so",
  direction="on",
 },
 [0x2727]={
  category="so",
  direction="on",
 },
 [0x2728]={
  category="so",
  direction="on",
 },
 [0x2729]={
  category="so",
  direction="on",
 },
 [0x272A]={
  category="so",
  direction="on",
 },
 [0x272B]={
  category="so",
  direction="on",
 },
 [0x272C]={
  category="so",
  direction="on",
 },
 [0x272D]={
  category="so",
  direction="on",
 },
 [0x272E]={
  category="so",
  direction="on",
 },
 [0x272F]={
  category="so",
  direction="on",
 },
 [0x2730]={
  category="so",
  direction="on",
 },
 [0x2731]={
  category="so",
  direction="on",
 },
 [0x2732]={
  category="so",
  direction="on",
 },
 [0x2733]={
  category="so",
  direction="on",
 },
 [0x2734]={
  category="so",
  direction="on",
 },
 [0x2735]={
  category="so",
  direction="on",
 },
 [0x2736]={
  category="so",
  direction="on",
 },
 [0x2737]={
  category="so",
  direction="on",
 },
 [0x2738]={
  category="so",
  direction="on",
 },
 [0x2739]={
  category="so",
  direction="on",
 },
 [0x273A]={
  category="so",
  direction="on",
 },
 [0x273B]={
  category="so",
  direction="on",
 },
 [0x273C]={
  category="so",
  direction="on",
 },
 [0x273D]={
  category="so",
  direction="on",
 },
 [0x273E]={
  category="so",
  direction="on",
 },
 [0x273F]={
  category="so",
  direction="on",
 },
 [0x2740]={
  category="so",
  direction="on",
 },
 [0x2741]={
  category="so",
  direction="on",
 },
 [0x2742]={
  category="so",
  direction="on",
 },
 [0x2743]={
  category="so",
  direction="on",
 },
 [0x2744]={
  category="so",
  direction="on",
 },
 [0x2745]={
  category="so",
  direction="on",
 },
 [0x2746]={
  category="so",
  direction="on",
 },
 [0x2747]={
  category="so",
  direction="on",
 },
 [0x2748]={
  category="so",
  direction="on",
 },
 [0x2749]={
  category="so",
  direction="on",
 },
 [0x274A]={
  category="so",
  direction="on",
 },
 [0x274B]={
  category="so",
  direction="on",
 },
 [0x274C]={
  category="so",
  direction="on",
 },
 [0x274D]={
  category="so",
  direction="on",
 },
 [0x274E]={
  category="so",
  direction="on",
 },
 [0x274F]={
  category="so",
  direction="on",
 },
 [0x2750]={
  category="so",
  direction="on",
 },
 [0x2751]={
  category="so",
  direction="on",
 },
 [0x2752]={
  category="so",
  direction="on",
 },
 [0x2753]={
  category="so",
  direction="on",
 },
 [0x2754]={
  category="so",
  direction="on",
 },
 [0x2755]={
  category="so",
  direction="on",
 },
 [0x2756]={
  category="so",
  direction="on",
 },
 [0x2757]={
  category="so",
  direction="on",
 },
 [0x2758]={
  category="so",
  direction="on",
 },
 [0x2759]={
  category="so",
  direction="on",
 },
 [0x275A]={
  category="so",
  direction="on",
 },
 [0x275B]={
  category="so",
  direction="on",
 },
 [0x275C]={
  category="so",
  direction="on",
 },
 [0x275D]={
  category="so",
  direction="on",
 },
 [0x275E]={
  category="so",
  direction="on",
 },
 [0x275F]={
  category="so",
  direction="on",
 },
 [0x2760]={
  category="so",
  direction="on",
 },
 [0x2761]={
  category="so",
  direction="on",
 },
 [0x2762]={
  category="so",
  direction="on",
 },
 [0x2763]={
  category="so",
  direction="on",
 },
 [0x2764]={
  category="so",
  direction="on",
 },
 [0x2765]={
  category="so",
  direction="on",
 },
 [0x2766]={
  category="so",
  direction="on",
 },
 [0x2767]={
  category="so",
  direction="on",
 },
 [0x2768]={
  category="ps",
  direction="on",
  mirror=0x2769,
 },
 [0x2769]={
  category="pe",
  direction="on",
  mirror=0x2768,
 },
 [0x276A]={
  category="ps",
  direction="on",
  mirror=0x276B,
 },
 [0x276B]={
  category="pe",
  direction="on",
  mirror=0x276A,
 },
 [0x276C]={
  category="ps",
  direction="on",
  mirror=0x276D,
 },
 [0x276D]={
  category="pe",
  direction="on",
  mirror=0x276C,
 },
 [0x276E]={
  category="ps",
  direction="on",
  mirror=0x276F,
 },
 [0x276F]={
  category="pe",
  direction="on",
  mirror=0x276E,
 },
 [0x2770]={
  category="ps",
  direction="on",
  mirror=0x2771,
 },
 [0x2771]={
  category="pe",
  direction="on",
  mirror=0x2770,
 },
 [0x2772]={
  category="ps",
  direction="on",
  mirror=0x2773,
 },
 [0x2773]={
  category="pe",
  direction="on",
  mirror=0x2772,
 },
 [0x2774]={
  category="ps",
  direction="on",
  mirror=0x2775,
 },
 [0x2775]={
  category="pe",
  direction="on",
  mirror=0x2774,
 },
 [0x2776]={
  category="no",
  direction="on",
 },
 [0x2777]={
  category="no",
  direction="on",
 },
 [0x2778]={
  category="no",
  direction="on",
 },
 [0x2779]={
  category="no",
  direction="on",
 },
 [0x277A]={
  category="no",
  direction="on",
 },
 [0x277B]={
  category="no",
  direction="on",
 },
 [0x277C]={
  category="no",
  direction="on",
 },
 [0x277D]={
  category="no",
  direction="on",
 },
 [0x277E]={
  category="no",
  direction="on",
 },
 [0x277F]={
  category="no",
  direction="on",
 },
 [0x2780]={
  category="no",
  direction="on",
 },
 [0x2781]={
  category="no",
  direction="on",
 },
 [0x2782]={
  category="no",
  direction="on",
 },
 [0x2783]={
  category="no",
  direction="on",
 },
 [0x2784]={
  category="no",
  direction="on",
 },
 [0x2785]={
  category="no",
  direction="on",
 },
 [0x2786]={
  category="no",
  direction="on",
 },
 [0x2787]={
  category="no",
  direction="on",
 },
 [0x2788]={
  category="no",
  direction="on",
 },
 [0x2789]={
  category="no",
  direction="on",
 },
 [0x278A]={
  category="no",
  direction="on",
 },
 [0x278B]={
  category="no",
  direction="on",
 },
 [0x278C]={
  category="no",
  direction="on",
 },
 [0x278D]={
  category="no",
  direction="on",
 },
 [0x278E]={
  category="no",
  direction="on",
 },
 [0x278F]={
  category="no",
  direction="on",
 },
 [0x2790]={
  category="no",
  direction="on",
 },
 [0x2791]={
  category="no",
  direction="on",
 },
 [0x2792]={
  category="no",
  direction="on",
 },
 [0x2793]={
  category="no",
  direction="on",
 },
 [0x2794]={
  category="so",
  direction="on",
 },
 [0x2795]={
  category="so",
  direction="on",
 },
 [0x2796]={
  category="so",
  direction="on",
 },
 [0x2797]={
  category="so",
  direction="on",
 },
 [0x2798]={
  category="so",
  direction="on",
 },
 [0x2799]={
  category="so",
  direction="on",
 },
 [0x279A]={
  category="so",
  direction="on",
 },
 [0x279B]={
  category="so",
  direction="on",
 },
 [0x279C]={
  category="so",
  direction="on",
 },
 [0x279D]={
  category="so",
  direction="on",
 },
 [0x279E]={
  category="so",
  direction="on",
 },
 [0x279F]={
  category="so",
  direction="on",
 },
 [0x27A0]={
  category="so",
  direction="on",
 },
 [0x27A1]={
  category="so",
  direction="on",
 },
 [0x27A2]={
  category="so",
  direction="on",
 },
 [0x27A3]={
  category="so",
  direction="on",
 },
 [0x27A4]={
  category="so",
  direction="on",
 },
 [0x27A5]={
  category="so",
  direction="on",
 },
 [0x27A6]={
  category="so",
  direction="on",
 },
 [0x27A7]={
  category="so",
  direction="on",
 },
 [0x27A8]={
  category="so",
  direction="on",
 },
 [0x27A9]={
  category="so",
  direction="on",
 },
 [0x27AA]={
  category="so",
  direction="on",
 },
 [0x27AB]={
  category="so",
  direction="on",
 },
 [0x27AC]={
  category="so",
  direction="on",
 },
 [0x27AD]={
  category="so",
  direction="on",
 },
 [0x27AE]={
  category="so",
  direction="on",
 },
 [0x27AF]={
  category="so",
  direction="on",
 },
 [0x27B0]={
  category="so",
  direction="on",
 },
 [0x27B1]={
  category="so",
  direction="on",
 },
 [0x27B2]={
  category="so",
  direction="on",
 },
 [0x27B3]={
  category="so",
  direction="on",
 },
 [0x27B4]={
  category="so",
  direction="on",
 },
 [0x27B5]={
  category="so",
  direction="on",
 },
 [0x27B6]={
  category="so",
  direction="on",
 },
 [0x27B7]={
  category="so",
  direction="on",
 },
 [0x27B8]={
  category="so",
  direction="on",
 },
 [0x27B9]={
  category="so",
  direction="on",
 },
 [0x27BA]={
  category="so",
  direction="on",
 },
 [0x27BB]={
  category="so",
  direction="on",
 },
 [0x27BC]={
  category="so",
  direction="on",
 },
 [0x27BD]={
  category="so",
  direction="on",
 },
 [0x27BE]={
  category="so",
  direction="on",
 },
 [0x27BF]={
  category="so",
  direction="on",
 },
 [0x27C0]={
  category="sm",
  direction="on",
 },
 [0x27C1]={
  category="sm",
  direction="on",
 },
 [0x27C2]={
  category="sm",
  direction="on",
 },
 [0x27C3]={
  category="sm",
  direction="on",
  mirror=0x27C4,
 },
 [0x27C4]={
  category="sm",
  direction="on",
  mirror=0x27C3,
 },
 [0x27C5]={
  category="ps",
  direction="on",
  mirror=0x27C6,
 },
 [0x27C6]={
  category="pe",
  direction="on",
  mirror=0x27C5,
 },
 [0x27C7]={
  category="sm",
  direction="on",
 },
 [0x27C8]={
  category="sm",
  direction="on",
  mirror=0x27C9,
 },
 [0x27C9]={
  category="sm",
  direction="on",
  mirror=0x27C8,
 },
 [0x27CA]={
  category="sm",
  direction="on",
 },
 [0x27CB]={
  category="sm",
  direction="on",
  mirror=0x27CD,
 },
 [0x27CC]={
  category="sm",
  direction="on",
 },
 [0x27CD]={
  category="sm",
  direction="on",
  mirror=0x27CB,
 },
 [0x27CE]={
  category="sm",
  direction="on",
 },
 [0x27CF]={
  category="sm",
  direction="on",
 },
 [0x27D0]={
  category="sm",
  direction="on",
 },
 [0x27D1]={
  category="sm",
  direction="on",
 },
 [0x27D2]={
  category="sm",
  direction="on",
 },
 [0x27D3]={
  category="sm",
  direction="on",
 },
 [0x27D4]={
  category="sm",
  direction="on",
 },
 [0x27D5]={
  category="sm",
  direction="on",
  mirror=0x27D6,
 },
 [0x27D6]={
  category="sm",
  direction="on",
  mirror=0x27D5,
 },
 [0x27D7]={
  category="sm",
  direction="on",
 },
 [0x27D8]={
  category="sm",
  direction="on",
 },
 [0x27D9]={
  category="sm",
  direction="on",
 },
 [0x27DA]={
  category="sm",
  direction="on",
 },
 [0x27DB]={
  category="sm",
  direction="on",
 },
 [0x27DC]={
  category="sm",
  direction="on",
 },
 [0x27DD]={
  category="sm",
  direction="on",
  mirror=0x27DE,
 },
 [0x27DE]={
  category="sm",
  direction="on",
  mirror=0x27DD,
 },
 [0x27DF]={
  category="sm",
  direction="on",
 },
 [0x27E0]={
  category="sm",
  direction="on",
 },
 [0x27E1]={
  category="sm",
  direction="on",
 },
 [0x27E2]={
  category="sm",
  direction="on",
  mirror=0x27E3,
 },
 [0x27E3]={
  category="sm",
  direction="on",
  mirror=0x27E2,
 },
 [0x27E4]={
  category="sm",
  direction="on",
  mirror=0x27E5,
 },
 [0x27E5]={
  category="sm",
  direction="on",
  mirror=0x27E4,
 },
 [0x27E6]={
  category="ps",
  direction="on",
  mirror=0x27E7,
 },
 [0x27E7]={
  category="pe",
  direction="on",
  mirror=0x27E6,
 },
 [0x27E8]={
  category="ps",
  direction="on",
  mirror=0x27E9,
 },
 [0x27E9]={
  category="pe",
  direction="on",
  mirror=0x27E8,
 },
 [0x27EA]={
  category="ps",
  direction="on",
  mirror=0x27EB,
 },
 [0x27EB]={
  category="pe",
  direction="on",
  mirror=0x27EA,
 },
 [0x27EC]={
  category="ps",
  direction="on",
  mirror=0x27ED,
 },
 [0x27ED]={
  category="pe",
  direction="on",
  mirror=0x27EC,
 },
 [0x27EE]={
  category="ps",
  direction="on",
  mirror=0x27EF,
 },
 [0x27EF]={
  category="pe",
  direction="on",
  mirror=0x27EE,
 },
 [0x27F0]={
  category="sm",
  direction="on",
 },
 [0x27F1]={
  category="sm",
  direction="on",
 },
 [0x27F2]={
  category="sm",
  direction="on",
 },
 [0x27F3]={
  category="sm",
  direction="on",
 },
 [0x27F4]={
  category="sm",
  direction="on",
 },
 [0x27F5]={
  category="sm",
  direction="on",
 },
 [0x27F6]={
  category="sm",
  direction="on",
 },
 [0x27F7]={
  category="sm",
  direction="on",
 },
 [0x27F8]={
  category="sm",
  direction="on",
 },
 [0x27F9]={
  category="sm",
  direction="on",
 },
 [0x27FA]={
  category="sm",
  direction="on",
 },
 [0x27FB]={
  category="sm",
  direction="on",
 },
 [0x27FC]={
  category="sm",
  direction="on",
 },
 [0x27FD]={
  category="sm",
  direction="on",
 },
 [0x27FE]={
  category="sm",
  direction="on",
 },
 [0x27FF]={
  category="sm",
  direction="on",
 },
 [0x2800]={
  category="so",
  direction="l",
 },
 [0x2801]={
  category="so",
  direction="l",
 },
 [0x2802]={
  category="so",
  direction="l",
 },
 [0x2803]={
  category="so",
  direction="l",
 },
 [0x2804]={
  category="so",
  direction="l",
 },
 [0x2805]={
  category="so",
  direction="l",
 },
 [0x2806]={
  category="so",
  direction="l",
 },
 [0x2807]={
  category="so",
  direction="l",
 },
 [0x2808]={
  category="so",
  direction="l",
 },
 [0x2809]={
  category="so",
  direction="l",
 },
 [0x280A]={
  category="so",
  direction="l",
 },
 [0x280B]={
  category="so",
  direction="l",
 },
 [0x280C]={
  category="so",
  direction="l",
 },
 [0x280D]={
  category="so",
  direction="l",
 },
 [0x280E]={
  category="so",
  direction="l",
 },
 [0x280F]={
  category="so",
  direction="l",
 },
 [0x2810]={
  category="so",
  direction="l",
 },
 [0x2811]={
  category="so",
  direction="l",
 },
 [0x2812]={
  category="so",
  direction="l",
 },
 [0x2813]={
  category="so",
  direction="l",
 },
 [0x2814]={
  category="so",
  direction="l",
 },
 [0x2815]={
  category="so",
  direction="l",
 },
 [0x2816]={
  category="so",
  direction="l",
 },
 [0x2817]={
  category="so",
  direction="l",
 },
 [0x2818]={
  category="so",
  direction="l",
 },
 [0x2819]={
  category="so",
  direction="l",
 },
 [0x281A]={
  category="so",
  direction="l",
 },
 [0x281B]={
  category="so",
  direction="l",
 },
 [0x281C]={
  category="so",
  direction="l",
 },
 [0x281D]={
  category="so",
  direction="l",
 },
 [0x281E]={
  category="so",
  direction="l",
 },
 [0x281F]={
  category="so",
  direction="l",
 },
 [0x2820]={
  category="so",
  direction="l",
 },
 [0x2821]={
  category="so",
  direction="l",
 },
 [0x2822]={
  category="so",
  direction="l",
 },
 [0x2823]={
  category="so",
  direction="l",
 },
 [0x2824]={
  category="so",
  direction="l",
 },
 [0x2825]={
  category="so",
  direction="l",
 },
 [0x2826]={
  category="so",
  direction="l",
 },
 [0x2827]={
  category="so",
  direction="l",
 },
 [0x2828]={
  category="so",
  direction="l",
 },
 [0x2829]={
  category="so",
  direction="l",
 },
 [0x282A]={
  category="so",
  direction="l",
 },
 [0x282B]={
  category="so",
  direction="l",
 },
 [0x282C]={
  category="so",
  direction="l",
 },
 [0x282D]={
  category="so",
  direction="l",
 },
 [0x282E]={
  category="so",
  direction="l",
 },
 [0x282F]={
  category="so",
  direction="l",
 },
 [0x2830]={
  category="so",
  direction="l",
 },
 [0x2831]={
  category="so",
  direction="l",
 },
 [0x2832]={
  category="so",
  direction="l",
 },
 [0x2833]={
  category="so",
  direction="l",
 },
 [0x2834]={
  category="so",
  direction="l",
 },
 [0x2835]={
  category="so",
  direction="l",
 },
 [0x2836]={
  category="so",
  direction="l",
 },
 [0x2837]={
  category="so",
  direction="l",
 },
 [0x2838]={
  category="so",
  direction="l",
 },
 [0x2839]={
  category="so",
  direction="l",
 },
 [0x283A]={
  category="so",
  direction="l",
 },
 [0x283B]={
  category="so",
  direction="l",
 },
 [0x283C]={
  category="so",
  direction="l",
 },
 [0x283D]={
  category="so",
  direction="l",
 },
 [0x283E]={
  category="so",
  direction="l",
 },
 [0x283F]={
  category="so",
  direction="l",
 },
 [0x2840]={
  category="so",
  direction="l",
 },
 [0x2841]={
  category="so",
  direction="l",
 },
 [0x2842]={
  category="so",
  direction="l",
 },
 [0x2843]={
  category="so",
  direction="l",
 },
 [0x2844]={
  category="so",
  direction="l",
 },
 [0x2845]={
  category="so",
  direction="l",
 },
 [0x2846]={
  category="so",
  direction="l",
 },
 [0x2847]={
  category="so",
  direction="l",
 },
 [0x2848]={
  category="so",
  direction="l",
 },
 [0x2849]={
  category="so",
  direction="l",
 },
 [0x284A]={
  category="so",
  direction="l",
 },
 [0x284B]={
  category="so",
  direction="l",
 },
 [0x284C]={
  category="so",
  direction="l",
 },
 [0x284D]={
  category="so",
  direction="l",
 },
 [0x284E]={
  category="so",
  direction="l",
 },
 [0x284F]={
  category="so",
  direction="l",
 },
 [0x2850]={
  category="so",
  direction="l",
 },
 [0x2851]={
  category="so",
  direction="l",
 },
 [0x2852]={
  category="so",
  direction="l",
 },
 [0x2853]={
  category="so",
  direction="l",
 },
 [0x2854]={
  category="so",
  direction="l",
 },
 [0x2855]={
  category="so",
  direction="l",
 },
 [0x2856]={
  category="so",
  direction="l",
 },
 [0x2857]={
  category="so",
  direction="l",
 },
 [0x2858]={
  category="so",
  direction="l",
 },
 [0x2859]={
  category="so",
  direction="l",
 },
 [0x285A]={
  category="so",
  direction="l",
 },
 [0x285B]={
  category="so",
  direction="l",
 },
 [0x285C]={
  category="so",
  direction="l",
 },
 [0x285D]={
  category="so",
  direction="l",
 },
 [0x285E]={
  category="so",
  direction="l",
 },
 [0x285F]={
  category="so",
  direction="l",
 },
 [0x2860]={
  category="so",
  direction="l",
 },
 [0x2861]={
  category="so",
  direction="l",
 },
 [0x2862]={
  category="so",
  direction="l",
 },
 [0x2863]={
  category="so",
  direction="l",
 },
 [0x2864]={
  category="so",
  direction="l",
 },
 [0x2865]={
  category="so",
  direction="l",
 },
 [0x2866]={
  category="so",
  direction="l",
 },
 [0x2867]={
  category="so",
  direction="l",
 },
 [0x2868]={
  category="so",
  direction="l",
 },
 [0x2869]={
  category="so",
  direction="l",
 },
 [0x286A]={
  category="so",
  direction="l",
 },
 [0x286B]={
  category="so",
  direction="l",
 },
 [0x286C]={
  category="so",
  direction="l",
 },
 [0x286D]={
  category="so",
  direction="l",
 },
 [0x286E]={
  category="so",
  direction="l",
 },
 [0x286F]={
  category="so",
  direction="l",
 },
 [0x2870]={
  category="so",
  direction="l",
 },
 [0x2871]={
  category="so",
  direction="l",
 },
 [0x2872]={
  category="so",
  direction="l",
 },
 [0x2873]={
  category="so",
  direction="l",
 },
 [0x2874]={
  category="so",
  direction="l",
 },
 [0x2875]={
  category="so",
  direction="l",
 },
 [0x2876]={
  category="so",
  direction="l",
 },
 [0x2877]={
  category="so",
  direction="l",
 },
 [0x2878]={
  category="so",
  direction="l",
 },
 [0x2879]={
  category="so",
  direction="l",
 },
 [0x287A]={
  category="so",
  direction="l",
 },
 [0x287B]={
  category="so",
  direction="l",
 },
 [0x287C]={
  category="so",
  direction="l",
 },
 [0x287D]={
  category="so",
  direction="l",
 },
 [0x287E]={
  category="so",
  direction="l",
 },
 [0x287F]={
  category="so",
  direction="l",
 },
 [0x2880]={
  category="so",
  direction="l",
 },
 [0x2881]={
  category="so",
  direction="l",
 },
 [0x2882]={
  category="so",
  direction="l",
 },
 [0x2883]={
  category="so",
  direction="l",
 },
 [0x2884]={
  category="so",
  direction="l",
 },
 [0x2885]={
  category="so",
  direction="l",
 },
 [0x2886]={
  category="so",
  direction="l",
 },
 [0x2887]={
  category="so",
  direction="l",
 },
 [0x2888]={
  category="so",
  direction="l",
 },
 [0x2889]={
  category="so",
  direction="l",
 },
 [0x288A]={
  category="so",
  direction="l",
 },
 [0x288B]={
  category="so",
  direction="l",
 },
 [0x288C]={
  category="so",
  direction="l",
 },
 [0x288D]={
  category="so",
  direction="l",
 },
 [0x288E]={
  category="so",
  direction="l",
 },
 [0x288F]={
  category="so",
  direction="l",
 },
 [0x2890]={
  category="so",
  direction="l",
 },
 [0x2891]={
  category="so",
  direction="l",
 },
 [0x2892]={
  category="so",
  direction="l",
 },
 [0x2893]={
  category="so",
  direction="l",
 },
 [0x2894]={
  category="so",
  direction="l",
 },
 [0x2895]={
  category="so",
  direction="l",
 },
 [0x2896]={
  category="so",
  direction="l",
 },
 [0x2897]={
  category="so",
  direction="l",
 },
 [0x2898]={
  category="so",
  direction="l",
 },
 [0x2899]={
  category="so",
  direction="l",
 },
 [0x289A]={
  category="so",
  direction="l",
 },
 [0x289B]={
  category="so",
  direction="l",
 },
 [0x289C]={
  category="so",
  direction="l",
 },
 [0x289D]={
  category="so",
  direction="l",
 },
 [0x289E]={
  category="so",
  direction="l",
 },
 [0x289F]={
  category="so",
  direction="l",
 },
 [0x28A0]={
  category="so",
  direction="l",
 },
 [0x28A1]={
  category="so",
  direction="l",
 },
 [0x28A2]={
  category="so",
  direction="l",
 },
 [0x28A3]={
  category="so",
  direction="l",
 },
 [0x28A4]={
  category="so",
  direction="l",
 },
 [0x28A5]={
  category="so",
  direction="l",
 },
 [0x28A6]={
  category="so",
  direction="l",
 },
 [0x28A7]={
  category="so",
  direction="l",
 },
 [0x28A8]={
  category="so",
  direction="l",
 },
 [0x28A9]={
  category="so",
  direction="l",
 },
 [0x28AA]={
  category="so",
  direction="l",
 },
 [0x28AB]={
  category="so",
  direction="l",
 },
 [0x28AC]={
  category="so",
  direction="l",
 },
 [0x28AD]={
  category="so",
  direction="l",
 },
 [0x28AE]={
  category="so",
  direction="l",
 },
 [0x28AF]={
  category="so",
  direction="l",
 },
 [0x28B0]={
  category="so",
  direction="l",
 },
 [0x28B1]={
  category="so",
  direction="l",
 },
 [0x28B2]={
  category="so",
  direction="l",
 },
 [0x28B3]={
  category="so",
  direction="l",
 },
 [0x28B4]={
  category="so",
  direction="l",
 },
 [0x28B5]={
  category="so",
  direction="l",
 },
 [0x28B6]={
  category="so",
  direction="l",
 },
 [0x28B7]={
  category="so",
  direction="l",
 },
 [0x28B8]={
  category="so",
  direction="l",
 },
 [0x28B9]={
  category="so",
  direction="l",
 },
 [0x28BA]={
  category="so",
  direction="l",
 },
 [0x28BB]={
  category="so",
  direction="l",
 },
 [0x28BC]={
  category="so",
  direction="l",
 },
 [0x28BD]={
  category="so",
  direction="l",
 },
 [0x28BE]={
  category="so",
  direction="l",
 },
 [0x28BF]={
  category="so",
  direction="l",
 },
 [0x28C0]={
  category="so",
  direction="l",
 },
 [0x28C1]={
  category="so",
  direction="l",
 },
 [0x28C2]={
  category="so",
  direction="l",
 },
 [0x28C3]={
  category="so",
  direction="l",
 },
 [0x28C4]={
  category="so",
  direction="l",
 },
 [0x28C5]={
  category="so",
  direction="l",
 },
 [0x28C6]={
  category="so",
  direction="l",
 },
 [0x28C7]={
  category="so",
  direction="l",
 },
 [0x28C8]={
  category="so",
  direction="l",
 },
 [0x28C9]={
  category="so",
  direction="l",
 },
 [0x28CA]={
  category="so",
  direction="l",
 },
 [0x28CB]={
  category="so",
  direction="l",
 },
 [0x28CC]={
  category="so",
  direction="l",
 },
 [0x28CD]={
  category="so",
  direction="l",
 },
 [0x28CE]={
  category="so",
  direction="l",
 },
 [0x28CF]={
  category="so",
  direction="l",
 },
 [0x28D0]={
  category="so",
  direction="l",
 },
 [0x28D1]={
  category="so",
  direction="l",
 },
 [0x28D2]={
  category="so",
  direction="l",
 },
 [0x28D3]={
  category="so",
  direction="l",
 },
 [0x28D4]={
  category="so",
  direction="l",
 },
 [0x28D5]={
  category="so",
  direction="l",
 },
 [0x28D6]={
  category="so",
  direction="l",
 },
 [0x28D7]={
  category="so",
  direction="l",
 },
 [0x28D8]={
  category="so",
  direction="l",
 },
 [0x28D9]={
  category="so",
  direction="l",
 },
 [0x28DA]={
  category="so",
  direction="l",
 },
 [0x28DB]={
  category="so",
  direction="l",
 },
 [0x28DC]={
  category="so",
  direction="l",
 },
 [0x28DD]={
  category="so",
  direction="l",
 },
 [0x28DE]={
  category="so",
  direction="l",
 },
 [0x28DF]={
  category="so",
  direction="l",
 },
 [0x28E0]={
  category="so",
  direction="l",
 },
 [0x28E1]={
  category="so",
  direction="l",
 },
 [0x28E2]={
  category="so",
  direction="l",
 },
 [0x28E3]={
  category="so",
  direction="l",
 },
 [0x28E4]={
  category="so",
  direction="l",
 },
 [0x28E5]={
  category="so",
  direction="l",
 },
 [0x28E6]={
  category="so",
  direction="l",
 },
 [0x28E7]={
  category="so",
  direction="l",
 },
 [0x28E8]={
  category="so",
  direction="l",
 },
 [0x28E9]={
  category="so",
  direction="l",
 },
 [0x28EA]={
  category="so",
  direction="l",
 },
 [0x28EB]={
  category="so",
  direction="l",
 },
 [0x28EC]={
  category="so",
  direction="l",
 },
 [0x28ED]={
  category="so",
  direction="l",
 },
 [0x28EE]={
  category="so",
  direction="l",
 },
 [0x28EF]={
  category="so",
  direction="l",
 },
 [0x28F0]={
  category="so",
  direction="l",
 },
 [0x28F1]={
  category="so",
  direction="l",
 },
 [0x28F2]={
  category="so",
  direction="l",
 },
 [0x28F3]={
  category="so",
  direction="l",
 },
 [0x28F4]={
  category="so",
  direction="l",
 },
 [0x28F5]={
  category="so",
  direction="l",
 },
 [0x28F6]={
  category="so",
  direction="l",
 },
 [0x28F7]={
  category="so",
  direction="l",
 },
 [0x28F8]={
  category="so",
  direction="l",
 },
 [0x28F9]={
  category="so",
  direction="l",
 },
 [0x28FA]={
  category="so",
  direction="l",
 },
 [0x28FB]={
  category="so",
  direction="l",
 },
 [0x28FC]={
  category="so",
  direction="l",
 },
 [0x28FD]={
  category="so",
  direction="l",
 },
 [0x28FE]={
  category="so",
  direction="l",
 },
 [0x28FF]={
  category="so",
  direction="l",
 },
 [0x2900]={
  category="sm",
  direction="on",
 },
 [0x2901]={
  category="sm",
  direction="on",
 },
 [0x2902]={
  category="sm",
  direction="on",
 },
 [0x2903]={
  category="sm",
  direction="on",
 },
 [0x2904]={
  category="sm",
  direction="on",
 },
 [0x2905]={
  category="sm",
  direction="on",
 },
 [0x2906]={
  category="sm",
  direction="on",
 },
 [0x2907]={
  category="sm",
  direction="on",
 },
 [0x2908]={
  category="sm",
  direction="on",
 },
 [0x2909]={
  category="sm",
  direction="on",
 },
 [0x290A]={
  category="sm",
  direction="on",
 },
 [0x290B]={
  category="sm",
  direction="on",
 },
 [0x290C]={
  category="sm",
  direction="on",
 },
 [0x290D]={
  category="sm",
  direction="on",
 },
 [0x290E]={
  category="sm",
  direction="on",
 },
 [0x290F]={
  category="sm",
  direction="on",
 },
 [0x2910]={
  category="sm",
  direction="on",
 },
 [0x2911]={
  category="sm",
  direction="on",
 },
 [0x2912]={
  category="sm",
  direction="on",
 },
 [0x2913]={
  category="sm",
  direction="on",
 },
 [0x2914]={
  category="sm",
  direction="on",
 },
 [0x2915]={
  category="sm",
  direction="on",
 },
 [0x2916]={
  category="sm",
  direction="on",
 },
 [0x2917]={
  category="sm",
  direction="on",
 },
 [0x2918]={
  category="sm",
  direction="on",
 },
 [0x2919]={
  category="sm",
  direction="on",
 },
 [0x291A]={
  category="sm",
  direction="on",
 },
 [0x291B]={
  category="sm",
  direction="on",
 },
 [0x291C]={
  category="sm",
  direction="on",
 },
 [0x291D]={
  category="sm",
  direction="on",
 },
 [0x291E]={
  category="sm",
  direction="on",
 },
 [0x291F]={
  category="sm",
  direction="on",
 },
 [0x2920]={
  category="sm",
  direction="on",
 },
 [0x2921]={
  category="sm",
  direction="on",
 },
 [0x2922]={
  category="sm",
  direction="on",
 },
 [0x2923]={
  category="sm",
  direction="on",
 },
 [0x2924]={
  category="sm",
  direction="on",
 },
 [0x2925]={
  category="sm",
  direction="on",
 },
 [0x2926]={
  category="sm",
  direction="on",
 },
 [0x2927]={
  category="sm",
  direction="on",
 },
 [0x2928]={
  category="sm",
  direction="on",
 },
 [0x2929]={
  category="sm",
  direction="on",
 },
 [0x292A]={
  category="sm",
  direction="on",
 },
 [0x292B]={
  category="sm",
  direction="on",
 },
 [0x292C]={
  category="sm",
  direction="on",
 },
 [0x292D]={
  category="sm",
  direction="on",
 },
 [0x292E]={
  category="sm",
  direction="on",
 },
 [0x292F]={
  category="sm",
  direction="on",
 },
 [0x2930]={
  category="sm",
  direction="on",
 },
 [0x2931]={
  category="sm",
  direction="on",
 },
 [0x2932]={
  category="sm",
  direction="on",
 },
 [0x2933]={
  category="sm",
  direction="on",
 },
 [0x2934]={
  category="sm",
  direction="on",
 },
 [0x2935]={
  category="sm",
  direction="on",
 },
 [0x2936]={
  category="sm",
  direction="on",
 },
 [0x2937]={
  category="sm",
  direction="on",
 },
 [0x2938]={
  category="sm",
  direction="on",
 },
 [0x2939]={
  category="sm",
  direction="on",
 },
 [0x293A]={
  category="sm",
  direction="on",
 },
 [0x293B]={
  category="sm",
  direction="on",
 },
 [0x293C]={
  category="sm",
  direction="on",
 },
 [0x293D]={
  category="sm",
  direction="on",
 },
 [0x293E]={
  category="sm",
  direction="on",
 },
 [0x293F]={
  category="sm",
  direction="on",
 },
 [0x2940]={
  category="sm",
  direction="on",
 },
 [0x2941]={
  category="sm",
  direction="on",
 },
 [0x2942]={
  category="sm",
  direction="on",
 },
 [0x2943]={
  category="sm",
  direction="on",
 },
 [0x2944]={
  category="sm",
  direction="on",
 },
 [0x2945]={
  category="sm",
  direction="on",
 },
 [0x2946]={
  category="sm",
  direction="on",
 },
 [0x2947]={
  category="sm",
  direction="on",
 },
 [0x2948]={
  category="sm",
  direction="on",
 },
 [0x2949]={
  category="sm",
  direction="on",
 },
 [0x294A]={
  category="sm",
  direction="on",
 },
 [0x294B]={
  category="sm",
  direction="on",
 },
 [0x294C]={
  category="sm",
  direction="on",
 },
 [0x294D]={
  category="sm",
  direction="on",
 },
 [0x294E]={
  category="sm",
  direction="on",
 },
 [0x294F]={
  category="sm",
  direction="on",
 },
 [0x2950]={
  category="sm",
  direction="on",
 },
 [0x2951]={
  category="sm",
  direction="on",
 },
 [0x2952]={
  category="sm",
  direction="on",
 },
 [0x2953]={
  category="sm",
  direction="on",
 },
 [0x2954]={
  category="sm",
  direction="on",
 },
 [0x2955]={
  category="sm",
  direction="on",
 },
 [0x2956]={
  category="sm",
  direction="on",
 },
 [0x2957]={
  category="sm",
  direction="on",
 },
 [0x2958]={
  category="sm",
  direction="on",
 },
 [0x2959]={
  category="sm",
  direction="on",
 },
 [0x295A]={
  category="sm",
  direction="on",
 },
 [0x295B]={
  category="sm",
  direction="on",
 },
 [0x295C]={
  category="sm",
  direction="on",
 },
 [0x295D]={
  category="sm",
  direction="on",
 },
 [0x295E]={
  category="sm",
  direction="on",
 },
 [0x295F]={
  category="sm",
  direction="on",
 },
 [0x2960]={
  category="sm",
  direction="on",
 },
 [0x2961]={
  category="sm",
  direction="on",
 },
 [0x2962]={
  category="sm",
  direction="on",
 },
 [0x2963]={
  category="sm",
  direction="on",
 },
 [0x2964]={
  category="sm",
  direction="on",
 },
 [0x2965]={
  category="sm",
  direction="on",
 },
 [0x2966]={
  category="sm",
  direction="on",
 },
 [0x2967]={
  category="sm",
  direction="on",
 },
 [0x2968]={
  category="sm",
  direction="on",
 },
 [0x2969]={
  category="sm",
  direction="on",
 },
 [0x296A]={
  category="sm",
  direction="on",
 },
 [0x296B]={
  category="sm",
  direction="on",
 },
 [0x296C]={
  category="sm",
  direction="on",
 },
 [0x296D]={
  category="sm",
  direction="on",
 },
 [0x296E]={
  category="sm",
  direction="on",
 },
 [0x296F]={
  category="sm",
  direction="on",
 },
 [0x2970]={
  category="sm",
  direction="on",
 },
 [0x2971]={
  category="sm",
  direction="on",
 },
 [0x2972]={
  category="sm",
  direction="on",
 },
 [0x2973]={
  category="sm",
  direction="on",
 },
 [0x2974]={
  category="sm",
  direction="on",
 },
 [0x2975]={
  category="sm",
  direction="on",
 },
 [0x2976]={
  category="sm",
  direction="on",
 },
 [0x2977]={
  category="sm",
  direction="on",
 },
 [0x2978]={
  category="sm",
  direction="on",
 },
 [0x2979]={
  category="sm",
  direction="on",
 },
 [0x297A]={
  category="sm",
  direction="on",
 },
 [0x297B]={
  category="sm",
  direction="on",
 },
 [0x297C]={
  category="sm",
  direction="on",
 },
 [0x297D]={
  category="sm",
  direction="on",
 },
 [0x297E]={
  category="sm",
  direction="on",
 },
 [0x297F]={
  category="sm",
  direction="on",
 },
 [0x2980]={
  category="sm",
  direction="on",
 },
 [0x2981]={
  category="sm",
  direction="on",
 },
 [0x2982]={
  category="sm",
  direction="on",
 },
 [0x2983]={
  category="ps",
  direction="on",
  mirror=0x2984,
 },
 [0x2984]={
  category="pe",
  direction="on",
  mirror=0x2983,
 },
 [0x2985]={
  category="ps",
  direction="on",
  mirror=0x2986,
 },
 [0x2986]={
  category="pe",
  direction="on",
  mirror=0x2985,
 },
 [0x2987]={
  category="ps",
  direction="on",
  mirror=0x2988,
 },
 [0x2988]={
  category="pe",
  direction="on",
  mirror=0x2987,
 },
 [0x2989]={
  category="ps",
  direction="on",
  mirror=0x298A,
 },
 [0x298A]={
  category="pe",
  direction="on",
  mirror=0x2989,
 },
 [0x298B]={
  category="ps",
  direction="on",
  mirror=0x298C,
 },
 [0x298C]={
  category="pe",
  direction="on",
  mirror=0x298B,
 },
 [0x298D]={
  category="ps",
  direction="on",
  mirror=0x2990,
 },
 [0x298E]={
  category="pe",
  direction="on",
  mirror=0x298F,
 },
 [0x298F]={
  category="ps",
  direction="on",
  mirror=0x298E,
 },
 [0x2990]={
  category="pe",
  direction="on",
  mirror=0x298D,
 },
 [0x2991]={
  category="ps",
  direction="on",
  mirror=0x2992,
 },
 [0x2992]={
  category="pe",
  direction="on",
  mirror=0x2991,
 },
 [0x2993]={
  category="ps",
  direction="on",
  mirror=0x2994,
 },
 [0x2994]={
  category="pe",
  direction="on",
  mirror=0x2993,
 },
 [0x2995]={
  category="ps",
  direction="on",
  mirror=0x2996,
 },
 [0x2996]={
  category="pe",
  direction="on",
  mirror=0x2995,
 },
 [0x2997]={
  category="ps",
  direction="on",
  mirror=0x2998,
 },
 [0x2998]={
  category="pe",
  direction="on",
  mirror=0x2997,
 },
 [0x2999]={
  category="sm",
  direction="on",
 },
 [0x299A]={
  category="sm",
  direction="on",
 },
 [0x299B]={
  category="sm",
  direction="on",
 },
 [0x299C]={
  category="sm",
  direction="on",
 },
 [0x299D]={
  category="sm",
  direction="on",
 },
 [0x299E]={
  category="sm",
  direction="on",
 },
 [0x299F]={
  category="sm",
  direction="on",
 },
 [0x29A0]={
  category="sm",
  direction="on",
 },
 [0x29A1]={
  category="sm",
  direction="on",
 },
 [0x29A2]={
  category="sm",
  direction="on",
 },
 [0x29A3]={
  category="sm",
  direction="on",
 },
 [0x29A4]={
  category="sm",
  direction="on",
 },
 [0x29A5]={
  category="sm",
  direction="on",
 },
 [0x29A6]={
  category="sm",
  direction="on",
 },
 [0x29A7]={
  category="sm",
  direction="on",
 },
 [0x29A8]={
  category="sm",
  direction="on",
 },
 [0x29A9]={
  category="sm",
  direction="on",
 },
 [0x29AA]={
  category="sm",
  direction="on",
 },
 [0x29AB]={
  category="sm",
  direction="on",
 },
 [0x29AC]={
  category="sm",
  direction="on",
 },
 [0x29AD]={
  category="sm",
  direction="on",
 },
 [0x29AE]={
  category="sm",
  direction="on",
 },
 [0x29AF]={
  category="sm",
  direction="on",
 },
 [0x29B0]={
  category="sm",
  direction="on",
 },
 [0x29B1]={
  category="sm",
  direction="on",
 },
 [0x29B2]={
  category="sm",
  direction="on",
 },
 [0x29B3]={
  category="sm",
  direction="on",
 },
 [0x29B4]={
  category="sm",
  direction="on",
 },
 [0x29B5]={
  category="sm",
  direction="on",
 },
 [0x29B6]={
  category="sm",
  direction="on",
 },
 [0x29B7]={
  category="sm",
  direction="on",
 },
 [0x29B8]={
  category="sm",
  direction="on",
  mirror=0x2298,
 },
 [0x29B9]={
  category="sm",
  direction="on",
 },
 [0x29BA]={
  category="sm",
  direction="on",
 },
 [0x29BB]={
  category="sm",
  direction="on",
 },
 [0x29BC]={
  category="sm",
  direction="on",
 },
 [0x29BD]={
  category="sm",
  direction="on",
 },
 [0x29BE]={
  category="sm",
  direction="on",
 },
 [0x29BF]={
  category="sm",
  direction="on",
 },
 [0x29C0]={
  category="sm",
  direction="on",
  mirror=0x29C1,
 },
 [0x29C1]={
  category="sm",
  direction="on",
  mirror=0x29C0,
 },
 [0x29C2]={
  category="sm",
  direction="on",
 },
 [0x29C3]={
  category="sm",
  direction="on",
 },
 [0x29C4]={
  category="sm",
  direction="on",
  mirror=0x29C5,
 },
 [0x29C5]={
  category="sm",
  direction="on",
  mirror=0x29C4,
 },
 [0x29C6]={
  category="sm",
  direction="on",
 },
 [0x29C7]={
  category="sm",
  direction="on",
 },
 [0x29C8]={
  category="sm",
  direction="on",
 },
 [0x29C9]={
  category="sm",
  direction="on",
 },
 [0x29CA]={
  category="sm",
  direction="on",
 },
 [0x29CB]={
  category="sm",
  direction="on",
 },
 [0x29CC]={
  category="sm",
  direction="on",
 },
 [0x29CD]={
  category="sm",
  direction="on",
 },
 [0x29CE]={
  category="sm",
  direction="on",
 },
 [0x29CF]={
  category="sm",
  direction="on",
  mirror=0x29D0,
 },
 [0x29D0]={
  category="sm",
  direction="on",
  mirror=0x29CF,
 },
 [0x29D1]={
  category="sm",
  direction="on",
  mirror=0x29D2,
 },
 [0x29D2]={
  category="sm",
  direction="on",
  mirror=0x29D1,
 },
 [0x29D3]={
  category="sm",
  direction="on",
 },
 [0x29D4]={
  category="sm",
  direction="on",
  mirror=0x29D5,
 },
 [0x29D5]={
  category="sm",
  direction="on",
  mirror=0x29D4,
 },
 [0x29D6]={
  category="sm",
  direction="on",
 },
 [0x29D7]={
  category="sm",
  direction="on",
 },
 [0x29D8]={
  category="ps",
  direction="on",
  mirror=0x29D9,
 },
 [0x29D9]={
  category="pe",
  direction="on",
  mirror=0x29D8,
 },
 [0x29DA]={
  category="ps",
  direction="on",
  mirror=0x29DB,
 },
 [0x29DB]={
  category="pe",
  direction="on",
  mirror=0x29DA,
 },
 [0x29DC]={
  category="sm",
  direction="on",
 },
 [0x29DD]={
  category="sm",
  direction="on",
 },
 [0x29DE]={
  category="sm",
  direction="on",
 },
 [0x29DF]={
  category="sm",
  direction="on",
 },
 [0x29E0]={
  category="sm",
  direction="on",
 },
 [0x29E1]={
  category="sm",
  direction="on",
 },
 [0x29E2]={
  category="sm",
  direction="on",
 },
 [0x29E3]={
  category="sm",
  direction="on",
 },
 [0x29E4]={
  category="sm",
  direction="on",
 },
 [0x29E5]={
  category="sm",
  direction="on",
 },
 [0x29E6]={
  category="sm",
  direction="on",
 },
 [0x29E7]={
  category="sm",
  direction="on",
 },
 [0x29E8]={
  category="sm",
  direction="on",
 },
 [0x29E9]={
  category="sm",
  direction="on",
 },
 [0x29EA]={
  category="sm",
  direction="on",
 },
 [0x29EB]={
  category="sm",
  direction="on",
 },
 [0x29EC]={
  category="sm",
  direction="on",
 },
 [0x29ED]={
  category="sm",
  direction="on",
 },
 [0x29EE]={
  category="sm",
  direction="on",
 },
 [0x29EF]={
  category="sm",
  direction="on",
 },
 [0x29F0]={
  category="sm",
  direction="on",
 },
 [0x29F1]={
  category="sm",
  direction="on",
 },
 [0x29F2]={
  category="sm",
  direction="on",
 },
 [0x29F3]={
  category="sm",
  direction="on",
 },
 [0x29F4]={
  category="sm",
  direction="on",
 },
 [0x29F5]={
  category="sm",
  direction="on",
  mirror=0x2215,
 },
 [0x29F6]={
  category="sm",
  direction="on",
 },
 [0x29F7]={
  category="sm",
  direction="on",
 },
 [0x29F8]={
  category="sm",
  direction="on",
  mirror=0x29F9,
 },
 [0x29F9]={
  category="sm",
  direction="on",
  mirror=0x29F8,
 },
 [0x29FA]={
  category="sm",
  direction="on",
 },
 [0x29FB]={
  category="sm",
  direction="on",
 },
 [0x29FC]={
  category="ps",
  direction="on",
  mirror=0x29FD,
 },
 [0x29FD]={
  category="pe",
  direction="on",
  mirror=0x29FC,
 },
 [0x29FE]={
  category="sm",
  direction="on",
 },
 [0x29FF]={
  category="sm",
  direction="on",
 },
 [0x2A00]={
  category="sm",
  direction="on",
 },
 [0x2A01]={
  category="sm",
  direction="on",
 },
 [0x2A02]={
  category="sm",
  direction="on",
 },
 [0x2A03]={
  category="sm",
  direction="on",
 },
 [0x2A04]={
  category="sm",
  direction="on",
 },
 [0x2A05]={
  category="sm",
  direction="on",
 },
 [0x2A06]={
  category="sm",
  direction="on",
 },
 [0x2A07]={
  category="sm",
  direction="on",
 },
 [0x2A08]={
  category="sm",
  direction="on",
 },
 [0x2A09]={
  category="sm",
  direction="on",
 },
 [0x2A0A]={
  category="sm",
  direction="on",
 },
 [0x2A0B]={
  category="sm",
  direction="on",
 },
 [0x2A0C]={
  category="sm",
  direction="on",
 },
 [0x2A0D]={
  category="sm",
  direction="on",
 },
 [0x2A0E]={
  category="sm",
  direction="on",
 },
 [0x2A0F]={
  category="sm",
  direction="on",
 },
 [0x2A10]={
  category="sm",
  direction="on",
 },
 [0x2A11]={
  category="sm",
  direction="on",
 },
 [0x2A12]={
  category="sm",
  direction="on",
 },
 [0x2A13]={
  category="sm",
  direction="on",
 },
 [0x2A14]={
  category="sm",
  direction="on",
 },
 [0x2A15]={
  category="sm",
  direction="on",
 },
 [0x2A16]={
  category="sm",
  direction="on",
 },
 [0x2A17]={
  category="sm",
  direction="on",
 },
 [0x2A18]={
  category="sm",
  direction="on",
 },
 [0x2A19]={
  category="sm",
  direction="on",
 },
 [0x2A1A]={
  category="sm",
  direction="on",
 },
 [0x2A1B]={
  category="sm",
  direction="on",
 },
 [0x2A1C]={
  category="sm",
  direction="on",
 },
 [0x2A1D]={
  category="sm",
  direction="on",
 },
 [0x2A1E]={
  category="sm",
  direction="on",
 },
 [0x2A1F]={
  category="sm",
  direction="on",
 },
 [0x2A20]={
  category="sm",
  direction="on",
 },
 [0x2A21]={
  category="sm",
  direction="on",
 },
 [0x2A22]={
  category="sm",
  direction="on",
 },
 [0x2A23]={
  category="sm",
  direction="on",
 },
 [0x2A24]={
  category="sm",
  direction="on",
 },
 [0x2A25]={
  category="sm",
  direction="on",
 },
 [0x2A26]={
  category="sm",
  direction="on",
 },
 [0x2A27]={
  category="sm",
  direction="on",
 },
 [0x2A28]={
  category="sm",
  direction="on",
 },
 [0x2A29]={
  category="sm",
  direction="on",
 },
 [0x2A2A]={
  category="sm",
  direction="on",
 },
 [0x2A2B]={
  category="sm",
  direction="on",
  mirror=0x2A2C,
 },
 [0x2A2C]={
  category="sm",
  direction="on",
  mirror=0x2A2B,
 },
 [0x2A2D]={
  category="sm",
  direction="on",
  mirror=0x2A2E,
 },
 [0x2A2E]={
  category="sm",
  direction="on",
  mirror=0x2A2D,
 },
 [0x2A2F]={
  category="sm",
  direction="on",
 },
 [0x2A30]={
  category="sm",
  direction="on",
 },
 [0x2A31]={
  category="sm",
  direction="on",
 },
 [0x2A32]={
  category="sm",
  direction="on",
 },
 [0x2A33]={
  category="sm",
  direction="on",
 },
 [0x2A34]={
  category="sm",
  direction="on",
  mirror=0x2A35,
 },
 [0x2A35]={
  category="sm",
  direction="on",
  mirror=0x2A34,
 },
 [0x2A36]={
  category="sm",
  direction="on",
 },
 [0x2A37]={
  category="sm",
  direction="on",
 },
 [0x2A38]={
  category="sm",
  direction="on",
 },
 [0x2A39]={
  category="sm",
  direction="on",
 },
 [0x2A3A]={
  category="sm",
  direction="on",
 },
 [0x2A3B]={
  category="sm",
  direction="on",
 },
 [0x2A3C]={
  category="sm",
  direction="on",
  mirror=0x2A3D,
 },
 [0x2A3D]={
  category="sm",
  direction="on",
  mirror=0x2A3C,
 },
 [0x2A3E]={
  category="sm",
  direction="on",
 },
 [0x2A3F]={
  category="sm",
  direction="on",
 },
 [0x2A40]={
  category="sm",
  direction="on",
 },
 [0x2A41]={
  category="sm",
  direction="on",
 },
 [0x2A42]={
  category="sm",
  direction="on",
 },
 [0x2A43]={
  category="sm",
  direction="on",
 },
 [0x2A44]={
  category="sm",
  direction="on",
 },
 [0x2A45]={
  category="sm",
  direction="on",
 },
 [0x2A46]={
  category="sm",
  direction="on",
 },
 [0x2A47]={
  category="sm",
  direction="on",
 },
 [0x2A48]={
  category="sm",
  direction="on",
 },
 [0x2A49]={
  category="sm",
  direction="on",
 },
 [0x2A4A]={
  category="sm",
  direction="on",
 },
 [0x2A4B]={
  category="sm",
  direction="on",
 },
 [0x2A4C]={
  category="sm",
  direction="on",
 },
 [0x2A4D]={
  category="sm",
  direction="on",
 },
 [0x2A4E]={
  category="sm",
  direction="on",
 },
 [0x2A4F]={
  category="sm",
  direction="on",
 },
 [0x2A50]={
  category="sm",
  direction="on",
 },
 [0x2A51]={
  category="sm",
  direction="on",
 },
 [0x2A52]={
  category="sm",
  direction="on",
 },
 [0x2A53]={
  category="sm",
  direction="on",
 },
 [0x2A54]={
  category="sm",
  direction="on",
 },
 [0x2A55]={
  category="sm",
  direction="on",
 },
 [0x2A56]={
  category="sm",
  direction="on",
 },
 [0x2A57]={
  category="sm",
  direction="on",
 },
 [0x2A58]={
  category="sm",
  direction="on",
 },
 [0x2A59]={
  category="sm",
  direction="on",
 },
 [0x2A5A]={
  category="sm",
  direction="on",
 },
 [0x2A5B]={
  category="sm",
  direction="on",
 },
 [0x2A5C]={
  category="sm",
  direction="on",
 },
 [0x2A5D]={
  category="sm",
  direction="on",
 },
 [0x2A5E]={
  category="sm",
  direction="on",
 },
 [0x2A5F]={
  category="sm",
  direction="on",
 },
 [0x2A60]={
  category="sm",
  direction="on",
 },
 [0x2A61]={
  category="sm",
  direction="on",
 },
 [0x2A62]={
  category="sm",
  direction="on",
 },
 [0x2A63]={
  category="sm",
  direction="on",
 },
 [0x2A64]={
  category="sm",
  direction="on",
  mirror=0x2A65,
 },
 [0x2A65]={
  category="sm",
  direction="on",
  mirror=0x2A64,
 },
 [0x2A66]={
  category="sm",
  direction="on",
 },
 [0x2A67]={
  category="sm",
  direction="on",
 },
 [0x2A68]={
  category="sm",
  direction="on",
 },
 [0x2A69]={
  category="sm",
  direction="on",
 },
 [0x2A6A]={
  category="sm",
  direction="on",
 },
 [0x2A6B]={
  category="sm",
  direction="on",
 },
 [0x2A6C]={
  category="sm",
  direction="on",
 },
 [0x2A6D]={
  category="sm",
  direction="on",
 },
 [0x2A6E]={
  category="sm",
  direction="on",
 },
 [0x2A6F]={
  category="sm",
  direction="on",
 },
 [0x2A70]={
  category="sm",
  direction="on",
 },
 [0x2A71]={
  category="sm",
  direction="on",
 },
 [0x2A72]={
  category="sm",
  direction="on",
 },
 [0x2A73]={
  category="sm",
  direction="on",
 },
 [0x2A74]={
  category="sm",
  direction="on",
 },
 [0x2A75]={
  category="sm",
  direction="on",
 },
 [0x2A76]={
  category="sm",
  direction="on",
 },
 [0x2A77]={
  category="sm",
  direction="on",
 },
 [0x2A78]={
  category="sm",
  direction="on",
 },
 [0x2A79]={
  category="sm",
  direction="on",
  mirror=0x2A7A,
 },
 [0x2A7A]={
  category="sm",
  direction="on",
  mirror=0x2A79,
 },
 [0x2A7B]={
  category="sm",
  direction="on",
 },
 [0x2A7C]={
  category="sm",
  direction="on",
 },
 [0x2A7D]={
  category="sm",
  direction="on",
  mirror=0x2A7E,
 },
 [0x2A7E]={
  category="sm",
  direction="on",
  mirror=0x2A7D,
 },
 [0x2A7F]={
  category="sm",
  direction="on",
  mirror=0x2A80,
 },
 [0x2A80]={
  category="sm",
  direction="on",
  mirror=0x2A7F,
 },
 [0x2A81]={
  category="sm",
  direction="on",
  mirror=0x2A82,
 },
 [0x2A82]={
  category="sm",
  direction="on",
  mirror=0x2A81,
 },
 [0x2A83]={
  category="sm",
  direction="on",
  mirror=0x2A84,
 },
 [0x2A84]={
  category="sm",
  direction="on",
  mirror=0x2A83,
 },
 [0x2A85]={
  category="sm",
  direction="on",
 },
 [0x2A86]={
  category="sm",
  direction="on",
 },
 [0x2A87]={
  category="sm",
  direction="on",
 },
 [0x2A88]={
  category="sm",
  direction="on",
 },
 [0x2A89]={
  category="sm",
  direction="on",
 },
 [0x2A8A]={
  category="sm",
  direction="on",
 },
 [0x2A8B]={
  category="sm",
  direction="on",
  mirror=0x2A8C,
 },
 [0x2A8C]={
  category="sm",
  direction="on",
  mirror=0x2A8B,
 },
 [0x2A8D]={
  category="sm",
  direction="on",
 },
 [0x2A8E]={
  category="sm",
  direction="on",
 },
 [0x2A8F]={
  category="sm",
  direction="on",
 },
 [0x2A90]={
  category="sm",
  direction="on",
 },
 [0x2A91]={
  category="sm",
  direction="on",
  mirror=0x2A92,
 },
 [0x2A92]={
  category="sm",
  direction="on",
  mirror=0x2A91,
 },
 [0x2A93]={
  category="sm",
  direction="on",
  mirror=0x2A94,
 },
 [0x2A94]={
  category="sm",
  direction="on",
  mirror=0x2A93,
 },
 [0x2A95]={
  category="sm",
  direction="on",
  mirror=0x2A96,
 },
 [0x2A96]={
  category="sm",
  direction="on",
  mirror=0x2A95,
 },
 [0x2A97]={
  category="sm",
  direction="on",
  mirror=0x2A98,
 },
 [0x2A98]={
  category="sm",
  direction="on",
  mirror=0x2A97,
 },
 [0x2A99]={
  category="sm",
  direction="on",
  mirror=0x2A9A,
 },
 [0x2A9A]={
  category="sm",
  direction="on",
  mirror=0x2A99,
 },
 [0x2A9B]={
  category="sm",
  direction="on",
  mirror=0x2A9C,
 },
 [0x2A9C]={
  category="sm",
  direction="on",
  mirror=0x2A9B,
 },
 [0x2A9D]={
  category="sm",
  direction="on",
 },
 [0x2A9E]={
  category="sm",
  direction="on",
 },
 [0x2A9F]={
  category="sm",
  direction="on",
 },
 [0x2AA0]={
  category="sm",
  direction="on",
 },
 [0x2AA1]={
  category="sm",
  direction="on",
  mirror=0x2AA2,
 },
 [0x2AA2]={
  category="sm",
  direction="on",
  mirror=0x2AA1,
 },
 [0x2AA3]={
  category="sm",
  direction="on",
 },
 [0x2AA4]={
  category="sm",
  direction="on",
 },
 [0x2AA5]={
  category="sm",
  direction="on",
 },
 [0x2AA6]={
  category="sm",
  direction="on",
  mirror=0x2AA7,
 },
 [0x2AA7]={
  category="sm",
  direction="on",
  mirror=0x2AA6,
 },
 [0x2AA8]={
  category="sm",
  direction="on",
  mirror=0x2AA9,
 },
 [0x2AA9]={
  category="sm",
  direction="on",
  mirror=0x2AA8,
 },
 [0x2AAA]={
  category="sm",
  direction="on",
  mirror=0x2AAB,
 },
 [0x2AAB]={
  category="sm",
  direction="on",
  mirror=0x2AAA,
 },
 [0x2AAC]={
  category="sm",
  direction="on",
  mirror=0x2AAD,
 },
 [0x2AAD]={
  category="sm",
  direction="on",
  mirror=0x2AAC,
 },
 [0x2AAE]={
  category="sm",
  direction="on",
 },
 [0x2AAF]={
  category="sm",
  direction="on",
  mirror=0x2AB0,
 },
 [0x2AB0]={
  category="sm",
  direction="on",
  mirror=0x2AAF,
 },
 [0x2AB1]={
  category="sm",
  direction="on",
 },
 [0x2AB2]={
  category="sm",
  direction="on",
 },
 [0x2AB3]={
  category="sm",
  direction="on",
  mirror=0x2AB4,
 },
 [0x2AB4]={
  category="sm",
  direction="on",
  mirror=0x2AB3,
 },
 [0x2AB5]={
  category="sm",
  direction="on",
 },
 [0x2AB6]={
  category="sm",
  direction="on",
 },
 [0x2AB7]={
  category="sm",
  direction="on",
 },
 [0x2AB8]={
  category="sm",
  direction="on",
 },
 [0x2AB9]={
  category="sm",
  direction="on",
 },
 [0x2ABA]={
  category="sm",
  direction="on",
 },
 [0x2ABB]={
  category="sm",
  direction="on",
  mirror=0x2ABC,
 },
 [0x2ABC]={
  category="sm",
  direction="on",
  mirror=0x2ABB,
 },
 [0x2ABD]={
  category="sm",
  direction="on",
  mirror=0x2ABE,
 },
 [0x2ABE]={
  category="sm",
  direction="on",
  mirror=0x2ABD,
 },
 [0x2ABF]={
  category="sm",
  direction="on",
  mirror=0x2AC0,
 },
 [0x2AC0]={
  category="sm",
  direction="on",
  mirror=0x2ABF,
 },
 [0x2AC1]={
  category="sm",
  direction="on",
  mirror=0x2AC2,
 },
 [0x2AC2]={
  category="sm",
  direction="on",
  mirror=0x2AC1,
 },
 [0x2AC3]={
  category="sm",
  direction="on",
  mirror=0x2AC4,
 },
 [0x2AC4]={
  category="sm",
  direction="on",
  mirror=0x2AC3,
 },
 [0x2AC5]={
  category="sm",
  direction="on",
  mirror=0x2AC6,
 },
 [0x2AC6]={
  category="sm",
  direction="on",
  mirror=0x2AC5,
 },
 [0x2AC7]={
  category="sm",
  direction="on",
 },
 [0x2AC8]={
  category="sm",
  direction="on",
 },
 [0x2AC9]={
  category="sm",
  direction="on",
 },
 [0x2ACA]={
  category="sm",
  direction="on",
 },
 [0x2ACB]={
  category="sm",
  direction="on",
 },
 [0x2ACC]={
  category="sm",
  direction="on",
 },
 [0x2ACD]={
  category="sm",
  direction="on",
  mirror=0x2ACE,
 },
 [0x2ACE]={
  category="sm",
  direction="on",
  mirror=0x2ACD,
 },
 [0x2ACF]={
  category="sm",
  direction="on",
  mirror=0x2AD0,
 },
 [0x2AD0]={
  category="sm",
  direction="on",
  mirror=0x2ACF,
 },
 [0x2AD1]={
  category="sm",
  direction="on",
  mirror=0x2AD2,
 },
 [0x2AD2]={
  category="sm",
  direction="on",
  mirror=0x2AD1,
 },
 [0x2AD3]={
  category="sm",
  direction="on",
  mirror=0x2AD4,
 },
 [0x2AD4]={
  category="sm",
  direction="on",
  mirror=0x2AD3,
 },
 [0x2AD5]={
  category="sm",
  direction="on",
  mirror=0x2AD6,
 },
 [0x2AD6]={
  category="sm",
  direction="on",
  mirror=0x2AD5,
 },
 [0x2AD7]={
  category="sm",
  direction="on",
 },
 [0x2AD8]={
  category="sm",
  direction="on",
 },
 [0x2AD9]={
  category="sm",
  direction="on",
 },
 [0x2ADA]={
  category="sm",
  direction="on",
 },
 [0x2ADB]={
  category="sm",
  direction="on",
 },
 [0x2ADC]={
  category="sm",
  direction="on",
 },
 [0x2ADD]={
  category="sm",
  direction="on",
 },
 [0x2ADE]={
  category="sm",
  direction="on",
  mirror=0x22A6,
 },
 [0x2ADF]={
  category="sm",
  direction="on",
 },
 [0x2AE0]={
  category="sm",
  direction="on",
 },
 [0x2AE1]={
  category="sm",
  direction="on",
 },
 [0x2AE2]={
  category="sm",
  direction="on",
 },
 [0x2AE3]={
  category="sm",
  direction="on",
  mirror=0x22A9,
 },
 [0x2AE4]={
  category="sm",
  direction="on",
  mirror=0x22A8,
 },
 [0x2AE5]={
  category="sm",
  direction="on",
  mirror=0x22AB,
 },
 [0x2AE6]={
  category="sm",
  direction="on",
 },
 [0x2AE7]={
  category="sm",
  direction="on",
 },
 [0x2AE8]={
  category="sm",
  direction="on",
 },
 [0x2AE9]={
  category="sm",
  direction="on",
 },
 [0x2AEA]={
  category="sm",
  direction="on",
 },
 [0x2AEB]={
  category="sm",
  direction="on",
 },
 [0x2AEC]={
  category="sm",
  direction="on",
  mirror=0x2AED,
 },
 [0x2AED]={
  category="sm",
  direction="on",
  mirror=0x2AEC,
 },
 [0x2AEE]={
  category="sm",
  direction="on",
 },
 [0x2AEF]={
  category="sm",
  direction="on",
 },
 [0x2AF0]={
  category="sm",
  direction="on",
 },
 [0x2AF1]={
  category="sm",
  direction="on",
 },
 [0x2AF2]={
  category="sm",
  direction="on",
 },
 [0x2AF3]={
  category="sm",
  direction="on",
 },
 [0x2AF4]={
  category="sm",
  direction="on",
 },
 [0x2AF5]={
  category="sm",
  direction="on",
 },
 [0x2AF6]={
  category="sm",
  direction="on",
 },
 [0x2AF7]={
  category="sm",
  direction="on",
  mirror=0x2AF8,
 },
 [0x2AF8]={
  category="sm",
  direction="on",
  mirror=0x2AF7,
 },
 [0x2AF9]={
  category="sm",
  direction="on",
  mirror=0x2AFA,
 },
 [0x2AFA]={
  category="sm",
  direction="on",
  mirror=0x2AF9,
 },
 [0x2AFB]={
  category="sm",
  direction="on",
 },
 [0x2AFC]={
  category="sm",
  direction="on",
 },
 [0x2AFD]={
  category="sm",
  direction="on",
 },
 [0x2AFE]={
  category="sm",
  direction="on",
 },
 [0x2AFF]={
  category="sm",
  direction="on",
 },
 [0x2B00]={
  category="so",
  direction="on",
 },
 [0x2B01]={
  category="so",
  direction="on",
 },
 [0x2B02]={
  category="so",
  direction="on",
 },
 [0x2B03]={
  category="so",
  direction="on",
 },
 [0x2B04]={
  category="so",
  direction="on",
 },
 [0x2B05]={
  category="so",
  direction="on",
 },
 [0x2B06]={
  category="so",
  direction="on",
 },
 [0x2B07]={
  category="so",
  direction="on",
 },
 [0x2B08]={
  category="so",
  direction="on",
 },
 [0x2B09]={
  category="so",
  direction="on",
 },
 [0x2B0A]={
  category="so",
  direction="on",
 },
 [0x2B0B]={
  category="so",
  direction="on",
 },
 [0x2B0C]={
  category="so",
  direction="on",
 },
 [0x2B0D]={
  category="so",
  direction="on",
 },
 [0x2B0E]={
  category="so",
  direction="on",
 },
 [0x2B0F]={
  category="so",
  direction="on",
 },
 [0x2B10]={
  category="so",
  direction="on",
 },
 [0x2B11]={
  category="so",
  direction="on",
 },
 [0x2B12]={
  category="so",
  direction="on",
 },
 [0x2B13]={
  category="so",
  direction="on",
 },
 [0x2B14]={
  category="so",
  direction="on",
 },
 [0x2B15]={
  category="so",
  direction="on",
 },
 [0x2B16]={
  category="so",
  direction="on",
 },
 [0x2B17]={
  category="so",
  direction="on",
 },
 [0x2B18]={
  category="so",
  direction="on",
 },
 [0x2B19]={
  category="so",
  direction="on",
 },
 [0x2B1A]={
  category="so",
  direction="on",
 },
 [0x2B1B]={
  category="so",
  direction="on",
 },
 [0x2B1C]={
  category="so",
  direction="on",
 },
 [0x2B1D]={
  category="so",
  direction="on",
 },
 [0x2B1E]={
  category="so",
  direction="on",
 },
 [0x2B1F]={
  category="so",
  direction="on",
 },
 [0x2B20]={
  category="so",
  direction="on",
 },
 [0x2B21]={
  category="so",
  direction="on",
 },
 [0x2B22]={
  category="so",
  direction="on",
 },
 [0x2B23]={
  category="so",
  direction="on",
 },
 [0x2B24]={
  category="so",
  direction="on",
 },
 [0x2B25]={
  category="so",
  direction="on",
 },
 [0x2B26]={
  category="so",
  direction="on",
 },
 [0x2B27]={
  category="so",
  direction="on",
 },
 [0x2B28]={
  category="so",
  direction="on",
 },
 [0x2B29]={
  category="so",
  direction="on",
 },
 [0x2B2A]={
  category="so",
  direction="on",
 },
 [0x2B2B]={
  category="so",
  direction="on",
 },
 [0x2B2C]={
  category="so",
  direction="on",
 },
 [0x2B2D]={
  category="so",
  direction="on",
 },
 [0x2B2E]={
  category="so",
  direction="on",
 },
 [0x2B2F]={
  category="so",
  direction="on",
 },
 [0x2B30]={
  category="sm",
  direction="on",
 },
 [0x2B31]={
  category="sm",
  direction="on",
 },
 [0x2B32]={
  category="sm",
  direction="on",
 },
 [0x2B33]={
  category="sm",
  direction="on",
 },
 [0x2B34]={
  category="sm",
  direction="on",
 },
 [0x2B35]={
  category="sm",
  direction="on",
 },
 [0x2B36]={
  category="sm",
  direction="on",
 },
 [0x2B37]={
  category="sm",
  direction="on",
 },
 [0x2B38]={
  category="sm",
  direction="on",
 },
 [0x2B39]={
  category="sm",
  direction="on",
 },
 [0x2B3A]={
  category="sm",
  direction="on",
 },
 [0x2B3B]={
  category="sm",
  direction="on",
 },
 [0x2B3C]={
  category="sm",
  direction="on",
 },
 [0x2B3D]={
  category="sm",
  direction="on",
 },
 [0x2B3E]={
  category="sm",
  direction="on",
 },
 [0x2B3F]={
  category="sm",
  direction="on",
 },
 [0x2B40]={
  category="sm",
  direction="on",
 },
 [0x2B41]={
  category="sm",
  direction="on",
 },
 [0x2B42]={
  category="sm",
  direction="on",
 },
 [0x2B43]={
  category="sm",
  direction="on",
 },
 [0x2B44]={
  category="sm",
  direction="on",
 },
 [0x2B45]={
  category="so",
  direction="on",
 },
 [0x2B46]={
  category="so",
  direction="on",
 },
 [0x2B47]={
  category="sm",
  direction="on",
 },
 [0x2B48]={
  category="sm",
  direction="on",
 },
 [0x2B49]={
  category="sm",
  direction="on",
 },
 [0x2B4A]={
  category="sm",
  direction="on",
 },
 [0x2B4B]={
  category="sm",
  direction="on",
 },
 [0x2B4C]={
  category="sm",
  direction="on",
 },
 [0x2B4D]={
  category="so",
  direction="on",
 },
 [0x2B4E]={
  category="so",
  direction="on",
 },
 [0x2B4F]={
  category="so",
  direction="on",
 },
 [0x2B50]={
  category="so",
  direction="on",
 },
 [0x2B51]={
  category="so",
  direction="on",
 },
 [0x2B52]={
  category="so",
  direction="on",
 },
 [0x2B53]={
  category="so",
  direction="on",
 },
 [0x2B54]={
  category="so",
  direction="on",
 },
 [0x2B55]={
  category="so",
  direction="on",
 },
 [0x2B56]={
  category="so",
  direction="on",
 },
 [0x2B57]={
  category="so",
  direction="on",
 },
 [0x2B58]={
  category="so",
  direction="on",
 },
 [0x2B59]={
  category="so",
  direction="on",
 },
 [0x2B5A]={
  category="so",
  direction="on",
 },
 [0x2B5B]={
  category="so",
  direction="on",
 },
 [0x2B5C]={
  category="so",
  direction="on",
 },
 [0x2B5D]={
  category="so",
  direction="on",
 },
 [0x2B5E]={
  category="so",
  direction="on",
 },
 [0x2B5F]={
  category="so",
  direction="on",
 },
 [0x2B60]={
  category="so",
  direction="on",
 },
 [0x2B61]={
  category="so",
  direction="on",
 },
 [0x2B62]={
  category="so",
  direction="on",
 },
 [0x2B63]={
  category="so",
  direction="on",
 },
 [0x2B64]={
  category="so",
  direction="on",
 },
 [0x2B65]={
  category="so",
  direction="on",
 },
 [0x2B66]={
  category="so",
  direction="on",
 },
 [0x2B67]={
  category="so",
  direction="on",
 },
 [0x2B68]={
  category="so",
  direction="on",
 },
 [0x2B69]={
  category="so",
  direction="on",
 },
 [0x2B6A]={
  category="so",
  direction="on",
 },
 [0x2B6B]={
  category="so",
  direction="on",
 },
 [0x2B6C]={
  category="so",
  direction="on",
 },
 [0x2B6D]={
  category="so",
  direction="on",
 },
 [0x2B6E]={
  category="so",
  direction="on",
 },
 [0x2B6F]={
  category="so",
  direction="on",
 },
 [0x2B70]={
  category="so",
  direction="on",
 },
 [0x2B71]={
  category="so",
  direction="on",
 },
 [0x2B72]={
  category="so",
  direction="on",
 },
 [0x2B73]={
  category="so",
  direction="on",
 },
 [0x2B76]={
  category="so",
  direction="on",
 },
 [0x2B77]={
  category="so",
  direction="on",
 },
 [0x2B78]={
  category="so",
  direction="on",
 },
 [0x2B79]={
  category="so",
  direction="on",
 },
 [0x2B7A]={
  category="so",
  direction="on",
 },
 [0x2B7B]={
  category="so",
  direction="on",
 },
 [0x2B7C]={
  category="so",
  direction="on",
 },
 [0x2B7D]={
  category="so",
  direction="on",
 },
 [0x2B7E]={
  category="so",
  direction="on",
 },
 [0x2B7F]={
  category="so",
  direction="on",
 },
 [0x2B80]={
  category="so",
  direction="on",
 },
 [0x2B81]={
  category="so",
  direction="on",
 },
 [0x2B82]={
  category="so",
  direction="on",
 },
 [0x2B83]={
  category="so",
  direction="on",
 },
 [0x2B84]={
  category="so",
  direction="on",
 },
 [0x2B85]={
  category="so",
  direction="on",
 },
 [0x2B86]={
  category="so",
  direction="on",
 },
 [0x2B87]={
  category="so",
  direction="on",
 },
 [0x2B88]={
  category="so",
  direction="on",
 },
 [0x2B89]={
  category="so",
  direction="on",
 },
 [0x2B8A]={
  category="so",
  direction="on",
 },
 [0x2B8B]={
  category="so",
  direction="on",
 },
 [0x2B8C]={
  category="so",
  direction="on",
 },
 [0x2B8D]={
  category="so",
  direction="on",
 },
 [0x2B8E]={
  category="so",
  direction="on",
 },
 [0x2B8F]={
  category="so",
  direction="on",
 },
 [0x2B90]={
  category="so",
  direction="on",
 },
 [0x2B91]={
  category="so",
  direction="on",
 },
 [0x2B92]={
  category="so",
  direction="on",
 },
 [0x2B93]={
  category="so",
  direction="on",
 },
 [0x2B94]={
  category="so",
  direction="on",
 },
 [0x2B95]={
  category="so",
  direction="on",
 },
 [0x2B98]={
  category="so",
  direction="on",
 },
 [0x2B99]={
  category="so",
  direction="on",
 },
 [0x2B9A]={
  category="so",
  direction="on",
 },
 [0x2B9B]={
  category="so",
  direction="on",
 },
 [0x2B9C]={
  category="so",
  direction="on",
 },
 [0x2B9D]={
  category="so",
  direction="on",
 },
 [0x2B9E]={
  category="so",
  direction="on",
 },
 [0x2B9F]={
  category="so",
  direction="on",
 },
 [0x2BA0]={
  category="so",
  direction="on",
 },
 [0x2BA1]={
  category="so",
  direction="on",
 },
 [0x2BA2]={
  category="so",
  direction="on",
 },
 [0x2BA3]={
  category="so",
  direction="on",
 },
 [0x2BA4]={
  category="so",
  direction="on",
 },
 [0x2BA5]={
  category="so",
  direction="on",
 },
 [0x2BA6]={
  category="so",
  direction="on",
 },
 [0x2BA7]={
  category="so",
  direction="on",
 },
 [0x2BA8]={
  category="so",
  direction="on",
 },
 [0x2BA9]={
  category="so",
  direction="on",
 },
 [0x2BAA]={
  category="so",
  direction="on",
 },
 [0x2BAB]={
  category="so",
  direction="on",
 },
 [0x2BAC]={
  category="so",
  direction="on",
 },
 [0x2BAD]={
  category="so",
  direction="on",
 },
 [0x2BAE]={
  category="so",
  direction="on",
 },
 [0x2BAF]={
  category="so",
  direction="on",
 },
 [0x2BB0]={
  category="so",
  direction="on",
 },
 [0x2BB1]={
  category="so",
  direction="on",
 },
 [0x2BB2]={
  category="so",
  direction="on",
 },
 [0x2BB3]={
  category="so",
  direction="on",
 },
 [0x2BB4]={
  category="so",
  direction="on",
 },
 [0x2BB5]={
  category="so",
  direction="on",
 },
 [0x2BB6]={
  category="so",
  direction="on",
 },
 [0x2BB7]={
  category="so",
  direction="on",
 },
 [0x2BB8]={
  category="so",
  direction="on",
 },
 [0x2BB9]={
  category="so",
  direction="on",
 },
 [0x2BBD]={
  category="so",
  direction="on",
 },
 [0x2BBE]={
  category="so",
  direction="on",
 },
 [0x2BBF]={
  category="so",
  direction="on",
 },
 [0x2BC0]={
  category="so",
  direction="on",
 },
 [0x2BC1]={
  category="so",
  direction="on",
 },
 [0x2BC2]={
  category="so",
  direction="on",
 },
 [0x2BC3]={
  category="so",
  direction="on",
 },
 [0x2BC4]={
  category="so",
  direction="on",
 },
 [0x2BC5]={
  category="so",
  direction="on",
 },
 [0x2BC6]={
  category="so",
  direction="on",
 },
 [0x2BC7]={
  category="so",
  direction="on",
 },
 [0x2BC8]={
  category="so",
  direction="on",
 },
 [0x2BCA]={
  category="so",
  direction="on",
 },
 [0x2BCB]={
  category="so",
  direction="on",
 },
 [0x2BCC]={
  category="so",
  direction="on",
 },
 [0x2BCD]={
  category="so",
  direction="on",
 },
 [0x2BCE]={
  category="so",
  direction="on",
 },
 [0x2BCF]={
  category="so",
  direction="on",
 },
 [0x2BD0]={
  category="so",
  direction="on",
 },
 [0x2BD1]={
  category="so",
  direction="on",
 },
 [0x2C00]={
  category="lu",
  direction="l",
 },
 [0x2C01]={
  category="lu",
  direction="l",
 },
 [0x2C02]={
  category="lu",
  direction="l",
 },
 [0x2C03]={
  category="lu",
  direction="l",
 },
 [0x2C04]={
  category="lu",
  direction="l",
 },
 [0x2C05]={
  category="lu",
  direction="l",
 },
 [0x2C06]={
  category="lu",
  direction="l",
 },
 [0x2C07]={
  category="lu",
  direction="l",
 },
 [0x2C08]={
  category="lu",
  direction="l",
 },
 [0x2C09]={
  category="lu",
  direction="l",
 },
 [0x2C0A]={
  category="lu",
  direction="l",
 },
 [0x2C0B]={
  category="lu",
  direction="l",
 },
 [0x2C0C]={
  category="lu",
  direction="l",
 },
 [0x2C0D]={
  category="lu",
  direction="l",
 },
 [0x2C0E]={
  category="lu",
  direction="l",
 },
 [0x2C0F]={
  category="lu",
  direction="l",
 },
 [0x2C10]={
  category="lu",
  direction="l",
 },
 [0x2C11]={
  category="lu",
  direction="l",
 },
 [0x2C12]={
  category="lu",
  direction="l",
 },
 [0x2C13]={
  category="lu",
  direction="l",
 },
 [0x2C14]={
  category="lu",
  direction="l",
 },
 [0x2C15]={
  category="lu",
  direction="l",
 },
 [0x2C16]={
  category="lu",
  direction="l",
 },
 [0x2C17]={
  category="lu",
  direction="l",
 },
 [0x2C18]={
  category="lu",
  direction="l",
 },
 [0x2C19]={
  category="lu",
  direction="l",
 },
 [0x2C1A]={
  category="lu",
  direction="l",
 },
 [0x2C1B]={
  category="lu",
  direction="l",
 },
 [0x2C1C]={
  category="lu",
  direction="l",
 },
 [0x2C1D]={
  category="lu",
  direction="l",
 },
 [0x2C1E]={
  category="lu",
  direction="l",
 },
 [0x2C1F]={
  category="lu",
  direction="l",
 },
 [0x2C20]={
  category="lu",
  direction="l",
 },
 [0x2C21]={
  category="lu",
  direction="l",
 },
 [0x2C22]={
  category="lu",
  direction="l",
 },
 [0x2C23]={
  category="lu",
  direction="l",
 },
 [0x2C24]={
  category="lu",
  direction="l",
 },
 [0x2C25]={
  category="lu",
  direction="l",
 },
 [0x2C26]={
  category="lu",
  direction="l",
 },
 [0x2C27]={
  category="lu",
  direction="l",
 },
 [0x2C28]={
  category="lu",
  direction="l",
 },
 [0x2C29]={
  category="lu",
  direction="l",
 },
 [0x2C2A]={
  category="lu",
  direction="l",
 },
 [0x2C2B]={
  category="lu",
  direction="l",
 },
 [0x2C2C]={
  category="lu",
  direction="l",
 },
 [0x2C2D]={
  category="lu",
  direction="l",
 },
 [0x2C2E]={
  category="lu",
  direction="l",
 },
 [0x2C30]={
  category="ll",
  direction="l",
 },
 [0x2C31]={
  category="ll",
  direction="l",
 },
 [0x2C32]={
  category="ll",
  direction="l",
 },
 [0x2C33]={
  category="ll",
  direction="l",
 },
 [0x2C34]={
  category="ll",
  direction="l",
 },
 [0x2C35]={
  category="ll",
  direction="l",
 },
 [0x2C36]={
  category="ll",
  direction="l",
 },
 [0x2C37]={
  category="ll",
  direction="l",
 },
 [0x2C38]={
  category="ll",
  direction="l",
 },
 [0x2C39]={
  category="ll",
  direction="l",
 },
 [0x2C3A]={
  category="ll",
  direction="l",
 },
 [0x2C3B]={
  category="ll",
  direction="l",
 },
 [0x2C3C]={
  category="ll",
  direction="l",
 },
 [0x2C3D]={
  category="ll",
  direction="l",
 },
 [0x2C3E]={
  category="ll",
  direction="l",
 },
 [0x2C3F]={
  category="ll",
  direction="l",
 },
 [0x2C40]={
  category="ll",
  direction="l",
 },
 [0x2C41]={
  category="ll",
  direction="l",
 },
 [0x2C42]={
  category="ll",
  direction="l",
 },
 [0x2C43]={
  category="ll",
  direction="l",
 },
 [0x2C44]={
  category="ll",
  direction="l",
 },
 [0x2C45]={
  category="ll",
  direction="l",
 },
 [0x2C46]={
  category="ll",
  direction="l",
 },
 [0x2C47]={
  category="ll",
  direction="l",
 },
 [0x2C48]={
  category="ll",
  direction="l",
 },
 [0x2C49]={
  category="ll",
  direction="l",
 },
 [0x2C4A]={
  category="ll",
  direction="l",
 },
 [0x2C4B]={
  category="ll",
  direction="l",
 },
 [0x2C4C]={
  category="ll",
  direction="l",
 },
 [0x2C4D]={
  category="ll",
  direction="l",
 },
 [0x2C4E]={
  category="ll",
  direction="l",
 },
 [0x2C4F]={
  category="ll",
  direction="l",
 },
 [0x2C50]={
  category="ll",
  direction="l",
 },
 [0x2C51]={
  category="ll",
  direction="l",
 },
 [0x2C52]={
  category="ll",
  direction="l",
 },
 [0x2C53]={
  category="ll",
  direction="l",
 },
 [0x2C54]={
  category="ll",
  direction="l",
 },
 [0x2C55]={
  category="ll",
  direction="l",
 },
 [0x2C56]={
  category="ll",
  direction="l",
 },
 [0x2C57]={
  category="ll",
  direction="l",
 },
 [0x2C58]={
  category="ll",
  direction="l",
 },
 [0x2C59]={
  category="ll",
  direction="l",
 },
 [0x2C5A]={
  category="ll",
  direction="l",
 },
 [0x2C5B]={
  category="ll",
  direction="l",
 },
 [0x2C5C]={
  category="ll",
  direction="l",
 },
 [0x2C5D]={
  category="ll",
  direction="l",
 },
 [0x2C5E]={
  category="ll",
  direction="l",
 },
 [0x2C60]={
  category="lu",
  direction="l",
 },
 [0x2C61]={
  category="ll",
  direction="l",
 },
 [0x2C62]={
  category="lu",
  direction="l",
 },
 [0x2C63]={
  category="lu",
  direction="l",
 },
 [0x2C64]={
  category="lu",
  direction="l",
 },
 [0x2C65]={
  category="ll",
  direction="l",
 },
 [0x2C66]={
  category="ll",
  direction="l",
 },
 [0x2C67]={
  category="lu",
  direction="l",
 },
 [0x2C68]={
  category="ll",
  direction="l",
 },
 [0x2C69]={
  category="lu",
  direction="l",
 },
 [0x2C6A]={
  category="ll",
  direction="l",
 },
 [0x2C6B]={
  category="lu",
  direction="l",
 },
 [0x2C6C]={
  category="ll",
  direction="l",
 },
 [0x2C6D]={
  category="lu",
  direction="l",
 },
 [0x2C6E]={
  category="lu",
  direction="l",
 },
 [0x2C6F]={
  category="lu",
  direction="l",
 },
 [0x2C70]={
  category="lu",
  direction="l",
 },
 [0x2C71]={
  category="ll",
  direction="l",
 },
 [0x2C72]={
  category="lu",
  direction="l",
 },
 [0x2C73]={
  category="ll",
  direction="l",
 },
 [0x2C74]={
  category="ll",
  direction="l",
 },
 [0x2C75]={
  category="lu",
  direction="l",
 },
 [0x2C76]={
  category="ll",
  direction="l",
 },
 [0x2C77]={
  category="ll",
  direction="l",
 },
 [0x2C78]={
  category="ll",
  direction="l",
 },
 [0x2C79]={
  category="ll",
  direction="l",
 },
 [0x2C7A]={
  category="ll",
  direction="l",
 },
 [0x2C7B]={
  category="ll",
  direction="l",
 },
 [0x2C7C]={
  category="ll",
  direction="l",
 },
 [0x2C7D]={
  category="lm",
  direction="l",
 },
 [0x2C7E]={
  category="lu",
  direction="l",
 },
 [0x2C7F]={
  category="lu",
  direction="l",
 },
 [0x2C80]={
  category="lu",
  direction="l",
 },
 [0x2C81]={
  category="ll",
  direction="l",
 },
 [0x2C82]={
  category="lu",
  direction="l",
 },
 [0x2C83]={
  category="ll",
  direction="l",
 },
 [0x2C84]={
  category="lu",
  direction="l",
 },
 [0x2C85]={
  category="ll",
  direction="l",
 },
 [0x2C86]={
  category="lu",
  direction="l",
 },
 [0x2C87]={
  category="ll",
  direction="l",
 },
 [0x2C88]={
  category="lu",
  direction="l",
 },
 [0x2C89]={
  category="ll",
  direction="l",
 },
 [0x2C8A]={
  category="lu",
  direction="l",
 },
 [0x2C8B]={
  category="ll",
  direction="l",
 },
 [0x2C8C]={
  category="lu",
  direction="l",
 },
 [0x2C8D]={
  category="ll",
  direction="l",
 },
 [0x2C8E]={
  category="lu",
  direction="l",
 },
 [0x2C8F]={
  category="ll",
  direction="l",
 },
 [0x2C90]={
  category="lu",
  direction="l",
 },
 [0x2C91]={
  category="ll",
  direction="l",
 },
 [0x2C92]={
  category="lu",
  direction="l",
 },
 [0x2C93]={
  category="ll",
  direction="l",
 },
 [0x2C94]={
  category="lu",
  direction="l",
 },
 [0x2C95]={
  category="ll",
  direction="l",
 },
 [0x2C96]={
  category="lu",
  direction="l",
 },
 [0x2C97]={
  category="ll",
  direction="l",
 },
 [0x2C98]={
  category="lu",
  direction="l",
 },
 [0x2C99]={
  category="ll",
  direction="l",
 },
 [0x2C9A]={
  category="lu",
  direction="l",
 },
 [0x2C9B]={
  category="ll",
  direction="l",
 },
 [0x2C9C]={
  category="lu",
  direction="l",
 },
 [0x2C9D]={
  category="ll",
  direction="l",
 },
 [0x2C9E]={
  category="lu",
  direction="l",
 },
 [0x2C9F]={
  category="ll",
  direction="l",
 },
 [0x2CA0]={
  category="lu",
  direction="l",
 },
 [0x2CA1]={
  category="ll",
  direction="l",
 },
 [0x2CA2]={
  category="lu",
  direction="l",
 },
 [0x2CA3]={
  category="ll",
  direction="l",
 },
 [0x2CA4]={
  category="lu",
  direction="l",
 },
 [0x2CA5]={
  category="ll",
  direction="l",
 },
 [0x2CA6]={
  category="lu",
  direction="l",
 },
 [0x2CA7]={
  category="ll",
  direction="l",
 },
 [0x2CA8]={
  category="lu",
  direction="l",
 },
 [0x2CA9]={
  category="ll",
  direction="l",
 },
 [0x2CAA]={
  category="lu",
  direction="l",
 },
 [0x2CAB]={
  category="ll",
  direction="l",
 },
 [0x2CAC]={
  category="lu",
  direction="l",
 },
 [0x2CAD]={
  category="ll",
  direction="l",
 },
 [0x2CAE]={
  category="lu",
  direction="l",
 },
 [0x2CAF]={
  category="ll",
  direction="l",
 },
 [0x2CB0]={
  category="lu",
  direction="l",
 },
 [0x2CB1]={
  category="ll",
  direction="l",
 },
 [0x2CB2]={
  category="lu",
  direction="l",
 },
 [0x2CB3]={
  category="ll",
  direction="l",
 },
 [0x2CB4]={
  category="lu",
  direction="l",
 },
 [0x2CB5]={
  category="ll",
  direction="l",
 },
 [0x2CB6]={
  category="lu",
  direction="l",
 },
 [0x2CB7]={
  category="ll",
  direction="l",
 },
 [0x2CB8]={
  category="lu",
  direction="l",
 },
 [0x2CB9]={
  category="ll",
  direction="l",
 },
 [0x2CBA]={
  category="lu",
  direction="l",
 },
 [0x2CBB]={
  category="ll",
  direction="l",
 },
 [0x2CBC]={
  category="lu",
  direction="l",
 },
 [0x2CBD]={
  category="ll",
  direction="l",
 },
 [0x2CBE]={
  category="lu",
  direction="l",
 },
 [0x2CBF]={
  category="ll",
  direction="l",
 },
 [0x2CC0]={
  category="lu",
  direction="l",
 },
 [0x2CC1]={
  category="ll",
  direction="l",
 },
 [0x2CC2]={
  category="lu",
  direction="l",
 },
 [0x2CC3]={
  category="ll",
  direction="l",
 },
 [0x2CC4]={
  category="lu",
  direction="l",
 },
 [0x2CC5]={
  category="ll",
  direction="l",
 },
 [0x2CC6]={
  category="lu",
  direction="l",
 },
 [0x2CC7]={
  category="ll",
  direction="l",
 },
 [0x2CC8]={
  category="lu",
  direction="l",
 },
 [0x2CC9]={
  category="ll",
  direction="l",
 },
 [0x2CCA]={
  category="lu",
  direction="l",
 },
 [0x2CCB]={
  category="ll",
  direction="l",
 },
 [0x2CCC]={
  category="lu",
  direction="l",
 },
 [0x2CCD]={
  category="ll",
  direction="l",
 },
 [0x2CCE]={
  category="lu",
  direction="l",
 },
 [0x2CCF]={
  category="ll",
  direction="l",
 },
 [0x2CD0]={
  category="lu",
  direction="l",
 },
 [0x2CD1]={
  category="ll",
  direction="l",
 },
 [0x2CD2]={
  category="lu",
  direction="l",
 },
 [0x2CD3]={
  category="ll",
  direction="l",
 },
 [0x2CD4]={
  category="lu",
  direction="l",
 },
 [0x2CD5]={
  category="ll",
  direction="l",
 },
 [0x2CD6]={
  category="lu",
  direction="l",
 },
 [0x2CD7]={
  category="ll",
  direction="l",
 },
 [0x2CD8]={
  category="lu",
  direction="l",
 },
 [0x2CD9]={
  category="ll",
  direction="l",
 },
 [0x2CDA]={
  category="lu",
  direction="l",
 },
 [0x2CDB]={
  category="ll",
  direction="l",
 },
 [0x2CDC]={
  category="lu",
  direction="l",
 },
 [0x2CDD]={
  category="ll",
  direction="l",
 },
 [0x2CDE]={
  category="lu",
  direction="l",
 },
 [0x2CDF]={
  category="ll",
  direction="l",
 },
 [0x2CE0]={
  category="lu",
  direction="l",
 },
 [0x2CE1]={
  category="ll",
  direction="l",
 },
 [0x2CE2]={
  category="lu",
  direction="l",
 },
 [0x2CE3]={
  category="ll",
  direction="l",
 },
 [0x2CE4]={
  category="ll",
  direction="l",
 },
 [0x2CE5]={
  category="so",
  direction="on",
 },
 [0x2CE6]={
  category="so",
  direction="on",
 },
 [0x2CE7]={
  category="so",
  direction="on",
 },
 [0x2CE8]={
  category="so",
  direction="on",
 },
 [0x2CE9]={
  category="so",
  direction="on",
 },
 [0x2CEA]={
  category="so",
  direction="on",
 },
 [0x2CEB]={
  category="lu",
  direction="l",
 },
 [0x2CEC]={
  category="ll",
  direction="l",
 },
 [0x2CED]={
  category="lu",
  direction="l",
 },
 [0x2CEE]={
  category="ll",
  direction="l",
 },
 [0x2CEF]={
  category="mn",
  direction="nsm",
 },
 [0x2CF0]={
  category="mn",
  direction="nsm",
 },
 [0x2CF1]={
  category="mn",
  direction="nsm",
 },
 [0x2CF2]={
  category="lu",
  direction="l",
 },
 [0x2CF3]={
  category="ll",
  direction="l",
 },
 [0x2CF9]={
  category="po",
  direction="on",
 },
 [0x2CFA]={
  category="po",
  direction="on",
 },
 [0x2CFB]={
  category="po",
  direction="on",
 },
 [0x2CFC]={
  category="po",
  direction="on",
 },
 [0x2CFD]={
  category="no",
  direction="on",
 },
 [0x2CFE]={
  category="po",
  direction="on",
 },
 [0x2CFF]={
  category="po",
  direction="on",
 },
 [0x2D00]={
  category="ll",
  direction="l",
 },
 [0x2D01]={
  category="ll",
  direction="l",
 },
 [0x2D02]={
  category="ll",
  direction="l",
 },
 [0x2D03]={
  category="ll",
  direction="l",
 },
 [0x2D04]={
  category="ll",
  direction="l",
 },
 [0x2D05]={
  category="ll",
  direction="l",
 },
 [0x2D06]={
  category="ll",
  direction="l",
 },
 [0x2D07]={
  category="ll",
  direction="l",
 },
 [0x2D08]={
  category="ll",
  direction="l",
 },
 [0x2D09]={
  category="ll",
  direction="l",
 },
 [0x2D0A]={
  category="ll",
  direction="l",
 },
 [0x2D0B]={
  category="ll",
  direction="l",
 },
 [0x2D0C]={
  category="ll",
  direction="l",
 },
 [0x2D0D]={
  category="ll",
  direction="l",
 },
 [0x2D0E]={
  category="ll",
  direction="l",
 },
 [0x2D0F]={
  category="ll",
  direction="l",
 },
 [0x2D10]={
  category="ll",
  direction="l",
 },
 [0x2D11]={
  category="ll",
  direction="l",
 },
 [0x2D12]={
  category="ll",
  direction="l",
 },
 [0x2D13]={
  category="ll",
  direction="l",
 },
 [0x2D14]={
  category="ll",
  direction="l",
 },
 [0x2D15]={
  category="ll",
  direction="l",
 },
 [0x2D16]={
  category="ll",
  direction="l",
 },
 [0x2D17]={
  category="ll",
  direction="l",
 },
 [0x2D18]={
  category="ll",
  direction="l",
 },
 [0x2D19]={
  category="ll",
  direction="l",
 },
 [0x2D1A]={
  category="ll",
  direction="l",
 },
 [0x2D1B]={
  category="ll",
  direction="l",
 },
 [0x2D1C]={
  category="ll",
  direction="l",
 },
 [0x2D1D]={
  category="ll",
  direction="l",
 },
 [0x2D1E]={
  category="ll",
  direction="l",
 },
 [0x2D1F]={
  category="ll",
  direction="l",
 },
 [0x2D20]={
  category="ll",
  direction="l",
 },
 [0x2D21]={
  category="ll",
  direction="l",
 },
 [0x2D22]={
  category="ll",
  direction="l",
 },
 [0x2D23]={
  category="ll",
  direction="l",
 },
 [0x2D24]={
  category="ll",
  direction="l",
 },
 [0x2D25]={
  category="ll",
  direction="l",
 },
 [0x2D27]={
  category="ll",
  direction="l",
 },
 [0x2D2D]={
  category="ll",
  direction="l",
 },
 [0x2D30]={
  category="lo",
  direction="l",
 },
 [0x2D31]={
  category="lo",
  direction="l",
 },
 [0x2D32]={
  category="lo",
  direction="l",
 },
 [0x2D33]={
  category="lo",
  direction="l",
 },
 [0x2D34]={
  category="lo",
  direction="l",
 },
 [0x2D35]={
  category="lo",
  direction="l",
 },
 [0x2D36]={
  category="lo",
  direction="l",
 },
 [0x2D37]={
  category="lo",
  direction="l",
 },
 [0x2D38]={
  category="lo",
  direction="l",
 },
 [0x2D39]={
  category="lo",
  direction="l",
 },
 [0x2D3A]={
  category="lo",
  direction="l",
 },
 [0x2D3B]={
  category="lo",
  direction="l",
 },
 [0x2D3C]={
  category="lo",
  direction="l",
 },
 [0x2D3D]={
  category="lo",
  direction="l",
 },
 [0x2D3E]={
  category="lo",
  direction="l",
 },
 [0x2D3F]={
  category="lo",
  direction="l",
 },
 [0x2D40]={
  category="lo",
  direction="l",
 },
 [0x2D41]={
  category="lo",
  direction="l",
 },
 [0x2D42]={
  category="lo",
  direction="l",
 },
 [0x2D43]={
  category="lo",
  direction="l",
 },
 [0x2D44]={
  category="lo",
  direction="l",
 },
 [0x2D45]={
  category="lo",
  direction="l",
 },
 [0x2D46]={
  category="lo",
  direction="l",
 },
 [0x2D47]={
  category="lo",
  direction="l",
 },
 [0x2D48]={
  category="lo",
  direction="l",
 },
 [0x2D49]={
  category="lo",
  direction="l",
 },
 [0x2D4A]={
  category="lo",
  direction="l",
 },
 [0x2D4B]={
  category="lo",
  direction="l",
 },
 [0x2D4C]={
  category="lo",
  direction="l",
 },
 [0x2D4D]={
  category="lo",
  direction="l",
 },
 [0x2D4E]={
  category="lo",
  direction="l",
 },
 [0x2D4F]={
  category="lo",
  direction="l",
 },
 [0x2D50]={
  category="lo",
  direction="l",
 },
 [0x2D51]={
  category="lo",
  direction="l",
 },
 [0x2D52]={
  category="lo",
  direction="l",
 },
 [0x2D53]={
  category="lo",
  direction="l",
 },
 [0x2D54]={
  category="lo",
  direction="l",
 },
 [0x2D55]={
  category="lo",
  direction="l",
 },
 [0x2D56]={
  category="lo",
  direction="l",
 },
 [0x2D57]={
  category="lo",
  direction="l",
 },
 [0x2D58]={
  category="lo",
  direction="l",
 },
 [0x2D59]={
  category="lo",
  direction="l",
 },
 [0x2D5A]={
  category="lo",
  direction="l",
 },
 [0x2D5B]={
  category="lo",
  direction="l",
 },
 [0x2D5C]={
  category="lo",
  direction="l",
 },
 [0x2D5D]={
  category="lo",
  direction="l",
 },
 [0x2D5E]={
  category="lo",
  direction="l",
 },
 [0x2D5F]={
  category="lo",
  direction="l",
 },
 [0x2D60]={
  category="lo",
  direction="l",
 },
 [0x2D61]={
  category="lo",
  direction="l",
 },
 [0x2D62]={
  category="lo",
  direction="l",
 },
 [0x2D63]={
  category="lo",
  direction="l",
 },
 [0x2D64]={
  category="lo",
  direction="l",
 },
 [0x2D65]={
  category="lo",
  direction="l",
 },
 [0x2D66]={
  category="lo",
  direction="l",
 },
 [0x2D67]={
  category="lo",
  direction="l",
 },
 [0x2D6F]={
  category="lm",
  direction="l",
 },
 [0x2D70]={
  category="po",
  direction="l",
 },
 [0x2D7F]={
  category="mn",
  direction="nsm",
 },
 [0x2D80]={
  category="lo",
  direction="l",
 },
 [0x2D81]={
  category="lo",
  direction="l",
 },
 [0x2D82]={
  category="lo",
  direction="l",
 },
 [0x2D83]={
  category="lo",
  direction="l",
 },
 [0x2D84]={
  category="lo",
  direction="l",
 },
 [0x2D85]={
  category="lo",
  direction="l",
 },
 [0x2D86]={
  category="lo",
  direction="l",
 },
 [0x2D87]={
  category="lo",
  direction="l",
 },
 [0x2D88]={
  category="lo",
  direction="l",
 },
 [0x2D89]={
  category="lo",
  direction="l",
 },
 [0x2D8A]={
  category="lo",
  direction="l",
 },
 [0x2D8B]={
  category="lo",
  direction="l",
 },
 [0x2D8C]={
  category="lo",
  direction="l",
 },
 [0x2D8D]={
  category="lo",
  direction="l",
 },
 [0x2D8E]={
  category="lo",
  direction="l",
 },
 [0x2D8F]={
  category="lo",
  direction="l",
 },
 [0x2D90]={
  category="lo",
  direction="l",
 },
 [0x2D91]={
  category="lo",
  direction="l",
 },
 [0x2D92]={
  category="lo",
  direction="l",
 },
 [0x2D93]={
  category="lo",
  direction="l",
 },
 [0x2D94]={
  category="lo",
  direction="l",
 },
 [0x2D95]={
  category="lo",
  direction="l",
 },
 [0x2D96]={
  category="lo",
  direction="l",
 },
 [0x2DA0]={
  category="lo",
  direction="l",
 },
 [0x2DA1]={
  category="lo",
  direction="l",
 },
 [0x2DA2]={
  category="lo",
  direction="l",
 },
 [0x2DA3]={
  category="lo",
  direction="l",
 },
 [0x2DA4]={
  category="lo",
  direction="l",
 },
 [0x2DA5]={
  category="lo",
  direction="l",
 },
 [0x2DA6]={
  category="lo",
  direction="l",
 },
 [0x2DA8]={
  category="lo",
  direction="l",
 },
 [0x2DA9]={
  category="lo",
  direction="l",
 },
 [0x2DAA]={
  category="lo",
  direction="l",
 },
 [0x2DAB]={
  category="lo",
  direction="l",
 },
 [0x2DAC]={
  category="lo",
  direction="l",
 },
 [0x2DAD]={
  category="lo",
  direction="l",
 },
 [0x2DAE]={
  category="lo",
  direction="l",
 },
 [0x2DB0]={
  category="lo",
  direction="l",
 },
 [0x2DB1]={
  category="lo",
  direction="l",
 },
 [0x2DB2]={
  category="lo",
  direction="l",
 },
 [0x2DB3]={
  category="lo",
  direction="l",
 },
 [0x2DB4]={
  category="lo",
  direction="l",
 },
 [0x2DB5]={
  category="lo",
  direction="l",
 },
 [0x2DB6]={
  category="lo",
  direction="l",
 },
 [0x2DB8]={
  category="lo",
  direction="l",
 },
 [0x2DB9]={
  category="lo",
  direction="l",
 },
 [0x2DBA]={
  category="lo",
  direction="l",
 },
 [0x2DBB]={
  category="lo",
  direction="l",
 },
 [0x2DBC]={
  category="lo",
  direction="l",
 },
 [0x2DBD]={
  category="lo",
  direction="l",
 },
 [0x2DBE]={
  category="lo",
  direction="l",
 },
 [0x2DC0]={
  category="lo",
  direction="l",
 },
 [0x2DC1]={
  category="lo",
  direction="l",
 },
 [0x2DC2]={
  category="lo",
  direction="l",
 },
 [0x2DC3]={
  category="lo",
  direction="l",
 },
 [0x2DC4]={
  category="lo",
  direction="l",
 },
 [0x2DC5]={
  category="lo",
  direction="l",
 },
 [0x2DC6]={
  category="lo",
  direction="l",
 },
 [0x2DC8]={
  category="lo",
  direction="l",
 },
 [0x2DC9]={
  category="lo",
  direction="l",
 },
 [0x2DCA]={
  category="lo",
  direction="l",
 },
 [0x2DCB]={
  category="lo",
  direction="l",
 },
 [0x2DCC]={
  category="lo",
  direction="l",
 },
 [0x2DCD]={
  category="lo",
  direction="l",
 },
 [0x2DCE]={
  category="lo",
  direction="l",
 },
 [0x2DD0]={
  category="lo",
  direction="l",
 },
 [0x2DD1]={
  category="lo",
  direction="l",
 },
 [0x2DD2]={
  category="lo",
  direction="l",
 },
 [0x2DD3]={
  category="lo",
  direction="l",
 },
 [0x2DD4]={
  category="lo",
  direction="l",
 },
 [0x2DD5]={
  category="lo",
  direction="l",
 },
 [0x2DD6]={
  category="lo",
  direction="l",
 },
 [0x2DD8]={
  category="lo",
  direction="l",
 },
 [0x2DD9]={
  category="lo",
  direction="l",
 },
 [0x2DDA]={
  category="lo",
  direction="l",
 },
 [0x2DDB]={
  category="lo",
  direction="l",
 },
 [0x2DDC]={
  category="lo",
  direction="l",
 },
 [0x2DDD]={
  category="lo",
  direction="l",
 },
 [0x2DDE]={
  category="lo",
  direction="l",
 },
 [0x2DE0]={
  category="mn",
  direction="nsm",
 },
 [0x2DE1]={
  category="mn",
  direction="nsm",
 },
 [0x2DE2]={
  category="mn",
  direction="nsm",
 },
 [0x2DE3]={
  category="mn",
  direction="nsm",
 },
 [0x2DE4]={
  category="mn",
  direction="nsm",
 },
 [0x2DE5]={
  category="mn",
  direction="nsm",
 },
 [0x2DE6]={
  category="mn",
  direction="nsm",
 },
 [0x2DE7]={
  category="mn",
  direction="nsm",
 },
 [0x2DE8]={
  category="mn",
  direction="nsm",
 },
 [0x2DE9]={
  category="mn",
  direction="nsm",
 },
 [0x2DEA]={
  category="mn",
  direction="nsm",
 },
 [0x2DEB]={
  category="mn",
  direction="nsm",
 },
 [0x2DEC]={
  category="mn",
  direction="nsm",
 },
 [0x2DED]={
  category="mn",
  direction="nsm",
 },
 [0x2DEE]={
  category="mn",
  direction="nsm",
 },
 [0x2DEF]={
  category="mn",
  direction="nsm",
 },
 [0x2DF0]={
  category="mn",
  direction="nsm",
 },
 [0x2DF1]={
  category="mn",
  direction="nsm",
 },
 [0x2DF2]={
  category="mn",
  direction="nsm",
 },
 [0x2DF3]={
  category="mn",
  direction="nsm",
 },
 [0x2DF4]={
  category="mn",
  direction="nsm",
 },
 [0x2DF5]={
  category="mn",
  direction="nsm",
 },
 [0x2DF6]={
  category="mn",
  direction="nsm",
 },
 [0x2DF7]={
  category="mn",
  direction="nsm",
 },
 [0x2DF8]={
  category="mn",
  direction="nsm",
 },
 [0x2DF9]={
  category="mn",
  direction="nsm",
 },
 [0x2DFA]={
  category="mn",
  direction="nsm",
 },
 [0x2DFB]={
  category="mn",
  direction="nsm",
 },
 [0x2DFC]={
  category="mn",
  direction="nsm",
 },
 [0x2DFD]={
  category="mn",
  direction="nsm",
 },
 [0x2DFE]={
  category="mn",
  direction="nsm",
 },
 [0x2DFF]={
  category="mn",
  direction="nsm",
 },
 [0x2E00]={
  category="po",
  direction="on",
 },
 [0x2E01]={
  category="po",
  direction="on",
 },
 [0x2E02]={
  category="pi",
  direction="on",
  mirror=0x2E03,
 },
 [0x2E03]={
  category="pf",
  direction="on",
  mirror=0x2E02,
 },
 [0x2E04]={
  category="pi",
  direction="on",
  mirror=0x2E05,
 },
 [0x2E05]={
  category="pf",
  direction="on",
  mirror=0x2E04,
 },
 [0x2E06]={
  category="po",
  direction="on",
 },
 [0x2E07]={
  category="po",
  direction="on",
 },
 [0x2E08]={
  category="po",
  direction="on",
 },
 [0x2E09]={
  category="pi",
  direction="on",
  mirror=0x2E0A,
 },
 [0x2E0A]={
  category="pf",
  direction="on",
  mirror=0x2E09,
 },
 [0x2E0B]={
  category="po",
  direction="on",
 },
 [0x2E0C]={
  category="pi",
  direction="on",
  mirror=0x2E0D,
 },
 [0x2E0D]={
  category="pf",
  direction="on",
  mirror=0x2E0C,
 },
 [0x2E0E]={
  category="po",
  direction="on",
 },
 [0x2E0F]={
  category="po",
  direction="on",
 },
 [0x2E10]={
  category="po",
  direction="on",
 },
 [0x2E11]={
  category="po",
  direction="on",
 },
 [0x2E12]={
  category="po",
  direction="on",
 },
 [0x2E13]={
  category="po",
  direction="on",
 },
 [0x2E14]={
  category="po",
  direction="on",
 },
 [0x2E15]={
  category="po",
  direction="on",
 },
 [0x2E16]={
  category="po",
  direction="on",
 },
 [0x2E17]={
  category="pd",
  direction="on",
 },
 [0x2E18]={
  category="po",
  direction="on",
 },
 [0x2E19]={
  category="po",
  direction="on",
 },
 [0x2E1A]={
  category="pd",
  direction="on",
 },
 [0x2E1B]={
  category="po",
  direction="on",
 },
 [0x2E1C]={
  category="pi",
  direction="on",
  mirror=0x2E1D,
 },
 [0x2E1D]={
  category="pf",
  direction="on",
  mirror=0x2E1C,
 },
 [0x2E1E]={
  category="po",
  direction="on",
 },
 [0x2E1F]={
  category="po",
  direction="on",
 },
 [0x2E20]={
  category="pi",
  direction="on",
  mirror=0x2E21,
 },
 [0x2E21]={
  category="pf",
  direction="on",
  mirror=0x2E20,
 },
 [0x2E22]={
  category="ps",
  direction="on",
  mirror=0x2E23,
 },
 [0x2E23]={
  category="pe",
  direction="on",
  mirror=0x2E22,
 },
 [0x2E24]={
  category="ps",
  direction="on",
  mirror=0x2E25,
 },
 [0x2E25]={
  category="pe",
  direction="on",
  mirror=0x2E24,
 },
 [0x2E26]={
  category="ps",
  direction="on",
  mirror=0x2E27,
 },
 [0x2E27]={
  category="pe",
  direction="on",
  mirror=0x2E26,
 },
 [0x2E28]={
  category="ps",
  direction="on",
  mirror=0x2E29,
 },
 [0x2E29]={
  category="pe",
  direction="on",
  mirror=0x2E28,
 },
 [0x2E2A]={
  category="po",
  direction="on",
 },
 [0x2E2B]={
  category="po",
  direction="on",
 },
 [0x2E2C]={
  category="po",
  direction="on",
 },
 [0x2E2D]={
  category="po",
  direction="on",
 },
 [0x2E2E]={
  category="po",
  direction="on",
 },
 [0x2E2F]={
  category="lm",
  direction="on",
 },
 [0x2E30]={
  category="po",
  direction="on",
 },
 [0x2E31]={
  category="po",
  direction="on",
 },
 [0x2E32]={
  category="po",
  direction="on",
 },
 [0x2E33]={
  category="po",
  direction="on",
 },
 [0x2E34]={
  category="po",
  direction="on",
 },
 [0x2E35]={
  category="po",
  direction="on",
 },
 [0x2E36]={
  category="po",
  direction="on",
 },
 [0x2E37]={
  category="po",
  direction="on",
 },
 [0x2E38]={
  category="po",
  direction="on",
 },
 [0x2E39]={
  category="po",
  direction="on",
 },
 [0x2E3A]={
  category="pd",
  direction="on",
 },
 [0x2E3B]={
  category="pd",
  direction="on",
 },
 [0x2E3C]={
  category="po",
  direction="on",
 },
 [0x2E3D]={
  category="po",
  direction="on",
 },
 [0x2E3E]={
  category="po",
  direction="on",
 },
 [0x2E3F]={
  category="po",
  direction="on",
 },
 [0x2E40]={
  category="pd",
  direction="on",
 },
 [0x2E41]={
  category="po",
  direction="on",
 },
 [0x2E42]={
  category="ps",
  direction="on",
 },
 [0x2E80]={
  category="so",
  direction="on",
 },
 [0x2E81]={
  category="so",
  direction="on",
 },
 [0x2E82]={
  category="so",
  direction="on",
 },
 [0x2E83]={
  category="so",
  direction="on",
 },
 [0x2E84]={
  category="so",
  direction="on",
 },
 [0x2E85]={
  category="so",
  direction="on",
 },
 [0x2E86]={
  category="so",
  direction="on",
 },
 [0x2E87]={
  category="so",
  direction="on",
 },
 [0x2E88]={
  category="so",
  direction="on",
 },
 [0x2E89]={
  category="so",
  direction="on",
 },
 [0x2E8A]={
  category="so",
  direction="on",
 },
 [0x2E8B]={
  category="so",
  direction="on",
 },
 [0x2E8C]={
  category="so",
  direction="on",
 },
 [0x2E8D]={
  category="so",
  direction="on",
 },
 [0x2E8E]={
  category="so",
  direction="on",
 },
 [0x2E8F]={
  category="so",
  direction="on",
 },
 [0x2E90]={
  category="so",
  direction="on",
 },
 [0x2E91]={
  category="so",
  direction="on",
 },
 [0x2E92]={
  category="so",
  direction="on",
 },
 [0x2E93]={
  category="so",
  direction="on",
 },
 [0x2E94]={
  category="so",
  direction="on",
 },
 [0x2E95]={
  category="so",
  direction="on",
 },
 [0x2E96]={
  category="so",
  direction="on",
 },
 [0x2E97]={
  category="so",
  direction="on",
 },
 [0x2E98]={
  category="so",
  direction="on",
 },
 [0x2E99]={
  category="so",
  direction="on",
 },
 [0x2E9B]={
  category="so",
  direction="on",
 },
 [0x2E9C]={
  category="so",
  direction="on",
 },
 [0x2E9D]={
  category="so",
  direction="on",
 },
 [0x2E9E]={
  category="so",
  direction="on",
 },
 [0x2E9F]={
  category="so",
  direction="on",
 },
 [0x2EA0]={
  category="so",
  direction="on",
 },
 [0x2EA1]={
  category="so",
  direction="on",
 },
 [0x2EA2]={
  category="so",
  direction="on",
 },
 [0x2EA3]={
  category="so",
  direction="on",
 },
 [0x2EA4]={
  category="so",
  direction="on",
 },
 [0x2EA5]={
  category="so",
  direction="on",
 },
 [0x2EA6]={
  category="so",
  direction="on",
 },
 [0x2EA7]={
  category="so",
  direction="on",
 },
 [0x2EA8]={
  category="so",
  direction="on",
 },
 [0x2EA9]={
  category="so",
  direction="on",
 },
 [0x2EAA]={
  category="so",
  direction="on",
 },
 [0x2EAB]={
  category="so",
  direction="on",
 },
 [0x2EAC]={
  category="so",
  direction="on",
 },
 [0x2EAD]={
  category="so",
  direction="on",
 },
 [0x2EAE]={
  category="so",
  direction="on",
 },
 [0x2EAF]={
  category="so",
  direction="on",
 },
 [0x2EB0]={
  category="so",
  direction="on",
 },
 [0x2EB1]={
  category="so",
  direction="on",
 },
 [0x2EB2]={
  category="so",
  direction="on",
 },
 [0x2EB3]={
  category="so",
  direction="on",
 },
 [0x2EB4]={
  category="so",
  direction="on",
 },
 [0x2EB5]={
  category="so",
  direction="on",
 },
 [0x2EB6]={
  category="so",
  direction="on",
 },
 [0x2EB7]={
  category="so",
  direction="on",
 },
 [0x2EB8]={
  category="so",
  direction="on",
 },
 [0x2EB9]={
  category="so",
  direction="on",
 },
 [0x2EBA]={
  category="so",
  direction="on",
 },
 [0x2EBB]={
  category="so",
  direction="on",
 },
 [0x2EBC]={
  category="so",
  direction="on",
 },
 [0x2EBD]={
  category="so",
  direction="on",
 },
 [0x2EBE]={
  category="so",
  direction="on",
 },
 [0x2EBF]={
  category="so",
  direction="on",
 },
 [0x2EC0]={
  category="so",
  direction="on",
 },
 [0x2EC1]={
  category="so",
  direction="on",
 },
 [0x2EC2]={
  category="so",
  direction="on",
 },
 [0x2EC3]={
  category="so",
  direction="on",
 },
 [0x2EC4]={
  category="so",
  direction="on",
 },
 [0x2EC5]={
  category="so",
  direction="on",
 },
 [0x2EC6]={
  category="so",
  direction="on",
 },
 [0x2EC7]={
  category="so",
  direction="on",
 },
 [0x2EC8]={
  category="so",
  direction="on",
 },
 [0x2EC9]={
  category="so",
  direction="on",
 },
 [0x2ECA]={
  category="so",
  direction="on",
 },
 [0x2ECB]={
  category="so",
  direction="on",
 },
 [0x2ECC]={
  category="so",
  direction="on",
 },
 [0x2ECD]={
  category="so",
  direction="on",
 },
 [0x2ECE]={
  category="so",
  direction="on",
 },
 [0x2ECF]={
  category="so",
  direction="on",
 },
 [0x2ED0]={
  category="so",
  direction="on",
 },
 [0x2ED1]={
  category="so",
  direction="on",
 },
 [0x2ED2]={
  category="so",
  direction="on",
 },
 [0x2ED3]={
  category="so",
  direction="on",
 },
 [0x2ED4]={
  category="so",
  direction="on",
 },
 [0x2ED5]={
  category="so",
  direction="on",
 },
 [0x2ED6]={
  category="so",
  direction="on",
 },
 [0x2ED7]={
  category="so",
  direction="on",
 },
 [0x2ED8]={
  category="so",
  direction="on",
 },
 [0x2ED9]={
  category="so",
  direction="on",
 },
 [0x2EDA]={
  category="so",
  direction="on",
 },
 [0x2EDB]={
  category="so",
  direction="on",
 },
 [0x2EDC]={
  category="so",
  direction="on",
 },
 [0x2EDD]={
  category="so",
  direction="on",
 },
 [0x2EDE]={
  category="so",
  direction="on",
 },
 [0x2EDF]={
  category="so",
  direction="on",
 },
 [0x2EE0]={
  category="so",
  direction="on",
 },
 [0x2EE1]={
  category="so",
  direction="on",
 },
 [0x2EE2]={
  category="so",
  direction="on",
 },
 [0x2EE3]={
  category="so",
  direction="on",
 },
 [0x2EE4]={
  category="so",
  direction="on",
 },
 [0x2EE5]={
  category="so",
  direction="on",
 },
 [0x2EE6]={
  category="so",
  direction="on",
 },
 [0x2EE7]={
  category="so",
  direction="on",
 },
 [0x2EE8]={
  category="so",
  direction="on",
 },
 [0x2EE9]={
  category="so",
  direction="on",
 },
 [0x2EEA]={
  category="so",
  direction="on",
 },
 [0x2EEB]={
  category="so",
  direction="on",
 },
 [0x2EEC]={
  category="so",
  direction="on",
 },
 [0x2EED]={
  category="so",
  direction="on",
 },
 [0x2EEE]={
  category="so",
  direction="on",
 },
 [0x2EEF]={
  category="so",
  direction="on",
 },
 [0x2EF0]={
  category="so",
  direction="on",
 },
 [0x2EF1]={
  category="so",
  direction="on",
 },
 [0x2EF2]={
  category="so",
  direction="on",
 },
 [0x2EF3]={
  category="so",
  direction="on",
 },
 [0x2F00]={
  category="so",
  direction="on",
 },
 [0x2F01]={
  category="so",
  direction="on",
 },
 [0x2F02]={
  category="so",
  direction="on",
 },
 [0x2F03]={
  category="so",
  direction="on",
 },
 [0x2F04]={
  category="so",
  direction="on",
 },
 [0x2F05]={
  category="so",
  direction="on",
 },
 [0x2F06]={
  category="so",
  direction="on",
 },
 [0x2F07]={
  category="so",
  direction="on",
 },
 [0x2F08]={
  category="so",
  direction="on",
 },
 [0x2F09]={
  category="so",
  direction="on",
 },
 [0x2F0A]={
  category="so",
  direction="on",
 },
 [0x2F0B]={
  category="so",
  direction="on",
 },
 [0x2F0C]={
  category="so",
  direction="on",
 },
 [0x2F0D]={
  category="so",
  direction="on",
 },
 [0x2F0E]={
  category="so",
  direction="on",
 },
 [0x2F0F]={
  category="so",
  direction="on",
 },
 [0x2F10]={
  category="so",
  direction="on",
 },
 [0x2F11]={
  category="so",
  direction="on",
 },
 [0x2F12]={
  category="so",
  direction="on",
 },
 [0x2F13]={
  category="so",
  direction="on",
 },
 [0x2F14]={
  category="so",
  direction="on",
 },
 [0x2F15]={
  category="so",
  direction="on",
 },
 [0x2F16]={
  category="so",
  direction="on",
 },
 [0x2F17]={
  category="so",
  direction="on",
 },
 [0x2F18]={
  category="so",
  direction="on",
 },
 [0x2F19]={
  category="so",
  direction="on",
 },
 [0x2F1A]={
  category="so",
  direction="on",
 },
 [0x2F1B]={
  category="so",
  direction="on",
 },
 [0x2F1C]={
  category="so",
  direction="on",
 },
 [0x2F1D]={
  category="so",
  direction="on",
 },
 [0x2F1E]={
  category="so",
  direction="on",
 },
 [0x2F1F]={
  category="so",
  direction="on",
 },
 [0x2F20]={
  category="so",
  direction="on",
 },
 [0x2F21]={
  category="so",
  direction="on",
 },
 [0x2F22]={
  category="so",
  direction="on",
 },
 [0x2F23]={
  category="so",
  direction="on",
 },
 [0x2F24]={
  category="so",
  direction="on",
 },
 [0x2F25]={
  category="so",
  direction="on",
 },
 [0x2F26]={
  category="so",
  direction="on",
 },
 [0x2F27]={
  category="so",
  direction="on",
 },
 [0x2F28]={
  category="so",
  direction="on",
 },
 [0x2F29]={
  category="so",
  direction="on",
 },
 [0x2F2A]={
  category="so",
  direction="on",
 },
 [0x2F2B]={
  category="so",
  direction="on",
 },
 [0x2F2C]={
  category="so",
  direction="on",
 },
 [0x2F2D]={
  category="so",
  direction="on",
 },
 [0x2F2E]={
  category="so",
  direction="on",
 },
 [0x2F2F]={
  category="so",
  direction="on",
 },
 [0x2F30]={
  category="so",
  direction="on",
 },
 [0x2F31]={
  category="so",
  direction="on",
 },
 [0x2F32]={
  category="so",
  direction="on",
 },
 [0x2F33]={
  category="so",
  direction="on",
 },
 [0x2F34]={
  category="so",
  direction="on",
 },
 [0x2F35]={
  category="so",
  direction="on",
 },
 [0x2F36]={
  category="so",
  direction="on",
 },
 [0x2F37]={
  category="so",
  direction="on",
 },
 [0x2F38]={
  category="so",
  direction="on",
 },
 [0x2F39]={
  category="so",
  direction="on",
 },
 [0x2F3A]={
  category="so",
  direction="on",
 },
 [0x2F3B]={
  category="so",
  direction="on",
 },
 [0x2F3C]={
  category="so",
  direction="on",
 },
 [0x2F3D]={
  category="so",
  direction="on",
 },
 [0x2F3E]={
  category="so",
  direction="on",
 },
 [0x2F3F]={
  category="so",
  direction="on",
 },
 [0x2F40]={
  category="so",
  direction="on",
 },
 [0x2F41]={
  category="so",
  direction="on",
 },
 [0x2F42]={
  category="so",
  direction="on",
 },
 [0x2F43]={
  category="so",
  direction="on",
 },
 [0x2F44]={
  category="so",
  direction="on",
 },
 [0x2F45]={
  category="so",
  direction="on",
 },
 [0x2F46]={
  category="so",
  direction="on",
 },
 [0x2F47]={
  category="so",
  direction="on",
 },
 [0x2F48]={
  category="so",
  direction="on",
 },
 [0x2F49]={
  category="so",
  direction="on",
 },
 [0x2F4A]={
  category="so",
  direction="on",
 },
 [0x2F4B]={
  category="so",
  direction="on",
 },
 [0x2F4C]={
  category="so",
  direction="on",
 },
 [0x2F4D]={
  category="so",
  direction="on",
 },
 [0x2F4E]={
  category="so",
  direction="on",
 },
 [0x2F4F]={
  category="so",
  direction="on",
 },
 [0x2F50]={
  category="so",
  direction="on",
 },
 [0x2F51]={
  category="so",
  direction="on",
 },
 [0x2F52]={
  category="so",
  direction="on",
 },
 [0x2F53]={
  category="so",
  direction="on",
 },
 [0x2F54]={
  category="so",
  direction="on",
 },
 [0x2F55]={
  category="so",
  direction="on",
 },
 [0x2F56]={
  category="so",
  direction="on",
 },
 [0x2F57]={
  category="so",
  direction="on",
 },
 [0x2F58]={
  category="so",
  direction="on",
 },
 [0x2F59]={
  category="so",
  direction="on",
 },
 [0x2F5A]={
  category="so",
  direction="on",
 },
 [0x2F5B]={
  category="so",
  direction="on",
 },
 [0x2F5C]={
  category="so",
  direction="on",
 },
 [0x2F5D]={
  category="so",
  direction="on",
 },
 [0x2F5E]={
  category="so",
  direction="on",
 },
 [0x2F5F]={
  category="so",
  direction="on",
 },
 [0x2F60]={
  category="so",
  direction="on",
 },
 [0x2F61]={
  category="so",
  direction="on",
 },
 [0x2F62]={
  category="so",
  direction="on",
 },
 [0x2F63]={
  category="so",
  direction="on",
 },
 [0x2F64]={
  category="so",
  direction="on",
 },
 [0x2F65]={
  category="so",
  direction="on",
 },
 [0x2F66]={
  category="so",
  direction="on",
 },
 [0x2F67]={
  category="so",
  direction="on",
 },
 [0x2F68]={
  category="so",
  direction="on",
 },
 [0x2F69]={
  category="so",
  direction="on",
 },
 [0x2F6A]={
  category="so",
  direction="on",
 },
 [0x2F6B]={
  category="so",
  direction="on",
 },
 [0x2F6C]={
  category="so",
  direction="on",
 },
 [0x2F6D]={
  category="so",
  direction="on",
 },
 [0x2F6E]={
  category="so",
  direction="on",
 },
 [0x2F6F]={
  category="so",
  direction="on",
 },
 [0x2F70]={
  category="so",
  direction="on",
 },
 [0x2F71]={
  category="so",
  direction="on",
 },
 [0x2F72]={
  category="so",
  direction="on",
 },
 [0x2F73]={
  category="so",
  direction="on",
 },
 [0x2F74]={
  category="so",
  direction="on",
 },
 [0x2F75]={
  category="so",
  direction="on",
 },
 [0x2F76]={
  category="so",
  direction="on",
 },
 [0x2F77]={
  category="so",
  direction="on",
 },
 [0x2F78]={
  category="so",
  direction="on",
 },
 [0x2F79]={
  category="so",
  direction="on",
 },
 [0x2F7A]={
  category="so",
  direction="on",
 },
 [0x2F7B]={
  category="so",
  direction="on",
 },
 [0x2F7C]={
  category="so",
  direction="on",
 },
 [0x2F7D]={
  category="so",
  direction="on",
 },
 [0x2F7E]={
  category="so",
  direction="on",
 },
 [0x2F7F]={
  category="so",
  direction="on",
 },
 [0x2F80]={
  category="so",
  direction="on",
 },
 [0x2F81]={
  category="so",
  direction="on",
 },
 [0x2F82]={
  category="so",
  direction="on",
 },
 [0x2F83]={
  category="so",
  direction="on",
 },
 [0x2F84]={
  category="so",
  direction="on",
 },
 [0x2F85]={
  category="so",
  direction="on",
 },
 [0x2F86]={
  category="so",
  direction="on",
 },
 [0x2F87]={
  category="so",
  direction="on",
 },
 [0x2F88]={
  category="so",
  direction="on",
 },
 [0x2F89]={
  category="so",
  direction="on",
 },
 [0x2F8A]={
  category="so",
  direction="on",
 },
 [0x2F8B]={
  category="so",
  direction="on",
 },
 [0x2F8C]={
  category="so",
  direction="on",
 },
 [0x2F8D]={
  category="so",
  direction="on",
 },
 [0x2F8E]={
  category="so",
  direction="on",
 },
 [0x2F8F]={
  category="so",
  direction="on",
 },
 [0x2F90]={
  category="so",
  direction="on",
 },
 [0x2F91]={
  category="so",
  direction="on",
 },
 [0x2F92]={
  category="so",
  direction="on",
 },
 [0x2F93]={
  category="so",
  direction="on",
 },
 [0x2F94]={
  category="so",
  direction="on",
 },
 [0x2F95]={
  category="so",
  direction="on",
 },
 [0x2F96]={
  category="so",
  direction="on",
 },
 [0x2F97]={
  category="so",
  direction="on",
 },
 [0x2F98]={
  category="so",
  direction="on",
 },
 [0x2F99]={
  category="so",
  direction="on",
 },
 [0x2F9A]={
  category="so",
  direction="on",
 },
 [0x2F9B]={
  category="so",
  direction="on",
 },
 [0x2F9C]={
  category="so",
  direction="on",
 },
 [0x2F9D]={
  category="so",
  direction="on",
 },
 [0x2F9E]={
  category="so",
  direction="on",
 },
 [0x2F9F]={
  category="so",
  direction="on",
 },
 [0x2FA0]={
  category="so",
  direction="on",
 },
 [0x2FA1]={
  category="so",
  direction="on",
 },
 [0x2FA2]={
  category="so",
  direction="on",
 },
 [0x2FA3]={
  category="so",
  direction="on",
 },
 [0x2FA4]={
  category="so",
  direction="on",
 },
 [0x2FA5]={
  category="so",
  direction="on",
 },
 [0x2FA6]={
  category="so",
  direction="on",
 },
 [0x2FA7]={
  category="so",
  direction="on",
 },
 [0x2FA8]={
  category="so",
  direction="on",
 },
 [0x2FA9]={
  category="so",
  direction="on",
 },
 [0x2FAA]={
  category="so",
  direction="on",
 },
 [0x2FAB]={
  category="so",
  direction="on",
 },
 [0x2FAC]={
  category="so",
  direction="on",
 },
 [0x2FAD]={
  category="so",
  direction="on",
 },
 [0x2FAE]={
  category="so",
  direction="on",
 },
 [0x2FAF]={
  category="so",
  direction="on",
 },
 [0x2FB0]={
  category="so",
  direction="on",
 },
 [0x2FB1]={
  category="so",
  direction="on",
 },
 [0x2FB2]={
  category="so",
  direction="on",
 },
 [0x2FB3]={
  category="so",
  direction="on",
 },
 [0x2FB4]={
  category="so",
  direction="on",
 },
 [0x2FB5]={
  category="so",
  direction="on",
 },
 [0x2FB6]={
  category="so",
  direction="on",
 },
 [0x2FB7]={
  category="so",
  direction="on",
 },
 [0x2FB8]={
  category="so",
  direction="on",
 },
 [0x2FB9]={
  category="so",
  direction="on",
 },
 [0x2FBA]={
  category="so",
  direction="on",
 },
 [0x2FBB]={
  category="so",
  direction="on",
 },
 [0x2FBC]={
  category="so",
  direction="on",
 },
 [0x2FBD]={
  category="so",
  direction="on",
 },
 [0x2FBE]={
  category="so",
  direction="on",
 },
 [0x2FBF]={
  category="so",
  direction="on",
 },
 [0x2FC0]={
  category="so",
  direction="on",
 },
 [0x2FC1]={
  category="so",
  direction="on",
 },
 [0x2FC2]={
  category="so",
  direction="on",
 },
 [0x2FC3]={
  category="so",
  direction="on",
 },
 [0x2FC4]={
  category="so",
  direction="on",
 },
 [0x2FC5]={
  category="so",
  direction="on",
 },
 [0x2FC6]={
  category="so",
  direction="on",
 },
 [0x2FC7]={
  category="so",
  direction="on",
 },
 [0x2FC8]={
  category="so",
  direction="on",
 },
 [0x2FC9]={
  category="so",
  direction="on",
 },
 [0x2FCA]={
  category="so",
  direction="on",
 },
 [0x2FCB]={
  category="so",
  direction="on",
 },
 [0x2FCC]={
  category="so",
  direction="on",
 },
 [0x2FCD]={
  category="so",
  direction="on",
 },
 [0x2FCE]={
  category="so",
  direction="on",
 },
 [0x2FCF]={
  category="so",
  direction="on",
 },
 [0x2FD0]={
  category="so",
  direction="on",
 },
 [0x2FD1]={
  category="so",
  direction="on",
 },
 [0x2FD2]={
  category="so",
  direction="on",
 },
 [0x2FD3]={
  category="so",
  direction="on",
 },
 [0x2FD4]={
  category="so",
  direction="on",
 },
 [0x2FD5]={
  category="so",
  direction="on",
 },
 [0x2FF0]={
  category="so",
  direction="on",
 },
 [0x2FF1]={
  category="so",
  direction="on",
 },
 [0x2FF2]={
  category="so",
  direction="on",
 },
 [0x2FF3]={
  category="so",
  direction="on",
 },
 [0x2FF4]={
  category="so",
  direction="on",
 },
 [0x2FF5]={
  category="so",
  direction="on",
 },
 [0x2FF6]={
  category="so",
  direction="on",
 },
 [0x2FF7]={
  category="so",
  direction="on",
 },
 [0x2FF8]={
  category="so",
  direction="on",
 },
 [0x2FF9]={
  category="so",
  direction="on",
 },
 [0x2FFA]={
  category="so",
  direction="on",
 },
 [0x2FFB]={
  category="so",
  direction="on",
 },
 [0x3000]={
  category="zs",
  direction="ws",
 },
 [0x3001]={
  category="po",
  direction="on",
 },
 [0x3002]={
  category="po",
  direction="on",
 },
 [0x3003]={
  category="po",
  direction="on",
 },
 [0x3004]={
  category="so",
  direction="on",
 },
 [0x3005]={
  category="lm",
  direction="l",
 },
 [0x3006]={
  category="lo",
  direction="l",
 },
 [0x3007]={
  category="nl",
  direction="l",
 },
 [0x3008]={
  category="ps",
  direction="on",
  mirror=0x3009,
 },
 [0x3009]={
  category="pe",
  direction="on",
  mirror=0x3008,
 },
 [0x300A]={
  category="ps",
  direction="on",
  mirror=0x300B,
 },
 [0x300B]={
  category="pe",
  direction="on",
  mirror=0x300A,
 },
 [0x300C]={
  category="ps",
  direction="on",
  mirror=0x300D,
 },
 [0x300D]={
  category="pe",
  direction="on",
  mirror=0x300C,
 },
 [0x300E]={
  category="ps",
  direction="on",
  mirror=0x300F,
 },
 [0x300F]={
  category="pe",
  direction="on",
  mirror=0x300E,
 },
 [0x3010]={
  category="ps",
  direction="on",
  mirror=0x3011,
 },
 [0x3011]={
  category="pe",
  direction="on",
  mirror=0x3010,
 },
 [0x3012]={
  category="so",
  direction="on",
 },
 [0x3013]={
  category="so",
  direction="on",
 },
 [0x3014]={
  category="ps",
  direction="on",
  mirror=0x3015,
 },
 [0x3015]={
  category="pe",
  direction="on",
  mirror=0x3014,
 },
 [0x3016]={
  category="ps",
  direction="on",
  mirror=0x3017,
 },
 [0x3017]={
  category="pe",
  direction="on",
  mirror=0x3016,
 },
 [0x3018]={
  category="ps",
  direction="on",
  mirror=0x3019,
 },
 [0x3019]={
  category="pe",
  direction="on",
  mirror=0x3018,
 },
 [0x301A]={
  category="ps",
  direction="on",
  mirror=0x301B,
 },
 [0x301B]={
  category="pe",
  direction="on",
  mirror=0x301A,
 },
 [0x301C]={
  category="pd",
  direction="on",
 },
 [0x301D]={
  category="ps",
  direction="on",
 },
 [0x301E]={
  category="pe",
  direction="on",
 },
 [0x301F]={
  category="pe",
  direction="on",
 },
 [0x3020]={
  category="so",
  direction="on",
 },
 [0x3021]={
  category="nl",
  direction="l",
 },
 [0x3022]={
  category="nl",
  direction="l",
 },
 [0x3023]={
  category="nl",
  direction="l",
 },
 [0x3024]={
  category="nl",
  direction="l",
 },
 [0x3025]={
  category="nl",
  direction="l",
 },
 [0x3026]={
  category="nl",
  direction="l",
 },
 [0x3027]={
  category="nl",
  direction="l",
 },
 [0x3028]={
  category="nl",
  direction="l",
 },
 [0x3029]={
  category="nl",
  direction="l",
 },
 [0x302A]={
  category="mn",
  direction="nsm",
 },
 [0x302B]={
  category="mn",
  direction="nsm",
 },
 [0x302C]={
  category="mn",
  direction="nsm",
 },
 [0x302D]={
  category="mn",
  direction="nsm",
 },
 [0x302E]={
  category="mn",
  direction="l",
 },
 [0x302F]={
  category="mn",
  direction="l",
 },
 [0x3030]={
  category="pd",
  direction="on",
 },
 [0x3031]={
  category="lm",
  direction="l",
 },
 [0x3032]={
  category="lm",
  direction="l",
 },
 [0x3033]={
  category="lm",
  direction="l",
 },
 [0x3034]={
  category="lm",
  direction="l",
 },
 [0x3035]={
  category="lm",
  direction="l",
 },
 [0x3036]={
  category="so",
  direction="on",
 },
 [0x3037]={
  category="so",
  direction="on",
 },
 [0x3038]={
  category="nl",
  direction="l",
 },
 [0x3039]={
  category="nl",
  direction="l",
 },
 [0x303A]={
  category="nl",
  direction="l",
 },
 [0x303B]={
  category="lm",
  direction="l",
 },
 [0x303C]={
  category="lo",
  direction="l",
 },
 [0x303D]={
  category="po",
  direction="on",
 },
 [0x303E]={
  category="so",
  direction="on",
 },
 [0x303F]={
  category="so",
  direction="on",
 },
 [0x3041]={
  category="lo",
  direction="l",
 },
 [0x3042]={
  category="lo",
  direction="l",
 },
 [0x3043]={
  category="lo",
  direction="l",
 },
 [0x3044]={
  category="lo",
  direction="l",
 },
 [0x3045]={
  category="lo",
  direction="l",
 },
 [0x3046]={
  category="lo",
  direction="l",
 },
 [0x3047]={
  category="lo",
  direction="l",
 },
 [0x3048]={
  category="lo",
  direction="l",
 },
 [0x3049]={
  category="lo",
  direction="l",
 },
 [0x304A]={
  category="lo",
  direction="l",
 },
 [0x304B]={
  category="lo",
  direction="l",
 },
 [0x304C]={
  category="lo",
  direction="l",
 },
 [0x304D]={
  category="lo",
  direction="l",
 },
 [0x304E]={
  category="lo",
  direction="l",
 },
 [0x304F]={
  category="lo",
  direction="l",
 },
 [0x3050]={
  category="lo",
  direction="l",
 },
 [0x3051]={
  category="lo",
  direction="l",
 },
 [0x3052]={
  category="lo",
  direction="l",
 },
 [0x3053]={
  category="lo",
  direction="l",
 },
 [0x3054]={
  category="lo",
  direction="l",
 },
 [0x3055]={
  category="lo",
  direction="l",
 },
 [0x3056]={
  category="lo",
  direction="l",
 },
 [0x3057]={
  category="lo",
  direction="l",
 },
 [0x3058]={
  category="lo",
  direction="l",
 },
 [0x3059]={
  category="lo",
  direction="l",
 },
 [0x305A]={
  category="lo",
  direction="l",
 },
 [0x305B]={
  category="lo",
  direction="l",
 },
 [0x305C]={
  category="lo",
  direction="l",
 },
 [0x305D]={
  category="lo",
  direction="l",
 },
 [0x305E]={
  category="lo",
  direction="l",
 },
 [0x305F]={
  category="lo",
  direction="l",
 },
 [0x3060]={
  category="lo",
  direction="l",
 },
 [0x3061]={
  category="lo",
  direction="l",
 },
 [0x3062]={
  category="lo",
  direction="l",
 },
 [0x3063]={
  category="lo",
  direction="l",
 },
 [0x3064]={
  category="lo",
  direction="l",
 },
 [0x3065]={
  category="lo",
  direction="l",
 },
 [0x3066]={
  category="lo",
  direction="l",
 },
 [0x3067]={
  category="lo",
  direction="l",
 },
 [0x3068]={
  category="lo",
  direction="l",
 },
 [0x3069]={
  category="lo",
  direction="l",
 },
 [0x306A]={
  category="lo",
  direction="l",
 },
 [0x306B]={
  category="lo",
  direction="l",
 },
 [0x306C]={
  category="lo",
  direction="l",
 },
 [0x306D]={
  category="lo",
  direction="l",
 },
 [0x306E]={
  category="lo",
  direction="l",
 },
 [0x306F]={
  category="lo",
  direction="l",
 },
 [0x3070]={
  category="lo",
  direction="l",
 },
 [0x3071]={
  category="lo",
  direction="l",
 },
 [0x3072]={
  category="lo",
  direction="l",
 },
 [0x3073]={
  category="lo",
  direction="l",
 },
 [0x3074]={
  category="lo",
  direction="l",
 },
 [0x3075]={
  category="lo",
  direction="l",
 },
 [0x3076]={
  category="lo",
  direction="l",
 },
 [0x3077]={
  category="lo",
  direction="l",
 },
 [0x3078]={
  category="lo",
  direction="l",
 },
 [0x3079]={
  category="lo",
  direction="l",
 },
 [0x307A]={
  category="lo",
  direction="l",
 },
 [0x307B]={
  category="lo",
  direction="l",
 },
 [0x307C]={
  category="lo",
  direction="l",
 },
 [0x307D]={
  category="lo",
  direction="l",
 },
 [0x307E]={
  category="lo",
  direction="l",
 },
 [0x307F]={
  category="lo",
  direction="l",
 },
 [0x3080]={
  category="lo",
  direction="l",
 },
 [0x3081]={
  category="lo",
  direction="l",
 },
 [0x3082]={
  category="lo",
  direction="l",
 },
 [0x3083]={
  category="lo",
  direction="l",
 },
 [0x3084]={
  category="lo",
  direction="l",
 },
 [0x3085]={
  category="lo",
  direction="l",
 },
 [0x3086]={
  category="lo",
  direction="l",
 },
 [0x3087]={
  category="lo",
  direction="l",
 },
 [0x3088]={
  category="lo",
  direction="l",
 },
 [0x3089]={
  category="lo",
  direction="l",
 },
 [0x308A]={
  category="lo",
  direction="l",
 },
 [0x308B]={
  category="lo",
  direction="l",
 },
 [0x308C]={
  category="lo",
  direction="l",
 },
 [0x308D]={
  category="lo",
  direction="l",
 },
 [0x308E]={
  category="lo",
  direction="l",
 },
 [0x308F]={
  category="lo",
  direction="l",
 },
 [0x3090]={
  category="lo",
  direction="l",
 },
 [0x3091]={
  category="lo",
  direction="l",
 },
 [0x3092]={
  category="lo",
  direction="l",
 },
 [0x3093]={
  category="lo",
  direction="l",
 },
 [0x3094]={
  category="lo",
  direction="l",
 },
 [0x3095]={
  category="lo",
  direction="l",
 },
 [0x3096]={
  category="lo",
  direction="l",
 },
 [0x3099]={
  category="mn",
  direction="nsm",
 },
 [0x309A]={
  category="mn",
  direction="nsm",
 },
 [0x309B]={
  category="sk",
  direction="on",
 },
 [0x309C]={
  category="sk",
  direction="on",
 },
 [0x309D]={
  category="lm",
  direction="l",
 },
 [0x309E]={
  category="lm",
  direction="l",
 },
 [0x309F]={
  category="lo",
  direction="l",
 },
 [0x30A0]={
  category="pd",
  direction="on",
 },
 [0x30A1]={
  category="lo",
  direction="l",
 },
 [0x30A2]={
  category="lo",
  direction="l",
 },
 [0x30A3]={
  category="lo",
  direction="l",
 },
 [0x30A4]={
  category="lo",
  direction="l",
 },
 [0x30A5]={
  category="lo",
  direction="l",
 },
 [0x30A6]={
  category="lo",
  direction="l",
 },
 [0x30A7]={
  category="lo",
  direction="l",
 },
 [0x30A8]={
  category="lo",
  direction="l",
 },
 [0x30A9]={
  category="lo",
  direction="l",
 },
 [0x30AA]={
  category="lo",
  direction="l",
 },
 [0x30AB]={
  category="lo",
  direction="l",
 },
 [0x30AC]={
  category="lo",
  direction="l",
 },
 [0x30AD]={
  category="lo",
  direction="l",
 },
 [0x30AE]={
  category="lo",
  direction="l",
 },
 [0x30AF]={
  category="lo",
  direction="l",
 },
 [0x30B0]={
  category="lo",
  direction="l",
 },
 [0x30B1]={
  category="lo",
  direction="l",
 },
 [0x30B2]={
  category="lo",
  direction="l",
 },
 [0x30B3]={
  category="lo",
  direction="l",
 },
 [0x30B4]={
  category="lo",
  direction="l",
 },
 [0x30B5]={
  category="lo",
  direction="l",
 },
 [0x30B6]={
  category="lo",
  direction="l",
 },
 [0x30B7]={
  category="lo",
  direction="l",
 },
 [0x30B8]={
  category="lo",
  direction="l",
 },
 [0x30B9]={
  category="lo",
  direction="l",
 },
 [0x30BA]={
  category="lo",
  direction="l",
 },
 [0x30BB]={
  category="lo",
  direction="l",
 },
 [0x30BC]={
  category="lo",
  direction="l",
 },
 [0x30BD]={
  category="lo",
  direction="l",
 },
 [0x30BE]={
  category="lo",
  direction="l",
 },
 [0x30BF]={
  category="lo",
  direction="l",
 },
 [0x30C0]={
  category="lo",
  direction="l",
 },
 [0x30C1]={
  category="lo",
  direction="l",
 },
 [0x30C2]={
  category="lo",
  direction="l",
 },
 [0x30C3]={
  category="lo",
  direction="l",
 },
 [0x30C4]={
  category="lo",
  direction="l",
 },
 [0x30C5]={
  category="lo",
  direction="l",
 },
 [0x30C6]={
  category="lo",
  direction="l",
 },
 [0x30C7]={
  category="lo",
  direction="l",
 },
 [0x30C8]={
  category="lo",
  direction="l",
 },
 [0x30C9]={
  category="lo",
  direction="l",
 },
 [0x30CA]={
  category="lo",
  direction="l",
 },
 [0x30CB]={
  category="lo",
  direction="l",
 },
 [0x30CC]={
  category="lo",
  direction="l",
 },
 [0x30CD]={
  category="lo",
  direction="l",
 },
 [0x30CE]={
  category="lo",
  direction="l",
 },
 [0x30CF]={
  category="lo",
  direction="l",
 },
 [0x30D0]={
  category="lo",
  direction="l",
 },
 [0x30D1]={
  category="lo",
  direction="l",
 },
 [0x30D2]={
  category="lo",
  direction="l",
 },
 [0x30D3]={
  category="lo",
  direction="l",
 },
 [0x30D4]={
  category="lo",
  direction="l",
 },
 [0x30D5]={
  category="lo",
  direction="l",
 },
 [0x30D6]={
  category="lo",
  direction="l",
 },
 [0x30D7]={
  category="lo",
  direction="l",
 },
 [0x30D8]={
  category="lo",
  direction="l",
 },
 [0x30D9]={
  category="lo",
  direction="l",
 },
 [0x30DA]={
  category="lo",
  direction="l",
 },
 [0x30DB]={
  category="lo",
  direction="l",
 },
 [0x30DC]={
  category="lo",
  direction="l",
 },
 [0x30DD]={
  category="lo",
  direction="l",
 },
 [0x30DE]={
  category="lo",
  direction="l",
 },
 [0x30DF]={
  category="lo",
  direction="l",
 },
 [0x30E0]={
  category="lo",
  direction="l",
 },
 [0x30E1]={
  category="lo",
  direction="l",
 },
 [0x30E2]={
  category="lo",
  direction="l",
 },
 [0x30E3]={
  category="lo",
  direction="l",
 },
 [0x30E4]={
  category="lo",
  direction="l",
 },
 [0x30E5]={
  category="lo",
  direction="l",
 },
 [0x30E6]={
  category="lo",
  direction="l",
 },
 [0x30E7]={
  category="lo",
  direction="l",
 },
 [0x30E8]={
  category="lo",
  direction="l",
 },
 [0x30E9]={
  category="lo",
  direction="l",
 },
 [0x30EA]={
  category="lo",
  direction="l",
 },
 [0x30EB]={
  category="lo",
  direction="l",
 },
 [0x30EC]={
  category="lo",
  direction="l",
 },
 [0x30ED]={
  category="lo",
  direction="l",
 },
 [0x30EE]={
  category="lo",
  direction="l",
 },
 [0x30EF]={
  category="lo",
  direction="l",
 },
 [0x30F0]={
  category="lo",
  direction="l",
 },
 [0x30F1]={
  category="lo",
  direction="l",
 },
 [0x30F2]={
  category="lo",
  direction="l",
 },
 [0x30F3]={
  category="lo",
  direction="l",
 },
 [0x30F4]={
  category="lo",
  direction="l",
 },
 [0x30F5]={
  category="lo",
  direction="l",
 },
 [0x30F6]={
  category="lo",
  direction="l",
 },
 [0x30F7]={
  category="lo",
  direction="l",
 },
 [0x30F8]={
  category="lo",
  direction="l",
 },
 [0x30F9]={
  category="lo",
  direction="l",
 },
 [0x30FA]={
  category="lo",
  direction="l",
 },
 [0x30FB]={
  category="po",
  direction="on",
 },
 [0x30FC]={
  category="lm",
  direction="l",
 },
 [0x30FD]={
  category="lm",
  direction="l",
 },
 [0x30FE]={
  category="lm",
  direction="l",
 },
 [0x30FF]={
  category="lo",
  direction="l",
 },
 [0x3105]={
  category="lo",
  direction="l",
 },
 [0x3106]={
  category="lo",
  direction="l",
 },
 [0x3107]={
  category="lo",
  direction="l",
 },
 [0x3108]={
  category="lo",
  direction="l",
 },
 [0x3109]={
  category="lo",
  direction="l",
 },
 [0x310A]={
  category="lo",
  direction="l",
 },
 [0x310B]={
  category="lo",
  direction="l",
 },
 [0x310C]={
  category="lo",
  direction="l",
 },
 [0x310D]={
  category="lo",
  direction="l",
 },
 [0x310E]={
  category="lo",
  direction="l",
 },
 [0x310F]={
  category="lo",
  direction="l",
 },
 [0x3110]={
  category="lo",
  direction="l",
 },
 [0x3111]={
  category="lo",
  direction="l",
 },
 [0x3112]={
  category="lo",
  direction="l",
 },
 [0x3113]={
  category="lo",
  direction="l",
 },
 [0x3114]={
  category="lo",
  direction="l",
 },
 [0x3115]={
  category="lo",
  direction="l",
 },
 [0x3116]={
  category="lo",
  direction="l",
 },
 [0x3117]={
  category="lo",
  direction="l",
 },
 [0x3118]={
  category="lo",
  direction="l",
 },
 [0x3119]={
  category="lo",
  direction="l",
 },
 [0x311A]={
  category="lo",
  direction="l",
 },
 [0x311B]={
  category="lo",
  direction="l",
 },
 [0x311C]={
  category="lo",
  direction="l",
 },
 [0x311D]={
  category="lo",
  direction="l",
 },
 [0x311E]={
  category="lo",
  direction="l",
 },
 [0x311F]={
  category="lo",
  direction="l",
 },
 [0x3120]={
  category="lo",
  direction="l",
 },
 [0x3121]={
  category="lo",
  direction="l",
 },
 [0x3122]={
  category="lo",
  direction="l",
 },
 [0x3123]={
  category="lo",
  direction="l",
 },
 [0x3124]={
  category="lo",
  direction="l",
 },
 [0x3125]={
  category="lo",
  direction="l",
 },
 [0x3126]={
  category="lo",
  direction="l",
 },
 [0x3127]={
  category="lo",
  direction="l",
 },
 [0x3128]={
  category="lo",
  direction="l",
 },
 [0x3129]={
  category="lo",
  direction="l",
 },
 [0x312A]={
  category="lo",
  direction="l",
 },
 [0x312B]={
  category="lo",
  direction="l",
 },
 [0x312C]={
  category="lo",
  direction="l",
 },
 [0x312D]={
  category="lo",
  direction="l",
 },
 [0x3131]={
  category="lo",
  direction="l",
 },
 [0x3132]={
  category="lo",
  direction="l",
 },
 [0x3133]={
  category="lo",
  direction="l",
 },
 [0x3134]={
  category="lo",
  direction="l",
 },
 [0x3135]={
  category="lo",
  direction="l",
 },
 [0x3136]={
  category="lo",
  direction="l",
 },
 [0x3137]={
  category="lo",
  direction="l",
 },
 [0x3138]={
  category="lo",
  direction="l",
 },
 [0x3139]={
  category="lo",
  direction="l",
 },
 [0x313A]={
  category="lo",
  direction="l",
 },
 [0x313B]={
  category="lo",
  direction="l",
 },
 [0x313C]={
  category="lo",
  direction="l",
 },
 [0x313D]={
  category="lo",
  direction="l",
 },
 [0x313E]={
  category="lo",
  direction="l",
 },
 [0x313F]={
  category="lo",
  direction="l",
 },
 [0x3140]={
  category="lo",
  direction="l",
 },
 [0x3141]={
  category="lo",
  direction="l",
 },
 [0x3142]={
  category="lo",
  direction="l",
 },
 [0x3143]={
  category="lo",
  direction="l",
 },
 [0x3144]={
  category="lo",
  direction="l",
 },
 [0x3145]={
  category="lo",
  direction="l",
 },
 [0x3146]={
  category="lo",
  direction="l",
 },
 [0x3147]={
  category="lo",
  direction="l",
 },
 [0x3148]={
  category="lo",
  direction="l",
 },
 [0x3149]={
  category="lo",
  direction="l",
 },
 [0x314A]={
  category="lo",
  direction="l",
 },
 [0x314B]={
  category="lo",
  direction="l",
 },
 [0x314C]={
  category="lo",
  direction="l",
 },
 [0x314D]={
  category="lo",
  direction="l",
 },
 [0x314E]={
  category="lo",
  direction="l",
 },
 [0x314F]={
  category="lo",
  direction="l",
 },
 [0x3150]={
  category="lo",
  direction="l",
 },
 [0x3151]={
  category="lo",
  direction="l",
 },
 [0x3152]={
  category="lo",
  direction="l",
 },
 [0x3153]={
  category="lo",
  direction="l",
 },
 [0x3154]={
  category="lo",
  direction="l",
 },
 [0x3155]={
  category="lo",
  direction="l",
 },
 [0x3156]={
  category="lo",
  direction="l",
 },
 [0x3157]={
  category="lo",
  direction="l",
 },
 [0x3158]={
  category="lo",
  direction="l",
 },
 [0x3159]={
  category="lo",
  direction="l",
 },
 [0x315A]={
  category="lo",
  direction="l",
 },
 [0x315B]={
  category="lo",
  direction="l",
 },
 [0x315C]={
  category="lo",
  direction="l",
 },
 [0x315D]={
  category="lo",
  direction="l",
 },
 [0x315E]={
  category="lo",
  direction="l",
 },
 [0x315F]={
  category="lo",
  direction="l",
 },
 [0x3160]={
  category="lo",
  direction="l",
 },
 [0x3161]={
  category="lo",
  direction="l",
 },
 [0x3162]={
  category="lo",
  direction="l",
 },
 [0x3163]={
  category="lo",
  direction="l",
 },
 [0x3164]={
  category="lo",
  direction="l",
 },
 [0x3165]={
  category="lo",
  direction="l",
 },
 [0x3166]={
  category="lo",
  direction="l",
 },
 [0x3167]={
  category="lo",
  direction="l",
 },
 [0x3168]={
  category="lo",
  direction="l",
 },
 [0x3169]={
  category="lo",
  direction="l",
 },
 [0x316A]={
  category="lo",
  direction="l",
 },
 [0x316B]={
  category="lo",
  direction="l",
 },
 [0x316C]={
  category="lo",
  direction="l",
 },
 [0x316D]={
  category="lo",
  direction="l",
 },
 [0x316E]={
  category="lo",
  direction="l",
 },
 [0x316F]={
  category="lo",
  direction="l",
 },
 [0x3170]={
  category="lo",
  direction="l",
 },
 [0x3171]={
  category="lo",
  direction="l",
 },
 [0x3172]={
  category="lo",
  direction="l",
 },
 [0x3173]={
  category="lo",
  direction="l",
 },
 [0x3174]={
  category="lo",
  direction="l",
 },
 [0x3175]={
  category="lo",
  direction="l",
 },
 [0x3176]={
  category="lo",
  direction="l",
 },
 [0x3177]={
  category="lo",
  direction="l",
 },
 [0x3178]={
  category="lo",
  direction="l",
 },
 [0x3179]={
  category="lo",
  direction="l",
 },
 [0x317A]={
  category="lo",
  direction="l",
 },
 [0x317B]={
  category="lo",
  direction="l",
 },
 [0x317C]={
  category="lo",
  direction="l",
 },
 [0x317D]={
  category="lo",
  direction="l",
 },
 [0x317E]={
  category="lo",
  direction="l",
 },
 [0x317F]={
  category="lo",
  direction="l",
 },
 [0x3180]={
  category="lo",
  direction="l",
 },
 [0x3181]={
  category="lo",
  direction="l",
 },
 [0x3182]={
  category="lo",
  direction="l",
 },
 [0x3183]={
  category="lo",
  direction="l",
 },
 [0x3184]={
  category="lo",
  direction="l",
 },
 [0x3185]={
  category="lo",
  direction="l",
 },
 [0x3186]={
  category="lo",
  direction="l",
 },
 [0x3187]={
  category="lo",
  direction="l",
 },
 [0x3188]={
  category="lo",
  direction="l",
 },
 [0x3189]={
  category="lo",
  direction="l",
 },
 [0x318A]={
  category="lo",
  direction="l",
 },
 [0x318B]={
  category="lo",
  direction="l",
 },
 [0x318C]={
  category="lo",
  direction="l",
 },
 [0x318D]={
  category="lo",
  direction="l",
 },
 [0x318E]={
  category="lo",
  direction="l",
 },
 [0x3190]={
  category="so",
  direction="l",
 },
 [0x3191]={
  category="so",
  direction="l",
 },
 [0x3192]={
  category="no",
  direction="l",
 },
 [0x3193]={
  category="no",
  direction="l",
 },
 [0x3194]={
  category="no",
  direction="l",
 },
 [0x3195]={
  category="no",
  direction="l",
 },
 [0x3196]={
  category="so",
  direction="l",
 },
 [0x3197]={
  category="so",
  direction="l",
 },
 [0x3198]={
  category="so",
  direction="l",
 },
 [0x3199]={
  category="so",
  direction="l",
 },
 [0x319A]={
  category="so",
  direction="l",
 },
 [0x319B]={
  category="so",
  direction="l",
 },
 [0x319C]={
  category="so",
  direction="l",
 },
 [0x319D]={
  category="so",
  direction="l",
 },
 [0x319E]={
  category="so",
  direction="l",
 },
 [0x319F]={
  category="so",
  direction="l",
 },
 [0x31A0]={
  category="lo",
  direction="l",
 },
 [0x31A1]={
  category="lo",
  direction="l",
 },
 [0x31A2]={
  category="lo",
  direction="l",
 },
 [0x31A3]={
  category="lo",
  direction="l",
 },
 [0x31A4]={
  category="lo",
  direction="l",
 },
 [0x31A5]={
  category="lo",
  direction="l",
 },
 [0x31A6]={
  category="lo",
  direction="l",
 },
 [0x31A7]={
  category="lo",
  direction="l",
 },
 [0x31A8]={
  category="lo",
  direction="l",
 },
 [0x31A9]={
  category="lo",
  direction="l",
 },
 [0x31AA]={
  category="lo",
  direction="l",
 },
 [0x31AB]={
  category="lo",
  direction="l",
 },
 [0x31AC]={
  category="lo",
  direction="l",
 },
 [0x31AD]={
  category="lo",
  direction="l",
 },
 [0x31AE]={
  category="lo",
  direction="l",
 },
 [0x31AF]={
  category="lo",
  direction="l",
 },
 [0x31B0]={
  category="lo",
  direction="l",
 },
 [0x31B1]={
  category="lo",
  direction="l",
 },
 [0x31B2]={
  category="lo",
  direction="l",
 },
 [0x31B3]={
  category="lo",
  direction="l",
 },
 [0x31B4]={
  category="lo",
  direction="l",
 },
 [0x31B5]={
  category="lo",
  direction="l",
 },
 [0x31B6]={
  category="lo",
  direction="l",
 },
 [0x31B7]={
  category="lo",
  direction="l",
 },
 [0x31B8]={
  category="lo",
  direction="l",
 },
 [0x31B9]={
  category="lo",
  direction="l",
 },
 [0x31BA]={
  category="lo",
  direction="l",
 },
 [0x31C0]={
  category="so",
  direction="on",
 },
 [0x31C1]={
  category="so",
  direction="on",
 },
 [0x31C2]={
  category="so",
  direction="on",
 },
 [0x31C3]={
  category="so",
  direction="on",
 },
 [0x31C4]={
  category="so",
  direction="on",
 },
 [0x31C5]={
  category="so",
  direction="on",
 },
 [0x31C6]={
  category="so",
  direction="on",
 },
 [0x31C7]={
  category="so",
  direction="on",
 },
 [0x31C8]={
  category="so",
  direction="on",
 },
 [0x31C9]={
  category="so",
  direction="on",
 },
 [0x31CA]={
  category="so",
  direction="on",
 },
 [0x31CB]={
  category="so",
  direction="on",
 },
 [0x31CC]={
  category="so",
  direction="on",
 },
 [0x31CD]={
  category="so",
  direction="on",
 },
 [0x31CE]={
  category="so",
  direction="on",
 },
 [0x31CF]={
  category="so",
  direction="on",
 },
 [0x31D0]={
  category="so",
  direction="on",
 },
 [0x31D1]={
  category="so",
  direction="on",
 },
 [0x31D2]={
  category="so",
  direction="on",
 },
 [0x31D3]={
  category="so",
  direction="on",
 },
 [0x31D4]={
  category="so",
  direction="on",
 },
 [0x31D5]={
  category="so",
  direction="on",
 },
 [0x31D6]={
  category="so",
  direction="on",
 },
 [0x31D7]={
  category="so",
  direction="on",
 },
 [0x31D8]={
  category="so",
  direction="on",
 },
 [0x31D9]={
  category="so",
  direction="on",
 },
 [0x31DA]={
  category="so",
  direction="on",
 },
 [0x31DB]={
  category="so",
  direction="on",
 },
 [0x31DC]={
  category="so",
  direction="on",
 },
 [0x31DD]={
  category="so",
  direction="on",
 },
 [0x31DE]={
  category="so",
  direction="on",
 },
 [0x31DF]={
  category="so",
  direction="on",
 },
 [0x31E0]={
  category="so",
  direction="on",
 },
 [0x31E1]={
  category="so",
  direction="on",
 },
 [0x31E2]={
  category="so",
  direction="on",
 },
 [0x31E3]={
  category="so",
  direction="on",
 },
 [0x31F0]={
  category="lo",
  direction="l",
 },
 [0x31F1]={
  category="lo",
  direction="l",
 },
 [0x31F2]={
  category="lo",
  direction="l",
 },
 [0x31F3]={
  category="lo",
  direction="l",
 },
 [0x31F4]={
  category="lo",
  direction="l",
 },
 [0x31F5]={
  category="lo",
  direction="l",
 },
 [0x31F6]={
  category="lo",
  direction="l",
 },
 [0x31F7]={
  category="lo",
  direction="l",
 },
 [0x31F8]={
  category="lo",
  direction="l",
 },
 [0x31F9]={
  category="lo",
  direction="l",
 },
 [0x31FA]={
  category="lo",
  direction="l",
 },
 [0x31FB]={
  category="lo",
  direction="l",
 },
 [0x31FC]={
  category="lo",
  direction="l",
 },
 [0x31FD]={
  category="lo",
  direction="l",
 },
 [0x31FE]={
  category="lo",
  direction="l",
 },
 [0x31FF]={
  category="lo",
  direction="l",
 },
 [0x3200]={
  category="so",
  direction="l",
 },
 [0x3201]={
  category="so",
  direction="l",
 },
 [0x3202]={
  category="so",
  direction="l",
 },
 [0x3203]={
  category="so",
  direction="l",
 },
 [0x3204]={
  category="so",
  direction="l",
 },
 [0x3205]={
  category="so",
  direction="l",
 },
 [0x3206]={
  category="so",
  direction="l",
 },
 [0x3207]={
  category="so",
  direction="l",
 },
 [0x3208]={
  category="so",
  direction="l",
 },
 [0x3209]={
  category="so",
  direction="l",
 },
 [0x320A]={
  category="so",
  direction="l",
 },
 [0x320B]={
  category="so",
  direction="l",
 },
 [0x320C]={
  category="so",
  direction="l",
 },
 [0x320D]={
  category="so",
  direction="l",
 },
 [0x320E]={
  category="so",
  direction="l",
 },
 [0x320F]={
  category="so",
  direction="l",
 },
 [0x3210]={
  category="so",
  direction="l",
 },
 [0x3211]={
  category="so",
  direction="l",
 },
 [0x3212]={
  category="so",
  direction="l",
 },
 [0x3213]={
  category="so",
  direction="l",
 },
 [0x3214]={
  category="so",
  direction="l",
 },
 [0x3215]={
  category="so",
  direction="l",
 },
 [0x3216]={
  category="so",
  direction="l",
 },
 [0x3217]={
  category="so",
  direction="l",
 },
 [0x3218]={
  category="so",
  direction="l",
 },
 [0x3219]={
  category="so",
  direction="l",
 },
 [0x321A]={
  category="so",
  direction="l",
 },
 [0x321B]={
  category="so",
  direction="l",
 },
 [0x321C]={
  category="so",
  direction="l",
 },
 [0x321D]={
  category="so",
  direction="on",
 },
 [0x321E]={
  category="so",
  direction="on",
 },
 [0x3220]={
  category="no",
  direction="l",
 },
 [0x3221]={
  category="no",
  direction="l",
 },
 [0x3222]={
  category="no",
  direction="l",
 },
 [0x3223]={
  category="no",
  direction="l",
 },
 [0x3224]={
  category="no",
  direction="l",
 },
 [0x3225]={
  category="no",
  direction="l",
 },
 [0x3226]={
  category="no",
  direction="l",
 },
 [0x3227]={
  category="no",
  direction="l",
 },
 [0x3228]={
  category="no",
  direction="l",
 },
 [0x3229]={
  category="no",
  direction="l",
 },
 [0x322A]={
  category="so",
  direction="l",
 },
 [0x322B]={
  category="so",
  direction="l",
 },
 [0x322C]={
  category="so",
  direction="l",
 },
 [0x322D]={
  category="so",
  direction="l",
 },
 [0x322E]={
  category="so",
  direction="l",
 },
 [0x322F]={
  category="so",
  direction="l",
 },
 [0x3230]={
  category="so",
  direction="l",
 },
 [0x3231]={
  category="so",
  direction="l",
 },
 [0x3232]={
  category="so",
  direction="l",
 },
 [0x3233]={
  category="so",
  direction="l",
 },
 [0x3234]={
  category="so",
  direction="l",
 },
 [0x3235]={
  category="so",
  direction="l",
 },
 [0x3236]={
  category="so",
  direction="l",
 },
 [0x3237]={
  category="so",
  direction="l",
 },
 [0x3238]={
  category="so",
  direction="l",
 },
 [0x3239]={
  category="so",
  direction="l",
 },
 [0x323A]={
  category="so",
  direction="l",
 },
 [0x323B]={
  category="so",
  direction="l",
 },
 [0x323C]={
  category="so",
  direction="l",
 },
 [0x323D]={
  category="so",
  direction="l",
 },
 [0x323E]={
  category="so",
  direction="l",
 },
 [0x323F]={
  category="so",
  direction="l",
 },
 [0x3240]={
  category="so",
  direction="l",
 },
 [0x3241]={
  category="so",
  direction="l",
 },
 [0x3242]={
  category="so",
  direction="l",
 },
 [0x3243]={
  category="so",
  direction="l",
 },
 [0x3244]={
  category="so",
  direction="l",
 },
 [0x3245]={
  category="so",
  direction="l",
 },
 [0x3246]={
  category="so",
  direction="l",
 },
 [0x3247]={
  category="so",
  direction="l",
 },
 [0x3248]={
  category="so",
  direction="l",
 },
 [0x3249]={
  category="so",
  direction="l",
 },
 [0x324A]={
  category="so",
  direction="l",
 },
 [0x324B]={
  category="so",
  direction="l",
 },
 [0x324C]={
  category="so",
  direction="l",
 },
 [0x324D]={
  category="so",
  direction="l",
 },
 [0x324E]={
  category="so",
  direction="l",
 },
 [0x324F]={
  category="so",
  direction="l",
 },
 [0x3250]={
  category="so",
  direction="on",
 },
 [0x3251]={
  category="no",
  direction="on",
 },
 [0x3252]={
  category="no",
  direction="on",
 },
 [0x3253]={
  category="no",
  direction="on",
 },
 [0x3254]={
  category="no",
  direction="on",
 },
 [0x3255]={
  category="no",
  direction="on",
 },
 [0x3256]={
  category="no",
  direction="on",
 },
 [0x3257]={
  category="no",
  direction="on",
 },
 [0x3258]={
  category="no",
  direction="on",
 },
 [0x3259]={
  category="no",
  direction="on",
 },
 [0x325A]={
  category="no",
  direction="on",
 },
 [0x325B]={
  category="no",
  direction="on",
 },
 [0x325C]={
  category="no",
  direction="on",
 },
 [0x325D]={
  category="no",
  direction="on",
 },
 [0x325E]={
  category="no",
  direction="on",
 },
 [0x325F]={
  category="no",
  direction="on",
 },
 [0x3260]={
  category="so",
  direction="l",
 },
 [0x3261]={
  category="so",
  direction="l",
 },
 [0x3262]={
  category="so",
  direction="l",
 },
 [0x3263]={
  category="so",
  direction="l",
 },
 [0x3264]={
  category="so",
  direction="l",
 },
 [0x3265]={
  category="so",
  direction="l",
 },
 [0x3266]={
  category="so",
  direction="l",
 },
 [0x3267]={
  category="so",
  direction="l",
 },
 [0x3268]={
  category="so",
  direction="l",
 },
 [0x3269]={
  category="so",
  direction="l",
 },
 [0x326A]={
  category="so",
  direction="l",
 },
 [0x326B]={
  category="so",
  direction="l",
 },
 [0x326C]={
  category="so",
  direction="l",
 },
 [0x326D]={
  category="so",
  direction="l",
 },
 [0x326E]={
  category="so",
  direction="l",
 },
 [0x326F]={
  category="so",
  direction="l",
 },
 [0x3270]={
  category="so",
  direction="l",
 },
 [0x3271]={
  category="so",
  direction="l",
 },
 [0x3272]={
  category="so",
  direction="l",
 },
 [0x3273]={
  category="so",
  direction="l",
 },
 [0x3274]={
  category="so",
  direction="l",
 },
 [0x3275]={
  category="so",
  direction="l",
 },
 [0x3276]={
  category="so",
  direction="l",
 },
 [0x3277]={
  category="so",
  direction="l",
 },
 [0x3278]={
  category="so",
  direction="l",
 },
 [0x3279]={
  category="so",
  direction="l",
 },
 [0x327A]={
  category="so",
  direction="l",
 },
 [0x327B]={
  category="so",
  direction="l",
 },
 [0x327C]={
  category="so",
  direction="on",
 },
 [0x327D]={
  category="so",
  direction="on",
 },
 [0x327E]={
  category="so",
  direction="on",
 },
 [0x327F]={
  category="so",
  direction="l",
 },
 [0x3280]={
  category="no",
  direction="l",
 },
 [0x3281]={
  category="no",
  direction="l",
 },
 [0x3282]={
  category="no",
  direction="l",
 },
 [0x3283]={
  category="no",
  direction="l",
 },
 [0x3284]={
  category="no",
  direction="l",
 },
 [0x3285]={
  category="no",
  direction="l",
 },
 [0x3286]={
  category="no",
  direction="l",
 },
 [0x3287]={
  category="no",
  direction="l",
 },
 [0x3288]={
  category="no",
  direction="l",
 },
 [0x3289]={
  category="no",
  direction="l",
 },
 [0x328A]={
  category="so",
  direction="l",
 },
 [0x328B]={
  category="so",
  direction="l",
 },
 [0x328C]={
  category="so",
  direction="l",
 },
 [0x328D]={
  category="so",
  direction="l",
 },
 [0x328E]={
  category="so",
  direction="l",
 },
 [0x328F]={
  category="so",
  direction="l",
 },
 [0x3290]={
  category="so",
  direction="l",
 },
 [0x3291]={
  category="so",
  direction="l",
 },
 [0x3292]={
  category="so",
  direction="l",
 },
 [0x3293]={
  category="so",
  direction="l",
 },
 [0x3294]={
  category="so",
  direction="l",
 },
 [0x3295]={
  category="so",
  direction="l",
 },
 [0x3296]={
  category="so",
  direction="l",
 },
 [0x3297]={
  category="so",
  direction="l",
 },
 [0x3298]={
  category="so",
  direction="l",
 },
 [0x3299]={
  category="so",
  direction="l",
 },
 [0x329A]={
  category="so",
  direction="l",
 },
 [0x329B]={
  category="so",
  direction="l",
 },
 [0x329C]={
  category="so",
  direction="l",
 },
 [0x329D]={
  category="so",
  direction="l",
 },
 [0x329E]={
  category="so",
  direction="l",
 },
 [0x329F]={
  category="so",
  direction="l",
 },
 [0x32A0]={
  category="so",
  direction="l",
 },
 [0x32A1]={
  category="so",
  direction="l",
 },
 [0x32A2]={
  category="so",
  direction="l",
 },
 [0x32A3]={
  category="so",
  direction="l",
 },
 [0x32A4]={
  category="so",
  direction="l",
 },
 [0x32A5]={
  category="so",
  direction="l",
 },
 [0x32A6]={
  category="so",
  direction="l",
 },
 [0x32A7]={
  category="so",
  direction="l",
 },
 [0x32A8]={
  category="so",
  direction="l",
 },
 [0x32A9]={
  category="so",
  direction="l",
 },
 [0x32AA]={
  category="so",
  direction="l",
 },
 [0x32AB]={
  category="so",
  direction="l",
 },
 [0x32AC]={
  category="so",
  direction="l",
 },
 [0x32AD]={
  category="so",
  direction="l",
 },
 [0x32AE]={
  category="so",
  direction="l",
 },
 [0x32AF]={
  category="so",
  direction="l",
 },
 [0x32B0]={
  category="so",
  direction="l",
 },
 [0x32B1]={
  category="no",
  direction="on",
 },
 [0x32B2]={
  category="no",
  direction="on",
 },
 [0x32B3]={
  category="no",
  direction="on",
 },
 [0x32B4]={
  category="no",
  direction="on",
 },
 [0x32B5]={
  category="no",
  direction="on",
 },
 [0x32B6]={
  category="no",
  direction="on",
 },
 [0x32B7]={
  category="no",
  direction="on",
 },
 [0x32B8]={
  category="no",
  direction="on",
 },
 [0x32B9]={
  category="no",
  direction="on",
 },
 [0x32BA]={
  category="no",
  direction="on",
 },
 [0x32BB]={
  category="no",
  direction="on",
 },
 [0x32BC]={
  category="no",
  direction="on",
 },
 [0x32BD]={
  category="no",
  direction="on",
 },
 [0x32BE]={
  category="no",
  direction="on",
 },
 [0x32BF]={
  category="no",
  direction="on",
 },
 [0x32C0]={
  category="so",
  direction="l",
 },
 [0x32C1]={
  category="so",
  direction="l",
 },
 [0x32C2]={
  category="so",
  direction="l",
 },
 [0x32C3]={
  category="so",
  direction="l",
 },
 [0x32C4]={
  category="so",
  direction="l",
 },
 [0x32C5]={
  category="so",
  direction="l",
 },
 [0x32C6]={
  category="so",
  direction="l",
 },
 [0x32C7]={
  category="so",
  direction="l",
 },
 [0x32C8]={
  category="so",
  direction="l",
 },
 [0x32C9]={
  category="so",
  direction="l",
 },
 [0x32CA]={
  category="so",
  direction="l",
 },
 [0x32CB]={
  category="so",
  direction="l",
 },
 [0x32CC]={
  category="so",
  direction="on",
 },
 [0x32CD]={
  category="so",
  direction="on",
 },
 [0x32CE]={
  category="so",
  direction="on",
 },
 [0x32CF]={
  category="so",
  direction="on",
 },
 [0x32D0]={
  category="so",
  direction="l",
 },
 [0x32D1]={
  category="so",
  direction="l",
 },
 [0x32D2]={
  category="so",
  direction="l",
 },
 [0x32D3]={
  category="so",
  direction="l",
 },
 [0x32D4]={
  category="so",
  direction="l",
 },
 [0x32D5]={
  category="so",
  direction="l",
 },
 [0x32D6]={
  category="so",
  direction="l",
 },
 [0x32D7]={
  category="so",
  direction="l",
 },
 [0x32D8]={
  category="so",
  direction="l",
 },
 [0x32D9]={
  category="so",
  direction="l",
 },
 [0x32DA]={
  category="so",
  direction="l",
 },
 [0x32DB]={
  category="so",
  direction="l",
 },
 [0x32DC]={
  category="so",
  direction="l",
 },
 [0x32DD]={
  category="so",
  direction="l",
 },
 [0x32DE]={
  category="so",
  direction="l",
 },
 [0x32DF]={
  category="so",
  direction="l",
 },
 [0x32E0]={
  category="so",
  direction="l",
 },
 [0x32E1]={
  category="so",
  direction="l",
 },
 [0x32E2]={
  category="so",
  direction="l",
 },
 [0x32E3]={
  category="so",
  direction="l",
 },
 [0x32E4]={
  category="so",
  direction="l",
 },
 [0x32E5]={
  category="so",
  direction="l",
 },
 [0x32E6]={
  category="so",
  direction="l",
 },
 [0x32E7]={
  category="so",
  direction="l",
 },
 [0x32E8]={
  category="so",
  direction="l",
 },
 [0x32E9]={
  category="so",
  direction="l",
 },
 [0x32EA]={
  category="so",
  direction="l",
 },
 [0x32EB]={
  category="so",
  direction="l",
 },
 [0x32EC]={
  category="so",
  direction="l",
 },
 [0x32ED]={
  category="so",
  direction="l",
 },
 [0x32EE]={
  category="so",
  direction="l",
 },
 [0x32EF]={
  category="so",
  direction="l",
 },
 [0x32F0]={
  category="so",
  direction="l",
 },
 [0x32F1]={
  category="so",
  direction="l",
 },
 [0x32F2]={
  category="so",
  direction="l",
 },
 [0x32F3]={
  category="so",
  direction="l",
 },
 [0x32F4]={
  category="so",
  direction="l",
 },
 [0x32F5]={
  category="so",
  direction="l",
 },
 [0x32F6]={
  category="so",
  direction="l",
 },
 [0x32F7]={
  category="so",
  direction="l",
 },
 [0x32F8]={
  category="so",
  direction="l",
 },
 [0x32F9]={
  category="so",
  direction="l",
 },
 [0x32FA]={
  category="so",
  direction="l",
 },
 [0x32FB]={
  category="so",
  direction="l",
 },
 [0x32FC]={
  category="so",
  direction="l",
 },
 [0x32FD]={
  category="so",
  direction="l",
 },
 [0x32FE]={
  category="so",
  direction="l",
 },
 [0x3300]={
  category="so",
  direction="l",
 },
 [0x3301]={
  category="so",
  direction="l",
 },
 [0x3302]={
  category="so",
  direction="l",
 },
 [0x3303]={
  category="so",
  direction="l",
 },
 [0x3304]={
  category="so",
  direction="l",
 },
 [0x3305]={
  category="so",
  direction="l",
 },
 [0x3306]={
  category="so",
  direction="l",
 },
 [0x3307]={
  category="so",
  direction="l",
 },
 [0x3308]={
  category="so",
  direction="l",
 },
 [0x3309]={
  category="so",
  direction="l",
 },
 [0x330A]={
  category="so",
  direction="l",
 },
 [0x330B]={
  category="so",
  direction="l",
 },
 [0x330C]={
  category="so",
  direction="l",
 },
 [0x330D]={
  category="so",
  direction="l",
 },
 [0x330E]={
  category="so",
  direction="l",
 },
 [0x330F]={
  category="so",
  direction="l",
 },
 [0x3310]={
  category="so",
  direction="l",
 },
 [0x3311]={
  category="so",
  direction="l",
 },
 [0x3312]={
  category="so",
  direction="l",
 },
 [0x3313]={
  category="so",
  direction="l",
 },
 [0x3314]={
  category="so",
  direction="l",
 },
 [0x3315]={
  category="so",
  direction="l",
 },
 [0x3316]={
  category="so",
  direction="l",
 },
 [0x3317]={
  category="so",
  direction="l",
 },
 [0x3318]={
  category="so",
  direction="l",
 },
 [0x3319]={
  category="so",
  direction="l",
 },
 [0x331A]={
  category="so",
  direction="l",
 },
 [0x331B]={
  category="so",
  direction="l",
 },
 [0x331C]={
  category="so",
  direction="l",
 },
 [0x331D]={
  category="so",
  direction="l",
 },
 [0x331E]={
  category="so",
  direction="l",
 },
 [0x331F]={
  category="so",
  direction="l",
 },
 [0x3320]={
  category="so",
  direction="l",
 },
 [0x3321]={
  category="so",
  direction="l",
 },
 [0x3322]={
  category="so",
  direction="l",
 },
 [0x3323]={
  category="so",
  direction="l",
 },
 [0x3324]={
  category="so",
  direction="l",
 },
 [0x3325]={
  category="so",
  direction="l",
 },
 [0x3326]={
  category="so",
  direction="l",
 },
 [0x3327]={
  category="so",
  direction="l",
 },
 [0x3328]={
  category="so",
  direction="l",
 },
 [0x3329]={
  category="so",
  direction="l",
 },
 [0x332A]={
  category="so",
  direction="l",
 },
 [0x332B]={
  category="so",
  direction="l",
 },
 [0x332C]={
  category="so",
  direction="l",
 },
 [0x332D]={
  category="so",
  direction="l",
 },
 [0x332E]={
  category="so",
  direction="l",
 },
 [0x332F]={
  category="so",
  direction="l",
 },
 [0x3330]={
  category="so",
  direction="l",
 },
 [0x3331]={
  category="so",
  direction="l",
 },
 [0x3332]={
  category="so",
  direction="l",
 },
 [0x3333]={
  category="so",
  direction="l",
 },
 [0x3334]={
  category="so",
  direction="l",
 },
 [0x3335]={
  category="so",
  direction="l",
 },
 [0x3336]={
  category="so",
  direction="l",
 },
 [0x3337]={
  category="so",
  direction="l",
 },
 [0x3338]={
  category="so",
  direction="l",
 },
 [0x3339]={
  category="so",
  direction="l",
 },
 [0x333A]={
  category="so",
  direction="l",
 },
 [0x333B]={
  category="so",
  direction="l",
 },
 [0x333C]={
  category="so",
  direction="l",
 },
 [0x333D]={
  category="so",
  direction="l",
 },
 [0x333E]={
  category="so",
  direction="l",
 },
 [0x333F]={
  category="so",
  direction="l",
 },
 [0x3340]={
  category="so",
  direction="l",
 },
 [0x3341]={
  category="so",
  direction="l",
 },
 [0x3342]={
  category="so",
  direction="l",
 },
 [0x3343]={
  category="so",
  direction="l",
 },
 [0x3344]={
  category="so",
  direction="l",
 },
 [0x3345]={
  category="so",
  direction="l",
 },
 [0x3346]={
  category="so",
  direction="l",
 },
 [0x3347]={
  category="so",
  direction="l",
 },
 [0x3348]={
  category="so",
  direction="l",
 },
 [0x3349]={
  category="so",
  direction="l",
 },
 [0x334A]={
  category="so",
  direction="l",
 },
 [0x334B]={
  category="so",
  direction="l",
 },
 [0x334C]={
  category="so",
  direction="l",
 },
 [0x334D]={
  category="so",
  direction="l",
 },
 [0x334E]={
  category="so",
  direction="l",
 },
 [0x334F]={
  category="so",
  direction="l",
 },
 [0x3350]={
  category="so",
  direction="l",
 },
 [0x3351]={
  category="so",
  direction="l",
 },
 [0x3352]={
  category="so",
  direction="l",
 },
 [0x3353]={
  category="so",
  direction="l",
 },
 [0x3354]={
  category="so",
  direction="l",
 },
 [0x3355]={
  category="so",
  direction="l",
 },
 [0x3356]={
  category="so",
  direction="l",
 },
 [0x3357]={
  category="so",
  direction="l",
 },
 [0x3358]={
  category="so",
  direction="l",
 },
 [0x3359]={
  category="so",
  direction="l",
 },
 [0x335A]={
  category="so",
  direction="l",
 },
 [0x335B]={
  category="so",
  direction="l",
 },
 [0x335C]={
  category="so",
  direction="l",
 },
 [0x335D]={
  category="so",
  direction="l",
 },
 [0x335E]={
  category="so",
  direction="l",
 },
 [0x335F]={
  category="so",
  direction="l",
 },
 [0x3360]={
  category="so",
  direction="l",
 },
 [0x3361]={
  category="so",
  direction="l",
 },
 [0x3362]={
  category="so",
  direction="l",
 },
 [0x3363]={
  category="so",
  direction="l",
 },
 [0x3364]={
  category="so",
  direction="l",
 },
 [0x3365]={
  category="so",
  direction="l",
 },
 [0x3366]={
  category="so",
  direction="l",
 },
 [0x3367]={
  category="so",
  direction="l",
 },
 [0x3368]={
  category="so",
  direction="l",
 },
 [0x3369]={
  category="so",
  direction="l",
 },
 [0x336A]={
  category="so",
  direction="l",
 },
 [0x336B]={
  category="so",
  direction="l",
 },
 [0x336C]={
  category="so",
  direction="l",
 },
 [0x336D]={
  category="so",
  direction="l",
 },
 [0x336E]={
  category="so",
  direction="l",
 },
 [0x336F]={
  category="so",
  direction="l",
 },
 [0x3370]={
  category="so",
  direction="l",
 },
 [0x3371]={
  category="so",
  direction="l",
 },
 [0x3372]={
  category="so",
  direction="l",
 },
 [0x3373]={
  category="so",
  direction="l",
 },
 [0x3374]={
  category="so",
  direction="l",
 },
 [0x3375]={
  category="so",
  direction="l",
 },
 [0x3376]={
  category="so",
  direction="l",
 },
 [0x3377]={
  category="so",
  direction="on",
 },
 [0x3378]={
  category="so",
  direction="on",
 },
 [0x3379]={
  category="so",
  direction="on",
 },
 [0x337A]={
  category="so",
  direction="on",
 },
 [0x337B]={
  category="so",
  direction="l",
 },
 [0x337C]={
  category="so",
  direction="l",
 },
 [0x337D]={
  category="so",
  direction="l",
 },
 [0x337E]={
  category="so",
  direction="l",
 },
 [0x337F]={
  category="so",
  direction="l",
 },
 [0x3380]={
  category="so",
  direction="l",
 },
 [0x3381]={
  category="so",
  direction="l",
 },
 [0x3382]={
  category="so",
  direction="l",
 },
 [0x3383]={
  category="so",
  direction="l",
 },
 [0x3384]={
  category="so",
  direction="l",
 },
 [0x3385]={
  category="so",
  direction="l",
 },
 [0x3386]={
  category="so",
  direction="l",
 },
 [0x3387]={
  category="so",
  direction="l",
 },
 [0x3388]={
  category="so",
  direction="l",
 },
 [0x3389]={
  category="so",
  direction="l",
 },
 [0x338A]={
  category="so",
  direction="l",
 },
 [0x338B]={
  category="so",
  direction="l",
 },
 [0x338C]={
  category="so",
  direction="l",
 },
 [0x338D]={
  category="so",
  direction="l",
 },
 [0x338E]={
  category="so",
  direction="l",
 },
 [0x338F]={
  category="so",
  direction="l",
 },
 [0x3390]={
  category="so",
  direction="l",
 },
 [0x3391]={
  category="so",
  direction="l",
 },
 [0x3392]={
  category="so",
  direction="l",
 },
 [0x3393]={
  category="so",
  direction="l",
 },
 [0x3394]={
  category="so",
  direction="l",
 },
 [0x3395]={
  category="so",
  direction="l",
 },
 [0x3396]={
  category="so",
  direction="l",
 },
 [0x3397]={
  category="so",
  direction="l",
 },
 [0x3398]={
  category="so",
  direction="l",
 },
 [0x3399]={
  category="so",
  direction="l",
 },
 [0x339A]={
  category="so",
  direction="l",
 },
 [0x339B]={
  category="so",
  direction="l",
 },
 [0x339C]={
  category="so",
  direction="l",
 },
 [0x339D]={
  category="so",
  direction="l",
 },
 [0x339E]={
  category="so",
  direction="l",
 },
 [0x339F]={
  category="so",
  direction="l",
 },
 [0x33A0]={
  category="so",
  direction="l",
 },
 [0x33A1]={
  category="so",
  direction="l",
 },
 [0x33A2]={
  category="so",
  direction="l",
 },
 [0x33A3]={
  category="so",
  direction="l",
 },
 [0x33A4]={
  category="so",
  direction="l",
 },
 [0x33A5]={
  category="so",
  direction="l",
 },
 [0x33A6]={
  category="so",
  direction="l",
 },
 [0x33A7]={
  category="so",
  direction="l",
 },
 [0x33A8]={
  category="so",
  direction="l",
 },
 [0x33A9]={
  category="so",
  direction="l",
 },
 [0x33AA]={
  category="so",
  direction="l",
 },
 [0x33AB]={
  category="so",
  direction="l",
 },
 [0x33AC]={
  category="so",
  direction="l",
 },
 [0x33AD]={
  category="so",
  direction="l",
 },
 [0x33AE]={
  category="so",
  direction="l",
 },
 [0x33AF]={
  category="so",
  direction="l",
 },
 [0x33B0]={
  category="so",
  direction="l",
 },
 [0x33B1]={
  category="so",
  direction="l",
 },
 [0x33B2]={
  category="so",
  direction="l",
 },
 [0x33B3]={
  category="so",
  direction="l",
 },
 [0x33B4]={
  category="so",
  direction="l",
 },
 [0x33B5]={
  category="so",
  direction="l",
 },
 [0x33B6]={
  category="so",
  direction="l",
 },
 [0x33B7]={
  category="so",
  direction="l",
 },
 [0x33B8]={
  category="so",
  direction="l",
 },
 [0x33B9]={
  category="so",
  direction="l",
 },
 [0x33BA]={
  category="so",
  direction="l",
 },
 [0x33BB]={
  category="so",
  direction="l",
 },
 [0x33BC]={
  category="so",
  direction="l",
 },
 [0x33BD]={
  category="so",
  direction="l",
 },
 [0x33BE]={
  category="so",
  direction="l",
 },
 [0x33BF]={
  category="so",
  direction="l",
 },
 [0x33C0]={
  category="so",
  direction="l",
 },
 [0x33C1]={
  category="so",
  direction="l",
 },
 [0x33C2]={
  category="so",
  direction="l",
 },
 [0x33C3]={
  category="so",
  direction="l",
 },
 [0x33C4]={
  category="so",
  direction="l",
 },
 [0x33C5]={
  category="so",
  direction="l",
 },
 [0x33C6]={
  category="so",
  direction="l",
 },
 [0x33C7]={
  category="so",
  direction="l",
 },
 [0x33C8]={
  category="so",
  direction="l",
 },
 [0x33C9]={
  category="so",
  direction="l",
 },
 [0x33CA]={
  category="so",
  direction="l",
 },
 [0x33CB]={
  category="so",
  direction="l",
 },
 [0x33CC]={
  category="so",
  direction="l",
 },
 [0x33CD]={
  category="so",
  direction="l",
 },
 [0x33CE]={
  category="so",
  direction="l",
 },
 [0x33CF]={
  category="so",
  direction="l",
 },
 [0x33D0]={
  category="so",
  direction="l",
 },
 [0x33D1]={
  category="so",
  direction="l",
 },
 [0x33D2]={
  category="so",
  direction="l",
 },
 [0x33D3]={
  category="so",
  direction="l",
 },
 [0x33D4]={
  category="so",
  direction="l",
 },
 [0x33D5]={
  category="so",
  direction="l",
 },
 [0x33D6]={
  category="so",
  direction="l",
 },
 [0x33D7]={
  category="so",
  direction="l",
 },
 [0x33D8]={
  category="so",
  direction="l",
 },
 [0x33D9]={
  category="so",
  direction="l",
 },
 [0x33DA]={
  category="so",
  direction="l",
 },
 [0x33DB]={
  category="so",
  direction="l",
 },
 [0x33DC]={
  category="so",
  direction="l",
 },
 [0x33DD]={
  category="so",
  direction="l",
 },
 [0x33DE]={
  category="so",
  direction="on",
 },
 [0x33DF]={
  category="so",
  direction="on",
 },
 [0x33E0]={
  category="so",
  direction="l",
 },
 [0x33E1]={
  category="so",
  direction="l",
 },
 [0x33E2]={
  category="so",
  direction="l",
 },
 [0x33E3]={
  category="so",
  direction="l",
 },
 [0x33E4]={
  category="so",
  direction="l",
 },
 [0x33E5]={
  category="so",
  direction="l",
 },
 [0x33E6]={
  category="so",
  direction="l",
 },
 [0x33E7]={
  category="so",
  direction="l",
 },
 [0x33E8]={
  category="so",
  direction="l",
 },
 [0x33E9]={
  category="so",
  direction="l",
 },
 [0x33EA]={
  category="so",
  direction="l",
 },
 [0x33EB]={
  category="so",
  direction="l",
 },
 [0x33EC]={
  category="so",
  direction="l",
 },
 [0x33ED]={
  category="so",
  direction="l",
 },
 [0x33EE]={
  category="so",
  direction="l",
 },
 [0x33EF]={
  category="so",
  direction="l",
 },
 [0x33F0]={
  category="so",
  direction="l",
 },
 [0x33F1]={
  category="so",
  direction="l",
 },
 [0x33F2]={
  category="so",
  direction="l",
 },
 [0x33F3]={
  category="so",
  direction="l",
 },
 [0x33F4]={
  category="so",
  direction="l",
 },
 [0x33F5]={
  category="so",
  direction="l",
 },
 [0x33F6]={
  category="so",
  direction="l",
 },
 [0x33F7]={
  category="so",
  direction="l",
 },
 [0x33F8]={
  category="so",
  direction="l",
 },
 [0x33F9]={
  category="so",
  direction="l",
 },
 [0x33FA]={
  category="so",
  direction="l",
 },
 [0x33FB]={
  category="so",
  direction="l",
 },
 [0x33FC]={
  category="so",
  direction="l",
 },
 [0x33FD]={
  category="so",
  direction="l",
 },
 [0x33FE]={
  category="so",
  direction="l",
 },
 [0x33FF]={
  category="so",
  direction="on",
 },
 [0x4DC0]={
  category="so",
  direction="on",
 },
 [0x4DC1]={
  category="so",
  direction="on",
 },
 [0x4DC2]={
  category="so",
  direction="on",
 },
 [0x4DC3]={
  category="so",
  direction="on",
 },
 [0x4DC4]={
  category="so",
  direction="on",
 },
 [0x4DC5]={
  category="so",
  direction="on",
 },
 [0x4DC6]={
  category="so",
  direction="on",
 },
 [0x4DC7]={
  category="so",
  direction="on",
 },
 [0x4DC8]={
  category="so",
  direction="on",
 },
 [0x4DC9]={
  category="so",
  direction="on",
 },
 [0x4DCA]={
  category="so",
  direction="on",
 },
 [0x4DCB]={
  category="so",
  direction="on",
 },
 [0x4DCC]={
  category="so",
  direction="on",
 },
 [0x4DCD]={
  category="so",
  direction="on",
 },
 [0x4DCE]={
  category="so",
  direction="on",
 },
 [0x4DCF]={
  category="so",
  direction="on",
 },
 [0x4DD0]={
  category="so",
  direction="on",
 },
 [0x4DD1]={
  category="so",
  direction="on",
 },
 [0x4DD2]={
  category="so",
  direction="on",
 },
 [0x4DD3]={
  category="so",
  direction="on",
 },
 [0x4DD4]={
  category="so",
  direction="on",
 },
 [0x4DD5]={
  category="so",
  direction="on",
 },
 [0x4DD6]={
  category="so",
  direction="on",
 },
 [0x4DD7]={
  category="so",
  direction="on",
 },
 [0x4DD8]={
  category="so",
  direction="on",
 },
 [0x4DD9]={
  category="so",
  direction="on",
 },
 [0x4DDA]={
  category="so",
  direction="on",
 },
 [0x4DDB]={
  category="so",
  direction="on",
 },
 [0x4DDC]={
  category="so",
  direction="on",
 },
 [0x4DDD]={
  category="so",
  direction="on",
 },
 [0x4DDE]={
  category="so",
  direction="on",
 },
 [0x4DDF]={
  category="so",
  direction="on",
 },
 [0x4DE0]={
  category="so",
  direction="on",
 },
 [0x4DE1]={
  category="so",
  direction="on",
 },
 [0x4DE2]={
  category="so",
  direction="on",
 },
 [0x4DE3]={
  category="so",
  direction="on",
 },
 [0x4DE4]={
  category="so",
  direction="on",
 },
 [0x4DE5]={
  category="so",
  direction="on",
 },
 [0x4DE6]={
  category="so",
  direction="on",
 },
 [0x4DE7]={
  category="so",
  direction="on",
 },
 [0x4DE8]={
  category="so",
  direction="on",
 },
 [0x4DE9]={
  category="so",
  direction="on",
 },
 [0x4DEA]={
  category="so",
  direction="on",
 },
 [0x4DEB]={
  category="so",
  direction="on",
 },
 [0x4DEC]={
  category="so",
  direction="on",
 },
 [0x4DED]={
  category="so",
  direction="on",
 },
 [0x4DEE]={
  category="so",
  direction="on",
 },
 [0x4DEF]={
  category="so",
  direction="on",
 },
 [0x4DF0]={
  category="so",
  direction="on",
 },
 [0x4DF1]={
  category="so",
  direction="on",
 },
 [0x4DF2]={
  category="so",
  direction="on",
 },
 [0x4DF3]={
  category="so",
  direction="on",
 },
 [0x4DF4]={
  category="so",
  direction="on",
 },
 [0x4DF5]={
  category="so",
  direction="on",
 },
 [0x4DF6]={
  category="so",
  direction="on",
 },
 [0x4DF7]={
  category="so",
  direction="on",
 },
 [0x4DF8]={
  category="so",
  direction="on",
 },
 [0x4DF9]={
  category="so",
  direction="on",
 },
 [0x4DFA]={
  category="so",
  direction="on",
 },
 [0x4DFB]={
  category="so",
  direction="on",
 },
 [0x4DFC]={
  category="so",
  direction="on",
 },
 [0x4DFD]={
  category="so",
  direction="on",
 },
 [0x4DFE]={
  category="so",
  direction="on",
 },
 [0x4DFF]={
  category="so",
  direction="on",
 },
 [0xA000]={
  category="lo",
  direction="l",
 },
 [0xA001]={
  category="lo",
  direction="l",
 },
 [0xA002]={
  category="lo",
  direction="l",
 },
 [0xA003]={
  category="lo",
  direction="l",
 },
 [0xA004]={
  category="lo",
  direction="l",
 },
 [0xA005]={
  category="lo",
  direction="l",
 },
 [0xA006]={
  category="lo",
  direction="l",
 },
 [0xA007]={
  category="lo",
  direction="l",
 },
 [0xA008]={
  category="lo",
  direction="l",
 },
 [0xA009]={
  category="lo",
  direction="l",
 },
 [0xA00A]={
  category="lo",
  direction="l",
 },
 [0xA00B]={
  category="lo",
  direction="l",
 },
 [0xA00C]={
  category="lo",
  direction="l",
 },
 [0xA00D]={
  category="lo",
  direction="l",
 },
 [0xA00E]={
  category="lo",
  direction="l",
 },
 [0xA00F]={
  category="lo",
  direction="l",
 },
 [0xA010]={
  category="lo",
  direction="l",
 },
 [0xA011]={
  category="lo",
  direction="l",
 },
 [0xA012]={
  category="lo",
  direction="l",
 },
 [0xA013]={
  category="lo",
  direction="l",
 },
 [0xA014]={
  category="lo",
  direction="l",
 },
 [0xA015]={
  category="lm",
  direction="l",
 },
 [0xA016]={
  category="lo",
  direction="l",
 },
 [0xA017]={
  category="lo",
  direction="l",
 },
 [0xA018]={
  category="lo",
  direction="l",
 },
 [0xA019]={
  category="lo",
  direction="l",
 },
 [0xA01A]={
  category="lo",
  direction="l",
 },
 [0xA01B]={
  category="lo",
  direction="l",
 },
 [0xA01C]={
  category="lo",
  direction="l",
 },
 [0xA01D]={
  category="lo",
  direction="l",
 },
 [0xA01E]={
  category="lo",
  direction="l",
 },
 [0xA01F]={
  category="lo",
  direction="l",
 },
 [0xA020]={
  category="lo",
  direction="l",
 },
 [0xA021]={
  category="lo",
  direction="l",
 },
 [0xA022]={
  category="lo",
  direction="l",
 },
 [0xA023]={
  category="lo",
  direction="l",
 },
 [0xA024]={
  category="lo",
  direction="l",
 },
 [0xA025]={
  category="lo",
  direction="l",
 },
 [0xA026]={
  category="lo",
  direction="l",
 },
 [0xA027]={
  category="lo",
  direction="l",
 },
 [0xA028]={
  category="lo",
  direction="l",
 },
 [0xA029]={
  category="lo",
  direction="l",
 },
 [0xA02A]={
  category="lo",
  direction="l",
 },
 [0xA02B]={
  category="lo",
  direction="l",
 },
 [0xA02C]={
  category="lo",
  direction="l",
 },
 [0xA02D]={
  category="lo",
  direction="l",
 },
 [0xA02E]={
  category="lo",
  direction="l",
 },
 [0xA02F]={
  category="lo",
  direction="l",
 },
 [0xA030]={
  category="lo",
  direction="l",
 },
 [0xA031]={
  category="lo",
  direction="l",
 },
 [0xA032]={
  category="lo",
  direction="l",
 },
 [0xA033]={
  category="lo",
  direction="l",
 },
 [0xA034]={
  category="lo",
  direction="l",
 },
 [0xA035]={
  category="lo",
  direction="l",
 },
 [0xA036]={
  category="lo",
  direction="l",
 },
 [0xA037]={
  category="lo",
  direction="l",
 },
 [0xA038]={
  category="lo",
  direction="l",
 },
 [0xA039]={
  category="lo",
  direction="l",
 },
 [0xA03A]={
  category="lo",
  direction="l",
 },
 [0xA03B]={
  category="lo",
  direction="l",
 },
 [0xA03C]={
  category="lo",
  direction="l",
 },
 [0xA03D]={
  category="lo",
  direction="l",
 },
 [0xA03E]={
  category="lo",
  direction="l",
 },
 [0xA03F]={
  category="lo",
  direction="l",
 },
 [0xA040]={
  category="lo",
  direction="l",
 },
 [0xA041]={
  category="lo",
  direction="l",
 },
 [0xA042]={
  category="lo",
  direction="l",
 },
 [0xA043]={
  category="lo",
  direction="l",
 },
 [0xA044]={
  category="lo",
  direction="l",
 },
 [0xA045]={
  category="lo",
  direction="l",
 },
 [0xA046]={
  category="lo",
  direction="l",
 },
 [0xA047]={
  category="lo",
  direction="l",
 },
 [0xA048]={
  category="lo",
  direction="l",
 },
 [0xA049]={
  category="lo",
  direction="l",
 },
 [0xA04A]={
  category="lo",
  direction="l",
 },
 [0xA04B]={
  category="lo",
  direction="l",
 },
 [0xA04C]={
  category="lo",
  direction="l",
 },
 [0xA04D]={
  category="lo",
  direction="l",
 },
 [0xA04E]={
  category="lo",
  direction="l",
 },
 [0xA04F]={
  category="lo",
  direction="l",
 },
 [0xA050]={
  category="lo",
  direction="l",
 },
 [0xA051]={
  category="lo",
  direction="l",
 },
 [0xA052]={
  category="lo",
  direction="l",
 },
 [0xA053]={
  category="lo",
  direction="l",
 },
 [0xA054]={
  category="lo",
  direction="l",
 },
 [0xA055]={
  category="lo",
  direction="l",
 },
 [0xA056]={
  category="lo",
  direction="l",
 },
 [0xA057]={
  category="lo",
  direction="l",
 },
 [0xA058]={
  category="lo",
  direction="l",
 },
 [0xA059]={
  category="lo",
  direction="l",
 },
 [0xA05A]={
  category="lo",
  direction="l",
 },
 [0xA05B]={
  category="lo",
  direction="l",
 },
 [0xA05C]={
  category="lo",
  direction="l",
 },
 [0xA05D]={
  category="lo",
  direction="l",
 },
 [0xA05E]={
  category="lo",
  direction="l",
 },
 [0xA05F]={
  category="lo",
  direction="l",
 },
 [0xA060]={
  category="lo",
  direction="l",
 },
 [0xA061]={
  category="lo",
  direction="l",
 },
 [0xA062]={
  category="lo",
  direction="l",
 },
 [0xA063]={
  category="lo",
  direction="l",
 },
 [0xA064]={
  category="lo",
  direction="l",
 },
 [0xA065]={
  category="lo",
  direction="l",
 },
 [0xA066]={
  category="lo",
  direction="l",
 },
 [0xA067]={
  category="lo",
  direction="l",
 },
 [0xA068]={
  category="lo",
  direction="l",
 },
 [0xA069]={
  category="lo",
  direction="l",
 },
 [0xA06A]={
  category="lo",
  direction="l",
 },
 [0xA06B]={
  category="lo",
  direction="l",
 },
 [0xA06C]={
  category="lo",
  direction="l",
 },
 [0xA06D]={
  category="lo",
  direction="l",
 },
 [0xA06E]={
  category="lo",
  direction="l",
 },
 [0xA06F]={
  category="lo",
  direction="l",
 },
 [0xA070]={
  category="lo",
  direction="l",
 },
 [0xA071]={
  category="lo",
  direction="l",
 },
 [0xA072]={
  category="lo",
  direction="l",
 },
 [0xA073]={
  category="lo",
  direction="l",
 },
 [0xA074]={
  category="lo",
  direction="l",
 },
 [0xA075]={
  category="lo",
  direction="l",
 },
 [0xA076]={
  category="lo",
  direction="l",
 },
 [0xA077]={
  category="lo",
  direction="l",
 },
 [0xA078]={
  category="lo",
  direction="l",
 },
 [0xA079]={
  category="lo",
  direction="l",
 },
 [0xA07A]={
  category="lo",
  direction="l",
 },
 [0xA07B]={
  category="lo",
  direction="l",
 },
 [0xA07C]={
  category="lo",
  direction="l",
 },
 [0xA07D]={
  category="lo",
  direction="l",
 },
 [0xA07E]={
  category="lo",
  direction="l",
 },
 [0xA07F]={
  category="lo",
  direction="l",
 },
 [0xA080]={
  category="lo",
  direction="l",
 },
 [0xA081]={
  category="lo",
  direction="l",
 },
 [0xA082]={
  category="lo",
  direction="l",
 },
 [0xA083]={
  category="lo",
  direction="l",
 },
 [0xA084]={
  category="lo",
  direction="l",
 },
 [0xA085]={
  category="lo",
  direction="l",
 },
 [0xA086]={
  category="lo",
  direction="l",
 },
 [0xA087]={
  category="lo",
  direction="l",
 },
 [0xA088]={
  category="lo",
  direction="l",
 },
 [0xA089]={
  category="lo",
  direction="l",
 },
 [0xA08A]={
  category="lo",
  direction="l",
 },
 [0xA08B]={
  category="lo",
  direction="l",
 },
 [0xA08C]={
  category="lo",
  direction="l",
 },
 [0xA08D]={
  category="lo",
  direction="l",
 },
 [0xA08E]={
  category="lo",
  direction="l",
 },
 [0xA08F]={
  category="lo",
  direction="l",
 },
 [0xA090]={
  category="lo",
  direction="l",
 },
 [0xA091]={
  category="lo",
  direction="l",
 },
 [0xA092]={
  category="lo",
  direction="l",
 },
 [0xA093]={
  category="lo",
  direction="l",
 },
 [0xA094]={
  category="lo",
  direction="l",
 },
 [0xA095]={
  category="lo",
  direction="l",
 },
 [0xA096]={
  category="lo",
  direction="l",
 },
 [0xA097]={
  category="lo",
  direction="l",
 },
 [0xA098]={
  category="lo",
  direction="l",
 },
 [0xA099]={
  category="lo",
  direction="l",
 },
 [0xA09A]={
  category="lo",
  direction="l",
 },
 [0xA09B]={
  category="lo",
  direction="l",
 },
 [0xA09C]={
  category="lo",
  direction="l",
 },
 [0xA09D]={
  category="lo",
  direction="l",
 },
 [0xA09E]={
  category="lo",
  direction="l",
 },
 [0xA09F]={
  category="lo",
  direction="l",
 },
 [0xA0A0]={
  category="lo",
  direction="l",
 },
 [0xA0A1]={
  category="lo",
  direction="l",
 },
 [0xA0A2]={
  category="lo",
  direction="l",
 },
 [0xA0A3]={
  category="lo",
  direction="l",
 },
 [0xA0A4]={
  category="lo",
  direction="l",
 },
 [0xA0A5]={
  category="lo",
  direction="l",
 },
 [0xA0A6]={
  category="lo",
  direction="l",
 },
 [0xA0A7]={
  category="lo",
  direction="l",
 },
 [0xA0A8]={
  category="lo",
  direction="l",
 },
 [0xA0A9]={
  category="lo",
  direction="l",
 },
 [0xA0AA]={
  category="lo",
  direction="l",
 },
 [0xA0AB]={
  category="lo",
  direction="l",
 },
 [0xA0AC]={
  category="lo",
  direction="l",
 },
 [0xA0AD]={
  category="lo",
  direction="l",
 },
 [0xA0AE]={
  category="lo",
  direction="l",
 },
 [0xA0AF]={
  category="lo",
  direction="l",
 },
 [0xA0B0]={
  category="lo",
  direction="l",
 },
 [0xA0B1]={
  category="lo",
  direction="l",
 },
 [0xA0B2]={
  category="lo",
  direction="l",
 },
 [0xA0B3]={
  category="lo",
  direction="l",
 },
 [0xA0B4]={
  category="lo",
  direction="l",
 },
 [0xA0B5]={
  category="lo",
  direction="l",
 },
 [0xA0B6]={
  category="lo",
  direction="l",
 },
 [0xA0B7]={
  category="lo",
  direction="l",
 },
 [0xA0B8]={
  category="lo",
  direction="l",
 },
 [0xA0B9]={
  category="lo",
  direction="l",
 },
 [0xA0BA]={
  category="lo",
  direction="l",
 },
 [0xA0BB]={
  category="lo",
  direction="l",
 },
 [0xA0BC]={
  category="lo",
  direction="l",
 },
 [0xA0BD]={
  category="lo",
  direction="l",
 },
 [0xA0BE]={
  category="lo",
  direction="l",
 },
 [0xA0BF]={
  category="lo",
  direction="l",
 },
 [0xA0C0]={
  category="lo",
  direction="l",
 },
 [0xA0C1]={
  category="lo",
  direction="l",
 },
 [0xA0C2]={
  category="lo",
  direction="l",
 },
 [0xA0C3]={
  category="lo",
  direction="l",
 },
 [0xA0C4]={
  category="lo",
  direction="l",
 },
 [0xA0C5]={
  category="lo",
  direction="l",
 },
 [0xA0C6]={
  category="lo",
  direction="l",
 },
 [0xA0C7]={
  category="lo",
  direction="l",
 },
 [0xA0C8]={
  category="lo",
  direction="l",
 },
 [0xA0C9]={
  category="lo",
  direction="l",
 },
 [0xA0CA]={
  category="lo",
  direction="l",
 },
 [0xA0CB]={
  category="lo",
  direction="l",
 },
 [0xA0CC]={
  category="lo",
  direction="l",
 },
 [0xA0CD]={
  category="lo",
  direction="l",
 },
 [0xA0CE]={
  category="lo",
  direction="l",
 },
 [0xA0CF]={
  category="lo",
  direction="l",
 },
 [0xA0D0]={
  category="lo",
  direction="l",
 },
 [0xA0D1]={
  category="lo",
  direction="l",
 },
 [0xA0D2]={
  category="lo",
  direction="l",
 },
 [0xA0D3]={
  category="lo",
  direction="l",
 },
 [0xA0D4]={
  category="lo",
  direction="l",
 },
 [0xA0D5]={
  category="lo",
  direction="l",
 },
 [0xA0D6]={
  category="lo",
  direction="l",
 },
 [0xA0D7]={
  category="lo",
  direction="l",
 },
 [0xA0D8]={
  category="lo",
  direction="l",
 },
 [0xA0D9]={
  category="lo",
  direction="l",
 },
 [0xA0DA]={
  category="lo",
  direction="l",
 },
 [0xA0DB]={
  category="lo",
  direction="l",
 },
 [0xA0DC]={
  category="lo",
  direction="l",
 },
 [0xA0DD]={
  category="lo",
  direction="l",
 },
 [0xA0DE]={
  category="lo",
  direction="l",
 },
 [0xA0DF]={
  category="lo",
  direction="l",
 },
 [0xA0E0]={
  category="lo",
  direction="l",
 },
 [0xA0E1]={
  category="lo",
  direction="l",
 },
 [0xA0E2]={
  category="lo",
  direction="l",
 },
 [0xA0E3]={
  category="lo",
  direction="l",
 },
 [0xA0E4]={
  category="lo",
  direction="l",
 },
 [0xA0E5]={
  category="lo",
  direction="l",
 },
 [0xA0E6]={
  category="lo",
  direction="l",
 },
 [0xA0E7]={
  category="lo",
  direction="l",
 },
 [0xA0E8]={
  category="lo",
  direction="l",
 },
 [0xA0E9]={
  category="lo",
  direction="l",
 },
 [0xA0EA]={
  category="lo",
  direction="l",
 },
 [0xA0EB]={
  category="lo",
  direction="l",
 },
 [0xA0EC]={
  category="lo",
  direction="l",
 },
 [0xA0ED]={
  category="lo",
  direction="l",
 },
 [0xA0EE]={
  category="lo",
  direction="l",
 },
 [0xA0EF]={
  category="lo",
  direction="l",
 },
 [0xA0F0]={
  category="lo",
  direction="l",
 },
 [0xA0F1]={
  category="lo",
  direction="l",
 },
 [0xA0F2]={
  category="lo",
  direction="l",
 },
 [0xA0F3]={
  category="lo",
  direction="l",
 },
 [0xA0F4]={
  category="lo",
  direction="l",
 },
 [0xA0F5]={
  category="lo",
  direction="l",
 },
 [0xA0F6]={
  category="lo",
  direction="l",
 },
 [0xA0F7]={
  category="lo",
  direction="l",
 },
 [0xA0F8]={
  category="lo",
  direction="l",
 },
 [0xA0F9]={
  category="lo",
  direction="l",
 },
 [0xA0FA]={
  category="lo",
  direction="l",
 },
 [0xA0FB]={
  category="lo",
  direction="l",
 },
 [0xA0FC]={
  category="lo",
  direction="l",
 },
 [0xA0FD]={
  category="lo",
  direction="l",
 },
 [0xA0FE]={
  category="lo",
  direction="l",
 },
 [0xA0FF]={
  category="lo",
  direction="l",
 },
 [0xA100]={
  category="lo",
  direction="l",
 },
 [0xA101]={
  category="lo",
  direction="l",
 },
 [0xA102]={
  category="lo",
  direction="l",
 },
 [0xA103]={
  category="lo",
  direction="l",
 },
 [0xA104]={
  category="lo",
  direction="l",
 },
 [0xA105]={
  category="lo",
  direction="l",
 },
 [0xA106]={
  category="lo",
  direction="l",
 },
 [0xA107]={
  category="lo",
  direction="l",
 },
 [0xA108]={
  category="lo",
  direction="l",
 },
 [0xA109]={
  category="lo",
  direction="l",
 },
 [0xA10A]={
  category="lo",
  direction="l",
 },
 [0xA10B]={
  category="lo",
  direction="l",
 },
 [0xA10C]={
  category="lo",
  direction="l",
 },
 [0xA10D]={
  category="lo",
  direction="l",
 },
 [0xA10E]={
  category="lo",
  direction="l",
 },
 [0xA10F]={
  category="lo",
  direction="l",
 },
 [0xA110]={
  category="lo",
  direction="l",
 },
 [0xA111]={
  category="lo",
  direction="l",
 },
 [0xA112]={
  category="lo",
  direction="l",
 },
 [0xA113]={
  category="lo",
  direction="l",
 },
 [0xA114]={
  category="lo",
  direction="l",
 },
 [0xA115]={
  category="lo",
  direction="l",
 },
 [0xA116]={
  category="lo",
  direction="l",
 },
 [0xA117]={
  category="lo",
  direction="l",
 },
 [0xA118]={
  category="lo",
  direction="l",
 },
 [0xA119]={
  category="lo",
  direction="l",
 },
 [0xA11A]={
  category="lo",
  direction="l",
 },
 [0xA11B]={
  category="lo",
  direction="l",
 },
 [0xA11C]={
  category="lo",
  direction="l",
 },
 [0xA11D]={
  category="lo",
  direction="l",
 },
 [0xA11E]={
  category="lo",
  direction="l",
 },
 [0xA11F]={
  category="lo",
  direction="l",
 },
 [0xA120]={
  category="lo",
  direction="l",
 },
 [0xA121]={
  category="lo",
  direction="l",
 },
 [0xA122]={
  category="lo",
  direction="l",
 },
 [0xA123]={
  category="lo",
  direction="l",
 },
 [0xA124]={
  category="lo",
  direction="l",
 },
 [0xA125]={
  category="lo",
  direction="l",
 },
 [0xA126]={
  category="lo",
  direction="l",
 },
 [0xA127]={
  category="lo",
  direction="l",
 },
 [0xA128]={
  category="lo",
  direction="l",
 },
 [0xA129]={
  category="lo",
  direction="l",
 },
 [0xA12A]={
  category="lo",
  direction="l",
 },
 [0xA12B]={
  category="lo",
  direction="l",
 },
 [0xA12C]={
  category="lo",
  direction="l",
 },
 [0xA12D]={
  category="lo",
  direction="l",
 },
 [0xA12E]={
  category="lo",
  direction="l",
 },
 [0xA12F]={
  category="lo",
  direction="l",
 },
 [0xA130]={
  category="lo",
  direction="l",
 },
 [0xA131]={
  category="lo",
  direction="l",
 },
 [0xA132]={
  category="lo",
  direction="l",
 },
 [0xA133]={
  category="lo",
  direction="l",
 },
 [0xA134]={
  category="lo",
  direction="l",
 },
 [0xA135]={
  category="lo",
  direction="l",
 },
 [0xA136]={
  category="lo",
  direction="l",
 },
 [0xA137]={
  category="lo",
  direction="l",
 },
 [0xA138]={
  category="lo",
  direction="l",
 },
 [0xA139]={
  category="lo",
  direction="l",
 },
 [0xA13A]={
  category="lo",
  direction="l",
 },
 [0xA13B]={
  category="lo",
  direction="l",
 },
 [0xA13C]={
  category="lo",
  direction="l",
 },
 [0xA13D]={
  category="lo",
  direction="l",
 },
 [0xA13E]={
  category="lo",
  direction="l",
 },
 [0xA13F]={
  category="lo",
  direction="l",
 },
 [0xA140]={
  category="lo",
  direction="l",
 },
 [0xA141]={
  category="lo",
  direction="l",
 },
 [0xA142]={
  category="lo",
  direction="l",
 },
 [0xA143]={
  category="lo",
  direction="l",
 },
 [0xA144]={
  category="lo",
  direction="l",
 },
 [0xA145]={
  category="lo",
  direction="l",
 },
 [0xA146]={
  category="lo",
  direction="l",
 },
 [0xA147]={
  category="lo",
  direction="l",
 },
 [0xA148]={
  category="lo",
  direction="l",
 },
 [0xA149]={
  category="lo",
  direction="l",
 },
 [0xA14A]={
  category="lo",
  direction="l",
 },
 [0xA14B]={
  category="lo",
  direction="l",
 },
 [0xA14C]={
  category="lo",
  direction="l",
 },
 [0xA14D]={
  category="lo",
  direction="l",
 },
 [0xA14E]={
  category="lo",
  direction="l",
 },
 [0xA14F]={
  category="lo",
  direction="l",
 },
 [0xA150]={
  category="lo",
  direction="l",
 },
 [0xA151]={
  category="lo",
  direction="l",
 },
 [0xA152]={
  category="lo",
  direction="l",
 },
 [0xA153]={
  category="lo",
  direction="l",
 },
 [0xA154]={
  category="lo",
  direction="l",
 },
 [0xA155]={
  category="lo",
  direction="l",
 },
 [0xA156]={
  category="lo",
  direction="l",
 },
 [0xA157]={
  category="lo",
  direction="l",
 },
 [0xA158]={
  category="lo",
  direction="l",
 },
 [0xA159]={
  category="lo",
  direction="l",
 },
 [0xA15A]={
  category="lo",
  direction="l",
 },
 [0xA15B]={
  category="lo",
  direction="l",
 },
 [0xA15C]={
  category="lo",
  direction="l",
 },
 [0xA15D]={
  category="lo",
  direction="l",
 },
 [0xA15E]={
  category="lo",
  direction="l",
 },
 [0xA15F]={
  category="lo",
  direction="l",
 },
 [0xA160]={
  category="lo",
  direction="l",
 },
 [0xA161]={
  category="lo",
  direction="l",
 },
 [0xA162]={
  category="lo",
  direction="l",
 },
 [0xA163]={
  category="lo",
  direction="l",
 },
 [0xA164]={
  category="lo",
  direction="l",
 },
 [0xA165]={
  category="lo",
  direction="l",
 },
 [0xA166]={
  category="lo",
  direction="l",
 },
 [0xA167]={
  category="lo",
  direction="l",
 },
 [0xA168]={
  category="lo",
  direction="l",
 },
 [0xA169]={
  category="lo",
  direction="l",
 },
 [0xA16A]={
  category="lo",
  direction="l",
 },
 [0xA16B]={
  category="lo",
  direction="l",
 },
 [0xA16C]={
  category="lo",
  direction="l",
 },
 [0xA16D]={
  category="lo",
  direction="l",
 },
 [0xA16E]={
  category="lo",
  direction="l",
 },
 [0xA16F]={
  category="lo",
  direction="l",
 },
 [0xA170]={
  category="lo",
  direction="l",
 },
 [0xA171]={
  category="lo",
  direction="l",
 },
 [0xA172]={
  category="lo",
  direction="l",
 },
 [0xA173]={
  category="lo",
  direction="l",
 },
 [0xA174]={
  category="lo",
  direction="l",
 },
 [0xA175]={
  category="lo",
  direction="l",
 },
 [0xA176]={
  category="lo",
  direction="l",
 },
 [0xA177]={
  category="lo",
  direction="l",
 },
 [0xA178]={
  category="lo",
  direction="l",
 },
 [0xA179]={
  category="lo",
  direction="l",
 },
 [0xA17A]={
  category="lo",
  direction="l",
 },
 [0xA17B]={
  category="lo",
  direction="l",
 },
 [0xA17C]={
  category="lo",
  direction="l",
 },
 [0xA17D]={
  category="lo",
  direction="l",
 },
 [0xA17E]={
  category="lo",
  direction="l",
 },
 [0xA17F]={
  category="lo",
  direction="l",
 },
 [0xA180]={
  category="lo",
  direction="l",
 },
 [0xA181]={
  category="lo",
  direction="l",
 },
 [0xA182]={
  category="lo",
  direction="l",
 },
 [0xA183]={
  category="lo",
  direction="l",
 },
 [0xA184]={
  category="lo",
  direction="l",
 },
 [0xA185]={
  category="lo",
  direction="l",
 },
 [0xA186]={
  category="lo",
  direction="l",
 },
 [0xA187]={
  category="lo",
  direction="l",
 },
 [0xA188]={
  category="lo",
  direction="l",
 },
 [0xA189]={
  category="lo",
  direction="l",
 },
 [0xA18A]={
  category="lo",
  direction="l",
 },
 [0xA18B]={
  category="lo",
  direction="l",
 },
 [0xA18C]={
  category="lo",
  direction="l",
 },
 [0xA18D]={
  category="lo",
  direction="l",
 },
 [0xA18E]={
  category="lo",
  direction="l",
 },
 [0xA18F]={
  category="lo",
  direction="l",
 },
 [0xA190]={
  category="lo",
  direction="l",
 },
 [0xA191]={
  category="lo",
  direction="l",
 },
 [0xA192]={
  category="lo",
  direction="l",
 },
 [0xA193]={
  category="lo",
  direction="l",
 },
 [0xA194]={
  category="lo",
  direction="l",
 },
 [0xA195]={
  category="lo",
  direction="l",
 },
 [0xA196]={
  category="lo",
  direction="l",
 },
 [0xA197]={
  category="lo",
  direction="l",
 },
 [0xA198]={
  category="lo",
  direction="l",
 },
 [0xA199]={
  category="lo",
  direction="l",
 },
 [0xA19A]={
  category="lo",
  direction="l",
 },
 [0xA19B]={
  category="lo",
  direction="l",
 },
 [0xA19C]={
  category="lo",
  direction="l",
 },
 [0xA19D]={
  category="lo",
  direction="l",
 },
 [0xA19E]={
  category="lo",
  direction="l",
 },
 [0xA19F]={
  category="lo",
  direction="l",
 },
 [0xA1A0]={
  category="lo",
  direction="l",
 },
 [0xA1A1]={
  category="lo",
  direction="l",
 },
 [0xA1A2]={
  category="lo",
  direction="l",
 },
 [0xA1A3]={
  category="lo",
  direction="l",
 },
 [0xA1A4]={
  category="lo",
  direction="l",
 },
 [0xA1A5]={
  category="lo",
  direction="l",
 },
 [0xA1A6]={
  category="lo",
  direction="l",
 },
 [0xA1A7]={
  category="lo",
  direction="l",
 },
 [0xA1A8]={
  category="lo",
  direction="l",
 },
 [0xA1A9]={
  category="lo",
  direction="l",
 },
 [0xA1AA]={
  category="lo",
  direction="l",
 },
 [0xA1AB]={
  category="lo",
  direction="l",
 },
 [0xA1AC]={
  category="lo",
  direction="l",
 },
 [0xA1AD]={
  category="lo",
  direction="l",
 },
 [0xA1AE]={
  category="lo",
  direction="l",
 },
 [0xA1AF]={
  category="lo",
  direction="l",
 },
 [0xA1B0]={
  category="lo",
  direction="l",
 },
 [0xA1B1]={
  category="lo",
  direction="l",
 },
 [0xA1B2]={
  category="lo",
  direction="l",
 },
 [0xA1B3]={
  category="lo",
  direction="l",
 },
 [0xA1B4]={
  category="lo",
  direction="l",
 },
 [0xA1B5]={
  category="lo",
  direction="l",
 },
 [0xA1B6]={
  category="lo",
  direction="l",
 },
 [0xA1B7]={
  category="lo",
  direction="l",
 },
 [0xA1B8]={
  category="lo",
  direction="l",
 },
 [0xA1B9]={
  category="lo",
  direction="l",
 },
 [0xA1BA]={
  category="lo",
  direction="l",
 },
 [0xA1BB]={
  category="lo",
  direction="l",
 },
 [0xA1BC]={
  category="lo",
  direction="l",
 },
 [0xA1BD]={
  category="lo",
  direction="l",
 },
 [0xA1BE]={
  category="lo",
  direction="l",
 },
 [0xA1BF]={
  category="lo",
  direction="l",
 },
 [0xA1C0]={
  category="lo",
  direction="l",
 },
 [0xA1C1]={
  category="lo",
  direction="l",
 },
 [0xA1C2]={
  category="lo",
  direction="l",
 },
 [0xA1C3]={
  category="lo",
  direction="l",
 },
 [0xA1C4]={
  category="lo",
  direction="l",
 },
 [0xA1C5]={
  category="lo",
  direction="l",
 },
 [0xA1C6]={
  category="lo",
  direction="l",
 },
 [0xA1C7]={
  category="lo",
  direction="l",
 },
 [0xA1C8]={
  category="lo",
  direction="l",
 },
 [0xA1C9]={
  category="lo",
  direction="l",
 },
 [0xA1CA]={
  category="lo",
  direction="l",
 },
 [0xA1CB]={
  category="lo",
  direction="l",
 },
 [0xA1CC]={
  category="lo",
  direction="l",
 },
 [0xA1CD]={
  category="lo",
  direction="l",
 },
 [0xA1CE]={
  category="lo",
  direction="l",
 },
 [0xA1CF]={
  category="lo",
  direction="l",
 },
 [0xA1D0]={
  category="lo",
  direction="l",
 },
 [0xA1D1]={
  category="lo",
  direction="l",
 },
 [0xA1D2]={
  category="lo",
  direction="l",
 },
 [0xA1D3]={
  category="lo",
  direction="l",
 },
 [0xA1D4]={
  category="lo",
  direction="l",
 },
 [0xA1D5]={
  category="lo",
  direction="l",
 },
 [0xA1D6]={
  category="lo",
  direction="l",
 },
 [0xA1D7]={
  category="lo",
  direction="l",
 },
 [0xA1D8]={
  category="lo",
  direction="l",
 },
 [0xA1D9]={
  category="lo",
  direction="l",
 },
 [0xA1DA]={
  category="lo",
  direction="l",
 },
 [0xA1DB]={
  category="lo",
  direction="l",
 },
 [0xA1DC]={
  category="lo",
  direction="l",
 },
 [0xA1DD]={
  category="lo",
  direction="l",
 },
 [0xA1DE]={
  category="lo",
  direction="l",
 },
 [0xA1DF]={
  category="lo",
  direction="l",
 },
 [0xA1E0]={
  category="lo",
  direction="l",
 },
 [0xA1E1]={
  category="lo",
  direction="l",
 },
 [0xA1E2]={
  category="lo",
  direction="l",
 },
 [0xA1E3]={
  category="lo",
  direction="l",
 },
 [0xA1E4]={
  category="lo",
  direction="l",
 },
 [0xA1E5]={
  category="lo",
  direction="l",
 },
 [0xA1E6]={
  category="lo",
  direction="l",
 },
 [0xA1E7]={
  category="lo",
  direction="l",
 },
 [0xA1E8]={
  category="lo",
  direction="l",
 },
 [0xA1E9]={
  category="lo",
  direction="l",
 },
 [0xA1EA]={
  category="lo",
  direction="l",
 },
 [0xA1EB]={
  category="lo",
  direction="l",
 },
 [0xA1EC]={
  category="lo",
  direction="l",
 },
 [0xA1ED]={
  category="lo",
  direction="l",
 },
 [0xA1EE]={
  category="lo",
  direction="l",
 },
 [0xA1EF]={
  category="lo",
  direction="l",
 },
 [0xA1F0]={
  category="lo",
  direction="l",
 },
 [0xA1F1]={
  category="lo",
  direction="l",
 },
 [0xA1F2]={
  category="lo",
  direction="l",
 },
 [0xA1F3]={
  category="lo",
  direction="l",
 },
 [0xA1F4]={
  category="lo",
  direction="l",
 },
 [0xA1F5]={
  category="lo",
  direction="l",
 },
 [0xA1F6]={
  category="lo",
  direction="l",
 },
 [0xA1F7]={
  category="lo",
  direction="l",
 },
 [0xA1F8]={
  category="lo",
  direction="l",
 },
 [0xA1F9]={
  category="lo",
  direction="l",
 },
 [0xA1FA]={
  category="lo",
  direction="l",
 },
 [0xA1FB]={
  category="lo",
  direction="l",
 },
 [0xA1FC]={
  category="lo",
  direction="l",
 },
 [0xA1FD]={
  category="lo",
  direction="l",
 },
 [0xA1FE]={
  category="lo",
  direction="l",
 },
 [0xA1FF]={
  category="lo",
  direction="l",
 },
 [0xA200]={
  category="lo",
  direction="l",
 },
 [0xA201]={
  category="lo",
  direction="l",
 },
 [0xA202]={
  category="lo",
  direction="l",
 },
 [0xA203]={
  category="lo",
  direction="l",
 },
 [0xA204]={
  category="lo",
  direction="l",
 },
 [0xA205]={
  category="lo",
  direction="l",
 },
 [0xA206]={
  category="lo",
  direction="l",
 },
 [0xA207]={
  category="lo",
  direction="l",
 },
 [0xA208]={
  category="lo",
  direction="l",
 },
 [0xA209]={
  category="lo",
  direction="l",
 },
 [0xA20A]={
  category="lo",
  direction="l",
 },
 [0xA20B]={
  category="lo",
  direction="l",
 },
 [0xA20C]={
  category="lo",
  direction="l",
 },
 [0xA20D]={
  category="lo",
  direction="l",
 },
 [0xA20E]={
  category="lo",
  direction="l",
 },
 [0xA20F]={
  category="lo",
  direction="l",
 },
 [0xA210]={
  category="lo",
  direction="l",
 },
 [0xA211]={
  category="lo",
  direction="l",
 },
 [0xA212]={
  category="lo",
  direction="l",
 },
 [0xA213]={
  category="lo",
  direction="l",
 },
 [0xA214]={
  category="lo",
  direction="l",
 },
 [0xA215]={
  category="lo",
  direction="l",
 },
 [0xA216]={
  category="lo",
  direction="l",
 },
 [0xA217]={
  category="lo",
  direction="l",
 },
 [0xA218]={
  category="lo",
  direction="l",
 },
 [0xA219]={
  category="lo",
  direction="l",
 },
 [0xA21A]={
  category="lo",
  direction="l",
 },
 [0xA21B]={
  category="lo",
  direction="l",
 },
 [0xA21C]={
  category="lo",
  direction="l",
 },
 [0xA21D]={
  category="lo",
  direction="l",
 },
 [0xA21E]={
  category="lo",
  direction="l",
 },
 [0xA21F]={
  category="lo",
  direction="l",
 },
 [0xA220]={
  category="lo",
  direction="l",
 },
 [0xA221]={
  category="lo",
  direction="l",
 },
 [0xA222]={
  category="lo",
  direction="l",
 },
 [0xA223]={
  category="lo",
  direction="l",
 },
 [0xA224]={
  category="lo",
  direction="l",
 },
 [0xA225]={
  category="lo",
  direction="l",
 },
 [0xA226]={
  category="lo",
  direction="l",
 },
 [0xA227]={
  category="lo",
  direction="l",
 },
 [0xA228]={
  category="lo",
  direction="l",
 },
 [0xA229]={
  category="lo",
  direction="l",
 },
 [0xA22A]={
  category="lo",
  direction="l",
 },
 [0xA22B]={
  category="lo",
  direction="l",
 },
 [0xA22C]={
  category="lo",
  direction="l",
 },
 [0xA22D]={
  category="lo",
  direction="l",
 },
 [0xA22E]={
  category="lo",
  direction="l",
 },
 [0xA22F]={
  category="lo",
  direction="l",
 },
 [0xA230]={
  category="lo",
  direction="l",
 },
 [0xA231]={
  category="lo",
  direction="l",
 },
 [0xA232]={
  category="lo",
  direction="l",
 },
 [0xA233]={
  category="lo",
  direction="l",
 },
 [0xA234]={
  category="lo",
  direction="l",
 },
 [0xA235]={
  category="lo",
  direction="l",
 },
 [0xA236]={
  category="lo",
  direction="l",
 },
 [0xA237]={
  category="lo",
  direction="l",
 },
 [0xA238]={
  category="lo",
  direction="l",
 },
 [0xA239]={
  category="lo",
  direction="l",
 },
 [0xA23A]={
  category="lo",
  direction="l",
 },
 [0xA23B]={
  category="lo",
  direction="l",
 },
 [0xA23C]={
  category="lo",
  direction="l",
 },
 [0xA23D]={
  category="lo",
  direction="l",
 },
 [0xA23E]={
  category="lo",
  direction="l",
 },
 [0xA23F]={
  category="lo",
  direction="l",
 },
 [0xA240]={
  category="lo",
  direction="l",
 },
 [0xA241]={
  category="lo",
  direction="l",
 },
 [0xA242]={
  category="lo",
  direction="l",
 },
 [0xA243]={
  category="lo",
  direction="l",
 },
 [0xA244]={
  category="lo",
  direction="l",
 },
 [0xA245]={
  category="lo",
  direction="l",
 },
 [0xA246]={
  category="lo",
  direction="l",
 },
 [0xA247]={
  category="lo",
  direction="l",
 },
 [0xA248]={
  category="lo",
  direction="l",
 },
 [0xA249]={
  category="lo",
  direction="l",
 },
 [0xA24A]={
  category="lo",
  direction="l",
 },
 [0xA24B]={
  category="lo",
  direction="l",
 },
 [0xA24C]={
  category="lo",
  direction="l",
 },
 [0xA24D]={
  category="lo",
  direction="l",
 },
 [0xA24E]={
  category="lo",
  direction="l",
 },
 [0xA24F]={
  category="lo",
  direction="l",
 },
 [0xA250]={
  category="lo",
  direction="l",
 },
 [0xA251]={
  category="lo",
  direction="l",
 },
 [0xA252]={
  category="lo",
  direction="l",
 },
 [0xA253]={
  category="lo",
  direction="l",
 },
 [0xA254]={
  category="lo",
  direction="l",
 },
 [0xA255]={
  category="lo",
  direction="l",
 },
 [0xA256]={
  category="lo",
  direction="l",
 },
 [0xA257]={
  category="lo",
  direction="l",
 },
 [0xA258]={
  category="lo",
  direction="l",
 },
 [0xA259]={
  category="lo",
  direction="l",
 },
 [0xA25A]={
  category="lo",
  direction="l",
 },
 [0xA25B]={
  category="lo",
  direction="l",
 },
 [0xA25C]={
  category="lo",
  direction="l",
 },
 [0xA25D]={
  category="lo",
  direction="l",
 },
 [0xA25E]={
  category="lo",
  direction="l",
 },
 [0xA25F]={
  category="lo",
  direction="l",
 },
 [0xA260]={
  category="lo",
  direction="l",
 },
 [0xA261]={
  category="lo",
  direction="l",
 },
 [0xA262]={
  category="lo",
  direction="l",
 },
 [0xA263]={
  category="lo",
  direction="l",
 },
 [0xA264]={
  category="lo",
  direction="l",
 },
 [0xA265]={
  category="lo",
  direction="l",
 },
 [0xA266]={
  category="lo",
  direction="l",
 },
 [0xA267]={
  category="lo",
  direction="l",
 },
 [0xA268]={
  category="lo",
  direction="l",
 },
 [0xA269]={
  category="lo",
  direction="l",
 },
 [0xA26A]={
  category="lo",
  direction="l",
 },
 [0xA26B]={
  category="lo",
  direction="l",
 },
 [0xA26C]={
  category="lo",
  direction="l",
 },
 [0xA26D]={
  category="lo",
  direction="l",
 },
 [0xA26E]={
  category="lo",
  direction="l",
 },
 [0xA26F]={
  category="lo",
  direction="l",
 },
 [0xA270]={
  category="lo",
  direction="l",
 },
 [0xA271]={
  category="lo",
  direction="l",
 },
 [0xA272]={
  category="lo",
  direction="l",
 },
 [0xA273]={
  category="lo",
  direction="l",
 },
 [0xA274]={
  category="lo",
  direction="l",
 },
 [0xA275]={
  category="lo",
  direction="l",
 },
 [0xA276]={
  category="lo",
  direction="l",
 },
 [0xA277]={
  category="lo",
  direction="l",
 },
 [0xA278]={
  category="lo",
  direction="l",
 },
 [0xA279]={
  category="lo",
  direction="l",
 },
 [0xA27A]={
  category="lo",
  direction="l",
 },
 [0xA27B]={
  category="lo",
  direction="l",
 },
 [0xA27C]={
  category="lo",
  direction="l",
 },
 [0xA27D]={
  category="lo",
  direction="l",
 },
 [0xA27E]={
  category="lo",
  direction="l",
 },
 [0xA27F]={
  category="lo",
  direction="l",
 },
 [0xA280]={
  category="lo",
  direction="l",
 },
 [0xA281]={
  category="lo",
  direction="l",
 },
 [0xA282]={
  category="lo",
  direction="l",
 },
 [0xA283]={
  category="lo",
  direction="l",
 },
 [0xA284]={
  category="lo",
  direction="l",
 },
 [0xA285]={
  category="lo",
  direction="l",
 },
 [0xA286]={
  category="lo",
  direction="l",
 },
 [0xA287]={
  category="lo",
  direction="l",
 },
 [0xA288]={
  category="lo",
  direction="l",
 },
 [0xA289]={
  category="lo",
  direction="l",
 },
 [0xA28A]={
  category="lo",
  direction="l",
 },
 [0xA28B]={
  category="lo",
  direction="l",
 },
 [0xA28C]={
  category="lo",
  direction="l",
 },
 [0xA28D]={
  category="lo",
  direction="l",
 },
 [0xA28E]={
  category="lo",
  direction="l",
 },
 [0xA28F]={
  category="lo",
  direction="l",
 },
 [0xA290]={
  category="lo",
  direction="l",
 },
 [0xA291]={
  category="lo",
  direction="l",
 },
 [0xA292]={
  category="lo",
  direction="l",
 },
 [0xA293]={
  category="lo",
  direction="l",
 },
 [0xA294]={
  category="lo",
  direction="l",
 },
 [0xA295]={
  category="lo",
  direction="l",
 },
 [0xA296]={
  category="lo",
  direction="l",
 },
 [0xA297]={
  category="lo",
  direction="l",
 },
 [0xA298]={
  category="lo",
  direction="l",
 },
 [0xA299]={
  category="lo",
  direction="l",
 },
 [0xA29A]={
  category="lo",
  direction="l",
 },
 [0xA29B]={
  category="lo",
  direction="l",
 },
 [0xA29C]={
  category="lo",
  direction="l",
 },
 [0xA29D]={
  category="lo",
  direction="l",
 },
 [0xA29E]={
  category="lo",
  direction="l",
 },
 [0xA29F]={
  category="lo",
  direction="l",
 },
 [0xA2A0]={
  category="lo",
  direction="l",
 },
 [0xA2A1]={
  category="lo",
  direction="l",
 },
 [0xA2A2]={
  category="lo",
  direction="l",
 },
 [0xA2A3]={
  category="lo",
  direction="l",
 },
 [0xA2A4]={
  category="lo",
  direction="l",
 },
 [0xA2A5]={
  category="lo",
  direction="l",
 },
 [0xA2A6]={
  category="lo",
  direction="l",
 },
 [0xA2A7]={
  category="lo",
  direction="l",
 },
 [0xA2A8]={
  category="lo",
  direction="l",
 },
 [0xA2A9]={
  category="lo",
  direction="l",
 },
 [0xA2AA]={
  category="lo",
  direction="l",
 },
 [0xA2AB]={
  category="lo",
  direction="l",
 },
 [0xA2AC]={
  category="lo",
  direction="l",
 },
 [0xA2AD]={
  category="lo",
  direction="l",
 },
 [0xA2AE]={
  category="lo",
  direction="l",
 },
 [0xA2AF]={
  category="lo",
  direction="l",
 },
 [0xA2B0]={
  category="lo",
  direction="l",
 },
 [0xA2B1]={
  category="lo",
  direction="l",
 },
 [0xA2B2]={
  category="lo",
  direction="l",
 },
 [0xA2B3]={
  category="lo",
  direction="l",
 },
 [0xA2B4]={
  category="lo",
  direction="l",
 },
 [0xA2B5]={
  category="lo",
  direction="l",
 },
 [0xA2B6]={
  category="lo",
  direction="l",
 },
 [0xA2B7]={
  category="lo",
  direction="l",
 },
 [0xA2B8]={
  category="lo",
  direction="l",
 },
 [0xA2B9]={
  category="lo",
  direction="l",
 },
 [0xA2BA]={
  category="lo",
  direction="l",
 },
 [0xA2BB]={
  category="lo",
  direction="l",
 },
 [0xA2BC]={
  category="lo",
  direction="l",
 },
 [0xA2BD]={
  category="lo",
  direction="l",
 },
 [0xA2BE]={
  category="lo",
  direction="l",
 },
 [0xA2BF]={
  category="lo",
  direction="l",
 },
 [0xA2C0]={
  category="lo",
  direction="l",
 },
 [0xA2C1]={
  category="lo",
  direction="l",
 },
 [0xA2C2]={
  category="lo",
  direction="l",
 },
 [0xA2C3]={
  category="lo",
  direction="l",
 },
 [0xA2C4]={
  category="lo",
  direction="l",
 },
 [0xA2C5]={
  category="lo",
  direction="l",
 },
 [0xA2C6]={
  category="lo",
  direction="l",
 },
 [0xA2C7]={
  category="lo",
  direction="l",
 },
 [0xA2C8]={
  category="lo",
  direction="l",
 },
 [0xA2C9]={
  category="lo",
  direction="l",
 },
 [0xA2CA]={
  category="lo",
  direction="l",
 },
 [0xA2CB]={
  category="lo",
  direction="l",
 },
 [0xA2CC]={
  category="lo",
  direction="l",
 },
 [0xA2CD]={
  category="lo",
  direction="l",
 },
 [0xA2CE]={
  category="lo",
  direction="l",
 },
 [0xA2CF]={
  category="lo",
  direction="l",
 },
 [0xA2D0]={
  category="lo",
  direction="l",
 },
 [0xA2D1]={
  category="lo",
  direction="l",
 },
 [0xA2D2]={
  category="lo",
  direction="l",
 },
 [0xA2D3]={
  category="lo",
  direction="l",
 },
 [0xA2D4]={
  category="lo",
  direction="l",
 },
 [0xA2D5]={
  category="lo",
  direction="l",
 },
 [0xA2D6]={
  category="lo",
  direction="l",
 },
 [0xA2D7]={
  category="lo",
  direction="l",
 },
 [0xA2D8]={
  category="lo",
  direction="l",
 },
 [0xA2D9]={
  category="lo",
  direction="l",
 },
 [0xA2DA]={
  category="lo",
  direction="l",
 },
 [0xA2DB]={
  category="lo",
  direction="l",
 },
 [0xA2DC]={
  category="lo",
  direction="l",
 },
 [0xA2DD]={
  category="lo",
  direction="l",
 },
 [0xA2DE]={
  category="lo",
  direction="l",
 },
 [0xA2DF]={
  category="lo",
  direction="l",
 },
 [0xA2E0]={
  category="lo",
  direction="l",
 },
 [0xA2E1]={
  category="lo",
  direction="l",
 },
 [0xA2E2]={
  category="lo",
  direction="l",
 },
 [0xA2E3]={
  category="lo",
  direction="l",
 },
 [0xA2E4]={
  category="lo",
  direction="l",
 },
 [0xA2E5]={
  category="lo",
  direction="l",
 },
 [0xA2E6]={
  category="lo",
  direction="l",
 },
 [0xA2E7]={
  category="lo",
  direction="l",
 },
 [0xA2E8]={
  category="lo",
  direction="l",
 },
 [0xA2E9]={
  category="lo",
  direction="l",
 },
 [0xA2EA]={
  category="lo",
  direction="l",
 },
 [0xA2EB]={
  category="lo",
  direction="l",
 },
 [0xA2EC]={
  category="lo",
  direction="l",
 },
 [0xA2ED]={
  category="lo",
  direction="l",
 },
 [0xA2EE]={
  category="lo",
  direction="l",
 },
 [0xA2EF]={
  category="lo",
  direction="l",
 },
 [0xA2F0]={
  category="lo",
  direction="l",
 },
 [0xA2F1]={
  category="lo",
  direction="l",
 },
 [0xA2F2]={
  category="lo",
  direction="l",
 },
 [0xA2F3]={
  category="lo",
  direction="l",
 },
 [0xA2F4]={
  category="lo",
  direction="l",
 },
 [0xA2F5]={
  category="lo",
  direction="l",
 },
 [0xA2F6]={
  category="lo",
  direction="l",
 },
 [0xA2F7]={
  category="lo",
  direction="l",
 },
 [0xA2F8]={
  category="lo",
  direction="l",
 },
 [0xA2F9]={
  category="lo",
  direction="l",
 },
 [0xA2FA]={
  category="lo",
  direction="l",
 },
 [0xA2FB]={
  category="lo",
  direction="l",
 },
 [0xA2FC]={
  category="lo",
  direction="l",
 },
 [0xA2FD]={
  category="lo",
  direction="l",
 },
 [0xA2FE]={
  category="lo",
  direction="l",
 },
 [0xA2FF]={
  category="lo",
  direction="l",
 },
 [0xA300]={
  category="lo",
  direction="l",
 },
 [0xA301]={
  category="lo",
  direction="l",
 },
 [0xA302]={
  category="lo",
  direction="l",
 },
 [0xA303]={
  category="lo",
  direction="l",
 },
 [0xA304]={
  category="lo",
  direction="l",
 },
 [0xA305]={
  category="lo",
  direction="l",
 },
 [0xA306]={
  category="lo",
  direction="l",
 },
 [0xA307]={
  category="lo",
  direction="l",
 },
 [0xA308]={
  category="lo",
  direction="l",
 },
 [0xA309]={
  category="lo",
  direction="l",
 },
 [0xA30A]={
  category="lo",
  direction="l",
 },
 [0xA30B]={
  category="lo",
  direction="l",
 },
 [0xA30C]={
  category="lo",
  direction="l",
 },
 [0xA30D]={
  category="lo",
  direction="l",
 },
 [0xA30E]={
  category="lo",
  direction="l",
 },
 [0xA30F]={
  category="lo",
  direction="l",
 },
 [0xA310]={
  category="lo",
  direction="l",
 },
 [0xA311]={
  category="lo",
  direction="l",
 },
 [0xA312]={
  category="lo",
  direction="l",
 },
 [0xA313]={
  category="lo",
  direction="l",
 },
 [0xA314]={
  category="lo",
  direction="l",
 },
 [0xA315]={
  category="lo",
  direction="l",
 },
 [0xA316]={
  category="lo",
  direction="l",
 },
 [0xA317]={
  category="lo",
  direction="l",
 },
 [0xA318]={
  category="lo",
  direction="l",
 },
 [0xA319]={
  category="lo",
  direction="l",
 },
 [0xA31A]={
  category="lo",
  direction="l",
 },
 [0xA31B]={
  category="lo",
  direction="l",
 },
 [0xA31C]={
  category="lo",
  direction="l",
 },
 [0xA31D]={
  category="lo",
  direction="l",
 },
 [0xA31E]={
  category="lo",
  direction="l",
 },
 [0xA31F]={
  category="lo",
  direction="l",
 },
 [0xA320]={
  category="lo",
  direction="l",
 },
 [0xA321]={
  category="lo",
  direction="l",
 },
 [0xA322]={
  category="lo",
  direction="l",
 },
 [0xA323]={
  category="lo",
  direction="l",
 },
 [0xA324]={
  category="lo",
  direction="l",
 },
 [0xA325]={
  category="lo",
  direction="l",
 },
 [0xA326]={
  category="lo",
  direction="l",
 },
 [0xA327]={
  category="lo",
  direction="l",
 },
 [0xA328]={
  category="lo",
  direction="l",
 },
 [0xA329]={
  category="lo",
  direction="l",
 },
 [0xA32A]={
  category="lo",
  direction="l",
 },
 [0xA32B]={
  category="lo",
  direction="l",
 },
 [0xA32C]={
  category="lo",
  direction="l",
 },
 [0xA32D]={
  category="lo",
  direction="l",
 },
 [0xA32E]={
  category="lo",
  direction="l",
 },
 [0xA32F]={
  category="lo",
  direction="l",
 },
 [0xA330]={
  category="lo",
  direction="l",
 },
 [0xA331]={
  category="lo",
  direction="l",
 },
 [0xA332]={
  category="lo",
  direction="l",
 },
 [0xA333]={
  category="lo",
  direction="l",
 },
 [0xA334]={
  category="lo",
  direction="l",
 },
 [0xA335]={
  category="lo",
  direction="l",
 },
 [0xA336]={
  category="lo",
  direction="l",
 },
 [0xA337]={
  category="lo",
  direction="l",
 },
 [0xA338]={
  category="lo",
  direction="l",
 },
 [0xA339]={
  category="lo",
  direction="l",
 },
 [0xA33A]={
  category="lo",
  direction="l",
 },
 [0xA33B]={
  category="lo",
  direction="l",
 },
 [0xA33C]={
  category="lo",
  direction="l",
 },
 [0xA33D]={
  category="lo",
  direction="l",
 },
 [0xA33E]={
  category="lo",
  direction="l",
 },
 [0xA33F]={
  category="lo",
  direction="l",
 },
 [0xA340]={
  category="lo",
  direction="l",
 },
 [0xA341]={
  category="lo",
  direction="l",
 },
 [0xA342]={
  category="lo",
  direction="l",
 },
 [0xA343]={
  category="lo",
  direction="l",
 },
 [0xA344]={
  category="lo",
  direction="l",
 },
 [0xA345]={
  category="lo",
  direction="l",
 },
 [0xA346]={
  category="lo",
  direction="l",
 },
 [0xA347]={
  category="lo",
  direction="l",
 },
 [0xA348]={
  category="lo",
  direction="l",
 },
 [0xA349]={
  category="lo",
  direction="l",
 },
 [0xA34A]={
  category="lo",
  direction="l",
 },
 [0xA34B]={
  category="lo",
  direction="l",
 },
 [0xA34C]={
  category="lo",
  direction="l",
 },
 [0xA34D]={
  category="lo",
  direction="l",
 },
 [0xA34E]={
  category="lo",
  direction="l",
 },
 [0xA34F]={
  category="lo",
  direction="l",
 },
 [0xA350]={
  category="lo",
  direction="l",
 },
 [0xA351]={
  category="lo",
  direction="l",
 },
 [0xA352]={
  category="lo",
  direction="l",
 },
 [0xA353]={
  category="lo",
  direction="l",
 },
 [0xA354]={
  category="lo",
  direction="l",
 },
 [0xA355]={
  category="lo",
  direction="l",
 },
 [0xA356]={
  category="lo",
  direction="l",
 },
 [0xA357]={
  category="lo",
  direction="l",
 },
 [0xA358]={
  category="lo",
  direction="l",
 },
 [0xA359]={
  category="lo",
  direction="l",
 },
 [0xA35A]={
  category="lo",
  direction="l",
 },
 [0xA35B]={
  category="lo",
  direction="l",
 },
 [0xA35C]={
  category="lo",
  direction="l",
 },
 [0xA35D]={
  category="lo",
  direction="l",
 },
 [0xA35E]={
  category="lo",
  direction="l",
 },
 [0xA35F]={
  category="lo",
  direction="l",
 },
 [0xA360]={
  category="lo",
  direction="l",
 },
 [0xA361]={
  category="lo",
  direction="l",
 },
 [0xA362]={
  category="lo",
  direction="l",
 },
 [0xA363]={
  category="lo",
  direction="l",
 },
 [0xA364]={
  category="lo",
  direction="l",
 },
 [0xA365]={
  category="lo",
  direction="l",
 },
 [0xA366]={
  category="lo",
  direction="l",
 },
 [0xA367]={
  category="lo",
  direction="l",
 },
 [0xA368]={
  category="lo",
  direction="l",
 },
 [0xA369]={
  category="lo",
  direction="l",
 },
 [0xA36A]={
  category="lo",
  direction="l",
 },
 [0xA36B]={
  category="lo",
  direction="l",
 },
 [0xA36C]={
  category="lo",
  direction="l",
 },
 [0xA36D]={
  category="lo",
  direction="l",
 },
 [0xA36E]={
  category="lo",
  direction="l",
 },
 [0xA36F]={
  category="lo",
  direction="l",
 },
 [0xA370]={
  category="lo",
  direction="l",
 },
 [0xA371]={
  category="lo",
  direction="l",
 },
 [0xA372]={
  category="lo",
  direction="l",
 },
 [0xA373]={
  category="lo",
  direction="l",
 },
 [0xA374]={
  category="lo",
  direction="l",
 },
 [0xA375]={
  category="lo",
  direction="l",
 },
 [0xA376]={
  category="lo",
  direction="l",
 },
 [0xA377]={
  category="lo",
  direction="l",
 },
 [0xA378]={
  category="lo",
  direction="l",
 },
 [0xA379]={
  category="lo",
  direction="l",
 },
 [0xA37A]={
  category="lo",
  direction="l",
 },
 [0xA37B]={
  category="lo",
  direction="l",
 },
 [0xA37C]={
  category="lo",
  direction="l",
 },
 [0xA37D]={
  category="lo",
  direction="l",
 },
 [0xA37E]={
  category="lo",
  direction="l",
 },
 [0xA37F]={
  category="lo",
  direction="l",
 },
 [0xA380]={
  category="lo",
  direction="l",
 },
 [0xA381]={
  category="lo",
  direction="l",
 },
 [0xA382]={
  category="lo",
  direction="l",
 },
 [0xA383]={
  category="lo",
  direction="l",
 },
 [0xA384]={
  category="lo",
  direction="l",
 },
 [0xA385]={
  category="lo",
  direction="l",
 },
 [0xA386]={
  category="lo",
  direction="l",
 },
 [0xA387]={
  category="lo",
  direction="l",
 },
 [0xA388]={
  category="lo",
  direction="l",
 },
 [0xA389]={
  category="lo",
  direction="l",
 },
 [0xA38A]={
  category="lo",
  direction="l",
 },
 [0xA38B]={
  category="lo",
  direction="l",
 },
 [0xA38C]={
  category="lo",
  direction="l",
 },
 [0xA38D]={
  category="lo",
  direction="l",
 },
 [0xA38E]={
  category="lo",
  direction="l",
 },
 [0xA38F]={
  category="lo",
  direction="l",
 },
 [0xA390]={
  category="lo",
  direction="l",
 },
 [0xA391]={
  category="lo",
  direction="l",
 },
 [0xA392]={
  category="lo",
  direction="l",
 },
 [0xA393]={
  category="lo",
  direction="l",
 },
 [0xA394]={
  category="lo",
  direction="l",
 },
 [0xA395]={
  category="lo",
  direction="l",
 },
 [0xA396]={
  category="lo",
  direction="l",
 },
 [0xA397]={
  category="lo",
  direction="l",
 },
 [0xA398]={
  category="lo",
  direction="l",
 },
 [0xA399]={
  category="lo",
  direction="l",
 },
 [0xA39A]={
  category="lo",
  direction="l",
 },
 [0xA39B]={
  category="lo",
  direction="l",
 },
 [0xA39C]={
  category="lo",
  direction="l",
 },
 [0xA39D]={
  category="lo",
  direction="l",
 },
 [0xA39E]={
  category="lo",
  direction="l",
 },
 [0xA39F]={
  category="lo",
  direction="l",
 },
 [0xA3A0]={
  category="lo",
  direction="l",
 },
 [0xA3A1]={
  category="lo",
  direction="l",
 },
 [0xA3A2]={
  category="lo",
  direction="l",
 },
 [0xA3A3]={
  category="lo",
  direction="l",
 },
 [0xA3A4]={
  category="lo",
  direction="l",
 },
 [0xA3A5]={
  category="lo",
  direction="l",
 },
 [0xA3A6]={
  category="lo",
  direction="l",
 },
 [0xA3A7]={
  category="lo",
  direction="l",
 },
 [0xA3A8]={
  category="lo",
  direction="l",
 },
 [0xA3A9]={
  category="lo",
  direction="l",
 },
 [0xA3AA]={
  category="lo",
  direction="l",
 },
 [0xA3AB]={
  category="lo",
  direction="l",
 },
 [0xA3AC]={
  category="lo",
  direction="l",
 },
 [0xA3AD]={
  category="lo",
  direction="l",
 },
 [0xA3AE]={
  category="lo",
  direction="l",
 },
 [0xA3AF]={
  category="lo",
  direction="l",
 },
 [0xA3B0]={
  category="lo",
  direction="l",
 },
 [0xA3B1]={
  category="lo",
  direction="l",
 },
 [0xA3B2]={
  category="lo",
  direction="l",
 },
 [0xA3B3]={
  category="lo",
  direction="l",
 },
 [0xA3B4]={
  category="lo",
  direction="l",
 },
 [0xA3B5]={
  category="lo",
  direction="l",
 },
 [0xA3B6]={
  category="lo",
  direction="l",
 },
 [0xA3B7]={
  category="lo",
  direction="l",
 },
 [0xA3B8]={
  category="lo",
  direction="l",
 },
 [0xA3B9]={
  category="lo",
  direction="l",
 },
 [0xA3BA]={
  category="lo",
  direction="l",
 },
 [0xA3BB]={
  category="lo",
  direction="l",
 },
 [0xA3BC]={
  category="lo",
  direction="l",
 },
 [0xA3BD]={
  category="lo",
  direction="l",
 },
 [0xA3BE]={
  category="lo",
  direction="l",
 },
 [0xA3BF]={
  category="lo",
  direction="l",
 },
 [0xA3C0]={
  category="lo",
  direction="l",
 },
 [0xA3C1]={
  category="lo",
  direction="l",
 },
 [0xA3C2]={
  category="lo",
  direction="l",
 },
 [0xA3C3]={
  category="lo",
  direction="l",
 },
 [0xA3C4]={
  category="lo",
  direction="l",
 },
 [0xA3C5]={
  category="lo",
  direction="l",
 },
 [0xA3C6]={
  category="lo",
  direction="l",
 },
 [0xA3C7]={
  category="lo",
  direction="l",
 },
 [0xA3C8]={
  category="lo",
  direction="l",
 },
 [0xA3C9]={
  category="lo",
  direction="l",
 },
 [0xA3CA]={
  category="lo",
  direction="l",
 },
 [0xA3CB]={
  category="lo",
  direction="l",
 },
 [0xA3CC]={
  category="lo",
  direction="l",
 },
 [0xA3CD]={
  category="lo",
  direction="l",
 },
 [0xA3CE]={
  category="lo",
  direction="l",
 },
 [0xA3CF]={
  category="lo",
  direction="l",
 },
 [0xA3D0]={
  category="lo",
  direction="l",
 },
 [0xA3D1]={
  category="lo",
  direction="l",
 },
 [0xA3D2]={
  category="lo",
  direction="l",
 },
 [0xA3D3]={
  category="lo",
  direction="l",
 },
 [0xA3D4]={
  category="lo",
  direction="l",
 },
 [0xA3D5]={
  category="lo",
  direction="l",
 },
 [0xA3D6]={
  category="lo",
  direction="l",
 },
 [0xA3D7]={
  category="lo",
  direction="l",
 },
 [0xA3D8]={
  category="lo",
  direction="l",
 },
 [0xA3D9]={
  category="lo",
  direction="l",
 },
 [0xA3DA]={
  category="lo",
  direction="l",
 },
 [0xA3DB]={
  category="lo",
  direction="l",
 },
 [0xA3DC]={
  category="lo",
  direction="l",
 },
 [0xA3DD]={
  category="lo",
  direction="l",
 },
 [0xA3DE]={
  category="lo",
  direction="l",
 },
 [0xA3DF]={
  category="lo",
  direction="l",
 },
 [0xA3E0]={
  category="lo",
  direction="l",
 },
 [0xA3E1]={
  category="lo",
  direction="l",
 },
 [0xA3E2]={
  category="lo",
  direction="l",
 },
 [0xA3E3]={
  category="lo",
  direction="l",
 },
 [0xA3E4]={
  category="lo",
  direction="l",
 },
 [0xA3E5]={
  category="lo",
  direction="l",
 },
 [0xA3E6]={
  category="lo",
  direction="l",
 },
 [0xA3E7]={
  category="lo",
  direction="l",
 },
 [0xA3E8]={
  category="lo",
  direction="l",
 },
 [0xA3E9]={
  category="lo",
  direction="l",
 },
 [0xA3EA]={
  category="lo",
  direction="l",
 },
 [0xA3EB]={
  category="lo",
  direction="l",
 },
 [0xA3EC]={
  category="lo",
  direction="l",
 },
 [0xA3ED]={
  category="lo",
  direction="l",
 },
 [0xA3EE]={
  category="lo",
  direction="l",
 },
 [0xA3EF]={
  category="lo",
  direction="l",
 },
 [0xA3F0]={
  category="lo",
  direction="l",
 },
 [0xA3F1]={
  category="lo",
  direction="l",
 },
 [0xA3F2]={
  category="lo",
  direction="l",
 },
 [0xA3F3]={
  category="lo",
  direction="l",
 },
 [0xA3F4]={
  category="lo",
  direction="l",
 },
 [0xA3F5]={
  category="lo",
  direction="l",
 },
 [0xA3F6]={
  category="lo",
  direction="l",
 },
 [0xA3F7]={
  category="lo",
  direction="l",
 },
 [0xA3F8]={
  category="lo",
  direction="l",
 },
 [0xA3F9]={
  category="lo",
  direction="l",
 },
 [0xA3FA]={
  category="lo",
  direction="l",
 },
 [0xA3FB]={
  category="lo",
  direction="l",
 },
 [0xA3FC]={
  category="lo",
  direction="l",
 },
 [0xA3FD]={
  category="lo",
  direction="l",
 },
 [0xA3FE]={
  category="lo",
  direction="l",
 },
 [0xA3FF]={
  category="lo",
  direction="l",
 },
 [0xA400]={
  category="lo",
  direction="l",
 },
 [0xA401]={
  category="lo",
  direction="l",
 },
 [0xA402]={
  category="lo",
  direction="l",
 },
 [0xA403]={
  category="lo",
  direction="l",
 },
 [0xA404]={
  category="lo",
  direction="l",
 },
 [0xA405]={
  category="lo",
  direction="l",
 },
 [0xA406]={
  category="lo",
  direction="l",
 },
 [0xA407]={
  category="lo",
  direction="l",
 },
 [0xA408]={
  category="lo",
  direction="l",
 },
 [0xA409]={
  category="lo",
  direction="l",
 },
 [0xA40A]={
  category="lo",
  direction="l",
 },
 [0xA40B]={
  category="lo",
  direction="l",
 },
 [0xA40C]={
  category="lo",
  direction="l",
 },
 [0xA40D]={
  category="lo",
  direction="l",
 },
 [0xA40E]={
  category="lo",
  direction="l",
 },
 [0xA40F]={
  category="lo",
  direction="l",
 },
 [0xA410]={
  category="lo",
  direction="l",
 },
 [0xA411]={
  category="lo",
  direction="l",
 },
 [0xA412]={
  category="lo",
  direction="l",
 },
 [0xA413]={
  category="lo",
  direction="l",
 },
 [0xA414]={
  category="lo",
  direction="l",
 },
 [0xA415]={
  category="lo",
  direction="l",
 },
 [0xA416]={
  category="lo",
  direction="l",
 },
 [0xA417]={
  category="lo",
  direction="l",
 },
 [0xA418]={
  category="lo",
  direction="l",
 },
 [0xA419]={
  category="lo",
  direction="l",
 },
 [0xA41A]={
  category="lo",
  direction="l",
 },
 [0xA41B]={
  category="lo",
  direction="l",
 },
 [0xA41C]={
  category="lo",
  direction="l",
 },
 [0xA41D]={
  category="lo",
  direction="l",
 },
 [0xA41E]={
  category="lo",
  direction="l",
 },
 [0xA41F]={
  category="lo",
  direction="l",
 },
 [0xA420]={
  category="lo",
  direction="l",
 },
 [0xA421]={
  category="lo",
  direction="l",
 },
 [0xA422]={
  category="lo",
  direction="l",
 },
 [0xA423]={
  category="lo",
  direction="l",
 },
 [0xA424]={
  category="lo",
  direction="l",
 },
 [0xA425]={
  category="lo",
  direction="l",
 },
 [0xA426]={
  category="lo",
  direction="l",
 },
 [0xA427]={
  category="lo",
  direction="l",
 },
 [0xA428]={
  category="lo",
  direction="l",
 },
 [0xA429]={
  category="lo",
  direction="l",
 },
 [0xA42A]={
  category="lo",
  direction="l",
 },
 [0xA42B]={
  category="lo",
  direction="l",
 },
 [0xA42C]={
  category="lo",
  direction="l",
 },
 [0xA42D]={
  category="lo",
  direction="l",
 },
 [0xA42E]={
  category="lo",
  direction="l",
 },
 [0xA42F]={
  category="lo",
  direction="l",
 },
 [0xA430]={
  category="lo",
  direction="l",
 },
 [0xA431]={
  category="lo",
  direction="l",
 },
 [0xA432]={
  category="lo",
  direction="l",
 },
 [0xA433]={
  category="lo",
  direction="l",
 },
 [0xA434]={
  category="lo",
  direction="l",
 },
 [0xA435]={
  category="lo",
  direction="l",
 },
 [0xA436]={
  category="lo",
  direction="l",
 },
 [0xA437]={
  category="lo",
  direction="l",
 },
 [0xA438]={
  category="lo",
  direction="l",
 },
 [0xA439]={
  category="lo",
  direction="l",
 },
 [0xA43A]={
  category="lo",
  direction="l",
 },
 [0xA43B]={
  category="lo",
  direction="l",
 },
 [0xA43C]={
  category="lo",
  direction="l",
 },
 [0xA43D]={
  category="lo",
  direction="l",
 },
 [0xA43E]={
  category="lo",
  direction="l",
 },
 [0xA43F]={
  category="lo",
  direction="l",
 },
 [0xA440]={
  category="lo",
  direction="l",
 },
 [0xA441]={
  category="lo",
  direction="l",
 },
 [0xA442]={
  category="lo",
  direction="l",
 },
 [0xA443]={
  category="lo",
  direction="l",
 },
 [0xA444]={
  category="lo",
  direction="l",
 },
 [0xA445]={
  category="lo",
  direction="l",
 },
 [0xA446]={
  category="lo",
  direction="l",
 },
 [0xA447]={
  category="lo",
  direction="l",
 },
 [0xA448]={
  category="lo",
  direction="l",
 },
 [0xA449]={
  category="lo",
  direction="l",
 },
 [0xA44A]={
  category="lo",
  direction="l",
 },
 [0xA44B]={
  category="lo",
  direction="l",
 },
 [0xA44C]={
  category="lo",
  direction="l",
 },
 [0xA44D]={
  category="lo",
  direction="l",
 },
 [0xA44E]={
  category="lo",
  direction="l",
 },
 [0xA44F]={
  category="lo",
  direction="l",
 },
 [0xA450]={
  category="lo",
  direction="l",
 },
 [0xA451]={
  category="lo",
  direction="l",
 },
 [0xA452]={
  category="lo",
  direction="l",
 },
 [0xA453]={
  category="lo",
  direction="l",
 },
 [0xA454]={
  category="lo",
  direction="l",
 },
 [0xA455]={
  category="lo",
  direction="l",
 },
 [0xA456]={
  category="lo",
  direction="l",
 },
 [0xA457]={
  category="lo",
  direction="l",
 },
 [0xA458]={
  category="lo",
  direction="l",
 },
 [0xA459]={
  category="lo",
  direction="l",
 },
 [0xA45A]={
  category="lo",
  direction="l",
 },
 [0xA45B]={
  category="lo",
  direction="l",
 },
 [0xA45C]={
  category="lo",
  direction="l",
 },
 [0xA45D]={
  category="lo",
  direction="l",
 },
 [0xA45E]={
  category="lo",
  direction="l",
 },
 [0xA45F]={
  category="lo",
  direction="l",
 },
 [0xA460]={
  category="lo",
  direction="l",
 },
 [0xA461]={
  category="lo",
  direction="l",
 },
 [0xA462]={
  category="lo",
  direction="l",
 },
 [0xA463]={
  category="lo",
  direction="l",
 },
 [0xA464]={
  category="lo",
  direction="l",
 },
 [0xA465]={
  category="lo",
  direction="l",
 },
 [0xA466]={
  category="lo",
  direction="l",
 },
 [0xA467]={
  category="lo",
  direction="l",
 },
 [0xA468]={
  category="lo",
  direction="l",
 },
 [0xA469]={
  category="lo",
  direction="l",
 },
 [0xA46A]={
  category="lo",
  direction="l",
 },
 [0xA46B]={
  category="lo",
  direction="l",
 },
 [0xA46C]={
  category="lo",
  direction="l",
 },
 [0xA46D]={
  category="lo",
  direction="l",
 },
 [0xA46E]={
  category="lo",
  direction="l",
 },
 [0xA46F]={
  category="lo",
  direction="l",
 },
 [0xA470]={
  category="lo",
  direction="l",
 },
 [0xA471]={
  category="lo",
  direction="l",
 },
 [0xA472]={
  category="lo",
  direction="l",
 },
 [0xA473]={
  category="lo",
  direction="l",
 },
 [0xA474]={
  category="lo",
  direction="l",
 },
 [0xA475]={
  category="lo",
  direction="l",
 },
 [0xA476]={
  category="lo",
  direction="l",
 },
 [0xA477]={
  category="lo",
  direction="l",
 },
 [0xA478]={
  category="lo",
  direction="l",
 },
 [0xA479]={
  category="lo",
  direction="l",
 },
 [0xA47A]={
  category="lo",
  direction="l",
 },
 [0xA47B]={
  category="lo",
  direction="l",
 },
 [0xA47C]={
  category="lo",
  direction="l",
 },
 [0xA47D]={
  category="lo",
  direction="l",
 },
 [0xA47E]={
  category="lo",
  direction="l",
 },
 [0xA47F]={
  category="lo",
  direction="l",
 },
 [0xA480]={
  category="lo",
  direction="l",
 },
 [0xA481]={
  category="lo",
  direction="l",
 },
 [0xA482]={
  category="lo",
  direction="l",
 },
 [0xA483]={
  category="lo",
  direction="l",
 },
 [0xA484]={
  category="lo",
  direction="l",
 },
 [0xA485]={
  category="lo",
  direction="l",
 },
 [0xA486]={
  category="lo",
  direction="l",
 },
 [0xA487]={
  category="lo",
  direction="l",
 },
 [0xA488]={
  category="lo",
  direction="l",
 },
 [0xA489]={
  category="lo",
  direction="l",
 },
 [0xA48A]={
  category="lo",
  direction="l",
 },
 [0xA48B]={
  category="lo",
  direction="l",
 },
 [0xA48C]={
  category="lo",
  direction="l",
 },
 [0xA490]={
  category="so",
  direction="on",
 },
 [0xA491]={
  category="so",
  direction="on",
 },
 [0xA492]={
  category="so",
  direction="on",
 },
 [0xA493]={
  category="so",
  direction="on",
 },
 [0xA494]={
  category="so",
  direction="on",
 },
 [0xA495]={
  category="so",
  direction="on",
 },
 [0xA496]={
  category="so",
  direction="on",
 },
 [0xA497]={
  category="so",
  direction="on",
 },
 [0xA498]={
  category="so",
  direction="on",
 },
 [0xA499]={
  category="so",
  direction="on",
 },
 [0xA49A]={
  category="so",
  direction="on",
 },
 [0xA49B]={
  category="so",
  direction="on",
 },
 [0xA49C]={
  category="so",
  direction="on",
 },
 [0xA49D]={
  category="so",
  direction="on",
 },
 [0xA49E]={
  category="so",
  direction="on",
 },
 [0xA49F]={
  category="so",
  direction="on",
 },
 [0xA4A0]={
  category="so",
  direction="on",
 },
 [0xA4A1]={
  category="so",
  direction="on",
 },
 [0xA4A2]={
  category="so",
  direction="on",
 },
 [0xA4A3]={
  category="so",
  direction="on",
 },
 [0xA4A4]={
  category="so",
  direction="on",
 },
 [0xA4A5]={
  category="so",
  direction="on",
 },
 [0xA4A6]={
  category="so",
  direction="on",
 },
 [0xA4A7]={
  category="so",
  direction="on",
 },
 [0xA4A8]={
  category="so",
  direction="on",
 },
 [0xA4A9]={
  category="so",
  direction="on",
 },
 [0xA4AA]={
  category="so",
  direction="on",
 },
 [0xA4AB]={
  category="so",
  direction="on",
 },
 [0xA4AC]={
  category="so",
  direction="on",
 },
 [0xA4AD]={
  category="so",
  direction="on",
 },
 [0xA4AE]={
  category="so",
  direction="on",
 },
 [0xA4AF]={
  category="so",
  direction="on",
 },
 [0xA4B0]={
  category="so",
  direction="on",
 },
 [0xA4B1]={
  category="so",
  direction="on",
 },
 [0xA4B2]={
  category="so",
  direction="on",
 },
 [0xA4B3]={
  category="so",
  direction="on",
 },
 [0xA4B4]={
  category="so",
  direction="on",
 },
 [0xA4B5]={
  category="so",
  direction="on",
 },
 [0xA4B6]={
  category="so",
  direction="on",
 },
 [0xA4B7]={
  category="so",
  direction="on",
 },
 [0xA4B8]={
  category="so",
  direction="on",
 },
 [0xA4B9]={
  category="so",
  direction="on",
 },
 [0xA4BA]={
  category="so",
  direction="on",
 },
 [0xA4BB]={
  category="so",
  direction="on",
 },
 [0xA4BC]={
  category="so",
  direction="on",
 },
 [0xA4BD]={
  category="so",
  direction="on",
 },
 [0xA4BE]={
  category="so",
  direction="on",
 },
 [0xA4BF]={
  category="so",
  direction="on",
 },
 [0xA4C0]={
  category="so",
  direction="on",
 },
 [0xA4C1]={
  category="so",
  direction="on",
 },
 [0xA4C2]={
  category="so",
  direction="on",
 },
 [0xA4C3]={
  category="so",
  direction="on",
 },
 [0xA4C4]={
  category="so",
  direction="on",
 },
 [0xA4C5]={
  category="so",
  direction="on",
 },
 [0xA4C6]={
  category="so",
  direction="on",
 },
 [0xA4D0]={
  category="lo",
  direction="l",
 },
 [0xA4D1]={
  category="lo",
  direction="l",
 },
 [0xA4D2]={
  category="lo",
  direction="l",
 },
 [0xA4D3]={
  category="lo",
  direction="l",
 },
 [0xA4D4]={
  category="lo",
  direction="l",
 },
 [0xA4D5]={
  category="lo",
  direction="l",
 },
 [0xA4D6]={
  category="lo",
  direction="l",
 },
 [0xA4D7]={
  category="lo",
  direction="l",
 },
 [0xA4D8]={
  category="lo",
  direction="l",
 },
 [0xA4D9]={
  category="lo",
  direction="l",
 },
 [0xA4DA]={
  category="lo",
  direction="l",
 },
 [0xA4DB]={
  category="lo",
  direction="l",
 },
 [0xA4DC]={
  category="lo",
  direction="l",
 },
 [0xA4DD]={
  category="lo",
  direction="l",
 },
 [0xA4DE]={
  category="lo",
  direction="l",
 },
 [0xA4DF]={
  category="lo",
  direction="l",
 },
 [0xA4E0]={
  category="lo",
  direction="l",
 },
 [0xA4E1]={
  category="lo",
  direction="l",
 },
 [0xA4E2]={
  category="lo",
  direction="l",
 },
 [0xA4E3]={
  category="lo",
  direction="l",
 },
 [0xA4E4]={
  category="lo",
  direction="l",
 },
 [0xA4E5]={
  category="lo",
  direction="l",
 },
 [0xA4E6]={
  category="lo",
  direction="l",
 },
 [0xA4E7]={
  category="lo",
  direction="l",
 },
 [0xA4E8]={
  category="lo",
  direction="l",
 },
 [0xA4E9]={
  category="lo",
  direction="l",
 },
 [0xA4EA]={
  category="lo",
  direction="l",
 },
 [0xA4EB]={
  category="lo",
  direction="l",
 },
 [0xA4EC]={
  category="lo",
  direction="l",
 },
 [0xA4ED]={
  category="lo",
  direction="l",
 },
 [0xA4EE]={
  category="lo",
  direction="l",
 },
 [0xA4EF]={
  category="lo",
  direction="l",
 },
 [0xA4F0]={
  category="lo",
  direction="l",
 },
 [0xA4F1]={
  category="lo",
  direction="l",
 },
 [0xA4F2]={
  category="lo",
  direction="l",
 },
 [0xA4F3]={
  category="lo",
  direction="l",
 },
 [0xA4F4]={
  category="lo",
  direction="l",
 },
 [0xA4F5]={
  category="lo",
  direction="l",
 },
 [0xA4F6]={
  category="lo",
  direction="l",
 },
 [0xA4F7]={
  category="lo",
  direction="l",
 },
 [0xA4F8]={
  category="lm",
  direction="l",
 },
 [0xA4F9]={
  category="lm",
  direction="l",
 },
 [0xA4FA]={
  category="lm",
  direction="l",
 },
 [0xA4FB]={
  category="lm",
  direction="l",
 },
 [0xA4FC]={
  category="lm",
  direction="l",
 },
 [0xA4FD]={
  category="lm",
  direction="l",
 },
 [0xA4FE]={
  category="po",
  direction="l",
 },
 [0xA4FF]={
  category="po",
  direction="l",
 },
 [0xA500]={
  category="lo",
  direction="l",
 },
 [0xA501]={
  category="lo",
  direction="l",
 },
 [0xA502]={
  category="lo",
  direction="l",
 },
 [0xA503]={
  category="lo",
  direction="l",
 },
 [0xA504]={
  category="lo",
  direction="l",
 },
 [0xA505]={
  category="lo",
  direction="l",
 },
 [0xA506]={
  category="lo",
  direction="l",
 },
 [0xA507]={
  category="lo",
  direction="l",
 },
 [0xA508]={
  category="lo",
  direction="l",
 },
 [0xA509]={
  category="lo",
  direction="l",
 },
 [0xA50A]={
  category="lo",
  direction="l",
 },
 [0xA50B]={
  category="lo",
  direction="l",
 },
 [0xA50C]={
  category="lo",
  direction="l",
 },
 [0xA50D]={
  category="lo",
  direction="l",
 },
 [0xA50E]={
  category="lo",
  direction="l",
 },
 [0xA50F]={
  category="lo",
  direction="l",
 },
 [0xA510]={
  category="lo",
  direction="l",
 },
 [0xA511]={
  category="lo",
  direction="l",
 },
 [0xA512]={
  category="lo",
  direction="l",
 },
 [0xA513]={
  category="lo",
  direction="l",
 },
 [0xA514]={
  category="lo",
  direction="l",
 },
 [0xA515]={
  category="lo",
  direction="l",
 },
 [0xA516]={
  category="lo",
  direction="l",
 },
 [0xA517]={
  category="lo",
  direction="l",
 },
 [0xA518]={
  category="lo",
  direction="l",
 },
 [0xA519]={
  category="lo",
  direction="l",
 },
 [0xA51A]={
  category="lo",
  direction="l",
 },
 [0xA51B]={
  category="lo",
  direction="l",
 },
 [0xA51C]={
  category="lo",
  direction="l",
 },
 [0xA51D]={
  category="lo",
  direction="l",
 },
 [0xA51E]={
  category="lo",
  direction="l",
 },
 [0xA51F]={
  category="lo",
  direction="l",
 },
 [0xA520]={
  category="lo",
  direction="l",
 },
 [0xA521]={
  category="lo",
  direction="l",
 },
 [0xA522]={
  category="lo",
  direction="l",
 },
 [0xA523]={
  category="lo",
  direction="l",
 },
 [0xA524]={
  category="lo",
  direction="l",
 },
 [0xA525]={
  category="lo",
  direction="l",
 },
 [0xA526]={
  category="lo",
  direction="l",
 },
 [0xA527]={
  category="lo",
  direction="l",
 },
 [0xA528]={
  category="lo",
  direction="l",
 },
 [0xA529]={
  category="lo",
  direction="l",
 },
 [0xA52A]={
  category="lo",
  direction="l",
 },
 [0xA52B]={
  category="lo",
  direction="l",
 },
 [0xA52C]={
  category="lo",
  direction="l",
 },
 [0xA52D]={
  category="lo",
  direction="l",
 },
 [0xA52E]={
  category="lo",
  direction="l",
 },
 [0xA52F]={
  category="lo",
  direction="l",
 },
 [0xA530]={
  category="lo",
  direction="l",
 },
 [0xA531]={
  category="lo",
  direction="l",
 },
 [0xA532]={
  category="lo",
  direction="l",
 },
 [0xA533]={
  category="lo",
  direction="l",
 },
 [0xA534]={
  category="lo",
  direction="l",
 },
 [0xA535]={
  category="lo",
  direction="l",
 },
 [0xA536]={
  category="lo",
  direction="l",
 },
 [0xA537]={
  category="lo",
  direction="l",
 },
 [0xA538]={
  category="lo",
  direction="l",
 },
 [0xA539]={
  category="lo",
  direction="l",
 },
 [0xA53A]={
  category="lo",
  direction="l",
 },
 [0xA53B]={
  category="lo",
  direction="l",
 },
 [0xA53C]={
  category="lo",
  direction="l",
 },
 [0xA53D]={
  category="lo",
  direction="l",
 },
 [0xA53E]={
  category="lo",
  direction="l",
 },
 [0xA53F]={
  category="lo",
  direction="l",
 },
 [0xA540]={
  category="lo",
  direction="l",
 },
 [0xA541]={
  category="lo",
  direction="l",
 },
 [0xA542]={
  category="lo",
  direction="l",
 },
 [0xA543]={
  category="lo",
  direction="l",
 },
 [0xA544]={
  category="lo",
  direction="l",
 },
 [0xA545]={
  category="lo",
  direction="l",
 },
 [0xA546]={
  category="lo",
  direction="l",
 },
 [0xA547]={
  category="lo",
  direction="l",
 },
 [0xA548]={
  category="lo",
  direction="l",
 },
 [0xA549]={
  category="lo",
  direction="l",
 },
 [0xA54A]={
  category="lo",
  direction="l",
 },
 [0xA54B]={
  category="lo",
  direction="l",
 },
 [0xA54C]={
  category="lo",
  direction="l",
 },
 [0xA54D]={
  category="lo",
  direction="l",
 },
 [0xA54E]={
  category="lo",
  direction="l",
 },
 [0xA54F]={
  category="lo",
  direction="l",
 },
 [0xA550]={
  category="lo",
  direction="l",
 },
 [0xA551]={
  category="lo",
  direction="l",
 },
 [0xA552]={
  category="lo",
  direction="l",
 },
 [0xA553]={
  category="lo",
  direction="l",
 },
 [0xA554]={
  category="lo",
  direction="l",
 },
 [0xA555]={
  category="lo",
  direction="l",
 },
 [0xA556]={
  category="lo",
  direction="l",
 },
 [0xA557]={
  category="lo",
  direction="l",
 },
 [0xA558]={
  category="lo",
  direction="l",
 },
 [0xA559]={
  category="lo",
  direction="l",
 },
 [0xA55A]={
  category="lo",
  direction="l",
 },
 [0xA55B]={
  category="lo",
  direction="l",
 },
 [0xA55C]={
  category="lo",
  direction="l",
 },
 [0xA55D]={
  category="lo",
  direction="l",
 },
 [0xA55E]={
  category="lo",
  direction="l",
 },
 [0xA55F]={
  category="lo",
  direction="l",
 },
 [0xA560]={
  category="lo",
  direction="l",
 },
 [0xA561]={
  category="lo",
  direction="l",
 },
 [0xA562]={
  category="lo",
  direction="l",
 },
 [0xA563]={
  category="lo",
  direction="l",
 },
 [0xA564]={
  category="lo",
  direction="l",
 },
 [0xA565]={
  category="lo",
  direction="l",
 },
 [0xA566]={
  category="lo",
  direction="l",
 },
 [0xA567]={
  category="lo",
  direction="l",
 },
 [0xA568]={
  category="lo",
  direction="l",
 },
 [0xA569]={
  category="lo",
  direction="l",
 },
 [0xA56A]={
  category="lo",
  direction="l",
 },
 [0xA56B]={
  category="lo",
  direction="l",
 },
 [0xA56C]={
  category="lo",
  direction="l",
 },
 [0xA56D]={
  category="lo",
  direction="l",
 },
 [0xA56E]={
  category="lo",
  direction="l",
 },
 [0xA56F]={
  category="lo",
  direction="l",
 },
 [0xA570]={
  category="lo",
  direction="l",
 },
 [0xA571]={
  category="lo",
  direction="l",
 },
 [0xA572]={
  category="lo",
  direction="l",
 },
 [0xA573]={
  category="lo",
  direction="l",
 },
 [0xA574]={
  category="lo",
  direction="l",
 },
 [0xA575]={
  category="lo",
  direction="l",
 },
 [0xA576]={
  category="lo",
  direction="l",
 },
 [0xA577]={
  category="lo",
  direction="l",
 },
 [0xA578]={
  category="lo",
  direction="l",
 },
 [0xA579]={
  category="lo",
  direction="l",
 },
 [0xA57A]={
  category="lo",
  direction="l",
 },
 [0xA57B]={
  category="lo",
  direction="l",
 },
 [0xA57C]={
  category="lo",
  direction="l",
 },
 [0xA57D]={
  category="lo",
  direction="l",
 },
 [0xA57E]={
  category="lo",
  direction="l",
 },
 [0xA57F]={
  category="lo",
  direction="l",
 },
 [0xA580]={
  category="lo",
  direction="l",
 },
 [0xA581]={
  category="lo",
  direction="l",
 },
 [0xA582]={
  category="lo",
  direction="l",
 },
 [0xA583]={
  category="lo",
  direction="l",
 },
 [0xA584]={
  category="lo",
  direction="l",
 },
 [0xA585]={
  category="lo",
  direction="l",
 },
 [0xA586]={
  category="lo",
  direction="l",
 },
 [0xA587]={
  category="lo",
  direction="l",
 },
 [0xA588]={
  category="lo",
  direction="l",
 },
 [0xA589]={
  category="lo",
  direction="l",
 },
 [0xA58A]={
  category="lo",
  direction="l",
 },
 [0xA58B]={
  category="lo",
  direction="l",
 },
 [0xA58C]={
  category="lo",
  direction="l",
 },
 [0xA58D]={
  category="lo",
  direction="l",
 },
 [0xA58E]={
  category="lo",
  direction="l",
 },
 [0xA58F]={
  category="lo",
  direction="l",
 },
 [0xA590]={
  category="lo",
  direction="l",
 },
 [0xA591]={
  category="lo",
  direction="l",
 },
 [0xA592]={
  category="lo",
  direction="l",
 },
 [0xA593]={
  category="lo",
  direction="l",
 },
 [0xA594]={
  category="lo",
  direction="l",
 },
 [0xA595]={
  category="lo",
  direction="l",
 },
 [0xA596]={
  category="lo",
  direction="l",
 },
 [0xA597]={
  category="lo",
  direction="l",
 },
 [0xA598]={
  category="lo",
  direction="l",
 },
 [0xA599]={
  category="lo",
  direction="l",
 },
 [0xA59A]={
  category="lo",
  direction="l",
 },
 [0xA59B]={
  category="lo",
  direction="l",
 },
 [0xA59C]={
  category="lo",
  direction="l",
 },
 [0xA59D]={
  category="lo",
  direction="l",
 },
 [0xA59E]={
  category="lo",
  direction="l",
 },
 [0xA59F]={
  category="lo",
  direction="l",
 },
 [0xA5A0]={
  category="lo",
  direction="l",
 },
 [0xA5A1]={
  category="lo",
  direction="l",
 },
 [0xA5A2]={
  category="lo",
  direction="l",
 },
 [0xA5A3]={
  category="lo",
  direction="l",
 },
 [0xA5A4]={
  category="lo",
  direction="l",
 },
 [0xA5A5]={
  category="lo",
  direction="l",
 },
 [0xA5A6]={
  category="lo",
  direction="l",
 },
 [0xA5A7]={
  category="lo",
  direction="l",
 },
 [0xA5A8]={
  category="lo",
  direction="l",
 },
 [0xA5A9]={
  category="lo",
  direction="l",
 },
 [0xA5AA]={
  category="lo",
  direction="l",
 },
 [0xA5AB]={
  category="lo",
  direction="l",
 },
 [0xA5AC]={
  category="lo",
  direction="l",
 },
 [0xA5AD]={
  category="lo",
  direction="l",
 },
 [0xA5AE]={
  category="lo",
  direction="l",
 },
 [0xA5AF]={
  category="lo",
  direction="l",
 },
 [0xA5B0]={
  category="lo",
  direction="l",
 },
 [0xA5B1]={
  category="lo",
  direction="l",
 },
 [0xA5B2]={
  category="lo",
  direction="l",
 },
 [0xA5B3]={
  category="lo",
  direction="l",
 },
 [0xA5B4]={
  category="lo",
  direction="l",
 },
 [0xA5B5]={
  category="lo",
  direction="l",
 },
 [0xA5B6]={
  category="lo",
  direction="l",
 },
 [0xA5B7]={
  category="lo",
  direction="l",
 },
 [0xA5B8]={
  category="lo",
  direction="l",
 },
 [0xA5B9]={
  category="lo",
  direction="l",
 },
 [0xA5BA]={
  category="lo",
  direction="l",
 },
 [0xA5BB]={
  category="lo",
  direction="l",
 },
 [0xA5BC]={
  category="lo",
  direction="l",
 },
 [0xA5BD]={
  category="lo",
  direction="l",
 },
 [0xA5BE]={
  category="lo",
  direction="l",
 },
 [0xA5BF]={
  category="lo",
  direction="l",
 },
 [0xA5C0]={
  category="lo",
  direction="l",
 },
 [0xA5C1]={
  category="lo",
  direction="l",
 },
 [0xA5C2]={
  category="lo",
  direction="l",
 },
 [0xA5C3]={
  category="lo",
  direction="l",
 },
 [0xA5C4]={
  category="lo",
  direction="l",
 },
 [0xA5C5]={
  category="lo",
  direction="l",
 },
 [0xA5C6]={
  category="lo",
  direction="l",
 },
 [0xA5C7]={
  category="lo",
  direction="l",
 },
 [0xA5C8]={
  category="lo",
  direction="l",
 },
 [0xA5C9]={
  category="lo",
  direction="l",
 },
 [0xA5CA]={
  category="lo",
  direction="l",
 },
 [0xA5CB]={
  category="lo",
  direction="l",
 },
 [0xA5CC]={
  category="lo",
  direction="l",
 },
 [0xA5CD]={
  category="lo",
  direction="l",
 },
 [0xA5CE]={
  category="lo",
  direction="l",
 },
 [0xA5CF]={
  category="lo",
  direction="l",
 },
 [0xA5D0]={
  category="lo",
  direction="l",
 },
 [0xA5D1]={
  category="lo",
  direction="l",
 },
 [0xA5D2]={
  category="lo",
  direction="l",
 },
 [0xA5D3]={
  category="lo",
  direction="l",
 },
 [0xA5D4]={
  category="lo",
  direction="l",
 },
 [0xA5D5]={
  category="lo",
  direction="l",
 },
 [0xA5D6]={
  category="lo",
  direction="l",
 },
 [0xA5D7]={
  category="lo",
  direction="l",
 },
 [0xA5D8]={
  category="lo",
  direction="l",
 },
 [0xA5D9]={
  category="lo",
  direction="l",
 },
 [0xA5DA]={
  category="lo",
  direction="l",
 },
 [0xA5DB]={
  category="lo",
  direction="l",
 },
 [0xA5DC]={
  category="lo",
  direction="l",
 },
 [0xA5DD]={
  category="lo",
  direction="l",
 },
 [0xA5DE]={
  category="lo",
  direction="l",
 },
 [0xA5DF]={
  category="lo",
  direction="l",
 },
 [0xA5E0]={
  category="lo",
  direction="l",
 },
 [0xA5E1]={
  category="lo",
  direction="l",
 },
 [0xA5E2]={
  category="lo",
  direction="l",
 },
 [0xA5E3]={
  category="lo",
  direction="l",
 },
 [0xA5E4]={
  category="lo",
  direction="l",
 },
 [0xA5E5]={
  category="lo",
  direction="l",
 },
 [0xA5E6]={
  category="lo",
  direction="l",
 },
 [0xA5E7]={
  category="lo",
  direction="l",
 },
 [0xA5E8]={
  category="lo",
  direction="l",
 },
 [0xA5E9]={
  category="lo",
  direction="l",
 },
 [0xA5EA]={
  category="lo",
  direction="l",
 },
 [0xA5EB]={
  category="lo",
  direction="l",
 },
 [0xA5EC]={
  category="lo",
  direction="l",
 },
 [0xA5ED]={
  category="lo",
  direction="l",
 },
 [0xA5EE]={
  category="lo",
  direction="l",
 },
 [0xA5EF]={
  category="lo",
  direction="l",
 },
 [0xA5F0]={
  category="lo",
  direction="l",
 },
 [0xA5F1]={
  category="lo",
  direction="l",
 },
 [0xA5F2]={
  category="lo",
  direction="l",
 },
 [0xA5F3]={
  category="lo",
  direction="l",
 },
 [0xA5F4]={
  category="lo",
  direction="l",
 },
 [0xA5F5]={
  category="lo",
  direction="l",
 },
 [0xA5F6]={
  category="lo",
  direction="l",
 },
 [0xA5F7]={
  category="lo",
  direction="l",
 },
 [0xA5F8]={
  category="lo",
  direction="l",
 },
 [0xA5F9]={
  category="lo",
  direction="l",
 },
 [0xA5FA]={
  category="lo",
  direction="l",
 },
 [0xA5FB]={
  category="lo",
  direction="l",
 },
 [0xA5FC]={
  category="lo",
  direction="l",
 },
 [0xA5FD]={
  category="lo",
  direction="l",
 },
 [0xA5FE]={
  category="lo",
  direction="l",
 },
 [0xA5FF]={
  category="lo",
  direction="l",
 },
 [0xA600]={
  category="lo",
  direction="l",
 },
 [0xA601]={
  category="lo",
  direction="l",
 },
 [0xA602]={
  category="lo",
  direction="l",
 },
 [0xA603]={
  category="lo",
  direction="l",
 },
 [0xA604]={
  category="lo",
  direction="l",
 },
 [0xA605]={
  category="lo",
  direction="l",
 },
 [0xA606]={
  category="lo",
  direction="l",
 },
 [0xA607]={
  category="lo",
  direction="l",
 },
 [0xA608]={
  category="lo",
  direction="l",
 },
 [0xA609]={
  category="lo",
  direction="l",
 },
 [0xA60A]={
  category="lo",
  direction="l",
 },
 [0xA60B]={
  category="lo",
  direction="l",
 },
 [0xA60C]={
  category="lm",
  direction="l",
 },
 [0xA60D]={
  category="po",
  direction="on",
 },
 [0xA60E]={
  category="po",
  direction="on",
 },
 [0xA60F]={
  category="po",
  direction="on",
 },
 [0xA610]={
  category="lo",
  direction="l",
 },
 [0xA611]={
  category="lo",
  direction="l",
 },
 [0xA612]={
  category="lo",
  direction="l",
 },
 [0xA613]={
  category="lo",
  direction="l",
 },
 [0xA614]={
  category="lo",
  direction="l",
 },
 [0xA615]={
  category="lo",
  direction="l",
 },
 [0xA616]={
  category="lo",
  direction="l",
 },
 [0xA617]={
  category="lo",
  direction="l",
 },
 [0xA618]={
  category="lo",
  direction="l",
 },
 [0xA619]={
  category="lo",
  direction="l",
 },
 [0xA61A]={
  category="lo",
  direction="l",
 },
 [0xA61B]={
  category="lo",
  direction="l",
 },
 [0xA61C]={
  category="lo",
  direction="l",
 },
 [0xA61D]={
  category="lo",
  direction="l",
 },
 [0xA61E]={
  category="lo",
  direction="l",
 },
 [0xA61F]={
  category="lo",
  direction="l",
 },
 [0xA620]={
  category="nd",
  direction="l",
 },
 [0xA621]={
  category="nd",
  direction="l",
 },
 [0xA622]={
  category="nd",
  direction="l",
 },
 [0xA623]={
  category="nd",
  direction="l",
 },
 [0xA624]={
  category="nd",
  direction="l",
 },
 [0xA625]={
  category="nd",
  direction="l",
 },
 [0xA626]={
  category="nd",
  direction="l",
 },
 [0xA627]={
  category="nd",
  direction="l",
 },
 [0xA628]={
  category="nd",
  direction="l",
 },
 [0xA629]={
  category="nd",
  direction="l",
 },
 [0xA62A]={
  category="lo",
  direction="l",
 },
 [0xA62B]={
  category="lo",
  direction="l",
 },
 [0xA640]={
  category="lu",
  direction="l",
 },
 [0xA641]={
  category="ll",
  direction="l",
 },
 [0xA642]={
  category="lu",
  direction="l",
 },
 [0xA643]={
  category="ll",
  direction="l",
 },
 [0xA644]={
  category="lu",
  direction="l",
 },
 [0xA645]={
  category="ll",
  direction="l",
 },
 [0xA646]={
  category="lu",
  direction="l",
 },
 [0xA647]={
  category="ll",
  direction="l",
 },
 [0xA648]={
  category="lu",
  direction="l",
 },
 [0xA649]={
  category="ll",
  direction="l",
 },
 [0xA64A]={
  category="lu",
  direction="l",
 },
 [0xA64B]={
  category="ll",
  direction="l",
 },
 [0xA64C]={
  category="lu",
  direction="l",
 },
 [0xA64D]={
  category="ll",
  direction="l",
 },
 [0xA64E]={
  category="lu",
  direction="l",
 },
 [0xA64F]={
  category="ll",
  direction="l",
 },
 [0xA650]={
  category="lu",
  direction="l",
 },
 [0xA651]={
  category="ll",
  direction="l",
 },
 [0xA652]={
  category="lu",
  direction="l",
 },
 [0xA653]={
  category="ll",
  direction="l",
 },
 [0xA654]={
  category="lu",
  direction="l",
 },
 [0xA655]={
  category="ll",
  direction="l",
 },
 [0xA656]={
  category="lu",
  direction="l",
 },
 [0xA657]={
  category="ll",
  direction="l",
 },
 [0xA658]={
  category="lu",
  direction="l",
 },
 [0xA659]={
  category="ll",
  direction="l",
 },
 [0xA65A]={
  category="lu",
  direction="l",
 },
 [0xA65B]={
  category="ll",
  direction="l",
 },
 [0xA65C]={
  category="lu",
  direction="l",
 },
 [0xA65D]={
  category="ll",
  direction="l",
 },
 [0xA65E]={
  category="lu",
  direction="l",
 },
 [0xA65F]={
  category="ll",
  direction="l",
 },
 [0xA660]={
  category="lu",
  direction="l",
 },
 [0xA661]={
  category="ll",
  direction="l",
 },
 [0xA662]={
  category="lu",
  direction="l",
 },
 [0xA663]={
  category="ll",
  direction="l",
 },
 [0xA664]={
  category="lu",
  direction="l",
 },
 [0xA665]={
  category="ll",
  direction="l",
 },
 [0xA666]={
  category="lu",
  direction="l",
 },
 [0xA667]={
  category="ll",
  direction="l",
 },
 [0xA668]={
  category="lu",
  direction="l",
 },
 [0xA669]={
  category="ll",
  direction="l",
 },
 [0xA66A]={
  category="lu",
  direction="l",
 },
 [0xA66B]={
  category="ll",
  direction="l",
 },
 [0xA66C]={
  category="lu",
  direction="l",
 },
 [0xA66D]={
  category="ll",
  direction="l",
 },
 [0xA66E]={
  category="lo",
  direction="l",
 },
 [0xA66F]={
  category="mn",
  direction="nsm",
 },
 [0xA670]={
  category="me",
  direction="nsm",
 },
 [0xA671]={
  category="me",
  direction="nsm",
 },
 [0xA672]={
  category="me",
  direction="nsm",
 },
 [0xA673]={
  category="po",
  direction="on",
 },
 [0xA674]={
  category="mn",
  direction="nsm",
 },
 [0xA675]={
  category="mn",
  direction="nsm",
 },
 [0xA676]={
  category="mn",
  direction="nsm",
 },
 [0xA677]={
  category="mn",
  direction="nsm",
 },
 [0xA678]={
  category="mn",
  direction="nsm",
 },
 [0xA679]={
  category="mn",
  direction="nsm",
 },
 [0xA67A]={
  category="mn",
  direction="nsm",
 },
 [0xA67B]={
  category="mn",
  direction="nsm",
 },
 [0xA67C]={
  category="mn",
  direction="nsm",
 },
 [0xA67D]={
  category="mn",
  direction="nsm",
 },
 [0xA67E]={
  category="po",
  direction="on",
 },
 [0xA67F]={
  category="lm",
  direction="on",
 },
 [0xA680]={
  category="lu",
  direction="l",
 },
 [0xA681]={
  category="ll",
  direction="l",
 },
 [0xA682]={
  category="lu",
  direction="l",
 },
 [0xA683]={
  category="ll",
  direction="l",
 },
 [0xA684]={
  category="lu",
  direction="l",
 },
 [0xA685]={
  category="ll",
  direction="l",
 },
 [0xA686]={
  category="lu",
  direction="l",
 },
 [0xA687]={
  category="ll",
  direction="l",
 },
 [0xA688]={
  category="lu",
  direction="l",
 },
 [0xA689]={
  category="ll",
  direction="l",
 },
 [0xA68A]={
  category="lu",
  direction="l",
 },
 [0xA68B]={
  category="ll",
  direction="l",
 },
 [0xA68C]={
  category="lu",
  direction="l",
 },
 [0xA68D]={
  category="ll",
  direction="l",
 },
 [0xA68E]={
  category="lu",
  direction="l",
 },
 [0xA68F]={
  category="ll",
  direction="l",
 },
 [0xA690]={
  category="lu",
  direction="l",
 },
 [0xA691]={
  category="ll",
  direction="l",
 },
 [0xA692]={
  category="lu",
  direction="l",
 },
 [0xA693]={
  category="ll",
  direction="l",
 },
 [0xA694]={
  category="lu",
  direction="l",
 },
 [0xA695]={
  category="ll",
  direction="l",
 },
 [0xA696]={
  category="lu",
  direction="l",
 },
 [0xA697]={
  category="ll",
  direction="l",
 },
 [0xA698]={
  category="lu",
  direction="l",
 },
 [0xA699]={
  category="ll",
  direction="l",
 },
 [0xA69A]={
  category="lu",
  direction="l",
 },
 [0xA69B]={
  category="ll",
  direction="l",
 },
 [0xA69C]={
  category="lm",
  direction="l",
 },
 [0xA69D]={
  category="lm",
  direction="l",
 },
 [0xA69F]={
  category="mn",
  direction="nsm",
 },
 [0xA6A0]={
  category="lo",
  direction="l",
 },
 [0xA6A1]={
  category="lo",
  direction="l",
 },
 [0xA6A2]={
  category="lo",
  direction="l",
 },
 [0xA6A3]={
  category="lo",
  direction="l",
 },
 [0xA6A4]={
  category="lo",
  direction="l",
 },
 [0xA6A5]={
  category="lo",
  direction="l",
 },
 [0xA6A6]={
  category="lo",
  direction="l",
 },
 [0xA6A7]={
  category="lo",
  direction="l",
 },
 [0xA6A8]={
  category="lo",
  direction="l",
 },
 [0xA6A9]={
  category="lo",
  direction="l",
 },
 [0xA6AA]={
  category="lo",
  direction="l",
 },
 [0xA6AB]={
  category="lo",
  direction="l",
 },
 [0xA6AC]={
  category="lo",
  direction="l",
 },
 [0xA6AD]={
  category="lo",
  direction="l",
 },
 [0xA6AE]={
  category="lo",
  direction="l",
 },
 [0xA6AF]={
  category="lo",
  direction="l",
 },
 [0xA6B0]={
  category="lo",
  direction="l",
 },
 [0xA6B1]={
  category="lo",
  direction="l",
 },
 [0xA6B2]={
  category="lo",
  direction="l",
 },
 [0xA6B3]={
  category="lo",
  direction="l",
 },
 [0xA6B4]={
  category="lo",
  direction="l",
 },
 [0xA6B5]={
  category="lo",
  direction="l",
 },
 [0xA6B6]={
  category="lo",
  direction="l",
 },
 [0xA6B7]={
  category="lo",
  direction="l",
 },
 [0xA6B8]={
  category="lo",
  direction="l",
 },
 [0xA6B9]={
  category="lo",
  direction="l",
 },
 [0xA6BA]={
  category="lo",
  direction="l",
 },
 [0xA6BB]={
  category="lo",
  direction="l",
 },
 [0xA6BC]={
  category="lo",
  direction="l",
 },
 [0xA6BD]={
  category="lo",
  direction="l",
 },
 [0xA6BE]={
  category="lo",
  direction="l",
 },
 [0xA6BF]={
  category="lo",
  direction="l",
 },
 [0xA6C0]={
  category="lo",
  direction="l",
 },
 [0xA6C1]={
  category="lo",
  direction="l",
 },
 [0xA6C2]={
  category="lo",
  direction="l",
 },
 [0xA6C3]={
  category="lo",
  direction="l",
 },
 [0xA6C4]={
  category="lo",
  direction="l",
 },
 [0xA6C5]={
  category="lo",
  direction="l",
 },
 [0xA6C6]={
  category="lo",
  direction="l",
 },
 [0xA6C7]={
  category="lo",
  direction="l",
 },
 [0xA6C8]={
  category="lo",
  direction="l",
 },
 [0xA6C9]={
  category="lo",
  direction="l",
 },
 [0xA6CA]={
  category="lo",
  direction="l",
 },
 [0xA6CB]={
  category="lo",
  direction="l",
 },
 [0xA6CC]={
  category="lo",
  direction="l",
 },
 [0xA6CD]={
  category="lo",
  direction="l",
 },
 [0xA6CE]={
  category="lo",
  direction="l",
 },
 [0xA6CF]={
  category="lo",
  direction="l",
 },
 [0xA6D0]={
  category="lo",
  direction="l",
 },
 [0xA6D1]={
  category="lo",
  direction="l",
 },
 [0xA6D2]={
  category="lo",
  direction="l",
 },
 [0xA6D3]={
  category="lo",
  direction="l",
 },
 [0xA6D4]={
  category="lo",
  direction="l",
 },
 [0xA6D5]={
  category="lo",
  direction="l",
 },
 [0xA6D6]={
  category="lo",
  direction="l",
 },
 [0xA6D7]={
  category="lo",
  direction="l",
 },
 [0xA6D8]={
  category="lo",
  direction="l",
 },
 [0xA6D9]={
  category="lo",
  direction="l",
 },
 [0xA6DA]={
  category="lo",
  direction="l",
 },
 [0xA6DB]={
  category="lo",
  direction="l",
 },
 [0xA6DC]={
  category="lo",
  direction="l",
 },
 [0xA6DD]={
  category="lo",
  direction="l",
 },
 [0xA6DE]={
  category="lo",
  direction="l",
 },
 [0xA6DF]={
  category="lo",
  direction="l",
 },
 [0xA6E0]={
  category="lo",
  direction="l",
 },
 [0xA6E1]={
  category="lo",
  direction="l",
 },
 [0xA6E2]={
  category="lo",
  direction="l",
 },
 [0xA6E3]={
  category="lo",
  direction="l",
 },
 [0xA6E4]={
  category="lo",
  direction="l",
 },
 [0xA6E5]={
  category="lo",
  direction="l",
 },
 [0xA6E6]={
  category="nl",
  direction="l",
 },
 [0xA6E7]={
  category="nl",
  direction="l",
 },
 [0xA6E8]={
  category="nl",
  direction="l",
 },
 [0xA6E9]={
  category="nl",
  direction="l",
 },
 [0xA6EA]={
  category="nl",
  direction="l",
 },
 [0xA6EB]={
  category="nl",
  direction="l",
 },
 [0xA6EC]={
  category="nl",
  direction="l",
 },
 [0xA6ED]={
  category="nl",
  direction="l",
 },
 [0xA6EE]={
  category="nl",
  direction="l",
 },
 [0xA6EF]={
  category="nl",
  direction="l",
 },
 [0xA6F0]={
  category="mn",
  direction="nsm",
 },
 [0xA6F1]={
  category="mn",
  direction="nsm",
 },
 [0xA6F2]={
  category="po",
  direction="l",
 },
 [0xA6F3]={
  category="po",
  direction="l",
 },
 [0xA6F4]={
  category="po",
  direction="l",
 },
 [0xA6F5]={
  category="po",
  direction="l",
 },
 [0xA6F6]={
  category="po",
  direction="l",
 },
 [0xA6F7]={
  category="po",
  direction="l",
 },
 [0xA700]={
  category="sk",
  direction="on",
 },
 [0xA701]={
  category="sk",
  direction="on",
 },
 [0xA702]={
  category="sk",
  direction="on",
 },
 [0xA703]={
  category="sk",
  direction="on",
 },
 [0xA704]={
  category="sk",
  direction="on",
 },
 [0xA705]={
  category="sk",
  direction="on",
 },
 [0xA706]={
  category="sk",
  direction="on",
 },
 [0xA707]={
  category="sk",
  direction="on",
 },
 [0xA708]={
  category="sk",
  direction="on",
 },
 [0xA709]={
  category="sk",
  direction="on",
 },
 [0xA70A]={
  category="sk",
  direction="on",
 },
 [0xA70B]={
  category="sk",
  direction="on",
 },
 [0xA70C]={
  category="sk",
  direction="on",
 },
 [0xA70D]={
  category="sk",
  direction="on",
 },
 [0xA70E]={
  category="sk",
  direction="on",
 },
 [0xA70F]={
  category="sk",
  direction="on",
 },
 [0xA710]={
  category="sk",
  direction="on",
 },
 [0xA711]={
  category="sk",
  direction="on",
 },
 [0xA712]={
  category="sk",
  direction="on",
 },
 [0xA713]={
  category="sk",
  direction="on",
 },
 [0xA714]={
  category="sk",
  direction="on",
 },
 [0xA715]={
  category="sk",
  direction="on",
 },
 [0xA716]={
  category="sk",
  direction="on",
 },
 [0xA717]={
  category="lm",
  direction="on",
 },
 [0xA718]={
  category="lm",
  direction="on",
 },
 [0xA719]={
  category="lm",
  direction="on",
 },
 [0xA71A]={
  category="lm",
  direction="on",
 },
 [0xA71B]={
  category="lm",
  direction="on",
 },
 [0xA71C]={
  category="lm",
  direction="on",
 },
 [0xA71D]={
  category="lm",
  direction="on",
 },
 [0xA71E]={
  category="lm",
  direction="on",
 },
 [0xA71F]={
  category="lm",
  direction="on",
 },
 [0xA720]={
  category="sk",
  direction="on",
 },
 [0xA721]={
  category="sk",
  direction="on",
 },
 [0xA722]={
  category="lu",
  direction="l",
 },
 [0xA723]={
  category="ll",
  direction="l",
 },
 [0xA724]={
  category="lu",
  direction="l",
 },
 [0xA725]={
  category="ll",
  direction="l",
 },
 [0xA726]={
  category="lu",
  direction="l",
 },
 [0xA727]={
  category="ll",
  direction="l",
 },
 [0xA728]={
  category="lu",
  direction="l",
 },
 [0xA729]={
  category="ll",
  direction="l",
 },
 [0xA72A]={
  category="lu",
  direction="l",
 },
 [0xA72B]={
  category="ll",
  direction="l",
 },
 [0xA72C]={
  category="lu",
  direction="l",
 },
 [0xA72D]={
  category="ll",
  direction="l",
 },
 [0xA72E]={
  category="lu",
  direction="l",
 },
 [0xA72F]={
  category="ll",
  direction="l",
 },
 [0xA730]={
  category="ll",
  direction="l",
 },
 [0xA731]={
  category="ll",
  direction="l",
 },
 [0xA732]={
  category="lu",
  direction="l",
 },
 [0xA733]={
  category="ll",
  direction="l",
 },
 [0xA734]={
  category="lu",
  direction="l",
 },
 [0xA735]={
  category="ll",
  direction="l",
 },
 [0xA736]={
  category="lu",
  direction="l",
 },
 [0xA737]={
  category="ll",
  direction="l",
 },
 [0xA738]={
  category="lu",
  direction="l",
 },
 [0xA739]={
  category="ll",
  direction="l",
 },
 [0xA73A]={
  category="lu",
  direction="l",
 },
 [0xA73B]={
  category="ll",
  direction="l",
 },
 [0xA73C]={
  category="lu",
  direction="l",
 },
 [0xA73D]={
  category="ll",
  direction="l",
 },
 [0xA73E]={
  category="lu",
  direction="l",
 },
 [0xA73F]={
  category="ll",
  direction="l",
 },
 [0xA740]={
  category="lu",
  direction="l",
 },
 [0xA741]={
  category="ll",
  direction="l",
 },
 [0xA742]={
  category="lu",
  direction="l",
 },
 [0xA743]={
  category="ll",
  direction="l",
 },
 [0xA744]={
  category="lu",
  direction="l",
 },
 [0xA745]={
  category="ll",
  direction="l",
 },
 [0xA746]={
  category="lu",
  direction="l",
 },
 [0xA747]={
  category="ll",
  direction="l",
 },
 [0xA748]={
  category="lu",
  direction="l",
 },
 [0xA749]={
  category="ll",
  direction="l",
 },
 [0xA74A]={
  category="lu",
  direction="l",
 },
 [0xA74B]={
  category="ll",
  direction="l",
 },
 [0xA74C]={
  category="lu",
  direction="l",
 },
 [0xA74D]={
  category="ll",
  direction="l",
 },
 [0xA74E]={
  category="lu",
  direction="l",
 },
 [0xA74F]={
  category="ll",
  direction="l",
 },
 [0xA750]={
  category="lu",
  direction="l",
 },
 [0xA751]={
  category="ll",
  direction="l",
 },
 [0xA752]={
  category="lu",
  direction="l",
 },
 [0xA753]={
  category="ll",
  direction="l",
 },
 [0xA754]={
  category="lu",
  direction="l",
 },
 [0xA755]={
  category="ll",
  direction="l",
 },
 [0xA756]={
  category="lu",
  direction="l",
 },
 [0xA757]={
  category="ll",
  direction="l",
 },
 [0xA758]={
  category="lu",
  direction="l",
 },
 [0xA759]={
  category="ll",
  direction="l",
 },
 [0xA75A]={
  category="lu",
  direction="l",
 },
 [0xA75B]={
  category="ll",
  direction="l",
 },
 [0xA75C]={
  category="lu",
  direction="l",
 },
 [0xA75D]={
  category="ll",
  direction="l",
 },
 [0xA75E]={
  category="lu",
  direction="l",
 },
 [0xA75F]={
  category="ll",
  direction="l",
 },
 [0xA760]={
  category="lu",
  direction="l",
 },
 [0xA761]={
  category="ll",
  direction="l",
 },
 [0xA762]={
  category="lu",
  direction="l",
 },
 [0xA763]={
  category="ll",
  direction="l",
 },
 [0xA764]={
  category="lu",
  direction="l",
 },
 [0xA765]={
  category="ll",
  direction="l",
 },
 [0xA766]={
  category="lu",
  direction="l",
 },
 [0xA767]={
  category="ll",
  direction="l",
 },
 [0xA768]={
  category="lu",
  direction="l",
 },
 [0xA769]={
  category="ll",
  direction="l",
 },
 [0xA76A]={
  category="lu",
  direction="l",
 },
 [0xA76B]={
  category="ll",
  direction="l",
 },
 [0xA76C]={
  category="lu",
  direction="l",
 },
 [0xA76D]={
  category="ll",
  direction="l",
 },
 [0xA76E]={
  category="lu",
  direction="l",
 },
 [0xA76F]={
  category="ll",
  direction="l",
 },
 [0xA770]={
  category="lm",
  direction="l",
 },
 [0xA771]={
  category="ll",
  direction="l",
 },
 [0xA772]={
  category="ll",
  direction="l",
 },
 [0xA773]={
  category="ll",
  direction="l",
 },
 [0xA774]={
  category="ll",
  direction="l",
 },
 [0xA775]={
  category="ll",
  direction="l",
 },
 [0xA776]={
  category="ll",
  direction="l",
 },
 [0xA777]={
  category="ll",
  direction="l",
 },
 [0xA778]={
  category="ll",
  direction="l",
 },
 [0xA779]={
  category="lu",
  direction="l",
 },
 [0xA77A]={
  category="ll",
  direction="l",
 },
 [0xA77B]={
  category="lu",
  direction="l",
 },
 [0xA77C]={
  category="ll",
  direction="l",
 },
 [0xA77D]={
  category="lu",
  direction="l",
 },
 [0xA77E]={
  category="lu",
  direction="l",
 },
 [0xA77F]={
  category="ll",
  direction="l",
 },
 [0xA780]={
  category="lu",
  direction="l",
 },
 [0xA781]={
  category="ll",
  direction="l",
 },
 [0xA782]={
  category="lu",
  direction="l",
 },
 [0xA783]={
  category="ll",
  direction="l",
 },
 [0xA784]={
  category="lu",
  direction="l",
 },
 [0xA785]={
  category="ll",
  direction="l",
 },
 [0xA786]={
  category="lu",
  direction="l",
 },
 [0xA787]={
  category="ll",
  direction="l",
 },
 [0xA788]={
  category="lm",
  direction="on",
 },
 [0xA789]={
  category="sk",
  direction="l",
 },
 [0xA78A]={
  category="sk",
  direction="l",
 },
 [0xA78B]={
  category="lu",
  direction="l",
 },
 [0xA78C]={
  category="ll",
  direction="l",
 },
 [0xA78D]={
  category="lu",
  direction="l",
 },
 [0xA78E]={
  category="ll",
  direction="l",
 },
 [0xA790]={
  category="lu",
  direction="l",
 },
 [0xA791]={
  category="ll",
  direction="l",
 },
 [0xA792]={
  category="lu",
  direction="l",
 },
 [0xA793]={
  category="ll",
  direction="l",
 },
 [0xA794]={
  category="ll",
  direction="l",
 },
 [0xA795]={
  category="ll",
  direction="l",
 },
 [0xA796]={
  category="lu",
  direction="l",
 },
 [0xA797]={
  category="ll",
  direction="l",
 },
 [0xA798]={
  category="lu",
  direction="l",
 },
 [0xA799]={
  category="ll",
  direction="l",
 },
 [0xA79A]={
  category="lu",
  direction="l",
 },
 [0xA79B]={
  category="ll",
  direction="l",
 },
 [0xA79C]={
  category="lu",
  direction="l",
 },
 [0xA79D]={
  category="ll",
  direction="l",
 },
 [0xA79E]={
  category="lu",
  direction="l",
 },
 [0xA79F]={
  category="ll",
  direction="l",
 },
 [0xA7A0]={
  category="lu",
  direction="l",
 },
 [0xA7A1]={
  category="ll",
  direction="l",
 },
 [0xA7A2]={
  category="lu",
  direction="l",
 },
 [0xA7A3]={
  category="ll",
  direction="l",
 },
 [0xA7A4]={
  category="lu",
  direction="l",
 },
 [0xA7A5]={
  category="ll",
  direction="l",
 },
 [0xA7A6]={
  category="lu",
  direction="l",
 },
 [0xA7A7]={
  category="ll",
  direction="l",
 },
 [0xA7A8]={
  category="lu",
  direction="l",
 },
 [0xA7A9]={
  category="ll",
  direction="l",
 },
 [0xA7AA]={
  category="lu",
  direction="l",
 },
 [0xA7AB]={
  category="lu",
  direction="l",
 },
 [0xA7AC]={
  category="lu",
  direction="l",
 },
 [0xA7AD]={
  category="lu",
  direction="l",
 },
 [0xA7B0]={
  category="lu",
  direction="l",
 },
 [0xA7B1]={
  category="lu",
  direction="l",
 },
 [0xA7F7]={
  category="lo",
  direction="l",
 },
 [0xA7F8]={
  category="lm",
  direction="l",
 },
 [0xA7F9]={
  category="lm",
  direction="l",
 },
 [0xA7FA]={
  category="ll",
  direction="l",
 },
 [0xA7FB]={
  category="lo",
  direction="l",
 },
 [0xA7FC]={
  category="lo",
  direction="l",
 },
 [0xA7FD]={
  category="lo",
  direction="l",
 },
 [0xA7FE]={
  category="lo",
  direction="l",
 },
 [0xA7FF]={
  category="lo",
  direction="l",
 },
 [0xA800]={
  category="lo",
  direction="l",
 },
 [0xA801]={
  category="lo",
  direction="l",
 },
 [0xA802]={
  category="mc",
  direction="nsm",
 },
 [0xA803]={
  category="lo",
  direction="l",
 },
 [0xA804]={
  category="lo",
  direction="l",
 },
 [0xA805]={
  category="lo",
  direction="l",
 },
 [0xA806]={
  category="mn",
  direction="nsm",
 },
 [0xA807]={
  category="lo",
  direction="l",
 },
 [0xA808]={
  category="lo",
  direction="l",
 },
 [0xA809]={
  category="lo",
  direction="l",
 },
 [0xA80A]={
  category="lo",
  direction="l",
 },
 [0xA80B]={
  category="mn",
  direction="nsm",
 },
 [0xA80C]={
  category="lo",
  direction="l",
 },
 [0xA80D]={
  category="lo",
  direction="l",
 },
 [0xA80E]={
  category="lo",
  direction="l",
 },
 [0xA80F]={
  category="lo",
  direction="l",
 },
 [0xA810]={
  category="lo",
  direction="l",
 },
 [0xA811]={
  category="lo",
  direction="l",
 },
 [0xA812]={
  category="lo",
  direction="l",
 },
 [0xA813]={
  category="lo",
  direction="l",
 },
 [0xA814]={
  category="lo",
  direction="l",
 },
 [0xA815]={
  category="lo",
  direction="l",
 },
 [0xA816]={
  category="lo",
  direction="l",
 },
 [0xA817]={
  category="lo",
  direction="l",
 },
 [0xA818]={
  category="lo",
  direction="l",
 },
 [0xA819]={
  category="lo",
  direction="l",
 },
 [0xA81A]={
  category="lo",
  direction="l",
 },
 [0xA81B]={
  category="lo",
  direction="l",
 },
 [0xA81C]={
  category="lo",
  direction="l",
 },
 [0xA81D]={
  category="lo",
  direction="l",
 },
 [0xA81E]={
  category="lo",
  direction="l",
 },
 [0xA81F]={
  category="lo",
  direction="l",
 },
 [0xA820]={
  category="lo",
  direction="l",
 },
 [0xA821]={
  category="lo",
  direction="l",
 },
 [0xA822]={
  category="lo",
  direction="l",
 },
 [0xA823]={
  category="mc",
  direction="l",
 },
 [0xA824]={
  category="mc",
  direction="l",
 },
 [0xA825]={
  category="mn",
  direction="nsm",
 },
 [0xA826]={
  category="mn",
  direction="nsm",
 },
 [0xA827]={
  category="mc",
  direction="l",
 },
 [0xA828]={
  category="so",
  direction="on",
 },
 [0xA829]={
  category="so",
  direction="on",
 },
 [0xA82A]={
  category="so",
  direction="on",
 },
 [0xA82B]={
  category="so",
  direction="on",
 },
 [0xA830]={
  category="no",
  direction="l",
 },
 [0xA831]={
  category="no",
  direction="l",
 },
 [0xA832]={
  category="no",
  direction="l",
 },
 [0xA833]={
  category="no",
  direction="l",
 },
 [0xA834]={
  category="no",
  direction="l",
 },
 [0xA835]={
  category="no",
  direction="l",
 },
 [0xA836]={
  category="so",
  direction="l",
 },
 [0xA837]={
  category="so",
  direction="l",
 },
 [0xA838]={
  category="sc",
  direction="et",
 },
 [0xA839]={
  category="so",
  direction="et",
 },
 [0xA840]={
  category="lo",
  direction="l",
 },
 [0xA841]={
  category="lo",
  direction="l",
 },
 [0xA842]={
  category="lo",
  direction="l",
 },
 [0xA843]={
  category="lo",
  direction="l",
 },
 [0xA844]={
  category="lo",
  direction="l",
 },
 [0xA845]={
  category="lo",
  direction="l",
 },
 [0xA846]={
  category="lo",
  direction="l",
 },
 [0xA847]={
  category="lo",
  direction="l",
 },
 [0xA848]={
  category="lo",
  direction="l",
 },
 [0xA849]={
  category="lo",
  direction="l",
 },
 [0xA84A]={
  category="lo",
  direction="l",
 },
 [0xA84B]={
  category="lo",
  direction="l",
 },
 [0xA84C]={
  category="lo",
  direction="l",
 },
 [0xA84D]={
  category="lo",
  direction="l",
 },
 [0xA84E]={
  category="lo",
  direction="l",
 },
 [0xA84F]={
  category="lo",
  direction="l",
 },
 [0xA850]={
  category="lo",
  direction="l",
 },
 [0xA851]={
  category="lo",
  direction="l",
 },
 [0xA852]={
  category="lo",
  direction="l",
 },
 [0xA853]={
  category="lo",
  direction="l",
 },
 [0xA854]={
  category="lo",
  direction="l",
 },
 [0xA855]={
  category="lo",
  direction="l",
 },
 [0xA856]={
  category="lo",
  direction="l",
 },
 [0xA857]={
  category="lo",
  direction="l",
 },
 [0xA858]={
  category="lo",
  direction="l",
 },
 [0xA859]={
  category="lo",
  direction="l",
 },
 [0xA85A]={
  category="lo",
  direction="l",
 },
 [0xA85B]={
  category="lo",
  direction="l",
 },
 [0xA85C]={
  category="lo",
  direction="l",
 },
 [0xA85D]={
  category="lo",
  direction="l",
 },
 [0xA85E]={
  category="lo",
  direction="l",
 },
 [0xA85F]={
  category="lo",
  direction="l",
 },
 [0xA860]={
  category="lo",
  direction="l",
 },
 [0xA861]={
  category="lo",
  direction="l",
 },
 [0xA862]={
  category="lo",
  direction="l",
 },
 [0xA863]={
  category="lo",
  direction="l",
 },
 [0xA864]={
  category="lo",
  direction="l",
 },
 [0xA865]={
  category="lo",
  direction="l",
 },
 [0xA866]={
  category="lo",
  direction="l",
 },
 [0xA867]={
  category="lo",
  direction="l",
 },
 [0xA868]={
  category="lo",
  direction="l",
 },
 [0xA869]={
  category="lo",
  direction="l",
 },
 [0xA86A]={
  category="lo",
  direction="l",
 },
 [0xA86B]={
  category="lo",
  direction="l",
 },
 [0xA86C]={
  category="lo",
  direction="l",
 },
 [0xA86D]={
  category="lo",
  direction="l",
 },
 [0xA86E]={
  category="lo",
  direction="l",
 },
 [0xA86F]={
  category="lo",
  direction="l",
 },
 [0xA870]={
  category="lo",
  direction="l",
 },
 [0xA871]={
  category="lo",
  direction="l",
 },
 [0xA872]={
  category="lo",
  direction="l",
 },
 [0xA873]={
  category="lo",
  direction="l",
 },
 [0xA874]={
  category="po",
  direction="on",
 },
 [0xA875]={
  category="po",
  direction="on",
 },
 [0xA876]={
  category="po",
  direction="on",
 },
 [0xA877]={
  category="po",
  direction="on",
 },
 [0xA880]={
  category="mc",
  direction="l",
 },
 [0xA881]={
  category="mc",
  direction="l",
 },
 [0xA882]={
  category="lo",
  direction="l",
 },
 [0xA883]={
  category="lo",
  direction="l",
 },
 [0xA884]={
  category="lo",
  direction="l",
 },
 [0xA885]={
  category="lo",
  direction="l",
 },
 [0xA886]={
  category="lo",
  direction="l",
 },
 [0xA887]={
  category="lo",
  direction="l",
 },
 [0xA888]={
  category="lo",
  direction="l",
 },
 [0xA889]={
  category="lo",
  direction="l",
 },
 [0xA88A]={
  category="lo",
  direction="l",
 },
 [0xA88B]={
  category="lo",
  direction="l",
 },
 [0xA88C]={
  category="lo",
  direction="l",
 },
 [0xA88D]={
  category="lo",
  direction="l",
 },
 [0xA88E]={
  category="lo",
  direction="l",
 },
 [0xA88F]={
  category="lo",
  direction="l",
 },
 [0xA890]={
  category="lo",
  direction="l",
 },
 [0xA891]={
  category="lo",
  direction="l",
 },
 [0xA892]={
  category="lo",
  direction="l",
 },
 [0xA893]={
  category="lo",
  direction="l",
 },
 [0xA894]={
  category="lo",
  direction="l",
 },
 [0xA895]={
  category="lo",
  direction="l",
 },
 [0xA896]={
  category="lo",
  direction="l",
 },
 [0xA897]={
  category="lo",
  direction="l",
 },
 [0xA898]={
  category="lo",
  direction="l",
 },
 [0xA899]={
  category="lo",
  direction="l",
 },
 [0xA89A]={
  category="lo",
  direction="l",
 },
 [0xA89B]={
  category="lo",
  direction="l",
 },
 [0xA89C]={
  category="lo",
  direction="l",
 },
 [0xA89D]={
  category="lo",
  direction="l",
 },
 [0xA89E]={
  category="lo",
  direction="l",
 },
 [0xA89F]={
  category="lo",
  direction="l",
 },
 [0xA8A0]={
  category="lo",
  direction="l",
 },
 [0xA8A1]={
  category="lo",
  direction="l",
 },
 [0xA8A2]={
  category="lo",
  direction="l",
 },
 [0xA8A3]={
  category="lo",
  direction="l",
 },
 [0xA8A4]={
  category="lo",
  direction="l",
 },
 [0xA8A5]={
  category="lo",
  direction="l",
 },
 [0xA8A6]={
  category="lo",
  direction="l",
 },
 [0xA8A7]={
  category="lo",
  direction="l",
 },
 [0xA8A8]={
  category="lo",
  direction="l",
 },
 [0xA8A9]={
  category="lo",
  direction="l",
 },
 [0xA8AA]={
  category="lo",
  direction="l",
 },
 [0xA8AB]={
  category="lo",
  direction="l",
 },
 [0xA8AC]={
  category="lo",
  direction="l",
 },
 [0xA8AD]={
  category="lo",
  direction="l",
 },
 [0xA8AE]={
  category="lo",
  direction="l",
 },
 [0xA8AF]={
  category="lo",
  direction="l",
 },
 [0xA8B0]={
  category="lo",
  direction="l",
 },
 [0xA8B1]={
  category="lo",
  direction="l",
 },
 [0xA8B2]={
  category="lo",
  direction="l",
 },
 [0xA8B3]={
  category="lo",
  direction="l",
 },
 [0xA8B4]={
  category="mc",
  direction="l",
 },
 [0xA8B5]={
  category="mc",
  direction="l",
 },
 [0xA8B6]={
  category="mc",
  direction="l",
 },
 [0xA8B7]={
  category="mc",
  direction="l",
 },
 [0xA8B8]={
  category="mc",
  direction="l",
 },
 [0xA8B9]={
  category="mc",
  direction="l",
 },
 [0xA8BA]={
  category="mc",
  direction="l",
 },
 [0xA8BB]={
  category="mc",
  direction="l",
 },
 [0xA8BC]={
  category="mc",
  direction="l",
 },
 [0xA8BD]={
  category="mc",
  direction="l",
 },
 [0xA8BE]={
  category="mc",
  direction="l",
 },
 [0xA8BF]={
  category="mc",
  direction="l",
 },
 [0xA8C0]={
  category="mc",
  direction="l",
 },
 [0xA8C1]={
  category="mc",
  direction="l",
 },
 [0xA8C2]={
  category="mc",
  direction="l",
 },
 [0xA8C3]={
  category="mc",
  direction="l",
 },
 [0xA8C4]={
  category="mn",
  direction="nsm",
 },
 [0xA8CE]={
  category="po",
  direction="l",
 },
 [0xA8CF]={
  category="po",
  direction="l",
 },
 [0xA8D0]={
  category="nd",
  direction="l",
 },
 [0xA8D1]={
  category="nd",
  direction="l",
 },
 [0xA8D2]={
  category="nd",
  direction="l",
 },
 [0xA8D3]={
  category="nd",
  direction="l",
 },
 [0xA8D4]={
  category="nd",
  direction="l",
 },
 [0xA8D5]={
  category="nd",
  direction="l",
 },
 [0xA8D6]={
  category="nd",
  direction="l",
 },
 [0xA8D7]={
  category="nd",
  direction="l",
 },
 [0xA8D8]={
  category="nd",
  direction="l",
 },
 [0xA8D9]={
  category="nd",
  direction="l",
 },
 [0xA8E0]={
  category="mn",
  direction="nsm",
 },
 [0xA8E1]={
  category="mn",
  direction="nsm",
 },
 [0xA8E2]={
  category="mn",
  direction="nsm",
 },
 [0xA8E3]={
  category="mn",
  direction="nsm",
 },
 [0xA8E4]={
  category="mn",
  direction="nsm",
 },
 [0xA8E5]={
  category="mn",
  direction="nsm",
 },
 [0xA8E6]={
  category="mn",
  direction="nsm",
 },
 [0xA8E7]={
  category="mn",
  direction="nsm",
 },
 [0xA8E8]={
  category="mn",
  direction="nsm",
 },
 [0xA8E9]={
  category="mn",
  direction="nsm",
 },
 [0xA8EA]={
  category="mn",
  direction="nsm",
 },
 [0xA8EB]={
  category="mn",
  direction="nsm",
 },
 [0xA8EC]={
  category="mn",
  direction="nsm",
 },
 [0xA8ED]={
  category="mn",
  direction="nsm",
 },
 [0xA8EE]={
  category="mn",
  direction="nsm",
 },
 [0xA8EF]={
  category="mn",
  direction="nsm",
 },
 [0xA8F0]={
  category="mn",
  direction="nsm",
 },
 [0xA8F1]={
  category="mn",
  direction="nsm",
 },
 [0xA8F2]={
  category="lo",
  direction="l",
 },
 [0xA8F3]={
  category="lo",
  direction="l",
 },
 [0xA8F4]={
  category="lo",
  direction="l",
 },
 [0xA8F5]={
  category="lo",
  direction="l",
 },
 [0xA8F6]={
  category="lo",
  direction="l",
 },
 [0xA8F7]={
  category="lo",
  direction="l",
 },
 [0xA8F8]={
  category="po",
  direction="l",
 },
 [0xA8F9]={
  category="po",
  direction="l",
 },
 [0xA8FA]={
  category="po",
  direction="l",
 },
 [0xA8FB]={
  category="lo",
  direction="l",
 },
 [0xA900]={
  category="nd",
  direction="l",
 },
 [0xA901]={
  category="nd",
  direction="l",
 },
 [0xA902]={
  category="nd",
  direction="l",
 },
 [0xA903]={
  category="nd",
  direction="l",
 },
 [0xA904]={
  category="nd",
  direction="l",
 },
 [0xA905]={
  category="nd",
  direction="l",
 },
 [0xA906]={
  category="nd",
  direction="l",
 },
 [0xA907]={
  category="nd",
  direction="l",
 },
 [0xA908]={
  category="nd",
  direction="l",
 },
 [0xA909]={
  category="nd",
  direction="l",
 },
 [0xA90A]={
  category="lo",
  direction="l",
 },
 [0xA90B]={
  category="lo",
  direction="l",
 },
 [0xA90C]={
  category="lo",
  direction="l",
 },
 [0xA90D]={
  category="lo",
  direction="l",
 },
 [0xA90E]={
  category="lo",
  direction="l",
 },
 [0xA90F]={
  category="lo",
  direction="l",
 },
 [0xA910]={
  category="lo",
  direction="l",
 },
 [0xA911]={
  category="lo",
  direction="l",
 },
 [0xA912]={
  category="lo",
  direction="l",
 },
 [0xA913]={
  category="lo",
  direction="l",
 },
 [0xA914]={
  category="lo",
  direction="l",
 },
 [0xA915]={
  category="lo",
  direction="l",
 },
 [0xA916]={
  category="lo",
  direction="l",
 },
 [0xA917]={
  category="lo",
  direction="l",
 },
 [0xA918]={
  category="lo",
  direction="l",
 },
 [0xA919]={
  category="lo",
  direction="l",
 },
 [0xA91A]={
  category="lo",
  direction="l",
 },
 [0xA91B]={
  category="lo",
  direction="l",
 },
 [0xA91C]={
  category="lo",
  direction="l",
 },
 [0xA91D]={
  category="lo",
  direction="l",
 },
 [0xA91E]={
  category="lo",
  direction="l",
 },
 [0xA91F]={
  category="lo",
  direction="l",
 },
 [0xA920]={
  category="lo",
  direction="l",
 },
 [0xA921]={
  category="lo",
  direction="l",
 },
 [0xA922]={
  category="lo",
  direction="l",
 },
 [0xA923]={
  category="lo",
  direction="l",
 },
 [0xA924]={
  category="lo",
  direction="l",
 },
 [0xA925]={
  category="lo",
  direction="l",
 },
 [0xA926]={
  category="mn",
  direction="nsm",
 },
 [0xA927]={
  category="mn",
  direction="nsm",
 },
 [0xA928]={
  category="mn",
  direction="nsm",
 },
 [0xA929]={
  category="mn",
  direction="nsm",
 },
 [0xA92A]={
  category="mn",
  direction="nsm",
 },
 [0xA92B]={
  category="mn",
  direction="nsm",
 },
 [0xA92C]={
  category="mn",
  direction="nsm",
 },
 [0xA92D]={
  category="mn",
  direction="nsm",
 },
 [0xA92E]={
  category="po",
  direction="l",
 },
 [0xA92F]={
  category="po",
  direction="l",
 },
 [0xA930]={
  category="lo",
  direction="l",
 },
 [0xA931]={
  category="lo",
  direction="l",
 },
 [0xA932]={
  category="lo",
  direction="l",
 },
 [0xA933]={
  category="lo",
  direction="l",
 },
 [0xA934]={
  category="lo",
  direction="l",
 },
 [0xA935]={
  category="lo",
  direction="l",
 },
 [0xA936]={
  category="lo",
  direction="l",
 },
 [0xA937]={
  category="lo",
  direction="l",
 },
 [0xA938]={
  category="lo",
  direction="l",
 },
 [0xA939]={
  category="lo",
  direction="l",
 },
 [0xA93A]={
  category="lo",
  direction="l",
 },
 [0xA93B]={
  category="lo",
  direction="l",
 },
 [0xA93C]={
  category="lo",
  direction="l",
 },
 [0xA93D]={
  category="lo",
  direction="l",
 },
 [0xA93E]={
  category="lo",
  direction="l",
 },
 [0xA93F]={
  category="lo",
  direction="l",
 },
 [0xA940]={
  category="lo",
  direction="l",
 },
 [0xA941]={
  category="lo",
  direction="l",
 },
 [0xA942]={
  category="lo",
  direction="l",
 },
 [0xA943]={
  category="lo",
  direction="l",
 },
 [0xA944]={
  category="lo",
  direction="l",
 },
 [0xA945]={
  category="lo",
  direction="l",
 },
 [0xA946]={
  category="lo",
  direction="l",
 },
 [0xA947]={
  category="mn",
  direction="nsm",
 },
 [0xA948]={
  category="mn",
  direction="nsm",
 },
 [0xA949]={
  category="mn",
  direction="nsm",
 },
 [0xA94A]={
  category="mn",
  direction="nsm",
 },
 [0xA94B]={
  category="mn",
  direction="nsm",
 },
 [0xA94C]={
  category="mn",
  direction="nsm",
 },
 [0xA94D]={
  category="mn",
  direction="nsm",
 },
 [0xA94E]={
  category="mn",
  direction="nsm",
 },
 [0xA94F]={
  category="mn",
  direction="nsm",
 },
 [0xA950]={
  category="mn",
  direction="nsm",
 },
 [0xA951]={
  category="mn",
  direction="nsm",
 },
 [0xA952]={
  category="mc",
  direction="l",
 },
 [0xA953]={
  category="mc",
  direction="l",
 },
 [0xA95F]={
  category="po",
  direction="l",
 },
 [0xA960]={
  category="lo",
  direction="l",
 },
 [0xA961]={
  category="lo",
  direction="l",
 },
 [0xA962]={
  category="lo",
  direction="l",
 },
 [0xA963]={
  category="lo",
  direction="l",
 },
 [0xA964]={
  category="lo",
  direction="l",
 },
 [0xA965]={
  category="lo",
  direction="l",
 },
 [0xA966]={
  category="lo",
  direction="l",
 },
 [0xA967]={
  category="lo",
  direction="l",
 },
 [0xA968]={
  category="lo",
  direction="l",
 },
 [0xA969]={
  category="lo",
  direction="l",
 },
 [0xA96A]={
  category="lo",
  direction="l",
 },
 [0xA96B]={
  category="lo",
  direction="l",
 },
 [0xA96C]={
  category="lo",
  direction="l",
 },
 [0xA96D]={
  category="lo",
  direction="l",
 },
 [0xA96E]={
  category="lo",
  direction="l",
 },
 [0xA96F]={
  category="lo",
  direction="l",
 },
 [0xA970]={
  category="lo",
  direction="l",
 },
 [0xA971]={
  category="lo",
  direction="l",
 },
 [0xA972]={
  category="lo",
  direction="l",
 },
 [0xA973]={
  category="lo",
  direction="l",
 },
 [0xA974]={
  category="lo",
  direction="l",
 },
 [0xA975]={
  category="lo",
  direction="l",
 },
 [0xA976]={
  category="lo",
  direction="l",
 },
 [0xA977]={
  category="lo",
  direction="l",
 },
 [0xA978]={
  category="lo",
  direction="l",
 },
 [0xA979]={
  category="lo",
  direction="l",
 },
 [0xA97A]={
  category="lo",
  direction="l",
 },
 [0xA97B]={
  category="lo",
  direction="l",
 },
 [0xA97C]={
  category="lo",
  direction="l",
 },
 [0xA980]={
  category="mn",
  direction="nsm",
 },
 [0xA981]={
  category="mn",
  direction="nsm",
 },
 [0xA982]={
  category="mn",
  direction="nsm",
 },
 [0xA983]={
  category="mc",
  direction="l",
 },
 [0xA984]={
  category="lo",
  direction="l",
 },
 [0xA985]={
  category="lo",
  direction="l",
 },
 [0xA986]={
  category="lo",
  direction="l",
 },
 [0xA987]={
  category="lo",
  direction="l",
 },
 [0xA988]={
  category="lo",
  direction="l",
 },
 [0xA989]={
  category="lo",
  direction="l",
 },
 [0xA98A]={
  category="lo",
  direction="l",
 },
 [0xA98B]={
  category="lo",
  direction="l",
 },
 [0xA98C]={
  category="lo",
  direction="l",
 },
 [0xA98D]={
  category="lo",
  direction="l",
 },
 [0xA98E]={
  category="lo",
  direction="l",
 },
 [0xA98F]={
  category="lo",
  direction="l",
 },
 [0xA990]={
  category="lo",
  direction="l",
 },
 [0xA991]={
  category="lo",
  direction="l",
 },
 [0xA992]={
  category="lo",
  direction="l",
 },
 [0xA993]={
  category="lo",
  direction="l",
 },
 [0xA994]={
  category="lo",
  direction="l",
 },
 [0xA995]={
  category="lo",
  direction="l",
 },
 [0xA996]={
  category="lo",
  direction="l",
 },
 [0xA997]={
  category="lo",
  direction="l",
 },
 [0xA998]={
  category="lo",
  direction="l",
 },
 [0xA999]={
  category="lo",
  direction="l",
 },
 [0xA99A]={
  category="lo",
  direction="l",
 },
 [0xA99B]={
  category="lo",
  direction="l",
 },
 [0xA99C]={
  category="lo",
  direction="l",
 },
 [0xA99D]={
  category="lo",
  direction="l",
 },
 [0xA99E]={
  category="lo",
  direction="l",
 },
 [0xA99F]={
  category="lo",
  direction="l",
 },
 [0xA9A0]={
  category="lo",
  direction="l",
 },
 [0xA9A1]={
  category="lo",
  direction="l",
 },
 [0xA9A2]={
  category="lo",
  direction="l",
 },
 [0xA9A3]={
  category="lo",
  direction="l",
 },
 [0xA9A4]={
  category="lo",
  direction="l",
 },
 [0xA9A5]={
  category="lo",
  direction="l",
 },
 [0xA9A6]={
  category="lo",
  direction="l",
 },
 [0xA9A7]={
  category="lo",
  direction="l",
 },
 [0xA9A8]={
  category="lo",
  direction="l",
 },
 [0xA9A9]={
  category="lo",
  direction="l",
 },
 [0xA9AA]={
  category="lo",
  direction="l",
 },
 [0xA9AB]={
  category="lo",
  direction="l",
 },
 [0xA9AC]={
  category="lo",
  direction="l",
 },
 [0xA9AD]={
  category="lo",
  direction="l",
 },
 [0xA9AE]={
  category="lo",
  direction="l",
 },
 [0xA9AF]={
  category="lo",
  direction="l",
 },
 [0xA9B0]={
  category="lo",
  direction="l",
 },
 [0xA9B1]={
  category="lo",
  direction="l",
 },
 [0xA9B2]={
  category="lo",
  direction="l",
 },
 [0xA9B3]={
  category="mn",
  direction="nsm",
 },
 [0xA9B4]={
  category="mc",
  direction="l",
 },
 [0xA9B5]={
  category="mc",
  direction="l",
 },
 [0xA9B6]={
  category="mn",
  direction="nsm",
 },
 [0xA9B7]={
  category="mn",
  direction="nsm",
 },
 [0xA9B8]={
  category="mn",
  direction="nsm",
 },
 [0xA9B9]={
  category="mn",
  direction="nsm",
 },
 [0xA9BA]={
  category="mc",
  direction="l",
 },
 [0xA9BB]={
  category="mc",
  direction="l",
 },
 [0xA9BC]={
  category="mn",
  direction="nsm",
 },
 [0xA9BD]={
  category="mc",
  direction="l",
 },
 [0xA9BE]={
  category="mc",
  direction="l",
 },
 [0xA9BF]={
  category="mc",
  direction="l",
 },
 [0xA9C0]={
  category="mc",
  direction="l",
 },
 [0xA9C1]={
  category="po",
  direction="l",
 },
 [0xA9C2]={
  category="po",
  direction="l",
 },
 [0xA9C3]={
  category="po",
  direction="l",
 },
 [0xA9C4]={
  category="po",
  direction="l",
 },
 [0xA9C5]={
  category="po",
  direction="l",
 },
 [0xA9C6]={
  category="po",
  direction="l",
 },
 [0xA9C7]={
  category="po",
  direction="l",
 },
 [0xA9C8]={
  category="po",
  direction="l",
 },
 [0xA9C9]={
  category="po",
  direction="l",
 },
 [0xA9CA]={
  category="po",
  direction="l",
 },
 [0xA9CB]={
  category="po",
  direction="l",
 },
 [0xA9CC]={
  category="po",
  direction="l",
 },
 [0xA9CD]={
  category="po",
  direction="l",
 },
 [0xA9CF]={
  category="lm",
  direction="l",
 },
 [0xA9D0]={
  category="nd",
  direction="l",
 },
 [0xA9D1]={
  category="nd",
  direction="l",
 },
 [0xA9D2]={
  category="nd",
  direction="l",
 },
 [0xA9D3]={
  category="nd",
  direction="l",
 },
 [0xA9D4]={
  category="nd",
  direction="l",
 },
 [0xA9D5]={
  category="nd",
  direction="l",
 },
 [0xA9D6]={
  category="nd",
  direction="l",
 },
 [0xA9D7]={
  category="nd",
  direction="l",
 },
 [0xA9D8]={
  category="nd",
  direction="l",
 },
 [0xA9D9]={
  category="nd",
  direction="l",
 },
 [0xA9DE]={
  category="po",
  direction="l",
 },
 [0xA9DF]={
  category="po",
  direction="l",
 },
 [0xA9E0]={
  category="lo",
  direction="l",
 },
 [0xA9E1]={
  category="lo",
  direction="l",
 },
 [0xA9E2]={
  category="lo",
  direction="l",
 },
 [0xA9E3]={
  category="lo",
  direction="l",
 },
 [0xA9E4]={
  category="lo",
  direction="l",
 },
 [0xA9E5]={
  category="mn",
  direction="nsm",
 },
 [0xA9E6]={
  category="lm",
  direction="l",
 },
 [0xA9E7]={
  category="lo",
  direction="l",
 },
 [0xA9E8]={
  category="lo",
  direction="l",
 },
 [0xA9E9]={
  category="lo",
  direction="l",
 },
 [0xA9EA]={
  category="lo",
  direction="l",
 },
 [0xA9EB]={
  category="lo",
  direction="l",
 },
 [0xA9EC]={
  category="lo",
  direction="l",
 },
 [0xA9ED]={
  category="lo",
  direction="l",
 },
 [0xA9EE]={
  category="lo",
  direction="l",
 },
 [0xA9EF]={
  category="lo",
  direction="l",
 },
 [0xA9F0]={
  category="nd",
  direction="l",
 },
 [0xA9F1]={
  category="nd",
  direction="l",
 },
 [0xA9F2]={
  category="nd",
  direction="l",
 },
 [0xA9F3]={
  category="nd",
  direction="l",
 },
 [0xA9F4]={
  category="nd",
  direction="l",
 },
 [0xA9F5]={
  category="nd",
  direction="l",
 },
 [0xA9F6]={
  category="nd",
  direction="l",
 },
 [0xA9F7]={
  category="nd",
  direction="l",
 },
 [0xA9F8]={
  category="nd",
  direction="l",
 },
 [0xA9F9]={
  category="nd",
  direction="l",
 },
 [0xA9FA]={
  category="lo",
  direction="l",
 },
 [0xA9FB]={
  category="lo",
  direction="l",
 },
 [0xA9FC]={
  category="lo",
  direction="l",
 },
 [0xA9FD]={
  category="lo",
  direction="l",
 },
 [0xA9FE]={
  category="lo",
  direction="l",
 },
 [0xAA00]={
  category="lo",
  direction="l",
 },
 [0xAA01]={
  category="lo",
  direction="l",
 },
 [0xAA02]={
  category="lo",
  direction="l",
 },
 [0xAA03]={
  category="lo",
  direction="l",
 },
 [0xAA04]={
  category="lo",
  direction="l",
 },
 [0xAA05]={
  category="lo",
  direction="l",
 },
 [0xAA06]={
  category="lo",
  direction="l",
 },
 [0xAA07]={
  category="lo",
  direction="l",
 },
 [0xAA08]={
  category="lo",
  direction="l",
 },
 [0xAA09]={
  category="lo",
  direction="l",
 },
 [0xAA0A]={
  category="lo",
  direction="l",
 },
 [0xAA0B]={
  category="lo",
  direction="l",
 },
 [0xAA0C]={
  category="lo",
  direction="l",
 },
 [0xAA0D]={
  category="lo",
  direction="l",
 },
 [0xAA0E]={
  category="lo",
  direction="l",
 },
 [0xAA0F]={
  category="lo",
  direction="l",
 },
 [0xAA10]={
  category="lo",
  direction="l",
 },
 [0xAA11]={
  category="lo",
  direction="l",
 },
 [0xAA12]={
  category="lo",
  direction="l",
 },
 [0xAA13]={
  category="lo",
  direction="l",
 },
 [0xAA14]={
  category="lo",
  direction="l",
 },
 [0xAA15]={
  category="lo",
  direction="l",
 },
 [0xAA16]={
  category="lo",
  direction="l",
 },
 [0xAA17]={
  category="lo",
  direction="l",
 },
 [0xAA18]={
  category="lo",
  direction="l",
 },
 [0xAA19]={
  category="lo",
  direction="l",
 },
 [0xAA1A]={
  category="lo",
  direction="l",
 },
 [0xAA1B]={
  category="lo",
  direction="l",
 },
 [0xAA1C]={
  category="lo",
  direction="l",
 },
 [0xAA1D]={
  category="lo",
  direction="l",
 },
 [0xAA1E]={
  category="lo",
  direction="l",
 },
 [0xAA1F]={
  category="lo",
  direction="l",
 },
 [0xAA20]={
  category="lo",
  direction="l",
 },
 [0xAA21]={
  category="lo",
  direction="l",
 },
 [0xAA22]={
  category="lo",
  direction="l",
 },
 [0xAA23]={
  category="lo",
  direction="l",
 },
 [0xAA24]={
  category="lo",
  direction="l",
 },
 [0xAA25]={
  category="lo",
  direction="l",
 },
 [0xAA26]={
  category="lo",
  direction="l",
 },
 [0xAA27]={
  category="lo",
  direction="l",
 },
 [0xAA28]={
  category="lo",
  direction="l",
 },
 [0xAA29]={
  category="mn",
  direction="nsm",
 },
 [0xAA2A]={
  category="mn",
  direction="nsm",
 },
 [0xAA2B]={
  category="mn",
  direction="nsm",
 },
 [0xAA2C]={
  category="mn",
  direction="nsm",
 },
 [0xAA2D]={
  category="mn",
  direction="nsm",
 },
 [0xAA2E]={
  category="mn",
  direction="nsm",
 },
 [0xAA2F]={
  category="mc",
  direction="l",
 },
 [0xAA30]={
  category="mc",
  direction="l",
 },
 [0xAA31]={
  category="mn",
  direction="nsm",
 },
 [0xAA32]={
  category="mn",
  direction="nsm",
 },
 [0xAA33]={
  category="mc",
  direction="l",
 },
 [0xAA34]={
  category="mc",
  direction="l",
 },
 [0xAA35]={
  category="mn",
  direction="nsm",
 },
 [0xAA36]={
  category="mn",
  direction="nsm",
 },
 [0xAA40]={
  category="lo",
  direction="l",
 },
 [0xAA41]={
  category="lo",
  direction="l",
 },
 [0xAA42]={
  category="lo",
  direction="l",
 },
 [0xAA43]={
  category="mn",
  direction="nsm",
 },
 [0xAA44]={
  category="lo",
  direction="l",
 },
 [0xAA45]={
  category="lo",
  direction="l",
 },
 [0xAA46]={
  category="lo",
  direction="l",
 },
 [0xAA47]={
  category="lo",
  direction="l",
 },
 [0xAA48]={
  category="lo",
  direction="l",
 },
 [0xAA49]={
  category="lo",
  direction="l",
 },
 [0xAA4A]={
  category="lo",
  direction="l",
 },
 [0xAA4B]={
  category="lo",
  direction="l",
 },
 [0xAA4C]={
  category="mn",
  direction="nsm",
 },
 [0xAA4D]={
  category="mc",
  direction="l",
 },
 [0xAA50]={
  category="nd",
  direction="l",
 },
 [0xAA51]={
  category="nd",
  direction="l",
 },
 [0xAA52]={
  category="nd",
  direction="l",
 },
 [0xAA53]={
  category="nd",
  direction="l",
 },
 [0xAA54]={
  category="nd",
  direction="l",
 },
 [0xAA55]={
  category="nd",
  direction="l",
 },
 [0xAA56]={
  category="nd",
  direction="l",
 },
 [0xAA57]={
  category="nd",
  direction="l",
 },
 [0xAA58]={
  category="nd",
  direction="l",
 },
 [0xAA59]={
  category="nd",
  direction="l",
 },
 [0xAA5C]={
  category="po",
  direction="l",
 },
 [0xAA5D]={
  category="po",
  direction="l",
 },
 [0xAA5E]={
  category="po",
  direction="l",
 },
 [0xAA5F]={
  category="po",
  direction="l",
 },
 [0xAA60]={
  category="lo",
  direction="l",
 },
 [0xAA61]={
  category="lo",
  direction="l",
 },
 [0xAA62]={
  category="lo",
  direction="l",
 },
 [0xAA63]={
  category="lo",
  direction="l",
 },
 [0xAA64]={
  category="lo",
  direction="l",
 },
 [0xAA65]={
  category="lo",
  direction="l",
 },
 [0xAA66]={
  category="lo",
  direction="l",
 },
 [0xAA67]={
  category="lo",
  direction="l",
 },
 [0xAA68]={
  category="lo",
  direction="l",
 },
 [0xAA69]={
  category="lo",
  direction="l",
 },
 [0xAA6A]={
  category="lo",
  direction="l",
 },
 [0xAA6B]={
  category="lo",
  direction="l",
 },
 [0xAA6C]={
  category="lo",
  direction="l",
 },
 [0xAA6D]={
  category="lo",
  direction="l",
 },
 [0xAA6E]={
  category="lo",
  direction="l",
 },
 [0xAA6F]={
  category="lo",
  direction="l",
 },
 [0xAA70]={
  category="lm",
  direction="l",
 },
 [0xAA71]={
  category="lo",
  direction="l",
 },
 [0xAA72]={
  category="lo",
  direction="l",
 },
 [0xAA73]={
  category="lo",
  direction="l",
 },
 [0xAA74]={
  category="lo",
  direction="l",
 },
 [0xAA75]={
  category="lo",
  direction="l",
 },
 [0xAA76]={
  category="lo",
  direction="l",
 },
 [0xAA77]={
  category="so",
  direction="l",
 },
 [0xAA78]={
  category="so",
  direction="l",
 },
 [0xAA79]={
  category="so",
  direction="l",
 },
 [0xAA7A]={
  category="lo",
  direction="l",
 },
 [0xAA7B]={
  category="mc",
  direction="l",
 },
 [0xAA7C]={
  category="mn",
  direction="nsm",
 },
 [0xAA7D]={
  category="mc",
  direction="l",
 },
 [0xAA7E]={
  category="lo",
  direction="l",
 },
 [0xAA7F]={
  category="lo",
  direction="l",
 },
 [0xAA80]={
  category="lo",
  direction="l",
 },
 [0xAA81]={
  category="lo",
  direction="l",
 },
 [0xAA82]={
  category="lo",
  direction="l",
 },
 [0xAA83]={
  category="lo",
  direction="l",
 },
 [0xAA84]={
  category="lo",
  direction="l",
 },
 [0xAA85]={
  category="lo",
  direction="l",
 },
 [0xAA86]={
  category="lo",
  direction="l",
 },
 [0xAA87]={
  category="lo",
  direction="l",
 },
 [0xAA88]={
  category="lo",
  direction="l",
 },
 [0xAA89]={
  category="lo",
  direction="l",
 },
 [0xAA8A]={
  category="lo",
  direction="l",
 },
 [0xAA8B]={
  category="lo",
  direction="l",
 },
 [0xAA8C]={
  category="lo",
  direction="l",
 },
 [0xAA8D]={
  category="lo",
  direction="l",
 },
 [0xAA8E]={
  category="lo",
  direction="l",
 },
 [0xAA8F]={
  category="lo",
  direction="l",
 },
 [0xAA90]={
  category="lo",
  direction="l",
 },
 [0xAA91]={
  category="lo",
  direction="l",
 },
 [0xAA92]={
  category="lo",
  direction="l",
 },
 [0xAA93]={
  category="lo",
  direction="l",
 },
 [0xAA94]={
  category="lo",
  direction="l",
 },
 [0xAA95]={
  category="lo",
  direction="l",
 },
 [0xAA96]={
  category="lo",
  direction="l",
 },
 [0xAA97]={
  category="lo",
  direction="l",
 },
 [0xAA98]={
  category="lo",
  direction="l",
 },
 [0xAA99]={
  category="lo",
  direction="l",
 },
 [0xAA9A]={
  category="lo",
  direction="l",
 },
 [0xAA9B]={
  category="lo",
  direction="l",
 },
 [0xAA9C]={
  category="lo",
  direction="l",
 },
 [0xAA9D]={
  category="lo",
  direction="l",
 },
 [0xAA9E]={
  category="lo",
  direction="l",
 },
 [0xAA9F]={
  category="lo",
  direction="l",
 },
 [0xAAA0]={
  category="lo",
  direction="l",
 },
 [0xAAA1]={
  category="lo",
  direction="l",
 },
 [0xAAA2]={
  category="lo",
  direction="l",
 },
 [0xAAA3]={
  category="lo",
  direction="l",
 },
 [0xAAA4]={
  category="lo",
  direction="l",
 },
 [0xAAA5]={
  category="lo",
  direction="l",
 },
 [0xAAA6]={
  category="lo",
  direction="l",
 },
 [0xAAA7]={
  category="lo",
  direction="l",
 },
 [0xAAA8]={
  category="lo",
  direction="l",
 },
 [0xAAA9]={
  category="lo",
  direction="l",
 },
 [0xAAAA]={
  category="lo",
  direction="l",
 },
 [0xAAAB]={
  category="lo",
  direction="l",
 },
 [0xAAAC]={
  category="lo",
  direction="l",
 },
 [0xAAAD]={
  category="lo",
  direction="l",
 },
 [0xAAAE]={
  category="lo",
  direction="l",
 },
 [0xAAAF]={
  category="lo",
  direction="l",
 },
 [0xAAB0]={
  category="mn",
  direction="nsm",
 },
 [0xAAB1]={
  category="lo",
  direction="l",
 },
 [0xAAB2]={
  category="mn",
  direction="nsm",
 },
 [0xAAB3]={
  category="mn",
  direction="nsm",
 },
 [0xAAB4]={
  category="mn",
  direction="nsm",
 },
 [0xAAB5]={
  category="lo",
  direction="l",
 },
 [0xAAB6]={
  category="lo",
  direction="l",
 },
 [0xAAB7]={
  category="mn",
  direction="nsm",
 },
 [0xAAB8]={
  category="mn",
  direction="nsm",
 },
 [0xAAB9]={
  category="lo",
  direction="l",
 },
 [0xAABA]={
  category="lo",
  direction="l",
 },
 [0xAABB]={
  category="lo",
  direction="l",
 },
 [0xAABC]={
  category="lo",
  direction="l",
 },
 [0xAABD]={
  category="lo",
  direction="l",
 },
 [0xAABE]={
  category="mn",
  direction="nsm",
 },
 [0xAABF]={
  category="mn",
  direction="nsm",
 },
 [0xAAC0]={
  category="lo",
  direction="l",
 },
 [0xAAC1]={
  category="mn",
  direction="nsm",
 },
 [0xAAC2]={
  category="lo",
  direction="l",
 },
 [0xAADB]={
  category="lo",
  direction="l",
 },
 [0xAADC]={
  category="lo",
  direction="l",
 },
 [0xAADD]={
  category="lm",
  direction="l",
 },
 [0xAADE]={
  category="po",
  direction="l",
 },
 [0xAADF]={
  category="po",
  direction="l",
 },
 [0xAAE0]={
  category="lo",
  direction="l",
 },
 [0xAAE1]={
  category="lo",
  direction="l",
 },
 [0xAAE2]={
  category="lo",
  direction="l",
 },
 [0xAAE3]={
  category="lo",
  direction="l",
 },
 [0xAAE4]={
  category="lo",
  direction="l",
 },
 [0xAAE5]={
  category="lo",
  direction="l",
 },
 [0xAAE6]={
  category="lo",
  direction="l",
 },
 [0xAAE7]={
  category="lo",
  direction="l",
 },
 [0xAAE8]={
  category="lo",
  direction="l",
 },
 [0xAAE9]={
  category="lo",
  direction="l",
 },
 [0xAAEA]={
  category="lo",
  direction="l",
 },
 [0xAAEB]={
  category="mc",
  direction="l",
 },
 [0xAAEC]={
  category="mn",
  direction="nsm",
 },
 [0xAAED]={
  category="mn",
  direction="nsm",
 },
 [0xAAEE]={
  category="mc",
  direction="l",
 },
 [0xAAEF]={
  category="mc",
  direction="l",
 },
 [0xAAF0]={
  category="po",
  direction="l",
 },
 [0xAAF1]={
  category="po",
  direction="l",
 },
 [0xAAF2]={
  category="lo",
  direction="l",
 },
 [0xAAF3]={
  category="lm",
  direction="l",
 },
 [0xAAF4]={
  category="lm",
  direction="l",
 },
 [0xAAF5]={
  category="mc",
  direction="l",
 },
 [0xAAF6]={
  category="mn",
  direction="nsm",
 },
 [0xAB01]={
  category="lo",
  direction="l",
 },
 [0xAB02]={
  category="lo",
  direction="l",
 },
 [0xAB03]={
  category="lo",
  direction="l",
 },
 [0xAB04]={
  category="lo",
  direction="l",
 },
 [0xAB05]={
  category="lo",
  direction="l",
 },
 [0xAB06]={
  category="lo",
  direction="l",
 },
 [0xAB09]={
  category="lo",
  direction="l",
 },
 [0xAB0A]={
  category="lo",
  direction="l",
 },
 [0xAB0B]={
  category="lo",
  direction="l",
 },
 [0xAB0C]={
  category="lo",
  direction="l",
 },
 [0xAB0D]={
  category="lo",
  direction="l",
 },
 [0xAB0E]={
  category="lo",
  direction="l",
 },
 [0xAB11]={
  category="lo",
  direction="l",
 },
 [0xAB12]={
  category="lo",
  direction="l",
 },
 [0xAB13]={
  category="lo",
  direction="l",
 },
 [0xAB14]={
  category="lo",
  direction="l",
 },
 [0xAB15]={
  category="lo",
  direction="l",
 },
 [0xAB16]={
  category="lo",
  direction="l",
 },
 [0xAB20]={
  category="lo",
  direction="l",
 },
 [0xAB21]={
  category="lo",
  direction="l",
 },
 [0xAB22]={
  category="lo",
  direction="l",
 },
 [0xAB23]={
  category="lo",
  direction="l",
 },
 [0xAB24]={
  category="lo",
  direction="l",
 },
 [0xAB25]={
  category="lo",
  direction="l",
 },
 [0xAB26]={
  category="lo",
  direction="l",
 },
 [0xAB28]={
  category="lo",
  direction="l",
 },
 [0xAB29]={
  category="lo",
  direction="l",
 },
 [0xAB2A]={
  category="lo",
  direction="l",
 },
 [0xAB2B]={
  category="lo",
  direction="l",
 },
 [0xAB2C]={
  category="lo",
  direction="l",
 },
 [0xAB2D]={
  category="lo",
  direction="l",
 },
 [0xAB2E]={
  category="lo",
  direction="l",
 },
 [0xAB30]={
  category="ll",
  direction="l",
 },
 [0xAB31]={
  category="ll",
  direction="l",
 },
 [0xAB32]={
  category="ll",
  direction="l",
 },
 [0xAB33]={
  category="ll",
  direction="l",
 },
 [0xAB34]={
  category="ll",
  direction="l",
 },
 [0xAB35]={
  category="ll",
  direction="l",
 },
 [0xAB36]={
  category="ll",
  direction="l",
 },
 [0xAB37]={
  category="ll",
  direction="l",
 },
 [0xAB38]={
  category="ll",
  direction="l",
 },
 [0xAB39]={
  category="ll",
  direction="l",
 },
 [0xAB3A]={
  category="ll",
  direction="l",
 },
 [0xAB3B]={
  category="ll",
  direction="l",
 },
 [0xAB3C]={
  category="ll",
  direction="l",
 },
 [0xAB3D]={
  category="ll",
  direction="l",
 },
 [0xAB3E]={
  category="ll",
  direction="l",
 },
 [0xAB3F]={
  category="ll",
  direction="l",
 },
 [0xAB40]={
  category="ll",
  direction="l",
 },
 [0xAB41]={
  category="ll",
  direction="l",
 },
 [0xAB42]={
  category="ll",
  direction="l",
 },
 [0xAB43]={
  category="ll",
  direction="l",
 },
 [0xAB44]={
  category="ll",
  direction="l",
 },
 [0xAB45]={
  category="ll",
  direction="l",
 },
 [0xAB46]={
  category="ll",
  direction="l",
 },
 [0xAB47]={
  category="ll",
  direction="l",
 },
 [0xAB48]={
  category="ll",
  direction="l",
 },
 [0xAB49]={
  category="ll",
  direction="l",
 },
 [0xAB4A]={
  category="ll",
  direction="l",
 },
 [0xAB4B]={
  category="ll",
  direction="l",
 },
 [0xAB4C]={
  category="ll",
  direction="l",
 },
 [0xAB4D]={
  category="ll",
  direction="l",
 },
 [0xAB4E]={
  category="ll",
  direction="l",
 },
 [0xAB4F]={
  category="ll",
  direction="l",
 },
 [0xAB50]={
  category="ll",
  direction="l",
 },
 [0xAB51]={
  category="ll",
  direction="l",
 },
 [0xAB52]={
  category="ll",
  direction="l",
 },
 [0xAB53]={
  category="ll",
  direction="l",
 },
 [0xAB54]={
  category="ll",
  direction="l",
 },
 [0xAB55]={
  category="ll",
  direction="l",
 },
 [0xAB56]={
  category="ll",
  direction="l",
 },
 [0xAB57]={
  category="ll",
  direction="l",
 },
 [0xAB58]={
  category="ll",
  direction="l",
 },
 [0xAB59]={
  category="ll",
  direction="l",
 },
 [0xAB5A]={
  category="ll",
  direction="l",
 },
 [0xAB5B]={
  category="sk",
  direction="l",
 },
 [0xAB5C]={
  category="lm",
  direction="l",
 },
 [0xAB5D]={
  category="lm",
  direction="l",
 },
 [0xAB5E]={
  category="lm",
  direction="l",
 },
 [0xAB5F]={
  category="lm",
  direction="l",
 },
 [0xAB64]={
  category="ll",
  direction="l",
 },
 [0xAB65]={
  category="ll",
  direction="l",
 },
 [0xABC0]={
  category="lo",
  direction="l",
 },
 [0xABC1]={
  category="lo",
  direction="l",
 },
 [0xABC2]={
  category="lo",
  direction="l",
 },
 [0xABC3]={
  category="lo",
  direction="l",
 },
 [0xABC4]={
  category="lo",
  direction="l",
 },
 [0xABC5]={
  category="lo",
  direction="l",
 },
 [0xABC6]={
  category="lo",
  direction="l",
 },
 [0xABC7]={
  category="lo",
  direction="l",
 },
 [0xABC8]={
  category="lo",
  direction="l",
 },
 [0xABC9]={
  category="lo",
  direction="l",
 },
 [0xABCA]={
  category="lo",
  direction="l",
 },
 [0xABCB]={
  category="lo",
  direction="l",
 },
 [0xABCC]={
  category="lo",
  direction="l",
 },
 [0xABCD]={
  category="lo",
  direction="l",
 },
 [0xABCE]={
  category="lo",
  direction="l",
 },
 [0xABCF]={
  category="lo",
  direction="l",
 },
 [0xABD0]={
  category="lo",
  direction="l",
 },
 [0xABD1]={
  category="lo",
  direction="l",
 },
 [0xABD2]={
  category="lo",
  direction="l",
 },
 [0xABD3]={
  category="lo",
  direction="l",
 },
 [0xABD4]={
  category="lo",
  direction="l",
 },
 [0xABD5]={
  category="lo",
  direction="l",
 },
 [0xABD6]={
  category="lo",
  direction="l",
 },
 [0xABD7]={
  category="lo",
  direction="l",
 },
 [0xABD8]={
  category="lo",
  direction="l",
 },
 [0xABD9]={
  category="lo",
  direction="l",
 },
 [0xABDA]={
  category="lo",
  direction="l",
 },
 [0xABDB]={
  category="lo",
  direction="l",
 },
 [0xABDC]={
  category="lo",
  direction="l",
 },
 [0xABDD]={
  category="lo",
  direction="l",
 },
 [0xABDE]={
  category="lo",
  direction="l",
 },
 [0xABDF]={
  category="lo",
  direction="l",
 },
 [0xABE0]={
  category="lo",
  direction="l",
 },
 [0xABE1]={
  category="lo",
  direction="l",
 },
 [0xABE2]={
  category="lo",
  direction="l",
 },
 [0xABE3]={
  category="mc",
  direction="l",
 },
 [0xABE4]={
  category="mc",
  direction="l",
 },
 [0xABE5]={
  category="mn",
  direction="nsm",
 },
 [0xABE6]={
  category="mc",
  direction="l",
 },
 [0xABE7]={
  category="mc",
  direction="l",
 },
 [0xABE8]={
  category="mn",
  direction="nsm",
 },
 [0xABE9]={
  category="mc",
  direction="l",
 },
 [0xABEA]={
  category="mc",
  direction="l",
 },
 [0xABEB]={
  category="po",
  direction="l",
 },
 [0xABEC]={
  category="mc",
  direction="l",
 },
 [0xABED]={
  category="mn",
  direction="nsm",
 },
 [0xABF0]={
  category="nd",
  direction="l",
 },
 [0xABF1]={
  category="nd",
  direction="l",
 },
 [0xABF2]={
  category="nd",
  direction="l",
 },
 [0xABF3]={
  category="nd",
  direction="l",
 },
 [0xABF4]={
  category="nd",
  direction="l",
 },
 [0xABF5]={
  category="nd",
  direction="l",
 },
 [0xABF6]={
  category="nd",
  direction="l",
 },
 [0xABF7]={
  category="nd",
  direction="l",
 },
 [0xABF8]={
  category="nd",
  direction="l",
 },
 [0xABF9]={
  category="nd",
  direction="l",
 },
 [0xD7B0]={
  category="lo",
  direction="l",
 },
 [0xD7B1]={
  category="lo",
  direction="l",
 },
 [0xD7B2]={
  category="lo",
  direction="l",
 },
 [0xD7B3]={
  category="lo",
  direction="l",
 },
 [0xD7B4]={
  category="lo",
  direction="l",
 },
 [0xD7B5]={
  category="lo",
  direction="l",
 },
 [0xD7B6]={
  category="lo",
  direction="l",
 },
 [0xD7B7]={
  category="lo",
  direction="l",
 },
 [0xD7B8]={
  category="lo",
  direction="l",
 },
 [0xD7B9]={
  category="lo",
  direction="l",
 },
 [0xD7BA]={
  category="lo",
  direction="l",
 },
 [0xD7BB]={
  category="lo",
  direction="l",
 },
 [0xD7BC]={
  category="lo",
  direction="l",
 },
 [0xD7BD]={
  category="lo",
  direction="l",
 },
 [0xD7BE]={
  category="lo",
  direction="l",
 },
 [0xD7BF]={
  category="lo",
  direction="l",
 },
 [0xD7C0]={
  category="lo",
  direction="l",
 },
 [0xD7C1]={
  category="lo",
  direction="l",
 },
 [0xD7C2]={
  category="lo",
  direction="l",
 },
 [0xD7C3]={
  category="lo",
  direction="l",
 },
 [0xD7C4]={
  category="lo",
  direction="l",
 },
 [0xD7C5]={
  category="lo",
  direction="l",
 },
 [0xD7C6]={
  category="lo",
  direction="l",
 },
 [0xD7CB]={
  category="lo",
  direction="l",
 },
 [0xD7CC]={
  category="lo",
  direction="l",
 },
 [0xD7CD]={
  category="lo",
  direction="l",
 },
 [0xD7CE]={
  category="lo",
  direction="l",
 },
 [0xD7CF]={
  category="lo",
  direction="l",
 },
 [0xD7D0]={
  category="lo",
  direction="l",
 },
 [0xD7D1]={
  category="lo",
  direction="l",
 },
 [0xD7D2]={
  category="lo",
  direction="l",
 },
 [0xD7D3]={
  category="lo",
  direction="l",
 },
 [0xD7D4]={
  category="lo",
  direction="l",
 },
 [0xD7D5]={
  category="lo",
  direction="l",
 },
 [0xD7D6]={
  category="lo",
  direction="l",
 },
 [0xD7D7]={
  category="lo",
  direction="l",
 },
 [0xD7D8]={
  category="lo",
  direction="l",
 },
 [0xD7D9]={
  category="lo",
  direction="l",
 },
 [0xD7DA]={
  category="lo",
  direction="l",
 },
 [0xD7DB]={
  category="lo",
  direction="l",
 },
 [0xD7DC]={
  category="lo",
  direction="l",
 },
 [0xD7DD]={
  category="lo",
  direction="l",
 },
 [0xD7DE]={
  category="lo",
  direction="l",
 },
 [0xD7DF]={
  category="lo",
  direction="l",
 },
 [0xD7E0]={
  category="lo",
  direction="l",
 },
 [0xD7E1]={
  category="lo",
  direction="l",
 },
 [0xD7E2]={
  category="lo",
  direction="l",
 },
 [0xD7E3]={
  category="lo",
  direction="l",
 },
 [0xD7E4]={
  category="lo",
  direction="l",
 },
 [0xD7E5]={
  category="lo",
  direction="l",
 },
 [0xD7E6]={
  category="lo",
  direction="l",
 },
 [0xD7E7]={
  category="lo",
  direction="l",
 },
 [0xD7E8]={
  category="lo",
  direction="l",
 },
 [0xD7E9]={
  category="lo",
  direction="l",
 },
 [0xD7EA]={
  category="lo",
  direction="l",
 },
 [0xD7EB]={
  category="lo",
  direction="l",
 },
 [0xD7EC]={
  category="lo",
  direction="l",
 },
 [0xD7ED]={
  category="lo",
  direction="l",
 },
 [0xD7EE]={
  category="lo",
  direction="l",
 },
 [0xD7EF]={
  category="lo",
  direction="l",
 },
 [0xD7F0]={
  category="lo",
  direction="l",
 },
 [0xD7F1]={
  category="lo",
  direction="l",
 },
 [0xD7F2]={
  category="lo",
  direction="l",
 },
 [0xD7F3]={
  category="lo",
  direction="l",
 },
 [0xD7F4]={
  category="lo",
  direction="l",
 },
 [0xD7F5]={
  category="lo",
  direction="l",
 },
 [0xD7F6]={
  category="lo",
  direction="l",
 },
 [0xD7F7]={
  category="lo",
  direction="l",
 },
 [0xD7F8]={
  category="lo",
  direction="l",
 },
 [0xD7F9]={
  category="lo",
  direction="l",
 },
 [0xD7FA]={
  category="lo",
  direction="l",
 },
 [0xD7FB]={
  category="lo",
  direction="l",
 },
 [0xD800]={
  category="cs",
  direction="l",
 },
 [0xDB7F]={
  category="cs",
  direction="l",
 },
 [0xDB80]={
  category="cs",
  direction="l",
 },
 [0xDBFF]={
  category="cs",
  direction="l",
 },
 [0xDC00]={
  category="cs",
  direction="l",
 },
 [0xDFFF]={
  category="cs",
  direction="l",
 },
 [0xE000]={
  category="co",
  direction="l",
 },
 [0xF8FF]={
  category="co",
  direction="l",
 },
 [0xF900]={
  category="lo",
  direction="l",
 },
 [0xF901]={
  category="lo",
  direction="l",
 },
 [0xF902]={
  category="lo",
  direction="l",
 },
 [0xF903]={
  category="lo",
  direction="l",
 },
 [0xF904]={
  category="lo",
  direction="l",
 },
 [0xF905]={
  category="lo",
  direction="l",
 },
 [0xF906]={
  category="lo",
  direction="l",
 },
 [0xF907]={
  category="lo",
  direction="l",
 },
 [0xF908]={
  category="lo",
  direction="l",
 },
 [0xF909]={
  category="lo",
  direction="l",
 },
 [0xF90A]={
  category="lo",
  direction="l",
 },
 [0xF90B]={
  category="lo",
  direction="l",
 },
 [0xF90C]={
  category="lo",
  direction="l",
 },
 [0xF90D]={
  category="lo",
  direction="l",
 },
 [0xF90E]={
  category="lo",
  direction="l",
 },
 [0xF90F]={
  category="lo",
  direction="l",
 },
 [0xF910]={
  category="lo",
  direction="l",
 },
 [0xF911]={
  category="lo",
  direction="l",
 },
 [0xF912]={
  category="lo",
  direction="l",
 },
 [0xF913]={
  category="lo",
  direction="l",
 },
 [0xF914]={
  category="lo",
  direction="l",
 },
 [0xF915]={
  category="lo",
  direction="l",
 },
 [0xF916]={
  category="lo",
  direction="l",
 },
 [0xF917]={
  category="lo",
  direction="l",
 },
 [0xF918]={
  category="lo",
  direction="l",
 },
 [0xF919]={
  category="lo",
  direction="l",
 },
 [0xF91A]={
  category="lo",
  direction="l",
 },
 [0xF91B]={
  category="lo",
  direction="l",
 },
 [0xF91C]={
  category="lo",
  direction="l",
 },
 [0xF91D]={
  category="lo",
  direction="l",
 },
 [0xF91E]={
  category="lo",
  direction="l",
 },
 [0xF91F]={
  category="lo",
  direction="l",
 },
 [0xF920]={
  category="lo",
  direction="l",
 },
 [0xF921]={
  category="lo",
  direction="l",
 },
 [0xF922]={
  category="lo",
  direction="l",
 },
 [0xF923]={
  category="lo",
  direction="l",
 },
 [0xF924]={
  category="lo",
  direction="l",
 },
 [0xF925]={
  category="lo",
  direction="l",
 },
 [0xF926]={
  category="lo",
  direction="l",
 },
 [0xF927]={
  category="lo",
  direction="l",
 },
 [0xF928]={
  category="lo",
  direction="l",
 },
 [0xF929]={
  category="lo",
  direction="l",
 },
 [0xF92A]={
  category="lo",
  direction="l",
 },
 [0xF92B]={
  category="lo",
  direction="l",
 },
 [0xF92C]={
  category="lo",
  direction="l",
 },
 [0xF92D]={
  category="lo",
  direction="l",
 },
 [0xF92E]={
  category="lo",
  direction="l",
 },
 [0xF92F]={
  category="lo",
  direction="l",
 },
 [0xF930]={
  category="lo",
  direction="l",
 },
 [0xF931]={
  category="lo",
  direction="l",
 },
 [0xF932]={
  category="lo",
  direction="l",
 },
 [0xF933]={
  category="lo",
  direction="l",
 },
 [0xF934]={
  category="lo",
  direction="l",
 },
 [0xF935]={
  category="lo",
  direction="l",
 },
 [0xF936]={
  category="lo",
  direction="l",
 },
 [0xF937]={
  category="lo",
  direction="l",
 },
 [0xF938]={
  category="lo",
  direction="l",
 },
 [0xF939]={
  category="lo",
  direction="l",
 },
 [0xF93A]={
  category="lo",
  direction="l",
 },
 [0xF93B]={
  category="lo",
  direction="l",
 },
 [0xF93C]={
  category="lo",
  direction="l",
 },
 [0xF93D]={
  category="lo",
  direction="l",
 },
 [0xF93E]={
  category="lo",
  direction="l",
 },
 [0xF93F]={
  category="lo",
  direction="l",
 },
 [0xF940]={
  category="lo",
  direction="l",
 },
 [0xF941]={
  category="lo",
  direction="l",
 },
 [0xF942]={
  category="lo",
  direction="l",
 },
 [0xF943]={
  category="lo",
  direction="l",
 },
 [0xF944]={
  category="lo",
  direction="l",
 },
 [0xF945]={
  category="lo",
  direction="l",
 },
 [0xF946]={
  category="lo",
  direction="l",
 },
 [0xF947]={
  category="lo",
  direction="l",
 },
 [0xF948]={
  category="lo",
  direction="l",
 },
 [0xF949]={
  category="lo",
  direction="l",
 },
 [0xF94A]={
  category="lo",
  direction="l",
 },
 [0xF94B]={
  category="lo",
  direction="l",
 },
 [0xF94C]={
  category="lo",
  direction="l",
 },
 [0xF94D]={
  category="lo",
  direction="l",
 },
 [0xF94E]={
  category="lo",
  direction="l",
 },
 [0xF94F]={
  category="lo",
  direction="l",
 },
 [0xF950]={
  category="lo",
  direction="l",
 },
 [0xF951]={
  category="lo",
  direction="l",
 },
 [0xF952]={
  category="lo",
  direction="l",
 },
 [0xF953]={
  category="lo",
  direction="l",
 },
 [0xF954]={
  category="lo",
  direction="l",
 },
 [0xF955]={
  category="lo",
  direction="l",
 },
 [0xF956]={
  category="lo",
  direction="l",
 },
 [0xF957]={
  category="lo",
  direction="l",
 },
 [0xF958]={
  category="lo",
  direction="l",
 },
 [0xF959]={
  category="lo",
  direction="l",
 },
 [0xF95A]={
  category="lo",
  direction="l",
 },
 [0xF95B]={
  category="lo",
  direction="l",
 },
 [0xF95C]={
  category="lo",
  direction="l",
 },
 [0xF95D]={
  category="lo",
  direction="l",
 },
 [0xF95E]={
  category="lo",
  direction="l",
 },
 [0xF95F]={
  category="lo",
  direction="l",
 },
 [0xF960]={
  category="lo",
  direction="l",
 },
 [0xF961]={
  category="lo",
  direction="l",
 },
 [0xF962]={
  category="lo",
  direction="l",
 },
 [0xF963]={
  category="lo",
  direction="l",
 },
 [0xF964]={
  category="lo",
  direction="l",
 },
 [0xF965]={
  category="lo",
  direction="l",
 },
 [0xF966]={
  category="lo",
  direction="l",
 },
 [0xF967]={
  category="lo",
  direction="l",
 },
 [0xF968]={
  category="lo",
  direction="l",
 },
 [0xF969]={
  category="lo",
  direction="l",
 },
 [0xF96A]={
  category="lo",
  direction="l",
 },
 [0xF96B]={
  category="lo",
  direction="l",
 },
 [0xF96C]={
  category="lo",
  direction="l",
 },
 [0xF96D]={
  category="lo",
  direction="l",
 },
 [0xF96E]={
  category="lo",
  direction="l",
 },
 [0xF96F]={
  category="lo",
  direction="l",
 },
 [0xF970]={
  category="lo",
  direction="l",
 },
 [0xF971]={
  category="lo",
  direction="l",
 },
 [0xF972]={
  category="lo",
  direction="l",
 },
 [0xF973]={
  category="lo",
  direction="l",
 },
 [0xF974]={
  category="lo",
  direction="l",
 },
 [0xF975]={
  category="lo",
  direction="l",
 },
 [0xF976]={
  category="lo",
  direction="l",
 },
 [0xF977]={
  category="lo",
  direction="l",
 },
 [0xF978]={
  category="lo",
  direction="l",
 },
 [0xF979]={
  category="lo",
  direction="l",
 },
 [0xF97A]={
  category="lo",
  direction="l",
 },
 [0xF97B]={
  category="lo",
  direction="l",
 },
 [0xF97C]={
  category="lo",
  direction="l",
 },
 [0xF97D]={
  category="lo",
  direction="l",
 },
 [0xF97E]={
  category="lo",
  direction="l",
 },
 [0xF97F]={
  category="lo",
  direction="l",
 },
 [0xF980]={
  category="lo",
  direction="l",
 },
 [0xF981]={
  category="lo",
  direction="l",
 },
 [0xF982]={
  category="lo",
  direction="l",
 },
 [0xF983]={
  category="lo",
  direction="l",
 },
 [0xF984]={
  category="lo",
  direction="l",
 },
 [0xF985]={
  category="lo",
  direction="l",
 },
 [0xF986]={
  category="lo",
  direction="l",
 },
 [0xF987]={
  category="lo",
  direction="l",
 },
 [0xF988]={
  category="lo",
  direction="l",
 },
 [0xF989]={
  category="lo",
  direction="l",
 },
 [0xF98A]={
  category="lo",
  direction="l",
 },
 [0xF98B]={
  category="lo",
  direction="l",
 },
 [0xF98C]={
  category="lo",
  direction="l",
 },
 [0xF98D]={
  category="lo",
  direction="l",
 },
 [0xF98E]={
  category="lo",
  direction="l",
 },
 [0xF98F]={
  category="lo",
  direction="l",
 },
 [0xF990]={
  category="lo",
  direction="l",
 },
 [0xF991]={
  category="lo",
  direction="l",
 },
 [0xF992]={
  category="lo",
  direction="l",
 },
 [0xF993]={
  category="lo",
  direction="l",
 },
 [0xF994]={
  category="lo",
  direction="l",
 },
 [0xF995]={
  category="lo",
  direction="l",
 },
 [0xF996]={
  category="lo",
  direction="l",
 },
 [0xF997]={
  category="lo",
  direction="l",
 },
 [0xF998]={
  category="lo",
  direction="l",
 },
 [0xF999]={
  category="lo",
  direction="l",
 },
 [0xF99A]={
  category="lo",
  direction="l",
 },
 [0xF99B]={
  category="lo",
  direction="l",
 },
 [0xF99C]={
  category="lo",
  direction="l",
 },
 [0xF99D]={
  category="lo",
  direction="l",
 },
 [0xF99E]={
  category="lo",
  direction="l",
 },
 [0xF99F]={
  category="lo",
  direction="l",
 },
 [0xF9A0]={
  category="lo",
  direction="l",
 },
 [0xF9A1]={
  category="lo",
  direction="l",
 },
 [0xF9A2]={
  category="lo",
  direction="l",
 },
 [0xF9A3]={
  category="lo",
  direction="l",
 },
 [0xF9A4]={
  category="lo",
  direction="l",
 },
 [0xF9A5]={
  category="lo",
  direction="l",
 },
 [0xF9A6]={
  category="lo",
  direction="l",
 },
 [0xF9A7]={
  category="lo",
  direction="l",
 },
 [0xF9A8]={
  category="lo",
  direction="l",
 },
 [0xF9A9]={
  category="lo",
  direction="l",
 },
 [0xF9AA]={
  category="lo",
  direction="l",
 },
 [0xF9AB]={
  category="lo",
  direction="l",
 },
 [0xF9AC]={
  category="lo",
  direction="l",
 },
 [0xF9AD]={
  category="lo",
  direction="l",
 },
 [0xF9AE]={
  category="lo",
  direction="l",
 },
 [0xF9AF]={
  category="lo",
  direction="l",
 },
 [0xF9B0]={
  category="lo",
  direction="l",
 },
 [0xF9B1]={
  category="lo",
  direction="l",
 },
 [0xF9B2]={
  category="lo",
  direction="l",
 },
 [0xF9B3]={
  category="lo",
  direction="l",
 },
 [0xF9B4]={
  category="lo",
  direction="l",
 },
 [0xF9B5]={
  category="lo",
  direction="l",
 },
 [0xF9B6]={
  category="lo",
  direction="l",
 },
 [0xF9B7]={
  category="lo",
  direction="l",
 },
 [0xF9B8]={
  category="lo",
  direction="l",
 },
 [0xF9B9]={
  category="lo",
  direction="l",
 },
 [0xF9BA]={
  category="lo",
  direction="l",
 },
 [0xF9BB]={
  category="lo",
  direction="l",
 },
 [0xF9BC]={
  category="lo",
  direction="l",
 },
 [0xF9BD]={
  category="lo",
  direction="l",
 },
 [0xF9BE]={
  category="lo",
  direction="l",
 },
 [0xF9BF]={
  category="lo",
  direction="l",
 },
 [0xF9C0]={
  category="lo",
  direction="l",
 },
 [0xF9C1]={
  category="lo",
  direction="l",
 },
 [0xF9C2]={
  category="lo",
  direction="l",
 },
 [0xF9C3]={
  category="lo",
  direction="l",
 },
 [0xF9C4]={
  category="lo",
  direction="l",
 },
 [0xF9C5]={
  category="lo",
  direction="l",
 },
 [0xF9C6]={
  category="lo",
  direction="l",
 },
 [0xF9C7]={
  category="lo",
  direction="l",
 },
 [0xF9C8]={
  category="lo",
  direction="l",
 },
 [0xF9C9]={
  category="lo",
  direction="l",
 },
 [0xF9CA]={
  category="lo",
  direction="l",
 },
 [0xF9CB]={
  category="lo",
  direction="l",
 },
 [0xF9CC]={
  category="lo",
  direction="l",
 },
 [0xF9CD]={
  category="lo",
  direction="l",
 },
 [0xF9CE]={
  category="lo",
  direction="l",
 },
 [0xF9CF]={
  category="lo",
  direction="l",
 },
 [0xF9D0]={
  category="lo",
  direction="l",
 },
 [0xF9D1]={
  category="lo",
  direction="l",
 },
 [0xF9D2]={
  category="lo",
  direction="l",
 },
 [0xF9D3]={
  category="lo",
  direction="l",
 },
 [0xF9D4]={
  category="lo",
  direction="l",
 },
 [0xF9D5]={
  category="lo",
  direction="l",
 },
 [0xF9D6]={
  category="lo",
  direction="l",
 },
 [0xF9D7]={
  category="lo",
  direction="l",
 },
 [0xF9D8]={
  category="lo",
  direction="l",
 },
 [0xF9D9]={
  category="lo",
  direction="l",
 },
 [0xF9DA]={
  category="lo",
  direction="l",
 },
 [0xF9DB]={
  category="lo",
  direction="l",
 },
 [0xF9DC]={
  category="lo",
  direction="l",
 },
 [0xF9DD]={
  category="lo",
  direction="l",
 },
 [0xF9DE]={
  category="lo",
  direction="l",
 },
 [0xF9DF]={
  category="lo",
  direction="l",
 },
 [0xF9E0]={
  category="lo",
  direction="l",
 },
 [0xF9E1]={
  category="lo",
  direction="l",
 },
 [0xF9E2]={
  category="lo",
  direction="l",
 },
 [0xF9E3]={
  category="lo",
  direction="l",
 },
 [0xF9E4]={
  category="lo",
  direction="l",
 },
 [0xF9E5]={
  category="lo",
  direction="l",
 },
 [0xF9E6]={
  category="lo",
  direction="l",
 },
 [0xF9E7]={
  category="lo",
  direction="l",
 },
 [0xF9E8]={
  category="lo",
  direction="l",
 },
 [0xF9E9]={
  category="lo",
  direction="l",
 },
 [0xF9EA]={
  category="lo",
  direction="l",
 },
 [0xF9EB]={
  category="lo",
  direction="l",
 },
 [0xF9EC]={
  category="lo",
  direction="l",
 },
 [0xF9ED]={
  category="lo",
  direction="l",
 },
 [0xF9EE]={
  category="lo",
  direction="l",
 },
 [0xF9EF]={
  category="lo",
  direction="l",
 },
 [0xF9F0]={
  category="lo",
  direction="l",
 },
 [0xF9F1]={
  category="lo",
  direction="l",
 },
 [0xF9F2]={
  category="lo",
  direction="l",
 },
 [0xF9F3]={
  category="lo",
  direction="l",
 },
 [0xF9F4]={
  category="lo",
  direction="l",
 },
 [0xF9F5]={
  category="lo",
  direction="l",
 },
 [0xF9F6]={
  category="lo",
  direction="l",
 },
 [0xF9F7]={
  category="lo",
  direction="l",
 },
 [0xF9F8]={
  category="lo",
  direction="l",
 },
 [0xF9F9]={
  category="lo",
  direction="l",
 },
 [0xF9FA]={
  category="lo",
  direction="l",
 },
 [0xF9FB]={
  category="lo",
  direction="l",
 },
 [0xF9FC]={
  category="lo",
  direction="l",
 },
 [0xF9FD]={
  category="lo",
  direction="l",
 },
 [0xF9FE]={
  category="lo",
  direction="l",
 },
 [0xF9FF]={
  category="lo",
  direction="l",
 },
 [0xFA00]={
  category="lo",
  direction="l",
 },
 [0xFA01]={
  category="lo",
  direction="l",
 },
 [0xFA02]={
  category="lo",
  direction="l",
 },
 [0xFA03]={
  category="lo",
  direction="l",
 },
 [0xFA04]={
  category="lo",
  direction="l",
 },
 [0xFA05]={
  category="lo",
  direction="l",
 },
 [0xFA06]={
  category="lo",
  direction="l",
 },
 [0xFA07]={
  category="lo",
  direction="l",
 },
 [0xFA08]={
  category="lo",
  direction="l",
 },
 [0xFA09]={
  category="lo",
  direction="l",
 },
 [0xFA0A]={
  category="lo",
  direction="l",
 },
 [0xFA0B]={
  category="lo",
  direction="l",
 },
 [0xFA0C]={
  category="lo",
  direction="l",
 },
 [0xFA0D]={
  category="lo",
  direction="l",
 },
 [0xFA0E]={
  category="lo",
  direction="l",
 },
 [0xFA0F]={
  category="lo",
  direction="l",
 },
 [0xFA10]={
  category="lo",
  direction="l",
 },
 [0xFA11]={
  category="lo",
  direction="l",
 },
 [0xFA12]={
  category="lo",
  direction="l",
 },
 [0xFA13]={
  category="lo",
  direction="l",
 },
 [0xFA14]={
  category="lo",
  direction="l",
 },
 [0xFA15]={
  category="lo",
  direction="l",
 },
 [0xFA16]={
  category="lo",
  direction="l",
 },
 [0xFA17]={
  category="lo",
  direction="l",
 },
 [0xFA18]={
  category="lo",
  direction="l",
 },
 [0xFA19]={
  category="lo",
  direction="l",
 },
 [0xFA1A]={
  category="lo",
  direction="l",
 },
 [0xFA1B]={
  category="lo",
  direction="l",
 },
 [0xFA1C]={
  category="lo",
  direction="l",
 },
 [0xFA1D]={
  category="lo",
  direction="l",
 },
 [0xFA1E]={
  category="lo",
  direction="l",
 },
 [0xFA1F]={
  category="lo",
  direction="l",
 },
 [0xFA20]={
  category="lo",
  direction="l",
 },
 [0xFA21]={
  category="lo",
  direction="l",
 },
 [0xFA22]={
  category="lo",
  direction="l",
 },
 [0xFA23]={
  category="lo",
  direction="l",
 },
 [0xFA24]={
  category="lo",
  direction="l",
 },
 [0xFA25]={
  category="lo",
  direction="l",
 },
 [0xFA26]={
  category="lo",
  direction="l",
 },
 [0xFA27]={
  category="lo",
  direction="l",
 },
 [0xFA28]={
  category="lo",
  direction="l",
 },
 [0xFA29]={
  category="lo",
  direction="l",
 },
 [0xFA2A]={
  category="lo",
  direction="l",
 },
 [0xFA2B]={
  category="lo",
  direction="l",
 },
 [0xFA2C]={
  category="lo",
  direction="l",
 },
 [0xFA2D]={
  category="lo",
  direction="l",
 },
 [0xFA2E]={
  category="lo",
  direction="l",
 },
 [0xFA2F]={
  category="lo",
  direction="l",
 },
 [0xFA30]={
  category="lo",
  direction="l",
 },
 [0xFA31]={
  category="lo",
  direction="l",
 },
 [0xFA32]={
  category="lo",
  direction="l",
 },
 [0xFA33]={
  category="lo",
  direction="l",
 },
 [0xFA34]={
  category="lo",
  direction="l",
 },
 [0xFA35]={
  category="lo",
  direction="l",
 },
 [0xFA36]={
  category="lo",
  direction="l",
 },
 [0xFA37]={
  category="lo",
  direction="l",
 },
 [0xFA38]={
  category="lo",
  direction="l",
 },
 [0xFA39]={
  category="lo",
  direction="l",
 },
 [0xFA3A]={
  category="lo",
  direction="l",
 },
 [0xFA3B]={
  category="lo",
  direction="l",
 },
 [0xFA3C]={
  category="lo",
  direction="l",
 },
 [0xFA3D]={
  category="lo",
  direction="l",
 },
 [0xFA3E]={
  category="lo",
  direction="l",
 },
 [0xFA3F]={
  category="lo",
  direction="l",
 },
 [0xFA40]={
  category="lo",
  direction="l",
 },
 [0xFA41]={
  category="lo",
  direction="l",
 },
 [0xFA42]={
  category="lo",
  direction="l",
 },
 [0xFA43]={
  category="lo",
  direction="l",
 },
 [0xFA44]={
  category="lo",
  direction="l",
 },
 [0xFA45]={
  category="lo",
  direction="l",
 },
 [0xFA46]={
  category="lo",
  direction="l",
 },
 [0xFA47]={
  category="lo",
  direction="l",
 },
 [0xFA48]={
  category="lo",
  direction="l",
 },
 [0xFA49]={
  category="lo",
  direction="l",
 },
 [0xFA4A]={
  category="lo",
  direction="l",
 },
 [0xFA4B]={
  category="lo",
  direction="l",
 },
 [0xFA4C]={
  category="lo",
  direction="l",
 },
 [0xFA4D]={
  category="lo",
  direction="l",
 },
 [0xFA4E]={
  category="lo",
  direction="l",
 },
 [0xFA4F]={
  category="lo",
  direction="l",
 },
 [0xFA50]={
  category="lo",
  direction="l",
 },
 [0xFA51]={
  category="lo",
  direction="l",
 },
 [0xFA52]={
  category="lo",
  direction="l",
 },
 [0xFA53]={
  category="lo",
  direction="l",
 },
 [0xFA54]={
  category="lo",
  direction="l",
 },
 [0xFA55]={
  category="lo",
  direction="l",
 },
 [0xFA56]={
  category="lo",
  direction="l",
 },
 [0xFA57]={
  category="lo",
  direction="l",
 },
 [0xFA58]={
  category="lo",
  direction="l",
 },
 [0xFA59]={
  category="lo",
  direction="l",
 },
 [0xFA5A]={
  category="lo",
  direction="l",
 },
 [0xFA5B]={
  category="lo",
  direction="l",
 },
 [0xFA5C]={
  category="lo",
  direction="l",
 },
 [0xFA5D]={
  category="lo",
  direction="l",
 },
 [0xFA5E]={
  category="lo",
  direction="l",
 },
 [0xFA5F]={
  category="lo",
  direction="l",
 },
 [0xFA60]={
  category="lo",
  direction="l",
 },
 [0xFA61]={
  category="lo",
  direction="l",
 },
 [0xFA62]={
  category="lo",
  direction="l",
 },
 [0xFA63]={
  category="lo",
  direction="l",
 },
 [0xFA64]={
  category="lo",
  direction="l",
 },
 [0xFA65]={
  category="lo",
  direction="l",
 },
 [0xFA66]={
  category="lo",
  direction="l",
 },
 [0xFA67]={
  category="lo",
  direction="l",
 },
 [0xFA68]={
  category="lo",
  direction="l",
 },
 [0xFA69]={
  category="lo",
  direction="l",
 },
 [0xFA6A]={
  category="lo",
  direction="l",
 },
 [0xFA6B]={
  category="lo",
  direction="l",
 },
 [0xFA6C]={
  category="lo",
  direction="l",
 },
 [0xFA6D]={
  category="lo",
  direction="l",
 },
 [0xFA70]={
  category="lo",
  direction="l",
 },
 [0xFA71]={
  category="lo",
  direction="l",
 },
 [0xFA72]={
  category="lo",
  direction="l",
 },
 [0xFA73]={
  category="lo",
  direction="l",
 },
 [0xFA74]={
  category="lo",
  direction="l",
 },
 [0xFA75]={
  category="lo",
  direction="l",
 },
 [0xFA76]={
  category="lo",
  direction="l",
 },
 [0xFA77]={
  category="lo",
  direction="l",
 },
 [0xFA78]={
  category="lo",
  direction="l",
 },
 [0xFA79]={
  category="lo",
  direction="l",
 },
 [0xFA7A]={
  category="lo",
  direction="l",
 },
 [0xFA7B]={
  category="lo",
  direction="l",
 },
 [0xFA7C]={
  category="lo",
  direction="l",
 },
 [0xFA7D]={
  category="lo",
  direction="l",
 },
 [0xFA7E]={
  category="lo",
  direction="l",
 },
 [0xFA7F]={
  category="lo",
  direction="l",
 },
 [0xFA80]={
  category="lo",
  direction="l",
 },
 [0xFA81]={
  category="lo",
  direction="l",
 },
 [0xFA82]={
  category="lo",
  direction="l",
 },
 [0xFA83]={
  category="lo",
  direction="l",
 },
 [0xFA84]={
  category="lo",
  direction="l",
 },
 [0xFA85]={
  category="lo",
  direction="l",
 },
 [0xFA86]={
  category="lo",
  direction="l",
 },
 [0xFA87]={
  category="lo",
  direction="l",
 },
 [0xFA88]={
  category="lo",
  direction="l",
 },
 [0xFA89]={
  category="lo",
  direction="l",
 },
 [0xFA8A]={
  category="lo",
  direction="l",
 },
 [0xFA8B]={
  category="lo",
  direction="l",
 },
 [0xFA8C]={
  category="lo",
  direction="l",
 },
 [0xFA8D]={
  category="lo",
  direction="l",
 },
 [0xFA8E]={
  category="lo",
  direction="l",
 },
 [0xFA8F]={
  category="lo",
  direction="l",
 },
 [0xFA90]={
  category="lo",
  direction="l",
 },
 [0xFA91]={
  category="lo",
  direction="l",
 },
 [0xFA92]={
  category="lo",
  direction="l",
 },
 [0xFA93]={
  category="lo",
  direction="l",
 },
 [0xFA94]={
  category="lo",
  direction="l",
 },
 [0xFA95]={
  category="lo",
  direction="l",
 },
 [0xFA96]={
  category="lo",
  direction="l",
 },
 [0xFA97]={
  category="lo",
  direction="l",
 },
 [0xFA98]={
  category="lo",
  direction="l",
 },
 [0xFA99]={
  category="lo",
  direction="l",
 },
 [0xFA9A]={
  category="lo",
  direction="l",
 },
 [0xFA9B]={
  category="lo",
  direction="l",
 },
 [0xFA9C]={
  category="lo",
  direction="l",
 },
 [0xFA9D]={
  category="lo",
  direction="l",
 },
 [0xFA9E]={
  category="lo",
  direction="l",
 },
 [0xFA9F]={
  category="lo",
  direction="l",
 },
 [0xFAA0]={
  category="lo",
  direction="l",
 },
 [0xFAA1]={
  category="lo",
  direction="l",
 },
 [0xFAA2]={
  category="lo",
  direction="l",
 },
 [0xFAA3]={
  category="lo",
  direction="l",
 },
 [0xFAA4]={
  category="lo",
  direction="l",
 },
 [0xFAA5]={
  category="lo",
  direction="l",
 },
 [0xFAA6]={
  category="lo",
  direction="l",
 },
 [0xFAA7]={
  category="lo",
  direction="l",
 },
 [0xFAA8]={
  category="lo",
  direction="l",
 },
 [0xFAA9]={
  category="lo",
  direction="l",
 },
 [0xFAAA]={
  category="lo",
  direction="l",
 },
 [0xFAAB]={
  category="lo",
  direction="l",
 },
 [0xFAAC]={
  category="lo",
  direction="l",
 },
 [0xFAAD]={
  category="lo",
  direction="l",
 },
 [0xFAAE]={
  category="lo",
  direction="l",
 },
 [0xFAAF]={
  category="lo",
  direction="l",
 },
 [0xFAB0]={
  category="lo",
  direction="l",
 },
 [0xFAB1]={
  category="lo",
  direction="l",
 },
 [0xFAB2]={
  category="lo",
  direction="l",
 },
 [0xFAB3]={
  category="lo",
  direction="l",
 },
 [0xFAB4]={
  category="lo",
  direction="l",
 },
 [0xFAB5]={
  category="lo",
  direction="l",
 },
 [0xFAB6]={
  category="lo",
  direction="l",
 },
 [0xFAB7]={
  category="lo",
  direction="l",
 },
 [0xFAB8]={
  category="lo",
  direction="l",
 },
 [0xFAB9]={
  category="lo",
  direction="l",
 },
 [0xFABA]={
  category="lo",
  direction="l",
 },
 [0xFABB]={
  category="lo",
  direction="l",
 },
 [0xFABC]={
  category="lo",
  direction="l",
 },
 [0xFABD]={
  category="lo",
  direction="l",
 },
 [0xFABE]={
  category="lo",
  direction="l",
 },
 [0xFABF]={
  category="lo",
  direction="l",
 },
 [0xFAC0]={
  category="lo",
  direction="l",
 },
 [0xFAC1]={
  category="lo",
  direction="l",
 },
 [0xFAC2]={
  category="lo",
  direction="l",
 },
 [0xFAC3]={
  category="lo",
  direction="l",
 },
 [0xFAC4]={
  category="lo",
  direction="l",
 },
 [0xFAC5]={
  category="lo",
  direction="l",
 },
 [0xFAC6]={
  category="lo",
  direction="l",
 },
 [0xFAC7]={
  category="lo",
  direction="l",
 },
 [0xFAC8]={
  category="lo",
  direction="l",
 },
 [0xFAC9]={
  category="lo",
  direction="l",
 },
 [0xFACA]={
  category="lo",
  direction="l",
 },
 [0xFACB]={
  category="lo",
  direction="l",
 },
 [0xFACC]={
  category="lo",
  direction="l",
 },
 [0xFACD]={
  category="lo",
  direction="l",
 },
 [0xFACE]={
  category="lo",
  direction="l",
 },
 [0xFACF]={
  category="lo",
  direction="l",
 },
 [0xFAD0]={
  category="lo",
  direction="l",
 },
 [0xFAD1]={
  category="lo",
  direction="l",
 },
 [0xFAD2]={
  category="lo",
  direction="l",
 },
 [0xFAD3]={
  category="lo",
  direction="l",
 },
 [0xFAD4]={
  category="lo",
  direction="l",
 },
 [0xFAD5]={
  category="lo",
  direction="l",
 },
 [0xFAD6]={
  category="lo",
  direction="l",
 },
 [0xFAD7]={
  category="lo",
  direction="l",
 },
 [0xFAD8]={
  category="lo",
  direction="l",
 },
 [0xFAD9]={
  category="lo",
  direction="l",
 },
 [0xFB00]={
  category="ll",
  direction="l",
 },
 [0xFB01]={
  category="ll",
  direction="l",
 },
 [0xFB02]={
  category="ll",
  direction="l",
 },
 [0xFB03]={
  category="ll",
  direction="l",
 },
 [0xFB04]={
  category="ll",
  direction="l",
 },
 [0xFB05]={
  category="ll",
  direction="l",
 },
 [0xFB06]={
  category="ll",
  direction="l",
 },
 [0xFB13]={
  category="ll",
  direction="l",
 },
 [0xFB14]={
  category="ll",
  direction="l",
 },
 [0xFB15]={
  category="ll",
  direction="l",
 },
 [0xFB16]={
  category="ll",
  direction="l",
 },
 [0xFB17]={
  category="ll",
  direction="l",
 },
 [0xFB1D]={
  category="lo",
  direction="r",
 },
 [0xFB1E]={
  category="mn",
  direction="nsm",
 },
 [0xFB1F]={
  category="lo",
  direction="r",
 },
 [0xFB20]={
  category="lo",
  direction="r",
 },
 [0xFB21]={
  category="lo",
  direction="r",
 },
 [0xFB22]={
  category="lo",
  direction="r",
 },
 [0xFB23]={
  category="lo",
  direction="r",
 },
 [0xFB24]={
  category="lo",
  direction="r",
 },
 [0xFB25]={
  category="lo",
  direction="r",
 },
 [0xFB26]={
  category="lo",
  direction="r",
 },
 [0xFB27]={
  category="lo",
  direction="r",
 },
 [0xFB28]={
  category="lo",
  direction="r",
 },
 [0xFB29]={
  category="sm",
  direction="es",
 },
 [0xFB2A]={
  category="lo",
  direction="r",
 },
 [0xFB2B]={
  category="lo",
  direction="r",
 },
 [0xFB2C]={
  category="lo",
  direction="r",
 },
 [0xFB2D]={
  category="lo",
  direction="r",
 },
 [0xFB2E]={
  category="lo",
  direction="r",
 },
 [0xFB2F]={
  category="lo",
  direction="r",
 },
 [0xFB30]={
  category="lo",
  direction="r",
 },
 [0xFB31]={
  category="lo",
  direction="r",
 },
 [0xFB32]={
  category="lo",
  direction="r",
 },
 [0xFB33]={
  category="lo",
  direction="r",
 },
 [0xFB34]={
  category="lo",
  direction="r",
 },
 [0xFB35]={
  category="lo",
  direction="r",
 },
 [0xFB36]={
  category="lo",
  direction="r",
 },
 [0xFB38]={
  category="lo",
  direction="r",
 },
 [0xFB39]={
  category="lo",
  direction="r",
 },
 [0xFB3A]={
  category="lo",
  direction="r",
 },
 [0xFB3B]={
  category="lo",
  direction="r",
 },
 [0xFB3C]={
  category="lo",
  direction="r",
 },
 [0xFB3E]={
  category="lo",
  direction="r",
 },
 [0xFB40]={
  category="lo",
  direction="r",
 },
 [0xFB41]={
  category="lo",
  direction="r",
 },
 [0xFB43]={
  category="lo",
  direction="r",
 },
 [0xFB44]={
  category="lo",
  direction="r",
 },
 [0xFB46]={
  category="lo",
  direction="r",
 },
 [0xFB47]={
  category="lo",
  direction="r",
 },
 [0xFB48]={
  category="lo",
  direction="r",
 },
 [0xFB49]={
  category="lo",
  direction="r",
 },
 [0xFB4A]={
  category="lo",
  direction="r",
 },
 [0xFB4B]={
  category="lo",
  direction="r",
 },
 [0xFB4C]={
  category="lo",
  direction="r",
 },
 [0xFB4D]={
  category="lo",
  direction="r",
 },
 [0xFB4E]={
  category="lo",
  direction="r",
 },
 [0xFB4F]={
  category="lo",
  direction="r",
 },
 [0xFB50]={
  category="lo",
  direction="al",
 },
 [0xFB51]={
  category="lo",
  direction="al",
 },
 [0xFB52]={
  category="lo",
  direction="al",
 },
 [0xFB53]={
  category="lo",
  direction="al",
 },
 [0xFB54]={
  category="lo",
  direction="al",
 },
 [0xFB55]={
  category="lo",
  direction="al",
 },
 [0xFB56]={
  category="lo",
  direction="al",
 },
 [0xFB57]={
  category="lo",
  direction="al",
 },
 [0xFB58]={
  category="lo",
  direction="al",
 },
 [0xFB59]={
  category="lo",
  direction="al",
 },
 [0xFB5A]={
  category="lo",
  direction="al",
 },
 [0xFB5B]={
  category="lo",
  direction="al",
 },
 [0xFB5C]={
  category="lo",
  direction="al",
 },
 [0xFB5D]={
  category="lo",
  direction="al",
 },
 [0xFB5E]={
  category="lo",
  direction="al",
 },
 [0xFB5F]={
  category="lo",
  direction="al",
 },
 [0xFB60]={
  category="lo",
  direction="al",
 },
 [0xFB61]={
  category="lo",
  direction="al",
 },
 [0xFB62]={
  category="lo",
  direction="al",
 },
 [0xFB63]={
  category="lo",
  direction="al",
 },
 [0xFB64]={
  category="lo",
  direction="al",
 },
 [0xFB65]={
  category="lo",
  direction="al",
 },
 [0xFB66]={
  category="lo",
  direction="al",
 },
 [0xFB67]={
  category="lo",
  direction="al",
 },
 [0xFB68]={
  category="lo",
  direction="al",
 },
 [0xFB69]={
  category="lo",
  direction="al",
 },
 [0xFB6A]={
  category="lo",
  direction="al",
 },
 [0xFB6B]={
  category="lo",
  direction="al",
 },
 [0xFB6C]={
  category="lo",
  direction="al",
 },
 [0xFB6D]={
  category="lo",
  direction="al",
 },
 [0xFB6E]={
  category="lo",
  direction="al",
 },
 [0xFB6F]={
  category="lo",
  direction="al",
 },
 [0xFB70]={
  category="lo",
  direction="al",
 },
 [0xFB71]={
  category="lo",
  direction="al",
 },
 [0xFB72]={
  category="lo",
  direction="al",
 },
 [0xFB73]={
  category="lo",
  direction="al",
 },
 [0xFB74]={
  category="lo",
  direction="al",
 },
 [0xFB75]={
  category="lo",
  direction="al",
 },
 [0xFB76]={
  category="lo",
  direction="al",
 },
 [0xFB77]={
  category="lo",
  direction="al",
 },
 [0xFB78]={
  category="lo",
  direction="al",
 },
 [0xFB79]={
  category="lo",
  direction="al",
 },
 [0xFB7A]={
  category="lo",
  direction="al",
 },
 [0xFB7B]={
  category="lo",
  direction="al",
 },
 [0xFB7C]={
  category="lo",
  direction="al",
 },
 [0xFB7D]={
  category="lo",
  direction="al",
 },
 [0xFB7E]={
  category="lo",
  direction="al",
 },
 [0xFB7F]={
  category="lo",
  direction="al",
 },
 [0xFB80]={
  category="lo",
  direction="al",
 },
 [0xFB81]={
  category="lo",
  direction="al",
 },
 [0xFB82]={
  category="lo",
  direction="al",
 },
 [0xFB83]={
  category="lo",
  direction="al",
 },
 [0xFB84]={
  category="lo",
  direction="al",
 },
 [0xFB85]={
  category="lo",
  direction="al",
 },
 [0xFB86]={
  category="lo",
  direction="al",
 },
 [0xFB87]={
  category="lo",
  direction="al",
 },
 [0xFB88]={
  category="lo",
  direction="al",
 },
 [0xFB89]={
  category="lo",
  direction="al",
 },
 [0xFB8A]={
  category="lo",
  direction="al",
 },
 [0xFB8B]={
  category="lo",
  direction="al",
 },
 [0xFB8C]={
  category="lo",
  direction="al",
 },
 [0xFB8D]={
  category="lo",
  direction="al",
 },
 [0xFB8E]={
  category="lo",
  direction="al",
 },
 [0xFB8F]={
  category="lo",
  direction="al",
 },
 [0xFB90]={
  category="lo",
  direction="al",
 },
 [0xFB91]={
  category="lo",
  direction="al",
 },
 [0xFB92]={
  category="lo",
  direction="al",
 },
 [0xFB93]={
  category="lo",
  direction="al",
 },
 [0xFB94]={
  category="lo",
  direction="al",
 },
 [0xFB95]={
  category="lo",
  direction="al",
 },
 [0xFB96]={
  category="lo",
  direction="al",
 },
 [0xFB97]={
  category="lo",
  direction="al",
 },
 [0xFB98]={
  category="lo",
  direction="al",
 },
 [0xFB99]={
  category="lo",
  direction="al",
 },
 [0xFB9A]={
  category="lo",
  direction="al",
 },
 [0xFB9B]={
  category="lo",
  direction="al",
 },
 [0xFB9C]={
  category="lo",
  direction="al",
 },
 [0xFB9D]={
  category="lo",
  direction="al",
 },
 [0xFB9E]={
  category="lo",
  direction="al",
 },
 [0xFB9F]={
  category="lo",
  direction="al",
 },
 [0xFBA0]={
  category="lo",
  direction="al",
 },
 [0xFBA1]={
  category="lo",
  direction="al",
 },
 [0xFBA2]={
  category="lo",
  direction="al",
 },
 [0xFBA3]={
  category="lo",
  direction="al",
 },
 [0xFBA4]={
  category="lo",
  direction="al",
 },
 [0xFBA5]={
  category="lo",
  direction="al",
 },
 [0xFBA6]={
  category="lo",
  direction="al",
 },
 [0xFBA7]={
  category="lo",
  direction="al",
 },
 [0xFBA8]={
  category="lo",
  direction="al",
 },
 [0xFBA9]={
  category="lo",
  direction="al",
 },
 [0xFBAA]={
  category="lo",
  direction="al",
 },
 [0xFBAB]={
  category="lo",
  direction="al",
 },
 [0xFBAC]={
  category="lo",
  direction="al",
 },
 [0xFBAD]={
  category="lo",
  direction="al",
 },
 [0xFBAE]={
  category="lo",
  direction="al",
 },
 [0xFBAF]={
  category="lo",
  direction="al",
 },
 [0xFBB0]={
  category="lo",
  direction="al",
 },
 [0xFBB1]={
  category="lo",
  direction="al",
 },
 [0xFBB2]={
  category="sk",
  direction="al",
 },
 [0xFBB3]={
  category="sk",
  direction="al",
 },
 [0xFBB4]={
  category="sk",
  direction="al",
 },
 [0xFBB5]={
  category="sk",
  direction="al",
 },
 [0xFBB6]={
  category="sk",
  direction="al",
 },
 [0xFBB7]={
  category="sk",
  direction="al",
 },
 [0xFBB8]={
  category="sk",
  direction="al",
 },
 [0xFBB9]={
  category="sk",
  direction="al",
 },
 [0xFBBA]={
  category="sk",
  direction="al",
 },
 [0xFBBB]={
  category="sk",
  direction="al",
 },
 [0xFBBC]={
  category="sk",
  direction="al",
 },
 [0xFBBD]={
  category="sk",
  direction="al",
 },
 [0xFBBE]={
  category="sk",
  direction="al",
 },
 [0xFBBF]={
  category="sk",
  direction="al",
 },
 [0xFBC0]={
  category="sk",
  direction="al",
 },
 [0xFBC1]={
  category="sk",
  direction="al",
 },
 [0xFBD3]={
  category="lo",
  direction="al",
 },
 [0xFBD4]={
  category="lo",
  direction="al",
 },
 [0xFBD5]={
  category="lo",
  direction="al",
 },
 [0xFBD6]={
  category="lo",
  direction="al",
 },
 [0xFBD7]={
  category="lo",
  direction="al",
 },
 [0xFBD8]={
  category="lo",
  direction="al",
 },
 [0xFBD9]={
  category="lo",
  direction="al",
 },
 [0xFBDA]={
  category="lo",
  direction="al",
 },
 [0xFBDB]={
  category="lo",
  direction="al",
 },
 [0xFBDC]={
  category="lo",
  direction="al",
 },
 [0xFBDD]={
  category="lo",
  direction="al",
 },
 [0xFBDE]={
  category="lo",
  direction="al",
 },
 [0xFBDF]={
  category="lo",
  direction="al",
 },
 [0xFBE0]={
  category="lo",
  direction="al",
 },
 [0xFBE1]={
  category="lo",
  direction="al",
 },
 [0xFBE2]={
  category="lo",
  direction="al",
 },
 [0xFBE3]={
  category="lo",
  direction="al",
 },
 [0xFBE4]={
  category="lo",
  direction="al",
 },
 [0xFBE5]={
  category="lo",
  direction="al",
 },
 [0xFBE6]={
  category="lo",
  direction="al",
 },
 [0xFBE7]={
  category="lo",
  direction="al",
 },
 [0xFBE8]={
  category="lo",
  direction="al",
 },
 [0xFBE9]={
  category="lo",
  direction="al",
 },
 [0xFBEA]={
  category="lo",
  direction="al",
 },
 [0xFBEB]={
  category="lo",
  direction="al",
 },
 [0xFBEC]={
  category="lo",
  direction="al",
 },
 [0xFBED]={
  category="lo",
  direction="al",
 },
 [0xFBEE]={
  category="lo",
  direction="al",
 },
 [0xFBEF]={
  category="lo",
  direction="al",
 },
 [0xFBF0]={
  category="lo",
  direction="al",
 },
 [0xFBF1]={
  category="lo",
  direction="al",
 },
 [0xFBF2]={
  category="lo",
  direction="al",
 },
 [0xFBF3]={
  category="lo",
  direction="al",
 },
 [0xFBF4]={
  category="lo",
  direction="al",
 },
 [0xFBF5]={
  category="lo",
  direction="al",
 },
 [0xFBF6]={
  category="lo",
  direction="al",
 },
 [0xFBF7]={
  category="lo",
  direction="al",
 },
 [0xFBF8]={
  category="lo",
  direction="al",
 },
 [0xFBF9]={
  category="lo",
  direction="al",
 },
 [0xFBFA]={
  category="lo",
  direction="al",
 },
 [0xFBFB]={
  category="lo",
  direction="al",
 },
 [0xFBFC]={
  category="lo",
  direction="al",
 },
 [0xFBFD]={
  category="lo",
  direction="al",
 },
 [0xFBFE]={
  category="lo",
  direction="al",
 },
 [0xFBFF]={
  category="lo",
  direction="al",
 },
 [0xFC00]={
  category="lo",
  direction="al",
 },
 [0xFC01]={
  category="lo",
  direction="al",
 },
 [0xFC02]={
  category="lo",
  direction="al",
 },
 [0xFC03]={
  category="lo",
  direction="al",
 },
 [0xFC04]={
  category="lo",
  direction="al",
 },
 [0xFC05]={
  category="lo",
  direction="al",
 },
 [0xFC06]={
  category="lo",
  direction="al",
 },
 [0xFC07]={
  category="lo",
  direction="al",
 },
 [0xFC08]={
  category="lo",
  direction="al",
 },
 [0xFC09]={
  category="lo",
  direction="al",
 },
 [0xFC0A]={
  category="lo",
  direction="al",
 },
 [0xFC0B]={
  category="lo",
  direction="al",
 },
 [0xFC0C]={
  category="lo",
  direction="al",
 },
 [0xFC0D]={
  category="lo",
  direction="al",
 },
 [0xFC0E]={
  category="lo",
  direction="al",
 },
 [0xFC0F]={
  category="lo",
  direction="al",
 },
 [0xFC10]={
  category="lo",
  direction="al",
 },
 [0xFC11]={
  category="lo",
  direction="al",
 },
 [0xFC12]={
  category="lo",
  direction="al",
 },
 [0xFC13]={
  category="lo",
  direction="al",
 },
 [0xFC14]={
  category="lo",
  direction="al",
 },
 [0xFC15]={
  category="lo",
  direction="al",
 },
 [0xFC16]={
  category="lo",
  direction="al",
 },
 [0xFC17]={
  category="lo",
  direction="al",
 },
 [0xFC18]={
  category="lo",
  direction="al",
 },
 [0xFC19]={
  category="lo",
  direction="al",
 },
 [0xFC1A]={
  category="lo",
  direction="al",
 },
 [0xFC1B]={
  category="lo",
  direction="al",
 },
 [0xFC1C]={
  category="lo",
  direction="al",
 },
 [0xFC1D]={
  category="lo",
  direction="al",
 },
 [0xFC1E]={
  category="lo",
  direction="al",
 },
 [0xFC1F]={
  category="lo",
  direction="al",
 },
 [0xFC20]={
  category="lo",
  direction="al",
 },
 [0xFC21]={
  category="lo",
  direction="al",
 },
 [0xFC22]={
  category="lo",
  direction="al",
 },
 [0xFC23]={
  category="lo",
  direction="al",
 },
 [0xFC24]={
  category="lo",
  direction="al",
 },
 [0xFC25]={
  category="lo",
  direction="al",
 },
 [0xFC26]={
  category="lo",
  direction="al",
 },
 [0xFC27]={
  category="lo",
  direction="al",
 },
 [0xFC28]={
  category="lo",
  direction="al",
 },
 [0xFC29]={
  category="lo",
  direction="al",
 },
 [0xFC2A]={
  category="lo",
  direction="al",
 },
 [0xFC2B]={
  category="lo",
  direction="al",
 },
 [0xFC2C]={
  category="lo",
  direction="al",
 },
 [0xFC2D]={
  category="lo",
  direction="al",
 },
 [0xFC2E]={
  category="lo",
  direction="al",
 },
 [0xFC2F]={
  category="lo",
  direction="al",
 },
 [0xFC30]={
  category="lo",
  direction="al",
 },
 [0xFC31]={
  category="lo",
  direction="al",
 },
 [0xFC32]={
  category="lo",
  direction="al",
 },
 [0xFC33]={
  category="lo",
  direction="al",
 },
 [0xFC34]={
  category="lo",
  direction="al",
 },
 [0xFC35]={
  category="lo",
  direction="al",
 },
 [0xFC36]={
  category="lo",
  direction="al",
 },
 [0xFC37]={
  category="lo",
  direction="al",
 },
 [0xFC38]={
  category="lo",
  direction="al",
 },
 [0xFC39]={
  category="lo",
  direction="al",
 },
 [0xFC3A]={
  category="lo",
  direction="al",
 },
 [0xFC3B]={
  category="lo",
  direction="al",
 },
 [0xFC3C]={
  category="lo",
  direction="al",
 },
 [0xFC3D]={
  category="lo",
  direction="al",
 },
 [0xFC3E]={
  category="lo",
  direction="al",
 },
 [0xFC3F]={
  category="lo",
  direction="al",
 },
 [0xFC40]={
  category="lo",
  direction="al",
 },
 [0xFC41]={
  category="lo",
  direction="al",
 },
 [0xFC42]={
  category="lo",
  direction="al",
 },
 [0xFC43]={
  category="lo",
  direction="al",
 },
 [0xFC44]={
  category="lo",
  direction="al",
 },
 [0xFC45]={
  category="lo",
  direction="al",
 },
 [0xFC46]={
  category="lo",
  direction="al",
 },
 [0xFC47]={
  category="lo",
  direction="al",
 },
 [0xFC48]={
  category="lo",
  direction="al",
 },
 [0xFC49]={
  category="lo",
  direction="al",
 },
 [0xFC4A]={
  category="lo",
  direction="al",
 },
 [0xFC4B]={
  category="lo",
  direction="al",
 },
 [0xFC4C]={
  category="lo",
  direction="al",
 },
 [0xFC4D]={
  category="lo",
  direction="al",
 },
 [0xFC4E]={
  category="lo",
  direction="al",
 },
 [0xFC4F]={
  category="lo",
  direction="al",
 },
 [0xFC50]={
  category="lo",
  direction="al",
 },
 [0xFC51]={
  category="lo",
  direction="al",
 },
 [0xFC52]={
  category="lo",
  direction="al",
 },
 [0xFC53]={
  category="lo",
  direction="al",
 },
 [0xFC54]={
  category="lo",
  direction="al",
 },
 [0xFC55]={
  category="lo",
  direction="al",
 },
 [0xFC56]={
  category="lo",
  direction="al",
 },
 [0xFC57]={
  category="lo",
  direction="al",
 },
 [0xFC58]={
  category="lo",
  direction="al",
 },
 [0xFC59]={
  category="lo",
  direction="al",
 },
 [0xFC5A]={
  category="lo",
  direction="al",
 },
 [0xFC5B]={
  category="lo",
  direction="al",
 },
 [0xFC5C]={
  category="lo",
  direction="al",
 },
 [0xFC5D]={
  category="lo",
  direction="al",
 },
 [0xFC5E]={
  category="lo",
  direction="al",
 },
 [0xFC5F]={
  category="lo",
  direction="al",
 },
 [0xFC60]={
  category="lo",
  direction="al",
 },
 [0xFC61]={
  category="lo",
  direction="al",
 },
 [0xFC62]={
  category="lo",
  direction="al",
 },
 [0xFC63]={
  category="lo",
  direction="al",
 },
 [0xFC64]={
  category="lo",
  direction="al",
 },
 [0xFC65]={
  category="lo",
  direction="al",
 },
 [0xFC66]={
  category="lo",
  direction="al",
 },
 [0xFC67]={
  category="lo",
  direction="al",
 },
 [0xFC68]={
  category="lo",
  direction="al",
 },
 [0xFC69]={
  category="lo",
  direction="al",
 },
 [0xFC6A]={
  category="lo",
  direction="al",
 },
 [0xFC6B]={
  category="lo",
  direction="al",
 },
 [0xFC6C]={
  category="lo",
  direction="al",
 },
 [0xFC6D]={
  category="lo",
  direction="al",
 },
 [0xFC6E]={
  category="lo",
  direction="al",
 },
 [0xFC6F]={
  category="lo",
  direction="al",
 },
 [0xFC70]={
  category="lo",
  direction="al",
 },
 [0xFC71]={
  category="lo",
  direction="al",
 },
 [0xFC72]={
  category="lo",
  direction="al",
 },
 [0xFC73]={
  category="lo",
  direction="al",
 },
 [0xFC74]={
  category="lo",
  direction="al",
 },
 [0xFC75]={
  category="lo",
  direction="al",
 },
 [0xFC76]={
  category="lo",
  direction="al",
 },
 [0xFC77]={
  category="lo",
  direction="al",
 },
 [0xFC78]={
  category="lo",
  direction="al",
 },
 [0xFC79]={
  category="lo",
  direction="al",
 },
 [0xFC7A]={
  category="lo",
  direction="al",
 },
 [0xFC7B]={
  category="lo",
  direction="al",
 },
 [0xFC7C]={
  category="lo",
  direction="al",
 },
 [0xFC7D]={
  category="lo",
  direction="al",
 },
 [0xFC7E]={
  category="lo",
  direction="al",
 },
 [0xFC7F]={
  category="lo",
  direction="al",
 },
 [0xFC80]={
  category="lo",
  direction="al",
 },
 [0xFC81]={
  category="lo",
  direction="al",
 },
 [0xFC82]={
  category="lo",
  direction="al",
 },
 [0xFC83]={
  category="lo",
  direction="al",
 },
 [0xFC84]={
  category="lo",
  direction="al",
 },
 [0xFC85]={
  category="lo",
  direction="al",
 },
 [0xFC86]={
  category="lo",
  direction="al",
 },
 [0xFC87]={
  category="lo",
  direction="al",
 },
 [0xFC88]={
  category="lo",
  direction="al",
 },
 [0xFC89]={
  category="lo",
  direction="al",
 },
 [0xFC8A]={
  category="lo",
  direction="al",
 },
 [0xFC8B]={
  category="lo",
  direction="al",
 },
 [0xFC8C]={
  category="lo",
  direction="al",
 },
 [0xFC8D]={
  category="lo",
  direction="al",
 },
 [0xFC8E]={
  category="lo",
  direction="al",
 },
 [0xFC8F]={
  category="lo",
  direction="al",
 },
 [0xFC90]={
  category="lo",
  direction="al",
 },
 [0xFC91]={
  category="lo",
  direction="al",
 },
 [0xFC92]={
  category="lo",
  direction="al",
 },
 [0xFC93]={
  category="lo",
  direction="al",
 },
 [0xFC94]={
  category="lo",
  direction="al",
 },
 [0xFC95]={
  category="lo",
  direction="al",
 },
 [0xFC96]={
  category="lo",
  direction="al",
 },
 [0xFC97]={
  category="lo",
  direction="al",
 },
 [0xFC98]={
  category="lo",
  direction="al",
 },
 [0xFC99]={
  category="lo",
  direction="al",
 },
 [0xFC9A]={
  category="lo",
  direction="al",
 },
 [0xFC9B]={
  category="lo",
  direction="al",
 },
 [0xFC9C]={
  category="lo",
  direction="al",
 },
 [0xFC9D]={
  category="lo",
  direction="al",
 },
 [0xFC9E]={
  category="lo",
  direction="al",
 },
 [0xFC9F]={
  category="lo",
  direction="al",
 },
 [0xFCA0]={
  category="lo",
  direction="al",
 },
 [0xFCA1]={
  category="lo",
  direction="al",
 },
 [0xFCA2]={
  category="lo",
  direction="al",
 },
 [0xFCA3]={
  category="lo",
  direction="al",
 },
 [0xFCA4]={
  category="lo",
  direction="al",
 },
 [0xFCA5]={
  category="lo",
  direction="al",
 },
 [0xFCA6]={
  category="lo",
  direction="al",
 },
 [0xFCA7]={
  category="lo",
  direction="al",
 },
 [0xFCA8]={
  category="lo",
  direction="al",
 },
 [0xFCA9]={
  category="lo",
  direction="al",
 },
 [0xFCAA]={
  category="lo",
  direction="al",
 },
 [0xFCAB]={
  category="lo",
  direction="al",
 },
 [0xFCAC]={
  category="lo",
  direction="al",
 },
 [0xFCAD]={
  category="lo",
  direction="al",
 },
 [0xFCAE]={
  category="lo",
  direction="al",
 },
 [0xFCAF]={
  category="lo",
  direction="al",
 },
 [0xFCB0]={
  category="lo",
  direction="al",
 },
 [0xFCB1]={
  category="lo",
  direction="al",
 },
 [0xFCB2]={
  category="lo",
  direction="al",
 },
 [0xFCB3]={
  category="lo",
  direction="al",
 },
 [0xFCB4]={
  category="lo",
  direction="al",
 },
 [0xFCB5]={
  category="lo",
  direction="al",
 },
 [0xFCB6]={
  category="lo",
  direction="al",
 },
 [0xFCB7]={
  category="lo",
  direction="al",
 },
 [0xFCB8]={
  category="lo",
  direction="al",
 },
 [0xFCB9]={
  category="lo",
  direction="al",
 },
 [0xFCBA]={
  category="lo",
  direction="al",
 },
 [0xFCBB]={
  category="lo",
  direction="al",
 },
 [0xFCBC]={
  category="lo",
  direction="al",
 },
 [0xFCBD]={
  category="lo",
  direction="al",
 },
 [0xFCBE]={
  category="lo",
  direction="al",
 },
 [0xFCBF]={
  category="lo",
  direction="al",
 },
 [0xFCC0]={
  category="lo",
  direction="al",
 },
 [0xFCC1]={
  category="lo",
  direction="al",
 },
 [0xFCC2]={
  category="lo",
  direction="al",
 },
 [0xFCC3]={
  category="lo",
  direction="al",
 },
 [0xFCC4]={
  category="lo",
  direction="al",
 },
 [0xFCC5]={
  category="lo",
  direction="al",
 },
 [0xFCC6]={
  category="lo",
  direction="al",
 },
 [0xFCC7]={
  category="lo",
  direction="al",
 },
 [0xFCC8]={
  category="lo",
  direction="al",
 },
 [0xFCC9]={
  category="lo",
  direction="al",
 },
 [0xFCCA]={
  category="lo",
  direction="al",
 },
 [0xFCCB]={
  category="lo",
  direction="al",
 },
 [0xFCCC]={
  category="lo",
  direction="al",
 },
 [0xFCCD]={
  category="lo",
  direction="al",
 },
 [0xFCCE]={
  category="lo",
  direction="al",
 },
 [0xFCCF]={
  category="lo",
  direction="al",
 },
 [0xFCD0]={
  category="lo",
  direction="al",
 },
 [0xFCD1]={
  category="lo",
  direction="al",
 },
 [0xFCD2]={
  category="lo",
  direction="al",
 },
 [0xFCD3]={
  category="lo",
  direction="al",
 },
 [0xFCD4]={
  category="lo",
  direction="al",
 },
 [0xFCD5]={
  category="lo",
  direction="al",
 },
 [0xFCD6]={
  category="lo",
  direction="al",
 },
 [0xFCD7]={
  category="lo",
  direction="al",
 },
 [0xFCD8]={
  category="lo",
  direction="al",
 },
 [0xFCD9]={
  category="lo",
  direction="al",
 },
 [0xFCDA]={
  category="lo",
  direction="al",
 },
 [0xFCDB]={
  category="lo",
  direction="al",
 },
 [0xFCDC]={
  category="lo",
  direction="al",
 },
 [0xFCDD]={
  category="lo",
  direction="al",
 },
 [0xFCDE]={
  category="lo",
  direction="al",
 },
 [0xFCDF]={
  category="lo",
  direction="al",
 },
 [0xFCE0]={
  category="lo",
  direction="al",
 },
 [0xFCE1]={
  category="lo",
  direction="al",
 },
 [0xFCE2]={
  category="lo",
  direction="al",
 },
 [0xFCE3]={
  category="lo",
  direction="al",
 },
 [0xFCE4]={
  category="lo",
  direction="al",
 },
 [0xFCE5]={
  category="lo",
  direction="al",
 },
 [0xFCE6]={
  category="lo",
  direction="al",
 },
 [0xFCE7]={
  category="lo",
  direction="al",
 },
 [0xFCE8]={
  category="lo",
  direction="al",
 },
 [0xFCE9]={
  category="lo",
  direction="al",
 },
 [0xFCEA]={
  category="lo",
  direction="al",
 },
 [0xFCEB]={
  category="lo",
  direction="al",
 },
 [0xFCEC]={
  category="lo",
  direction="al",
 },
 [0xFCED]={
  category="lo",
  direction="al",
 },
 [0xFCEE]={
  category="lo",
  direction="al",
 },
 [0xFCEF]={
  category="lo",
  direction="al",
 },
 [0xFCF0]={
  category="lo",
  direction="al",
 },
 [0xFCF1]={
  category="lo",
  direction="al",
 },
 [0xFCF2]={
  category="lo",
  direction="al",
 },
 [0xFCF3]={
  category="lo",
  direction="al",
 },
 [0xFCF4]={
  category="lo",
  direction="al",
 },
 [0xFCF5]={
  category="lo",
  direction="al",
 },
 [0xFCF6]={
  category="lo",
  direction="al",
 },
 [0xFCF7]={
  category="lo",
  direction="al",
 },
 [0xFCF8]={
  category="lo",
  direction="al",
 },
 [0xFCF9]={
  category="lo",
  direction="al",
 },
 [0xFCFA]={
  category="lo",
  direction="al",
 },
 [0xFCFB]={
  category="lo",
  direction="al",
 },
 [0xFCFC]={
  category="lo",
  direction="al",
 },
 [0xFCFD]={
  category="lo",
  direction="al",
 },
 [0xFCFE]={
  category="lo",
  direction="al",
 },
 [0xFCFF]={
  category="lo",
  direction="al",
 },
 [0xFD00]={
  category="lo",
  direction="al",
 },
 [0xFD01]={
  category="lo",
  direction="al",
 },
 [0xFD02]={
  category="lo",
  direction="al",
 },
 [0xFD03]={
  category="lo",
  direction="al",
 },
 [0xFD04]={
  category="lo",
  direction="al",
 },
 [0xFD05]={
  category="lo",
  direction="al",
 },
 [0xFD06]={
  category="lo",
  direction="al",
 },
 [0xFD07]={
  category="lo",
  direction="al",
 },
 [0xFD08]={
  category="lo",
  direction="al",
 },
 [0xFD09]={
  category="lo",
  direction="al",
 },
 [0xFD0A]={
  category="lo",
  direction="al",
 },
 [0xFD0B]={
  category="lo",
  direction="al",
 },
 [0xFD0C]={
  category="lo",
  direction="al",
 },
 [0xFD0D]={
  category="lo",
  direction="al",
 },
 [0xFD0E]={
  category="lo",
  direction="al",
 },
 [0xFD0F]={
  category="lo",
  direction="al",
 },
 [0xFD10]={
  category="lo",
  direction="al",
 },
 [0xFD11]={
  category="lo",
  direction="al",
 },
 [0xFD12]={
  category="lo",
  direction="al",
 },
 [0xFD13]={
  category="lo",
  direction="al",
 },
 [0xFD14]={
  category="lo",
  direction="al",
 },
 [0xFD15]={
  category="lo",
  direction="al",
 },
 [0xFD16]={
  category="lo",
  direction="al",
 },
 [0xFD17]={
  category="lo",
  direction="al",
 },
 [0xFD18]={
  category="lo",
  direction="al",
 },
 [0xFD19]={
  category="lo",
  direction="al",
 },
 [0xFD1A]={
  category="lo",
  direction="al",
 },
 [0xFD1B]={
  category="lo",
  direction="al",
 },
 [0xFD1C]={
  category="lo",
  direction="al",
 },
 [0xFD1D]={
  category="lo",
  direction="al",
 },
 [0xFD1E]={
  category="lo",
  direction="al",
 },
 [0xFD1F]={
  category="lo",
  direction="al",
 },
 [0xFD20]={
  category="lo",
  direction="al",
 },
 [0xFD21]={
  category="lo",
  direction="al",
 },
 [0xFD22]={
  category="lo",
  direction="al",
 },
 [0xFD23]={
  category="lo",
  direction="al",
 },
 [0xFD24]={
  category="lo",
  direction="al",
 },
 [0xFD25]={
  category="lo",
  direction="al",
 },
 [0xFD26]={
  category="lo",
  direction="al",
 },
 [0xFD27]={
  category="lo",
  direction="al",
 },
 [0xFD28]={
  category="lo",
  direction="al",
 },
 [0xFD29]={
  category="lo",
  direction="al",
 },
 [0xFD2A]={
  category="lo",
  direction="al",
 },
 [0xFD2B]={
  category="lo",
  direction="al",
 },
 [0xFD2C]={
  category="lo",
  direction="al",
 },
 [0xFD2D]={
  category="lo",
  direction="al",
 },
 [0xFD2E]={
  category="lo",
  direction="al",
 },
 [0xFD2F]={
  category="lo",
  direction="al",
 },
 [0xFD30]={
  category="lo",
  direction="al",
 },
 [0xFD31]={
  category="lo",
  direction="al",
 },
 [0xFD32]={
  category="lo",
  direction="al",
 },
 [0xFD33]={
  category="lo",
  direction="al",
 },
 [0xFD34]={
  category="lo",
  direction="al",
 },
 [0xFD35]={
  category="lo",
  direction="al",
 },
 [0xFD36]={
  category="lo",
  direction="al",
 },
 [0xFD37]={
  category="lo",
  direction="al",
 },
 [0xFD38]={
  category="lo",
  direction="al",
 },
 [0xFD39]={
  category="lo",
  direction="al",
 },
 [0xFD3A]={
  category="lo",
  direction="al",
 },
 [0xFD3B]={
  category="lo",
  direction="al",
 },
 [0xFD3C]={
  category="lo",
  direction="al",
 },
 [0xFD3D]={
  category="lo",
  direction="al",
 },
 [0xFD3E]={
  category="ps",
  direction="on",
 },
 [0xFD3F]={
  category="pe",
  direction="on",
 },
 [0xFD50]={
  category="lo",
  direction="al",
 },
 [0xFD51]={
  category="lo",
  direction="al",
 },
 [0xFD52]={
  category="lo",
  direction="al",
 },
 [0xFD53]={
  category="lo",
  direction="al",
 },
 [0xFD54]={
  category="lo",
  direction="al",
 },
 [0xFD55]={
  category="lo",
  direction="al",
 },
 [0xFD56]={
  category="lo",
  direction="al",
 },
 [0xFD57]={
  category="lo",
  direction="al",
 },
 [0xFD58]={
  category="lo",
  direction="al",
 },
 [0xFD59]={
  category="lo",
  direction="al",
 },
 [0xFD5A]={
  category="lo",
  direction="al",
 },
 [0xFD5B]={
  category="lo",
  direction="al",
 },
 [0xFD5C]={
  category="lo",
  direction="al",
 },
 [0xFD5D]={
  category="lo",
  direction="al",
 },
 [0xFD5E]={
  category="lo",
  direction="al",
 },
 [0xFD5F]={
  category="lo",
  direction="al",
 },
 [0xFD60]={
  category="lo",
  direction="al",
 },
 [0xFD61]={
  category="lo",
  direction="al",
 },
 [0xFD62]={
  category="lo",
  direction="al",
 },
 [0xFD63]={
  category="lo",
  direction="al",
 },
 [0xFD64]={
  category="lo",
  direction="al",
 },
 [0xFD65]={
  category="lo",
  direction="al",
 },
 [0xFD66]={
  category="lo",
  direction="al",
 },
 [0xFD67]={
  category="lo",
  direction="al",
 },
 [0xFD68]={
  category="lo",
  direction="al",
 },
 [0xFD69]={
  category="lo",
  direction="al",
 },
 [0xFD6A]={
  category="lo",
  direction="al",
 },
 [0xFD6B]={
  category="lo",
  direction="al",
 },
 [0xFD6C]={
  category="lo",
  direction="al",
 },
 [0xFD6D]={
  category="lo",
  direction="al",
 },
 [0xFD6E]={
  category="lo",
  direction="al",
 },
 [0xFD6F]={
  category="lo",
  direction="al",
 },
 [0xFD70]={
  category="lo",
  direction="al",
 },
 [0xFD71]={
  category="lo",
  direction="al",
 },
 [0xFD72]={
  category="lo",
  direction="al",
 },
 [0xFD73]={
  category="lo",
  direction="al",
 },
 [0xFD74]={
  category="lo",
  direction="al",
 },
 [0xFD75]={
  category="lo",
  direction="al",
 },
 [0xFD76]={
  category="lo",
  direction="al",
 },
 [0xFD77]={
  category="lo",
  direction="al",
 },
 [0xFD78]={
  category="lo",
  direction="al",
 },
 [0xFD79]={
  category="lo",
  direction="al",
 },
 [0xFD7A]={
  category="lo",
  direction="al",
 },
 [0xFD7B]={
  category="lo",
  direction="al",
 },
 [0xFD7C]={
  category="lo",
  direction="al",
 },
 [0xFD7D]={
  category="lo",
  direction="al",
 },
 [0xFD7E]={
  category="lo",
  direction="al",
 },
 [0xFD7F]={
  category="lo",
  direction="al",
 },
 [0xFD80]={
  category="lo",
  direction="al",
 },
 [0xFD81]={
  category="lo",
  direction="al",
 },
 [0xFD82]={
  category="lo",
  direction="al",
 },
 [0xFD83]={
  category="lo",
  direction="al",
 },
 [0xFD84]={
  category="lo",
  direction="al",
 },
 [0xFD85]={
  category="lo",
  direction="al",
 },
 [0xFD86]={
  category="lo",
  direction="al",
 },
 [0xFD87]={
  category="lo",
  direction="al",
 },
 [0xFD88]={
  category="lo",
  direction="al",
 },
 [0xFD89]={
  category="lo",
  direction="al",
 },
 [0xFD8A]={
  category="lo",
  direction="al",
 },
 [0xFD8B]={
  category="lo",
  direction="al",
 },
 [0xFD8C]={
  category="lo",
  direction="al",
 },
 [0xFD8D]={
  category="lo",
  direction="al",
 },
 [0xFD8E]={
  category="lo",
  direction="al",
 },
 [0xFD8F]={
  category="lo",
  direction="al",
 },
 [0xFD92]={
  category="lo",
  direction="al",
 },
 [0xFD93]={
  category="lo",
  direction="al",
 },
 [0xFD94]={
  category="lo",
  direction="al",
 },
 [0xFD95]={
  category="lo",
  direction="al",
 },
 [0xFD96]={
  category="lo",
  direction="al",
 },
 [0xFD97]={
  category="lo",
  direction="al",
 },
 [0xFD98]={
  category="lo",
  direction="al",
 },
 [0xFD99]={
  category="lo",
  direction="al",
 },
 [0xFD9A]={
  category="lo",
  direction="al",
 },
 [0xFD9B]={
  category="lo",
  direction="al",
 },
 [0xFD9C]={
  category="lo",
  direction="al",
 },
 [0xFD9D]={
  category="lo",
  direction="al",
 },
 [0xFD9E]={
  category="lo",
  direction="al",
 },
 [0xFD9F]={
  category="lo",
  direction="al",
 },
 [0xFDA0]={
  category="lo",
  direction="al",
 },
 [0xFDA1]={
  category="lo",
  direction="al",
 },
 [0xFDA2]={
  category="lo",
  direction="al",
 },
 [0xFDA3]={
  category="lo",
  direction="al",
 },
 [0xFDA4]={
  category="lo",
  direction="al",
 },
 [0xFDA5]={
  category="lo",
  direction="al",
 },
 [0xFDA6]={
  category="lo",
  direction="al",
 },
 [0xFDA7]={
  category="lo",
  direction="al",
 },
 [0xFDA8]={
  category="lo",
  direction="al",
 },
 [0xFDA9]={
  category="lo",
  direction="al",
 },
 [0xFDAA]={
  category="lo",
  direction="al",
 },
 [0xFDAB]={
  category="lo",
  direction="al",
 },
 [0xFDAC]={
  category="lo",
  direction="al",
 },
 [0xFDAD]={
  category="lo",
  direction="al",
 },
 [0xFDAE]={
  category="lo",
  direction="al",
 },
 [0xFDAF]={
  category="lo",
  direction="al",
 },
 [0xFDB0]={
  category="lo",
  direction="al",
 },
 [0xFDB1]={
  category="lo",
  direction="al",
 },
 [0xFDB2]={
  category="lo",
  direction="al",
 },
 [0xFDB3]={
  category="lo",
  direction="al",
 },
 [0xFDB4]={
  category="lo",
  direction="al",
 },
 [0xFDB5]={
  category="lo",
  direction="al",
 },
 [0xFDB6]={
  category="lo",
  direction="al",
 },
 [0xFDB7]={
  category="lo",
  direction="al",
 },
 [0xFDB8]={
  category="lo",
  direction="al",
 },
 [0xFDB9]={
  category="lo",
  direction="al",
 },
 [0xFDBA]={
  category="lo",
  direction="al",
 },
 [0xFDBB]={
  category="lo",
  direction="al",
 },
 [0xFDBC]={
  category="lo",
  direction="al",
 },
 [0xFDBD]={
  category="lo",
  direction="al",
 },
 [0xFDBE]={
  category="lo",
  direction="al",
 },
 [0xFDBF]={
  category="lo",
  direction="al",
 },
 [0xFDC0]={
  category="lo",
  direction="al",
 },
 [0xFDC1]={
  category="lo",
  direction="al",
 },
 [0xFDC2]={
  category="lo",
  direction="al",
 },
 [0xFDC3]={
  category="lo",
  direction="al",
 },
 [0xFDC4]={
  category="lo",
  direction="al",
 },
 [0xFDC5]={
  category="lo",
  direction="al",
 },
 [0xFDC6]={
  category="lo",
  direction="al",
 },
 [0xFDC7]={
  category="lo",
  direction="al",
 },
 [0xFDF0]={
  category="lo",
  direction="al",
 },
 [0xFDF1]={
  category="lo",
  direction="al",
 },
 [0xFDF2]={
  category="lo",
  direction="al",
 },
 [0xFDF3]={
  category="lo",
  direction="al",
 },
 [0xFDF4]={
  category="lo",
  direction="al",
 },
 [0xFDF5]={
  category="lo",
  direction="al",
 },
 [0xFDF6]={
  category="lo",
  direction="al",
 },
 [0xFDF7]={
  category="lo",
  direction="al",
 },
 [0xFDF8]={
  category="lo",
  direction="al",
 },
 [0xFDF9]={
  category="lo",
  direction="al",
 },
 [0xFDFA]={
  category="lo",
  direction="al",
 },
 [0xFDFB]={
  category="lo",
  direction="al",
 },
 [0xFDFC]={
  category="sc",
  direction="al",
 },
 [0xFDFD]={
  category="so",
  direction="on",
 },
 [0xFE00]={
  category="mn",
  direction="nsm",
 },
 [0xFE01]={
  category="mn",
  direction="nsm",
 },
 [0xFE02]={
  category="mn",
  direction="nsm",
 },
 [0xFE03]={
  category="mn",
  direction="nsm",
 },
 [0xFE04]={
  category="mn",
  direction="nsm",
 },
 [0xFE05]={
  category="mn",
  direction="nsm",
 },
 [0xFE06]={
  category="mn",
  direction="nsm",
 },
 [0xFE07]={
  category="mn",
  direction="nsm",
 },
 [0xFE08]={
  category="mn",
  direction="nsm",
 },
 [0xFE09]={
  category="mn",
  direction="nsm",
 },
 [0xFE0A]={
  category="mn",
  direction="nsm",
 },
 [0xFE0B]={
  category="mn",
  direction="nsm",
 },
 [0xFE0C]={
  category="mn",
  direction="nsm",
 },
 [0xFE0D]={
  category="mn",
  direction="nsm",
 },
 [0xFE0E]={
  category="mn",
  direction="nsm",
 },
 [0xFE0F]={
  category="mn",
  direction="nsm",
 },
 [0xFE10]={
  category="po",
  direction="on",
 },
 [0xFE11]={
  category="po",
  direction="on",
 },
 [0xFE12]={
  category="po",
  direction="on",
 },
 [0xFE13]={
  category="po",
  direction="on",
 },
 [0xFE14]={
  category="po",
  direction="on",
 },
 [0xFE15]={
  category="po",
  direction="on",
 },
 [0xFE16]={
  category="po",
  direction="on",
 },
 [0xFE17]={
  category="ps",
  direction="on",
 },
 [0xFE18]={
  category="pe",
  direction="on",
 },
 [0xFE19]={
  category="po",
  direction="on",
 },
 [0xFE20]={
  category="mn",
  direction="nsm",
 },
 [0xFE21]={
  category="mn",
  direction="nsm",
 },
 [0xFE22]={
  category="mn",
  direction="nsm",
 },
 [0xFE23]={
  category="mn",
  direction="nsm",
 },
 [0xFE24]={
  category="mn",
  direction="nsm",
 },
 [0xFE25]={
  category="mn",
  direction="nsm",
 },
 [0xFE26]={
  category="mn",
  direction="nsm",
 },
 [0xFE27]={
  category="mn",
  direction="nsm",
 },
 [0xFE28]={
  category="mn",
  direction="nsm",
 },
 [0xFE29]={
  category="mn",
  direction="nsm",
 },
 [0xFE2A]={
  category="mn",
  direction="nsm",
 },
 [0xFE2B]={
  category="mn",
  direction="nsm",
 },
 [0xFE2C]={
  category="mn",
  direction="nsm",
 },
 [0xFE2D]={
  category="mn",
  direction="nsm",
 },
 [0xFE30]={
  category="po",
  direction="on",
 },
 [0xFE31]={
  category="pd",
  direction="on",
 },
 [0xFE32]={
  category="pd",
  direction="on",
 },
 [0xFE33]={
  category="pc",
  direction="on",
 },
 [0xFE34]={
  category="pc",
  direction="on",
 },
 [0xFE35]={
  category="ps",
  direction="on",
 },
 [0xFE36]={
  category="pe",
  direction="on",
 },
 [0xFE37]={
  category="ps",
  direction="on",
 },
 [0xFE38]={
  category="pe",
  direction="on",
 },
 [0xFE39]={
  category="ps",
  direction="on",
 },
 [0xFE3A]={
  category="pe",
  direction="on",
 },
 [0xFE3B]={
  category="ps",
  direction="on",
 },
 [0xFE3C]={
  category="pe",
  direction="on",
 },
 [0xFE3D]={
  category="ps",
  direction="on",
 },
 [0xFE3E]={
  category="pe",
  direction="on",
 },
 [0xFE3F]={
  category="ps",
  direction="on",
 },
 [0xFE40]={
  category="pe",
  direction="on",
 },
 [0xFE41]={
  category="ps",
  direction="on",
 },
 [0xFE42]={
  category="pe",
  direction="on",
 },
 [0xFE43]={
  category="ps",
  direction="on",
 },
 [0xFE44]={
  category="pe",
  direction="on",
 },
 [0xFE45]={
  category="po",
  direction="on",
 },
 [0xFE46]={
  category="po",
  direction="on",
 },
 [0xFE47]={
  category="ps",
  direction="on",
 },
 [0xFE48]={
  category="pe",
  direction="on",
 },
 [0xFE49]={
  category="po",
  direction="on",
 },
 [0xFE4A]={
  category="po",
  direction="on",
 },
 [0xFE4B]={
  category="po",
  direction="on",
 },
 [0xFE4C]={
  category="po",
  direction="on",
 },
 [0xFE4D]={
  category="pc",
  direction="on",
 },
 [0xFE4E]={
  category="pc",
  direction="on",
 },
 [0xFE4F]={
  category="pc",
  direction="on",
 },
 [0xFE50]={
  category="po",
  direction="cs",
 },
 [0xFE51]={
  category="po",
  direction="on",
 },
 [0xFE52]={
  category="po",
  direction="cs",
 },
 [0xFE54]={
  category="po",
  direction="on",
 },
 [0xFE55]={
  category="po",
  direction="cs",
 },
 [0xFE56]={
  category="po",
  direction="on",
 },
 [0xFE57]={
  category="po",
  direction="on",
 },
 [0xFE58]={
  category="pd",
  direction="on",
 },
 [0xFE59]={
  category="ps",
  direction="on",
  mirror=0xFE5A,
  textclass="open",
 },
 [0xFE5A]={
  category="pe",
  direction="on",
  mirror=0xFE59,
  textclass="close",
 },
 [0xFE5B]={
  category="ps",
  direction="on",
  mirror=0xFE5C,
  textclass="open",
 },
 [0xFE5C]={
  category="pe",
  direction="on",
  mirror=0xFE5B,
  textclass="close",
 },
 [0xFE5D]={
  category="ps",
  direction="on",
  mirror=0xFE5E,
  textclass="open",
 },
 [0xFE5E]={
  category="pe",
  direction="on",
  mirror=0xFE5D,
  textclass="close",
 },
 [0xFE5F]={
  category="po",
  direction="et",
 },
 [0xFE60]={
  category="po",
  direction="on",
 },
 [0xFE61]={
  category="po",
  direction="on",
 },
 [0xFE62]={
  category="sm",
  direction="es",
 },
 [0xFE63]={
  category="pd",
  direction="es",
 },
 [0xFE64]={
  category="sm",
  direction="on",
  mirror=0xFE65,
  textclass="open",
 },
 [0xFE65]={
  category="sm",
  direction="on",
  mirror=0xFE64,
  textclass="close",
 },
 [0xFE66]={
  category="sm",
  direction="on",
 },
 [0xFE68]={
  category="po",
  direction="on",
 },
 [0xFE69]={
  category="sc",
  direction="et",
 },
 [0xFE6A]={
  category="po",
  direction="et",
 },
 [0xFE6B]={
  category="po",
  direction="on",
 },
 [0xFE70]={
  category="lo",
  direction="al",
 },
 [0xFE71]={
  category="lo",
  direction="al",
 },
 [0xFE72]={
  category="lo",
  direction="al",
 },
 [0xFE73]={
  category="lo",
  direction="al",
 },
 [0xFE74]={
  category="lo",
  direction="al",
 },
 [0xFE76]={
  category="lo",
  direction="al",
 },
 [0xFE77]={
  category="lo",
  direction="al",
 },
 [0xFE78]={
  category="lo",
  direction="al",
 },
 [0xFE79]={
  category="lo",
  direction="al",
 },
 [0xFE7A]={
  category="lo",
  direction="al",
 },
 [0xFE7B]={
  category="lo",
  direction="al",
 },
 [0xFE7C]={
  category="lo",
  direction="al",
 },
 [0xFE7D]={
  category="lo",
  direction="al",
 },
 [0xFE7E]={
  category="lo",
  direction="al",
 },
 [0xFE7F]={
  category="lo",
  direction="al",
 },
 [0xFE80]={
  category="lo",
  direction="al",
 },
 [0xFE81]={
  category="lo",
  direction="al",
 },
 [0xFE82]={
  category="lo",
  direction="al",
 },
 [0xFE83]={
  category="lo",
  direction="al",
 },
 [0xFE84]={
  category="lo",
  direction="al",
 },
 [0xFE85]={
  category="lo",
  direction="al",
 },
 [0xFE86]={
  category="lo",
  direction="al",
 },
 [0xFE87]={
  category="lo",
  direction="al",
 },
 [0xFE88]={
  category="lo",
  direction="al",
 },
 [0xFE89]={
  category="lo",
  direction="al",
 },
 [0xFE8A]={
  category="lo",
  direction="al",
 },
 [0xFE8B]={
  category="lo",
  direction="al",
 },
 [0xFE8C]={
  category="lo",
  direction="al",
 },
 [0xFE8D]={
  category="lo",
  direction="al",
 },
 [0xFE8E]={
  category="lo",
  direction="al",
 },
 [0xFE8F]={
  category="lo",
  direction="al",
 },
 [0xFE90]={
  category="lo",
  direction="al",
 },
 [0xFE91]={
  category="lo",
  direction="al",
 },
 [0xFE92]={
  category="lo",
  direction="al",
 },
 [0xFE93]={
  category="lo",
  direction="al",
 },
 [0xFE94]={
  category="lo",
  direction="al",
 },
 [0xFE95]={
  category="lo",
  direction="al",
 },
 [0xFE96]={
  category="lo",
  direction="al",
 },
 [0xFE97]={
  category="lo",
  direction="al",
 },
 [0xFE98]={
  category="lo",
  direction="al",
 },
 [0xFE99]={
  category="lo",
  direction="al",
 },
 [0xFE9A]={
  category="lo",
  direction="al",
 },
 [0xFE9B]={
  category="lo",
  direction="al",
 },
 [0xFE9C]={
  category="lo",
  direction="al",
 },
 [0xFE9D]={
  category="lo",
  direction="al",
 },
 [0xFE9E]={
  category="lo",
  direction="al",
 },
 [0xFE9F]={
  category="lo",
  direction="al",
 },
 [0xFEA0]={
  category="lo",
  direction="al",
 },
 [0xFEA1]={
  category="lo",
  direction="al",
 },
 [0xFEA2]={
  category="lo",
  direction="al",
 },
 [0xFEA3]={
  category="lo",
  direction="al",
 },
 [0xFEA4]={
  category="lo",
  direction="al",
 },
 [0xFEA5]={
  category="lo",
  direction="al",
 },
 [0xFEA6]={
  category="lo",
  direction="al",
 },
 [0xFEA7]={
  category="lo",
  direction="al",
 },
 [0xFEA8]={
  category="lo",
  direction="al",
 },
 [0xFEA9]={
  category="lo",
  direction="al",
 },
 [0xFEAA]={
  category="lo",
  direction="al",
 },
 [0xFEAB]={
  category="lo",
  direction="al",
 },
 [0xFEAC]={
  category="lo",
  direction="al",
 },
 [0xFEAD]={
  category="lo",
  direction="al",
 },
 [0xFEAE]={
  category="lo",
  direction="al",
 },
 [0xFEAF]={
  category="lo",
  direction="al",
 },
 [0xFEB0]={
  category="lo",
  direction="al",
 },
 [0xFEB1]={
  category="lo",
  direction="al",
 },
 [0xFEB2]={
  category="lo",
  direction="al",
 },
 [0xFEB3]={
  category="lo",
  direction="al",
 },
 [0xFEB4]={
  category="lo",
  direction="al",
 },
 [0xFEB5]={
  category="lo",
  direction="al",
 },
 [0xFEB6]={
  category="lo",
  direction="al",
 },
 [0xFEB7]={
  category="lo",
  direction="al",
 },
 [0xFEB8]={
  category="lo",
  direction="al",
 },
 [0xFEB9]={
  category="lo",
  direction="al",
 },
 [0xFEBA]={
  category="lo",
  direction="al",
 },
 [0xFEBB]={
  category="lo",
  direction="al",
 },
 [0xFEBC]={
  category="lo",
  direction="al",
 },
 [0xFEBD]={
  category="lo",
  direction="al",
 },
 [0xFEBE]={
  category="lo",
  direction="al",
 },
 [0xFEBF]={
  category="lo",
  direction="al",
 },
 [0xFEC0]={
  category="lo",
  direction="al",
 },
 [0xFEC1]={
  category="lo",
  direction="al",
 },
 [0xFEC2]={
  category="lo",
  direction="al",
 },
 [0xFEC3]={
  category="lo",
  direction="al",
 },
 [0xFEC4]={
  category="lo",
  direction="al",
 },
 [0xFEC5]={
  category="lo",
  direction="al",
 },
 [0xFEC6]={
  category="lo",
  direction="al",
 },
 [0xFEC7]={
  category="lo",
  direction="al",
 },
 [0xFEC8]={
  category="lo",
  direction="al",
 },
 [0xFEC9]={
  category="lo",
  direction="al",
 },
 [0xFECA]={
  category="lo",
  direction="al",
 },
 [0xFECB]={
  category="lo",
  direction="al",
 },
 [0xFECC]={
  category="lo",
  direction="al",
 },
 [0xFECD]={
  category="lo",
  direction="al",
 },
 [0xFECE]={
  category="lo",
  direction="al",
 },
 [0xFECF]={
  category="lo",
  direction="al",
 },
 [0xFED0]={
  category="lo",
  direction="al",
 },
 [0xFED1]={
  category="lo",
  direction="al",
 },
 [0xFED2]={
  category="lo",
  direction="al",
 },
 [0xFED3]={
  category="lo",
  direction="al",
 },
 [0xFED4]={
  category="lo",
  direction="al",
 },
 [0xFED5]={
  category="lo",
  direction="al",
 },
 [0xFED6]={
  category="lo",
  direction="al",
 },
 [0xFED7]={
  category="lo",
  direction="al",
 },
 [0xFED8]={
  category="lo",
  direction="al",
 },
 [0xFED9]={
  category="lo",
  direction="al",
 },
 [0xFEDA]={
  category="lo",
  direction="al",
 },
 [0xFEDB]={
  category="lo",
  direction="al",
 },
 [0xFEDC]={
  category="lo",
  direction="al",
 },
 [0xFEDD]={
  category="lo",
  direction="al",
 },
 [0xFEDE]={
  category="lo",
  direction="al",
 },
 [0xFEDF]={
  category="lo",
  direction="al",
 },
 [0xFEE0]={
  category="lo",
  direction="al",
 },
 [0xFEE1]={
  category="lo",
  direction="al",
 },
 [0xFEE2]={
  category="lo",
  direction="al",
 },
 [0xFEE3]={
  category="lo",
  direction="al",
 },
 [0xFEE4]={
  category="lo",
  direction="al",
 },
 [0xFEE5]={
  category="lo",
  direction="al",
 },
 [0xFEE6]={
  category="lo",
  direction="al",
 },
 [0xFEE7]={
  category="lo",
  direction="al",
 },
 [0xFEE8]={
  category="lo",
  direction="al",
 },
 [0xFEE9]={
  category="lo",
  direction="al",
 },
 [0xFEEA]={
  category="lo",
  direction="al",
 },
 [0xFEEB]={
  category="lo",
  direction="al",
 },
 [0xFEEC]={
  category="lo",
  direction="al",
 },
 [0xFEED]={
  category="lo",
  direction="al",
 },
 [0xFEEE]={
  category="lo",
  direction="al",
 },
 [0xFEEF]={
  category="lo",
  direction="al",
 },
 [0xFEF0]={
  category="lo",
  direction="al",
 },
 [0xFEF1]={
  category="lo",
  direction="al",
 },
 [0xFEF2]={
  category="lo",
  direction="al",
 },
 [0xFEF3]={
  category="lo",
  direction="al",
 },
 [0xFEF4]={
  category="lo",
  direction="al",
 },
 [0xFEF5]={
  category="lo",
  direction="al",
 },
 [0xFEF6]={
  category="lo",
  direction="al",
 },
 [0xFEF7]={
  category="lo",
  direction="al",
 },
 [0xFEF8]={
  category="lo",
  direction="al",
 },
 [0xFEF9]={
  category="lo",
  direction="al",
 },
 [0xFEFA]={
  category="lo",
  direction="al",
 },
 [0xFEFB]={
  category="lo",
  direction="al",
 },
 [0xFEFC]={
  category="lo",
  direction="al",
 },
 [0xFEFF]={
  category="cf",
  direction="bn",
 },
 [0xFF01]={
  category="po",
  direction="on",
 },
 [0xFF02]={
  category="po",
  direction="on",
 },
 [0xFF03]={
  category="po",
  direction="et",
 },
 [0xFF04]={
  category="sc",
  direction="et",
 },
 [0xFF05]={
  category="po",
  direction="et",
 },
 [0xFF06]={
  category="po",
  direction="on",
 },
 [0xFF07]={
  category="po",
  direction="on",
 },
 [0xFF08]={
  category="ps",
  direction="on",
  mirror=0xFF09,
  textclass="open",
 },
 [0xFF09]={
  category="pe",
  direction="on",
  mirror=0xFF08,
  textclass="close",
 },
 [0xFF0A]={
  category="po",
  direction="on",
 },
 [0xFF0B]={
  category="sm",
  direction="es",
 },
 [0xFF0C]={
  category="po",
  direction="cs",
 },
 [0xFF0D]={
  category="pd",
  direction="es",
 },
 [0xFF0E]={
  category="po",
  direction="cs",
 },
 [0xFF0F]={
  category="po",
  direction="cs",
 },
 [0xFF10]={
  category="nd",
  direction="en",
 },
 [0xFF11]={
  category="nd",
  direction="en",
 },
 [0xFF12]={
  category="nd",
  direction="en",
 },
 [0xFF13]={
  category="nd",
  direction="en",
 },
 [0xFF14]={
  category="nd",
  direction="en",
 },
 [0xFF15]={
  category="nd",
  direction="en",
 },
 [0xFF16]={
  category="nd",
  direction="en",
 },
 [0xFF17]={
  category="nd",
  direction="en",
 },
 [0xFF18]={
  category="nd",
  direction="en",
 },
 [0xFF19]={
  category="nd",
  direction="en",
 },
 [0xFF1A]={
  category="po",
  direction="cs",
 },
 [0xFF1B]={
  category="po",
  direction="on",
 },
 [0xFF1C]={
  category="sm",
  direction="on",
  mirror=0xFF1E,
  textclass="open",
 },
 [0xFF1D]={
  category="sm",
  direction="on",
 },
 [0xFF1E]={
  category="sm",
  direction="on",
  mirror=0xFF1C,
  textclass="close",
 },
 [0xFF1F]={
  category="po",
  direction="on",
 },
 [0xFF20]={
  category="po",
  direction="on",
 },
 [0xFF21]={
  category="lu",
  direction="l",
 },
 [0xFF22]={
  category="lu",
  direction="l",
 },
 [0xFF23]={
  category="lu",
  direction="l",
 },
 [0xFF24]={
  category="lu",
  direction="l",
 },
 [0xFF25]={
  category="lu",
  direction="l",
 },
 [0xFF26]={
  category="lu",
  direction="l",
 },
 [0xFF27]={
  category="lu",
  direction="l",
 },
 [0xFF28]={
  category="lu",
  direction="l",
 },
 [0xFF29]={
  category="lu",
  direction="l",
 },
 [0xFF2A]={
  category="lu",
  direction="l",
 },
 [0xFF2B]={
  category="lu",
  direction="l",
 },
 [0xFF2C]={
  category="lu",
  direction="l",
 },
 [0xFF2D]={
  category="lu",
  direction="l",
 },
 [0xFF2E]={
  category="lu",
  direction="l",
 },
 [0xFF2F]={
  category="lu",
  direction="l",
 },
 [0xFF30]={
  category="lu",
  direction="l",
 },
 [0xFF31]={
  category="lu",
  direction="l",
 },
 [0xFF32]={
  category="lu",
  direction="l",
 },
 [0xFF33]={
  category="lu",
  direction="l",
 },
 [0xFF34]={
  category="lu",
  direction="l",
 },
 [0xFF35]={
  category="lu",
  direction="l",
 },
 [0xFF36]={
  category="lu",
  direction="l",
 },
 [0xFF37]={
  category="lu",
  direction="l",
 },
 [0xFF38]={
  category="lu",
  direction="l",
 },
 [0xFF39]={
  category="lu",
  direction="l",
 },
 [0xFF3A]={
  category="lu",
  direction="l",
 },
 [0xFF3B]={
  category="ps",
  direction="on",
  mirror=0xFF3D,
  textclass="open",
 },
 [0xFF3C]={
  category="po",
  direction="on",
 },
 [0xFF3D]={
  category="pe",
  direction="on",
  mirror=0xFF3B,
  textclass="close",
 },
 [0xFF3E]={
  category="sk",
  direction="on",
 },
 [0xFF3F]={
  category="pc",
  direction="on",
 },
 [0xFF40]={
  category="sk",
  direction="on",
 },
 [0xFF41]={
  category="ll",
  direction="l",
 },
 [0xFF42]={
  category="ll",
  direction="l",
 },
 [0xFF43]={
  category="ll",
  direction="l",
 },
 [0xFF44]={
  category="ll",
  direction="l",
 },
 [0xFF45]={
  category="ll",
  direction="l",
 },
 [0xFF46]={
  category="ll",
  direction="l",
 },
 [0xFF47]={
  category="ll",
  direction="l",
 },
 [0xFF48]={
  category="ll",
  direction="l",
 },
 [0xFF49]={
  category="ll",
  direction="l",
 },
 [0xFF4A]={
  category="ll",
  direction="l",
 },
 [0xFF4B]={
  category="ll",
  direction="l",
 },
 [0xFF4C]={
  category="ll",
  direction="l",
 },
 [0xFF4D]={
  category="ll",
  direction="l",
 },
 [0xFF4E]={
  category="ll",
  direction="l",
 },
 [0xFF4F]={
  category="ll",
  direction="l",
 },
 [0xFF50]={
  category="ll",
  direction="l",
 },
 [0xFF51]={
  category="ll",
  direction="l",
 },
 [0xFF52]={
  category="ll",
  direction="l",
 },
 [0xFF53]={
  category="ll",
  direction="l",
 },
 [0xFF54]={
  category="ll",
  direction="l",
 },
 [0xFF55]={
  category="ll",
  direction="l",
 },
 [0xFF56]={
  category="ll",
  direction="l",
 },
 [0xFF57]={
  category="ll",
  direction="l",
 },
 [0xFF58]={
  category="ll",
  direction="l",
 },
 [0xFF59]={
  category="ll",
  direction="l",
 },
 [0xFF5A]={
  category="ll",
  direction="l",
 },
 [0xFF5B]={
  category="ps",
  direction="on",
  mirror=0xFF5D,
  textclass="open",
 },
 [0xFF5C]={
  category="sm",
  direction="on",
 },
 [0xFF5D]={
  category="pe",
  direction="on",
  mirror=0xFF5B,
  textclass="close",
 },
 [0xFF5E]={
  category="sm",
  direction="on",
 },
 [0xFF5F]={
  category="ps",
  direction="on",
  mirror=0xFF60,
  textclass="open",
 },
 [0xFF60]={
  category="pe",
  direction="on",
  mirror=0xFF5F,
  textclass="close",
 },
 [0xFF61]={
  category="po",
  direction="on",
 },
 [0xFF62]={
  category="ps",
  direction="on",
  mirror=0xFF63,
  textclass="open",
 },
 [0xFF63]={
  category="pe",
  direction="on",
  mirror=0xFF62,
  textclass="close",
 },
 [0xFF64]={
  category="po",
  direction="on",
 },
 [0xFF65]={
  category="po",
  direction="on",
 },
 [0xFF66]={
  category="lo",
  direction="l",
 },
 [0xFF67]={
  category="lo",
  direction="l",
 },
 [0xFF68]={
  category="lo",
  direction="l",
 },
 [0xFF69]={
  category="lo",
  direction="l",
 },
 [0xFF6A]={
  category="lo",
  direction="l",
 },
 [0xFF6B]={
  category="lo",
  direction="l",
 },
 [0xFF6C]={
  category="lo",
  direction="l",
 },
 [0xFF6D]={
  category="lo",
  direction="l",
 },
 [0xFF6E]={
  category="lo",
  direction="l",
 },
 [0xFF6F]={
  category="lo",
  direction="l",
 },
 [0xFF70]={
  category="lm",
  direction="l",
 },
 [0xFF71]={
  category="lo",
  direction="l",
 },
 [0xFF72]={
  category="lo",
  direction="l",
 },
 [0xFF73]={
  category="lo",
  direction="l",
 },
 [0xFF74]={
  category="lo",
  direction="l",
 },
 [0xFF75]={
  category="lo",
  direction="l",
 },
 [0xFF76]={
  category="lo",
  direction="l",
 },
 [0xFF77]={
  category="lo",
  direction="l",
 },
 [0xFF78]={
  category="lo",
  direction="l",
 },
 [0xFF79]={
  category="lo",
  direction="l",
 },
 [0xFF7A]={
  category="lo",
  direction="l",
 },
 [0xFF7B]={
  category="lo",
  direction="l",
 },
 [0xFF7C]={
  category="lo",
  direction="l",
 },
 [0xFF7D]={
  category="lo",
  direction="l",
 },
 [0xFF7E]={
  category="lo",
  direction="l",
 },
 [0xFF7F]={
  category="lo",
  direction="l",
 },
 [0xFF80]={
  category="lo",
  direction="l",
 },
 [0xFF81]={
  category="lo",
  direction="l",
 },
 [0xFF82]={
  category="lo",
  direction="l",
 },
 [0xFF83]={
  category="lo",
  direction="l",
 },
 [0xFF84]={
  category="lo",
  direction="l",
 },
 [0xFF85]={
  category="lo",
  direction="l",
 },
 [0xFF86]={
  category="lo",
  direction="l",
 },
 [0xFF87]={
  category="lo",
  direction="l",
 },
 [0xFF88]={
  category="lo",
  direction="l",
 },
 [0xFF89]={
  category="lo",
  direction="l",
 },
 [0xFF8A]={
  category="lo",
  direction="l",
 },
 [0xFF8B]={
  category="lo",
  direction="l",
 },
 [0xFF8C]={
  category="lo",
  direction="l",
 },
 [0xFF8D]={
  category="lo",
  direction="l",
 },
 [0xFF8E]={
  category="lo",
  direction="l",
 },
 [0xFF8F]={
  category="lo",
  direction="l",
 },
 [0xFF90]={
  category="lo",
  direction="l",
 },
 [0xFF91]={
  category="lo",
  direction="l",
 },
 [0xFF92]={
  category="lo",
  direction="l",
 },
 [0xFF93]={
  category="lo",
  direction="l",
 },
 [0xFF94]={
  category="lo",
  direction="l",
 },
 [0xFF95]={
  category="lo",
  direction="l",
 },
 [0xFF96]={
  category="lo",
  direction="l",
 },
 [0xFF97]={
  category="lo",
  direction="l",
 },
 [0xFF98]={
  category="lo",
  direction="l",
 },
 [0xFF99]={
  category="lo",
  direction="l",
 },
 [0xFF9A]={
  category="lo",
  direction="l",
 },
 [0xFF9B]={
  category="lo",
  direction="l",
 },
 [0xFF9C]={
  category="lo",
  direction="l",
 },
 [0xFF9D]={
  category="lo",
  direction="l",
 },
 [0xFF9E]={
  category="lm",
  direction="l",
 },
 [0xFF9F]={
  category="lm",
  direction="l",
 },
 [0xFFA0]={
  category="lo",
  direction="l",
 },
 [0xFFA1]={
  category="lo",
  direction="l",
 },
 [0xFFA2]={
  category="lo",
  direction="l",
 },
 [0xFFA3]={
  category="lo",
  direction="l",
 },
 [0xFFA4]={
  category="lo",
  direction="l",
 },
 [0xFFA5]={
  category="lo",
  direction="l",
 },
 [0xFFA6]={
  category="lo",
  direction="l",
 },
 [0xFFA7]={
  category="lo",
  direction="l",
 },
 [0xFFA8]={
  category="lo",
  direction="l",
 },
 [0xFFA9]={
  category="lo",
  direction="l",
 },
 [0xFFAA]={
  category="lo",
  direction="l",
 },
 [0xFFAB]={
  category="lo",
  direction="l",
 },
 [0xFFAC]={
  category="lo",
  direction="l",
 },
 [0xFFAD]={
  category="lo",
  direction="l",
 },
 [0xFFAE]={
  category="lo",
  direction="l",
 },
 [0xFFAF]={
  category="lo",
  direction="l",
 },
 [0xFFB0]={
  category="lo",
  direction="l",
 },
 [0xFFB1]={
  category="lo",
  direction="l",
 },
 [0xFFB2]={
  category="lo",
  direction="l",
 },
 [0xFFB3]={
  category="lo",
  direction="l",
 },
 [0xFFB4]={
  category="lo",
  direction="l",
 },
 [0xFFB5]={
  category="lo",
  direction="l",
 },
 [0xFFB6]={
  category="lo",
  direction="l",
 },
 [0xFFB7]={
  category="lo",
  direction="l",
 },
 [0xFFB8]={
  category="lo",
  direction="l",
 },
 [0xFFB9]={
  category="lo",
  direction="l",
 },
 [0xFFBA]={
  category="lo",
  direction="l",
 },
 [0xFFBB]={
  category="lo",
  direction="l",
 },
 [0xFFBC]={
  category="lo",
  direction="l",
 },
 [0xFFBD]={
  category="lo",
  direction="l",
 },
 [0xFFBE]={
  category="lo",
  direction="l",
 },
 [0xFFC2]={
  category="lo",
  direction="l",
 },
 [0xFFC3]={
  category="lo",
  direction="l",
 },
 [0xFFC4]={
  category="lo",
  direction="l",
 },
 [0xFFC5]={
  category="lo",
  direction="l",
 },
 [0xFFC6]={
  category="lo",
  direction="l",
 },
 [0xFFC7]={
  category="lo",
  direction="l",
 },
 [0xFFCA]={
  category="lo",
  direction="l",
 },
 [0xFFCB]={
  category="lo",
  direction="l",
 },
 [0xFFCC]={
  category="lo",
  direction="l",
 },
 [0xFFCD]={
  category="lo",
  direction="l",
 },
 [0xFFCE]={
  category="lo",
  direction="l",
 },
 [0xFFCF]={
  category="lo",
  direction="l",
 },
 [0xFFD2]={
  category="lo",
  direction="l",
 },
 [0xFFD3]={
  category="lo",
  direction="l",
 },
 [0xFFD4]={
  category="lo",
  direction="l",
 },
 [0xFFD5]={
  category="lo",
  direction="l",
 },
 [0xFFD6]={
  category="lo",
  direction="l",
 },
 [0xFFD7]={
  category="lo",
  direction="l",
 },
 [0xFFDA]={
  category="lo",
  direction="l",
 },
 [0xFFDB]={
  category="lo",
  direction="l",
 },
 [0xFFDC]={
  category="lo",
  direction="l",
 },
 [0xFFE0]={
  category="sc",
  direction="et",
 },
 [0xFFE1]={
  category="sc",
  direction="et",
 },
 [0xFFE2]={
  category="sm",
  direction="on",
 },
 [0xFFE3]={
  category="sk",
  direction="on",
 },
 [0xFFE4]={
  category="so",
  direction="on",
 },
 [0xFFE5]={
  category="sc",
  direction="et",
 },
 [0xFFE6]={
  category="sc",
  direction="et",
 },
 [0xFFE8]={
  category="so",
  direction="on",
 },
 [0xFFE9]={
  category="sm",
  direction="on",
 },
 [0xFFEA]={
  category="sm",
  direction="on",
 },
 [0xFFEB]={
  category="sm",
  direction="on",
 },
 [0xFFEC]={
  category="sm",
  direction="on",
 },
 [0xFFED]={
  category="so",
  direction="on",
 },
 [0xFFEE]={
  category="so",
  direction="on",
 },
 [0xFFF9]={
  category="cf",
  direction="on",
 },
 [0xFFFA]={
  category="cf",
  direction="on",
 },
 [0xFFFB]={
  category="cf",
  direction="on",
 },
 [0xFFFC]={
  category="so",
  direction="on",
 },
 [0xFFFD]={
  category="so",
  direction="on",
 },
 [0x10000]={
  category="lo",
  direction="l",
 },
 [0x10001]={
  category="lo",
  direction="l",
 },
 [0x10002]={
  category="lo",
  direction="l",
 },
 [0x10003]={
  category="lo",
  direction="l",
 },
 [0x10004]={
  category="lo",
  direction="l",
 },
 [0x10005]={
  category="lo",
  direction="l",
 },
 [0x10006]={
  category="lo",
  direction="l",
 },
 [0x10007]={
  category="lo",
  direction="l",
 },
 [0x10008]={
  category="lo",
  direction="l",
 },
 [0x10009]={
  category="lo",
  direction="l",
 },
 [0x1000A]={
  category="lo",
  direction="l",
 },
 [0x1000B]={
  category="lo",
  direction="l",
 },
 [0x1000D]={
  category="lo",
  direction="l",
 },
 [0x1000E]={
  category="lo",
  direction="l",
 },
 [0x1000F]={
  category="lo",
  direction="l",
 },
 [0x10010]={
  category="lo",
  direction="l",
 },
 [0x10011]={
  category="lo",
  direction="l",
 },
 [0x10012]={
  category="lo",
  direction="l",
 },
 [0x10013]={
  category="lo",
  direction="l",
 },
 [0x10014]={
  category="lo",
  direction="l",
 },
 [0x10015]={
  category="lo",
  direction="l",
 },
 [0x10016]={
  category="lo",
  direction="l",
 },
 [0x10017]={
  category="lo",
  direction="l",
 },
 [0x10018]={
  category="lo",
  direction="l",
 },
 [0x10019]={
  category="lo",
  direction="l",
 },
 [0x1001A]={
  category="lo",
  direction="l",
 },
 [0x1001B]={
  category="lo",
  direction="l",
 },
 [0x1001C]={
  category="lo",
  direction="l",
 },
 [0x1001D]={
  category="lo",
  direction="l",
 },
 [0x1001E]={
  category="lo",
  direction="l",
 },
 [0x1001F]={
  category="lo",
  direction="l",
 },
 [0x10020]={
  category="lo",
  direction="l",
 },
 [0x10021]={
  category="lo",
  direction="l",
 },
 [0x10022]={
  category="lo",
  direction="l",
 },
 [0x10023]={
  category="lo",
  direction="l",
 },
 [0x10024]={
  category="lo",
  direction="l",
 },
 [0x10025]={
  category="lo",
  direction="l",
 },
 [0x10026]={
  category="lo",
  direction="l",
 },
 [0x10028]={
  category="lo",
  direction="l",
 },
 [0x10029]={
  category="lo",
  direction="l",
 },
 [0x1002A]={
  category="lo",
  direction="l",
 },
 [0x1002B]={
  category="lo",
  direction="l",
 },
 [0x1002C]={
  category="lo",
  direction="l",
 },
 [0x1002D]={
  category="lo",
  direction="l",
 },
 [0x1002E]={
  category="lo",
  direction="l",
 },
 [0x1002F]={
  category="lo",
  direction="l",
 },
 [0x10030]={
  category="lo",
  direction="l",
 },
 [0x10031]={
  category="lo",
  direction="l",
 },
 [0x10032]={
  category="lo",
  direction="l",
 },
 [0x10033]={
  category="lo",
  direction="l",
 },
 [0x10034]={
  category="lo",
  direction="l",
 },
 [0x10035]={
  category="lo",
  direction="l",
 },
 [0x10036]={
  category="lo",
  direction="l",
 },
 [0x10037]={
  category="lo",
  direction="l",
 },
 [0x10038]={
  category="lo",
  direction="l",
 },
 [0x10039]={
  category="lo",
  direction="l",
 },
 [0x1003A]={
  category="lo",
  direction="l",
 },
 [0x1003C]={
  category="lo",
  direction="l",
 },
 [0x1003D]={
  category="lo",
  direction="l",
 },
 [0x1003F]={
  category="lo",
  direction="l",
 },
 [0x10040]={
  category="lo",
  direction="l",
 },
 [0x10041]={
  category="lo",
  direction="l",
 },
 [0x10042]={
  category="lo",
  direction="l",
 },
 [0x10043]={
  category="lo",
  direction="l",
 },
 [0x10044]={
  category="lo",
  direction="l",
 },
 [0x10045]={
  category="lo",
  direction="l",
 },
 [0x10046]={
  category="lo",
  direction="l",
 },
 [0x10047]={
  category="lo",
  direction="l",
 },
 [0x10048]={
  category="lo",
  direction="l",
 },
 [0x10049]={
  category="lo",
  direction="l",
 },
 [0x1004A]={
  category="lo",
  direction="l",
 },
 [0x1004B]={
  category="lo",
  direction="l",
 },
 [0x1004C]={
  category="lo",
  direction="l",
 },
 [0x1004D]={
  category="lo",
  direction="l",
 },
 [0x10050]={
  category="lo",
  direction="l",
 },
 [0x10051]={
  category="lo",
  direction="l",
 },
 [0x10052]={
  category="lo",
  direction="l",
 },
 [0x10053]={
  category="lo",
  direction="l",
 },
 [0x10054]={
  category="lo",
  direction="l",
 },
 [0x10055]={
  category="lo",
  direction="l",
 },
 [0x10056]={
  category="lo",
  direction="l",
 },
 [0x10057]={
  category="lo",
  direction="l",
 },
 [0x10058]={
  category="lo",
  direction="l",
 },
 [0x10059]={
  category="lo",
  direction="l",
 },
 [0x1005A]={
  category="lo",
  direction="l",
 },
 [0x1005B]={
  category="lo",
  direction="l",
 },
 [0x1005C]={
  category="lo",
  direction="l",
 },
 [0x1005D]={
  category="lo",
  direction="l",
 },
 [0x10080]={
  category="lo",
  direction="l",
 },
 [0x10081]={
  category="lo",
  direction="l",
 },
 [0x10082]={
  category="lo",
  direction="l",
 },
 [0x10083]={
  category="lo",
  direction="l",
 },
 [0x10084]={
  category="lo",
  direction="l",
 },
 [0x10085]={
  category="lo",
  direction="l",
 },
 [0x10086]={
  category="lo",
  direction="l",
 },
 [0x10087]={
  category="lo",
  direction="l",
 },
 [0x10088]={
  category="lo",
  direction="l",
 },
 [0x10089]={
  category="lo",
  direction="l",
 },
 [0x1008A]={
  category="lo",
  direction="l",
 },
 [0x1008B]={
  category="lo",
  direction="l",
 },
 [0x1008C]={
  category="lo",
  direction="l",
 },
 [0x1008D]={
  category="lo",
  direction="l",
 },
 [0x1008E]={
  category="lo",
  direction="l",
 },
 [0x1008F]={
  category="lo",
  direction="l",
 },
 [0x10090]={
  category="lo",
  direction="l",
 },
 [0x10091]={
  category="lo",
  direction="l",
 },
 [0x10092]={
  category="lo",
  direction="l",
 },
 [0x10093]={
  category="lo",
  direction="l",
 },
 [0x10094]={
  category="lo",
  direction="l",
 },
 [0x10095]={
  category="lo",
  direction="l",
 },
 [0x10096]={
  category="lo",
  direction="l",
 },
 [0x10097]={
  category="lo",
  direction="l",
 },
 [0x10098]={
  category="lo",
  direction="l",
 },
 [0x10099]={
  category="lo",
  direction="l",
 },
 [0x1009A]={
  category="lo",
  direction="l",
 },
 [0x1009B]={
  category="lo",
  direction="l",
 },
 [0x1009C]={
  category="lo",
  direction="l",
 },
 [0x1009D]={
  category="lo",
  direction="l",
 },
 [0x1009E]={
  category="lo",
  direction="l",
 },
 [0x1009F]={
  category="lo",
  direction="l",
 },
 [0x100A0]={
  category="lo",
  direction="l",
 },
 [0x100A1]={
  category="lo",
  direction="l",
 },
 [0x100A2]={
  category="lo",
  direction="l",
 },
 [0x100A3]={
  category="lo",
  direction="l",
 },
 [0x100A4]={
  category="lo",
  direction="l",
 },
 [0x100A5]={
  category="lo",
  direction="l",
 },
 [0x100A6]={
  category="lo",
  direction="l",
 },
 [0x100A7]={
  category="lo",
  direction="l",
 },
 [0x100A8]={
  category="lo",
  direction="l",
 },
 [0x100A9]={
  category="lo",
  direction="l",
 },
 [0x100AA]={
  category="lo",
  direction="l",
 },
 [0x100AB]={
  category="lo",
  direction="l",
 },
 [0x100AC]={
  category="lo",
  direction="l",
 },
 [0x100AD]={
  category="lo",
  direction="l",
 },
 [0x100AE]={
  category="lo",
  direction="l",
 },
 [0x100AF]={
  category="lo",
  direction="l",
 },
 [0x100B0]={
  category="lo",
  direction="l",
 },
 [0x100B1]={
  category="lo",
  direction="l",
 },
 [0x100B2]={
  category="lo",
  direction="l",
 },
 [0x100B3]={
  category="lo",
  direction="l",
 },
 [0x100B4]={
  category="lo",
  direction="l",
 },
 [0x100B5]={
  category="lo",
  direction="l",
 },
 [0x100B6]={
  category="lo",
  direction="l",
 },
 [0x100B7]={
  category="lo",
  direction="l",
 },
 [0x100B8]={
  category="lo",
  direction="l",
 },
 [0x100B9]={
  category="lo",
  direction="l",
 },
 [0x100BA]={
  category="lo",
  direction="l",
 },
 [0x100BB]={
  category="lo",
  direction="l",
 },
 [0x100BC]={
  category="lo",
  direction="l",
 },
 [0x100BD]={
  category="lo",
  direction="l",
 },
 [0x100BE]={
  category="lo",
  direction="l",
 },
 [0x100BF]={
  category="lo",
  direction="l",
 },
 [0x100C0]={
  category="lo",
  direction="l",
 },
 [0x100C1]={
  category="lo",
  direction="l",
 },
 [0x100C2]={
  category="lo",
  direction="l",
 },
 [0x100C3]={
  category="lo",
  direction="l",
 },
 [0x100C4]={
  category="lo",
  direction="l",
 },
 [0x100C5]={
  category="lo",
  direction="l",
 },
 [0x100C6]={
  category="lo",
  direction="l",
 },
 [0x100C7]={
  category="lo",
  direction="l",
 },
 [0x100C8]={
  category="lo",
  direction="l",
 },
 [0x100C9]={
  category="lo",
  direction="l",
 },
 [0x100CA]={
  category="lo",
  direction="l",
 },
 [0x100CB]={
  category="lo",
  direction="l",
 },
 [0x100CC]={
  category="lo",
  direction="l",
 },
 [0x100CD]={
  category="lo",
  direction="l",
 },
 [0x100CE]={
  category="lo",
  direction="l",
 },
 [0x100CF]={
  category="lo",
  direction="l",
 },
 [0x100D0]={
  category="lo",
  direction="l",
 },
 [0x100D1]={
  category="lo",
  direction="l",
 },
 [0x100D2]={
  category="lo",
  direction="l",
 },
 [0x100D3]={
  category="lo",
  direction="l",
 },
 [0x100D4]={
  category="lo",
  direction="l",
 },
 [0x100D5]={
  category="lo",
  direction="l",
 },
 [0x100D6]={
  category="lo",
  direction="l",
 },
 [0x100D7]={
  category="lo",
  direction="l",
 },
 [0x100D8]={
  category="lo",
  direction="l",
 },
 [0x100D9]={
  category="lo",
  direction="l",
 },
 [0x100DA]={
  category="lo",
  direction="l",
 },
 [0x100DB]={
  category="lo",
  direction="l",
 },
 [0x100DC]={
  category="lo",
  direction="l",
 },
 [0x100DD]={
  category="lo",
  direction="l",
 },
 [0x100DE]={
  category="lo",
  direction="l",
 },
 [0x100DF]={
  category="lo",
  direction="l",
 },
 [0x100E0]={
  category="lo",
  direction="l",
 },
 [0x100E1]={
  category="lo",
  direction="l",
 },
 [0x100E2]={
  category="lo",
  direction="l",
 },
 [0x100E3]={
  category="lo",
  direction="l",
 },
 [0x100E4]={
  category="lo",
  direction="l",
 },
 [0x100E5]={
  category="lo",
  direction="l",
 },
 [0x100E6]={
  category="lo",
  direction="l",
 },
 [0x100E7]={
  category="lo",
  direction="l",
 },
 [0x100E8]={
  category="lo",
  direction="l",
 },
 [0x100E9]={
  category="lo",
  direction="l",
 },
 [0x100EA]={
  category="lo",
  direction="l",
 },
 [0x100EB]={
  category="lo",
  direction="l",
 },
 [0x100EC]={
  category="lo",
  direction="l",
 },
 [0x100ED]={
  category="lo",
  direction="l",
 },
 [0x100EE]={
  category="lo",
  direction="l",
 },
 [0x100EF]={
  category="lo",
  direction="l",
 },
 [0x100F0]={
  category="lo",
  direction="l",
 },
 [0x100F1]={
  category="lo",
  direction="l",
 },
 [0x100F2]={
  category="lo",
  direction="l",
 },
 [0x100F3]={
  category="lo",
  direction="l",
 },
 [0x100F4]={
  category="lo",
  direction="l",
 },
 [0x100F5]={
  category="lo",
  direction="l",
 },
 [0x100F6]={
  category="lo",
  direction="l",
 },
 [0x100F7]={
  category="lo",
  direction="l",
 },
 [0x100F8]={
  category="lo",
  direction="l",
 },
 [0x100F9]={
  category="lo",
  direction="l",
 },
 [0x100FA]={
  category="lo",
  direction="l",
 },
 [0x10100]={
  category="po",
  direction="l",
 },
 [0x10101]={
  category="po",
  direction="on",
 },
 [0x10102]={
  category="so",
  direction="l",
 },
 [0x10107]={
  category="no",
  direction="l",
 },
 [0x10108]={
  category="no",
  direction="l",
 },
 [0x10109]={
  category="no",
  direction="l",
 },
 [0x1010A]={
  category="no",
  direction="l",
 },
 [0x1010B]={
  category="no",
  direction="l",
 },
 [0x1010C]={
  category="no",
  direction="l",
 },
 [0x1010D]={
  category="no",
  direction="l",
 },
 [0x1010E]={
  category="no",
  direction="l",
 },
 [0x1010F]={
  category="no",
  direction="l",
 },
 [0x10110]={
  category="no",
  direction="l",
 },
 [0x10111]={
  category="no",
  direction="l",
 },
 [0x10112]={
  category="no",
  direction="l",
 },
 [0x10113]={
  category="no",
  direction="l",
 },
 [0x10114]={
  category="no",
  direction="l",
 },
 [0x10115]={
  category="no",
  direction="l",
 },
 [0x10116]={
  category="no",
  direction="l",
 },
 [0x10117]={
  category="no",
  direction="l",
 },
 [0x10118]={
  category="no",
  direction="l",
 },
 [0x10119]={
  category="no",
  direction="l",
 },
 [0x1011A]={
  category="no",
  direction="l",
 },
 [0x1011B]={
  category="no",
  direction="l",
 },
 [0x1011C]={
  category="no",
  direction="l",
 },
 [0x1011D]={
  category="no",
  direction="l",
 },
 [0x1011E]={
  category="no",
  direction="l",
 },
 [0x1011F]={
  category="no",
  direction="l",
 },
 [0x10120]={
  category="no",
  direction="l",
 },
 [0x10121]={
  category="no",
  direction="l",
 },
 [0x10122]={
  category="no",
  direction="l",
 },
 [0x10123]={
  category="no",
  direction="l",
 },
 [0x10124]={
  category="no",
  direction="l",
 },
 [0x10125]={
  category="no",
  direction="l",
 },
 [0x10126]={
  category="no",
  direction="l",
 },
 [0x10127]={
  category="no",
  direction="l",
 },
 [0x10128]={
  category="no",
  direction="l",
 },
 [0x10129]={
  category="no",
  direction="l",
 },
 [0x1012A]={
  category="no",
  direction="l",
 },
 [0x1012B]={
  category="no",
  direction="l",
 },
 [0x1012C]={
  category="no",
  direction="l",
 },
 [0x1012D]={
  category="no",
  direction="l",
 },
 [0x1012E]={
  category="no",
  direction="l",
 },
 [0x1012F]={
  category="no",
  direction="l",
 },
 [0x10130]={
  category="no",
  direction="l",
 },
 [0x10131]={
  category="no",
  direction="l",
 },
 [0x10132]={
  category="no",
  direction="l",
 },
 [0x10133]={
  category="no",
  direction="l",
 },
 [0x10137]={
  category="so",
  direction="l",
 },
 [0x10138]={
  category="so",
  direction="l",
 },
 [0x10139]={
  category="so",
  direction="l",
 },
 [0x1013A]={
  category="so",
  direction="l",
 },
 [0x1013B]={
  category="so",
  direction="l",
 },
 [0x1013C]={
  category="so",
  direction="l",
 },
 [0x1013D]={
  category="so",
  direction="l",
 },
 [0x1013E]={
  category="so",
  direction="l",
 },
 [0x1013F]={
  category="so",
  direction="l",
 },
 [0x10140]={
  category="nl",
  direction="on",
 },
 [0x10141]={
  category="nl",
  direction="on",
 },
 [0x10142]={
  category="nl",
  direction="on",
 },
 [0x10143]={
  category="nl",
  direction="on",
 },
 [0x10144]={
  category="nl",
  direction="on",
 },
 [0x10145]={
  category="nl",
  direction="on",
 },
 [0x10146]={
  category="nl",
  direction="on",
 },
 [0x10147]={
  category="nl",
  direction="on",
 },
 [0x10148]={
  category="nl",
  direction="on",
 },
 [0x10149]={
  category="nl",
  direction="on",
 },
 [0x1014A]={
  category="nl",
  direction="on",
 },
 [0x1014B]={
  category="nl",
  direction="on",
 },
 [0x1014C]={
  category="nl",
  direction="on",
 },
 [0x1014D]={
  category="nl",
  direction="on",
 },
 [0x1014E]={
  category="nl",
  direction="on",
 },
 [0x1014F]={
  category="nl",
  direction="on",
 },
 [0x10150]={
  category="nl",
  direction="on",
 },
 [0x10151]={
  category="nl",
  direction="on",
 },
 [0x10152]={
  category="nl",
  direction="on",
 },
 [0x10153]={
  category="nl",
  direction="on",
 },
 [0x10154]={
  category="nl",
  direction="on",
 },
 [0x10155]={
  category="nl",
  direction="on",
 },
 [0x10156]={
  category="nl",
  direction="on",
 },
 [0x10157]={
  category="nl",
  direction="on",
 },
 [0x10158]={
  category="nl",
  direction="on",
 },
 [0x10159]={
  category="nl",
  direction="on",
 },
 [0x1015A]={
  category="nl",
  direction="on",
 },
 [0x1015B]={
  category="nl",
  direction="on",
 },
 [0x1015C]={
  category="nl",
  direction="on",
 },
 [0x1015D]={
  category="nl",
  direction="on",
 },
 [0x1015E]={
  category="nl",
  direction="on",
 },
 [0x1015F]={
  category="nl",
  direction="on",
 },
 [0x10160]={
  category="nl",
  direction="on",
 },
 [0x10161]={
  category="nl",
  direction="on",
 },
 [0x10162]={
  category="nl",
  direction="on",
 },
 [0x10163]={
  category="nl",
  direction="on",
 },
 [0x10164]={
  category="nl",
  direction="on",
 },
 [0x10165]={
  category="nl",
  direction="on",
 },
 [0x10166]={
  category="nl",
  direction="on",
 },
 [0x10167]={
  category="nl",
  direction="on",
 },
 [0x10168]={
  category="nl",
  direction="on",
 },
 [0x10169]={
  category="nl",
  direction="on",
 },
 [0x1016A]={
  category="nl",
  direction="on",
 },
 [0x1016B]={
  category="nl",
  direction="on",
 },
 [0x1016C]={
  category="nl",
  direction="on",
 },
 [0x1016D]={
  category="nl",
  direction="on",
 },
 [0x1016E]={
  category="nl",
  direction="on",
 },
 [0x1016F]={
  category="nl",
  direction="on",
 },
 [0x10170]={
  category="nl",
  direction="on",
 },
 [0x10171]={
  category="nl",
  direction="on",
 },
 [0x10172]={
  category="nl",
  direction="on",
 },
 [0x10173]={
  category="nl",
  direction="on",
 },
 [0x10174]={
  category="nl",
  direction="on",
 },
 [0x10175]={
  category="no",
  direction="on",
 },
 [0x10176]={
  category="no",
  direction="on",
 },
 [0x10177]={
  category="no",
  direction="on",
 },
 [0x10178]={
  category="no",
  direction="on",
 },
 [0x10179]={
  category="so",
  direction="on",
 },
 [0x1017A]={
  category="so",
  direction="on",
 },
 [0x1017B]={
  category="so",
  direction="on",
 },
 [0x1017C]={
  category="so",
  direction="on",
 },
 [0x1017D]={
  category="so",
  direction="on",
 },
 [0x1017E]={
  category="so",
  direction="on",
 },
 [0x1017F]={
  category="so",
  direction="on",
 },
 [0x10180]={
  category="so",
  direction="on",
 },
 [0x10181]={
  category="so",
  direction="on",
 },
 [0x10182]={
  category="so",
  direction="on",
 },
 [0x10183]={
  category="so",
  direction="on",
 },
 [0x10184]={
  category="so",
  direction="on",
 },
 [0x10185]={
  category="so",
  direction="on",
 },
 [0x10186]={
  category="so",
  direction="on",
 },
 [0x10187]={
  category="so",
  direction="on",
 },
 [0x10188]={
  category="so",
  direction="on",
 },
 [0x10189]={
  category="so",
  direction="on",
 },
 [0x1018A]={
  category="no",
  direction="on",
 },
 [0x1018B]={
  category="no",
  direction="on",
 },
 [0x1018C]={
  category="so",
  direction="on",
 },
 [0x10190]={
  category="so",
  direction="on",
 },
 [0x10191]={
  category="so",
  direction="on",
 },
 [0x10192]={
  category="so",
  direction="on",
 },
 [0x10193]={
  category="so",
  direction="on",
 },
 [0x10194]={
  category="so",
  direction="on",
 },
 [0x10195]={
  category="so",
  direction="on",
 },
 [0x10196]={
  category="so",
  direction="on",
 },
 [0x10197]={
  category="so",
  direction="on",
 },
 [0x10198]={
  category="so",
  direction="on",
 },
 [0x10199]={
  category="so",
  direction="on",
 },
 [0x1019A]={
  category="so",
  direction="on",
 },
 [0x1019B]={
  category="so",
  direction="on",
 },
 [0x101A0]={
  category="so",
  direction="on",
 },
 [0x101D0]={
  category="so",
  direction="l",
 },
 [0x101D1]={
  category="so",
  direction="l",
 },
 [0x101D2]={
  category="so",
  direction="l",
 },
 [0x101D3]={
  category="so",
  direction="l",
 },
 [0x101D4]={
  category="so",
  direction="l",
 },
 [0x101D5]={
  category="so",
  direction="l",
 },
 [0x101D6]={
  category="so",
  direction="l",
 },
 [0x101D7]={
  category="so",
  direction="l",
 },
 [0x101D8]={
  category="so",
  direction="l",
 },
 [0x101D9]={
  category="so",
  direction="l",
 },
 [0x101DA]={
  category="so",
  direction="l",
 },
 [0x101DB]={
  category="so",
  direction="l",
 },
 [0x101DC]={
  category="so",
  direction="l",
 },
 [0x101DD]={
  category="so",
  direction="l",
 },
 [0x101DE]={
  category="so",
  direction="l",
 },
 [0x101DF]={
  category="so",
  direction="l",
 },
 [0x101E0]={
  category="so",
  direction="l",
 },
 [0x101E1]={
  category="so",
  direction="l",
 },
 [0x101E2]={
  category="so",
  direction="l",
 },
 [0x101E3]={
  category="so",
  direction="l",
 },
 [0x101E4]={
  category="so",
  direction="l",
 },
 [0x101E5]={
  category="so",
  direction="l",
 },
 [0x101E6]={
  category="so",
  direction="l",
 },
 [0x101E7]={
  category="so",
  direction="l",
 },
 [0x101E8]={
  category="so",
  direction="l",
 },
 [0x101E9]={
  category="so",
  direction="l",
 },
 [0x101EA]={
  category="so",
  direction="l",
 },
 [0x101EB]={
  category="so",
  direction="l",
 },
 [0x101EC]={
  category="so",
  direction="l",
 },
 [0x101ED]={
  category="so",
  direction="l",
 },
 [0x101EE]={
  category="so",
  direction="l",
 },
 [0x101EF]={
  category="so",
  direction="l",
 },
 [0x101F0]={
  category="so",
  direction="l",
 },
 [0x101F1]={
  category="so",
  direction="l",
 },
 [0x101F2]={
  category="so",
  direction="l",
 },
 [0x101F3]={
  category="so",
  direction="l",
 },
 [0x101F4]={
  category="so",
  direction="l",
 },
 [0x101F5]={
  category="so",
  direction="l",
 },
 [0x101F6]={
  category="so",
  direction="l",
 },
 [0x101F7]={
  category="so",
  direction="l",
 },
 [0x101F8]={
  category="so",
  direction="l",
 },
 [0x101F9]={
  category="so",
  direction="l",
 },
 [0x101FA]={
  category="so",
  direction="l",
 },
 [0x101FB]={
  category="so",
  direction="l",
 },
 [0x101FC]={
  category="so",
  direction="l",
 },
 [0x101FD]={
  category="mn",
  direction="nsm",
 },
 [0x10280]={
  category="lo",
  direction="l",
 },
 [0x10281]={
  category="lo",
  direction="l",
 },
 [0x10282]={
  category="lo",
  direction="l",
 },
 [0x10283]={
  category="lo",
  direction="l",
 },
 [0x10284]={
  category="lo",
  direction="l",
 },
 [0x10285]={
  category="lo",
  direction="l",
 },
 [0x10286]={
  category="lo",
  direction="l",
 },
 [0x10287]={
  category="lo",
  direction="l",
 },
 [0x10288]={
  category="lo",
  direction="l",
 },
 [0x10289]={
  category="lo",
  direction="l",
 },
 [0x1028A]={
  category="lo",
  direction="l",
 },
 [0x1028B]={
  category="lo",
  direction="l",
 },
 [0x1028C]={
  category="lo",
  direction="l",
 },
 [0x1028D]={
  category="lo",
  direction="l",
 },
 [0x1028E]={
  category="lo",
  direction="l",
 },
 [0x1028F]={
  category="lo",
  direction="l",
 },
 [0x10290]={
  category="lo",
  direction="l",
 },
 [0x10291]={
  category="lo",
  direction="l",
 },
 [0x10292]={
  category="lo",
  direction="l",
 },
 [0x10293]={
  category="lo",
  direction="l",
 },
 [0x10294]={
  category="lo",
  direction="l",
 },
 [0x10295]={
  category="lo",
  direction="l",
 },
 [0x10296]={
  category="lo",
  direction="l",
 },
 [0x10297]={
  category="lo",
  direction="l",
 },
 [0x10298]={
  category="lo",
  direction="l",
 },
 [0x10299]={
  category="lo",
  direction="l",
 },
 [0x1029A]={
  category="lo",
  direction="l",
 },
 [0x1029B]={
  category="lo",
  direction="l",
 },
 [0x1029C]={
  category="lo",
  direction="l",
 },
 [0x102A0]={
  category="lo",
  direction="l",
 },
 [0x102A1]={
  category="lo",
  direction="l",
 },
 [0x102A2]={
  category="lo",
  direction="l",
 },
 [0x102A3]={
  category="lo",
  direction="l",
 },
 [0x102A4]={
  category="lo",
  direction="l",
 },
 [0x102A5]={
  category="lo",
  direction="l",
 },
 [0x102A6]={
  category="lo",
  direction="l",
 },
 [0x102A7]={
  category="lo",
  direction="l",
 },
 [0x102A8]={
  category="lo",
  direction="l",
 },
 [0x102A9]={
  category="lo",
  direction="l",
 },
 [0x102AA]={
  category="lo",
  direction="l",
 },
 [0x102AB]={
  category="lo",
  direction="l",
 },
 [0x102AC]={
  category="lo",
  direction="l",
 },
 [0x102AD]={
  category="lo",
  direction="l",
 },
 [0x102AE]={
  category="lo",
  direction="l",
 },
 [0x102AF]={
  category="lo",
  direction="l",
 },
 [0x102B0]={
  category="lo",
  direction="l",
 },
 [0x102B1]={
  category="lo",
  direction="l",
 },
 [0x102B2]={
  category="lo",
  direction="l",
 },
 [0x102B3]={
  category="lo",
  direction="l",
 },
 [0x102B4]={
  category="lo",
  direction="l",
 },
 [0x102B5]={
  category="lo",
  direction="l",
 },
 [0x102B6]={
  category="lo",
  direction="l",
 },
 [0x102B7]={
  category="lo",
  direction="l",
 },
 [0x102B8]={
  category="lo",
  direction="l",
 },
 [0x102B9]={
  category="lo",
  direction="l",
 },
 [0x102BA]={
  category="lo",
  direction="l",
 },
 [0x102BB]={
  category="lo",
  direction="l",
 },
 [0x102BC]={
  category="lo",
  direction="l",
 },
 [0x102BD]={
  category="lo",
  direction="l",
 },
 [0x102BE]={
  category="lo",
  direction="l",
 },
 [0x102BF]={
  category="lo",
  direction="l",
 },
 [0x102C0]={
  category="lo",
  direction="l",
 },
 [0x102C1]={
  category="lo",
  direction="l",
 },
 [0x102C2]={
  category="lo",
  direction="l",
 },
 [0x102C3]={
  category="lo",
  direction="l",
 },
 [0x102C4]={
  category="lo",
  direction="l",
 },
 [0x102C5]={
  category="lo",
  direction="l",
 },
 [0x102C6]={
  category="lo",
  direction="l",
 },
 [0x102C7]={
  category="lo",
  direction="l",
 },
 [0x102C8]={
  category="lo",
  direction="l",
 },
 [0x102C9]={
  category="lo",
  direction="l",
 },
 [0x102CA]={
  category="lo",
  direction="l",
 },
 [0x102CB]={
  category="lo",
  direction="l",
 },
 [0x102CC]={
  category="lo",
  direction="l",
 },
 [0x102CD]={
  category="lo",
  direction="l",
 },
 [0x102CE]={
  category="lo",
  direction="l",
 },
 [0x102CF]={
  category="lo",
  direction="l",
 },
 [0x102D0]={
  category="lo",
  direction="l",
 },
 [0x102E0]={
  category="mn",
  direction="nsm",
 },
 [0x102E1]={
  category="no",
  direction="en",
 },
 [0x102E2]={
  category="no",
  direction="en",
 },
 [0x102E3]={
  category="no",
  direction="en",
 },
 [0x102E4]={
  category="no",
  direction="en",
 },
 [0x102E5]={
  category="no",
  direction="en",
 },
 [0x102E6]={
  category="no",
  direction="en",
 },
 [0x102E7]={
  category="no",
  direction="en",
 },
 [0x102E8]={
  category="no",
  direction="en",
 },
 [0x102E9]={
  category="no",
  direction="en",
 },
 [0x102EA]={
  category="no",
  direction="en",
 },
 [0x102EB]={
  category="no",
  direction="en",
 },
 [0x102EC]={
  category="no",
  direction="en",
 },
 [0x102ED]={
  category="no",
  direction="en",
 },
 [0x102EE]={
  category="no",
  direction="en",
 },
 [0x102EF]={
  category="no",
  direction="en",
 },
 [0x102F0]={
  category="no",
  direction="en",
 },
 [0x102F1]={
  category="no",
  direction="en",
 },
 [0x102F2]={
  category="no",
  direction="en",
 },
 [0x102F3]={
  category="no",
  direction="en",
 },
 [0x102F4]={
  category="no",
  direction="en",
 },
 [0x102F5]={
  category="no",
  direction="en",
 },
 [0x102F6]={
  category="no",
  direction="en",
 },
 [0x102F7]={
  category="no",
  direction="en",
 },
 [0x102F8]={
  category="no",
  direction="en",
 },
 [0x102F9]={
  category="no",
  direction="en",
 },
 [0x102FA]={
  category="no",
  direction="en",
 },
 [0x102FB]={
  category="no",
  direction="en",
 },
 [0x10300]={
  category="lo",
  direction="l",
 },
 [0x10301]={
  category="lo",
  direction="l",
 },
 [0x10302]={
  category="lo",
  direction="l",
 },
 [0x10303]={
  category="lo",
  direction="l",
 },
 [0x10304]={
  category="lo",
  direction="l",
 },
 [0x10305]={
  category="lo",
  direction="l",
 },
 [0x10306]={
  category="lo",
  direction="l",
 },
 [0x10307]={
  category="lo",
  direction="l",
 },
 [0x10308]={
  category="lo",
  direction="l",
 },
 [0x10309]={
  category="lo",
  direction="l",
 },
 [0x1030A]={
  category="lo",
  direction="l",
 },
 [0x1030B]={
  category="lo",
  direction="l",
 },
 [0x1030C]={
  category="lo",
  direction="l",
 },
 [0x1030D]={
  category="lo",
  direction="l",
 },
 [0x1030E]={
  category="lo",
  direction="l",
 },
 [0x1030F]={
  category="lo",
  direction="l",
 },
 [0x10310]={
  category="lo",
  direction="l",
 },
 [0x10311]={
  category="lo",
  direction="l",
 },
 [0x10312]={
  category="lo",
  direction="l",
 },
 [0x10313]={
  category="lo",
  direction="l",
 },
 [0x10314]={
  category="lo",
  direction="l",
 },
 [0x10315]={
  category="lo",
  direction="l",
 },
 [0x10316]={
  category="lo",
  direction="l",
 },
 [0x10317]={
  category="lo",
  direction="l",
 },
 [0x10318]={
  category="lo",
  direction="l",
 },
 [0x10319]={
  category="lo",
  direction="l",
 },
 [0x1031A]={
  category="lo",
  direction="l",
 },
 [0x1031B]={
  category="lo",
  direction="l",
 },
 [0x1031C]={
  category="lo",
  direction="l",
 },
 [0x1031D]={
  category="lo",
  direction="l",
 },
 [0x1031E]={
  category="lo",
  direction="l",
 },
 [0x1031F]={
  category="lo",
  direction="l",
 },
 [0x10320]={
  category="no",
  direction="l",
 },
 [0x10321]={
  category="no",
  direction="l",
 },
 [0x10322]={
  category="no",
  direction="l",
 },
 [0x10323]={
  category="no",
  direction="l",
 },
 [0x10330]={
  category="lo",
  direction="l",
 },
 [0x10331]={
  category="lo",
  direction="l",
 },
 [0x10332]={
  category="lo",
  direction="l",
 },
 [0x10333]={
  category="lo",
  direction="l",
 },
 [0x10334]={
  category="lo",
  direction="l",
 },
 [0x10335]={
  category="lo",
  direction="l",
 },
 [0x10336]={
  category="lo",
  direction="l",
 },
 [0x10337]={
  category="lo",
  direction="l",
 },
 [0x10338]={
  category="lo",
  direction="l",
 },
 [0x10339]={
  category="lo",
  direction="l",
 },
 [0x1033A]={
  category="lo",
  direction="l",
 },
 [0x1033B]={
  category="lo",
  direction="l",
 },
 [0x1033C]={
  category="lo",
  direction="l",
 },
 [0x1033D]={
  category="lo",
  direction="l",
 },
 [0x1033E]={
  category="lo",
  direction="l",
 },
 [0x1033F]={
  category="lo",
  direction="l",
 },
 [0x10340]={
  category="lo",
  direction="l",
 },
 [0x10341]={
  category="nl",
  direction="l",
 },
 [0x10342]={
  category="lo",
  direction="l",
 },
 [0x10343]={
  category="lo",
  direction="l",
 },
 [0x10344]={
  category="lo",
  direction="l",
 },
 [0x10345]={
  category="lo",
  direction="l",
 },
 [0x10346]={
  category="lo",
  direction="l",
 },
 [0x10347]={
  category="lo",
  direction="l",
 },
 [0x10348]={
  category="lo",
  direction="l",
 },
 [0x10349]={
  category="lo",
  direction="l",
 },
 [0x1034A]={
  category="nl",
  direction="l",
 },
 [0x10350]={
  category="lo",
  direction="l",
 },
 [0x10351]={
  category="lo",
  direction="l",
 },
 [0x10352]={
  category="lo",
  direction="l",
 },
 [0x10353]={
  category="lo",
  direction="l",
 },
 [0x10354]={
  category="lo",
  direction="l",
 },
 [0x10355]={
  category="lo",
  direction="l",
 },
 [0x10356]={
  category="lo",
  direction="l",
 },
 [0x10357]={
  category="lo",
  direction="l",
 },
 [0x10358]={
  category="lo",
  direction="l",
 },
 [0x10359]={
  category="lo",
  direction="l",
 },
 [0x1035A]={
  category="lo",
  direction="l",
 },
 [0x1035B]={
  category="lo",
  direction="l",
 },
 [0x1035C]={
  category="lo",
  direction="l",
 },
 [0x1035D]={
  category="lo",
  direction="l",
 },
 [0x1035E]={
  category="lo",
  direction="l",
 },
 [0x1035F]={
  category="lo",
  direction="l",
 },
 [0x10360]={
  category="lo",
  direction="l",
 },
 [0x10361]={
  category="lo",
  direction="l",
 },
 [0x10362]={
  category="lo",
  direction="l",
 },
 [0x10363]={
  category="lo",
  direction="l",
 },
 [0x10364]={
  category="lo",
  direction="l",
 },
 [0x10365]={
  category="lo",
  direction="l",
 },
 [0x10366]={
  category="lo",
  direction="l",
 },
 [0x10367]={
  category="lo",
  direction="l",
 },
 [0x10368]={
  category="lo",
  direction="l",
 },
 [0x10369]={
  category="lo",
  direction="l",
 },
 [0x1036A]={
  category="lo",
  direction="l",
 },
 [0x1036B]={
  category="lo",
  direction="l",
 },
 [0x1036C]={
  category="lo",
  direction="l",
 },
 [0x1036D]={
  category="lo",
  direction="l",
 },
 [0x1036E]={
  category="lo",
  direction="l",
 },
 [0x1036F]={
  category="lo",
  direction="l",
 },
 [0x10370]={
  category="lo",
  direction="l",
 },
 [0x10371]={
  category="lo",
  direction="l",
 },
 [0x10372]={
  category="lo",
  direction="l",
 },
 [0x10373]={
  category="lo",
  direction="l",
 },
 [0x10374]={
  category="lo",
  direction="l",
 },
 [0x10375]={
  category="lo",
  direction="l",
 },
 [0x10376]={
  category="mn",
  direction="nsm",
 },
 [0x10377]={
  category="mn",
  direction="nsm",
 },
 [0x10378]={
  category="mn",
  direction="nsm",
 },
 [0x10379]={
  category="mn",
  direction="nsm",
 },
 [0x1037A]={
  category="mn",
  direction="nsm",
 },
 [0x10380]={
  category="lo",
  direction="l",
 },
 [0x10381]={
  category="lo",
  direction="l",
 },
 [0x10382]={
  category="lo",
  direction="l",
 },
 [0x10383]={
  category="lo",
  direction="l",
 },
 [0x10384]={
  category="lo",
  direction="l",
 },
 [0x10385]={
  category="lo",
  direction="l",
 },
 [0x10386]={
  category="lo",
  direction="l",
 },
 [0x10387]={
  category="lo",
  direction="l",
 },
 [0x10388]={
  category="lo",
  direction="l",
 },
 [0x10389]={
  category="lo",
  direction="l",
 },
 [0x1038A]={
  category="lo",
  direction="l",
 },
 [0x1038B]={
  category="lo",
  direction="l",
 },
 [0x1038C]={
  category="lo",
  direction="l",
 },
 [0x1038D]={
  category="lo",
  direction="l",
 },
 [0x1038E]={
  category="lo",
  direction="l",
 },
 [0x1038F]={
  category="lo",
  direction="l",
 },
 [0x10390]={
  category="lo",
  direction="l",
 },
 [0x10391]={
  category="lo",
  direction="l",
 },
 [0x10392]={
  category="lo",
  direction="l",
 },
 [0x10393]={
  category="lo",
  direction="l",
 },
 [0x10394]={
  category="lo",
  direction="l",
 },
 [0x10395]={
  category="lo",
  direction="l",
 },
 [0x10396]={
  category="lo",
  direction="l",
 },
 [0x10397]={
  category="lo",
  direction="l",
 },
 [0x10398]={
  category="lo",
  direction="l",
 },
 [0x10399]={
  category="lo",
  direction="l",
 },
 [0x1039A]={
  category="lo",
  direction="l",
 },
 [0x1039B]={
  category="lo",
  direction="l",
 },
 [0x1039C]={
  category="lo",
  direction="l",
 },
 [0x1039D]={
  category="lo",
  direction="l",
 },
 [0x1039F]={
  category="po",
  direction="l",
 },
 [0x103A0]={
  category="lo",
  direction="l",
 },
 [0x103A1]={
  category="lo",
  direction="l",
 },
 [0x103A2]={
  category="lo",
  direction="l",
 },
 [0x103A3]={
  category="lo",
  direction="l",
 },
 [0x103A4]={
  category="lo",
  direction="l",
 },
 [0x103A5]={
  category="lo",
  direction="l",
 },
 [0x103A6]={
  category="lo",
  direction="l",
 },
 [0x103A7]={
  category="lo",
  direction="l",
 },
 [0x103A8]={
  category="lo",
  direction="l",
 },
 [0x103A9]={
  category="lo",
  direction="l",
 },
 [0x103AA]={
  category="lo",
  direction="l",
 },
 [0x103AB]={
  category="lo",
  direction="l",
 },
 [0x103AC]={
  category="lo",
  direction="l",
 },
 [0x103AD]={
  category="lo",
  direction="l",
 },
 [0x103AE]={
  category="lo",
  direction="l",
 },
 [0x103AF]={
  category="lo",
  direction="l",
 },
 [0x103B0]={
  category="lo",
  direction="l",
 },
 [0x103B1]={
  category="lo",
  direction="l",
 },
 [0x103B2]={
  category="lo",
  direction="l",
 },
 [0x103B3]={
  category="lo",
  direction="l",
 },
 [0x103B4]={
  category="lo",
  direction="l",
 },
 [0x103B5]={
  category="lo",
  direction="l",
 },
 [0x103B6]={
  category="lo",
  direction="l",
 },
 [0x103B7]={
  category="lo",
  direction="l",
 },
 [0x103B8]={
  category="lo",
  direction="l",
 },
 [0x103B9]={
  category="lo",
  direction="l",
 },
 [0x103BA]={
  category="lo",
  direction="l",
 },
 [0x103BB]={
  category="lo",
  direction="l",
 },
 [0x103BC]={
  category="lo",
  direction="l",
 },
 [0x103BD]={
  category="lo",
  direction="l",
 },
 [0x103BE]={
  category="lo",
  direction="l",
 },
 [0x103BF]={
  category="lo",
  direction="l",
 },
 [0x103C0]={
  category="lo",
  direction="l",
 },
 [0x103C1]={
  category="lo",
  direction="l",
 },
 [0x103C2]={
  category="lo",
  direction="l",
 },
 [0x103C3]={
  category="lo",
  direction="l",
 },
 [0x103C8]={
  category="lo",
  direction="l",
 },
 [0x103C9]={
  category="lo",
  direction="l",
 },
 [0x103CA]={
  category="lo",
  direction="l",
 },
 [0x103CB]={
  category="lo",
  direction="l",
 },
 [0x103CC]={
  category="lo",
  direction="l",
 },
 [0x103CD]={
  category="lo",
  direction="l",
 },
 [0x103CE]={
  category="lo",
  direction="l",
 },
 [0x103CF]={
  category="lo",
  direction="l",
 },
 [0x103D0]={
  category="po",
  direction="l",
 },
 [0x103D1]={
  category="nl",
  direction="l",
 },
 [0x103D2]={
  category="nl",
  direction="l",
 },
 [0x103D3]={
  category="nl",
  direction="l",
 },
 [0x103D4]={
  category="nl",
  direction="l",
 },
 [0x103D5]={
  category="nl",
  direction="l",
 },
 [0x10400]={
  category="lu",
  direction="l",
 },
 [0x10401]={
  category="lu",
  direction="l",
 },
 [0x10402]={
  category="lu",
  direction="l",
 },
 [0x10403]={
  category="lu",
  direction="l",
 },
 [0x10404]={
  category="lu",
  direction="l",
 },
 [0x10405]={
  category="lu",
  direction="l",
 },
 [0x10406]={
  category="lu",
  direction="l",
 },
 [0x10407]={
  category="lu",
  direction="l",
 },
 [0x10408]={
  category="lu",
  direction="l",
 },
 [0x10409]={
  category="lu",
  direction="l",
 },
 [0x1040A]={
  category="lu",
  direction="l",
 },
 [0x1040B]={
  category="lu",
  direction="l",
 },
 [0x1040C]={
  category="lu",
  direction="l",
 },
 [0x1040D]={
  category="lu",
  direction="l",
 },
 [0x1040E]={
  category="lu",
  direction="l",
 },
 [0x1040F]={
  category="lu",
  direction="l",
 },
 [0x10410]={
  category="lu",
  direction="l",
 },
 [0x10411]={
  category="lu",
  direction="l",
 },
 [0x10412]={
  category="lu",
  direction="l",
 },
 [0x10413]={
  category="lu",
  direction="l",
 },
 [0x10414]={
  category="lu",
  direction="l",
 },
 [0x10415]={
  category="lu",
  direction="l",
 },
 [0x10416]={
  category="lu",
  direction="l",
 },
 [0x10417]={
  category="lu",
  direction="l",
 },
 [0x10418]={
  category="lu",
  direction="l",
 },
 [0x10419]={
  category="lu",
  direction="l",
 },
 [0x1041A]={
  category="lu",
  direction="l",
 },
 [0x1041B]={
  category="lu",
  direction="l",
 },
 [0x1041C]={
  category="lu",
  direction="l",
 },
 [0x1041D]={
  category="lu",
  direction="l",
 },
 [0x1041E]={
  category="lu",
  direction="l",
 },
 [0x1041F]={
  category="lu",
  direction="l",
 },
 [0x10420]={
  category="lu",
  direction="l",
 },
 [0x10421]={
  category="lu",
  direction="l",
 },
 [0x10422]={
  category="lu",
  direction="l",
 },
 [0x10423]={
  category="lu",
  direction="l",
 },
 [0x10424]={
  category="lu",
  direction="l",
 },
 [0x10425]={
  category="lu",
  direction="l",
 },
 [0x10426]={
  category="lu",
  direction="l",
 },
 [0x10427]={
  category="lu",
  direction="l",
 },
 [0x10428]={
  category="ll",
  direction="l",
 },
 [0x10429]={
  category="ll",
  direction="l",
 },
 [0x1042A]={
  category="ll",
  direction="l",
 },
 [0x1042B]={
  category="ll",
  direction="l",
 },
 [0x1042C]={
  category="ll",
  direction="l",
 },
 [0x1042D]={
  category="ll",
  direction="l",
 },
 [0x1042E]={
  category="ll",
  direction="l",
 },
 [0x1042F]={
  category="ll",
  direction="l",
 },
 [0x10430]={
  category="ll",
  direction="l",
 },
 [0x10431]={
  category="ll",
  direction="l",
 },
 [0x10432]={
  category="ll",
  direction="l",
 },
 [0x10433]={
  category="ll",
  direction="l",
 },
 [0x10434]={
  category="ll",
  direction="l",
 },
 [0x10435]={
  category="ll",
  direction="l",
 },
 [0x10436]={
  category="ll",
  direction="l",
 },
 [0x10437]={
  category="ll",
  direction="l",
 },
 [0x10438]={
  category="ll",
  direction="l",
 },
 [0x10439]={
  category="ll",
  direction="l",
 },
 [0x1043A]={
  category="ll",
  direction="l",
 },
 [0x1043B]={
  category="ll",
  direction="l",
 },
 [0x1043C]={
  category="ll",
  direction="l",
 },
 [0x1043D]={
  category="ll",
  direction="l",
 },
 [0x1043E]={
  category="ll",
  direction="l",
 },
 [0x1043F]={
  category="ll",
  direction="l",
 },
 [0x10440]={
  category="ll",
  direction="l",
 },
 [0x10441]={
  category="ll",
  direction="l",
 },
 [0x10442]={
  category="ll",
  direction="l",
 },
 [0x10443]={
  category="ll",
  direction="l",
 },
 [0x10444]={
  category="ll",
  direction="l",
 },
 [0x10445]={
  category="ll",
  direction="l",
 },
 [0x10446]={
  category="ll",
  direction="l",
 },
 [0x10447]={
  category="ll",
  direction="l",
 },
 [0x10448]={
  category="ll",
  direction="l",
 },
 [0x10449]={
  category="ll",
  direction="l",
 },
 [0x1044A]={
  category="ll",
  direction="l",
 },
 [0x1044B]={
  category="ll",
  direction="l",
 },
 [0x1044C]={
  category="ll",
  direction="l",
 },
 [0x1044D]={
  category="ll",
  direction="l",
 },
 [0x1044E]={
  category="ll",
  direction="l",
 },
 [0x1044F]={
  category="ll",
  direction="l",
 },
 [0x10450]={
  category="lo",
  direction="l",
 },
 [0x10451]={
  category="lo",
  direction="l",
 },
 [0x10452]={
  category="lo",
  direction="l",
 },
 [0x10453]={
  category="lo",
  direction="l",
 },
 [0x10454]={
  category="lo",
  direction="l",
 },
 [0x10455]={
  category="lo",
  direction="l",
 },
 [0x10456]={
  category="lo",
  direction="l",
 },
 [0x10457]={
  category="lo",
  direction="l",
 },
 [0x10458]={
  category="lo",
  direction="l",
 },
 [0x10459]={
  category="lo",
  direction="l",
 },
 [0x1045A]={
  category="lo",
  direction="l",
 },
 [0x1045B]={
  category="lo",
  direction="l",
 },
 [0x1045C]={
  category="lo",
  direction="l",
 },
 [0x1045D]={
  category="lo",
  direction="l",
 },
 [0x1045E]={
  category="lo",
  direction="l",
 },
 [0x1045F]={
  category="lo",
  direction="l",
 },
 [0x10460]={
  category="lo",
  direction="l",
 },
 [0x10461]={
  category="lo",
  direction="l",
 },
 [0x10462]={
  category="lo",
  direction="l",
 },
 [0x10463]={
  category="lo",
  direction="l",
 },
 [0x10464]={
  category="lo",
  direction="l",
 },
 [0x10465]={
  category="lo",
  direction="l",
 },
 [0x10466]={
  category="lo",
  direction="l",
 },
 [0x10467]={
  category="lo",
  direction="l",
 },
 [0x10468]={
  category="lo",
  direction="l",
 },
 [0x10469]={
  category="lo",
  direction="l",
 },
 [0x1046A]={
  category="lo",
  direction="l",
 },
 [0x1046B]={
  category="lo",
  direction="l",
 },
 [0x1046C]={
  category="lo",
  direction="l",
 },
 [0x1046D]={
  category="lo",
  direction="l",
 },
 [0x1046E]={
  category="lo",
  direction="l",
 },
 [0x1046F]={
  category="lo",
  direction="l",
 },
 [0x10470]={
  category="lo",
  direction="l",
 },
 [0x10471]={
  category="lo",
  direction="l",
 },
 [0x10472]={
  category="lo",
  direction="l",
 },
 [0x10473]={
  category="lo",
  direction="l",
 },
 [0x10474]={
  category="lo",
  direction="l",
 },
 [0x10475]={
  category="lo",
  direction="l",
 },
 [0x10476]={
  category="lo",
  direction="l",
 },
 [0x10477]={
  category="lo",
  direction="l",
 },
 [0x10478]={
  category="lo",
  direction="l",
 },
 [0x10479]={
  category="lo",
  direction="l",
 },
 [0x1047A]={
  category="lo",
  direction="l",
 },
 [0x1047B]={
  category="lo",
  direction="l",
 },
 [0x1047C]={
  category="lo",
  direction="l",
 },
 [0x1047D]={
  category="lo",
  direction="l",
 },
 [0x1047E]={
  category="lo",
  direction="l",
 },
 [0x1047F]={
  category="lo",
  direction="l",
 },
 [0x10480]={
  category="lo",
  direction="l",
 },
 [0x10481]={
  category="lo",
  direction="l",
 },
 [0x10482]={
  category="lo",
  direction="l",
 },
 [0x10483]={
  category="lo",
  direction="l",
 },
 [0x10484]={
  category="lo",
  direction="l",
 },
 [0x10485]={
  category="lo",
  direction="l",
 },
 [0x10486]={
  category="lo",
  direction="l",
 },
 [0x10487]={
  category="lo",
  direction="l",
 },
 [0x10488]={
  category="lo",
  direction="l",
 },
 [0x10489]={
  category="lo",
  direction="l",
 },
 [0x1048A]={
  category="lo",
  direction="l",
 },
 [0x1048B]={
  category="lo",
  direction="l",
 },
 [0x1048C]={
  category="lo",
  direction="l",
 },
 [0x1048D]={
  category="lo",
  direction="l",
 },
 [0x1048E]={
  category="lo",
  direction="l",
 },
 [0x1048F]={
  category="lo",
  direction="l",
 },
 [0x10490]={
  category="lo",
  direction="l",
 },
 [0x10491]={
  category="lo",
  direction="l",
 },
 [0x10492]={
  category="lo",
  direction="l",
 },
 [0x10493]={
  category="lo",
  direction="l",
 },
 [0x10494]={
  category="lo",
  direction="l",
 },
 [0x10495]={
  category="lo",
  direction="l",
 },
 [0x10496]={
  category="lo",
  direction="l",
 },
 [0x10497]={
  category="lo",
  direction="l",
 },
 [0x10498]={
  category="lo",
  direction="l",
 },
 [0x10499]={
  category="lo",
  direction="l",
 },
 [0x1049A]={
  category="lo",
  direction="l",
 },
 [0x1049B]={
  category="lo",
  direction="l",
 },
 [0x1049C]={
  category="lo",
  direction="l",
 },
 [0x1049D]={
  category="lo",
  direction="l",
 },
 [0x104A0]={
  category="nd",
  direction="l",
 },
 [0x104A1]={
  category="nd",
  direction="l",
 },
 [0x104A2]={
  category="nd",
  direction="l",
 },
 [0x104A3]={
  category="nd",
  direction="l",
 },
 [0x104A4]={
  category="nd",
  direction="l",
 },
 [0x104A5]={
  category="nd",
  direction="l",
 },
 [0x104A6]={
  category="nd",
  direction="l",
 },
 [0x104A7]={
  category="nd",
  direction="l",
 },
 [0x104A8]={
  category="nd",
  direction="l",
 },
 [0x104A9]={
  category="nd",
  direction="l",
 },
 [0x10500]={
  category="lo",
  direction="l",
 },
 [0x10501]={
  category="lo",
  direction="l",
 },
 [0x10502]={
  category="lo",
  direction="l",
 },
 [0x10503]={
  category="lo",
  direction="l",
 },
 [0x10504]={
  category="lo",
  direction="l",
 },
 [0x10505]={
  category="lo",
  direction="l",
 },
 [0x10506]={
  category="lo",
  direction="l",
 },
 [0x10507]={
  category="lo",
  direction="l",
 },
 [0x10508]={
  category="lo",
  direction="l",
 },
 [0x10509]={
  category="lo",
  direction="l",
 },
 [0x1050A]={
  category="lo",
  direction="l",
 },
 [0x1050B]={
  category="lo",
  direction="l",
 },
 [0x1050C]={
  category="lo",
  direction="l",
 },
 [0x1050D]={
  category="lo",
  direction="l",
 },
 [0x1050E]={
  category="lo",
  direction="l",
 },
 [0x1050F]={
  category="lo",
  direction="l",
 },
 [0x10510]={
  category="lo",
  direction="l",
 },
 [0x10511]={
  category="lo",
  direction="l",
 },
 [0x10512]={
  category="lo",
  direction="l",
 },
 [0x10513]={
  category="lo",
  direction="l",
 },
 [0x10514]={
  category="lo",
  direction="l",
 },
 [0x10515]={
  category="lo",
  direction="l",
 },
 [0x10516]={
  category="lo",
  direction="l",
 },
 [0x10517]={
  category="lo",
  direction="l",
 },
 [0x10518]={
  category="lo",
  direction="l",
 },
 [0x10519]={
  category="lo",
  direction="l",
 },
 [0x1051A]={
  category="lo",
  direction="l",
 },
 [0x1051B]={
  category="lo",
  direction="l",
 },
 [0x1051C]={
  category="lo",
  direction="l",
 },
 [0x1051D]={
  category="lo",
  direction="l",
 },
 [0x1051E]={
  category="lo",
  direction="l",
 },
 [0x1051F]={
  category="lo",
  direction="l",
 },
 [0x10520]={
  category="lo",
  direction="l",
 },
 [0x10521]={
  category="lo",
  direction="l",
 },
 [0x10522]={
  category="lo",
  direction="l",
 },
 [0x10523]={
  category="lo",
  direction="l",
 },
 [0x10524]={
  category="lo",
  direction="l",
 },
 [0x10525]={
  category="lo",
  direction="l",
 },
 [0x10526]={
  category="lo",
  direction="l",
 },
 [0x10527]={
  category="lo",
  direction="l",
 },
 [0x10530]={
  category="lo",
  direction="l",
 },
 [0x10531]={
  category="lo",
  direction="l",
 },
 [0x10532]={
  category="lo",
  direction="l",
 },
 [0x10533]={
  category="lo",
  direction="l",
 },
 [0x10534]={
  category="lo",
  direction="l",
 },
 [0x10535]={
  category="lo",
  direction="l",
 },
 [0x10536]={
  category="lo",
  direction="l",
 },
 [0x10537]={
  category="lo",
  direction="l",
 },
 [0x10538]={
  category="lo",
  direction="l",
 },
 [0x10539]={
  category="lo",
  direction="l",
 },
 [0x1053A]={
  category="lo",
  direction="l",
 },
 [0x1053B]={
  category="lo",
  direction="l",
 },
 [0x1053C]={
  category="lo",
  direction="l",
 },
 [0x1053D]={
  category="lo",
  direction="l",
 },
 [0x1053E]={
  category="lo",
  direction="l",
 },
 [0x1053F]={
  category="lo",
  direction="l",
 },
 [0x10540]={
  category="lo",
  direction="l",
 },
 [0x10541]={
  category="lo",
  direction="l",
 },
 [0x10542]={
  category="lo",
  direction="l",
 },
 [0x10543]={
  category="lo",
  direction="l",
 },
 [0x10544]={
  category="lo",
  direction="l",
 },
 [0x10545]={
  category="lo",
  direction="l",
 },
 [0x10546]={
  category="lo",
  direction="l",
 },
 [0x10547]={
  category="lo",
  direction="l",
 },
 [0x10548]={
  category="lo",
  direction="l",
 },
 [0x10549]={
  category="lo",
  direction="l",
 },
 [0x1054A]={
  category="lo",
  direction="l",
 },
 [0x1054B]={
  category="lo",
  direction="l",
 },
 [0x1054C]={
  category="lo",
  direction="l",
 },
 [0x1054D]={
  category="lo",
  direction="l",
 },
 [0x1054E]={
  category="lo",
  direction="l",
 },
 [0x1054F]={
  category="lo",
  direction="l",
 },
 [0x10550]={
  category="lo",
  direction="l",
 },
 [0x10551]={
  category="lo",
  direction="l",
 },
 [0x10552]={
  category="lo",
  direction="l",
 },
 [0x10553]={
  category="lo",
  direction="l",
 },
 [0x10554]={
  category="lo",
  direction="l",
 },
 [0x10555]={
  category="lo",
  direction="l",
 },
 [0x10556]={
  category="lo",
  direction="l",
 },
 [0x10557]={
  category="lo",
  direction="l",
 },
 [0x10558]={
  category="lo",
  direction="l",
 },
 [0x10559]={
  category="lo",
  direction="l",
 },
 [0x1055A]={
  category="lo",
  direction="l",
 },
 [0x1055B]={
  category="lo",
  direction="l",
 },
 [0x1055C]={
  category="lo",
  direction="l",
 },
 [0x1055D]={
  category="lo",
  direction="l",
 },
 [0x1055E]={
  category="lo",
  direction="l",
 },
 [0x1055F]={
  category="lo",
  direction="l",
 },
 [0x10560]={
  category="lo",
  direction="l",
 },
 [0x10561]={
  category="lo",
  direction="l",
 },
 [0x10562]={
  category="lo",
  direction="l",
 },
 [0x10563]={
  category="lo",
  direction="l",
 },
 [0x1056F]={
  category="po",
  direction="l",
 },
 [0x10600]={
  category="lo",
  direction="l",
 },
 [0x10601]={
  category="lo",
  direction="l",
 },
 [0x10602]={
  category="lo",
  direction="l",
 },
 [0x10603]={
  category="lo",
  direction="l",
 },
 [0x10604]={
  category="lo",
  direction="l",
 },
 [0x10605]={
  category="lo",
  direction="l",
 },
 [0x10606]={
  category="lo",
  direction="l",
 },
 [0x10607]={
  category="lo",
  direction="l",
 },
 [0x10608]={
  category="lo",
  direction="l",
 },
 [0x10609]={
  category="lo",
  direction="l",
 },
 [0x1060A]={
  category="lo",
  direction="l",
 },
 [0x1060B]={
  category="lo",
  direction="l",
 },
 [0x1060C]={
  category="lo",
  direction="l",
 },
 [0x1060D]={
  category="lo",
  direction="l",
 },
 [0x1060E]={
  category="lo",
  direction="l",
 },
 [0x1060F]={
  category="lo",
  direction="l",
 },
 [0x10610]={
  category="lo",
  direction="l",
 },
 [0x10611]={
  category="lo",
  direction="l",
 },
 [0x10612]={
  category="lo",
  direction="l",
 },
 [0x10613]={
  category="lo",
  direction="l",
 },
 [0x10614]={
  category="lo",
  direction="l",
 },
 [0x10615]={
  category="lo",
  direction="l",
 },
 [0x10616]={
  category="lo",
  direction="l",
 },
 [0x10617]={
  category="lo",
  direction="l",
 },
 [0x10618]={
  category="lo",
  direction="l",
 },
 [0x10619]={
  category="lo",
  direction="l",
 },
 [0x1061A]={
  category="lo",
  direction="l",
 },
 [0x1061B]={
  category="lo",
  direction="l",
 },
 [0x1061C]={
  category="lo",
  direction="l",
 },
 [0x1061D]={
  category="lo",
  direction="l",
 },
 [0x1061E]={
  category="lo",
  direction="l",
 },
 [0x1061F]={
  category="lo",
  direction="l",
 },
 [0x10620]={
  category="lo",
  direction="l",
 },
 [0x10621]={
  category="lo",
  direction="l",
 },
 [0x10622]={
  category="lo",
  direction="l",
 },
 [0x10623]={
  category="lo",
  direction="l",
 },
 [0x10624]={
  category="lo",
  direction="l",
 },
 [0x10625]={
  category="lo",
  direction="l",
 },
 [0x10626]={
  category="lo",
  direction="l",
 },
 [0x10627]={
  category="lo",
  direction="l",
 },
 [0x10628]={
  category="lo",
  direction="l",
 },
 [0x10629]={
  category="lo",
  direction="l",
 },
 [0x1062A]={
  category="lo",
  direction="l",
 },
 [0x1062B]={
  category="lo",
  direction="l",
 },
 [0x1062C]={
  category="lo",
  direction="l",
 },
 [0x1062D]={
  category="lo",
  direction="l",
 },
 [0x1062E]={
  category="lo",
  direction="l",
 },
 [0x1062F]={
  category="lo",
  direction="l",
 },
 [0x10630]={
  category="lo",
  direction="l",
 },
 [0x10631]={
  category="lo",
  direction="l",
 },
 [0x10632]={
  category="lo",
  direction="l",
 },
 [0x10633]={
  category="lo",
  direction="l",
 },
 [0x10634]={
  category="lo",
  direction="l",
 },
 [0x10635]={
  category="lo",
  direction="l",
 },
 [0x10636]={
  category="lo",
  direction="l",
 },
 [0x10637]={
  category="lo",
  direction="l",
 },
 [0x10638]={
  category="lo",
  direction="l",
 },
 [0x10639]={
  category="lo",
  direction="l",
 },
 [0x1063A]={
  category="lo",
  direction="l",
 },
 [0x1063B]={
  category="lo",
  direction="l",
 },
 [0x1063C]={
  category="lo",
  direction="l",
 },
 [0x1063D]={
  category="lo",
  direction="l",
 },
 [0x1063E]={
  category="lo",
  direction="l",
 },
 [0x1063F]={
  category="lo",
  direction="l",
 },
 [0x10640]={
  category="lo",
  direction="l",
 },
 [0x10641]={
  category="lo",
  direction="l",
 },
 [0x10642]={
  category="lo",
  direction="l",
 },
 [0x10643]={
  category="lo",
  direction="l",
 },
 [0x10644]={
  category="lo",
  direction="l",
 },
 [0x10645]={
  category="lo",
  direction="l",
 },
 [0x10646]={
  category="lo",
  direction="l",
 },
 [0x10647]={
  category="lo",
  direction="l",
 },
 [0x10648]={
  category="lo",
  direction="l",
 },
 [0x10649]={
  category="lo",
  direction="l",
 },
 [0x1064A]={
  category="lo",
  direction="l",
 },
 [0x1064B]={
  category="lo",
  direction="l",
 },
 [0x1064C]={
  category="lo",
  direction="l",
 },
 [0x1064D]={
  category="lo",
  direction="l",
 },
 [0x1064E]={
  category="lo",
  direction="l",
 },
 [0x1064F]={
  category="lo",
  direction="l",
 },
 [0x10650]={
  category="lo",
  direction="l",
 },
 [0x10651]={
  category="lo",
  direction="l",
 },
 [0x10652]={
  category="lo",
  direction="l",
 },
 [0x10653]={
  category="lo",
  direction="l",
 },
 [0x10654]={
  category="lo",
  direction="l",
 },
 [0x10655]={
  category="lo",
  direction="l",
 },
 [0x10656]={
  category="lo",
  direction="l",
 },
 [0x10657]={
  category="lo",
  direction="l",
 },
 [0x10658]={
  category="lo",
  direction="l",
 },
 [0x10659]={
  category="lo",
  direction="l",
 },
 [0x1065A]={
  category="lo",
  direction="l",
 },
 [0x1065B]={
  category="lo",
  direction="l",
 },
 [0x1065C]={
  category="lo",
  direction="l",
 },
 [0x1065D]={
  category="lo",
  direction="l",
 },
 [0x1065E]={
  category="lo",
  direction="l",
 },
 [0x1065F]={
  category="lo",
  direction="l",
 },
 [0x10660]={
  category="lo",
  direction="l",
 },
 [0x10661]={
  category="lo",
  direction="l",
 },
 [0x10662]={
  category="lo",
  direction="l",
 },
 [0x10663]={
  category="lo",
  direction="l",
 },
 [0x10664]={
  category="lo",
  direction="l",
 },
 [0x10665]={
  category="lo",
  direction="l",
 },
 [0x10666]={
  category="lo",
  direction="l",
 },
 [0x10667]={
  category="lo",
  direction="l",
 },
 [0x10668]={
  category="lo",
  direction="l",
 },
 [0x10669]={
  category="lo",
  direction="l",
 },
 [0x1066A]={
  category="lo",
  direction="l",
 },
 [0x1066B]={
  category="lo",
  direction="l",
 },
 [0x1066C]={
  category="lo",
  direction="l",
 },
 [0x1066D]={
  category="lo",
  direction="l",
 },
 [0x1066E]={
  category="lo",
  direction="l",
 },
 [0x1066F]={
  category="lo",
  direction="l",
 },
 [0x10670]={
  category="lo",
  direction="l",
 },
 [0x10671]={
  category="lo",
  direction="l",
 },
 [0x10672]={
  category="lo",
  direction="l",
 },
 [0x10673]={
  category="lo",
  direction="l",
 },
 [0x10674]={
  category="lo",
  direction="l",
 },
 [0x10675]={
  category="lo",
  direction="l",
 },
 [0x10676]={
  category="lo",
  direction="l",
 },
 [0x10677]={
  category="lo",
  direction="l",
 },
 [0x10678]={
  category="lo",
  direction="l",
 },
 [0x10679]={
  category="lo",
  direction="l",
 },
 [0x1067A]={
  category="lo",
  direction="l",
 },
 [0x1067B]={
  category="lo",
  direction="l",
 },
 [0x1067C]={
  category="lo",
  direction="l",
 },
 [0x1067D]={
  category="lo",
  direction="l",
 },
 [0x1067E]={
  category="lo",
  direction="l",
 },
 [0x1067F]={
  category="lo",
  direction="l",
 },
 [0x10680]={
  category="lo",
  direction="l",
 },
 [0x10681]={
  category="lo",
  direction="l",
 },
 [0x10682]={
  category="lo",
  direction="l",
 },
 [0x10683]={
  category="lo",
  direction="l",
 },
 [0x10684]={
  category="lo",
  direction="l",
 },
 [0x10685]={
  category="lo",
  direction="l",
 },
 [0x10686]={
  category="lo",
  direction="l",
 },
 [0x10687]={
  category="lo",
  direction="l",
 },
 [0x10688]={
  category="lo",
  direction="l",
 },
 [0x10689]={
  category="lo",
  direction="l",
 },
 [0x1068A]={
  category="lo",
  direction="l",
 },
 [0x1068B]={
  category="lo",
  direction="l",
 },
 [0x1068C]={
  category="lo",
  direction="l",
 },
 [0x1068D]={
  category="lo",
  direction="l",
 },
 [0x1068E]={
  category="lo",
  direction="l",
 },
 [0x1068F]={
  category="lo",
  direction="l",
 },
 [0x10690]={
  category="lo",
  direction="l",
 },
 [0x10691]={
  category="lo",
  direction="l",
 },
 [0x10692]={
  category="lo",
  direction="l",
 },
 [0x10693]={
  category="lo",
  direction="l",
 },
 [0x10694]={
  category="lo",
  direction="l",
 },
 [0x10695]={
  category="lo",
  direction="l",
 },
 [0x10696]={
  category="lo",
  direction="l",
 },
 [0x10697]={
  category="lo",
  direction="l",
 },
 [0x10698]={
  category="lo",
  direction="l",
 },
 [0x10699]={
  category="lo",
  direction="l",
 },
 [0x1069A]={
  category="lo",
  direction="l",
 },
 [0x1069B]={
  category="lo",
  direction="l",
 },
 [0x1069C]={
  category="lo",
  direction="l",
 },
 [0x1069D]={
  category="lo",
  direction="l",
 },
 [0x1069E]={
  category="lo",
  direction="l",
 },
 [0x1069F]={
  category="lo",
  direction="l",
 },
 [0x106A0]={
  category="lo",
  direction="l",
 },
 [0x106A1]={
  category="lo",
  direction="l",
 },
 [0x106A2]={
  category="lo",
  direction="l",
 },
 [0x106A3]={
  category="lo",
  direction="l",
 },
 [0x106A4]={
  category="lo",
  direction="l",
 },
 [0x106A5]={
  category="lo",
  direction="l",
 },
 [0x106A6]={
  category="lo",
  direction="l",
 },
 [0x106A7]={
  category="lo",
  direction="l",
 },
 [0x106A8]={
  category="lo",
  direction="l",
 },
 [0x106A9]={
  category="lo",
  direction="l",
 },
 [0x106AA]={
  category="lo",
  direction="l",
 },
 [0x106AB]={
  category="lo",
  direction="l",
 },
 [0x106AC]={
  category="lo",
  direction="l",
 },
 [0x106AD]={
  category="lo",
  direction="l",
 },
 [0x106AE]={
  category="lo",
  direction="l",
 },
 [0x106AF]={
  category="lo",
  direction="l",
 },
 [0x106B0]={
  category="lo",
  direction="l",
 },
 [0x106B1]={
  category="lo",
  direction="l",
 },
 [0x106B2]={
  category="lo",
  direction="l",
 },
 [0x106B3]={
  category="lo",
  direction="l",
 },
 [0x106B4]={
  category="lo",
  direction="l",
 },
 [0x106B5]={
  category="lo",
  direction="l",
 },
 [0x106B6]={
  category="lo",
  direction="l",
 },
 [0x106B7]={
  category="lo",
  direction="l",
 },
 [0x106B8]={
  category="lo",
  direction="l",
 },
 [0x106B9]={
  category="lo",
  direction="l",
 },
 [0x106BA]={
  category="lo",
  direction="l",
 },
 [0x106BB]={
  category="lo",
  direction="l",
 },
 [0x106BC]={
  category="lo",
  direction="l",
 },
 [0x106BD]={
  category="lo",
  direction="l",
 },
 [0x106BE]={
  category="lo",
  direction="l",
 },
 [0x106BF]={
  category="lo",
  direction="l",
 },
 [0x106C0]={
  category="lo",
  direction="l",
 },
 [0x106C1]={
  category="lo",
  direction="l",
 },
 [0x106C2]={
  category="lo",
  direction="l",
 },
 [0x106C3]={
  category="lo",
  direction="l",
 },
 [0x106C4]={
  category="lo",
  direction="l",
 },
 [0x106C5]={
  category="lo",
  direction="l",
 },
 [0x106C6]={
  category="lo",
  direction="l",
 },
 [0x106C7]={
  category="lo",
  direction="l",
 },
 [0x106C8]={
  category="lo",
  direction="l",
 },
 [0x106C9]={
  category="lo",
  direction="l",
 },
 [0x106CA]={
  category="lo",
  direction="l",
 },
 [0x106CB]={
  category="lo",
  direction="l",
 },
 [0x106CC]={
  category="lo",
  direction="l",
 },
 [0x106CD]={
  category="lo",
  direction="l",
 },
 [0x106CE]={
  category="lo",
  direction="l",
 },
 [0x106CF]={
  category="lo",
  direction="l",
 },
 [0x106D0]={
  category="lo",
  direction="l",
 },
 [0x106D1]={
  category="lo",
  direction="l",
 },
 [0x106D2]={
  category="lo",
  direction="l",
 },
 [0x106D3]={
  category="lo",
  direction="l",
 },
 [0x106D4]={
  category="lo",
  direction="l",
 },
 [0x106D5]={
  category="lo",
  direction="l",
 },
 [0x106D6]={
  category="lo",
  direction="l",
 },
 [0x106D7]={
  category="lo",
  direction="l",
 },
 [0x106D8]={
  category="lo",
  direction="l",
 },
 [0x106D9]={
  category="lo",
  direction="l",
 },
 [0x106DA]={
  category="lo",
  direction="l",
 },
 [0x106DB]={
  category="lo",
  direction="l",
 },
 [0x106DC]={
  category="lo",
  direction="l",
 },
 [0x106DD]={
  category="lo",
  direction="l",
 },
 [0x106DE]={
  category="lo",
  direction="l",
 },
 [0x106DF]={
  category="lo",
  direction="l",
 },
 [0x106E0]={
  category="lo",
  direction="l",
 },
 [0x106E1]={
  category="lo",
  direction="l",
 },
 [0x106E2]={
  category="lo",
  direction="l",
 },
 [0x106E3]={
  category="lo",
  direction="l",
 },
 [0x106E4]={
  category="lo",
  direction="l",
 },
 [0x106E5]={
  category="lo",
  direction="l",
 },
 [0x106E6]={
  category="lo",
  direction="l",
 },
 [0x106E7]={
  category="lo",
  direction="l",
 },
 [0x106E8]={
  category="lo",
  direction="l",
 },
 [0x106E9]={
  category="lo",
  direction="l",
 },
 [0x106EA]={
  category="lo",
  direction="l",
 },
 [0x106EB]={
  category="lo",
  direction="l",
 },
 [0x106EC]={
  category="lo",
  direction="l",
 },
 [0x106ED]={
  category="lo",
  direction="l",
 },
 [0x106EE]={
  category="lo",
  direction="l",
 },
 [0x106EF]={
  category="lo",
  direction="l",
 },
 [0x106F0]={
  category="lo",
  direction="l",
 },
 [0x106F1]={
  category="lo",
  direction="l",
 },
 [0x106F2]={
  category="lo",
  direction="l",
 },
 [0x106F3]={
  category="lo",
  direction="l",
 },
 [0x106F4]={
  category="lo",
  direction="l",
 },
 [0x106F5]={
  category="lo",
  direction="l",
 },
 [0x106F6]={
  category="lo",
  direction="l",
 },
 [0x106F7]={
  category="lo",
  direction="l",
 },
 [0x106F8]={
  category="lo",
  direction="l",
 },
 [0x106F9]={
  category="lo",
  direction="l",
 },
 [0x106FA]={
  category="lo",
  direction="l",
 },
 [0x106FB]={
  category="lo",
  direction="l",
 },
 [0x106FC]={
  category="lo",
  direction="l",
 },
 [0x106FD]={
  category="lo",
  direction="l",
 },
 [0x106FE]={
  category="lo",
  direction="l",
 },
 [0x106FF]={
  category="lo",
  direction="l",
 },
 [0x10700]={
  category="lo",
  direction="l",
 },
 [0x10701]={
  category="lo",
  direction="l",
 },
 [0x10702]={
  category="lo",
  direction="l",
 },
 [0x10703]={
  category="lo",
  direction="l",
 },
 [0x10704]={
  category="lo",
  direction="l",
 },
 [0x10705]={
  category="lo",
  direction="l",
 },
 [0x10706]={
  category="lo",
  direction="l",
 },
 [0x10707]={
  category="lo",
  direction="l",
 },
 [0x10708]={
  category="lo",
  direction="l",
 },
 [0x10709]={
  category="lo",
  direction="l",
 },
 [0x1070A]={
  category="lo",
  direction="l",
 },
 [0x1070B]={
  category="lo",
  direction="l",
 },
 [0x1070C]={
  category="lo",
  direction="l",
 },
 [0x1070D]={
  category="lo",
  direction="l",
 },
 [0x1070E]={
  category="lo",
  direction="l",
 },
 [0x1070F]={
  category="lo",
  direction="l",
 },
 [0x10710]={
  category="lo",
  direction="l",
 },
 [0x10711]={
  category="lo",
  direction="l",
 },
 [0x10712]={
  category="lo",
  direction="l",
 },
 [0x10713]={
  category="lo",
  direction="l",
 },
 [0x10714]={
  category="lo",
  direction="l",
 },
 [0x10715]={
  category="lo",
  direction="l",
 },
 [0x10716]={
  category="lo",
  direction="l",
 },
 [0x10717]={
  category="lo",
  direction="l",
 },
 [0x10718]={
  category="lo",
  direction="l",
 },
 [0x10719]={
  category="lo",
  direction="l",
 },
 [0x1071A]={
  category="lo",
  direction="l",
 },
 [0x1071B]={
  category="lo",
  direction="l",
 },
 [0x1071C]={
  category="lo",
  direction="l",
 },
 [0x1071D]={
  category="lo",
  direction="l",
 },
 [0x1071E]={
  category="lo",
  direction="l",
 },
 [0x1071F]={
  category="lo",
  direction="l",
 },
 [0x10720]={
  category="lo",
  direction="l",
 },
 [0x10721]={
  category="lo",
  direction="l",
 },
 [0x10722]={
  category="lo",
  direction="l",
 },
 [0x10723]={
  category="lo",
  direction="l",
 },
 [0x10724]={
  category="lo",
  direction="l",
 },
 [0x10725]={
  category="lo",
  direction="l",
 },
 [0x10726]={
  category="lo",
  direction="l",
 },
 [0x10727]={
  category="lo",
  direction="l",
 },
 [0x10728]={
  category="lo",
  direction="l",
 },
 [0x10729]={
  category="lo",
  direction="l",
 },
 [0x1072A]={
  category="lo",
  direction="l",
 },
 [0x1072B]={
  category="lo",
  direction="l",
 },
 [0x1072C]={
  category="lo",
  direction="l",
 },
 [0x1072D]={
  category="lo",
  direction="l",
 },
 [0x1072E]={
  category="lo",
  direction="l",
 },
 [0x1072F]={
  category="lo",
  direction="l",
 },
 [0x10730]={
  category="lo",
  direction="l",
 },
 [0x10731]={
  category="lo",
  direction="l",
 },
 [0x10732]={
  category="lo",
  direction="l",
 },
 [0x10733]={
  category="lo",
  direction="l",
 },
 [0x10734]={
  category="lo",
  direction="l",
 },
 [0x10735]={
  category="lo",
  direction="l",
 },
 [0x10736]={
  category="lo",
  direction="l",
 },
 [0x10740]={
  category="lo",
  direction="l",
 },
 [0x10741]={
  category="lo",
  direction="l",
 },
 [0x10742]={
  category="lo",
  direction="l",
 },
 [0x10743]={
  category="lo",
  direction="l",
 },
 [0x10744]={
  category="lo",
  direction="l",
 },
 [0x10745]={
  category="lo",
  direction="l",
 },
 [0x10746]={
  category="lo",
  direction="l",
 },
 [0x10747]={
  category="lo",
  direction="l",
 },
 [0x10748]={
  category="lo",
  direction="l",
 },
 [0x10749]={
  category="lo",
  direction="l",
 },
 [0x1074A]={
  category="lo",
  direction="l",
 },
 [0x1074B]={
  category="lo",
  direction="l",
 },
 [0x1074C]={
  category="lo",
  direction="l",
 },
 [0x1074D]={
  category="lo",
  direction="l",
 },
 [0x1074E]={
  category="lo",
  direction="l",
 },
 [0x1074F]={
  category="lo",
  direction="l",
 },
 [0x10750]={
  category="lo",
  direction="l",
 },
 [0x10751]={
  category="lo",
  direction="l",
 },
 [0x10752]={
  category="lo",
  direction="l",
 },
 [0x10753]={
  category="lo",
  direction="l",
 },
 [0x10754]={
  category="lo",
  direction="l",
 },
 [0x10755]={
  category="lo",
  direction="l",
 },
 [0x10760]={
  category="lo",
  direction="l",
 },
 [0x10761]={
  category="lo",
  direction="l",
 },
 [0x10762]={
  category="lo",
  direction="l",
 },
 [0x10763]={
  category="lo",
  direction="l",
 },
 [0x10764]={
  category="lo",
  direction="l",
 },
 [0x10765]={
  category="lo",
  direction="l",
 },
 [0x10766]={
  category="lo",
  direction="l",
 },
 [0x10767]={
  category="lo",
  direction="l",
 },
 [0x10800]={
  category="lo",
  direction="r",
 },
 [0x10801]={
  category="lo",
  direction="r",
 },
 [0x10802]={
  category="lo",
  direction="r",
 },
 [0x10803]={
  category="lo",
  direction="r",
 },
 [0x10804]={
  category="lo",
  direction="r",
 },
 [0x10805]={
  category="lo",
  direction="r",
 },
 [0x10808]={
  category="lo",
  direction="r",
 },
 [0x1080A]={
  category="lo",
  direction="r",
 },
 [0x1080B]={
  category="lo",
  direction="r",
 },
 [0x1080C]={
  category="lo",
  direction="r",
 },
 [0x1080D]={
  category="lo",
  direction="r",
 },
 [0x1080E]={
  category="lo",
  direction="r",
 },
 [0x1080F]={
  category="lo",
  direction="r",
 },
 [0x10810]={
  category="lo",
  direction="r",
 },
 [0x10811]={
  category="lo",
  direction="r",
 },
 [0x10812]={
  category="lo",
  direction="r",
 },
 [0x10813]={
  category="lo",
  direction="r",
 },
 [0x10814]={
  category="lo",
  direction="r",
 },
 [0x10815]={
  category="lo",
  direction="r",
 },
 [0x10816]={
  category="lo",
  direction="r",
 },
 [0x10817]={
  category="lo",
  direction="r",
 },
 [0x10818]={
  category="lo",
  direction="r",
 },
 [0x10819]={
  category="lo",
  direction="r",
 },
 [0x1081A]={
  category="lo",
  direction="r",
 },
 [0x1081B]={
  category="lo",
  direction="r",
 },
 [0x1081C]={
  category="lo",
  direction="r",
 },
 [0x1081D]={
  category="lo",
  direction="r",
 },
 [0x1081E]={
  category="lo",
  direction="r",
 },
 [0x1081F]={
  category="lo",
  direction="r",
 },
 [0x10820]={
  category="lo",
  direction="r",
 },
 [0x10821]={
  category="lo",
  direction="r",
 },
 [0x10822]={
  category="lo",
  direction="r",
 },
 [0x10823]={
  category="lo",
  direction="r",
 },
 [0x10824]={
  category="lo",
  direction="r",
 },
 [0x10825]={
  category="lo",
  direction="r",
 },
 [0x10826]={
  category="lo",
  direction="r",
 },
 [0x10827]={
  category="lo",
  direction="r",
 },
 [0x10828]={
  category="lo",
  direction="r",
 },
 [0x10829]={
  category="lo",
  direction="r",
 },
 [0x1082A]={
  category="lo",
  direction="r",
 },
 [0x1082B]={
  category="lo",
  direction="r",
 },
 [0x1082C]={
  category="lo",
  direction="r",
 },
 [0x1082D]={
  category="lo",
  direction="r",
 },
 [0x1082E]={
  category="lo",
  direction="r",
 },
 [0x1082F]={
  category="lo",
  direction="r",
 },
 [0x10830]={
  category="lo",
  direction="r",
 },
 [0x10831]={
  category="lo",
  direction="r",
 },
 [0x10832]={
  category="lo",
  direction="r",
 },
 [0x10833]={
  category="lo",
  direction="r",
 },
 [0x10834]={
  category="lo",
  direction="r",
 },
 [0x10835]={
  category="lo",
  direction="r",
 },
 [0x10837]={
  category="lo",
  direction="r",
 },
 [0x10838]={
  category="lo",
  direction="r",
 },
 [0x1083C]={
  category="lo",
  direction="r",
 },
 [0x1083F]={
  category="lo",
  direction="r",
 },
 [0x10840]={
  category="lo",
  direction="r",
 },
 [0x10841]={
  category="lo",
  direction="r",
 },
 [0x10842]={
  category="lo",
  direction="r",
 },
 [0x10843]={
  category="lo",
  direction="r",
 },
 [0x10844]={
  category="lo",
  direction="r",
 },
 [0x10845]={
  category="lo",
  direction="r",
 },
 [0x10846]={
  category="lo",
  direction="r",
 },
 [0x10847]={
  category="lo",
  direction="r",
 },
 [0x10848]={
  category="lo",
  direction="r",
 },
 [0x10849]={
  category="lo",
  direction="r",
 },
 [0x1084A]={
  category="lo",
  direction="r",
 },
 [0x1084B]={
  category="lo",
  direction="r",
 },
 [0x1084C]={
  category="lo",
  direction="r",
 },
 [0x1084D]={
  category="lo",
  direction="r",
 },
 [0x1084E]={
  category="lo",
  direction="r",
 },
 [0x1084F]={
  category="lo",
  direction="r",
 },
 [0x10850]={
  category="lo",
  direction="r",
 },
 [0x10851]={
  category="lo",
  direction="r",
 },
 [0x10852]={
  category="lo",
  direction="r",
 },
 [0x10853]={
  category="lo",
  direction="r",
 },
 [0x10854]={
  category="lo",
  direction="r",
 },
 [0x10855]={
  category="lo",
  direction="r",
 },
 [0x10857]={
  category="po",
  direction="r",
 },
 [0x10858]={
  category="no",
  direction="r",
 },
 [0x10859]={
  category="no",
  direction="r",
 },
 [0x1085A]={
  category="no",
  direction="r",
 },
 [0x1085B]={
  category="no",
  direction="r",
 },
 [0x1085C]={
  category="no",
  direction="r",
 },
 [0x1085D]={
  category="no",
  direction="r",
 },
 [0x1085E]={
  category="no",
  direction="r",
 },
 [0x1085F]={
  category="no",
  direction="r",
 },
 [0x10860]={
  category="lo",
  direction="r",
 },
 [0x10861]={
  category="lo",
  direction="r",
 },
 [0x10862]={
  category="lo",
  direction="r",
 },
 [0x10863]={
  category="lo",
  direction="r",
 },
 [0x10864]={
  category="lo",
  direction="r",
 },
 [0x10865]={
  category="lo",
  direction="r",
 },
 [0x10866]={
  category="lo",
  direction="r",
 },
 [0x10867]={
  category="lo",
  direction="r",
 },
 [0x10868]={
  category="lo",
  direction="r",
 },
 [0x10869]={
  category="lo",
  direction="r",
 },
 [0x1086A]={
  category="lo",
  direction="r",
 },
 [0x1086B]={
  category="lo",
  direction="r",
 },
 [0x1086C]={
  category="lo",
  direction="r",
 },
 [0x1086D]={
  category="lo",
  direction="r",
 },
 [0x1086E]={
  category="lo",
  direction="r",
 },
 [0x1086F]={
  category="lo",
  direction="r",
 },
 [0x10870]={
  category="lo",
  direction="r",
 },
 [0x10871]={
  category="lo",
  direction="r",
 },
 [0x10872]={
  category="lo",
  direction="r",
 },
 [0x10873]={
  category="lo",
  direction="r",
 },
 [0x10874]={
  category="lo",
  direction="r",
 },
 [0x10875]={
  category="lo",
  direction="r",
 },
 [0x10876]={
  category="lo",
  direction="r",
 },
 [0x10877]={
  category="so",
  direction="r",
 },
 [0x10878]={
  category="so",
  direction="r",
 },
 [0x10879]={
  category="no",
  direction="r",
 },
 [0x1087A]={
  category="no",
  direction="r",
 },
 [0x1087B]={
  category="no",
  direction="r",
 },
 [0x1087C]={
  category="no",
  direction="r",
 },
 [0x1087D]={
  category="no",
  direction="r",
 },
 [0x1087E]={
  category="no",
  direction="r",
 },
 [0x1087F]={
  category="no",
  direction="r",
 },
 [0x10880]={
  category="lo",
  direction="r",
 },
 [0x10881]={
  category="lo",
  direction="r",
 },
 [0x10882]={
  category="lo",
  direction="r",
 },
 [0x10883]={
  category="lo",
  direction="r",
 },
 [0x10884]={
  category="lo",
  direction="r",
 },
 [0x10885]={
  category="lo",
  direction="r",
 },
 [0x10886]={
  category="lo",
  direction="r",
 },
 [0x10887]={
  category="lo",
  direction="r",
 },
 [0x10888]={
  category="lo",
  direction="r",
 },
 [0x10889]={
  category="lo",
  direction="r",
 },
 [0x1088A]={
  category="lo",
  direction="r",
 },
 [0x1088B]={
  category="lo",
  direction="r",
 },
 [0x1088C]={
  category="lo",
  direction="r",
 },
 [0x1088D]={
  category="lo",
  direction="r",
 },
 [0x1088E]={
  category="lo",
  direction="r",
 },
 [0x1088F]={
  category="lo",
  direction="r",
 },
 [0x10890]={
  category="lo",
  direction="r",
 },
 [0x10891]={
  category="lo",
  direction="r",
 },
 [0x10892]={
  category="lo",
  direction="r",
 },
 [0x10893]={
  category="lo",
  direction="r",
 },
 [0x10894]={
  category="lo",
  direction="r",
 },
 [0x10895]={
  category="lo",
  direction="r",
 },
 [0x10896]={
  category="lo",
  direction="r",
 },
 [0x10897]={
  category="lo",
  direction="r",
 },
 [0x10898]={
  category="lo",
  direction="r",
 },
 [0x10899]={
  category="lo",
  direction="r",
 },
 [0x1089A]={
  category="lo",
  direction="r",
 },
 [0x1089B]={
  category="lo",
  direction="r",
 },
 [0x1089C]={
  category="lo",
  direction="r",
 },
 [0x1089D]={
  category="lo",
  direction="r",
 },
 [0x1089E]={
  category="lo",
  direction="r",
 },
 [0x108A7]={
  category="no",
  direction="r",
 },
 [0x108A8]={
  category="no",
  direction="r",
 },
 [0x108A9]={
  category="no",
  direction="r",
 },
 [0x108AA]={
  category="no",
  direction="r",
 },
 [0x108AB]={
  category="no",
  direction="r",
 },
 [0x108AC]={
  category="no",
  direction="r",
 },
 [0x108AD]={
  category="no",
  direction="r",
 },
 [0x108AE]={
  category="no",
  direction="r",
 },
 [0x108AF]={
  category="no",
  direction="r",
 },
 [0x10900]={
  category="lo",
  direction="r",
 },
 [0x10901]={
  category="lo",
  direction="r",
 },
 [0x10902]={
  category="lo",
  direction="r",
 },
 [0x10903]={
  category="lo",
  direction="r",
 },
 [0x10904]={
  category="lo",
  direction="r",
 },
 [0x10905]={
  category="lo",
  direction="r",
 },
 [0x10906]={
  category="lo",
  direction="r",
 },
 [0x10907]={
  category="lo",
  direction="r",
 },
 [0x10908]={
  category="lo",
  direction="r",
 },
 [0x10909]={
  category="lo",
  direction="r",
 },
 [0x1090A]={
  category="lo",
  direction="r",
 },
 [0x1090B]={
  category="lo",
  direction="r",
 },
 [0x1090C]={
  category="lo",
  direction="r",
 },
 [0x1090D]={
  category="lo",
  direction="r",
 },
 [0x1090E]={
  category="lo",
  direction="r",
 },
 [0x1090F]={
  category="lo",
  direction="r",
 },
 [0x10910]={
  category="lo",
  direction="r",
 },
 [0x10911]={
  category="lo",
  direction="r",
 },
 [0x10912]={
  category="lo",
  direction="r",
 },
 [0x10913]={
  category="lo",
  direction="r",
 },
 [0x10914]={
  category="lo",
  direction="r",
 },
 [0x10915]={
  category="lo",
  direction="r",
 },
 [0x10916]={
  category="no",
  direction="r",
 },
 [0x10917]={
  category="no",
  direction="r",
 },
 [0x10918]={
  category="no",
  direction="r",
 },
 [0x10919]={
  category="no",
  direction="r",
 },
 [0x1091A]={
  category="no",
  direction="r",
 },
 [0x1091B]={
  category="no",
  direction="r",
 },
 [0x1091F]={
  category="po",
  direction="on",
 },
 [0x10920]={
  category="lo",
  direction="r",
 },
 [0x10921]={
  category="lo",
  direction="r",
 },
 [0x10922]={
  category="lo",
  direction="r",
 },
 [0x10923]={
  category="lo",
  direction="r",
 },
 [0x10924]={
  category="lo",
  direction="r",
 },
 [0x10925]={
  category="lo",
  direction="r",
 },
 [0x10926]={
  category="lo",
  direction="r",
 },
 [0x10927]={
  category="lo",
  direction="r",
 },
 [0x10928]={
  category="lo",
  direction="r",
 },
 [0x10929]={
  category="lo",
  direction="r",
 },
 [0x1092A]={
  category="lo",
  direction="r",
 },
 [0x1092B]={
  category="lo",
  direction="r",
 },
 [0x1092C]={
  category="lo",
  direction="r",
 },
 [0x1092D]={
  category="lo",
  direction="r",
 },
 [0x1092E]={
  category="lo",
  direction="r",
 },
 [0x1092F]={
  category="lo",
  direction="r",
 },
 [0x10930]={
  category="lo",
  direction="r",
 },
 [0x10931]={
  category="lo",
  direction="r",
 },
 [0x10932]={
  category="lo",
  direction="r",
 },
 [0x10933]={
  category="lo",
  direction="r",
 },
 [0x10934]={
  category="lo",
  direction="r",
 },
 [0x10935]={
  category="lo",
  direction="r",
 },
 [0x10936]={
  category="lo",
  direction="r",
 },
 [0x10937]={
  category="lo",
  direction="r",
 },
 [0x10938]={
  category="lo",
  direction="r",
 },
 [0x10939]={
  category="lo",
  direction="r",
 },
 [0x1093F]={
  category="po",
  direction="r",
 },
 [0x10980]={
  category="lo",
  direction="r",
 },
 [0x10981]={
  category="lo",
  direction="r",
 },
 [0x10982]={
  category="lo",
  direction="r",
 },
 [0x10983]={
  category="lo",
  direction="r",
 },
 [0x10984]={
  category="lo",
  direction="r",
 },
 [0x10985]={
  category="lo",
  direction="r",
 },
 [0x10986]={
  category="lo",
  direction="r",
 },
 [0x10987]={
  category="lo",
  direction="r",
 },
 [0x10988]={
  category="lo",
  direction="r",
 },
 [0x10989]={
  category="lo",
  direction="r",
 },
 [0x1098A]={
  category="lo",
  direction="r",
 },
 [0x1098B]={
  category="lo",
  direction="r",
 },
 [0x1098C]={
  category="lo",
  direction="r",
 },
 [0x1098D]={
  category="lo",
  direction="r",
 },
 [0x1098E]={
  category="lo",
  direction="r",
 },
 [0x1098F]={
  category="lo",
  direction="r",
 },
 [0x10990]={
  category="lo",
  direction="r",
 },
 [0x10991]={
  category="lo",
  direction="r",
 },
 [0x10992]={
  category="lo",
  direction="r",
 },
 [0x10993]={
  category="lo",
  direction="r",
 },
 [0x10994]={
  category="lo",
  direction="r",
 },
 [0x10995]={
  category="lo",
  direction="r",
 },
 [0x10996]={
  category="lo",
  direction="r",
 },
 [0x10997]={
  category="lo",
  direction="r",
 },
 [0x10998]={
  category="lo",
  direction="r",
 },
 [0x10999]={
  category="lo",
  direction="r",
 },
 [0x1099A]={
  category="lo",
  direction="r",
 },
 [0x1099B]={
  category="lo",
  direction="r",
 },
 [0x1099C]={
  category="lo",
  direction="r",
 },
 [0x1099D]={
  category="lo",
  direction="r",
 },
 [0x1099E]={
  category="lo",
  direction="r",
 },
 [0x1099F]={
  category="lo",
  direction="r",
 },
 [0x109A0]={
  category="lo",
  direction="r",
 },
 [0x109A1]={
  category="lo",
  direction="r",
 },
 [0x109A2]={
  category="lo",
  direction="r",
 },
 [0x109A3]={
  category="lo",
  direction="r",
 },
 [0x109A4]={
  category="lo",
  direction="r",
 },
 [0x109A5]={
  category="lo",
  direction="r",
 },
 [0x109A6]={
  category="lo",
  direction="r",
 },
 [0x109A7]={
  category="lo",
  direction="r",
 },
 [0x109A8]={
  category="lo",
  direction="r",
 },
 [0x109A9]={
  category="lo",
  direction="r",
 },
 [0x109AA]={
  category="lo",
  direction="r",
 },
 [0x109AB]={
  category="lo",
  direction="r",
 },
 [0x109AC]={
  category="lo",
  direction="r",
 },
 [0x109AD]={
  category="lo",
  direction="r",
 },
 [0x109AE]={
  category="lo",
  direction="r",
 },
 [0x109AF]={
  category="lo",
  direction="r",
 },
 [0x109B0]={
  category="lo",
  direction="r",
 },
 [0x109B1]={
  category="lo",
  direction="r",
 },
 [0x109B2]={
  category="lo",
  direction="r",
 },
 [0x109B3]={
  category="lo",
  direction="r",
 },
 [0x109B4]={
  category="lo",
  direction="r",
 },
 [0x109B5]={
  category="lo",
  direction="r",
 },
 [0x109B6]={
  category="lo",
  direction="r",
 },
 [0x109B7]={
  category="lo",
  direction="r",
 },
 [0x109BE]={
  category="lo",
  direction="r",
 },
 [0x109BF]={
  category="lo",
  direction="r",
 },
 [0x10A00]={
  category="lo",
  direction="r",
 },
 [0x10A01]={
  category="mn",
  direction="nsm",
 },
 [0x10A02]={
  category="mn",
  direction="nsm",
 },
 [0x10A03]={
  category="mn",
  direction="nsm",
 },
 [0x10A05]={
  category="mn",
  direction="nsm",
 },
 [0x10A06]={
  category="mn",
  direction="nsm",
 },
 [0x10A0C]={
  category="mn",
  direction="nsm",
 },
 [0x10A0D]={
  category="mn",
  direction="nsm",
 },
 [0x10A0E]={
  category="mn",
  direction="nsm",
 },
 [0x10A0F]={
  category="mn",
  direction="nsm",
 },
 [0x10A10]={
  category="lo",
  direction="r",
 },
 [0x10A11]={
  category="lo",
  direction="r",
 },
 [0x10A12]={
  category="lo",
  direction="r",
 },
 [0x10A13]={
  category="lo",
  direction="r",
 },
 [0x10A15]={
  category="lo",
  direction="r",
 },
 [0x10A16]={
  category="lo",
  direction="r",
 },
 [0x10A17]={
  category="lo",
  direction="r",
 },
 [0x10A19]={
  category="lo",
  direction="r",
 },
 [0x10A1A]={
  category="lo",
  direction="r",
 },
 [0x10A1B]={
  category="lo",
  direction="r",
 },
 [0x10A1C]={
  category="lo",
  direction="r",
 },
 [0x10A1D]={
  category="lo",
  direction="r",
 },
 [0x10A1E]={
  category="lo",
  direction="r",
 },
 [0x10A1F]={
  category="lo",
  direction="r",
 },
 [0x10A20]={
  category="lo",
  direction="r",
 },
 [0x10A21]={
  category="lo",
  direction="r",
 },
 [0x10A22]={
  category="lo",
  direction="r",
 },
 [0x10A23]={
  category="lo",
  direction="r",
 },
 [0x10A24]={
  category="lo",
  direction="r",
 },
 [0x10A25]={
  category="lo",
  direction="r",
 },
 [0x10A26]={
  category="lo",
  direction="r",
 },
 [0x10A27]={
  category="lo",
  direction="r",
 },
 [0x10A28]={
  category="lo",
  direction="r",
 },
 [0x10A29]={
  category="lo",
  direction="r",
 },
 [0x10A2A]={
  category="lo",
  direction="r",
 },
 [0x10A2B]={
  category="lo",
  direction="r",
 },
 [0x10A2C]={
  category="lo",
  direction="r",
 },
 [0x10A2D]={
  category="lo",
  direction="r",
 },
 [0x10A2E]={
  category="lo",
  direction="r",
 },
 [0x10A2F]={
  category="lo",
  direction="r",
 },
 [0x10A30]={
  category="lo",
  direction="r",
 },
 [0x10A31]={
  category="lo",
  direction="r",
 },
 [0x10A32]={
  category="lo",
  direction="r",
 },
 [0x10A33]={
  category="lo",
  direction="r",
 },
 [0x10A38]={
  category="mn",
  direction="nsm",
 },
 [0x10A39]={
  category="mn",
  direction="nsm",
 },
 [0x10A3A]={
  category="mn",
  direction="nsm",
 },
 [0x10A3F]={
  category="mn",
  direction="nsm",
 },
 [0x10A40]={
  category="no",
  direction="r",
 },
 [0x10A41]={
  category="no",
  direction="r",
 },
 [0x10A42]={
  category="no",
  direction="r",
 },
 [0x10A43]={
  category="no",
  direction="r",
 },
 [0x10A44]={
  category="no",
  direction="r",
 },
 [0x10A45]={
  category="no",
  direction="r",
 },
 [0x10A46]={
  category="no",
  direction="r",
 },
 [0x10A47]={
  category="no",
  direction="r",
 },
 [0x10A50]={
  category="po",
  direction="r",
 },
 [0x10A51]={
  category="po",
  direction="r",
 },
 [0x10A52]={
  category="po",
  direction="r",
 },
 [0x10A53]={
  category="po",
  direction="r",
 },
 [0x10A54]={
  category="po",
  direction="r",
 },
 [0x10A55]={
  category="po",
  direction="r",
 },
 [0x10A56]={
  category="po",
  direction="r",
 },
 [0x10A57]={
  category="po",
  direction="r",
 },
 [0x10A58]={
  category="po",
  direction="r",
 },
 [0x10A60]={
  category="lo",
  direction="r",
 },
 [0x10A61]={
  category="lo",
  direction="r",
 },
 [0x10A62]={
  category="lo",
  direction="r",
 },
 [0x10A63]={
  category="lo",
  direction="r",
 },
 [0x10A64]={
  category="lo",
  direction="r",
 },
 [0x10A65]={
  category="lo",
  direction="r",
 },
 [0x10A66]={
  category="lo",
  direction="r",
 },
 [0x10A67]={
  category="lo",
  direction="r",
 },
 [0x10A68]={
  category="lo",
  direction="r",
 },
 [0x10A69]={
  category="lo",
  direction="r",
 },
 [0x10A6A]={
  category="lo",
  direction="r",
 },
 [0x10A6B]={
  category="lo",
  direction="r",
 },
 [0x10A6C]={
  category="lo",
  direction="r",
 },
 [0x10A6D]={
  category="lo",
  direction="r",
 },
 [0x10A6E]={
  category="lo",
  direction="r",
 },
 [0x10A6F]={
  category="lo",
  direction="r",
 },
 [0x10A70]={
  category="lo",
  direction="r",
 },
 [0x10A71]={
  category="lo",
  direction="r",
 },
 [0x10A72]={
  category="lo",
  direction="r",
 },
 [0x10A73]={
  category="lo",
  direction="r",
 },
 [0x10A74]={
  category="lo",
  direction="r",
 },
 [0x10A75]={
  category="lo",
  direction="r",
 },
 [0x10A76]={
  category="lo",
  direction="r",
 },
 [0x10A77]={
  category="lo",
  direction="r",
 },
 [0x10A78]={
  category="lo",
  direction="r",
 },
 [0x10A79]={
  category="lo",
  direction="r",
 },
 [0x10A7A]={
  category="lo",
  direction="r",
 },
 [0x10A7B]={
  category="lo",
  direction="r",
 },
 [0x10A7C]={
  category="lo",
  direction="r",
 },
 [0x10A7D]={
  category="no",
  direction="r",
 },
 [0x10A7E]={
  category="no",
  direction="r",
 },
 [0x10A7F]={
  category="po",
  direction="r",
 },
 [0x10A80]={
  category="lo",
  direction="r",
 },
 [0x10A81]={
  category="lo",
  direction="r",
 },
 [0x10A82]={
  category="lo",
  direction="r",
 },
 [0x10A83]={
  category="lo",
  direction="r",
 },
 [0x10A84]={
  category="lo",
  direction="r",
 },
 [0x10A85]={
  category="lo",
  direction="r",
 },
 [0x10A86]={
  category="lo",
  direction="r",
 },
 [0x10A87]={
  category="lo",
  direction="r",
 },
 [0x10A88]={
  category="lo",
  direction="r",
 },
 [0x10A89]={
  category="lo",
  direction="r",
 },
 [0x10A8A]={
  category="lo",
  direction="r",
 },
 [0x10A8B]={
  category="lo",
  direction="r",
 },
 [0x10A8C]={
  category="lo",
  direction="r",
 },
 [0x10A8D]={
  category="lo",
  direction="r",
 },
 [0x10A8E]={
  category="lo",
  direction="r",
 },
 [0x10A8F]={
  category="lo",
  direction="r",
 },
 [0x10A90]={
  category="lo",
  direction="r",
 },
 [0x10A91]={
  category="lo",
  direction="r",
 },
 [0x10A92]={
  category="lo",
  direction="r",
 },
 [0x10A93]={
  category="lo",
  direction="r",
 },
 [0x10A94]={
  category="lo",
  direction="r",
 },
 [0x10A95]={
  category="lo",
  direction="r",
 },
 [0x10A96]={
  category="lo",
  direction="r",
 },
 [0x10A97]={
  category="lo",
  direction="r",
 },
 [0x10A98]={
  category="lo",
  direction="r",
 },
 [0x10A99]={
  category="lo",
  direction="r",
 },
 [0x10A9A]={
  category="lo",
  direction="r",
 },
 [0x10A9B]={
  category="lo",
  direction="r",
 },
 [0x10A9C]={
  category="lo",
  direction="r",
 },
 [0x10A9D]={
  category="no",
  direction="r",
 },
 [0x10A9E]={
  category="no",
  direction="r",
 },
 [0x10A9F]={
  category="no",
  direction="r",
 },
 [0x10AC0]={
  category="lo",
  direction="r",
 },
 [0x10AC1]={
  category="lo",
  direction="r",
 },
 [0x10AC2]={
  category="lo",
  direction="r",
 },
 [0x10AC3]={
  category="lo",
  direction="r",
 },
 [0x10AC4]={
  category="lo",
  direction="r",
 },
 [0x10AC5]={
  category="lo",
  direction="r",
 },
 [0x10AC6]={
  category="lo",
  direction="r",
 },
 [0x10AC7]={
  category="lo",
  direction="r",
 },
 [0x10AC8]={
  category="so",
  direction="r",
 },
 [0x10AC9]={
  category="lo",
  direction="r",
 },
 [0x10ACA]={
  category="lo",
  direction="r",
 },
 [0x10ACB]={
  category="lo",
  direction="r",
 },
 [0x10ACC]={
  category="lo",
  direction="r",
 },
 [0x10ACD]={
  category="lo",
  direction="r",
 },
 [0x10ACE]={
  category="lo",
  direction="r",
 },
 [0x10ACF]={
  category="lo",
  direction="r",
 },
 [0x10AD0]={
  category="lo",
  direction="r",
 },
 [0x10AD1]={
  category="lo",
  direction="r",
 },
 [0x10AD2]={
  category="lo",
  direction="r",
 },
 [0x10AD3]={
  category="lo",
  direction="r",
 },
 [0x10AD4]={
  category="lo",
  direction="r",
 },
 [0x10AD5]={
  category="lo",
  direction="r",
 },
 [0x10AD6]={
  category="lo",
  direction="r",
 },
 [0x10AD7]={
  category="lo",
  direction="r",
 },
 [0x10AD8]={
  category="lo",
  direction="r",
 },
 [0x10AD9]={
  category="lo",
  direction="r",
 },
 [0x10ADA]={
  category="lo",
  direction="r",
 },
 [0x10ADB]={
  category="lo",
  direction="r",
 },
 [0x10ADC]={
  category="lo",
  direction="r",
 },
 [0x10ADD]={
  category="lo",
  direction="r",
 },
 [0x10ADE]={
  category="lo",
  direction="r",
 },
 [0x10ADF]={
  category="lo",
  direction="r",
 },
 [0x10AE0]={
  category="lo",
  direction="r",
 },
 [0x10AE1]={
  category="lo",
  direction="r",
 },
 [0x10AE2]={
  category="lo",
  direction="r",
 },
 [0x10AE3]={
  category="lo",
  direction="r",
 },
 [0x10AE4]={
  category="lo",
  direction="r",
 },
 [0x10AE5]={
  category="mn",
  direction="nsm",
 },
 [0x10AE6]={
  category="mn",
  direction="nsm",
 },
 [0x10AEB]={
  category="no",
  direction="r",
 },
 [0x10AEC]={
  category="no",
  direction="r",
 },
 [0x10AED]={
  category="no",
  direction="r",
 },
 [0x10AEE]={
  category="no",
  direction="r",
 },
 [0x10AEF]={
  category="no",
  direction="r",
 },
 [0x10AF0]={
  category="po",
  direction="r",
 },
 [0x10AF1]={
  category="po",
  direction="r",
 },
 [0x10AF2]={
  category="po",
  direction="r",
 },
 [0x10AF3]={
  category="po",
  direction="r",
 },
 [0x10AF4]={
  category="po",
  direction="r",
 },
 [0x10AF5]={
  category="po",
  direction="r",
 },
 [0x10AF6]={
  category="po",
  direction="r",
 },
 [0x10B00]={
  category="lo",
  direction="r",
 },
 [0x10B01]={
  category="lo",
  direction="r",
 },
 [0x10B02]={
  category="lo",
  direction="r",
 },
 [0x10B03]={
  category="lo",
  direction="r",
 },
 [0x10B04]={
  category="lo",
  direction="r",
 },
 [0x10B05]={
  category="lo",
  direction="r",
 },
 [0x10B06]={
  category="lo",
  direction="r",
 },
 [0x10B07]={
  category="lo",
  direction="r",
 },
 [0x10B08]={
  category="lo",
  direction="r",
 },
 [0x10B09]={
  category="lo",
  direction="r",
 },
 [0x10B0A]={
  category="lo",
  direction="r",
 },
 [0x10B0B]={
  category="lo",
  direction="r",
 },
 [0x10B0C]={
  category="lo",
  direction="r",
 },
 [0x10B0D]={
  category="lo",
  direction="r",
 },
 [0x10B0E]={
  category="lo",
  direction="r",
 },
 [0x10B0F]={
  category="lo",
  direction="r",
 },
 [0x10B10]={
  category="lo",
  direction="r",
 },
 [0x10B11]={
  category="lo",
  direction="r",
 },
 [0x10B12]={
  category="lo",
  direction="r",
 },
 [0x10B13]={
  category="lo",
  direction="r",
 },
 [0x10B14]={
  category="lo",
  direction="r",
 },
 [0x10B15]={
  category="lo",
  direction="r",
 },
 [0x10B16]={
  category="lo",
  direction="r",
 },
 [0x10B17]={
  category="lo",
  direction="r",
 },
 [0x10B18]={
  category="lo",
  direction="r",
 },
 [0x10B19]={
  category="lo",
  direction="r",
 },
 [0x10B1A]={
  category="lo",
  direction="r",
 },
 [0x10B1B]={
  category="lo",
  direction="r",
 },
 [0x10B1C]={
  category="lo",
  direction="r",
 },
 [0x10B1D]={
  category="lo",
  direction="r",
 },
 [0x10B1E]={
  category="lo",
  direction="r",
 },
 [0x10B1F]={
  category="lo",
  direction="r",
 },
 [0x10B20]={
  category="lo",
  direction="r",
 },
 [0x10B21]={
  category="lo",
  direction="r",
 },
 [0x10B22]={
  category="lo",
  direction="r",
 },
 [0x10B23]={
  category="lo",
  direction="r",
 },
 [0x10B24]={
  category="lo",
  direction="r",
 },
 [0x10B25]={
  category="lo",
  direction="r",
 },
 [0x10B26]={
  category="lo",
  direction="r",
 },
 [0x10B27]={
  category="lo",
  direction="r",
 },
 [0x10B28]={
  category="lo",
  direction="r",
 },
 [0x10B29]={
  category="lo",
  direction="r",
 },
 [0x10B2A]={
  category="lo",
  direction="r",
 },
 [0x10B2B]={
  category="lo",
  direction="r",
 },
 [0x10B2C]={
  category="lo",
  direction="r",
 },
 [0x10B2D]={
  category="lo",
  direction="r",
 },
 [0x10B2E]={
  category="lo",
  direction="r",
 },
 [0x10B2F]={
  category="lo",
  direction="r",
 },
 [0x10B30]={
  category="lo",
  direction="r",
 },
 [0x10B31]={
  category="lo",
  direction="r",
 },
 [0x10B32]={
  category="lo",
  direction="r",
 },
 [0x10B33]={
  category="lo",
  direction="r",
 },
 [0x10B34]={
  category="lo",
  direction="r",
 },
 [0x10B35]={
  category="lo",
  direction="r",
 },
 [0x10B39]={
  category="po",
  direction="on",
 },
 [0x10B3A]={
  category="po",
  direction="on",
 },
 [0x10B3B]={
  category="po",
  direction="on",
 },
 [0x10B3C]={
  category="po",
  direction="on",
 },
 [0x10B3D]={
  category="po",
  direction="on",
 },
 [0x10B3E]={
  category="po",
  direction="on",
 },
 [0x10B3F]={
  category="po",
  direction="on",
 },
 [0x10B40]={
  category="lo",
  direction="r",
 },
 [0x10B41]={
  category="lo",
  direction="r",
 },
 [0x10B42]={
  category="lo",
  direction="r",
 },
 [0x10B43]={
  category="lo",
  direction="r",
 },
 [0x10B44]={
  category="lo",
  direction="r",
 },
 [0x10B45]={
  category="lo",
  direction="r",
 },
 [0x10B46]={
  category="lo",
  direction="r",
 },
 [0x10B47]={
  category="lo",
  direction="r",
 },
 [0x10B48]={
  category="lo",
  direction="r",
 },
 [0x10B49]={
  category="lo",
  direction="r",
 },
 [0x10B4A]={
  category="lo",
  direction="r",
 },
 [0x10B4B]={
  category="lo",
  direction="r",
 },
 [0x10B4C]={
  category="lo",
  direction="r",
 },
 [0x10B4D]={
  category="lo",
  direction="r",
 },
 [0x10B4E]={
  category="lo",
  direction="r",
 },
 [0x10B4F]={
  category="lo",
  direction="r",
 },
 [0x10B50]={
  category="lo",
  direction="r",
 },
 [0x10B51]={
  category="lo",
  direction="r",
 },
 [0x10B52]={
  category="lo",
  direction="r",
 },
 [0x10B53]={
  category="lo",
  direction="r",
 },
 [0x10B54]={
  category="lo",
  direction="r",
 },
 [0x10B55]={
  category="lo",
  direction="r",
 },
 [0x10B58]={
  category="no",
  direction="r",
 },
 [0x10B59]={
  category="no",
  direction="r",
 },
 [0x10B5A]={
  category="no",
  direction="r",
 },
 [0x10B5B]={
  category="no",
  direction="r",
 },
 [0x10B5C]={
  category="no",
  direction="r",
 },
 [0x10B5D]={
  category="no",
  direction="r",
 },
 [0x10B5E]={
  category="no",
  direction="r",
 },
 [0x10B5F]={
  category="no",
  direction="r",
 },
 [0x10B60]={
  category="lo",
  direction="r",
 },
 [0x10B61]={
  category="lo",
  direction="r",
 },
 [0x10B62]={
  category="lo",
  direction="r",
 },
 [0x10B63]={
  category="lo",
  direction="r",
 },
 [0x10B64]={
  category="lo",
  direction="r",
 },
 [0x10B65]={
  category="lo",
  direction="r",
 },
 [0x10B66]={
  category="lo",
  direction="r",
 },
 [0x10B67]={
  category="lo",
  direction="r",
 },
 [0x10B68]={
  category="lo",
  direction="r",
 },
 [0x10B69]={
  category="lo",
  direction="r",
 },
 [0x10B6A]={
  category="lo",
  direction="r",
 },
 [0x10B6B]={
  category="lo",
  direction="r",
 },
 [0x10B6C]={
  category="lo",
  direction="r",
 },
 [0x10B6D]={
  category="lo",
  direction="r",
 },
 [0x10B6E]={
  category="lo",
  direction="r",
 },
 [0x10B6F]={
  category="lo",
  direction="r",
 },
 [0x10B70]={
  category="lo",
  direction="r",
 },
 [0x10B71]={
  category="lo",
  direction="r",
 },
 [0x10B72]={
  category="lo",
  direction="r",
 },
 [0x10B78]={
  category="no",
  direction="r",
 },
 [0x10B79]={
  category="no",
  direction="r",
 },
 [0x10B7A]={
  category="no",
  direction="r",
 },
 [0x10B7B]={
  category="no",
  direction="r",
 },
 [0x10B7C]={
  category="no",
  direction="r",
 },
 [0x10B7D]={
  category="no",
  direction="r",
 },
 [0x10B7E]={
  category="no",
  direction="r",
 },
 [0x10B7F]={
  category="no",
  direction="r",
 },
 [0x10B80]={
  category="lo",
  direction="r",
 },
 [0x10B81]={
  category="lo",
  direction="r",
 },
 [0x10B82]={
  category="lo",
  direction="r",
 },
 [0x10B83]={
  category="lo",
  direction="r",
 },
 [0x10B84]={
  category="lo",
  direction="r",
 },
 [0x10B85]={
  category="lo",
  direction="r",
 },
 [0x10B86]={
  category="lo",
  direction="r",
 },
 [0x10B87]={
  category="lo",
  direction="r",
 },
 [0x10B88]={
  category="lo",
  direction="r",
 },
 [0x10B89]={
  category="lo",
  direction="r",
 },
 [0x10B8A]={
  category="lo",
  direction="r",
 },
 [0x10B8B]={
  category="lo",
  direction="r",
 },
 [0x10B8C]={
  category="lo",
  direction="r",
 },
 [0x10B8D]={
  category="lo",
  direction="r",
 },
 [0x10B8E]={
  category="lo",
  direction="r",
 },
 [0x10B8F]={
  category="lo",
  direction="r",
 },
 [0x10B90]={
  category="lo",
  direction="r",
 },
 [0x10B91]={
  category="lo",
  direction="r",
 },
 [0x10B99]={
  category="po",
  direction="r",
 },
 [0x10B9A]={
  category="po",
  direction="r",
 },
 [0x10B9B]={
  category="po",
  direction="r",
 },
 [0x10B9C]={
  category="po",
  direction="r",
 },
 [0x10BA9]={
  category="no",
  direction="r",
 },
 [0x10BAA]={
  category="no",
  direction="r",
 },
 [0x10BAB]={
  category="no",
  direction="r",
 },
 [0x10BAC]={
  category="no",
  direction="r",
 },
 [0x10BAD]={
  category="no",
  direction="r",
 },
 [0x10BAE]={
  category="no",
  direction="r",
 },
 [0x10BAF]={
  category="no",
  direction="r",
 },
 [0x10C00]={
  category="lo",
  direction="r",
 },
 [0x10C01]={
  category="lo",
  direction="r",
 },
 [0x10C02]={
  category="lo",
  direction="r",
 },
 [0x10C03]={
  category="lo",
  direction="r",
 },
 [0x10C04]={
  category="lo",
  direction="r",
 },
 [0x10C05]={
  category="lo",
  direction="r",
 },
 [0x10C06]={
  category="lo",
  direction="r",
 },
 [0x10C07]={
  category="lo",
  direction="r",
 },
 [0x10C08]={
  category="lo",
  direction="r",
 },
 [0x10C09]={
  category="lo",
  direction="r",
 },
 [0x10C0A]={
  category="lo",
  direction="r",
 },
 [0x10C0B]={
  category="lo",
  direction="r",
 },
 [0x10C0C]={
  category="lo",
  direction="r",
 },
 [0x10C0D]={
  category="lo",
  direction="r",
 },
 [0x10C0E]={
  category="lo",
  direction="r",
 },
 [0x10C0F]={
  category="lo",
  direction="r",
 },
 [0x10C10]={
  category="lo",
  direction="r",
 },
 [0x10C11]={
  category="lo",
  direction="r",
 },
 [0x10C12]={
  category="lo",
  direction="r",
 },
 [0x10C13]={
  category="lo",
  direction="r",
 },
 [0x10C14]={
  category="lo",
  direction="r",
 },
 [0x10C15]={
  category="lo",
  direction="r",
 },
 [0x10C16]={
  category="lo",
  direction="r",
 },
 [0x10C17]={
  category="lo",
  direction="r",
 },
 [0x10C18]={
  category="lo",
  direction="r",
 },
 [0x10C19]={
  category="lo",
  direction="r",
 },
 [0x10C1A]={
  category="lo",
  direction="r",
 },
 [0x10C1B]={
  category="lo",
  direction="r",
 },
 [0x10C1C]={
  category="lo",
  direction="r",
 },
 [0x10C1D]={
  category="lo",
  direction="r",
 },
 [0x10C1E]={
  category="lo",
  direction="r",
 },
 [0x10C1F]={
  category="lo",
  direction="r",
 },
 [0x10C20]={
  category="lo",
  direction="r",
 },
 [0x10C21]={
  category="lo",
  direction="r",
 },
 [0x10C22]={
  category="lo",
  direction="r",
 },
 [0x10C23]={
  category="lo",
  direction="r",
 },
 [0x10C24]={
  category="lo",
  direction="r",
 },
 [0x10C25]={
  category="lo",
  direction="r",
 },
 [0x10C26]={
  category="lo",
  direction="r",
 },
 [0x10C27]={
  category="lo",
  direction="r",
 },
 [0x10C28]={
  category="lo",
  direction="r",
 },
 [0x10C29]={
  category="lo",
  direction="r",
 },
 [0x10C2A]={
  category="lo",
  direction="r",
 },
 [0x10C2B]={
  category="lo",
  direction="r",
 },
 [0x10C2C]={
  category="lo",
  direction="r",
 },
 [0x10C2D]={
  category="lo",
  direction="r",
 },
 [0x10C2E]={
  category="lo",
  direction="r",
 },
 [0x10C2F]={
  category="lo",
  direction="r",
 },
 [0x10C30]={
  category="lo",
  direction="r",
 },
 [0x10C31]={
  category="lo",
  direction="r",
 },
 [0x10C32]={
  category="lo",
  direction="r",
 },
 [0x10C33]={
  category="lo",
  direction="r",
 },
 [0x10C34]={
  category="lo",
  direction="r",
 },
 [0x10C35]={
  category="lo",
  direction="r",
 },
 [0x10C36]={
  category="lo",
  direction="r",
 },
 [0x10C37]={
  category="lo",
  direction="r",
 },
 [0x10C38]={
  category="lo",
  direction="r",
 },
 [0x10C39]={
  category="lo",
  direction="r",
 },
 [0x10C3A]={
  category="lo",
  direction="r",
 },
 [0x10C3B]={
  category="lo",
  direction="r",
 },
 [0x10C3C]={
  category="lo",
  direction="r",
 },
 [0x10C3D]={
  category="lo",
  direction="r",
 },
 [0x10C3E]={
  category="lo",
  direction="r",
 },
 [0x10C3F]={
  category="lo",
  direction="r",
 },
 [0x10C40]={
  category="lo",
  direction="r",
 },
 [0x10C41]={
  category="lo",
  direction="r",
 },
 [0x10C42]={
  category="lo",
  direction="r",
 },
 [0x10C43]={
  category="lo",
  direction="r",
 },
 [0x10C44]={
  category="lo",
  direction="r",
 },
 [0x10C45]={
  category="lo",
  direction="r",
 },
 [0x10C46]={
  category="lo",
  direction="r",
 },
 [0x10C47]={
  category="lo",
  direction="r",
 },
 [0x10C48]={
  category="lo",
  direction="r",
 },
 [0x10E60]={
  category="no",
  direction="an",
 },
 [0x10E61]={
  category="no",
  direction="an",
 },
 [0x10E62]={
  category="no",
  direction="an",
 },
 [0x10E63]={
  category="no",
  direction="an",
 },
 [0x10E64]={
  category="no",
  direction="an",
 },
 [0x10E65]={
  category="no",
  direction="an",
 },
 [0x10E66]={
  category="no",
  direction="an",
 },
 [0x10E67]={
  category="no",
  direction="an",
 },
 [0x10E68]={
  category="no",
  direction="an",
 },
 [0x10E69]={
  category="no",
  direction="an",
 },
 [0x10E6A]={
  category="no",
  direction="an",
 },
 [0x10E6B]={
  category="no",
  direction="an",
 },
 [0x10E6C]={
  category="no",
  direction="an",
 },
 [0x10E6D]={
  category="no",
  direction="an",
 },
 [0x10E6E]={
  category="no",
  direction="an",
 },
 [0x10E6F]={
  category="no",
  direction="an",
 },
 [0x10E70]={
  category="no",
  direction="an",
 },
 [0x10E71]={
  category="no",
  direction="an",
 },
 [0x10E72]={
  category="no",
  direction="an",
 },
 [0x10E73]={
  category="no",
  direction="an",
 },
 [0x10E74]={
  category="no",
  direction="an",
 },
 [0x10E75]={
  category="no",
  direction="an",
 },
 [0x10E76]={
  category="no",
  direction="an",
 },
 [0x10E77]={
  category="no",
  direction="an",
 },
 [0x10E78]={
  category="no",
  direction="an",
 },
 [0x10E79]={
  category="no",
  direction="an",
 },
 [0x10E7A]={
  category="no",
  direction="an",
 },
 [0x10E7B]={
  category="no",
  direction="an",
 },
 [0x10E7C]={
  category="no",
  direction="an",
 },
 [0x10E7D]={
  category="no",
  direction="an",
 },
 [0x10E7E]={
  category="no",
  direction="an",
 },
 [0x11000]={
  category="mc",
  direction="l",
 },
 [0x11001]={
  category="mn",
  direction="nsm",
 },
 [0x11002]={
  category="mc",
  direction="l",
 },
 [0x11003]={
  category="lo",
  direction="l",
 },
 [0x11004]={
  category="lo",
  direction="l",
 },
 [0x11005]={
  category="lo",
  direction="l",
 },
 [0x11006]={
  category="lo",
  direction="l",
 },
 [0x11007]={
  category="lo",
  direction="l",
 },
 [0x11008]={
  category="lo",
  direction="l",
 },
 [0x11009]={
  category="lo",
  direction="l",
 },
 [0x1100A]={
  category="lo",
  direction="l",
 },
 [0x1100B]={
  category="lo",
  direction="l",
 },
 [0x1100C]={
  category="lo",
  direction="l",
 },
 [0x1100D]={
  category="lo",
  direction="l",
 },
 [0x1100E]={
  category="lo",
  direction="l",
 },
 [0x1100F]={
  category="lo",
  direction="l",
 },
 [0x11010]={
  category="lo",
  direction="l",
 },
 [0x11011]={
  category="lo",
  direction="l",
 },
 [0x11012]={
  category="lo",
  direction="l",
 },
 [0x11013]={
  category="lo",
  direction="l",
 },
 [0x11014]={
  category="lo",
  direction="l",
 },
 [0x11015]={
  category="lo",
  direction="l",
 },
 [0x11016]={
  category="lo",
  direction="l",
 },
 [0x11017]={
  category="lo",
  direction="l",
 },
 [0x11018]={
  category="lo",
  direction="l",
 },
 [0x11019]={
  category="lo",
  direction="l",
 },
 [0x1101A]={
  category="lo",
  direction="l",
 },
 [0x1101B]={
  category="lo",
  direction="l",
 },
 [0x1101C]={
  category="lo",
  direction="l",
 },
 [0x1101D]={
  category="lo",
  direction="l",
 },
 [0x1101E]={
  category="lo",
  direction="l",
 },
 [0x1101F]={
  category="lo",
  direction="l",
 },
 [0x11020]={
  category="lo",
  direction="l",
 },
 [0x11021]={
  category="lo",
  direction="l",
 },
 [0x11022]={
  category="lo",
  direction="l",
 },
 [0x11023]={
  category="lo",
  direction="l",
 },
 [0x11024]={
  category="lo",
  direction="l",
 },
 [0x11025]={
  category="lo",
  direction="l",
 },
 [0x11026]={
  category="lo",
  direction="l",
 },
 [0x11027]={
  category="lo",
  direction="l",
 },
 [0x11028]={
  category="lo",
  direction="l",
 },
 [0x11029]={
  category="lo",
  direction="l",
 },
 [0x1102A]={
  category="lo",
  direction="l",
 },
 [0x1102B]={
  category="lo",
  direction="l",
 },
 [0x1102C]={
  category="lo",
  direction="l",
 },
 [0x1102D]={
  category="lo",
  direction="l",
 },
 [0x1102E]={
  category="lo",
  direction="l",
 },
 [0x1102F]={
  category="lo",
  direction="l",
 },
 [0x11030]={
  category="lo",
  direction="l",
 },
 [0x11031]={
  category="lo",
  direction="l",
 },
 [0x11032]={
  category="lo",
  direction="l",
 },
 [0x11033]={
  category="lo",
  direction="l",
 },
 [0x11034]={
  category="lo",
  direction="l",
 },
 [0x11035]={
  category="lo",
  direction="l",
 },
 [0x11036]={
  category="lo",
  direction="l",
 },
 [0x11037]={
  category="lo",
  direction="l",
 },
 [0x11038]={
  category="mn",
  direction="nsm",
 },
 [0x11039]={
  category="mn",
  direction="nsm",
 },
 [0x1103A]={
  category="mn",
  direction="nsm",
 },
 [0x1103B]={
  category="mn",
  direction="nsm",
 },
 [0x1103C]={
  category="mn",
  direction="nsm",
 },
 [0x1103D]={
  category="mn",
  direction="nsm",
 },
 [0x1103E]={
  category="mn",
  direction="nsm",
 },
 [0x1103F]={
  category="mn",
  direction="nsm",
 },
 [0x11040]={
  category="mn",
  direction="nsm",
 },
 [0x11041]={
  category="mn",
  direction="nsm",
 },
 [0x11042]={
  category="mn",
  direction="nsm",
 },
 [0x11043]={
  category="mn",
  direction="nsm",
 },
 [0x11044]={
  category="mn",
  direction="nsm",
 },
 [0x11045]={
  category="mn",
  direction="nsm",
 },
 [0x11046]={
  category="mn",
  direction="nsm",
 },
 [0x11047]={
  category="po",
  direction="l",
 },
 [0x11048]={
  category="po",
  direction="l",
 },
 [0x11049]={
  category="po",
  direction="l",
 },
 [0x1104A]={
  category="po",
  direction="l",
 },
 [0x1104B]={
  category="po",
  direction="l",
 },
 [0x1104C]={
  category="po",
  direction="l",
 },
 [0x1104D]={
  category="po",
  direction="l",
 },
 [0x11052]={
  category="no",
  direction="on",
 },
 [0x11053]={
  category="no",
  direction="on",
 },
 [0x11054]={
  category="no",
  direction="on",
 },
 [0x11055]={
  category="no",
  direction="on",
 },
 [0x11056]={
  category="no",
  direction="on",
 },
 [0x11057]={
  category="no",
  direction="on",
 },
 [0x11058]={
  category="no",
  direction="on",
 },
 [0x11059]={
  category="no",
  direction="on",
 },
 [0x1105A]={
  category="no",
  direction="on",
 },
 [0x1105B]={
  category="no",
  direction="on",
 },
 [0x1105C]={
  category="no",
  direction="on",
 },
 [0x1105D]={
  category="no",
  direction="on",
 },
 [0x1105E]={
  category="no",
  direction="on",
 },
 [0x1105F]={
  category="no",
  direction="on",
 },
 [0x11060]={
  category="no",
  direction="on",
 },
 [0x11061]={
  category="no",
  direction="on",
 },
 [0x11062]={
  category="no",
  direction="on",
 },
 [0x11063]={
  category="no",
  direction="on",
 },
 [0x11064]={
  category="no",
  direction="on",
 },
 [0x11065]={
  category="no",
  direction="on",
 },
 [0x11066]={
  category="nd",
  direction="l",
 },
 [0x11067]={
  category="nd",
  direction="l",
 },
 [0x11068]={
  category="nd",
  direction="l",
 },
 [0x11069]={
  category="nd",
  direction="l",
 },
 [0x1106A]={
  category="nd",
  direction="l",
 },
 [0x1106B]={
  category="nd",
  direction="l",
 },
 [0x1106C]={
  category="nd",
  direction="l",
 },
 [0x1106D]={
  category="nd",
  direction="l",
 },
 [0x1106E]={
  category="nd",
  direction="l",
 },
 [0x1106F]={
  category="nd",
  direction="l",
 },
 [0x1107F]={
  category="mn",
  direction="nsm",
 },
 [0x11080]={
  category="mn",
  direction="nsm",
 },
 [0x11081]={
  category="mn",
  direction="nsm",
 },
 [0x11082]={
  category="mc",
  direction="l",
 },
 [0x11083]={
  category="lo",
  direction="l",
 },
 [0x11084]={
  category="lo",
  direction="l",
 },
 [0x11085]={
  category="lo",
  direction="l",
 },
 [0x11086]={
  category="lo",
  direction="l",
 },
 [0x11087]={
  category="lo",
  direction="l",
 },
 [0x11088]={
  category="lo",
  direction="l",
 },
 [0x11089]={
  category="lo",
  direction="l",
 },
 [0x1108A]={
  category="lo",
  direction="l",
 },
 [0x1108B]={
  category="lo",
  direction="l",
 },
 [0x1108C]={
  category="lo",
  direction="l",
 },
 [0x1108D]={
  category="lo",
  direction="l",
 },
 [0x1108E]={
  category="lo",
  direction="l",
 },
 [0x1108F]={
  category="lo",
  direction="l",
 },
 [0x11090]={
  category="lo",
  direction="l",
 },
 [0x11091]={
  category="lo",
  direction="l",
 },
 [0x11092]={
  category="lo",
  direction="l",
 },
 [0x11093]={
  category="lo",
  direction="l",
 },
 [0x11094]={
  category="lo",
  direction="l",
 },
 [0x11095]={
  category="lo",
  direction="l",
 },
 [0x11096]={
  category="lo",
  direction="l",
 },
 [0x11097]={
  category="lo",
  direction="l",
 },
 [0x11098]={
  category="lo",
  direction="l",
 },
 [0x11099]={
  category="lo",
  direction="l",
 },
 [0x1109A]={
  category="lo",
  direction="l",
 },
 [0x1109B]={
  category="lo",
  direction="l",
 },
 [0x1109C]={
  category="lo",
  direction="l",
 },
 [0x1109D]={
  category="lo",
  direction="l",
 },
 [0x1109E]={
  category="lo",
  direction="l",
 },
 [0x1109F]={
  category="lo",
  direction="l",
 },
 [0x110A0]={
  category="lo",
  direction="l",
 },
 [0x110A1]={
  category="lo",
  direction="l",
 },
 [0x110A2]={
  category="lo",
  direction="l",
 },
 [0x110A3]={
  category="lo",
  direction="l",
 },
 [0x110A4]={
  category="lo",
  direction="l",
 },
 [0x110A5]={
  category="lo",
  direction="l",
 },
 [0x110A6]={
  category="lo",
  direction="l",
 },
 [0x110A7]={
  category="lo",
  direction="l",
 },
 [0x110A8]={
  category="lo",
  direction="l",
 },
 [0x110A9]={
  category="lo",
  direction="l",
 },
 [0x110AA]={
  category="lo",
  direction="l",
 },
 [0x110AB]={
  category="lo",
  direction="l",
 },
 [0x110AC]={
  category="lo",
  direction="l",
 },
 [0x110AD]={
  category="lo",
  direction="l",
 },
 [0x110AE]={
  category="lo",
  direction="l",
 },
 [0x110AF]={
  category="lo",
  direction="l",
 },
 [0x110B0]={
  category="mc",
  direction="l",
 },
 [0x110B1]={
  category="mc",
  direction="l",
 },
 [0x110B2]={
  category="mc",
  direction="l",
 },
 [0x110B3]={
  category="mn",
  direction="nsm",
 },
 [0x110B4]={
  category="mn",
  direction="nsm",
 },
 [0x110B5]={
  category="mn",
  direction="nsm",
 },
 [0x110B6]={
  category="mn",
  direction="nsm",
 },
 [0x110B7]={
  category="mc",
  direction="l",
 },
 [0x110B8]={
  category="mc",
  direction="l",
 },
 [0x110B9]={
  category="mn",
  direction="nsm",
 },
 [0x110BA]={
  category="mn",
  direction="nsm",
 },
 [0x110BB]={
  category="po",
  direction="l",
 },
 [0x110BC]={
  category="po",
  direction="l",
 },
 [0x110BD]={
  category="cf",
  direction="l",
 },
 [0x110BE]={
  category="po",
  direction="l",
 },
 [0x110BF]={
  category="po",
  direction="l",
 },
 [0x110C0]={
  category="po",
  direction="l",
 },
 [0x110C1]={
  category="po",
  direction="l",
 },
 [0x110D0]={
  category="lo",
  direction="l",
 },
 [0x110D1]={
  category="lo",
  direction="l",
 },
 [0x110D2]={
  category="lo",
  direction="l",
 },
 [0x110D3]={
  category="lo",
  direction="l",
 },
 [0x110D4]={
  category="lo",
  direction="l",
 },
 [0x110D5]={
  category="lo",
  direction="l",
 },
 [0x110D6]={
  category="lo",
  direction="l",
 },
 [0x110D7]={
  category="lo",
  direction="l",
 },
 [0x110D8]={
  category="lo",
  direction="l",
 },
 [0x110D9]={
  category="lo",
  direction="l",
 },
 [0x110DA]={
  category="lo",
  direction="l",
 },
 [0x110DB]={
  category="lo",
  direction="l",
 },
 [0x110DC]={
  category="lo",
  direction="l",
 },
 [0x110DD]={
  category="lo",
  direction="l",
 },
 [0x110DE]={
  category="lo",
  direction="l",
 },
 [0x110DF]={
  category="lo",
  direction="l",
 },
 [0x110E0]={
  category="lo",
  direction="l",
 },
 [0x110E1]={
  category="lo",
  direction="l",
 },
 [0x110E2]={
  category="lo",
  direction="l",
 },
 [0x110E3]={
  category="lo",
  direction="l",
 },
 [0x110E4]={
  category="lo",
  direction="l",
 },
 [0x110E5]={
  category="lo",
  direction="l",
 },
 [0x110E6]={
  category="lo",
  direction="l",
 },
 [0x110E7]={
  category="lo",
  direction="l",
 },
 [0x110E8]={
  category="lo",
  direction="l",
 },
 [0x110F0]={
  category="nd",
  direction="l",
 },
 [0x110F1]={
  category="nd",
  direction="l",
 },
 [0x110F2]={
  category="nd",
  direction="l",
 },
 [0x110F3]={
  category="nd",
  direction="l",
 },
 [0x110F4]={
  category="nd",
  direction="l",
 },
 [0x110F5]={
  category="nd",
  direction="l",
 },
 [0x110F6]={
  category="nd",
  direction="l",
 },
 [0x110F7]={
  category="nd",
  direction="l",
 },
 [0x110F8]={
  category="nd",
  direction="l",
 },
 [0x110F9]={
  category="nd",
  direction="l",
 },
 [0x11100]={
  category="mn",
  direction="nsm",
 },
 [0x11101]={
  category="mn",
  direction="nsm",
 },
 [0x11102]={
  category="mn",
  direction="nsm",
 },
 [0x11103]={
  category="lo",
  direction="l",
 },
 [0x11104]={
  category="lo",
  direction="l",
 },
 [0x11105]={
  category="lo",
  direction="l",
 },
 [0x11106]={
  category="lo",
  direction="l",
 },
 [0x11107]={
  category="lo",
  direction="l",
 },
 [0x11108]={
  category="lo",
  direction="l",
 },
 [0x11109]={
  category="lo",
  direction="l",
 },
 [0x1110A]={
  category="lo",
  direction="l",
 },
 [0x1110B]={
  category="lo",
  direction="l",
 },
 [0x1110C]={
  category="lo",
  direction="l",
 },
 [0x1110D]={
  category="lo",
  direction="l",
 },
 [0x1110E]={
  category="lo",
  direction="l",
 },
 [0x1110F]={
  category="lo",
  direction="l",
 },
 [0x11110]={
  category="lo",
  direction="l",
 },
 [0x11111]={
  category="lo",
  direction="l",
 },
 [0x11112]={
  category="lo",
  direction="l",
 },
 [0x11113]={
  category="lo",
  direction="l",
 },
 [0x11114]={
  category="lo",
  direction="l",
 },
 [0x11115]={
  category="lo",
  direction="l",
 },
 [0x11116]={
  category="lo",
  direction="l",
 },
 [0x11117]={
  category="lo",
  direction="l",
 },
 [0x11118]={
  category="lo",
  direction="l",
 },
 [0x11119]={
  category="lo",
  direction="l",
 },
 [0x1111A]={
  category="lo",
  direction="l",
 },
 [0x1111B]={
  category="lo",
  direction="l",
 },
 [0x1111C]={
  category="lo",
  direction="l",
 },
 [0x1111D]={
  category="lo",
  direction="l",
 },
 [0x1111E]={
  category="lo",
  direction="l",
 },
 [0x1111F]={
  category="lo",
  direction="l",
 },
 [0x11120]={
  category="lo",
  direction="l",
 },
 [0x11121]={
  category="lo",
  direction="l",
 },
 [0x11122]={
  category="lo",
  direction="l",
 },
 [0x11123]={
  category="lo",
  direction="l",
 },
 [0x11124]={
  category="lo",
  direction="l",
 },
 [0x11125]={
  category="lo",
  direction="l",
 },
 [0x11126]={
  category="lo",
  direction="l",
 },
 [0x11127]={
  category="mn",
  direction="nsm",
 },
 [0x11128]={
  category="mn",
  direction="nsm",
 },
 [0x11129]={
  category="mn",
  direction="nsm",
 },
 [0x1112A]={
  category="mn",
  direction="nsm",
 },
 [0x1112B]={
  category="mn",
  direction="nsm",
 },
 [0x1112C]={
  category="mc",
  direction="l",
 },
 [0x1112D]={
  category="mn",
  direction="nsm",
 },
 [0x1112E]={
  category="mn",
  direction="nsm",
 },
 [0x1112F]={
  category="mn",
  direction="nsm",
 },
 [0x11130]={
  category="mn",
  direction="nsm",
 },
 [0x11131]={
  category="mn",
  direction="nsm",
 },
 [0x11132]={
  category="mn",
  direction="nsm",
 },
 [0x11133]={
  category="mn",
  direction="nsm",
 },
 [0x11134]={
  category="mn",
  direction="nsm",
 },
 [0x11136]={
  category="nd",
  direction="l",
 },
 [0x11137]={
  category="nd",
  direction="l",
 },
 [0x11138]={
  category="nd",
  direction="l",
 },
 [0x11139]={
  category="nd",
  direction="l",
 },
 [0x1113A]={
  category="nd",
  direction="l",
 },
 [0x1113B]={
  category="nd",
  direction="l",
 },
 [0x1113C]={
  category="nd",
  direction="l",
 },
 [0x1113D]={
  category="nd",
  direction="l",
 },
 [0x1113E]={
  category="nd",
  direction="l",
 },
 [0x1113F]={
  category="nd",
  direction="l",
 },
 [0x11140]={
  category="po",
  direction="l",
 },
 [0x11141]={
  category="po",
  direction="l",
 },
 [0x11142]={
  category="po",
  direction="l",
 },
 [0x11143]={
  category="po",
  direction="l",
 },
 [0x11150]={
  category="lo",
  direction="l",
 },
 [0x11151]={
  category="lo",
  direction="l",
 },
 [0x11152]={
  category="lo",
  direction="l",
 },
 [0x11153]={
  category="lo",
  direction="l",
 },
 [0x11154]={
  category="lo",
  direction="l",
 },
 [0x11155]={
  category="lo",
  direction="l",
 },
 [0x11156]={
  category="lo",
  direction="l",
 },
 [0x11157]={
  category="lo",
  direction="l",
 },
 [0x11158]={
  category="lo",
  direction="l",
 },
 [0x11159]={
  category="lo",
  direction="l",
 },
 [0x1115A]={
  category="lo",
  direction="l",
 },
 [0x1115B]={
  category="lo",
  direction="l",
 },
 [0x1115C]={
  category="lo",
  direction="l",
 },
 [0x1115D]={
  category="lo",
  direction="l",
 },
 [0x1115E]={
  category="lo",
  direction="l",
 },
 [0x1115F]={
  category="lo",
  direction="l",
 },
 [0x11160]={
  category="lo",
  direction="l",
 },
 [0x11161]={
  category="lo",
  direction="l",
 },
 [0x11162]={
  category="lo",
  direction="l",
 },
 [0x11163]={
  category="lo",
  direction="l",
 },
 [0x11164]={
  category="lo",
  direction="l",
 },
 [0x11165]={
  category="lo",
  direction="l",
 },
 [0x11166]={
  category="lo",
  direction="l",
 },
 [0x11167]={
  category="lo",
  direction="l",
 },
 [0x11168]={
  category="lo",
  direction="l",
 },
 [0x11169]={
  category="lo",
  direction="l",
 },
 [0x1116A]={
  category="lo",
  direction="l",
 },
 [0x1116B]={
  category="lo",
  direction="l",
 },
 [0x1116C]={
  category="lo",
  direction="l",
 },
 [0x1116D]={
  category="lo",
  direction="l",
 },
 [0x1116E]={
  category="lo",
  direction="l",
 },
 [0x1116F]={
  category="lo",
  direction="l",
 },
 [0x11170]={
  category="lo",
  direction="l",
 },
 [0x11171]={
  category="lo",
  direction="l",
 },
 [0x11172]={
  category="lo",
  direction="l",
 },
 [0x11173]={
  category="mn",
  direction="nsm",
 },
 [0x11174]={
  category="po",
  direction="l",
 },
 [0x11175]={
  category="po",
  direction="l",
 },
 [0x11176]={
  category="lo",
  direction="l",
 },
 [0x11180]={
  category="mn",
  direction="nsm",
 },
 [0x11181]={
  category="mn",
  direction="nsm",
 },
 [0x11182]={
  category="mc",
  direction="l",
 },
 [0x11183]={
  category="lo",
  direction="l",
 },
 [0x11184]={
  category="lo",
  direction="l",
 },
 [0x11185]={
  category="lo",
  direction="l",
 },
 [0x11186]={
  category="lo",
  direction="l",
 },
 [0x11187]={
  category="lo",
  direction="l",
 },
 [0x11188]={
  category="lo",
  direction="l",
 },
 [0x11189]={
  category="lo",
  direction="l",
 },
 [0x1118A]={
  category="lo",
  direction="l",
 },
 [0x1118B]={
  category="lo",
  direction="l",
 },
 [0x1118C]={
  category="lo",
  direction="l",
 },
 [0x1118D]={
  category="lo",
  direction="l",
 },
 [0x1118E]={
  category="lo",
  direction="l",
 },
 [0x1118F]={
  category="lo",
  direction="l",
 },
 [0x11190]={
  category="lo",
  direction="l",
 },
 [0x11191]={
  category="lo",
  direction="l",
 },
 [0x11192]={
  category="lo",
  direction="l",
 },
 [0x11193]={
  category="lo",
  direction="l",
 },
 [0x11194]={
  category="lo",
  direction="l",
 },
 [0x11195]={
  category="lo",
  direction="l",
 },
 [0x11196]={
  category="lo",
  direction="l",
 },
 [0x11197]={
  category="lo",
  direction="l",
 },
 [0x11198]={
  category="lo",
  direction="l",
 },
 [0x11199]={
  category="lo",
  direction="l",
 },
 [0x1119A]={
  category="lo",
  direction="l",
 },
 [0x1119B]={
  category="lo",
  direction="l",
 },
 [0x1119C]={
  category="lo",
  direction="l",
 },
 [0x1119D]={
  category="lo",
  direction="l",
 },
 [0x1119E]={
  category="lo",
  direction="l",
 },
 [0x1119F]={
  category="lo",
  direction="l",
 },
 [0x111A0]={
  category="lo",
  direction="l",
 },
 [0x111A1]={
  category="lo",
  direction="l",
 },
 [0x111A2]={
  category="lo",
  direction="l",
 },
 [0x111A3]={
  category="lo",
  direction="l",
 },
 [0x111A4]={
  category="lo",
  direction="l",
 },
 [0x111A5]={
  category="lo",
  direction="l",
 },
 [0x111A6]={
  category="lo",
  direction="l",
 },
 [0x111A7]={
  category="lo",
  direction="l",
 },
 [0x111A8]={
  category="lo",
  direction="l",
 },
 [0x111A9]={
  category="lo",
  direction="l",
 },
 [0x111AA]={
  category="lo",
  direction="l",
 },
 [0x111AB]={
  category="lo",
  direction="l",
 },
 [0x111AC]={
  category="lo",
  direction="l",
 },
 [0x111AD]={
  category="lo",
  direction="l",
 },
 [0x111AE]={
  category="lo",
  direction="l",
 },
 [0x111AF]={
  category="lo",
  direction="l",
 },
 [0x111B0]={
  category="lo",
  direction="l",
 },
 [0x111B1]={
  category="lo",
  direction="l",
 },
 [0x111B2]={
  category="lo",
  direction="l",
 },
 [0x111B3]={
  category="mc",
  direction="l",
 },
 [0x111B4]={
  category="mc",
  direction="l",
 },
 [0x111B5]={
  category="mc",
  direction="l",
 },
 [0x111B6]={
  category="mn",
  direction="nsm",
 },
 [0x111B7]={
  category="mn",
  direction="nsm",
 },
 [0x111B8]={
  category="mn",
  direction="nsm",
 },
 [0x111B9]={
  category="mn",
  direction="nsm",
 },
 [0x111BA]={
  category="mn",
  direction="nsm",
 },
 [0x111BB]={
  category="mn",
  direction="nsm",
 },
 [0x111BC]={
  category="mn",
  direction="nsm",
 },
 [0x111BD]={
  category="mn",
  direction="nsm",
 },
 [0x111BE]={
  category="mn",
  direction="nsm",
 },
 [0x111BF]={
  category="mc",
  direction="l",
 },
 [0x111C0]={
  category="mc",
  direction="l",
 },
 [0x111C1]={
  category="lo",
  direction="l",
 },
 [0x111C2]={
  category="lo",
  direction="l",
 },
 [0x111C3]={
  category="lo",
  direction="l",
 },
 [0x111C4]={
  category="lo",
  direction="l",
 },
 [0x111C5]={
  category="po",
  direction="l",
 },
 [0x111C6]={
  category="po",
  direction="l",
 },
 [0x111C7]={
  category="po",
  direction="l",
 },
 [0x111C8]={
  category="po",
  direction="l",
 },
 [0x111CD]={
  category="po",
  direction="l",
 },
 [0x111D0]={
  category="nd",
  direction="l",
 },
 [0x111D1]={
  category="nd",
  direction="l",
 },
 [0x111D2]={
  category="nd",
  direction="l",
 },
 [0x111D3]={
  category="nd",
  direction="l",
 },
 [0x111D4]={
  category="nd",
  direction="l",
 },
 [0x111D5]={
  category="nd",
  direction="l",
 },
 [0x111D6]={
  category="nd",
  direction="l",
 },
 [0x111D7]={
  category="nd",
  direction="l",
 },
 [0x111D8]={
  category="nd",
  direction="l",
 },
 [0x111D9]={
  category="nd",
  direction="l",
 },
 [0x111DA]={
  category="lo",
  direction="l",
 },
 [0x111E1]={
  category="no",
  direction="l",
 },
 [0x111E2]={
  category="no",
  direction="l",
 },
 [0x111E3]={
  category="no",
  direction="l",
 },
 [0x111E4]={
  category="no",
  direction="l",
 },
 [0x111E5]={
  category="no",
  direction="l",
 },
 [0x111E6]={
  category="no",
  direction="l",
 },
 [0x111E7]={
  category="no",
  direction="l",
 },
 [0x111E8]={
  category="no",
  direction="l",
 },
 [0x111E9]={
  category="no",
  direction="l",
 },
 [0x111EA]={
  category="no",
  direction="l",
 },
 [0x111EB]={
  category="no",
  direction="l",
 },
 [0x111EC]={
  category="no",
  direction="l",
 },
 [0x111ED]={
  category="no",
  direction="l",
 },
 [0x111EE]={
  category="no",
  direction="l",
 },
 [0x111EF]={
  category="no",
  direction="l",
 },
 [0x111F0]={
  category="no",
  direction="l",
 },
 [0x111F1]={
  category="no",
  direction="l",
 },
 [0x111F2]={
  category="no",
  direction="l",
 },
 [0x111F3]={
  category="no",
  direction="l",
 },
 [0x111F4]={
  category="no",
  direction="l",
 },
 [0x11200]={
  category="lo",
  direction="l",
 },
 [0x11201]={
  category="lo",
  direction="l",
 },
 [0x11202]={
  category="lo",
  direction="l",
 },
 [0x11203]={
  category="lo",
  direction="l",
 },
 [0x11204]={
  category="lo",
  direction="l",
 },
 [0x11205]={
  category="lo",
  direction="l",
 },
 [0x11206]={
  category="lo",
  direction="l",
 },
 [0x11207]={
  category="lo",
  direction="l",
 },
 [0x11208]={
  category="lo",
  direction="l",
 },
 [0x11209]={
  category="lo",
  direction="l",
 },
 [0x1120A]={
  category="lo",
  direction="l",
 },
 [0x1120B]={
  category="lo",
  direction="l",
 },
 [0x1120C]={
  category="lo",
  direction="l",
 },
 [0x1120D]={
  category="lo",
  direction="l",
 },
 [0x1120E]={
  category="lo",
  direction="l",
 },
 [0x1120F]={
  category="lo",
  direction="l",
 },
 [0x11210]={
  category="lo",
  direction="l",
 },
 [0x11211]={
  category="lo",
  direction="l",
 },
 [0x11213]={
  category="lo",
  direction="l",
 },
 [0x11214]={
  category="lo",
  direction="l",
 },
 [0x11215]={
  category="lo",
  direction="l",
 },
 [0x11216]={
  category="lo",
  direction="l",
 },
 [0x11217]={
  category="lo",
  direction="l",
 },
 [0x11218]={
  category="lo",
  direction="l",
 },
 [0x11219]={
  category="lo",
  direction="l",
 },
 [0x1121A]={
  category="lo",
  direction="l",
 },
 [0x1121B]={
  category="lo",
  direction="l",
 },
 [0x1121C]={
  category="lo",
  direction="l",
 },
 [0x1121D]={
  category="lo",
  direction="l",
 },
 [0x1121E]={
  category="lo",
  direction="l",
 },
 [0x1121F]={
  category="lo",
  direction="l",
 },
 [0x11220]={
  category="lo",
  direction="l",
 },
 [0x11221]={
  category="lo",
  direction="l",
 },
 [0x11222]={
  category="lo",
  direction="l",
 },
 [0x11223]={
  category="lo",
  direction="l",
 },
 [0x11224]={
  category="lo",
  direction="l",
 },
 [0x11225]={
  category="lo",
  direction="l",
 },
 [0x11226]={
  category="lo",
  direction="l",
 },
 [0x11227]={
  category="lo",
  direction="l",
 },
 [0x11228]={
  category="lo",
  direction="l",
 },
 [0x11229]={
  category="lo",
  direction="l",
 },
 [0x1122A]={
  category="lo",
  direction="l",
 },
 [0x1122B]={
  category="lo",
  direction="l",
 },
 [0x1122C]={
  category="mc",
  direction="l",
 },
 [0x1122D]={
  category="mc",
  direction="l",
 },
 [0x1122E]={
  category="mc",
  direction="l",
 },
 [0x1122F]={
  category="mn",
  direction="nsm",
 },
 [0x11230]={
  category="mn",
  direction="nsm",
 },
 [0x11231]={
  category="mn",
  direction="nsm",
 },
 [0x11232]={
  category="mc",
  direction="l",
 },
 [0x11233]={
  category="mc",
  direction="l",
 },
 [0x11234]={
  category="mn",
  direction="nsm",
 },
 [0x11235]={
  category="mc",
  direction="l",
 },
 [0x11236]={
  category="mn",
  direction="nsm",
 },
 [0x11237]={
  category="mn",
  direction="nsm",
 },
 [0x11238]={
  category="po",
  direction="l",
 },
 [0x11239]={
  category="po",
  direction="l",
 },
 [0x1123A]={
  category="po",
  direction="l",
 },
 [0x1123B]={
  category="po",
  direction="l",
 },
 [0x1123C]={
  category="po",
  direction="l",
 },
 [0x1123D]={
  category="po",
  direction="l",
 },
 [0x112B0]={
  category="lo",
  direction="l",
 },
 [0x112B1]={
  category="lo",
  direction="l",
 },
 [0x112B2]={
  category="lo",
  direction="l",
 },
 [0x112B3]={
  category="lo",
  direction="l",
 },
 [0x112B4]={
  category="lo",
  direction="l",
 },
 [0x112B5]={
  category="lo",
  direction="l",
 },
 [0x112B6]={
  category="lo",
  direction="l",
 },
 [0x112B7]={
  category="lo",
  direction="l",
 },
 [0x112B8]={
  category="lo",
  direction="l",
 },
 [0x112B9]={
  category="lo",
  direction="l",
 },
 [0x112BA]={
  category="lo",
  direction="l",
 },
 [0x112BB]={
  category="lo",
  direction="l",
 },
 [0x112BC]={
  category="lo",
  direction="l",
 },
 [0x112BD]={
  category="lo",
  direction="l",
 },
 [0x112BE]={
  category="lo",
  direction="l",
 },
 [0x112BF]={
  category="lo",
  direction="l",
 },
 [0x112C0]={
  category="lo",
  direction="l",
 },
 [0x112C1]={
  category="lo",
  direction="l",
 },
 [0x112C2]={
  category="lo",
  direction="l",
 },
 [0x112C3]={
  category="lo",
  direction="l",
 },
 [0x112C4]={
  category="lo",
  direction="l",
 },
 [0x112C5]={
  category="lo",
  direction="l",
 },
 [0x112C6]={
  category="lo",
  direction="l",
 },
 [0x112C7]={
  category="lo",
  direction="l",
 },
 [0x112C8]={
  category="lo",
  direction="l",
 },
 [0x112C9]={
  category="lo",
  direction="l",
 },
 [0x112CA]={
  category="lo",
  direction="l",
 },
 [0x112CB]={
  category="lo",
  direction="l",
 },
 [0x112CC]={
  category="lo",
  direction="l",
 },
 [0x112CD]={
  category="lo",
  direction="l",
 },
 [0x112CE]={
  category="lo",
  direction="l",
 },
 [0x112CF]={
  category="lo",
  direction="l",
 },
 [0x112D0]={
  category="lo",
  direction="l",
 },
 [0x112D1]={
  category="lo",
  direction="l",
 },
 [0x112D2]={
  category="lo",
  direction="l",
 },
 [0x112D3]={
  category="lo",
  direction="l",
 },
 [0x112D4]={
  category="lo",
  direction="l",
 },
 [0x112D5]={
  category="lo",
  direction="l",
 },
 [0x112D6]={
  category="lo",
  direction="l",
 },
 [0x112D7]={
  category="lo",
  direction="l",
 },
 [0x112D8]={
  category="lo",
  direction="l",
 },
 [0x112D9]={
  category="lo",
  direction="l",
 },
 [0x112DA]={
  category="lo",
  direction="l",
 },
 [0x112DB]={
  category="lo",
  direction="l",
 },
 [0x112DC]={
  category="lo",
  direction="l",
 },
 [0x112DD]={
  category="lo",
  direction="l",
 },
 [0x112DE]={
  category="lo",
  direction="l",
 },
 [0x112DF]={
  category="mn",
  direction="nsm",
 },
 [0x112E0]={
  category="mc",
  direction="l",
 },
 [0x112E1]={
  category="mc",
  direction="l",
 },
 [0x112E2]={
  category="mc",
  direction="l",
 },
 [0x112E3]={
  category="mn",
  direction="nsm",
 },
 [0x112E4]={
  category="mn",
  direction="nsm",
 },
 [0x112E5]={
  category="mn",
  direction="nsm",
 },
 [0x112E6]={
  category="mn",
  direction="nsm",
 },
 [0x112E7]={
  category="mn",
  direction="nsm",
 },
 [0x112E8]={
  category="mn",
  direction="nsm",
 },
 [0x112E9]={
  category="mn",
  direction="nsm",
 },
 [0x112EA]={
  category="mn",
  direction="nsm",
 },
 [0x112F0]={
  category="nd",
  direction="l",
 },
 [0x112F1]={
  category="nd",
  direction="l",
 },
 [0x112F2]={
  category="nd",
  direction="l",
 },
 [0x112F3]={
  category="nd",
  direction="l",
 },
 [0x112F4]={
  category="nd",
  direction="l",
 },
 [0x112F5]={
  category="nd",
  direction="l",
 },
 [0x112F6]={
  category="nd",
  direction="l",
 },
 [0x112F7]={
  category="nd",
  direction="l",
 },
 [0x112F8]={
  category="nd",
  direction="l",
 },
 [0x112F9]={
  category="nd",
  direction="l",
 },
 [0x11301]={
  category="mn",
  direction="nsm",
 },
 [0x11302]={
  category="mc",
  direction="l",
 },
 [0x11303]={
  category="mc",
  direction="l",
 },
 [0x11305]={
  category="lo",
  direction="l",
 },
 [0x11306]={
  category="lo",
  direction="l",
 },
 [0x11307]={
  category="lo",
  direction="l",
 },
 [0x11308]={
  category="lo",
  direction="l",
 },
 [0x11309]={
  category="lo",
  direction="l",
 },
 [0x1130A]={
  category="lo",
  direction="l",
 },
 [0x1130B]={
  category="lo",
  direction="l",
 },
 [0x1130C]={
  category="lo",
  direction="l",
 },
 [0x1130F]={
  category="lo",
  direction="l",
 },
 [0x11310]={
  category="lo",
  direction="l",
 },
 [0x11313]={
  category="lo",
  direction="l",
 },
 [0x11314]={
  category="lo",
  direction="l",
 },
 [0x11315]={
  category="lo",
  direction="l",
 },
 [0x11316]={
  category="lo",
  direction="l",
 },
 [0x11317]={
  category="lo",
  direction="l",
 },
 [0x11318]={
  category="lo",
  direction="l",
 },
 [0x11319]={
  category="lo",
  direction="l",
 },
 [0x1131A]={
  category="lo",
  direction="l",
 },
 [0x1131B]={
  category="lo",
  direction="l",
 },
 [0x1131C]={
  category="lo",
  direction="l",
 },
 [0x1131D]={
  category="lo",
  direction="l",
 },
 [0x1131E]={
  category="lo",
  direction="l",
 },
 [0x1131F]={
  category="lo",
  direction="l",
 },
 [0x11320]={
  category="lo",
  direction="l",
 },
 [0x11321]={
  category="lo",
  direction="l",
 },
 [0x11322]={
  category="lo",
  direction="l",
 },
 [0x11323]={
  category="lo",
  direction="l",
 },
 [0x11324]={
  category="lo",
  direction="l",
 },
 [0x11325]={
  category="lo",
  direction="l",
 },
 [0x11326]={
  category="lo",
  direction="l",
 },
 [0x11327]={
  category="lo",
  direction="l",
 },
 [0x11328]={
  category="lo",
  direction="l",
 },
 [0x1132A]={
  category="lo",
  direction="l",
 },
 [0x1132B]={
  category="lo",
  direction="l",
 },
 [0x1132C]={
  category="lo",
  direction="l",
 },
 [0x1132D]={
  category="lo",
  direction="l",
 },
 [0x1132E]={
  category="lo",
  direction="l",
 },
 [0x1132F]={
  category="lo",
  direction="l",
 },
 [0x11330]={
  category="lo",
  direction="l",
 },
 [0x11332]={
  category="lo",
  direction="l",
 },
 [0x11333]={
  category="lo",
  direction="l",
 },
 [0x11335]={
  category="lo",
  direction="l",
 },
 [0x11336]={
  category="lo",
  direction="l",
 },
 [0x11337]={
  category="lo",
  direction="l",
 },
 [0x11338]={
  category="lo",
  direction="l",
 },
 [0x11339]={
  category="lo",
  direction="l",
 },
 [0x1133C]={
  category="mn",
  direction="nsm",
 },
 [0x1133D]={
  category="lo",
  direction="l",
 },
 [0x1133E]={
  category="mc",
  direction="l",
 },
 [0x1133F]={
  category="mc",
  direction="l",
 },
 [0x11340]={
  category="mn",
  direction="nsm",
 },
 [0x11341]={
  category="mc",
  direction="l",
 },
 [0x11342]={
  category="mc",
  direction="l",
 },
 [0x11343]={
  category="mc",
  direction="l",
 },
 [0x11344]={
  category="mc",
  direction="l",
 },
 [0x11347]={
  category="mc",
  direction="l",
 },
 [0x11348]={
  category="mc",
  direction="l",
 },
 [0x1134B]={
  category="mc",
  direction="l",
 },
 [0x1134C]={
  category="mc",
  direction="l",
 },
 [0x1134D]={
  category="mc",
  direction="l",
 },
 [0x11357]={
  category="mc",
  direction="l",
 },
 [0x1135D]={
  category="lo",
  direction="l",
 },
 [0x1135E]={
  category="lo",
  direction="l",
 },
 [0x1135F]={
  category="lo",
  direction="l",
 },
 [0x11360]={
  category="lo",
  direction="l",
 },
 [0x11361]={
  category="lo",
  direction="l",
 },
 [0x11362]={
  category="mc",
  direction="l",
 },
 [0x11363]={
  category="mc",
  direction="l",
 },
 [0x11366]={
  category="mn",
  direction="nsm",
 },
 [0x11367]={
  category="mn",
  direction="nsm",
 },
 [0x11368]={
  category="mn",
  direction="nsm",
 },
 [0x11369]={
  category="mn",
  direction="nsm",
 },
 [0x1136A]={
  category="mn",
  direction="nsm",
 },
 [0x1136B]={
  category="mn",
  direction="nsm",
 },
 [0x1136C]={
  category="mn",
  direction="nsm",
 },
 [0x11370]={
  category="mn",
  direction="nsm",
 },
 [0x11371]={
  category="mn",
  direction="nsm",
 },
 [0x11372]={
  category="mn",
  direction="nsm",
 },
 [0x11373]={
  category="mn",
  direction="nsm",
 },
 [0x11374]={
  category="mn",
  direction="nsm",
 },
 [0x11480]={
  category="lo",
  direction="l",
 },
 [0x11481]={
  category="lo",
  direction="l",
 },
 [0x11482]={
  category="lo",
  direction="l",
 },
 [0x11483]={
  category="lo",
  direction="l",
 },
 [0x11484]={
  category="lo",
  direction="l",
 },
 [0x11485]={
  category="lo",
  direction="l",
 },
 [0x11486]={
  category="lo",
  direction="l",
 },
 [0x11487]={
  category="lo",
  direction="l",
 },
 [0x11488]={
  category="lo",
  direction="l",
 },
 [0x11489]={
  category="lo",
  direction="l",
 },
 [0x1148A]={
  category="lo",
  direction="l",
 },
 [0x1148B]={
  category="lo",
  direction="l",
 },
 [0x1148C]={
  category="lo",
  direction="l",
 },
 [0x1148D]={
  category="lo",
  direction="l",
 },
 [0x1148E]={
  category="lo",
  direction="l",
 },
 [0x1148F]={
  category="lo",
  direction="l",
 },
 [0x11490]={
  category="lo",
  direction="l",
 },
 [0x11491]={
  category="lo",
  direction="l",
 },
 [0x11492]={
  category="lo",
  direction="l",
 },
 [0x11493]={
  category="lo",
  direction="l",
 },
 [0x11494]={
  category="lo",
  direction="l",
 },
 [0x11495]={
  category="lo",
  direction="l",
 },
 [0x11496]={
  category="lo",
  direction="l",
 },
 [0x11497]={
  category="lo",
  direction="l",
 },
 [0x11498]={
  category="lo",
  direction="l",
 },
 [0x11499]={
  category="lo",
  direction="l",
 },
 [0x1149A]={
  category="lo",
  direction="l",
 },
 [0x1149B]={
  category="lo",
  direction="l",
 },
 [0x1149C]={
  category="lo",
  direction="l",
 },
 [0x1149D]={
  category="lo",
  direction="l",
 },
 [0x1149E]={
  category="lo",
  direction="l",
 },
 [0x1149F]={
  category="lo",
  direction="l",
 },
 [0x114A0]={
  category="lo",
  direction="l",
 },
 [0x114A1]={
  category="lo",
  direction="l",
 },
 [0x114A2]={
  category="lo",
  direction="l",
 },
 [0x114A3]={
  category="lo",
  direction="l",
 },
 [0x114A4]={
  category="lo",
  direction="l",
 },
 [0x114A5]={
  category="lo",
  direction="l",
 },
 [0x114A6]={
  category="lo",
  direction="l",
 },
 [0x114A7]={
  category="lo",
  direction="l",
 },
 [0x114A8]={
  category="lo",
  direction="l",
 },
 [0x114A9]={
  category="lo",
  direction="l",
 },
 [0x114AA]={
  category="lo",
  direction="l",
 },
 [0x114AB]={
  category="lo",
  direction="l",
 },
 [0x114AC]={
  category="lo",
  direction="l",
 },
 [0x114AD]={
  category="lo",
  direction="l",
 },
 [0x114AE]={
  category="lo",
  direction="l",
 },
 [0x114AF]={
  category="lo",
  direction="l",
 },
 [0x114B0]={
  category="mc",
  direction="l",
 },
 [0x114B1]={
  category="mc",
  direction="l",
 },
 [0x114B2]={
  category="mc",
  direction="l",
 },
 [0x114B3]={
  category="mn",
  direction="nsm",
 },
 [0x114B4]={
  category="mn",
  direction="nsm",
 },
 [0x114B5]={
  category="mn",
  direction="nsm",
 },
 [0x114B6]={
  category="mn",
  direction="nsm",
 },
 [0x114B7]={
  category="mn",
  direction="nsm",
 },
 [0x114B8]={
  category="mn",
  direction="nsm",
 },
 [0x114B9]={
  category="mc",
  direction="l",
 },
 [0x114BA]={
  category="mn",
  direction="nsm",
 },
 [0x114BB]={
  category="mc",
  direction="l",
 },
 [0x114BC]={
  category="mc",
  direction="l",
 },
 [0x114BD]={
  category="mc",
  direction="l",
 },
 [0x114BE]={
  category="mc",
  direction="l",
 },
 [0x114BF]={
  category="mn",
  direction="nsm",
 },
 [0x114C0]={
  category="mn",
  direction="nsm",
 },
 [0x114C1]={
  category="mc",
  direction="l",
 },
 [0x114C2]={
  category="mn",
  direction="nsm",
 },
 [0x114C3]={
  category="mn",
  direction="nsm",
 },
 [0x114C4]={
  category="lo",
  direction="l",
 },
 [0x114C5]={
  category="lo",
  direction="l",
 },
 [0x114C6]={
  category="po",
  direction="l",
 },
 [0x114C7]={
  category="lo",
  direction="l",
 },
 [0x114D0]={
  category="nd",
  direction="l",
 },
 [0x114D1]={
  category="nd",
  direction="l",
 },
 [0x114D2]={
  category="nd",
  direction="l",
 },
 [0x114D3]={
  category="nd",
  direction="l",
 },
 [0x114D4]={
  category="nd",
  direction="l",
 },
 [0x114D5]={
  category="nd",
  direction="l",
 },
 [0x114D6]={
  category="nd",
  direction="l",
 },
 [0x114D7]={
  category="nd",
  direction="l",
 },
 [0x114D8]={
  category="nd",
  direction="l",
 },
 [0x114D9]={
  category="nd",
  direction="l",
 },
 [0x11580]={
  category="lo",
  direction="l",
 },
 [0x11581]={
  category="lo",
  direction="l",
 },
 [0x11582]={
  category="lo",
  direction="l",
 },
 [0x11583]={
  category="lo",
  direction="l",
 },
 [0x11584]={
  category="lo",
  direction="l",
 },
 [0x11585]={
  category="lo",
  direction="l",
 },
 [0x11586]={
  category="lo",
  direction="l",
 },
 [0x11587]={
  category="lo",
  direction="l",
 },
 [0x11588]={
  category="lo",
  direction="l",
 },
 [0x11589]={
  category="lo",
  direction="l",
 },
 [0x1158A]={
  category="lo",
  direction="l",
 },
 [0x1158B]={
  category="lo",
  direction="l",
 },
 [0x1158C]={
  category="lo",
  direction="l",
 },
 [0x1158D]={
  category="lo",
  direction="l",
 },
 [0x1158E]={
  category="lo",
  direction="l",
 },
 [0x1158F]={
  category="lo",
  direction="l",
 },
 [0x11590]={
  category="lo",
  direction="l",
 },
 [0x11591]={
  category="lo",
  direction="l",
 },
 [0x11592]={
  category="lo",
  direction="l",
 },
 [0x11593]={
  category="lo",
  direction="l",
 },
 [0x11594]={
  category="lo",
  direction="l",
 },
 [0x11595]={
  category="lo",
  direction="l",
 },
 [0x11596]={
  category="lo",
  direction="l",
 },
 [0x11597]={
  category="lo",
  direction="l",
 },
 [0x11598]={
  category="lo",
  direction="l",
 },
 [0x11599]={
  category="lo",
  direction="l",
 },
 [0x1159A]={
  category="lo",
  direction="l",
 },
 [0x1159B]={
  category="lo",
  direction="l",
 },
 [0x1159C]={
  category="lo",
  direction="l",
 },
 [0x1159D]={
  category="lo",
  direction="l",
 },
 [0x1159E]={
  category="lo",
  direction="l",
 },
 [0x1159F]={
  category="lo",
  direction="l",
 },
 [0x115A0]={
  category="lo",
  direction="l",
 },
 [0x115A1]={
  category="lo",
  direction="l",
 },
 [0x115A2]={
  category="lo",
  direction="l",
 },
 [0x115A3]={
  category="lo",
  direction="l",
 },
 [0x115A4]={
  category="lo",
  direction="l",
 },
 [0x115A5]={
  category="lo",
  direction="l",
 },
 [0x115A6]={
  category="lo",
  direction="l",
 },
 [0x115A7]={
  category="lo",
  direction="l",
 },
 [0x115A8]={
  category="lo",
  direction="l",
 },
 [0x115A9]={
  category="lo",
  direction="l",
 },
 [0x115AA]={
  category="lo",
  direction="l",
 },
 [0x115AB]={
  category="lo",
  direction="l",
 },
 [0x115AC]={
  category="lo",
  direction="l",
 },
 [0x115AD]={
  category="lo",
  direction="l",
 },
 [0x115AE]={
  category="lo",
  direction="l",
 },
 [0x115AF]={
  category="mc",
  direction="l",
 },
 [0x115B0]={
  category="mc",
  direction="l",
 },
 [0x115B1]={
  category="mc",
  direction="l",
 },
 [0x115B2]={
  category="mn",
  direction="nsm",
 },
 [0x115B3]={
  category="mn",
  direction="nsm",
 },
 [0x115B4]={
  category="mn",
  direction="nsm",
 },
 [0x115B5]={
  category="mn",
  direction="nsm",
 },
 [0x115B8]={
  category="mc",
  direction="l",
 },
 [0x115B9]={
  category="mc",
  direction="l",
 },
 [0x115BA]={
  category="mc",
  direction="l",
 },
 [0x115BB]={
  category="mc",
  direction="l",
 },
 [0x115BC]={
  category="mn",
  direction="nsm",
 },
 [0x115BD]={
  category="mn",
  direction="nsm",
 },
 [0x115BE]={
  category="mc",
  direction="l",
 },
 [0x115BF]={
  category="mn",
  direction="nsm",
 },
 [0x115C0]={
  category="mn",
  direction="nsm",
 },
 [0x115C1]={
  category="po",
  direction="l",
 },
 [0x115C2]={
  category="po",
  direction="l",
 },
 [0x115C3]={
  category="po",
  direction="l",
 },
 [0x115C4]={
  category="po",
  direction="l",
 },
 [0x115C5]={
  category="po",
  direction="l",
 },
 [0x115C6]={
  category="po",
  direction="l",
 },
 [0x115C7]={
  category="po",
  direction="l",
 },
 [0x115C8]={
  category="po",
  direction="l",
 },
 [0x115C9]={
  category="po",
  direction="l",
 },
 [0x11600]={
  category="lo",
  direction="l",
 },
 [0x11601]={
  category="lo",
  direction="l",
 },
 [0x11602]={
  category="lo",
  direction="l",
 },
 [0x11603]={
  category="lo",
  direction="l",
 },
 [0x11604]={
  category="lo",
  direction="l",
 },
 [0x11605]={
  category="lo",
  direction="l",
 },
 [0x11606]={
  category="lo",
  direction="l",
 },
 [0x11607]={
  category="lo",
  direction="l",
 },
 [0x11608]={
  category="lo",
  direction="l",
 },
 [0x11609]={
  category="lo",
  direction="l",
 },
 [0x1160A]={
  category="lo",
  direction="l",
 },
 [0x1160B]={
  category="lo",
  direction="l",
 },
 [0x1160C]={
  category="lo",
  direction="l",
 },
 [0x1160D]={
  category="lo",
  direction="l",
 },
 [0x1160E]={
  category="lo",
  direction="l",
 },
 [0x1160F]={
  category="lo",
  direction="l",
 },
 [0x11610]={
  category="lo",
  direction="l",
 },
 [0x11611]={
  category="lo",
  direction="l",
 },
 [0x11612]={
  category="lo",
  direction="l",
 },
 [0x11613]={
  category="lo",
  direction="l",
 },
 [0x11614]={
  category="lo",
  direction="l",
 },
 [0x11615]={
  category="lo",
  direction="l",
 },
 [0x11616]={
  category="lo",
  direction="l",
 },
 [0x11617]={
  category="lo",
  direction="l",
 },
 [0x11618]={
  category="lo",
  direction="l",
 },
 [0x11619]={
  category="lo",
  direction="l",
 },
 [0x1161A]={
  category="lo",
  direction="l",
 },
 [0x1161B]={
  category="lo",
  direction="l",
 },
 [0x1161C]={
  category="lo",
  direction="l",
 },
 [0x1161D]={
  category="lo",
  direction="l",
 },
 [0x1161E]={
  category="lo",
  direction="l",
 },
 [0x1161F]={
  category="lo",
  direction="l",
 },
 [0x11620]={
  category="lo",
  direction="l",
 },
 [0x11621]={
  category="lo",
  direction="l",
 },
 [0x11622]={
  category="lo",
  direction="l",
 },
 [0x11623]={
  category="lo",
  direction="l",
 },
 [0x11624]={
  category="lo",
  direction="l",
 },
 [0x11625]={
  category="lo",
  direction="l",
 },
 [0x11626]={
  category="lo",
  direction="l",
 },
 [0x11627]={
  category="lo",
  direction="l",
 },
 [0x11628]={
  category="lo",
  direction="l",
 },
 [0x11629]={
  category="lo",
  direction="l",
 },
 [0x1162A]={
  category="lo",
  direction="l",
 },
 [0x1162B]={
  category="lo",
  direction="l",
 },
 [0x1162C]={
  category="lo",
  direction="l",
 },
 [0x1162D]={
  category="lo",
  direction="l",
 },
 [0x1162E]={
  category="lo",
  direction="l",
 },
 [0x1162F]={
  category="lo",
  direction="l",
 },
 [0x11630]={
  category="mc",
  direction="l",
 },
 [0x11631]={
  category="mc",
  direction="l",
 },
 [0x11632]={
  category="mc",
  direction="l",
 },
 [0x11633]={
  category="mn",
  direction="nsm",
 },
 [0x11634]={
  category="mn",
  direction="nsm",
 },
 [0x11635]={
  category="mn",
  direction="nsm",
 },
 [0x11636]={
  category="mn",
  direction="nsm",
 },
 [0x11637]={
  category="mn",
  direction="nsm",
 },
 [0x11638]={
  category="mn",
  direction="nsm",
 },
 [0x11639]={
  category="mn",
  direction="nsm",
 },
 [0x1163A]={
  category="mn",
  direction="nsm",
 },
 [0x1163B]={
  category="mc",
  direction="l",
 },
 [0x1163C]={
  category="mc",
  direction="l",
 },
 [0x1163D]={
  category="mn",
  direction="nsm",
 },
 [0x1163E]={
  category="mc",
  direction="l",
 },
 [0x1163F]={
  category="mn",
  direction="nsm",
 },
 [0x11640]={
  category="mn",
  direction="nsm",
 },
 [0x11641]={
  category="po",
  direction="l",
 },
 [0x11642]={
  category="po",
  direction="l",
 },
 [0x11643]={
  category="po",
  direction="l",
 },
 [0x11644]={
  category="lo",
  direction="l",
 },
 [0x11650]={
  category="nd",
  direction="l",
 },
 [0x11651]={
  category="nd",
  direction="l",
 },
 [0x11652]={
  category="nd",
  direction="l",
 },
 [0x11653]={
  category="nd",
  direction="l",
 },
 [0x11654]={
  category="nd",
  direction="l",
 },
 [0x11655]={
  category="nd",
  direction="l",
 },
 [0x11656]={
  category="nd",
  direction="l",
 },
 [0x11657]={
  category="nd",
  direction="l",
 },
 [0x11658]={
  category="nd",
  direction="l",
 },
 [0x11659]={
  category="nd",
  direction="l",
 },
 [0x11680]={
  category="lo",
  direction="l",
 },
 [0x11681]={
  category="lo",
  direction="l",
 },
 [0x11682]={
  category="lo",
  direction="l",
 },
 [0x11683]={
  category="lo",
  direction="l",
 },
 [0x11684]={
  category="lo",
  direction="l",
 },
 [0x11685]={
  category="lo",
  direction="l",
 },
 [0x11686]={
  category="lo",
  direction="l",
 },
 [0x11687]={
  category="lo",
  direction="l",
 },
 [0x11688]={
  category="lo",
  direction="l",
 },
 [0x11689]={
  category="lo",
  direction="l",
 },
 [0x1168A]={
  category="lo",
  direction="l",
 },
 [0x1168B]={
  category="lo",
  direction="l",
 },
 [0x1168C]={
  category="lo",
  direction="l",
 },
 [0x1168D]={
  category="lo",
  direction="l",
 },
 [0x1168E]={
  category="lo",
  direction="l",
 },
 [0x1168F]={
  category="lo",
  direction="l",
 },
 [0x11690]={
  category="lo",
  direction="l",
 },
 [0x11691]={
  category="lo",
  direction="l",
 },
 [0x11692]={
  category="lo",
  direction="l",
 },
 [0x11693]={
  category="lo",
  direction="l",
 },
 [0x11694]={
  category="lo",
  direction="l",
 },
 [0x11695]={
  category="lo",
  direction="l",
 },
 [0x11696]={
  category="lo",
  direction="l",
 },
 [0x11697]={
  category="lo",
  direction="l",
 },
 [0x11698]={
  category="lo",
  direction="l",
 },
 [0x11699]={
  category="lo",
  direction="l",
 },
 [0x1169A]={
  category="lo",
  direction="l",
 },
 [0x1169B]={
  category="lo",
  direction="l",
 },
 [0x1169C]={
  category="lo",
  direction="l",
 },
 [0x1169D]={
  category="lo",
  direction="l",
 },
 [0x1169E]={
  category="lo",
  direction="l",
 },
 [0x1169F]={
  category="lo",
  direction="l",
 },
 [0x116A0]={
  category="lo",
  direction="l",
 },
 [0x116A1]={
  category="lo",
  direction="l",
 },
 [0x116A2]={
  category="lo",
  direction="l",
 },
 [0x116A3]={
  category="lo",
  direction="l",
 },
 [0x116A4]={
  category="lo",
  direction="l",
 },
 [0x116A5]={
  category="lo",
  direction="l",
 },
 [0x116A6]={
  category="lo",
  direction="l",
 },
 [0x116A7]={
  category="lo",
  direction="l",
 },
 [0x116A8]={
  category="lo",
  direction="l",
 },
 [0x116A9]={
  category="lo",
  direction="l",
 },
 [0x116AA]={
  category="lo",
  direction="l",
 },
 [0x116AB]={
  category="mn",
  direction="nsm",
 },
 [0x116AC]={
  category="mc",
  direction="l",
 },
 [0x116AD]={
  category="mn",
  direction="nsm",
 },
 [0x116AE]={
  category="mc",
  direction="l",
 },
 [0x116AF]={
  category="mc",
  direction="l",
 },
 [0x116B0]={
  category="mn",
  direction="nsm",
 },
 [0x116B1]={
  category="mn",
  direction="nsm",
 },
 [0x116B2]={
  category="mn",
  direction="nsm",
 },
 [0x116B3]={
  category="mn",
  direction="nsm",
 },
 [0x116B4]={
  category="mn",
  direction="nsm",
 },
 [0x116B5]={
  category="mn",
  direction="nsm",
 },
 [0x116B6]={
  category="mc",
  direction="l",
 },
 [0x116B7]={
  category="mn",
  direction="nsm",
 },
 [0x116C0]={
  category="nd",
  direction="l",
 },
 [0x116C1]={
  category="nd",
  direction="l",
 },
 [0x116C2]={
  category="nd",
  direction="l",
 },
 [0x116C3]={
  category="nd",
  direction="l",
 },
 [0x116C4]={
  category="nd",
  direction="l",
 },
 [0x116C5]={
  category="nd",
  direction="l",
 },
 [0x116C6]={
  category="nd",
  direction="l",
 },
 [0x116C7]={
  category="nd",
  direction="l",
 },
 [0x116C8]={
  category="nd",
  direction="l",
 },
 [0x116C9]={
  category="nd",
  direction="l",
 },
 [0x118A0]={
  category="lu",
  direction="l",
 },
 [0x118A1]={
  category="lu",
  direction="l",
 },
 [0x118A2]={
  category="lu",
  direction="l",
 },
 [0x118A3]={
  category="lu",
  direction="l",
 },
 [0x118A4]={
  category="lu",
  direction="l",
 },
 [0x118A5]={
  category="lu",
  direction="l",
 },
 [0x118A6]={
  category="lu",
  direction="l",
 },
 [0x118A7]={
  category="lu",
  direction="l",
 },
 [0x118A8]={
  category="lu",
  direction="l",
 },
 [0x118A9]={
  category="lu",
  direction="l",
 },
 [0x118AA]={
  category="lu",
  direction="l",
 },
 [0x118AB]={
  category="lu",
  direction="l",
 },
 [0x118AC]={
  category="lu",
  direction="l",
 },
 [0x118AD]={
  category="lu",
  direction="l",
 },
 [0x118AE]={
  category="lu",
  direction="l",
 },
 [0x118AF]={
  category="lu",
  direction="l",
 },
 [0x118B0]={
  category="lu",
  direction="l",
 },
 [0x118B1]={
  category="lu",
  direction="l",
 },
 [0x118B2]={
  category="lu",
  direction="l",
 },
 [0x118B3]={
  category="lu",
  direction="l",
 },
 [0x118B4]={
  category="lu",
  direction="l",
 },
 [0x118B5]={
  category="lu",
  direction="l",
 },
 [0x118B6]={
  category="lu",
  direction="l",
 },
 [0x118B7]={
  category="lu",
  direction="l",
 },
 [0x118B8]={
  category="lu",
  direction="l",
 },
 [0x118B9]={
  category="lu",
  direction="l",
 },
 [0x118BA]={
  category="lu",
  direction="l",
 },
 [0x118BB]={
  category="lu",
  direction="l",
 },
 [0x118BC]={
  category="lu",
  direction="l",
 },
 [0x118BD]={
  category="lu",
  direction="l",
 },
 [0x118BE]={
  category="lu",
  direction="l",
 },
 [0x118BF]={
  category="lu",
  direction="l",
 },
 [0x118C0]={
  category="ll",
  direction="l",
 },
 [0x118C1]={
  category="ll",
  direction="l",
 },
 [0x118C2]={
  category="ll",
  direction="l",
 },
 [0x118C3]={
  category="ll",
  direction="l",
 },
 [0x118C4]={
  category="ll",
  direction="l",
 },
 [0x118C5]={
  category="ll",
  direction="l",
 },
 [0x118C6]={
  category="ll",
  direction="l",
 },
 [0x118C7]={
  category="ll",
  direction="l",
 },
 [0x118C8]={
  category="ll",
  direction="l",
 },
 [0x118C9]={
  category="ll",
  direction="l",
 },
 [0x118CA]={
  category="ll",
  direction="l",
 },
 [0x118CB]={
  category="ll",
  direction="l",
 },
 [0x118CC]={
  category="ll",
  direction="l",
 },
 [0x118CD]={
  category="ll",
  direction="l",
 },
 [0x118CE]={
  category="ll",
  direction="l",
 },
 [0x118CF]={
  category="ll",
  direction="l",
 },
 [0x118D0]={
  category="ll",
  direction="l",
 },
 [0x118D1]={
  category="ll",
  direction="l",
 },
 [0x118D2]={
  category="ll",
  direction="l",
 },
 [0x118D3]={
  category="ll",
  direction="l",
 },
 [0x118D4]={
  category="ll",
  direction="l",
 },
 [0x118D5]={
  category="ll",
  direction="l",
 },
 [0x118D6]={
  category="ll",
  direction="l",
 },
 [0x118D7]={
  category="ll",
  direction="l",
 },
 [0x118D8]={
  category="ll",
  direction="l",
 },
 [0x118D9]={
  category="ll",
  direction="l",
 },
 [0x118DA]={
  category="ll",
  direction="l",
 },
 [0x118DB]={
  category="ll",
  direction="l",
 },
 [0x118DC]={
  category="ll",
  direction="l",
 },
 [0x118DD]={
  category="ll",
  direction="l",
 },
 [0x118DE]={
  category="ll",
  direction="l",
 },
 [0x118DF]={
  category="ll",
  direction="l",
 },
 [0x118E0]={
  category="nd",
  direction="l",
 },
 [0x118E1]={
  category="nd",
  direction="l",
 },
 [0x118E2]={
  category="nd",
  direction="l",
 },
 [0x118E3]={
  category="nd",
  direction="l",
 },
 [0x118E4]={
  category="nd",
  direction="l",
 },
 [0x118E5]={
  category="nd",
  direction="l",
 },
 [0x118E6]={
  category="nd",
  direction="l",
 },
 [0x118E7]={
  category="nd",
  direction="l",
 },
 [0x118E8]={
  category="nd",
  direction="l",
 },
 [0x118E9]={
  category="nd",
  direction="l",
 },
 [0x118EA]={
  category="no",
  direction="l",
 },
 [0x118EB]={
  category="no",
  direction="l",
 },
 [0x118EC]={
  category="no",
  direction="l",
 },
 [0x118ED]={
  category="no",
  direction="l",
 },
 [0x118EE]={
  category="no",
  direction="l",
 },
 [0x118EF]={
  category="no",
  direction="l",
 },
 [0x118F0]={
  category="no",
  direction="l",
 },
 [0x118F1]={
  category="no",
  direction="l",
 },
 [0x118F2]={
  category="no",
  direction="l",
 },
 [0x118FF]={
  category="lo",
  direction="l",
 },
 [0x11AC0]={
  category="lo",
  direction="l",
 },
 [0x11AC1]={
  category="lo",
  direction="l",
 },
 [0x11AC2]={
  category="lo",
  direction="l",
 },
 [0x11AC3]={
  category="lo",
  direction="l",
 },
 [0x11AC4]={
  category="lo",
  direction="l",
 },
 [0x11AC5]={
  category="lo",
  direction="l",
 },
 [0x11AC6]={
  category="lo",
  direction="l",
 },
 [0x11AC7]={
  category="lo",
  direction="l",
 },
 [0x11AC8]={
  category="lo",
  direction="l",
 },
 [0x11AC9]={
  category="lo",
  direction="l",
 },
 [0x11ACA]={
  category="lo",
  direction="l",
 },
 [0x11ACB]={
  category="lo",
  direction="l",
 },
 [0x11ACC]={
  category="lo",
  direction="l",
 },
 [0x11ACD]={
  category="lo",
  direction="l",
 },
 [0x11ACE]={
  category="lo",
  direction="l",
 },
 [0x11ACF]={
  category="lo",
  direction="l",
 },
 [0x11AD0]={
  category="lo",
  direction="l",
 },
 [0x11AD1]={
  category="lo",
  direction="l",
 },
 [0x11AD2]={
  category="lo",
  direction="l",
 },
 [0x11AD3]={
  category="lo",
  direction="l",
 },
 [0x11AD4]={
  category="lo",
  direction="l",
 },
 [0x11AD5]={
  category="lo",
  direction="l",
 },
 [0x11AD6]={
  category="lo",
  direction="l",
 },
 [0x11AD7]={
  category="lo",
  direction="l",
 },
 [0x11AD8]={
  category="lo",
  direction="l",
 },
 [0x11AD9]={
  category="lo",
  direction="l",
 },
 [0x11ADA]={
  category="lo",
  direction="l",
 },
 [0x11ADB]={
  category="lo",
  direction="l",
 },
 [0x11ADC]={
  category="lo",
  direction="l",
 },
 [0x11ADD]={
  category="lo",
  direction="l",
 },
 [0x11ADE]={
  category="lo",
  direction="l",
 },
 [0x11ADF]={
  category="lo",
  direction="l",
 },
 [0x11AE0]={
  category="lo",
  direction="l",
 },
 [0x11AE1]={
  category="lo",
  direction="l",
 },
 [0x11AE2]={
  category="lo",
  direction="l",
 },
 [0x11AE3]={
  category="lo",
  direction="l",
 },
 [0x11AE4]={
  category="lo",
  direction="l",
 },
 [0x11AE5]={
  category="lo",
  direction="l",
 },
 [0x11AE6]={
  category="lo",
  direction="l",
 },
 [0x11AE7]={
  category="lo",
  direction="l",
 },
 [0x11AE8]={
  category="lo",
  direction="l",
 },
 [0x11AE9]={
  category="lo",
  direction="l",
 },
 [0x11AEA]={
  category="lo",
  direction="l",
 },
 [0x11AEB]={
  category="lo",
  direction="l",
 },
 [0x11AEC]={
  category="lo",
  direction="l",
 },
 [0x11AED]={
  category="lo",
  direction="l",
 },
 [0x11AEE]={
  category="lo",
  direction="l",
 },
 [0x11AEF]={
  category="lo",
  direction="l",
 },
 [0x11AF0]={
  category="lo",
  direction="l",
 },
 [0x11AF1]={
  category="lo",
  direction="l",
 },
 [0x11AF2]={
  category="lo",
  direction="l",
 },
 [0x11AF3]={
  category="lo",
  direction="l",
 },
 [0x11AF4]={
  category="lo",
  direction="l",
 },
 [0x11AF5]={
  category="lo",
  direction="l",
 },
 [0x11AF6]={
  category="lo",
  direction="l",
 },
 [0x11AF7]={
  category="lo",
  direction="l",
 },
 [0x11AF8]={
  category="lo",
  direction="l",
 },
 [0x12000]={
  category="lo",
  direction="l",
 },
 [0x12001]={
  category="lo",
  direction="l",
 },
 [0x12002]={
  category="lo",
  direction="l",
 },
 [0x12003]={
  category="lo",
  direction="l",
 },
 [0x12004]={
  category="lo",
  direction="l",
 },
 [0x12005]={
  category="lo",
  direction="l",
 },
 [0x12006]={
  category="lo",
  direction="l",
 },
 [0x12007]={
  category="lo",
  direction="l",
 },
 [0x12008]={
  category="lo",
  direction="l",
 },
 [0x12009]={
  category="lo",
  direction="l",
 },
 [0x1200A]={
  category="lo",
  direction="l",
 },
 [0x1200B]={
  category="lo",
  direction="l",
 },
 [0x1200C]={
  category="lo",
  direction="l",
 },
 [0x1200D]={
  category="lo",
  direction="l",
 },
 [0x1200E]={
  category="lo",
  direction="l",
 },
 [0x1200F]={
  category="lo",
  direction="l",
 },
 [0x12010]={
  category="lo",
  direction="l",
 },
 [0x12011]={
  category="lo",
  direction="l",
 },
 [0x12012]={
  category="lo",
  direction="l",
 },
 [0x12013]={
  category="lo",
  direction="l",
 },
 [0x12014]={
  category="lo",
  direction="l",
 },
 [0x12015]={
  category="lo",
  direction="l",
 },
 [0x12016]={
  category="lo",
  direction="l",
 },
 [0x12017]={
  category="lo",
  direction="l",
 },
 [0x12018]={
  category="lo",
  direction="l",
 },
 [0x12019]={
  category="lo",
  direction="l",
 },
 [0x1201A]={
  category="lo",
  direction="l",
 },
 [0x1201B]={
  category="lo",
  direction="l",
 },
 [0x1201C]={
  category="lo",
  direction="l",
 },
 [0x1201D]={
  category="lo",
  direction="l",
 },
 [0x1201E]={
  category="lo",
  direction="l",
 },
 [0x1201F]={
  category="lo",
  direction="l",
 },
 [0x12020]={
  category="lo",
  direction="l",
 },
 [0x12021]={
  category="lo",
  direction="l",
 },
 [0x12022]={
  category="lo",
  direction="l",
 },
 [0x12023]={
  category="lo",
  direction="l",
 },
 [0x12024]={
  category="lo",
  direction="l",
 },
 [0x12025]={
  category="lo",
  direction="l",
 },
 [0x12026]={
  category="lo",
  direction="l",
 },
 [0x12027]={
  category="lo",
  direction="l",
 },
 [0x12028]={
  category="lo",
  direction="l",
 },
 [0x12029]={
  category="lo",
  direction="l",
 },
 [0x1202A]={
  category="lo",
  direction="l",
 },
 [0x1202B]={
  category="lo",
  direction="l",
 },
 [0x1202C]={
  category="lo",
  direction="l",
 },
 [0x1202D]={
  category="lo",
  direction="l",
 },
 [0x1202E]={
  category="lo",
  direction="l",
 },
 [0x1202F]={
  category="lo",
  direction="l",
 },
 [0x12030]={
  category="lo",
  direction="l",
 },
 [0x12031]={
  category="lo",
  direction="l",
 },
 [0x12032]={
  category="lo",
  direction="l",
 },
 [0x12033]={
  category="lo",
  direction="l",
 },
 [0x12034]={
  category="lo",
  direction="l",
 },
 [0x12035]={
  category="lo",
  direction="l",
 },
 [0x12036]={
  category="lo",
  direction="l",
 },
 [0x12037]={
  category="lo",
  direction="l",
 },
 [0x12038]={
  category="lo",
  direction="l",
 },
 [0x12039]={
  category="lo",
  direction="l",
 },
 [0x1203A]={
  category="lo",
  direction="l",
 },
 [0x1203B]={
  category="lo",
  direction="l",
 },
 [0x1203C]={
  category="lo",
  direction="l",
 },
 [0x1203D]={
  category="lo",
  direction="l",
 },
 [0x1203E]={
  category="lo",
  direction="l",
 },
 [0x1203F]={
  category="lo",
  direction="l",
 },
 [0x12040]={
  category="lo",
  direction="l",
 },
 [0x12041]={
  category="lo",
  direction="l",
 },
 [0x12042]={
  category="lo",
  direction="l",
 },
 [0x12043]={
  category="lo",
  direction="l",
 },
 [0x12044]={
  category="lo",
  direction="l",
 },
 [0x12045]={
  category="lo",
  direction="l",
 },
 [0x12046]={
  category="lo",
  direction="l",
 },
 [0x12047]={
  category="lo",
  direction="l",
 },
 [0x12048]={
  category="lo",
  direction="l",
 },
 [0x12049]={
  category="lo",
  direction="l",
 },
 [0x1204A]={
  category="lo",
  direction="l",
 },
 [0x1204B]={
  category="lo",
  direction="l",
 },
 [0x1204C]={
  category="lo",
  direction="l",
 },
 [0x1204D]={
  category="lo",
  direction="l",
 },
 [0x1204E]={
  category="lo",
  direction="l",
 },
 [0x1204F]={
  category="lo",
  direction="l",
 },
 [0x12050]={
  category="lo",
  direction="l",
 },
 [0x12051]={
  category="lo",
  direction="l",
 },
 [0x12052]={
  category="lo",
  direction="l",
 },
 [0x12053]={
  category="lo",
  direction="l",
 },
 [0x12054]={
  category="lo",
  direction="l",
 },
 [0x12055]={
  category="lo",
  direction="l",
 },
 [0x12056]={
  category="lo",
  direction="l",
 },
 [0x12057]={
  category="lo",
  direction="l",
 },
 [0x12058]={
  category="lo",
  direction="l",
 },
 [0x12059]={
  category="lo",
  direction="l",
 },
 [0x1205A]={
  category="lo",
  direction="l",
 },
 [0x1205B]={
  category="lo",
  direction="l",
 },
 [0x1205C]={
  category="lo",
  direction="l",
 },
 [0x1205D]={
  category="lo",
  direction="l",
 },
 [0x1205E]={
  category="lo",
  direction="l",
 },
 [0x1205F]={
  category="lo",
  direction="l",
 },
 [0x12060]={
  category="lo",
  direction="l",
 },
 [0x12061]={
  category="lo",
  direction="l",
 },
 [0x12062]={
  category="lo",
  direction="l",
 },
 [0x12063]={
  category="lo",
  direction="l",
 },
 [0x12064]={
  category="lo",
  direction="l",
 },
 [0x12065]={
  category="lo",
  direction="l",
 },
 [0x12066]={
  category="lo",
  direction="l",
 },
 [0x12067]={
  category="lo",
  direction="l",
 },
 [0x12068]={
  category="lo",
  direction="l",
 },
 [0x12069]={
  category="lo",
  direction="l",
 },
 [0x1206A]={
  category="lo",
  direction="l",
 },
 [0x1206B]={
  category="lo",
  direction="l",
 },
 [0x1206C]={
  category="lo",
  direction="l",
 },
 [0x1206D]={
  category="lo",
  direction="l",
 },
 [0x1206E]={
  category="lo",
  direction="l",
 },
 [0x1206F]={
  category="lo",
  direction="l",
 },
 [0x12070]={
  category="lo",
  direction="l",
 },
 [0x12071]={
  category="lo",
  direction="l",
 },
 [0x12072]={
  category="lo",
  direction="l",
 },
 [0x12073]={
  category="lo",
  direction="l",
 },
 [0x12074]={
  category="lo",
  direction="l",
 },
 [0x12075]={
  category="lo",
  direction="l",
 },
 [0x12076]={
  category="lo",
  direction="l",
 },
 [0x12077]={
  category="lo",
  direction="l",
 },
 [0x12078]={
  category="lo",
  direction="l",
 },
 [0x12079]={
  category="lo",
  direction="l",
 },
 [0x1207A]={
  category="lo",
  direction="l",
 },
 [0x1207B]={
  category="lo",
  direction="l",
 },
 [0x1207C]={
  category="lo",
  direction="l",
 },
 [0x1207D]={
  category="lo",
  direction="l",
 },
 [0x1207E]={
  category="lo",
  direction="l",
 },
 [0x1207F]={
  category="lo",
  direction="l",
 },
 [0x12080]={
  category="lo",
  direction="l",
 },
 [0x12081]={
  category="lo",
  direction="l",
 },
 [0x12082]={
  category="lo",
  direction="l",
 },
 [0x12083]={
  category="lo",
  direction="l",
 },
 [0x12084]={
  category="lo",
  direction="l",
 },
 [0x12085]={
  category="lo",
  direction="l",
 },
 [0x12086]={
  category="lo",
  direction="l",
 },
 [0x12087]={
  category="lo",
  direction="l",
 },
 [0x12088]={
  category="lo",
  direction="l",
 },
 [0x12089]={
  category="lo",
  direction="l",
 },
 [0x1208A]={
  category="lo",
  direction="l",
 },
 [0x1208B]={
  category="lo",
  direction="l",
 },
 [0x1208C]={
  category="lo",
  direction="l",
 },
 [0x1208D]={
  category="lo",
  direction="l",
 },
 [0x1208E]={
  category="lo",
  direction="l",
 },
 [0x1208F]={
  category="lo",
  direction="l",
 },
 [0x12090]={
  category="lo",
  direction="l",
 },
 [0x12091]={
  category="lo",
  direction="l",
 },
 [0x12092]={
  category="lo",
  direction="l",
 },
 [0x12093]={
  category="lo",
  direction="l",
 },
 [0x12094]={
  category="lo",
  direction="l",
 },
 [0x12095]={
  category="lo",
  direction="l",
 },
 [0x12096]={
  category="lo",
  direction="l",
 },
 [0x12097]={
  category="lo",
  direction="l",
 },
 [0x12098]={
  category="lo",
  direction="l",
 },
 [0x12099]={
  category="lo",
  direction="l",
 },
 [0x1209A]={
  category="lo",
  direction="l",
 },
 [0x1209B]={
  category="lo",
  direction="l",
 },
 [0x1209C]={
  category="lo",
  direction="l",
 },
 [0x1209D]={
  category="lo",
  direction="l",
 },
 [0x1209E]={
  category="lo",
  direction="l",
 },
 [0x1209F]={
  category="lo",
  direction="l",
 },
 [0x120A0]={
  category="lo",
  direction="l",
 },
 [0x120A1]={
  category="lo",
  direction="l",
 },
 [0x120A2]={
  category="lo",
  direction="l",
 },
 [0x120A3]={
  category="lo",
  direction="l",
 },
 [0x120A4]={
  category="lo",
  direction="l",
 },
 [0x120A5]={
  category="lo",
  direction="l",
 },
 [0x120A6]={
  category="lo",
  direction="l",
 },
 [0x120A7]={
  category="lo",
  direction="l",
 },
 [0x120A8]={
  category="lo",
  direction="l",
 },
 [0x120A9]={
  category="lo",
  direction="l",
 },
 [0x120AA]={
  category="lo",
  direction="l",
 },
 [0x120AB]={
  category="lo",
  direction="l",
 },
 [0x120AC]={
  category="lo",
  direction="l",
 },
 [0x120AD]={
  category="lo",
  direction="l",
 },
 [0x120AE]={
  category="lo",
  direction="l",
 },
 [0x120AF]={
  category="lo",
  direction="l",
 },
 [0x120B0]={
  category="lo",
  direction="l",
 },
 [0x120B1]={
  category="lo",
  direction="l",
 },
 [0x120B2]={
  category="lo",
  direction="l",
 },
 [0x120B3]={
  category="lo",
  direction="l",
 },
 [0x120B4]={
  category="lo",
  direction="l",
 },
 [0x120B5]={
  category="lo",
  direction="l",
 },
 [0x120B6]={
  category="lo",
  direction="l",
 },
 [0x120B7]={
  category="lo",
  direction="l",
 },
 [0x120B8]={
  category="lo",
  direction="l",
 },
 [0x120B9]={
  category="lo",
  direction="l",
 },
 [0x120BA]={
  category="lo",
  direction="l",
 },
 [0x120BB]={
  category="lo",
  direction="l",
 },
 [0x120BC]={
  category="lo",
  direction="l",
 },
 [0x120BD]={
  category="lo",
  direction="l",
 },
 [0x120BE]={
  category="lo",
  direction="l",
 },
 [0x120BF]={
  category="lo",
  direction="l",
 },
 [0x120C0]={
  category="lo",
  direction="l",
 },
 [0x120C1]={
  category="lo",
  direction="l",
 },
 [0x120C2]={
  category="lo",
  direction="l",
 },
 [0x120C3]={
  category="lo",
  direction="l",
 },
 [0x120C4]={
  category="lo",
  direction="l",
 },
 [0x120C5]={
  category="lo",
  direction="l",
 },
 [0x120C6]={
  category="lo",
  direction="l",
 },
 [0x120C7]={
  category="lo",
  direction="l",
 },
 [0x120C8]={
  category="lo",
  direction="l",
 },
 [0x120C9]={
  category="lo",
  direction="l",
 },
 [0x120CA]={
  category="lo",
  direction="l",
 },
 [0x120CB]={
  category="lo",
  direction="l",
 },
 [0x120CC]={
  category="lo",
  direction="l",
 },
 [0x120CD]={
  category="lo",
  direction="l",
 },
 [0x120CE]={
  category="lo",
  direction="l",
 },
 [0x120CF]={
  category="lo",
  direction="l",
 },
 [0x120D0]={
  category="lo",
  direction="l",
 },
 [0x120D1]={
  category="lo",
  direction="l",
 },
 [0x120D2]={
  category="lo",
  direction="l",
 },
 [0x120D3]={
  category="lo",
  direction="l",
 },
 [0x120D4]={
  category="lo",
  direction="l",
 },
 [0x120D5]={
  category="lo",
  direction="l",
 },
 [0x120D6]={
  category="lo",
  direction="l",
 },
 [0x120D7]={
  category="lo",
  direction="l",
 },
 [0x120D8]={
  category="lo",
  direction="l",
 },
 [0x120D9]={
  category="lo",
  direction="l",
 },
 [0x120DA]={
  category="lo",
  direction="l",
 },
 [0x120DB]={
  category="lo",
  direction="l",
 },
 [0x120DC]={
  category="lo",
  direction="l",
 },
 [0x120DD]={
  category="lo",
  direction="l",
 },
 [0x120DE]={
  category="lo",
  direction="l",
 },
 [0x120DF]={
  category="lo",
  direction="l",
 },
 [0x120E0]={
  category="lo",
  direction="l",
 },
 [0x120E1]={
  category="lo",
  direction="l",
 },
 [0x120E2]={
  category="lo",
  direction="l",
 },
 [0x120E3]={
  category="lo",
  direction="l",
 },
 [0x120E4]={
  category="lo",
  direction="l",
 },
 [0x120E5]={
  category="lo",
  direction="l",
 },
 [0x120E6]={
  category="lo",
  direction="l",
 },
 [0x120E7]={
  category="lo",
  direction="l",
 },
 [0x120E8]={
  category="lo",
  direction="l",
 },
 [0x120E9]={
  category="lo",
  direction="l",
 },
 [0x120EA]={
  category="lo",
  direction="l",
 },
 [0x120EB]={
  category="lo",
  direction="l",
 },
 [0x120EC]={
  category="lo",
  direction="l",
 },
 [0x120ED]={
  category="lo",
  direction="l",
 },
 [0x120EE]={
  category="lo",
  direction="l",
 },
 [0x120EF]={
  category="lo",
  direction="l",
 },
 [0x120F0]={
  category="lo",
  direction="l",
 },
 [0x120F1]={
  category="lo",
  direction="l",
 },
 [0x120F2]={
  category="lo",
  direction="l",
 },
 [0x120F3]={
  category="lo",
  direction="l",
 },
 [0x120F4]={
  category="lo",
  direction="l",
 },
 [0x120F5]={
  category="lo",
  direction="l",
 },
 [0x120F6]={
  category="lo",
  direction="l",
 },
 [0x120F7]={
  category="lo",
  direction="l",
 },
 [0x120F8]={
  category="lo",
  direction="l",
 },
 [0x120F9]={
  category="lo",
  direction="l",
 },
 [0x120FA]={
  category="lo",
  direction="l",
 },
 [0x120FB]={
  category="lo",
  direction="l",
 },
 [0x120FC]={
  category="lo",
  direction="l",
 },
 [0x120FD]={
  category="lo",
  direction="l",
 },
 [0x120FE]={
  category="lo",
  direction="l",
 },
 [0x120FF]={
  category="lo",
  direction="l",
 },
 [0x12100]={
  category="lo",
  direction="l",
 },
 [0x12101]={
  category="lo",
  direction="l",
 },
 [0x12102]={
  category="lo",
  direction="l",
 },
 [0x12103]={
  category="lo",
  direction="l",
 },
 [0x12104]={
  category="lo",
  direction="l",
 },
 [0x12105]={
  category="lo",
  direction="l",
 },
 [0x12106]={
  category="lo",
  direction="l",
 },
 [0x12107]={
  category="lo",
  direction="l",
 },
 [0x12108]={
  category="lo",
  direction="l",
 },
 [0x12109]={
  category="lo",
  direction="l",
 },
 [0x1210A]={
  category="lo",
  direction="l",
 },
 [0x1210B]={
  category="lo",
  direction="l",
 },
 [0x1210C]={
  category="lo",
  direction="l",
 },
 [0x1210D]={
  category="lo",
  direction="l",
 },
 [0x1210E]={
  category="lo",
  direction="l",
 },
 [0x1210F]={
  category="lo",
  direction="l",
 },
 [0x12110]={
  category="lo",
  direction="l",
 },
 [0x12111]={
  category="lo",
  direction="l",
 },
 [0x12112]={
  category="lo",
  direction="l",
 },
 [0x12113]={
  category="lo",
  direction="l",
 },
 [0x12114]={
  category="lo",
  direction="l",
 },
 [0x12115]={
  category="lo",
  direction="l",
 },
 [0x12116]={
  category="lo",
  direction="l",
 },
 [0x12117]={
  category="lo",
  direction="l",
 },
 [0x12118]={
  category="lo",
  direction="l",
 },
 [0x12119]={
  category="lo",
  direction="l",
 },
 [0x1211A]={
  category="lo",
  direction="l",
 },
 [0x1211B]={
  category="lo",
  direction="l",
 },
 [0x1211C]={
  category="lo",
  direction="l",
 },
 [0x1211D]={
  category="lo",
  direction="l",
 },
 [0x1211E]={
  category="lo",
  direction="l",
 },
 [0x1211F]={
  category="lo",
  direction="l",
 },
 [0x12120]={
  category="lo",
  direction="l",
 },
 [0x12121]={
  category="lo",
  direction="l",
 },
 [0x12122]={
  category="lo",
  direction="l",
 },
 [0x12123]={
  category="lo",
  direction="l",
 },
 [0x12124]={
  category="lo",
  direction="l",
 },
 [0x12125]={
  category="lo",
  direction="l",
 },
 [0x12126]={
  category="lo",
  direction="l",
 },
 [0x12127]={
  category="lo",
  direction="l",
 },
 [0x12128]={
  category="lo",
  direction="l",
 },
 [0x12129]={
  category="lo",
  direction="l",
 },
 [0x1212A]={
  category="lo",
  direction="l",
 },
 [0x1212B]={
  category="lo",
  direction="l",
 },
 [0x1212C]={
  category="lo",
  direction="l",
 },
 [0x1212D]={
  category="lo",
  direction="l",
 },
 [0x1212E]={
  category="lo",
  direction="l",
 },
 [0x1212F]={
  category="lo",
  direction="l",
 },
 [0x12130]={
  category="lo",
  direction="l",
 },
 [0x12131]={
  category="lo",
  direction="l",
 },
 [0x12132]={
  category="lo",
  direction="l",
 },
 [0x12133]={
  category="lo",
  direction="l",
 },
 [0x12134]={
  category="lo",
  direction="l",
 },
 [0x12135]={
  category="lo",
  direction="l",
 },
 [0x12136]={
  category="lo",
  direction="l",
 },
 [0x12137]={
  category="lo",
  direction="l",
 },
 [0x12138]={
  category="lo",
  direction="l",
 },
 [0x12139]={
  category="lo",
  direction="l",
 },
 [0x1213A]={
  category="lo",
  direction="l",
 },
 [0x1213B]={
  category="lo",
  direction="l",
 },
 [0x1213C]={
  category="lo",
  direction="l",
 },
 [0x1213D]={
  category="lo",
  direction="l",
 },
 [0x1213E]={
  category="lo",
  direction="l",
 },
 [0x1213F]={
  category="lo",
  direction="l",
 },
 [0x12140]={
  category="lo",
  direction="l",
 },
 [0x12141]={
  category="lo",
  direction="l",
 },
 [0x12142]={
  category="lo",
  direction="l",
 },
 [0x12143]={
  category="lo",
  direction="l",
 },
 [0x12144]={
  category="lo",
  direction="l",
 },
 [0x12145]={
  category="lo",
  direction="l",
 },
 [0x12146]={
  category="lo",
  direction="l",
 },
 [0x12147]={
  category="lo",
  direction="l",
 },
 [0x12148]={
  category="lo",
  direction="l",
 },
 [0x12149]={
  category="lo",
  direction="l",
 },
 [0x1214A]={
  category="lo",
  direction="l",
 },
 [0x1214B]={
  category="lo",
  direction="l",
 },
 [0x1214C]={
  category="lo",
  direction="l",
 },
 [0x1214D]={
  category="lo",
  direction="l",
 },
 [0x1214E]={
  category="lo",
  direction="l",
 },
 [0x1214F]={
  category="lo",
  direction="l",
 },
 [0x12150]={
  category="lo",
  direction="l",
 },
 [0x12151]={
  category="lo",
  direction="l",
 },
 [0x12152]={
  category="lo",
  direction="l",
 },
 [0x12153]={
  category="lo",
  direction="l",
 },
 [0x12154]={
  category="lo",
  direction="l",
 },
 [0x12155]={
  category="lo",
  direction="l",
 },
 [0x12156]={
  category="lo",
  direction="l",
 },
 [0x12157]={
  category="lo",
  direction="l",
 },
 [0x12158]={
  category="lo",
  direction="l",
 },
 [0x12159]={
  category="lo",
  direction="l",
 },
 [0x1215A]={
  category="lo",
  direction="l",
 },
 [0x1215B]={
  category="lo",
  direction="l",
 },
 [0x1215C]={
  category="lo",
  direction="l",
 },
 [0x1215D]={
  category="lo",
  direction="l",
 },
 [0x1215E]={
  category="lo",
  direction="l",
 },
 [0x1215F]={
  category="lo",
  direction="l",
 },
 [0x12160]={
  category="lo",
  direction="l",
 },
 [0x12161]={
  category="lo",
  direction="l",
 },
 [0x12162]={
  category="lo",
  direction="l",
 },
 [0x12163]={
  category="lo",
  direction="l",
 },
 [0x12164]={
  category="lo",
  direction="l",
 },
 [0x12165]={
  category="lo",
  direction="l",
 },
 [0x12166]={
  category="lo",
  direction="l",
 },
 [0x12167]={
  category="lo",
  direction="l",
 },
 [0x12168]={
  category="lo",
  direction="l",
 },
 [0x12169]={
  category="lo",
  direction="l",
 },
 [0x1216A]={
  category="lo",
  direction="l",
 },
 [0x1216B]={
  category="lo",
  direction="l",
 },
 [0x1216C]={
  category="lo",
  direction="l",
 },
 [0x1216D]={
  category="lo",
  direction="l",
 },
 [0x1216E]={
  category="lo",
  direction="l",
 },
 [0x1216F]={
  category="lo",
  direction="l",
 },
 [0x12170]={
  category="lo",
  direction="l",
 },
 [0x12171]={
  category="lo",
  direction="l",
 },
 [0x12172]={
  category="lo",
  direction="l",
 },
 [0x12173]={
  category="lo",
  direction="l",
 },
 [0x12174]={
  category="lo",
  direction="l",
 },
 [0x12175]={
  category="lo",
  direction="l",
 },
 [0x12176]={
  category="lo",
  direction="l",
 },
 [0x12177]={
  category="lo",
  direction="l",
 },
 [0x12178]={
  category="lo",
  direction="l",
 },
 [0x12179]={
  category="lo",
  direction="l",
 },
 [0x1217A]={
  category="lo",
  direction="l",
 },
 [0x1217B]={
  category="lo",
  direction="l",
 },
 [0x1217C]={
  category="lo",
  direction="l",
 },
 [0x1217D]={
  category="lo",
  direction="l",
 },
 [0x1217E]={
  category="lo",
  direction="l",
 },
 [0x1217F]={
  category="lo",
  direction="l",
 },
 [0x12180]={
  category="lo",
  direction="l",
 },
 [0x12181]={
  category="lo",
  direction="l",
 },
 [0x12182]={
  category="lo",
  direction="l",
 },
 [0x12183]={
  category="lo",
  direction="l",
 },
 [0x12184]={
  category="lo",
  direction="l",
 },
 [0x12185]={
  category="lo",
  direction="l",
 },
 [0x12186]={
  category="lo",
  direction="l",
 },
 [0x12187]={
  category="lo",
  direction="l",
 },
 [0x12188]={
  category="lo",
  direction="l",
 },
 [0x12189]={
  category="lo",
  direction="l",
 },
 [0x1218A]={
  category="lo",
  direction="l",
 },
 [0x1218B]={
  category="lo",
  direction="l",
 },
 [0x1218C]={
  category="lo",
  direction="l",
 },
 [0x1218D]={
  category="lo",
  direction="l",
 },
 [0x1218E]={
  category="lo",
  direction="l",
 },
 [0x1218F]={
  category="lo",
  direction="l",
 },
 [0x12190]={
  category="lo",
  direction="l",
 },
 [0x12191]={
  category="lo",
  direction="l",
 },
 [0x12192]={
  category="lo",
  direction="l",
 },
 [0x12193]={
  category="lo",
  direction="l",
 },
 [0x12194]={
  category="lo",
  direction="l",
 },
 [0x12195]={
  category="lo",
  direction="l",
 },
 [0x12196]={
  category="lo",
  direction="l",
 },
 [0x12197]={
  category="lo",
  direction="l",
 },
 [0x12198]={
  category="lo",
  direction="l",
 },
 [0x12199]={
  category="lo",
  direction="l",
 },
 [0x1219A]={
  category="lo",
  direction="l",
 },
 [0x1219B]={
  category="lo",
  direction="l",
 },
 [0x1219C]={
  category="lo",
  direction="l",
 },
 [0x1219D]={
  category="lo",
  direction="l",
 },
 [0x1219E]={
  category="lo",
  direction="l",
 },
 [0x1219F]={
  category="lo",
  direction="l",
 },
 [0x121A0]={
  category="lo",
  direction="l",
 },
 [0x121A1]={
  category="lo",
  direction="l",
 },
 [0x121A2]={
  category="lo",
  direction="l",
 },
 [0x121A3]={
  category="lo",
  direction="l",
 },
 [0x121A4]={
  category="lo",
  direction="l",
 },
 [0x121A5]={
  category="lo",
  direction="l",
 },
 [0x121A6]={
  category="lo",
  direction="l",
 },
 [0x121A7]={
  category="lo",
  direction="l",
 },
 [0x121A8]={
  category="lo",
  direction="l",
 },
 [0x121A9]={
  category="lo",
  direction="l",
 },
 [0x121AA]={
  category="lo",
  direction="l",
 },
 [0x121AB]={
  category="lo",
  direction="l",
 },
 [0x121AC]={
  category="lo",
  direction="l",
 },
 [0x121AD]={
  category="lo",
  direction="l",
 },
 [0x121AE]={
  category="lo",
  direction="l",
 },
 [0x121AF]={
  category="lo",
  direction="l",
 },
 [0x121B0]={
  category="lo",
  direction="l",
 },
 [0x121B1]={
  category="lo",
  direction="l",
 },
 [0x121B2]={
  category="lo",
  direction="l",
 },
 [0x121B3]={
  category="lo",
  direction="l",
 },
 [0x121B4]={
  category="lo",
  direction="l",
 },
 [0x121B5]={
  category="lo",
  direction="l",
 },
 [0x121B6]={
  category="lo",
  direction="l",
 },
 [0x121B7]={
  category="lo",
  direction="l",
 },
 [0x121B8]={
  category="lo",
  direction="l",
 },
 [0x121B9]={
  category="lo",
  direction="l",
 },
 [0x121BA]={
  category="lo",
  direction="l",
 },
 [0x121BB]={
  category="lo",
  direction="l",
 },
 [0x121BC]={
  category="lo",
  direction="l",
 },
 [0x121BD]={
  category="lo",
  direction="l",
 },
 [0x121BE]={
  category="lo",
  direction="l",
 },
 [0x121BF]={
  category="lo",
  direction="l",
 },
 [0x121C0]={
  category="lo",
  direction="l",
 },
 [0x121C1]={
  category="lo",
  direction="l",
 },
 [0x121C2]={
  category="lo",
  direction="l",
 },
 [0x121C3]={
  category="lo",
  direction="l",
 },
 [0x121C4]={
  category="lo",
  direction="l",
 },
 [0x121C5]={
  category="lo",
  direction="l",
 },
 [0x121C6]={
  category="lo",
  direction="l",
 },
 [0x121C7]={
  category="lo",
  direction="l",
 },
 [0x121C8]={
  category="lo",
  direction="l",
 },
 [0x121C9]={
  category="lo",
  direction="l",
 },
 [0x121CA]={
  category="lo",
  direction="l",
 },
 [0x121CB]={
  category="lo",
  direction="l",
 },
 [0x121CC]={
  category="lo",
  direction="l",
 },
 [0x121CD]={
  category="lo",
  direction="l",
 },
 [0x121CE]={
  category="lo",
  direction="l",
 },
 [0x121CF]={
  category="lo",
  direction="l",
 },
 [0x121D0]={
  category="lo",
  direction="l",
 },
 [0x121D1]={
  category="lo",
  direction="l",
 },
 [0x121D2]={
  category="lo",
  direction="l",
 },
 [0x121D3]={
  category="lo",
  direction="l",
 },
 [0x121D4]={
  category="lo",
  direction="l",
 },
 [0x121D5]={
  category="lo",
  direction="l",
 },
 [0x121D6]={
  category="lo",
  direction="l",
 },
 [0x121D7]={
  category="lo",
  direction="l",
 },
 [0x121D8]={
  category="lo",
  direction="l",
 },
 [0x121D9]={
  category="lo",
  direction="l",
 },
 [0x121DA]={
  category="lo",
  direction="l",
 },
 [0x121DB]={
  category="lo",
  direction="l",
 },
 [0x121DC]={
  category="lo",
  direction="l",
 },
 [0x121DD]={
  category="lo",
  direction="l",
 },
 [0x121DE]={
  category="lo",
  direction="l",
 },
 [0x121DF]={
  category="lo",
  direction="l",
 },
 [0x121E0]={
  category="lo",
  direction="l",
 },
 [0x121E1]={
  category="lo",
  direction="l",
 },
 [0x121E2]={
  category="lo",
  direction="l",
 },
 [0x121E3]={
  category="lo",
  direction="l",
 },
 [0x121E4]={
  category="lo",
  direction="l",
 },
 [0x121E5]={
  category="lo",
  direction="l",
 },
 [0x121E6]={
  category="lo",
  direction="l",
 },
 [0x121E7]={
  category="lo",
  direction="l",
 },
 [0x121E8]={
  category="lo",
  direction="l",
 },
 [0x121E9]={
  category="lo",
  direction="l",
 },
 [0x121EA]={
  category="lo",
  direction="l",
 },
 [0x121EB]={
  category="lo",
  direction="l",
 },
 [0x121EC]={
  category="lo",
  direction="l",
 },
 [0x121ED]={
  category="lo",
  direction="l",
 },
 [0x121EE]={
  category="lo",
  direction="l",
 },
 [0x121EF]={
  category="lo",
  direction="l",
 },
 [0x121F0]={
  category="lo",
  direction="l",
 },
 [0x121F1]={
  category="lo",
  direction="l",
 },
 [0x121F2]={
  category="lo",
  direction="l",
 },
 [0x121F3]={
  category="lo",
  direction="l",
 },
 [0x121F4]={
  category="lo",
  direction="l",
 },
 [0x121F5]={
  category="lo",
  direction="l",
 },
 [0x121F6]={
  category="lo",
  direction="l",
 },
 [0x121F7]={
  category="lo",
  direction="l",
 },
 [0x121F8]={
  category="lo",
  direction="l",
 },
 [0x121F9]={
  category="lo",
  direction="l",
 },
 [0x121FA]={
  category="lo",
  direction="l",
 },
 [0x121FB]={
  category="lo",
  direction="l",
 },
 [0x121FC]={
  category="lo",
  direction="l",
 },
 [0x121FD]={
  category="lo",
  direction="l",
 },
 [0x121FE]={
  category="lo",
  direction="l",
 },
 [0x121FF]={
  category="lo",
  direction="l",
 },
 [0x12200]={
  category="lo",
  direction="l",
 },
 [0x12201]={
  category="lo",
  direction="l",
 },
 [0x12202]={
  category="lo",
  direction="l",
 },
 [0x12203]={
  category="lo",
  direction="l",
 },
 [0x12204]={
  category="lo",
  direction="l",
 },
 [0x12205]={
  category="lo",
  direction="l",
 },
 [0x12206]={
  category="lo",
  direction="l",
 },
 [0x12207]={
  category="lo",
  direction="l",
 },
 [0x12208]={
  category="lo",
  direction="l",
 },
 [0x12209]={
  category="lo",
  direction="l",
 },
 [0x1220A]={
  category="lo",
  direction="l",
 },
 [0x1220B]={
  category="lo",
  direction="l",
 },
 [0x1220C]={
  category="lo",
  direction="l",
 },
 [0x1220D]={
  category="lo",
  direction="l",
 },
 [0x1220E]={
  category="lo",
  direction="l",
 },
 [0x1220F]={
  category="lo",
  direction="l",
 },
 [0x12210]={
  category="lo",
  direction="l",
 },
 [0x12211]={
  category="lo",
  direction="l",
 },
 [0x12212]={
  category="lo",
  direction="l",
 },
 [0x12213]={
  category="lo",
  direction="l",
 },
 [0x12214]={
  category="lo",
  direction="l",
 },
 [0x12215]={
  category="lo",
  direction="l",
 },
 [0x12216]={
  category="lo",
  direction="l",
 },
 [0x12217]={
  category="lo",
  direction="l",
 },
 [0x12218]={
  category="lo",
  direction="l",
 },
 [0x12219]={
  category="lo",
  direction="l",
 },
 [0x1221A]={
  category="lo",
  direction="l",
 },
 [0x1221B]={
  category="lo",
  direction="l",
 },
 [0x1221C]={
  category="lo",
  direction="l",
 },
 [0x1221D]={
  category="lo",
  direction="l",
 },
 [0x1221E]={
  category="lo",
  direction="l",
 },
 [0x1221F]={
  category="lo",
  direction="l",
 },
 [0x12220]={
  category="lo",
  direction="l",
 },
 [0x12221]={
  category="lo",
  direction="l",
 },
 [0x12222]={
  category="lo",
  direction="l",
 },
 [0x12223]={
  category="lo",
  direction="l",
 },
 [0x12224]={
  category="lo",
  direction="l",
 },
 [0x12225]={
  category="lo",
  direction="l",
 },
 [0x12226]={
  category="lo",
  direction="l",
 },
 [0x12227]={
  category="lo",
  direction="l",
 },
 [0x12228]={
  category="lo",
  direction="l",
 },
 [0x12229]={
  category="lo",
  direction="l",
 },
 [0x1222A]={
  category="lo",
  direction="l",
 },
 [0x1222B]={
  category="lo",
  direction="l",
 },
 [0x1222C]={
  category="lo",
  direction="l",
 },
 [0x1222D]={
  category="lo",
  direction="l",
 },
 [0x1222E]={
  category="lo",
  direction="l",
 },
 [0x1222F]={
  category="lo",
  direction="l",
 },
 [0x12230]={
  category="lo",
  direction="l",
 },
 [0x12231]={
  category="lo",
  direction="l",
 },
 [0x12232]={
  category="lo",
  direction="l",
 },
 [0x12233]={
  category="lo",
  direction="l",
 },
 [0x12234]={
  category="lo",
  direction="l",
 },
 [0x12235]={
  category="lo",
  direction="l",
 },
 [0x12236]={
  category="lo",
  direction="l",
 },
 [0x12237]={
  category="lo",
  direction="l",
 },
 [0x12238]={
  category="lo",
  direction="l",
 },
 [0x12239]={
  category="lo",
  direction="l",
 },
 [0x1223A]={
  category="lo",
  direction="l",
 },
 [0x1223B]={
  category="lo",
  direction="l",
 },
 [0x1223C]={
  category="lo",
  direction="l",
 },
 [0x1223D]={
  category="lo",
  direction="l",
 },
 [0x1223E]={
  category="lo",
  direction="l",
 },
 [0x1223F]={
  category="lo",
  direction="l",
 },
 [0x12240]={
  category="lo",
  direction="l",
 },
 [0x12241]={
  category="lo",
  direction="l",
 },
 [0x12242]={
  category="lo",
  direction="l",
 },
 [0x12243]={
  category="lo",
  direction="l",
 },
 [0x12244]={
  category="lo",
  direction="l",
 },
 [0x12245]={
  category="lo",
  direction="l",
 },
 [0x12246]={
  category="lo",
  direction="l",
 },
 [0x12247]={
  category="lo",
  direction="l",
 },
 [0x12248]={
  category="lo",
  direction="l",
 },
 [0x12249]={
  category="lo",
  direction="l",
 },
 [0x1224A]={
  category="lo",
  direction="l",
 },
 [0x1224B]={
  category="lo",
  direction="l",
 },
 [0x1224C]={
  category="lo",
  direction="l",
 },
 [0x1224D]={
  category="lo",
  direction="l",
 },
 [0x1224E]={
  category="lo",
  direction="l",
 },
 [0x1224F]={
  category="lo",
  direction="l",
 },
 [0x12250]={
  category="lo",
  direction="l",
 },
 [0x12251]={
  category="lo",
  direction="l",
 },
 [0x12252]={
  category="lo",
  direction="l",
 },
 [0x12253]={
  category="lo",
  direction="l",
 },
 [0x12254]={
  category="lo",
  direction="l",
 },
 [0x12255]={
  category="lo",
  direction="l",
 },
 [0x12256]={
  category="lo",
  direction="l",
 },
 [0x12257]={
  category="lo",
  direction="l",
 },
 [0x12258]={
  category="lo",
  direction="l",
 },
 [0x12259]={
  category="lo",
  direction="l",
 },
 [0x1225A]={
  category="lo",
  direction="l",
 },
 [0x1225B]={
  category="lo",
  direction="l",
 },
 [0x1225C]={
  category="lo",
  direction="l",
 },
 [0x1225D]={
  category="lo",
  direction="l",
 },
 [0x1225E]={
  category="lo",
  direction="l",
 },
 [0x1225F]={
  category="lo",
  direction="l",
 },
 [0x12260]={
  category="lo",
  direction="l",
 },
 [0x12261]={
  category="lo",
  direction="l",
 },
 [0x12262]={
  category="lo",
  direction="l",
 },
 [0x12263]={
  category="lo",
  direction="l",
 },
 [0x12264]={
  category="lo",
  direction="l",
 },
 [0x12265]={
  category="lo",
  direction="l",
 },
 [0x12266]={
  category="lo",
  direction="l",
 },
 [0x12267]={
  category="lo",
  direction="l",
 },
 [0x12268]={
  category="lo",
  direction="l",
 },
 [0x12269]={
  category="lo",
  direction="l",
 },
 [0x1226A]={
  category="lo",
  direction="l",
 },
 [0x1226B]={
  category="lo",
  direction="l",
 },
 [0x1226C]={
  category="lo",
  direction="l",
 },
 [0x1226D]={
  category="lo",
  direction="l",
 },
 [0x1226E]={
  category="lo",
  direction="l",
 },
 [0x1226F]={
  category="lo",
  direction="l",
 },
 [0x12270]={
  category="lo",
  direction="l",
 },
 [0x12271]={
  category="lo",
  direction="l",
 },
 [0x12272]={
  category="lo",
  direction="l",
 },
 [0x12273]={
  category="lo",
  direction="l",
 },
 [0x12274]={
  category="lo",
  direction="l",
 },
 [0x12275]={
  category="lo",
  direction="l",
 },
 [0x12276]={
  category="lo",
  direction="l",
 },
 [0x12277]={
  category="lo",
  direction="l",
 },
 [0x12278]={
  category="lo",
  direction="l",
 },
 [0x12279]={
  category="lo",
  direction="l",
 },
 [0x1227A]={
  category="lo",
  direction="l",
 },
 [0x1227B]={
  category="lo",
  direction="l",
 },
 [0x1227C]={
  category="lo",
  direction="l",
 },
 [0x1227D]={
  category="lo",
  direction="l",
 },
 [0x1227E]={
  category="lo",
  direction="l",
 },
 [0x1227F]={
  category="lo",
  direction="l",
 },
 [0x12280]={
  category="lo",
  direction="l",
 },
 [0x12281]={
  category="lo",
  direction="l",
 },
 [0x12282]={
  category="lo",
  direction="l",
 },
 [0x12283]={
  category="lo",
  direction="l",
 },
 [0x12284]={
  category="lo",
  direction="l",
 },
 [0x12285]={
  category="lo",
  direction="l",
 },
 [0x12286]={
  category="lo",
  direction="l",
 },
 [0x12287]={
  category="lo",
  direction="l",
 },
 [0x12288]={
  category="lo",
  direction="l",
 },
 [0x12289]={
  category="lo",
  direction="l",
 },
 [0x1228A]={
  category="lo",
  direction="l",
 },
 [0x1228B]={
  category="lo",
  direction="l",
 },
 [0x1228C]={
  category="lo",
  direction="l",
 },
 [0x1228D]={
  category="lo",
  direction="l",
 },
 [0x1228E]={
  category="lo",
  direction="l",
 },
 [0x1228F]={
  category="lo",
  direction="l",
 },
 [0x12290]={
  category="lo",
  direction="l",
 },
 [0x12291]={
  category="lo",
  direction="l",
 },
 [0x12292]={
  category="lo",
  direction="l",
 },
 [0x12293]={
  category="lo",
  direction="l",
 },
 [0x12294]={
  category="lo",
  direction="l",
 },
 [0x12295]={
  category="lo",
  direction="l",
 },
 [0x12296]={
  category="lo",
  direction="l",
 },
 [0x12297]={
  category="lo",
  direction="l",
 },
 [0x12298]={
  category="lo",
  direction="l",
 },
 [0x12299]={
  category="lo",
  direction="l",
 },
 [0x1229A]={
  category="lo",
  direction="l",
 },
 [0x1229B]={
  category="lo",
  direction="l",
 },
 [0x1229C]={
  category="lo",
  direction="l",
 },
 [0x1229D]={
  category="lo",
  direction="l",
 },
 [0x1229E]={
  category="lo",
  direction="l",
 },
 [0x1229F]={
  category="lo",
  direction="l",
 },
 [0x122A0]={
  category="lo",
  direction="l",
 },
 [0x122A1]={
  category="lo",
  direction="l",
 },
 [0x122A2]={
  category="lo",
  direction="l",
 },
 [0x122A3]={
  category="lo",
  direction="l",
 },
 [0x122A4]={
  category="lo",
  direction="l",
 },
 [0x122A5]={
  category="lo",
  direction="l",
 },
 [0x122A6]={
  category="lo",
  direction="l",
 },
 [0x122A7]={
  category="lo",
  direction="l",
 },
 [0x122A8]={
  category="lo",
  direction="l",
 },
 [0x122A9]={
  category="lo",
  direction="l",
 },
 [0x122AA]={
  category="lo",
  direction="l",
 },
 [0x122AB]={
  category="lo",
  direction="l",
 },
 [0x122AC]={
  category="lo",
  direction="l",
 },
 [0x122AD]={
  category="lo",
  direction="l",
 },
 [0x122AE]={
  category="lo",
  direction="l",
 },
 [0x122AF]={
  category="lo",
  direction="l",
 },
 [0x122B0]={
  category="lo",
  direction="l",
 },
 [0x122B1]={
  category="lo",
  direction="l",
 },
 [0x122B2]={
  category="lo",
  direction="l",
 },
 [0x122B3]={
  category="lo",
  direction="l",
 },
 [0x122B4]={
  category="lo",
  direction="l",
 },
 [0x122B5]={
  category="lo",
  direction="l",
 },
 [0x122B6]={
  category="lo",
  direction="l",
 },
 [0x122B7]={
  category="lo",
  direction="l",
 },
 [0x122B8]={
  category="lo",
  direction="l",
 },
 [0x122B9]={
  category="lo",
  direction="l",
 },
 [0x122BA]={
  category="lo",
  direction="l",
 },
 [0x122BB]={
  category="lo",
  direction="l",
 },
 [0x122BC]={
  category="lo",
  direction="l",
 },
 [0x122BD]={
  category="lo",
  direction="l",
 },
 [0x122BE]={
  category="lo",
  direction="l",
 },
 [0x122BF]={
  category="lo",
  direction="l",
 },
 [0x122C0]={
  category="lo",
  direction="l",
 },
 [0x122C1]={
  category="lo",
  direction="l",
 },
 [0x122C2]={
  category="lo",
  direction="l",
 },
 [0x122C3]={
  category="lo",
  direction="l",
 },
 [0x122C4]={
  category="lo",
  direction="l",
 },
 [0x122C5]={
  category="lo",
  direction="l",
 },
 [0x122C6]={
  category="lo",
  direction="l",
 },
 [0x122C7]={
  category="lo",
  direction="l",
 },
 [0x122C8]={
  category="lo",
  direction="l",
 },
 [0x122C9]={
  category="lo",
  direction="l",
 },
 [0x122CA]={
  category="lo",
  direction="l",
 },
 [0x122CB]={
  category="lo",
  direction="l",
 },
 [0x122CC]={
  category="lo",
  direction="l",
 },
 [0x122CD]={
  category="lo",
  direction="l",
 },
 [0x122CE]={
  category="lo",
  direction="l",
 },
 [0x122CF]={
  category="lo",
  direction="l",
 },
 [0x122D0]={
  category="lo",
  direction="l",
 },
 [0x122D1]={
  category="lo",
  direction="l",
 },
 [0x122D2]={
  category="lo",
  direction="l",
 },
 [0x122D3]={
  category="lo",
  direction="l",
 },
 [0x122D4]={
  category="lo",
  direction="l",
 },
 [0x122D5]={
  category="lo",
  direction="l",
 },
 [0x122D6]={
  category="lo",
  direction="l",
 },
 [0x122D7]={
  category="lo",
  direction="l",
 },
 [0x122D8]={
  category="lo",
  direction="l",
 },
 [0x122D9]={
  category="lo",
  direction="l",
 },
 [0x122DA]={
  category="lo",
  direction="l",
 },
 [0x122DB]={
  category="lo",
  direction="l",
 },
 [0x122DC]={
  category="lo",
  direction="l",
 },
 [0x122DD]={
  category="lo",
  direction="l",
 },
 [0x122DE]={
  category="lo",
  direction="l",
 },
 [0x122DF]={
  category="lo",
  direction="l",
 },
 [0x122E0]={
  category="lo",
  direction="l",
 },
 [0x122E1]={
  category="lo",
  direction="l",
 },
 [0x122E2]={
  category="lo",
  direction="l",
 },
 [0x122E3]={
  category="lo",
  direction="l",
 },
 [0x122E4]={
  category="lo",
  direction="l",
 },
 [0x122E5]={
  category="lo",
  direction="l",
 },
 [0x122E6]={
  category="lo",
  direction="l",
 },
 [0x122E7]={
  category="lo",
  direction="l",
 },
 [0x122E8]={
  category="lo",
  direction="l",
 },
 [0x122E9]={
  category="lo",
  direction="l",
 },
 [0x122EA]={
  category="lo",
  direction="l",
 },
 [0x122EB]={
  category="lo",
  direction="l",
 },
 [0x122EC]={
  category="lo",
  direction="l",
 },
 [0x122ED]={
  category="lo",
  direction="l",
 },
 [0x122EE]={
  category="lo",
  direction="l",
 },
 [0x122EF]={
  category="lo",
  direction="l",
 },
 [0x122F0]={
  category="lo",
  direction="l",
 },
 [0x122F1]={
  category="lo",
  direction="l",
 },
 [0x122F2]={
  category="lo",
  direction="l",
 },
 [0x122F3]={
  category="lo",
  direction="l",
 },
 [0x122F4]={
  category="lo",
  direction="l",
 },
 [0x122F5]={
  category="lo",
  direction="l",
 },
 [0x122F6]={
  category="lo",
  direction="l",
 },
 [0x122F7]={
  category="lo",
  direction="l",
 },
 [0x122F8]={
  category="lo",
  direction="l",
 },
 [0x122F9]={
  category="lo",
  direction="l",
 },
 [0x122FA]={
  category="lo",
  direction="l",
 },
 [0x122FB]={
  category="lo",
  direction="l",
 },
 [0x122FC]={
  category="lo",
  direction="l",
 },
 [0x122FD]={
  category="lo",
  direction="l",
 },
 [0x122FE]={
  category="lo",
  direction="l",
 },
 [0x122FF]={
  category="lo",
  direction="l",
 },
 [0x12300]={
  category="lo",
  direction="l",
 },
 [0x12301]={
  category="lo",
  direction="l",
 },
 [0x12302]={
  category="lo",
  direction="l",
 },
 [0x12303]={
  category="lo",
  direction="l",
 },
 [0x12304]={
  category="lo",
  direction="l",
 },
 [0x12305]={
  category="lo",
  direction="l",
 },
 [0x12306]={
  category="lo",
  direction="l",
 },
 [0x12307]={
  category="lo",
  direction="l",
 },
 [0x12308]={
  category="lo",
  direction="l",
 },
 [0x12309]={
  category="lo",
  direction="l",
 },
 [0x1230A]={
  category="lo",
  direction="l",
 },
 [0x1230B]={
  category="lo",
  direction="l",
 },
 [0x1230C]={
  category="lo",
  direction="l",
 },
 [0x1230D]={
  category="lo",
  direction="l",
 },
 [0x1230E]={
  category="lo",
  direction="l",
 },
 [0x1230F]={
  category="lo",
  direction="l",
 },
 [0x12310]={
  category="lo",
  direction="l",
 },
 [0x12311]={
  category="lo",
  direction="l",
 },
 [0x12312]={
  category="lo",
  direction="l",
 },
 [0x12313]={
  category="lo",
  direction="l",
 },
 [0x12314]={
  category="lo",
  direction="l",
 },
 [0x12315]={
  category="lo",
  direction="l",
 },
 [0x12316]={
  category="lo",
  direction="l",
 },
 [0x12317]={
  category="lo",
  direction="l",
 },
 [0x12318]={
  category="lo",
  direction="l",
 },
 [0x12319]={
  category="lo",
  direction="l",
 },
 [0x1231A]={
  category="lo",
  direction="l",
 },
 [0x1231B]={
  category="lo",
  direction="l",
 },
 [0x1231C]={
  category="lo",
  direction="l",
 },
 [0x1231D]={
  category="lo",
  direction="l",
 },
 [0x1231E]={
  category="lo",
  direction="l",
 },
 [0x1231F]={
  category="lo",
  direction="l",
 },
 [0x12320]={
  category="lo",
  direction="l",
 },
 [0x12321]={
  category="lo",
  direction="l",
 },
 [0x12322]={
  category="lo",
  direction="l",
 },
 [0x12323]={
  category="lo",
  direction="l",
 },
 [0x12324]={
  category="lo",
  direction="l",
 },
 [0x12325]={
  category="lo",
  direction="l",
 },
 [0x12326]={
  category="lo",
  direction="l",
 },
 [0x12327]={
  category="lo",
  direction="l",
 },
 [0x12328]={
  category="lo",
  direction="l",
 },
 [0x12329]={
  category="lo",
  direction="l",
 },
 [0x1232A]={
  category="lo",
  direction="l",
 },
 [0x1232B]={
  category="lo",
  direction="l",
 },
 [0x1232C]={
  category="lo",
  direction="l",
 },
 [0x1232D]={
  category="lo",
  direction="l",
 },
 [0x1232E]={
  category="lo",
  direction="l",
 },
 [0x1232F]={
  category="lo",
  direction="l",
 },
 [0x12330]={
  category="lo",
  direction="l",
 },
 [0x12331]={
  category="lo",
  direction="l",
 },
 [0x12332]={
  category="lo",
  direction="l",
 },
 [0x12333]={
  category="lo",
  direction="l",
 },
 [0x12334]={
  category="lo",
  direction="l",
 },
 [0x12335]={
  category="lo",
  direction="l",
 },
 [0x12336]={
  category="lo",
  direction="l",
 },
 [0x12337]={
  category="lo",
  direction="l",
 },
 [0x12338]={
  category="lo",
  direction="l",
 },
 [0x12339]={
  category="lo",
  direction="l",
 },
 [0x1233A]={
  category="lo",
  direction="l",
 },
 [0x1233B]={
  category="lo",
  direction="l",
 },
 [0x1233C]={
  category="lo",
  direction="l",
 },
 [0x1233D]={
  category="lo",
  direction="l",
 },
 [0x1233E]={
  category="lo",
  direction="l",
 },
 [0x1233F]={
  category="lo",
  direction="l",
 },
 [0x12340]={
  category="lo",
  direction="l",
 },
 [0x12341]={
  category="lo",
  direction="l",
 },
 [0x12342]={
  category="lo",
  direction="l",
 },
 [0x12343]={
  category="lo",
  direction="l",
 },
 [0x12344]={
  category="lo",
  direction="l",
 },
 [0x12345]={
  category="lo",
  direction="l",
 },
 [0x12346]={
  category="lo",
  direction="l",
 },
 [0x12347]={
  category="lo",
  direction="l",
 },
 [0x12348]={
  category="lo",
  direction="l",
 },
 [0x12349]={
  category="lo",
  direction="l",
 },
 [0x1234A]={
  category="lo",
  direction="l",
 },
 [0x1234B]={
  category="lo",
  direction="l",
 },
 [0x1234C]={
  category="lo",
  direction="l",
 },
 [0x1234D]={
  category="lo",
  direction="l",
 },
 [0x1234E]={
  category="lo",
  direction="l",
 },
 [0x1234F]={
  category="lo",
  direction="l",
 },
 [0x12350]={
  category="lo",
  direction="l",
 },
 [0x12351]={
  category="lo",
  direction="l",
 },
 [0x12352]={
  category="lo",
  direction="l",
 },
 [0x12353]={
  category="lo",
  direction="l",
 },
 [0x12354]={
  category="lo",
  direction="l",
 },
 [0x12355]={
  category="lo",
  direction="l",
 },
 [0x12356]={
  category="lo",
  direction="l",
 },
 [0x12357]={
  category="lo",
  direction="l",
 },
 [0x12358]={
  category="lo",
  direction="l",
 },
 [0x12359]={
  category="lo",
  direction="l",
 },
 [0x1235A]={
  category="lo",
  direction="l",
 },
 [0x1235B]={
  category="lo",
  direction="l",
 },
 [0x1235C]={
  category="lo",
  direction="l",
 },
 [0x1235D]={
  category="lo",
  direction="l",
 },
 [0x1235E]={
  category="lo",
  direction="l",
 },
 [0x1235F]={
  category="lo",
  direction="l",
 },
 [0x12360]={
  category="lo",
  direction="l",
 },
 [0x12361]={
  category="lo",
  direction="l",
 },
 [0x12362]={
  category="lo",
  direction="l",
 },
 [0x12363]={
  category="lo",
  direction="l",
 },
 [0x12364]={
  category="lo",
  direction="l",
 },
 [0x12365]={
  category="lo",
  direction="l",
 },
 [0x12366]={
  category="lo",
  direction="l",
 },
 [0x12367]={
  category="lo",
  direction="l",
 },
 [0x12368]={
  category="lo",
  direction="l",
 },
 [0x12369]={
  category="lo",
  direction="l",
 },
 [0x1236A]={
  category="lo",
  direction="l",
 },
 [0x1236B]={
  category="lo",
  direction="l",
 },
 [0x1236C]={
  category="lo",
  direction="l",
 },
 [0x1236D]={
  category="lo",
  direction="l",
 },
 [0x1236E]={
  category="lo",
  direction="l",
 },
 [0x1236F]={
  category="lo",
  direction="l",
 },
 [0x12370]={
  category="lo",
  direction="l",
 },
 [0x12371]={
  category="lo",
  direction="l",
 },
 [0x12372]={
  category="lo",
  direction="l",
 },
 [0x12373]={
  category="lo",
  direction="l",
 },
 [0x12374]={
  category="lo",
  direction="l",
 },
 [0x12375]={
  category="lo",
  direction="l",
 },
 [0x12376]={
  category="lo",
  direction="l",
 },
 [0x12377]={
  category="lo",
  direction="l",
 },
 [0x12378]={
  category="lo",
  direction="l",
 },
 [0x12379]={
  category="lo",
  direction="l",
 },
 [0x1237A]={
  category="lo",
  direction="l",
 },
 [0x1237B]={
  category="lo",
  direction="l",
 },
 [0x1237C]={
  category="lo",
  direction="l",
 },
 [0x1237D]={
  category="lo",
  direction="l",
 },
 [0x1237E]={
  category="lo",
  direction="l",
 },
 [0x1237F]={
  category="lo",
  direction="l",
 },
 [0x12380]={
  category="lo",
  direction="l",
 },
 [0x12381]={
  category="lo",
  direction="l",
 },
 [0x12382]={
  category="lo",
  direction="l",
 },
 [0x12383]={
  category="lo",
  direction="l",
 },
 [0x12384]={
  category="lo",
  direction="l",
 },
 [0x12385]={
  category="lo",
  direction="l",
 },
 [0x12386]={
  category="lo",
  direction="l",
 },
 [0x12387]={
  category="lo",
  direction="l",
 },
 [0x12388]={
  category="lo",
  direction="l",
 },
 [0x12389]={
  category="lo",
  direction="l",
 },
 [0x1238A]={
  category="lo",
  direction="l",
 },
 [0x1238B]={
  category="lo",
  direction="l",
 },
 [0x1238C]={
  category="lo",
  direction="l",
 },
 [0x1238D]={
  category="lo",
  direction="l",
 },
 [0x1238E]={
  category="lo",
  direction="l",
 },
 [0x1238F]={
  category="lo",
  direction="l",
 },
 [0x12390]={
  category="lo",
  direction="l",
 },
 [0x12391]={
  category="lo",
  direction="l",
 },
 [0x12392]={
  category="lo",
  direction="l",
 },
 [0x12393]={
  category="lo",
  direction="l",
 },
 [0x12394]={
  category="lo",
  direction="l",
 },
 [0x12395]={
  category="lo",
  direction="l",
 },
 [0x12396]={
  category="lo",
  direction="l",
 },
 [0x12397]={
  category="lo",
  direction="l",
 },
 [0x12398]={
  category="lo",
  direction="l",
 },
 [0x12400]={
  category="nl",
  direction="l",
 },
 [0x12401]={
  category="nl",
  direction="l",
 },
 [0x12402]={
  category="nl",
  direction="l",
 },
 [0x12403]={
  category="nl",
  direction="l",
 },
 [0x12404]={
  category="nl",
  direction="l",
 },
 [0x12405]={
  category="nl",
  direction="l",
 },
 [0x12406]={
  category="nl",
  direction="l",
 },
 [0x12407]={
  category="nl",
  direction="l",
 },
 [0x12408]={
  category="nl",
  direction="l",
 },
 [0x12409]={
  category="nl",
  direction="l",
 },
 [0x1240A]={
  category="nl",
  direction="l",
 },
 [0x1240B]={
  category="nl",
  direction="l",
 },
 [0x1240C]={
  category="nl",
  direction="l",
 },
 [0x1240D]={
  category="nl",
  direction="l",
 },
 [0x1240E]={
  category="nl",
  direction="l",
 },
 [0x1240F]={
  category="nl",
  direction="l",
 },
 [0x12410]={
  category="nl",
  direction="l",
 },
 [0x12411]={
  category="nl",
  direction="l",
 },
 [0x12412]={
  category="nl",
  direction="l",
 },
 [0x12413]={
  category="nl",
  direction="l",
 },
 [0x12414]={
  category="nl",
  direction="l",
 },
 [0x12415]={
  category="nl",
  direction="l",
 },
 [0x12416]={
  category="nl",
  direction="l",
 },
 [0x12417]={
  category="nl",
  direction="l",
 },
 [0x12418]={
  category="nl",
  direction="l",
 },
 [0x12419]={
  category="nl",
  direction="l",
 },
 [0x1241A]={
  category="nl",
  direction="l",
 },
 [0x1241B]={
  category="nl",
  direction="l",
 },
 [0x1241C]={
  category="nl",
  direction="l",
 },
 [0x1241D]={
  category="nl",
  direction="l",
 },
 [0x1241E]={
  category="nl",
  direction="l",
 },
 [0x1241F]={
  category="nl",
  direction="l",
 },
 [0x12420]={
  category="nl",
  direction="l",
 },
 [0x12421]={
  category="nl",
  direction="l",
 },
 [0x12422]={
  category="nl",
  direction="l",
 },
 [0x12423]={
  category="nl",
  direction="l",
 },
 [0x12424]={
  category="nl",
  direction="l",
 },
 [0x12425]={
  category="nl",
  direction="l",
 },
 [0x12426]={
  category="nl",
  direction="l",
 },
 [0x12427]={
  category="nl",
  direction="l",
 },
 [0x12428]={
  category="nl",
  direction="l",
 },
 [0x12429]={
  category="nl",
  direction="l",
 },
 [0x1242A]={
  category="nl",
  direction="l",
 },
 [0x1242B]={
  category="nl",
  direction="l",
 },
 [0x1242C]={
  category="nl",
  direction="l",
 },
 [0x1242D]={
  category="nl",
  direction="l",
 },
 [0x1242E]={
  category="nl",
  direction="l",
 },
 [0x1242F]={
  category="nl",
  direction="l",
 },
 [0x12430]={
  category="nl",
  direction="l",
 },
 [0x12431]={
  category="nl",
  direction="l",
 },
 [0x12432]={
  category="nl",
  direction="l",
 },
 [0x12433]={
  category="nl",
  direction="l",
 },
 [0x12434]={
  category="nl",
  direction="l",
 },
 [0x12435]={
  category="nl",
  direction="l",
 },
 [0x12436]={
  category="nl",
  direction="l",
 },
 [0x12437]={
  category="nl",
  direction="l",
 },
 [0x12438]={
  category="nl",
  direction="l",
 },
 [0x12439]={
  category="nl",
  direction="l",
 },
 [0x1243A]={
  category="nl",
  direction="l",
 },
 [0x1243B]={
  category="nl",
  direction="l",
 },
 [0x1243C]={
  category="nl",
  direction="l",
 },
 [0x1243D]={
  category="nl",
  direction="l",
 },
 [0x1243E]={
  category="nl",
  direction="l",
 },
 [0x1243F]={
  category="nl",
  direction="l",
 },
 [0x12440]={
  category="nl",
  direction="l",
 },
 [0x12441]={
  category="nl",
  direction="l",
 },
 [0x12442]={
  category="nl",
  direction="l",
 },
 [0x12443]={
  category="nl",
  direction="l",
 },
 [0x12444]={
  category="nl",
  direction="l",
 },
 [0x12445]={
  category="nl",
  direction="l",
 },
 [0x12446]={
  category="nl",
  direction="l",
 },
 [0x12447]={
  category="nl",
  direction="l",
 },
 [0x12448]={
  category="nl",
  direction="l",
 },
 [0x12449]={
  category="nl",
  direction="l",
 },
 [0x1244A]={
  category="nl",
  direction="l",
 },
 [0x1244B]={
  category="nl",
  direction="l",
 },
 [0x1244C]={
  category="nl",
  direction="l",
 },
 [0x1244D]={
  category="nl",
  direction="l",
 },
 [0x1244E]={
  category="nl",
  direction="l",
 },
 [0x1244F]={
  category="nl",
  direction="l",
 },
 [0x12450]={
  category="nl",
  direction="l",
 },
 [0x12451]={
  category="nl",
  direction="l",
 },
 [0x12452]={
  category="nl",
  direction="l",
 },
 [0x12453]={
  category="nl",
  direction="l",
 },
 [0x12454]={
  category="nl",
  direction="l",
 },
 [0x12455]={
  category="nl",
  direction="l",
 },
 [0x12456]={
  category="nl",
  direction="l",
 },
 [0x12457]={
  category="nl",
  direction="l",
 },
 [0x12458]={
  category="nl",
  direction="l",
 },
 [0x12459]={
  category="nl",
  direction="l",
 },
 [0x1245A]={
  category="nl",
  direction="l",
 },
 [0x1245B]={
  category="nl",
  direction="l",
 },
 [0x1245C]={
  category="nl",
  direction="l",
 },
 [0x1245D]={
  category="nl",
  direction="l",
 },
 [0x1245E]={
  category="nl",
  direction="l",
 },
 [0x1245F]={
  category="nl",
  direction="l",
 },
 [0x12460]={
  category="nl",
  direction="l",
 },
 [0x12461]={
  category="nl",
  direction="l",
 },
 [0x12462]={
  category="nl",
  direction="l",
 },
 [0x12463]={
  category="nl",
  direction="l",
 },
 [0x12464]={
  category="nl",
  direction="l",
 },
 [0x12465]={
  category="nl",
  direction="l",
 },
 [0x12466]={
  category="nl",
  direction="l",
 },
 [0x12467]={
  category="nl",
  direction="l",
 },
 [0x12468]={
  category="nl",
  direction="l",
 },
 [0x12469]={
  category="nl",
  direction="l",
 },
 [0x1246A]={
  category="nl",
  direction="l",
 },
 [0x1246B]={
  category="nl",
  direction="l",
 },
 [0x1246C]={
  category="nl",
  direction="l",
 },
 [0x1246D]={
  category="nl",
  direction="l",
 },
 [0x1246E]={
  category="nl",
  direction="l",
 },
 [0x12470]={
  category="po",
  direction="l",
 },
 [0x12471]={
  category="po",
  direction="l",
 },
 [0x12472]={
  category="po",
  direction="l",
 },
 [0x12473]={
  category="po",
  direction="l",
 },
 [0x12474]={
  category="po",
  direction="l",
 },
 [0x13000]={
  category="lo",
  direction="l",
 },
 [0x13001]={
  category="lo",
  direction="l",
 },
 [0x13002]={
  category="lo",
  direction="l",
 },
 [0x13003]={
  category="lo",
  direction="l",
 },
 [0x13004]={
  category="lo",
  direction="l",
 },
 [0x13005]={
  category="lo",
  direction="l",
 },
 [0x13006]={
  category="lo",
  direction="l",
 },
 [0x13007]={
  category="lo",
  direction="l",
 },
 [0x13008]={
  category="lo",
  direction="l",
 },
 [0x13009]={
  category="lo",
  direction="l",
 },
 [0x1300A]={
  category="lo",
  direction="l",
 },
 [0x1300B]={
  category="lo",
  direction="l",
 },
 [0x1300C]={
  category="lo",
  direction="l",
 },
 [0x1300D]={
  category="lo",
  direction="l",
 },
 [0x1300E]={
  category="lo",
  direction="l",
 },
 [0x1300F]={
  category="lo",
  direction="l",
 },
 [0x13010]={
  category="lo",
  direction="l",
 },
 [0x13011]={
  category="lo",
  direction="l",
 },
 [0x13012]={
  category="lo",
  direction="l",
 },
 [0x13013]={
  category="lo",
  direction="l",
 },
 [0x13014]={
  category="lo",
  direction="l",
 },
 [0x13015]={
  category="lo",
  direction="l",
 },
 [0x13016]={
  category="lo",
  direction="l",
 },
 [0x13017]={
  category="lo",
  direction="l",
 },
 [0x13018]={
  category="lo",
  direction="l",
 },
 [0x13019]={
  category="lo",
  direction="l",
 },
 [0x1301A]={
  category="lo",
  direction="l",
 },
 [0x1301B]={
  category="lo",
  direction="l",
 },
 [0x1301C]={
  category="lo",
  direction="l",
 },
 [0x1301D]={
  category="lo",
  direction="l",
 },
 [0x1301E]={
  category="lo",
  direction="l",
 },
 [0x1301F]={
  category="lo",
  direction="l",
 },
 [0x13020]={
  category="lo",
  direction="l",
 },
 [0x13021]={
  category="lo",
  direction="l",
 },
 [0x13022]={
  category="lo",
  direction="l",
 },
 [0x13023]={
  category="lo",
  direction="l",
 },
 [0x13024]={
  category="lo",
  direction="l",
 },
 [0x13025]={
  category="lo",
  direction="l",
 },
 [0x13026]={
  category="lo",
  direction="l",
 },
 [0x13027]={
  category="lo",
  direction="l",
 },
 [0x13028]={
  category="lo",
  direction="l",
 },
 [0x13029]={
  category="lo",
  direction="l",
 },
 [0x1302A]={
  category="lo",
  direction="l",
 },
 [0x1302B]={
  category="lo",
  direction="l",
 },
 [0x1302C]={
  category="lo",
  direction="l",
 },
 [0x1302D]={
  category="lo",
  direction="l",
 },
 [0x1302E]={
  category="lo",
  direction="l",
 },
 [0x1302F]={
  category="lo",
  direction="l",
 },
 [0x13030]={
  category="lo",
  direction="l",
 },
 [0x13031]={
  category="lo",
  direction="l",
 },
 [0x13032]={
  category="lo",
  direction="l",
 },
 [0x13033]={
  category="lo",
  direction="l",
 },
 [0x13034]={
  category="lo",
  direction="l",
 },
 [0x13035]={
  category="lo",
  direction="l",
 },
 [0x13036]={
  category="lo",
  direction="l",
 },
 [0x13037]={
  category="lo",
  direction="l",
 },
 [0x13038]={
  category="lo",
  direction="l",
 },
 [0x13039]={
  category="lo",
  direction="l",
 },
 [0x1303A]={
  category="lo",
  direction="l",
 },
 [0x1303B]={
  category="lo",
  direction="l",
 },
 [0x1303C]={
  category="lo",
  direction="l",
 },
 [0x1303D]={
  category="lo",
  direction="l",
 },
 [0x1303E]={
  category="lo",
  direction="l",
 },
 [0x1303F]={
  category="lo",
  direction="l",
 },
 [0x13040]={
  category="lo",
  direction="l",
 },
 [0x13041]={
  category="lo",
  direction="l",
 },
 [0x13042]={
  category="lo",
  direction="l",
 },
 [0x13043]={
  category="lo",
  direction="l",
 },
 [0x13044]={
  category="lo",
  direction="l",
 },
 [0x13045]={
  category="lo",
  direction="l",
 },
 [0x13046]={
  category="lo",
  direction="l",
 },
 [0x13047]={
  category="lo",
  direction="l",
 },
 [0x13048]={
  category="lo",
  direction="l",
 },
 [0x13049]={
  category="lo",
  direction="l",
 },
 [0x1304A]={
  category="lo",
  direction="l",
 },
 [0x1304B]={
  category="lo",
  direction="l",
 },
 [0x1304C]={
  category="lo",
  direction="l",
 },
 [0x1304D]={
  category="lo",
  direction="l",
 },
 [0x1304E]={
  category="lo",
  direction="l",
 },
 [0x1304F]={
  category="lo",
  direction="l",
 },
 [0x13050]={
  category="lo",
  direction="l",
 },
 [0x13051]={
  category="lo",
  direction="l",
 },
 [0x13052]={
  category="lo",
  direction="l",
 },
 [0x13053]={
  category="lo",
  direction="l",
 },
 [0x13054]={
  category="lo",
  direction="l",
 },
 [0x13055]={
  category="lo",
  direction="l",
 },
 [0x13056]={
  category="lo",
  direction="l",
 },
 [0x13057]={
  category="lo",
  direction="l",
 },
 [0x13058]={
  category="lo",
  direction="l",
 },
 [0x13059]={
  category="lo",
  direction="l",
 },
 [0x1305A]={
  category="lo",
  direction="l",
 },
 [0x1305B]={
  category="lo",
  direction="l",
 },
 [0x1305C]={
  category="lo",
  direction="l",
 },
 [0x1305D]={
  category="lo",
  direction="l",
 },
 [0x1305E]={
  category="lo",
  direction="l",
 },
 [0x1305F]={
  category="lo",
  direction="l",
 },
 [0x13060]={
  category="lo",
  direction="l",
 },
 [0x13061]={
  category="lo",
  direction="l",
 },
 [0x13062]={
  category="lo",
  direction="l",
 },
 [0x13063]={
  category="lo",
  direction="l",
 },
 [0x13064]={
  category="lo",
  direction="l",
 },
 [0x13065]={
  category="lo",
  direction="l",
 },
 [0x13066]={
  category="lo",
  direction="l",
 },
 [0x13067]={
  category="lo",
  direction="l",
 },
 [0x13068]={
  category="lo",
  direction="l",
 },
 [0x13069]={
  category="lo",
  direction="l",
 },
 [0x1306A]={
  category="lo",
  direction="l",
 },
 [0x1306B]={
  category="lo",
  direction="l",
 },
 [0x1306C]={
  category="lo",
  direction="l",
 },
 [0x1306D]={
  category="lo",
  direction="l",
 },
 [0x1306E]={
  category="lo",
  direction="l",
 },
 [0x1306F]={
  category="lo",
  direction="l",
 },
 [0x13070]={
  category="lo",
  direction="l",
 },
 [0x13071]={
  category="lo",
  direction="l",
 },
 [0x13072]={
  category="lo",
  direction="l",
 },
 [0x13073]={
  category="lo",
  direction="l",
 },
 [0x13074]={
  category="lo",
  direction="l",
 },
 [0x13075]={
  category="lo",
  direction="l",
 },
 [0x13076]={
  category="lo",
  direction="l",
 },
 [0x13077]={
  category="lo",
  direction="l",
 },
 [0x13078]={
  category="lo",
  direction="l",
 },
 [0x13079]={
  category="lo",
  direction="l",
 },
 [0x1307A]={
  category="lo",
  direction="l",
 },
 [0x1307B]={
  category="lo",
  direction="l",
 },
 [0x1307C]={
  category="lo",
  direction="l",
 },
 [0x1307D]={
  category="lo",
  direction="l",
 },
 [0x1307E]={
  category="lo",
  direction="l",
 },
 [0x1307F]={
  category="lo",
  direction="l",
 },
 [0x13080]={
  category="lo",
  direction="l",
 },
 [0x13081]={
  category="lo",
  direction="l",
 },
 [0x13082]={
  category="lo",
  direction="l",
 },
 [0x13083]={
  category="lo",
  direction="l",
 },
 [0x13084]={
  category="lo",
  direction="l",
 },
 [0x13085]={
  category="lo",
  direction="l",
 },
 [0x13086]={
  category="lo",
  direction="l",
 },
 [0x13087]={
  category="lo",
  direction="l",
 },
 [0x13088]={
  category="lo",
  direction="l",
 },
 [0x13089]={
  category="lo",
  direction="l",
 },
 [0x1308A]={
  category="lo",
  direction="l",
 },
 [0x1308B]={
  category="lo",
  direction="l",
 },
 [0x1308C]={
  category="lo",
  direction="l",
 },
 [0x1308D]={
  category="lo",
  direction="l",
 },
 [0x1308E]={
  category="lo",
  direction="l",
 },
 [0x1308F]={
  category="lo",
  direction="l",
 },
 [0x13090]={
  category="lo",
  direction="l",
 },
 [0x13091]={
  category="lo",
  direction="l",
 },
 [0x13092]={
  category="lo",
  direction="l",
 },
 [0x13093]={
  category="lo",
  direction="l",
 },
 [0x13094]={
  category="lo",
  direction="l",
 },
 [0x13095]={
  category="lo",
  direction="l",
 },
 [0x13096]={
  category="lo",
  direction="l",
 },
 [0x13097]={
  category="lo",
  direction="l",
 },
 [0x13098]={
  category="lo",
  direction="l",
 },
 [0x13099]={
  category="lo",
  direction="l",
 },
 [0x1309A]={
  category="lo",
  direction="l",
 },
 [0x1309B]={
  category="lo",
  direction="l",
 },
 [0x1309C]={
  category="lo",
  direction="l",
 },
 [0x1309D]={
  category="lo",
  direction="l",
 },
 [0x1309E]={
  category="lo",
  direction="l",
 },
 [0x1309F]={
  category="lo",
  direction="l",
 },
 [0x130A0]={
  category="lo",
  direction="l",
 },
 [0x130A1]={
  category="lo",
  direction="l",
 },
 [0x130A2]={
  category="lo",
  direction="l",
 },
 [0x130A3]={
  category="lo",
  direction="l",
 },
 [0x130A4]={
  category="lo",
  direction="l",
 },
 [0x130A5]={
  category="lo",
  direction="l",
 },
 [0x130A6]={
  category="lo",
  direction="l",
 },
 [0x130A7]={
  category="lo",
  direction="l",
 },
 [0x130A8]={
  category="lo",
  direction="l",
 },
 [0x130A9]={
  category="lo",
  direction="l",
 },
 [0x130AA]={
  category="lo",
  direction="l",
 },
 [0x130AB]={
  category="lo",
  direction="l",
 },
 [0x130AC]={
  category="lo",
  direction="l",
 },
 [0x130AD]={
  category="lo",
  direction="l",
 },
 [0x130AE]={
  category="lo",
  direction="l",
 },
 [0x130AF]={
  category="lo",
  direction="l",
 },
 [0x130B0]={
  category="lo",
  direction="l",
 },
 [0x130B1]={
  category="lo",
  direction="l",
 },
 [0x130B2]={
  category="lo",
  direction="l",
 },
 [0x130B3]={
  category="lo",
  direction="l",
 },
 [0x130B4]={
  category="lo",
  direction="l",
 },
 [0x130B5]={
  category="lo",
  direction="l",
 },
 [0x130B6]={
  category="lo",
  direction="l",
 },
 [0x130B7]={
  category="lo",
  direction="l",
 },
 [0x130B8]={
  category="lo",
  direction="l",
 },
 [0x130B9]={
  category="lo",
  direction="l",
 },
 [0x130BA]={
  category="lo",
  direction="l",
 },
 [0x130BB]={
  category="lo",
  direction="l",
 },
 [0x130BC]={
  category="lo",
  direction="l",
 },
 [0x130BD]={
  category="lo",
  direction="l",
 },
 [0x130BE]={
  category="lo",
  direction="l",
 },
 [0x130BF]={
  category="lo",
  direction="l",
 },
 [0x130C0]={
  category="lo",
  direction="l",
 },
 [0x130C1]={
  category="lo",
  direction="l",
 },
 [0x130C2]={
  category="lo",
  direction="l",
 },
 [0x130C3]={
  category="lo",
  direction="l",
 },
 [0x130C4]={
  category="lo",
  direction="l",
 },
 [0x130C5]={
  category="lo",
  direction="l",
 },
 [0x130C6]={
  category="lo",
  direction="l",
 },
 [0x130C7]={
  category="lo",
  direction="l",
 },
 [0x130C8]={
  category="lo",
  direction="l",
 },
 [0x130C9]={
  category="lo",
  direction="l",
 },
 [0x130CA]={
  category="lo",
  direction="l",
 },
 [0x130CB]={
  category="lo",
  direction="l",
 },
 [0x130CC]={
  category="lo",
  direction="l",
 },
 [0x130CD]={
  category="lo",
  direction="l",
 },
 [0x130CE]={
  category="lo",
  direction="l",
 },
 [0x130CF]={
  category="lo",
  direction="l",
 },
 [0x130D0]={
  category="lo",
  direction="l",
 },
 [0x130D1]={
  category="lo",
  direction="l",
 },
 [0x130D2]={
  category="lo",
  direction="l",
 },
 [0x130D3]={
  category="lo",
  direction="l",
 },
 [0x130D4]={
  category="lo",
  direction="l",
 },
 [0x130D5]={
  category="lo",
  direction="l",
 },
 [0x130D6]={
  category="lo",
  direction="l",
 },
 [0x130D7]={
  category="lo",
  direction="l",
 },
 [0x130D8]={
  category="lo",
  direction="l",
 },
 [0x130D9]={
  category="lo",
  direction="l",
 },
 [0x130DA]={
  category="lo",
  direction="l",
 },
 [0x130DB]={
  category="lo",
  direction="l",
 },
 [0x130DC]={
  category="lo",
  direction="l",
 },
 [0x130DD]={
  category="lo",
  direction="l",
 },
 [0x130DE]={
  category="lo",
  direction="l",
 },
 [0x130DF]={
  category="lo",
  direction="l",
 },
 [0x130E0]={
  category="lo",
  direction="l",
 },
 [0x130E1]={
  category="lo",
  direction="l",
 },
 [0x130E2]={
  category="lo",
  direction="l",
 },
 [0x130E3]={
  category="lo",
  direction="l",
 },
 [0x130E4]={
  category="lo",
  direction="l",
 },
 [0x130E5]={
  category="lo",
  direction="l",
 },
 [0x130E6]={
  category="lo",
  direction="l",
 },
 [0x130E7]={
  category="lo",
  direction="l",
 },
 [0x130E8]={
  category="lo",
  direction="l",
 },
 [0x130E9]={
  category="lo",
  direction="l",
 },
 [0x130EA]={
  category="lo",
  direction="l",
 },
 [0x130EB]={
  category="lo",
  direction="l",
 },
 [0x130EC]={
  category="lo",
  direction="l",
 },
 [0x130ED]={
  category="lo",
  direction="l",
 },
 [0x130EE]={
  category="lo",
  direction="l",
 },
 [0x130EF]={
  category="lo",
  direction="l",
 },
 [0x130F0]={
  category="lo",
  direction="l",
 },
 [0x130F1]={
  category="lo",
  direction="l",
 },
 [0x130F2]={
  category="lo",
  direction="l",
 },
 [0x130F3]={
  category="lo",
  direction="l",
 },
 [0x130F4]={
  category="lo",
  direction="l",
 },
 [0x130F5]={
  category="lo",
  direction="l",
 },
 [0x130F6]={
  category="lo",
  direction="l",
 },
 [0x130F7]={
  category="lo",
  direction="l",
 },
 [0x130F8]={
  category="lo",
  direction="l",
 },
 [0x130F9]={
  category="lo",
  direction="l",
 },
 [0x130FA]={
  category="lo",
  direction="l",
 },
 [0x130FB]={
  category="lo",
  direction="l",
 },
 [0x130FC]={
  category="lo",
  direction="l",
 },
 [0x130FD]={
  category="lo",
  direction="l",
 },
 [0x130FE]={
  category="lo",
  direction="l",
 },
 [0x130FF]={
  category="lo",
  direction="l",
 },
 [0x13100]={
  category="lo",
  direction="l",
 },
 [0x13101]={
  category="lo",
  direction="l",
 },
 [0x13102]={
  category="lo",
  direction="l",
 },
 [0x13103]={
  category="lo",
  direction="l",
 },
 [0x13104]={
  category="lo",
  direction="l",
 },
 [0x13105]={
  category="lo",
  direction="l",
 },
 [0x13106]={
  category="lo",
  direction="l",
 },
 [0x13107]={
  category="lo",
  direction="l",
 },
 [0x13108]={
  category="lo",
  direction="l",
 },
 [0x13109]={
  category="lo",
  direction="l",
 },
 [0x1310A]={
  category="lo",
  direction="l",
 },
 [0x1310B]={
  category="lo",
  direction="l",
 },
 [0x1310C]={
  category="lo",
  direction="l",
 },
 [0x1310D]={
  category="lo",
  direction="l",
 },
 [0x1310E]={
  category="lo",
  direction="l",
 },
 [0x1310F]={
  category="lo",
  direction="l",
 },
 [0x13110]={
  category="lo",
  direction="l",
 },
 [0x13111]={
  category="lo",
  direction="l",
 },
 [0x13112]={
  category="lo",
  direction="l",
 },
 [0x13113]={
  category="lo",
  direction="l",
 },
 [0x13114]={
  category="lo",
  direction="l",
 },
 [0x13115]={
  category="lo",
  direction="l",
 },
 [0x13116]={
  category="lo",
  direction="l",
 },
 [0x13117]={
  category="lo",
  direction="l",
 },
 [0x13118]={
  category="lo",
  direction="l",
 },
 [0x13119]={
  category="lo",
  direction="l",
 },
 [0x1311A]={
  category="lo",
  direction="l",
 },
 [0x1311B]={
  category="lo",
  direction="l",
 },
 [0x1311C]={
  category="lo",
  direction="l",
 },
 [0x1311D]={
  category="lo",
  direction="l",
 },
 [0x1311E]={
  category="lo",
  direction="l",
 },
 [0x1311F]={
  category="lo",
  direction="l",
 },
 [0x13120]={
  category="lo",
  direction="l",
 },
 [0x13121]={
  category="lo",
  direction="l",
 },
 [0x13122]={
  category="lo",
  direction="l",
 },
 [0x13123]={
  category="lo",
  direction="l",
 },
 [0x13124]={
  category="lo",
  direction="l",
 },
 [0x13125]={
  category="lo",
  direction="l",
 },
 [0x13126]={
  category="lo",
  direction="l",
 },
 [0x13127]={
  category="lo",
  direction="l",
 },
 [0x13128]={
  category="lo",
  direction="l",
 },
 [0x13129]={
  category="lo",
  direction="l",
 },
 [0x1312A]={
  category="lo",
  direction="l",
 },
 [0x1312B]={
  category="lo",
  direction="l",
 },
 [0x1312C]={
  category="lo",
  direction="l",
 },
 [0x1312D]={
  category="lo",
  direction="l",
 },
 [0x1312E]={
  category="lo",
  direction="l",
 },
 [0x1312F]={
  category="lo",
  direction="l",
 },
 [0x13130]={
  category="lo",
  direction="l",
 },
 [0x13131]={
  category="lo",
  direction="l",
 },
 [0x13132]={
  category="lo",
  direction="l",
 },
 [0x13133]={
  category="lo",
  direction="l",
 },
 [0x13134]={
  category="lo",
  direction="l",
 },
 [0x13135]={
  category="lo",
  direction="l",
 },
 [0x13136]={
  category="lo",
  direction="l",
 },
 [0x13137]={
  category="lo",
  direction="l",
 },
 [0x13138]={
  category="lo",
  direction="l",
 },
 [0x13139]={
  category="lo",
  direction="l",
 },
 [0x1313A]={
  category="lo",
  direction="l",
 },
 [0x1313B]={
  category="lo",
  direction="l",
 },
 [0x1313C]={
  category="lo",
  direction="l",
 },
 [0x1313D]={
  category="lo",
  direction="l",
 },
 [0x1313E]={
  category="lo",
  direction="l",
 },
 [0x1313F]={
  category="lo",
  direction="l",
 },
 [0x13140]={
  category="lo",
  direction="l",
 },
 [0x13141]={
  category="lo",
  direction="l",
 },
 [0x13142]={
  category="lo",
  direction="l",
 },
 [0x13143]={
  category="lo",
  direction="l",
 },
 [0x13144]={
  category="lo",
  direction="l",
 },
 [0x13145]={
  category="lo",
  direction="l",
 },
 [0x13146]={
  category="lo",
  direction="l",
 },
 [0x13147]={
  category="lo",
  direction="l",
 },
 [0x13148]={
  category="lo",
  direction="l",
 },
 [0x13149]={
  category="lo",
  direction="l",
 },
 [0x1314A]={
  category="lo",
  direction="l",
 },
 [0x1314B]={
  category="lo",
  direction="l",
 },
 [0x1314C]={
  category="lo",
  direction="l",
 },
 [0x1314D]={
  category="lo",
  direction="l",
 },
 [0x1314E]={
  category="lo",
  direction="l",
 },
 [0x1314F]={
  category="lo",
  direction="l",
 },
 [0x13150]={
  category="lo",
  direction="l",
 },
 [0x13151]={
  category="lo",
  direction="l",
 },
 [0x13152]={
  category="lo",
  direction="l",
 },
 [0x13153]={
  category="lo",
  direction="l",
 },
 [0x13154]={
  category="lo",
  direction="l",
 },
 [0x13155]={
  category="lo",
  direction="l",
 },
 [0x13156]={
  category="lo",
  direction="l",
 },
 [0x13157]={
  category="lo",
  direction="l",
 },
 [0x13158]={
  category="lo",
  direction="l",
 },
 [0x13159]={
  category="lo",
  direction="l",
 },
 [0x1315A]={
  category="lo",
  direction="l",
 },
 [0x1315B]={
  category="lo",
  direction="l",
 },
 [0x1315C]={
  category="lo",
  direction="l",
 },
 [0x1315D]={
  category="lo",
  direction="l",
 },
 [0x1315E]={
  category="lo",
  direction="l",
 },
 [0x1315F]={
  category="lo",
  direction="l",
 },
 [0x13160]={
  category="lo",
  direction="l",
 },
 [0x13161]={
  category="lo",
  direction="l",
 },
 [0x13162]={
  category="lo",
  direction="l",
 },
 [0x13163]={
  category="lo",
  direction="l",
 },
 [0x13164]={
  category="lo",
  direction="l",
 },
 [0x13165]={
  category="lo",
  direction="l",
 },
 [0x13166]={
  category="lo",
  direction="l",
 },
 [0x13167]={
  category="lo",
  direction="l",
 },
 [0x13168]={
  category="lo",
  direction="l",
 },
 [0x13169]={
  category="lo",
  direction="l",
 },
 [0x1316A]={
  category="lo",
  direction="l",
 },
 [0x1316B]={
  category="lo",
  direction="l",
 },
 [0x1316C]={
  category="lo",
  direction="l",
 },
 [0x1316D]={
  category="lo",
  direction="l",
 },
 [0x1316E]={
  category="lo",
  direction="l",
 },
 [0x1316F]={
  category="lo",
  direction="l",
 },
 [0x13170]={
  category="lo",
  direction="l",
 },
 [0x13171]={
  category="lo",
  direction="l",
 },
 [0x13172]={
  category="lo",
  direction="l",
 },
 [0x13173]={
  category="lo",
  direction="l",
 },
 [0x13174]={
  category="lo",
  direction="l",
 },
 [0x13175]={
  category="lo",
  direction="l",
 },
 [0x13176]={
  category="lo",
  direction="l",
 },
 [0x13177]={
  category="lo",
  direction="l",
 },
 [0x13178]={
  category="lo",
  direction="l",
 },
 [0x13179]={
  category="lo",
  direction="l",
 },
 [0x1317A]={
  category="lo",
  direction="l",
 },
 [0x1317B]={
  category="lo",
  direction="l",
 },
 [0x1317C]={
  category="lo",
  direction="l",
 },
 [0x1317D]={
  category="lo",
  direction="l",
 },
 [0x1317E]={
  category="lo",
  direction="l",
 },
 [0x1317F]={
  category="lo",
  direction="l",
 },
 [0x13180]={
  category="lo",
  direction="l",
 },
 [0x13181]={
  category="lo",
  direction="l",
 },
 [0x13182]={
  category="lo",
  direction="l",
 },
 [0x13183]={
  category="lo",
  direction="l",
 },
 [0x13184]={
  category="lo",
  direction="l",
 },
 [0x13185]={
  category="lo",
  direction="l",
 },
 [0x13186]={
  category="lo",
  direction="l",
 },
 [0x13187]={
  category="lo",
  direction="l",
 },
 [0x13188]={
  category="lo",
  direction="l",
 },
 [0x13189]={
  category="lo",
  direction="l",
 },
 [0x1318A]={
  category="lo",
  direction="l",
 },
 [0x1318B]={
  category="lo",
  direction="l",
 },
 [0x1318C]={
  category="lo",
  direction="l",
 },
 [0x1318D]={
  category="lo",
  direction="l",
 },
 [0x1318E]={
  category="lo",
  direction="l",
 },
 [0x1318F]={
  category="lo",
  direction="l",
 },
 [0x13190]={
  category="lo",
  direction="l",
 },
 [0x13191]={
  category="lo",
  direction="l",
 },
 [0x13192]={
  category="lo",
  direction="l",
 },
 [0x13193]={
  category="lo",
  direction="l",
 },
 [0x13194]={
  category="lo",
  direction="l",
 },
 [0x13195]={
  category="lo",
  direction="l",
 },
 [0x13196]={
  category="lo",
  direction="l",
 },
 [0x13197]={
  category="lo",
  direction="l",
 },
 [0x13198]={
  category="lo",
  direction="l",
 },
 [0x13199]={
  category="lo",
  direction="l",
 },
 [0x1319A]={
  category="lo",
  direction="l",
 },
 [0x1319B]={
  category="lo",
  direction="l",
 },
 [0x1319C]={
  category="lo",
  direction="l",
 },
 [0x1319D]={
  category="lo",
  direction="l",
 },
 [0x1319E]={
  category="lo",
  direction="l",
 },
 [0x1319F]={
  category="lo",
  direction="l",
 },
 [0x131A0]={
  category="lo",
  direction="l",
 },
 [0x131A1]={
  category="lo",
  direction="l",
 },
 [0x131A2]={
  category="lo",
  direction="l",
 },
 [0x131A3]={
  category="lo",
  direction="l",
 },
 [0x131A4]={
  category="lo",
  direction="l",
 },
 [0x131A5]={
  category="lo",
  direction="l",
 },
 [0x131A6]={
  category="lo",
  direction="l",
 },
 [0x131A7]={
  category="lo",
  direction="l",
 },
 [0x131A8]={
  category="lo",
  direction="l",
 },
 [0x131A9]={
  category="lo",
  direction="l",
 },
 [0x131AA]={
  category="lo",
  direction="l",
 },
 [0x131AB]={
  category="lo",
  direction="l",
 },
 [0x131AC]={
  category="lo",
  direction="l",
 },
 [0x131AD]={
  category="lo",
  direction="l",
 },
 [0x131AE]={
  category="lo",
  direction="l",
 },
 [0x131AF]={
  category="lo",
  direction="l",
 },
 [0x131B0]={
  category="lo",
  direction="l",
 },
 [0x131B1]={
  category="lo",
  direction="l",
 },
 [0x131B2]={
  category="lo",
  direction="l",
 },
 [0x131B3]={
  category="lo",
  direction="l",
 },
 [0x131B4]={
  category="lo",
  direction="l",
 },
 [0x131B5]={
  category="lo",
  direction="l",
 },
 [0x131B6]={
  category="lo",
  direction="l",
 },
 [0x131B7]={
  category="lo",
  direction="l",
 },
 [0x131B8]={
  category="lo",
  direction="l",
 },
 [0x131B9]={
  category="lo",
  direction="l",
 },
 [0x131BA]={
  category="lo",
  direction="l",
 },
 [0x131BB]={
  category="lo",
  direction="l",
 },
 [0x131BC]={
  category="lo",
  direction="l",
 },
 [0x131BD]={
  category="lo",
  direction="l",
 },
 [0x131BE]={
  category="lo",
  direction="l",
 },
 [0x131BF]={
  category="lo",
  direction="l",
 },
 [0x131C0]={
  category="lo",
  direction="l",
 },
 [0x131C1]={
  category="lo",
  direction="l",
 },
 [0x131C2]={
  category="lo",
  direction="l",
 },
 [0x131C3]={
  category="lo",
  direction="l",
 },
 [0x131C4]={
  category="lo",
  direction="l",
 },
 [0x131C5]={
  category="lo",
  direction="l",
 },
 [0x131C6]={
  category="lo",
  direction="l",
 },
 [0x131C7]={
  category="lo",
  direction="l",
 },
 [0x131C8]={
  category="lo",
  direction="l",
 },
 [0x131C9]={
  category="lo",
  direction="l",
 },
 [0x131CA]={
  category="lo",
  direction="l",
 },
 [0x131CB]={
  category="lo",
  direction="l",
 },
 [0x131CC]={
  category="lo",
  direction="l",
 },
 [0x131CD]={
  category="lo",
  direction="l",
 },
 [0x131CE]={
  category="lo",
  direction="l",
 },
 [0x131CF]={
  category="lo",
  direction="l",
 },
 [0x131D0]={
  category="lo",
  direction="l",
 },
 [0x131D1]={
  category="lo",
  direction="l",
 },
 [0x131D2]={
  category="lo",
  direction="l",
 },
 [0x131D3]={
  category="lo",
  direction="l",
 },
 [0x131D4]={
  category="lo",
  direction="l",
 },
 [0x131D5]={
  category="lo",
  direction="l",
 },
 [0x131D6]={
  category="lo",
  direction="l",
 },
 [0x131D7]={
  category="lo",
  direction="l",
 },
 [0x131D8]={
  category="lo",
  direction="l",
 },
 [0x131D9]={
  category="lo",
  direction="l",
 },
 [0x131DA]={
  category="lo",
  direction="l",
 },
 [0x131DB]={
  category="lo",
  direction="l",
 },
 [0x131DC]={
  category="lo",
  direction="l",
 },
 [0x131DD]={
  category="lo",
  direction="l",
 },
 [0x131DE]={
  category="lo",
  direction="l",
 },
 [0x131DF]={
  category="lo",
  direction="l",
 },
 [0x131E0]={
  category="lo",
  direction="l",
 },
 [0x131E1]={
  category="lo",
  direction="l",
 },
 [0x131E2]={
  category="lo",
  direction="l",
 },
 [0x131E3]={
  category="lo",
  direction="l",
 },
 [0x131E4]={
  category="lo",
  direction="l",
 },
 [0x131E5]={
  category="lo",
  direction="l",
 },
 [0x131E6]={
  category="lo",
  direction="l",
 },
 [0x131E7]={
  category="lo",
  direction="l",
 },
 [0x131E8]={
  category="lo",
  direction="l",
 },
 [0x131E9]={
  category="lo",
  direction="l",
 },
 [0x131EA]={
  category="lo",
  direction="l",
 },
 [0x131EB]={
  category="lo",
  direction="l",
 },
 [0x131EC]={
  category="lo",
  direction="l",
 },
 [0x131ED]={
  category="lo",
  direction="l",
 },
 [0x131EE]={
  category="lo",
  direction="l",
 },
 [0x131EF]={
  category="lo",
  direction="l",
 },
 [0x131F0]={
  category="lo",
  direction="l",
 },
 [0x131F1]={
  category="lo",
  direction="l",
 },
 [0x131F2]={
  category="lo",
  direction="l",
 },
 [0x131F3]={
  category="lo",
  direction="l",
 },
 [0x131F4]={
  category="lo",
  direction="l",
 },
 [0x131F5]={
  category="lo",
  direction="l",
 },
 [0x131F6]={
  category="lo",
  direction="l",
 },
 [0x131F7]={
  category="lo",
  direction="l",
 },
 [0x131F8]={
  category="lo",
  direction="l",
 },
 [0x131F9]={
  category="lo",
  direction="l",
 },
 [0x131FA]={
  category="lo",
  direction="l",
 },
 [0x131FB]={
  category="lo",
  direction="l",
 },
 [0x131FC]={
  category="lo",
  direction="l",
 },
 [0x131FD]={
  category="lo",
  direction="l",
 },
 [0x131FE]={
  category="lo",
  direction="l",
 },
 [0x131FF]={
  category="lo",
  direction="l",
 },
 [0x13200]={
  category="lo",
  direction="l",
 },
 [0x13201]={
  category="lo",
  direction="l",
 },
 [0x13202]={
  category="lo",
  direction="l",
 },
 [0x13203]={
  category="lo",
  direction="l",
 },
 [0x13204]={
  category="lo",
  direction="l",
 },
 [0x13205]={
  category="lo",
  direction="l",
 },
 [0x13206]={
  category="lo",
  direction="l",
 },
 [0x13207]={
  category="lo",
  direction="l",
 },
 [0x13208]={
  category="lo",
  direction="l",
 },
 [0x13209]={
  category="lo",
  direction="l",
 },
 [0x1320A]={
  category="lo",
  direction="l",
 },
 [0x1320B]={
  category="lo",
  direction="l",
 },
 [0x1320C]={
  category="lo",
  direction="l",
 },
 [0x1320D]={
  category="lo",
  direction="l",
 },
 [0x1320E]={
  category="lo",
  direction="l",
 },
 [0x1320F]={
  category="lo",
  direction="l",
 },
 [0x13210]={
  category="lo",
  direction="l",
 },
 [0x13211]={
  category="lo",
  direction="l",
 },
 [0x13212]={
  category="lo",
  direction="l",
 },
 [0x13213]={
  category="lo",
  direction="l",
 },
 [0x13214]={
  category="lo",
  direction="l",
 },
 [0x13215]={
  category="lo",
  direction="l",
 },
 [0x13216]={
  category="lo",
  direction="l",
 },
 [0x13217]={
  category="lo",
  direction="l",
 },
 [0x13218]={
  category="lo",
  direction="l",
 },
 [0x13219]={
  category="lo",
  direction="l",
 },
 [0x1321A]={
  category="lo",
  direction="l",
 },
 [0x1321B]={
  category="lo",
  direction="l",
 },
 [0x1321C]={
  category="lo",
  direction="l",
 },
 [0x1321D]={
  category="lo",
  direction="l",
 },
 [0x1321E]={
  category="lo",
  direction="l",
 },
 [0x1321F]={
  category="lo",
  direction="l",
 },
 [0x13220]={
  category="lo",
  direction="l",
 },
 [0x13221]={
  category="lo",
  direction="l",
 },
 [0x13222]={
  category="lo",
  direction="l",
 },
 [0x13223]={
  category="lo",
  direction="l",
 },
 [0x13224]={
  category="lo",
  direction="l",
 },
 [0x13225]={
  category="lo",
  direction="l",
 },
 [0x13226]={
  category="lo",
  direction="l",
 },
 [0x13227]={
  category="lo",
  direction="l",
 },
 [0x13228]={
  category="lo",
  direction="l",
 },
 [0x13229]={
  category="lo",
  direction="l",
 },
 [0x1322A]={
  category="lo",
  direction="l",
 },
 [0x1322B]={
  category="lo",
  direction="l",
 },
 [0x1322C]={
  category="lo",
  direction="l",
 },
 [0x1322D]={
  category="lo",
  direction="l",
 },
 [0x1322E]={
  category="lo",
  direction="l",
 },
 [0x1322F]={
  category="lo",
  direction="l",
 },
 [0x13230]={
  category="lo",
  direction="l",
 },
 [0x13231]={
  category="lo",
  direction="l",
 },
 [0x13232]={
  category="lo",
  direction="l",
 },
 [0x13233]={
  category="lo",
  direction="l",
 },
 [0x13234]={
  category="lo",
  direction="l",
 },
 [0x13235]={
  category="lo",
  direction="l",
 },
 [0x13236]={
  category="lo",
  direction="l",
 },
 [0x13237]={
  category="lo",
  direction="l",
 },
 [0x13238]={
  category="lo",
  direction="l",
 },
 [0x13239]={
  category="lo",
  direction="l",
 },
 [0x1323A]={
  category="lo",
  direction="l",
 },
 [0x1323B]={
  category="lo",
  direction="l",
 },
 [0x1323C]={
  category="lo",
  direction="l",
 },
 [0x1323D]={
  category="lo",
  direction="l",
 },
 [0x1323E]={
  category="lo",
  direction="l",
 },
 [0x1323F]={
  category="lo",
  direction="l",
 },
 [0x13240]={
  category="lo",
  direction="l",
 },
 [0x13241]={
  category="lo",
  direction="l",
 },
 [0x13242]={
  category="lo",
  direction="l",
 },
 [0x13243]={
  category="lo",
  direction="l",
 },
 [0x13244]={
  category="lo",
  direction="l",
 },
 [0x13245]={
  category="lo",
  direction="l",
 },
 [0x13246]={
  category="lo",
  direction="l",
 },
 [0x13247]={
  category="lo",
  direction="l",
 },
 [0x13248]={
  category="lo",
  direction="l",
 },
 [0x13249]={
  category="lo",
  direction="l",
 },
 [0x1324A]={
  category="lo",
  direction="l",
 },
 [0x1324B]={
  category="lo",
  direction="l",
 },
 [0x1324C]={
  category="lo",
  direction="l",
 },
 [0x1324D]={
  category="lo",
  direction="l",
 },
 [0x1324E]={
  category="lo",
  direction="l",
 },
 [0x1324F]={
  category="lo",
  direction="l",
 },
 [0x13250]={
  category="lo",
  direction="l",
 },
 [0x13251]={
  category="lo",
  direction="l",
 },
 [0x13252]={
  category="lo",
  direction="l",
 },
 [0x13253]={
  category="lo",
  direction="l",
 },
 [0x13254]={
  category="lo",
  direction="l",
 },
 [0x13255]={
  category="lo",
  direction="l",
 },
 [0x13256]={
  category="lo",
  direction="l",
 },
 [0x13257]={
  category="lo",
  direction="l",
 },
 [0x13258]={
  category="lo",
  direction="l",
 },
 [0x13259]={
  category="lo",
  direction="l",
 },
 [0x1325A]={
  category="lo",
  direction="l",
 },
 [0x1325B]={
  category="lo",
  direction="l",
 },
 [0x1325C]={
  category="lo",
  direction="l",
 },
 [0x1325D]={
  category="lo",
  direction="l",
 },
 [0x1325E]={
  category="lo",
  direction="l",
 },
 [0x1325F]={
  category="lo",
  direction="l",
 },
 [0x13260]={
  category="lo",
  direction="l",
 },
 [0x13261]={
  category="lo",
  direction="l",
 },
 [0x13262]={
  category="lo",
  direction="l",
 },
 [0x13263]={
  category="lo",
  direction="l",
 },
 [0x13264]={
  category="lo",
  direction="l",
 },
 [0x13265]={
  category="lo",
  direction="l",
 },
 [0x13266]={
  category="lo",
  direction="l",
 },
 [0x13267]={
  category="lo",
  direction="l",
 },
 [0x13268]={
  category="lo",
  direction="l",
 },
 [0x13269]={
  category="lo",
  direction="l",
 },
 [0x1326A]={
  category="lo",
  direction="l",
 },
 [0x1326B]={
  category="lo",
  direction="l",
 },
 [0x1326C]={
  category="lo",
  direction="l",
 },
 [0x1326D]={
  category="lo",
  direction="l",
 },
 [0x1326E]={
  category="lo",
  direction="l",
 },
 [0x1326F]={
  category="lo",
  direction="l",
 },
 [0x13270]={
  category="lo",
  direction="l",
 },
 [0x13271]={
  category="lo",
  direction="l",
 },
 [0x13272]={
  category="lo",
  direction="l",
 },
 [0x13273]={
  category="lo",
  direction="l",
 },
 [0x13274]={
  category="lo",
  direction="l",
 },
 [0x13275]={
  category="lo",
  direction="l",
 },
 [0x13276]={
  category="lo",
  direction="l",
 },
 [0x13277]={
  category="lo",
  direction="l",
 },
 [0x13278]={
  category="lo",
  direction="l",
 },
 [0x13279]={
  category="lo",
  direction="l",
 },
 [0x1327A]={
  category="lo",
  direction="l",
 },
 [0x1327B]={
  category="lo",
  direction="l",
 },
 [0x1327C]={
  category="lo",
  direction="l",
 },
 [0x1327D]={
  category="lo",
  direction="l",
 },
 [0x1327E]={
  category="lo",
  direction="l",
 },
 [0x1327F]={
  category="lo",
  direction="l",
 },
 [0x13280]={
  category="lo",
  direction="l",
 },
 [0x13281]={
  category="lo",
  direction="l",
 },
 [0x13282]={
  category="lo",
  direction="l",
 },
 [0x13283]={
  category="lo",
  direction="l",
 },
 [0x13284]={
  category="lo",
  direction="l",
 },
 [0x13285]={
  category="lo",
  direction="l",
 },
 [0x13286]={
  category="lo",
  direction="l",
 },
 [0x13287]={
  category="lo",
  direction="l",
 },
 [0x13288]={
  category="lo",
  direction="l",
 },
 [0x13289]={
  category="lo",
  direction="l",
 },
 [0x1328A]={
  category="lo",
  direction="l",
 },
 [0x1328B]={
  category="lo",
  direction="l",
 },
 [0x1328C]={
  category="lo",
  direction="l",
 },
 [0x1328D]={
  category="lo",
  direction="l",
 },
 [0x1328E]={
  category="lo",
  direction="l",
 },
 [0x1328F]={
  category="lo",
  direction="l",
 },
 [0x13290]={
  category="lo",
  direction="l",
 },
 [0x13291]={
  category="lo",
  direction="l",
 },
 [0x13292]={
  category="lo",
  direction="l",
 },
 [0x13293]={
  category="lo",
  direction="l",
 },
 [0x13294]={
  category="lo",
  direction="l",
 },
 [0x13295]={
  category="lo",
  direction="l",
 },
 [0x13296]={
  category="lo",
  direction="l",
 },
 [0x13297]={
  category="lo",
  direction="l",
 },
 [0x13298]={
  category="lo",
  direction="l",
 },
 [0x13299]={
  category="lo",
  direction="l",
 },
 [0x1329A]={
  category="lo",
  direction="l",
 },
 [0x1329B]={
  category="lo",
  direction="l",
 },
 [0x1329C]={
  category="lo",
  direction="l",
 },
 [0x1329D]={
  category="lo",
  direction="l",
 },
 [0x1329E]={
  category="lo",
  direction="l",
 },
 [0x1329F]={
  category="lo",
  direction="l",
 },
 [0x132A0]={
  category="lo",
  direction="l",
 },
 [0x132A1]={
  category="lo",
  direction="l",
 },
 [0x132A2]={
  category="lo",
  direction="l",
 },
 [0x132A3]={
  category="lo",
  direction="l",
 },
 [0x132A4]={
  category="lo",
  direction="l",
 },
 [0x132A5]={
  category="lo",
  direction="l",
 },
 [0x132A6]={
  category="lo",
  direction="l",
 },
 [0x132A7]={
  category="lo",
  direction="l",
 },
 [0x132A8]={
  category="lo",
  direction="l",
 },
 [0x132A9]={
  category="lo",
  direction="l",
 },
 [0x132AA]={
  category="lo",
  direction="l",
 },
 [0x132AB]={
  category="lo",
  direction="l",
 },
 [0x132AC]={
  category="lo",
  direction="l",
 },
 [0x132AD]={
  category="lo",
  direction="l",
 },
 [0x132AE]={
  category="lo",
  direction="l",
 },
 [0x132AF]={
  category="lo",
  direction="l",
 },
 [0x132B0]={
  category="lo",
  direction="l",
 },
 [0x132B1]={
  category="lo",
  direction="l",
 },
 [0x132B2]={
  category="lo",
  direction="l",
 },
 [0x132B3]={
  category="lo",
  direction="l",
 },
 [0x132B4]={
  category="lo",
  direction="l",
 },
 [0x132B5]={
  category="lo",
  direction="l",
 },
 [0x132B6]={
  category="lo",
  direction="l",
 },
 [0x132B7]={
  category="lo",
  direction="l",
 },
 [0x132B8]={
  category="lo",
  direction="l",
 },
 [0x132B9]={
  category="lo",
  direction="l",
 },
 [0x132BA]={
  category="lo",
  direction="l",
 },
 [0x132BB]={
  category="lo",
  direction="l",
 },
 [0x132BC]={
  category="lo",
  direction="l",
 },
 [0x132BD]={
  category="lo",
  direction="l",
 },
 [0x132BE]={
  category="lo",
  direction="l",
 },
 [0x132BF]={
  category="lo",
  direction="l",
 },
 [0x132C0]={
  category="lo",
  direction="l",
 },
 [0x132C1]={
  category="lo",
  direction="l",
 },
 [0x132C2]={
  category="lo",
  direction="l",
 },
 [0x132C3]={
  category="lo",
  direction="l",
 },
 [0x132C4]={
  category="lo",
  direction="l",
 },
 [0x132C5]={
  category="lo",
  direction="l",
 },
 [0x132C6]={
  category="lo",
  direction="l",
 },
 [0x132C7]={
  category="lo",
  direction="l",
 },
 [0x132C8]={
  category="lo",
  direction="l",
 },
 [0x132C9]={
  category="lo",
  direction="l",
 },
 [0x132CA]={
  category="lo",
  direction="l",
 },
 [0x132CB]={
  category="lo",
  direction="l",
 },
 [0x132CC]={
  category="lo",
  direction="l",
 },
 [0x132CD]={
  category="lo",
  direction="l",
 },
 [0x132CE]={
  category="lo",
  direction="l",
 },
 [0x132CF]={
  category="lo",
  direction="l",
 },
 [0x132D0]={
  category="lo",
  direction="l",
 },
 [0x132D1]={
  category="lo",
  direction="l",
 },
 [0x132D2]={
  category="lo",
  direction="l",
 },
 [0x132D3]={
  category="lo",
  direction="l",
 },
 [0x132D4]={
  category="lo",
  direction="l",
 },
 [0x132D5]={
  category="lo",
  direction="l",
 },
 [0x132D6]={
  category="lo",
  direction="l",
 },
 [0x132D7]={
  category="lo",
  direction="l",
 },
 [0x132D8]={
  category="lo",
  direction="l",
 },
 [0x132D9]={
  category="lo",
  direction="l",
 },
 [0x132DA]={
  category="lo",
  direction="l",
 },
 [0x132DB]={
  category="lo",
  direction="l",
 },
 [0x132DC]={
  category="lo",
  direction="l",
 },
 [0x132DD]={
  category="lo",
  direction="l",
 },
 [0x132DE]={
  category="lo",
  direction="l",
 },
 [0x132DF]={
  category="lo",
  direction="l",
 },
 [0x132E0]={
  category="lo",
  direction="l",
 },
 [0x132E1]={
  category="lo",
  direction="l",
 },
 [0x132E2]={
  category="lo",
  direction="l",
 },
 [0x132E3]={
  category="lo",
  direction="l",
 },
 [0x132E4]={
  category="lo",
  direction="l",
 },
 [0x132E5]={
  category="lo",
  direction="l",
 },
 [0x132E6]={
  category="lo",
  direction="l",
 },
 [0x132E7]={
  category="lo",
  direction="l",
 },
 [0x132E8]={
  category="lo",
  direction="l",
 },
 [0x132E9]={
  category="lo",
  direction="l",
 },
 [0x132EA]={
  category="lo",
  direction="l",
 },
 [0x132EB]={
  category="lo",
  direction="l",
 },
 [0x132EC]={
  category="lo",
  direction="l",
 },
 [0x132ED]={
  category="lo",
  direction="l",
 },
 [0x132EE]={
  category="lo",
  direction="l",
 },
 [0x132EF]={
  category="lo",
  direction="l",
 },
 [0x132F0]={
  category="lo",
  direction="l",
 },
 [0x132F1]={
  category="lo",
  direction="l",
 },
 [0x132F2]={
  category="lo",
  direction="l",
 },
 [0x132F3]={
  category="lo",
  direction="l",
 },
 [0x132F4]={
  category="lo",
  direction="l",
 },
 [0x132F5]={
  category="lo",
  direction="l",
 },
 [0x132F6]={
  category="lo",
  direction="l",
 },
 [0x132F7]={
  category="lo",
  direction="l",
 },
 [0x132F8]={
  category="lo",
  direction="l",
 },
 [0x132F9]={
  category="lo",
  direction="l",
 },
 [0x132FA]={
  category="lo",
  direction="l",
 },
 [0x132FB]={
  category="lo",
  direction="l",
 },
 [0x132FC]={
  category="lo",
  direction="l",
 },
 [0x132FD]={
  category="lo",
  direction="l",
 },
 [0x132FE]={
  category="lo",
  direction="l",
 },
 [0x132FF]={
  category="lo",
  direction="l",
 },
 [0x13300]={
  category="lo",
  direction="l",
 },
 [0x13301]={
  category="lo",
  direction="l",
 },
 [0x13302]={
  category="lo",
  direction="l",
 },
 [0x13303]={
  category="lo",
  direction="l",
 },
 [0x13304]={
  category="lo",
  direction="l",
 },
 [0x13305]={
  category="lo",
  direction="l",
 },
 [0x13306]={
  category="lo",
  direction="l",
 },
 [0x13307]={
  category="lo",
  direction="l",
 },
 [0x13308]={
  category="lo",
  direction="l",
 },
 [0x13309]={
  category="lo",
  direction="l",
 },
 [0x1330A]={
  category="lo",
  direction="l",
 },
 [0x1330B]={
  category="lo",
  direction="l",
 },
 [0x1330C]={
  category="lo",
  direction="l",
 },
 [0x1330D]={
  category="lo",
  direction="l",
 },
 [0x1330E]={
  category="lo",
  direction="l",
 },
 [0x1330F]={
  category="lo",
  direction="l",
 },
 [0x13310]={
  category="lo",
  direction="l",
 },
 [0x13311]={
  category="lo",
  direction="l",
 },
 [0x13312]={
  category="lo",
  direction="l",
 },
 [0x13313]={
  category="lo",
  direction="l",
 },
 [0x13314]={
  category="lo",
  direction="l",
 },
 [0x13315]={
  category="lo",
  direction="l",
 },
 [0x13316]={
  category="lo",
  direction="l",
 },
 [0x13317]={
  category="lo",
  direction="l",
 },
 [0x13318]={
  category="lo",
  direction="l",
 },
 [0x13319]={
  category="lo",
  direction="l",
 },
 [0x1331A]={
  category="lo",
  direction="l",
 },
 [0x1331B]={
  category="lo",
  direction="l",
 },
 [0x1331C]={
  category="lo",
  direction="l",
 },
 [0x1331D]={
  category="lo",
  direction="l",
 },
 [0x1331E]={
  category="lo",
  direction="l",
 },
 [0x1331F]={
  category="lo",
  direction="l",
 },
 [0x13320]={
  category="lo",
  direction="l",
 },
 [0x13321]={
  category="lo",
  direction="l",
 },
 [0x13322]={
  category="lo",
  direction="l",
 },
 [0x13323]={
  category="lo",
  direction="l",
 },
 [0x13324]={
  category="lo",
  direction="l",
 },
 [0x13325]={
  category="lo",
  direction="l",
 },
 [0x13326]={
  category="lo",
  direction="l",
 },
 [0x13327]={
  category="lo",
  direction="l",
 },
 [0x13328]={
  category="lo",
  direction="l",
 },
 [0x13329]={
  category="lo",
  direction="l",
 },
 [0x1332A]={
  category="lo",
  direction="l",
 },
 [0x1332B]={
  category="lo",
  direction="l",
 },
 [0x1332C]={
  category="lo",
  direction="l",
 },
 [0x1332D]={
  category="lo",
  direction="l",
 },
 [0x1332E]={
  category="lo",
  direction="l",
 },
 [0x1332F]={
  category="lo",
  direction="l",
 },
 [0x13330]={
  category="lo",
  direction="l",
 },
 [0x13331]={
  category="lo",
  direction="l",
 },
 [0x13332]={
  category="lo",
  direction="l",
 },
 [0x13333]={
  category="lo",
  direction="l",
 },
 [0x13334]={
  category="lo",
  direction="l",
 },
 [0x13335]={
  category="lo",
  direction="l",
 },
 [0x13336]={
  category="lo",
  direction="l",
 },
 [0x13337]={
  category="lo",
  direction="l",
 },
 [0x13338]={
  category="lo",
  direction="l",
 },
 [0x13339]={
  category="lo",
  direction="l",
 },
 [0x1333A]={
  category="lo",
  direction="l",
 },
 [0x1333B]={
  category="lo",
  direction="l",
 },
 [0x1333C]={
  category="lo",
  direction="l",
 },
 [0x1333D]={
  category="lo",
  direction="l",
 },
 [0x1333E]={
  category="lo",
  direction="l",
 },
 [0x1333F]={
  category="lo",
  direction="l",
 },
 [0x13340]={
  category="lo",
  direction="l",
 },
 [0x13341]={
  category="lo",
  direction="l",
 },
 [0x13342]={
  category="lo",
  direction="l",
 },
 [0x13343]={
  category="lo",
  direction="l",
 },
 [0x13344]={
  category="lo",
  direction="l",
 },
 [0x13345]={
  category="lo",
  direction="l",
 },
 [0x13346]={
  category="lo",
  direction="l",
 },
 [0x13347]={
  category="lo",
  direction="l",
 },
 [0x13348]={
  category="lo",
  direction="l",
 },
 [0x13349]={
  category="lo",
  direction="l",
 },
 [0x1334A]={
  category="lo",
  direction="l",
 },
 [0x1334B]={
  category="lo",
  direction="l",
 },
 [0x1334C]={
  category="lo",
  direction="l",
 },
 [0x1334D]={
  category="lo",
  direction="l",
 },
 [0x1334E]={
  category="lo",
  direction="l",
 },
 [0x1334F]={
  category="lo",
  direction="l",
 },
 [0x13350]={
  category="lo",
  direction="l",
 },
 [0x13351]={
  category="lo",
  direction="l",
 },
 [0x13352]={
  category="lo",
  direction="l",
 },
 [0x13353]={
  category="lo",
  direction="l",
 },
 [0x13354]={
  category="lo",
  direction="l",
 },
 [0x13355]={
  category="lo",
  direction="l",
 },
 [0x13356]={
  category="lo",
  direction="l",
 },
 [0x13357]={
  category="lo",
  direction="l",
 },
 [0x13358]={
  category="lo",
  direction="l",
 },
 [0x13359]={
  category="lo",
  direction="l",
 },
 [0x1335A]={
  category="lo",
  direction="l",
 },
 [0x1335B]={
  category="lo",
  direction="l",
 },
 [0x1335C]={
  category="lo",
  direction="l",
 },
 [0x1335D]={
  category="lo",
  direction="l",
 },
 [0x1335E]={
  category="lo",
  direction="l",
 },
 [0x1335F]={
  category="lo",
  direction="l",
 },
 [0x13360]={
  category="lo",
  direction="l",
 },
 [0x13361]={
  category="lo",
  direction="l",
 },
 [0x13362]={
  category="lo",
  direction="l",
 },
 [0x13363]={
  category="lo",
  direction="l",
 },
 [0x13364]={
  category="lo",
  direction="l",
 },
 [0x13365]={
  category="lo",
  direction="l",
 },
 [0x13366]={
  category="lo",
  direction="l",
 },
 [0x13367]={
  category="lo",
  direction="l",
 },
 [0x13368]={
  category="lo",
  direction="l",
 },
 [0x13369]={
  category="lo",
  direction="l",
 },
 [0x1336A]={
  category="lo",
  direction="l",
 },
 [0x1336B]={
  category="lo",
  direction="l",
 },
 [0x1336C]={
  category="lo",
  direction="l",
 },
 [0x1336D]={
  category="lo",
  direction="l",
 },
 [0x1336E]={
  category="lo",
  direction="l",
 },
 [0x1336F]={
  category="lo",
  direction="l",
 },
 [0x13370]={
  category="lo",
  direction="l",
 },
 [0x13371]={
  category="lo",
  direction="l",
 },
 [0x13372]={
  category="lo",
  direction="l",
 },
 [0x13373]={
  category="lo",
  direction="l",
 },
 [0x13374]={
  category="lo",
  direction="l",
 },
 [0x13375]={
  category="lo",
  direction="l",
 },
 [0x13376]={
  category="lo",
  direction="l",
 },
 [0x13377]={
  category="lo",
  direction="l",
 },
 [0x13378]={
  category="lo",
  direction="l",
 },
 [0x13379]={
  category="lo",
  direction="l",
 },
 [0x1337A]={
  category="lo",
  direction="l",
 },
 [0x1337B]={
  category="lo",
  direction="l",
 },
 [0x1337C]={
  category="lo",
  direction="l",
 },
 [0x1337D]={
  category="lo",
  direction="l",
 },
 [0x1337E]={
  category="lo",
  direction="l",
 },
 [0x1337F]={
  category="lo",
  direction="l",
 },
 [0x13380]={
  category="lo",
  direction="l",
 },
 [0x13381]={
  category="lo",
  direction="l",
 },
 [0x13382]={
  category="lo",
  direction="l",
 },
 [0x13383]={
  category="lo",
  direction="l",
 },
 [0x13384]={
  category="lo",
  direction="l",
 },
 [0x13385]={
  category="lo",
  direction="l",
 },
 [0x13386]={
  category="lo",
  direction="l",
 },
 [0x13387]={
  category="lo",
  direction="l",
 },
 [0x13388]={
  category="lo",
  direction="l",
 },
 [0x13389]={
  category="lo",
  direction="l",
 },
 [0x1338A]={
  category="lo",
  direction="l",
 },
 [0x1338B]={
  category="lo",
  direction="l",
 },
 [0x1338C]={
  category="lo",
  direction="l",
 },
 [0x1338D]={
  category="lo",
  direction="l",
 },
 [0x1338E]={
  category="lo",
  direction="l",
 },
 [0x1338F]={
  category="lo",
  direction="l",
 },
 [0x13390]={
  category="lo",
  direction="l",
 },
 [0x13391]={
  category="lo",
  direction="l",
 },
 [0x13392]={
  category="lo",
  direction="l",
 },
 [0x13393]={
  category="lo",
  direction="l",
 },
 [0x13394]={
  category="lo",
  direction="l",
 },
 [0x13395]={
  category="lo",
  direction="l",
 },
 [0x13396]={
  category="lo",
  direction="l",
 },
 [0x13397]={
  category="lo",
  direction="l",
 },
 [0x13398]={
  category="lo",
  direction="l",
 },
 [0x13399]={
  category="lo",
  direction="l",
 },
 [0x1339A]={
  category="lo",
  direction="l",
 },
 [0x1339B]={
  category="lo",
  direction="l",
 },
 [0x1339C]={
  category="lo",
  direction="l",
 },
 [0x1339D]={
  category="lo",
  direction="l",
 },
 [0x1339E]={
  category="lo",
  direction="l",
 },
 [0x1339F]={
  category="lo",
  direction="l",
 },
 [0x133A0]={
  category="lo",
  direction="l",
 },
 [0x133A1]={
  category="lo",
  direction="l",
 },
 [0x133A2]={
  category="lo",
  direction="l",
 },
 [0x133A3]={
  category="lo",
  direction="l",
 },
 [0x133A4]={
  category="lo",
  direction="l",
 },
 [0x133A5]={
  category="lo",
  direction="l",
 },
 [0x133A6]={
  category="lo",
  direction="l",
 },
 [0x133A7]={
  category="lo",
  direction="l",
 },
 [0x133A8]={
  category="lo",
  direction="l",
 },
 [0x133A9]={
  category="lo",
  direction="l",
 },
 [0x133AA]={
  category="lo",
  direction="l",
 },
 [0x133AB]={
  category="lo",
  direction="l",
 },
 [0x133AC]={
  category="lo",
  direction="l",
 },
 [0x133AD]={
  category="lo",
  direction="l",
 },
 [0x133AE]={
  category="lo",
  direction="l",
 },
 [0x133AF]={
  category="lo",
  direction="l",
 },
 [0x133B0]={
  category="lo",
  direction="l",
 },
 [0x133B1]={
  category="lo",
  direction="l",
 },
 [0x133B2]={
  category="lo",
  direction="l",
 },
 [0x133B3]={
  category="lo",
  direction="l",
 },
 [0x133B4]={
  category="lo",
  direction="l",
 },
 [0x133B5]={
  category="lo",
  direction="l",
 },
 [0x133B6]={
  category="lo",
  direction="l",
 },
 [0x133B7]={
  category="lo",
  direction="l",
 },
 [0x133B8]={
  category="lo",
  direction="l",
 },
 [0x133B9]={
  category="lo",
  direction="l",
 },
 [0x133BA]={
  category="lo",
  direction="l",
 },
 [0x133BB]={
  category="lo",
  direction="l",
 },
 [0x133BC]={
  category="lo",
  direction="l",
 },
 [0x133BD]={
  category="lo",
  direction="l",
 },
 [0x133BE]={
  category="lo",
  direction="l",
 },
 [0x133BF]={
  category="lo",
  direction="l",
 },
 [0x133C0]={
  category="lo",
  direction="l",
 },
 [0x133C1]={
  category="lo",
  direction="l",
 },
 [0x133C2]={
  category="lo",
  direction="l",
 },
 [0x133C3]={
  category="lo",
  direction="l",
 },
 [0x133C4]={
  category="lo",
  direction="l",
 },
 [0x133C5]={
  category="lo",
  direction="l",
 },
 [0x133C6]={
  category="lo",
  direction="l",
 },
 [0x133C7]={
  category="lo",
  direction="l",
 },
 [0x133C8]={
  category="lo",
  direction="l",
 },
 [0x133C9]={
  category="lo",
  direction="l",
 },
 [0x133CA]={
  category="lo",
  direction="l",
 },
 [0x133CB]={
  category="lo",
  direction="l",
 },
 [0x133CC]={
  category="lo",
  direction="l",
 },
 [0x133CD]={
  category="lo",
  direction="l",
 },
 [0x133CE]={
  category="lo",
  direction="l",
 },
 [0x133CF]={
  category="lo",
  direction="l",
 },
 [0x133D0]={
  category="lo",
  direction="l",
 },
 [0x133D1]={
  category="lo",
  direction="l",
 },
 [0x133D2]={
  category="lo",
  direction="l",
 },
 [0x133D3]={
  category="lo",
  direction="l",
 },
 [0x133D4]={
  category="lo",
  direction="l",
 },
 [0x133D5]={
  category="lo",
  direction="l",
 },
 [0x133D6]={
  category="lo",
  direction="l",
 },
 [0x133D7]={
  category="lo",
  direction="l",
 },
 [0x133D8]={
  category="lo",
  direction="l",
 },
 [0x133D9]={
  category="lo",
  direction="l",
 },
 [0x133DA]={
  category="lo",
  direction="l",
 },
 [0x133DB]={
  category="lo",
  direction="l",
 },
 [0x133DC]={
  category="lo",
  direction="l",
 },
 [0x133DD]={
  category="lo",
  direction="l",
 },
 [0x133DE]={
  category="lo",
  direction="l",
 },
 [0x133DF]={
  category="lo",
  direction="l",
 },
 [0x133E0]={
  category="lo",
  direction="l",
 },
 [0x133E1]={
  category="lo",
  direction="l",
 },
 [0x133E2]={
  category="lo",
  direction="l",
 },
 [0x133E3]={
  category="lo",
  direction="l",
 },
 [0x133E4]={
  category="lo",
  direction="l",
 },
 [0x133E5]={
  category="lo",
  direction="l",
 },
 [0x133E6]={
  category="lo",
  direction="l",
 },
 [0x133E7]={
  category="lo",
  direction="l",
 },
 [0x133E8]={
  category="lo",
  direction="l",
 },
 [0x133E9]={
  category="lo",
  direction="l",
 },
 [0x133EA]={
  category="lo",
  direction="l",
 },
 [0x133EB]={
  category="lo",
  direction="l",
 },
 [0x133EC]={
  category="lo",
  direction="l",
 },
 [0x133ED]={
  category="lo",
  direction="l",
 },
 [0x133EE]={
  category="lo",
  direction="l",
 },
 [0x133EF]={
  category="lo",
  direction="l",
 },
 [0x133F0]={
  category="lo",
  direction="l",
 },
 [0x133F1]={
  category="lo",
  direction="l",
 },
 [0x133F2]={
  category="lo",
  direction="l",
 },
 [0x133F3]={
  category="lo",
  direction="l",
 },
 [0x133F4]={
  category="lo",
  direction="l",
 },
 [0x133F5]={
  category="lo",
  direction="l",
 },
 [0x133F6]={
  category="lo",
  direction="l",
 },
 [0x133F7]={
  category="lo",
  direction="l",
 },
 [0x133F8]={
  category="lo",
  direction="l",
 },
 [0x133F9]={
  category="lo",
  direction="l",
 },
 [0x133FA]={
  category="lo",
  direction="l",
 },
 [0x133FB]={
  category="lo",
  direction="l",
 },
 [0x133FC]={
  category="lo",
  direction="l",
 },
 [0x133FD]={
  category="lo",
  direction="l",
 },
 [0x133FE]={
  category="lo",
  direction="l",
 },
 [0x133FF]={
  category="lo",
  direction="l",
 },
 [0x13400]={
  category="lo",
  direction="l",
 },
 [0x13401]={
  category="lo",
  direction="l",
 },
 [0x13402]={
  category="lo",
  direction="l",
 },
 [0x13403]={
  category="lo",
  direction="l",
 },
 [0x13404]={
  category="lo",
  direction="l",
 },
 [0x13405]={
  category="lo",
  direction="l",
 },
 [0x13406]={
  category="lo",
  direction="l",
 },
 [0x13407]={
  category="lo",
  direction="l",
 },
 [0x13408]={
  category="lo",
  direction="l",
 },
 [0x13409]={
  category="lo",
  direction="l",
 },
 [0x1340A]={
  category="lo",
  direction="l",
 },
 [0x1340B]={
  category="lo",
  direction="l",
 },
 [0x1340C]={
  category="lo",
  direction="l",
 },
 [0x1340D]={
  category="lo",
  direction="l",
 },
 [0x1340E]={
  category="lo",
  direction="l",
 },
 [0x1340F]={
  category="lo",
  direction="l",
 },
 [0x13410]={
  category="lo",
  direction="l",
 },
 [0x13411]={
  category="lo",
  direction="l",
 },
 [0x13412]={
  category="lo",
  direction="l",
 },
 [0x13413]={
  category="lo",
  direction="l",
 },
 [0x13414]={
  category="lo",
  direction="l",
 },
 [0x13415]={
  category="lo",
  direction="l",
 },
 [0x13416]={
  category="lo",
  direction="l",
 },
 [0x13417]={
  category="lo",
  direction="l",
 },
 [0x13418]={
  category="lo",
  direction="l",
 },
 [0x13419]={
  category="lo",
  direction="l",
 },
 [0x1341A]={
  category="lo",
  direction="l",
 },
 [0x1341B]={
  category="lo",
  direction="l",
 },
 [0x1341C]={
  category="lo",
  direction="l",
 },
 [0x1341D]={
  category="lo",
  direction="l",
 },
 [0x1341E]={
  category="lo",
  direction="l",
 },
 [0x1341F]={
  category="lo",
  direction="l",
 },
 [0x13420]={
  category="lo",
  direction="l",
 },
 [0x13421]={
  category="lo",
  direction="l",
 },
 [0x13422]={
  category="lo",
  direction="l",
 },
 [0x13423]={
  category="lo",
  direction="l",
 },
 [0x13424]={
  category="lo",
  direction="l",
 },
 [0x13425]={
  category="lo",
  direction="l",
 },
 [0x13426]={
  category="lo",
  direction="l",
 },
 [0x13427]={
  category="lo",
  direction="l",
 },
 [0x13428]={
  category="lo",
  direction="l",
 },
 [0x13429]={
  category="lo",
  direction="l",
 },
 [0x1342A]={
  category="lo",
  direction="l",
 },
 [0x1342B]={
  category="lo",
  direction="l",
 },
 [0x1342C]={
  category="lo",
  direction="l",
 },
 [0x1342D]={
  category="lo",
  direction="l",
 },
 [0x1342E]={
  category="lo",
  direction="l",
 },
 [0x16800]={
  category="lo",
  direction="l",
 },
 [0x16801]={
  category="lo",
  direction="l",
 },
 [0x16802]={
  category="lo",
  direction="l",
 },
 [0x16803]={
  category="lo",
  direction="l",
 },
 [0x16804]={
  category="lo",
  direction="l",
 },
 [0x16805]={
  category="lo",
  direction="l",
 },
 [0x16806]={
  category="lo",
  direction="l",
 },
 [0x16807]={
  category="lo",
  direction="l",
 },
 [0x16808]={
  category="lo",
  direction="l",
 },
 [0x16809]={
  category="lo",
  direction="l",
 },
 [0x1680A]={
  category="lo",
  direction="l",
 },
 [0x1680B]={
  category="lo",
  direction="l",
 },
 [0x1680C]={
  category="lo",
  direction="l",
 },
 [0x1680D]={
  category="lo",
  direction="l",
 },
 [0x1680E]={
  category="lo",
  direction="l",
 },
 [0x1680F]={
  category="lo",
  direction="l",
 },
 [0x16810]={
  category="lo",
  direction="l",
 },
 [0x16811]={
  category="lo",
  direction="l",
 },
 [0x16812]={
  category="lo",
  direction="l",
 },
 [0x16813]={
  category="lo",
  direction="l",
 },
 [0x16814]={
  category="lo",
  direction="l",
 },
 [0x16815]={
  category="lo",
  direction="l",
 },
 [0x16816]={
  category="lo",
  direction="l",
 },
 [0x16817]={
  category="lo",
  direction="l",
 },
 [0x16818]={
  category="lo",
  direction="l",
 },
 [0x16819]={
  category="lo",
  direction="l",
 },
 [0x1681A]={
  category="lo",
  direction="l",
 },
 [0x1681B]={
  category="lo",
  direction="l",
 },
 [0x1681C]={
  category="lo",
  direction="l",
 },
 [0x1681D]={
  category="lo",
  direction="l",
 },
 [0x1681E]={
  category="lo",
  direction="l",
 },
 [0x1681F]={
  category="lo",
  direction="l",
 },
 [0x16820]={
  category="lo",
  direction="l",
 },
 [0x16821]={
  category="lo",
  direction="l",
 },
 [0x16822]={
  category="lo",
  direction="l",
 },
 [0x16823]={
  category="lo",
  direction="l",
 },
 [0x16824]={
  category="lo",
  direction="l",
 },
 [0x16825]={
  category="lo",
  direction="l",
 },
 [0x16826]={
  category="lo",
  direction="l",
 },
 [0x16827]={
  category="lo",
  direction="l",
 },
 [0x16828]={
  category="lo",
  direction="l",
 },
 [0x16829]={
  category="lo",
  direction="l",
 },
 [0x1682A]={
  category="lo",
  direction="l",
 },
 [0x1682B]={
  category="lo",
  direction="l",
 },
 [0x1682C]={
  category="lo",
  direction="l",
 },
 [0x1682D]={
  category="lo",
  direction="l",
 },
 [0x1682E]={
  category="lo",
  direction="l",
 },
 [0x1682F]={
  category="lo",
  direction="l",
 },
 [0x16830]={
  category="lo",
  direction="l",
 },
 [0x16831]={
  category="lo",
  direction="l",
 },
 [0x16832]={
  category="lo",
  direction="l",
 },
 [0x16833]={
  category="lo",
  direction="l",
 },
 [0x16834]={
  category="lo",
  direction="l",
 },
 [0x16835]={
  category="lo",
  direction="l",
 },
 [0x16836]={
  category="lo",
  direction="l",
 },
 [0x16837]={
  category="lo",
  direction="l",
 },
 [0x16838]={
  category="lo",
  direction="l",
 },
 [0x16839]={
  category="lo",
  direction="l",
 },
 [0x1683A]={
  category="lo",
  direction="l",
 },
 [0x1683B]={
  category="lo",
  direction="l",
 },
 [0x1683C]={
  category="lo",
  direction="l",
 },
 [0x1683D]={
  category="lo",
  direction="l",
 },
 [0x1683E]={
  category="lo",
  direction="l",
 },
 [0x1683F]={
  category="lo",
  direction="l",
 },
 [0x16840]={
  category="lo",
  direction="l",
 },
 [0x16841]={
  category="lo",
  direction="l",
 },
 [0x16842]={
  category="lo",
  direction="l",
 },
 [0x16843]={
  category="lo",
  direction="l",
 },
 [0x16844]={
  category="lo",
  direction="l",
 },
 [0x16845]={
  category="lo",
  direction="l",
 },
 [0x16846]={
  category="lo",
  direction="l",
 },
 [0x16847]={
  category="lo",
  direction="l",
 },
 [0x16848]={
  category="lo",
  direction="l",
 },
 [0x16849]={
  category="lo",
  direction="l",
 },
 [0x1684A]={
  category="lo",
  direction="l",
 },
 [0x1684B]={
  category="lo",
  direction="l",
 },
 [0x1684C]={
  category="lo",
  direction="l",
 },
 [0x1684D]={
  category="lo",
  direction="l",
 },
 [0x1684E]={
  category="lo",
  direction="l",
 },
 [0x1684F]={
  category="lo",
  direction="l",
 },
 [0x16850]={
  category="lo",
  direction="l",
 },
 [0x16851]={
  category="lo",
  direction="l",
 },
 [0x16852]={
  category="lo",
  direction="l",
 },
 [0x16853]={
  category="lo",
  direction="l",
 },
 [0x16854]={
  category="lo",
  direction="l",
 },
 [0x16855]={
  category="lo",
  direction="l",
 },
 [0x16856]={
  category="lo",
  direction="l",
 },
 [0x16857]={
  category="lo",
  direction="l",
 },
 [0x16858]={
  category="lo",
  direction="l",
 },
 [0x16859]={
  category="lo",
  direction="l",
 },
 [0x1685A]={
  category="lo",
  direction="l",
 },
 [0x1685B]={
  category="lo",
  direction="l",
 },
 [0x1685C]={
  category="lo",
  direction="l",
 },
 [0x1685D]={
  category="lo",
  direction="l",
 },
 [0x1685E]={
  category="lo",
  direction="l",
 },
 [0x1685F]={
  category="lo",
  direction="l",
 },
 [0x16860]={
  category="lo",
  direction="l",
 },
 [0x16861]={
  category="lo",
  direction="l",
 },
 [0x16862]={
  category="lo",
  direction="l",
 },
 [0x16863]={
  category="lo",
  direction="l",
 },
 [0x16864]={
  category="lo",
  direction="l",
 },
 [0x16865]={
  category="lo",
  direction="l",
 },
 [0x16866]={
  category="lo",
  direction="l",
 },
 [0x16867]={
  category="lo",
  direction="l",
 },
 [0x16868]={
  category="lo",
  direction="l",
 },
 [0x16869]={
  category="lo",
  direction="l",
 },
 [0x1686A]={
  category="lo",
  direction="l",
 },
 [0x1686B]={
  category="lo",
  direction="l",
 },
 [0x1686C]={
  category="lo",
  direction="l",
 },
 [0x1686D]={
  category="lo",
  direction="l",
 },
 [0x1686E]={
  category="lo",
  direction="l",
 },
 [0x1686F]={
  category="lo",
  direction="l",
 },
 [0x16870]={
  category="lo",
  direction="l",
 },
 [0x16871]={
  category="lo",
  direction="l",
 },
 [0x16872]={
  category="lo",
  direction="l",
 },
 [0x16873]={
  category="lo",
  direction="l",
 },
 [0x16874]={
  category="lo",
  direction="l",
 },
 [0x16875]={
  category="lo",
  direction="l",
 },
 [0x16876]={
  category="lo",
  direction="l",
 },
 [0x16877]={
  category="lo",
  direction="l",
 },
 [0x16878]={
  category="lo",
  direction="l",
 },
 [0x16879]={
  category="lo",
  direction="l",
 },
 [0x1687A]={
  category="lo",
  direction="l",
 },
 [0x1687B]={
  category="lo",
  direction="l",
 },
 [0x1687C]={
  category="lo",
  direction="l",
 },
 [0x1687D]={
  category="lo",
  direction="l",
 },
 [0x1687E]={
  category="lo",
  direction="l",
 },
 [0x1687F]={
  category="lo",
  direction="l",
 },
 [0x16880]={
  category="lo",
  direction="l",
 },
 [0x16881]={
  category="lo",
  direction="l",
 },
 [0x16882]={
  category="lo",
  direction="l",
 },
 [0x16883]={
  category="lo",
  direction="l",
 },
 [0x16884]={
  category="lo",
  direction="l",
 },
 [0x16885]={
  category="lo",
  direction="l",
 },
 [0x16886]={
  category="lo",
  direction="l",
 },
 [0x16887]={
  category="lo",
  direction="l",
 },
 [0x16888]={
  category="lo",
  direction="l",
 },
 [0x16889]={
  category="lo",
  direction="l",
 },
 [0x1688A]={
  category="lo",
  direction="l",
 },
 [0x1688B]={
  category="lo",
  direction="l",
 },
 [0x1688C]={
  category="lo",
  direction="l",
 },
 [0x1688D]={
  category="lo",
  direction="l",
 },
 [0x1688E]={
  category="lo",
  direction="l",
 },
 [0x1688F]={
  category="lo",
  direction="l",
 },
 [0x16890]={
  category="lo",
  direction="l",
 },
 [0x16891]={
  category="lo",
  direction="l",
 },
 [0x16892]={
  category="lo",
  direction="l",
 },
 [0x16893]={
  category="lo",
  direction="l",
 },
 [0x16894]={
  category="lo",
  direction="l",
 },
 [0x16895]={
  category="lo",
  direction="l",
 },
 [0x16896]={
  category="lo",
  direction="l",
 },
 [0x16897]={
  category="lo",
  direction="l",
 },
 [0x16898]={
  category="lo",
  direction="l",
 },
 [0x16899]={
  category="lo",
  direction="l",
 },
 [0x1689A]={
  category="lo",
  direction="l",
 },
 [0x1689B]={
  category="lo",
  direction="l",
 },
 [0x1689C]={
  category="lo",
  direction="l",
 },
 [0x1689D]={
  category="lo",
  direction="l",
 },
 [0x1689E]={
  category="lo",
  direction="l",
 },
 [0x1689F]={
  category="lo",
  direction="l",
 },
 [0x168A0]={
  category="lo",
  direction="l",
 },
 [0x168A1]={
  category="lo",
  direction="l",
 },
 [0x168A2]={
  category="lo",
  direction="l",
 },
 [0x168A3]={
  category="lo",
  direction="l",
 },
 [0x168A4]={
  category="lo",
  direction="l",
 },
 [0x168A5]={
  category="lo",
  direction="l",
 },
 [0x168A6]={
  category="lo",
  direction="l",
 },
 [0x168A7]={
  category="lo",
  direction="l",
 },
 [0x168A8]={
  category="lo",
  direction="l",
 },
 [0x168A9]={
  category="lo",
  direction="l",
 },
 [0x168AA]={
  category="lo",
  direction="l",
 },
 [0x168AB]={
  category="lo",
  direction="l",
 },
 [0x168AC]={
  category="lo",
  direction="l",
 },
 [0x168AD]={
  category="lo",
  direction="l",
 },
 [0x168AE]={
  category="lo",
  direction="l",
 },
 [0x168AF]={
  category="lo",
  direction="l",
 },
 [0x168B0]={
  category="lo",
  direction="l",
 },
 [0x168B1]={
  category="lo",
  direction="l",
 },
 [0x168B2]={
  category="lo",
  direction="l",
 },
 [0x168B3]={
  category="lo",
  direction="l",
 },
 [0x168B4]={
  category="lo",
  direction="l",
 },
 [0x168B5]={
  category="lo",
  direction="l",
 },
 [0x168B6]={
  category="lo",
  direction="l",
 },
 [0x168B7]={
  category="lo",
  direction="l",
 },
 [0x168B8]={
  category="lo",
  direction="l",
 },
 [0x168B9]={
  category="lo",
  direction="l",
 },
 [0x168BA]={
  category="lo",
  direction="l",
 },
 [0x168BB]={
  category="lo",
  direction="l",
 },
 [0x168BC]={
  category="lo",
  direction="l",
 },
 [0x168BD]={
  category="lo",
  direction="l",
 },
 [0x168BE]={
  category="lo",
  direction="l",
 },
 [0x168BF]={
  category="lo",
  direction="l",
 },
 [0x168C0]={
  category="lo",
  direction="l",
 },
 [0x168C1]={
  category="lo",
  direction="l",
 },
 [0x168C2]={
  category="lo",
  direction="l",
 },
 [0x168C3]={
  category="lo",
  direction="l",
 },
 [0x168C4]={
  category="lo",
  direction="l",
 },
 [0x168C5]={
  category="lo",
  direction="l",
 },
 [0x168C6]={
  category="lo",
  direction="l",
 },
 [0x168C7]={
  category="lo",
  direction="l",
 },
 [0x168C8]={
  category="lo",
  direction="l",
 },
 [0x168C9]={
  category="lo",
  direction="l",
 },
 [0x168CA]={
  category="lo",
  direction="l",
 },
 [0x168CB]={
  category="lo",
  direction="l",
 },
 [0x168CC]={
  category="lo",
  direction="l",
 },
 [0x168CD]={
  category="lo",
  direction="l",
 },
 [0x168CE]={
  category="lo",
  direction="l",
 },
 [0x168CF]={
  category="lo",
  direction="l",
 },
 [0x168D0]={
  category="lo",
  direction="l",
 },
 [0x168D1]={
  category="lo",
  direction="l",
 },
 [0x168D2]={
  category="lo",
  direction="l",
 },
 [0x168D3]={
  category="lo",
  direction="l",
 },
 [0x168D4]={
  category="lo",
  direction="l",
 },
 [0x168D5]={
  category="lo",
  direction="l",
 },
 [0x168D6]={
  category="lo",
  direction="l",
 },
 [0x168D7]={
  category="lo",
  direction="l",
 },
 [0x168D8]={
  category="lo",
  direction="l",
 },
 [0x168D9]={
  category="lo",
  direction="l",
 },
 [0x168DA]={
  category="lo",
  direction="l",
 },
 [0x168DB]={
  category="lo",
  direction="l",
 },
 [0x168DC]={
  category="lo",
  direction="l",
 },
 [0x168DD]={
  category="lo",
  direction="l",
 },
 [0x168DE]={
  category="lo",
  direction="l",
 },
 [0x168DF]={
  category="lo",
  direction="l",
 },
 [0x168E0]={
  category="lo",
  direction="l",
 },
 [0x168E1]={
  category="lo",
  direction="l",
 },
 [0x168E2]={
  category="lo",
  direction="l",
 },
 [0x168E3]={
  category="lo",
  direction="l",
 },
 [0x168E4]={
  category="lo",
  direction="l",
 },
 [0x168E5]={
  category="lo",
  direction="l",
 },
 [0x168E6]={
  category="lo",
  direction="l",
 },
 [0x168E7]={
  category="lo",
  direction="l",
 },
 [0x168E8]={
  category="lo",
  direction="l",
 },
 [0x168E9]={
  category="lo",
  direction="l",
 },
 [0x168EA]={
  category="lo",
  direction="l",
 },
 [0x168EB]={
  category="lo",
  direction="l",
 },
 [0x168EC]={
  category="lo",
  direction="l",
 },
 [0x168ED]={
  category="lo",
  direction="l",
 },
 [0x168EE]={
  category="lo",
  direction="l",
 },
 [0x168EF]={
  category="lo",
  direction="l",
 },
 [0x168F0]={
  category="lo",
  direction="l",
 },
 [0x168F1]={
  category="lo",
  direction="l",
 },
 [0x168F2]={
  category="lo",
  direction="l",
 },
 [0x168F3]={
  category="lo",
  direction="l",
 },
 [0x168F4]={
  category="lo",
  direction="l",
 },
 [0x168F5]={
  category="lo",
  direction="l",
 },
 [0x168F6]={
  category="lo",
  direction="l",
 },
 [0x168F7]={
  category="lo",
  direction="l",
 },
 [0x168F8]={
  category="lo",
  direction="l",
 },
 [0x168F9]={
  category="lo",
  direction="l",
 },
 [0x168FA]={
  category="lo",
  direction="l",
 },
 [0x168FB]={
  category="lo",
  direction="l",
 },
 [0x168FC]={
  category="lo",
  direction="l",
 },
 [0x168FD]={
  category="lo",
  direction="l",
 },
 [0x168FE]={
  category="lo",
  direction="l",
 },
 [0x168FF]={
  category="lo",
  direction="l",
 },
 [0x16900]={
  category="lo",
  direction="l",
 },
 [0x16901]={
  category="lo",
  direction="l",
 },
 [0x16902]={
  category="lo",
  direction="l",
 },
 [0x16903]={
  category="lo",
  direction="l",
 },
 [0x16904]={
  category="lo",
  direction="l",
 },
 [0x16905]={
  category="lo",
  direction="l",
 },
 [0x16906]={
  category="lo",
  direction="l",
 },
 [0x16907]={
  category="lo",
  direction="l",
 },
 [0x16908]={
  category="lo",
  direction="l",
 },
 [0x16909]={
  category="lo",
  direction="l",
 },
 [0x1690A]={
  category="lo",
  direction="l",
 },
 [0x1690B]={
  category="lo",
  direction="l",
 },
 [0x1690C]={
  category="lo",
  direction="l",
 },
 [0x1690D]={
  category="lo",
  direction="l",
 },
 [0x1690E]={
  category="lo",
  direction="l",
 },
 [0x1690F]={
  category="lo",
  direction="l",
 },
 [0x16910]={
  category="lo",
  direction="l",
 },
 [0x16911]={
  category="lo",
  direction="l",
 },
 [0x16912]={
  category="lo",
  direction="l",
 },
 [0x16913]={
  category="lo",
  direction="l",
 },
 [0x16914]={
  category="lo",
  direction="l",
 },
 [0x16915]={
  category="lo",
  direction="l",
 },
 [0x16916]={
  category="lo",
  direction="l",
 },
 [0x16917]={
  category="lo",
  direction="l",
 },
 [0x16918]={
  category="lo",
  direction="l",
 },
 [0x16919]={
  category="lo",
  direction="l",
 },
 [0x1691A]={
  category="lo",
  direction="l",
 },
 [0x1691B]={
  category="lo",
  direction="l",
 },
 [0x1691C]={
  category="lo",
  direction="l",
 },
 [0x1691D]={
  category="lo",
  direction="l",
 },
 [0x1691E]={
  category="lo",
  direction="l",
 },
 [0x1691F]={
  category="lo",
  direction="l",
 },
 [0x16920]={
  category="lo",
  direction="l",
 },
 [0x16921]={
  category="lo",
  direction="l",
 },
 [0x16922]={
  category="lo",
  direction="l",
 },
 [0x16923]={
  category="lo",
  direction="l",
 },
 [0x16924]={
  category="lo",
  direction="l",
 },
 [0x16925]={
  category="lo",
  direction="l",
 },
 [0x16926]={
  category="lo",
  direction="l",
 },
 [0x16927]={
  category="lo",
  direction="l",
 },
 [0x16928]={
  category="lo",
  direction="l",
 },
 [0x16929]={
  category="lo",
  direction="l",
 },
 [0x1692A]={
  category="lo",
  direction="l",
 },
 [0x1692B]={
  category="lo",
  direction="l",
 },
 [0x1692C]={
  category="lo",
  direction="l",
 },
 [0x1692D]={
  category="lo",
  direction="l",
 },
 [0x1692E]={
  category="lo",
  direction="l",
 },
 [0x1692F]={
  category="lo",
  direction="l",
 },
 [0x16930]={
  category="lo",
  direction="l",
 },
 [0x16931]={
  category="lo",
  direction="l",
 },
 [0x16932]={
  category="lo",
  direction="l",
 },
 [0x16933]={
  category="lo",
  direction="l",
 },
 [0x16934]={
  category="lo",
  direction="l",
 },
 [0x16935]={
  category="lo",
  direction="l",
 },
 [0x16936]={
  category="lo",
  direction="l",
 },
 [0x16937]={
  category="lo",
  direction="l",
 },
 [0x16938]={
  category="lo",
  direction="l",
 },
 [0x16939]={
  category="lo",
  direction="l",
 },
 [0x1693A]={
  category="lo",
  direction="l",
 },
 [0x1693B]={
  category="lo",
  direction="l",
 },
 [0x1693C]={
  category="lo",
  direction="l",
 },
 [0x1693D]={
  category="lo",
  direction="l",
 },
 [0x1693E]={
  category="lo",
  direction="l",
 },
 [0x1693F]={
  category="lo",
  direction="l",
 },
 [0x16940]={
  category="lo",
  direction="l",
 },
 [0x16941]={
  category="lo",
  direction="l",
 },
 [0x16942]={
  category="lo",
  direction="l",
 },
 [0x16943]={
  category="lo",
  direction="l",
 },
 [0x16944]={
  category="lo",
  direction="l",
 },
 [0x16945]={
  category="lo",
  direction="l",
 },
 [0x16946]={
  category="lo",
  direction="l",
 },
 [0x16947]={
  category="lo",
  direction="l",
 },
 [0x16948]={
  category="lo",
  direction="l",
 },
 [0x16949]={
  category="lo",
  direction="l",
 },
 [0x1694A]={
  category="lo",
  direction="l",
 },
 [0x1694B]={
  category="lo",
  direction="l",
 },
 [0x1694C]={
  category="lo",
  direction="l",
 },
 [0x1694D]={
  category="lo",
  direction="l",
 },
 [0x1694E]={
  category="lo",
  direction="l",
 },
 [0x1694F]={
  category="lo",
  direction="l",
 },
 [0x16950]={
  category="lo",
  direction="l",
 },
 [0x16951]={
  category="lo",
  direction="l",
 },
 [0x16952]={
  category="lo",
  direction="l",
 },
 [0x16953]={
  category="lo",
  direction="l",
 },
 [0x16954]={
  category="lo",
  direction="l",
 },
 [0x16955]={
  category="lo",
  direction="l",
 },
 [0x16956]={
  category="lo",
  direction="l",
 },
 [0x16957]={
  category="lo",
  direction="l",
 },
 [0x16958]={
  category="lo",
  direction="l",
 },
 [0x16959]={
  category="lo",
  direction="l",
 },
 [0x1695A]={
  category="lo",
  direction="l",
 },
 [0x1695B]={
  category="lo",
  direction="l",
 },
 [0x1695C]={
  category="lo",
  direction="l",
 },
 [0x1695D]={
  category="lo",
  direction="l",
 },
 [0x1695E]={
  category="lo",
  direction="l",
 },
 [0x1695F]={
  category="lo",
  direction="l",
 },
 [0x16960]={
  category="lo",
  direction="l",
 },
 [0x16961]={
  category="lo",
  direction="l",
 },
 [0x16962]={
  category="lo",
  direction="l",
 },
 [0x16963]={
  category="lo",
  direction="l",
 },
 [0x16964]={
  category="lo",
  direction="l",
 },
 [0x16965]={
  category="lo",
  direction="l",
 },
 [0x16966]={
  category="lo",
  direction="l",
 },
 [0x16967]={
  category="lo",
  direction="l",
 },
 [0x16968]={
  category="lo",
  direction="l",
 },
 [0x16969]={
  category="lo",
  direction="l",
 },
 [0x1696A]={
  category="lo",
  direction="l",
 },
 [0x1696B]={
  category="lo",
  direction="l",
 },
 [0x1696C]={
  category="lo",
  direction="l",
 },
 [0x1696D]={
  category="lo",
  direction="l",
 },
 [0x1696E]={
  category="lo",
  direction="l",
 },
 [0x1696F]={
  category="lo",
  direction="l",
 },
 [0x16970]={
  category="lo",
  direction="l",
 },
 [0x16971]={
  category="lo",
  direction="l",
 },
 [0x16972]={
  category="lo",
  direction="l",
 },
 [0x16973]={
  category="lo",
  direction="l",
 },
 [0x16974]={
  category="lo",
  direction="l",
 },
 [0x16975]={
  category="lo",
  direction="l",
 },
 [0x16976]={
  category="lo",
  direction="l",
 },
 [0x16977]={
  category="lo",
  direction="l",
 },
 [0x16978]={
  category="lo",
  direction="l",
 },
 [0x16979]={
  category="lo",
  direction="l",
 },
 [0x1697A]={
  category="lo",
  direction="l",
 },
 [0x1697B]={
  category="lo",
  direction="l",
 },
 [0x1697C]={
  category="lo",
  direction="l",
 },
 [0x1697D]={
  category="lo",
  direction="l",
 },
 [0x1697E]={
  category="lo",
  direction="l",
 },
 [0x1697F]={
  category="lo",
  direction="l",
 },
 [0x16980]={
  category="lo",
  direction="l",
 },
 [0x16981]={
  category="lo",
  direction="l",
 },
 [0x16982]={
  category="lo",
  direction="l",
 },
 [0x16983]={
  category="lo",
  direction="l",
 },
 [0x16984]={
  category="lo",
  direction="l",
 },
 [0x16985]={
  category="lo",
  direction="l",
 },
 [0x16986]={
  category="lo",
  direction="l",
 },
 [0x16987]={
  category="lo",
  direction="l",
 },
 [0x16988]={
  category="lo",
  direction="l",
 },
 [0x16989]={
  category="lo",
  direction="l",
 },
 [0x1698A]={
  category="lo",
  direction="l",
 },
 [0x1698B]={
  category="lo",
  direction="l",
 },
 [0x1698C]={
  category="lo",
  direction="l",
 },
 [0x1698D]={
  category="lo",
  direction="l",
 },
 [0x1698E]={
  category="lo",
  direction="l",
 },
 [0x1698F]={
  category="lo",
  direction="l",
 },
 [0x16990]={
  category="lo",
  direction="l",
 },
 [0x16991]={
  category="lo",
  direction="l",
 },
 [0x16992]={
  category="lo",
  direction="l",
 },
 [0x16993]={
  category="lo",
  direction="l",
 },
 [0x16994]={
  category="lo",
  direction="l",
 },
 [0x16995]={
  category="lo",
  direction="l",
 },
 [0x16996]={
  category="lo",
  direction="l",
 },
 [0x16997]={
  category="lo",
  direction="l",
 },
 [0x16998]={
  category="lo",
  direction="l",
 },
 [0x16999]={
  category="lo",
  direction="l",
 },
 [0x1699A]={
  category="lo",
  direction="l",
 },
 [0x1699B]={
  category="lo",
  direction="l",
 },
 [0x1699C]={
  category="lo",
  direction="l",
 },
 [0x1699D]={
  category="lo",
  direction="l",
 },
 [0x1699E]={
  category="lo",
  direction="l",
 },
 [0x1699F]={
  category="lo",
  direction="l",
 },
 [0x169A0]={
  category="lo",
  direction="l",
 },
 [0x169A1]={
  category="lo",
  direction="l",
 },
 [0x169A2]={
  category="lo",
  direction="l",
 },
 [0x169A3]={
  category="lo",
  direction="l",
 },
 [0x169A4]={
  category="lo",
  direction="l",
 },
 [0x169A5]={
  category="lo",
  direction="l",
 },
 [0x169A6]={
  category="lo",
  direction="l",
 },
 [0x169A7]={
  category="lo",
  direction="l",
 },
 [0x169A8]={
  category="lo",
  direction="l",
 },
 [0x169A9]={
  category="lo",
  direction="l",
 },
 [0x169AA]={
  category="lo",
  direction="l",
 },
 [0x169AB]={
  category="lo",
  direction="l",
 },
 [0x169AC]={
  category="lo",
  direction="l",
 },
 [0x169AD]={
  category="lo",
  direction="l",
 },
 [0x169AE]={
  category="lo",
  direction="l",
 },
 [0x169AF]={
  category="lo",
  direction="l",
 },
 [0x169B0]={
  category="lo",
  direction="l",
 },
 [0x169B1]={
  category="lo",
  direction="l",
 },
 [0x169B2]={
  category="lo",
  direction="l",
 },
 [0x169B3]={
  category="lo",
  direction="l",
 },
 [0x169B4]={
  category="lo",
  direction="l",
 },
 [0x169B5]={
  category="lo",
  direction="l",
 },
 [0x169B6]={
  category="lo",
  direction="l",
 },
 [0x169B7]={
  category="lo",
  direction="l",
 },
 [0x169B8]={
  category="lo",
  direction="l",
 },
 [0x169B9]={
  category="lo",
  direction="l",
 },
 [0x169BA]={
  category="lo",
  direction="l",
 },
 [0x169BB]={
  category="lo",
  direction="l",
 },
 [0x169BC]={
  category="lo",
  direction="l",
 },
 [0x169BD]={
  category="lo",
  direction="l",
 },
 [0x169BE]={
  category="lo",
  direction="l",
 },
 [0x169BF]={
  category="lo",
  direction="l",
 },
 [0x169C0]={
  category="lo",
  direction="l",
 },
 [0x169C1]={
  category="lo",
  direction="l",
 },
 [0x169C2]={
  category="lo",
  direction="l",
 },
 [0x169C3]={
  category="lo",
  direction="l",
 },
 [0x169C4]={
  category="lo",
  direction="l",
 },
 [0x169C5]={
  category="lo",
  direction="l",
 },
 [0x169C6]={
  category="lo",
  direction="l",
 },
 [0x169C7]={
  category="lo",
  direction="l",
 },
 [0x169C8]={
  category="lo",
  direction="l",
 },
 [0x169C9]={
  category="lo",
  direction="l",
 },
 [0x169CA]={
  category="lo",
  direction="l",
 },
 [0x169CB]={
  category="lo",
  direction="l",
 },
 [0x169CC]={
  category="lo",
  direction="l",
 },
 [0x169CD]={
  category="lo",
  direction="l",
 },
 [0x169CE]={
  category="lo",
  direction="l",
 },
 [0x169CF]={
  category="lo",
  direction="l",
 },
 [0x169D0]={
  category="lo",
  direction="l",
 },
 [0x169D1]={
  category="lo",
  direction="l",
 },
 [0x169D2]={
  category="lo",
  direction="l",
 },
 [0x169D3]={
  category="lo",
  direction="l",
 },
 [0x169D4]={
  category="lo",
  direction="l",
 },
 [0x169D5]={
  category="lo",
  direction="l",
 },
 [0x169D6]={
  category="lo",
  direction="l",
 },
 [0x169D7]={
  category="lo",
  direction="l",
 },
 [0x169D8]={
  category="lo",
  direction="l",
 },
 [0x169D9]={
  category="lo",
  direction="l",
 },
 [0x169DA]={
  category="lo",
  direction="l",
 },
 [0x169DB]={
  category="lo",
  direction="l",
 },
 [0x169DC]={
  category="lo",
  direction="l",
 },
 [0x169DD]={
  category="lo",
  direction="l",
 },
 [0x169DE]={
  category="lo",
  direction="l",
 },
 [0x169DF]={
  category="lo",
  direction="l",
 },
 [0x169E0]={
  category="lo",
  direction="l",
 },
 [0x169E1]={
  category="lo",
  direction="l",
 },
 [0x169E2]={
  category="lo",
  direction="l",
 },
 [0x169E3]={
  category="lo",
  direction="l",
 },
 [0x169E4]={
  category="lo",
  direction="l",
 },
 [0x169E5]={
  category="lo",
  direction="l",
 },
 [0x169E6]={
  category="lo",
  direction="l",
 },
 [0x169E7]={
  category="lo",
  direction="l",
 },
 [0x169E8]={
  category="lo",
  direction="l",
 },
 [0x169E9]={
  category="lo",
  direction="l",
 },
 [0x169EA]={
  category="lo",
  direction="l",
 },
 [0x169EB]={
  category="lo",
  direction="l",
 },
 [0x169EC]={
  category="lo",
  direction="l",
 },
 [0x169ED]={
  category="lo",
  direction="l",
 },
 [0x169EE]={
  category="lo",
  direction="l",
 },
 [0x169EF]={
  category="lo",
  direction="l",
 },
 [0x169F0]={
  category="lo",
  direction="l",
 },
 [0x169F1]={
  category="lo",
  direction="l",
 },
 [0x169F2]={
  category="lo",
  direction="l",
 },
 [0x169F3]={
  category="lo",
  direction="l",
 },
 [0x169F4]={
  category="lo",
  direction="l",
 },
 [0x169F5]={
  category="lo",
  direction="l",
 },
 [0x169F6]={
  category="lo",
  direction="l",
 },
 [0x169F7]={
  category="lo",
  direction="l",
 },
 [0x169F8]={
  category="lo",
  direction="l",
 },
 [0x169F9]={
  category="lo",
  direction="l",
 },
 [0x169FA]={
  category="lo",
  direction="l",
 },
 [0x169FB]={
  category="lo",
  direction="l",
 },
 [0x169FC]={
  category="lo",
  direction="l",
 },
 [0x169FD]={
  category="lo",
  direction="l",
 },
 [0x169FE]={
  category="lo",
  direction="l",
 },
 [0x169FF]={
  category="lo",
  direction="l",
 },
 [0x16A00]={
  category="lo",
  direction="l",
 },
 [0x16A01]={
  category="lo",
  direction="l",
 },
 [0x16A02]={
  category="lo",
  direction="l",
 },
 [0x16A03]={
  category="lo",
  direction="l",
 },
 [0x16A04]={
  category="lo",
  direction="l",
 },
 [0x16A05]={
  category="lo",
  direction="l",
 },
 [0x16A06]={
  category="lo",
  direction="l",
 },
 [0x16A07]={
  category="lo",
  direction="l",
 },
 [0x16A08]={
  category="lo",
  direction="l",
 },
 [0x16A09]={
  category="lo",
  direction="l",
 },
 [0x16A0A]={
  category="lo",
  direction="l",
 },
 [0x16A0B]={
  category="lo",
  direction="l",
 },
 [0x16A0C]={
  category="lo",
  direction="l",
 },
 [0x16A0D]={
  category="lo",
  direction="l",
 },
 [0x16A0E]={
  category="lo",
  direction="l",
 },
 [0x16A0F]={
  category="lo",
  direction="l",
 },
 [0x16A10]={
  category="lo",
  direction="l",
 },
 [0x16A11]={
  category="lo",
  direction="l",
 },
 [0x16A12]={
  category="lo",
  direction="l",
 },
 [0x16A13]={
  category="lo",
  direction="l",
 },
 [0x16A14]={
  category="lo",
  direction="l",
 },
 [0x16A15]={
  category="lo",
  direction="l",
 },
 [0x16A16]={
  category="lo",
  direction="l",
 },
 [0x16A17]={
  category="lo",
  direction="l",
 },
 [0x16A18]={
  category="lo",
  direction="l",
 },
 [0x16A19]={
  category="lo",
  direction="l",
 },
 [0x16A1A]={
  category="lo",
  direction="l",
 },
 [0x16A1B]={
  category="lo",
  direction="l",
 },
 [0x16A1C]={
  category="lo",
  direction="l",
 },
 [0x16A1D]={
  category="lo",
  direction="l",
 },
 [0x16A1E]={
  category="lo",
  direction="l",
 },
 [0x16A1F]={
  category="lo",
  direction="l",
 },
 [0x16A20]={
  category="lo",
  direction="l",
 },
 [0x16A21]={
  category="lo",
  direction="l",
 },
 [0x16A22]={
  category="lo",
  direction="l",
 },
 [0x16A23]={
  category="lo",
  direction="l",
 },
 [0x16A24]={
  category="lo",
  direction="l",
 },
 [0x16A25]={
  category="lo",
  direction="l",
 },
 [0x16A26]={
  category="lo",
  direction="l",
 },
 [0x16A27]={
  category="lo",
  direction="l",
 },
 [0x16A28]={
  category="lo",
  direction="l",
 },
 [0x16A29]={
  category="lo",
  direction="l",
 },
 [0x16A2A]={
  category="lo",
  direction="l",
 },
 [0x16A2B]={
  category="lo",
  direction="l",
 },
 [0x16A2C]={
  category="lo",
  direction="l",
 },
 [0x16A2D]={
  category="lo",
  direction="l",
 },
 [0x16A2E]={
  category="lo",
  direction="l",
 },
 [0x16A2F]={
  category="lo",
  direction="l",
 },
 [0x16A30]={
  category="lo",
  direction="l",
 },
 [0x16A31]={
  category="lo",
  direction="l",
 },
 [0x16A32]={
  category="lo",
  direction="l",
 },
 [0x16A33]={
  category="lo",
  direction="l",
 },
 [0x16A34]={
  category="lo",
  direction="l",
 },
 [0x16A35]={
  category="lo",
  direction="l",
 },
 [0x16A36]={
  category="lo",
  direction="l",
 },
 [0x16A37]={
  category="lo",
  direction="l",
 },
 [0x16A38]={
  category="lo",
  direction="l",
 },
 [0x16A40]={
  category="lo",
  direction="l",
 },
 [0x16A41]={
  category="lo",
  direction="l",
 },
 [0x16A42]={
  category="lo",
  direction="l",
 },
 [0x16A43]={
  category="lo",
  direction="l",
 },
 [0x16A44]={
  category="lo",
  direction="l",
 },
 [0x16A45]={
  category="lo",
  direction="l",
 },
 [0x16A46]={
  category="lo",
  direction="l",
 },
 [0x16A47]={
  category="lo",
  direction="l",
 },
 [0x16A48]={
  category="lo",
  direction="l",
 },
 [0x16A49]={
  category="lo",
  direction="l",
 },
 [0x16A4A]={
  category="lo",
  direction="l",
 },
 [0x16A4B]={
  category="lo",
  direction="l",
 },
 [0x16A4C]={
  category="lo",
  direction="l",
 },
 [0x16A4D]={
  category="lo",
  direction="l",
 },
 [0x16A4E]={
  category="lo",
  direction="l",
 },
 [0x16A4F]={
  category="lo",
  direction="l",
 },
 [0x16A50]={
  category="lo",
  direction="l",
 },
 [0x16A51]={
  category="lo",
  direction="l",
 },
 [0x16A52]={
  category="lo",
  direction="l",
 },
 [0x16A53]={
  category="lo",
  direction="l",
 },
 [0x16A54]={
  category="lo",
  direction="l",
 },
 [0x16A55]={
  category="lo",
  direction="l",
 },
 [0x16A56]={
  category="lo",
  direction="l",
 },
 [0x16A57]={
  category="lo",
  direction="l",
 },
 [0x16A58]={
  category="lo",
  direction="l",
 },
 [0x16A59]={
  category="lo",
  direction="l",
 },
 [0x16A5A]={
  category="lo",
  direction="l",
 },
 [0x16A5B]={
  category="lo",
  direction="l",
 },
 [0x16A5C]={
  category="lo",
  direction="l",
 },
 [0x16A5D]={
  category="lo",
  direction="l",
 },
 [0x16A5E]={
  category="lo",
  direction="l",
 },
 [0x16A60]={
  category="nd",
  direction="l",
 },
 [0x16A61]={
  category="nd",
  direction="l",
 },
 [0x16A62]={
  category="nd",
  direction="l",
 },
 [0x16A63]={
  category="nd",
  direction="l",
 },
 [0x16A64]={
  category="nd",
  direction="l",
 },
 [0x16A65]={
  category="nd",
  direction="l",
 },
 [0x16A66]={
  category="nd",
  direction="l",
 },
 [0x16A67]={
  category="nd",
  direction="l",
 },
 [0x16A68]={
  category="nd",
  direction="l",
 },
 [0x16A69]={
  category="nd",
  direction="l",
 },
 [0x16A6E]={
  category="po",
  direction="l",
 },
 [0x16A6F]={
  category="po",
  direction="l",
 },
 [0x16AD0]={
  category="lo",
  direction="l",
 },
 [0x16AD1]={
  category="lo",
  direction="l",
 },
 [0x16AD2]={
  category="lo",
  direction="l",
 },
 [0x16AD3]={
  category="lo",
  direction="l",
 },
 [0x16AD4]={
  category="lo",
  direction="l",
 },
 [0x16AD5]={
  category="lo",
  direction="l",
 },
 [0x16AD6]={
  category="lo",
  direction="l",
 },
 [0x16AD7]={
  category="lo",
  direction="l",
 },
 [0x16AD8]={
  category="lo",
  direction="l",
 },
 [0x16AD9]={
  category="lo",
  direction="l",
 },
 [0x16ADA]={
  category="lo",
  direction="l",
 },
 [0x16ADB]={
  category="lo",
  direction="l",
 },
 [0x16ADC]={
  category="lo",
  direction="l",
 },
 [0x16ADD]={
  category="lo",
  direction="l",
 },
 [0x16ADE]={
  category="lo",
  direction="l",
 },
 [0x16ADF]={
  category="lo",
  direction="l",
 },
 [0x16AE0]={
  category="lo",
  direction="l",
 },
 [0x16AE1]={
  category="lo",
  direction="l",
 },
 [0x16AE2]={
  category="lo",
  direction="l",
 },
 [0x16AE3]={
  category="lo",
  direction="l",
 },
 [0x16AE4]={
  category="lo",
  direction="l",
 },
 [0x16AE5]={
  category="lo",
  direction="l",
 },
 [0x16AE6]={
  category="lo",
  direction="l",
 },
 [0x16AE7]={
  category="lo",
  direction="l",
 },
 [0x16AE8]={
  category="lo",
  direction="l",
 },
 [0x16AE9]={
  category="lo",
  direction="l",
 },
 [0x16AEA]={
  category="lo",
  direction="l",
 },
 [0x16AEB]={
  category="lo",
  direction="l",
 },
 [0x16AEC]={
  category="lo",
  direction="l",
 },
 [0x16AED]={
  category="lo",
  direction="l",
 },
 [0x16AF0]={
  category="mn",
  direction="nsm",
 },
 [0x16AF1]={
  category="mn",
  direction="nsm",
 },
 [0x16AF2]={
  category="mn",
  direction="nsm",
 },
 [0x16AF3]={
  category="mn",
  direction="nsm",
 },
 [0x16AF4]={
  category="mn",
  direction="nsm",
 },
 [0x16AF5]={
  category="po",
  direction="l",
 },
 [0x16B00]={
  category="lo",
  direction="l",
 },
 [0x16B01]={
  category="lo",
  direction="l",
 },
 [0x16B02]={
  category="lo",
  direction="l",
 },
 [0x16B03]={
  category="lo",
  direction="l",
 },
 [0x16B04]={
  category="lo",
  direction="l",
 },
 [0x16B05]={
  category="lo",
  direction="l",
 },
 [0x16B06]={
  category="lo",
  direction="l",
 },
 [0x16B07]={
  category="lo",
  direction="l",
 },
 [0x16B08]={
  category="lo",
  direction="l",
 },
 [0x16B09]={
  category="lo",
  direction="l",
 },
 [0x16B0A]={
  category="lo",
  direction="l",
 },
 [0x16B0B]={
  category="lo",
  direction="l",
 },
 [0x16B0C]={
  category="lo",
  direction="l",
 },
 [0x16B0D]={
  category="lo",
  direction="l",
 },
 [0x16B0E]={
  category="lo",
  direction="l",
 },
 [0x16B0F]={
  category="lo",
  direction="l",
 },
 [0x16B10]={
  category="lo",
  direction="l",
 },
 [0x16B11]={
  category="lo",
  direction="l",
 },
 [0x16B12]={
  category="lo",
  direction="l",
 },
 [0x16B13]={
  category="lo",
  direction="l",
 },
 [0x16B14]={
  category="lo",
  direction="l",
 },
 [0x16B15]={
  category="lo",
  direction="l",
 },
 [0x16B16]={
  category="lo",
  direction="l",
 },
 [0x16B17]={
  category="lo",
  direction="l",
 },
 [0x16B18]={
  category="lo",
  direction="l",
 },
 [0x16B19]={
  category="lo",
  direction="l",
 },
 [0x16B1A]={
  category="lo",
  direction="l",
 },
 [0x16B1B]={
  category="lo",
  direction="l",
 },
 [0x16B1C]={
  category="lo",
  direction="l",
 },
 [0x16B1D]={
  category="lo",
  direction="l",
 },
 [0x16B1E]={
  category="lo",
  direction="l",
 },
 [0x16B1F]={
  category="lo",
  direction="l",
 },
 [0x16B20]={
  category="lo",
  direction="l",
 },
 [0x16B21]={
  category="lo",
  direction="l",
 },
 [0x16B22]={
  category="lo",
  direction="l",
 },
 [0x16B23]={
  category="lo",
  direction="l",
 },
 [0x16B24]={
  category="lo",
  direction="l",
 },
 [0x16B25]={
  category="lo",
  direction="l",
 },
 [0x16B26]={
  category="lo",
  direction="l",
 },
 [0x16B27]={
  category="lo",
  direction="l",
 },
 [0x16B28]={
  category="lo",
  direction="l",
 },
 [0x16B29]={
  category="lo",
  direction="l",
 },
 [0x16B2A]={
  category="lo",
  direction="l",
 },
 [0x16B2B]={
  category="lo",
  direction="l",
 },
 [0x16B2C]={
  category="lo",
  direction="l",
 },
 [0x16B2D]={
  category="lo",
  direction="l",
 },
 [0x16B2E]={
  category="lo",
  direction="l",
 },
 [0x16B2F]={
  category="lo",
  direction="l",
 },
 [0x16B30]={
  category="mn",
  direction="nsm",
 },
 [0x16B31]={
  category="mn",
  direction="nsm",
 },
 [0x16B32]={
  category="mn",
  direction="nsm",
 },
 [0x16B33]={
  category="mn",
  direction="nsm",
 },
 [0x16B34]={
  category="mn",
  direction="nsm",
 },
 [0x16B35]={
  category="mn",
  direction="nsm",
 },
 [0x16B36]={
  category="mn",
  direction="nsm",
 },
 [0x16B37]={
  category="po",
  direction="l",
 },
 [0x16B38]={
  category="po",
  direction="l",
 },
 [0x16B39]={
  category="po",
  direction="l",
 },
 [0x16B3A]={
  category="po",
  direction="l",
 },
 [0x16B3B]={
  category="po",
  direction="l",
 },
 [0x16B3C]={
  category="so",
  direction="l",
 },
 [0x16B3D]={
  category="so",
  direction="l",
 },
 [0x16B3E]={
  category="so",
  direction="l",
 },
 [0x16B3F]={
  category="so",
  direction="l",
 },
 [0x16B40]={
  category="lm",
  direction="l",
 },
 [0x16B41]={
  category="lm",
  direction="l",
 },
 [0x16B42]={
  category="lm",
  direction="l",
 },
 [0x16B43]={
  category="lm",
  direction="l",
 },
 [0x16B44]={
  category="po",
  direction="l",
 },
 [0x16B45]={
  category="so",
  direction="l",
 },
 [0x16B50]={
  category="nd",
  direction="l",
 },
 [0x16B51]={
  category="nd",
  direction="l",
 },
 [0x16B52]={
  category="nd",
  direction="l",
 },
 [0x16B53]={
  category="nd",
  direction="l",
 },
 [0x16B54]={
  category="nd",
  direction="l",
 },
 [0x16B55]={
  category="nd",
  direction="l",
 },
 [0x16B56]={
  category="nd",
  direction="l",
 },
 [0x16B57]={
  category="nd",
  direction="l",
 },
 [0x16B58]={
  category="nd",
  direction="l",
 },
 [0x16B59]={
  category="nd",
  direction="l",
 },
 [0x16B5B]={
  category="no",
  direction="l",
 },
 [0x16B5C]={
  category="no",
  direction="l",
 },
 [0x16B5D]={
  category="no",
  direction="l",
 },
 [0x16B5E]={
  category="no",
  direction="l",
 },
 [0x16B5F]={
  category="no",
  direction="l",
 },
 [0x16B60]={
  category="no",
  direction="l",
 },
 [0x16B61]={
  category="no",
  direction="l",
 },
 [0x16B63]={
  category="lo",
  direction="l",
 },
 [0x16B64]={
  category="lo",
  direction="l",
 },
 [0x16B65]={
  category="lo",
  direction="l",
 },
 [0x16B66]={
  category="lo",
  direction="l",
 },
 [0x16B67]={
  category="lo",
  direction="l",
 },
 [0x16B68]={
  category="lo",
  direction="l",
 },
 [0x16B69]={
  category="lo",
  direction="l",
 },
 [0x16B6A]={
  category="lo",
  direction="l",
 },
 [0x16B6B]={
  category="lo",
  direction="l",
 },
 [0x16B6C]={
  category="lo",
  direction="l",
 },
 [0x16B6D]={
  category="lo",
  direction="l",
 },
 [0x16B6E]={
  category="lo",
  direction="l",
 },
 [0x16B6F]={
  category="lo",
  direction="l",
 },
 [0x16B70]={
  category="lo",
  direction="l",
 },
 [0x16B71]={
  category="lo",
  direction="l",
 },
 [0x16B72]={
  category="lo",
  direction="l",
 },
 [0x16B73]={
  category="lo",
  direction="l",
 },
 [0x16B74]={
  category="lo",
  direction="l",
 },
 [0x16B75]={
  category="lo",
  direction="l",
 },
 [0x16B76]={
  category="lo",
  direction="l",
 },
 [0x16B77]={
  category="lo",
  direction="l",
 },
 [0x16B7D]={
  category="lo",
  direction="l",
 },
 [0x16B7E]={
  category="lo",
  direction="l",
 },
 [0x16B7F]={
  category="lo",
  direction="l",
 },
 [0x16B80]={
  category="lo",
  direction="l",
 },
 [0x16B81]={
  category="lo",
  direction="l",
 },
 [0x16B82]={
  category="lo",
  direction="l",
 },
 [0x16B83]={
  category="lo",
  direction="l",
 },
 [0x16B84]={
  category="lo",
  direction="l",
 },
 [0x16B85]={
  category="lo",
  direction="l",
 },
 [0x16B86]={
  category="lo",
  direction="l",
 },
 [0x16B87]={
  category="lo",
  direction="l",
 },
 [0x16B88]={
  category="lo",
  direction="l",
 },
 [0x16B89]={
  category="lo",
  direction="l",
 },
 [0x16B8A]={
  category="lo",
  direction="l",
 },
 [0x16B8B]={
  category="lo",
  direction="l",
 },
 [0x16B8C]={
  category="lo",
  direction="l",
 },
 [0x16B8D]={
  category="lo",
  direction="l",
 },
 [0x16B8E]={
  category="lo",
  direction="l",
 },
 [0x16B8F]={
  category="lo",
  direction="l",
 },
 [0x16F00]={
  category="lo",
  direction="l",
 },
 [0x16F01]={
  category="lo",
  direction="l",
 },
 [0x16F02]={
  category="lo",
  direction="l",
 },
 [0x16F03]={
  category="lo",
  direction="l",
 },
 [0x16F04]={
  category="lo",
  direction="l",
 },
 [0x16F05]={
  category="lo",
  direction="l",
 },
 [0x16F06]={
  category="lo",
  direction="l",
 },
 [0x16F07]={
  category="lo",
  direction="l",
 },
 [0x16F08]={
  category="lo",
  direction="l",
 },
 [0x16F09]={
  category="lo",
  direction="l",
 },
 [0x16F0A]={
  category="lo",
  direction="l",
 },
 [0x16F0B]={
  category="lo",
  direction="l",
 },
 [0x16F0C]={
  category="lo",
  direction="l",
 },
 [0x16F0D]={
  category="lo",
  direction="l",
 },
 [0x16F0E]={
  category="lo",
  direction="l",
 },
 [0x16F0F]={
  category="lo",
  direction="l",
 },
 [0x16F10]={
  category="lo",
  direction="l",
 },
 [0x16F11]={
  category="lo",
  direction="l",
 },
 [0x16F12]={
  category="lo",
  direction="l",
 },
 [0x16F13]={
  category="lo",
  direction="l",
 },
 [0x16F14]={
  category="lo",
  direction="l",
 },
 [0x16F15]={
  category="lo",
  direction="l",
 },
 [0x16F16]={
  category="lo",
  direction="l",
 },
 [0x16F17]={
  category="lo",
  direction="l",
 },
 [0x16F18]={
  category="lo",
  direction="l",
 },
 [0x16F19]={
  category="lo",
  direction="l",
 },
 [0x16F1A]={
  category="lo",
  direction="l",
 },
 [0x16F1B]={
  category="lo",
  direction="l",
 },
 [0x16F1C]={
  category="lo",
  direction="l",
 },
 [0x16F1D]={
  category="lo",
  direction="l",
 },
 [0x16F1E]={
  category="lo",
  direction="l",
 },
 [0x16F1F]={
  category="lo",
  direction="l",
 },
 [0x16F20]={
  category="lo",
  direction="l",
 },
 [0x16F21]={
  category="lo",
  direction="l",
 },
 [0x16F22]={
  category="lo",
  direction="l",
 },
 [0x16F23]={
  category="lo",
  direction="l",
 },
 [0x16F24]={
  category="lo",
  direction="l",
 },
 [0x16F25]={
  category="lo",
  direction="l",
 },
 [0x16F26]={
  category="lo",
  direction="l",
 },
 [0x16F27]={
  category="lo",
  direction="l",
 },
 [0x16F28]={
  category="lo",
  direction="l",
 },
 [0x16F29]={
  category="lo",
  direction="l",
 },
 [0x16F2A]={
  category="lo",
  direction="l",
 },
 [0x16F2B]={
  category="lo",
  direction="l",
 },
 [0x16F2C]={
  category="lo",
  direction="l",
 },
 [0x16F2D]={
  category="lo",
  direction="l",
 },
 [0x16F2E]={
  category="lo",
  direction="l",
 },
 [0x16F2F]={
  category="lo",
  direction="l",
 },
 [0x16F30]={
  category="lo",
  direction="l",
 },
 [0x16F31]={
  category="lo",
  direction="l",
 },
 [0x16F32]={
  category="lo",
  direction="l",
 },
 [0x16F33]={
  category="lo",
  direction="l",
 },
 [0x16F34]={
  category="lo",
  direction="l",
 },
 [0x16F35]={
  category="lo",
  direction="l",
 },
 [0x16F36]={
  category="lo",
  direction="l",
 },
 [0x16F37]={
  category="lo",
  direction="l",
 },
 [0x16F38]={
  category="lo",
  direction="l",
 },
 [0x16F39]={
  category="lo",
  direction="l",
 },
 [0x16F3A]={
  category="lo",
  direction="l",
 },
 [0x16F3B]={
  category="lo",
  direction="l",
 },
 [0x16F3C]={
  category="lo",
  direction="l",
 },
 [0x16F3D]={
  category="lo",
  direction="l",
 },
 [0x16F3E]={
  category="lo",
  direction="l",
 },
 [0x16F3F]={
  category="lo",
  direction="l",
 },
 [0x16F40]={
  category="lo",
  direction="l",
 },
 [0x16F41]={
  category="lo",
  direction="l",
 },
 [0x16F42]={
  category="lo",
  direction="l",
 },
 [0x16F43]={
  category="lo",
  direction="l",
 },
 [0x16F44]={
  category="lo",
  direction="l",
 },
 [0x16F50]={
  category="lo",
  direction="l",
 },
 [0x16F51]={
  category="mc",
  direction="l",
 },
 [0x16F52]={
  category="mc",
  direction="l",
 },
 [0x16F53]={
  category="mc",
  direction="l",
 },
 [0x16F54]={
  category="mc",
  direction="l",
 },
 [0x16F55]={
  category="mc",
  direction="l",
 },
 [0x16F56]={
  category="mc",
  direction="l",
 },
 [0x16F57]={
  category="mc",
  direction="l",
 },
 [0x16F58]={
  category="mc",
  direction="l",
 },
 [0x16F59]={
  category="mc",
  direction="l",
 },
 [0x16F5A]={
  category="mc",
  direction="l",
 },
 [0x16F5B]={
  category="mc",
  direction="l",
 },
 [0x16F5C]={
  category="mc",
  direction="l",
 },
 [0x16F5D]={
  category="mc",
  direction="l",
 },
 [0x16F5E]={
  category="mc",
  direction="l",
 },
 [0x16F5F]={
  category="mc",
  direction="l",
 },
 [0x16F60]={
  category="mc",
  direction="l",
 },
 [0x16F61]={
  category="mc",
  direction="l",
 },
 [0x16F62]={
  category="mc",
  direction="l",
 },
 [0x16F63]={
  category="mc",
  direction="l",
 },
 [0x16F64]={
  category="mc",
  direction="l",
 },
 [0x16F65]={
  category="mc",
  direction="l",
 },
 [0x16F66]={
  category="mc",
  direction="l",
 },
 [0x16F67]={
  category="mc",
  direction="l",
 },
 [0x16F68]={
  category="mc",
  direction="l",
 },
 [0x16F69]={
  category="mc",
  direction="l",
 },
 [0x16F6A]={
  category="mc",
  direction="l",
 },
 [0x16F6B]={
  category="mc",
  direction="l",
 },
 [0x16F6C]={
  category="mc",
  direction="l",
 },
 [0x16F6D]={
  category="mc",
  direction="l",
 },
 [0x16F6E]={
  category="mc",
  direction="l",
 },
 [0x16F6F]={
  category="mc",
  direction="l",
 },
 [0x16F70]={
  category="mc",
  direction="l",
 },
 [0x16F71]={
  category="mc",
  direction="l",
 },
 [0x16F72]={
  category="mc",
  direction="l",
 },
 [0x16F73]={
  category="mc",
  direction="l",
 },
 [0x16F74]={
  category="mc",
  direction="l",
 },
 [0x16F75]={
  category="mc",
  direction="l",
 },
 [0x16F76]={
  category="mc",
  direction="l",
 },
 [0x16F77]={
  category="mc",
  direction="l",
 },
 [0x16F78]={
  category="mc",
  direction="l",
 },
 [0x16F79]={
  category="mc",
  direction="l",
 },
 [0x16F7A]={
  category="mc",
  direction="l",
 },
 [0x16F7B]={
  category="mc",
  direction="l",
 },
 [0x16F7C]={
  category="mc",
  direction="l",
 },
 [0x16F7D]={
  category="mc",
  direction="l",
 },
 [0x16F7E]={
  category="mc",
  direction="l",
 },
 [0x16F8F]={
  category="mn",
  direction="nsm",
 },
 [0x16F90]={
  category="mn",
  direction="nsm",
 },
 [0x16F91]={
  category="mn",
  direction="nsm",
 },
 [0x16F92]={
  category="mn",
  direction="nsm",
 },
 [0x16F93]={
  category="lm",
  direction="l",
 },
 [0x16F94]={
  category="lm",
  direction="l",
 },
 [0x16F95]={
  category="lm",
  direction="l",
 },
 [0x16F96]={
  category="lm",
  direction="l",
 },
 [0x16F97]={
  category="lm",
  direction="l",
 },
 [0x16F98]={
  category="lm",
  direction="l",
 },
 [0x16F99]={
  category="lm",
  direction="l",
 },
 [0x16F9A]={
  category="lm",
  direction="l",
 },
 [0x16F9B]={
  category="lm",
  direction="l",
 },
 [0x16F9C]={
  category="lm",
  direction="l",
 },
 [0x16F9D]={
  category="lm",
  direction="l",
 },
 [0x16F9E]={
  category="lm",
  direction="l",
 },
 [0x16F9F]={
  category="lm",
  direction="l",
 },
 [0x1B000]={
  category="lo",
  direction="l",
 },
 [0x1B001]={
  category="lo",
  direction="l",
 },
 [0x1BC00]={
  category="lo",
  direction="l",
 },
 [0x1BC01]={
  category="lo",
  direction="l",
 },
 [0x1BC02]={
  category="lo",
  direction="l",
 },
 [0x1BC03]={
  category="lo",
  direction="l",
 },
 [0x1BC04]={
  category="lo",
  direction="l",
 },
 [0x1BC05]={
  category="lo",
  direction="l",
 },
 [0x1BC06]={
  category="lo",
  direction="l",
 },
 [0x1BC07]={
  category="lo",
  direction="l",
 },
 [0x1BC08]={
  category="lo",
  direction="l",
 },
 [0x1BC09]={
  category="lo",
  direction="l",
 },
 [0x1BC0A]={
  category="lo",
  direction="l",
 },
 [0x1BC0B]={
  category="lo",
  direction="l",
 },
 [0x1BC0C]={
  category="lo",
  direction="l",
 },
 [0x1BC0D]={
  category="lo",
  direction="l",
 },
 [0x1BC0E]={
  category="lo",
  direction="l",
 },
 [0x1BC0F]={
  category="lo",
  direction="l",
 },
 [0x1BC10]={
  category="lo",
  direction="l",
 },
 [0x1BC11]={
  category="lo",
  direction="l",
 },
 [0x1BC12]={
  category="lo",
  direction="l",
 },
 [0x1BC13]={
  category="lo",
  direction="l",
 },
 [0x1BC14]={
  category="lo",
  direction="l",
 },
 [0x1BC15]={
  category="lo",
  direction="l",
 },
 [0x1BC16]={
  category="lo",
  direction="l",
 },
 [0x1BC17]={
  category="lo",
  direction="l",
 },
 [0x1BC18]={
  category="lo",
  direction="l",
 },
 [0x1BC19]={
  category="lo",
  direction="l",
 },
 [0x1BC1A]={
  category="lo",
  direction="l",
 },
 [0x1BC1B]={
  category="lo",
  direction="l",
 },
 [0x1BC1C]={
  category="lo",
  direction="l",
 },
 [0x1BC1D]={
  category="lo",
  direction="l",
 },
 [0x1BC1E]={
  category="lo",
  direction="l",
 },
 [0x1BC1F]={
  category="lo",
  direction="l",
 },
 [0x1BC20]={
  category="lo",
  direction="l",
 },
 [0x1BC21]={
  category="lo",
  direction="l",
 },
 [0x1BC22]={
  category="lo",
  direction="l",
 },
 [0x1BC23]={
  category="lo",
  direction="l",
 },
 [0x1BC24]={
  category="lo",
  direction="l",
 },
 [0x1BC25]={
  category="lo",
  direction="l",
 },
 [0x1BC26]={
  category="lo",
  direction="l",
 },
 [0x1BC27]={
  category="lo",
  direction="l",
 },
 [0x1BC28]={
  category="lo",
  direction="l",
 },
 [0x1BC29]={
  category="lo",
  direction="l",
 },
 [0x1BC2A]={
  category="lo",
  direction="l",
 },
 [0x1BC2B]={
  category="lo",
  direction="l",
 },
 [0x1BC2C]={
  category="lo",
  direction="l",
 },
 [0x1BC2D]={
  category="lo",
  direction="l",
 },
 [0x1BC2E]={
  category="lo",
  direction="l",
 },
 [0x1BC2F]={
  category="lo",
  direction="l",
 },
 [0x1BC30]={
  category="lo",
  direction="l",
 },
 [0x1BC31]={
  category="lo",
  direction="l",
 },
 [0x1BC32]={
  category="lo",
  direction="l",
 },
 [0x1BC33]={
  category="lo",
  direction="l",
 },
 [0x1BC34]={
  category="lo",
  direction="l",
 },
 [0x1BC35]={
  category="lo",
  direction="l",
 },
 [0x1BC36]={
  category="lo",
  direction="l",
 },
 [0x1BC37]={
  category="lo",
  direction="l",
 },
 [0x1BC38]={
  category="lo",
  direction="l",
 },
 [0x1BC39]={
  category="lo",
  direction="l",
 },
 [0x1BC3A]={
  category="lo",
  direction="l",
 },
 [0x1BC3B]={
  category="lo",
  direction="l",
 },
 [0x1BC3C]={
  category="lo",
  direction="l",
 },
 [0x1BC3D]={
  category="lo",
  direction="l",
 },
 [0x1BC3E]={
  category="lo",
  direction="l",
 },
 [0x1BC3F]={
  category="lo",
  direction="l",
 },
 [0x1BC40]={
  category="lo",
  direction="l",
 },
 [0x1BC41]={
  category="lo",
  direction="l",
 },
 [0x1BC42]={
  category="lo",
  direction="l",
 },
 [0x1BC43]={
  category="lo",
  direction="l",
 },
 [0x1BC44]={
  category="lo",
  direction="l",
 },
 [0x1BC45]={
  category="lo",
  direction="l",
 },
 [0x1BC46]={
  category="lo",
  direction="l",
 },
 [0x1BC47]={
  category="lo",
  direction="l",
 },
 [0x1BC48]={
  category="lo",
  direction="l",
 },
 [0x1BC49]={
  category="lo",
  direction="l",
 },
 [0x1BC4A]={
  category="lo",
  direction="l",
 },
 [0x1BC4B]={
  category="lo",
  direction="l",
 },
 [0x1BC4C]={
  category="lo",
  direction="l",
 },
 [0x1BC4D]={
  category="lo",
  direction="l",
 },
 [0x1BC4E]={
  category="lo",
  direction="l",
 },
 [0x1BC4F]={
  category="lo",
  direction="l",
 },
 [0x1BC50]={
  category="lo",
  direction="l",
 },
 [0x1BC51]={
  category="lo",
  direction="l",
 },
 [0x1BC52]={
  category="lo",
  direction="l",
 },
 [0x1BC53]={
  category="lo",
  direction="l",
 },
 [0x1BC54]={
  category="lo",
  direction="l",
 },
 [0x1BC55]={
  category="lo",
  direction="l",
 },
 [0x1BC56]={
  category="lo",
  direction="l",
 },
 [0x1BC57]={
  category="lo",
  direction="l",
 },
 [0x1BC58]={
  category="lo",
  direction="l",
 },
 [0x1BC59]={
  category="lo",
  direction="l",
 },
 [0x1BC5A]={
  category="lo",
  direction="l",
 },
 [0x1BC5B]={
  category="lo",
  direction="l",
 },
 [0x1BC5C]={
  category="lo",
  direction="l",
 },
 [0x1BC5D]={
  category="lo",
  direction="l",
 },
 [0x1BC5E]={
  category="lo",
  direction="l",
 },
 [0x1BC5F]={
  category="lo",
  direction="l",
 },
 [0x1BC60]={
  category="lo",
  direction="l",
 },
 [0x1BC61]={
  category="lo",
  direction="l",
 },
 [0x1BC62]={
  category="lo",
  direction="l",
 },
 [0x1BC63]={
  category="lo",
  direction="l",
 },
 [0x1BC64]={
  category="lo",
  direction="l",
 },
 [0x1BC65]={
  category="lo",
  direction="l",
 },
 [0x1BC66]={
  category="lo",
  direction="l",
 },
 [0x1BC67]={
  category="lo",
  direction="l",
 },
 [0x1BC68]={
  category="lo",
  direction="l",
 },
 [0x1BC69]={
  category="lo",
  direction="l",
 },
 [0x1BC6A]={
  category="lo",
  direction="l",
 },
 [0x1BC70]={
  category="lo",
  direction="l",
 },
 [0x1BC71]={
  category="lo",
  direction="l",
 },
 [0x1BC72]={
  category="lo",
  direction="l",
 },
 [0x1BC73]={
  category="lo",
  direction="l",
 },
 [0x1BC74]={
  category="lo",
  direction="l",
 },
 [0x1BC75]={
  category="lo",
  direction="l",
 },
 [0x1BC76]={
  category="lo",
  direction="l",
 },
 [0x1BC77]={
  category="lo",
  direction="l",
 },
 [0x1BC78]={
  category="lo",
  direction="l",
 },
 [0x1BC79]={
  category="lo",
  direction="l",
 },
 [0x1BC7A]={
  category="lo",
  direction="l",
 },
 [0x1BC7B]={
  category="lo",
  direction="l",
 },
 [0x1BC7C]={
  category="lo",
  direction="l",
 },
 [0x1BC80]={
  category="lo",
  direction="l",
 },
 [0x1BC81]={
  category="lo",
  direction="l",
 },
 [0x1BC82]={
  category="lo",
  direction="l",
 },
 [0x1BC83]={
  category="lo",
  direction="l",
 },
 [0x1BC84]={
  category="lo",
  direction="l",
 },
 [0x1BC85]={
  category="lo",
  direction="l",
 },
 [0x1BC86]={
  category="lo",
  direction="l",
 },
 [0x1BC87]={
  category="lo",
  direction="l",
 },
 [0x1BC88]={
  category="lo",
  direction="l",
 },
 [0x1BC90]={
  category="lo",
  direction="l",
 },
 [0x1BC91]={
  category="lo",
  direction="l",
 },
 [0x1BC92]={
  category="lo",
  direction="l",
 },
 [0x1BC93]={
  category="lo",
  direction="l",
 },
 [0x1BC94]={
  category="lo",
  direction="l",
 },
 [0x1BC95]={
  category="lo",
  direction="l",
 },
 [0x1BC96]={
  category="lo",
  direction="l",
 },
 [0x1BC97]={
  category="lo",
  direction="l",
 },
 [0x1BC98]={
  category="lo",
  direction="l",
 },
 [0x1BC99]={
  category="lo",
  direction="l",
 },
 [0x1BC9C]={
  category="so",
  direction="l",
 },
 [0x1BC9D]={
  category="mn",
  direction="nsm",
 },
 [0x1BC9E]={
  category="mn",
  direction="nsm",
 },
 [0x1BC9F]={
  category="po",
  direction="l",
 },
 [0x1BCA0]={
  category="cf",
  direction="bn",
 },
 [0x1BCA1]={
  category="cf",
  direction="bn",
 },
 [0x1BCA2]={
  category="cf",
  direction="bn",
 },
 [0x1BCA3]={
  category="cf",
  direction="bn",
 },
 [0x1D000]={
  category="so",
  direction="l",
 },
 [0x1D001]={
  category="so",
  direction="l",
 },
 [0x1D002]={
  category="so",
  direction="l",
 },
 [0x1D003]={
  category="so",
  direction="l",
 },
 [0x1D004]={
  category="so",
  direction="l",
 },
 [0x1D005]={
  category="so",
  direction="l",
 },
 [0x1D006]={
  category="so",
  direction="l",
 },
 [0x1D007]={
  category="so",
  direction="l",
 },
 [0x1D008]={
  category="so",
  direction="l",
 },
 [0x1D009]={
  category="so",
  direction="l",
 },
 [0x1D00A]={
  category="so",
  direction="l",
 },
 [0x1D00B]={
  category="so",
  direction="l",
 },
 [0x1D00C]={
  category="so",
  direction="l",
 },
 [0x1D00D]={
  category="so",
  direction="l",
 },
 [0x1D00E]={
  category="so",
  direction="l",
 },
 [0x1D00F]={
  category="so",
  direction="l",
 },
 [0x1D010]={
  category="so",
  direction="l",
 },
 [0x1D011]={
  category="so",
  direction="l",
 },
 [0x1D012]={
  category="so",
  direction="l",
 },
 [0x1D013]={
  category="so",
  direction="l",
 },
 [0x1D014]={
  category="so",
  direction="l",
 },
 [0x1D015]={
  category="so",
  direction="l",
 },
 [0x1D016]={
  category="so",
  direction="l",
 },
 [0x1D017]={
  category="so",
  direction="l",
 },
 [0x1D018]={
  category="so",
  direction="l",
 },
 [0x1D019]={
  category="so",
  direction="l",
 },
 [0x1D01A]={
  category="so",
  direction="l",
 },
 [0x1D01B]={
  category="so",
  direction="l",
 },
 [0x1D01C]={
  category="so",
  direction="l",
 },
 [0x1D01D]={
  category="so",
  direction="l",
 },
 [0x1D01E]={
  category="so",
  direction="l",
 },
 [0x1D01F]={
  category="so",
  direction="l",
 },
 [0x1D020]={
  category="so",
  direction="l",
 },
 [0x1D021]={
  category="so",
  direction="l",
 },
 [0x1D022]={
  category="so",
  direction="l",
 },
 [0x1D023]={
  category="so",
  direction="l",
 },
 [0x1D024]={
  category="so",
  direction="l",
 },
 [0x1D025]={
  category="so",
  direction="l",
 },
 [0x1D026]={
  category="so",
  direction="l",
 },
 [0x1D027]={
  category="so",
  direction="l",
 },
 [0x1D028]={
  category="so",
  direction="l",
 },
 [0x1D029]={
  category="so",
  direction="l",
 },
 [0x1D02A]={
  category="so",
  direction="l",
 },
 [0x1D02B]={
  category="so",
  direction="l",
 },
 [0x1D02C]={
  category="so",
  direction="l",
 },
 [0x1D02D]={
  category="so",
  direction="l",
 },
 [0x1D02E]={
  category="so",
  direction="l",
 },
 [0x1D02F]={
  category="so",
  direction="l",
 },
 [0x1D030]={
  category="so",
  direction="l",
 },
 [0x1D031]={
  category="so",
  direction="l",
 },
 [0x1D032]={
  category="so",
  direction="l",
 },
 [0x1D033]={
  category="so",
  direction="l",
 },
 [0x1D034]={
  category="so",
  direction="l",
 },
 [0x1D035]={
  category="so",
  direction="l",
 },
 [0x1D036]={
  category="so",
  direction="l",
 },
 [0x1D037]={
  category="so",
  direction="l",
 },
 [0x1D038]={
  category="so",
  direction="l",
 },
 [0x1D039]={
  category="so",
  direction="l",
 },
 [0x1D03A]={
  category="so",
  direction="l",
 },
 [0x1D03B]={
  category="so",
  direction="l",
 },
 [0x1D03C]={
  category="so",
  direction="l",
 },
 [0x1D03D]={
  category="so",
  direction="l",
 },
 [0x1D03E]={
  category="so",
  direction="l",
 },
 [0x1D03F]={
  category="so",
  direction="l",
 },
 [0x1D040]={
  category="so",
  direction="l",
 },
 [0x1D041]={
  category="so",
  direction="l",
 },
 [0x1D042]={
  category="so",
  direction="l",
 },
 [0x1D043]={
  category="so",
  direction="l",
 },
 [0x1D044]={
  category="so",
  direction="l",
 },
 [0x1D045]={
  category="so",
  direction="l",
 },
 [0x1D046]={
  category="so",
  direction="l",
 },
 [0x1D047]={
  category="so",
  direction="l",
 },
 [0x1D048]={
  category="so",
  direction="l",
 },
 [0x1D049]={
  category="so",
  direction="l",
 },
 [0x1D04A]={
  category="so",
  direction="l",
 },
 [0x1D04B]={
  category="so",
  direction="l",
 },
 [0x1D04C]={
  category="so",
  direction="l",
 },
 [0x1D04D]={
  category="so",
  direction="l",
 },
 [0x1D04E]={
  category="so",
  direction="l",
 },
 [0x1D04F]={
  category="so",
  direction="l",
 },
 [0x1D050]={
  category="so",
  direction="l",
 },
 [0x1D051]={
  category="so",
  direction="l",
 },
 [0x1D052]={
  category="so",
  direction="l",
 },
 [0x1D053]={
  category="so",
  direction="l",
 },
 [0x1D054]={
  category="so",
  direction="l",
 },
 [0x1D055]={
  category="so",
  direction="l",
 },
 [0x1D056]={
  category="so",
  direction="l",
 },
 [0x1D057]={
  category="so",
  direction="l",
 },
 [0x1D058]={
  category="so",
  direction="l",
 },
 [0x1D059]={
  category="so",
  direction="l",
 },
 [0x1D05A]={
  category="so",
  direction="l",
 },
 [0x1D05B]={
  category="so",
  direction="l",
 },
 [0x1D05C]={
  category="so",
  direction="l",
 },
 [0x1D05D]={
  category="so",
  direction="l",
 },
 [0x1D05E]={
  category="so",
  direction="l",
 },
 [0x1D05F]={
  category="so",
  direction="l",
 },
 [0x1D060]={
  category="so",
  direction="l",
 },
 [0x1D061]={
  category="so",
  direction="l",
 },
 [0x1D062]={
  category="so",
  direction="l",
 },
 [0x1D063]={
  category="so",
  direction="l",
 },
 [0x1D064]={
  category="so",
  direction="l",
 },
 [0x1D065]={
  category="so",
  direction="l",
 },
 [0x1D066]={
  category="so",
  direction="l",
 },
 [0x1D067]={
  category="so",
  direction="l",
 },
 [0x1D068]={
  category="so",
  direction="l",
 },
 [0x1D069]={
  category="so",
  direction="l",
 },
 [0x1D06A]={
  category="so",
  direction="l",
 },
 [0x1D06B]={
  category="so",
  direction="l",
 },
 [0x1D06C]={
  category="so",
  direction="l",
 },
 [0x1D06D]={
  category="so",
  direction="l",
 },
 [0x1D06E]={
  category="so",
  direction="l",
 },
 [0x1D06F]={
  category="so",
  direction="l",
 },
 [0x1D070]={
  category="so",
  direction="l",
 },
 [0x1D071]={
  category="so",
  direction="l",
 },
 [0x1D072]={
  category="so",
  direction="l",
 },
 [0x1D073]={
  category="so",
  direction="l",
 },
 [0x1D074]={
  category="so",
  direction="l",
 },
 [0x1D075]={
  category="so",
  direction="l",
 },
 [0x1D076]={
  category="so",
  direction="l",
 },
 [0x1D077]={
  category="so",
  direction="l",
 },
 [0x1D078]={
  category="so",
  direction="l",
 },
 [0x1D079]={
  category="so",
  direction="l",
 },
 [0x1D07A]={
  category="so",
  direction="l",
 },
 [0x1D07B]={
  category="so",
  direction="l",
 },
 [0x1D07C]={
  category="so",
  direction="l",
 },
 [0x1D07D]={
  category="so",
  direction="l",
 },
 [0x1D07E]={
  category="so",
  direction="l",
 },
 [0x1D07F]={
  category="so",
  direction="l",
 },
 [0x1D080]={
  category="so",
  direction="l",
 },
 [0x1D081]={
  category="so",
  direction="l",
 },
 [0x1D082]={
  category="so",
  direction="l",
 },
 [0x1D083]={
  category="so",
  direction="l",
 },
 [0x1D084]={
  category="so",
  direction="l",
 },
 [0x1D085]={
  category="so",
  direction="l",
 },
 [0x1D086]={
  category="so",
  direction="l",
 },
 [0x1D087]={
  category="so",
  direction="l",
 },
 [0x1D088]={
  category="so",
  direction="l",
 },
 [0x1D089]={
  category="so",
  direction="l",
 },
 [0x1D08A]={
  category="so",
  direction="l",
 },
 [0x1D08B]={
  category="so",
  direction="l",
 },
 [0x1D08C]={
  category="so",
  direction="l",
 },
 [0x1D08D]={
  category="so",
  direction="l",
 },
 [0x1D08E]={
  category="so",
  direction="l",
 },
 [0x1D08F]={
  category="so",
  direction="l",
 },
 [0x1D090]={
  category="so",
  direction="l",
 },
 [0x1D091]={
  category="so",
  direction="l",
 },
 [0x1D092]={
  category="so",
  direction="l",
 },
 [0x1D093]={
  category="so",
  direction="l",
 },
 [0x1D094]={
  category="so",
  direction="l",
 },
 [0x1D095]={
  category="so",
  direction="l",
 },
 [0x1D096]={
  category="so",
  direction="l",
 },
 [0x1D097]={
  category="so",
  direction="l",
 },
 [0x1D098]={
  category="so",
  direction="l",
 },
 [0x1D099]={
  category="so",
  direction="l",
 },
 [0x1D09A]={
  category="so",
  direction="l",
 },
 [0x1D09B]={
  category="so",
  direction="l",
 },
 [0x1D09C]={
  category="so",
  direction="l",
 },
 [0x1D09D]={
  category="so",
  direction="l",
 },
 [0x1D09E]={
  category="so",
  direction="l",
 },
 [0x1D09F]={
  category="so",
  direction="l",
 },
 [0x1D0A0]={
  category="so",
  direction="l",
 },
 [0x1D0A1]={
  category="so",
  direction="l",
 },
 [0x1D0A2]={
  category="so",
  direction="l",
 },
 [0x1D0A3]={
  category="so",
  direction="l",
 },
 [0x1D0A4]={
  category="so",
  direction="l",
 },
 [0x1D0A5]={
  category="so",
  direction="l",
 },
 [0x1D0A6]={
  category="so",
  direction="l",
 },
 [0x1D0A7]={
  category="so",
  direction="l",
 },
 [0x1D0A8]={
  category="so",
  direction="l",
 },
 [0x1D0A9]={
  category="so",
  direction="l",
 },
 [0x1D0AA]={
  category="so",
  direction="l",
 },
 [0x1D0AB]={
  category="so",
  direction="l",
 },
 [0x1D0AC]={
  category="so",
  direction="l",
 },
 [0x1D0AD]={
  category="so",
  direction="l",
 },
 [0x1D0AE]={
  category="so",
  direction="l",
 },
 [0x1D0AF]={
  category="so",
  direction="l",
 },
 [0x1D0B0]={
  category="so",
  direction="l",
 },
 [0x1D0B1]={
  category="so",
  direction="l",
 },
 [0x1D0B2]={
  category="so",
  direction="l",
 },
 [0x1D0B3]={
  category="so",
  direction="l",
 },
 [0x1D0B4]={
  category="so",
  direction="l",
 },
 [0x1D0B5]={
  category="so",
  direction="l",
 },
 [0x1D0B6]={
  category="so",
  direction="l",
 },
 [0x1D0B7]={
  category="so",
  direction="l",
 },
 [0x1D0B8]={
  category="so",
  direction="l",
 },
 [0x1D0B9]={
  category="so",
  direction="l",
 },
 [0x1D0BA]={
  category="so",
  direction="l",
 },
 [0x1D0BB]={
  category="so",
  direction="l",
 },
 [0x1D0BC]={
  category="so",
  direction="l",
 },
 [0x1D0BD]={
  category="so",
  direction="l",
 },
 [0x1D0BE]={
  category="so",
  direction="l",
 },
 [0x1D0BF]={
  category="so",
  direction="l",
 },
 [0x1D0C0]={
  category="so",
  direction="l",
 },
 [0x1D0C1]={
  category="so",
  direction="l",
 },
 [0x1D0C2]={
  category="so",
  direction="l",
 },
 [0x1D0C3]={
  category="so",
  direction="l",
 },
 [0x1D0C4]={
  category="so",
  direction="l",
 },
 [0x1D0C5]={
  category="so",
  direction="l",
 },
 [0x1D0C6]={
  category="so",
  direction="l",
 },
 [0x1D0C7]={
  category="so",
  direction="l",
 },
 [0x1D0C8]={
  category="so",
  direction="l",
 },
 [0x1D0C9]={
  category="so",
  direction="l",
 },
 [0x1D0CA]={
  category="so",
  direction="l",
 },
 [0x1D0CB]={
  category="so",
  direction="l",
 },
 [0x1D0CC]={
  category="so",
  direction="l",
 },
 [0x1D0CD]={
  category="so",
  direction="l",
 },
 [0x1D0CE]={
  category="so",
  direction="l",
 },
 [0x1D0CF]={
  category="so",
  direction="l",
 },
 [0x1D0D0]={
  category="so",
  direction="l",
 },
 [0x1D0D1]={
  category="so",
  direction="l",
 },
 [0x1D0D2]={
  category="so",
  direction="l",
 },
 [0x1D0D3]={
  category="so",
  direction="l",
 },
 [0x1D0D4]={
  category="so",
  direction="l",
 },
 [0x1D0D5]={
  category="so",
  direction="l",
 },
 [0x1D0D6]={
  category="so",
  direction="l",
 },
 [0x1D0D7]={
  category="so",
  direction="l",
 },
 [0x1D0D8]={
  category="so",
  direction="l",
 },
 [0x1D0D9]={
  category="so",
  direction="l",
 },
 [0x1D0DA]={
  category="so",
  direction="l",
 },
 [0x1D0DB]={
  category="so",
  direction="l",
 },
 [0x1D0DC]={
  category="so",
  direction="l",
 },
 [0x1D0DD]={
  category="so",
  direction="l",
 },
 [0x1D0DE]={
  category="so",
  direction="l",
 },
 [0x1D0DF]={
  category="so",
  direction="l",
 },
 [0x1D0E0]={
  category="so",
  direction="l",
 },
 [0x1D0E1]={
  category="so",
  direction="l",
 },
 [0x1D0E2]={
  category="so",
  direction="l",
 },
 [0x1D0E3]={
  category="so",
  direction="l",
 },
 [0x1D0E4]={
  category="so",
  direction="l",
 },
 [0x1D0E5]={
  category="so",
  direction="l",
 },
 [0x1D0E6]={
  category="so",
  direction="l",
 },
 [0x1D0E7]={
  category="so",
  direction="l",
 },
 [0x1D0E8]={
  category="so",
  direction="l",
 },
 [0x1D0E9]={
  category="so",
  direction="l",
 },
 [0x1D0EA]={
  category="so",
  direction="l",
 },
 [0x1D0EB]={
  category="so",
  direction="l",
 },
 [0x1D0EC]={
  category="so",
  direction="l",
 },
 [0x1D0ED]={
  category="so",
  direction="l",
 },
 [0x1D0EE]={
  category="so",
  direction="l",
 },
 [0x1D0EF]={
  category="so",
  direction="l",
 },
 [0x1D0F0]={
  category="so",
  direction="l",
 },
 [0x1D0F1]={
  category="so",
  direction="l",
 },
 [0x1D0F2]={
  category="so",
  direction="l",
 },
 [0x1D0F3]={
  category="so",
  direction="l",
 },
 [0x1D0F4]={
  category="so",
  direction="l",
 },
 [0x1D0F5]={
  category="so",
  direction="l",
 },
 [0x1D100]={
  category="so",
  direction="l",
 },
 [0x1D101]={
  category="so",
  direction="l",
 },
 [0x1D102]={
  category="so",
  direction="l",
 },
 [0x1D103]={
  category="so",
  direction="l",
 },
 [0x1D104]={
  category="so",
  direction="l",
 },
 [0x1D105]={
  category="so",
  direction="l",
 },
 [0x1D106]={
  category="so",
  direction="l",
 },
 [0x1D107]={
  category="so",
  direction="l",
 },
 [0x1D108]={
  category="so",
  direction="l",
 },
 [0x1D109]={
  category="so",
  direction="l",
 },
 [0x1D10A]={
  category="so",
  direction="l",
 },
 [0x1D10B]={
  category="so",
  direction="l",
 },
 [0x1D10C]={
  category="so",
  direction="l",
 },
 [0x1D10D]={
  category="so",
  direction="l",
 },
 [0x1D10E]={
  category="so",
  direction="l",
 },
 [0x1D10F]={
  category="so",
  direction="l",
 },
 [0x1D110]={
  category="so",
  direction="l",
 },
 [0x1D111]={
  category="so",
  direction="l",
 },
 [0x1D112]={
  category="so",
  direction="l",
 },
 [0x1D113]={
  category="so",
  direction="l",
 },
 [0x1D114]={
  category="so",
  direction="l",
 },
 [0x1D115]={
  category="so",
  direction="l",
 },
 [0x1D116]={
  category="so",
  direction="l",
 },
 [0x1D117]={
  category="so",
  direction="l",
 },
 [0x1D118]={
  category="so",
  direction="l",
 },
 [0x1D119]={
  category="so",
  direction="l",
 },
 [0x1D11A]={
  category="so",
  direction="l",
 },
 [0x1D11B]={
  category="so",
  direction="l",
 },
 [0x1D11C]={
  category="so",
  direction="l",
 },
 [0x1D11D]={
  category="so",
  direction="l",
 },
 [0x1D11E]={
  category="so",
  direction="l",
 },
 [0x1D11F]={
  category="so",
  direction="l",
 },
 [0x1D120]={
  category="so",
  direction="l",
 },
 [0x1D121]={
  category="so",
  direction="l",
 },
 [0x1D122]={
  category="so",
  direction="l",
 },
 [0x1D123]={
  category="so",
  direction="l",
 },
 [0x1D124]={
  category="so",
  direction="l",
 },
 [0x1D125]={
  category="so",
  direction="l",
 },
 [0x1D126]={
  category="so",
  direction="l",
 },
 [0x1D129]={
  category="so",
  direction="l",
 },
 [0x1D12A]={
  category="so",
  direction="l",
 },
 [0x1D12B]={
  category="so",
  direction="l",
 },
 [0x1D12C]={
  category="so",
  direction="l",
 },
 [0x1D12D]={
  category="so",
  direction="l",
 },
 [0x1D12E]={
  category="so",
  direction="l",
 },
 [0x1D12F]={
  category="so",
  direction="l",
 },
 [0x1D130]={
  category="so",
  direction="l",
 },
 [0x1D131]={
  category="so",
  direction="l",
 },
 [0x1D132]={
  category="so",
  direction="l",
 },
 [0x1D133]={
  category="so",
  direction="l",
 },
 [0x1D134]={
  category="so",
  direction="l",
 },
 [0x1D135]={
  category="so",
  direction="l",
 },
 [0x1D136]={
  category="so",
  direction="l",
 },
 [0x1D137]={
  category="so",
  direction="l",
 },
 [0x1D138]={
  category="so",
  direction="l",
 },
 [0x1D139]={
  category="so",
  direction="l",
 },
 [0x1D13A]={
  category="so",
  direction="l",
 },
 [0x1D13B]={
  category="so",
  direction="l",
 },
 [0x1D13C]={
  category="so",
  direction="l",
 },
 [0x1D13D]={
  category="so",
  direction="l",
 },
 [0x1D13E]={
  category="so",
  direction="l",
 },
 [0x1D13F]={
  category="so",
  direction="l",
 },
 [0x1D140]={
  category="so",
  direction="l",
 },
 [0x1D141]={
  category="so",
  direction="l",
 },
 [0x1D142]={
  category="so",
  direction="l",
 },
 [0x1D143]={
  category="so",
  direction="l",
 },
 [0x1D144]={
  category="so",
  direction="l",
 },
 [0x1D145]={
  category="so",
  direction="l",
 },
 [0x1D146]={
  category="so",
  direction="l",
 },
 [0x1D147]={
  category="so",
  direction="l",
 },
 [0x1D148]={
  category="so",
  direction="l",
 },
 [0x1D149]={
  category="so",
  direction="l",
 },
 [0x1D14A]={
  category="so",
  direction="l",
 },
 [0x1D14B]={
  category="so",
  direction="l",
 },
 [0x1D14C]={
  category="so",
  direction="l",
 },
 [0x1D14D]={
  category="so",
  direction="l",
 },
 [0x1D14E]={
  category="so",
  direction="l",
 },
 [0x1D14F]={
  category="so",
  direction="l",
 },
 [0x1D150]={
  category="so",
  direction="l",
 },
 [0x1D151]={
  category="so",
  direction="l",
 },
 [0x1D152]={
  category="so",
  direction="l",
 },
 [0x1D153]={
  category="so",
  direction="l",
 },
 [0x1D154]={
  category="so",
  direction="l",
 },
 [0x1D155]={
  category="so",
  direction="l",
 },
 [0x1D156]={
  category="so",
  direction="l",
 },
 [0x1D157]={
  category="so",
  direction="l",
 },
 [0x1D158]={
  category="so",
  direction="l",
 },
 [0x1D159]={
  category="so",
  direction="l",
 },
 [0x1D15A]={
  category="so",
  direction="l",
 },
 [0x1D15B]={
  category="so",
  direction="l",
 },
 [0x1D15C]={
  category="so",
  direction="l",
 },
 [0x1D15D]={
  category="so",
  direction="l",
 },
 [0x1D15E]={
  category="so",
  direction="l",
 },
 [0x1D15F]={
  category="so",
  direction="l",
 },
 [0x1D160]={
  category="so",
  direction="l",
 },
 [0x1D161]={
  category="so",
  direction="l",
 },
 [0x1D162]={
  category="so",
  direction="l",
 },
 [0x1D163]={
  category="so",
  direction="l",
 },
 [0x1D164]={
  category="so",
  direction="l",
 },
 [0x1D165]={
  category="mc",
  direction="l",
 },
 [0x1D166]={
  category="mc",
  direction="l",
 },
 [0x1D167]={
  category="mn",
  direction="nsm",
 },
 [0x1D168]={
  category="mn",
  direction="nsm",
 },
 [0x1D169]={
  category="mn",
  direction="nsm",
 },
 [0x1D16A]={
  category="so",
  direction="l",
 },
 [0x1D16B]={
  category="so",
  direction="l",
 },
 [0x1D16C]={
  category="so",
  direction="l",
 },
 [0x1D16D]={
  category="mc",
  direction="l",
 },
 [0x1D16E]={
  category="mc",
  direction="l",
 },
 [0x1D16F]={
  category="mc",
  direction="l",
 },
 [0x1D170]={
  category="mc",
  direction="l",
 },
 [0x1D171]={
  category="mc",
  direction="l",
 },
 [0x1D172]={
  category="mc",
  direction="l",
 },
 [0x1D173]={
  category="cf",
  direction="bn",
 },
 [0x1D174]={
  category="cf",
  direction="bn",
 },
 [0x1D175]={
  category="cf",
  direction="bn",
 },
 [0x1D176]={
  category="cf",
  direction="bn",
 },
 [0x1D177]={
  category="cf",
  direction="bn",
 },
 [0x1D178]={
  category="cf",
  direction="bn",
 },
 [0x1D179]={
  category="cf",
  direction="bn",
 },
 [0x1D17A]={
  category="cf",
  direction="bn",
 },
 [0x1D17B]={
  category="mn",
  direction="nsm",
 },
 [0x1D17C]={
  category="mn",
  direction="nsm",
 },
 [0x1D17D]={
  category="mn",
  direction="nsm",
 },
 [0x1D17E]={
  category="mn",
  direction="nsm",
 },
 [0x1D17F]={
  category="mn",
  direction="nsm",
 },
 [0x1D180]={
  category="mn",
  direction="nsm",
 },
 [0x1D181]={
  category="mn",
  direction="nsm",
 },
 [0x1D182]={
  category="mn",
  direction="nsm",
 },
 [0x1D183]={
  category="so",
  direction="l",
 },
 [0x1D184]={
  category="so",
  direction="l",
 },
 [0x1D185]={
  category="mn",
  direction="nsm",
 },
 [0x1D186]={
  category="mn",
  direction="nsm",
 },
 [0x1D187]={
  category="mn",
  direction="nsm",
 },
 [0x1D188]={
  category="mn",
  direction="nsm",
 },
 [0x1D189]={
  category="mn",
  direction="nsm",
 },
 [0x1D18A]={
  category="mn",
  direction="nsm",
 },
 [0x1D18B]={
  category="mn",
  direction="nsm",
 },
 [0x1D18C]={
  category="so",
  direction="l",
 },
 [0x1D18D]={
  category="so",
  direction="l",
 },
 [0x1D18E]={
  category="so",
  direction="l",
 },
 [0x1D18F]={
  category="so",
  direction="l",
 },
 [0x1D190]={
  category="so",
  direction="l",
 },
 [0x1D191]={
  category="so",
  direction="l",
 },
 [0x1D192]={
  category="so",
  direction="l",
 },
 [0x1D193]={
  category="so",
  direction="l",
 },
 [0x1D194]={
  category="so",
  direction="l",
 },
 [0x1D195]={
  category="so",
  direction="l",
 },
 [0x1D196]={
  category="so",
  direction="l",
 },
 [0x1D197]={
  category="so",
  direction="l",
 },
 [0x1D198]={
  category="so",
  direction="l",
 },
 [0x1D199]={
  category="so",
  direction="l",
 },
 [0x1D19A]={
  category="so",
  direction="l",
 },
 [0x1D19B]={
  category="so",
  direction="l",
 },
 [0x1D19C]={
  category="so",
  direction="l",
 },
 [0x1D19D]={
  category="so",
  direction="l",
 },
 [0x1D19E]={
  category="so",
  direction="l",
 },
 [0x1D19F]={
  category="so",
  direction="l",
 },
 [0x1D1A0]={
  category="so",
  direction="l",
 },
 [0x1D1A1]={
  category="so",
  direction="l",
 },
 [0x1D1A2]={
  category="so",
  direction="l",
 },
 [0x1D1A3]={
  category="so",
  direction="l",
 },
 [0x1D1A4]={
  category="so",
  direction="l",
 },
 [0x1D1A5]={
  category="so",
  direction="l",
 },
 [0x1D1A6]={
  category="so",
  direction="l",
 },
 [0x1D1A7]={
  category="so",
  direction="l",
 },
 [0x1D1A8]={
  category="so",
  direction="l",
 },
 [0x1D1A9]={
  category="so",
  direction="l",
 },
 [0x1D1AA]={
  category="mn",
  direction="nsm",
 },
 [0x1D1AB]={
  category="mn",
  direction="nsm",
 },
 [0x1D1AC]={
  category="mn",
  direction="nsm",
 },
 [0x1D1AD]={
  category="mn",
  direction="nsm",
 },
 [0x1D1AE]={
  category="so",
  direction="l",
 },
 [0x1D1AF]={
  category="so",
  direction="l",
 },
 [0x1D1B0]={
  category="so",
  direction="l",
 },
 [0x1D1B1]={
  category="so",
  direction="l",
 },
 [0x1D1B2]={
  category="so",
  direction="l",
 },
 [0x1D1B3]={
  category="so",
  direction="l",
 },
 [0x1D1B4]={
  category="so",
  direction="l",
 },
 [0x1D1B5]={
  category="so",
  direction="l",
 },
 [0x1D1B6]={
  category="so",
  direction="l",
 },
 [0x1D1B7]={
  category="so",
  direction="l",
 },
 [0x1D1B8]={
  category="so",
  direction="l",
 },
 [0x1D1B9]={
  category="so",
  direction="l",
 },
 [0x1D1BA]={
  category="so",
  direction="l",
 },
 [0x1D1BB]={
  category="so",
  direction="l",
 },
 [0x1D1BC]={
  category="so",
  direction="l",
 },
 [0x1D1BD]={
  category="so",
  direction="l",
 },
 [0x1D1BE]={
  category="so",
  direction="l",
 },
 [0x1D1BF]={
  category="so",
  direction="l",
 },
 [0x1D1C0]={
  category="so",
  direction="l",
 },
 [0x1D1C1]={
  category="so",
  direction="l",
 },
 [0x1D1C2]={
  category="so",
  direction="l",
 },
 [0x1D1C3]={
  category="so",
  direction="l",
 },
 [0x1D1C4]={
  category="so",
  direction="l",
 },
 [0x1D1C5]={
  category="so",
  direction="l",
 },
 [0x1D1C6]={
  category="so",
  direction="l",
 },
 [0x1D1C7]={
  category="so",
  direction="l",
 },
 [0x1D1C8]={
  category="so",
  direction="l",
 },
 [0x1D1C9]={
  category="so",
  direction="l",
 },
 [0x1D1CA]={
  category="so",
  direction="l",
 },
 [0x1D1CB]={
  category="so",
  direction="l",
 },
 [0x1D1CC]={
  category="so",
  direction="l",
 },
 [0x1D1CD]={
  category="so",
  direction="l",
 },
 [0x1D1CE]={
  category="so",
  direction="l",
 },
 [0x1D1CF]={
  category="so",
  direction="l",
 },
 [0x1D1D0]={
  category="so",
  direction="l",
 },
 [0x1D1D1]={
  category="so",
  direction="l",
 },
 [0x1D1D2]={
  category="so",
  direction="l",
 },
 [0x1D1D3]={
  category="so",
  direction="l",
 },
 [0x1D1D4]={
  category="so",
  direction="l",
 },
 [0x1D1D5]={
  category="so",
  direction="l",
 },
 [0x1D1D6]={
  category="so",
  direction="l",
 },
 [0x1D1D7]={
  category="so",
  direction="l",
 },
 [0x1D1D8]={
  category="so",
  direction="l",
 },
 [0x1D1D9]={
  category="so",
  direction="l",
 },
 [0x1D1DA]={
  category="so",
  direction="l",
 },
 [0x1D1DB]={
  category="so",
  direction="l",
 },
 [0x1D1DC]={
  category="so",
  direction="l",
 },
 [0x1D1DD]={
  category="so",
  direction="l",
 },
 [0x1D200]={
  category="so",
  direction="on",
 },
 [0x1D201]={
  category="so",
  direction="on",
 },
 [0x1D202]={
  category="so",
  direction="on",
 },
 [0x1D203]={
  category="so",
  direction="on",
 },
 [0x1D204]={
  category="so",
  direction="on",
 },
 [0x1D205]={
  category="so",
  direction="on",
 },
 [0x1D206]={
  category="so",
  direction="on",
 },
 [0x1D207]={
  category="so",
  direction="on",
 },
 [0x1D208]={
  category="so",
  direction="on",
 },
 [0x1D209]={
  category="so",
  direction="on",
 },
 [0x1D20A]={
  category="so",
  direction="on",
 },
 [0x1D20B]={
  category="so",
  direction="on",
 },
 [0x1D20C]={
  category="so",
  direction="on",
 },
 [0x1D20D]={
  category="so",
  direction="on",
 },
 [0x1D20E]={
  category="so",
  direction="on",
 },
 [0x1D20F]={
  category="so",
  direction="on",
 },
 [0x1D210]={
  category="so",
  direction="on",
 },
 [0x1D211]={
  category="so",
  direction="on",
 },
 [0x1D212]={
  category="so",
  direction="on",
 },
 [0x1D213]={
  category="so",
  direction="on",
 },
 [0x1D214]={
  category="so",
  direction="on",
 },
 [0x1D215]={
  category="so",
  direction="on",
 },
 [0x1D216]={
  category="so",
  direction="on",
 },
 [0x1D217]={
  category="so",
  direction="on",
 },
 [0x1D218]={
  category="so",
  direction="on",
 },
 [0x1D219]={
  category="so",
  direction="on",
 },
 [0x1D21A]={
  category="so",
  direction="on",
 },
 [0x1D21B]={
  category="so",
  direction="on",
 },
 [0x1D21C]={
  category="so",
  direction="on",
 },
 [0x1D21D]={
  category="so",
  direction="on",
 },
 [0x1D21E]={
  category="so",
  direction="on",
 },
 [0x1D21F]={
  category="so",
  direction="on",
 },
 [0x1D220]={
  category="so",
  direction="on",
 },
 [0x1D221]={
  category="so",
  direction="on",
 },
 [0x1D222]={
  category="so",
  direction="on",
 },
 [0x1D223]={
  category="so",
  direction="on",
 },
 [0x1D224]={
  category="so",
  direction="on",
 },
 [0x1D225]={
  category="so",
  direction="on",
 },
 [0x1D226]={
  category="so",
  direction="on",
 },
 [0x1D227]={
  category="so",
  direction="on",
 },
 [0x1D228]={
  category="so",
  direction="on",
 },
 [0x1D229]={
  category="so",
  direction="on",
 },
 [0x1D22A]={
  category="so",
  direction="on",
 },
 [0x1D22B]={
  category="so",
  direction="on",
 },
 [0x1D22C]={
  category="so",
  direction="on",
 },
 [0x1D22D]={
  category="so",
  direction="on",
 },
 [0x1D22E]={
  category="so",
  direction="on",
 },
 [0x1D22F]={
  category="so",
  direction="on",
 },
 [0x1D230]={
  category="so",
  direction="on",
 },
 [0x1D231]={
  category="so",
  direction="on",
 },
 [0x1D232]={
  category="so",
  direction="on",
 },
 [0x1D233]={
  category="so",
  direction="on",
 },
 [0x1D234]={
  category="so",
  direction="on",
 },
 [0x1D235]={
  category="so",
  direction="on",
 },
 [0x1D236]={
  category="so",
  direction="on",
 },
 [0x1D237]={
  category="so",
  direction="on",
 },
 [0x1D238]={
  category="so",
  direction="on",
 },
 [0x1D239]={
  category="so",
  direction="on",
 },
 [0x1D23A]={
  category="so",
  direction="on",
 },
 [0x1D23B]={
  category="so",
  direction="on",
 },
 [0x1D23C]={
  category="so",
  direction="on",
 },
 [0x1D23D]={
  category="so",
  direction="on",
 },
 [0x1D23E]={
  category="so",
  direction="on",
 },
 [0x1D23F]={
  category="so",
  direction="on",
 },
 [0x1D240]={
  category="so",
  direction="on",
 },
 [0x1D241]={
  category="so",
  direction="on",
 },
 [0x1D242]={
  category="mn",
  direction="nsm",
 },
 [0x1D243]={
  category="mn",
  direction="nsm",
 },
 [0x1D244]={
  category="mn",
  direction="nsm",
 },
 [0x1D245]={
  category="so",
  direction="on",
 },
 [0x1D300]={
  category="so",
  direction="on",
 },
 [0x1D301]={
  category="so",
  direction="on",
 },
 [0x1D302]={
  category="so",
  direction="on",
 },
 [0x1D303]={
  category="so",
  direction="on",
 },
 [0x1D304]={
  category="so",
  direction="on",
 },
 [0x1D305]={
  category="so",
  direction="on",
 },
 [0x1D306]={
  category="so",
  direction="on",
 },
 [0x1D307]={
  category="so",
  direction="on",
 },
 [0x1D308]={
  category="so",
  direction="on",
 },
 [0x1D309]={
  category="so",
  direction="on",
 },
 [0x1D30A]={
  category="so",
  direction="on",
 },
 [0x1D30B]={
  category="so",
  direction="on",
 },
 [0x1D30C]={
  category="so",
  direction="on",
 },
 [0x1D30D]={
  category="so",
  direction="on",
 },
 [0x1D30E]={
  category="so",
  direction="on",
 },
 [0x1D30F]={
  category="so",
  direction="on",
 },
 [0x1D310]={
  category="so",
  direction="on",
 },
 [0x1D311]={
  category="so",
  direction="on",
 },
 [0x1D312]={
  category="so",
  direction="on",
 },
 [0x1D313]={
  category="so",
  direction="on",
 },
 [0x1D314]={
  category="so",
  direction="on",
 },
 [0x1D315]={
  category="so",
  direction="on",
 },
 [0x1D316]={
  category="so",
  direction="on",
 },
 [0x1D317]={
  category="so",
  direction="on",
 },
 [0x1D318]={
  category="so",
  direction="on",
 },
 [0x1D319]={
  category="so",
  direction="on",
 },
 [0x1D31A]={
  category="so",
  direction="on",
 },
 [0x1D31B]={
  category="so",
  direction="on",
 },
 [0x1D31C]={
  category="so",
  direction="on",
 },
 [0x1D31D]={
  category="so",
  direction="on",
 },
 [0x1D31E]={
  category="so",
  direction="on",
 },
 [0x1D31F]={
  category="so",
  direction="on",
 },
 [0x1D320]={
  category="so",
  direction="on",
 },
 [0x1D321]={
  category="so",
  direction="on",
 },
 [0x1D322]={
  category="so",
  direction="on",
 },
 [0x1D323]={
  category="so",
  direction="on",
 },
 [0x1D324]={
  category="so",
  direction="on",
 },
 [0x1D325]={
  category="so",
  direction="on",
 },
 [0x1D326]={
  category="so",
  direction="on",
 },
 [0x1D327]={
  category="so",
  direction="on",
 },
 [0x1D328]={
  category="so",
  direction="on",
 },
 [0x1D329]={
  category="so",
  direction="on",
 },
 [0x1D32A]={
  category="so",
  direction="on",
 },
 [0x1D32B]={
  category="so",
  direction="on",
 },
 [0x1D32C]={
  category="so",
  direction="on",
 },
 [0x1D32D]={
  category="so",
  direction="on",
 },
 [0x1D32E]={
  category="so",
  direction="on",
 },
 [0x1D32F]={
  category="so",
  direction="on",
 },
 [0x1D330]={
  category="so",
  direction="on",
 },
 [0x1D331]={
  category="so",
  direction="on",
 },
 [0x1D332]={
  category="so",
  direction="on",
 },
 [0x1D333]={
  category="so",
  direction="on",
 },
 [0x1D334]={
  category="so",
  direction="on",
 },
 [0x1D335]={
  category="so",
  direction="on",
 },
 [0x1D336]={
  category="so",
  direction="on",
 },
 [0x1D337]={
  category="so",
  direction="on",
 },
 [0x1D338]={
  category="so",
  direction="on",
 },
 [0x1D339]={
  category="so",
  direction="on",
 },
 [0x1D33A]={
  category="so",
  direction="on",
 },
 [0x1D33B]={
  category="so",
  direction="on",
 },
 [0x1D33C]={
  category="so",
  direction="on",
 },
 [0x1D33D]={
  category="so",
  direction="on",
 },
 [0x1D33E]={
  category="so",
  direction="on",
 },
 [0x1D33F]={
  category="so",
  direction="on",
 },
 [0x1D340]={
  category="so",
  direction="on",
 },
 [0x1D341]={
  category="so",
  direction="on",
 },
 [0x1D342]={
  category="so",
  direction="on",
 },
 [0x1D343]={
  category="so",
  direction="on",
 },
 [0x1D344]={
  category="so",
  direction="on",
 },
 [0x1D345]={
  category="so",
  direction="on",
 },
 [0x1D346]={
  category="so",
  direction="on",
 },
 [0x1D347]={
  category="so",
  direction="on",
 },
 [0x1D348]={
  category="so",
  direction="on",
 },
 [0x1D349]={
  category="so",
  direction="on",
 },
 [0x1D34A]={
  category="so",
  direction="on",
 },
 [0x1D34B]={
  category="so",
  direction="on",
 },
 [0x1D34C]={
  category="so",
  direction="on",
 },
 [0x1D34D]={
  category="so",
  direction="on",
 },
 [0x1D34E]={
  category="so",
  direction="on",
 },
 [0x1D34F]={
  category="so",
  direction="on",
 },
 [0x1D350]={
  category="so",
  direction="on",
 },
 [0x1D351]={
  category="so",
  direction="on",
 },
 [0x1D352]={
  category="so",
  direction="on",
 },
 [0x1D353]={
  category="so",
  direction="on",
 },
 [0x1D354]={
  category="so",
  direction="on",
 },
 [0x1D355]={
  category="so",
  direction="on",
 },
 [0x1D356]={
  category="so",
  direction="on",
 },
 [0x1D360]={
  category="no",
  direction="l",
 },
 [0x1D361]={
  category="no",
  direction="l",
 },
 [0x1D362]={
  category="no",
  direction="l",
 },
 [0x1D363]={
  category="no",
  direction="l",
 },
 [0x1D364]={
  category="no",
  direction="l",
 },
 [0x1D365]={
  category="no",
  direction="l",
 },
 [0x1D366]={
  category="no",
  direction="l",
 },
 [0x1D367]={
  category="no",
  direction="l",
 },
 [0x1D368]={
  category="no",
  direction="l",
 },
 [0x1D369]={
  category="no",
  direction="l",
 },
 [0x1D36A]={
  category="no",
  direction="l",
 },
 [0x1D36B]={
  category="no",
  direction="l",
 },
 [0x1D36C]={
  category="no",
  direction="l",
 },
 [0x1D36D]={
  category="no",
  direction="l",
 },
 [0x1D36E]={
  category="no",
  direction="l",
 },
 [0x1D36F]={
  category="no",
  direction="l",
 },
 [0x1D370]={
  category="no",
  direction="l",
 },
 [0x1D371]={
  category="no",
  direction="l",
 },
 [0x1D400]={
  category="lu",
  direction="l",
 },
 [0x1D401]={
  category="lu",
  direction="l",
 },
 [0x1D402]={
  category="lu",
  direction="l",
 },
 [0x1D403]={
  category="lu",
  direction="l",
 },
 [0x1D404]={
  category="lu",
  direction="l",
 },
 [0x1D405]={
  category="lu",
  direction="l",
 },
 [0x1D406]={
  category="lu",
  direction="l",
 },
 [0x1D407]={
  category="lu",
  direction="l",
 },
 [0x1D408]={
  category="lu",
  direction="l",
 },
 [0x1D409]={
  category="lu",
  direction="l",
 },
 [0x1D40A]={
  category="lu",
  direction="l",
 },
 [0x1D40B]={
  category="lu",
  direction="l",
 },
 [0x1D40C]={
  category="lu",
  direction="l",
 },
 [0x1D40D]={
  category="lu",
  direction="l",
 },
 [0x1D40E]={
  category="lu",
  direction="l",
 },
 [0x1D40F]={
  category="lu",
  direction="l",
 },
 [0x1D410]={
  category="lu",
  direction="l",
 },
 [0x1D411]={
  category="lu",
  direction="l",
 },
 [0x1D412]={
  category="lu",
  direction="l",
 },
 [0x1D413]={
  category="lu",
  direction="l",
 },
 [0x1D414]={
  category="lu",
  direction="l",
 },
 [0x1D415]={
  category="lu",
  direction="l",
 },
 [0x1D416]={
  category="lu",
  direction="l",
 },
 [0x1D417]={
  category="lu",
  direction="l",
 },
 [0x1D418]={
  category="lu",
  direction="l",
 },
 [0x1D419]={
  category="lu",
  direction="l",
 },
 [0x1D41A]={
  category="ll",
  direction="l",
 },
 [0x1D41B]={
  category="ll",
  direction="l",
 },
 [0x1D41C]={
  category="ll",
  direction="l",
 },
 [0x1D41D]={
  category="ll",
  direction="l",
 },
 [0x1D41E]={
  category="ll",
  direction="l",
 },
 [0x1D41F]={
  category="ll",
  direction="l",
 },
 [0x1D420]={
  category="ll",
  direction="l",
 },
 [0x1D421]={
  category="ll",
  direction="l",
 },
 [0x1D422]={
  category="ll",
  direction="l",
 },
 [0x1D423]={
  category="ll",
  direction="l",
 },
 [0x1D424]={
  category="ll",
  direction="l",
 },
 [0x1D425]={
  category="ll",
  direction="l",
 },
 [0x1D426]={
  category="ll",
  direction="l",
 },
 [0x1D427]={
  category="ll",
  direction="l",
 },
 [0x1D428]={
  category="ll",
  direction="l",
 },
 [0x1D429]={
  category="ll",
  direction="l",
 },
 [0x1D42A]={
  category="ll",
  direction="l",
 },
 [0x1D42B]={
  category="ll",
  direction="l",
 },
 [0x1D42C]={
  category="ll",
  direction="l",
 },
 [0x1D42D]={
  category="ll",
  direction="l",
 },
 [0x1D42E]={
  category="ll",
  direction="l",
 },
 [0x1D42F]={
  category="ll",
  direction="l",
 },
 [0x1D430]={
  category="ll",
  direction="l",
 },
 [0x1D431]={
  category="ll",
  direction="l",
 },
 [0x1D432]={
  category="ll",
  direction="l",
 },
 [0x1D433]={
  category="ll",
  direction="l",
 },
 [0x1D434]={
  category="lu",
  direction="l",
 },
 [0x1D435]={
  category="lu",
  direction="l",
 },
 [0x1D436]={
  category="lu",
  direction="l",
 },
 [0x1D437]={
  category="lu",
  direction="l",
 },
 [0x1D438]={
  category="lu",
  direction="l",
 },
 [0x1D439]={
  category="lu",
  direction="l",
 },
 [0x1D43A]={
  category="lu",
  direction="l",
 },
 [0x1D43B]={
  category="lu",
  direction="l",
 },
 [0x1D43C]={
  category="lu",
  direction="l",
 },
 [0x1D43D]={
  category="lu",
  direction="l",
 },
 [0x1D43E]={
  category="lu",
  direction="l",
 },
 [0x1D43F]={
  category="lu",
  direction="l",
 },
 [0x1D440]={
  category="lu",
  direction="l",
 },
 [0x1D441]={
  category="lu",
  direction="l",
 },
 [0x1D442]={
  category="lu",
  direction="l",
 },
 [0x1D443]={
  category="lu",
  direction="l",
 },
 [0x1D444]={
  category="lu",
  direction="l",
 },
 [0x1D445]={
  category="lu",
  direction="l",
 },
 [0x1D446]={
  category="lu",
  direction="l",
 },
 [0x1D447]={
  category="lu",
  direction="l",
 },
 [0x1D448]={
  category="lu",
  direction="l",
 },
 [0x1D449]={
  category="lu",
  direction="l",
 },
 [0x1D44A]={
  category="lu",
  direction="l",
 },
 [0x1D44B]={
  category="lu",
  direction="l",
 },
 [0x1D44C]={
  category="lu",
  direction="l",
 },
 [0x1D44D]={
  category="lu",
  direction="l",
 },
 [0x1D44E]={
  category="ll",
  direction="l",
 },
 [0x1D44F]={
  category="ll",
  direction="l",
 },
 [0x1D450]={
  category="ll",
  direction="l",
 },
 [0x1D451]={
  category="ll",
  direction="l",
 },
 [0x1D452]={
  category="ll",
  direction="l",
 },
 [0x1D453]={
  category="ll",
  direction="l",
 },
 [0x1D454]={
  category="ll",
  direction="l",
 },
 [0x1D456]={
  category="ll",
  direction="l",
 },
 [0x1D457]={
  category="ll",
  direction="l",
 },
 [0x1D458]={
  category="ll",
  direction="l",
 },
 [0x1D459]={
  category="ll",
  direction="l",
 },
 [0x1D45A]={
  category="ll",
  direction="l",
 },
 [0x1D45B]={
  category="ll",
  direction="l",
 },
 [0x1D45C]={
  category="ll",
  direction="l",
 },
 [0x1D45D]={
  category="ll",
  direction="l",
 },
 [0x1D45E]={
  category="ll",
  direction="l",
 },
 [0x1D45F]={
  category="ll",
  direction="l",
 },
 [0x1D460]={
  category="ll",
  direction="l",
 },
 [0x1D461]={
  category="ll",
  direction="l",
 },
 [0x1D462]={
  category="ll",
  direction="l",
 },
 [0x1D463]={
  category="ll",
  direction="l",
 },
 [0x1D464]={
  category="ll",
  direction="l",
 },
 [0x1D465]={
  category="ll",
  direction="l",
 },
 [0x1D466]={
  category="ll",
  direction="l",
 },
 [0x1D467]={
  category="ll",
  direction="l",
 },
 [0x1D468]={
  category="lu",
  direction="l",
 },
 [0x1D469]={
  category="lu",
  direction="l",
 },
 [0x1D46A]={
  category="lu",
  direction="l",
 },
 [0x1D46B]={
  category="lu",
  direction="l",
 },
 [0x1D46C]={
  category="lu",
  direction="l",
 },
 [0x1D46D]={
  category="lu",
  direction="l",
 },
 [0x1D46E]={
  category="lu",
  direction="l",
 },
 [0x1D46F]={
  category="lu",
  direction="l",
 },
 [0x1D470]={
  category="lu",
  direction="l",
 },
 [0x1D471]={
  category="lu",
  direction="l",
 },
 [0x1D472]={
  category="lu",
  direction="l",
 },
 [0x1D473]={
  category="lu",
  direction="l",
 },
 [0x1D474]={
  category="lu",
  direction="l",
 },
 [0x1D475]={
  category="lu",
  direction="l",
 },
 [0x1D476]={
  category="lu",
  direction="l",
 },
 [0x1D477]={
  category="lu",
  direction="l",
 },
 [0x1D478]={
  category="lu",
  direction="l",
 },
 [0x1D479]={
  category="lu",
  direction="l",
 },
 [0x1D47A]={
  category="lu",
  direction="l",
 },
 [0x1D47B]={
  category="lu",
  direction="l",
 },
 [0x1D47C]={
  category="lu",
  direction="l",
 },
 [0x1D47D]={
  category="lu",
  direction="l",
 },
 [0x1D47E]={
  category="lu",
  direction="l",
 },
 [0x1D47F]={
  category="lu",
  direction="l",
 },
 [0x1D480]={
  category="lu",
  direction="l",
 },
 [0x1D481]={
  category="lu",
  direction="l",
 },
 [0x1D482]={
  category="ll",
  direction="l",
 },
 [0x1D483]={
  category="ll",
  direction="l",
 },
 [0x1D484]={
  category="ll",
  direction="l",
 },
 [0x1D485]={
  category="ll",
  direction="l",
 },
 [0x1D486]={
  category="ll",
  direction="l",
 },
 [0x1D487]={
  category="ll",
  direction="l",
 },
 [0x1D488]={
  category="ll",
  direction="l",
 },
 [0x1D489]={
  category="ll",
  direction="l",
 },
 [0x1D48A]={
  category="ll",
  direction="l",
 },
 [0x1D48B]={
  category="ll",
  direction="l",
 },
 [0x1D48C]={
  category="ll",
  direction="l",
 },
 [0x1D48D]={
  category="ll",
  direction="l",
 },
 [0x1D48E]={
  category="ll",
  direction="l",
 },
 [0x1D48F]={
  category="ll",
  direction="l",
 },
 [0x1D490]={
  category="ll",
  direction="l",
 },
 [0x1D491]={
  category="ll",
  direction="l",
 },
 [0x1D492]={
  category="ll",
  direction="l",
 },
 [0x1D493]={
  category="ll",
  direction="l",
 },
 [0x1D494]={
  category="ll",
  direction="l",
 },
 [0x1D495]={
  category="ll",
  direction="l",
 },
 [0x1D496]={
  category="ll",
  direction="l",
 },
 [0x1D497]={
  category="ll",
  direction="l",
 },
 [0x1D498]={
  category="ll",
  direction="l",
 },
 [0x1D499]={
  category="ll",
  direction="l",
 },
 [0x1D49A]={
  category="ll",
  direction="l",
 },
 [0x1D49B]={
  category="ll",
  direction="l",
 },
 [0x1D49C]={
  category="lu",
  direction="l",
 },
 [0x1D49E]={
  category="lu",
  direction="l",
 },
 [0x1D49F]={
  category="lu",
  direction="l",
 },
 [0x1D4A2]={
  category="lu",
  direction="l",
 },
 [0x1D4A5]={
  category="lu",
  direction="l",
 },
 [0x1D4A6]={
  category="lu",
  direction="l",
 },
 [0x1D4A9]={
  category="lu",
  direction="l",
 },
 [0x1D4AA]={
  category="lu",
  direction="l",
 },
 [0x1D4AB]={
  category="lu",
  direction="l",
 },
 [0x1D4AC]={
  category="lu",
  direction="l",
 },
 [0x1D4AE]={
  category="lu",
  direction="l",
 },
 [0x1D4AF]={
  category="lu",
  direction="l",
 },
 [0x1D4B0]={
  category="lu",
  direction="l",
 },
 [0x1D4B1]={
  category="lu",
  direction="l",
 },
 [0x1D4B2]={
  category="lu",
  direction="l",
 },
 [0x1D4B3]={
  category="lu",
  direction="l",
 },
 [0x1D4B4]={
  category="lu",
  direction="l",
 },
 [0x1D4B5]={
  category="lu",
  direction="l",
 },
 [0x1D4B6]={
  category="ll",
  direction="l",
 },
 [0x1D4B7]={
  category="ll",
  direction="l",
 },
 [0x1D4B8]={
  category="ll",
  direction="l",
 },
 [0x1D4B9]={
  category="ll",
  direction="l",
 },
 [0x1D4BB]={
  category="ll",
  direction="l",
 },
 [0x1D4BD]={
  category="ll",
  direction="l",
 },
 [0x1D4BE]={
  category="ll",
  direction="l",
 },
 [0x1D4BF]={
  category="ll",
  direction="l",
 },
 [0x1D4C0]={
  category="ll",
  direction="l",
 },
 [0x1D4C1]={
  category="ll",
  direction="l",
 },
 [0x1D4C2]={
  category="ll",
  direction="l",
 },
 [0x1D4C3]={
  category="ll",
  direction="l",
 },
 [0x1D4C5]={
  category="ll",
  direction="l",
 },
 [0x1D4C6]={
  category="ll",
  direction="l",
 },
 [0x1D4C7]={
  category="ll",
  direction="l",
 },
 [0x1D4C8]={
  category="ll",
  direction="l",
 },
 [0x1D4C9]={
  category="ll",
  direction="l",
 },
 [0x1D4CA]={
  category="ll",
  direction="l",
 },
 [0x1D4CB]={
  category="ll",
  direction="l",
 },
 [0x1D4CC]={
  category="ll",
  direction="l",
 },
 [0x1D4CD]={
  category="ll",
  direction="l",
 },
 [0x1D4CE]={
  category="ll",
  direction="l",
 },
 [0x1D4CF]={
  category="ll",
  direction="l",
 },
 [0x1D4D0]={
  category="lu",
  direction="l",
 },
 [0x1D4D1]={
  category="lu",
  direction="l",
 },
 [0x1D4D2]={
  category="lu",
  direction="l",
 },
 [0x1D4D3]={
  category="lu",
  direction="l",
 },
 [0x1D4D4]={
  category="lu",
  direction="l",
 },
 [0x1D4D5]={
  category="lu",
  direction="l",
 },
 [0x1D4D6]={
  category="lu",
  direction="l",
 },
 [0x1D4D7]={
  category="lu",
  direction="l",
 },
 [0x1D4D8]={
  category="lu",
  direction="l",
 },
 [0x1D4D9]={
  category="lu",
  direction="l",
 },
 [0x1D4DA]={
  category="lu",
  direction="l",
 },
 [0x1D4DB]={
  category="lu",
  direction="l",
 },
 [0x1D4DC]={
  category="lu",
  direction="l",
 },
 [0x1D4DD]={
  category="lu",
  direction="l",
 },
 [0x1D4DE]={
  category="lu",
  direction="l",
 },
 [0x1D4DF]={
  category="lu",
  direction="l",
 },
 [0x1D4E0]={
  category="lu",
  direction="l",
 },
 [0x1D4E1]={
  category="lu",
  direction="l",
 },
 [0x1D4E2]={
  category="lu",
  direction="l",
 },
 [0x1D4E3]={
  category="lu",
  direction="l",
 },
 [0x1D4E4]={
  category="lu",
  direction="l",
 },
 [0x1D4E5]={
  category="lu",
  direction="l",
 },
 [0x1D4E6]={
  category="lu",
  direction="l",
 },
 [0x1D4E7]={
  category="lu",
  direction="l",
 },
 [0x1D4E8]={
  category="lu",
  direction="l",
 },
 [0x1D4E9]={
  category="lu",
  direction="l",
 },
 [0x1D4EA]={
  category="ll",
  direction="l",
 },
 [0x1D4EB]={
  category="ll",
  direction="l",
 },
 [0x1D4EC]={
  category="ll",
  direction="l",
 },
 [0x1D4ED]={
  category="ll",
  direction="l",
 },
 [0x1D4EE]={
  category="ll",
  direction="l",
 },
 [0x1D4EF]={
  category="ll",
  direction="l",
 },
 [0x1D4F0]={
  category="ll",
  direction="l",
 },
 [0x1D4F1]={
  category="ll",
  direction="l",
 },
 [0x1D4F2]={
  category="ll",
  direction="l",
 },
 [0x1D4F3]={
  category="ll",
  direction="l",
 },
 [0x1D4F4]={
  category="ll",
  direction="l",
 },
 [0x1D4F5]={
  category="ll",
  direction="l",
 },
 [0x1D4F6]={
  category="ll",
  direction="l",
 },
 [0x1D4F7]={
  category="ll",
  direction="l",
 },
 [0x1D4F8]={
  category="ll",
  direction="l",
 },
 [0x1D4F9]={
  category="ll",
  direction="l",
 },
 [0x1D4FA]={
  category="ll",
  direction="l",
 },
 [0x1D4FB]={
  category="ll",
  direction="l",
 },
 [0x1D4FC]={
  category="ll",
  direction="l",
 },
 [0x1D4FD]={
  category="ll",
  direction="l",
 },
 [0x1D4FE]={
  category="ll",
  direction="l",
 },
 [0x1D4FF]={
  category="ll",
  direction="l",
 },
 [0x1D500]={
  category="ll",
  direction="l",
 },
 [0x1D501]={
  category="ll",
  direction="l",
 },
 [0x1D502]={
  category="ll",
  direction="l",
 },
 [0x1D503]={
  category="ll",
  direction="l",
 },
 [0x1D504]={
  category="lu",
  direction="l",
 },
 [0x1D505]={
  category="lu",
  direction="l",
 },
 [0x1D507]={
  category="lu",
  direction="l",
 },
 [0x1D508]={
  category="lu",
  direction="l",
 },
 [0x1D509]={
  category="lu",
  direction="l",
 },
 [0x1D50A]={
  category="lu",
  direction="l",
 },
 [0x1D50D]={
  category="lu",
  direction="l",
 },
 [0x1D50E]={
  category="lu",
  direction="l",
 },
 [0x1D50F]={
  category="lu",
  direction="l",
 },
 [0x1D510]={
  category="lu",
  direction="l",
 },
 [0x1D511]={
  category="lu",
  direction="l",
 },
 [0x1D512]={
  category="lu",
  direction="l",
 },
 [0x1D513]={
  category="lu",
  direction="l",
 },
 [0x1D514]={
  category="lu",
  direction="l",
 },
 [0x1D516]={
  category="lu",
  direction="l",
 },
 [0x1D517]={
  category="lu",
  direction="l",
 },
 [0x1D518]={
  category="lu",
  direction="l",
 },
 [0x1D519]={
  category="lu",
  direction="l",
 },
 [0x1D51A]={
  category="lu",
  direction="l",
 },
 [0x1D51B]={
  category="lu",
  direction="l",
 },
 [0x1D51C]={
  category="lu",
  direction="l",
 },
 [0x1D51E]={
  category="ll",
  direction="l",
 },
 [0x1D51F]={
  category="ll",
  direction="l",
 },
 [0x1D520]={
  category="ll",
  direction="l",
 },
 [0x1D521]={
  category="ll",
  direction="l",
 },
 [0x1D522]={
  category="ll",
  direction="l",
 },
 [0x1D523]={
  category="ll",
  direction="l",
 },
 [0x1D524]={
  category="ll",
  direction="l",
 },
 [0x1D525]={
  category="ll",
  direction="l",
 },
 [0x1D526]={
  category="ll",
  direction="l",
 },
 [0x1D527]={
  category="ll",
  direction="l",
 },
 [0x1D528]={
  category="ll",
  direction="l",
 },
 [0x1D529]={
  category="ll",
  direction="l",
 },
 [0x1D52A]={
  category="ll",
  direction="l",
 },
 [0x1D52B]={
  category="ll",
  direction="l",
 },
 [0x1D52C]={
  category="ll",
  direction="l",
 },
 [0x1D52D]={
  category="ll",
  direction="l",
 },
 [0x1D52E]={
  category="ll",
  direction="l",
 },
 [0x1D52F]={
  category="ll",
  direction="l",
 },
 [0x1D530]={
  category="ll",
  direction="l",
 },
 [0x1D531]={
  category="ll",
  direction="l",
 },
 [0x1D532]={
  category="ll",
  direction="l",
 },
 [0x1D533]={
  category="ll",
  direction="l",
 },
 [0x1D534]={
  category="ll",
  direction="l",
 },
 [0x1D535]={
  category="ll",
  direction="l",
 },
 [0x1D536]={
  category="ll",
  direction="l",
 },
 [0x1D537]={
  category="ll",
  direction="l",
 },
 [0x1D538]={
  category="lu",
  direction="l",
 },
 [0x1D539]={
  category="lu",
  direction="l",
 },
 [0x1D53B]={
  category="lu",
  direction="l",
 },
 [0x1D53C]={
  category="lu",
  direction="l",
 },
 [0x1D53D]={
  category="lu",
  direction="l",
 },
 [0x1D53E]={
  category="lu",
  direction="l",
 },
 [0x1D540]={
  category="lu",
  direction="l",
 },
 [0x1D541]={
  category="lu",
  direction="l",
 },
 [0x1D542]={
  category="lu",
  direction="l",
 },
 [0x1D543]={
  category="lu",
  direction="l",
 },
 [0x1D544]={
  category="lu",
  direction="l",
 },
 [0x1D546]={
  category="lu",
  direction="l",
 },
 [0x1D54A]={
  category="lu",
  direction="l",
 },
 [0x1D54B]={
  category="lu",
  direction="l",
 },
 [0x1D54C]={
  category="lu",
  direction="l",
 },
 [0x1D54D]={
  category="lu",
  direction="l",
 },
 [0x1D54E]={
  category="lu",
  direction="l",
 },
 [0x1D54F]={
  category="lu",
  direction="l",
 },
 [0x1D550]={
  category="lu",
  direction="l",
 },
 [0x1D552]={
  category="ll",
  direction="l",
 },
 [0x1D553]={
  category="ll",
  direction="l",
 },
 [0x1D554]={
  category="ll",
  direction="l",
 },
 [0x1D555]={
  category="ll",
  direction="l",
 },
 [0x1D556]={
  category="ll",
  direction="l",
 },
 [0x1D557]={
  category="ll",
  direction="l",
 },
 [0x1D558]={
  category="ll",
  direction="l",
 },
 [0x1D559]={
  category="ll",
  direction="l",
 },
 [0x1D55A]={
  category="ll",
  direction="l",
 },
 [0x1D55B]={
  category="ll",
  direction="l",
 },
 [0x1D55C]={
  category="ll",
  direction="l",
 },
 [0x1D55D]={
  category="ll",
  direction="l",
 },
 [0x1D55E]={
  category="ll",
  direction="l",
 },
 [0x1D55F]={
  category="ll",
  direction="l",
 },
 [0x1D560]={
  category="ll",
  direction="l",
 },
 [0x1D561]={
  category="ll",
  direction="l",
 },
 [0x1D562]={
  category="ll",
  direction="l",
 },
 [0x1D563]={
  category="ll",
  direction="l",
 },
 [0x1D564]={
  category="ll",
  direction="l",
 },
 [0x1D565]={
  category="ll",
  direction="l",
 },
 [0x1D566]={
  category="ll",
  direction="l",
 },
 [0x1D567]={
  category="ll",
  direction="l",
 },
 [0x1D568]={
  category="ll",
  direction="l",
 },
 [0x1D569]={
  category="ll",
  direction="l",
 },
 [0x1D56A]={
  category="ll",
  direction="l",
 },
 [0x1D56B]={
  category="ll",
  direction="l",
 },
 [0x1D56C]={
  category="lu",
  direction="l",
 },
 [0x1D56D]={
  category="lu",
  direction="l",
 },
 [0x1D56E]={
  category="lu",
  direction="l",
 },
 [0x1D56F]={
  category="lu",
  direction="l",
 },
 [0x1D570]={
  category="lu",
  direction="l",
 },
 [0x1D571]={
  category="lu",
  direction="l",
 },
 [0x1D572]={
  category="lu",
  direction="l",
 },
 [0x1D573]={
  category="lu",
  direction="l",
 },
 [0x1D574]={
  category="lu",
  direction="l",
 },
 [0x1D575]={
  category="lu",
  direction="l",
 },
 [0x1D576]={
  category="lu",
  direction="l",
 },
 [0x1D577]={
  category="lu",
  direction="l",
 },
 [0x1D578]={
  category="lu",
  direction="l",
 },
 [0x1D579]={
  category="lu",
  direction="l",
 },
 [0x1D57A]={
  category="lu",
  direction="l",
 },
 [0x1D57B]={
  category="lu",
  direction="l",
 },
 [0x1D57C]={
  category="lu",
  direction="l",
 },
 [0x1D57D]={
  category="lu",
  direction="l",
 },
 [0x1D57E]={
  category="lu",
  direction="l",
 },
 [0x1D57F]={
  category="lu",
  direction="l",
 },
 [0x1D580]={
  category="lu",
  direction="l",
 },
 [0x1D581]={
  category="lu",
  direction="l",
 },
 [0x1D582]={
  category="lu",
  direction="l",
 },
 [0x1D583]={
  category="lu",
  direction="l",
 },
 [0x1D584]={
  category="lu",
  direction="l",
 },
 [0x1D585]={
  category="lu",
  direction="l",
 },
 [0x1D586]={
  category="ll",
  direction="l",
 },
 [0x1D587]={
  category="ll",
  direction="l",
 },
 [0x1D588]={
  category="ll",
  direction="l",
 },
 [0x1D589]={
  category="ll",
  direction="l",
 },
 [0x1D58A]={
  category="ll",
  direction="l",
 },
 [0x1D58B]={
  category="ll",
  direction="l",
 },
 [0x1D58C]={
  category="ll",
  direction="l",
 },
 [0x1D58D]={
  category="ll",
  direction="l",
 },
 [0x1D58E]={
  category="ll",
  direction="l",
 },
 [0x1D58F]={
  category="ll",
  direction="l",
 },
 [0x1D590]={
  category="ll",
  direction="l",
 },
 [0x1D591]={
  category="ll",
  direction="l",
 },
 [0x1D592]={
  category="ll",
  direction="l",
 },
 [0x1D593]={
  category="ll",
  direction="l",
 },
 [0x1D594]={
  category="ll",
  direction="l",
 },
 [0x1D595]={
  category="ll",
  direction="l",
 },
 [0x1D596]={
  category="ll",
  direction="l",
 },
 [0x1D597]={
  category="ll",
  direction="l",
 },
 [0x1D598]={
  category="ll",
  direction="l",
 },
 [0x1D599]={
  category="ll",
  direction="l",
 },
 [0x1D59A]={
  category="ll",
  direction="l",
 },
 [0x1D59B]={
  category="ll",
  direction="l",
 },
 [0x1D59C]={
  category="ll",
  direction="l",
 },
 [0x1D59D]={
  category="ll",
  direction="l",
 },
 [0x1D59E]={
  category="ll",
  direction="l",
 },
 [0x1D59F]={
  category="ll",
  direction="l",
 },
 [0x1D5A0]={
  category="lu",
  direction="l",
 },
 [0x1D5A1]={
  category="lu",
  direction="l",
 },
 [0x1D5A2]={
  category="lu",
  direction="l",
 },
 [0x1D5A3]={
  category="lu",
  direction="l",
 },
 [0x1D5A4]={
  category="lu",
  direction="l",
 },
 [0x1D5A5]={
  category="lu",
  direction="l",
 },
 [0x1D5A6]={
  category="lu",
  direction="l",
 },
 [0x1D5A7]={
  category="lu",
  direction="l",
 },
 [0x1D5A8]={
  category="lu",
  direction="l",
 },
 [0x1D5A9]={
  category="lu",
  direction="l",
 },
 [0x1D5AA]={
  category="lu",
  direction="l",
 },
 [0x1D5AB]={
  category="lu",
  direction="l",
 },
 [0x1D5AC]={
  category="lu",
  direction="l",
 },
 [0x1D5AD]={
  category="lu",
  direction="l",
 },
 [0x1D5AE]={
  category="lu",
  direction="l",
 },
 [0x1D5AF]={
  category="lu",
  direction="l",
 },
 [0x1D5B0]={
  category="lu",
  direction="l",
 },
 [0x1D5B1]={
  category="lu",
  direction="l",
 },
 [0x1D5B2]={
  category="lu",
  direction="l",
 },
 [0x1D5B3]={
  category="lu",
  direction="l",
 },
 [0x1D5B4]={
  category="lu",
  direction="l",
 },
 [0x1D5B5]={
  category="lu",
  direction="l",
 },
 [0x1D5B6]={
  category="lu",
  direction="l",
 },
 [0x1D5B7]={
  category="lu",
  direction="l",
 },
 [0x1D5B8]={
  category="lu",
  direction="l",
 },
 [0x1D5B9]={
  category="lu",
  direction="l",
 },
 [0x1D5BA]={
  category="ll",
  direction="l",
 },
 [0x1D5BB]={
  category="ll",
  direction="l",
 },
 [0x1D5BC]={
  category="ll",
  direction="l",
 },
 [0x1D5BD]={
  category="ll",
  direction="l",
 },
 [0x1D5BE]={
  category="ll",
  direction="l",
 },
 [0x1D5BF]={
  category="ll",
  direction="l",
 },
 [0x1D5C0]={
  category="ll",
  direction="l",
 },
 [0x1D5C1]={
  category="ll",
  direction="l",
 },
 [0x1D5C2]={
  category="ll",
  direction="l",
 },
 [0x1D5C3]={
  category="ll",
  direction="l",
 },
 [0x1D5C4]={
  category="ll",
  direction="l",
 },
 [0x1D5C5]={
  category="ll",
  direction="l",
 },
 [0x1D5C6]={
  category="ll",
  direction="l",
 },
 [0x1D5C7]={
  category="ll",
  direction="l",
 },
 [0x1D5C8]={
  category="ll",
  direction="l",
 },
 [0x1D5C9]={
  category="ll",
  direction="l",
 },
 [0x1D5CA]={
  category="ll",
  direction="l",
 },
 [0x1D5CB]={
  category="ll",
  direction="l",
 },
 [0x1D5CC]={
  category="ll",
  direction="l",
 },
 [0x1D5CD]={
  category="ll",
  direction="l",
 },
 [0x1D5CE]={
  category="ll",
  direction="l",
 },
 [0x1D5CF]={
  category="ll",
  direction="l",
 },
 [0x1D5D0]={
  category="ll",
  direction="l",
 },
 [0x1D5D1]={
  category="ll",
  direction="l",
 },
 [0x1D5D2]={
  category="ll",
  direction="l",
 },
 [0x1D5D3]={
  category="ll",
  direction="l",
 },
 [0x1D5D4]={
  category="lu",
  direction="l",
 },
 [0x1D5D5]={
  category="lu",
  direction="l",
 },
 [0x1D5D6]={
  category="lu",
  direction="l",
 },
 [0x1D5D7]={
  category="lu",
  direction="l",
 },
 [0x1D5D8]={
  category="lu",
  direction="l",
 },
 [0x1D5D9]={
  category="lu",
  direction="l",
 },
 [0x1D5DA]={
  category="lu",
  direction="l",
 },
 [0x1D5DB]={
  category="lu",
  direction="l",
 },
 [0x1D5DC]={
  category="lu",
  direction="l",
 },
 [0x1D5DD]={
  category="lu",
  direction="l",
 },
 [0x1D5DE]={
  category="lu",
  direction="l",
 },
 [0x1D5DF]={
  category="lu",
  direction="l",
 },
 [0x1D5E0]={
  category="lu",
  direction="l",
 },
 [0x1D5E1]={
  category="lu",
  direction="l",
 },
 [0x1D5E2]={
  category="lu",
  direction="l",
 },
 [0x1D5E3]={
  category="lu",
  direction="l",
 },
 [0x1D5E4]={
  category="lu",
  direction="l",
 },
 [0x1D5E5]={
  category="lu",
  direction="l",
 },
 [0x1D5E6]={
  category="lu",
  direction="l",
 },
 [0x1D5E7]={
  category="lu",
  direction="l",
 },
 [0x1D5E8]={
  category="lu",
  direction="l",
 },
 [0x1D5E9]={
  category="lu",
  direction="l",
 },
 [0x1D5EA]={
  category="lu",
  direction="l",
 },
 [0x1D5EB]={
  category="lu",
  direction="l",
 },
 [0x1D5EC]={
  category="lu",
  direction="l",
 },
 [0x1D5ED]={
  category="lu",
  direction="l",
 },
 [0x1D5EE]={
  category="ll",
  direction="l",
 },
 [0x1D5EF]={
  category="ll",
  direction="l",
 },
 [0x1D5F0]={
  category="ll",
  direction="l",
 },
 [0x1D5F1]={
  category="ll",
  direction="l",
 },
 [0x1D5F2]={
  category="ll",
  direction="l",
 },
 [0x1D5F3]={
  category="ll",
  direction="l",
 },
 [0x1D5F4]={
  category="ll",
  direction="l",
 },
 [0x1D5F5]={
  category="ll",
  direction="l",
 },
 [0x1D5F6]={
  category="ll",
  direction="l",
 },
 [0x1D5F7]={
  category="ll",
  direction="l",
 },
 [0x1D5F8]={
  category="ll",
  direction="l",
 },
 [0x1D5F9]={
  category="ll",
  direction="l",
 },
 [0x1D5FA]={
  category="ll",
  direction="l",
 },
 [0x1D5FB]={
  category="ll",
  direction="l",
 },
 [0x1D5FC]={
  category="ll",
  direction="l",
 },
 [0x1D5FD]={
  category="ll",
  direction="l",
 },
 [0x1D5FE]={
  category="ll",
  direction="l",
 },
 [0x1D5FF]={
  category="ll",
  direction="l",
 },
 [0x1D600]={
  category="ll",
  direction="l",
 },
 [0x1D601]={
  category="ll",
  direction="l",
 },
 [0x1D602]={
  category="ll",
  direction="l",
 },
 [0x1D603]={
  category="ll",
  direction="l",
 },
 [0x1D604]={
  category="ll",
  direction="l",
 },
 [0x1D605]={
  category="ll",
  direction="l",
 },
 [0x1D606]={
  category="ll",
  direction="l",
 },
 [0x1D607]={
  category="ll",
  direction="l",
 },
 [0x1D608]={
  category="lu",
  direction="l",
 },
 [0x1D609]={
  category="lu",
  direction="l",
 },
 [0x1D60A]={
  category="lu",
  direction="l",
 },
 [0x1D60B]={
  category="lu",
  direction="l",
 },
 [0x1D60C]={
  category="lu",
  direction="l",
 },
 [0x1D60D]={
  category="lu",
  direction="l",
 },
 [0x1D60E]={
  category="lu",
  direction="l",
 },
 [0x1D60F]={
  category="lu",
  direction="l",
 },
 [0x1D610]={
  category="lu",
  direction="l",
 },
 [0x1D611]={
  category="lu",
  direction="l",
 },
 [0x1D612]={
  category="lu",
  direction="l",
 },
 [0x1D613]={
  category="lu",
  direction="l",
 },
 [0x1D614]={
  category="lu",
  direction="l",
 },
 [0x1D615]={
  category="lu",
  direction="l",
 },
 [0x1D616]={
  category="lu",
  direction="l",
 },
 [0x1D617]={
  category="lu",
  direction="l",
 },
 [0x1D618]={
  category="lu",
  direction="l",
 },
 [0x1D619]={
  category="lu",
  direction="l",
 },
 [0x1D61A]={
  category="lu",
  direction="l",
 },
 [0x1D61B]={
  category="lu",
  direction="l",
 },
 [0x1D61C]={
  category="lu",
  direction="l",
 },
 [0x1D61D]={
  category="lu",
  direction="l",
 },
 [0x1D61E]={
  category="lu",
  direction="l",
 },
 [0x1D61F]={
  category="lu",
  direction="l",
 },
 [0x1D620]={
  category="lu",
  direction="l",
 },
 [0x1D621]={
  category="lu",
  direction="l",
 },
 [0x1D622]={
  category="ll",
  direction="l",
 },
 [0x1D623]={
  category="ll",
  direction="l",
 },
 [0x1D624]={
  category="ll",
  direction="l",
 },
 [0x1D625]={
  category="ll",
  direction="l",
 },
 [0x1D626]={
  category="ll",
  direction="l",
 },
 [0x1D627]={
  category="ll",
  direction="l",
 },
 [0x1D628]={
  category="ll",
  direction="l",
 },
 [0x1D629]={
  category="ll",
  direction="l",
 },
 [0x1D62A]={
  category="ll",
  direction="l",
 },
 [0x1D62B]={
  category="ll",
  direction="l",
 },
 [0x1D62C]={
  category="ll",
  direction="l",
 },
 [0x1D62D]={
  category="ll",
  direction="l",
 },
 [0x1D62E]={
  category="ll",
  direction="l",
 },
 [0x1D62F]={
  category="ll",
  direction="l",
 },
 [0x1D630]={
  category="ll",
  direction="l",
 },
 [0x1D631]={
  category="ll",
  direction="l",
 },
 [0x1D632]={
  category="ll",
  direction="l",
 },
 [0x1D633]={
  category="ll",
  direction="l",
 },
 [0x1D634]={
  category="ll",
  direction="l",
 },
 [0x1D635]={
  category="ll",
  direction="l",
 },
 [0x1D636]={
  category="ll",
  direction="l",
 },
 [0x1D637]={
  category="ll",
  direction="l",
 },
 [0x1D638]={
  category="ll",
  direction="l",
 },
 [0x1D639]={
  category="ll",
  direction="l",
 },
 [0x1D63A]={
  category="ll",
  direction="l",
 },
 [0x1D63B]={
  category="ll",
  direction="l",
 },
 [0x1D63C]={
  category="lu",
  direction="l",
 },
 [0x1D63D]={
  category="lu",
  direction="l",
 },
 [0x1D63E]={
  category="lu",
  direction="l",
 },
 [0x1D63F]={
  category="lu",
  direction="l",
 },
 [0x1D640]={
  category="lu",
  direction="l",
 },
 [0x1D641]={
  category="lu",
  direction="l",
 },
 [0x1D642]={
  category="lu",
  direction="l",
 },
 [0x1D643]={
  category="lu",
  direction="l",
 },
 [0x1D644]={
  category="lu",
  direction="l",
 },
 [0x1D645]={
  category="lu",
  direction="l",
 },
 [0x1D646]={
  category="lu",
  direction="l",
 },
 [0x1D647]={
  category="lu",
  direction="l",
 },
 [0x1D648]={
  category="lu",
  direction="l",
 },
 [0x1D649]={
  category="lu",
  direction="l",
 },
 [0x1D64A]={
  category="lu",
  direction="l",
 },
 [0x1D64B]={
  category="lu",
  direction="l",
 },
 [0x1D64C]={
  category="lu",
  direction="l",
 },
 [0x1D64D]={
  category="lu",
  direction="l",
 },
 [0x1D64E]={
  category="lu",
  direction="l",
 },
 [0x1D64F]={
  category="lu",
  direction="l",
 },
 [0x1D650]={
  category="lu",
  direction="l",
 },
 [0x1D651]={
  category="lu",
  direction="l",
 },
 [0x1D652]={
  category="lu",
  direction="l",
 },
 [0x1D653]={
  category="lu",
  direction="l",
 },
 [0x1D654]={
  category="lu",
  direction="l",
 },
 [0x1D655]={
  category="lu",
  direction="l",
 },
 [0x1D656]={
  category="ll",
  direction="l",
 },
 [0x1D657]={
  category="ll",
  direction="l",
 },
 [0x1D658]={
  category="ll",
  direction="l",
 },
 [0x1D659]={
  category="ll",
  direction="l",
 },
 [0x1D65A]={
  category="ll",
  direction="l",
 },
 [0x1D65B]={
  category="ll",
  direction="l",
 },
 [0x1D65C]={
  category="ll",
  direction="l",
 },
 [0x1D65D]={
  category="ll",
  direction="l",
 },
 [0x1D65E]={
  category="ll",
  direction="l",
 },
 [0x1D65F]={
  category="ll",
  direction="l",
 },
 [0x1D660]={
  category="ll",
  direction="l",
 },
 [0x1D661]={
  category="ll",
  direction="l",
 },
 [0x1D662]={
  category="ll",
  direction="l",
 },
 [0x1D663]={
  category="ll",
  direction="l",
 },
 [0x1D664]={
  category="ll",
  direction="l",
 },
 [0x1D665]={
  category="ll",
  direction="l",
 },
 [0x1D666]={
  category="ll",
  direction="l",
 },
 [0x1D667]={
  category="ll",
  direction="l",
 },
 [0x1D668]={
  category="ll",
  direction="l",
 },
 [0x1D669]={
  category="ll",
  direction="l",
 },
 [0x1D66A]={
  category="ll",
  direction="l",
 },
 [0x1D66B]={
  category="ll",
  direction="l",
 },
 [0x1D66C]={
  category="ll",
  direction="l",
 },
 [0x1D66D]={
  category="ll",
  direction="l",
 },
 [0x1D66E]={
  category="ll",
  direction="l",
 },
 [0x1D66F]={
  category="ll",
  direction="l",
 },
 [0x1D670]={
  category="lu",
  direction="l",
 },
 [0x1D671]={
  category="lu",
  direction="l",
 },
 [0x1D672]={
  category="lu",
  direction="l",
 },
 [0x1D673]={
  category="lu",
  direction="l",
 },
 [0x1D674]={
  category="lu",
  direction="l",
 },
 [0x1D675]={
  category="lu",
  direction="l",
 },
 [0x1D676]={
  category="lu",
  direction="l",
 },
 [0x1D677]={
  category="lu",
  direction="l",
 },
 [0x1D678]={
  category="lu",
  direction="l",
 },
 [0x1D679]={
  category="lu",
  direction="l",
 },
 [0x1D67A]={
  category="lu",
  direction="l",
 },
 [0x1D67B]={
  category="lu",
  direction="l",
 },
 [0x1D67C]={
  category="lu",
  direction="l",
 },
 [0x1D67D]={
  category="lu",
  direction="l",
 },
 [0x1D67E]={
  category="lu",
  direction="l",
 },
 [0x1D67F]={
  category="lu",
  direction="l",
 },
 [0x1D680]={
  category="lu",
  direction="l",
 },
 [0x1D681]={
  category="lu",
  direction="l",
 },
 [0x1D682]={
  category="lu",
  direction="l",
 },
 [0x1D683]={
  category="lu",
  direction="l",
 },
 [0x1D684]={
  category="lu",
  direction="l",
 },
 [0x1D685]={
  category="lu",
  direction="l",
 },
 [0x1D686]={
  category="lu",
  direction="l",
 },
 [0x1D687]={
  category="lu",
  direction="l",
 },
 [0x1D688]={
  category="lu",
  direction="l",
 },
 [0x1D689]={
  category="lu",
  direction="l",
 },
 [0x1D68A]={
  category="ll",
  direction="l",
 },
 [0x1D68B]={
  category="ll",
  direction="l",
 },
 [0x1D68C]={
  category="ll",
  direction="l",
 },
 [0x1D68D]={
  category="ll",
  direction="l",
 },
 [0x1D68E]={
  category="ll",
  direction="l",
 },
 [0x1D68F]={
  category="ll",
  direction="l",
 },
 [0x1D690]={
  category="ll",
  direction="l",
 },
 [0x1D691]={
  category="ll",
  direction="l",
 },
 [0x1D692]={
  category="ll",
  direction="l",
 },
 [0x1D693]={
  category="ll",
  direction="l",
 },
 [0x1D694]={
  category="ll",
  direction="l",
 },
 [0x1D695]={
  category="ll",
  direction="l",
 },
 [0x1D696]={
  category="ll",
  direction="l",
 },
 [0x1D697]={
  category="ll",
  direction="l",
 },
 [0x1D698]={
  category="ll",
  direction="l",
 },
 [0x1D699]={
  category="ll",
  direction="l",
 },
 [0x1D69A]={
  category="ll",
  direction="l",
 },
 [0x1D69B]={
  category="ll",
  direction="l",
 },
 [0x1D69C]={
  category="ll",
  direction="l",
 },
 [0x1D69D]={
  category="ll",
  direction="l",
 },
 [0x1D69E]={
  category="ll",
  direction="l",
 },
 [0x1D69F]={
  category="ll",
  direction="l",
 },
 [0x1D6A0]={
  category="ll",
  direction="l",
 },
 [0x1D6A1]={
  category="ll",
  direction="l",
 },
 [0x1D6A2]={
  category="ll",
  direction="l",
 },
 [0x1D6A3]={
  category="ll",
  direction="l",
 },
 [0x1D6A4]={
  category="ll",
  direction="l",
 },
 [0x1D6A5]={
  category="ll",
  direction="l",
 },
 [0x1D6A8]={
  category="lu",
  direction="l",
 },
 [0x1D6A9]={
  category="lu",
  direction="l",
 },
 [0x1D6AA]={
  category="lu",
  direction="l",
 },
 [0x1D6AB]={
  category="lu",
  direction="l",
 },
 [0x1D6AC]={
  category="lu",
  direction="l",
 },
 [0x1D6AD]={
  category="lu",
  direction="l",
 },
 [0x1D6AE]={
  category="lu",
  direction="l",
 },
 [0x1D6AF]={
  category="lu",
  direction="l",
 },
 [0x1D6B0]={
  category="lu",
  direction="l",
 },
 [0x1D6B1]={
  category="lu",
  direction="l",
 },
 [0x1D6B2]={
  category="lu",
  direction="l",
 },
 [0x1D6B3]={
  category="lu",
  direction="l",
 },
 [0x1D6B4]={
  category="lu",
  direction="l",
 },
 [0x1D6B5]={
  category="lu",
  direction="l",
 },
 [0x1D6B6]={
  category="lu",
  direction="l",
 },
 [0x1D6B7]={
  category="lu",
  direction="l",
 },
 [0x1D6B8]={
  category="lu",
  direction="l",
 },
 [0x1D6B9]={
  category="lu",
  direction="l",
 },
 [0x1D6BA]={
  category="lu",
  direction="l",
 },
 [0x1D6BB]={
  category="lu",
  direction="l",
 },
 [0x1D6BC]={
  category="lu",
  direction="l",
 },
 [0x1D6BD]={
  category="lu",
  direction="l",
 },
 [0x1D6BE]={
  category="lu",
  direction="l",
 },
 [0x1D6BF]={
  category="lu",
  direction="l",
 },
 [0x1D6C0]={
  category="lu",
  direction="l",
 },
 [0x1D6C1]={
  category="sm",
  direction="l",
 },
 [0x1D6C2]={
  category="ll",
  direction="l",
 },
 [0x1D6C3]={
  category="ll",
  direction="l",
 },
 [0x1D6C4]={
  category="ll",
  direction="l",
 },
 [0x1D6C5]={
  category="ll",
  direction="l",
 },
 [0x1D6C6]={
  category="ll",
  direction="l",
 },
 [0x1D6C7]={
  category="ll",
  direction="l",
 },
 [0x1D6C8]={
  category="ll",
  direction="l",
 },
 [0x1D6C9]={
  category="ll",
  direction="l",
 },
 [0x1D6CA]={
  category="ll",
  direction="l",
 },
 [0x1D6CB]={
  category="ll",
  direction="l",
 },
 [0x1D6CC]={
  category="ll",
  direction="l",
 },
 [0x1D6CD]={
  category="ll",
  direction="l",
 },
 [0x1D6CE]={
  category="ll",
  direction="l",
 },
 [0x1D6CF]={
  category="ll",
  direction="l",
 },
 [0x1D6D0]={
  category="ll",
  direction="l",
 },
 [0x1D6D1]={
  category="ll",
  direction="l",
 },
 [0x1D6D2]={
  category="ll",
  direction="l",
 },
 [0x1D6D3]={
  category="ll",
  direction="l",
 },
 [0x1D6D4]={
  category="ll",
  direction="l",
 },
 [0x1D6D5]={
  category="ll",
  direction="l",
 },
 [0x1D6D6]={
  category="ll",
  direction="l",
 },
 [0x1D6D7]={
  category="ll",
  direction="l",
 },
 [0x1D6D8]={
  category="ll",
  direction="l",
 },
 [0x1D6D9]={
  category="ll",
  direction="l",
 },
 [0x1D6DA]={
  category="ll",
  direction="l",
 },
 [0x1D6DB]={
  category="sm",
  direction="on",
 },
 [0x1D6DC]={
  category="ll",
  direction="l",
 },
 [0x1D6DD]={
  category="ll",
  direction="l",
 },
 [0x1D6DE]={
  category="ll",
  direction="l",
 },
 [0x1D6DF]={
  category="ll",
  direction="l",
 },
 [0x1D6E0]={
  category="ll",
  direction="l",
 },
 [0x1D6E1]={
  category="ll",
  direction="l",
 },
 [0x1D6E2]={
  category="lu",
  direction="l",
 },
 [0x1D6E3]={
  category="lu",
  direction="l",
 },
 [0x1D6E4]={
  category="lu",
  direction="l",
 },
 [0x1D6E5]={
  category="lu",
  direction="l",
 },
 [0x1D6E6]={
  category="lu",
  direction="l",
 },
 [0x1D6E7]={
  category="lu",
  direction="l",
 },
 [0x1D6E8]={
  category="lu",
  direction="l",
 },
 [0x1D6E9]={
  category="lu",
  direction="l",
 },
 [0x1D6EA]={
  category="lu",
  direction="l",
 },
 [0x1D6EB]={
  category="lu",
  direction="l",
 },
 [0x1D6EC]={
  category="lu",
  direction="l",
 },
 [0x1D6ED]={
  category="lu",
  direction="l",
 },
 [0x1D6EE]={
  category="lu",
  direction="l",
 },
 [0x1D6EF]={
  category="lu",
  direction="l",
 },
 [0x1D6F0]={
  category="lu",
  direction="l",
 },
 [0x1D6F1]={
  category="lu",
  direction="l",
 },
 [0x1D6F2]={
  category="lu",
  direction="l",
 },
 [0x1D6F3]={
  category="lu",
  direction="l",
 },
 [0x1D6F4]={
  category="lu",
  direction="l",
 },
 [0x1D6F5]={
  category="lu",
  direction="l",
 },
 [0x1D6F6]={
  category="lu",
  direction="l",
 },
 [0x1D6F7]={
  category="lu",
  direction="l",
 },
 [0x1D6F8]={
  category="lu",
  direction="l",
 },
 [0x1D6F9]={
  category="lu",
  direction="l",
 },
 [0x1D6FA]={
  category="lu",
  direction="l",
 },
 [0x1D6FB]={
  category="sm",
  direction="l",
 },
 [0x1D6FC]={
  category="ll",
  direction="l",
 },
 [0x1D6FD]={
  category="ll",
  direction="l",
 },
 [0x1D6FE]={
  category="ll",
  direction="l",
 },
 [0x1D6FF]={
  category="ll",
  direction="l",
 },
 [0x1D700]={
  category="ll",
  direction="l",
 },
 [0x1D701]={
  category="ll",
  direction="l",
 },
 [0x1D702]={
  category="ll",
  direction="l",
 },
 [0x1D703]={
  category="ll",
  direction="l",
 },
 [0x1D704]={
  category="ll",
  direction="l",
 },
 [0x1D705]={
  category="ll",
  direction="l",
 },
 [0x1D706]={
  category="ll",
  direction="l",
 },
 [0x1D707]={
  category="ll",
  direction="l",
 },
 [0x1D708]={
  category="ll",
  direction="l",
 },
 [0x1D709]={
  category="ll",
  direction="l",
 },
 [0x1D70A]={
  category="ll",
  direction="l",
 },
 [0x1D70B]={
  category="ll",
  direction="l",
 },
 [0x1D70C]={
  category="ll",
  direction="l",
 },
 [0x1D70D]={
  category="ll",
  direction="l",
 },
 [0x1D70E]={
  category="ll",
  direction="l",
 },
 [0x1D70F]={
  category="ll",
  direction="l",
 },
 [0x1D710]={
  category="ll",
  direction="l",
 },
 [0x1D711]={
  category="ll",
  direction="l",
 },
 [0x1D712]={
  category="ll",
  direction="l",
 },
 [0x1D713]={
  category="ll",
  direction="l",
 },
 [0x1D714]={
  category="ll",
  direction="l",
 },
 [0x1D715]={
  category="sm",
  direction="on",
 },
 [0x1D716]={
  category="ll",
  direction="l",
 },
 [0x1D717]={
  category="ll",
  direction="l",
 },
 [0x1D718]={
  category="ll",
  direction="l",
 },
 [0x1D719]={
  category="ll",
  direction="l",
 },
 [0x1D71A]={
  category="ll",
  direction="l",
 },
 [0x1D71B]={
  category="ll",
  direction="l",
 },
 [0x1D71C]={
  category="lu",
  direction="l",
 },
 [0x1D71D]={
  category="lu",
  direction="l",
 },
 [0x1D71E]={
  category="lu",
  direction="l",
 },
 [0x1D71F]={
  category="lu",
  direction="l",
 },
 [0x1D720]={
  category="lu",
  direction="l",
 },
 [0x1D721]={
  category="lu",
  direction="l",
 },
 [0x1D722]={
  category="lu",
  direction="l",
 },
 [0x1D723]={
  category="lu",
  direction="l",
 },
 [0x1D724]={
  category="lu",
  direction="l",
 },
 [0x1D725]={
  category="lu",
  direction="l",
 },
 [0x1D726]={
  category="lu",
  direction="l",
 },
 [0x1D727]={
  category="lu",
  direction="l",
 },
 [0x1D728]={
  category="lu",
  direction="l",
 },
 [0x1D729]={
  category="lu",
  direction="l",
 },
 [0x1D72A]={
  category="lu",
  direction="l",
 },
 [0x1D72B]={
  category="lu",
  direction="l",
 },
 [0x1D72C]={
  category="lu",
  direction="l",
 },
 [0x1D72D]={
  category="lu",
  direction="l",
 },
 [0x1D72E]={
  category="lu",
  direction="l",
 },
 [0x1D72F]={
  category="lu",
  direction="l",
 },
 [0x1D730]={
  category="lu",
  direction="l",
 },
 [0x1D731]={
  category="lu",
  direction="l",
 },
 [0x1D732]={
  category="lu",
  direction="l",
 },
 [0x1D733]={
  category="lu",
  direction="l",
 },
 [0x1D734]={
  category="lu",
  direction="l",
 },
 [0x1D735]={
  category="sm",
  direction="l",
 },
 [0x1D736]={
  category="ll",
  direction="l",
 },
 [0x1D737]={
  category="ll",
  direction="l",
 },
 [0x1D738]={
  category="ll",
  direction="l",
 },
 [0x1D739]={
  category="ll",
  direction="l",
 },
 [0x1D73A]={
  category="ll",
  direction="l",
 },
 [0x1D73B]={
  category="ll",
  direction="l",
 },
 [0x1D73C]={
  category="ll",
  direction="l",
 },
 [0x1D73D]={
  category="ll",
  direction="l",
 },
 [0x1D73E]={
  category="ll",
  direction="l",
 },
 [0x1D73F]={
  category="ll",
  direction="l",
 },
 [0x1D740]={
  category="ll",
  direction="l",
 },
 [0x1D741]={
  category="ll",
  direction="l",
 },
 [0x1D742]={
  category="ll",
  direction="l",
 },
 [0x1D743]={
  category="ll",
  direction="l",
 },
 [0x1D744]={
  category="ll",
  direction="l",
 },
 [0x1D745]={
  category="ll",
  direction="l",
 },
 [0x1D746]={
  category="ll",
  direction="l",
 },
 [0x1D747]={
  category="ll",
  direction="l",
 },
 [0x1D748]={
  category="ll",
  direction="l",
 },
 [0x1D749]={
  category="ll",
  direction="l",
 },
 [0x1D74A]={
  category="ll",
  direction="l",
 },
 [0x1D74B]={
  category="ll",
  direction="l",
 },
 [0x1D74C]={
  category="ll",
  direction="l",
 },
 [0x1D74D]={
  category="ll",
  direction="l",
 },
 [0x1D74E]={
  category="ll",
  direction="l",
 },
 [0x1D74F]={
  category="sm",
  direction="on",
 },
 [0x1D750]={
  category="ll",
  direction="l",
 },
 [0x1D751]={
  category="ll",
  direction="l",
 },
 [0x1D752]={
  category="ll",
  direction="l",
 },
 [0x1D753]={
  category="ll",
  direction="l",
 },
 [0x1D754]={
  category="ll",
  direction="l",
 },
 [0x1D755]={
  category="ll",
  direction="l",
 },
 [0x1D756]={
  category="lu",
  direction="l",
 },
 [0x1D757]={
  category="lu",
  direction="l",
 },
 [0x1D758]={
  category="lu",
  direction="l",
 },
 [0x1D759]={
  category="lu",
  direction="l",
 },
 [0x1D75A]={
  category="lu",
  direction="l",
 },
 [0x1D75B]={
  category="lu",
  direction="l",
 },
 [0x1D75C]={
  category="lu",
  direction="l",
 },
 [0x1D75D]={
  category="lu",
  direction="l",
 },
 [0x1D75E]={
  category="lu",
  direction="l",
 },
 [0x1D75F]={
  category="lu",
  direction="l",
 },
 [0x1D760]={
  category="lu",
  direction="l",
 },
 [0x1D761]={
  category="lu",
  direction="l",
 },
 [0x1D762]={
  category="lu",
  direction="l",
 },
 [0x1D763]={
  category="lu",
  direction="l",
 },
 [0x1D764]={
  category="lu",
  direction="l",
 },
 [0x1D765]={
  category="lu",
  direction="l",
 },
 [0x1D766]={
  category="lu",
  direction="l",
 },
 [0x1D767]={
  category="lu",
  direction="l",
 },
 [0x1D768]={
  category="lu",
  direction="l",
 },
 [0x1D769]={
  category="lu",
  direction="l",
 },
 [0x1D76A]={
  category="lu",
  direction="l",
 },
 [0x1D76B]={
  category="lu",
  direction="l",
 },
 [0x1D76C]={
  category="lu",
  direction="l",
 },
 [0x1D76D]={
  category="lu",
  direction="l",
 },
 [0x1D76E]={
  category="lu",
  direction="l",
 },
 [0x1D76F]={
  category="sm",
  direction="l",
 },
 [0x1D770]={
  category="ll",
  direction="l",
 },
 [0x1D771]={
  category="ll",
  direction="l",
 },
 [0x1D772]={
  category="ll",
  direction="l",
 },
 [0x1D773]={
  category="ll",
  direction="l",
 },
 [0x1D774]={
  category="ll",
  direction="l",
 },
 [0x1D775]={
  category="ll",
  direction="l",
 },
 [0x1D776]={
  category="ll",
  direction="l",
 },
 [0x1D777]={
  category="ll",
  direction="l",
 },
 [0x1D778]={
  category="ll",
  direction="l",
 },
 [0x1D779]={
  category="ll",
  direction="l",
 },
 [0x1D77A]={
  category="ll",
  direction="l",
 },
 [0x1D77B]={
  category="ll",
  direction="l",
 },
 [0x1D77C]={
  category="ll",
  direction="l",
 },
 [0x1D77D]={
  category="ll",
  direction="l",
 },
 [0x1D77E]={
  category="ll",
  direction="l",
 },
 [0x1D77F]={
  category="ll",
  direction="l",
 },
 [0x1D780]={
  category="ll",
  direction="l",
 },
 [0x1D781]={
  category="ll",
  direction="l",
 },
 [0x1D782]={
  category="ll",
  direction="l",
 },
 [0x1D783]={
  category="ll",
  direction="l",
 },
 [0x1D784]={
  category="ll",
  direction="l",
 },
 [0x1D785]={
  category="ll",
  direction="l",
 },
 [0x1D786]={
  category="ll",
  direction="l",
 },
 [0x1D787]={
  category="ll",
  direction="l",
 },
 [0x1D788]={
  category="ll",
  direction="l",
 },
 [0x1D789]={
  category="sm",
  direction="on",
 },
 [0x1D78A]={
  category="ll",
  direction="l",
 },
 [0x1D78B]={
  category="ll",
  direction="l",
 },
 [0x1D78C]={
  category="ll",
  direction="l",
 },
 [0x1D78D]={
  category="ll",
  direction="l",
 },
 [0x1D78E]={
  category="ll",
  direction="l",
 },
 [0x1D78F]={
  category="ll",
  direction="l",
 },
 [0x1D790]={
  category="lu",
  direction="l",
 },
 [0x1D791]={
  category="lu",
  direction="l",
 },
 [0x1D792]={
  category="lu",
  direction="l",
 },
 [0x1D793]={
  category="lu",
  direction="l",
 },
 [0x1D794]={
  category="lu",
  direction="l",
 },
 [0x1D795]={
  category="lu",
  direction="l",
 },
 [0x1D796]={
  category="lu",
  direction="l",
 },
 [0x1D797]={
  category="lu",
  direction="l",
 },
 [0x1D798]={
  category="lu",
  direction="l",
 },
 [0x1D799]={
  category="lu",
  direction="l",
 },
 [0x1D79A]={
  category="lu",
  direction="l",
 },
 [0x1D79B]={
  category="lu",
  direction="l",
 },
 [0x1D79C]={
  category="lu",
  direction="l",
 },
 [0x1D79D]={
  category="lu",
  direction="l",
 },
 [0x1D79E]={
  category="lu",
  direction="l",
 },
 [0x1D79F]={
  category="lu",
  direction="l",
 },
 [0x1D7A0]={
  category="lu",
  direction="l",
 },
 [0x1D7A1]={
  category="lu",
  direction="l",
 },
 [0x1D7A2]={
  category="lu",
  direction="l",
 },
 [0x1D7A3]={
  category="lu",
  direction="l",
 },
 [0x1D7A4]={
  category="lu",
  direction="l",
 },
 [0x1D7A5]={
  category="lu",
  direction="l",
 },
 [0x1D7A6]={
  category="lu",
  direction="l",
 },
 [0x1D7A7]={
  category="lu",
  direction="l",
 },
 [0x1D7A8]={
  category="lu",
  direction="l",
 },
 [0x1D7A9]={
  category="sm",
  direction="l",
 },
 [0x1D7AA]={
  category="ll",
  direction="l",
 },
 [0x1D7AB]={
  category="ll",
  direction="l",
 },
 [0x1D7AC]={
  category="ll",
  direction="l",
 },
 [0x1D7AD]={
  category="ll",
  direction="l",
 },
 [0x1D7AE]={
  category="ll",
  direction="l",
 },
 [0x1D7AF]={
  category="ll",
  direction="l",
 },
 [0x1D7B0]={
  category="ll",
  direction="l",
 },
 [0x1D7B1]={
  category="ll",
  direction="l",
 },
 [0x1D7B2]={
  category="ll",
  direction="l",
 },
 [0x1D7B3]={
  category="ll",
  direction="l",
 },
 [0x1D7B4]={
  category="ll",
  direction="l",
 },
 [0x1D7B5]={
  category="ll",
  direction="l",
 },
 [0x1D7B6]={
  category="ll",
  direction="l",
 },
 [0x1D7B7]={
  category="ll",
  direction="l",
 },
 [0x1D7B8]={
  category="ll",
  direction="l",
 },
 [0x1D7B9]={
  category="ll",
  direction="l",
 },
 [0x1D7BA]={
  category="ll",
  direction="l",
 },
 [0x1D7BB]={
  category="ll",
  direction="l",
 },
 [0x1D7BC]={
  category="ll",
  direction="l",
 },
 [0x1D7BD]={
  category="ll",
  direction="l",
 },
 [0x1D7BE]={
  category="ll",
  direction="l",
 },
 [0x1D7BF]={
  category="ll",
  direction="l",
 },
 [0x1D7C0]={
  category="ll",
  direction="l",
 },
 [0x1D7C1]={
  category="ll",
  direction="l",
 },
 [0x1D7C2]={
  category="ll",
  direction="l",
 },
 [0x1D7C3]={
  category="sm",
  direction="on",
 },
 [0x1D7C4]={
  category="ll",
  direction="l",
 },
 [0x1D7C5]={
  category="ll",
  direction="l",
 },
 [0x1D7C6]={
  category="ll",
  direction="l",
 },
 [0x1D7C7]={
  category="ll",
  direction="l",
 },
 [0x1D7C8]={
  category="ll",
  direction="l",
 },
 [0x1D7C9]={
  category="ll",
  direction="l",
 },
 [0x1D7CA]={
  category="lu",
  direction="l",
 },
 [0x1D7CB]={
  category="ll",
  direction="l",
 },
 [0x1D7CE]={
  category="nd",
  direction="en",
 },
 [0x1D7CF]={
  category="nd",
  direction="en",
 },
 [0x1D7D0]={
  category="nd",
  direction="en",
 },
 [0x1D7D1]={
  category="nd",
  direction="en",
 },
 [0x1D7D2]={
  category="nd",
  direction="en",
 },
 [0x1D7D3]={
  category="nd",
  direction="en",
 },
 [0x1D7D4]={
  category="nd",
  direction="en",
 },
 [0x1D7D5]={
  category="nd",
  direction="en",
 },
 [0x1D7D6]={
  category="nd",
  direction="en",
 },
 [0x1D7D7]={
  category="nd",
  direction="en",
 },
 [0x1D7D8]={
  category="nd",
  direction="en",
 },
 [0x1D7D9]={
  category="nd",
  direction="en",
 },
 [0x1D7DA]={
  category="nd",
  direction="en",
 },
 [0x1D7DB]={
  category="nd",
  direction="en",
 },
 [0x1D7DC]={
  category="nd",
  direction="en",
 },
 [0x1D7DD]={
  category="nd",
  direction="en",
 },
 [0x1D7DE]={
  category="nd",
  direction="en",
 },
 [0x1D7DF]={
  category="nd",
  direction="en",
 },
 [0x1D7E0]={
  category="nd",
  direction="en",
 },
 [0x1D7E1]={
  category="nd",
  direction="en",
 },
 [0x1D7E2]={
  category="nd",
  direction="en",
 },
 [0x1D7E3]={
  category="nd",
  direction="en",
 },
 [0x1D7E4]={
  category="nd",
  direction="en",
 },
 [0x1D7E5]={
  category="nd",
  direction="en",
 },
 [0x1D7E6]={
  category="nd",
  direction="en",
 },
 [0x1D7E7]={
  category="nd",
  direction="en",
 },
 [0x1D7E8]={
  category="nd",
  direction="en",
 },
 [0x1D7E9]={
  category="nd",
  direction="en",
 },
 [0x1D7EA]={
  category="nd",
  direction="en",
 },
 [0x1D7EB]={
  category="nd",
  direction="en",
 },
 [0x1D7EC]={
  category="nd",
  direction="en",
 },
 [0x1D7ED]={
  category="nd",
  direction="en",
 },
 [0x1D7EE]={
  category="nd",
  direction="en",
 },
 [0x1D7EF]={
  category="nd",
  direction="en",
 },
 [0x1D7F0]={
  category="nd",
  direction="en",
 },
 [0x1D7F1]={
  category="nd",
  direction="en",
 },
 [0x1D7F2]={
  category="nd",
  direction="en",
 },
 [0x1D7F3]={
  category="nd",
  direction="en",
 },
 [0x1D7F4]={
  category="nd",
  direction="en",
 },
 [0x1D7F5]={
  category="nd",
  direction="en",
 },
 [0x1D7F6]={
  category="nd",
  direction="en",
 },
 [0x1D7F7]={
  category="nd",
  direction="en",
 },
 [0x1D7F8]={
  category="nd",
  direction="en",
 },
 [0x1D7F9]={
  category="nd",
  direction="en",
 },
 [0x1D7FA]={
  category="nd",
  direction="en",
 },
 [0x1D7FB]={
  category="nd",
  direction="en",
 },
 [0x1D7FC]={
  category="nd",
  direction="en",
 },
 [0x1D7FD]={
  category="nd",
  direction="en",
 },
 [0x1D7FE]={
  category="nd",
  direction="en",
 },
 [0x1D7FF]={
  category="nd",
  direction="en",
 },
 [0x1E800]={
  category="lo",
  direction="r",
 },
 [0x1E801]={
  category="lo",
  direction="r",
 },
 [0x1E802]={
  category="lo",
  direction="r",
 },
 [0x1E803]={
  category="lo",
  direction="r",
 },
 [0x1E804]={
  category="lo",
  direction="r",
 },
 [0x1E805]={
  category="lo",
  direction="r",
 },
 [0x1E806]={
  category="lo",
  direction="r",
 },
 [0x1E807]={
  category="lo",
  direction="r",
 },
 [0x1E808]={
  category="lo",
  direction="r",
 },
 [0x1E809]={
  category="lo",
  direction="r",
 },
 [0x1E80A]={
  category="lo",
  direction="r",
 },
 [0x1E80B]={
  category="lo",
  direction="r",
 },
 [0x1E80C]={
  category="lo",
  direction="r",
 },
 [0x1E80D]={
  category="lo",
  direction="r",
 },
 [0x1E80E]={
  category="lo",
  direction="r",
 },
 [0x1E80F]={
  category="lo",
  direction="r",
 },
 [0x1E810]={
  category="lo",
  direction="r",
 },
 [0x1E811]={
  category="lo",
  direction="r",
 },
 [0x1E812]={
  category="lo",
  direction="r",
 },
 [0x1E813]={
  category="lo",
  direction="r",
 },
 [0x1E814]={
  category="lo",
  direction="r",
 },
 [0x1E815]={
  category="lo",
  direction="r",
 },
 [0x1E816]={
  category="lo",
  direction="r",
 },
 [0x1E817]={
  category="lo",
  direction="r",
 },
 [0x1E818]={
  category="lo",
  direction="r",
 },
 [0x1E819]={
  category="lo",
  direction="r",
 },
 [0x1E81A]={
  category="lo",
  direction="r",
 },
 [0x1E81B]={
  category="lo",
  direction="r",
 },
 [0x1E81C]={
  category="lo",
  direction="r",
 },
 [0x1E81D]={
  category="lo",
  direction="r",
 },
 [0x1E81E]={
  category="lo",
  direction="r",
 },
 [0x1E81F]={
  category="lo",
  direction="r",
 },
 [0x1E820]={
  category="lo",
  direction="r",
 },
 [0x1E821]={
  category="lo",
  direction="r",
 },
 [0x1E822]={
  category="lo",
  direction="r",
 },
 [0x1E823]={
  category="lo",
  direction="r",
 },
 [0x1E824]={
  category="lo",
  direction="r",
 },
 [0x1E825]={
  category="lo",
  direction="r",
 },
 [0x1E826]={
  category="lo",
  direction="r",
 },
 [0x1E827]={
  category="lo",
  direction="r",
 },
 [0x1E828]={
  category="lo",
  direction="r",
 },
 [0x1E829]={
  category="lo",
  direction="r",
 },
 [0x1E82A]={
  category="lo",
  direction="r",
 },
 [0x1E82B]={
  category="lo",
  direction="r",
 },
 [0x1E82C]={
  category="lo",
  direction="r",
 },
 [0x1E82D]={
  category="lo",
  direction="r",
 },
 [0x1E82E]={
  category="lo",
  direction="r",
 },
 [0x1E82F]={
  category="lo",
  direction="r",
 },
 [0x1E830]={
  category="lo",
  direction="r",
 },
 [0x1E831]={
  category="lo",
  direction="r",
 },
 [0x1E832]={
  category="lo",
  direction="r",
 },
 [0x1E833]={
  category="lo",
  direction="r",
 },
 [0x1E834]={
  category="lo",
  direction="r",
 },
 [0x1E835]={
  category="lo",
  direction="r",
 },
 [0x1E836]={
  category="lo",
  direction="r",
 },
 [0x1E837]={
  category="lo",
  direction="r",
 },
 [0x1E838]={
  category="lo",
  direction="r",
 },
 [0x1E839]={
  category="lo",
  direction="r",
 },
 [0x1E83A]={
  category="lo",
  direction="r",
 },
 [0x1E83B]={
  category="lo",
  direction="r",
 },
 [0x1E83C]={
  category="lo",
  direction="r",
 },
 [0x1E83D]={
  category="lo",
  direction="r",
 },
 [0x1E83E]={
  category="lo",
  direction="r",
 },
 [0x1E83F]={
  category="lo",
  direction="r",
 },
 [0x1E840]={
  category="lo",
  direction="r",
 },
 [0x1E841]={
  category="lo",
  direction="r",
 },
 [0x1E842]={
  category="lo",
  direction="r",
 },
 [0x1E843]={
  category="lo",
  direction="r",
 },
 [0x1E844]={
  category="lo",
  direction="r",
 },
 [0x1E845]={
  category="lo",
  direction="r",
 },
 [0x1E846]={
  category="lo",
  direction="r",
 },
 [0x1E847]={
  category="lo",
  direction="r",
 },
 [0x1E848]={
  category="lo",
  direction="r",
 },
 [0x1E849]={
  category="lo",
  direction="r",
 },
 [0x1E84A]={
  category="lo",
  direction="r",
 },
 [0x1E84B]={
  category="lo",
  direction="r",
 },
 [0x1E84C]={
  category="lo",
  direction="r",
 },
 [0x1E84D]={
  category="lo",
  direction="r",
 },
 [0x1E84E]={
  category="lo",
  direction="r",
 },
 [0x1E84F]={
  category="lo",
  direction="r",
 },
 [0x1E850]={
  category="lo",
  direction="r",
 },
 [0x1E851]={
  category="lo",
  direction="r",
 },
 [0x1E852]={
  category="lo",
  direction="r",
 },
 [0x1E853]={
  category="lo",
  direction="r",
 },
 [0x1E854]={
  category="lo",
  direction="r",
 },
 [0x1E855]={
  category="lo",
  direction="r",
 },
 [0x1E856]={
  category="lo",
  direction="r",
 },
 [0x1E857]={
  category="lo",
  direction="r",
 },
 [0x1E858]={
  category="lo",
  direction="r",
 },
 [0x1E859]={
  category="lo",
  direction="r",
 },
 [0x1E85A]={
  category="lo",
  direction="r",
 },
 [0x1E85B]={
  category="lo",
  direction="r",
 },
 [0x1E85C]={
  category="lo",
  direction="r",
 },
 [0x1E85D]={
  category="lo",
  direction="r",
 },
 [0x1E85E]={
  category="lo",
  direction="r",
 },
 [0x1E85F]={
  category="lo",
  direction="r",
 },
 [0x1E860]={
  category="lo",
  direction="r",
 },
 [0x1E861]={
  category="lo",
  direction="r",
 },
 [0x1E862]={
  category="lo",
  direction="r",
 },
 [0x1E863]={
  category="lo",
  direction="r",
 },
 [0x1E864]={
  category="lo",
  direction="r",
 },
 [0x1E865]={
  category="lo",
  direction="r",
 },
 [0x1E866]={
  category="lo",
  direction="r",
 },
 [0x1E867]={
  category="lo",
  direction="r",
 },
 [0x1E868]={
  category="lo",
  direction="r",
 },
 [0x1E869]={
  category="lo",
  direction="r",
 },
 [0x1E86A]={
  category="lo",
  direction="r",
 },
 [0x1E86B]={
  category="lo",
  direction="r",
 },
 [0x1E86C]={
  category="lo",
  direction="r",
 },
 [0x1E86D]={
  category="lo",
  direction="r",
 },
 [0x1E86E]={
  category="lo",
  direction="r",
 },
 [0x1E86F]={
  category="lo",
  direction="r",
 },
 [0x1E870]={
  category="lo",
  direction="r",
 },
 [0x1E871]={
  category="lo",
  direction="r",
 },
 [0x1E872]={
  category="lo",
  direction="r",
 },
 [0x1E873]={
  category="lo",
  direction="r",
 },
 [0x1E874]={
  category="lo",
  direction="r",
 },
 [0x1E875]={
  category="lo",
  direction="r",
 },
 [0x1E876]={
  category="lo",
  direction="r",
 },
 [0x1E877]={
  category="lo",
  direction="r",
 },
 [0x1E878]={
  category="lo",
  direction="r",
 },
 [0x1E879]={
  category="lo",
  direction="r",
 },
 [0x1E87A]={
  category="lo",
  direction="r",
 },
 [0x1E87B]={
  category="lo",
  direction="r",
 },
 [0x1E87C]={
  category="lo",
  direction="r",
 },
 [0x1E87D]={
  category="lo",
  direction="r",
 },
 [0x1E87E]={
  category="lo",
  direction="r",
 },
 [0x1E87F]={
  category="lo",
  direction="r",
 },
 [0x1E880]={
  category="lo",
  direction="r",
 },
 [0x1E881]={
  category="lo",
  direction="r",
 },
 [0x1E882]={
  category="lo",
  direction="r",
 },
 [0x1E883]={
  category="lo",
  direction="r",
 },
 [0x1E884]={
  category="lo",
  direction="r",
 },
 [0x1E885]={
  category="lo",
  direction="r",
 },
 [0x1E886]={
  category="lo",
  direction="r",
 },
 [0x1E887]={
  category="lo",
  direction="r",
 },
 [0x1E888]={
  category="lo",
  direction="r",
 },
 [0x1E889]={
  category="lo",
  direction="r",
 },
 [0x1E88A]={
  category="lo",
  direction="r",
 },
 [0x1E88B]={
  category="lo",
  direction="r",
 },
 [0x1E88C]={
  category="lo",
  direction="r",
 },
 [0x1E88D]={
  category="lo",
  direction="r",
 },
 [0x1E88E]={
  category="lo",
  direction="r",
 },
 [0x1E88F]={
  category="lo",
  direction="r",
 },
 [0x1E890]={
  category="lo",
  direction="r",
 },
 [0x1E891]={
  category="lo",
  direction="r",
 },
 [0x1E892]={
  category="lo",
  direction="r",
 },
 [0x1E893]={
  category="lo",
  direction="r",
 },
 [0x1E894]={
  category="lo",
  direction="r",
 },
 [0x1E895]={
  category="lo",
  direction="r",
 },
 [0x1E896]={
  category="lo",
  direction="r",
 },
 [0x1E897]={
  category="lo",
  direction="r",
 },
 [0x1E898]={
  category="lo",
  direction="r",
 },
 [0x1E899]={
  category="lo",
  direction="r",
 },
 [0x1E89A]={
  category="lo",
  direction="r",
 },
 [0x1E89B]={
  category="lo",
  direction="r",
 },
 [0x1E89C]={
  category="lo",
  direction="r",
 },
 [0x1E89D]={
  category="lo",
  direction="r",
 },
 [0x1E89E]={
  category="lo",
  direction="r",
 },
 [0x1E89F]={
  category="lo",
  direction="r",
 },
 [0x1E8A0]={
  category="lo",
  direction="r",
 },
 [0x1E8A1]={
  category="lo",
  direction="r",
 },
 [0x1E8A2]={
  category="lo",
  direction="r",
 },
 [0x1E8A3]={
  category="lo",
  direction="r",
 },
 [0x1E8A4]={
  category="lo",
  direction="r",
 },
 [0x1E8A5]={
  category="lo",
  direction="r",
 },
 [0x1E8A6]={
  category="lo",
  direction="r",
 },
 [0x1E8A7]={
  category="lo",
  direction="r",
 },
 [0x1E8A8]={
  category="lo",
  direction="r",
 },
 [0x1E8A9]={
  category="lo",
  direction="r",
 },
 [0x1E8AA]={
  category="lo",
  direction="r",
 },
 [0x1E8AB]={
  category="lo",
  direction="r",
 },
 [0x1E8AC]={
  category="lo",
  direction="r",
 },
 [0x1E8AD]={
  category="lo",
  direction="r",
 },
 [0x1E8AE]={
  category="lo",
  direction="r",
 },
 [0x1E8AF]={
  category="lo",
  direction="r",
 },
 [0x1E8B0]={
  category="lo",
  direction="r",
 },
 [0x1E8B1]={
  category="lo",
  direction="r",
 },
 [0x1E8B2]={
  category="lo",
  direction="r",
 },
 [0x1E8B3]={
  category="lo",
  direction="r",
 },
 [0x1E8B4]={
  category="lo",
  direction="r",
 },
 [0x1E8B5]={
  category="lo",
  direction="r",
 },
 [0x1E8B6]={
  category="lo",
  direction="r",
 },
 [0x1E8B7]={
  category="lo",
  direction="r",
 },
 [0x1E8B8]={
  category="lo",
  direction="r",
 },
 [0x1E8B9]={
  category="lo",
  direction="r",
 },
 [0x1E8BA]={
  category="lo",
  direction="r",
 },
 [0x1E8BB]={
  category="lo",
  direction="r",
 },
 [0x1E8BC]={
  category="lo",
  direction="r",
 },
 [0x1E8BD]={
  category="lo",
  direction="r",
 },
 [0x1E8BE]={
  category="lo",
  direction="r",
 },
 [0x1E8BF]={
  category="lo",
  direction="r",
 },
 [0x1E8C0]={
  category="lo",
  direction="r",
 },
 [0x1E8C1]={
  category="lo",
  direction="r",
 },
 [0x1E8C2]={
  category="lo",
  direction="r",
 },
 [0x1E8C3]={
  category="lo",
  direction="r",
 },
 [0x1E8C4]={
  category="lo",
  direction="r",
 },
 [0x1E8C7]={
  category="no",
  direction="r",
 },
 [0x1E8C8]={
  category="no",
  direction="r",
 },
 [0x1E8C9]={
  category="no",
  direction="r",
 },
 [0x1E8CA]={
  category="no",
  direction="r",
 },
 [0x1E8CB]={
  category="no",
  direction="r",
 },
 [0x1E8CC]={
  category="no",
  direction="r",
 },
 [0x1E8CD]={
  category="no",
  direction="r",
 },
 [0x1E8CE]={
  category="no",
  direction="r",
 },
 [0x1E8CF]={
  category="no",
  direction="r",
 },
 [0x1E8D0]={
  category="mn",
  direction="nsm",
 },
 [0x1E8D1]={
  category="mn",
  direction="nsm",
 },
 [0x1E8D2]={
  category="mn",
  direction="nsm",
 },
 [0x1E8D3]={
  category="mn",
  direction="nsm",
 },
 [0x1E8D4]={
  category="mn",
  direction="nsm",
 },
 [0x1E8D5]={
  category="mn",
  direction="nsm",
 },
 [0x1E8D6]={
  category="mn",
  direction="nsm",
 },
 [0x1EE00]={
  category="lo",
  direction="al",
 },
 [0x1EE01]={
  category="lo",
  direction="al",
 },
 [0x1EE02]={
  category="lo",
  direction="al",
 },
 [0x1EE03]={
  category="lo",
  direction="al",
 },
 [0x1EE05]={
  category="lo",
  direction="al",
 },
 [0x1EE06]={
  category="lo",
  direction="al",
 },
 [0x1EE07]={
  category="lo",
  direction="al",
 },
 [0x1EE08]={
  category="lo",
  direction="al",
 },
 [0x1EE09]={
  category="lo",
  direction="al",
 },
 [0x1EE0A]={
  category="lo",
  direction="al",
 },
 [0x1EE0B]={
  category="lo",
  direction="al",
 },
 [0x1EE0C]={
  category="lo",
  direction="al",
 },
 [0x1EE0D]={
  category="lo",
  direction="al",
 },
 [0x1EE0E]={
  category="lo",
  direction="al",
 },
 [0x1EE0F]={
  category="lo",
  direction="al",
 },
 [0x1EE10]={
  category="lo",
  direction="al",
 },
 [0x1EE11]={
  category="lo",
  direction="al",
 },
 [0x1EE12]={
  category="lo",
  direction="al",
 },
 [0x1EE13]={
  category="lo",
  direction="al",
 },
 [0x1EE14]={
  category="lo",
  direction="al",
 },
 [0x1EE15]={
  category="lo",
  direction="al",
 },
 [0x1EE16]={
  category="lo",
  direction="al",
 },
 [0x1EE17]={
  category="lo",
  direction="al",
 },
 [0x1EE18]={
  category="lo",
  direction="al",
 },
 [0x1EE19]={
  category="lo",
  direction="al",
 },
 [0x1EE1A]={
  category="lo",
  direction="al",
 },
 [0x1EE1B]={
  category="lo",
  direction="al",
 },
 [0x1EE1C]={
  category="lo",
  direction="al",
 },
 [0x1EE1D]={
  category="lo",
  direction="al",
 },
 [0x1EE1E]={
  category="lo",
  direction="al",
 },
 [0x1EE1F]={
  category="lo",
  direction="al",
 },
 [0x1EE21]={
  category="lo",
  direction="al",
 },
 [0x1EE22]={
  category="lo",
  direction="al",
 },
 [0x1EE24]={
  category="lo",
  direction="al",
 },
 [0x1EE27]={
  category="lo",
  direction="al",
 },
 [0x1EE29]={
  category="lo",
  direction="al",
 },
 [0x1EE2A]={
  category="lo",
  direction="al",
 },
 [0x1EE2B]={
  category="lo",
  direction="al",
 },
 [0x1EE2C]={
  category="lo",
  direction="al",
 },
 [0x1EE2D]={
  category="lo",
  direction="al",
 },
 [0x1EE2E]={
  category="lo",
  direction="al",
 },
 [0x1EE2F]={
  category="lo",
  direction="al",
 },
 [0x1EE30]={
  category="lo",
  direction="al",
 },
 [0x1EE31]={
  category="lo",
  direction="al",
 },
 [0x1EE32]={
  category="lo",
  direction="al",
 },
 [0x1EE34]={
  category="lo",
  direction="al",
 },
 [0x1EE35]={
  category="lo",
  direction="al",
 },
 [0x1EE36]={
  category="lo",
  direction="al",
 },
 [0x1EE37]={
  category="lo",
  direction="al",
 },
 [0x1EE39]={
  category="lo",
  direction="al",
 },
 [0x1EE3B]={
  category="lo",
  direction="al",
 },
 [0x1EE42]={
  category="lo",
  direction="al",
 },
 [0x1EE47]={
  category="lo",
  direction="al",
 },
 [0x1EE49]={
  category="lo",
  direction="al",
 },
 [0x1EE4B]={
  category="lo",
  direction="al",
 },
 [0x1EE4D]={
  category="lo",
  direction="al",
 },
 [0x1EE4E]={
  category="lo",
  direction="al",
 },
 [0x1EE4F]={
  category="lo",
  direction="al",
 },
 [0x1EE51]={
  category="lo",
  direction="al",
 },
 [0x1EE52]={
  category="lo",
  direction="al",
 },
 [0x1EE54]={
  category="lo",
  direction="al",
 },
 [0x1EE57]={
  category="lo",
  direction="al",
 },
 [0x1EE59]={
  category="lo",
  direction="al",
 },
 [0x1EE5B]={
  category="lo",
  direction="al",
 },
 [0x1EE5D]={
  category="lo",
  direction="al",
 },
 [0x1EE5F]={
  category="lo",
  direction="al",
 },
 [0x1EE61]={
  category="lo",
  direction="al",
 },
 [0x1EE62]={
  category="lo",
  direction="al",
 },
 [0x1EE64]={
  category="lo",
  direction="al",
 },
 [0x1EE67]={
  category="lo",
  direction="al",
 },
 [0x1EE68]={
  category="lo",
  direction="al",
 },
 [0x1EE69]={
  category="lo",
  direction="al",
 },
 [0x1EE6A]={
  category="lo",
  direction="al",
 },
 [0x1EE6C]={
  category="lo",
  direction="al",
 },
 [0x1EE6D]={
  category="lo",
  direction="al",
 },
 [0x1EE6E]={
  category="lo",
  direction="al",
 },
 [0x1EE6F]={
  category="lo",
  direction="al",
 },
 [0x1EE70]={
  category="lo",
  direction="al",
 },
 [0x1EE71]={
  category="lo",
  direction="al",
 },
 [0x1EE72]={
  category="lo",
  direction="al",
 },
 [0x1EE74]={
  category="lo",
  direction="al",
 },
 [0x1EE75]={
  category="lo",
  direction="al",
 },
 [0x1EE76]={
  category="lo",
  direction="al",
 },
 [0x1EE77]={
  category="lo",
  direction="al",
 },
 [0x1EE79]={
  category="lo",
  direction="al",
 },
 [0x1EE7A]={
  category="lo",
  direction="al",
 },
 [0x1EE7B]={
  category="lo",
  direction="al",
 },
 [0x1EE7C]={
  category="lo",
  direction="al",
 },
 [0x1EE7E]={
  category="lo",
  direction="al",
 },
 [0x1EE80]={
  category="lo",
  direction="al",
 },
 [0x1EE81]={
  category="lo",
  direction="al",
 },
 [0x1EE82]={
  category="lo",
  direction="al",
 },
 [0x1EE83]={
  category="lo",
  direction="al",
 },
 [0x1EE84]={
  category="lo",
  direction="al",
 },
 [0x1EE85]={
  category="lo",
  direction="al",
 },
 [0x1EE86]={
  category="lo",
  direction="al",
 },
 [0x1EE87]={
  category="lo",
  direction="al",
 },
 [0x1EE88]={
  category="lo",
  direction="al",
 },
 [0x1EE89]={
  category="lo",
  direction="al",
 },
 [0x1EE8B]={
  category="lo",
  direction="al",
 },
 [0x1EE8C]={
  category="lo",
  direction="al",
 },
 [0x1EE8D]={
  category="lo",
  direction="al",
 },
 [0x1EE8E]={
  category="lo",
  direction="al",
 },
 [0x1EE8F]={
  category="lo",
  direction="al",
 },
 [0x1EE90]={
  category="lo",
  direction="al",
 },
 [0x1EE91]={
  category="lo",
  direction="al",
 },
 [0x1EE92]={
  category="lo",
  direction="al",
 },
 [0x1EE93]={
  category="lo",
  direction="al",
 },
 [0x1EE94]={
  category="lo",
  direction="al",
 },
 [0x1EE95]={
  category="lo",
  direction="al",
 },
 [0x1EE96]={
  category="lo",
  direction="al",
 },
 [0x1EE97]={
  category="lo",
  direction="al",
 },
 [0x1EE98]={
  category="lo",
  direction="al",
 },
 [0x1EE99]={
  category="lo",
  direction="al",
 },
 [0x1EE9A]={
  category="lo",
  direction="al",
 },
 [0x1EE9B]={
  category="lo",
  direction="al",
 },
 [0x1EEA1]={
  category="lo",
  direction="al",
 },
 [0x1EEA2]={
  category="lo",
  direction="al",
 },
 [0x1EEA3]={
  category="lo",
  direction="al",
 },
 [0x1EEA5]={
  category="lo",
  direction="al",
 },
 [0x1EEA6]={
  category="lo",
  direction="al",
 },
 [0x1EEA7]={
  category="lo",
  direction="al",
 },
 [0x1EEA8]={
  category="lo",
  direction="al",
 },
 [0x1EEA9]={
  category="lo",
  direction="al",
 },
 [0x1EEAB]={
  category="lo",
  direction="al",
 },
 [0x1EEAC]={
  category="lo",
  direction="al",
 },
 [0x1EEAD]={
  category="lo",
  direction="al",
 },
 [0x1EEAE]={
  category="lo",
  direction="al",
 },
 [0x1EEAF]={
  category="lo",
  direction="al",
 },
 [0x1EEB0]={
  category="lo",
  direction="al",
 },
 [0x1EEB1]={
  category="lo",
  direction="al",
 },
 [0x1EEB2]={
  category="lo",
  direction="al",
 },
 [0x1EEB3]={
  category="lo",
  direction="al",
 },
 [0x1EEB4]={
  category="lo",
  direction="al",
 },
 [0x1EEB5]={
  category="lo",
  direction="al",
 },
 [0x1EEB6]={
  category="lo",
  direction="al",
 },
 [0x1EEB7]={
  category="lo",
  direction="al",
 },
 [0x1EEB8]={
  category="lo",
  direction="al",
 },
 [0x1EEB9]={
  category="lo",
  direction="al",
 },
 [0x1EEBA]={
  category="lo",
  direction="al",
 },
 [0x1EEBB]={
  category="lo",
  direction="al",
 },
 [0x1EEF0]={
  category="sm",
  direction="on",
 },
 [0x1EEF1]={
  category="sm",
  direction="on",
 },
 [0x1F000]={
  category="so",
  direction="on",
 },
 [0x1F001]={
  category="so",
  direction="on",
 },
 [0x1F002]={
  category="so",
  direction="on",
 },
 [0x1F003]={
  category="so",
  direction="on",
 },
 [0x1F004]={
  category="so",
  direction="on",
 },
 [0x1F005]={
  category="so",
  direction="on",
 },
 [0x1F006]={
  category="so",
  direction="on",
 },
 [0x1F007]={
  category="so",
  direction="on",
 },
 [0x1F008]={
  category="so",
  direction="on",
 },
 [0x1F009]={
  category="so",
  direction="on",
 },
 [0x1F00A]={
  category="so",
  direction="on",
 },
 [0x1F00B]={
  category="so",
  direction="on",
 },
 [0x1F00C]={
  category="so",
  direction="on",
 },
 [0x1F00D]={
  category="so",
  direction="on",
 },
 [0x1F00E]={
  category="so",
  direction="on",
 },
 [0x1F00F]={
  category="so",
  direction="on",
 },
 [0x1F010]={
  category="so",
  direction="on",
 },
 [0x1F011]={
  category="so",
  direction="on",
 },
 [0x1F012]={
  category="so",
  direction="on",
 },
 [0x1F013]={
  category="so",
  direction="on",
 },
 [0x1F014]={
  category="so",
  direction="on",
 },
 [0x1F015]={
  category="so",
  direction="on",
 },
 [0x1F016]={
  category="so",
  direction="on",
 },
 [0x1F017]={
  category="so",
  direction="on",
 },
 [0x1F018]={
  category="so",
  direction="on",
 },
 [0x1F019]={
  category="so",
  direction="on",
 },
 [0x1F01A]={
  category="so",
  direction="on",
 },
 [0x1F01B]={
  category="so",
  direction="on",
 },
 [0x1F01C]={
  category="so",
  direction="on",
 },
 [0x1F01D]={
  category="so",
  direction="on",
 },
 [0x1F01E]={
  category="so",
  direction="on",
 },
 [0x1F01F]={
  category="so",
  direction="on",
 },
 [0x1F020]={
  category="so",
  direction="on",
 },
 [0x1F021]={
  category="so",
  direction="on",
 },
 [0x1F022]={
  category="so",
  direction="on",
 },
 [0x1F023]={
  category="so",
  direction="on",
 },
 [0x1F024]={
  category="so",
  direction="on",
 },
 [0x1F025]={
  category="so",
  direction="on",
 },
 [0x1F026]={
  category="so",
  direction="on",
 },
 [0x1F027]={
  category="so",
  direction="on",
 },
 [0x1F028]={
  category="so",
  direction="on",
 },
 [0x1F029]={
  category="so",
  direction="on",
 },
 [0x1F02A]={
  category="so",
  direction="on",
 },
 [0x1F02B]={
  category="so",
  direction="on",
 },
 [0x1F030]={
  category="so",
  direction="on",
 },
 [0x1F031]={
  category="so",
  direction="on",
 },
 [0x1F032]={
  category="so",
  direction="on",
 },
 [0x1F033]={
  category="so",
  direction="on",
 },
 [0x1F034]={
  category="so",
  direction="on",
 },
 [0x1F035]={
  category="so",
  direction="on",
 },
 [0x1F036]={
  category="so",
  direction="on",
 },
 [0x1F037]={
  category="so",
  direction="on",
 },
 [0x1F038]={
  category="so",
  direction="on",
 },
 [0x1F039]={
  category="so",
  direction="on",
 },
 [0x1F03A]={
  category="so",
  direction="on",
 },
 [0x1F03B]={
  category="so",
  direction="on",
 },
 [0x1F03C]={
  category="so",
  direction="on",
 },
 [0x1F03D]={
  category="so",
  direction="on",
 },
 [0x1F03E]={
  category="so",
  direction="on",
 },
 [0x1F03F]={
  category="so",
  direction="on",
 },
 [0x1F040]={
  category="so",
  direction="on",
 },
 [0x1F041]={
  category="so",
  direction="on",
 },
 [0x1F042]={
  category="so",
  direction="on",
 },
 [0x1F043]={
  category="so",
  direction="on",
 },
 [0x1F044]={
  category="so",
  direction="on",
 },
 [0x1F045]={
  category="so",
  direction="on",
 },
 [0x1F046]={
  category="so",
  direction="on",
 },
 [0x1F047]={
  category="so",
  direction="on",
 },
 [0x1F048]={
  category="so",
  direction="on",
 },
 [0x1F049]={
  category="so",
  direction="on",
 },
 [0x1F04A]={
  category="so",
  direction="on",
 },
 [0x1F04B]={
  category="so",
  direction="on",
 },
 [0x1F04C]={
  category="so",
  direction="on",
 },
 [0x1F04D]={
  category="so",
  direction="on",
 },
 [0x1F04E]={
  category="so",
  direction="on",
 },
 [0x1F04F]={
  category="so",
  direction="on",
 },
 [0x1F050]={
  category="so",
  direction="on",
 },
 [0x1F051]={
  category="so",
  direction="on",
 },
 [0x1F052]={
  category="so",
  direction="on",
 },
 [0x1F053]={
  category="so",
  direction="on",
 },
 [0x1F054]={
  category="so",
  direction="on",
 },
 [0x1F055]={
  category="so",
  direction="on",
 },
 [0x1F056]={
  category="so",
  direction="on",
 },
 [0x1F057]={
  category="so",
  direction="on",
 },
 [0x1F058]={
  category="so",
  direction="on",
 },
 [0x1F059]={
  category="so",
  direction="on",
 },
 [0x1F05A]={
  category="so",
  direction="on",
 },
 [0x1F05B]={
  category="so",
  direction="on",
 },
 [0x1F05C]={
  category="so",
  direction="on",
 },
 [0x1F05D]={
  category="so",
  direction="on",
 },
 [0x1F05E]={
  category="so",
  direction="on",
 },
 [0x1F05F]={
  category="so",
  direction="on",
 },
 [0x1F060]={
  category="so",
  direction="on",
 },
 [0x1F061]={
  category="so",
  direction="on",
 },
 [0x1F062]={
  category="so",
  direction="on",
 },
 [0x1F063]={
  category="so",
  direction="on",
 },
 [0x1F064]={
  category="so",
  direction="on",
 },
 [0x1F065]={
  category="so",
  direction="on",
 },
 [0x1F066]={
  category="so",
  direction="on",
 },
 [0x1F067]={
  category="so",
  direction="on",
 },
 [0x1F068]={
  category="so",
  direction="on",
 },
 [0x1F069]={
  category="so",
  direction="on",
 },
 [0x1F06A]={
  category="so",
  direction="on",
 },
 [0x1F06B]={
  category="so",
  direction="on",
 },
 [0x1F06C]={
  category="so",
  direction="on",
 },
 [0x1F06D]={
  category="so",
  direction="on",
 },
 [0x1F06E]={
  category="so",
  direction="on",
 },
 [0x1F06F]={
  category="so",
  direction="on",
 },
 [0x1F070]={
  category="so",
  direction="on",
 },
 [0x1F071]={
  category="so",
  direction="on",
 },
 [0x1F072]={
  category="so",
  direction="on",
 },
 [0x1F073]={
  category="so",
  direction="on",
 },
 [0x1F074]={
  category="so",
  direction="on",
 },
 [0x1F075]={
  category="so",
  direction="on",
 },
 [0x1F076]={
  category="so",
  direction="on",
 },
 [0x1F077]={
  category="so",
  direction="on",
 },
 [0x1F078]={
  category="so",
  direction="on",
 },
 [0x1F079]={
  category="so",
  direction="on",
 },
 [0x1F07A]={
  category="so",
  direction="on",
 },
 [0x1F07B]={
  category="so",
  direction="on",
 },
 [0x1F07C]={
  category="so",
  direction="on",
 },
 [0x1F07D]={
  category="so",
  direction="on",
 },
 [0x1F07E]={
  category="so",
  direction="on",
 },
 [0x1F07F]={
  category="so",
  direction="on",
 },
 [0x1F080]={
  category="so",
  direction="on",
 },
 [0x1F081]={
  category="so",
  direction="on",
 },
 [0x1F082]={
  category="so",
  direction="on",
 },
 [0x1F083]={
  category="so",
  direction="on",
 },
 [0x1F084]={
  category="so",
  direction="on",
 },
 [0x1F085]={
  category="so",
  direction="on",
 },
 [0x1F086]={
  category="so",
  direction="on",
 },
 [0x1F087]={
  category="so",
  direction="on",
 },
 [0x1F088]={
  category="so",
  direction="on",
 },
 [0x1F089]={
  category="so",
  direction="on",
 },
 [0x1F08A]={
  category="so",
  direction="on",
 },
 [0x1F08B]={
  category="so",
  direction="on",
 },
 [0x1F08C]={
  category="so",
  direction="on",
 },
 [0x1F08D]={
  category="so",
  direction="on",
 },
 [0x1F08E]={
  category="so",
  direction="on",
 },
 [0x1F08F]={
  category="so",
  direction="on",
 },
 [0x1F090]={
  category="so",
  direction="on",
 },
 [0x1F091]={
  category="so",
  direction="on",
 },
 [0x1F092]={
  category="so",
  direction="on",
 },
 [0x1F093]={
  category="so",
  direction="on",
 },
 [0x1F0A0]={
  category="so",
  direction="on",
 },
 [0x1F0A1]={
  category="so",
  direction="on",
 },
 [0x1F0A2]={
  category="so",
  direction="on",
 },
 [0x1F0A3]={
  category="so",
  direction="on",
 },
 [0x1F0A4]={
  category="so",
  direction="on",
 },
 [0x1F0A5]={
  category="so",
  direction="on",
 },
 [0x1F0A6]={
  category="so",
  direction="on",
 },
 [0x1F0A7]={
  category="so",
  direction="on",
 },
 [0x1F0A8]={
  category="so",
  direction="on",
 },
 [0x1F0A9]={
  category="so",
  direction="on",
 },
 [0x1F0AA]={
  category="so",
  direction="on",
 },
 [0x1F0AB]={
  category="so",
  direction="on",
 },
 [0x1F0AC]={
  category="so",
  direction="on",
 },
 [0x1F0AD]={
  category="so",
  direction="on",
 },
 [0x1F0AE]={
  category="so",
  direction="on",
 },
 [0x1F0B1]={
  category="so",
  direction="on",
 },
 [0x1F0B2]={
  category="so",
  direction="on",
 },
 [0x1F0B3]={
  category="so",
  direction="on",
 },
 [0x1F0B4]={
  category="so",
  direction="on",
 },
 [0x1F0B5]={
  category="so",
  direction="on",
 },
 [0x1F0B6]={
  category="so",
  direction="on",
 },
 [0x1F0B7]={
  category="so",
  direction="on",
 },
 [0x1F0B8]={
  category="so",
  direction="on",
 },
 [0x1F0B9]={
  category="so",
  direction="on",
 },
 [0x1F0BA]={
  category="so",
  direction="on",
 },
 [0x1F0BB]={
  category="so",
  direction="on",
 },
 [0x1F0BC]={
  category="so",
  direction="on",
 },
 [0x1F0BD]={
  category="so",
  direction="on",
 },
 [0x1F0BE]={
  category="so",
  direction="on",
 },
 [0x1F0BF]={
  category="so",
  direction="on",
 },
 [0x1F0C1]={
  category="so",
  direction="on",
 },
 [0x1F0C2]={
  category="so",
  direction="on",
 },
 [0x1F0C3]={
  category="so",
  direction="on",
 },
 [0x1F0C4]={
  category="so",
  direction="on",
 },
 [0x1F0C5]={
  category="so",
  direction="on",
 },
 [0x1F0C6]={
  category="so",
  direction="on",
 },
 [0x1F0C7]={
  category="so",
  direction="on",
 },
 [0x1F0C8]={
  category="so",
  direction="on",
 },
 [0x1F0C9]={
  category="so",
  direction="on",
 },
 [0x1F0CA]={
  category="so",
  direction="on",
 },
 [0x1F0CB]={
  category="so",
  direction="on",
 },
 [0x1F0CC]={
  category="so",
  direction="on",
 },
 [0x1F0CD]={
  category="so",
  direction="on",
 },
 [0x1F0CE]={
  category="so",
  direction="on",
 },
 [0x1F0CF]={
  category="so",
  direction="on",
 },
 [0x1F0D1]={
  category="so",
  direction="on",
 },
 [0x1F0D2]={
  category="so",
  direction="on",
 },
 [0x1F0D3]={
  category="so",
  direction="on",
 },
 [0x1F0D4]={
  category="so",
  direction="on",
 },
 [0x1F0D5]={
  category="so",
  direction="on",
 },
 [0x1F0D6]={
  category="so",
  direction="on",
 },
 [0x1F0D7]={
  category="so",
  direction="on",
 },
 [0x1F0D8]={
  category="so",
  direction="on",
 },
 [0x1F0D9]={
  category="so",
  direction="on",
 },
 [0x1F0DA]={
  category="so",
  direction="on",
 },
 [0x1F0DB]={
  category="so",
  direction="on",
 },
 [0x1F0DC]={
  category="so",
  direction="on",
 },
 [0x1F0DD]={
  category="so",
  direction="on",
 },
 [0x1F0DE]={
  category="so",
  direction="on",
 },
 [0x1F0DF]={
  category="so",
  direction="on",
 },
 [0x1F0E0]={
  category="so",
  direction="on",
 },
 [0x1F0E1]={
  category="so",
  direction="on",
 },
 [0x1F0E2]={
  category="so",
  direction="on",
 },
 [0x1F0E3]={
  category="so",
  direction="on",
 },
 [0x1F0E4]={
  category="so",
  direction="on",
 },
 [0x1F0E5]={
  category="so",
  direction="on",
 },
 [0x1F0E6]={
  category="so",
  direction="on",
 },
 [0x1F0E7]={
  category="so",
  direction="on",
 },
 [0x1F0E8]={
  category="so",
  direction="on",
 },
 [0x1F0E9]={
  category="so",
  direction="on",
 },
 [0x1F0EA]={
  category="so",
  direction="on",
 },
 [0x1F0EB]={
  category="so",
  direction="on",
 },
 [0x1F0EC]={
  category="so",
  direction="on",
 },
 [0x1F0ED]={
  category="so",
  direction="on",
 },
 [0x1F0EE]={
  category="so",
  direction="on",
 },
 [0x1F0EF]={
  category="so",
  direction="on",
 },
 [0x1F0F0]={
  category="so",
  direction="on",
 },
 [0x1F0F1]={
  category="so",
  direction="on",
 },
 [0x1F0F2]={
  category="so",
  direction="on",
 },
 [0x1F0F3]={
  category="so",
  direction="on",
 },
 [0x1F0F4]={
  category="so",
  direction="on",
 },
 [0x1F0F5]={
  category="so",
  direction="on",
 },
 [0x1F100]={
  category="no",
  direction="en",
 },
 [0x1F101]={
  category="no",
  direction="en",
 },
 [0x1F102]={
  category="no",
  direction="en",
 },
 [0x1F103]={
  category="no",
  direction="en",
 },
 [0x1F104]={
  category="no",
  direction="en",
 },
 [0x1F105]={
  category="no",
  direction="en",
 },
 [0x1F106]={
  category="no",
  direction="en",
 },
 [0x1F107]={
  category="no",
  direction="en",
 },
 [0x1F108]={
  category="no",
  direction="en",
 },
 [0x1F109]={
  category="no",
  direction="en",
 },
 [0x1F10A]={
  category="no",
  direction="en",
 },
 [0x1F10B]={
  category="no",
  direction="on",
 },
 [0x1F10C]={
  category="no",
  direction="on",
 },
 [0x1F110]={
  category="so",
  direction="l",
 },
 [0x1F111]={
  category="so",
  direction="l",
 },
 [0x1F112]={
  category="so",
  direction="l",
 },
 [0x1F113]={
  category="so",
  direction="l",
 },
 [0x1F114]={
  category="so",
  direction="l",
 },
 [0x1F115]={
  category="so",
  direction="l",
 },
 [0x1F116]={
  category="so",
  direction="l",
 },
 [0x1F117]={
  category="so",
  direction="l",
 },
 [0x1F118]={
  category="so",
  direction="l",
 },
 [0x1F119]={
  category="so",
  direction="l",
 },
 [0x1F11A]={
  category="so",
  direction="l",
 },
 [0x1F11B]={
  category="so",
  direction="l",
 },
 [0x1F11C]={
  category="so",
  direction="l",
 },
 [0x1F11D]={
  category="so",
  direction="l",
 },
 [0x1F11E]={
  category="so",
  direction="l",
 },
 [0x1F11F]={
  category="so",
  direction="l",
 },
 [0x1F120]={
  category="so",
  direction="l",
 },
 [0x1F121]={
  category="so",
  direction="l",
 },
 [0x1F122]={
  category="so",
  direction="l",
 },
 [0x1F123]={
  category="so",
  direction="l",
 },
 [0x1F124]={
  category="so",
  direction="l",
 },
 [0x1F125]={
  category="so",
  direction="l",
 },
 [0x1F126]={
  category="so",
  direction="l",
 },
 [0x1F127]={
  category="so",
  direction="l",
 },
 [0x1F128]={
  category="so",
  direction="l",
 },
 [0x1F129]={
  category="so",
  direction="l",
 },
 [0x1F12A]={
  category="so",
  direction="l",
 },
 [0x1F12B]={
  category="so",
  direction="l",
 },
 [0x1F12C]={
  category="so",
  direction="l",
 },
 [0x1F12D]={
  category="so",
  direction="l",
 },
 [0x1F12E]={
  category="so",
  direction="l",
 },
 [0x1F130]={
  category="so",
  direction="l",
 },
 [0x1F131]={
  category="so",
  direction="l",
 },
 [0x1F132]={
  category="so",
  direction="l",
 },
 [0x1F133]={
  category="so",
  direction="l",
 },
 [0x1F134]={
  category="so",
  direction="l",
 },
 [0x1F135]={
  category="so",
  direction="l",
 },
 [0x1F136]={
  category="so",
  direction="l",
 },
 [0x1F137]={
  category="so",
  direction="l",
 },
 [0x1F138]={
  category="so",
  direction="l",
 },
 [0x1F139]={
  category="so",
  direction="l",
 },
 [0x1F13A]={
  category="so",
  direction="l",
 },
 [0x1F13B]={
  category="so",
  direction="l",
 },
 [0x1F13C]={
  category="so",
  direction="l",
 },
 [0x1F13D]={
  category="so",
  direction="l",
 },
 [0x1F13E]={
  category="so",
  direction="l",
 },
 [0x1F13F]={
  category="so",
  direction="l",
 },
 [0x1F140]={
  category="so",
  direction="l",
 },
 [0x1F141]={
  category="so",
  direction="l",
 },
 [0x1F142]={
  category="so",
  direction="l",
 },
 [0x1F143]={
  category="so",
  direction="l",
 },
 [0x1F144]={
  category="so",
  direction="l",
 },
 [0x1F145]={
  category="so",
  direction="l",
 },
 [0x1F146]={
  category="so",
  direction="l",
 },
 [0x1F147]={
  category="so",
  direction="l",
 },
 [0x1F148]={
  category="so",
  direction="l",
 },
 [0x1F149]={
  category="so",
  direction="l",
 },
 [0x1F14A]={
  category="so",
  direction="l",
 },
 [0x1F14B]={
  category="so",
  direction="l",
 },
 [0x1F14C]={
  category="so",
  direction="l",
 },
 [0x1F14D]={
  category="so",
  direction="l",
 },
 [0x1F14E]={
  category="so",
  direction="l",
 },
 [0x1F14F]={
  category="so",
  direction="l",
 },
 [0x1F150]={
  category="so",
  direction="l",
 },
 [0x1F151]={
  category="so",
  direction="l",
 },
 [0x1F152]={
  category="so",
  direction="l",
 },
 [0x1F153]={
  category="so",
  direction="l",
 },
 [0x1F154]={
  category="so",
  direction="l",
 },
 [0x1F155]={
  category="so",
  direction="l",
 },
 [0x1F156]={
  category="so",
  direction="l",
 },
 [0x1F157]={
  category="so",
  direction="l",
 },
 [0x1F158]={
  category="so",
  direction="l",
 },
 [0x1F159]={
  category="so",
  direction="l",
 },
 [0x1F15A]={
  category="so",
  direction="l",
 },
 [0x1F15B]={
  category="so",
  direction="l",
 },
 [0x1F15C]={
  category="so",
  direction="l",
 },
 [0x1F15D]={
  category="so",
  direction="l",
 },
 [0x1F15E]={
  category="so",
  direction="l",
 },
 [0x1F15F]={
  category="so",
  direction="l",
 },
 [0x1F160]={
  category="so",
  direction="l",
 },
 [0x1F161]={
  category="so",
  direction="l",
 },
 [0x1F162]={
  category="so",
  direction="l",
 },
 [0x1F163]={
  category="so",
  direction="l",
 },
 [0x1F164]={
  category="so",
  direction="l",
 },
 [0x1F165]={
  category="so",
  direction="l",
 },
 [0x1F166]={
  category="so",
  direction="l",
 },
 [0x1F167]={
  category="so",
  direction="l",
 },
 [0x1F168]={
  category="so",
  direction="l",
 },
 [0x1F169]={
  category="so",
  direction="l",
 },
 [0x1F16A]={
  category="so",
  direction="on",
 },
 [0x1F16B]={
  category="so",
  direction="on",
 },
 [0x1F170]={
  category="so",
  direction="l",
 },
 [0x1F171]={
  category="so",
  direction="l",
 },
 [0x1F172]={
  category="so",
  direction="l",
 },
 [0x1F173]={
  category="so",
  direction="l",
 },
 [0x1F174]={
  category="so",
  direction="l",
 },
 [0x1F175]={
  category="so",
  direction="l",
 },
 [0x1F176]={
  category="so",
  direction="l",
 },
 [0x1F177]={
  category="so",
  direction="l",
 },
 [0x1F178]={
  category="so",
  direction="l",
 },
 [0x1F179]={
  category="so",
  direction="l",
 },
 [0x1F17A]={
  category="so",
  direction="l",
 },
 [0x1F17B]={
  category="so",
  direction="l",
 },
 [0x1F17C]={
  category="so",
  direction="l",
 },
 [0x1F17D]={
  category="so",
  direction="l",
 },
 [0x1F17E]={
  category="so",
  direction="l",
 },
 [0x1F17F]={
  category="so",
  direction="l",
 },
 [0x1F180]={
  category="so",
  direction="l",
 },
 [0x1F181]={
  category="so",
  direction="l",
 },
 [0x1F182]={
  category="so",
  direction="l",
 },
 [0x1F183]={
  category="so",
  direction="l",
 },
 [0x1F184]={
  category="so",
  direction="l",
 },
 [0x1F185]={
  category="so",
  direction="l",
 },
 [0x1F186]={
  category="so",
  direction="l",
 },
 [0x1F187]={
  category="so",
  direction="l",
 },
 [0x1F188]={
  category="so",
  direction="l",
 },
 [0x1F189]={
  category="so",
  direction="l",
 },
 [0x1F18A]={
  category="so",
  direction="l",
 },
 [0x1F18B]={
  category="so",
  direction="l",
 },
 [0x1F18C]={
  category="so",
  direction="l",
 },
 [0x1F18D]={
  category="so",
  direction="l",
 },
 [0x1F18E]={
  category="so",
  direction="l",
 },
 [0x1F18F]={
  category="so",
  direction="l",
 },
 [0x1F190]={
  category="so",
  direction="l",
 },
 [0x1F191]={
  category="so",
  direction="l",
 },
 [0x1F192]={
  category="so",
  direction="l",
 },
 [0x1F193]={
  category="so",
  direction="l",
 },
 [0x1F194]={
  category="so",
  direction="l",
 },
 [0x1F195]={
  category="so",
  direction="l",
 },
 [0x1F196]={
  category="so",
  direction="l",
 },
 [0x1F197]={
  category="so",
  direction="l",
 },
 [0x1F198]={
  category="so",
  direction="l",
 },
 [0x1F199]={
  category="so",
  direction="l",
 },
 [0x1F19A]={
  category="so",
  direction="l",
 },
 [0x1F1E6]={
  category="so",
  direction="l",
 },
 [0x1F1E7]={
  category="so",
  direction="l",
 },
 [0x1F1E8]={
  category="so",
  direction="l",
 },
 [0x1F1E9]={
  category="so",
  direction="l",
 },
 [0x1F1EA]={
  category="so",
  direction="l",
 },
 [0x1F1EB]={
  category="so",
  direction="l",
 },
 [0x1F1EC]={
  category="so",
  direction="l",
 },
 [0x1F1ED]={
  category="so",
  direction="l",
 },
 [0x1F1EE]={
  category="so",
  direction="l",
 },
 [0x1F1EF]={
  category="so",
  direction="l",
 },
 [0x1F1F0]={
  category="so",
  direction="l",
 },
 [0x1F1F1]={
  category="so",
  direction="l",
 },
 [0x1F1F2]={
  category="so",
  direction="l",
 },
 [0x1F1F3]={
  category="so",
  direction="l",
 },
 [0x1F1F4]={
  category="so",
  direction="l",
 },
 [0x1F1F5]={
  category="so",
  direction="l",
 },
 [0x1F1F6]={
  category="so",
  direction="l",
 },
 [0x1F1F7]={
  category="so",
  direction="l",
 },
 [0x1F1F8]={
  category="so",
  direction="l",
 },
 [0x1F1F9]={
  category="so",
  direction="l",
 },
 [0x1F1FA]={
  category="so",
  direction="l",
 },
 [0x1F1FB]={
  category="so",
  direction="l",
 },
 [0x1F1FC]={
  category="so",
  direction="l",
 },
 [0x1F1FD]={
  category="so",
  direction="l",
 },
 [0x1F1FE]={
  category="so",
  direction="l",
 },
 [0x1F1FF]={
  category="so",
  direction="l",
 },
 [0x1F200]={
  category="so",
  direction="l",
 },
 [0x1F201]={
  category="so",
  direction="l",
 },
 [0x1F202]={
  category="so",
  direction="l",
 },
 [0x1F210]={
  category="so",
  direction="l",
 },
 [0x1F211]={
  category="so",
  direction="l",
 },
 [0x1F212]={
  category="so",
  direction="l",
 },
 [0x1F213]={
  category="so",
  direction="l",
 },
 [0x1F214]={
  category="so",
  direction="l",
 },
 [0x1F215]={
  category="so",
  direction="l",
 },
 [0x1F216]={
  category="so",
  direction="l",
 },
 [0x1F217]={
  category="so",
  direction="l",
 },
 [0x1F218]={
  category="so",
  direction="l",
 },
 [0x1F219]={
  category="so",
  direction="l",
 },
 [0x1F21A]={
  category="so",
  direction="l",
 },
 [0x1F21B]={
  category="so",
  direction="l",
 },
 [0x1F21C]={
  category="so",
  direction="l",
 },
 [0x1F21D]={
  category="so",
  direction="l",
 },
 [0x1F21E]={
  category="so",
  direction="l",
 },
 [0x1F21F]={
  category="so",
  direction="l",
 },
 [0x1F220]={
  category="so",
  direction="l",
 },
 [0x1F221]={
  category="so",
  direction="l",
 },
 [0x1F222]={
  category="so",
  direction="l",
 },
 [0x1F223]={
  category="so",
  direction="l",
 },
 [0x1F224]={
  category="so",
  direction="l",
 },
 [0x1F225]={
  category="so",
  direction="l",
 },
 [0x1F226]={
  category="so",
  direction="l",
 },
 [0x1F227]={
  category="so",
  direction="l",
 },
 [0x1F228]={
  category="so",
  direction="l",
 },
 [0x1F229]={
  category="so",
  direction="l",
 },
 [0x1F22A]={
  category="so",
  direction="l",
 },
 [0x1F22B]={
  category="so",
  direction="l",
 },
 [0x1F22C]={
  category="so",
  direction="l",
 },
 [0x1F22D]={
  category="so",
  direction="l",
 },
 [0x1F22E]={
  category="so",
  direction="l",
 },
 [0x1F22F]={
  category="so",
  direction="l",
 },
 [0x1F230]={
  category="so",
  direction="l",
 },
 [0x1F231]={
  category="so",
  direction="l",
 },
 [0x1F232]={
  category="so",
  direction="l",
 },
 [0x1F233]={
  category="so",
  direction="l",
 },
 [0x1F234]={
  category="so",
  direction="l",
 },
 [0x1F235]={
  category="so",
  direction="l",
 },
 [0x1F236]={
  category="so",
  direction="l",
 },
 [0x1F237]={
  category="so",
  direction="l",
 },
 [0x1F238]={
  category="so",
  direction="l",
 },
 [0x1F239]={
  category="so",
  direction="l",
 },
 [0x1F23A]={
  category="so",
  direction="l",
 },
 [0x1F240]={
  category="so",
  direction="l",
 },
 [0x1F241]={
  category="so",
  direction="l",
 },
 [0x1F242]={
  category="so",
  direction="l",
 },
 [0x1F243]={
  category="so",
  direction="l",
 },
 [0x1F244]={
  category="so",
  direction="l",
 },
 [0x1F245]={
  category="so",
  direction="l",
 },
 [0x1F246]={
  category="so",
  direction="l",
 },
 [0x1F247]={
  category="so",
  direction="l",
 },
 [0x1F248]={
  category="so",
  direction="l",
 },
 [0x1F250]={
  category="so",
  direction="l",
 },
 [0x1F251]={
  category="so",
  direction="l",
 },
 [0x1F300]={
  category="so",
  direction="on",
 },
 [0x1F301]={
  category="so",
  direction="on",
 },
 [0x1F302]={
  category="so",
  direction="on",
 },
 [0x1F303]={
  category="so",
  direction="on",
 },
 [0x1F304]={
  category="so",
  direction="on",
 },
 [0x1F305]={
  category="so",
  direction="on",
 },
 [0x1F306]={
  category="so",
  direction="on",
 },
 [0x1F307]={
  category="so",
  direction="on",
 },
 [0x1F308]={
  category="so",
  direction="on",
 },
 [0x1F309]={
  category="so",
  direction="on",
 },
 [0x1F30A]={
  category="so",
  direction="on",
 },
 [0x1F30B]={
  category="so",
  direction="on",
 },
 [0x1F30C]={
  category="so",
  direction="on",
 },
 [0x1F30D]={
  category="so",
  direction="on",
 },
 [0x1F30E]={
  category="so",
  direction="on",
 },
 [0x1F30F]={
  category="so",
  direction="on",
 },
 [0x1F310]={
  category="so",
  direction="on",
 },
 [0x1F311]={
  category="so",
  direction="on",
 },
 [0x1F312]={
  category="so",
  direction="on",
 },
 [0x1F313]={
  category="so",
  direction="on",
 },
 [0x1F314]={
  category="so",
  direction="on",
 },
 [0x1F315]={
  category="so",
  direction="on",
 },
 [0x1F316]={
  category="so",
  direction="on",
 },
 [0x1F317]={
  category="so",
  direction="on",
 },
 [0x1F318]={
  category="so",
  direction="on",
 },
 [0x1F319]={
  category="so",
  direction="on",
 },
 [0x1F31A]={
  category="so",
  direction="on",
 },
 [0x1F31B]={
  category="so",
  direction="on",
 },
 [0x1F31C]={
  category="so",
  direction="on",
 },
 [0x1F31D]={
  category="so",
  direction="on",
 },
 [0x1F31E]={
  category="so",
  direction="on",
 },
 [0x1F31F]={
  category="so",
  direction="on",
 },
 [0x1F320]={
  category="so",
  direction="on",
 },
 [0x1F321]={
  category="so",
  direction="on",
 },
 [0x1F322]={
  category="so",
  direction="on",
 },
 [0x1F323]={
  category="so",
  direction="on",
 },
 [0x1F324]={
  category="so",
  direction="on",
 },
 [0x1F325]={
  category="so",
  direction="on",
 },
 [0x1F326]={
  category="so",
  direction="on",
 },
 [0x1F327]={
  category="so",
  direction="on",
 },
 [0x1F328]={
  category="so",
  direction="on",
 },
 [0x1F329]={
  category="so",
  direction="on",
 },
 [0x1F32A]={
  category="so",
  direction="on",
 },
 [0x1F32B]={
  category="so",
  direction="on",
 },
 [0x1F32C]={
  category="so",
  direction="on",
 },
 [0x1F330]={
  category="so",
  direction="on",
 },
 [0x1F331]={
  category="so",
  direction="on",
 },
 [0x1F332]={
  category="so",
  direction="on",
 },
 [0x1F333]={
  category="so",
  direction="on",
 },
 [0x1F334]={
  category="so",
  direction="on",
 },
 [0x1F335]={
  category="so",
  direction="on",
 },
 [0x1F336]={
  category="so",
  direction="on",
 },
 [0x1F337]={
  category="so",
  direction="on",
 },
 [0x1F338]={
  category="so",
  direction="on",
 },
 [0x1F339]={
  category="so",
  direction="on",
 },
 [0x1F33A]={
  category="so",
  direction="on",
 },
 [0x1F33B]={
  category="so",
  direction="on",
 },
 [0x1F33C]={
  category="so",
  direction="on",
 },
 [0x1F33D]={
  category="so",
  direction="on",
 },
 [0x1F33E]={
  category="so",
  direction="on",
 },
 [0x1F33F]={
  category="so",
  direction="on",
 },
 [0x1F340]={
  category="so",
  direction="on",
 },
 [0x1F341]={
  category="so",
  direction="on",
 },
 [0x1F342]={
  category="so",
  direction="on",
 },
 [0x1F343]={
  category="so",
  direction="on",
 },
 [0x1F344]={
  category="so",
  direction="on",
 },
 [0x1F345]={
  category="so",
  direction="on",
 },
 [0x1F346]={
  category="so",
  direction="on",
 },
 [0x1F347]={
  category="so",
  direction="on",
 },
 [0x1F348]={
  category="so",
  direction="on",
 },
 [0x1F349]={
  category="so",
  direction="on",
 },
 [0x1F34A]={
  category="so",
  direction="on",
 },
 [0x1F34B]={
  category="so",
  direction="on",
 },
 [0x1F34C]={
  category="so",
  direction="on",
 },
 [0x1F34D]={
  category="so",
  direction="on",
 },
 [0x1F34E]={
  category="so",
  direction="on",
 },
 [0x1F34F]={
  category="so",
  direction="on",
 },
 [0x1F350]={
  category="so",
  direction="on",
 },
 [0x1F351]={
  category="so",
  direction="on",
 },
 [0x1F352]={
  category="so",
  direction="on",
 },
 [0x1F353]={
  category="so",
  direction="on",
 },
 [0x1F354]={
  category="so",
  direction="on",
 },
 [0x1F355]={
  category="so",
  direction="on",
 },
 [0x1F356]={
  category="so",
  direction="on",
 },
 [0x1F357]={
  category="so",
  direction="on",
 },
 [0x1F358]={
  category="so",
  direction="on",
 },
 [0x1F359]={
  category="so",
  direction="on",
 },
 [0x1F35A]={
  category="so",
  direction="on",
 },
 [0x1F35B]={
  category="so",
  direction="on",
 },
 [0x1F35C]={
  category="so",
  direction="on",
 },
 [0x1F35D]={
  category="so",
  direction="on",
 },
 [0x1F35E]={
  category="so",
  direction="on",
 },
 [0x1F35F]={
  category="so",
  direction="on",
 },
 [0x1F360]={
  category="so",
  direction="on",
 },
 [0x1F361]={
  category="so",
  direction="on",
 },
 [0x1F362]={
  category="so",
  direction="on",
 },
 [0x1F363]={
  category="so",
  direction="on",
 },
 [0x1F364]={
  category="so",
  direction="on",
 },
 [0x1F365]={
  category="so",
  direction="on",
 },
 [0x1F366]={
  category="so",
  direction="on",
 },
 [0x1F367]={
  category="so",
  direction="on",
 },
 [0x1F368]={
  category="so",
  direction="on",
 },
 [0x1F369]={
  category="so",
  direction="on",
 },
 [0x1F36A]={
  category="so",
  direction="on",
 },
 [0x1F36B]={
  category="so",
  direction="on",
 },
 [0x1F36C]={
  category="so",
  direction="on",
 },
 [0x1F36D]={
  category="so",
  direction="on",
 },
 [0x1F36E]={
  category="so",
  direction="on",
 },
 [0x1F36F]={
  category="so",
  direction="on",
 },
 [0x1F370]={
  category="so",
  direction="on",
 },
 [0x1F371]={
  category="so",
  direction="on",
 },
 [0x1F372]={
  category="so",
  direction="on",
 },
 [0x1F373]={
  category="so",
  direction="on",
 },
 [0x1F374]={
  category="so",
  direction="on",
 },
 [0x1F375]={
  category="so",
  direction="on",
 },
 [0x1F376]={
  category="so",
  direction="on",
 },
 [0x1F377]={
  category="so",
  direction="on",
 },
 [0x1F378]={
  category="so",
  direction="on",
 },
 [0x1F379]={
  category="so",
  direction="on",
 },
 [0x1F37A]={
  category="so",
  direction="on",
 },
 [0x1F37B]={
  category="so",
  direction="on",
 },
 [0x1F37C]={
  category="so",
  direction="on",
 },
 [0x1F37D]={
  category="so",
  direction="on",
 },
 [0x1F380]={
  category="so",
  direction="on",
 },
 [0x1F381]={
  category="so",
  direction="on",
 },
 [0x1F382]={
  category="so",
  direction="on",
 },
 [0x1F383]={
  category="so",
  direction="on",
 },
 [0x1F384]={
  category="so",
  direction="on",
 },
 [0x1F385]={
  category="so",
  direction="on",
 },
 [0x1F386]={
  category="so",
  direction="on",
 },
 [0x1F387]={
  category="so",
  direction="on",
 },
 [0x1F388]={
  category="so",
  direction="on",
 },
 [0x1F389]={
  category="so",
  direction="on",
 },
 [0x1F38A]={
  category="so",
  direction="on",
 },
 [0x1F38B]={
  category="so",
  direction="on",
 },
 [0x1F38C]={
  category="so",
  direction="on",
 },
 [0x1F38D]={
  category="so",
  direction="on",
 },
 [0x1F38E]={
  category="so",
  direction="on",
 },
 [0x1F38F]={
  category="so",
  direction="on",
 },
 [0x1F390]={
  category="so",
  direction="on",
 },
 [0x1F391]={
  category="so",
  direction="on",
 },
 [0x1F392]={
  category="so",
  direction="on",
 },
 [0x1F393]={
  category="so",
  direction="on",
 },
 [0x1F394]={
  category="so",
  direction="on",
 },
 [0x1F395]={
  category="so",
  direction="on",
 },
 [0x1F396]={
  category="so",
  direction="on",
 },
 [0x1F397]={
  category="so",
  direction="on",
 },
 [0x1F398]={
  category="so",
  direction="on",
 },
 [0x1F399]={
  category="so",
  direction="on",
 },
 [0x1F39A]={
  category="so",
  direction="on",
 },
 [0x1F39B]={
  category="so",
  direction="on",
 },
 [0x1F39C]={
  category="so",
  direction="on",
 },
 [0x1F39D]={
  category="so",
  direction="on",
 },
 [0x1F39E]={
  category="so",
  direction="on",
 },
 [0x1F39F]={
  category="so",
  direction="on",
 },
 [0x1F3A0]={
  category="so",
  direction="on",
 },
 [0x1F3A1]={
  category="so",
  direction="on",
 },
 [0x1F3A2]={
  category="so",
  direction="on",
 },
 [0x1F3A3]={
  category="so",
  direction="on",
 },
 [0x1F3A4]={
  category="so",
  direction="on",
 },
 [0x1F3A5]={
  category="so",
  direction="on",
 },
 [0x1F3A6]={
  category="so",
  direction="on",
 },
 [0x1F3A7]={
  category="so",
  direction="on",
 },
 [0x1F3A8]={
  category="so",
  direction="on",
 },
 [0x1F3A9]={
  category="so",
  direction="on",
 },
 [0x1F3AA]={
  category="so",
  direction="on",
 },
 [0x1F3AB]={
  category="so",
  direction="on",
 },
 [0x1F3AC]={
  category="so",
  direction="on",
 },
 [0x1F3AD]={
  category="so",
  direction="on",
 },
 [0x1F3AE]={
  category="so",
  direction="on",
 },
 [0x1F3AF]={
  category="so",
  direction="on",
 },
 [0x1F3B0]={
  category="so",
  direction="on",
 },
 [0x1F3B1]={
  category="so",
  direction="on",
 },
 [0x1F3B2]={
  category="so",
  direction="on",
 },
 [0x1F3B3]={
  category="so",
  direction="on",
 },
 [0x1F3B4]={
  category="so",
  direction="on",
 },
 [0x1F3B5]={
  category="so",
  direction="on",
 },
 [0x1F3B6]={
  category="so",
  direction="on",
 },
 [0x1F3B7]={
  category="so",
  direction="on",
 },
 [0x1F3B8]={
  category="so",
  direction="on",
 },
 [0x1F3B9]={
  category="so",
  direction="on",
 },
 [0x1F3BA]={
  category="so",
  direction="on",
 },
 [0x1F3BB]={
  category="so",
  direction="on",
 },
 [0x1F3BC]={
  category="so",
  direction="on",
 },
 [0x1F3BD]={
  category="so",
  direction="on",
 },
 [0x1F3BE]={
  category="so",
  direction="on",
 },
 [0x1F3BF]={
  category="so",
  direction="on",
 },
 [0x1F3C0]={
  category="so",
  direction="on",
 },
 [0x1F3C1]={
  category="so",
  direction="on",
 },
 [0x1F3C2]={
  category="so",
  direction="on",
 },
 [0x1F3C3]={
  category="so",
  direction="on",
 },
 [0x1F3C4]={
  category="so",
  direction="on",
 },
 [0x1F3C5]={
  category="so",
  direction="on",
 },
 [0x1F3C6]={
  category="so",
  direction="on",
 },
 [0x1F3C7]={
  category="so",
  direction="on",
 },
 [0x1F3C8]={
  category="so",
  direction="on",
 },
 [0x1F3C9]={
  category="so",
  direction="on",
 },
 [0x1F3CA]={
  category="so",
  direction="on",
 },
 [0x1F3CB]={
  category="so",
  direction="on",
 },
 [0x1F3CC]={
  category="so",
  direction="on",
 },
 [0x1F3CD]={
  category="so",
  direction="on",
 },
 [0x1F3CE]={
  category="so",
  direction="on",
 },
 [0x1F3D4]={
  category="so",
  direction="on",
 },
 [0x1F3D5]={
  category="so",
  direction="on",
 },
 [0x1F3D6]={
  category="so",
  direction="on",
 },
 [0x1F3D7]={
  category="so",
  direction="on",
 },
 [0x1F3D8]={
  category="so",
  direction="on",
 },
 [0x1F3D9]={
  category="so",
  direction="on",
 },
 [0x1F3DA]={
  category="so",
  direction="on",
 },
 [0x1F3DB]={
  category="so",
  direction="on",
 },
 [0x1F3DC]={
  category="so",
  direction="on",
 },
 [0x1F3DD]={
  category="so",
  direction="on",
 },
 [0x1F3DE]={
  category="so",
  direction="on",
 },
 [0x1F3DF]={
  category="so",
  direction="on",
 },
 [0x1F3E0]={
  category="so",
  direction="on",
 },
 [0x1F3E1]={
  category="so",
  direction="on",
 },
 [0x1F3E2]={
  category="so",
  direction="on",
 },
 [0x1F3E3]={
  category="so",
  direction="on",
 },
 [0x1F3E4]={
  category="so",
  direction="on",
 },
 [0x1F3E5]={
  category="so",
  direction="on",
 },
 [0x1F3E6]={
  category="so",
  direction="on",
 },
 [0x1F3E7]={
  category="so",
  direction="on",
 },
 [0x1F3E8]={
  category="so",
  direction="on",
 },
 [0x1F3E9]={
  category="so",
  direction="on",
 },
 [0x1F3EA]={
  category="so",
  direction="on",
 },
 [0x1F3EB]={
  category="so",
  direction="on",
 },
 [0x1F3EC]={
  category="so",
  direction="on",
 },
 [0x1F3ED]={
  category="so",
  direction="on",
 },
 [0x1F3EE]={
  category="so",
  direction="on",
 },
 [0x1F3EF]={
  category="so",
  direction="on",
 },
 [0x1F3F0]={
  category="so",
  direction="on",
 },
 [0x1F3F1]={
  category="so",
  direction="on",
 },
 [0x1F3F2]={
  category="so",
  direction="on",
 },
 [0x1F3F3]={
  category="so",
  direction="on",
 },
 [0x1F3F4]={
  category="so",
  direction="on",
 },
 [0x1F3F5]={
  category="so",
  direction="on",
 },
 [0x1F3F6]={
  category="so",
  direction="on",
 },
 [0x1F3F7]={
  category="so",
  direction="on",
 },
 [0x1F400]={
  category="so",
  direction="on",
 },
 [0x1F401]={
  category="so",
  direction="on",
 },
 [0x1F402]={
  category="so",
  direction="on",
 },
 [0x1F403]={
  category="so",
  direction="on",
 },
 [0x1F404]={
  category="so",
  direction="on",
 },
 [0x1F405]={
  category="so",
  direction="on",
 },
 [0x1F406]={
  category="so",
  direction="on",
 },
 [0x1F407]={
  category="so",
  direction="on",
 },
 [0x1F408]={
  category="so",
  direction="on",
 },
 [0x1F409]={
  category="so",
  direction="on",
 },
 [0x1F40A]={
  category="so",
  direction="on",
 },
 [0x1F40B]={
  category="so",
  direction="on",
 },
 [0x1F40C]={
  category="so",
  direction="on",
 },
 [0x1F40D]={
  category="so",
  direction="on",
 },
 [0x1F40E]={
  category="so",
  direction="on",
 },
 [0x1F40F]={
  category="so",
  direction="on",
 },
 [0x1F410]={
  category="so",
  direction="on",
 },
 [0x1F411]={
  category="so",
  direction="on",
 },
 [0x1F412]={
  category="so",
  direction="on",
 },
 [0x1F413]={
  category="so",
  direction="on",
 },
 [0x1F414]={
  category="so",
  direction="on",
 },
 [0x1F415]={
  category="so",
  direction="on",
 },
 [0x1F416]={
  category="so",
  direction="on",
 },
 [0x1F417]={
  category="so",
  direction="on",
 },
 [0x1F418]={
  category="so",
  direction="on",
 },
 [0x1F419]={
  category="so",
  direction="on",
 },
 [0x1F41A]={
  category="so",
  direction="on",
 },
 [0x1F41B]={
  category="so",
  direction="on",
 },
 [0x1F41C]={
  category="so",
  direction="on",
 },
 [0x1F41D]={
  category="so",
  direction="on",
 },
 [0x1F41E]={
  category="so",
  direction="on",
 },
 [0x1F41F]={
  category="so",
  direction="on",
 },
 [0x1F420]={
  category="so",
  direction="on",
 },
 [0x1F421]={
  category="so",
  direction="on",
 },
 [0x1F422]={
  category="so",
  direction="on",
 },
 [0x1F423]={
  category="so",
  direction="on",
 },
 [0x1F424]={
  category="so",
  direction="on",
 },
 [0x1F425]={
  category="so",
  direction="on",
 },
 [0x1F426]={
  category="so",
  direction="on",
 },
 [0x1F427]={
  category="so",
  direction="on",
 },
 [0x1F428]={
  category="so",
  direction="on",
 },
 [0x1F429]={
  category="so",
  direction="on",
 },
 [0x1F42A]={
  category="so",
  direction="on",
 },
 [0x1F42B]={
  category="so",
  direction="on",
 },
 [0x1F42C]={
  category="so",
  direction="on",
 },
 [0x1F42D]={
  category="so",
  direction="on",
 },
 [0x1F42E]={
  category="so",
  direction="on",
 },
 [0x1F42F]={
  category="so",
  direction="on",
 },
 [0x1F430]={
  category="so",
  direction="on",
 },
 [0x1F431]={
  category="so",
  direction="on",
 },
 [0x1F432]={
  category="so",
  direction="on",
 },
 [0x1F433]={
  category="so",
  direction="on",
 },
 [0x1F434]={
  category="so",
  direction="on",
 },
 [0x1F435]={
  category="so",
  direction="on",
 },
 [0x1F436]={
  category="so",
  direction="on",
 },
 [0x1F437]={
  category="so",
  direction="on",
 },
 [0x1F438]={
  category="so",
  direction="on",
 },
 [0x1F439]={
  category="so",
  direction="on",
 },
 [0x1F43A]={
  category="so",
  direction="on",
 },
 [0x1F43B]={
  category="so",
  direction="on",
 },
 [0x1F43C]={
  category="so",
  direction="on",
 },
 [0x1F43D]={
  category="so",
  direction="on",
 },
 [0x1F43E]={
  category="so",
  direction="on",
 },
 [0x1F43F]={
  category="so",
  direction="on",
 },
 [0x1F440]={
  category="so",
  direction="on",
 },
 [0x1F441]={
  category="so",
  direction="on",
 },
 [0x1F442]={
  category="so",
  direction="on",
 },
 [0x1F443]={
  category="so",
  direction="on",
 },
 [0x1F444]={
  category="so",
  direction="on",
 },
 [0x1F445]={
  category="so",
  direction="on",
 },
 [0x1F446]={
  category="so",
  direction="on",
 },
 [0x1F447]={
  category="so",
  direction="on",
 },
 [0x1F448]={
  category="so",
  direction="on",
 },
 [0x1F449]={
  category="so",
  direction="on",
 },
 [0x1F44A]={
  category="so",
  direction="on",
 },
 [0x1F44B]={
  category="so",
  direction="on",
 },
 [0x1F44C]={
  category="so",
  direction="on",
 },
 [0x1F44D]={
  category="so",
  direction="on",
 },
 [0x1F44E]={
  category="so",
  direction="on",
 },
 [0x1F44F]={
  category="so",
  direction="on",
 },
 [0x1F450]={
  category="so",
  direction="on",
 },
 [0x1F451]={
  category="so",
  direction="on",
 },
 [0x1F452]={
  category="so",
  direction="on",
 },
 [0x1F453]={
  category="so",
  direction="on",
 },
 [0x1F454]={
  category="so",
  direction="on",
 },
 [0x1F455]={
  category="so",
  direction="on",
 },
 [0x1F456]={
  category="so",
  direction="on",
 },
 [0x1F457]={
  category="so",
  direction="on",
 },
 [0x1F458]={
  category="so",
  direction="on",
 },
 [0x1F459]={
  category="so",
  direction="on",
 },
 [0x1F45A]={
  category="so",
  direction="on",
 },
 [0x1F45B]={
  category="so",
  direction="on",
 },
 [0x1F45C]={
  category="so",
  direction="on",
 },
 [0x1F45D]={
  category="so",
  direction="on",
 },
 [0x1F45E]={
  category="so",
  direction="on",
 },
 [0x1F45F]={
  category="so",
  direction="on",
 },
 [0x1F460]={
  category="so",
  direction="on",
 },
 [0x1F461]={
  category="so",
  direction="on",
 },
 [0x1F462]={
  category="so",
  direction="on",
 },
 [0x1F463]={
  category="so",
  direction="on",
 },
 [0x1F464]={
  category="so",
  direction="on",
 },
 [0x1F465]={
  category="so",
  direction="on",
 },
 [0x1F466]={
  category="so",
  direction="on",
 },
 [0x1F467]={
  category="so",
  direction="on",
 },
 [0x1F468]={
  category="so",
  direction="on",
 },
 [0x1F469]={
  category="so",
  direction="on",
 },
 [0x1F46A]={
  category="so",
  direction="on",
 },
 [0x1F46B]={
  category="so",
  direction="on",
 },
 [0x1F46C]={
  category="so",
  direction="on",
 },
 [0x1F46D]={
  category="so",
  direction="on",
 },
 [0x1F46E]={
  category="so",
  direction="on",
 },
 [0x1F46F]={
  category="so",
  direction="on",
 },
 [0x1F470]={
  category="so",
  direction="on",
 },
 [0x1F471]={
  category="so",
  direction="on",
 },
 [0x1F472]={
  category="so",
  direction="on",
 },
 [0x1F473]={
  category="so",
  direction="on",
 },
 [0x1F474]={
  category="so",
  direction="on",
 },
 [0x1F475]={
  category="so",
  direction="on",
 },
 [0x1F476]={
  category="so",
  direction="on",
 },
 [0x1F477]={
  category="so",
  direction="on",
 },
 [0x1F478]={
  category="so",
  direction="on",
 },
 [0x1F479]={
  category="so",
  direction="on",
 },
 [0x1F47A]={
  category="so",
  direction="on",
 },
 [0x1F47B]={
  category="so",
  direction="on",
 },
 [0x1F47C]={
  category="so",
  direction="on",
 },
 [0x1F47D]={
  category="so",
  direction="on",
 },
 [0x1F47E]={
  category="so",
  direction="on",
 },
 [0x1F47F]={
  category="so",
  direction="on",
 },
 [0x1F480]={
  category="so",
  direction="on",
 },
 [0x1F481]={
  category="so",
  direction="on",
 },
 [0x1F482]={
  category="so",
  direction="on",
 },
 [0x1F483]={
  category="so",
  direction="on",
 },
 [0x1F484]={
  category="so",
  direction="on",
 },
 [0x1F485]={
  category="so",
  direction="on",
 },
 [0x1F486]={
  category="so",
  direction="on",
 },
 [0x1F487]={
  category="so",
  direction="on",
 },
 [0x1F488]={
  category="so",
  direction="on",
 },
 [0x1F489]={
  category="so",
  direction="on",
 },
 [0x1F48A]={
  category="so",
  direction="on",
 },
 [0x1F48B]={
  category="so",
  direction="on",
 },
 [0x1F48C]={
  category="so",
  direction="on",
 },
 [0x1F48D]={
  category="so",
  direction="on",
 },
 [0x1F48E]={
  category="so",
  direction="on",
 },
 [0x1F48F]={
  category="so",
  direction="on",
 },
 [0x1F490]={
  category="so",
  direction="on",
 },
 [0x1F491]={
  category="so",
  direction="on",
 },
 [0x1F492]={
  category="so",
  direction="on",
 },
 [0x1F493]={
  category="so",
  direction="on",
 },
 [0x1F494]={
  category="so",
  direction="on",
 },
 [0x1F495]={
  category="so",
  direction="on",
 },
 [0x1F496]={
  category="so",
  direction="on",
 },
 [0x1F497]={
  category="so",
  direction="on",
 },
 [0x1F498]={
  category="so",
  direction="on",
 },
 [0x1F499]={
  category="so",
  direction="on",
 },
 [0x1F49A]={
  category="so",
  direction="on",
 },
 [0x1F49B]={
  category="so",
  direction="on",
 },
 [0x1F49C]={
  category="so",
  direction="on",
 },
 [0x1F49D]={
  category="so",
  direction="on",
 },
 [0x1F49E]={
  category="so",
  direction="on",
 },
 [0x1F49F]={
  category="so",
  direction="on",
 },
 [0x1F4A0]={
  category="so",
  direction="on",
 },
 [0x1F4A1]={
  category="so",
  direction="on",
 },
 [0x1F4A2]={
  category="so",
  direction="on",
 },
 [0x1F4A3]={
  category="so",
  direction="on",
 },
 [0x1F4A4]={
  category="so",
  direction="on",
 },
 [0x1F4A5]={
  category="so",
  direction="on",
 },
 [0x1F4A6]={
  category="so",
  direction="on",
 },
 [0x1F4A7]={
  category="so",
  direction="on",
 },
 [0x1F4A8]={
  category="so",
  direction="on",
 },
 [0x1F4A9]={
  category="so",
  direction="on",
 },
 [0x1F4AA]={
  category="so",
  direction="on",
 },
 [0x1F4AB]={
  category="so",
  direction="on",
 },
 [0x1F4AC]={
  category="so",
  direction="on",
 },
 [0x1F4AD]={
  category="so",
  direction="on",
 },
 [0x1F4AE]={
  category="so",
  direction="on",
 },
 [0x1F4AF]={
  category="so",
  direction="on",
 },
 [0x1F4B0]={
  category="so",
  direction="on",
 },
 [0x1F4B1]={
  category="so",
  direction="on",
 },
 [0x1F4B2]={
  category="so",
  direction="on",
 },
 [0x1F4B3]={
  category="so",
  direction="on",
 },
 [0x1F4B4]={
  category="so",
  direction="on",
 },
 [0x1F4B5]={
  category="so",
  direction="on",
 },
 [0x1F4B6]={
  category="so",
  direction="on",
 },
 [0x1F4B7]={
  category="so",
  direction="on",
 },
 [0x1F4B8]={
  category="so",
  direction="on",
 },
 [0x1F4B9]={
  category="so",
  direction="on",
 },
 [0x1F4BA]={
  category="so",
  direction="on",
 },
 [0x1F4BB]={
  category="so",
  direction="on",
 },
 [0x1F4BC]={
  category="so",
  direction="on",
 },
 [0x1F4BD]={
  category="so",
  direction="on",
 },
 [0x1F4BE]={
  category="so",
  direction="on",
 },
 [0x1F4BF]={
  category="so",
  direction="on",
 },
 [0x1F4C0]={
  category="so",
  direction="on",
 },
 [0x1F4C1]={
  category="so",
  direction="on",
 },
 [0x1F4C2]={
  category="so",
  direction="on",
 },
 [0x1F4C3]={
  category="so",
  direction="on",
 },
 [0x1F4C4]={
  category="so",
  direction="on",
 },
 [0x1F4C5]={
  category="so",
  direction="on",
 },
 [0x1F4C6]={
  category="so",
  direction="on",
 },
 [0x1F4C7]={
  category="so",
  direction="on",
 },
 [0x1F4C8]={
  category="so",
  direction="on",
 },
 [0x1F4C9]={
  category="so",
  direction="on",
 },
 [0x1F4CA]={
  category="so",
  direction="on",
 },
 [0x1F4CB]={
  category="so",
  direction="on",
 },
 [0x1F4CC]={
  category="so",
  direction="on",
 },
 [0x1F4CD]={
  category="so",
  direction="on",
 },
 [0x1F4CE]={
  category="so",
  direction="on",
 },
 [0x1F4CF]={
  category="so",
  direction="on",
 },
 [0x1F4D0]={
  category="so",
  direction="on",
 },
 [0x1F4D1]={
  category="so",
  direction="on",
 },
 [0x1F4D2]={
  category="so",
  direction="on",
 },
 [0x1F4D3]={
  category="so",
  direction="on",
 },
 [0x1F4D4]={
  category="so",
  direction="on",
 },
 [0x1F4D5]={
  category="so",
  direction="on",
 },
 [0x1F4D6]={
  category="so",
  direction="on",
 },
 [0x1F4D7]={
  category="so",
  direction="on",
 },
 [0x1F4D8]={
  category="so",
  direction="on",
 },
 [0x1F4D9]={
  category="so",
  direction="on",
 },
 [0x1F4DA]={
  category="so",
  direction="on",
 },
 [0x1F4DB]={
  category="so",
  direction="on",
 },
 [0x1F4DC]={
  category="so",
  direction="on",
 },
 [0x1F4DD]={
  category="so",
  direction="on",
 },
 [0x1F4DE]={
  category="so",
  direction="on",
 },
 [0x1F4DF]={
  category="so",
  direction="on",
 },
 [0x1F4E0]={
  category="so",
  direction="on",
 },
 [0x1F4E1]={
  category="so",
  direction="on",
 },
 [0x1F4E2]={
  category="so",
  direction="on",
 },
 [0x1F4E3]={
  category="so",
  direction="on",
 },
 [0x1F4E4]={
  category="so",
  direction="on",
 },
 [0x1F4E5]={
  category="so",
  direction="on",
 },
 [0x1F4E6]={
  category="so",
  direction="on",
 },
 [0x1F4E7]={
  category="so",
  direction="on",
 },
 [0x1F4E8]={
  category="so",
  direction="on",
 },
 [0x1F4E9]={
  category="so",
  direction="on",
 },
 [0x1F4EA]={
  category="so",
  direction="on",
 },
 [0x1F4EB]={
  category="so",
  direction="on",
 },
 [0x1F4EC]={
  category="so",
  direction="on",
 },
 [0x1F4ED]={
  category="so",
  direction="on",
 },
 [0x1F4EE]={
  category="so",
  direction="on",
 },
 [0x1F4EF]={
  category="so",
  direction="on",
 },
 [0x1F4F0]={
  category="so",
  direction="on",
 },
 [0x1F4F1]={
  category="so",
  direction="on",
 },
 [0x1F4F2]={
  category="so",
  direction="on",
 },
 [0x1F4F3]={
  category="so",
  direction="on",
 },
 [0x1F4F4]={
  category="so",
  direction="on",
 },
 [0x1F4F5]={
  category="so",
  direction="on",
 },
 [0x1F4F6]={
  category="so",
  direction="on",
 },
 [0x1F4F7]={
  category="so",
  direction="on",
 },
 [0x1F4F8]={
  category="so",
  direction="on",
 },
 [0x1F4F9]={
  category="so",
  direction="on",
 },
 [0x1F4FA]={
  category="so",
  direction="on",
 },
 [0x1F4FB]={
  category="so",
  direction="on",
 },
 [0x1F4FC]={
  category="so",
  direction="on",
 },
 [0x1F4FD]={
  category="so",
  direction="on",
 },
 [0x1F4FE]={
  category="so",
  direction="on",
 },
 [0x1F500]={
  category="so",
  direction="on",
 },
 [0x1F501]={
  category="so",
  direction="on",
 },
 [0x1F502]={
  category="so",
  direction="on",
 },
 [0x1F503]={
  category="so",
  direction="on",
 },
 [0x1F504]={
  category="so",
  direction="on",
 },
 [0x1F505]={
  category="so",
  direction="on",
 },
 [0x1F506]={
  category="so",
  direction="on",
 },
 [0x1F507]={
  category="so",
  direction="on",
 },
 [0x1F508]={
  category="so",
  direction="on",
 },
 [0x1F509]={
  category="so",
  direction="on",
 },
 [0x1F50A]={
  category="so",
  direction="on",
 },
 [0x1F50B]={
  category="so",
  direction="on",
 },
 [0x1F50C]={
  category="so",
  direction="on",
 },
 [0x1F50D]={
  category="so",
  direction="on",
 },
 [0x1F50E]={
  category="so",
  direction="on",
 },
 [0x1F50F]={
  category="so",
  direction="on",
 },
 [0x1F510]={
  category="so",
  direction="on",
 },
 [0x1F511]={
  category="so",
  direction="on",
 },
 [0x1F512]={
  category="so",
  direction="on",
 },
 [0x1F513]={
  category="so",
  direction="on",
 },
 [0x1F514]={
  category="so",
  direction="on",
 },
 [0x1F515]={
  category="so",
  direction="on",
 },
 [0x1F516]={
  category="so",
  direction="on",
 },
 [0x1F517]={
  category="so",
  direction="on",
 },
 [0x1F518]={
  category="so",
  direction="on",
 },
 [0x1F519]={
  category="so",
  direction="on",
 },
 [0x1F51A]={
  category="so",
  direction="on",
 },
 [0x1F51B]={
  category="so",
  direction="on",
 },
 [0x1F51C]={
  category="so",
  direction="on",
 },
 [0x1F51D]={
  category="so",
  direction="on",
 },
 [0x1F51E]={
  category="so",
  direction="on",
 },
 [0x1F51F]={
  category="so",
  direction="on",
 },
 [0x1F520]={
  category="so",
  direction="on",
 },
 [0x1F521]={
  category="so",
  direction="on",
 },
 [0x1F522]={
  category="so",
  direction="on",
 },
 [0x1F523]={
  category="so",
  direction="on",
 },
 [0x1F524]={
  category="so",
  direction="on",
 },
 [0x1F525]={
  category="so",
  direction="on",
 },
 [0x1F526]={
  category="so",
  direction="on",
 },
 [0x1F527]={
  category="so",
  direction="on",
 },
 [0x1F528]={
  category="so",
  direction="on",
 },
 [0x1F529]={
  category="so",
  direction="on",
 },
 [0x1F52A]={
  category="so",
  direction="on",
 },
 [0x1F52B]={
  category="so",
  direction="on",
 },
 [0x1F52C]={
  category="so",
  direction="on",
 },
 [0x1F52D]={
  category="so",
  direction="on",
 },
 [0x1F52E]={
  category="so",
  direction="on",
 },
 [0x1F52F]={
  category="so",
  direction="on",
 },
 [0x1F530]={
  category="so",
  direction="on",
 },
 [0x1F531]={
  category="so",
  direction="on",
 },
 [0x1F532]={
  category="so",
  direction="on",
 },
 [0x1F533]={
  category="so",
  direction="on",
 },
 [0x1F534]={
  category="so",
  direction="on",
 },
 [0x1F535]={
  category="so",
  direction="on",
 },
 [0x1F536]={
  category="so",
  direction="on",
 },
 [0x1F537]={
  category="so",
  direction="on",
 },
 [0x1F538]={
  category="so",
  direction="on",
 },
 [0x1F539]={
  category="so",
  direction="on",
 },
 [0x1F53A]={
  category="so",
  direction="on",
 },
 [0x1F53B]={
  category="so",
  direction="on",
 },
 [0x1F53C]={
  category="so",
  direction="on",
 },
 [0x1F53D]={
  category="so",
  direction="on",
 },
 [0x1F53E]={
  category="so",
  direction="on",
 },
 [0x1F53F]={
  category="so",
  direction="on",
 },
 [0x1F540]={
  category="so",
  direction="on",
 },
 [0x1F541]={
  category="so",
  direction="on",
 },
 [0x1F542]={
  category="so",
  direction="on",
 },
 [0x1F543]={
  category="so",
  direction="on",
 },
 [0x1F544]={
  category="so",
  direction="on",
 },
 [0x1F545]={
  category="so",
  direction="on",
 },
 [0x1F546]={
  category="so",
  direction="on",
 },
 [0x1F547]={
  category="so",
  direction="on",
 },
 [0x1F548]={
  category="so",
  direction="on",
 },
 [0x1F549]={
  category="so",
  direction="on",
 },
 [0x1F54A]={
  category="so",
  direction="on",
 },
 [0x1F550]={
  category="so",
  direction="on",
 },
 [0x1F551]={
  category="so",
  direction="on",
 },
 [0x1F552]={
  category="so",
  direction="on",
 },
 [0x1F553]={
  category="so",
  direction="on",
 },
 [0x1F554]={
  category="so",
  direction="on",
 },
 [0x1F555]={
  category="so",
  direction="on",
 },
 [0x1F556]={
  category="so",
  direction="on",
 },
 [0x1F557]={
  category="so",
  direction="on",
 },
 [0x1F558]={
  category="so",
  direction="on",
 },
 [0x1F559]={
  category="so",
  direction="on",
 },
 [0x1F55A]={
  category="so",
  direction="on",
 },
 [0x1F55B]={
  category="so",
  direction="on",
 },
 [0x1F55C]={
  category="so",
  direction="on",
 },
 [0x1F55D]={
  category="so",
  direction="on",
 },
 [0x1F55E]={
  category="so",
  direction="on",
 },
 [0x1F55F]={
  category="so",
  direction="on",
 },
 [0x1F560]={
  category="so",
  direction="on",
 },
 [0x1F561]={
  category="so",
  direction="on",
 },
 [0x1F562]={
  category="so",
  direction="on",
 },
 [0x1F563]={
  category="so",
  direction="on",
 },
 [0x1F564]={
  category="so",
  direction="on",
 },
 [0x1F565]={
  category="so",
  direction="on",
 },
 [0x1F566]={
  category="so",
  direction="on",
 },
 [0x1F567]={
  category="so",
  direction="on",
 },
 [0x1F568]={
  category="so",
  direction="on",
 },
 [0x1F569]={
  category="so",
  direction="on",
 },
 [0x1F56A]={
  category="so",
  direction="on",
 },
 [0x1F56B]={
  category="so",
  direction="on",
 },
 [0x1F56C]={
  category="so",
  direction="on",
 },
 [0x1F56D]={
  category="so",
  direction="on",
 },
 [0x1F56E]={
  category="so",
  direction="on",
 },
 [0x1F56F]={
  category="so",
  direction="on",
 },
 [0x1F570]={
  category="so",
  direction="on",
 },
 [0x1F571]={
  category="so",
  direction="on",
 },
 [0x1F572]={
  category="so",
  direction="on",
 },
 [0x1F573]={
  category="so",
  direction="on",
 },
 [0x1F574]={
  category="so",
  direction="on",
 },
 [0x1F575]={
  category="so",
  direction="on",
 },
 [0x1F576]={
  category="so",
  direction="on",
 },
 [0x1F577]={
  category="so",
  direction="on",
 },
 [0x1F578]={
  category="so",
  direction="on",
 },
 [0x1F579]={
  category="so",
  direction="on",
 },
 [0x1F57B]={
  category="so",
  direction="on",
 },
 [0x1F57C]={
  category="so",
  direction="on",
 },
 [0x1F57D]={
  category="so",
  direction="on",
 },
 [0x1F57E]={
  category="so",
  direction="on",
 },
 [0x1F57F]={
  category="so",
  direction="on",
 },
 [0x1F580]={
  category="so",
  direction="on",
 },
 [0x1F581]={
  category="so",
  direction="on",
 },
 [0x1F582]={
  category="so",
  direction="on",
 },
 [0x1F583]={
  category="so",
  direction="on",
 },
 [0x1F584]={
  category="so",
  direction="on",
 },
 [0x1F585]={
  category="so",
  direction="on",
 },
 [0x1F586]={
  category="so",
  direction="on",
 },
 [0x1F587]={
  category="so",
  direction="on",
 },
 [0x1F588]={
  category="so",
  direction="on",
 },
 [0x1F589]={
  category="so",
  direction="on",
 },
 [0x1F58A]={
  category="so",
  direction="on",
 },
 [0x1F58B]={
  category="so",
  direction="on",
 },
 [0x1F58C]={
  category="so",
  direction="on",
 },
 [0x1F58D]={
  category="so",
  direction="on",
 },
 [0x1F58E]={
  category="so",
  direction="on",
 },
 [0x1F58F]={
  category="so",
  direction="on",
 },
 [0x1F590]={
  category="so",
  direction="on",
 },
 [0x1F591]={
  category="so",
  direction="on",
 },
 [0x1F592]={
  category="so",
  direction="on",
 },
 [0x1F593]={
  category="so",
  direction="on",
 },
 [0x1F594]={
  category="so",
  direction="on",
 },
 [0x1F595]={
  category="so",
  direction="on",
 },
 [0x1F596]={
  category="so",
  direction="on",
 },
 [0x1F597]={
  category="so",
  direction="on",
 },
 [0x1F598]={
  category="so",
  direction="on",
 },
 [0x1F599]={
  category="so",
  direction="on",
 },
 [0x1F59A]={
  category="so",
  direction="on",
 },
 [0x1F59B]={
  category="so",
  direction="on",
 },
 [0x1F59C]={
  category="so",
  direction="on",
 },
 [0x1F59D]={
  category="so",
  direction="on",
 },
 [0x1F59E]={
  category="so",
  direction="on",
 },
 [0x1F59F]={
  category="so",
  direction="on",
 },
 [0x1F5A0]={
  category="so",
  direction="on",
 },
 [0x1F5A1]={
  category="so",
  direction="on",
 },
 [0x1F5A2]={
  category="so",
  direction="on",
 },
 [0x1F5A3]={
  category="so",
  direction="on",
 },
 [0x1F5A5]={
  category="so",
  direction="on",
 },
 [0x1F5A6]={
  category="so",
  direction="on",
 },
 [0x1F5A7]={
  category="so",
  direction="on",
 },
 [0x1F5A8]={
  category="so",
  direction="on",
 },
 [0x1F5A9]={
  category="so",
  direction="on",
 },
 [0x1F5AA]={
  category="so",
  direction="on",
 },
 [0x1F5AB]={
  category="so",
  direction="on",
 },
 [0x1F5AC]={
  category="so",
  direction="on",
 },
 [0x1F5AD]={
  category="so",
  direction="on",
 },
 [0x1F5AE]={
  category="so",
  direction="on",
 },
 [0x1F5AF]={
  category="so",
  direction="on",
 },
 [0x1F5B0]={
  category="so",
  direction="on",
 },
 [0x1F5B1]={
  category="so",
  direction="on",
 },
 [0x1F5B2]={
  category="so",
  direction="on",
 },
 [0x1F5B3]={
  category="so",
  direction="on",
 },
 [0x1F5B4]={
  category="so",
  direction="on",
 },
 [0x1F5B5]={
  category="so",
  direction="on",
 },
 [0x1F5B6]={
  category="so",
  direction="on",
 },
 [0x1F5B7]={
  category="so",
  direction="on",
 },
 [0x1F5B8]={
  category="so",
  direction="on",
 },
 [0x1F5B9]={
  category="so",
  direction="on",
 },
 [0x1F5BA]={
  category="so",
  direction="on",
 },
 [0x1F5BB]={
  category="so",
  direction="on",
 },
 [0x1F5BC]={
  category="so",
  direction="on",
 },
 [0x1F5BD]={
  category="so",
  direction="on",
 },
 [0x1F5BE]={
  category="so",
  direction="on",
 },
 [0x1F5BF]={
  category="so",
  direction="on",
 },
 [0x1F5C0]={
  category="so",
  direction="on",
 },
 [0x1F5C1]={
  category="so",
  direction="on",
 },
 [0x1F5C2]={
  category="so",
  direction="on",
 },
 [0x1F5C3]={
  category="so",
  direction="on",
 },
 [0x1F5C4]={
  category="so",
  direction="on",
 },
 [0x1F5C5]={
  category="so",
  direction="on",
 },
 [0x1F5C6]={
  category="so",
  direction="on",
 },
 [0x1F5C7]={
  category="so",
  direction="on",
 },
 [0x1F5C8]={
  category="so",
  direction="on",
 },
 [0x1F5C9]={
  category="so",
  direction="on",
 },
 [0x1F5CA]={
  category="so",
  direction="on",
 },
 [0x1F5CB]={
  category="so",
  direction="on",
 },
 [0x1F5CC]={
  category="so",
  direction="on",
 },
 [0x1F5CD]={
  category="so",
  direction="on",
 },
 [0x1F5CE]={
  category="so",
  direction="on",
 },
 [0x1F5CF]={
  category="so",
  direction="on",
 },
 [0x1F5D0]={
  category="so",
  direction="on",
 },
 [0x1F5D1]={
  category="so",
  direction="on",
 },
 [0x1F5D2]={
  category="so",
  direction="on",
 },
 [0x1F5D3]={
  category="so",
  direction="on",
 },
 [0x1F5D4]={
  category="so",
  direction="on",
 },
 [0x1F5D5]={
  category="so",
  direction="on",
 },
 [0x1F5D6]={
  category="so",
  direction="on",
 },
 [0x1F5D7]={
  category="so",
  direction="on",
 },
 [0x1F5D8]={
  category="so",
  direction="on",
 },
 [0x1F5D9]={
  category="so",
  direction="on",
 },
 [0x1F5DA]={
  category="so",
  direction="on",
 },
 [0x1F5DB]={
  category="so",
  direction="on",
 },
 [0x1F5DC]={
  category="so",
  direction="on",
 },
 [0x1F5DD]={
  category="so",
  direction="on",
 },
 [0x1F5DE]={
  category="so",
  direction="on",
 },
 [0x1F5DF]={
  category="so",
  direction="on",
 },
 [0x1F5E0]={
  category="so",
  direction="on",
 },
 [0x1F5E1]={
  category="so",
  direction="on",
 },
 [0x1F5E2]={
  category="so",
  direction="on",
 },
 [0x1F5E3]={
  category="so",
  direction="on",
 },
 [0x1F5E4]={
  category="so",
  direction="on",
 },
 [0x1F5E5]={
  category="so",
  direction="on",
 },
 [0x1F5E6]={
  category="so",
  direction="on",
 },
 [0x1F5E7]={
  category="so",
  direction="on",
 },
 [0x1F5E8]={
  category="so",
  direction="on",
 },
 [0x1F5E9]={
  category="so",
  direction="on",
 },
 [0x1F5EA]={
  category="so",
  direction="on",
 },
 [0x1F5EB]={
  category="so",
  direction="on",
 },
 [0x1F5EC]={
  category="so",
  direction="on",
 },
 [0x1F5ED]={
  category="so",
  direction="on",
 },
 [0x1F5EE]={
  category="so",
  direction="on",
 },
 [0x1F5EF]={
  category="so",
  direction="on",
 },
 [0x1F5F0]={
  category="so",
  direction="on",
 },
 [0x1F5F1]={
  category="so",
  direction="on",
 },
 [0x1F5F2]={
  category="so",
  direction="on",
 },
 [0x1F5F3]={
  category="so",
  direction="on",
 },
 [0x1F5F4]={
  category="so",
  direction="on",
 },
 [0x1F5F5]={
  category="so",
  direction="on",
 },
 [0x1F5F6]={
  category="so",
  direction="on",
 },
 [0x1F5F7]={
  category="so",
  direction="on",
 },
 [0x1F5F8]={
  category="so",
  direction="on",
 },
 [0x1F5F9]={
  category="so",
  direction="on",
 },
 [0x1F5FA]={
  category="so",
  direction="on",
 },
 [0x1F5FB]={
  category="so",
  direction="on",
 },
 [0x1F5FC]={
  category="so",
  direction="on",
 },
 [0x1F5FD]={
  category="so",
  direction="on",
 },
 [0x1F5FE]={
  category="so",
  direction="on",
 },
 [0x1F5FF]={
  category="so",
  direction="on",
 },
 [0x1F600]={
  category="so",
  direction="on",
 },
 [0x1F601]={
  category="so",
  direction="on",
 },
 [0x1F602]={
  category="so",
  direction="on",
 },
 [0x1F603]={
  category="so",
  direction="on",
 },
 [0x1F604]={
  category="so",
  direction="on",
 },
 [0x1F605]={
  category="so",
  direction="on",
 },
 [0x1F606]={
  category="so",
  direction="on",
 },
 [0x1F607]={
  category="so",
  direction="on",
 },
 [0x1F608]={
  category="so",
  direction="on",
 },
 [0x1F609]={
  category="so",
  direction="on",
 },
 [0x1F60A]={
  category="so",
  direction="on",
 },
 [0x1F60B]={
  category="so",
  direction="on",
 },
 [0x1F60C]={
  category="so",
  direction="on",
 },
 [0x1F60D]={
  category="so",
  direction="on",
 },
 [0x1F60E]={
  category="so",
  direction="on",
 },
 [0x1F60F]={
  category="so",
  direction="on",
 },
 [0x1F610]={
  category="so",
  direction="on",
 },
 [0x1F611]={
  category="so",
  direction="on",
 },
 [0x1F612]={
  category="so",
  direction="on",
 },
 [0x1F613]={
  category="so",
  direction="on",
 },
 [0x1F614]={
  category="so",
  direction="on",
 },
 [0x1F615]={
  category="so",
  direction="on",
 },
 [0x1F616]={
  category="so",
  direction="on",
 },
 [0x1F617]={
  category="so",
  direction="on",
 },
 [0x1F618]={
  category="so",
  direction="on",
 },
 [0x1F619]={
  category="so",
  direction="on",
 },
 [0x1F61A]={
  category="so",
  direction="on",
 },
 [0x1F61B]={
  category="so",
  direction="on",
 },
 [0x1F61C]={
  category="so",
  direction="on",
 },
 [0x1F61D]={
  category="so",
  direction="on",
 },
 [0x1F61E]={
  category="so",
  direction="on",
 },
 [0x1F61F]={
  category="so",
  direction="on",
 },
 [0x1F620]={
  category="so",
  direction="on",
 },
 [0x1F621]={
  category="so",
  direction="on",
 },
 [0x1F622]={
  category="so",
  direction="on",
 },
 [0x1F623]={
  category="so",
  direction="on",
 },
 [0x1F624]={
  category="so",
  direction="on",
 },
 [0x1F625]={
  category="so",
  direction="on",
 },
 [0x1F626]={
  category="so",
  direction="on",
 },
 [0x1F627]={
  category="so",
  direction="on",
 },
 [0x1F628]={
  category="so",
  direction="on",
 },
 [0x1F629]={
  category="so",
  direction="on",
 },
 [0x1F62A]={
  category="so",
  direction="on",
 },
 [0x1F62B]={
  category="so",
  direction="on",
 },
 [0x1F62C]={
  category="so",
  direction="on",
 },
 [0x1F62D]={
  category="so",
  direction="on",
 },
 [0x1F62E]={
  category="so",
  direction="on",
 },
 [0x1F62F]={
  category="so",
  direction="on",
 },
 [0x1F630]={
  category="so",
  direction="on",
 },
 [0x1F631]={
  category="so",
  direction="on",
 },
 [0x1F632]={
  category="so",
  direction="on",
 },
 [0x1F633]={
  category="so",
  direction="on",
 },
 [0x1F634]={
  category="so",
  direction="on",
 },
 [0x1F635]={
  category="so",
  direction="on",
 },
 [0x1F636]={
  category="so",
  direction="on",
 },
 [0x1F637]={
  category="so",
  direction="on",
 },
 [0x1F638]={
  category="so",
  direction="on",
 },
 [0x1F639]={
  category="so",
  direction="on",
 },
 [0x1F63A]={
  category="so",
  direction="on",
 },
 [0x1F63B]={
  category="so",
  direction="on",
 },
 [0x1F63C]={
  category="so",
  direction="on",
 },
 [0x1F63D]={
  category="so",
  direction="on",
 },
 [0x1F63E]={
  category="so",
  direction="on",
 },
 [0x1F63F]={
  category="so",
  direction="on",
 },
 [0x1F640]={
  category="so",
  direction="on",
 },
 [0x1F641]={
  category="so",
  direction="on",
 },
 [0x1F642]={
  category="so",
  direction="on",
 },
 [0x1F645]={
  category="so",
  direction="on",
 },
 [0x1F646]={
  category="so",
  direction="on",
 },
 [0x1F647]={
  category="so",
  direction="on",
 },
 [0x1F648]={
  category="so",
  direction="on",
 },
 [0x1F649]={
  category="so",
  direction="on",
 },
 [0x1F64A]={
  category="so",
  direction="on",
 },
 [0x1F64B]={
  category="so",
  direction="on",
 },
 [0x1F64C]={
  category="so",
  direction="on",
 },
 [0x1F64D]={
  category="so",
  direction="on",
 },
 [0x1F64E]={
  category="so",
  direction="on",
 },
 [0x1F64F]={
  category="so",
  direction="on",
 },
 [0x1F650]={
  category="so",
  direction="on",
 },
 [0x1F651]={
  category="so",
  direction="on",
 },
 [0x1F652]={
  category="so",
  direction="on",
 },
 [0x1F653]={
  category="so",
  direction="on",
 },
 [0x1F654]={
  category="so",
  direction="on",
 },
 [0x1F655]={
  category="so",
  direction="on",
 },
 [0x1F656]={
  category="so",
  direction="on",
 },
 [0x1F657]={
  category="so",
  direction="on",
 },
 [0x1F658]={
  category="so",
  direction="on",
 },
 [0x1F659]={
  category="so",
  direction="on",
 },
 [0x1F65A]={
  category="so",
  direction="on",
 },
 [0x1F65B]={
  category="so",
  direction="on",
 },
 [0x1F65C]={
  category="so",
  direction="on",
 },
 [0x1F65D]={
  category="so",
  direction="on",
 },
 [0x1F65E]={
  category="so",
  direction="on",
 },
 [0x1F65F]={
  category="so",
  direction="on",
 },
 [0x1F660]={
  category="so",
  direction="on",
 },
 [0x1F661]={
  category="so",
  direction="on",
 },
 [0x1F662]={
  category="so",
  direction="on",
 },
 [0x1F663]={
  category="so",
  direction="on",
 },
 [0x1F664]={
  category="so",
  direction="on",
 },
 [0x1F665]={
  category="so",
  direction="on",
 },
 [0x1F666]={
  category="so",
  direction="on",
 },
 [0x1F667]={
  category="so",
  direction="on",
 },
 [0x1F668]={
  category="so",
  direction="on",
 },
 [0x1F669]={
  category="so",
  direction="on",
 },
 [0x1F66A]={
  category="so",
  direction="on",
 },
 [0x1F66B]={
  category="so",
  direction="on",
 },
 [0x1F66C]={
  category="so",
  direction="on",
 },
 [0x1F66D]={
  category="so",
  direction="on",
 },
 [0x1F66E]={
  category="so",
  direction="on",
 },
 [0x1F66F]={
  category="so",
  direction="on",
 },
 [0x1F670]={
  category="so",
  direction="on",
 },
 [0x1F671]={
  category="so",
  direction="on",
 },
 [0x1F672]={
  category="so",
  direction="on",
 },
 [0x1F673]={
  category="so",
  direction="on",
 },
 [0x1F674]={
  category="so",
  direction="on",
 },
 [0x1F675]={
  category="so",
  direction="on",
 },
 [0x1F676]={
  category="so",
  direction="on",
 },
 [0x1F677]={
  category="so",
  direction="on",
 },
 [0x1F678]={
  category="so",
  direction="on",
 },
 [0x1F679]={
  category="so",
  direction="on",
 },
 [0x1F67A]={
  category="so",
  direction="on",
 },
 [0x1F67B]={
  category="so",
  direction="on",
 },
 [0x1F67C]={
  category="so",
  direction="on",
 },
 [0x1F67D]={
  category="so",
  direction="on",
 },
 [0x1F67E]={
  category="so",
  direction="on",
 },
 [0x1F67F]={
  category="so",
  direction="on",
 },
 [0x1F680]={
  category="so",
  direction="on",
 },
 [0x1F681]={
  category="so",
  direction="on",
 },
 [0x1F682]={
  category="so",
  direction="on",
 },
 [0x1F683]={
  category="so",
  direction="on",
 },
 [0x1F684]={
  category="so",
  direction="on",
 },
 [0x1F685]={
  category="so",
  direction="on",
 },
 [0x1F686]={
  category="so",
  direction="on",
 },
 [0x1F687]={
  category="so",
  direction="on",
 },
 [0x1F688]={
  category="so",
  direction="on",
 },
 [0x1F689]={
  category="so",
  direction="on",
 },
 [0x1F68A]={
  category="so",
  direction="on",
 },
 [0x1F68B]={
  category="so",
  direction="on",
 },
 [0x1F68C]={
  category="so",
  direction="on",
 },
 [0x1F68D]={
  category="so",
  direction="on",
 },
 [0x1F68E]={
  category="so",
  direction="on",
 },
 [0x1F68F]={
  category="so",
  direction="on",
 },
 [0x1F690]={
  category="so",
  direction="on",
 },
 [0x1F691]={
  category="so",
  direction="on",
 },
 [0x1F692]={
  category="so",
  direction="on",
 },
 [0x1F693]={
  category="so",
  direction="on",
 },
 [0x1F694]={
  category="so",
  direction="on",
 },
 [0x1F695]={
  category="so",
  direction="on",
 },
 [0x1F696]={
  category="so",
  direction="on",
 },
 [0x1F697]={
  category="so",
  direction="on",
 },
 [0x1F698]={
  category="so",
  direction="on",
 },
 [0x1F699]={
  category="so",
  direction="on",
 },
 [0x1F69A]={
  category="so",
  direction="on",
 },
 [0x1F69B]={
  category="so",
  direction="on",
 },
 [0x1F69C]={
  category="so",
  direction="on",
 },
 [0x1F69D]={
  category="so",
  direction="on",
 },
 [0x1F69E]={
  category="so",
  direction="on",
 },
 [0x1F69F]={
  category="so",
  direction="on",
 },
 [0x1F6A0]={
  category="so",
  direction="on",
 },
 [0x1F6A1]={
  category="so",
  direction="on",
 },
 [0x1F6A2]={
  category="so",
  direction="on",
 },
 [0x1F6A3]={
  category="so",
  direction="on",
 },
 [0x1F6A4]={
  category="so",
  direction="on",
 },
 [0x1F6A5]={
  category="so",
  direction="on",
 },
 [0x1F6A6]={
  category="so",
  direction="on",
 },
 [0x1F6A7]={
  category="so",
  direction="on",
 },
 [0x1F6A8]={
  category="so",
  direction="on",
 },
 [0x1F6A9]={
  category="so",
  direction="on",
 },
 [0x1F6AA]={
  category="so",
  direction="on",
 },
 [0x1F6AB]={
  category="so",
  direction="on",
 },
 [0x1F6AC]={
  category="so",
  direction="on",
 },
 [0x1F6AD]={
  category="so",
  direction="on",
 },
 [0x1F6AE]={
  category="so",
  direction="on",
 },
 [0x1F6AF]={
  category="so",
  direction="on",
 },
 [0x1F6B0]={
  category="so",
  direction="on",
 },
 [0x1F6B1]={
  category="so",
  direction="on",
 },
 [0x1F6B2]={
  category="so",
  direction="on",
 },
 [0x1F6B3]={
  category="so",
  direction="on",
 },
 [0x1F6B4]={
  category="so",
  direction="on",
 },
 [0x1F6B5]={
  category="so",
  direction="on",
 },
 [0x1F6B6]={
  category="so",
  direction="on",
 },
 [0x1F6B7]={
  category="so",
  direction="on",
 },
 [0x1F6B8]={
  category="so",
  direction="on",
 },
 [0x1F6B9]={
  category="so",
  direction="on",
 },
 [0x1F6BA]={
  category="so",
  direction="on",
 },
 [0x1F6BB]={
  category="so",
  direction="on",
 },
 [0x1F6BC]={
  category="so",
  direction="on",
 },
 [0x1F6BD]={
  category="so",
  direction="on",
 },
 [0x1F6BE]={
  category="so",
  direction="on",
 },
 [0x1F6BF]={
  category="so",
  direction="on",
 },
 [0x1F6C0]={
  category="so",
  direction="on",
 },
 [0x1F6C1]={
  category="so",
  direction="on",
 },
 [0x1F6C2]={
  category="so",
  direction="on",
 },
 [0x1F6C3]={
  category="so",
  direction="on",
 },
 [0x1F6C4]={
  category="so",
  direction="on",
 },
 [0x1F6C5]={
  category="so",
  direction="on",
 },
 [0x1F6C6]={
  category="so",
  direction="on",
 },
 [0x1F6C7]={
  category="so",
  direction="on",
 },
 [0x1F6C8]={
  category="so",
  direction="on",
 },
 [0x1F6C9]={
  category="so",
  direction="on",
 },
 [0x1F6CA]={
  category="so",
  direction="on",
 },
 [0x1F6CB]={
  category="so",
  direction="on",
 },
 [0x1F6CC]={
  category="so",
  direction="on",
 },
 [0x1F6CD]={
  category="so",
  direction="on",
 },
 [0x1F6CE]={
  category="so",
  direction="on",
 },
 [0x1F6CF]={
  category="so",
  direction="on",
 },
 [0x1F6E0]={
  category="so",
  direction="on",
 },
 [0x1F6E1]={
  category="so",
  direction="on",
 },
 [0x1F6E2]={
  category="so",
  direction="on",
 },
 [0x1F6E3]={
  category="so",
  direction="on",
 },
 [0x1F6E4]={
  category="so",
  direction="on",
 },
 [0x1F6E5]={
  category="so",
  direction="on",
 },
 [0x1F6E6]={
  category="so",
  direction="on",
 },
 [0x1F6E7]={
  category="so",
  direction="on",
 },
 [0x1F6E8]={
  category="so",
  direction="on",
 },
 [0x1F6E9]={
  category="so",
  direction="on",
 },
 [0x1F6EA]={
  category="so",
  direction="on",
 },
 [0x1F6EB]={
  category="so",
  direction="on",
 },
 [0x1F6EC]={
  category="so",
  direction="on",
 },
 [0x1F6F0]={
  category="so",
  direction="on",
 },
 [0x1F6F1]={
  category="so",
  direction="on",
 },
 [0x1F6F2]={
  category="so",
  direction="on",
 },
 [0x1F6F3]={
  category="so",
  direction="on",
 },
 [0x1F700]={
  category="so",
  direction="on",
 },
 [0x1F701]={
  category="so",
  direction="on",
 },
 [0x1F702]={
  category="so",
  direction="on",
 },
 [0x1F703]={
  category="so",
  direction="on",
 },
 [0x1F704]={
  category="so",
  direction="on",
 },
 [0x1F705]={
  category="so",
  direction="on",
 },
 [0x1F706]={
  category="so",
  direction="on",
 },
 [0x1F707]={
  category="so",
  direction="on",
 },
 [0x1F708]={
  category="so",
  direction="on",
 },
 [0x1F709]={
  category="so",
  direction="on",
 },
 [0x1F70A]={
  category="so",
  direction="on",
 },
 [0x1F70B]={
  category="so",
  direction="on",
 },
 [0x1F70C]={
  category="so",
  direction="on",
 },
 [0x1F70D]={
  category="so",
  direction="on",
 },
 [0x1F70E]={
  category="so",
  direction="on",
 },
 [0x1F70F]={
  category="so",
  direction="on",
 },
 [0x1F710]={
  category="so",
  direction="on",
 },
 [0x1F711]={
  category="so",
  direction="on",
 },
 [0x1F712]={
  category="so",
  direction="on",
 },
 [0x1F713]={
  category="so",
  direction="on",
 },
 [0x1F714]={
  category="so",
  direction="on",
 },
 [0x1F715]={
  category="so",
  direction="on",
 },
 [0x1F716]={
  category="so",
  direction="on",
 },
 [0x1F717]={
  category="so",
  direction="on",
 },
 [0x1F718]={
  category="so",
  direction="on",
 },
 [0x1F719]={
  category="so",
  direction="on",
 },
 [0x1F71A]={
  category="so",
  direction="on",
 },
 [0x1F71B]={
  category="so",
  direction="on",
 },
 [0x1F71C]={
  category="so",
  direction="on",
 },
 [0x1F71D]={
  category="so",
  direction="on",
 },
 [0x1F71E]={
  category="so",
  direction="on",
 },
 [0x1F71F]={
  category="so",
  direction="on",
 },
 [0x1F720]={
  category="so",
  direction="on",
 },
 [0x1F721]={
  category="so",
  direction="on",
 },
 [0x1F722]={
  category="so",
  direction="on",
 },
 [0x1F723]={
  category="so",
  direction="on",
 },
 [0x1F724]={
  category="so",
  direction="on",
 },
 [0x1F725]={
  category="so",
  direction="on",
 },
 [0x1F726]={
  category="so",
  direction="on",
 },
 [0x1F727]={
  category="so",
  direction="on",
 },
 [0x1F728]={
  category="so",
  direction="on",
 },
 [0x1F729]={
  category="so",
  direction="on",
 },
 [0x1F72A]={
  category="so",
  direction="on",
 },
 [0x1F72B]={
  category="so",
  direction="on",
 },
 [0x1F72C]={
  category="so",
  direction="on",
 },
 [0x1F72D]={
  category="so",
  direction="on",
 },
 [0x1F72E]={
  category="so",
  direction="on",
 },
 [0x1F72F]={
  category="so",
  direction="on",
 },
 [0x1F730]={
  category="so",
  direction="on",
 },
 [0x1F731]={
  category="so",
  direction="on",
 },
 [0x1F732]={
  category="so",
  direction="on",
 },
 [0x1F733]={
  category="so",
  direction="on",
 },
 [0x1F734]={
  category="so",
  direction="on",
 },
 [0x1F735]={
  category="so",
  direction="on",
 },
 [0x1F736]={
  category="so",
  direction="on",
 },
 [0x1F737]={
  category="so",
  direction="on",
 },
 [0x1F738]={
  category="so",
  direction="on",
 },
 [0x1F739]={
  category="so",
  direction="on",
 },
 [0x1F73A]={
  category="so",
  direction="on",
 },
 [0x1F73B]={
  category="so",
  direction="on",
 },
 [0x1F73C]={
  category="so",
  direction="on",
 },
 [0x1F73D]={
  category="so",
  direction="on",
 },
 [0x1F73E]={
  category="so",
  direction="on",
 },
 [0x1F73F]={
  category="so",
  direction="on",
 },
 [0x1F740]={
  category="so",
  direction="on",
 },
 [0x1F741]={
  category="so",
  direction="on",
 },
 [0x1F742]={
  category="so",
  direction="on",
 },
 [0x1F743]={
  category="so",
  direction="on",
 },
 [0x1F744]={
  category="so",
  direction="on",
 },
 [0x1F745]={
  category="so",
  direction="on",
 },
 [0x1F746]={
  category="so",
  direction="on",
 },
 [0x1F747]={
  category="so",
  direction="on",
 },
 [0x1F748]={
  category="so",
  direction="on",
 },
 [0x1F749]={
  category="so",
  direction="on",
 },
 [0x1F74A]={
  category="so",
  direction="on",
 },
 [0x1F74B]={
  category="so",
  direction="on",
 },
 [0x1F74C]={
  category="so",
  direction="on",
 },
 [0x1F74D]={
  category="so",
  direction="on",
 },
 [0x1F74E]={
  category="so",
  direction="on",
 },
 [0x1F74F]={
  category="so",
  direction="on",
 },
 [0x1F750]={
  category="so",
  direction="on",
 },
 [0x1F751]={
  category="so",
  direction="on",
 },
 [0x1F752]={
  category="so",
  direction="on",
 },
 [0x1F753]={
  category="so",
  direction="on",
 },
 [0x1F754]={
  category="so",
  direction="on",
 },
 [0x1F755]={
  category="so",
  direction="on",
 },
 [0x1F756]={
  category="so",
  direction="on",
 },
 [0x1F757]={
  category="so",
  direction="on",
 },
 [0x1F758]={
  category="so",
  direction="on",
 },
 [0x1F759]={
  category="so",
  direction="on",
 },
 [0x1F75A]={
  category="so",
  direction="on",
 },
 [0x1F75B]={
  category="so",
  direction="on",
 },
 [0x1F75C]={
  category="so",
  direction="on",
 },
 [0x1F75D]={
  category="so",
  direction="on",
 },
 [0x1F75E]={
  category="so",
  direction="on",
 },
 [0x1F75F]={
  category="so",
  direction="on",
 },
 [0x1F760]={
  category="so",
  direction="on",
 },
 [0x1F761]={
  category="so",
  direction="on",
 },
 [0x1F762]={
  category="so",
  direction="on",
 },
 [0x1F763]={
  category="so",
  direction="on",
 },
 [0x1F764]={
  category="so",
  direction="on",
 },
 [0x1F765]={
  category="so",
  direction="on",
 },
 [0x1F766]={
  category="so",
  direction="on",
 },
 [0x1F767]={
  category="so",
  direction="on",
 },
 [0x1F768]={
  category="so",
  direction="on",
 },
 [0x1F769]={
  category="so",
  direction="on",
 },
 [0x1F76A]={
  category="so",
  direction="on",
 },
 [0x1F76B]={
  category="so",
  direction="on",
 },
 [0x1F76C]={
  category="so",
  direction="on",
 },
 [0x1F76D]={
  category="so",
  direction="on",
 },
 [0x1F76E]={
  category="so",
  direction="on",
 },
 [0x1F76F]={
  category="so",
  direction="on",
 },
 [0x1F770]={
  category="so",
  direction="on",
 },
 [0x1F771]={
  category="so",
  direction="on",
 },
 [0x1F772]={
  category="so",
  direction="on",
 },
 [0x1F773]={
  category="so",
  direction="on",
 },
 [0x1F780]={
  category="so",
  direction="on",
 },
 [0x1F781]={
  category="so",
  direction="on",
 },
 [0x1F782]={
  category="so",
  direction="on",
 },
 [0x1F783]={
  category="so",
  direction="on",
 },
 [0x1F784]={
  category="so",
  direction="on",
 },
 [0x1F785]={
  category="so",
  direction="on",
 },
 [0x1F786]={
  category="so",
  direction="on",
 },
 [0x1F787]={
  category="so",
  direction="on",
 },
 [0x1F788]={
  category="so",
  direction="on",
 },
 [0x1F789]={
  category="so",
  direction="on",
 },
 [0x1F78A]={
  category="so",
  direction="on",
 },
 [0x1F78B]={
  category="so",
  direction="on",
 },
 [0x1F78C]={
  category="so",
  direction="on",
 },
 [0x1F78D]={
  category="so",
  direction="on",
 },
 [0x1F78E]={
  category="so",
  direction="on",
 },
 [0x1F78F]={
  category="so",
  direction="on",
 },
 [0x1F790]={
  category="so",
  direction="on",
 },
 [0x1F791]={
  category="so",
  direction="on",
 },
 [0x1F792]={
  category="so",
  direction="on",
 },
 [0x1F793]={
  category="so",
  direction="on",
 },
 [0x1F794]={
  category="so",
  direction="on",
 },
 [0x1F795]={
  category="so",
  direction="on",
 },
 [0x1F796]={
  category="so",
  direction="on",
 },
 [0x1F797]={
  category="so",
  direction="on",
 },
 [0x1F798]={
  category="so",
  direction="on",
 },
 [0x1F799]={
  category="so",
  direction="on",
 },
 [0x1F79A]={
  category="so",
  direction="on",
 },
 [0x1F79B]={
  category="so",
  direction="on",
 },
 [0x1F79C]={
  category="so",
  direction="on",
 },
 [0x1F79D]={
  category="so",
  direction="on",
 },
 [0x1F79E]={
  category="so",
  direction="on",
 },
 [0x1F79F]={
  category="so",
  direction="on",
 },
 [0x1F7A0]={
  category="so",
  direction="on",
 },
 [0x1F7A1]={
  category="so",
  direction="on",
 },
 [0x1F7A2]={
  category="so",
  direction="on",
 },
 [0x1F7A3]={
  category="so",
  direction="on",
 },
 [0x1F7A4]={
  category="so",
  direction="on",
 },
 [0x1F7A5]={
  category="so",
  direction="on",
 },
 [0x1F7A6]={
  category="so",
  direction="on",
 },
 [0x1F7A7]={
  category="so",
  direction="on",
 },
 [0x1F7A8]={
  category="so",
  direction="on",
 },
 [0x1F7A9]={
  category="so",
  direction="on",
 },
 [0x1F7AA]={
  category="so",
  direction="on",
 },
 [0x1F7AB]={
  category="so",
  direction="on",
 },
 [0x1F7AC]={
  category="so",
  direction="on",
 },
 [0x1F7AD]={
  category="so",
  direction="on",
 },
 [0x1F7AE]={
  category="so",
  direction="on",
 },
 [0x1F7AF]={
  category="so",
  direction="on",
 },
 [0x1F7B0]={
  category="so",
  direction="on",
 },
 [0x1F7B1]={
  category="so",
  direction="on",
 },
 [0x1F7B2]={
  category="so",
  direction="on",
 },
 [0x1F7B3]={
  category="so",
  direction="on",
 },
 [0x1F7B4]={
  category="so",
  direction="on",
 },
 [0x1F7B5]={
  category="so",
  direction="on",
 },
 [0x1F7B6]={
  category="so",
  direction="on",
 },
 [0x1F7B7]={
  category="so",
  direction="on",
 },
 [0x1F7B8]={
  category="so",
  direction="on",
 },
 [0x1F7B9]={
  category="so",
  direction="on",
 },
 [0x1F7BA]={
  category="so",
  direction="on",
 },
 [0x1F7BB]={
  category="so",
  direction="on",
 },
 [0x1F7BC]={
  category="so",
  direction="on",
 },
 [0x1F7BD]={
  category="so",
  direction="on",
 },
 [0x1F7BE]={
  category="so",
  direction="on",
 },
 [0x1F7BF]={
  category="so",
  direction="on",
 },
 [0x1F7C0]={
  category="so",
  direction="on",
 },
 [0x1F7C1]={
  category="so",
  direction="on",
 },
 [0x1F7C2]={
  category="so",
  direction="on",
 },
 [0x1F7C3]={
  category="so",
  direction="on",
 },
 [0x1F7C4]={
  category="so",
  direction="on",
 },
 [0x1F7C5]={
  category="so",
  direction="on",
 },
 [0x1F7C6]={
  category="so",
  direction="on",
 },
 [0x1F7C7]={
  category="so",
  direction="on",
 },
 [0x1F7C8]={
  category="so",
  direction="on",
 },
 [0x1F7C9]={
  category="so",
  direction="on",
 },
 [0x1F7CA]={
  category="so",
  direction="on",
 },
 [0x1F7CB]={
  category="so",
  direction="on",
 },
 [0x1F7CC]={
  category="so",
  direction="on",
 },
 [0x1F7CD]={
  category="so",
  direction="on",
 },
 [0x1F7CE]={
  category="so",
  direction="on",
 },
 [0x1F7CF]={
  category="so",
  direction="on",
 },
 [0x1F7D0]={
  category="so",
  direction="on",
 },
 [0x1F7D1]={
  category="so",
  direction="on",
 },
 [0x1F7D2]={
  category="so",
  direction="on",
 },
 [0x1F7D3]={
  category="so",
  direction="on",
 },
 [0x1F7D4]={
  category="so",
  direction="on",
 },
 [0x1F800]={
  category="so",
  direction="on",
 },
 [0x1F801]={
  category="so",
  direction="on",
 },
 [0x1F802]={
  category="so",
  direction="on",
 },
 [0x1F803]={
  category="so",
  direction="on",
 },
 [0x1F804]={
  category="so",
  direction="on",
 },
 [0x1F805]={
  category="so",
  direction="on",
 },
 [0x1F806]={
  category="so",
  direction="on",
 },
 [0x1F807]={
  category="so",
  direction="on",
 },
 [0x1F808]={
  category="so",
  direction="on",
 },
 [0x1F809]={
  category="so",
  direction="on",
 },
 [0x1F80A]={
  category="so",
  direction="on",
 },
 [0x1F80B]={
  category="so",
  direction="on",
 },
 [0x1F810]={
  category="so",
  direction="on",
 },
 [0x1F811]={
  category="so",
  direction="on",
 },
 [0x1F812]={
  category="so",
  direction="on",
 },
 [0x1F813]={
  category="so",
  direction="on",
 },
 [0x1F814]={
  category="so",
  direction="on",
 },
 [0x1F815]={
  category="so",
  direction="on",
 },
 [0x1F816]={
  category="so",
  direction="on",
 },
 [0x1F817]={
  category="so",
  direction="on",
 },
 [0x1F818]={
  category="so",
  direction="on",
 },
 [0x1F819]={
  category="so",
  direction="on",
 },
 [0x1F81A]={
  category="so",
  direction="on",
 },
 [0x1F81B]={
  category="so",
  direction="on",
 },
 [0x1F81C]={
  category="so",
  direction="on",
 },
 [0x1F81D]={
  category="so",
  direction="on",
 },
 [0x1F81E]={
  category="so",
  direction="on",
 },
 [0x1F81F]={
  category="so",
  direction="on",
 },
 [0x1F820]={
  category="so",
  direction="on",
 },
 [0x1F821]={
  category="so",
  direction="on",
 },
 [0x1F822]={
  category="so",
  direction="on",
 },
 [0x1F823]={
  category="so",
  direction="on",
 },
 [0x1F824]={
  category="so",
  direction="on",
 },
 [0x1F825]={
  category="so",
  direction="on",
 },
 [0x1F826]={
  category="so",
  direction="on",
 },
 [0x1F827]={
  category="so",
  direction="on",
 },
 [0x1F828]={
  category="so",
  direction="on",
 },
 [0x1F829]={
  category="so",
  direction="on",
 },
 [0x1F82A]={
  category="so",
  direction="on",
 },
 [0x1F82B]={
  category="so",
  direction="on",
 },
 [0x1F82C]={
  category="so",
  direction="on",
 },
 [0x1F82D]={
  category="so",
  direction="on",
 },
 [0x1F82E]={
  category="so",
  direction="on",
 },
 [0x1F82F]={
  category="so",
  direction="on",
 },
 [0x1F830]={
  category="so",
  direction="on",
 },
 [0x1F831]={
  category="so",
  direction="on",
 },
 [0x1F832]={
  category="so",
  direction="on",
 },
 [0x1F833]={
  category="so",
  direction="on",
 },
 [0x1F834]={
  category="so",
  direction="on",
 },
 [0x1F835]={
  category="so",
  direction="on",
 },
 [0x1F836]={
  category="so",
  direction="on",
 },
 [0x1F837]={
  category="so",
  direction="on",
 },
 [0x1F838]={
  category="so",
  direction="on",
 },
 [0x1F839]={
  category="so",
  direction="on",
 },
 [0x1F83A]={
  category="so",
  direction="on",
 },
 [0x1F83B]={
  category="so",
  direction="on",
 },
 [0x1F83C]={
  category="so",
  direction="on",
 },
 [0x1F83D]={
  category="so",
  direction="on",
 },
 [0x1F83E]={
  category="so",
  direction="on",
 },
 [0x1F83F]={
  category="so",
  direction="on",
 },
 [0x1F840]={
  category="so",
  direction="on",
 },
 [0x1F841]={
  category="so",
  direction="on",
 },
 [0x1F842]={
  category="so",
  direction="on",
 },
 [0x1F843]={
  category="so",
  direction="on",
 },
 [0x1F844]={
  category="so",
  direction="on",
 },
 [0x1F845]={
  category="so",
  direction="on",
 },
 [0x1F846]={
  category="so",
  direction="on",
 },
 [0x1F847]={
  category="so",
  direction="on",
 },
 [0x1F850]={
  category="so",
  direction="on",
 },
 [0x1F851]={
  category="so",
  direction="on",
 },
 [0x1F852]={
  category="so",
  direction="on",
 },
 [0x1F853]={
  category="so",
  direction="on",
 },
 [0x1F854]={
  category="so",
  direction="on",
 },
 [0x1F855]={
  category="so",
  direction="on",
 },
 [0x1F856]={
  category="so",
  direction="on",
 },
 [0x1F857]={
  category="so",
  direction="on",
 },
 [0x1F858]={
  category="so",
  direction="on",
 },
 [0x1F859]={
  category="so",
  direction="on",
 },
 [0x1F860]={
  category="so",
  direction="on",
 },
 [0x1F861]={
  category="so",
  direction="on",
 },
 [0x1F862]={
  category="so",
  direction="on",
 },
 [0x1F863]={
  category="so",
  direction="on",
 },
 [0x1F864]={
  category="so",
  direction="on",
 },
 [0x1F865]={
  category="so",
  direction="on",
 },
 [0x1F866]={
  category="so",
  direction="on",
 },
 [0x1F867]={
  category="so",
  direction="on",
 },
 [0x1F868]={
  category="so",
  direction="on",
 },
 [0x1F869]={
  category="so",
  direction="on",
 },
 [0x1F86A]={
  category="so",
  direction="on",
 },
 [0x1F86B]={
  category="so",
  direction="on",
 },
 [0x1F86C]={
  category="so",
  direction="on",
 },
 [0x1F86D]={
  category="so",
  direction="on",
 },
 [0x1F86E]={
  category="so",
  direction="on",
 },
 [0x1F86F]={
  category="so",
  direction="on",
 },
 [0x1F870]={
  category="so",
  direction="on",
 },
 [0x1F871]={
  category="so",
  direction="on",
 },
 [0x1F872]={
  category="so",
  direction="on",
 },
 [0x1F873]={
  category="so",
  direction="on",
 },
 [0x1F874]={
  category="so",
  direction="on",
 },
 [0x1F875]={
  category="so",
  direction="on",
 },
 [0x1F876]={
  category="so",
  direction="on",
 },
 [0x1F877]={
  category="so",
  direction="on",
 },
 [0x1F878]={
  category="so",
  direction="on",
 },
 [0x1F879]={
  category="so",
  direction="on",
 },
 [0x1F87A]={
  category="so",
  direction="on",
 },
 [0x1F87B]={
  category="so",
  direction="on",
 },
 [0x1F87C]={
  category="so",
  direction="on",
 },
 [0x1F87D]={
  category="so",
  direction="on",
 },
 [0x1F87E]={
  category="so",
  direction="on",
 },
 [0x1F87F]={
  category="so",
  direction="on",
 },
 [0x1F880]={
  category="so",
  direction="on",
 },
 [0x1F881]={
  category="so",
  direction="on",
 },
 [0x1F882]={
  category="so",
  direction="on",
 },
 [0x1F883]={
  category="so",
  direction="on",
 },
 [0x1F884]={
  category="so",
  direction="on",
 },
 [0x1F885]={
  category="so",
  direction="on",
 },
 [0x1F886]={
  category="so",
  direction="on",
 },
 [0x1F887]={
  category="so",
  direction="on",
 },
 [0x1F890]={
  category="so",
  direction="on",
 },
 [0x1F891]={
  category="so",
  direction="on",
 },
 [0x1F892]={
  category="so",
  direction="on",
 },
 [0x1F893]={
  category="so",
  direction="on",
 },
 [0x1F894]={
  category="so",
  direction="on",
 },
 [0x1F895]={
  category="so",
  direction="on",
 },
 [0x1F896]={
  category="so",
  direction="on",
 },
 [0x1F897]={
  category="so",
  direction="on",
 },
 [0x1F898]={
  category="so",
  direction="on",
 },
 [0x1F899]={
  category="so",
  direction="on",
 },
 [0x1F89A]={
  category="so",
  direction="on",
 },
 [0x1F89B]={
  category="so",
  direction="on",
 },
 [0x1F89C]={
  category="so",
  direction="on",
 },
 [0x1F89D]={
  category="so",
  direction="on",
 },
 [0x1F89E]={
  category="so",
  direction="on",
 },
 [0x1F89F]={
  category="so",
  direction="on",
 },
 [0x1F8A0]={
  category="so",
  direction="on",
 },
 [0x1F8A1]={
  category="so",
  direction="on",
 },
 [0x1F8A2]={
  category="so",
  direction="on",
 },
 [0x1F8A3]={
  category="so",
  direction="on",
 },
 [0x1F8A4]={
  category="so",
  direction="on",
 },
 [0x1F8A5]={
  category="so",
  direction="on",
 },
 [0x1F8A6]={
  category="so",
  direction="on",
 },
 [0x1F8A7]={
  category="so",
  direction="on",
 },
 [0x1F8A8]={
  category="so",
  direction="on",
 },
 [0x1F8A9]={
  category="so",
  direction="on",
 },
 [0x1F8AA]={
  category="so",
  direction="on",
 },
 [0x1F8AB]={
  category="so",
  direction="on",
 },
 [0x1F8AC]={
  category="so",
  direction="on",
 },
 [0x1F8AD]={
  category="so",
  direction="on",
 },
 [0x2F800]={
  category="lo",
  direction="l",
 },
 [0x2F801]={
  category="lo",
  direction="l",
 },
 [0x2F802]={
  category="lo",
  direction="l",
 },
 [0x2F803]={
  category="lo",
  direction="l",
 },
 [0x2F804]={
  category="lo",
  direction="l",
 },
 [0x2F805]={
  category="lo",
  direction="l",
 },
 [0x2F806]={
  category="lo",
  direction="l",
 },
 [0x2F807]={
  category="lo",
  direction="l",
 },
 [0x2F808]={
  category="lo",
  direction="l",
 },
 [0x2F809]={
  category="lo",
  direction="l",
 },
 [0x2F80A]={
  category="lo",
  direction="l",
 },
 [0x2F80B]={
  category="lo",
  direction="l",
 },
 [0x2F80C]={
  category="lo",
  direction="l",
 },
 [0x2F80D]={
  category="lo",
  direction="l",
 },
 [0x2F80E]={
  category="lo",
  direction="l",
 },
 [0x2F80F]={
  category="lo",
  direction="l",
 },
 [0x2F810]={
  category="lo",
  direction="l",
 },
 [0x2F811]={
  category="lo",
  direction="l",
 },
 [0x2F812]={
  category="lo",
  direction="l",
 },
 [0x2F813]={
  category="lo",
  direction="l",
 },
 [0x2F814]={
  category="lo",
  direction="l",
 },
 [0x2F815]={
  category="lo",
  direction="l",
 },
 [0x2F816]={
  category="lo",
  direction="l",
 },
 [0x2F817]={
  category="lo",
  direction="l",
 },
 [0x2F818]={
  category="lo",
  direction="l",
 },
 [0x2F819]={
  category="lo",
  direction="l",
 },
 [0x2F81A]={
  category="lo",
  direction="l",
 },
 [0x2F81B]={
  category="lo",
  direction="l",
 },
 [0x2F81C]={
  category="lo",
  direction="l",
 },
 [0x2F81D]={
  category="lo",
  direction="l",
 },
 [0x2F81E]={
  category="lo",
  direction="l",
 },
 [0x2F81F]={
  category="lo",
  direction="l",
 },
 [0x2F820]={
  category="lo",
  direction="l",
 },
 [0x2F821]={
  category="lo",
  direction="l",
 },
 [0x2F822]={
  category="lo",
  direction="l",
 },
 [0x2F823]={
  category="lo",
  direction="l",
 },
 [0x2F824]={
  category="lo",
  direction="l",
 },
 [0x2F825]={
  category="lo",
  direction="l",
 },
 [0x2F826]={
  category="lo",
  direction="l",
 },
 [0x2F827]={
  category="lo",
  direction="l",
 },
 [0x2F828]={
  category="lo",
  direction="l",
 },
 [0x2F829]={
  category="lo",
  direction="l",
 },
 [0x2F82A]={
  category="lo",
  direction="l",
 },
 [0x2F82B]={
  category="lo",
  direction="l",
 },
 [0x2F82C]={
  category="lo",
  direction="l",
 },
 [0x2F82D]={
  category="lo",
  direction="l",
 },
 [0x2F82E]={
  category="lo",
  direction="l",
 },
 [0x2F82F]={
  category="lo",
  direction="l",
 },
 [0x2F830]={
  category="lo",
  direction="l",
 },
 [0x2F831]={
  category="lo",
  direction="l",
 },
 [0x2F832]={
  category="lo",
  direction="l",
 },
 [0x2F833]={
  category="lo",
  direction="l",
 },
 [0x2F834]={
  category="lo",
  direction="l",
 },
 [0x2F835]={
  category="lo",
  direction="l",
 },
 [0x2F836]={
  category="lo",
  direction="l",
 },
 [0x2F837]={
  category="lo",
  direction="l",
 },
 [0x2F838]={
  category="lo",
  direction="l",
 },
 [0x2F839]={
  category="lo",
  direction="l",
 },
 [0x2F83A]={
  category="lo",
  direction="l",
 },
 [0x2F83B]={
  category="lo",
  direction="l",
 },
 [0x2F83C]={
  category="lo",
  direction="l",
 },
 [0x2F83D]={
  category="lo",
  direction="l",
 },
 [0x2F83E]={
  category="lo",
  direction="l",
 },
 [0x2F83F]={
  category="lo",
  direction="l",
 },
 [0x2F840]={
  category="lo",
  direction="l",
 },
 [0x2F841]={
  category="lo",
  direction="l",
 },
 [0x2F842]={
  category="lo",
  direction="l",
 },
 [0x2F843]={
  category="lo",
  direction="l",
 },
 [0x2F844]={
  category="lo",
  direction="l",
 },
 [0x2F845]={
  category="lo",
  direction="l",
 },
 [0x2F846]={
  category="lo",
  direction="l",
 },
 [0x2F847]={
  category="lo",
  direction="l",
 },
 [0x2F848]={
  category="lo",
  direction="l",
 },
 [0x2F849]={
  category="lo",
  direction="l",
 },
 [0x2F84A]={
  category="lo",
  direction="l",
 },
 [0x2F84B]={
  category="lo",
  direction="l",
 },
 [0x2F84C]={
  category="lo",
  direction="l",
 },
 [0x2F84D]={
  category="lo",
  direction="l",
 },
 [0x2F84E]={
  category="lo",
  direction="l",
 },
 [0x2F84F]={
  category="lo",
  direction="l",
 },
 [0x2F850]={
  category="lo",
  direction="l",
 },
 [0x2F851]={
  category="lo",
  direction="l",
 },
 [0x2F852]={
  category="lo",
  direction="l",
 },
 [0x2F853]={
  category="lo",
  direction="l",
 },
 [0x2F854]={
  category="lo",
  direction="l",
 },
 [0x2F855]={
  category="lo",
  direction="l",
 },
 [0x2F856]={
  category="lo",
  direction="l",
 },
 [0x2F857]={
  category="lo",
  direction="l",
 },
 [0x2F858]={
  category="lo",
  direction="l",
 },
 [0x2F859]={
  category="lo",
  direction="l",
 },
 [0x2F85A]={
  category="lo",
  direction="l",
 },
 [0x2F85B]={
  category="lo",
  direction="l",
 },
 [0x2F85C]={
  category="lo",
  direction="l",
 },
 [0x2F85D]={
  category="lo",
  direction="l",
 },
 [0x2F85E]={
  category="lo",
  direction="l",
 },
 [0x2F85F]={
  category="lo",
  direction="l",
 },
 [0x2F860]={
  category="lo",
  direction="l",
 },
 [0x2F861]={
  category="lo",
  direction="l",
 },
 [0x2F862]={
  category="lo",
  direction="l",
 },
 [0x2F863]={
  category="lo",
  direction="l",
 },
 [0x2F864]={
  category="lo",
  direction="l",
 },
 [0x2F865]={
  category="lo",
  direction="l",
 },
 [0x2F866]={
  category="lo",
  direction="l",
 },
 [0x2F867]={
  category="lo",
  direction="l",
 },
 [0x2F868]={
  category="lo",
  direction="l",
 },
 [0x2F869]={
  category="lo",
  direction="l",
 },
 [0x2F86A]={
  category="lo",
  direction="l",
 },
 [0x2F86B]={
  category="lo",
  direction="l",
 },
 [0x2F86C]={
  category="lo",
  direction="l",
 },
 [0x2F86D]={
  category="lo",
  direction="l",
 },
 [0x2F86E]={
  category="lo",
  direction="l",
 },
 [0x2F86F]={
  category="lo",
  direction="l",
 },
 [0x2F870]={
  category="lo",
  direction="l",
 },
 [0x2F871]={
  category="lo",
  direction="l",
 },
 [0x2F872]={
  category="lo",
  direction="l",
 },
 [0x2F873]={
  category="lo",
  direction="l",
 },
 [0x2F874]={
  category="lo",
  direction="l",
 },
 [0x2F875]={
  category="lo",
  direction="l",
 },
 [0x2F876]={
  category="lo",
  direction="l",
 },
 [0x2F877]={
  category="lo",
  direction="l",
 },
 [0x2F878]={
  category="lo",
  direction="l",
 },
 [0x2F879]={
  category="lo",
  direction="l",
 },
 [0x2F87A]={
  category="lo",
  direction="l",
 },
 [0x2F87B]={
  category="lo",
  direction="l",
 },
 [0x2F87C]={
  category="lo",
  direction="l",
 },
 [0x2F87D]={
  category="lo",
  direction="l",
 },
 [0x2F87E]={
  category="lo",
  direction="l",
 },
 [0x2F87F]={
  category="lo",
  direction="l",
 },
 [0x2F880]={
  category="lo",
  direction="l",
 },
 [0x2F881]={
  category="lo",
  direction="l",
 },
 [0x2F882]={
  category="lo",
  direction="l",
 },
 [0x2F883]={
  category="lo",
  direction="l",
 },
 [0x2F884]={
  category="lo",
  direction="l",
 },
 [0x2F885]={
  category="lo",
  direction="l",
 },
 [0x2F886]={
  category="lo",
  direction="l",
 },
 [0x2F887]={
  category="lo",
  direction="l",
 },
 [0x2F888]={
  category="lo",
  direction="l",
 },
 [0x2F889]={
  category="lo",
  direction="l",
 },
 [0x2F88A]={
  category="lo",
  direction="l",
 },
 [0x2F88B]={
  category="lo",
  direction="l",
 },
 [0x2F88C]={
  category="lo",
  direction="l",
 },
 [0x2F88D]={
  category="lo",
  direction="l",
 },
 [0x2F88E]={
  category="lo",
  direction="l",
 },
 [0x2F88F]={
  category="lo",
  direction="l",
 },
 [0x2F890]={
  category="lo",
  direction="l",
 },
 [0x2F891]={
  category="lo",
  direction="l",
 },
 [0x2F892]={
  category="lo",
  direction="l",
 },
 [0x2F893]={
  category="lo",
  direction="l",
 },
 [0x2F894]={
  category="lo",
  direction="l",
 },
 [0x2F895]={
  category="lo",
  direction="l",
 },
 [0x2F896]={
  category="lo",
  direction="l",
 },
 [0x2F897]={
  category="lo",
  direction="l",
 },
 [0x2F898]={
  category="lo",
  direction="l",
 },
 [0x2F899]={
  category="lo",
  direction="l",
 },
 [0x2F89A]={
  category="lo",
  direction="l",
 },
 [0x2F89B]={
  category="lo",
  direction="l",
 },
 [0x2F89C]={
  category="lo",
  direction="l",
 },
 [0x2F89D]={
  category="lo",
  direction="l",
 },
 [0x2F89E]={
  category="lo",
  direction="l",
 },
 [0x2F89F]={
  category="lo",
  direction="l",
 },
 [0x2F8A0]={
  category="lo",
  direction="l",
 },
 [0x2F8A1]={
  category="lo",
  direction="l",
 },
 [0x2F8A2]={
  category="lo",
  direction="l",
 },
 [0x2F8A3]={
  category="lo",
  direction="l",
 },
 [0x2F8A4]={
  category="lo",
  direction="l",
 },
 [0x2F8A5]={
  category="lo",
  direction="l",
 },
 [0x2F8A6]={
  category="lo",
  direction="l",
 },
 [0x2F8A7]={
  category="lo",
  direction="l",
 },
 [0x2F8A8]={
  category="lo",
  direction="l",
 },
 [0x2F8A9]={
  category="lo",
  direction="l",
 },
 [0x2F8AA]={
  category="lo",
  direction="l",
 },
 [0x2F8AB]={
  category="lo",
  direction="l",
 },
 [0x2F8AC]={
  category="lo",
  direction="l",
 },
 [0x2F8AD]={
  category="lo",
  direction="l",
 },
 [0x2F8AE]={
  category="lo",
  direction="l",
 },
 [0x2F8AF]={
  category="lo",
  direction="l",
 },
 [0x2F8B0]={
  category="lo",
  direction="l",
 },
 [0x2F8B1]={
  category="lo",
  direction="l",
 },
 [0x2F8B2]={
  category="lo",
  direction="l",
 },
 [0x2F8B3]={
  category="lo",
  direction="l",
 },
 [0x2F8B4]={
  category="lo",
  direction="l",
 },
 [0x2F8B5]={
  category="lo",
  direction="l",
 },
 [0x2F8B6]={
  category="lo",
  direction="l",
 },
 [0x2F8B7]={
  category="lo",
  direction="l",
 },
 [0x2F8B8]={
  category="lo",
  direction="l",
 },
 [0x2F8B9]={
  category="lo",
  direction="l",
 },
 [0x2F8BA]={
  category="lo",
  direction="l",
 },
 [0x2F8BB]={
  category="lo",
  direction="l",
 },
 [0x2F8BC]={
  category="lo",
  direction="l",
 },
 [0x2F8BD]={
  category="lo",
  direction="l",
 },
 [0x2F8BE]={
  category="lo",
  direction="l",
 },
 [0x2F8BF]={
  category="lo",
  direction="l",
 },
 [0x2F8C0]={
  category="lo",
  direction="l",
 },
 [0x2F8C1]={
  category="lo",
  direction="l",
 },
 [0x2F8C2]={
  category="lo",
  direction="l",
 },
 [0x2F8C3]={
  category="lo",
  direction="l",
 },
 [0x2F8C4]={
  category="lo",
  direction="l",
 },
 [0x2F8C5]={
  category="lo",
  direction="l",
 },
 [0x2F8C6]={
  category="lo",
  direction="l",
 },
 [0x2F8C7]={
  category="lo",
  direction="l",
 },
 [0x2F8C8]={
  category="lo",
  direction="l",
 },
 [0x2F8C9]={
  category="lo",
  direction="l",
 },
 [0x2F8CA]={
  category="lo",
  direction="l",
 },
 [0x2F8CB]={
  category="lo",
  direction="l",
 },
 [0x2F8CC]={
  category="lo",
  direction="l",
 },
 [0x2F8CD]={
  category="lo",
  direction="l",
 },
 [0x2F8CE]={
  category="lo",
  direction="l",
 },
 [0x2F8CF]={
  category="lo",
  direction="l",
 },
 [0x2F8D0]={
  category="lo",
  direction="l",
 },
 [0x2F8D1]={
  category="lo",
  direction="l",
 },
 [0x2F8D2]={
  category="lo",
  direction="l",
 },
 [0x2F8D3]={
  category="lo",
  direction="l",
 },
 [0x2F8D4]={
  category="lo",
  direction="l",
 },
 [0x2F8D5]={
  category="lo",
  direction="l",
 },
 [0x2F8D6]={
  category="lo",
  direction="l",
 },
 [0x2F8D7]={
  category="lo",
  direction="l",
 },
 [0x2F8D8]={
  category="lo",
  direction="l",
 },
 [0x2F8D9]={
  category="lo",
  direction="l",
 },
 [0x2F8DA]={
  category="lo",
  direction="l",
 },
 [0x2F8DB]={
  category="lo",
  direction="l",
 },
 [0x2F8DC]={
  category="lo",
  direction="l",
 },
 [0x2F8DD]={
  category="lo",
  direction="l",
 },
 [0x2F8DE]={
  category="lo",
  direction="l",
 },
 [0x2F8DF]={
  category="lo",
  direction="l",
 },
 [0x2F8E0]={
  category="lo",
  direction="l",
 },
 [0x2F8E1]={
  category="lo",
  direction="l",
 },
 [0x2F8E2]={
  category="lo",
  direction="l",
 },
 [0x2F8E3]={
  category="lo",
  direction="l",
 },
 [0x2F8E4]={
  category="lo",
  direction="l",
 },
 [0x2F8E5]={
  category="lo",
  direction="l",
 },
 [0x2F8E6]={
  category="lo",
  direction="l",
 },
 [0x2F8E7]={
  category="lo",
  direction="l",
 },
 [0x2F8E8]={
  category="lo",
  direction="l",
 },
 [0x2F8E9]={
  category="lo",
  direction="l",
 },
 [0x2F8EA]={
  category="lo",
  direction="l",
 },
 [0x2F8EB]={
  category="lo",
  direction="l",
 },
 [0x2F8EC]={
  category="lo",
  direction="l",
 },
 [0x2F8ED]={
  category="lo",
  direction="l",
 },
 [0x2F8EE]={
  category="lo",
  direction="l",
 },
 [0x2F8EF]={
  category="lo",
  direction="l",
 },
 [0x2F8F0]={
  category="lo",
  direction="l",
 },
 [0x2F8F1]={
  category="lo",
  direction="l",
 },
 [0x2F8F2]={
  category="lo",
  direction="l",
 },
 [0x2F8F3]={
  category="lo",
  direction="l",
 },
 [0x2F8F4]={
  category="lo",
  direction="l",
 },
 [0x2F8F5]={
  category="lo",
  direction="l",
 },
 [0x2F8F6]={
  category="lo",
  direction="l",
 },
 [0x2F8F7]={
  category="lo",
  direction="l",
 },
 [0x2F8F8]={
  category="lo",
  direction="l",
 },
 [0x2F8F9]={
  category="lo",
  direction="l",
 },
 [0x2F8FA]={
  category="lo",
  direction="l",
 },
 [0x2F8FB]={
  category="lo",
  direction="l",
 },
 [0x2F8FC]={
  category="lo",
  direction="l",
 },
 [0x2F8FD]={
  category="lo",
  direction="l",
 },
 [0x2F8FE]={
  category="lo",
  direction="l",
 },
 [0x2F8FF]={
  category="lo",
  direction="l",
 },
 [0x2F900]={
  category="lo",
  direction="l",
 },
 [0x2F901]={
  category="lo",
  direction="l",
 },
 [0x2F902]={
  category="lo",
  direction="l",
 },
 [0x2F903]={
  category="lo",
  direction="l",
 },
 [0x2F904]={
  category="lo",
  direction="l",
 },
 [0x2F905]={
  category="lo",
  direction="l",
 },
 [0x2F906]={
  category="lo",
  direction="l",
 },
 [0x2F907]={
  category="lo",
  direction="l",
 },
 [0x2F908]={
  category="lo",
  direction="l",
 },
 [0x2F909]={
  category="lo",
  direction="l",
 },
 [0x2F90A]={
  category="lo",
  direction="l",
 },
 [0x2F90B]={
  category="lo",
  direction="l",
 },
 [0x2F90C]={
  category="lo",
  direction="l",
 },
 [0x2F90D]={
  category="lo",
  direction="l",
 },
 [0x2F90E]={
  category="lo",
  direction="l",
 },
 [0x2F90F]={
  category="lo",
  direction="l",
 },
 [0x2F910]={
  category="lo",
  direction="l",
 },
 [0x2F911]={
  category="lo",
  direction="l",
 },
 [0x2F912]={
  category="lo",
  direction="l",
 },
 [0x2F913]={
  category="lo",
  direction="l",
 },
 [0x2F914]={
  category="lo",
  direction="l",
 },
 [0x2F915]={
  category="lo",
  direction="l",
 },
 [0x2F916]={
  category="lo",
  direction="l",
 },
 [0x2F917]={
  category="lo",
  direction="l",
 },
 [0x2F918]={
  category="lo",
  direction="l",
 },
 [0x2F919]={
  category="lo",
  direction="l",
 },
 [0x2F91A]={
  category="lo",
  direction="l",
 },
 [0x2F91B]={
  category="lo",
  direction="l",
 },
 [0x2F91C]={
  category="lo",
  direction="l",
 },
 [0x2F91D]={
  category="lo",
  direction="l",
 },
 [0x2F91E]={
  category="lo",
  direction="l",
 },
 [0x2F91F]={
  category="lo",
  direction="l",
 },
 [0x2F920]={
  category="lo",
  direction="l",
 },
 [0x2F921]={
  category="lo",
  direction="l",
 },
 [0x2F922]={
  category="lo",
  direction="l",
 },
 [0x2F923]={
  category="lo",
  direction="l",
 },
 [0x2F924]={
  category="lo",
  direction="l",
 },
 [0x2F925]={
  category="lo",
  direction="l",
 },
 [0x2F926]={
  category="lo",
  direction="l",
 },
 [0x2F927]={
  category="lo",
  direction="l",
 },
 [0x2F928]={
  category="lo",
  direction="l",
 },
 [0x2F929]={
  category="lo",
  direction="l",
 },
 [0x2F92A]={
  category="lo",
  direction="l",
 },
 [0x2F92B]={
  category="lo",
  direction="l",
 },
 [0x2F92C]={
  category="lo",
  direction="l",
 },
 [0x2F92D]={
  category="lo",
  direction="l",
 },
 [0x2F92E]={
  category="lo",
  direction="l",
 },
 [0x2F92F]={
  category="lo",
  direction="l",
 },
 [0x2F930]={
  category="lo",
  direction="l",
 },
 [0x2F931]={
  category="lo",
  direction="l",
 },
 [0x2F932]={
  category="lo",
  direction="l",
 },
 [0x2F933]={
  category="lo",
  direction="l",
 },
 [0x2F934]={
  category="lo",
  direction="l",
 },
 [0x2F935]={
  category="lo",
  direction="l",
 },
 [0x2F936]={
  category="lo",
  direction="l",
 },
 [0x2F937]={
  category="lo",
  direction="l",
 },
 [0x2F938]={
  category="lo",
  direction="l",
 },
 [0x2F939]={
  category="lo",
  direction="l",
 },
 [0x2F93A]={
  category="lo",
  direction="l",
 },
 [0x2F93B]={
  category="lo",
  direction="l",
 },
 [0x2F93C]={
  category="lo",
  direction="l",
 },
 [0x2F93D]={
  category="lo",
  direction="l",
 },
 [0x2F93E]={
  category="lo",
  direction="l",
 },
 [0x2F93F]={
  category="lo",
  direction="l",
 },
 [0x2F940]={
  category="lo",
  direction="l",
 },
 [0x2F941]={
  category="lo",
  direction="l",
 },
 [0x2F942]={
  category="lo",
  direction="l",
 },
 [0x2F943]={
  category="lo",
  direction="l",
 },
 [0x2F944]={
  category="lo",
  direction="l",
 },
 [0x2F945]={
  category="lo",
  direction="l",
 },
 [0x2F946]={
  category="lo",
  direction="l",
 },
 [0x2F947]={
  category="lo",
  direction="l",
 },
 [0x2F948]={
  category="lo",
  direction="l",
 },
 [0x2F949]={
  category="lo",
  direction="l",
 },
 [0x2F94A]={
  category="lo",
  direction="l",
 },
 [0x2F94B]={
  category="lo",
  direction="l",
 },
 [0x2F94C]={
  category="lo",
  direction="l",
 },
 [0x2F94D]={
  category="lo",
  direction="l",
 },
 [0x2F94E]={
  category="lo",
  direction="l",
 },
 [0x2F94F]={
  category="lo",
  direction="l",
 },
 [0x2F950]={
  category="lo",
  direction="l",
 },
 [0x2F951]={
  category="lo",
  direction="l",
 },
 [0x2F952]={
  category="lo",
  direction="l",
 },
 [0x2F953]={
  category="lo",
  direction="l",
 },
 [0x2F954]={
  category="lo",
  direction="l",
 },
 [0x2F955]={
  category="lo",
  direction="l",
 },
 [0x2F956]={
  category="lo",
  direction="l",
 },
 [0x2F957]={
  category="lo",
  direction="l",
 },
 [0x2F958]={
  category="lo",
  direction="l",
 },
 [0x2F959]={
  category="lo",
  direction="l",
 },
 [0x2F95A]={
  category="lo",
  direction="l",
 },
 [0x2F95B]={
  category="lo",
  direction="l",
 },
 [0x2F95C]={
  category="lo",
  direction="l",
 },
 [0x2F95D]={
  category="lo",
  direction="l",
 },
 [0x2F95E]={
  category="lo",
  direction="l",
 },
 [0x2F95F]={
  category="lo",
  direction="l",
 },
 [0x2F960]={
  category="lo",
  direction="l",
 },
 [0x2F961]={
  category="lo",
  direction="l",
 },
 [0x2F962]={
  category="lo",
  direction="l",
 },
 [0x2F963]={
  category="lo",
  direction="l",
 },
 [0x2F964]={
  category="lo",
  direction="l",
 },
 [0x2F965]={
  category="lo",
  direction="l",
 },
 [0x2F966]={
  category="lo",
  direction="l",
 },
 [0x2F967]={
  category="lo",
  direction="l",
 },
 [0x2F968]={
  category="lo",
  direction="l",
 },
 [0x2F969]={
  category="lo",
  direction="l",
 },
 [0x2F96A]={
  category="lo",
  direction="l",
 },
 [0x2F96B]={
  category="lo",
  direction="l",
 },
 [0x2F96C]={
  category="lo",
  direction="l",
 },
 [0x2F96D]={
  category="lo",
  direction="l",
 },
 [0x2F96E]={
  category="lo",
  direction="l",
 },
 [0x2F96F]={
  category="lo",
  direction="l",
 },
 [0x2F970]={
  category="lo",
  direction="l",
 },
 [0x2F971]={
  category="lo",
  direction="l",
 },
 [0x2F972]={
  category="lo",
  direction="l",
 },
 [0x2F973]={
  category="lo",
  direction="l",
 },
 [0x2F974]={
  category="lo",
  direction="l",
 },
 [0x2F975]={
  category="lo",
  direction="l",
 },
 [0x2F976]={
  category="lo",
  direction="l",
 },
 [0x2F977]={
  category="lo",
  direction="l",
 },
 [0x2F978]={
  category="lo",
  direction="l",
 },
 [0x2F979]={
  category="lo",
  direction="l",
 },
 [0x2F97A]={
  category="lo",
  direction="l",
 },
 [0x2F97B]={
  category="lo",
  direction="l",
 },
 [0x2F97C]={
  category="lo",
  direction="l",
 },
 [0x2F97D]={
  category="lo",
  direction="l",
 },
 [0x2F97E]={
  category="lo",
  direction="l",
 },
 [0x2F97F]={
  category="lo",
  direction="l",
 },
 [0x2F980]={
  category="lo",
  direction="l",
 },
 [0x2F981]={
  category="lo",
  direction="l",
 },
 [0x2F982]={
  category="lo",
  direction="l",
 },
 [0x2F983]={
  category="lo",
  direction="l",
 },
 [0x2F984]={
  category="lo",
  direction="l",
 },
 [0x2F985]={
  category="lo",
  direction="l",
 },
 [0x2F986]={
  category="lo",
  direction="l",
 },
 [0x2F987]={
  category="lo",
  direction="l",
 },
 [0x2F988]={
  category="lo",
  direction="l",
 },
 [0x2F989]={
  category="lo",
  direction="l",
 },
 [0x2F98A]={
  category="lo",
  direction="l",
 },
 [0x2F98B]={
  category="lo",
  direction="l",
 },
 [0x2F98C]={
  category="lo",
  direction="l",
 },
 [0x2F98D]={
  category="lo",
  direction="l",
 },
 [0x2F98E]={
  category="lo",
  direction="l",
 },
 [0x2F98F]={
  category="lo",
  direction="l",
 },
 [0x2F990]={
  category="lo",
  direction="l",
 },
 [0x2F991]={
  category="lo",
  direction="l",
 },
 [0x2F992]={
  category="lo",
  direction="l",
 },
 [0x2F993]={
  category="lo",
  direction="l",
 },
 [0x2F994]={
  category="lo",
  direction="l",
 },
 [0x2F995]={
  category="lo",
  direction="l",
 },
 [0x2F996]={
  category="lo",
  direction="l",
 },
 [0x2F997]={
  category="lo",
  direction="l",
 },
 [0x2F998]={
  category="lo",
  direction="l",
 },
 [0x2F999]={
  category="lo",
  direction="l",
 },
 [0x2F99A]={
  category="lo",
  direction="l",
 },
 [0x2F99B]={
  category="lo",
  direction="l",
 },
 [0x2F99C]={
  category="lo",
  direction="l",
 },
 [0x2F99D]={
  category="lo",
  direction="l",
 },
 [0x2F99E]={
  category="lo",
  direction="l",
 },
 [0x2F99F]={
  category="lo",
  direction="l",
 },
 [0x2F9A0]={
  category="lo",
  direction="l",
 },
 [0x2F9A1]={
  category="lo",
  direction="l",
 },
 [0x2F9A2]={
  category="lo",
  direction="l",
 },
 [0x2F9A3]={
  category="lo",
  direction="l",
 },
 [0x2F9A4]={
  category="lo",
  direction="l",
 },
 [0x2F9A5]={
  category="lo",
  direction="l",
 },
 [0x2F9A6]={
  category="lo",
  direction="l",
 },
 [0x2F9A7]={
  category="lo",
  direction="l",
 },
 [0x2F9A8]={
  category="lo",
  direction="l",
 },
 [0x2F9A9]={
  category="lo",
  direction="l",
 },
 [0x2F9AA]={
  category="lo",
  direction="l",
 },
 [0x2F9AB]={
  category="lo",
  direction="l",
 },
 [0x2F9AC]={
  category="lo",
  direction="l",
 },
 [0x2F9AD]={
  category="lo",
  direction="l",
 },
 [0x2F9AE]={
  category="lo",
  direction="l",
 },
 [0x2F9AF]={
  category="lo",
  direction="l",
 },
 [0x2F9B0]={
  category="lo",
  direction="l",
 },
 [0x2F9B1]={
  category="lo",
  direction="l",
 },
 [0x2F9B2]={
  category="lo",
  direction="l",
 },
 [0x2F9B3]={
  category="lo",
  direction="l",
 },
 [0x2F9B4]={
  category="lo",
  direction="l",
 },
 [0x2F9B5]={
  category="lo",
  direction="l",
 },
 [0x2F9B6]={
  category="lo",
  direction="l",
 },
 [0x2F9B7]={
  category="lo",
  direction="l",
 },
 [0x2F9B8]={
  category="lo",
  direction="l",
 },
 [0x2F9B9]={
  category="lo",
  direction="l",
 },
 [0x2F9BA]={
  category="lo",
  direction="l",
 },
 [0x2F9BB]={
  category="lo",
  direction="l",
 },
 [0x2F9BC]={
  category="lo",
  direction="l",
 },
 [0x2F9BD]={
  category="lo",
  direction="l",
 },
 [0x2F9BE]={
  category="lo",
  direction="l",
 },
 [0x2F9BF]={
  category="lo",
  direction="l",
 },
 [0x2F9C0]={
  category="lo",
  direction="l",
 },
 [0x2F9C1]={
  category="lo",
  direction="l",
 },
 [0x2F9C2]={
  category="lo",
  direction="l",
 },
 [0x2F9C3]={
  category="lo",
  direction="l",
 },
 [0x2F9C4]={
  category="lo",
  direction="l",
 },
 [0x2F9C5]={
  category="lo",
  direction="l",
 },
 [0x2F9C6]={
  category="lo",
  direction="l",
 },
 [0x2F9C7]={
  category="lo",
  direction="l",
 },
 [0x2F9C8]={
  category="lo",
  direction="l",
 },
 [0x2F9C9]={
  category="lo",
  direction="l",
 },
 [0x2F9CA]={
  category="lo",
  direction="l",
 },
 [0x2F9CB]={
  category="lo",
  direction="l",
 },
 [0x2F9CC]={
  category="lo",
  direction="l",
 },
 [0x2F9CD]={
  category="lo",
  direction="l",
 },
 [0x2F9CE]={
  category="lo",
  direction="l",
 },
 [0x2F9CF]={
  category="lo",
  direction="l",
 },
 [0x2F9D0]={
  category="lo",
  direction="l",
 },
 [0x2F9D1]={
  category="lo",
  direction="l",
 },
 [0x2F9D2]={
  category="lo",
  direction="l",
 },
 [0x2F9D3]={
  category="lo",
  direction="l",
 },
 [0x2F9D4]={
  category="lo",
  direction="l",
 },
 [0x2F9D5]={
  category="lo",
  direction="l",
 },
 [0x2F9D6]={
  category="lo",
  direction="l",
 },
 [0x2F9D7]={
  category="lo",
  direction="l",
 },
 [0x2F9D8]={
  category="lo",
  direction="l",
 },
 [0x2F9D9]={
  category="lo",
  direction="l",
 },
 [0x2F9DA]={
  category="lo",
  direction="l",
 },
 [0x2F9DB]={
  category="lo",
  direction="l",
 },
 [0x2F9DC]={
  category="lo",
  direction="l",
 },
 [0x2F9DD]={
  category="lo",
  direction="l",
 },
 [0x2F9DE]={
  category="lo",
  direction="l",
 },
 [0x2F9DF]={
  category="lo",
  direction="l",
 },
 [0x2F9E0]={
  category="lo",
  direction="l",
 },
 [0x2F9E1]={
  category="lo",
  direction="l",
 },
 [0x2F9E2]={
  category="lo",
  direction="l",
 },
 [0x2F9E3]={
  category="lo",
  direction="l",
 },
 [0x2F9E4]={
  category="lo",
  direction="l",
 },
 [0x2F9E5]={
  category="lo",
  direction="l",
 },
 [0x2F9E6]={
  category="lo",
  direction="l",
 },
 [0x2F9E7]={
  category="lo",
  direction="l",
 },
 [0x2F9E8]={
  category="lo",
  direction="l",
 },
 [0x2F9E9]={
  category="lo",
  direction="l",
 },
 [0x2F9EA]={
  category="lo",
  direction="l",
 },
 [0x2F9EB]={
  category="lo",
  direction="l",
 },
 [0x2F9EC]={
  category="lo",
  direction="l",
 },
 [0x2F9ED]={
  category="lo",
  direction="l",
 },
 [0x2F9EE]={
  category="lo",
  direction="l",
 },
 [0x2F9EF]={
  category="lo",
  direction="l",
 },
 [0x2F9F0]={
  category="lo",
  direction="l",
 },
 [0x2F9F1]={
  category="lo",
  direction="l",
 },
 [0x2F9F2]={
  category="lo",
  direction="l",
 },
 [0x2F9F3]={
  category="lo",
  direction="l",
 },
 [0x2F9F4]={
  category="lo",
  direction="l",
 },
 [0x2F9F5]={
  category="lo",
  direction="l",
 },
 [0x2F9F6]={
  category="lo",
  direction="l",
 },
 [0x2F9F7]={
  category="lo",
  direction="l",
 },
 [0x2F9F8]={
  category="lo",
  direction="l",
 },
 [0x2F9F9]={
  category="lo",
  direction="l",
 },
 [0x2F9FA]={
  category="lo",
  direction="l",
 },
 [0x2F9FB]={
  category="lo",
  direction="l",
 },
 [0x2F9FC]={
  category="lo",
  direction="l",
 },
 [0x2F9FD]={
  category="lo",
  direction="l",
 },
 [0x2F9FE]={
  category="lo",
  direction="l",
 },
 [0x2F9FF]={
  category="lo",
  direction="l",
 },
 [0x2FA00]={
  category="lo",
  direction="l",
 },
 [0x2FA01]={
  category="lo",
  direction="l",
 },
 [0x2FA02]={
  category="lo",
  direction="l",
 },
 [0x2FA03]={
  category="lo",
  direction="l",
 },
 [0x2FA04]={
  category="lo",
  direction="l",
 },
 [0x2FA05]={
  category="lo",
  direction="l",
 },
 [0x2FA06]={
  category="lo",
  direction="l",
 },
 [0x2FA07]={
  category="lo",
  direction="l",
 },
 [0x2FA08]={
  category="lo",
  direction="l",
 },
 [0x2FA09]={
  category="lo",
  direction="l",
 },
 [0x2FA0A]={
  category="lo",
  direction="l",
 },
 [0x2FA0B]={
  category="lo",
  direction="l",
 },
 [0x2FA0C]={
  category="lo",
  direction="l",
 },
 [0x2FA0D]={
  category="lo",
  direction="l",
 },
 [0x2FA0E]={
  category="lo",
  direction="l",
 },
 [0x2FA0F]={
  category="lo",
  direction="l",
 },
 [0x2FA10]={
  category="lo",
  direction="l",
 },
 [0x2FA11]={
  category="lo",
  direction="l",
 },
 [0x2FA12]={
  category="lo",
  direction="l",
 },
 [0x2FA13]={
  category="lo",
  direction="l",
 },
 [0x2FA14]={
  category="lo",
  direction="l",
 },
 [0x2FA15]={
  category="lo",
  direction="l",
 },
 [0x2FA16]={
  category="lo",
  direction="l",
 },
 [0x2FA17]={
  category="lo",
  direction="l",
 },
 [0x2FA18]={
  category="lo",
  direction="l",
 },
 [0x2FA19]={
  category="lo",
  direction="l",
 },
 [0x2FA1A]={
  category="lo",
  direction="l",
 },
 [0x2FA1B]={
  category="lo",
  direction="l",
 },
 [0x2FA1C]={
  category="lo",
  direction="l",
 },
 [0x2FA1D]={
  category="lo",
  direction="l",
 },
 [0xE0001]={
  category="cf",
  direction="bn",
 },
 [0xE0020]={
  category="cf",
  direction="bn",
 },
 [0xE0021]={
  category="cf",
  direction="bn",
 },
 [0xE0022]={
  category="cf",
  direction="bn",
 },
 [0xE0023]={
  category="cf",
  direction="bn",
 },
 [0xE0024]={
  category="cf",
  direction="bn",
 },
 [0xE0025]={
  category="cf",
  direction="bn",
 },
 [0xE0026]={
  category="cf",
  direction="bn",
 },
 [0xE0027]={
  category="cf",
  direction="bn",
 },
 [0xE0028]={
  category="cf",
  direction="bn",
 },
 [0xE0029]={
  category="cf",
  direction="bn",
 },
 [0xE002A]={
  category="cf",
  direction="bn",
 },
 [0xE002B]={
  category="cf",
  direction="bn",
 },
 [0xE002C]={
  category="cf",
  direction="bn",
 },
 [0xE002D]={
  category="cf",
  direction="bn",
 },
 [0xE002E]={
  category="cf",
  direction="bn",
 },
 [0xE002F]={
  category="cf",
  direction="bn",
 },
 [0xE0030]={
  category="cf",
  direction="bn",
 },
 [0xE0031]={
  category="cf",
  direction="bn",
 },
 [0xE0032]={
  category="cf",
  direction="bn",
 },
 [0xE0033]={
  category="cf",
  direction="bn",
 },
 [0xE0034]={
  category="cf",
  direction="bn",
 },
 [0xE0035]={
  category="cf",
  direction="bn",
 },
 [0xE0036]={
  category="cf",
  direction="bn",
 },
 [0xE0037]={
  category="cf",
  direction="bn",
 },
 [0xE0038]={
  category="cf",
  direction="bn",
 },
 [0xE0039]={
  category="cf",
  direction="bn",
 },
 [0xE003A]={
  category="cf",
  direction="bn",
 },
 [0xE003B]={
  category="cf",
  direction="bn",
 },
 [0xE003C]={
  category="cf",
  direction="bn",
 },
 [0xE003D]={
  category="cf",
  direction="bn",
 },
 [0xE003E]={
  category="cf",
  direction="bn",
 },
 [0xE003F]={
  category="cf",
  direction="bn",
 },
 [0xE0040]={
  category="cf",
  direction="bn",
 },
 [0xE0041]={
  category="cf",
  direction="bn",
 },
 [0xE0042]={
  category="cf",
  direction="bn",
 },
 [0xE0043]={
  category="cf",
  direction="bn",
 },
 [0xE0044]={
  category="cf",
  direction="bn",
 },
 [0xE0045]={
  category="cf",
  direction="bn",
 },
 [0xE0046]={
  category="cf",
  direction="bn",
 },
 [0xE0047]={
  category="cf",
  direction="bn",
 },
 [0xE0048]={
  category="cf",
  direction="bn",
 },
 [0xE0049]={
  category="cf",
  direction="bn",
 },
 [0xE004A]={
  category="cf",
  direction="bn",
 },
 [0xE004B]={
  category="cf",
  direction="bn",
 },
 [0xE004C]={
  category="cf",
  direction="bn",
 },
 [0xE004D]={
  category="cf",
  direction="bn",
 },
 [0xE004E]={
  category="cf",
  direction="bn",
 },
 [0xE004F]={
  category="cf",
  direction="bn",
 },
 [0xE0050]={
  category="cf",
  direction="bn",
 },
 [0xE0051]={
  category="cf",
  direction="bn",
 },
 [0xE0052]={
  category="cf",
  direction="bn",
 },
 [0xE0053]={
  category="cf",
  direction="bn",
 },
 [0xE0054]={
  category="cf",
  direction="bn",
 },
 [0xE0055]={
  category="cf",
  direction="bn",
 },
 [0xE0056]={
  category="cf",
  direction="bn",
 },
 [0xE0057]={
  category="cf",
  direction="bn",
 },
 [0xE0058]={
  category="cf",
  direction="bn",
 },
 [0xE0059]={
  category="cf",
  direction="bn",
 },
 [0xE005A]={
  category="cf",
  direction="bn",
 },
 [0xE005B]={
  category="cf",
  direction="bn",
 },
 [0xE005C]={
  category="cf",
  direction="bn",
 },
 [0xE005D]={
  category="cf",
  direction="bn",
 },
 [0xE005E]={
  category="cf",
  direction="bn",
 },
 [0xE005F]={
  category="cf",
  direction="bn",
 },
 [0xE0060]={
  category="cf",
  direction="bn",
 },
 [0xE0061]={
  category="cf",
  direction="bn",
 },
 [0xE0062]={
  category="cf",
  direction="bn",
 },
 [0xE0063]={
  category="cf",
  direction="bn",
 },
 [0xE0064]={
  category="cf",
  direction="bn",
 },
 [0xE0065]={
  category="cf",
  direction="bn",
 },
 [0xE0066]={
  category="cf",
  direction="bn",
 },
 [0xE0067]={
  category="cf",
  direction="bn",
 },
 [0xE0068]={
  category="cf",
  direction="bn",
 },
 [0xE0069]={
  category="cf",
  direction="bn",
 },
 [0xE006A]={
  category="cf",
  direction="bn",
 },
 [0xE006B]={
  category="cf",
  direction="bn",
 },
 [0xE006C]={
  category="cf",
  direction="bn",
 },
 [0xE006D]={
  category="cf",
  direction="bn",
 },
 [0xE006E]={
  category="cf",
  direction="bn",
 },
 [0xE006F]={
  category="cf",
  direction="bn",
 },
 [0xE0070]={
  category="cf",
  direction="bn",
 },
 [0xE0071]={
  category="cf",
  direction="bn",
 },
 [0xE0072]={
  category="cf",
  direction="bn",
 },
 [0xE0073]={
  category="cf",
  direction="bn",
 },
 [0xE0074]={
  category="cf",
  direction="bn",
 },
 [0xE0075]={
  category="cf",
  direction="bn",
 },
 [0xE0076]={
  category="cf",
  direction="bn",
 },
 [0xE0077]={
  category="cf",
  direction="bn",
 },
 [0xE0078]={
  category="cf",
  direction="bn",
 },
 [0xE0079]={
  category="cf",
  direction="bn",
 },
 [0xE007A]={
  category="cf",
  direction="bn",
 },
 [0xE007B]={
  category="cf",
  direction="bn",
 },
 [0xE007C]={
  category="cf",
  direction="bn",
 },
 [0xE007D]={
  category="cf",
  direction="bn",
 },
 [0xE007E]={
  category="cf",
  direction="bn",
 },
 [0xE007F]={
  category="cf",
  direction="bn",
 },
 [0xE0100]={
  category="mn",
  direction="nsm",
 },
 [0xE0101]={
  category="mn",
  direction="nsm",
 },
 [0xE0102]={
  category="mn",
  direction="nsm",
 },
 [0xE0103]={
  category="mn",
  direction="nsm",
 },
 [0xE0104]={
  category="mn",
  direction="nsm",
 },
 [0xE0105]={
  category="mn",
  direction="nsm",
 },
 [0xE0106]={
  category="mn",
  direction="nsm",
 },
 [0xE0107]={
  category="mn",
  direction="nsm",
 },
 [0xE0108]={
  category="mn",
  direction="nsm",
 },
 [0xE0109]={
  category="mn",
  direction="nsm",
 },
 [0xE010A]={
  category="mn",
  direction="nsm",
 },
 [0xE010B]={
  category="mn",
  direction="nsm",
 },
 [0xE010C]={
  category="mn",
  direction="nsm",
 },
 [0xE010D]={
  category="mn",
  direction="nsm",
 },
 [0xE010E]={
  category="mn",
  direction="nsm",
 },
 [0xE010F]={
  category="mn",
  direction="nsm",
 },
 [0xE0110]={
  category="mn",
  direction="nsm",
 },
 [0xE0111]={
  category="mn",
  direction="nsm",
 },
 [0xE0112]={
  category="mn",
  direction="nsm",
 },
 [0xE0113]={
  category="mn",
  direction="nsm",
 },
 [0xE0114]={
  category="mn",
  direction="nsm",
 },
 [0xE0115]={
  category="mn",
  direction="nsm",
 },
 [0xE0116]={
  category="mn",
  direction="nsm",
 },
 [0xE0117]={
  category="mn",
  direction="nsm",
 },
 [0xE0118]={
  category="mn",
  direction="nsm",
 },
 [0xE0119]={
  category="mn",
  direction="nsm",
 },
 [0xE011A]={
  category="mn",
  direction="nsm",
 },
 [0xE011B]={
  category="mn",
  direction="nsm",
 },
 [0xE011C]={
  category="mn",
  direction="nsm",
 },
 [0xE011D]={
  category="mn",
  direction="nsm",
 },
 [0xE011E]={
  category="mn",
  direction="nsm",
 },
 [0xE011F]={
  category="mn",
  direction="nsm",
 },
 [0xE0120]={
  category="mn",
  direction="nsm",
 },
 [0xE0121]={
  category="mn",
  direction="nsm",
 },
 [0xE0122]={
  category="mn",
  direction="nsm",
 },
 [0xE0123]={
  category="mn",
  direction="nsm",
 },
 [0xE0124]={
  category="mn",
  direction="nsm",
 },
 [0xE0125]={
  category="mn",
  direction="nsm",
 },
 [0xE0126]={
  category="mn",
  direction="nsm",
 },
 [0xE0127]={
  category="mn",
  direction="nsm",
 },
 [0xE0128]={
  category="mn",
  direction="nsm",
 },
 [0xE0129]={
  category="mn",
  direction="nsm",
 },
 [0xE012A]={
  category="mn",
  direction="nsm",
 },
 [0xE012B]={
  category="mn",
  direction="nsm",
 },
 [0xE012C]={
  category="mn",
  direction="nsm",
 },
 [0xE012D]={
  category="mn",
  direction="nsm",
 },
 [0xE012E]={
  category="mn",
  direction="nsm",
 },
 [0xE012F]={
  category="mn",
  direction="nsm",
 },
 [0xE0130]={
  category="mn",
  direction="nsm",
 },
 [0xE0131]={
  category="mn",
  direction="nsm",
 },
 [0xE0132]={
  category="mn",
  direction="nsm",
 },
 [0xE0133]={
  category="mn",
  direction="nsm",
 },
 [0xE0134]={
  category="mn",
  direction="nsm",
 },
 [0xE0135]={
  category="mn",
  direction="nsm",
 },
 [0xE0136]={
  category="mn",
  direction="nsm",
 },
 [0xE0137]={
  category="mn",
  direction="nsm",
 },
 [0xE0138]={
  category="mn",
  direction="nsm",
 },
 [0xE0139]={
  category="mn",
  direction="nsm",
 },
 [0xE013A]={
  category="mn",
  direction="nsm",
 },
 [0xE013B]={
  category="mn",
  direction="nsm",
 },
 [0xE013C]={
  category="mn",
  direction="nsm",
 },
 [0xE013D]={
  category="mn",
  direction="nsm",
 },
 [0xE013E]={
  category="mn",
  direction="nsm",
 },
 [0xE013F]={
  category="mn",
  direction="nsm",
 },
 [0xE0140]={
  category="mn",
  direction="nsm",
 },
 [0xE0141]={
  category="mn",
  direction="nsm",
 },
 [0xE0142]={
  category="mn",
  direction="nsm",
 },
 [0xE0143]={
  category="mn",
  direction="nsm",
 },
 [0xE0144]={
  category="mn",
  direction="nsm",
 },
 [0xE0145]={
  category="mn",
  direction="nsm",
 },
 [0xE0146]={
  category="mn",
  direction="nsm",
 },
 [0xE0147]={
  category="mn",
  direction="nsm",
 },
 [0xE0148]={
  category="mn",
  direction="nsm",
 },
 [0xE0149]={
  category="mn",
  direction="nsm",
 },
 [0xE014A]={
  category="mn",
  direction="nsm",
 },
 [0xE014B]={
  category="mn",
  direction="nsm",
 },
 [0xE014C]={
  category="mn",
  direction="nsm",
 },
 [0xE014D]={
  category="mn",
  direction="nsm",
 },
 [0xE014E]={
  category="mn",
  direction="nsm",
 },
 [0xE014F]={
  category="mn",
  direction="nsm",
 },
 [0xE0150]={
  category="mn",
  direction="nsm",
 },
 [0xE0151]={
  category="mn",
  direction="nsm",
 },
 [0xE0152]={
  category="mn",
  direction="nsm",
 },
 [0xE0153]={
  category="mn",
  direction="nsm",
 },
 [0xE0154]={
  category="mn",
  direction="nsm",
 },
 [0xE0155]={
  category="mn",
  direction="nsm",
 },
 [0xE0156]={
  category="mn",
  direction="nsm",
 },
 [0xE0157]={
  category="mn",
  direction="nsm",
 },
 [0xE0158]={
  category="mn",
  direction="nsm",
 },
 [0xE0159]={
  category="mn",
  direction="nsm",
 },
 [0xE015A]={
  category="mn",
  direction="nsm",
 },
 [0xE015B]={
  category="mn",
  direction="nsm",
 },
 [0xE015C]={
  category="mn",
  direction="nsm",
 },
 [0xE015D]={
  category="mn",
  direction="nsm",
 },
 [0xE015E]={
  category="mn",
  direction="nsm",
 },
 [0xE015F]={
  category="mn",
  direction="nsm",
 },
 [0xE0160]={
  category="mn",
  direction="nsm",
 },
 [0xE0161]={
  category="mn",
  direction="nsm",
 },
 [0xE0162]={
  category="mn",
  direction="nsm",
 },
 [0xE0163]={
  category="mn",
  direction="nsm",
 },
 [0xE0164]={
  category="mn",
  direction="nsm",
 },
 [0xE0165]={
  category="mn",
  direction="nsm",
 },
 [0xE0166]={
  category="mn",
  direction="nsm",
 },
 [0xE0167]={
  category="mn",
  direction="nsm",
 },
 [0xE0168]={
  category="mn",
  direction="nsm",
 },
 [0xE0169]={
  category="mn",
  direction="nsm",
 },
 [0xE016A]={
  category="mn",
  direction="nsm",
 },
 [0xE016B]={
  category="mn",
  direction="nsm",
 },
 [0xE016C]={
  category="mn",
  direction="nsm",
 },
 [0xE016D]={
  category="mn",
  direction="nsm",
 },
 [0xE016E]={
  category="mn",
  direction="nsm",
 },
 [0xE016F]={
  category="mn",
  direction="nsm",
 },
 [0xE0170]={
  category="mn",
  direction="nsm",
 },
 [0xE0171]={
  category="mn",
  direction="nsm",
 },
 [0xE0172]={
  category="mn",
  direction="nsm",
 },
 [0xE0173]={
  category="mn",
  direction="nsm",
 },
 [0xE0174]={
  category="mn",
  direction="nsm",
 },
 [0xE0175]={
  category="mn",
  direction="nsm",
 },
 [0xE0176]={
  category="mn",
  direction="nsm",
 },
 [0xE0177]={
  category="mn",
  direction="nsm",
 },
 [0xE0178]={
  category="mn",
  direction="nsm",
 },
 [0xE0179]={
  category="mn",
  direction="nsm",
 },
 [0xE017A]={
  category="mn",
  direction="nsm",
 },
 [0xE017B]={
  category="mn",
  direction="nsm",
 },
 [0xE017C]={
  category="mn",
  direction="nsm",
 },
 [0xE017D]={
  category="mn",
  direction="nsm",
 },
 [0xE017E]={
  category="mn",
  direction="nsm",
 },
 [0xE017F]={
  category="mn",
  direction="nsm",
 },
 [0xE0180]={
  category="mn",
  direction="nsm",
 },
 [0xE0181]={
  category="mn",
  direction="nsm",
 },
 [0xE0182]={
  category="mn",
  direction="nsm",
 },
 [0xE0183]={
  category="mn",
  direction="nsm",
 },
 [0xE0184]={
  category="mn",
  direction="nsm",
 },
 [0xE0185]={
  category="mn",
  direction="nsm",
 },
 [0xE0186]={
  category="mn",
  direction="nsm",
 },
 [0xE0187]={
  category="mn",
  direction="nsm",
 },
 [0xE0188]={
  category="mn",
  direction="nsm",
 },
 [0xE0189]={
  category="mn",
  direction="nsm",
 },
 [0xE018A]={
  category="mn",
  direction="nsm",
 },
 [0xE018B]={
  category="mn",
  direction="nsm",
 },
 [0xE018C]={
  category="mn",
  direction="nsm",
 },
 [0xE018D]={
  category="mn",
  direction="nsm",
 },
 [0xE018E]={
  category="mn",
  direction="nsm",
 },
 [0xE018F]={
  category="mn",
  direction="nsm",
 },
 [0xE0190]={
  category="mn",
  direction="nsm",
 },
 [0xE0191]={
  category="mn",
  direction="nsm",
 },
 [0xE0192]={
  category="mn",
  direction="nsm",
 },
 [0xE0193]={
  category="mn",
  direction="nsm",
 },
 [0xE0194]={
  category="mn",
  direction="nsm",
 },
 [0xE0195]={
  category="mn",
  direction="nsm",
 },
 [0xE0196]={
  category="mn",
  direction="nsm",
 },
 [0xE0197]={
  category="mn",
  direction="nsm",
 },
 [0xE0198]={
  category="mn",
  direction="nsm",
 },
 [0xE0199]={
  category="mn",
  direction="nsm",
 },
 [0xE019A]={
  category="mn",
  direction="nsm",
 },
 [0xE019B]={
  category="mn",
  direction="nsm",
 },
 [0xE019C]={
  category="mn",
  direction="nsm",
 },
 [0xE019D]={
  category="mn",
  direction="nsm",
 },
 [0xE019E]={
  category="mn",
  direction="nsm",
 },
 [0xE019F]={
  category="mn",
  direction="nsm",
 },
 [0xE01A0]={
  category="mn",
  direction="nsm",
 },
 [0xE01A1]={
  category="mn",
  direction="nsm",
 },
 [0xE01A2]={
  category="mn",
  direction="nsm",
 },
 [0xE01A3]={
  category="mn",
  direction="nsm",
 },
 [0xE01A4]={
  category="mn",
  direction="nsm",
 },
 [0xE01A5]={
  category="mn",
  direction="nsm",
 },
 [0xE01A6]={
  category="mn",
  direction="nsm",
 },
 [0xE01A7]={
  category="mn",
  direction="nsm",
 },
 [0xE01A8]={
  category="mn",
  direction="nsm",
 },
 [0xE01A9]={
  category="mn",
  direction="nsm",
 },
 [0xE01AA]={
  category="mn",
  direction="nsm",
 },
 [0xE01AB]={
  category="mn",
  direction="nsm",
 },
 [0xE01AC]={
  category="mn",
  direction="nsm",
 },
 [0xE01AD]={
  category="mn",
  direction="nsm",
 },
 [0xE01AE]={
  category="mn",
  direction="nsm",
 },
 [0xE01AF]={
  category="mn",
  direction="nsm",
 },
 [0xE01B0]={
  category="mn",
  direction="nsm",
 },
 [0xE01B1]={
  category="mn",
  direction="nsm",
 },
 [0xE01B2]={
  category="mn",
  direction="nsm",
 },
 [0xE01B3]={
  category="mn",
  direction="nsm",
 },
 [0xE01B4]={
  category="mn",
  direction="nsm",
 },
 [0xE01B5]={
  category="mn",
  direction="nsm",
 },
 [0xE01B6]={
  category="mn",
  direction="nsm",
 },
 [0xE01B7]={
  category="mn",
  direction="nsm",
 },
 [0xE01B8]={
  category="mn",
  direction="nsm",
 },
 [0xE01B9]={
  category="mn",
  direction="nsm",
 },
 [0xE01BA]={
  category="mn",
  direction="nsm",
 },
 [0xE01BB]={
  category="mn",
  direction="nsm",
 },
 [0xE01BC]={
  category="mn",
  direction="nsm",
 },
 [0xE01BD]={
  category="mn",
  direction="nsm",
 },
 [0xE01BE]={
  category="mn",
  direction="nsm",
 },
 [0xE01BF]={
  category="mn",
  direction="nsm",
 },
 [0xE01C0]={
  category="mn",
  direction="nsm",
 },
 [0xE01C1]={
  category="mn",
  direction="nsm",
 },
 [0xE01C2]={
  category="mn",
  direction="nsm",
 },
 [0xE01C3]={
  category="mn",
  direction="nsm",
 },
 [0xE01C4]={
  category="mn",
  direction="nsm",
 },
 [0xE01C5]={
  category="mn",
  direction="nsm",
 },
 [0xE01C6]={
  category="mn",
  direction="nsm",
 },
 [0xE01C7]={
  category="mn",
  direction="nsm",
 },
 [0xE01C8]={
  category="mn",
  direction="nsm",
 },
 [0xE01C9]={
  category="mn",
  direction="nsm",
 },
 [0xE01CA]={
  category="mn",
  direction="nsm",
 },
 [0xE01CB]={
  category="mn",
  direction="nsm",
 },
 [0xE01CC]={
  category="mn",
  direction="nsm",
 },
 [0xE01CD]={
  category="mn",
  direction="nsm",
 },
 [0xE01CE]={
  category="mn",
  direction="nsm",
 },
 [0xE01CF]={
  category="mn",
  direction="nsm",
 },
 [0xE01D0]={
  category="mn",
  direction="nsm",
 },
 [0xE01D1]={
  category="mn",
  direction="nsm",
 },
 [0xE01D2]={
  category="mn",
  direction="nsm",
 },
 [0xE01D3]={
  category="mn",
  direction="nsm",
 },
 [0xE01D4]={
  category="mn",
  direction="nsm",
 },
 [0xE01D5]={
  category="mn",
  direction="nsm",
 },
 [0xE01D6]={
  category="mn",
  direction="nsm",
 },
 [0xE01D7]={
  category="mn",
  direction="nsm",
 },
 [0xE01D8]={
  category="mn",
  direction="nsm",
 },
 [0xE01D9]={
  category="mn",
  direction="nsm",
 },
 [0xE01DA]={
  category="mn",
  direction="nsm",
 },
 [0xE01DB]={
  category="mn",
  direction="nsm",
 },
 [0xE01DC]={
  category="mn",
  direction="nsm",
 },
 [0xE01DD]={
  category="mn",
  direction="nsm",
 },
 [0xE01DE]={
  category="mn",
  direction="nsm",
 },
 [0xE01DF]={
  category="mn",
  direction="nsm",
 },
 [0xE01E0]={
  category="mn",
  direction="nsm",
 },
 [0xE01E1]={
  category="mn",
  direction="nsm",
 },
 [0xE01E2]={
  category="mn",
  direction="nsm",
 },
 [0xE01E3]={
  category="mn",
  direction="nsm",
 },
 [0xE01E4]={
  category="mn",
  direction="nsm",
 },
 [0xE01E5]={
  category="mn",
  direction="nsm",
 },
 [0xE01E6]={
  category="mn",
  direction="nsm",
 },
 [0xE01E7]={
  category="mn",
  direction="nsm",
 },
 [0xE01E8]={
  category="mn",
  direction="nsm",
 },
 [0xE01E9]={
  category="mn",
  direction="nsm",
 },
 [0xE01EA]={
  category="mn",
  direction="nsm",
 },
 [0xE01EB]={
  category="mn",
  direction="nsm",
 },
 [0xE01EC]={
  category="mn",
  direction="nsm",
 },
 [0xE01ED]={
  category="mn",
  direction="nsm",
 },
 [0xE01EE]={
  category="mn",
  direction="nsm",
 },
 [0xE01EF]={
  category="mn",
  direction="nsm",
 },
}